/*
 * © Copyright (c) 2004-2019 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <stdarg.h>

#include <infiniband/ibdiag/ibdiag.h>


/****************************************************/
#include <infiniband/ibdiag/ibdiag_version.h>

string get_ibdiag_version()
{
    string version = "\"";
    #ifdef IBDIAG_VERSION
        version += IBDIAG_VERSION;
    #else
        version += "unknown";
    #endif

    //add dependency libraries
    version += "\",\"";
    version += get_ibdm_version();
    version += "\",\"";
    version += get_ibis_version();
    version += "\"";
    return version;
}

u_int8_t get_operational_vl_num(u_int8_t opvl_code)
{
    if ((opvl_code < 1) | (opvl_code > 5))
        return 0;

    u_int8_t operational_vl = 1;
    opvl_code--;
    operational_vl = (u_int8_t)(operational_vl << opvl_code);
    return operational_vl;
}
/****************************************************/
/******************private methods*******************/
/****************************************************/
void IBDiag::SetLastError(const char *fmt, ...)
{
    IBDIAG_ENTER;
    char buff[IBDIAG_ERR_SIZE];
    va_list args;

    memset(buff, 0, sizeof(buff));
    va_start(args, fmt);
    vsprintf(buff, fmt, args);
    va_end(args);
    this->last_error.assign(buff);
    IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "-E- %s\n", this->last_error.c_str());
    IBDIAG_RETURN_VOID;
}


void IBDiag::CleanUpInternalDB()
{
    IBDIAG_ENTER;
    list_p_direct_route::iterator it;
    list_p_bad_direct_route::iterator bad_dr_it;

    //clean good direct routes
    for (it = this->good_direct_routes.begin(); it != this->good_direct_routes.end(); ++it) {
        delete (*it);
    }
    //clean bad direct routes
    for (bad_dr_it = this->bad_direct_routes.begin(); bad_dr_it != this->bad_direct_routes.end(); ++bad_dr_it) {
        delete((*bad_dr_it)->direct_route);
        delete (*bad_dr_it);
    }
    //clean loop direct routes
    for (it = this->loop_direct_routes.begin(); it != this->loop_direct_routes.end(); ++it) {
        delete (*it);
    }
    //clean bfs list
    for (it = this->bfs_list.begin(); it != this->bfs_list.end(); ++it) {
        delete (*it);
    }

    //no need to free map of nodes to direct routes - these direct routes were freed
    //when we cleaned the good_direct_routes list
    this->ibdiag_discovery_status = DISCOVERY_NOT_DONE;
    this->root_node = NULL;
    this->root_port_num = 0;
    CLEAR_STRUCT(this->discover_progress_bar_nodes);
    CLEAR_STRUCT(this->discover_progress_bar_ports);
    this->fabric_extended_info.CleanUpInternalDB();

    //make sure all lists + maps are empty
    this->bfs_list.clear();
    this->good_direct_routes.clear();
    this->bad_direct_routes.clear();
    this->loop_direct_routes.clear();
    this->duplicated_guids_detection_errs.clear();
    this->bfs_known_node_guids.clear();
    this->bfs_known_port_guids.clear();

    IBDIAG_RETURN_VOID;
}

//Returns non zero on error
int IBDiag::OpenFile(const char *file_name, ofstream& sout, bool to_append, bool add_header)
{
    IBDIAG_ENTER;

    string err_message;
    int rc = IBFabric::OpenFile(file_name, sout, to_append, err_message, add_header);
    if (rc && !err_message.empty())
        this->SetLastError(err_message.c_str());

    IBDIAG_RETURN(rc);
}

// Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::GetReverseDirectRoute(INOUT direct_route_t *p_reverse_direct_route,
        IN direct_route_t *p_direct_route)
{
    IBDIAG_ENTER;
    CLEAR_STRUCT(*p_reverse_direct_route);
    IBNode *p_curr_node = this->root_node;
    if (!p_curr_node) {         //don't have root node
        this->SetLastError("DB error - can't find reverse direct route for direct route=%s - null root node",
                        this->ibis_obj.ConvertDirPathToStr(p_direct_route).c_str());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }
    int i, j;
    //TODO - validate p_direct_route->length
    p_reverse_direct_route->length = (u_int8_t)(p_direct_route->length - 1);    //without 0 at first byte
    for (i = 1, j = p_direct_route->length - 2; j >= 0; ++i, --j) {
        u_int8_t curr_port = p_direct_route->path.BYTE[i];
        if ((curr_port <= 0) || (curr_port > p_curr_node->numPorts)) {
            this->SetLastError("DB error - can't find reverse direct route for direct route=%s - original direct route with port out of range",
                    this->ibis_obj.ConvertDirPathToStr(p_direct_route).c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }
        if (!p_curr_node->getPort(curr_port) || !p_curr_node->getPort(curr_port)->p_remotePort) {
            this->SetLastError("DB error - can't find reverse direct route for direct route=%s - reached null port",
                    this->ibis_obj.ConvertDirPathToStr(p_direct_route).c_str());
            if (!p_curr_node->getPort(curr_port)) {
                IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Node %s port:%d is null\n",
                           p_curr_node->name.c_str(), curr_port);
            } else {
                IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Node %s the peer of port:%d is null\n",
                           p_curr_node->name.c_str(), curr_port);
            }
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }
        p_reverse_direct_route->path.BYTE[j] = p_curr_node->getPort(curr_port)->p_remotePort->num;
        p_curr_node = p_curr_node->getPort(curr_port)->p_remotePort->p_node;
        if (!p_curr_node) {
            this->SetLastError("DB error - can't find reverse direct route for direct route=%s - reached null node",
                    this->ibis_obj.ConvertDirPathToStr(p_direct_route).c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


// Returns: SUCCESS_CODE / ERR_CODE_EXCEEDS_MAX_HOPS
int IBDiag::ConcatDirectRoutes(IN direct_route_t *p_direct_route1,
        IN direct_route_t *p_direct_route2,
        OUT direct_route_t *p_direct_route_result)
{
    IBDIAG_ENTER;
    CLEAR_STRUCT(*p_direct_route_result);
    int i;
    if ((p_direct_route1->length + p_direct_route2->length) > IBDIAG_MAX_HOPS) {
        this->SetLastError("Concatenate of direct route1=%s with direct route2=%s exceeded max hops",
                this->ibis_obj.ConvertDirPathToStr(p_direct_route1).c_str(),
                this->ibis_obj.ConvertDirPathToStr(p_direct_route2).c_str());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_EXCEEDS_MAX_HOPS)
    }
    //insert first direct route elements at the beginning
    for (i = 0; i < p_direct_route1->length; ++i)
        p_direct_route_result->path.BYTE[i] = p_direct_route1->path.BYTE[i];
    //insert second direct route elements at the end
    for (i = 0; i < p_direct_route2->length; ++i)
        p_direct_route_result->path.BYTE[i + p_direct_route1->length] = p_direct_route2->path.BYTE[i];
    //update length
    p_direct_route_result->length = (u_int8_t)(p_direct_route1->length + p_direct_route2->length);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void IBDiag::ResetAppData(bool force)
{
    static bool done = false;
    if (!force && done) {
        return;
    }
    //mark the node as not checked (xmit wait, port counters extended capability, ext speed counters)
    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end();
            ++nI) {
        IBNode *p_node = (*nI).second;
        if (!p_node)
            continue;
        p_node->appData1.val = 0;
    }
    done = true;
}


/****************************************************/
/*******************public methods*******************/
/****************************************************/
const char* IBDiag::GetLastError()
{
    IBDIAG_ENTER;
    if (this->last_error != "")
        IBDIAG_RETURN(this->last_error.c_str());
    IBDIAG_RETURN("Unknown");
}

IBDiag::IBDiag():
		ibdiag_status(NOT_INITILIAZED), ibdiag_discovery_status(DISCOVERY_NOT_DONE),
		check_duplicated_guids(true), root_node(NULL), root_port_num(0), no_mepi(false),
		llr_active_cell_size(RETRANS_NO_RETRANS), ber_threshold(DEFAULT_BER_THRESHOLD)
{
    IBDIAG_ENTER;
    CLEAR_STRUCT(this->discover_progress_bar_nodes);
    CLEAR_STRUCT(this->discover_progress_bar_ports);
    IBDIAG_RETURN_VOID;
}

IBDiag::~IBDiag()
{
    IBDIAG_ENTER;
    //Make sure that there are no pending and on the wire MADs.
    this->ibis_obj.MadRecAll();

    this->CleanUpInternalDB();
    IBDIAG_RETURN_VOID;
}


// Returns: SUCCESS_CODE / ERR_CODE_INIT_FAILED
int IBDiag::Init()
{
    IBDIAG_ENTER;
    if (this->IsInit() == false) {
        IBDIAG_LOG(TT_LOG_LEVEL_INFO, "Making IBDM using internal log\n");
        ibdmUseInternalLog();

        IBDIAG_LOG(TT_LOG_LEVEL_INFO, "Initialize IBIS\n");
        if (this->ibis_obj.Init()) {
            this->SetLastError("Failed to init ibis object, err=%s", this->ibis_obj.GetLastError());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_INIT_FAILED);
        }

        if (this->capability_module.Init(&(this->ibis_obj))) {
            this->SetLastError("Failed to init capability_module object");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_INIT_FAILED);
        }
        this->ibdiag_status = NOT_SET_PORT;
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


// Returns: SUCCESS_CODE / ERR_CODE_INIT_FAILED
int IBDiag::SetPort(const char* device_name, phys_port_t port_num)
{
    IBDIAG_ENTER;
    if (this->IsInit() == false) {
        this->SetLastError("IBDiag initialize wasn't done");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INIT_FAILED);
    }
    if (this->IsReady() == true) {
        this->SetLastError("IBDiag set_port was already done");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INIT_FAILED);
    }
    IBDIAG_LOG(TT_LOG_LEVEL_INFO, "Setting IBDiag with (Device name=%s, Port number=%u)\n",
            device_name, port_num);
    if (this->ibis_obj.SetPort(device_name, port_num)) {
        this->SetLastError("Failed to set port of ibis object, err=%s", this->ibis_obj.GetLastError());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INIT_FAILED);
    }
    this->ibdiag_status = READY;
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


// Returns: SUCCESS_CODE / ERR_CODE_INIT_FAILED
int IBDiag::SetPort(u_int64_t port_guid)
{
    IBDIAG_ENTER;
    if (this->IsInit() == false) {
        this->SetLastError("IBDiag initialize wasn't done");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INIT_FAILED);
    }
    if (this->IsReady() == true) {
        this->SetLastError("IBDiag set_port was already done");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INIT_FAILED);
    }
    IBDIAG_LOG(TT_LOG_LEVEL_INFO, "Setting IBDiag with (GUID=" U64H_FMT ")\n",
            port_guid);
    if (this->ibis_obj.SetPort(ADB2C_CPU_TO_BE64(port_guid))) {
        this->SetLastError("Failed to set port of ibis object, err=%s", this->ibis_obj.GetLastError());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INIT_FAILED);
    }
    this->ibdiag_status = READY;
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


IBNode * IBDiag::GetNodeByDirectRoute(IN const direct_route_t *p_direct_route)
{
    IBDIAG_ENTER;
    //set root node
    IBNode *p_curr_node = this->root_node;
    if (!p_curr_node)
        IBDIAG_RETURN(NULL);

    for (int length = 1; length < p_direct_route->length; ++length) {
        u_int8_t curr_port = p_direct_route->path.BYTE[length];
        if ((curr_port <= 0) || (curr_port > p_curr_node->numPorts))
            IBDIAG_RETURN(NULL);
        if (!p_curr_node->getPort(curr_port) || !p_curr_node->getPort(curr_port)->p_remotePort)
            IBDIAG_RETURN(NULL);

        p_curr_node = p_curr_node->getPort(curr_port)->p_remotePort->p_node;
        if (!p_curr_node)
            IBDIAG_RETURN(NULL);
    }
    IBDIAG_RETURN(p_curr_node);
}


direct_route_t * IBDiag::GetDirectRouteByNodeGuid(IN u_int64_t guid)
{
    IBDIAG_ENTER;
    list_p_direct_route list_direct_routes = this->GetDirectRoutesByNodeGuid(guid);
    if (list_direct_routes.empty())
        IBDIAG_RETURN(NULL);
    IBDIAG_RETURN(list_direct_routes.front());
}


direct_route_t * IBDiag::GetDirectRouteByPortGuid(IN u_int64_t guid)
{
    IBDIAG_ENTER;
    list_p_direct_route list_direct_routes = this->GetDirectRoutesByPortGuid(guid);
    if (list_direct_routes.empty()) {
        IBDIAG_RETURN(NULL);
    }
    IBDIAG_RETURN(list_direct_routes.front());
}


IBPort * IBDiag::GetRootPort()
{
    IBDIAG_ENTER;
    IBPort *p_port = NULL;

    //get root port
    if (!this->root_node) {
        this->SetLastError("DB error - root node is null");
        IBDIAG_RETURN(NULL);
    }
    p_port = this->root_node->getPort(this->root_port_num);
    if (!p_port) {
        this->SetLastError("DB error - root port is null");
        IBDIAG_RETURN(NULL);
    }
    IBDIAG_RETURN(p_port);
}


// Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::GetLocalPortState(OUT u_int8_t& state)
{
    IBDIAG_ENTER;
    IBPort *p_root_port = this->GetRootPort();
    if (!p_root_port) {
        this->SetLastError("DB error - failed to found root port");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    //check if logical state in init
    struct SMP_PortInfo *p_root_port_info = this->fabric_extended_info.getSMPPortInfo(p_root_port->createIndex);
    if (!p_root_port_info) {
        this->SetLastError("DB error - failed to found port info for root port");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }
    state = p_root_port_info->PortState;
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


int IBDiag::GetAllLocalPortGUIDs(OUT local_port_t local_ports_array[IBIS_MAX_LOCAL_PORTS],
            OUT u_int32_t *p_local_ports_num)
{
    IBDIAG_ENTER;
    if (this->ibis_obj.GetAllLocalPortGUIDs(local_ports_array, p_local_ports_num)) {
        this->SetLastError("Failed to get all local ports from ibis object, err=%s",
                this->ibis_obj.GetLastError());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::getLatestSupportedVersion(int page_number,
                                      unsigned int &latest_version)
{
    IBDIAG_ENTER;
    switch (page_number) {
        case VS_MLNX_CNTRS_PAGE0:
            latest_version = PAGE0_LATEST_VER;
            break;
        case VS_MLNX_CNTRS_PAGE1:
            latest_version = PAGE1_LATEST_VER;
            break;
        case VS_MLNX_CNTRS_PAGE255:
            latest_version = PAGE255_LATEST_VER;
            break;
        default:
            this->SetLastError("Failed to found latest version"  \
                               " for page %d",page_number);
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            break;
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::DiscoverFabricFromFile(IN const string &csv_file,
        IN progress_func_discovered_t discover_progress_func)
{
    int rc = IBDIAG_SUCCESS_CODE;

    IBDiagFabric diag_fabric(discovered_fabric, fabric_extended_info);

    try {
        if((rc = diag_fabric.UpdateFabric(csv_file))) {
            this->SetLastError("Failed to parse CSV file");
            IBDIAG_RETURN(rc);
        }
    } catch (exception &e) {
        this->SetLastError("Failed to parse CSV file, error: %s", e.what());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_PARSE_FILE_FAILED);
    } catch (...) {
        this->SetLastError("Failed to parse CSV file");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_PARSE_FILE_FAILED);
    }

    if((rc = BuildDirectRoutesDB()))
        IBDIAG_RETURN(rc);

    discover_progress_bar_nodes.nodes_found = diag_fabric.getNodesFound();
    discover_progress_bar_nodes.sw_found = diag_fabric.getSWFound();
    discover_progress_bar_nodes.ca_found = diag_fabric.getCAFound();
    discover_progress_bar_ports.ports_found = diag_fabric.getPortsFound();

    discover_progress_func(&discover_progress_bar_nodes);

    IBDIAG_RETURN(rc);
}

IBSpecialPortType IBDiag::GetSpecialPortType(IBPort *p_port)
{
    if (!capability_module.IsSupportedSMPCapability(
            p_port->p_node, EnSMPCapIsSpecialPortsMarkingSupported))
        return NotSpecial;

    struct SMP_MlnxExtPortInfo *p_mepi = fabric_extended_info.
            getSMPMlnxExtPortInfo(p_port->createIndex);

    if (p_mepi && p_mepi->IsSpecialPort)
       return (IBSpecialPortType)p_mepi->SpecialPortType;

    return NotSpecial;
}

IBSpecialPortType IBDiag::GetSpecialCAPortType(IBNode *p_node)
{
    if (p_node->type != IB_CA_NODE)
        return NotSpecial;

    // go over all ports of this node:
    for (u_int32_t i = 1; i <= p_node->numPorts; ++i) {
        IBPort *p_port = p_node->getPort((phys_port_t)i);
        if (!p_port || p_port->get_internal_state() !=  IB_PORT_STATE_ACTIVE)
            continue;

        if (!p_port->p_remotePort ||
            p_port->p_remotePort->get_internal_state() !=  IB_PORT_STATE_ACTIVE)
            continue;

        return GetSpecialPortType(p_port->p_remotePort);
    }

    return NotSpecial;
}
