/*
 * Copyright (c) 2004-2013 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <infiniband/ibdiag/ibdiag_types.h>
#include <ibis/ibis.h>
#include <infiniband/ibdm/Regexp.h>
#include <arpa/inet.h>
#include <algorithm>

//mask to look at MSB prefix len bits
#define CAPABILITY_PREFIX_MASK(prefix_len) \
    ((u_int64_t)-1 << (sizeof(u_int64_t)*8 - prefix_len))

/*****************************************************************************/
/*                        capability_mask_t                                  */
/*****************************************************************************/
//bit is a number between 0 and 127, to set bit k in mask[w] -> bit=w*32+k
int capability_mask_t::set(u_int8_t bit)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    if (bit >= sizeof(this->mask)* 8)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);
    u_int8_t mask_index = (u_int8_t)(bit / (u_int8_t)CAPABILITY_MASK_FIELD_SIZE);
    u_int8_t bit_in_field = (u_int8_t)(bit % (u_int8_t)CAPABILITY_MASK_FIELD_SIZE);
    u_int32_t new_mask = 0x1 << bit_in_field;
    this->mask[mask_index] |= new_mask;

    IBDIAG_RETURN(rc);
}

//convert from network order to 'our' order: host order in each capability index,
//so that the specific word CapabilityMask[w] that stores a specific capability
//bit can be calculated by: w=CapabilityBit/32 and
//its offset b=CapabilityBit mod 32.
void capability_mask_t::ntoh()
{
    capability_mask_t original = *this;
    unsigned int mask_len = (unsigned int)arrsize(mask);

    for (unsigned int i = 0; i < mask_len; i++)
        mask[i] = ntohl(original.mask[mask_len -1 - i]);
}

//convert to network order
void capability_mask_t::hton()
{
    capability_mask_t original = *this;
    unsigned int mask_len = (unsigned int)arrsize(mask);

    for (unsigned int i = 0; i < mask_len; i++)
        mask[i] = htonl(original.mask[mask_len -1 - i]);
}

std::ostream& operator<< (std::ostream& stream,
                          const struct capability_mask& mask)
{
    capability_mask_t lmask = mask;
    //convert it
    lmask.hton();
    char str[INET6_ADDRSTRLEN];

    if (inet_ntop(AF_INET6, lmask.mask, str, INET6_ADDRSTRLEN))
        stream << str;
    else
        stream << "mask[0] " << mask.mask[0] << " mask[1] " << mask.mask[1]
               << " mask[2] " << mask.mask[2] << " mask[3] " << mask.mask[3];

    return stream;
}

/*****************************************************************************/
/*                        query_or_mask_t                                    */
/*****************************************************************************/
std::ostream& operator<< (std::ostream& stream,
                          const struct query_or_mask& qmask)
{
    stream << "to query: " << qmask.to_query <<", mask: " << qmask.mask;
    return stream;
}

/*****************************************************************************/
/*                        prefix_guid_data_t                                 */
/*****************************************************************************/
std::ostream& operator<< (std::ostream& stream,
                          const struct prefix_guid_data& guid_data)
{
    stream << "original guid: 0x" << std::hex << guid_data.original_guid
           << ", to query: " << std::boolalpha << guid_data.qmask.to_query
           << ", mask: " << guid_data.qmask.mask;
    return stream;
}

/*****************************************************************************/
/*                        CapabilityMaskConfig                               */
/*****************************************************************************/
int CapabilityMaskConfig::Init(Ibis* ibis_obj)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    device_id_list_t old_dev_ids_list, mlnx_devs, bull_devs;
    device_id_list_t::iterator it;
    fw_version_obj_t fw, fw_zero;

    fw_zero.major = fw_zero.minor = fw_zero.sub_minor = 0;

    //update old devices: Anafa,BridgeXIB,Tavor,Sinai,Arbel with mask 0
    //as they don't support GI queries at all, and don't support VS capabilities
    ibis_obj->GetAnafaDevIds(old_dev_ids_list);
    ibis_obj->GetBridgeXIBDevIds(old_dev_ids_list);
    ibis_obj->GetTavorDevIds(old_dev_ids_list);
    ibis_obj->GetSinaiDevIds(old_dev_ids_list);
    ibis_obj->GetArbelDevIds(old_dev_ids_list);
    capability_mask_t mask;
    mask.clear();

    for(it = old_dev_ids_list.begin(); it != old_dev_ids_list.end(); it++)
        this->AddUnsupportMadDevice(MELLANOX_VEN_ID, *it, mask);

    query_or_mask_t qmask, query;

    qmask.mask.clear();
    qmask.to_query = false;
    //GMP InitialMask: LLRStatistics
    //SMP InitialMask: CableInfo, EyeOpen, ExtendedPortInfo
    InitMask(qmask.mask);

    query.to_query = true;

    //ConnectX3 devices
    InitFWConnectX3(fw);
    mlnx_devs.clear();
    bull_devs.clear();
    ibis_obj->GetConnectX_3IBDevIds(mlnx_devs, bull_devs);

    //SMP and GMP: Mellanox, connectX3 devices, fw: 0.0.0 mask: InitialMask
    //SMP and GMP: Mellanox, connectX3 devices, fw: 2.33.5100 query
    for(it = mlnx_devs.begin(); it != mlnx_devs.end(); it++) {
        this->AddFwDevice(MELLANOX_VEN_ID, *it, fw_zero, qmask);
        this->AddFwDevice(MELLANOX_VEN_ID, *it, fw, query);
    }

    //SMP and GMP: Bull, connectX3 devices, fw: 0.0.0 mask: InitialMask
    //SMP and GMP: Bull, connectX3 devices, fw: 2.33.5100 query
    for(it = bull_devs.begin(); it != bull_devs.end(); it++) {
        this->AddFwDevice(BULL_VEN_ID, *it, fw_zero, qmask);
        this->AddFwDevice(BULL_VEN_ID, *it, fw, query);
    }

    //ConnectIB/Golan devices
    InitFWConnectIB(fw);
    mlnx_devs.clear();
    //SMP and GMP: Mellanox, connectIB(Golan) devices, fw: 0.0.0 mask: initialMask
    //GMP: Mellanox, connectIB(Golan) devices, fw: 10.10.4000 query
    //SMP: Mellanox, connectIB(Golan) devices, fw: 10.10.3220 query
    ibis_obj->GetGolanDevIds(mlnx_devs);
    for(it = mlnx_devs.begin(); it != mlnx_devs.end(); it++) {
        this->AddFwDevice(MELLANOX_VEN_ID, *it, fw_zero, qmask);
        this->AddFwDevice(MELLANOX_VEN_ID, *it, fw, query);
    }


    mlnx_devs.clear();
    bull_devs.clear();
    fw.major = 0x00000009;//9
    fw.minor = 0x00000002;//2
    fw.sub_minor = 0x00001772;//6002
    ibis_obj->GetSwitchXIBDevIds(mlnx_devs, bull_devs);
    //SMP and GMP: Mellanox, switchX devices, fw: 0.0.0 mask: InitialMask
    //SMP and GMP: Mellanox, switchX devices, fw: 9.2.6002 query
    for(it = mlnx_devs.begin(); it != mlnx_devs.end(); it++) {
        this->AddFwDevice(MELLANOX_VEN_ID, *it, fw_zero, qmask);
        this->AddFwDevice(MELLANOX_VEN_ID, *it, fw, query);
    }
    //the same goes for BULL switchX devices
    for(it = bull_devs.begin(); it != bull_devs.end(); it++) {
        this->AddFwDevice(BULL_VEN_ID, *it, fw_zero, qmask);
        this->AddFwDevice(BULL_VEN_ID, *it, fw, query);
    }

    IBDIAG_RETURN(rc);
}

int CapabilityMaskConfig::AddFw(u_int64_t guid, fw_version_obj_t &fw)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    m_guid_2_fw.insert( std::pair<u_int64_t, fw_version_obj_t>(guid, fw) );

    IBDIAG_RETURN(rc);
}

int CapabilityMaskConfig::GetFw(u_int64_t guid, fw_version_obj_t &fw)
{
    int rc = IBDIAG_SUCCESS_CODE;

    map_uint64_fw_version_obj_t::iterator it = m_guid_2_fw.find(guid);

    if (it != m_guid_2_fw.end())
        fw = it->second;
    else
        rc = IBDIAG_ERR_CODE_CHECK_FAILED;
    return rc;
}

int CapabilityMaskConfig::AddCapabilityMask(u_int64_t guid, capability_mask_t &mask)
{
    int rc = IBDIAG_SUCCESS_CODE;
    map_uint64_capability_mask_t::iterator it = m_guid_2_mask.find(guid);

    //if we already have an entry, then it's an error
    if (it != m_guid_2_mask.end())
        rc = IBDIAG_ERR_CODE_DB_ERR;

    //override if it exists
    //example: SMP GeneralInfo MAD, when we send fw atribute modifier,
    //that can fail, then send capability atribute modifier that successfully,
    //then get fw from GMP GI MAD, and overide capability with configured
    //mask per fw.
    m_guid_2_mask[guid] = mask;
    return rc;
}

void CapabilityMaskConfig::AddFwDevice(u_int32_t ven_id, device_id_t dev_id, fw_version_obj_t &fw, query_or_mask_t &query_or_mask)
{
    pair_uint32_device_id_t key(ven_id, dev_id);
    map_ven_id_dev_id_2_fw_data_t::iterator it = m_fw_devices.find(key);

    if (it != m_fw_devices.end())
        //overwrite
        it->second[fw] = query_or_mask;
    else {
        map_fw_to_query_or_mask_t fw_2_query_map;
        fw_2_query_map.insert(std::pair<fw_version_obj_t, query_or_mask_t>(fw, query_or_mask));
        m_fw_devices.insert(std::pair<pair_uint32_device_id_t, map_fw_to_query_or_mask_t>(key, fw_2_query_map));
    }
}

void CapabilityMaskConfig::RemoveFwDevice(u_int32_t ven_id, device_id_t dev_id)
{
    map_ven_id_dev_id_2_fw_data_t::iterator it =
        m_fw_devices.find(pair_uint32_device_id_t(ven_id, dev_id));

    if (it != m_fw_devices.end()) {
        cout
            << "-I- Removing in " << this->m_what_mask
            << " section mad supported device from data base: VenID: "
            << std::hex << it->first.first << " DevID: " << it->first.second
            << std::dec << endl;
        m_fw_devices.erase(it);
    }
}

//receives vendor id, device id, fw and returns mask
//if there is such entry <ven id, dev id> <fw1, mask>,
//where fw1 is the greatest lower bound,
//i.e fw1 < fw and if there exists another fw2 <= fw, then fw2 <= fw1
//example: fw=6.0.0, fw1 = 5.0.0, fw2 = 0.0.0
// <ven id, dev id>, <<0,0,0>, mask1>
// <ven id, dev id>, <<5.0.0>, mask2>, for querying fw=<6.0.0>, it will receive
// mask2 as answer
//also returns if this is the only fw entry for this <ven id, dev id>
int CapabilityMaskConfig::GetFwConfiguredMask(uint32_t ven_id,
                                              device_id_t dev_id,
                                              fw_version_obj_t &fw,
                                              capability_mask_t &mask,
                                              bool *is_only_fw)
{
    map_ven_id_dev_id_2_fw_data_t::iterator it =
        m_fw_devices.find(pair_uint32_device_id_t(ven_id, dev_id));
    if (it != m_fw_devices.end()) {
        //returns the iterator to the first element whose key is not greater than k.
        map_fw_to_query_or_mask_t::iterator it1 = it->second.lower_bound(fw);
        if (it1 != it->second.end() && it1->second.to_query == false) {
            mask = it1->second.mask;
            if (is_only_fw)
                *is_only_fw = it->second.size() == 1 ? true : false;
            return IBDIAG_SUCCESS_CODE;
        }
    }
    return IBDIAG_ERR_CODE_DB_ERR;
}

void CapabilityMaskConfig::AddUnsupportMadDevice(u_int32_t ven_id,
                                                 device_id_t dev_id,
                                                 capability_mask_t &mask)
{
    pair_uint32_device_id_t key(ven_id, dev_id);

    //set a new value or override existing
    m_unsupported_mad_devices[key] = mask;
}

void CapabilityMaskConfig::RemoveUnsupportMadDevice(u_int32_t ven_id,
                                                    device_id_t dev_id)
{
    map_ven_dev_to_capability_mask_t::iterator it =
        m_unsupported_mad_devices.find(pair_uint32_device_id_t(ven_id, dev_id));

    if (it != m_unsupported_mad_devices.end()) {
        cout
            << "-I- Removing in " << this->m_what_mask
            << " section unsupported mad device from data base: VenID: "
            << std::hex << it->first.first << " DevID: " << it->first.second
            << " mask: " << it->second << std::dec << endl;
        m_unsupported_mad_devices.erase(it);
    }
}

//prefix_len - number of prefix bits
int CapabilityMaskConfig::AddPrefixGuid(u_int8_t prefix_len, u_int64_t guid,
                                        query_or_mask_t &qmask, string &output)
{
    IBDIAG_ENTER;
    u_int64_t prefix_mask = CAPABILITY_PREFIX_MASK(prefix_len);
    u_int64_t masked_guid = guid & prefix_mask;
    u_int8_t anyErr = 0;
    prefix_guid_data_t guid_data;

    guid_data.original_guid = guid;
    guid_data.qmask = qmask;

    map_prefix_2_guids_data_t::iterator it = m_prefix_guids_2_mask.find(prefix_len);

    if (it != m_prefix_guids_2_mask.end()) {
        map_guid_to_prefix_data_t::iterator it1 = it->second.find(masked_guid);
        stringstream tmp_stream;
        if (it1 != it->second.end()) {
            //such element already existed
            tmp_stream << "guid prefix 0x" << std::hex << masked_guid
                << " (prefix: " << std::dec << (int)prefix_len
                << ") already exists with value of "
                << it1->second;
            output = tmp_stream.str();
            anyErr++;

        }
        //insert the data
        it->second[masked_guid] = guid_data;
    } else {
        map_guid_to_prefix_data_t quids_2_prefix_data_map;
        quids_2_prefix_data_map.insert(std::pair<u_int64_t, prefix_guid_data_t>
                                       (masked_guid, guid_data));
        m_prefix_guids_2_mask.insert(
            std::pair<u_int8_t, map_guid_to_prefix_data_t>
            (prefix_len, quids_2_prefix_data_map));
    }
    IBDIAG_RETURN(anyErr);
}

bool CapabilityMaskConfig::IsUnsupportedMadDevice(uint32_t ven_id,
                                                  device_id_t dev_id,
                                                  capability_mask_t &mask)
{
    pair_uint32_device_id_t key = pair_uint32_device_id_t(ven_id, dev_id);
    map_ven_dev_to_capability_mask_t::iterator it = m_unsupported_mad_devices.find(key);
    if (it !=  m_unsupported_mad_devices.end()) {
        //found
        mask = it->second;
        return true;
    }

    return false;
}

bool CapabilityMaskConfig::IsMaskKnown(u_int64_t guid)
{
    map_uint64_capability_mask_t::iterator it = m_guid_2_mask.find(guid);
    return it != m_guid_2_mask.end();
}

bool CapabilityMaskConfig::IsSupportedCapability(IBNode *node, u_int8_t cap_bit)
{
    IBDIAG_ENTER;

    //check for illegal bit
    if (cap_bit < m_mask_first_bit || cap_bit > m_mask_last_bit)
        IBDIAG_RETURN(false);
    u_int32_t result = 0;

    map_uint64_capability_mask_t::iterator it = m_guid_2_mask.find(node->guid_get());

    if (it != m_guid_2_mask.end()) {
        capability_mask_t mask = it->second;

        // index = cap_bit/32
        u_int8_t capability_index = (u_int8_t)(cap_bit / (u_int8_t)CAPABILITY_MASK_FIELD_SIZE);
        //check valid index and bit
        if (capability_index < NUM_CAPABILITY_FIELDS) {
            // bit inside capability field = cap_bit mod 32
            u_int8_t bit = (u_int8_t)(cap_bit % (u_int8_t)CAPABILITY_MASK_FIELD_SIZE);
            u_int32_t bit_mask = 0x1 << bit;
            result = mask.mask[capability_index] & bit_mask;
        }
    }
    //printf("IsSupportedCapability:%d node:%s result:%d\n",
    //       cap_bit, node->getName().c_str(), result != 0);
    IBDIAG_RETURN(result != 0);
}

bool CapabilityMaskConfig::IsLongestPrefixMatch(u_int64_t guid,
                                                u_int8_t &prefix_len,
                                                u_int64_t &matched_guid,
                                                query_or_mask &qmask)
{
    IBDIAG_ENTER;
    map_prefix_2_guids_data_t::reverse_iterator it,
        itEnd = m_prefix_guids_2_mask.rend();
    bool found = false;

    //start from the longest prefix length
    for (it = m_prefix_guids_2_mask.rbegin(); it != itEnd; it++) {
        u_int64_t prefix_mask = CAPABILITY_PREFIX_MASK(it->first);
        u_int64_t masked_guid = guid & prefix_mask;
        map_guid_to_prefix_data_t::iterator it1 =
            it->second.find(masked_guid);
        if (it1 != it->second.end()) {
            //matched prefix found
            prefix_len = it->first;
            matched_guid = it1->second.original_guid;
            qmask = it1->second.qmask;
            found = true;
            break;
        }
    }
    IBDIAG_RETURN(found);
}

int CapabilityMaskConfig::DumpCapabilityMaskFile(ofstream& sout)
{
    IBDIAG_ENTER;
    char buffer[1024] = {'\0'};
    capability_mask_t mask;
    char mask_str[INET6_ADDRSTRLEN];
    int anyErr = 0;

    sout << m_section_header << endl
         << m_what_mask << endl;

    //section header
    sout << endl << m_unsupported_mad_devs_header << endl;

    //go over devices that don't support the GI mad (no fw)
    map_ven_dev_to_capability_mask_t::iterator unsupported_mad_it =
        m_unsupported_mad_devices.begin();

    for (; unsupported_mad_it != m_unsupported_mad_devices.end();
          unsupported_mad_it++) {
        mask = unsupported_mad_it->second;
        mask.hton();
        if (!inet_ntop(AF_INET6, mask.mask, mask_str, INET6_ADDRSTRLEN)) {
            cout << "-E- failed to print mask for VenID: 0x"
                 << std::hex << unsupported_mad_it->first.first
                 << " DevID: 0x" << std::hex << unsupported_mad_it->first.second << endl;
            anyErr ++;
            continue;
        }
        sout << "VenID: 0x" << std::hex << unsupported_mad_it->first.first
             << " DevID: 0x" << std::hex << unsupported_mad_it->first.second
             << " mask: " << mask_str << endl;
    }

    //section header
    sout << endl << m_fw_devs_header << endl;

    //go over devices that support the GI mad, have different mask per fw
    map_ven_id_dev_id_2_fw_data_t::iterator fw_it = m_fw_devices.begin();
    for (; fw_it != m_fw_devices.end(); fw_it++) {
        map_fw_to_query_or_mask_t fw_2_qmask_map = fw_it->second;
        map_fw_to_query_or_mask_t::reverse_iterator it = fw_2_qmask_map.rbegin();
        for (; it != fw_2_qmask_map.rend(); it++) {
         if (it->second.to_query)
             sprintf(buffer, "query");
         else {
             mask = it->second.mask;
             mask.hton();
             if (inet_ntop(AF_INET6, mask.mask, mask_str, INET6_ADDRSTRLEN))
                 sprintf(buffer, "mask: %s", mask_str);
             else {
                 cout << "-E- failed to print mask for VenID: 0x"
                      << std::hex << fw_it->first.first
                      << " DevID: 0x" << std::hex
                      << fw_it->first.second
                      << " FW: " << std::dec << it->first.major << "."
                      << it->first.minor << "." << it->first.sub_minor << endl;
                 anyErr ++;
                 continue;
             }
         }
         sout << "VenID: 0x" << std::hex << fw_it->first.first
              << " DevID: 0x" << std::hex << fw_it->first.second
              << " FW: " << std::dec << it->first.major << "."
                         << it->first.minor << "."
                         << it->first.sub_minor << " "
              << buffer << endl;
        }
    }

    //section header
    sout << endl << m_prefix_guids_2_mask_header << endl;
    //go over devices with different mask per guid
    map_prefix_2_guids_data_t::iterator guid_it =
         m_prefix_guids_2_mask.begin();
    for (; guid_it != m_prefix_guids_2_mask.end(); guid_it++) {
        map_guid_to_prefix_data_t guid_2_data_map = guid_it->second;
        map_guid_to_prefix_data_t::iterator data_it = guid_2_data_map.begin();
        for (; data_it != guid_2_data_map.end(); data_it++) {
            if (data_it->second.qmask.to_query)
                sprintf(buffer, "query");
            else {
                mask = data_it->second.qmask.mask;
                mask.hton();
                if (inet_ntop(AF_INET6, mask.mask, mask_str, INET6_ADDRSTRLEN))
                    sprintf(buffer, "mask: %s", mask_str);
                else {
                    cout << "-E- failed to print mask for NodeGUID: 0x"
                         << std::hex << data_it->second.original_guid
                         << "/" << guid_it->first << endl;
                    anyErr ++;
                    continue;
                }
            }
            sout << "NodeGUID: 0x" << std::hex << data_it->second.original_guid
                 << "/" << std::dec << (u_int32_t)guid_it->first << " " << buffer << endl;
        }
    }
    sout << endl << m_section_footer << endl << endl;
    IBDIAG_RETURN(anyErr);
}

int CapabilityMaskConfig::DumpGuid2Mask(ofstream& sout,
                                        IBFabric *p_discovered_fabric)
{
    IBDIAG_ENTER;
    map_uint64_capability_mask_t::iterator it = m_guid_2_mask.begin();
    capability_mask_t mask;
    char mask_str[INET6_ADDRSTRLEN];
    int anyErr = 0;

    sout << m_what_mask << endl;
    sout << "# node name, node guid,  discovered capability mask" << endl;
    for (; it != m_guid_2_mask.end(); it++) {
        IBNode *p_node = p_discovered_fabric->getNodeByGuid(it->first);
        sout << (p_node ? p_node->getName() : "NA")
             << ", 0x" << std::hex << it->first << ", ";
        mask = it->second;
        mask.hton();
        if (inet_ntop(AF_INET6, mask.mask, mask_str, INET6_ADDRSTRLEN))
            sout << mask_str << endl;
        else {
            sout << "NA" << endl;
            anyErr ++;
        }
    }
    //return decimal output as a default for numeric values
    sout << std::dec << m_section_footer << endl;

    IBDIAG_RETURN(anyErr);
}

int CapabilityMaskConfig::GetCapability(IBNode *node, capability_mask_t& mask)
{
    IBDIAG_ENTER;
    map_uint64_capability_mask_t::iterator it = m_guid_2_mask.find(node->guid_get());

    //if we don't have an entry
    if (it == m_guid_2_mask.end())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_NOT_READY);

    mask = it->second;
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

/*****************************************************************************/
/*                        SmpMask class                                      */
/*****************************************************************************/
SmpMask::SmpMask() : CapabilityMaskConfig(EnSMPCapFirst, EnSMPCapLast)
{
    m_what_mask = "SMP";
    m_section_header = "# Starting of " + m_what_mask + " Vendor Specific Attributes section";
    m_section_footer = "END" + m_what_mask;
    m_unsupported_mad_devs_header =
        "# Capability Mask per vendor id and device id\n"
        "#---------------------------------------------\n"
        "# Here come devices that don't support GeneralInfo MAD.\n"
        "# The Capability Mask is presented in IPv6 format\n"
        "# NOTE: It is illegal to define a rule for a DevID with no FW and then define one with FW.\n"
        "# Example: VenID: 0x2c9 DevID: 0xc738 mask: ::3\n"
        "#          meaning devices of vendor 0x2c9 and device 0xc738\n"
        "#          the supported SMP vendor specific attributes are\n"
        "#          those of capability with mask bit 1 and 2";

    m_fw_devs_header =
        "# Capability Mask per vendor id, device id and firmware:\n"
        "#------------------------------------------------------\n"
        "# * Use this section only for devices that support GeneralInfoSMP\n"
        "# * If the device started supporting the CapabilityMask feature on GeneralInfoSMP\n"
        "#   instead of provding the mask, write 'query'\n"
        "# * The firmware version provided is the minimal version from which the given\n"
        "#   Capability Mask is valid.\n"
        "#\n"
        "# Example 1: Providing Capability Masks explicitly:\n"
        "#   Given the following lines:\n"
        "#     VenID: 0x2c9 DevID: 0x245 FW: 0.0.0   mask: C1\n"
        "#     VenID: 0x2c9 DevID: 0x245 FW: 1.1.100 mask: C2\n"
        "#     VenID: 0x2c9 DevID: 0x245 FW: 1.65.0  mask: C3\n"
        "#     VenID: 0x2c9 DevID: 0x245 FW: 2.0.1   mask: C4\n"
        "#   We can conclude the CapabilityMask according to a FW version:\n"
        "#     C1 for version 1.0.0\n"
        "#     C2 for version 1.1.105\n"
        "#     C4 for version 2.1.1\n"
        "#\n"
        "# Example 2: Providing initial Capability Mask explicitly and\n"
        "# allowing query for the mask from some version:\n"
        "#   Given the following lines:\n"
        "#     VenID: 0x2c9 DevID: 0x245 FW: 0.0.0   mask: ::3\n"
        "#     VenID: 0x2c9 DevID: 0x245 FW: 1.0.100 query\n"
        "#";

    m_prefix_guids_2_mask_header =
        "# Capability Mask per Node GUID or per a set of Node GUIDs.\n"
        "#----------------------------------------------------------\n"
        "# Use this section if you have devices that do not support GeneralInfoSMP\n"
        "# and some of them differ from the others - so different Capability Masks\n"
        "# are supported by different devices.\n"
        "#\n"
        "# * Matching is performed by longest prefix. So if a device' NodeGUID is\n"
        "#   matching both a full 64-bit provided rule and a rule of 24-bit\n"
        "#   (just by the OUI), the 64-bit rule wins and Capability Mask provided\n"
        "#   by it will be used.\n"
        "# * Node Guid format is either 0x[0-9a-f]+ or [0-9a-f]+ or xxxx:xxxx:xxxx:xxxx\n"
        "# * Node Guid can be followed by 'query' or mask: <CapabilityMask>.\n"
        "#  'query' means that Capability Mask can be queried via GeneralInfoSMP MAD.\n"
        "# Example: NodeGUID: 0x0029100000000/24 mask: ::1\n"
        "# Example: NodeGUID: 0002:c903:005a:aa57/64 query";
}

//Init mask for SwitchX, ConnectX3, Golan.
//These devices have old fw that doesn't support capability mask,
//so store the mask for them, that tells what capabilities are supported
void SmpMask::InitMask(capability_mask_t &mask)
{
    IBDIAG_ENTER;
    mask.set(EnSMPCapIsCableInfoSupported);
    mask.set(EnSMPCapIsSMPEyeOpenSupported);
    mask.set(EnSMPCapIsExtendedPortInfoSupported);
    IBDIAG_RETURN_VOID;
}

//Init FW version, from which connectX3 supports capability mask
void SmpMask::InitFWConnectX3(fw_version_obj_t &fw)
{
    //starting 2.33.5100 connectX3 supports capability mask in SMP
    fw.major = 0x00000002;//2
    fw.minor = 0x00000021;//33
    fw.sub_minor = 0x00013ec;//5100
}

//Init FW version, from which connectIB/Golan supports capability mask
void SmpMask::InitFWConnectIB(fw_version_obj_t &fw)
{
    //starting 10.10.3220 ConnectIB/Golan supports capability mask in SMP
    fw.major = 0x0000000a;//10
    fw.minor = 0x0000000a;//10
    fw.sub_minor = 0x00000c94;//3220
}

int SmpMask::Init(Ibis* ibis_obj)
{
    IBDIAG_ENTER;

    //call the common Init to SMP and GMP
    int rc = CapabilityMaskConfig::Init(ibis_obj);

    device_id_list_t mlnx_devs, volt_shaldag_devs, bull_devs;
    device_id_list_t::iterator it;
    capability_mask_t mask;

    mask.clear();

    //add Shaldag devices
    ibis_obj->GetShaldagDevIds(mlnx_devs, volt_shaldag_devs);
    //Voltaire, shaldag devices mask: 0
    for(it = volt_shaldag_devs.begin(); it != volt_shaldag_devs.end(); it++)
        this->AddUnsupportMadDevice(VOLTAIRE_VEN_ID, *it, mask);
    //Mellanox, shaldag devices mask: CableInfo, EyeOpen, other zero
    mask.set(EnSMPCapIsCableInfoSupported);
    mask.set(EnSMPCapIsSMPEyeOpenSupported);
    for(it = mlnx_devs.begin(); it != mlnx_devs.end(); it++)
        this->AddUnsupportMadDevice(MELLANOX_VEN_ID, *it, mask);

    mlnx_devs.clear();

    //connectX devices that don't support MAD, support CableInfo and EyeOpen
    ibis_obj->GetConnectXDevIds(mlnx_devs);
    ibis_obj->GetConnectX_2DevIds(mlnx_devs);
    ibis_obj->GetConnectX_2_ENtDevIds(mlnx_devs);
    ibis_obj->GetConnectX_2_LxDevIds(mlnx_devs);
    for(it = mlnx_devs.begin(); it != mlnx_devs.end(); it++)
        this->AddUnsupportMadDevice(MELLANOX_VEN_ID, *it, mask);

    IBDIAG_RETURN(rc);
}

/*****************************************************************************/
/*                        GmpMask class                                      */
/*****************************************************************************/
GmpMask::GmpMask(): CapabilityMaskConfig(EnGMPCapFirst, EnGMPCapLast)
{
    m_what_mask = "GMP";
    m_section_header = "# Starting of " + m_what_mask + " Vendor Specific Class 0x000A section";
    m_section_footer = "END" + m_what_mask;
    m_unsupported_mad_devs_header =
        "# The same syntax and usage model provided for SMP's is\n"
        "# applicable for GMPs\n"
        "\n"
        "# Capability Mask per vendor id and device id.\n"
        "# Example: VenID: 0x2c9 DevID: 0xc738 mask: ::2\n";

    m_fw_devs_header =
        "# Capability Mask per vendor id, device id and firmware.\n"
        "# Example: VenID: 0x2c9 DevID: 0x245 FW: 0.0.0 mask: ::1\n"
        "# Example: VenID: 0x2c9 DevID: 0x245 FW: 1.0.0 query.\n"
        "# 'query' means Capability mask can be queried via GeneralInfo MAD.";

    m_prefix_guids_2_mask_header =
        "# Capability Mask per Node GUID or per a set of Node GUIDs.\n"
        "# Example: NodeGUID: 0x0029100000000/24 mask: ::1\n"
        "# Example: NodeGUID: 0002:c903:005a:aa57/64 query";
}

//Init mask for SwitchX, ConnectX3, Golan.
//These devices have old fw that doesn't support capability mask,
//so store the mask for them, that tells what capabilities are supported
void GmpMask::InitMask(capability_mask_t &mask)
{
    IBDIAG_ENTER;
    mask.set(EnGMPCapIsPortLLRStatisticsSupported);
    IBDIAG_RETURN_VOID;
}

//Init FW version, from which connectX3 supports capability mask
void GmpMask::InitFWConnectX3(fw_version_obj_t &fw)
{
    //starting 2.33.5100 connectX3 supports capability mask in GMP
    fw.major = 0x00000002;//2
    fw.minor = 0x00000021;//33
    fw.sub_minor = 0x00013ec;//5100
}

//Init FW version, from which connectIB/Golan supports capability mask
void GmpMask::InitFWConnectIB(fw_version_obj_t &fw)
{
    //starting 10.10.4000 ConnectIB/Golan supports capability mask in GMP
    fw.major = 0x0000000a;//10
    fw.minor = 0x0000000a;//10
    fw.sub_minor = 0x0000fa0;//4000
}

int GmpMask::Init(Ibis* ibis_obj)
{
    IBDIAG_ENTER;

    //call the common Init to SMP and GMP
    int rc = CapabilityMaskConfig::Init(ibis_obj);

    device_id_list_t mlnx_devs, volt_shaldag_devs, bull_devs;
    device_id_list_t::iterator it;
    capability_mask_t mask;

    mask.clear();

    //add Shaldag devices
    ibis_obj->GetShaldagDevIds(mlnx_devs, volt_shaldag_devs);
    //Voltaire, shaldag devices mask: 0
    for(it = volt_shaldag_devs.begin(); it != volt_shaldag_devs.end(); it++)
        this->AddUnsupportMadDevice(VOLTAIRE_VEN_ID, *it, mask);
    //Mellanox shaldag devices mask: 0
    for(it = mlnx_devs.begin(); it != mlnx_devs.end(); it++)
        this->AddUnsupportMadDevice(MELLANOX_VEN_ID, *it, mask);

    IBDIAG_RETURN(rc);
}


/*****************************************************************************/
/*                       CapabilityModule class                              */
/*****************************************************************************/

int CapabilityModule::Init(Ibis *ibis_obj)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    rc = smp_mask.Init(ibis_obj);
    if (rc != IBDIAG_SUCCESS_CODE)
        return rc;
    rc = gmp_mask.Init(ibis_obj);
    IBDIAG_RETURN(rc);
}

int CapabilityModule::DumpCapabilityMaskFile(const char *file_name)
{
    IBDIAG_ENTER;
    ofstream sout;
    int anyErr = 0;

    sout.open(file_name, ios_base::out);
    if (sout.fail()) {
        cout << "-E- Failed to open file " << file_name << " for writing"
             << endl;
        IBDIAG_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
    }

    sout << "# This capability masks configuration file was automatically"
            " generated by ibdiagnet" << endl << endl;

    //go over smp data structures
    anyErr = smp_mask.DumpCapabilityMaskFile(sout);
    sout << endl;
    anyErr += gmp_mask.DumpCapabilityMaskFile(sout);

    IBDIAG_RETURN(anyErr);
}

//for debug purposes
int CapabilityModule::DumpGuid2Mask(const char *file_name,
                                    IBFabric *p_discovered_fabric)
{
    IBDIAG_ENTER;
    ofstream sout;
    int anyErr = 0;

    sout.open(file_name, ios_base::app);
    if (sout.fail()) {
        cout << "-E- Failed to open file " << file_name << " for writing"
             << endl;
        IBDIAG_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
    }

    //go over smp data structures
    anyErr = smp_mask.DumpGuid2Mask(sout, p_discovered_fabric);
    sout << endl;
    anyErr += gmp_mask.DumpGuid2Mask(sout, p_discovered_fabric);

    IBDIAG_RETURN(anyErr);
}

int CapabilityModule::ParseCapabilityMaskFile(const char *file_name)
{
    IBDIAG_ENTER;
    unsigned int lineNum = 0;
    char sLine[1024] = {'\0'};
    int smp_start = -1, smp_end = -1, gmp_start = -1, gmp_end = -1;
    //mask is presented in IPv6 address format
    //devices that don't support the General Info mads, so no fw available
    //VenId: 0x02c9 DevID: 0x1003 mask: 2001:0db8:85a3:0000:0000:8a2e:0370:7334
    regExp noFwDevLine("^(VenID:)\\s*(0x[0-9a-fA-F]+|[0-9a-fA-F]+)\\s*(DevID:)\\s*(0x[0-9a-fA-F]+|[0-9a-fA-F]+)\\s*"
                       "(mask:)\\s*([0-9a-fA-F:.]+)\\s*");
    //VenId: 0x02c9 DevID: 0xc738 FW: 0.0.0 mask: 2001:0db8:85a3::8a2e:0370:7334
    //VenId: 0x02f1 DevID: 0xc738 FW: 0.0.0 query
    regExp fwDevLine("^(VenID:)\\s*(0x[0-9a-fA-F]+|[0-9a-fA-F]+)\\s*(DevID:)\\s*(0x[0-9a-fA-F]+|[0-9a-fA-F]+)\\s*"
                     "(FW:)\\s*([0-9]+[.][0-9]+[.][0-9]+)\\s*((query)|(mask:)\\s*([0-9a-fA-F:.]+))\\s*");
    //NodeGUID: 0x0029100000000/24 mask: ::5
    //NodeGUID: 0x0029100000000/64 query
    regExp nodeGuidLine("^(NodeGUID:)\\s*(0x[0-9a-fA-F]+/[1-6]?[0-9]|[0-9a-fA-F]+/[1-6]?[0-9]|[0-9a-fA-F:]+/[1-6]?[0-9])\\s*"
                        "((query)|(mask:)\\s*([0-9a-fA-F:.]+))\\s*");
    regExp ignoreLine("^\\s*(#|$)");
    rexMatch *p_rexRes = NULL;
    u_int32_t ven_id = 0;
    device_id_t dev_id = 0;
    //map <<ven id, dev id>, line number in file>
    map<pair_uint32_device_id_t, unsigned int> no_fw_dev_map_gmp, fw_dev_map_gmp;
    map<pair_uint32_device_id_t, unsigned int> no_fw_dev_map_smp, fw_dev_map_smp;
    map<pair_uint32_device_id_t, unsigned int>::iterator it;
    bool illegal_line;
    pair_uint32_device_id_t ven_dev_id;

    ifstream fs(file_name);

    if (fs.fail()) {
        cout << "-E- Failed to open file:" << file_name << endl;
        return IBDIAG_ERR_CODE_CHECK_FAILED;
    }

    cout << "-I- Parsing Capability Mask file: " << file_name << endl;
    unsigned int anyErr = 0;

    //parse file
    while (fs.good()) {
        fs.getline(sLine, 1024);
        lineNum++;

        if (!strcmp(sLine, "SMP")) {
            if (smp_start != -1) {
                cout << "-E- SMP was found at line " << lineNum
                     << ", previous occurrence at line " << smp_start << endl;
                anyErr++;
            }
            smp_start = lineNum;
            if (smp_end != -1) {
                cout << "-E- SMP was found at line " << lineNum
                     << ", SMP END was already found at line " << smp_end << endl;
                anyErr++;
            }
        } else if (!strcmp(sLine, "ENDSMP")) {
            if (smp_end != -1) {
                cout << "-E- ENDSMP was found at line " << lineNum
                     << ", previous occurrence at line " << smp_end << endl;
                anyErr++;
            }
            smp_end = lineNum;
            if (smp_start == -1) {
                cout << "-E- ENDSMP was found at line " << lineNum
                     << ", there was no occurrence of SMP yet" << endl;
                anyErr++;
            }
        } else if (!strcmp(sLine, "GMP")) {
            if (gmp_start != -1) {
                cout << "-E- GMP was found at line " << lineNum
                     << ", previous occurrence at line " << gmp_start << endl;
                anyErr++;
            }
            gmp_start = lineNum;
            if (gmp_end != -1) {
                cout << "-E- GMP was found at line " << lineNum
                     << ", ENDGMP was already found at line " << gmp_end << endl;
                anyErr++;
            }
        } else if (!strcmp(sLine, "ENDGMP")) {
            if (gmp_end != -1) {
                cout << "-E- ENDGMP was found at line " << lineNum
                     << ", previous occurrence at line " << gmp_end << endl;
                anyErr++;
            }
            gmp_end = lineNum;
            if (gmp_start == -1) {
                cout << "-E- ENDGMP was found at line " << lineNum
                     << ", there was no occurrence of GMP yet" << endl;
                anyErr++;
            }
        } else if ((p_rexRes = noFwDevLine.apply(sLine))) {
            ven_id = (u_int32_t)strtoul(p_rexRes->field(2).c_str(), NULL, 16);
            dev_id = (device_id_t)strtoul(p_rexRes->field(4).c_str(), NULL, 16);
            ven_dev_id = pair_uint32_device_id_t(ven_id, dev_id);

            illegal_line = false;

            //remember that we have this ven_id + dev_id as unsupported mad device
            if (smp_start != -1 && smp_end == -1) {
                no_fw_dev_map_smp.insert(
                    std::pair<pair_uint32_device_id_t, u_int32_t>(
                        pair_uint32_device_id_t(ven_id, dev_id), lineNum));
                it = fw_dev_map_smp.find(ven_dev_id);
                illegal_line = (it != fw_dev_map_smp.end()) ? true : false;

            } else if (gmp_start != -1 && gmp_end == -1) {
                no_fw_dev_map_gmp.insert(
                    std::pair<pair_uint32_device_id_t, u_int32_t>(
                        pair_uint32_device_id_t(ven_id, dev_id), lineNum));
                it = fw_dev_map_gmp.find(ven_dev_id);
                illegal_line = (it != fw_dev_map_gmp.end()) ? true : false;
            }
            //check if this ven_id+dev_id was also in the 'fw' devices
            if (illegal_line) {
                cout << "-E- Illegal configuration. Vendor id: " <<  p_rexRes->field(2).c_str()
                     << " device id: " << p_rexRes->field(4).c_str()
                     << " at line " << lineNum << " appear without FW, already appeared with FW at line " << it->second
                     << ". Can't have both" << endl;
                anyErr++;
            }
        } else if ((p_rexRes = fwDevLine.apply(sLine))) {
            //VenId: 0x02c9 DevID: 0xc738 FW: 0.0.0 mask: 0x8000,1101,0000,0000,0000
            ven_id = (u_int32_t)strtoul(p_rexRes->field(2).c_str(), NULL, 16);
            dev_id = (device_id_t)strtoul(p_rexRes->field(4).c_str(), NULL, 16);
            ven_dev_id = pair_uint32_device_id_t(ven_id, dev_id);

            illegal_line = false;

            //remember that we have this ven_id + dev_id as fw device
            if (smp_start != -1 && smp_end == -1) {
                fw_dev_map_smp.insert(
                    std::pair<pair_uint32_device_id_t, u_int32_t>(
                        pair_uint32_device_id_t(ven_id, dev_id), lineNum));
                it = no_fw_dev_map_smp.find(ven_dev_id);
                illegal_line = (it != no_fw_dev_map_smp.end()) ? true : false;

            } else if (gmp_start != -1 && gmp_end == -1) {
                fw_dev_map_gmp.insert(
                    std::pair<pair_uint32_device_id_t, u_int32_t>(
                        pair_uint32_device_id_t(ven_id, dev_id), lineNum));
                it = no_fw_dev_map_gmp.find(ven_dev_id);
                illegal_line = (it != no_fw_dev_map_gmp.end()) ? true : false;

            }
            //check if this ven_id+dev_id was also in the 'no mad supported' devices
            if (illegal_line) {
                cout << "-E- Illegal configuration. Vendor id: " <<  p_rexRes->field(2).c_str()
                     << " device id: " << p_rexRes->field(4).c_str()
                     << " at line " << lineNum << " appear with FW, already appeared without FW at line " << it->second
                     << ". Can't have both" << endl;
                anyErr++;
            }

        } else if ((p_rexRes = nodeGuidLine.apply(sLine))) {
            ;
        } else if  ((p_rexRes = ignoreLine.apply(sLine))) {
            ;
        } else {
            cout << "-E- Unsupported syntax at line: " << lineNum << endl;
            anyErr++;
        }

        if (p_rexRes) {
            delete p_rexRes;
            p_rexRes = NULL;
        }
    }
    fs.close();

    if (anyErr)
        return IBDIAG_ERR_CODE_CHECK_FAILED;

    // Make second pass and build the tables
    fs.open(file_name, ifstream::in);
    if (fs.fail()) {
        cout << "-E- Fail to open file:" << file_name << endl;
        return IBDIAG_ERR_CODE_CHECK_FAILED;
    }
    lineNum = 0;
    p_rexRes = NULL;
    smp_start = smp_end = gmp_start = gmp_end = -1;
    string output;

    while (fs.good()) {
        fs.getline(sLine,1024);
        lineNum++;

        if (!strcmp(sLine, "SMP"))
            smp_start = lineNum;
        else if (!strcmp(sLine, "ENDSMP"))
            smp_end = lineNum;
        else if (!strcmp(sLine, "GMP"))
            gmp_start = lineNum;
        else if (!strcmp(sLine, "ENDGMP"))
            gmp_end = lineNum;
        else if ((p_rexRes = noFwDevLine.apply(sLine))) {
            //VenId: 0x02c9 DevID: 0x1003 mask: 2001:0db8:85a3:0000:0000:8a2e:0370:7334
            ven_id = (u_int32_t)strtoul(p_rexRes->field(2).c_str(), NULL, 16);
            dev_id = (device_id_t)strtoul(p_rexRes->field(4).c_str(), NULL, 16);
            capability_mask_t mask;

            mask.clear();
            if (inet_pton(AF_INET6, p_rexRes->field(6).c_str(), &mask) == 1) {
                //convert network to host order
                mask.ntoh();
                //Add the data to the data base.
                //if internal default data base holds this device with fw version, then remove it
                //as user now adds it without fw, i.e unsupported, and it can't be both in the data base.
                //Also user doesn't have this device with fw in its file,
                //it was validated on the first pass over the file.

                //in smp section
                if (smp_start != -1 && (smp_end == -1 || (unsigned int)smp_end == lineNum)) {
                    smp_mask.AddUnsupportMadDevice(ven_id, dev_id, mask);
                    smp_mask.RemoveFwDevice(ven_id, dev_id);
                } else if (gmp_start != -1 && (gmp_end == -1 || (unsigned int)gmp_end == lineNum)) {
                    gmp_mask.AddUnsupportMadDevice(ven_id, dev_id, mask);
                    gmp_mask.RemoveFwDevice(ven_id, dev_id);
                } else {
                    cout << "-E- device at line " << lineNum << " is out of SMP or GMP section. Skipping" << endl;
                    anyErr++;
                }
            } else {
                cout << "-E- device at line " << lineNum << " has unsupported mask format" << endl;
                anyErr++;
            }
        } else if ((p_rexRes = fwDevLine.apply(sLine))) {
            //VenId: 0x02c9 DevID: 0xc738 FW: 0.0.0 mask: 2001:0db8:85a3::8a2e:0370:7334
            //VenId: 0x02f1 DevID: 0xc738 FW: 0.0.0 query
            ven_id = (u_int32_t)strtoul(p_rexRes->field(2).c_str(), NULL, 16);
            dev_id = (device_id_t)strtoul(p_rexRes->field(4).c_str(), NULL, 16);

            fw_version_obj_t fw;

            if (fw.set(p_rexRes->field(6).c_str())) {
                cout << "-E device at line " << lineNum << " has unsupported fw format" << endl;
                anyErr++;
            } else {
                query_or_mask_t qmask;

                if (!strcmp(p_rexRes->field(9).c_str(), "mask:"))
                    if (inet_pton(AF_INET6, p_rexRes->field(10).c_str(), &qmask.mask) == 1) {
                        qmask.to_query = false;
                        qmask.mask.ntoh();
                    } else {
                        cout << "-E- device at line " << lineNum << " has unsupported mask format" << endl;
                        anyErr++;
                        delete p_rexRes;
                        p_rexRes = NULL;
                        continue;
                    }
                else
                    //query
                    qmask.to_query = true;

                //Add the data to the data base.
                //if internal default data base holds this device as unsupported, then remove it
                //as user now adds it with fw, i.e supported, and it can't be both in the data base.
                //Also user doesn't have this device as unsupported in its file,
                //it was validated on the first pass over the file.
                if (smp_start != -1 && (smp_end == -1 || (unsigned int)smp_end == lineNum)) {
                    smp_mask.AddFwDevice(ven_id, dev_id, fw, qmask);
                    smp_mask.RemoveUnsupportMadDevice(ven_id, dev_id);
                } else if (gmp_start != -1 && (gmp_end == -1 || (unsigned int)gmp_end == lineNum)) {
                    gmp_mask.AddFwDevice(ven_id, dev_id, fw, qmask);
                    gmp_mask.RemoveUnsupportMadDevice(ven_id, dev_id);
                } else {
                    cout << "-E- device at line " << lineNum << " is out of SMP or GMP section. Skipping" << endl;
                    anyErr++;
                }
            }
        } else if ((p_rexRes = nodeGuidLine.apply(sLine))) {
            //NodeGUID: 0x0029100000000/24 mask: ::5
            //NodeGUID: 0x0029100000000/64 query
            //NodeGUID: 0000:0291:0000:0000/64 mask: ::1
            string guid_str = p_rexRes->field(2);
            //remove ':' if hsa such in representation
            guid_str.erase(std::remove(guid_str.begin(),
                                       guid_str.end(),
                                       ':'),
                           guid_str.end()) ;

            u_int64_t guid = strtoull(guid_str.c_str(), NULL, 16);
            char *prefix = strstr((char*)guid_str.c_str(), "/");
            u_int8_t prefix_len = 0;

            if (prefix) {
                prefix_len = (u_int8_t)strtoul(prefix + 1, NULL, 10);
                if (prefix_len == 0 || prefix_len > 64) {
                    cout << "-E- NodeGuid at line " << lineNum << " has illegal prefix length" << endl;
                    anyErr++;
                    delete p_rexRes;
                    p_rexRes = NULL;
                    continue;
                }
                query_or_mask_t qmask;
                if (!strcmp(p_rexRes->field(5).c_str(), "mask:"))
                    if (inet_pton(AF_INET6, p_rexRes->field(6).c_str(), &qmask.mask) == 1) {
                        qmask.to_query = false;
                        qmask.mask.ntoh();
                    } else {
                        cout << "-E- NodeGUID at line " << lineNum << " has unsupported mask format" << endl;
                        anyErr++;
                        delete p_rexRes;
                        p_rexRes = NULL;
                        continue;
                    }
                else
                    qmask.to_query = true;

                //Add guid,qmask to prefixes map
                output.clear();
                //int AddPrefixGuid(u_int8_t prefix_len, u_int64_t guid,query_or_mask_t qmask);
                if (smp_start != -1 &&
                    (smp_end == -1 || (unsigned int)smp_end == lineNum)) {
                    if (smp_mask.AddPrefixGuid(prefix_len, guid, qmask, output)) {
                        cout << "-W- at line " << lineNum << ": " << output << endl;
                        anyErr++;
                    }
                } else if (gmp_start != -1 &&
                         (gmp_end == -1 || (unsigned int)gmp_end == lineNum)) {
                    if (gmp_mask.AddPrefixGuid(prefix_len, guid, qmask, output)) {
                            cout << "-W- at line " << lineNum << ": " << output << endl;
                            anyErr++;
                    }
                } else {
                    cout << "-E- NodeGUID at line " << lineNum << " is out of SMP or GMP section. Skipping" << endl;
                    anyErr++;
                }


            } else {
                cout << "-E- Unsupported NodeGUID format at line " << lineNum << endl;
                anyErr++;
            }
        }
        if (p_rexRes) {
            delete p_rexRes;
            p_rexRes = NULL;
        }
    }
    fs.close();
    if (anyErr)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int CapabilityModule::AddSMPFw(u_int64_t guid, fw_version_obj_t &fw)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    rc = smp_mask.AddFw(guid, fw);

    IBDIAG_RETURN(rc);
}

int CapabilityModule::AddSMPCapabilityMask(u_int64_t guid, capability_mask_t &mask)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    rc = smp_mask.AddCapabilityMask(guid, mask);

    IBDIAG_RETURN(rc);
}

void CapabilityModule::AddSMPFwDevice(u_int32_t ven_id,
                                      device_id_t dev_id,
                                      fw_version_obj_t &fw,
                                      query_or_mask_t &query_or_mask)
{
    smp_mask.AddFwDevice(ven_id, dev_id, fw, query_or_mask);
}

bool CapabilityModule::IsSMPUnsupportedMadDevice(uint32_t ven_id,
                                                 device_id_t dev_id,
                                                 capability_mask_t &mask)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(smp_mask.IsUnsupportedMadDevice(ven_id, dev_id, mask));
}

bool CapabilityModule::IsSMPMaskKnown(u_int64_t guid)
{
    return smp_mask.IsMaskKnown(guid);
}

bool CapabilityModule::IsSupportedSMPCapability(IBNode *node, u_int8_t cap_bit)
{
    return smp_mask.IsSupportedCapability(node, cap_bit);
}

int CapabilityModule::AddGMPFw(u_int64_t guid, fw_version_obj_t &fw)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(gmp_mask.AddFw(guid, fw));
}

int CapabilityModule::GetGMPFw(u_int64_t guid, fw_version_obj_t &fw)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(gmp_mask.GetFw(guid, fw));
}

int CapabilityModule::AddGMPCapabilityMask(u_int64_t guid, capability_mask_t &mask)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    rc = gmp_mask.AddCapabilityMask(guid, mask);

    IBDIAG_RETURN(rc);
}



int CapabilityModule::GetSMPFw(u_int64_t guid, fw_version_obj_t &fw)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(smp_mask.GetFw(guid, fw));
}

bool CapabilityModule::IsLongestSMPPrefixMatch(u_int64_t guid,
                                               u_int8_t &prefix_len,
                                               u_int64_t &matched_guid,
                                               query_or_mask &qmask)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(smp_mask.IsLongestPrefixMatch(guid, prefix_len,
                                                matched_guid, qmask));
}

int CapabilityModule::GetSMPFwConfiguredMask(uint32_t ven_id,
                                             device_id_t dev_id,
                                             fw_version_obj_t &fw,
                                             capability_mask_t &mask,
                                             bool *is_only_fw)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(smp_mask.GetFwConfiguredMask(ven_id, dev_id,
                                               fw, mask, is_only_fw));
}

bool CapabilityModule::IsGMPUnsupportedMadDevice(uint32_t ven_id,
                                                 device_id_t dev_id,
                                                 capability_mask_t &mask)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(gmp_mask.IsUnsupportedMadDevice(ven_id, dev_id, mask));
}

bool CapabilityModule::IsGMPMaskKnown(u_int64_t guid)
{
    return gmp_mask.IsMaskKnown(guid);
}

bool CapabilityModule::IsLongestGMPPrefixMatch(u_int64_t guid,
                                               u_int8_t &prefix_len,
                                               u_int64_t &matched_guid,
                                               query_or_mask &qmask)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(gmp_mask.IsLongestPrefixMatch(guid, prefix_len,
                                                matched_guid, qmask));
}

int CapabilityModule::GetGMPFwConfiguredMask(uint32_t ven_id,
                                             device_id_t dev_id,
                                             fw_version_obj_t &fw,
                                             capability_mask_t &mask,
                                             bool *is_only_fw)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(gmp_mask.GetFwConfiguredMask(ven_id, dev_id,
                                               fw, mask, is_only_fw));
}

bool CapabilityModule::IsSupportedGMPCapability(IBNode *node, u_int8_t cap_bit)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(gmp_mask.IsSupportedCapability(node, cap_bit));
}

int CapabilityModule::GetFw(u_int64_t guid, fw_version_obj_t &fw)
{
    IBDIAG_ENTER;
    int rc;

    if ((rc = GetSMPFw(guid, fw)) == IBDIAG_SUCCESS_CODE)
        IBDIAG_RETURN(rc);
    IBDIAG_RETURN(GetGMPFw(guid, fw));
}

int CapabilityModule::GetCapability(IBNode *node, bool is_gmp,
                                    capability_mask_t& mask)
{
    if (is_gmp)
        return gmp_mask.GetCapability(node, mask);

    return smp_mask.GetCapability(node, mask);
}
