/*
 *  Copyright (c) 2004-2019 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <stdarg.h>
#include "infiniband/ibdiag/ibdiag_clbck.h"

IBDiagClbck ibDiagClbck;



void IBDiagClbck::SetLastError(const char *fmt, ...)
{
    IBDIAG_ENTER;
    char buff[IBDIAG_ERR_SIZE];
    va_list args;

    memset(buff, 0, sizeof(buff));
    va_start(args, fmt);
    vsprintf(buff, fmt, args);
    va_end(args);
    m_LastError.assign(buff);
    IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "-E- %s\n", m_LastError.c_str());
    IBDIAG_RETURN_VOID;
}

const char* IBDiagClbck::GetLastError()
{
    IBDIAG_ENTER;
    if (m_LastError != "")
        IBDIAG_RETURN(m_LastError.c_str());
    IBDIAG_RETURN("Unknown");
}

void IBDiagPMCapMaskClbck(const clbck_data_t &clbck_data,
                          int rec_status,
                          void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        PMCapMaskClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::PMCapMaskClbck(
                    const clbck_data_t &clbck_data,
                    int rec_status,
                    void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    IBNode *p_node = (IBNode *)clbck_data.m_data1;
    if (rec_status) {
        FabricErrNodeNotRespond *p_fabric_pm_err =
                new FabricErrNodeNotRespond(p_node, "PMClassPortInfoGet");
        if (!p_fabric_pm_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_fabric_pm_err);
    } else {
        struct IB_ClassPortInfo *p_class_port_info =
            (struct IB_ClassPortInfo *)p_attribute_data;
        int rc = m_pFabricExtendedInfo->addPMCapMask(p_node, p_class_port_info->CapMsk);
        if (rc) {
            SetLastError("Failed to add PMCapMask for node=%s, err=%s",
                    p_node->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagPMOptionMaskClbck(const clbck_data_t &clbck_data,
                             int rec_status,
                             void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        PMOptionMaskClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::PMOptionMaskClbck(
                    const clbck_data_t &clbck_data,
                    int rec_status,
                    void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    IBNode *p_node = (IBNode *)clbck_data.m_data1;
    if (rec_status) {
        FabricErrNodeNotRespond *p_fabric_pm_err =
                new FabricErrNodeNotRespond(p_node, "PMPortSamplesControl");
        if (!p_fabric_pm_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_fabric_pm_err);
    } else {
        struct PM_PortSamplesControl *p_sample_control =
            (struct PM_PortSamplesControl *)p_attribute_data;
        int rc = m_pFabricExtendedInfo->addPMOptionMask(
                p_node,
                p_sample_control->PortSampleControlOptionMask);
        if (rc) {
            SetLastError("Failed to add PMOptionMask for node=%s, err=%s",
                    p_node->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagPMPortCountersGetClbck(
                          const clbck_data_t &clbck_data,
                          int rec_status,
                          void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        PMPortCountersGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::PMPortCountersGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    if (rec_status) {
        FabricErrPortNotRespond *p_fabric_pm_err =
                new FabricErrPortNotRespond(p_port, "PMPortCountersGet");
        if (!p_fabric_pm_err) {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_fabric_pm_err);
    } else {
        struct PM_PortCounters *p_port_counters=
            (struct PM_PortCounters *)p_attribute_data;
        int rc = m_pFabricExtendedInfo->addPMPortCounters(p_port, *p_port_counters);
        if (rc) {
            SetLastError("Failed to add PMPortCounters for port=%s, err=%s",
                    p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagPMPortCountersClearClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        PMPortCountersClearClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::PMPortCountersClearClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    if (rec_status) {
        FabricErrPortNotRespond *p_fabric_pm_err =
                new FabricErrPortNotRespond(p_port, "PMPortCountersClear");
        if (!p_fabric_pm_err) {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_fabric_pm_err);
    }
}

void IBDiagPMPortCountersExtendedGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        PMPortCountersExtendedGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::PMPortCountersExtendedGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    if (rec_status) {

        FabricErrPortNotRespond *p_fabric_pm_err =
                new FabricErrPortNotRespond(p_port, "PMPortCountersExtendedGet");
        if (!p_fabric_pm_err) {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_fabric_pm_err);
    } else {
        struct PM_PortCountersExtended *p_port_counters=
            (struct PM_PortCountersExtended *)p_attribute_data;
        int rc = m_pFabricExtendedInfo->addPMPortCountersExtended(p_port, *p_port_counters);
        if (rc) {
            SetLastError("Failed to add PMPortCountersExtended for port=%s, err=%s",
                    p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagPMPortCountersExtendedClearClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        PMPortCountersExtendedClearClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::PMPortCountersExtendedClearClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    if (rec_status) {

        FabricErrPortNotRespond *p_fabric_pm_err =
                new FabricErrPortNotRespond(p_port, "PMPortCountersExtendedClear");
        if (!p_fabric_pm_err) {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_fabric_pm_err);
    }
}

void IBDiagPMPortExtendedSpeedsGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        PMPortExtendedSpeedsGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::PMPortExtendedSpeedsGetClbck(
    const clbck_data_t &clbck_data,
    int rec_status,
    void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    if (rec_status) {
        FabricErrPortNotRespond *p_fabric_pm_err =
                new FabricErrPortNotRespond(p_port, "PMPortExtendedSpeedsCountersGet");
        if (!p_fabric_pm_err) {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_fabric_pm_err);
    } else {
        struct PM_PortExtendedSpeedsCounters *p_port_counters=
            (struct PM_PortExtendedSpeedsCounters *)p_attribute_data;
        int rc = m_pFabricExtendedInfo->addPMPortExtSpeedsCounters(p_port, *p_port_counters);
        if (rc) {
            SetLastError("Failed to add PMPortExtendedSpeedsCounters for port=%s, err=%s",
                    p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagPMPortExtendedSpeedsClearClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        PMPortExtendedSpeedsClearClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::PMPortExtendedSpeedsClearClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    if (rec_status) {
        FabricErrPortNotRespond *p_fabric_pm_err =
                new FabricErrPortNotRespond(p_port, "PMPortExtendedSpeedsCountersClear");
        if (!p_fabric_pm_err) {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_fabric_pm_err);
    }
}

void IBDiagPMPortExtendedSpeedsRSFECGetClbck(const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        PMPortExtendedSpeedsRSFECGetClbck(clbck_data, rec_status,
                                          p_attribute_data);
}

void IBDiagClbck::PMPortExtendedSpeedsRSFECGetClbck(
    const clbck_data_t &clbck_data,
    int rec_status,
    void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific-don't need them here

    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    if (rec_status) {
        if (IS_SUPPORT_EXT_SPEEDS_RSFEC_FAILED(p_port->p_node->appData1.val))
            //already reported error
            IBDIAG_RETURN_VOID;
        //mark we already checked for support
        p_port->p_node->appData1.val |= NOT_SUPPORT_EXT_SPEEDS_RSFEC_COUNTERS;

        FabricErrPortNotRespond *p_fabric_pm_err =
                new FabricErrPortNotRespond(
                    p_port, "PMPortExtendedSpeedsRSFECCountersGet");
        if (!p_fabric_pm_err) {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_fabric_pm_err);
    } else {
        struct PM_PortExtendedSpeedsRSFECCounters *p_port_counters =
            (struct PM_PortExtendedSpeedsRSFECCounters *)p_attribute_data;
        int rc = m_pFabricExtendedInfo->addPMPortExtSpeedsRSFECCounters(
                                        p_port, *p_port_counters);
        if (rc) {
            SetLastError("Failed to add PMPortExtendedSpeedsRSFECCounters "
                         "for port=%s, err=%s",
                         p_port->getName().c_str(),
                         m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagPMPortExtendedSpeedsRSFECClearClbck(const clbck_data_t &clbck_data,
                          int rec_status,
                          void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        PMPortExtendedSpeedsRSFECClearClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::PMPortExtendedSpeedsRSFECClearClbck(
    const clbck_data_t &clbck_data,
    int rec_status,
    void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific-don't need them here

    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    if (rec_status &&
        !IS_SUPPORT_EXT_SPEEDS_RSFEC_FAILED(p_port->p_node->appData1.val)) {
        //mark we already checked for support
        p_port->p_node->appData1.val |= NOT_SUPPORT_EXT_SPEEDS_RSFEC_COUNTERS;
        FabricErrPortNotRespond *p_fabric_pm_err =
                new FabricErrPortNotRespond(
                    p_port, "PMPortExtendedSpeedsRSFECCountersClear");
        if (!p_fabric_pm_err) {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_fabric_pm_err);
    }
}

void IBDiagVSPortLLRStatisticsGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        VSPortLLRStatisticsGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::VSPortLLRStatisticsGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    if (rec_status) {
        if (IS_SUPPORT_LLR_FAILED(p_port->p_node->appData1.val))
            //already reported error
            IBDIAG_RETURN_VOID
        //mark we already checked for support
        p_port->p_node->appData1.val |= NOT_SUPPORT_LLR_COUNTERS;
        FabricErrPortNotRespond *p_fabric_pm_err =
                new FabricErrPortNotRespond(p_port, "VSPortLLRStatisticsGet");
        if (!p_fabric_pm_err) {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_fabric_pm_err);
    } else {
        struct VendorSpec_PortLLRStatistics *p_port_counters=
            (struct VendorSpec_PortLLRStatistics *)p_attribute_data;
        int rc = m_pFabricExtendedInfo->addVSPortLLRStatistics(p_port, *p_port_counters);
        if (rc) {
            SetLastError("Failed to add VSPortLLRStatistics for port=%s, err=%s",
                    p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
    IBDIAG_RETURN_VOID
}

void IBDiagVSPortLLRStatisticsClearClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        VSPortLLRStatisticsClearClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::VSPortLLRStatisticsClearClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    if (rec_status && !IS_SUPPORT_LLR_FAILED(p_port->p_node->appData1.val)) {
        //mark we already checked for support
        p_port->p_node->appData1.val |= NOT_SUPPORT_LLR_COUNTERS;
        if ((bool)(intptr_t)(clbck_data.m_data2)) {
            FabricErrPortNotRespond *p_fabric_pm_err =
                    new FabricErrPortNotRespond(p_port, "VSPortLLRStatisticsClear");
            if (!p_fabric_pm_err) {
                SetLastError("Failed to allocate FabricErrPortNotRespond");
                m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
            } else
                m_pErrors->push_back(p_fabric_pm_err);
        }
    }
    IBDIAG_RETURN_VOID
}

void IBDiagPMPortRcvErrorDetailsGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        PMPortRcvErrorDetailsGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::PMPortRcvErrorDetailsGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    if (rec_status)
    {
        FabricErrPortNotRespond *p_fabric_pm_err =
            new FabricErrPortNotRespond(p_port, "PMPortRcvErrorDetailsGet");
        if (!p_fabric_pm_err)
        {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        }
        else
            m_pErrors->push_back(p_fabric_pm_err);
    }
    else
    {
        struct PM_PortRcvErrorDetails *p_port_rcv_error_details;
        p_port_rcv_error_details = (struct PM_PortRcvErrorDetails *)p_attribute_data;
        int rc = m_pFabricExtendedInfo->addPMPortRcvErrorDetails(p_port, *p_port_rcv_error_details);
        if (rc)
        {
            SetLastError("Failed to add PM_PortRcvErrorDetails for port=%s, err=%s",
                p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagPMPortRcvErrorDetailsClearClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        PMPortRcvErrorDetailsClearClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::PMPortRcvErrorDetailsClearClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    if (rec_status)
    {
        FabricErrPortNotRespond *p_fabric_pm_err =
                new FabricErrPortNotRespond(p_port, "PMPortRcvErrorDetailsClear");
        if (!p_fabric_pm_err)
        {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        }
        else
            m_pErrors->push_back(p_fabric_pm_err);
    }
}

void IBDiagPMPortXmitDiscardDetailsGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        PMPortXmitDiscardDetailsGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::PMPortXmitDiscardDetailsGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    if (rec_status)
    {
        FabricErrPortNotRespond *p_fabric_pm_err =
                new FabricErrPortNotRespond(p_port, "PMPortXmitDiscardDetailsGet");
        if (!p_fabric_pm_err)
        {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        }
        else
            m_pErrors->push_back(p_fabric_pm_err);
    }
    else
    {
        struct PM_PortXmitDiscardDetails *p_port_xmit_discard_details;
        p_port_xmit_discard_details = (struct PM_PortXmitDiscardDetails *)p_attribute_data;
        int rc = m_pFabricExtendedInfo->addPMPortXmitDiscardDetails(p_port,
                                                                    *p_port_xmit_discard_details);
        if (rc)
        {
            SetLastError("Failed to add PM_PortXmitDiscardDetails for port=%s, err=%s",
            p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}

void IBDiagPMPortXmitDiscardDetailsClearClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        PMPortXmitDiscardDetailsClearClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::PMPortXmitDiscardDetailsClearClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    if (rec_status)
    {
        FabricErrPortNotRespond *p_fabric_pm_err =
                new FabricErrPortNotRespond(p_port, "PMPortXmitDiscardDetailsClear");
        if (!p_fabric_pm_err)
        {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        }
        else
            m_pErrors->push_back(p_fabric_pm_err);
    }
}

//Callback to GMP GeneralInfo MAD
void IBDiagVSGeneralInfoGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        VSGeneralInfoGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::VSGeneralInfoGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag || !m_p_capability_module)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    IBNode *p_node = (IBNode *)clbck_data.m_data1;

    if (rec_status == IBIS_MAD_STATUS_UNSUP_METHOD_ATTR) {         //not supported for this device
        FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                new FabricErrNodeNotSupportCap(p_node,
                        "The firmware of this device does not support general info GMP capability");
        if (!p_curr_fabric_node_err) {
            SetLastError("Failed to allocate FabricErrNodeNotSupportCap");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_curr_fabric_node_err);
    } else if (rec_status) {
        FabricErrNodeNotRespond *p_curr_fabric_node_err =
                new FabricErrNodeNotRespond(p_node, "VSGeneralInfo");
        if (!p_curr_fabric_node_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_curr_fabric_node_err);
    } else {
        struct VendorSpec_GeneralInfo *p_general_info=
            (struct VendorSpec_GeneralInfo *)p_attribute_data;
        u_int64_t guid = p_node->guid_get();

        m_ErrorState = m_pFabricExtendedInfo->addVSGeneralInfo(p_node, *p_general_info);
        if (m_ErrorState)
            SetLastError("Failed to add VSGeneralInfo for node=%s, err=%s",
                    p_node->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
        fw_version_obj_t gmp_fw;
        gmp_fw = *p_general_info;
        //update the fw in the cache
        m_ErrorState = m_p_capability_module->AddGMPFw(guid, gmp_fw);
        if (m_ErrorState)
            SetLastError("Failed to add GMP Fw Info for node=%s",
                    p_node->getName().c_str());
        capability_mask_t gmp_mask, smp_mask;

        u_int8_t prefix_len = 0;
        u_int64_t matched_guid = 0;
        query_or_mask_t qmask;
        bool gmp_query_per_matched_guid = false;

        //update the guid 2 gmp mask cache only if gmp mask is unknown
        if (!m_p_capability_module->IsGMPMaskKnown(guid)) {
            if (m_p_capability_module->IsLongestGMPPrefixMatch(guid,
                                                               prefix_len,
                                                               matched_guid,
                                                               qmask) &&
                qmask.to_query == true)
                gmp_query_per_matched_guid = true;

            //take the mask from the MAD, if either guid - 'query'
            //configuration match was found or there is NO fw configured device
            if (gmp_query_per_matched_guid ||
                m_p_capability_module->GetGMPFwConfiguredMask(p_node->vendId,
                                                              p_node->devId,
                                                              gmp_fw,
                                                              gmp_mask) !=
                IBDIAG_SUCCESS_CODE)
                gmp_mask = p_general_info->CapabilityMask;

            //update the guid 2 mask cache
            if (m_p_capability_module->AddGMPCapabilityMask(guid, gmp_mask) !=
                IBDIAG_SUCCESS_CODE) {
                FabricErrSmpGmpCapMaskExist *p_gmp_mask_exist_err =
                        new FabricErrSmpGmpCapMaskExist(p_node, false, gmp_mask);
                if (!p_gmp_mask_exist_err) {
                    SetLastError(
                        "Failed to allocate FabricErrSmpGmpCapMaskExist");
                    m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
                } else
                    m_pErrors->push_back(p_gmp_mask_exist_err);
            }
        }

        //try to update SMP mask for this node that fw dependant
        if (!m_p_capability_module->IsSMPMaskKnown(guid))
        {
            bool smp_query_per_matched_guid = false;
            if (m_p_capability_module->IsLongestSMPPrefixMatch(guid,
                                                               prefix_len,
                                                               matched_guid,
                                                               qmask) &&
                qmask.to_query == true)
                smp_query_per_matched_guid = true;
            //if no 'guid to query' configuration exists,
            //check for fw configured mask
            if (!smp_query_per_matched_guid &&
                m_p_capability_module->GetSMPFwConfiguredMask(p_node->vendId,
                                                              p_node->devId,
                                                              gmp_fw,
                                                              smp_mask) ==
                IBDIAG_SUCCESS_CODE)
                    //update the SMP guid 2 mask cache
                    if (m_p_capability_module->AddSMPCapabilityMask(guid, smp_mask)) {

                        FabricErrSmpGmpCapMaskExist *p_smp_mask_exist_err =
                            new FabricErrSmpGmpCapMaskExist(p_node, true, smp_mask);

                        if (!p_smp_mask_exist_err) {
                            SetLastError("Failed to allocate FabricErrSmpGmpCapMaskExist",
                                         p_node->getName().c_str());
                            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
                        } else
                            m_pErrors->push_back(p_smp_mask_exist_err);
                    }
        }
    }

    IBDIAG_RETURN_VOID
}

void IBDiagSMPGUIDInfoTableGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SMPGUIDInfoTableGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SMPGUIDInfoTableGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    IBPort *p_port = (IBPort *)clbck_data.m_data1;

    if (rec_status) {
        if (p_port->p_node->appData1.val)
            IBDIAG_RETURN_VOID;

        FabricErrNodeNotRespond *p_curr_fabric_guid_err =
                new FabricErrNodeNotRespond(p_port->p_node, "SMPGUIDInfoTableGetByLid");
        if (!p_curr_fabric_guid_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else {
            m_pErrors->push_back(p_curr_fabric_guid_err);
            p_port->p_node->appData1.val = 1;
        }
    } else {
        struct SMP_GUIDInfo *p_guid_info =
            (struct SMP_GUIDInfo *)p_attribute_data;
        u_int8_t block = (u_int8_t)(intptr_t)clbck_data.m_data2;

        m_ErrorState = m_pFabricExtendedInfo->addSMPGUIDInfo(p_port, *p_guid_info, block);
        if (m_ErrorState) {
            this->SetLastError("Failed to add SMPGuidInfo for port=%s, err=%s",
                    p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
        }
    }

    IBDIAG_RETURN_VOID
}

void IBDiagSMPPkeyTableGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SMPPkeyTableGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SMPPkeyTableGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    IBPort *p_port = (IBPort *)clbck_data.m_data1;

    if (rec_status) {
        if (p_port->p_node->appData1.val)
            IBDIAG_RETURN_VOID;

        FabricErrNodeNotRespond *p_curr_fabric_guid_err =
                new FabricErrNodeNotRespond(p_port->p_node, "SMPPKeyTableGetByDirect");
        if (!p_curr_fabric_guid_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else {
            m_pErrors->push_back(p_curr_fabric_guid_err);
            p_port->p_node->appData1.val = 1;
        }
    } else {
        struct SMP_PKeyTable *p_pkey_table =
            (struct SMP_PKeyTable *)p_attribute_data;
        u_int8_t block = (u_int8_t)(intptr_t)clbck_data.m_data2;

        m_ErrorState = m_pFabricExtendedInfo->addSMPPKeyTable(p_port, *p_pkey_table, block);
        if (m_ErrorState) {
            this->SetLastError("Failed to add SMPPKeyTable for port=%s, err=%s",
                    p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
        }
    }

    IBDIAG_RETURN_VOID
}

void IBDiagSMPSMInfoMadGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SMPSMInfoMadGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SMPSMInfoMadGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    IBPort *p_port = (IBPort *)clbck_data.m_data1;

    if (rec_status) {
        FabricErrPortNotRespond *p_curr_fabric_guid_err =
                new FabricErrPortNotRespond(p_port, "SMPSMInfoMadGet");
        if (!p_curr_fabric_guid_err) {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_curr_fabric_guid_err);
    } else {
        struct SMP_SMInfo *p_sm_info =
            (struct SMP_SMInfo *)p_attribute_data;

        m_ErrorState = m_pFabricExtendedInfo->addSMPSMInfoObj(p_port, *p_sm_info);
        if (m_ErrorState) {
            this->SetLastError("Failed to add SMInfoObj for port=%s, err=%s",
                    p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
        }
    }

    IBDIAG_RETURN_VOID
}

void IBDiagSMPLinearForwardingTableGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SMPLinearForwardingTableGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SMPLinearForwardingTableGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    IBNode *p_node = (IBNode *)clbck_data.m_data1;
    u_int16_t block = (u_int16_t)(intptr_t)clbck_data.m_data2;

    if (rec_status) {
        if (p_node->appData1.val)
            IBDIAG_RETURN_VOID;

        p_node->appData1.val = 1;
        char buff[512];
        sprintf(buff, "SMPLinearForwardingTable (block=%u)", block);
        FabricErrNodeNotRespond *p_curr_fabric_node_err =
                new FabricErrNodeNotRespond(p_node, buff);
        if (!p_curr_fabric_node_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_curr_fabric_node_err);
    } else {
        //build DB
        struct SMP_LinearForwardingTable *p_linear_forwarding_table =
            (struct SMP_LinearForwardingTable *)p_attribute_data;

        for (int curr_entry_in_block = 0; curr_entry_in_block < IBIS_IB_MAD_SMP_LFT_NUM_BLOCKS; ++curr_entry_in_block) {
            lid_t lid = (lid_t)(block * IBIS_IB_MAD_SMP_LFT_NUM_BLOCKS + curr_entry_in_block);
            p_node->setLFTPortForLid(lid, p_linear_forwarding_table->Port[curr_entry_in_block]);
        }
    }

    IBDIAG_RETURN_VOID
}

void IBDiagSMPMulticastForwardingTableGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SMPMulticastForwardingTableGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SMPMulticastForwardingTableGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    IBNode *p_node = (IBNode *)clbck_data.m_data1;
    u_int16_t block = (u_int16_t)(intptr_t)clbck_data.m_data2;
    u_int8_t port_group = (u_int8_t)(intptr_t)clbck_data.m_data3;

    if (rec_status) {
        if (p_node->appData1.val)
            IBDIAG_RETURN_VOID;

        p_node->appData1.val = 1;
        char buff[512];
        sprintf(buff, "SMPMulticastForwardingTable (block=%u, group=%u)", block, port_group);
        FabricErrNodeNotRespond *p_curr_fabric_node_err =
                new FabricErrNodeNotRespond(p_node, buff);
        if (!p_curr_fabric_node_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_curr_fabric_node_err);
    } else {
        //build DB
        struct SMP_MulticastForwardingTable *p_multicast_forwarding_table =
            (struct SMP_MulticastForwardingTable *)p_attribute_data;

        for (int curr_entry_in_block = 0; curr_entry_in_block < IBIS_IB_MAD_SMP_MFT_NUM_BLOCKS;
                ++curr_entry_in_block) {
            lid_t lid = (lid_t)(0xc000 + block * IBIS_IB_MAD_SMP_MFT_NUM_BLOCKS + curr_entry_in_block);
            if (p_multicast_forwarding_table->PortMask[curr_entry_in_block])
                p_node->setMFTPortForMLid(lid,
                        p_multicast_forwarding_table->PortMask[curr_entry_in_block],
                        port_group);
        }
    }

    IBDIAG_RETURN_VOID
}

void IBDiagSMPSLToVLMappingTableGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SMPSLToVLMappingTableGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SMPSLToVLMappingTableGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag || !m_p_sout)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    IBNode *p_node = (IBNode *)clbck_data.m_data1;
    u_int8_t in_port = (u_int8_t)(intptr_t)clbck_data.m_data2;
    u_int8_t out_port = (u_int8_t)(intptr_t)clbck_data.m_data3;

    if (rec_status) {

        FabricErrNodeNotRespond *p_curr_fabric_node_err =
                new FabricErrNodeNotRespond(p_node, "SMPSLToVLMappingTableGetByDirect");
        if (!p_curr_fabric_node_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_curr_fabric_node_err);
    } else {
        char sLine[1024];

        struct SMP_SLToVLMappingTable *p_slvl_mapping =
            (struct SMP_SLToVLMappingTable *)p_attribute_data;

        sprintf(sLine, U64H_FMT " %u %u"
                " 0x%x%x 0x%x%x 0x%x%x 0x%x%x"
                " 0x%x%x 0x%x%x 0x%x%x 0x%x%x\n",
                p_node->guid_get(),
                in_port,
                out_port,
                p_slvl_mapping->SL0ToVL, p_slvl_mapping->SL1ToVL,
                p_slvl_mapping->SL2ToVL, p_slvl_mapping->SL3ToVL,
                p_slvl_mapping->SL4ToVL, p_slvl_mapping->SL5ToVL,
                p_slvl_mapping->SL6ToVL, p_slvl_mapping->SL7ToVL,
                p_slvl_mapping->SL8ToVL, p_slvl_mapping->SL9ToVL,
                p_slvl_mapping->SL10ToVL, p_slvl_mapping->SL11ToVL,
                p_slvl_mapping->SL12ToVL, p_slvl_mapping->SL13ToVL,
                p_slvl_mapping->SL14ToVL, p_slvl_mapping->SL15ToVL);

        (*m_p_sout) << sLine;
    }

    IBDIAG_RETURN_VOID
}

void IBDiagSMPVSGeneralInfoFwInfoGetClbck(const clbck_data_t &clbck_data,
                                          int rec_status,
                                          void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SMPVSGeneralInfoFwInfoGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SMPVSGeneralInfoFwInfoGetClbck(const clbck_data_t &clbck_data,
                                                 int rec_status,
                                                 void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag || !m_p_capability_module)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    IBNode *p_node = (IBNode *)clbck_data.m_data1;

    if (rec_status == IBIS_MAD_STATUS_UNSUP_METHOD_ATTR) {         //not supported for this device
        FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
            new FabricErrNodeNotSupportCap(p_node,
                                           "The firmware of this device "
                                           "does not support GeneralInfoSMP MAD (Fw)");
        if (!p_curr_fabric_node_err) {
            SetLastError("Failed to allocate FabricErrNodeNotSupportCap");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_curr_fabric_node_err);
    } else if (rec_status) {

        FabricErrNodeNotRespond *p_curr_fabric_node_err =
                new FabricErrNodeNotRespond(p_node, "SMPVSGeneralInfoFwInfoGet");
        if (!p_curr_fabric_node_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_curr_fabric_node_err);
    } else {
        struct FWInfo_Block_Element *p_fw_info =
            (struct FWInfo_Block_Element *)p_attribute_data;
        fw_version_obj_t smp_fw;
        smp_fw.major = p_fw_info->Extended_Major;
        smp_fw.minor = p_fw_info->Extended_Minor;
        smp_fw.sub_minor = p_fw_info->Extended_SubMinor;
        m_p_capability_module->AddSMPFw(p_node->guid_get(), smp_fw);
        capability_mask_t mask;
        u_int8_t prefix_len = 0;
        u_int64_t matched_guid = 0;
        query_or_mask_t qmask;
        uint64_t guid = p_node->guid_get();

        //if this node has not mask yet, check if we can get it now
        if (!m_p_capability_module->IsSMPMaskKnown(guid)) {
            //check if there is configuration per GUID that says query
            //this case has precedence over fw configurations
            bool query_per_matched_guid = false;
            qmask.to_query = false;

            if (m_p_capability_module->IsLongestSMPPrefixMatch(
                p_node->guid_get(), prefix_len, matched_guid, qmask) &&
                qmask.to_query == true)
                query_per_matched_guid = true;

            //no match by guid with 'query'
            if (!query_per_matched_guid) {
                //check if this node has configured mask based on fw version
                //(<ven id, dev id> -> <fw, mask>)
                mask.clear();

                //check if it matched by fw configured device then
                if (m_p_capability_module->GetSMPFwConfiguredMask(
                    p_node->vendId, p_node->devId, smp_fw, mask) ==
                    IBDIAG_SUCCESS_CODE)
                    //update the cache
                    if (m_p_capability_module->AddSMPCapabilityMask(
                        p_node->guid_get(), mask) != IBDIAG_SUCCESS_CODE) {
                        FabricErrSmpGmpCapMaskExist *p_smp_mask_exist_err =
                            new FabricErrSmpGmpCapMaskExist(p_node, true, mask);
                        if (!p_smp_mask_exist_err) {
                            SetLastError(
                            "Failed to allocate FabricErrSmpGmpCapMaskExist");
                            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
                        } else
                            m_pErrors->push_back(p_smp_mask_exist_err);
                    }
            }
        }

        //try to update gmp mask for this guid
        //update the guid 2 gmp mask cache only if gmp mask is unknown
        if (!m_p_capability_module->IsGMPMaskKnown(guid)) {
            bool gmp_query_per_matched_guid = false;
            qmask.to_query = false;
            if (m_p_capability_module->IsLongestGMPPrefixMatch(guid,
                                                               prefix_len,
                                                               matched_guid,
                                                               qmask) &&
                qmask.to_query == true)
                gmp_query_per_matched_guid = true;

            mask.clear();
            //if no 'guid to query' configuration exists,
            //check for fw configured mask
            if (!gmp_query_per_matched_guid &&
                m_p_capability_module->GetGMPFwConfiguredMask(p_node->vendId,
                                                              p_node->devId,
                                                              smp_fw,
                                                              mask) ==
                IBDIAG_SUCCESS_CODE)
                //update the guid 2 mask cache
                if (m_p_capability_module->AddGMPCapabilityMask(guid, mask) !=
                    IBDIAG_SUCCESS_CODE) {
                    FabricErrSmpGmpCapMaskExist *p_gmp_mask_exist_err =
                            new FabricErrSmpGmpCapMaskExist(p_node, false, mask);
                    if (!p_gmp_mask_exist_err) {
                        SetLastError(
                            "Failed to allocate FabricErrSmpGmpCapMaskExist");
                        m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
                    } else
                        m_pErrors->push_back(p_gmp_mask_exist_err);
                }
        }

        //check if it's the same fw reported in smp and gmp
        fw_version_obj_t gmp_fw;
        if (m_p_capability_module->GetGMPFw(guid, gmp_fw) == IBDIAG_SUCCESS_CODE &&
            smp_fw != gmp_fw) {
            FabricErrSmpGmpFwMismatch *p_curr_fabric_node_err =
                new FabricErrSmpGmpFwMismatch(p_node, smp_fw, gmp_fw);
            if (!p_curr_fabric_node_err) {
                SetLastError("Failed to allocate FabricErrNodeNotRespond");
                m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
            } else
                m_pErrors->push_back(p_curr_fabric_node_err);
        }
    }
    IBDIAG_RETURN_VOID
}

void IBDiagSMPVSGeneralInfoCapabilityMaskGetClbck(const clbck_data_t &clbck_data,
                                                  int rec_status,
                                                  void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SMPVSGeneralInfoCapabilityMaskGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SMPVSGeneralInfoCapabilityMaskGetClbck(const clbck_data_t &clbck_data,
                                                    int rec_status,
                                                    void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag || !m_p_capability_module)
        return;
    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    IBNode *p_node = (IBNode *)clbck_data.m_data1;

    if (rec_status == IBIS_MAD_STATUS_UNSUP_METHOD_ATTR) {         //not supported for this device
        FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
            new FabricErrNodeNotSupportCap(p_node,
                                           "The firmware of this device "
                                           "does not support GeneralInfoSMP MAD (Capability)");
        if (!p_curr_fabric_node_err) {
            SetLastError("Failed to allocate FabricErrNodeNotSupportCap");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_curr_fabric_node_err);
    } else if (rec_status) {

        FabricErrNodeNotRespond *p_curr_fabric_node_err =
                new FabricErrNodeNotRespond(p_node, "SMPVSGeneralInfoCapabilityMaskGet");
        if (!p_curr_fabric_node_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_curr_fabric_node_err);
    } else {
        struct GeneralInfoCapabilityMask *p_general_info =
            (struct GeneralInfoCapabilityMask *)p_attribute_data;
        capability_mask_t mask;
        mask = *p_general_info;
        m_ErrorState = m_p_capability_module->AddSMPCapabilityMask(p_node->guid_get(), mask);
        if (m_ErrorState)
            SetLastError("Failed to add SMP Capability Mask for node=%s",
                         p_node->getName().c_str());
    }
    IBDIAG_RETURN_VOID
}

void IBDiagGSIPerSLVLGetClbck(const clbck_data_t &clbck_data,
                              int rec_status,
                              void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        GSIPerSLVLGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::GSIPerSLVLGetClbck(const clbck_data_t &clbck_data,
                                     int rec_status,
                                     void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag || !m_pFabricExtendedInfo ||
        !m_p_capability_module)
        return;
    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    class CountersPerSLVL * p_cntrs_per_slvl =
        (class CountersPerSLVL *)clbck_data.m_data2;

    if (rec_status == IBIS_MAD_STATUS_UNSUP_METHOD_ATTR) {         //not supported for this device
        FabricErrPortNotSupportCap *p_curr_fabric_port_err =
            new FabricErrPortNotSupportCap(p_port,
                                           "This port does not support PM " +
                                           p_cntrs_per_slvl->GetCntrHeader() +
                                           " MAD although capability bit is on");
        if (!p_curr_fabric_port_err) {
            SetLastError("Failed to allocate FabricErrNodeNotSupportCap");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else {
            p_curr_fabric_port_err->SetLevel(EN_FABRIC_ERR_WARNING);
            m_pErrors->push_back(p_curr_fabric_port_err);
        }
    } else if (rec_status) {
        FabricErrPortNotRespond *p_fabric_pm_err =
                new FabricErrPortNotRespond(p_port, p_cntrs_per_slvl->GetCntrHeader());
        if (!p_fabric_pm_err) {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_fabric_pm_err);
    } else {
        bool is_reset_cntr = (bool)clbck_data.m_data3;
        if (is_reset_cntr)
            IBDIAG_RETURN_VOID;
        struct PM_PortRcvXmitCntrsSlVl *p_per_slvl_ctrs =
            (PM_PortRcvXmitCntrsSlVl *)p_attribute_data;

        pair_ibport_slvl_cntr_data_t sl_vl_data(p_port, *p_per_slvl_ctrs);

        p_cntrs_per_slvl->m_set_port_data_update.insert(sl_vl_data);

        if (m_ErrorState)
            this->SetLastError("Failed to store pm per slvl counter"
                               " for port %s, err=%s",
                    p_port->getName().c_str(),
                    m_pFabricExtendedInfo->GetLastError());
    }
    IBDIAG_RETURN_VOID
}

void IBDiagSMPVSExtendedPortInfoGetClbck(const clbck_data_t &clbck_data,
                                         int rec_status,
                                         void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SMPVSExtendedPortInfoGetClbck(clbck_data, rec_status, p_attribute_data);
}

#define EXT_FEC_MODE_CONTROL_AND_REPORT_SUPPORT(capability)\
    ((capability) & 0x10)

void IBDiagClbck::SMPVSExtendedPortInfoGetClbck(const clbck_data_t &clbck_data,
                                                int rec_status,
                                                void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag || !m_pFabricExtendedInfo ||
        !m_p_capability_module)
        return;
    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    IBPort *p_port = (IBPort *)clbck_data.m_data1;

    if (rec_status == IBIS_MAD_STATUS_UNSUP_METHOD_ATTR) {         //not supported for this device
        FabricErrPortNotSupportCap *p_curr_fabric_port_err =
            new FabricErrPortNotSupportCap(p_port,
                                           "The firmware of this device "
                                           "does not support ExtendedPortInfoSMP MAD");
        if (!p_curr_fabric_port_err) {
            SetLastError("Failed to allocate FabricErrNodeNotSupportCap");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_curr_fabric_port_err);
    } else if (rec_status) {

        FabricErrPortNotRespond *p_fabric_pm_err =
                new FabricErrPortNotRespond(p_port, "SMPVSExtendedPortInfoGet");
        if (!p_fabric_pm_err) {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_fabric_pm_err);
    } else {
        SMP_MlnxExtPortInfo *p_mlnx_ext_port_info =
            (SMP_MlnxExtPortInfo *)p_attribute_data;

        /*
         * Check the MAD result - maybe we need to change speed to MLNX speed
         */
        IBLinkSpeed speed = (p_mlnx_ext_port_info->LinkSpeedActive ? \
                             mlnxspeed2speed(p_mlnx_ext_port_info->LinkSpeedActive) : \
                             p_port->get_common_speed());
        p_port->set_internal_speed(speed);

        u_int8_t active_cell_size = m_pIBDiag->GetLLRActiveCellSize();
        //assume the LLR active cell size, that was given by the user
        //for FDR/FDR10 and higher speeds
        if (speed >= IB_LINK_SPEED_14 && active_cell_size)
            p_mlnx_ext_port_info->RetransMode = active_cell_size;

        //check if can determine the FEC Mode
        if (EXT_FEC_MODE_CONTROL_AND_REPORT_SUPPORT(
                p_mlnx_ext_port_info->CapabilityMask))
            p_port->set_fec_mode((IBFECMode)p_mlnx_ext_port_info->FECModeActive);

        if (p_mlnx_ext_port_info->IsSpecialPort)
            p_port->setSpecialPortType(
                       (IBSpecialPortType)p_mlnx_ext_port_info->SpecialPortType);

        /*
         * Save smp mlnx ext port info attribute in the fabric
         */
        m_ErrorState = m_pFabricExtendedInfo->addSMPMlnxExtPortInfo(
                                                p_port,
                                                *p_mlnx_ext_port_info);
        if (m_ErrorState)
            this->SetLastError("Failed to store vs extended port info"
                               " for port %s, err=%s",
                    p_port->getName().c_str(),
                    m_pFabricExtendedInfo->GetLastError());
    }
    IBDIAG_RETURN_VOID
}

void IBDiagSMPExtendedSwitchInfoGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SMPExtendedSwitchInfoGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SMPExtendedSwitchInfoGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    IBNode *p_node = (IBNode *)clbck_data.m_data1;

    if (rec_status) {
        char buff[512];
        sprintf(buff, "SMPExtendedSwitchInfoGet");
        FabricErrNodeNotRespond *p_curr_fabric_node_err =
                new FabricErrNodeNotRespond(p_node, buff);
        if (!p_curr_fabric_node_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_curr_fabric_node_err);
    } else {
        //build DB
        struct ib_extended_switch_info *p_ext_sw_info =
            (struct ib_extended_switch_info *)p_attribute_data;

        if (p_ext_sw_info->sl2vl_act)
            p_node->setSL2VLAct(p_ext_sw_info->sl2vl_act);
    }

    IBDIAG_RETURN_VOID
}

void IBDiagSMPPLFTInfoGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SMPPLFTInfoGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SMPPLFTInfoGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    IBNode *p_node = (IBNode *)clbck_data.m_data1;

    if (rec_status) {
        char buff[512];
        sprintf(buff, "SMPPLFTInfoGet");
        FabricErrNodeNotRespond *p_curr_fabric_node_err =
                new FabricErrNodeNotRespond(p_node, buff);
        if (!p_curr_fabric_node_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_curr_fabric_node_err);
    } else {
        //build DB
        ib_private_lft_info * p_plft_info =
            (struct ib_private_lft_info *)p_attribute_data;

        if (p_plft_info->Active_Mode != 0)
            p_node->setPLFTEnabled();
    }

    IBDIAG_RETURN_VOID
}

void IBDiagSMPPortSLToPrivateLFTMapGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SMPPortSLToPrivateLFTMapGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SMPPortSLToPrivateLFTMapGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    IBNode *p_node = (IBNode *)clbck_data.m_data1;
    u_int16_t block = (u_int16_t)(intptr_t)clbck_data.m_data2;

    if (rec_status) {
        char buff[512];
        sprintf(buff, "SMPPortSLToPrivateLFTMapGet");
        FabricErrNodeNotRespond *p_curr_fabric_node_err =
                new FabricErrNodeNotRespond(p_node, buff);
        if (!p_curr_fabric_node_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_curr_fabric_node_err);
    } else {
        //build DB
        ib_port_sl_to_private_lft_map * p_plft_map =
            (struct ib_port_sl_to_private_lft_map *)p_attribute_data;

        ib_port_sl_to_plft * pPortSLToPLFT = NULL;

        for (int curr_entry_in_block = 0;
              curr_entry_in_block < IBIS_IB_MAD_PLFT_MAP_PORT_BLOCK_SIZE;
              ++curr_entry_in_block) {
            phys_port_t port =
                (phys_port_t)(block * IBIS_IB_MAD_PLFT_MAP_PORT_BLOCK_SIZE +
                                      curr_entry_in_block);

            if (port > p_node->numPorts)
                break;

            pPortSLToPLFT =
                &p_plft_map->PortSLToPLFT[curr_entry_in_block];

            p_node->setPLFTMapping(port, 0, pPortSLToPLFT->PLFTToPortSL0);
            p_node->setPLFTMapping(port, 1, pPortSLToPLFT->PLFTToPortSL1);
            p_node->setPLFTMapping(port, 2, pPortSLToPLFT->PLFTToPortSL2);
            p_node->setPLFTMapping(port, 3, pPortSLToPLFT->PLFTToPortSL3);
            p_node->setPLFTMapping(port, 4, pPortSLToPLFT->PLFTToPortSL4);
            p_node->setPLFTMapping(port, 5, pPortSLToPLFT->PLFTToPortSL5);
            p_node->setPLFTMapping(port, 6, pPortSLToPLFT->PLFTToPortSL6);
            p_node->setPLFTMapping(port, 7, pPortSLToPLFT->PLFTToPortSL7);
            p_node->setPLFTMapping(port, 8, pPortSLToPLFT->PLFTToPortSL8);
            p_node->setPLFTMapping(port, 9, pPortSLToPLFT->PLFTToPortSL9);
            p_node->setPLFTMapping(port, 10, pPortSLToPLFT->PLFTToPortSL10);
            p_node->setPLFTMapping(port, 11, pPortSLToPLFT->PLFTToPortSL11);
            p_node->setPLFTMapping(port, 12, pPortSLToPLFT->PLFTToPortSL12);
            p_node->setPLFTMapping(port, 13, pPortSLToPLFT->PLFTToPortSL13);
            p_node->setPLFTMapping(port, 14, pPortSLToPLFT->PLFTToPortSL14);
            p_node->setPLFTMapping(port, 15, pPortSLToPLFT->PLFTToPortSL15);
        }

        if (p_node->getMaxPLFT() >= MAX_PLFT_NUM) {
            char buff[512];
            sprintf(buff, "SMPARInfoGet unsupported PLFT number=%u",
                    p_node->getMaxPLFT() );
            FabricErrNodeWrongConfig *p_curr_fabric_node_err =
                    new FabricErrNodeWrongConfig(p_node, buff);
            if (!p_curr_fabric_node_err) {
                SetLastError("Failed to allocate FabricErrNodeNotRespond");
                m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
            } else
                m_pErrors->push_back(p_curr_fabric_node_err);

            p_node->setMaxPLFT((u_int8_t)(MAX_PLFT_NUM - 1));
        }

        //printf("guid:"U64H_FMT" MaxPLFT:%d\n",p_node->guid_get(), (int)p_node->getMaxPLFT());

    }

    IBDIAG_RETURN_VOID
}

void IBDiagSMPPrivateLFTTopGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SMPPrivateLFTTopGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SMPPrivateLFTTopGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    IBNode *p_node = (IBNode *)clbck_data.m_data1;
    u_int8_t pLFT = (u_int8_t)(intptr_t)clbck_data.m_data2;

    if (rec_status) {
        char buff[512];
        sprintf(buff, "SMPPrivateLFTTopGet");
        FabricErrNodeNotRespond *p_curr_fabric_node_err =
                new FabricErrNodeNotRespond(p_node, buff);
        if (!p_curr_fabric_node_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_curr_fabric_node_err);
    } else {
        //build DB
        struct ib_private_lft_map * p_plft_map =
            (struct ib_private_lft_map *)p_attribute_data;

        p_node->setLFDBTop(pLFT, p_plft_map->LFT_Top);
    }

    IBDIAG_RETURN_VOID
}

///
// AR
///

void IBDiagSMPARInfoGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SMPARInfoGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SMPARInfoGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    IBNode *p_node = (IBNode *)clbck_data.m_data1;

    if (rec_status) {
        char buff[512];
        sprintf(buff, "SMPARInfoGet");
        FabricErrNodeNotRespond *p_curr_fabric_node_err =
                new FabricErrNodeNotRespond(p_node, buff);
        if (!p_curr_fabric_node_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_curr_fabric_node_err);
    } else {
        //build DB
        adaptive_routing_info * p_ar_info =
            (struct adaptive_routing_info *)p_attribute_data;

        if (p_ar_info->e) {
            if (p_ar_info->glb_groups != 1) {
                char buff[512];
                sprintf(buff, "SMPARInfoGet unsupported non global groups");
                FabricErrNodeWrongConfig *p_curr_fabric_node_err =
                        new FabricErrNodeWrongConfig(p_node, buff);
                if (!p_curr_fabric_node_err) {
                    SetLastError("Failed to allocate FabricErrNodeNotRespond");
                    m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
                } else
                    m_pErrors->push_back(p_curr_fabric_node_err);
            } else if (p_ar_info->is4_mode) {
                    char buff[512];
                    sprintf(buff, "SMPARInfoGet unsupported IS4Mode");
                    FabricErrNodeWrongConfig *p_curr_fabric_node_err =
                            new FabricErrNodeWrongConfig(p_node, buff);
                    if (!p_curr_fabric_node_err) {
                        SetLastError("Failed to allocate FabricErrNodeNotRespond");
                        m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
                    } else
                        m_pErrors->push_back(p_curr_fabric_node_err);
            } else {
                u_int16_t arEnableBySLMask = \
                    ((p_ar_info->by_sl_cap &&  p_ar_info->by_sl_en) ? \
                      p_ar_info->enable_by_sl_mask : (u_int16_t)0xFFFF);

                p_node->setAREnabled(arEnableBySLMask,
                                     p_ar_info->sub_grps_active,
                                     p_ar_info->fr_enabled,
                                     p_ar_info->by_transp_cap,
                                     p_ar_info->by_transport_disable);

                p_node->setARGroupTop(p_ar_info->group_top);

                AdditionalRoutingDataMap *p_routing_data_map =
                    (AdditionalRoutingDataMap *)clbck_data.m_data2;

                if (p_routing_data_map)
                    m_pIBDiag->AddRNDataMapEntry(p_routing_data_map,
                                                 p_node,
                                                 (direct_route_t *)clbck_data.m_data3,
                                                 p_ar_info);
                else
                    p_node->appData3.ptr = NULL;
            }
        }
    }

    IBDIAG_RETURN_VOID
}

void IBDiagSMPARGroupTableGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SMPARGroupTableGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SMPARGroupTableGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    IBNode *p_node = (IBNode *)clbck_data.m_data1;

    if (rec_status) {
        char buff[512];
        sprintf(buff, "SMPARGroupTableGet");
        FabricErrNodeNotRespond *p_curr_fabric_node_err =
                new FabricErrNodeNotRespond(p_node, buff);
        if (!p_curr_fabric_node_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_curr_fabric_node_err);

        p_node->appData2.val = RETRIEVE_STAGE_REC_DONE;

    } else {
        //build DB
        ib_ar_group_table *p_ar_group_table =
            (struct ib_ar_group_table *)p_attribute_data;

        u_int16_t groupBlockNum = (u_int16_t)p_node->appData1.val;

        list_phys_ports portsList;
        bool set = false;
        for (u_int8_t blockEntry = 0;
              blockEntry < IBIS_IB_MAD_SMP_AR_GROUP_TABLE_BLOCK_SIZE ;
              blockEntry++) {
            portsList.clear();
            getPortsList(p_ar_group_table->Group[blockEntry], portsList);
            u_int16_t groupTop = p_node->getARGroupTop();

            //Old ar plugin or old fw do not support group top
            //(top equals 0) for this case the end condition is
            //empty portsList.
            if (!portsList.empty() || groupTop > 0) {
                u_int16_t groupNumber =
                    (u_int16_t)((blockEntry + groupBlockNum *
                     IBIS_IB_MAD_SMP_AR_GROUP_TABLE_BLOCK_SIZE) /
                     (p_node->getARSubGrpsActive()+1));

                if (groupTop != 0) {
                    if (groupTop < groupNumber) {
                        break;
                    } else if (groupTop == groupNumber) {
                        p_node->appData2.val = RETRIEVE_STAGE_REC_DONE;
                    }
                }

                p_node->setARPortGroup(groupNumber, portsList);
                set = true;
            }

        }
        if (set) {
            p_node->appData1.val++;
            if (p_node->appData2.val != RETRIEVE_STAGE_REC_DONE) {
                p_node->appData2.val = RETRIEVE_STAGE_SEND;
            }

            if (AdditionalRoutingData::dump_full_ar) {

                AdditionalRoutingData *p_routing_data =
                    (AdditionalRoutingData *)p_node->appData3.ptr;

                if(p_routing_data)
                    p_routing_data->AddGroupTable(
                        groupBlockNum, p_ar_group_table);
            }

        } else
            p_node->appData2.val = RETRIEVE_STAGE_REC_DONE;
    }

    IBDIAG_RETURN_VOID
}

void IBDiagClbck::getPortsList(u_int64_t pgSubBlockElement,
                               phys_port_t portOffset,
                               list_phys_ports &portsList)
{
    for (phys_port_t port = 0 ; port < 64 ; ++port) {
        if ( pgSubBlockElement & ((u_int64_t)1 << port) )
            portsList.push_back((phys_port_t)(port+portOffset));
    }
}

void IBDiagClbck::getPortsList(ib_portgroup_block_element &pgBlockElement,
                               list_phys_ports &portsList)
{
    if (pgBlockElement.SubGroup_0)
        getPortsList(pgBlockElement.SubGroup_0,
                      (phys_port_t)0, portsList);
    if (pgBlockElement.SubGroup_1)
        getPortsList(pgBlockElement.SubGroup_1,
                     (phys_port_t)(1*64), portsList);
    if (pgBlockElement.SubGroup_2)
        getPortsList(pgBlockElement.SubGroup_2,
                     (phys_port_t)(2*64), portsList);
    if (pgBlockElement.SubGroup_3)
        getPortsList(pgBlockElement.SubGroup_3,
                     (phys_port_t)(3*64), portsList);
}

void IBDiagSMPARLinearForwardingTableGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SMPARLinearForwardingTableGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SMPARLinearForwardingTableGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    IBNode *p_node = (IBNode *)clbck_data.m_data1;
    u_int16_t block = (u_int16_t)(intptr_t)clbck_data.m_data2;
    u_int8_t pLFT = (u_int8_t)(intptr_t)clbck_data.m_data3;

    if (rec_status) {
        if (p_node->appData1.val)
            IBDIAG_RETURN_VOID;

        p_node->appData1.val = 1;
        char buff[512];
        sprintf(buff, "SMPARLinearForwardingTableGet (block=%u pLFT=%u)",
                block, (int)pLFT);
        FabricErrNodeNotRespond *p_curr_fabric_node_err =
                new FabricErrNodeNotRespond(p_node, buff);
        if (!p_curr_fabric_node_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_curr_fabric_node_err);
    } else {
        //build DB
        struct ib_ar_linear_forwarding_table_sx *p_ar_linear_forwarding_table =
            (struct ib_ar_linear_forwarding_table_sx *)p_attribute_data;

        ib_ar_lft_block_element_sx *pLidEntry = NULL;

        for (int curr_entry_in_block = 0;
              curr_entry_in_block < IBIS_IB_MAD_SMP_AR_LFT_NUM_BLOCKS_SX;
              ++curr_entry_in_block) {
            lid_t lid = (lid_t)(block * IBIS_IB_MAD_SMP_AR_LFT_NUM_BLOCKS_SX +
                                      curr_entry_in_block);

            // check that this lid is exist.
            IBPort *p_dstPort = p_node->p_fabric->getPortByLid(lid);
            IBVPort *p_dstVPort = p_node->p_fabric->getVPortByLid(lid);
            if (!p_dstPort && !p_dstVPort)
                continue;

            pLidEntry =
                &p_ar_linear_forwarding_table->LidEntry[curr_entry_in_block];

            p_node->setLFTPortForLid(lid,
                                     pLidEntry->DefaultPort,
                                     pLFT);

            if (pLidEntry->LidState == AR_IB_LID_STATE_BOUNDED || //AR
                pLidEntry->LidState == AR_IB_LID_STATE_FREE ||           //AR
                p_node->isFREnabled()) {// Fault Routing (FLFR)

                if (pLidEntry->GroupNumber > p_node->getMaxARGroupNumber()) {
                    char buff[512];
                    sprintf(buff,
                            "SMPARLFTGet undefined group number:%u for lid:%u pLFT:%u",
                            pLidEntry->GroupNumber , lid, pLFT);
                    FabricErrNodeWrongConfig *p_curr_fabric_node_err =
                            new FabricErrNodeWrongConfig(p_node, buff);
                    if (!p_curr_fabric_node_err) {
                        SetLastError("Failed to allocate FabricErrNodeNotRespond");
                        m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
                    } else
                        m_pErrors->push_back(p_curr_fabric_node_err);
                } else {
                    p_node->setARLFTPortGroupForLid(lid,
                                             pLidEntry->GroupNumber,
                                             pLFT);
                }
            }
        }

        if (AdditionalRoutingData::dump_full_ar) {

            AdditionalRoutingData *p_routing_data =
                (AdditionalRoutingData *)p_node->appData3.ptr;

            if(p_routing_data)
                p_routing_data->AddARLFT(pLFT,
                    block, p_ar_linear_forwarding_table);
        }


    }

    IBDIAG_RETURN_VOID
}

void IBDiagSMPRNSubGroupDirectionTableGetClbck(const clbck_data_t &clbck_data,
                        int rec_status,
                        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SMPRNSubGroupDirectionTableGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SMPRNSubGroupDirectionTableGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    AdditionalRoutingData *p_routing_data =
        (AdditionalRoutingData *)clbck_data.m_data1;

    if (rec_status) {
        char buff[512];
        sprintf(buff, "SMPRNSubGroupDirectionTableGet");
        FabricErrNodeNotRespond *p_curr_fabric_node_err =
                new FabricErrNodeNotRespond(
                    p_routing_data->p_node, buff);
        if (!p_curr_fabric_node_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_curr_fabric_node_err);
    } else {
        struct rn_sub_group_direction_tbl *p_rn_sub_group_direction_table =
            (struct rn_sub_group_direction_tbl *)p_attribute_data;

        u_int16_t block = (u_int16_t)(intptr_t)clbck_data.m_data2;

        memcpy(&p_routing_data->sub_group_direction_table_vec[block],
               p_rn_sub_group_direction_table,
               sizeof(rn_sub_group_direction_tbl));
    }

    IBDIAG_RETURN_VOID
}

void IBDiagSMPRNGenStringTableGetClbck(const clbck_data_t &clbck_data,
                        int rec_status,
                        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SMPRNGenStringTableGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SMPRNGenStringTableGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    AdditionalRoutingData *p_routing_data =
        (AdditionalRoutingData *)clbck_data.m_data1;

    if (rec_status) {
        char buff[512];
        sprintf(buff, "SMPRNGenStringTableGet");
        FabricErrNodeNotRespond *p_curr_fabric_node_err =
                new FabricErrNodeNotRespond(
                    p_routing_data->p_node, buff);
        if (!p_curr_fabric_node_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_curr_fabric_node_err);
    } else {
        struct rn_gen_string_tbl *p_rn_gen_string_table =
            (struct rn_gen_string_tbl *)p_attribute_data;

        u_int16_t block = (u_int16_t)(intptr_t)clbck_data.m_data2;
        u_int8_t pLFT = (u_int8_t)(intptr_t)clbck_data.m_data3;

        memcpy(&p_routing_data->gen_string_table_vec[pLFT][block],
               p_rn_gen_string_table,
               sizeof(rn_gen_string_tbl));
    }

    IBDIAG_RETURN_VOID
}

void IBDiagSMPRNGenBySubGroupPriorityGetClbck(const clbck_data_t &clbck_data,
                        int rec_status,
                        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SMPRNGenBySubGroupPriorityGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SMPRNGenBySubGroupPriorityGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    AdditionalRoutingData *p_routing_data =
        (AdditionalRoutingData *)clbck_data.m_data1;

    if (rec_status) {
        char buff[512];
        sprintf(buff, "SMPRNGenBySubGroupPriorityGet");
        FabricErrNodeNotRespond *p_curr_fabric_node_err =
                new FabricErrNodeNotRespond(
                    p_routing_data->p_node, buff);
        if (!p_curr_fabric_node_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_curr_fabric_node_err);
    } else {
        struct rn_gen_by_sub_group_prio *p_rn_gen_by_sg_priority =
            (struct rn_gen_by_sub_group_prio *)p_attribute_data;

        memcpy(&p_routing_data->gen_by_sub_group_priority,
               p_rn_gen_by_sg_priority,
               sizeof(rn_gen_by_sub_group_prio));
    }

    IBDIAG_RETURN_VOID
}

void IBDiagSMPRNRcvStringGetClbck(const clbck_data_t &clbck_data,
                        int rec_status,
                        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SMPRNRcvStringGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SMPRNRcvStringGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    AdditionalRoutingData *p_routing_data =
        (AdditionalRoutingData *)clbck_data.m_data1;

    if (rec_status) {
        char buff[512];
        sprintf(buff, "SMPRNRcvStringGet");
        FabricErrNodeNotRespond *p_curr_fabric_node_err =
                new FabricErrNodeNotRespond(
                    p_routing_data->p_node, buff);
        if (!p_curr_fabric_node_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_curr_fabric_node_err);
    } else {
        struct rn_rcv_string *p_rn_rcv_string =
            (struct rn_rcv_string *)p_attribute_data;

        u_int16_t block = (u_int16_t)(intptr_t)clbck_data.m_data2;

        memcpy(&p_routing_data->rcv_string_vec[block],
               p_rn_rcv_string,
               sizeof(rn_rcv_string));
    }

    IBDIAG_RETURN_VOID
}

void IBDiagSMPRNXmitPortMaskGetClbck(const clbck_data_t &clbck_data,
                        int rec_status,
                        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SMPRNXmitPortMaskGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SMPRNXmitPortMaskGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    AdditionalRoutingData *p_routing_data =
        (AdditionalRoutingData *)clbck_data.m_data1;

    if (rec_status) {
        char buff[512];
        sprintf(buff, "SMPRNXmitPortMaskGet");
        FabricErrNodeNotRespond *p_curr_fabric_node_err =
                new FabricErrNodeNotRespond(
                    p_routing_data->p_node, buff);
        if (!p_curr_fabric_node_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_curr_fabric_node_err);
    } else {
        struct rn_xmit_port_mask *p_rn_xmit_port_mask =
            (struct rn_xmit_port_mask *)p_attribute_data;

        u_int16_t block = (u_int16_t)(intptr_t)clbck_data.m_data2;

        memcpy(&p_routing_data->xmit_port_mask_vec[block],
               p_rn_xmit_port_mask,
               sizeof(rn_xmit_port_mask));
    }

    IBDIAG_RETURN_VOID
}

void IBDiagVSPortRNCountersClearClbck(const clbck_data_t &clbck_data,
                        int rec_status,
                        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        VSPortRNCountersClearClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::VSPortRNCountersClearClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    AdditionalRoutingData *p_routing_data =
        (AdditionalRoutingData *)clbck_data.m_data1;

    if (rec_status) {
        char buff[512];
        sprintf(buff, "VSPortRNCountersClear");
        FabricErrNodeNotRespond *p_curr_fabric_node_err =
                new FabricErrNodeNotRespond(
                    p_routing_data->p_node, buff);
        if (!p_curr_fabric_node_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_curr_fabric_node_err);
    }

    IBDIAG_RETURN_VOID
}

void IBDiagVSPortRNCountersGetClbck(const clbck_data_t &clbck_data,
                        int rec_status,
                        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        VSPortRNCountersGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::VSPortRNCountersGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    AdditionalRoutingData *p_routing_data =
        (AdditionalRoutingData *)clbck_data.m_data1;

    if (rec_status) {
        char buff[512];
        sprintf(buff, "VSPortRNCountersGet");
        FabricErrNodeNotRespond *p_curr_fabric_node_err =
                new FabricErrNodeNotRespond(
                    p_routing_data->p_node, buff);
        if (!p_curr_fabric_node_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_curr_fabric_node_err);
    } else {
        struct port_rn_counters *p_rn_counters =
            (struct port_rn_counters *)p_attribute_data;

        u_int8_t port = (u_int8_t)(intptr_t)clbck_data.m_data2;

        memcpy(&p_routing_data->port_rn_counters_vec[port],
               p_rn_counters,
               sizeof(port_rn_counters));
    }

    IBDIAG_RETURN_VOID
}

void IBDiagVSDiagnosticCountersClearClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        VSDiagnosticCountersClearClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::VSDiagnosticCountersClearClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    if (rec_status) {
        FabricErrPortNotRespond *p_fabric_pm_err =
                new FabricErrPortNotRespond(p_port, "VSDiagnosticCountersClear");
        if (!p_fabric_pm_err) {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_fabric_pm_err);
    }
}

void IBDiagVSDiagnosticCountersPage0GetClbck(
                          const clbck_data_t &clbck_data,
                          int rec_status,
                          void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        VSDiagnosticCountersPage0GetClbck(clbck_data, rec_status,
                                          p_attribute_data);
}

void IBDiagClbck::VSDiagnosticCountersPage0GetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;
    unsigned int latest_version = 0;
    rec_status &= 0x00ff;//bits [8-15] are class specific - don't need them here
    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    int rc = 0;
    if (rec_status) {
        FabricErrPortNotRespond *p_fabric_mlnx_cntrs_err =
            new FabricErrPortNotRespond(p_port, "VSDiagnosticDataGet");
        if (!p_fabric_mlnx_cntrs_err) {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_fabric_mlnx_cntrs_err);
    } else if (m_pIBDiag->getLatestSupportedVersion(VS_MLNX_CNTRS_PAGE0,
                                                    latest_version)) {
        this->SetLastError("Failed to get latest version for "  \
                           "TransportErrorsAndFlows counters");
        m_ErrorState = IBDIAG_ERR_CODE_DB_ERR;
    } else {
        struct VS_DiagnosticData *p_mlnx_cntrs =
            (struct VS_DiagnosticData *)p_attribute_data;
         // Between Backward and Current revisions only new fields can be added and
         // no old fields can be changed (by name,offset,print location and etc.),
         // so the provided version is compatible with ibdiagnet latest
         // supported version.
        if (p_mlnx_cntrs->BackwardRevision <= latest_version &&
            latest_version <= p_mlnx_cntrs->CurrentRevision) {
            struct VS_DC_TransportErrorsAndFlowsV2 p_page0_rev1;
            VS_DC_TransportErrorsAndFlowsV2_unpack(
                    &p_page0_rev1,
                    (uint8_t *)&(p_mlnx_cntrs->data_set));
            p_mlnx_cntrs->data_set.TransportErrorsAndFlowsV2  =
                p_page0_rev1;
            rc = m_pFabricExtendedInfo->
                addVSDiagnosticCountersPage0(p_port, *p_mlnx_cntrs);
            if (rc) {
                SetLastError("Failed to add DiagnosticCounters"  \
                             " TransportErrorsAndFlowsV2 for port=%s, err=%s",
                             p_port->getName().c_str(),
                             m_pFabricExtendedInfo->GetLastError());
                m_ErrorState = rc;
            }
        } else {
            FabricErrNodeMlnxCountersPageVer *p_curr_fabric_node_err =
                new FabricErrNodeMlnxCountersPageVer(
                        p_port->p_node,
                        VS_MLNX_CNTRS_PAGE0,
                        p_mlnx_cntrs->CurrentRevision,
                        latest_version);
            if (!p_curr_fabric_node_err) {
                this->SetLastError("Failed to allocate"  \
                                   " FabricErrNodeMlnxCountersPageVer");
                m_ErrorState =IBDIAG_ERR_CODE_NO_MEM;
            }
            p_curr_fabric_node_err->SetLevel(EN_FABRIC_ERR_WARNING);
            m_pErrors->push_back(p_curr_fabric_node_err);
        }
    }
}

void IBDiagVSDiagnosticCountersPage1GetClbck(
                          const clbck_data_t &clbck_data,
                          int rec_status,
                          void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        VSDiagnosticCountersPage1GetClbck(clbck_data, rec_status,
                                          p_attribute_data);
}

void IBDiagClbck::VSDiagnosticCountersPage1GetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;
    unsigned int latest_version;
    rec_status &= 0x00ff;//bits [8-15] are class specific - don't need them here
    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    int rc = 0;
    if (rec_status) {
        FabricErrPortNotRespond *p_fabric_mlnx_cntrs_err =
            new FabricErrPortNotRespond(p_port, "VSDiagnosticDataGet");
        if (!p_fabric_mlnx_cntrs_err) {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_fabric_mlnx_cntrs_err);

    } else if (m_pIBDiag->getLatestSupportedVersion(VS_MLNX_CNTRS_PAGE1, latest_version)) {

        this->SetLastError("Failed to get latest version for HCAExtendedFlows counters");
        m_ErrorState = IBDIAG_ERR_CODE_DB_ERR;
    } else {
        struct VS_DiagnosticData *p_mlnx_cntrs =
            (struct VS_DiagnosticData *)p_attribute_data;

         if (p_mlnx_cntrs->CurrentRevision == 0) {
             // Page not supported on device
             FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                 new FabricErrNodeNotSupportCap(p_port->p_node,
                     "This device does not support Diagnostic Counters Page 1");
             if (!p_curr_fabric_node_err) {
                 SetLastError("Failed to allocate FabricErrNodeNotSupportCap");
                 m_ErrorState =IBDIAG_ERR_CODE_NO_MEM;
             }
             p_curr_fabric_node_err->SetLevel(EN_FABRIC_ERR_WARNING);
             m_pErrors->push_back(p_curr_fabric_node_err);

         } else {

            struct VS_DC_Page1LatestVersion p_page1_last_ver;
            VS_DC_Page1LatestVersion_unpack(
                    &p_page1_last_ver, (uint8_t *)&(p_mlnx_cntrs->data_set));
            p_mlnx_cntrs->data_set.Page1LatestVersion  = p_page1_last_ver;
            rc = m_pFabricExtendedInfo->addVSDiagnosticCountersPage1(
                    p_port, *p_mlnx_cntrs);
            if (rc) {
                SetLastError("Failed to add DiagnosticCounters Page1" \
                             " for port=%s, err=%s",
                             p_port->getName().c_str(),
                             m_pFabricExtendedInfo->GetLastError());
                rc = m_ErrorState;
            }

            if (p_mlnx_cntrs->BackwardRevision > latest_version ||
                    latest_version > p_mlnx_cntrs->CurrentRevision) {
                // Different versions of the page between node FW and ibdiagnet
                FabricErrNodeMlnxCountersPageVer *p_curr_fabric_node_err =
                    new FabricErrNodeMlnxCountersPageVer(
                            p_port->p_node,
                            VS_MLNX_CNTRS_PAGE1,
                            p_mlnx_cntrs->CurrentRevision,
                            latest_version);
                if (!p_curr_fabric_node_err) {
                    this->SetLastError("Failed to allocate FabricErrNodeMlnxCountersPageVer");
                    m_ErrorState =IBDIAG_ERR_CODE_NO_MEM;
                }
                p_curr_fabric_node_err->SetLevel(EN_FABRIC_ERR_WARNING);
                m_pErrors->push_back(p_curr_fabric_node_err);
            }
        }
    }
}

void IBDiagVSDiagnosticCountersPage255GetClbck(
                          const clbck_data_t &clbck_data,
                          int rec_status,
                          void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        VSDiagnosticCountersPage255GetClbck(clbck_data, rec_status,
                                            p_attribute_data);
}

void IBDiagClbck::VSDiagnosticCountersPage255GetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;
    unsigned int latest_version;
    rec_status &= 0x00ff;//bits [8-15] are class specific - don't need them here
    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    int rc = 0;
    if (rec_status) {
        FabricErrPortNotRespond *p_fabric_mlnx_cntrs_err =
            new FabricErrPortNotRespond(p_port, "VSDiagnosticDataGet");
        if (!p_fabric_mlnx_cntrs_err) {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_fabric_mlnx_cntrs_err);

    } else if (m_pIBDiag->getLatestSupportedVersion(VS_MLNX_CNTRS_PAGE255, latest_version)) {

        this->SetLastError("Failed to get latest version for DiagnosticCounters Page255");
        m_ErrorState = IBDIAG_ERR_CODE_DB_ERR;
    } else {
        struct VS_DiagnosticData *p_mlnx_cntrs = (struct VS_DiagnosticData *)p_attribute_data;

        if (p_mlnx_cntrs->CurrentRevision == 0) {
            // Page not supported on device
            FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                new FabricErrNodeNotSupportCap(p_port->p_node,
                    "This device does not support Diagnostic Counters Page 255");
            if (!p_curr_fabric_node_err) {
                SetLastError("Failed to allocate FabricErrNodeNotSupportCap");
                m_ErrorState =IBDIAG_ERR_CODE_NO_MEM;
            }
            p_curr_fabric_node_err->SetLevel(EN_FABRIC_ERR_WARNING);
            m_pErrors->push_back(p_curr_fabric_node_err);

        } else {

            struct VS_DC_Page255LatestVersion p_page255;
            VS_DC_Page255LatestVersion_unpack(&p_page255, (uint8_t *)&(p_mlnx_cntrs->data_set));
            p_mlnx_cntrs->data_set.Page255LatestVersion = p_page255;

            rc = m_pFabricExtendedInfo->addVSDiagnosticCountersPage255(p_port, *p_mlnx_cntrs);
            if (rc) {
                SetLastError("Failed to add DiagnosticCounters Page255 for port=%s, err=%s",
                        p_port->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
                rc = m_ErrorState;
            }

            if (p_mlnx_cntrs->BackwardRevision > latest_version ||
                    latest_version > p_mlnx_cntrs->CurrentRevision) {
                // Different versions of the page between node FW and ibdiagnet
                FabricErrNodeMlnxCountersPageVer *p_curr_fabric_node_err =
                    new FabricErrNodeMlnxCountersPageVer(p_port->p_node,
                                                         VS_MLNX_CNTRS_PAGE255,
                                                         p_mlnx_cntrs->CurrentRevision,
                                                         latest_version);
                if (!p_curr_fabric_node_err) {
                    this->SetLastError("Failed to allocate FabricErrNodeMlnxCountersPageVer");
                    m_ErrorState =IBDIAG_ERR_CODE_NO_MEM;
                }
                p_curr_fabric_node_err->SetLevel(EN_FABRIC_ERR_WARNING);
                m_pErrors->push_back(p_curr_fabric_node_err);
            }
        }
    }
}

#define PORT_INFO_EXT_FEC_MODE_SUPPORT(capability) \
    ((capability) & 0x1)

void IBDiagSMPPortInfoExtendedGetClbck(const clbck_data_t &clbck_data,
                                       int rec_status,
                                       void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SMPPortInfoExtendedGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SMPPortInfoExtendedGetClbck(const clbck_data_t &clbck_data,
                                              int rec_status,
                                              void *p_attribute_data)
{
    IBDIAG_ENTER;

    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific-don't need them here
    IBPort *p_port = (IBPort *)clbck_data.m_data1;

    if (rec_status) {
        if (IS_SUPPORT_PORT_INFO_EXTENDED_FAILED(p_port->p_node->appData1.val))
            //already reported error
            IBDIAG_RETURN_VOID;
        //mark we already checked for support
        p_port->p_node->appData1.val |= NOT_SUPPORT_PORT_INFO_EXTENDED;
        FabricErrPortNotRespond *p_curr_fabric_node_err =
                new FabricErrPortNotRespond(p_port, "SMPPortInfoExtendedGet");
        if (!p_curr_fabric_node_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_curr_fabric_node_err);
    } else {
        SMP_PortInfoExtended *p_port_info_ext =
            (SMP_PortInfoExtended *)p_attribute_data;

        //update the fec mode of the port
        if (PORT_INFO_EXT_FEC_MODE_SUPPORT(p_port_info_ext->CapMsk)) {
            // convert from bit mask to real value
            IBFECMode fec_value = fec_mask2value(p_port_info_ext->FECModeActive);

            if (fec_value == IB_FEC_NA) {
                FabricErrPortInvalidValue *p_curr_fabric_port_err =
                        new FabricErrPortInvalidValue(p_port, "Got wrong fec_mode_act from FW");

                if (!p_curr_fabric_port_err) {
                    SetLastError("Failed to allocate FabricErrPortInvalidValue");
                    m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
                    IBDIAG_RETURN_VOID
                } else
                    m_pErrors->push_back(p_curr_fabric_port_err);
            }
            p_port->set_fec_mode(fec_value);
        }

        //Save smp port info extended attribute in the fabric
        m_ErrorState = m_pFabricExtendedInfo->addSMPPortInfoExtended(
                                                p_port, *p_port_info_ext);
        if (m_ErrorState)
            this->SetLastError("Failed to store port info extended"
                               " for port %s, err=%s",
                    p_port->getName().c_str(),
                    m_pFabricExtendedInfo->GetLastError());
    }
    IBDIAG_RETURN_VOID;
}

void IBDiagSMPVirtualizationInfoGetClbck(
                          const clbck_data_t &clbck_data,
                          int rec_status,
                          void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
    IBDiagSMPVirtualizationInfoGetClbck(clbck_data, rec_status,
                                        p_attribute_data);
}

void IBDiagClbck::IBDiagSMPVirtualizationInfoGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff;//bits [8-15] are class specific - don't need them here
    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    int rc = 0;
    if (rec_status) {
        FabricErrPortNotRespond *p_fabric_virtual_info_err =
            new FabricErrPortNotRespond(p_port, "SMPVirtualizationInfoGet");
        if (!p_fabric_virtual_info_err) {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_fabric_virtual_info_err);
    } else {
        struct SMP_VirtualizationInfo *p_virtual_info =
            (struct SMP_VirtualizationInfo *)p_attribute_data;
        if (p_virtual_info->vport_index_top > p_virtual_info->vport_cap) {
           FabricErrVPortIvalidTopIndex *p_fabric_invalid_cap_idx_err =
                 new FabricErrVPortIvalidTopIndex(p_port,
                                                  p_virtual_info->vport_cap,
                                                  p_virtual_info->vport_index_top);
            if (!p_fabric_invalid_cap_idx_err) {
                SetLastError("Failed to allocate FabricErrVPortIvalidTopIndex");
                m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
            } else
                m_pErrors->push_back(p_fabric_invalid_cap_idx_err);
        } else {
            rc = m_pFabricExtendedInfo->addSMPVirtualizationInfo(
                    p_port, *p_virtual_info);
            if (rc) {
                SetLastError("Failed to add Virtualization Info" \
                        " for port=%s, err=%s",
                        p_port->getName().c_str(),
                        m_pFabricExtendedInfo->GetLastError());
                rc = m_ErrorState;
            }
        }
    }
}

void IBDiagSMPVPortStateGetClbck(
                          const clbck_data_t &clbck_data,
                          int rec_status,
                          void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
    IBDiagSMPVPortStateGetClbck(clbck_data, rec_status,
                                        p_attribute_data);
}

void IBDiagClbck::IBDiagSMPVPortStateGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff;//bits [8-15] are class specific - don't need them here
    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    u_int8_t block = (u_int8_t)(intptr_t)clbck_data.m_data2;
    int rc = 0;
    if (rec_status) {
        FabricErrPortNotRespond *p_fabric_virtual_info_err =
            new FabricErrPortNotRespond(p_port, "SMPVPortStateGet");
        if (!p_fabric_virtual_info_err) {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_fabric_virtual_info_err);
    } else {
        struct SMP_VPortState *p_vport_state_block =
            (struct SMP_VPortState *)p_attribute_data;
        rc = m_pFabricExtendedInfo->addSMPVPortState(
                p_port, *p_vport_state_block, block);
        if (rc) {
            SetLastError("Failed to add VPort State" \
                    " for port=%s, err=%s",
                    p_port->getName().c_str(),
                    m_pFabricExtendedInfo->GetLastError());
            rc = m_ErrorState;
        }
    }
}

void IBDiagSMPVPortInfoGetClbck(
                          const clbck_data_t &clbck_data,
                          int rec_status,
                          void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
    IBDiagSMPVPortInfoGetClbck(clbck_data, rec_status,
                                        p_attribute_data);
}

void IBDiagClbck::IBDiagSMPVPortInfoGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff;//bits [8-15] are class specific - don't need them here
    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    int rc = 0;
    if (rec_status) {
        FabricErrPortNotRespond *p_fabric_virtual_info_err =
            new FabricErrPortNotRespond(p_port, "SMPVPortInfoGet");
        if (!p_fabric_virtual_info_err) {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_fabric_virtual_info_err);
    } else {
        struct SMP_VPortInfo *p_vport_info =
            (struct SMP_VPortInfo *)p_attribute_data;
        //need the vport num to insert the IBVPort into VPorts map.
        virtual_port_t vportnum = (virtual_port_t)(intptr_t)clbck_data.m_data2;

        //insert pair port-number and  vport.
        IBVPort *p_vport = m_pIBDiag->GetDiscoverFabricPtr()->makeVPort(
                p_port,
                vportnum,
                p_vport_info->vport_guid,
                (IBPortState)p_vport_info->vport_state);

        if (!p_vport) {
            SetLastError("Failed to allocate new vport for port=%s", p_port->getName().c_str());
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
            IBDIAG_RETURN_VOID;
        }
        if (p_vport_info->lid_required) {
            p_vport->set_vlid(p_vport_info->vport_lid);
            p_vport->getIBFabricPtr()->setLidVPort(p_vport_info->vport_lid, p_vport);
        }

        p_port->VPorts.insert(pair_vportnum_vport(vportnum, p_vport));
        rc = m_pFabricExtendedInfo->addSMPVPortInfo(p_vport, *p_vport_info);
        if (rc) {
            SetLastError("Failed to add VPort Info for port=%s, vport"  \
                         " number=%d, err=%s",
                         p_port->getName().c_str(),
                         vportnum,
                         m_pFabricExtendedInfo->GetLastError());
            rc = m_ErrorState;
        }
    }
}

void IBDiagSMPVNodeInfoGetClbck(
                          const clbck_data_t &clbck_data,
                          int rec_status,
                          void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
    IBDiagSMPVNodeInfoGetClbck(clbck_data, rec_status,
                                        p_attribute_data);
}

void IBDiagClbck::IBDiagSMPVNodeInfoGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff;//bits [8-15] are class specific - don't need them here
    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    int rc = 0;
    if (rec_status) {
        FabricErrPortNotRespond *p_fabric_vnode_info_err =
            new FabricErrPortNotRespond(p_port, "SMPVNodeInfoGet");
        if (!p_fabric_vnode_info_err) {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_fabric_vnode_info_err);
    } else {

        IBVPort *p_vport = (IBVPort *)clbck_data.m_data2;
        struct SMP_VNodeInfo *p_vnode_info =
            (struct SMP_VNodeInfo *)p_attribute_data;

        IBVNode *p_vnode = m_pIBDiag->GetDiscoverFabricPtr()->makeVNode(
                                                p_vnode_info->vnode_guid,
                                                p_vnode_info->vnum_ports,
                                                p_vport,
                                                p_vnode_info->vlocal_port_num);
        if (!p_vnode) {
            SetLastError("Failed to create new vnode for port=%s vport=%s",
                         p_port->getName().c_str(),
                         p_vport->getName().c_str());
            IBDIAG_RETURN_VOID;
        }
        p_vport->setVNodePtr(p_vnode);
        rc = m_pFabricExtendedInfo->addSMPVNodeInfo(p_vnode, *p_vnode_info);
        if (rc) {
            SetLastError("Failed to add VNode Info" \
                         " for port=%s, vport=%s, err=%s",
                        p_port->getName().c_str(),
                        p_vport->getName().c_str(),
                        m_pFabricExtendedInfo->GetLastError());
        }
    }
    IBDIAG_RETURN_VOID;
}

void IBDiagSMPVNodeDescriptionGetClbck(
                          const clbck_data_t &clbck_data,
                          int rec_status,
                          void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
    IBDiagSMPVNodeDescriptionGetClbck(clbck_data, rec_status,
                                        p_attribute_data);
}

void IBDiagClbck::IBDiagSMPVNodeDescriptionGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff;//bits [8-15] are class specific - don't need them here
    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    if (rec_status) {
        FabricErrPortNotRespond *p_fabric_vnode_description_err =
            new FabricErrPortNotRespond(p_port, "SMPVNodeDescriptionGet");
        if (!p_fabric_vnode_description_err) {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_fabric_vnode_description_err);
    } else {
        IBVNode *p_vnode = (IBVNode *)clbck_data.m_data2;
        struct SMP_NodeDesc *p_vnode_description =
            (struct SMP_NodeDesc *)p_attribute_data;

        p_vnode->setDescription((char *)p_vnode_description->Byte);
    }
    IBDIAG_RETURN_VOID;
}

void IBDiagSMPVPortPKeyGetClbck(
                          const clbck_data_t &clbck_data,
                          int rec_status,
                          void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
    IBDiagSMPVPortPKeyGetClbck(clbck_data, rec_status,
                                        p_attribute_data);
}

void IBDiagClbck::IBDiagSMPVPortPKeyGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff;//bits [8-15] are class specific - don't need them here
    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    int rc = 0;
    if (rec_status) {
        FabricErrPortNotRespond *p_fabric_pkey_table_err =
            new FabricErrPortNotRespond(p_port, "SMPVPortPKeyTableGet");
        if (!p_fabric_pkey_table_err) {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_fabric_pkey_table_err);
    } else {
        IBVPort *p_vport = (IBVPort *)clbck_data.m_data2;
        struct SMP_PKeyTable *p_vport_pkey_tbl =
            (struct SMP_PKeyTable *)p_attribute_data;
        u_int16_t pkey_block = (u_int16_t)(intptr_t)clbck_data.m_data3;
        rc = m_pFabricExtendedInfo->addSMPVPortPKeyTable(p_vport,
                                                         *p_vport_pkey_tbl,
                                                         pkey_block);
        if (rc) {
            SetLastError("Failed to add VPort PKey Table" \
                    " for port=%s, err=%s",
                    p_vport->getName().c_str(),
                    m_pFabricExtendedInfo->GetLastError());
            rc = m_ErrorState;
        }
    }
}

void IBDiagSMPQoSConfigSLGetClbck(const clbck_data_t &clbck_data,
                                  int rec_status,
                                  void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
    IBDiagSMPQoSConfigSLGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::IBDiagSMPQoSConfigSLGetClbck(const clbck_data_t &clbck_data,
                                               int rec_status,
                                               void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff;//bits [8-15] are class specific - don't need them here
    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    int rc = 0;

    if (rec_status) {

        FabricErrPortNotRespond *p_qos_config_sl_err =
            new FabricErrPortNotRespond(p_port, "SMPQosConfigSLGet");

        if (!p_qos_config_sl_err) {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_qos_config_sl_err);

    } else {

        struct SMP_QosConfigSL *p_qos_config_sl = (struct SMP_QosConfigSL *)p_attribute_data;
        rc = m_pFabricExtendedInfo->addSMPQosConfigSL(p_port, *p_qos_config_sl);

        if (rc) {
            SetLastError("Failed to add SMP_QosConfigSL for node=%s, port=%u, err=%s",
                          p_port->p_node->getName().c_str(),
                          (u_int8_t)p_port->num,
                          m_pFabricExtendedInfo->GetLastError());
            rc = m_ErrorState;
        }
    }
}

void IBDiagSMPVPortQoSConfigSLGetClbck(const clbck_data_t &clbck_data,
                                       int rec_status,
                                       void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
    IBDiagSMPVPortQoSConfigSLGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::IBDiagSMPVPortQoSConfigSLGetClbck(const clbck_data_t &clbck_data,
                                                    int rec_status,
                                                    void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff;//bits [8-15] are class specific - don't need them here
    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    IBVPort *p_vport = (IBVPort *)clbck_data.m_data2;
    int rc = 0;

    if (rec_status) {
        char buff[256] = {0};
        sprintf(buff, "SMPQosConfigSLGet for vport: " U32D_FMT, (u_int32_t)p_vport->getVPortNum());
        FabricErrPortNotRespond *p_qos_config_sl_err = new FabricErrPortNotRespond(p_port, buff);

        if (!p_qos_config_sl_err) {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_qos_config_sl_err);

    } else {

        struct SMP_QosConfigSL *p_qos_config_sl = (struct SMP_QosConfigSL *)p_attribute_data;
        rc = m_pFabricExtendedInfo->addSMPVPortQosConfigSL(p_vport, *p_qos_config_sl);

        if (rc) {
            SetLastError("Failed to add SMP_QosConfigSL for vport=%hu node=%s, port=%u, err=%s",
                          p_vport->getVPortNum(),
                          p_port->p_node->getName().c_str(),
                          (u_int8_t)p_port->num,
                          m_pFabricExtendedInfo->GetLastError());
            rc = m_ErrorState;
        }
    }
}

void IBDiagSMPTempSensingGetClbck(
                          const clbck_data_t &clbck_data,
                          int rec_status,
                          void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
    IBDiagSMPTempSensingGetClbck(clbck_data, rec_status,
                                        p_attribute_data);
}

void IBDiagClbck::IBDiagSMPTempSensingGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff;//bits [8-15] are class specific - don't need them here
    IBNode *p_node = (IBNode *)clbck_data.m_data1;
    int rc = 0;
    if (rec_status) {
        FabricErrNodeNotRespond *p_fabric_virtual_info_err =
            new FabricErrNodeNotRespond(p_node, "SMPTempSensingGet");
        if (!p_fabric_virtual_info_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_fabric_virtual_info_err);
    } else {
        struct SMP_TempSensing *p_temp_sense =
            (struct SMP_TempSensing *)p_attribute_data;
        rc = m_pFabricExtendedInfo->addSMPTempSensing(
                p_node, *p_temp_sense);
        if (rc) {
            SetLastError("Failed to add Virtualization Info" \
                    " for node=%s, err=%s",
                    p_node->getName().c_str(),
                    m_pFabricExtendedInfo->GetLastError());
            rc = m_ErrorState;
        }
    }
}

void IBDiagSMPRouterInfoGetClbck(const clbck_data_t &clbck_data,
                                 int rec_status,
                                 void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
    SMPRouterInfoGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SMPRouterInfoGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff;//bits [8-15] are class specific - don't need them here
    IBNode *p_node = (IBNode *)clbck_data.m_data1;
    int rc = 0;
    if (rec_status) {
        FabricErrNodeNotRespond *p_fabric_info_err =
            new FabricErrNodeNotRespond(p_node, "SMPRouterInfoGet");
        if (!p_fabric_info_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_fabric_info_err);
    } else {
        struct SMP_RouterInfo *p_router_info =
            (struct SMP_RouterInfo *)p_attribute_data;
        rc = m_pFabricExtendedInfo->addSMPRouterInfo(
                p_node, *p_router_info);
        if (rc) {
            SetLastError("Failed to add Router Info" \
                    " for node=%s, err=%s",
                    p_node->getName().c_str(),
                    m_pFabricExtendedInfo->GetLastError());
            rc = m_ErrorState;
        }
    }
}

void IBDiagSMPAdjRouterTableGetClbck(const clbck_data_t &clbck_data,
                                     int rec_status,
                                     void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SMPAdjRouterTableGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SMPAdjRouterTableGetClbck(const clbck_data_t &clbck_data,
                                            int rec_status,
                                            void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff;//bits [8-15] are class specific - don't need them here
    IBNode *p_node = (IBNode *)clbck_data.m_data1;
    int rc = 0;
    if (rec_status) {
        FabricErrNodeNotRespond *p_fabric_info_err =
            new FabricErrNodeNotRespond(p_node, "SMPAdjSiteLocalSubnTblGet");
        if (!p_fabric_info_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_fabric_info_err);
    } else {
        struct SMP_AdjSiteLocalSubnTbl *p_router_tbl =
            (struct SMP_AdjSiteLocalSubnTbl *)p_attribute_data;
        u_int8_t rtr_block = (u_int8_t)(intptr_t)clbck_data.m_data2;

        rc = m_pFabricExtendedInfo->addSMPAdjSiteLocalSubnTbl(p_node,
                                                              *p_router_tbl,
                                                              rtr_block);
        if (rc) {
            SetLastError("Failed to add SMPAdjSiteLocalSubnTbl router table"
                         " for node=%s, err=%s",
                         p_node->getName().c_str(),
                         m_pFabricExtendedInfo->GetLastError());
            rc = m_ErrorState;
        }
    }
}

void IBDiagSMPNextHopRouterTableGetClbck(const clbck_data_t &clbck_data,
                                         int rec_status,
                                         void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SMPNextHopRouterTableGetClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SMPNextHopRouterTableGetClbck(const clbck_data_t &clbck_data,
                                                int rec_status,
                                                void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff;//bits [8-15] are class specific - don't need them here
    IBNode *p_node = (IBNode *)clbck_data.m_data1;
    int rc = 0;
    if (rec_status) {
        FabricErrNodeNotRespond *p_fabric_info_err =
            new FabricErrNodeNotRespond(p_node, "SMPNextHopRouterTableGet");
        if (!p_fabric_info_err) {
            SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            m_pErrors->push_back(p_fabric_info_err);
    } else {
        struct SMP_NextHopTbl *p_router_tbl =
            (struct SMP_NextHopTbl*)p_attribute_data;
        u_int32_t rtr_block = (u_int32_t)(intptr_t)clbck_data.m_data2;

        rc = m_pFabricExtendedInfo->addSMPNextHopTbl(p_node,
                                                     *p_router_tbl,
                                                     rtr_block);
        if (rc) {
            SetLastError("Failed to add SMPNextHop router table"
                         " for node=%s, err=%s",
                         p_node->getName().c_str(),
                         m_pFabricExtendedInfo->GetLastError());
            rc = m_ErrorState;
        }
    }
}


void IBDiagAggMngrClassPortInfoClbck(const clbck_data_t &clbck_data,
                                     int rec_status,
                                     void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SharpMngrClassPortInfoClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SharpMngrClassPortInfoClbck(
                    const clbck_data_t &clbck_data,
                    int rec_status,
                    void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    if (!rec_status) {
        // Not checking the return status as it cam
        IBNode *p_node = (IBNode *)clbck_data.m_data1;
        if (!m_p_sharp_mngr) {
            SetLastError("Failed to get SharpMngr ptr");
            m_ErrorState = IBDIAG_ERR_CODE_DB_ERR;
        } else {
            m_p_sharp_mngr->AddSharpSupportedNodes(p_node);
        }
    }
}


void IBDiagSharpMngrTreeConfigClbck(const clbck_data_t &clbck_data,
                                    int rec_status,
                                    void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SharpMngrTreeConfigClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SharpMngrTreeConfigClbck(
                    const clbck_data_t &clbck_data,
                    int rec_status,
                    void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    SharpAggNode *p_sharp_agg_node =
          (SharpAggNode *)clbck_data.m_data1;

    IBPort *p_port = p_sharp_agg_node->GetIBPort();
    if (!p_port) {
            SetLastError("Failed to get IBPort for Aggregation Node");
            m_ErrorState = IBDIAG_ERR_CODE_DB_ERR;
            IBDIAG_RETURN_VOID;
    }

    if (rec_status) {
        FabricErrNodeNotRespond *p_fabric_pm_err =
                new FabricErrNodeNotRespond(p_port->p_node, "AMTreeConfigGet");
        if (!p_fabric_pm_err) {
           SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else {
            m_num_errors++;
            m_pErrors->push_back(p_fabric_pm_err);
        }
    } else {
        if (!m_p_sharp_mngr) {
            SetLastError("Failed to get SharpMngr ptr");
            m_ErrorState = IBDIAG_ERR_CODE_DB_ERR;
        }

        struct AM_TreeConfig *p_tree_config =
            (struct AM_TreeConfig *)p_attribute_data;

        u_int16_t tree_id = (u_int16_t)(uintptr_t)clbck_data.m_data2;
        //if tree_state is DISEBLED we don't need to create new TreeNode for this TreeID
        if (!p_tree_config->tree_state) {
            IBDIAG_RETURN_VOID;
        }

        if (tree_id != p_tree_config->tree_id) {
            SharpErrNodeTreeIDNotMatchGetRespondTreeID *sharp_treeid_err =
                new SharpErrNodeTreeIDNotMatchGetRespondTreeID(p_port->p_node, tree_id);
            if (!sharp_treeid_err) {
                SetLastError("Failed to allocate FabricErrNodeNotRespond");
                m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
            } else {
                m_num_errors++;
                m_pErrors->push_back(sharp_treeid_err);
            }
        }
        // The number of children that was read in the previous calls
        u_int8_t total_children_query_num = (u_int8_t)(uintptr_t)clbck_data.m_data3;

        SharpTreeNode *p_sharp_tree_node =
            new SharpTreeNode(p_sharp_agg_node, tree_id);

        int rc  = p_sharp_agg_node->AddSharpTreeNode(p_sharp_tree_node, tree_id);
        if (rc) {
            SetLastError("Failed to add SharpTreeNode for node=%s",
                    p_port->p_node->getName().c_str());
            m_ErrorState = rc;
        }

        m_p_sharp_mngr->UpdateFabricMaxTreeIdx(tree_id);

        if (p_tree_config->parent_qpn)
            p_sharp_tree_node->SetSharpParentTreeEdge(new SharpTreeEdge(p_tree_config->parent_qpn, 0));
        else {
            int rc = m_p_sharp_mngr->AddTreeRoot(tree_id, p_sharp_tree_node);
            if (rc) {
                SharpErrRootTreeNodeAlreadyExistsForTreeID *sharp_treeid_err =
                    new SharpErrRootTreeNodeAlreadyExistsForTreeID(p_port->p_node, tree_id);
                if (!sharp_treeid_err) {
                    SetLastError("Failed to allocate SharpErrRootTreeNodeAlreadyExistsForTreeID");
                    m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
                } else {
                    m_num_warnings++;
                    m_pErrors->push_back(sharp_treeid_err);
                }
            }
        }

        // total_children_query_num - The total number of children for all the call
        // for this node and tree_id.
        // Total_children_query_num value need to go to the next call of this
        // Clbck function.
        u_int8_t i = 0;
        for (i = 0;
             (i < p_tree_config->num_of_children) && (i < MAX_CHILD_IDX_IN_TREE_CONFIG_MAD);
             i++) {
            SharpTreeEdge *p_sharp_tree_eage =
                new SharpTreeEdge(p_tree_config->child_qp[i].child_qpn,
                                  p_tree_config->child_qp[i].child_index);
            rc = p_sharp_tree_node->AddSharpTreeEdge(p_sharp_tree_eage,
                                                     (u_int8_t)(i + total_children_query_num));
        }

        total_children_query_num = (u_int8_t)(total_children_query_num + i);
        // need to check against the num_of_children
        if (p_tree_config->record_locator != p_sharp_agg_node->GetMaxNumQps()) {
            struct AM_TreeConfig tree_config;
            CLEAR_STRUCT(tree_config);
            clbck_data_t clbck_data;
            CLEAR_STRUCT(clbck_data);
            clbck_data.m_p_obj = &ibDiagClbck;
            clbck_data.m_handle_data_func = IBDiagSharpMngrTreeConfigClbck;
            clbck_data.m_data1 = p_sharp_agg_node;
            clbck_data.m_data2 = (void*)(uintptr_t)tree_id;
            clbck_data.m_data3 = (void*)(uintptr_t)total_children_query_num;

            tree_config.record_locator = p_tree_config->record_locator;
            tree_config.tree_id = tree_id;
            tree_config.num_of_children = MAX_CHILD_IDX_IN_TREE_CONFIG_MAD;

            m_pIBDiag->GetIbisPtr()->AMTreeConfigGet(p_port->base_lid, DEFAULT_SL,
                    DEFAULT_AM_KEY, DEFAULT_AM_CLASS_VERSION, &tree_config, &clbck_data);
        }

        if (rc) {
          SetLastError("Failed to add PMCapMask for node=%s, err=%s",
                  p_port->p_node->getName().c_str(), m_pFabricExtendedInfo->GetLastError());
            m_ErrorState = rc;
        }
    }
}


void IBDiagSharpMngrQPCConfigClbck(const clbck_data_t &clbck_data,
                                    int rec_status,
                                    void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SharpMngrQPCConfigClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SharpMngrQPCConfigClbck(
                    const clbck_data_t &clbck_data,
                    int rec_status,
                    void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    SharpAggNode *p_sharp_agg_node =
          (SharpAggNode *)clbck_data.m_data1;

    IBPort *p_port = p_sharp_agg_node->GetIBPort();
    if (!p_port) {
            SetLastError("Failed to get IBPort for Aggregation Node");
            m_ErrorState = IBDIAG_ERR_CODE_DB_ERR;
            IBDIAG_RETURN_VOID;
    }

    if (rec_status) {
        FabricErrNodeNotRespond *p_fabric_pm_err =
                new FabricErrNodeNotRespond(p_port->p_node, "AMQPCConfigGet");
        if (!p_fabric_pm_err) {
           SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else {
            m_num_errors++;
            m_pErrors->push_back(p_fabric_pm_err);
        }
    } else {
        struct AM_QPCConfig *p_qpc_config =
            (struct AM_QPCConfig *)p_attribute_data;

         SharpTreeEdge *pSharpTreeEdge = (SharpTreeEdge *)clbck_data.m_data2;
         pSharpTreeEdge->SetQPCConfig(*p_qpc_config);
    }
}

void IBDiagSharpMngrResetPerfCountersClbck(const clbck_data_t &clbck_data,
                                           int rec_status,
                                           void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SharpMngrResetPerfCountersClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SharpMngrResetPerfCountersClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    if (rec_status) {
        SharpAggNode *p_sharp_agg_node =
              (SharpAggNode *)clbck_data.m_data1;

        IBPort *p_port = p_sharp_agg_node->GetIBPort();
        if (!p_port) {
                SetLastError("Failed to get IBPort for Aggregation Node");
                m_ErrorState = IBDIAG_ERR_CODE_DB_ERR;
                IBDIAG_RETURN_VOID;
        }

        FabricErrNodeNotRespond *p_fabric_pm_err =
                new FabricErrNodeNotRespond(p_port->p_node, "AMPerfCountersSet");
        if (!p_fabric_pm_err) {
           SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else {
            m_num_errors++;
            m_pErrors->push_back(p_fabric_pm_err);
        }
    }
}

void IBDiagSharpMngrPerfCountersClbck(const clbck_data_t &clbck_data,
                                      int rec_status,
                                      void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SharpMngrPerfCountersClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SharpMngrPerfCountersClbck(
                    const clbck_data_t &clbck_data,
                    int rec_status,
                    void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    SharpAggNode *p_sharp_agg_node =
          (SharpAggNode *)clbck_data.m_data1;

    IBPort *p_port = p_sharp_agg_node->GetIBPort();
    if (!p_port) {
            SetLastError("Failed to get IBPort for Aggregation Node");
            m_ErrorState = IBDIAG_ERR_CODE_DB_ERR;
            IBDIAG_RETURN_VOID;
    }

    if (rec_status) {
        FabricErrNodeNotRespond *p_fabric_pm_err =
                new FabricErrNodeNotRespond(p_port->p_node, "AMPerfCountersGet");
        if (!p_fabric_pm_err) {
           SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else {
            m_num_errors++;
            m_pErrors->push_back(p_fabric_pm_err);
        }
    } else {
        struct AM_PerformanceCounters *p_perf_counters =
            (struct AM_PerformanceCounters *)p_attribute_data;

        p_sharp_agg_node->SetPerfCounters(p_perf_counters);
    }
}

void IBDiagSharpMngrANInfoClbck(const clbck_data_t &clbck_data,
                                int rec_status,
                                void *p_attribute_data)
{
    reinterpret_cast<IBDiagClbck*>(clbck_data.m_p_obj)->
        SharpMngrANInfoClbck(clbck_data, rec_status, p_attribute_data);
}

void IBDiagClbck::SharpMngrANInfoClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    if (m_ErrorState || !m_pErrors || !m_pIBDiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    SharpAggNode *p_sharp_agg_node =
          (SharpAggNode *)clbck_data.m_data1;

    IBPort *p_port = p_sharp_agg_node->GetIBPort();
    if (!p_port) {
            SetLastError("Failed to get IBPort for Aggregation Node");
            m_ErrorState = IBDIAG_ERR_CODE_DB_ERR;
            IBDIAG_RETURN_VOID;
    }

    if (rec_status) {
        FabricErrNodeNotRespond *p_fabric_pm_err =
                new FabricErrNodeNotRespond(p_port->p_node, "AMANInfoGet");
        if (!p_fabric_pm_err) {
           SetLastError("Failed to allocate FabricErrNodeNotRespond");
            m_ErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else {
            m_num_errors++;
            m_pErrors->push_back(p_fabric_pm_err);
        }
    } else {
        struct AM_ANInfo *p_an_info =
            (struct AM_ANInfo *)p_attribute_data;

        p_sharp_agg_node->SetANInfo(p_an_info);
    }
}
