/*
 *  Copyright (c) 2004-2019 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <infiniband/ibdiag/ibdiag.h>


/****************************************************/
/******************private methods*******************/
/****************************************************/
void IBDiag::DumpCSVNodesTable(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    csv_out.DumpStart(SECTION_NODES);

    stringstream sstream;
    sstream << "NodeDesc," << "NumPorts," << "NodeType," << "ClassVersion," << "BaseVersion," << "SystemImageGUID,"
            << "NodeGUID," << "PortGUID," << "DeviceID," << "PartitionCap," << "revision,"
            << "VendorID," << "LocalPortNum"
            << endl;
    csv_out.WriteLine(sstream.str());

    // go over all nodes:
    struct SMP_NodeInfo * p_curr_node_info = NULL;
    char curr_node_line[1024];
    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node)               //node not exist
            continue;
        p_curr_node_info = this->fabric_extended_info.getSMPNodeInfo(i);
        if (!p_curr_node_info)          //node info not exist
            continue;
        memset(curr_node_line, 0, sizeof(curr_node_line));
        sstream.str("");
        sprintf(curr_node_line, "\"" STR_FMT "\"" ","   \
                U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT ","  \
                U64H_FMT "," U64H_FMT "," U64H_FMT ","  \
                U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT,
                //p_curr_node->getName().c_str(),
                p_curr_node->description.c_str(),
                p_curr_node_info->NumPorts,
                p_curr_node_info->NodeType,
                p_curr_node_info->ClassVersion,
                p_curr_node_info->BaseVersion,
                p_curr_node_info->SystemImageGUID,
                p_curr_node_info->NodeGUID,
                p_curr_node_info->PortGUID,
                p_curr_node_info->DeviceID,
                p_curr_node_info->PartitionCap,
                p_curr_node_info->revision,
                p_curr_node_info->VendorID,
                p_curr_node_info->LocalPortNum);
        sstream << curr_node_line << endl;
        csv_out.WriteLine(sstream.str());
    }
    csv_out.DumpEnd(SECTION_NODES);

    IBDIAG_RETURN_VOID;
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::DumpCSVPortsTable(CSVOut &csv_out, bool show_ports_data_extra)
{
    IBDIAG_ENTER;

    stringstream sstream;
    sstream << "# In the next section the values of the fields LinkSpeedEn/LinkSpeedActv/LinkSpeedSup "
            << "are not real raw data from the wire but as follow: " << endl
            << "# They are 32 bit fields." << endl
            << "# Byte1 is regular speeds," << endl
            << "# Byte2 is extended speeds," << endl
            << "# Byte3 is mlnx ext speeds," << endl
            << "# Byte4 is reserved."
            << endl << endl;
    csv_out.WriteBuf(sstream.str());

    csv_out.DumpStart(SECTION_PORTS);

    sstream.str("");
    sstream << "NodeGuid," << "PortGuid," << "PortNum,"   //these first values are used to match port to node
            << "MKey," << "GIDPrfx," << "MSMLID," << "LID," << "CapMsk,"
            << "M_KeyLeasePeriod," << "DiagCode," << "LinkWidthActv," << "LinkWidthSup," << "LinkWidthEn,"
            << "LocalPortNum," << "LinkSpeedEn," << "LinkSpeedActv," << "LMC,"
            << "MKeyProtBits," << "LinkDownDefState," << "PortPhyState," << "PortState," << "LinkSpeedSup,"
            << "VLArbHighCap," << "VLHighLimit," << "InitType," << "VLCap," << "MSMSL,"
            << "NMTU," << "FilterRawOutb," << "FilterRawInb," << "PartEnfOutb," << "PartEnfInb,"
            << "OpVLs," << "HoQLife," << "VLStallCnt," << "MTUCap," << "InitTypeReply,"
            << "VLArbLowCap," << "PKeyViolations," << "MKeyViolations," << "SubnTmo,"
            << "ClientReregister," << "GUIDCap," << "QKeyViolations," << "MaxCreditHint," << "OverrunErrs,"
            << "LocalPhyError," << "RespTimeValue," << "LinkRoundTripLatency," << "OOOSLMask";
    if (show_ports_data_extra)
        sstream << ",CapMsk2,FECActv,RetransActv";
    sstream << endl;
    csv_out.WriteLine(sstream.str());

    struct SMP_PortInfo *p_curr_port_info = NULL;
    struct SMP_MlnxExtPortInfo *p_curr_mlnx_ext_port_info = NULL;
    char curr_port_line[1024];
    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port)           //port not exist
            continue;
        p_curr_port_info = this->fabric_extended_info.getSMPPortInfo(i);
        if (!p_curr_port_info)      //port info not exist
            continue;

        u_int32_t cap_mask                  = p_curr_port_info->CapMsk;
        u_int32_t correct_link_speed_en;
        u_int32_t correct_link_speed_actv;
        u_int32_t correct_link_speed_sup;

        //if it switch than take cap mask from port 0
        if ((p_curr_port->p_node->type == IB_SW_NODE) &&
                p_curr_port->num != 0) {
            IBPort *p_zero_port = p_curr_port->p_node->getPort(0);
            if (!p_zero_port) {
                this->SetLastError("DB error - can not found manage port for switch=%s\n",
                        p_curr_port->p_node->getName().c_str());
                IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            }
            struct SMP_PortInfo *p_zero_port_info =
                    this->fabric_extended_info.getSMPPortInfo(p_zero_port->createIndex);
            if (!p_zero_port_info) {
                this->SetLastError("DB error - can not found port info for port=%s\n",
                        p_zero_port->getName().c_str());
                IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            }
            cap_mask = p_zero_port_info->CapMsk;
        }

        //if the port supports mlnx extended speeds
        p_curr_mlnx_ext_port_info = this->fabric_extended_info.getSMPMlnxExtPortInfo(i);

        //regular speeds + extended + mellanox
        GetTotalSpeeds(p_curr_port_info, p_curr_mlnx_ext_port_info, cap_mask,
                       &correct_link_speed_sup, &correct_link_speed_en,
                       &correct_link_speed_actv);

        stringstream OOOSLMask_Output("N/A");
        if (p_curr_mlnx_ext_port_info) {
            if (this->capability_module.IsSupportedSMPCapability(
                                           p_curr_port->p_node,
                                           EnSMPCapIsGlobalOOOSupported)) {
                OOOSLMask_Output.str("");
                OOOSLMask_Output << "0x" << std::hex << p_curr_mlnx_ext_port_info->OOOSLMask;
            }
        }

        memset(curr_port_line, 0, sizeof(curr_port_line));
        sstream.str("");
        sprintf(curr_port_line, U64H_FMT "," U64H_FMT "," U32D_FMT ","    \
                U64H_FMT "," U64H_FMT ","    \
                U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT ","    \
                U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT ","    \
                U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT ","    \
                U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT ","    \
                U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT ","    \
                U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT ","    \
                U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT ","    \
                U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT ","    \
                U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT ",%s" ,
                p_curr_port->p_node->guid_get(),
                p_curr_port->guid_get(),
                p_curr_port->num,
                p_curr_port_info->MKey,
                p_curr_port_info->GIDPrfx,
                p_curr_port_info->MSMLID,
                p_curr_port_info->LID,
                p_curr_port_info->CapMsk,
                p_curr_port_info->M_KeyLeasePeriod,
                p_curr_port_info->DiagCode,
                p_curr_port_info->LinkWidthActv,
                p_curr_port_info->LinkWidthSup,
                p_curr_port_info->LinkWidthEn,
                p_curr_port_info->LocalPortNum,
                correct_link_speed_en,
                correct_link_speed_actv,
                p_curr_port_info->LMC,
                p_curr_port_info->MKeyProtBits,
                p_curr_port_info->LinkDownDefState,
                p_curr_port_info->PortPhyState,
                p_curr_port_info->PortState,
                correct_link_speed_sup,
                p_curr_port_info->VLArbHighCap,
                p_curr_port_info->VLHighLimit,
                p_curr_port_info->InitType,
                p_curr_port_info->VLCap,
                p_curr_port_info->MSMSL,
                p_curr_port_info->NMTU,
                p_curr_port_info->FilterRawOutb,
                p_curr_port_info->FilterRawInb,
                p_curr_port_info->PartEnfOutb,
                p_curr_port_info->PartEnfInb,
                p_curr_port_info->OpVLs,
                p_curr_port_info->HoQLife,
                p_curr_port_info->VLStallCnt,
                p_curr_port_info->MTUCap,
                p_curr_port_info->InitTypeReply,
                p_curr_port_info->VLArbLowCap,
                p_curr_port_info->PKeyViolations,
                p_curr_port_info->MKeyViolations,
                p_curr_port_info->SubnTmo,
                p_curr_port_info->ClientReregister,
                p_curr_port_info->GUIDCap,
                p_curr_port_info->QKeyViolations,
                p_curr_port_info->MaxCreditHint,
                p_curr_port_info->OverrunErrs,
                p_curr_port_info->LocalPhyError,
                p_curr_port_info->RespTimeValue,
                p_curr_port_info->LinkRoundTripLatency,
                OOOSLMask_Output.str().c_str());
        sstream << curr_port_line;
        if (show_ports_data_extra) {
            //PortInfoCapMsk2, FEC, Retransmission
            if (IS_SUPPORT_PORT_INFO_CAP_MASK2(p_curr_port_info->CapMsk))
                sstream << "," << p_curr_port_info->CapMsk2;
            else
                sstream << ",N/A";
            IBFECMode fec = p_curr_port->get_fec_mode();
            if (fec == IB_FEC_NA)
                sstream << "," <<  IB_FEC_NA_STR << ",";
            else
                sstream << "," << fec << ",";

            if (!p_curr_mlnx_ext_port_info)
                sstream << "N/A";
            else
                sstream << (unsigned int)p_curr_mlnx_ext_port_info->RetransMode;
        }
        sstream << endl;
        csv_out.WriteLine(sstream.str());
    }
    csv_out.DumpEnd(SECTION_PORTS);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void IBDiag::DumpCSV_FECModeTable(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    csv_out.DumpStart(SECTION_FEC_MODE);

    stringstream sstream;
    sstream << "NodeGuid,PortGuid,PortNum,FECActv,FDRFECSup,FDRFECEn,EDRFECSup,EDRFECEn,"
            "HDRFECSup,HDRFECEn,FDR10FECSup,FDR10FECEn,EDR20FECSup,EDR20FECEn" << endl;
    csv_out.WriteLine(sstream.str());

    char curr_fec_line[1024];
    IBPort *p_curr_port = NULL;
    struct SMP_PortInfoExtended *p_port_info_ext = NULL;
    struct SMP_MlnxExtPortInfo *p_mepi = NULL;
    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port)    // port not exist
            continue;

        p_port_info_ext = fabric_extended_info.getSMPPortInfoExtended(i);
        p_mepi = fabric_extended_info.getSMPMlnxExtPortInfo(i);

        bool advancedFECModes = this->capability_module.IsSupportedSMPCapability(
                p_curr_port->p_node, EnSMPCapIsAdvancedRetransAndFECModesSupported);

        if ((!p_mepi || !advancedFECModes) && !p_port_info_ext)    // no FEC info
            continue;

        memset(curr_fec_line, 0, sizeof(curr_fec_line));
        sstream.str("");
        sprintf(curr_fec_line, U64H_FMT "," U64H_FMT "," U32D_FMT "," U32D_FMT ",",
                p_curr_port->p_node->guid_get(),
                p_curr_port->guid_get(),
                p_curr_port->num,
                p_curr_port->get_fec_mode());

        sstream << curr_fec_line;
        memset(curr_fec_line, 0, sizeof(curr_fec_line));

        if (p_mepi && advancedFECModes) {
            sprintf(curr_fec_line,
                    U16H_FMT "," U16H_FMT "," U16H_FMT "," U16H_FMT "," U16H_FMT ","
                    U16H_FMT "," U16H_FMT "," U16H_FMT "," U16H_FMT "," U16H_FMT,
                    p_mepi->FDRFECModeSupported,
                    p_mepi->FDRFECModeEnabled,
                    p_mepi->EDRFECModeSupported,
                    p_mepi->EDRFECModeEnabled,
                    p_mepi->HDRFECModeSupported,
                    p_mepi->HDRFECModeEnabled,
                    p_mepi->FDR10FECModeSupported,
                    p_mepi->FDR10FECModeEnabled,
                    p_mepi->EDR20FECModeSupported,
                    p_mepi->EDR20FECModeEnabled);
        } else {    // no mepi, use port_info_ext.
            sprintf(curr_fec_line,
                    U16H_FMT "," U16H_FMT "," U16H_FMT "," U16H_FMT "," U16H_FMT ","
                    U16H_FMT ",N/A,N/A,N/A,N/A",
                    p_port_info_ext->FDRFECModeSupported,
                    p_port_info_ext->FDRFECModeEnabled,
                    p_port_info_ext->EDRFECModeSupported,
                    p_port_info_ext->EDRFECModeEnabled,
                    p_port_info_ext->HDRFECModeSupported,
                    p_port_info_ext->HDRFECModeEnabled);
        }

        sstream << curr_fec_line << endl;
        csv_out.WriteLine(sstream.str());
    }

    csv_out.DumpEnd(SECTION_FEC_MODE);
    IBDIAG_RETURN_VOID
}

void IBDiag::DumpCSVSwitchesTable(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    csv_out.DumpStart(SECTION_SWITCHES);

    stringstream sstream;
    sstream << "NodeGUID," << "LinearFDBCap," << "RandomFDBCap," << "MCastFDBCap,"
            << "LinearFDBTop," << "DefPort," << "DefMCastPriPort," << "DefMCastNotPriPort,"
            << "LifeTimeValue," << "PortStateChange," << "OptimizedSLVLMapping," << "LidsPerPort,"
            << "PartEnfCap," << "InbEnfCap," << "OutbEnfCap," << "FilterRawInbCap,"
            << "FilterRawOutbCap," << "ENP0," << "MCastFDBTop"
            << endl;
    csv_out.WriteLine(sstream.str());

    char curr_switch_line[1024];
    IBNode *p_curr_node = NULL;
    SMP_SwitchInfo *p_curr_switch_info = NULL;

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node)               // node not exist
            continue;

        p_curr_switch_info = this->fabric_extended_info.getSMPSwitchInfo(i);
        if (!p_curr_switch_info)        // switch info not exist
            continue;

        memset(curr_switch_line, 0, sizeof(curr_switch_line));
        sstream.str("");

        sprintf(curr_switch_line, U64H_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," \
                U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," \
                U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," \
                U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT,
                p_curr_node->guid_get(),
                p_curr_switch_info->LinearFDBCap,
                p_curr_switch_info->RandomFDBCap,
                p_curr_switch_info->MCastFDBCap,
                p_curr_switch_info->LinearFDBTop,
                p_curr_switch_info->DefPort,
                p_curr_switch_info->DefMCastPriPort,
                p_curr_switch_info->DefMCastNotPriPort,
                p_curr_switch_info->LifeTimeValue,
                p_curr_switch_info->PortStateChange,
                p_curr_switch_info->OptimizedSLVLMapping,
                p_curr_switch_info->LidsPerPort,
                p_curr_switch_info->PartEnfCap,
                p_curr_switch_info->InbEnfCap,
                p_curr_switch_info->OutbEnfCap,
                p_curr_switch_info->FilterRawInbCap,
                p_curr_switch_info->FilterRawOutbCap,
                p_curr_switch_info->ENP0,
                p_curr_switch_info->MCastFDBTop);
        sstream << curr_switch_line << endl;
        csv_out.WriteLine(sstream.str());
    }

    csv_out.DumpEnd(SECTION_SWITCHES);
    IBDIAG_RETURN_VOID
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::DumpCSVLinksTable(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    //first mark in each port some field we will use latter
    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end();
            ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }
        // go over all ports of this node:
        for (int i = 1; i <= p_curr_node->numPorts; ++i) {    //start from port 1, no need to dump port 0
                IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
                if (!p_curr_port)       //port not exist
                    continue;
                p_curr_port->counter1 = 0;
        }
    }

    csv_out.DumpStart(SECTION_LINKS);

    stringstream sstream;
    sstream << "NodeGuid1," << "PortNum1,"
            << "NodeGuid2," << "PortNum2"
            << endl;
    csv_out.WriteLine(sstream.str());

    // go over all nodes:
    char curr_link_line[1024];
    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end();
            ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());
            csv_out.DumpEnd(SECTION_LINKS);
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        // go over all ports of this node:
        for (u_int32_t i = 1; i <= p_curr_node->numPorts; ++i) { //start from port 1, no need to dump port 0
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            if (!p_curr_port)       //port not exist
                continue;
            IBPort *p_remote_port = p_curr_port->p_remotePort;
            if (!p_remote_port)     //port not connected
                continue;
            IBNode *p_remote_node = p_remote_port->p_node;
            if (!p_remote_node) {
                this->SetLastError("DB error - found port with no node %s",
                                   p_remote_port->getName().c_str());
                csv_out.DumpEnd(SECTION_LINKS);
                IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            }

            //don't write same connection more than one time
            if ((p_curr_port->counter1 == 1) || (p_remote_port->counter1 == 1))
                continue;
            p_curr_port->counter1 = 1;
            p_remote_port->counter1 = 1;

            sstream.str("");
            sprintf(curr_link_line, U64H_FMT "," U32D_FMT ","   \
                    U64H_FMT "," U32D_FMT,
                    p_curr_node->guid_get(), p_curr_port->num,
                    p_remote_node->guid_get(), p_remote_port->num);
            sstream << curr_link_line << endl;
            csv_out.WriteLine(sstream.str());
        }
    }

    csv_out.DumpEnd(SECTION_LINKS);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::DumpNetwork(ofstream &sout)
{
    IBDIAG_ENTER;
    char port_line[900] = {'\0'};
    char port_line_header[900] = {'\0'};

    sprintf(port_line_header,
            "  %-4s : %-3s : %-4s : %-10s : MTU : %-7s : " \
            "%-7s : %-19s : %-6s : %-18s : %-4s : %-4s : %s",
            "#", "IB#", "Sta", "PhysSta", "LWA", "LSA", "FEC mode", "Retran",
            "Neighbor Guid", "N#", "NLID", "Neighbor Description");

    // go over all nodes:
    struct SMP_NodeInfo * p_curr_node_info = NULL;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
         nI != this->discovered_fabric.NodeByName.end(); ++nI) {

        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName " \
                               "map for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        p_curr_node_info =
           this->fabric_extended_info.getSMPNodeInfo(p_curr_node->createIndex);
        if (!p_curr_node_info)          //node info not exist
            continue;

        if (p_curr_node->type != IB_SW_NODE)
            continue;

        sout << '\"' << p_curr_node->description << "\", "
             << (ibis_obj.IsVenMellanox(p_curr_node_info->VendorID)? "" : "Non ") << "Mellanox"
             << ", 0x" << hex << p_curr_node->guid_get() << dec
             << ", LID " << p_curr_node->getPort((phys_port_t)1)->base_lid
             << endl;

        sout << port_line_header << endl;

        // go over all ports of this node:
        // start from port 1, no need to dump port 0
        for (int i = 1; i <= p_curr_node->numPorts; ++i) {
                IBPort *p_curr_port =
                    p_curr_node->getPort((phys_port_t)i);
                //port not exist or not in the scope
                if (!p_curr_port || !p_curr_port->getInSubFabric())
                    continue;

                if (!p_curr_port->isValid())
                    continue;

                struct SMP_PortInfo * p_curr_port_info =
                    this->fabric_extended_info.getSMPPortInfo(
                                                p_curr_port->createIndex);
                if (!p_curr_port_info)
                    continue;

                struct SMP_MlnxExtPortInfo *p_mlnx_ext_port_info =
                    this->fabric_extended_info.getSMPMlnxExtPortInfo(
                                                p_curr_port->createIndex);

                stringstream mtu_str;
                mtu_str << +p_curr_port_info->NMTU;

                stringstream remote_guid;
                stringstream remote_num;
                stringstream remote_lid;
                stringstream remote_desc;
                if (p_curr_port->p_remotePort) {
                    remote_guid << "0x" << hex << p_curr_port->p_remotePort->guid_get();
                    remote_num << p_curr_port->p_remotePort->numAsString();
                    remote_lid << p_curr_port->p_remotePort->base_lid;
                    remote_desc << '\"'
                            << p_curr_port->p_remotePort->p_node->description.c_str() << '\"';
                }
                sprintf(port_line,
                        "  %-4s : %-3u : %-4s : %-10s : %-3s : %-7s : " \
                        "%-7s : %-19s : %-6s : %-18s : %-4s : %-4s : %s\n",
                        p_curr_port->numAsString().c_str(),
                        p_curr_port->num,
                        portstate2char(p_curr_port->get_internal_state()),
                        portphysstate2char(
                            (IBPortPhysState)p_curr_port_info->PortPhyState),
                        p_curr_port->get_internal_state() == IB_PORT_STATE_DOWN ?
                                NOT_AVAILABLE : mtu_str.str().c_str(),
                        p_curr_port->get_internal_state() == IB_PORT_STATE_DOWN ?
                                NOT_AVAILABLE : width2char(p_curr_port->get_internal_width()),
                        p_curr_port->get_internal_state() == IB_PORT_STATE_DOWN ?
                                NOT_AVAILABLE : speed2char(p_curr_port->get_internal_speed()),
                        fec2char(p_curr_port->get_fec_mode()),
                        !p_mlnx_ext_port_info? RETRANS_MODE_NA : retransmission2char(
                                (EnRetransmissionMode)p_mlnx_ext_port_info->RetransMode),
                        remote_guid.str().c_str(),
                        remote_num.str().c_str(),
                        remote_lid.str().c_str(),
                        remote_desc.str().c_str()
                        );
                sout << port_line;
        }
        sout << endl;
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

/****************************************************/
/*******************public methods*******************/
/****************************************************/
//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR / ERR_CODE_DB_ERR
int IBDiag::DumpInternalDBCSVTable(CSVOut &csv_out, bool show_ports_data_extra, bool to_dump_fec_mode)
{
    IBDIAG_ENTER;
    //if (!this->IsDiscoveryDone())
    //    IBDIAG_RETURN(ERR_CODE_DISCOVERY_NOT_SUCCESS);

    this->DumpCSVNodesTable(csv_out);
    int rc = this->DumpCSVPortsTable(csv_out, show_ports_data_extra); //here we can fail
    if (rc)
        IBDIAG_RETURN(rc);
    if (to_dump_fec_mode)
        this->DumpCSV_FECModeTable(csv_out);
    this->DumpCSVSwitchesTable(csv_out);
    rc = this->DumpCSVLinksTable(csv_out); //here we can fail because we are get the links information
                                        //only in this stage and after that we dump it
    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IBDM_ERR
int IBDiag::WriteLSTFile(const char *file_path, bool write_with_lmc)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    if (this->discovered_fabric.dumpLSTFile((char *)file_path, write_with_lmc)) {
        this->SetLastError("Failed to dump lst file");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::WriteNetDumpFile(const char* file_path)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile(file_path, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);
    this->DumpNetwork(sout);
    sout.close();
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}
