/*
 *  Copyright (c) 2004-2013 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <algorithm>
#include <infiniband/ibdiag/ibdiag.h>
#include <infiniband/ibdiag/ibdiag_types.h>
#include <infiniband/ibdiag/ibdiag_clbck.h>

extern IBDiagClbck ibDiagClbck;

/****************************************************/
/******************private methods*******************/
/****************************************************/
inline bool IBDiag::IsBFSKnownPortGuid(IN u_int64_t guid)
{
    map_guid_list_p_direct_route::iterator nI = this->bfs_known_port_guids.find(guid);
    if (nI == this->bfs_known_port_guids.end())
            return false;
    return ((*nI).second.size() != 0);
}

inline bool IBDiag::IsBFSKnownNodeGuid(IN u_int64_t guid)
{
    map_guid_list_p_direct_route::iterator nI = this->bfs_known_node_guids.find(guid);
    if (nI == this->bfs_known_node_guids.end())
            return false;
    return ((*nI).second.size() != 0);
}

inline void IBDiag::MarkPortGuidAsBFSKnown(IN u_int64_t guid, IN direct_route_t *p_direct_route)
{
    this->bfs_known_port_guids[guid].push_back(p_direct_route);
}

inline void IBDiag::MarkNodeGuidAsBFSKnown(IN u_int64_t guid, IN direct_route_t *p_direct_route)
{
    this->bfs_known_node_guids[guid].push_back(p_direct_route);
}


inline void IBDiag::BFSPushPath(IN direct_route_t *p_direct_route)
{
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "BFSPushPath %s\n",
               ibis_obj.ConvertDirPathToStr(p_direct_route).c_str());
    this->bfs_list.push_back(p_direct_route);   //push to the end of the list
}

inline direct_route_t * IBDiag::BFSPopPath()
{
    direct_route_t *current = this->bfs_list.front();
    this->bfs_list.pop_front();                 //pop from the begin of the list
    return current;
}


inline void IBDiag::AddGoodPath(IN direct_route_t *p_direct_route)
{
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "AddGoodPath %s\n",
              ibis_obj.ConvertDirPathToStr(p_direct_route).c_str());

    this->good_direct_routes.push_back(p_direct_route);
}

inline int IBDiag::AddBadPath(IN IbdiagBadDirectRoute_t *bad_direct_route,
                               IN direct_route_t *direct_route)
{
    IbdiagBadDirectRoute_t *p_bad_direct_route = NULL;

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "AddBadPath %s, fail reason \"%s\", port num: %d\n",
              ibis_obj.ConvertDirPathToStr(direct_route).c_str(),
               bad_direct_route_reasons[bad_direct_route->reason],
               bad_direct_route->port_num);

    p_bad_direct_route = new IbdiagBadDirectRoute_t;
    if (!p_bad_direct_route) {
        this->SetLastError("Failed to allocate bad direct");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
    }
    //copy the info
    *p_bad_direct_route = *bad_direct_route;
    p_bad_direct_route->direct_route = direct_route;
    this->bad_direct_routes.push_back(p_bad_direct_route);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

inline void IBDiag::AddLoopPath(IN direct_route_t *p_direct_route)
{
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "AddLoopPath %s\n",
              ibis_obj.ConvertDirPathToStr(p_direct_route).c_str());

    this->loop_direct_routes.push_back(p_direct_route);
}


void IBDiag::PostDiscoverFabricProcess()
{
    IBDIAG_ENTER;
    if (this->ibdiag_discovery_status == DISCOVERY_DUPLICATED_GUIDS) {
        this->SetLastError("Duplicated GUIDs were found during discovery process");
        IBDIAG_RETURN_VOID;
    }
    this->ibdiag_discovery_status = DISCOVERY_SUCCESS;
    IBDIAG_RETURN_VOID;
}


// Returns: SUCCESS_CODE / ERR_CODE_FABRIC_ERROR / ERR_CODE_IBDM_ERR / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDiag::DiscoverFabricOpenCAPorts(IN IBNode *p_node,
        IN direct_route_t *p_direct_route,
        IN SMP_NodeInfo *p_node_info,
        IN bool is_root,
        OUT IbdiagBadDirectRoute_t *p_bad_direct_route_info,
        IN bool push_new_direct_route)
{
    IBDIAG_ENTER;
    SMP_PortInfo curr_port_info;
    //sending only to entry port because we can't exit from CA
    if (this->ibis_obj.SMPPortInfoMadGetByDirect(p_direct_route,
            p_node_info->LocalPortNum,
            &curr_port_info)) {
        /*this->SetLastError("Failed to get port info for direct route %s (port=%u), err=%s",
                this->ibis_obj.ConvertDirPathToStr(p_direct_route).c_str(),
                p_node_info->LocalPortNum,
                this->ibis_obj.GetLastError());*/
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Failed to get port info for direct route %s (port=%u), err=%s\n",
                this->ibis_obj.ConvertDirPathToStr(p_direct_route).c_str(),
                p_node_info->LocalPortNum,
                this->ibis_obj.GetLastError());
        p_bad_direct_route_info->reason = IBDIAG_BAD_DR_PORT_PORT_INFO;
        p_bad_direct_route_info->port_num = p_node_info->LocalPortNum;
        IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
    }

    //set this port data to the node port
    IBLinkSpeed correct_speed = (IBLinkSpeed)curr_port_info.LinkSpeedActv;
    if ((curr_port_info.CapMsk & IBIS_IB_PORT_CAP_HAS_EXT_SPEEDS) &&
            curr_port_info.LinkSpeedExtActv) {
        /* port supports extended speeds and active extended speed
           is not zero ==> we have extended speed */
        correct_speed = extspeed2speed(curr_port_info.LinkSpeedExtActv);
    }

    int numLidsPerPort = 1 << curr_port_info.LMC;
    if ( curr_port_info.LID >= 0xc000 || ((int)curr_port_info.LID + numLidsPerPort) >= 0xc000) {
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "node %s invalid lid:%u lmc:%u\n",
                p_node->name.c_str(),curr_port_info.LID, curr_port_info.LMC);

        p_bad_direct_route_info->reason = IBDIAG_BAD_DR_PORT_INVALID_LID;
        p_bad_direct_route_info->port_num = p_node_info->LocalPortNum;

        FabricErrNodeInvalidLid *p_curr_fabric_discover_err =
                new FabricErrNodeInvalidLid(p_node, p_node_info->LocalPortNum, 
                                            curr_port_info.LID, curr_port_info.LMC);
        if (!p_curr_fabric_discover_err) {
            this->SetLastError("Failed to allocate FabricErrNodeInvalidLid");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
        }
        errors.push_back(p_curr_fabric_discover_err);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
    }

    IBPort *p_port = this->discovered_fabric.setNodePort(p_node,
            p_node_info->PortGUID,
            curr_port_info.LID,
            curr_port_info.LMC,
            p_node_info->LocalPortNum,
            (IBLinkWidth)curr_port_info.LinkWidthActv,
            correct_speed,
            (IBPortState)curr_port_info.PortState);
    if (!p_port) {
        this->SetLastError("Failed to set port data for port=%u of node=%s", p_node_info->LocalPortNum, p_node->getName().c_str());
        p_bad_direct_route_info->reason = IBDIAG_BAD_DR_PORT_INCONSISTENT;
        p_bad_direct_route_info->port_num = p_node_info->LocalPortNum;
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }

    //save smp port info attribute in the fabric
    int rc = this->fabric_extended_info.addSMPPortInfo(p_port, curr_port_info);
    if (rc) {
        this->SetLastError("Failed to set smp_port_info for port %u of node in direct route %s, err=%s",
                p_port->num, this->ibis_obj.ConvertDirPathToStr(p_direct_route).c_str(),
                this->fabric_extended_info.GetLastError());
        p_bad_direct_route_info->reason = IBDIAG_BAD_DR_PORT_INCONSISTENT;
        p_bad_direct_route_info->port_num = p_node_info->LocalPortNum;
        IBDIAG_RETURN(rc);
    }
    if (is_root == true) {  //if starting discovery from hca
        if (curr_port_info.PortPhyState != IB_PORT_PHYS_STATE_LINK_UP)    //not active
            IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

        if (push_new_direct_route) {
            //next bfs step
            direct_route_t *p_next_bfs_direct_route = new direct_route_t;
            if (!p_next_bfs_direct_route) {
                this->SetLastError("Failed to allocate next bfs direct route");
                p_bad_direct_route_info->reason = IBDIAG_BAD_DR_PORT_INCONSISTENT;
                p_bad_direct_route_info->port_num = p_node_info->LocalPortNum;
                IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
            }
            *p_next_bfs_direct_route =  *p_direct_route;
            p_next_bfs_direct_route->path.BYTE[p_direct_route->length] = p_node_info->LocalPortNum;
            p_next_bfs_direct_route->length = (u_int8_t)(p_direct_route->length + 1); //TODO check length
            this->BFSPushPath(p_next_bfs_direct_route);
        }
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


// Returns: SUCCESS_CODE / ERR_CODE_FABRIC_ERROR / ERR_CODE_IBDM_ERR / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDiag::DiscoverFabricOpenSWPorts(IN IBNode *p_node,
        IN direct_route_t *p_direct_route,
        IN SMP_NodeInfo *p_node_info,
        IN bool is_root,
        OUT IbdiagBadDirectRoute_t *p_bad_direct_route_info,
        IN bool push_new_direct_route)
{
    IBDIAG_ENTER;

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
               "DiscoverFabricOpenSWPorts SW GUID " U64H_FMT
               " Name: %s direct_route: %s\n",
               p_node_info->NodeGUID, p_node->name.c_str(),
               ibis_obj.ConvertDirPathToStr(p_direct_route).c_str());

    SMP_PortInfo curr_port_info;
    u_int16_t base_lid = 0;
    u_int32_t cap_mask = 0;
    u_int8_t lmc = 0;
    //going over all ports of this node (SW)
    // use unsigned int index to prevent overwrap
    for (unsigned int i = 0; i <= p_node_info->NumPorts; ++i) {
        phys_port_t port_num = (phys_port_t)i;
        //send port info mad to check if this port is active
        if (this->ibis_obj.SMPPortInfoMadGetByDirect(p_direct_route, port_num, &curr_port_info)) {
            /*this->SetLastError("Failed to get port info for direct route %s (port=%u), err=%s",
                    this->ibis_obj.ConvertDirPathToStr(p_direct_route).c_str(),
                    port_num,
                    this->ibis_obj.GetLastError());*/
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Failed to get port info for direct route %s (port=%u), err=%s\n",
                    this->ibis_obj.ConvertDirPathToStr(p_direct_route).c_str(),
                    port_num,
                    this->ibis_obj.GetLastError());

            p_bad_direct_route_info->reason = IBDIAG_BAD_DR_PORT_PORT_INFO;
            p_bad_direct_route_info->port_num = port_num;
            FabricErrPortInfoFail *p_curr_fabric_discover_err =
                    new FabricErrPortInfoFail(
                        p_node, port_num, ibis_obj.GetLastError());
            if (!p_curr_fabric_discover_err) {
                this->SetLastError("Failed to allocate FabricErrPortInfoFail");
                IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
            }
            errors.push_back(p_curr_fabric_discover_err);

            p_node->toIgnore=true;

            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "node %s (%p) toIgnore:%d\n",
                    p_node->name.c_str(),p_node,p_node->toIgnore);


            IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
        }

        if (port_num == 0)  {
            //manage port -
            //  we start loop from port_num = 0, therefore we will reach here first thing
            base_lid = curr_port_info.LID;
            cap_mask = curr_port_info.CapMsk;
            lmc = curr_port_info.LMC;
        }

        //add this port to discovered fabric or get ref to exist one
        IBLinkSpeed correct_speed = (IBLinkSpeed)curr_port_info.LinkSpeedActv;
        if ((cap_mask & IBIS_IB_PORT_CAP_HAS_EXT_SPEEDS) &&
                curr_port_info.LinkSpeedExtActv) {
            /* port supports extended speeds and active extended speed
               is not zero ==> we have extended speed */
            correct_speed = extspeed2speed(curr_port_info.LinkSpeedExtActv);
        }

        int numLidsPerPort = 1 << lmc;    
        if ( base_lid >= 0xc000 || ((int)base_lid + numLidsPerPort) >= 0xc000) {
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "node %s invalid lid:%u lmc:%u\n",
                    p_node->name.c_str(),base_lid, lmc);
            p_bad_direct_route_info->reason = IBDIAG_BAD_DR_PORT_INVALID_LID;
            p_bad_direct_route_info->port_num = port_num;

            FabricErrNodeInvalidLid *p_curr_fabric_discover_err =
                    new FabricErrNodeInvalidLid(p_node, port_num, base_lid, lmc);
            if (!p_curr_fabric_discover_err) {
                this->SetLastError("Failed to allocate FabricErrNodeInvalidLid");
                IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
            }
            errors.push_back(p_curr_fabric_discover_err);
            IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
        }

        IBPort *p_port = this->discovered_fabric.setNodePort(p_node,
                p_node_info->PortGUID,
                base_lid,
                lmc,
                port_num,
                (IBLinkWidth)curr_port_info.LinkWidthActv,
                correct_speed,
                (IBPortState)curr_port_info.PortState);
        if (!p_port) {
            this->SetLastError("Failed to store port data for port=%u of node=%s", p_node_info->LocalPortNum, p_node->getName().c_str());
            p_bad_direct_route_info->reason = IBDIAG_BAD_DR_PORT_INCONSISTENT;
            p_bad_direct_route_info->port_num = port_num;
            IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
        }

        //save smp port info attribute in the fabric
        int rc = this->fabric_extended_info.addSMPPortInfo(p_port, curr_port_info);
        if (rc) {
            this->SetLastError("Failed to store smp_port_info for port %u of node in direct route %s, err=%s",
                    port_num, this->ibis_obj.ConvertDirPathToStr(p_direct_route).c_str(),
                    this->fabric_extended_info.GetLastError());
            p_bad_direct_route_info->reason = IBDIAG_BAD_DR_PORT_INCONSISTENT;
            p_bad_direct_route_info->port_num = port_num;
            IBDIAG_RETURN(rc);
        }

        //manage port, can't go out from it
        if (port_num == 0)
            continue;
        //entry port, no sense to go out from it
        if ((is_root == false) && (port_num == p_node_info->LocalPortNum))
            continue;
        //not active
        if (curr_port_info.PortPhyState != IB_PORT_PHYS_STATE_LINK_UP)
            continue;

        if (push_new_direct_route) {
            //next bfs step
            direct_route_t *p_next_bfs_direct_route = new direct_route_t;
            if (!p_next_bfs_direct_route) {
                this->SetLastError("Failed to allocate next bfs direct route");
                p_bad_direct_route_info->reason = IBDIAG_BAD_DR_PORT_INCONSISTENT;
                p_bad_direct_route_info->port_num = port_num;
                IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
            }
            *p_next_bfs_direct_route =  *p_direct_route;
            p_next_bfs_direct_route->path.BYTE[p_direct_route->length] = port_num;
            p_next_bfs_direct_route->length = (u_int8_t)(p_direct_route->length + 1); //TODO Fix type
            this->BFSPushPath(p_next_bfs_direct_route);
        }
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


// Returns: SUCCESS_CODE / ERR_CODE_FABRIC_ERROR / ERR_CODE_IBDM_ERR / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDiag::DiscoverFabricBFSOpenPorts(IN direct_route_t * p_direct_route,
        IN IBNode *p_node,
        IN struct SMP_NodeInfo *p_node_info,
        IN bool is_visited_node,
        IN bool is_root,
        OUT IbdiagBadDirectRoute_t *p_bad_direct_route_info,
        IN bool push_new_direct_route)
{
    IBDIAG_ENTER;
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
               "DiscoverFabricBFSOpenPorts node GUID " U64H_FMT
               " Name: %s direct_route: %s\n",
               p_node_info->NodeGUID, p_node->name.c_str(),
               ibis_obj.ConvertDirPathToStr(p_direct_route).c_str());
    int rc;
    if ((p_node_info->NodeType == IB_SW_NODE) && (is_visited_node == false)) {  //not visited Switch
                                            //if visited than we already opened all it's ports
        rc = this->DiscoverFabricOpenSWPorts(p_node, p_direct_route, p_node_info, is_root, p_bad_direct_route_info,
                                             push_new_direct_route);
        if (rc)
            IBDIAG_RETURN(rc);
    } else if (p_node_info->NodeType == IB_CA_NODE ||
               p_node_info->NodeType == IB_RTR_NODE) { //for now RTR is like HCA
                                            //hca and therefore if visited already than
                                            //it means that now we are reaching it from it's other port
                                            //so we will open this port. In the first time we reached
                                            //it we didn't opened all the hca ports because we don't
                                            //know if we are in the same subnet
        rc = this->DiscoverFabricOpenCAPorts(p_node, p_direct_route, p_node_info, is_root, p_bad_direct_route_info,
                                             push_new_direct_route);
        if (rc)
            IBDIAG_RETURN(rc);
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


// Returns: SUCCESS_CODE / ERR_CODE_DB_ERR / ERR_CODE_IBDM_ERR / ERR_CODE_TRY_TO_DISCONNECT_CONNECTED_PORT
int IBDiag::DiscoverFabricBFSCreateLink(IN direct_route_t * p_direct_route,
        IN IBPort * p_port)
{
    IBDIAG_ENTER;

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
               "DiscoverFabricBFSCreateLink  direct_route:%s"
               "  ports node name%s, lid:%d port number:%d\n",
               ibis_obj.ConvertDirPathToStr(p_direct_route).c_str(),
               p_port->p_node->name.c_str(), p_port->base_lid, p_port->num);

    direct_route_t prev_direct_route = *p_direct_route;
    prev_direct_route.path.BYTE[p_direct_route->length - 1] = 0;
    --prev_direct_route.length;
    IBNode *p_prev_node = this->GetNodeByDirectRoute(&prev_direct_route);
    if (!p_prev_node) {
        this->SetLastError("DB error - can't find node for prev direct route = %s", this->ibis_obj.ConvertDirPathToStr(&prev_direct_route).c_str());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
               "prev_node of direct_route:%s is %s\n",
               ibis_obj.ConvertDirPathToStr(p_direct_route).c_str(),
               p_prev_node->name.c_str());

    IBPort* p_prev_port = p_prev_node->getPort(p_direct_route->path.BYTE[p_direct_route->length - 1]);
    if (!p_prev_port) {
        this->SetLastError("DB error - can't find port=%u for prev node=%s",
                p_direct_route->path.BYTE[p_direct_route->length - 1],
                p_prev_node->getName().c_str());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    //check if not connected already
    if (p_prev_port->p_remotePort) {
        this->SetLastError("Failure - reached connected port when trying to connect it to another one"  \
                " port=%s is already connected to port=%s, therefore we can't connect it to port=%s\n",
                p_prev_port->getName().c_str(), p_prev_port->p_remotePort->getName().c_str(), p_port->getName().c_str());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_TRY_TO_DISCONNECT_CONNECTED_PORT);
    }
    if (p_port->p_remotePort) {
        this->SetLastError("Failure - reached connected port when trying to connect it to another one"  \
                " port=%s is already connected to port=%s, therefore we can't connect it to port=%s\n",
                p_port->getName().c_str(), p_port->p_remotePort->getName().c_str(), p_prev_port->getName().c_str());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_TRY_TO_DISCONNECT_CONNECTED_PORT);
    }

    if (this->discovered_fabric.makeLinkBetweenPorts(p_prev_port, p_port)) {
        this->SetLastError("Failed to create a link");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

// Returns: SUCCESS_CODE / ERR_CODE_FABRIC_ERROR / ERR_CODE_IBDM_ERR / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDiag::DiscoverFabricBFSOpenNode(IN direct_route_t *p_direct_route,
        IN bool is_root,
        OUT IBNode **p_pnode,
        OUT struct SMP_NodeInfo *p_node_info,
        OUT bool *is_visited_node,
        IN progress_func_discovered_t discover_progress_func,
        OUT IbdiagBadDirectRoute_t *p_bad_direct_route_info)
{
    IBDIAG_ENTER;

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
               "DiscoverFabricBFSOpenNode  direct_route:%s is_root:%d\n",
               ibis_obj.ConvertDirPathToStr(p_direct_route).c_str(), is_root);

    if (this->ibis_obj.SMPNodeInfoMadGetByDirect(p_direct_route, p_node_info)) {
        /*this->SetLastError("Failed to get node information for direct route %s, err=%s",
                this->ibis_obj.ConvertDirPathToStr(p_direct_route).c_str(),
                this->ibis_obj.GetLastError());*/
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Failed to get node information for direct route %s, err=%s\n",
                this->ibis_obj.ConvertDirPathToStr(p_direct_route).c_str(),
                this->ibis_obj.GetLastError());
        p_bad_direct_route_info->reason = IBDIAG_BAD_DR_NODE_NODE_INFO;
        IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
    }
    if (is_root)
        this->root_port_num = p_node_info->LocalPortNum;

    //////////////////////////////////////////////////
    //Duplicated GUID algorithm
    //////////////////////////////////////////////////
    direct_route_t *p_old_direct_route;
    bool is_visited_port;
    /* is_visited_node is a income param.
     * we will use it latter to determine if we want to open a switch ports */
    if (check_duplicated_guids) {
        bool duplicated_node_guid;
        bool duplicated_port_guid;
        int rc = this->IsDuplicatedGuids(p_direct_route, p_node_info,
                &duplicated_node_guid, &duplicated_port_guid,
                is_visited_node, &is_visited_port, &p_old_direct_route,
                p_bad_direct_route_info);
        if (rc) {
            if (p_bad_direct_route_info->reason == IBDIAG_BAD_DR_NONE) {
                p_bad_direct_route_info->reason = IBDIAG_BAD_DR_NODE_DUPLICATE_CHECK;
            }
            IBDIAG_RETURN(rc);
        }
        if (duplicated_node_guid || !*is_visited_node)
            this->MarkNodeGuidAsBFSKnown(p_node_info->NodeGUID, p_direct_route);
        if (duplicated_port_guid || !is_visited_port) {
            this->MarkPortGuidAsBFSKnown(p_node_info->PortGUID, p_direct_route);
        }

        IBDIAG_LOG(TT_LOG_LEVEL_DISCOVER, "\nDuplicated node GUID check - result=%s\n Duplicated port GUID check - result=%s\n Visited Node - result=%s\n",
                        duplicated_node_guid ? "True" : "False", duplicated_port_guid ? "True" : "False", *is_visited_node ? "True" : "False");
    } else {    //assume excellent fabric
        *is_visited_node = false;
        list_p_direct_route old_p_direct_routers_list =
                this->GetDirectRoutesByNodeGuid(p_node_info->NodeGUID);
        if (!old_p_direct_routers_list.size()) {
            this->MarkNodeGuidAsBFSKnown(p_node_info->NodeGUID, p_direct_route);
        } else {
            *is_visited_node = true;
            p_old_direct_route = old_p_direct_routers_list.front();
        }

        old_p_direct_routers_list =
                this->GetDirectRoutesByPortGuid(p_node_info->PortGUID);
        if (!old_p_direct_routers_list.size()) {
            this->MarkPortGuidAsBFSKnown(p_node_info->PortGUID, p_direct_route);
        }
    }

    SMP_NodeDesc curr_node_desc;
    if (*is_visited_node == false) {
        if (this->ibis_obj.SMPNodeDescMadGetByDirect(p_direct_route, &curr_node_desc)) {
            /*this->SetLastError("Failed to get node description for direct route %s, err=%s",
                    this->ibis_obj.ConvertDirPathToStr(p_direct_route).c_str(),
                    this->ibis_obj.GetLastError());*/
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Failed to get node description for direct route %s, err=%s\n",
                    this->ibis_obj.ConvertDirPathToStr(p_direct_route).c_str(),
                    this->ibis_obj.GetLastError());

            p_bad_direct_route_info->reason = IBDIAG_BAD_DR_NODE_NODE_DESCRIPTION;
            IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
        }

        string additional_info;
        if (!IsValidNodeInfoData(p_node_info, additional_info)) {
            p_bad_direct_route_info->reason = IBDIAG_BAD_DR_NODE_BAD_NODE_INFO;
            p_bad_direct_route_info->message = additional_info;
            IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
        }
        //take care of progress bar
        ++this->discover_progress_bar_nodes.nodes_found;
        if (p_node_info->NodeType == IB_CA_NODE) {
            ++this->discover_progress_bar_nodes.ca_found;
        } else {
            ++this->discover_progress_bar_nodes.sw_found;
        }
        this->discover_progress_bar_ports.ports_found += p_node_info->NumPorts;
        if (discover_progress_func)
            discover_progress_func(&this->discover_progress_bar_nodes);

        //add this node to discovered fabric
        *p_pnode = this->discovered_fabric.makeNode((IBNodeType)p_node_info->NodeType,
                p_node_info->NumPorts,
                p_node_info->SystemImageGUID,
                p_node_info->NodeGUID,
                p_node_info->VendorID,
                p_node_info->DeviceID,
                p_node_info->revision,
                (*is_visited_node == false) ? (char *)curr_node_desc.Byte: "");
        if (!*p_pnode) {
            this->SetLastError("Failed to store new node for direct route=%s", this->ibis_obj.ConvertDirPathToStr(p_direct_route).c_str());
            p_bad_direct_route_info->reason = IBDIAG_BAD_DR_NODE_INCONSISTENT;
            IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
        }

        //save smp node info attribute in the fabric
        int rc = this->fabric_extended_info.addSMPNodeInfo(*p_pnode, *p_node_info);
        if (rc) {
            this->SetLastError("Failed to store smp_node_info for node in direct route %s, err=%s",
                     this->ibis_obj.ConvertDirPathToStr(p_direct_route).c_str(),
                     this->fabric_extended_info.GetLastError());
            p_bad_direct_route_info->reason = IBDIAG_BAD_DR_NODE_INCONSISTENT;
            IBDIAG_RETURN(rc);
        }
    } else {
        //get ref to exist one without consider all data given as args
        *p_pnode = this->GetNodeByDirectRoute(p_old_direct_route);
        //it's possible, if the node doesn't respond to 'node description' MAD,
        //then we won't create a node for it
        if (!*p_pnode) {
            p_bad_direct_route_info->reason = IBDIAG_BAD_DR_NODE_INCONSISTENT;
            //keep discovering fabric despite the fail
            IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
        }
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

// This function checks for valid node info data
bool IBDiag::IsValidNodeInfoData(IN struct SMP_NodeInfo *p_node_info,
                                 OUT string &additional_info)
{
    IBDIAG_ENTER;
    stringstream sstm;
    bool valid_node_info_data;

    //check that the number of port is valid
    if (!p_node_info)
        IBDIAG_RETURN(false);

    valid_node_info_data = true;
    if (p_node_info->NumPorts > IB_MAX_PHYS_NUM_PORTS ||
        p_node_info->NumPorts < IB_MIN_PHYS_NUM_PORTS) {
        valid_node_info_data = false;
        sstm << "Wrong number of ports " << (unsigned int)(p_node_info->NumPorts);
        additional_info = sstm.str();
    }
    IBDIAG_RETURN(valid_node_info_data);
}

u_int16_t IBDiag::GetPathNextNode(IN IBNode **p_node,
        IN lid_t dest_lid,
        INOUT direct_route_t **p_direct_route,
        IN direct_route_t *dr_path,
        IN struct SMP_NodeInfo &curr_node_info)
{
    struct SMP_LinearForwardingTable linear_forwarding_table;
    CLEAR_STRUCT(linear_forwarding_table);

    direct_route_t *new_p_direct_route = new direct_route_t;
    *new_p_direct_route = **p_direct_route;
    *p_direct_route = new_p_direct_route;

    //if ibdiagpath by direct route.
    if (dr_path->length > 0) {
        new_p_direct_route->path.BYTE[new_p_direct_route->length] =
            dr_path->path.BYTE[new_p_direct_route->length];
        new_p_direct_route->length++;
    } else {
        //must be the first local node or the source node
        if((*p_node)->type == IB_CA_NODE ) {
            //if it is the first HCA node
            if (new_p_direct_route->length == 1) {
                new_p_direct_route->path.BYTE[new_p_direct_route->length] =
                    curr_node_info.LocalPortNum;
                new_p_direct_route->length++;
            //if we got to HCA source we need to got back to the switch
            } else {
                new_p_direct_route->path.BYTE[new_p_direct_route->length] = 0;
                new_p_direct_route->length--;
            }
        } else if ((*p_node)->type == IB_SW_NODE) {
            if (this->ibis_obj.
                    SMPLinearForwardingTableGetByDirect(new_p_direct_route,
                        (lid_t)(dest_lid/IBIS_IB_MAD_SMP_LFT_NUM_BLOCKS),
                        &linear_forwarding_table,
                        NULL))
                IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
            u_int8_t next_port = linear_forwarding_table.Port[dest_lid % 64];
            new_p_direct_route->path.BYTE[new_p_direct_route->length] = next_port;
            new_p_direct_route->length++;
        }
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

bool IBDiag::IsVirtualLidForNode(IBNode *p_node,
                                 lid_t lid,
                                 stringstream *pss)
{
    IBDIAG_ENTER;
    int rc = 0;
    char strbuff[256];

    CLEAR_STRUCT(strbuff);
    sprintf(strbuff, "-I- Looking lid=%d in node %s vlid list\n",
            lid, p_node->name.c_str());
    (*pss) << strbuff;

    ibDiagClbck.Set(this, &fabric_extended_info, &errors, NULL);

    //TODO chage to collecting virtual lid data in REAL asynchronous way.
    //First "for" on all ports and than "for" on all lids.
    rc = BuildVirtualizationInfoDB(p_node);
    if(rc != IBDIAG_SUCCESS_CODE ) {
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,"Error during get on Virtualization" \
                   " Info on node %s"  \
                   " skipping sending VPortState and VPortInfo\n",
                   p_node->name.c_str());
        IBDIAG_RETURN(rc);
    }

    // second stage to get only the not down vport.
    rc = BuildVPortStateDB(p_node);
    if(rc != IBDIAG_SUCCESS_CODE) {
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,"Error during get on VPort" \
                   " State on node %s"  \
                   " skipping sending VPortInfo\n",
                   p_node->name.c_str());
        IBDIAG_RETURN(rc);
    }

    // third stage to get the VPort Info only for not down vports
    rc = BuildVPortInfoDB(p_node);
    if(rc != IBDIAG_SUCCESS_CODE) {
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,"Error during get on VPort" \
                   " Info on node %s skipping to next node\n",
                   p_node->name.c_str());
        IBDIAG_RETURN(rc);
    }

    // fourth stage to get the VNode Info
    rc = BuildVNodeInfoDB(p_node);
    if(rc != IBDIAG_SUCCESS_CODE) {
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,"Error during get on VNode" \
                   " Info on node %s skipping to next node\n",
                   p_node->name.c_str());
        IBDIAG_RETURN(rc);
    }

    // fifth stage to get the VNode Description
    rc = BuildVNodeDescriptionDB(p_node);
    if(rc != IBDIAG_SUCCESS_CODE) {
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,"Error during get on VNode" \
                   " Description on node %s skipping to next node\n",
                   p_node->name.c_str());
        IBDIAG_RETURN(rc);
    }

    rc = CheckAndSetVPortLid(errors);
    if(rc != IBDIAG_SUCCESS_CODE) {
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                   "Error during get on CheckAndSetVPortLid" \
                   " Description on node %s skipping to next node\n",
                   p_node->name.c_str());
        IBDIAG_RETURN(rc);
    }

    for (unsigned int i = 1 ; i <=  p_node->numPorts ; i++) {
        IBPort *p_curr_port = p_node->getPort((phys_port_t)i);
        if (!p_curr_port || !p_curr_port->is_data_worthy()) //port not exist
            continue;

        map_vportnum_vport vports = p_curr_port->VPorts;
        for (map_vportnum_vport::iterator vpI = vports.begin();
                vpI != vports.end(); ++vpI) {
            IBVPort *p_vport = vpI->second;
            if(!p_vport)
                continue;
            if(p_vport->get_vlid() == lid) {
                CLEAR_STRUCT(strbuff);
                sprintf(strbuff, "-I- Found vlid=%d in node %s\n",
                        lid, p_node->name.c_str());
                (*pss) << strbuff;

                IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
            }
        }
    }

    IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
}

int IBDiag::DiscoverPath(IN progress_func_discovered_t discover_progress_func,
                         IN u_int8_t max_hops, IN lid_t src_lid,
                         IN lid_t dest_lid, direct_route_t &dr_path,
                         OUT stringstream *pss)
{
    IBDIAG_ENTER;
    direct_route_t *p_current_direct_route = NULL;
    IbdiagBadDirectRoute_t bad_direct_route;
    struct SMP_NodeInfo curr_node_info;
    bool is_visited_node;
    IBNode *p_node;
    int rc,rc2;
    u_int8_t current_hop = 0;
    lid_t target_lid;
    bool found_src = false;
    IBPort *p_port;
    vector <IBNode *> sub_nodes;
    vector <IBPort *> sub_ports;
    char strbuff[256];

    if (dr_path.length == 0 && (!src_lid || !dest_lid)) {
        this->SetLastError("Options src_lid and dest_lid or dr_path are"  \
                  " mandatory to run ibdiagapth");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);
    }

    if (this->IsReady() == false) {
        this->SetLastError("IBDiag initialize wasn't done");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_NOT_READY);
    }
    if (max_hops > IBDIAG_MAX_HOPS) {
        this->SetLastError("MAX HOPS given is invalid, valid range is between 0 and 64");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);
    }
    //clean previous data
    this->CleanUpInternalDB();

    //initial step
    p_current_direct_route = new direct_route_t;
    if (!p_current_direct_route) {
        this->SetLastError("Failed to allocate root bfs direct route");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
    }
    CLEAR_STRUCT(*p_current_direct_route);
    p_current_direct_route->length = 1;

    bad_direct_route.clear();

    rc = this->DiscoverFabricBFSOpenNode(p_current_direct_route, true,
                                         &p_node, &curr_node_info,
                                         &is_visited_node,
                                         discover_progress_func,
                                         &bad_direct_route);
    if (rc) {
        this->SetLastError("Failed to open root node");
        rc2 = this->AddBadPath(&bad_direct_route, p_current_direct_route);
        if (rc2) {
            IBDIAG_RETURN(rc2);
        }
        if (bad_direct_route.reason == IBDIAG_BAD_DR_NONE) {
            this->SetLastError("Sanity check failed for root node discovery "
                               "failure reason");
        }
        IBDIAG_RETURN(rc);
    }
    this->root_node = p_node;
    rc = this->DiscoverFabricBFSOpenPorts(p_current_direct_route,
                                          p_node, &curr_node_info,
                                          false, true, &bad_direct_route,
                                          false);
    if (rc) {
        this->SetLastError("Failed to open ports of root node");
        rc2 = this->AddBadPath(&bad_direct_route, p_current_direct_route);
        if (rc2) {
            IBDIAG_RETURN(rc2);
        }
        if (bad_direct_route.reason == IBDIAG_BAD_DR_NONE) {
            this->SetLastError("Sanity check failed for root node ports"
                               " opening failure reason");
        }
        IBDIAG_RETURN(rc);
    }

    //first the target lid is the source lid and once the source is
    //found the target lid is the destination lid
    target_lid = src_lid;
    p_port = p_node->getPort(curr_node_info.LocalPortNum);
    if (!p_port) {
        this->SetLastError("DB error - failed to get port instance");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    /*if we run ibdiagpath using lid route and the root node is the source
     * or if we run ibdiagpath with direct route so the root node must be
     * the source*/
    if (p_port->base_lid == src_lid || dr_path.length > 0) {
        found_src = true;
        target_lid = dest_lid;
        sub_nodes.push_back(p_node);
        sub_ports.push_back(p_port);
          (*pss) << TRAVERSE_SRC_TO_DEST;
    } else
          (*pss) << TRAVERSE_LOCAL_TO_SRC;

    rc = this->GetPathNextNode(&p_node,
                               (u_int32_t)target_lid, &p_current_direct_route,
                               &dr_path, curr_node_info);

    //discover path
    while (current_hop < max_hops) {

        IBDIAG_LOG(TT_LOG_LEVEL_DISCOVER, "\nDiscovering ... %u nodes"
                   " (%u Switches & %u CA-s) discovered.\n"    \
                   "Current BFS direct route = %s\n",
                   this->discover_progress_bar_nodes.nodes_found,
                   this->discover_progress_bar_nodes.sw_found,
                   this->discover_progress_bar_nodes.ca_found,
                   this->ibis_obj.
                         ConvertDirPathToStr(p_current_direct_route).
                         c_str());

        //open the node we reached to
        rc2 = this->DiscoverFabricBFSOpenNode(p_current_direct_route, false,
                &p_node, &curr_node_info, &is_visited_node,
                discover_progress_func, &bad_direct_route);
        if (rc2) {
            this->SetLastError("Discovery failed for Lid %d when Quering LFT " \
                    "in Node %s", target_lid, p_node->getName().c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);
        }
        //open the ports of this node
        rc2 = this->DiscoverFabricBFSOpenPorts(p_current_direct_route, p_node,
                                               &curr_node_info, is_visited_node,
                                               false, &bad_direct_route, false);
        if (rc2) {
            this->SetLastError("Sanity check failed for ports discovery"
                               " failure reason (DR %s) ",
                    (this->ConvertDirPathToStr(p_current_direct_route)).c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);
        }

        /*
         *  TO: p_port FROM: p_port->remote.
         *  p_port - current port, entry port to the current node.
         *  p_port->remote - previous port, exit port from previous node.
         *  When we get to the src, we may need to go back to already visited
         *  node. If the src is switch, then the new diect route holds the exit
         *  port from the switch. If it's HCA, then the new direct route is an
         *  old one without last entry, thus we get nodeinfo of visited node,
         *  on its old entry port. To get the right p_port, we take the exit
         *  port of that node, through which we got to current HCA, i.e current
         *  p_port->remote.
         */
        if (found_src && p_port->p_node->type == IB_CA_NODE &&
            p_port->p_remotePort)
            p_port = p_port->p_remotePort;
        else
            p_port = p_node->getPort(curr_node_info.LocalPortNum);

        if (!p_port) {
            this->SetLastError("DB error - failed to get port instance");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        /* Check that the link doesn't exist yet.
         * It is possible, that we already created the link while traversing
         * from local to source, so don't create it again when traversing
         * from source to destination.
         * This also covers case, where link is created between current node
         * and another already visited node from previous traversing
         * (example: running from dest to src, and then back).
         */
        if (!p_port->p_remotePort)
            rc2 = this->DiscoverFabricBFSCreateLink(p_current_direct_route, p_port);
        if (rc2)
            IBDIAG_RETURN(rc2);

        if (found_src) {
            sub_nodes.push_back(p_node);
            sub_ports.push_back(p_port);
            sub_ports.push_back(p_port->p_remotePort);
        }

        CLEAR_STRUCT(strbuff);
        sprintf(strbuff, "-I- From: lid=%02d port guid=" U64H_FMT
                " dev=%d %s Port=%d\n",
                p_port->p_remotePort->base_lid, p_port->p_remotePort->guid_get(),
                p_port->p_remotePort->p_node->devId,
                p_port->p_remotePort->p_node->name.c_str(), p_port->p_remotePort->num);
        (*pss) << strbuff;
        CLEAR_STRUCT(strbuff);
        sprintf(strbuff, "-I- To : lid=%02d port guid=" U64H_FMT
                " dev=%d %s Port=%d\n",
                p_port->base_lid, p_port->guid_get(), p_port->p_node->devId,
                p_port->p_node->name.c_str(), p_port->num);
        (*pss) << strbuff;

        if (dr_path.length > 0 && p_current_direct_route->length == dr_path.length)
            break;
        else if(found_src) {
            if (p_port->base_lid == dest_lid)
                break;
            //if we already found source and we got to HCA that isn't the
            //distination it should be virtual lid or routing error from the switch LFT
            else if (p_node->type == IB_CA_NODE) {
                //check that this is not virtual node.
                if(IsVirtualLidForNode(p_node, dest_lid, pss))
                    IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
                break;
            }
        } else if (p_port->base_lid == src_lid ||
                   (p_port->base_lid != src_lid && p_node->type == IB_CA_NODE &&
                    !IsVirtualLidForNode(p_node, src_lid, pss))) {
                    found_src = true;
                    target_lid = dest_lid;
                    (*pss) << TRAVERSE_SRC_TO_DEST;
                    sub_nodes.push_back(p_node);
        }

        current_hop++;
        //////////////////////////////////////////////////
        //stop condition
        //////////////////////////////////////////////////
        if (current_hop >= max_hops) {
             IBDIAG_LOG(TT_LOG_LEVEL_DISCOVER,"\nReached maximum hops=%u during"
                       " discovery", max_hops);
            FabricErrDiscovery *p_curr_fabric_err =
                new FabricErrDiscovery(
                        p_port->p_node,
                        max_hops);
            if (!p_curr_fabric_err) {
                this->SetLastError("Failed to allocate"  \
                                   " FabricErrNodeMlnxCountersPageVer");
                IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            }
            p_curr_fabric_err->SetLevel(EN_FABRIC_ERR_WARNING);
            errors.push_back(p_curr_fabric_err);

            goto finished_discovery;
        }

        rc = this->GetPathNextNode(&p_node,
                                   (u_int32_t)target_lid,
                                   &p_current_direct_route, &dr_path,
                                   curr_node_info);
    }       /* BFS while */

    this->ApplySubCluster(sub_nodes, sub_ports);

finished_discovery:
    this->PostDiscoverFabricProcess();
    IBDIAG_RETURN(rc);
}

/*
 * Remove all nodes that are in subCluster from Fabric::NodeByName and extended
 * fabric db.
 * Those are the two main places, that our stages check for nodes in fabric.
 * We don't delete the actual objects from db, since they are needed to find
 * nodes by direct route in function GetNodeByDirectRoute (otherwise there
 * would be "holes" in db).
 */
int IBDiag::ApplySubCluster(IN vector_p_node sub_nodes,
                            IN vector_p_port sub_ports)
{
    //for all nodes mark with "in_sub_fabric = true" only the ones in the path
    this->discover_progress_bar_nodes.nodes_found = 0;
    this->discover_progress_bar_nodes.ca_found = 0;
    this->discover_progress_bar_nodes.sw_found = 0;
    this->discover_progress_bar_ports.ports_found = 0;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {
        IBNode *p_node = (*nI).second;
        if (!p_node) {
            cout << "-E-  found null node in nodes_vector " << endl;
            return 1;
        }
        //if the node isn't in sub_nodes list it's isnt in sub cluster.
        if(std::find(sub_nodes.begin(),
                    sub_nodes.end(),
                    p_node) == sub_nodes.end())
            p_node->setInSubFabric(false);
        else {
            ++this->discover_progress_bar_nodes.nodes_found;
            if (p_node->type == IB_CA_NODE)
                ++this->discover_progress_bar_nodes.ca_found;
            else
                ++this->discover_progress_bar_nodes.sw_found;
            uint8_t i = 0;
            if (p_node->type == IB_CA_NODE)
                i = 1;
            //set in_sub_cluster = false to unrelevant ports to the path
            for (;i <= p_node->numPorts; i++) {
                IBPort *p_port = p_node->getPort(i);
                if(!p_port) {
                    cout << "-E-  found null port in ports_vector " << endl;
                    return 1;
                }
                if(std::find(sub_ports.begin(),
                            sub_ports.end(),
                            p_port) == sub_ports.end())
                    p_port->setInSubFabric(false);
                else
                    this->discover_progress_bar_ports.ports_found++;
            }
        }
    }

    //remove all nodes that not in the path from NodeByName
    this->discovered_fabric.applySubCluster();

    //remove all node that not in the path from nodes_vector & ports_vector
    this->fabric_extended_info.applySubCluster();

    return 0;
}


/****************************************************/
/*******************public methods*******************/
/****************************************************/
// Returns: SUCCESS_CODE / ERR_CODE_FABRIC_ERROR / ERR_CODE_IBDM_ERR / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM / ERR_CODE_DB_ERR / ERR_CODE_TRY_TO_DISCONNECT_CONNECTED_PORT / ERR_CODE_INCORRECT_ARGS
int IBDiag::DiscoverFabric(IN progress_func_discovered_t discover_progress_func, IN u_int8_t max_hops)
{
    IBDIAG_ENTER;
    direct_route_t *p_bfs_direct_route = NULL;
    IbdiagBadDirectRoute_t bad_direct_route;
    struct SMP_NodeInfo curr_node_info;
    bool is_visited_node;
    IBNode *p_node;
    int rc,rc2, rc3;

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Start DiscoverFabric max_hops:%d\n",
               max_hops);

    if (this->IsReady() == false) {
        this->SetLastError("IBDiag initialize wasn't done");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_NOT_READY);
    }
    if (max_hops > IBDIAG_MAX_HOPS) {
        this->SetLastError("MAX HOPS given is invalid, valid range is between 0 and 64");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);
    }

    //clean previous data
    this->CleanUpInternalDB();

    //initial step
    p_bfs_direct_route = new direct_route_t;
    if (!p_bfs_direct_route) {
        this->SetLastError("Failed to allocate root bfs direct route");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
    }
    CLEAR_STRUCT(*p_bfs_direct_route);
    p_bfs_direct_route->length = 1;

    bad_direct_route.clear();

    rc = this->DiscoverFabricBFSOpenNode(p_bfs_direct_route, true,
            &p_node, &curr_node_info, &is_visited_node,
            discover_progress_func, &bad_direct_route);
    if (rc) {
        this->SetLastError("Failed to open root node");
        rc2 = this->AddBadPath(&bad_direct_route, p_bfs_direct_route);
        if (rc2) {
            IBDIAG_RETURN(rc2);
        }
        if (bad_direct_route.reason == IBDIAG_BAD_DR_NONE) {
            this->SetLastError("Sanity check failed for root node discovery failure reason");
        }
        IBDIAG_RETURN(rc);
    }
    this->root_node = p_node;
    rc = this->DiscoverFabricBFSOpenPorts(p_bfs_direct_route, p_node, &curr_node_info, false, true, &bad_direct_route);
    if (rc) {
        this->SetLastError("Failed to open ports of root node");
        rc2 = this->AddBadPath(&bad_direct_route, p_bfs_direct_route);
        if (rc2) {
            IBDIAG_RETURN(rc2);
        }
        if (bad_direct_route.reason == IBDIAG_BAD_DR_NONE) {
            this->SetLastError("Sanity check failed for root node ports opening failure reason");
        }
        IBDIAG_RETURN(rc);
    }
    this->AddGoodPath(p_bfs_direct_route);

    //bfs on all the fabric
    while (!this->bfs_list.empty()) {
        p_bfs_direct_route = this->BFSPopPath();

        IBDIAG_LOG(TT_LOG_LEVEL_DISCOVER, "\nDiscovering ... %u nodes (%u Switches & %u CA-s) discovered.\n"    \
                "Current BFS direct route = %s\n",
                this->discover_progress_bar_nodes.nodes_found,
                this->discover_progress_bar_nodes.sw_found,
                this->discover_progress_bar_nodes.ca_found,
                this->ibis_obj.ConvertDirPathToStr(p_bfs_direct_route).c_str());

        //////////////////////////////////////////////////
        //stop condition
        //////////////////////////////////////////////////
        if (p_bfs_direct_route->length > max_hops) {
            IBDIAG_LOG(TT_LOG_LEVEL_DISCOVER, "\nReached maximum hops=%u during discovery", max_hops);
            goto finished_discovery;
        }

        //optimization - this direct route can be reached already from other side
        if (this->GetNodeByDirectRoute(p_bfs_direct_route)) {
            this->AddLoopPath(p_bfs_direct_route);
            continue;
        }

        //open the node we reached to
        rc2 = this->DiscoverFabricBFSOpenNode(p_bfs_direct_route, false,
                &p_node, &curr_node_info, &is_visited_node,
                discover_progress_func, &bad_direct_route);
        if (rc2 == IBDIAG_ERR_CODE_FABRIC_ERROR) {       //bad link, we will try to coninue
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
            rc3 = this->AddBadPath(&bad_direct_route, p_bfs_direct_route);
            if (rc3) {
                IBDIAG_RETURN(rc3);
            }
            if (bad_direct_route.reason == IBDIAG_BAD_DR_NONE) {
                this->SetLastError("Sanity check failed for node discovery failure reason (DR %s) ",
                                   (this->ConvertDirPathToStr(p_bfs_direct_route)).c_str());
                IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);
            }
            // clear for next time
            bad_direct_route.clear();
            continue;
        } else if (rc2) {    //error
            this->AddBadPath(&bad_direct_route, p_bfs_direct_route);
            IBDIAG_RETURN(rc2);
        }

        //open the ports of this node
        rc2 = this->DiscoverFabricBFSOpenPorts(p_bfs_direct_route, p_node,
                &curr_node_info, is_visited_node, false, &bad_direct_route);
        if (rc2 == IBDIAG_ERR_CODE_FABRIC_ERROR) {       //bad link, we will try to coninue
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
            rc3 = this->AddBadPath(&bad_direct_route, p_bfs_direct_route);
            if (rc3) {
                IBDIAG_RETURN(rc3);
            }
            if (bad_direct_route.reason == IBDIAG_BAD_DR_NONE) {
                this->SetLastError("Sanity check failed for ports discovery failure reason (DR %s) ",
                                   (this->ConvertDirPathToStr(p_bfs_direct_route)).c_str());
                IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);
            }
            // clear for next time
            bad_direct_route.clear();
            continue;
        } else if (rc2) {    //error
            this->AddBadPath(&bad_direct_route, p_bfs_direct_route);
            IBDIAG_RETURN(rc2);
        }

        //add link to this node - from previous port to enter port
        IBPort *p_port = p_node->getPort(curr_node_info.LocalPortNum);
        if (!p_port) {
            this->SetLastError("DB error - failed to get port instance");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }
        rc2 = this->DiscoverFabricBFSCreateLink(p_bfs_direct_route, p_port);
        if (rc2) {
            this->AddBadPath(&bad_direct_route, p_bfs_direct_route);
            IBDIAG_RETURN(rc2);
        }

        //mark good path or loop path if node already visited
        if (is_visited_node == false) {
            this->AddGoodPath(p_bfs_direct_route);
            continue;
        } else {
            this->AddLoopPath(p_bfs_direct_route);
            continue;
        }
    }       /* BFS while */


finished_discovery:
    this->PostDiscoverFabricProcess();
    IBDIAG_RETURN(rc);
}


void IBDiag::PrintAllRoutes()
{
    IBDIAG_ENTER;
    list_p_direct_route::iterator it;
    list_p_bad_direct_route::iterator bad_dr_it;

    printf("Good Direct Routes:\n");
    for (it = this->good_direct_routes.begin(); it != this->good_direct_routes.end(); ++it) {
        printf("%s", this->ibis_obj.ConvertDirPathToStr(*it).c_str());
        printf("\t");
    }
    printf("\n\n");
    printf("Bad Direct Routes:\n");
    for (bad_dr_it = this->bad_direct_routes.begin(); bad_dr_it != this->bad_direct_routes.end(); ++bad_dr_it) {
        printf("%s", this->ibis_obj.ConvertDirPathToStr((*bad_dr_it)->direct_route).c_str());
        printf("\t");
    }
    printf("\n\n");
    printf("Loop Direct Routes:\n");
    for (it = this->loop_direct_routes.begin(); it != this->loop_direct_routes.end(); ++it) {
        printf("%s", this->ibis_obj.ConvertDirPathToStr(*it).c_str());
        printf("\t");
    }
    printf("\n\n");
/*
    this->discovered_fabric.dump(cout);
    printf("\n\n");
*/
    IBDIAG_RETURN_VOID;
}

void IBDiag::PrintAllDirectRoutes() // useful for debug
{
    map_guid_list_p_direct_route::iterator map_it;
    list_p_direct_route::iterator list_it;

    printf("NODES:\n");

    for (map_it = bfs_known_node_guids.begin(); map_it != bfs_known_node_guids.end(); ++map_it) {
        printf("GUID: " U64H_FMT ", DR: ", map_it->first);
        for (list_it = map_it->second.begin(); list_it != map_it->second.end(); ++list_it)
            printf("%s", this->ibis_obj.ConvertDirPathToStr(*list_it).c_str());
        printf("\n");
    }

    printf("\nPORTS:\n");

    for (map_it = bfs_known_port_guids.begin(); map_it != bfs_known_port_guids.end(); ++map_it) {
        printf("GUID: " U64H_FMT ", DR: ", map_it->first);
        for (list_it = map_it->second.begin(); list_it != map_it->second.end(); ++list_it)
            printf("%s", this->ibis_obj.ConvertDirPathToStr(*list_it).c_str());
        printf("\n");
    }
    printf("\n");
}

void IBDiag::GetGoodDirectRoutes(OUT list_string& good_direct_routes)
{
    string curr_direct_route;
    for (list_p_direct_route::iterator it = this->good_direct_routes.begin();
            it != this->good_direct_routes.end();
            ++it) {
        curr_direct_route = this->ibis_obj.ConvertDirPathToStr(*it);
        good_direct_routes.push_back(curr_direct_route);
    }
}

void IBDiag::GetLoopDirectRoutes(OUT list_string& loop_direct_routes)
{
    string curr_direct_route;
    for (list_p_direct_route::iterator it = this->loop_direct_routes.begin();
            it != this->loop_direct_routes.end();
            ++it) {
        curr_direct_route = this->ibis_obj.ConvertDirPathToStr(*it);
        loop_direct_routes.push_back(curr_direct_route);
    }
}

void IBDiag::GetBadDirectRoutes(OUT list_string& bad_direct_routes)
{
    string curr_direct_route;
    for (list_p_bad_direct_route::iterator it = this->bad_direct_routes.begin();
         it != this->bad_direct_routes.end();
         ++it) {
        curr_direct_route = this->ibis_obj.ConvertDirPathToStr((*it)->direct_route);
        bad_direct_routes.push_back(curr_direct_route);
    }
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::BuildSwitchInfoDB(list_p_fabric_general_err& retrieve_errors, progress_func_nodes_t progress_func)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    progress_bar_nodes_t progress_bar_nodes;
    CLEAR_STRUCT(progress_bar_nodes);

    // go over all nodes:
    struct SMP_SwitchInfo curr_switch_info;
    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        ++progress_bar_nodes.nodes_found;
        if (p_curr_node->type == IB_SW_NODE)
            ++progress_bar_nodes.sw_found;
        else
            ++progress_bar_nodes.ca_found;
        if (progress_func)
            progress_func(&progress_bar_nodes, &this->discover_progress_bar_nodes);

        //check if hca
        if (p_curr_node->type == IB_CA_NODE)
            continue;

        //get direct route for this node
        direct_route_t *p_curr_direct_route = this->GetDirectRouteByNodeGuid(p_curr_node->guid_get());
        if (!p_curr_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s", p_curr_node->getName().c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        //send switch info mad for this switch
        if (this->ibis_obj.SMPSwitchInfoMadGetByDirect(p_curr_direct_route, &curr_switch_info)) {
            FabricErrNodeNotRespond *p_curr_fabric_node_err =
                    new FabricErrNodeNotRespond(p_curr_node, "SMPSwitchInfoMad");
            if (!p_curr_fabric_node_err) {
                this->SetLastError("Failed to allocate FabricErrNodeNotRespond");
                IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
            }
            retrieve_errors.push_back(p_curr_fabric_node_err);
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
            continue;   //try next switch
        }
        int rc2 = this->fabric_extended_info.addSMPSwitchInfo(p_curr_node, curr_switch_info);
        if (rc2) {
            this->SetLastError("Failed to add SMPSwitchInfo for switch=%s, err=%s",
                    p_curr_node->getName().c_str(), this->fabric_extended_info.GetLastError());
            IBDIAG_RETURN(rc2);
        }
    }
    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR /
// ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::BuildPortInfoExtendedDB(list_p_fabric_general_err& retrieve_errors,
                                    progress_func_nodes_t progress_func)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE, rc2;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    //mark nodes clean of "non-supportiveness"
    ResetAppData();
    u_int32_t port_info_cap_mask = 0;
    u_int16_t port_info_cap_mask2 = 0;

    progress_bar_nodes_t progress_bar_nodes;
    CLEAR_STRUCT(progress_bar_nodes);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = IBDiagSMPPortInfoExtendedGetClbck;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName " \
                               "map for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        //get cap mask for this node
        bool to_read_cap = true;

        ++progress_bar_nodes.nodes_found;
        if (p_curr_node->type == IB_SW_NODE)
            ++progress_bar_nodes.sw_found;
        else
            ++progress_bar_nodes.ca_found;
        if (progress_func)
            progress_func(&progress_bar_nodes, &this->discover_progress_bar_nodes);

        // go over all ports of this node:
        //start from port 1, no need to read port 0
        for (u_int32_t i = 1; i <= p_curr_node->numPorts; ++i) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            if (!p_curr_port)                       //port not exist
                continue;
            if (!p_curr_port->p_remotePort)         //port not connected
                continue;
            SMP_PortInfo *curr_port_info =
                this->fabric_extended_info.getSMPPortInfo(
                                            p_curr_port->createIndex);
            if (!curr_port_info) {
                this->SetLastError("DB error - found connected port=%s"
                                   " without SMPPortInfo",
                                   p_curr_port->getName().c_str());
                IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            }
            //don't query non active ports
            if (curr_port_info->PortPhyState != IB_PORT_PHYS_STATE_LINK_UP)
                continue;

            //if we already have FEC for this port, no need to send the MAD
            if (p_curr_port->get_fec_mode() != IB_FEC_NA)
                continue;

            //for switch read cap only on port 0, for others read on every port
            if (to_read_cap) {
                rc2 = ReadPortInfoCapMask(p_curr_node, p_curr_port,
                                port_info_cap_mask, &port_info_cap_mask2);
                if (rc2) {
                    rc = rc2;
                    goto exit;
                }
                IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "node: %s port_info_cap_mask: 0x%x port_info_cap_mask2: 0x%x\n",
                           (*nI).first.c_str(), port_info_cap_mask, port_info_cap_mask2);
                if (p_curr_node->type == IB_SW_NODE)
                    to_read_cap = false;
            }

            //is this device supports Port Info Extended Port Info mad
            if (!IS_SUPPORT_PORT_INFO_EXTENDED_FAILED(p_curr_node->appData1.val)) {
                if (IS_SUPPORT_PORT_INFO_CAP_MASK2(port_info_cap_mask) &&
                    IS_SUPPORT_PORT_INFO_EXTENDED(port_info_cap_mask2)) {
                    SMP_PortInfoExtended curr_port_info_ext;

                    //Send the MAD
                    clbck_data.m_data1 = p_curr_port;

                    this->ibis_obj.SMPPortInfoExtMadGetByDirect(
                        this->GetDirectRouteByNodeGuid(
                            p_curr_port->p_node->guid_get()),
                        p_curr_port->num,
                        &curr_port_info_ext,
                        &clbck_data);

                } else {
                    //don't count it as an error, this is an old fw
                    //that doesn't support yet
                    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                               "The node %s doesn't support " \
                               "PortInfoExtended MAD\n",
                               p_curr_node->getName().c_str());

                    //mark we already checked for support
                    p_curr_node->appData1.val |= NOT_SUPPORT_PORT_INFO_EXTENDED;
                }
            }

            if(ibDiagClbck.GetState())
                goto exit;
        }
    }
exit:
    this->ibis_obj.MadRecAll();

    if (rc) {
        if (last_error.empty())
            this->SetLastError("Retrieve of PortInfoExtended Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "Retrieve of PortInfoExtended Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::ParsePSLFile(const string & file_name, string& output)
{
    IBDIAG_ENTER;

    ibdmClearInternalLog();

    int rc = discovered_fabric.parsePSLFile(file_name);

    char *buffer= ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    output += buffer;
    free(buffer);

    if (rc)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    if(!root_node->isValidPSLInfo()) {
        this->SetLastError("Can not find path SL info for local node.");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    ibis_obj.SetPSLTable(root_node->PSL);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::ParseSADumpFile(const string & file_name, string& output)
{
    IBDIAG_ENTER;

    ibdmClearInternalLog();

    int rc = discovered_fabric.parseSADumpFile(file_name);

    char *buffer= ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    output += buffer;
    free(buffer);

    if (rc)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR / ERR_CODE_IBDM_ERR
int IBDiag::DumpCapabilityMaskFile(const char* file_name, string& output)
{
    IBDIAG_ENTER;

    ibdmClearInternalLog();

    int rc = capability_module.DumpCapabilityMaskFile(file_name);

    char *buffer = ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output"
                           " for capability masks output file");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    output += buffer;
    free(buffer);

    if (rc)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//for debug purposes
//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR / ERR_CODE_IBDM_ERR
int IBDiag::DumpGuid2Mask(const char *file_name, string& output)
{
    IBDIAG_ENTER;

    ibdmClearInternalLog();

    int rc = capability_module.DumpGuid2Mask(file_name,
                                             this->GetDiscoverFabricPtr());

    char *buffer = ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output"
                           " for capability masks output file");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    output += buffer;
    free(buffer);

    if (rc)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::ParseCapabilityMaskFile(const char *file_name, string& output)
{
    IBDIAG_ENTER;

    ibdmClearInternalLog();

    int rc = capability_module.ParseCapabilityMaskFile(file_name);

    char *buffer = ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output for capability use");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    output += buffer;
    free(buffer);

    if (rc)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / IBDIAG_ERR_CODE_IBDM_ERR/ ERR_CODE_DB_ERR
int IBDiag::ParseNodeNameMapFile(const char* file_name, string& output)
{
    IBDIAG_ENTER;

    ibdmClearInternalLog();

    int rc = discovered_fabric.parseNodeNameMapFile(file_name);

    char *buffer = ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output"
                           " for node name map use");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    output += buffer;
    free(buffer);

    if (rc)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::BuildCapabilityCache(string& output)
{
    IBDIAG_ENTER;

    bool is_smp_only_fw, is_gmp_only_fw;
    fw_version_obj_t fw;
    fw.major = fw.minor = fw.sub_minor = 0;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node "
                               "in NodeByName map for key = %s",
                               (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }
        //check if the node has a longest guid prefix matched in the capability
        //mask configuration
        u_int64_t curr_node_guid = p_curr_node->guid_get();
        u_int8_t prefix_len = 0;
        u_int64_t matched_guid = 0;
        query_or_mask_t qmask;
        capability_mask_t mask;

        mask.clear();

        //SMP
        bool smp_prefix_match =
            capability_module.IsLongestSMPPrefixMatch(curr_node_guid,
                                                      prefix_len,
                                                      matched_guid,
                                                      qmask);
        //first look if it matched by longest prefix match, query or mask
        //if mask is found, it will be added now, if query, then the mask will be added later
        //anyway don't check other configurations
        if (smp_prefix_match) {
            if (!qmask.to_query)
                //store the found mask in the cache
                capability_module.AddSMPCapabilityMask(curr_node_guid, qmask.mask);
        }
        //look for match by ven id + dev id that don't support query by mad
        else if (capability_module.IsSMPUnsupportedMadDevice(p_curr_node->vendId,
                                                             p_curr_node->devId,
                                                             mask))
            capability_module.AddSMPCapabilityMask(curr_node_guid, mask);
        //see if there a device configured only with fw = 0.0.0,
        //then the mask can be already added
        //no need to ask for device's fw in the fabric
        else {
            is_smp_only_fw = false;
            if (capability_module.GetSMPFwConfiguredMask(p_curr_node->vendId,
                                                         p_curr_node->devId,
                                                         fw,
                                                         mask,
                                                         &is_smp_only_fw) == IBDIAG_SUCCESS_CODE &&
                is_smp_only_fw)
                capability_module.AddSMPCapabilityMask(curr_node_guid, mask);
        }

        //GMP
        mask.clear();
        bool gmp_prefix_match = capability_module.IsLongestGMPPrefixMatch(
                                                    curr_node_guid,
                                                    prefix_len,
                                                    matched_guid,
                                                    qmask);

        if (gmp_prefix_match) {
            if (!qmask.to_query)
                //store the found mask in the cache
                capability_module.AddGMPCapabilityMask(curr_node_guid, qmask.mask);
        }
        else if (capability_module.IsGMPUnsupportedMadDevice(p_curr_node->vendId,
                                                             p_curr_node->devId,
                                                             mask))
            //look for match by ven id + dev id that don't support query by mad
            capability_module.AddGMPCapabilityMask(curr_node_guid, mask);
        //see if there a device configured only with fw = 0.0.0,
        //then the mask can be already added
        //no need to ask for device's fw in the fabric
        else {
            is_gmp_only_fw = false;
            if (capability_module.GetGMPFwConfiguredMask(p_curr_node->vendId,
                                                         p_curr_node->devId,
                                                         fw,
                                                         mask,
                                                         &is_gmp_only_fw) == IBDIAG_SUCCESS_CODE &&
                is_gmp_only_fw)
                capability_module.AddGMPCapabilityMask(curr_node_guid, mask);
        }
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR / IBDIAG_ERR_CODE_IBDM_ERR
int IBDiag::ParseScopePortGuidsFile(const string & file_name,
                                    string& output,
                                    bool include_in_scope)
{
    IBDIAG_ENTER;

    ibdmClearInternalLog();

    //parse and apply scope
    int rc = discovered_fabric.parseScopePortGuidsFile(file_name,
                                                       include_in_scope);

    char *buffer = ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    output += buffer;
    free(buffer);

    //apply scope marked in fabric to extended info
    this->fabric_extended_info.applySubCluster();

    if (rc)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

