/*
 *  Copyright (c) 2004-2013 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include "infiniband/ibdiag/ibdiag.h"
#include "infiniband/ibdiag/ibdiag_clbck.h"

extern IBDiagClbck ibDiagClbck;

/****************************************************/

typedef pair < u_int64_t, IBPort * > pair_guid_port;
typedef pair < u_int64_t, IBNode * > pair_guid_node;
typedef pair < u_int64_t, IBSystem * > pair_guid_system;

//typedef map_guid2port::iterator it_map_guid2port;

/****************************************************/
static void readPortGUIDsToVec(IBDMExtendedInfo *extended_info,
        IBPort *p_port, u_int16_t guid_cap,
        vec_guids &vec_guids)
{
    IBDIAG_ENTER;
    int num_of_blocks;
    int entries;
    u_int64_t aguid;

    vec_guids.clear();
    num_of_blocks = (guid_cap + IBIS_IB_MAD_SMP_GUIDS_TABLE_NUM_BLOCKS - 1) /
            IBIS_IB_MAD_SMP_GUIDS_TABLE_NUM_BLOCKS;
    entries = IBIS_IB_MAD_SMP_GUIDS_TABLE_NUM_BLOCKS;

    for (int i = 0; i < num_of_blocks; ++i) {
        SMP_GUIDInfo *p_guid_info =
                extended_info->getSMPGUIDInfo(p_port->createIndex, i);
        if (!p_guid_info)
            continue;

        if ((i + 1) * IBIS_IB_MAD_SMP_GUIDS_TABLE_NUM_BLOCKS  > guid_cap)
            entries = guid_cap % IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS;

        for (int j = 0; j < entries; ++j) {
            aguid = p_guid_info->GUIDBlock.GUID[j].High;
            aguid = aguid << 32;
            aguid |= 0xffffffff & p_guid_info->GUIDBlock.GUID[j].Low;
            vec_guids.push_back(aguid);
        }
    }
    IBDIAG_RETURN_VOID;
}


/****************************************************/
/******************private methods*******************/
/****************************************************/
void IBDiag::DumpAliasGUID(ofstream &sout)
{
    IBDIAG_ENTER;
    char line[2096];
    phys_port_t ports_range_start, ports_range_end;
    vec_guids vec_guids;

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node)   //node not exists
            continue;

        if (p_curr_node->type == IB_SW_NODE) {
            ports_range_start   = 0;
            ports_range_end     = 0;
        } else {
            ports_range_start   = 1;
            ports_range_end     = p_curr_node->numPorts;
        }

        // use unsigned int index to prevent overwrap
        for (unsigned int i = ports_range_start; i <= ports_range_end; i++) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            if (!p_curr_port)
                continue;

            SMP_PortInfo *p_port_info =
                    this->fabric_extended_info.getSMPPortInfo(p_curr_port->createIndex);
            if (!p_port_info)
                continue;

            memset(line, 0, sizeof(line));
            sprintf(line, "Port Name=%s, Primary GUID=" U64H_FMT,
                    p_curr_port->getName().c_str(), p_curr_port->guid_get());
            sout << line << endl;
            u_int8_t guid_cap = p_port_info->GUIDCap;
            readPortGUIDsToVec(&this->fabric_extended_info, p_curr_port,
                guid_cap, vec_guids);

            for (vec_guids::iterator it = vec_guids.begin();
                    it != vec_guids.end(); ++it) {
                u_int64_t guid = *it;
                if (!guid)
                    continue;

                sprintf(line, "\talias guid=" U64H_FMT, guid);
                sout << line << endl;
            }
            sout << endl;
        }
    }
    IBDIAG_RETURN_VOID;
}

/****************************************************/
/*******************public methods*******************/
/****************************************************/
//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::BuildAliasGuidsDB(list_p_fabric_general_err &aguid_errors,
        progress_func_nodes_t progress_func)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    phys_port_t ports_range_start, ports_range_end;
    SMP_GUIDInfo guid_info;

    ibDiagClbck.Set(this, &fabric_extended_info, &aguid_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = IBDiagSMPGUIDInfoTableGetClbck;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    progress_bar_nodes_t progress_bar_nodes;
    CLEAR_STRUCT(progress_bar_nodes);

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node)   //node not exists
            continue;

        p_curr_node->appData1.val = 0;

        ++progress_bar_nodes.nodes_found;
        if (p_curr_node->type == IB_SW_NODE)
            ++progress_bar_nodes.sw_found;
        else
            ++progress_bar_nodes.ca_found;
        if (progress_func)
            progress_func(&progress_bar_nodes, &this->discover_progress_bar_nodes);

        if (p_curr_node->type == IB_SW_NODE) {    // if switch only use port 0
            ports_range_start = 0;
            ports_range_end = 0;
        } else {
            ports_range_start = 1;
            ports_range_end = p_curr_node->numPorts;
        }

        for (unsigned int i = ports_range_start; i <= ports_range_end; ++i) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || (i && !p_curr_port->is_data_worthy()))
                continue;

            // get port info
            SMP_PortInfo *p_port_info =
                    this->fabric_extended_info.getSMPPortInfo(p_curr_port->createIndex);
            if (!p_port_info)
                continue;

            // read GUIDs
            u_int8_t guid_cap    = p_port_info->GUIDCap;
            int num_of_blocks    =
                    (guid_cap + IBIS_IB_MAD_SMP_GUIDS_TABLE_NUM_BLOCKS - 1) /
                    IBIS_IB_MAD_SMP_GUIDS_TABLE_NUM_BLOCKS;
            u_int16_t lid = p_curr_port->base_lid;

            for (int blk = 0; blk < num_of_blocks; ++blk) {
                clbck_data.m_data1 = p_curr_port;
                clbck_data.m_data2 = (void *)(intptr_t)blk;
                this->ibis_obj.SMPGUIDInfoTableGetByLid(lid,
                        blk, &guid_info, &clbck_data);
                /*
                if (this->ibis_obj.SMPGUIDInfoTableGetByLid(lid,
                        blk, &guid_info)) {
                    FabricErrNodeNotRespond *p_curr_fabric_guid_err =
                            new FabricErrNodeNotRespond(p_curr_node, "SMPGUIDInfoTableGetByLid");
                    if (!p_curr_fabric_guid_err) {
                        this->SetLastError("Failed to allocate FabricErrNodeNotRespond");
                        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
                    }
                    aguid_errors.push_back(p_curr_fabric_guid_err);
                    rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
                    continue;
                }
                int rc2 = this->fabric_extended_info.addSMPGUIDInfo(p_curr_port, guid_info, blk);
                if (rc2) {
                    this->SetLastError("Failed to add SMPGuidInfo for port=%s, err=%s",
                            p_curr_port->getName().c_str(), this->fabric_extended_info.GetLastError());
                    IBDIAG_RETURN(rc2);
                }
                */

                if(ibDiagClbck.GetState())
                    goto exit;
                if (p_curr_node->appData1.val != 0)
                    continue;
            }    //block loop
        }  //port loop
    }  //node loop

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!aguid_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_NO_MEM / ERR_CODE_CHECK_FAILED
int IBDiag::CheckDuplicatedAliasGuids(list_p_fabric_general_err& aguid_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    phys_port_t ports_range_start, ports_range_end;
    vec_guids vec_guids;
    //init the port dictionary with the ports from the discovery
    //then add only alias guids to it
    this->port_aguids = this->discovered_fabric.PortByGuid;
    map_guid_pnode::iterator guid2NodeEnd = this->discovered_fabric.NodeByGuid.end();
    map_guid_pnode::iterator sysguid2NodeEnd = this->discovered_fabric.NodeBySystemGuid.end();
    u_int64_t first_alias_guid;
    uint64_t port_guid;
    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node)   //node not exists
            continue;

        if (p_curr_node->type == IB_SW_NODE) {
            ports_range_start = 0;
            ports_range_end   = 0;
        } else {
            ports_range_start = 1;
            ports_range_end   = p_curr_node->numPorts;
        }

        for (unsigned int i = ports_range_start; i <= ports_range_end; ++i) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || (i && !p_curr_port->is_data_worthy()))
                continue;

            // get port info
            SMP_PortInfo *p_port_info =
                    this->fabric_extended_info.getSMPPortInfo(p_curr_port->createIndex);
            if (!p_port_info)
                continue;

            // read port alias GUIDs
            u_int8_t guid_cap = p_port_info->GUIDCap;
            readPortGUIDsToVec(&this->fabric_extended_info,
                    p_curr_port, guid_cap, vec_guids);

            vec_guids::iterator first_guid_it = vec_guids.begin();
            //the table is empty, nothing to do then
            if (first_guid_it == vec_guids.end())
                continue;

            first_alias_guid = *first_guid_it;
            port_guid = p_curr_port->guid_get();
            if (first_alias_guid != port_guid) {
                FabricErrAGUIDInvalidFirstEntry *p_err = new FabricErrAGUIDInvalidFirstEntry(p_curr_port, first_alias_guid);
                if (!p_err) {
                    this->SetLastError("Failed to allocate FabricErrAGUIDInvalidFirstEntry");
                    IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
                }
                aguid_errors.push_back(p_err);
                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            }

            // check duplicated aguids, starting from second guid
            for (vec_guids::iterator it = ++first_guid_it;
                    it != vec_guids.end(); ++it) {
                u_int64_t guid = *it;
                if (!guid)
                    continue;

                FabricErrAGUID *p_error;
                map_guid_pport::iterator findPort = this->port_aguids.find(guid);
                if (findPort != this->port_aguids.end()) {
                    p_error = new FabricErrAGUIDPortGuidDuplicated(
                            p_curr_port,
                            findPort->second,
                            guid);
                    if (!p_error) {
                        this->SetLastError("Failed to allocate FabricErrAGUIDPortGuidDuplicated");
                        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
                    }
                    aguid_errors.push_back(p_error);
                    rc = IBDIAG_ERR_CODE_CHECK_FAILED;
                } else {
                    //add this port guid to maps, only if didn't exist before
                    //in order to print the first port details that has this guid
                    this->port_aguids.insert(pair_guid_port(*it, p_curr_port));
                    this->discovered_fabric.PortByAGuid.insert(pair_guid_port(*it, p_curr_port));
                }

                map_guid_pnode::iterator findSys = this->discovered_fabric.NodeBySystemGuid.find(guid);
                if (findSys != sysguid2NodeEnd) {
                    p_error = new FabricErrAGUIDSysGuidDuplicated(
                            p_curr_port,
                            findSys->second->p_system,
                            guid);
                    if (!p_error) {
                        this->SetLastError("Failed to allocate FabricErrAGUIDSysGuidDuplicated");
                        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
                    }
                    aguid_errors.push_back(p_error);
                    rc = IBDIAG_ERR_CODE_CHECK_FAILED;
                }
                map_guid_pnode::iterator findNode = this->discovered_fabric.NodeByGuid.find(guid);
                if (findNode != guid2NodeEnd) {
                    p_error = new FabricErrAGUIDNodeGuidDuplicated(
                            p_curr_port,
                            findNode->second,
                            guid);
                    if (!p_error) {
                        this->SetLastError("Failed to allocate FabricErrAGUIDNodeGuidDuplicated");
                        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
                    }
                    aguid_errors.push_back(p_error);
                    rc = IBDIAG_ERR_CODE_CHECK_FAILED;
                }
            }   //alias guids loop
        }   //ports loop
    }   //nodes loop

    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR
int IBDiag::WriteAliasGUIDFile(const char *file_name)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);
    ofstream sout;
    int rc = this->OpenFile(file_name, sout,
                            false, //no append
                            true); //add_header
    if (rc)
        IBDIAG_RETURN(rc);
    this->DumpAliasGUID(sout);
    sout.close();
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::DumpAliasGUIDCSVTable(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    char line[2096];
    phys_port_t ports_range_start, ports_range_end;
    vec_guids vec_guids;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    csv_out.DumpStart(SECTION_AGUID);

    stringstream sstream;
    sstream << "NodeGUID," << "PortGUID," "PortNum," << "Index," << "AGUID" << endl;
    csv_out.WriteLine(sstream.str());

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node)   //node not exists
            continue;

        if (p_curr_node->type == IB_SW_NODE) {
            ports_range_start = 0;
            ports_range_end   = 0;
        } else {
            ports_range_start = 1;
            ports_range_end   = p_curr_node->numPorts;
        }

        for (unsigned int i = ports_range_start; i <= ports_range_end; ++i) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            if (!p_curr_port)
                continue;

            // get port info
            SMP_PortInfo *p_port_info =
                    this->fabric_extended_info.getSMPPortInfo(p_curr_port->createIndex);
            if (!p_port_info)
                continue;

            // read port alias GUIDs
            u_int8_t guid_cap = p_port_info->GUIDCap;
            readPortGUIDsToVec(&this->fabric_extended_info,
                    p_curr_port, guid_cap, vec_guids);

            memset(line, 0, sizeof(line));
            for (u_int32_t idx = 0; idx < vec_guids.size(); ++idx) {
                u_int64_t guid = vec_guids[idx];
                if (!guid)
                    continue;

                sstream.str("");
                sprintf(line, U64H_FMT "," U64H_FMT ",%d,%d," U64H_FMT,
                        p_curr_node->guid_get(),
                        p_curr_port->guid_get(),
                        i,
                        idx,
                        guid);
                sstream << line << endl;
                csv_out.WriteLine(sstream.str());
            }
        }
    }

    csv_out.DumpEnd(SECTION_AGUID);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}
