/*
 *  Copyright (c) 2004-2013 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <infiniband/ibdiag/ibdiag.h>


/****************************************************/
/******************private methods*******************/
/****************************************************/
#define CHECK_CA_GUID_STR "Duplicated CA node GUID check - "
#define CHECK_SW_GUID_STR "Duplicated SW node GUID check - "
#define CHECK_PORT_GUID_STR "Duplicated port GUID check - "

static const string whitespaces =  " \t\f\v\n\r";
static void trim_last_whitespaces(string & desc)
{
    int found = (int)desc.find_last_not_of(whitespaces);
    if (found != (int)string::npos)
        desc.erase(found + 1);
}

void IBDiag::AddDupGUIDDetectError(IN direct_route_t *p_direct_route_checked_node,
        IN u_int64_t checked_node_guid,
        IN u_int8_t checked_node_type,
        IN direct_route_t *p_direct_route_got_err,
        IN bool no_response_err,
        IN bool max_hops_err,
        IN string err_desc)
{
    IBDIAG_ENTER;
    char buffer[512];
    memset(buffer, 0, sizeof(buffer));
    if (no_response_err) {
        sprintf(buffer, "no response in DR=%s while %s",
                this->ibis_obj.ConvertDirPathToStr(p_direct_route_got_err).c_str(),
                err_desc.c_str());
    }
    if (max_hops_err) {
        sprintf(buffer, "exceeds maximum hops in DR=%s + DR=%s while %s",
                this->ibis_obj.ConvertDirPathToStr(p_direct_route_checked_node).c_str(),
                this->ibis_obj.ConvertDirPathToStr(p_direct_route_got_err).c_str(),
                err_desc.c_str());
    }
    sprintf(buffer, "%s", err_desc.c_str());

    char buffer2[1024];
    memset(buffer2, 0, sizeof(buffer2));
    sprintf(buffer2, "Can not check Node in DR=%s(Type=%s, GUID=" U64H_FMT ") for duplicated GUID because %s",
            this->ibis_obj.ConvertDirPathToStr(p_direct_route_checked_node).c_str(),
            nodetype2char((IBNodeType)checked_node_type), checked_node_guid, buffer);
    this->duplicated_guids_detection_errs.push_back(string(buffer2));
    IBDIAG_RETURN_VOID;
}



// Returns: SUCCESS_CODE / ERR_CODE_DB_ERR / ERR_CODE_DUPLICATED_GUID / ERR_CODE_FABRIC_ERROR
int IBDiag::CheckIfSameCADevice(IN direct_route_t *p_new_direct_route,
        IN direct_route_t *p_old_direct_route,
        IN struct SMP_NodeInfo *p_new_node_info,
        OUT IbdiagBadDirectRoute_t *p_bad_direct_route_info)
{
    IBDIAG_ENTER;
    //get direct route to prev neighbor from old direct route
    direct_route_t direct_route_to_prev_neigbor_node;
    direct_route_to_prev_neigbor_node = *p_old_direct_route;
    --direct_route_to_prev_neigbor_node.length;
    IBDIAG_LOG(TT_LOG_LEVEL_DISCOVER, CHECK_CA_GUID_STR "Remote node direct route = %s\n",
            this->ibis_obj.ConvertDirPathToStr(&direct_route_to_prev_neigbor_node).c_str());

    //if user map file (node guid to name) was given,
    //refer to original node description, instead of one provided by user
    string node_desc = string("");

    //Stage 1 - we reached CA with same guid as the CA we started to run from
    if (!direct_route_to_prev_neigbor_node.length) {
        IBDIAG_LOG(TT_LOG_LEVEL_DISCOVER, CHECK_CA_GUID_STR "we reached CA with same GUID as the CA we started to run discovery from\n");
        //Stage 1.1 - check entry port
        if (this->root_port_num == p_new_node_info->LocalPortNum) {
            //we returned to the same port we got out ==> can't be
            this->SetLastError(CHECK_CA_GUID_STR "duplicated node GUID was found (reached CA with same GUID as the CA we started to run discovery from but through same entry port as starting port ==> can not be)");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DUPLICATED_GUID);
        }

        //Stage 1.2 - check system guid
        if (!this->root_node || !this->root_node->p_system) {
            this->SetLastError(CHECK_CA_GUID_STR "DB error - root node error");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }
        if (this->root_node->system_guid_get() != p_new_node_info->SystemImageGUID) {
            this->SetLastError(CHECK_CA_GUID_STR "duplicated node GUID was found (reached CA with same node GUID as the CA we started to run discovery from but different system GUID: "   \
                    "old system GUID=" U64H_FMT ", new system GUID= " U64H_FMT ")", this->root_node->system_guid_get(), p_new_node_info->SystemImageGUID);
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DUPLICATED_GUID);
        }

        //Stage 1.3 - try to compare node description
        struct SMP_NodeDesc curr_node_desc;
        if (this->ibis_obj.SMPNodeDescMadGetByDirect(p_new_direct_route, &curr_node_desc)) {
            /*this->SetLastError(CHECK_CA_GUID_STR "Failed to get node description for direct route %s, err=%s",
                    this->ibis_obj.ConvertDirPathToStr(p_new_direct_route).c_str(),
                    this->ibis_obj.GetLastError());*/
            p_bad_direct_route_info->reason = IBDIAG_BAD_DR_NODE_NODE_DESCRIPTION;
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, CHECK_CA_GUID_STR "Failed to get node description for direct route %s, err=%s\n",
                    this->ibis_obj.ConvertDirPathToStr(p_new_direct_route).c_str(),
                    this->ibis_obj.GetLastError());
            this->AddDupGUIDDetectError(p_new_direct_route, p_new_node_info->NodeGUID,
                    p_new_node_info->NodeType, p_new_direct_route,
                    true, false, "try to compare node description");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
        }
        string desc((char *)curr_node_desc.Byte);
        trim_last_whitespaces(desc);

        node_desc = this->root_node->getOrigNodeDescription();

        if (node_desc != desc) {
            this->SetLastError(CHECK_CA_GUID_STR "duplicated node GUID was found (reached CA with same node GUID as the CA we started to run discovery from but different node description: "   \
                    "old node description=%s, new node description=%s)", node_desc.c_str(), (char *)curr_node_desc.Byte);
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DUPLICATED_GUID);
        }

        //Stage 1.4 - nothing more to do ==> therefore we will assume everything is good
        string err_str = "reached CA with same node / system GUID / node description as the CA we started to run discovery from "  \
                "but different entry point ==> nothing to do, assuming no duplicated GUID\n";
        /*this->AddDupGUIDDetectError(p_new_direct_route, p_new_node_info->NodeGUID,
                p_new_node_info->NodeType, p_new_direct_route,
                false, false, err_str);*/
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, CHECK_CA_GUID_STR "%s", err_str.c_str());
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
    }

    //Stage 2 - we reached CA with known guid (but not guid of CA we started to run from)
    //get previous neighbor node
    IBNode *p_prev_neighbor_node = this->GetNodeByDirectRoute(&direct_route_to_prev_neigbor_node);
    if (!p_prev_neighbor_node) {
        this->SetLastError(CHECK_CA_GUID_STR "DB error - can't find node for direct route = %s",
                this->ibis_obj.ConvertDirPathToStr(&direct_route_to_prev_neigbor_node).c_str());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }
    //get previous neighbor port
    u_int8_t prev_neighbor_port_index = p_old_direct_route->path.BYTE[p_old_direct_route->length - 1];
    IBPort *p_prev_neighbor_port = p_prev_neighbor_node->getPort(prev_neighbor_port_index);
    if (!p_prev_neighbor_port) {
        this->SetLastError(CHECK_CA_GUID_STR "DB error - can't find port=%u for known node=%s",
                prev_neighbor_port_index, p_prev_neighbor_node->getName().c_str());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }
    if (!p_prev_neighbor_port->p_remotePort ||
            !p_prev_neighbor_port->p_remotePort->p_node ||
            !p_prev_neighbor_port->p_remotePort->p_node->p_system) {
        this->SetLastError(CHECK_CA_GUID_STR "DB error - previous neighbor port error, port=%s",
                p_prev_neighbor_port->getName().c_str());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    //Stage 2.1 - check same system guid for this CA
    if (p_prev_neighbor_port->p_remotePort->p_node->system_guid_get() != p_new_node_info->SystemImageGUID) {
        this->SetLastError(CHECK_CA_GUID_STR "duplicated node GUID was found (reached CA with already known node GUID but different system GUID: "   \
                "old system GUID=" U64H_FMT ", new system GUID= " U64H_FMT ")",
                p_prev_neighbor_port->p_remotePort->p_node->system_guid_get(), p_new_node_info->SystemImageGUID);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DUPLICATED_GUID);
    }

    //Stage 2.2 - try to compare node description
    struct SMP_NodeDesc curr_node_desc;
    if (this->ibis_obj.SMPNodeDescMadGetByDirect(p_new_direct_route, &curr_node_desc)) {
        /*this->SetLastError(CHECK_CA_GUID_STR "Failed to get node description for direct route %s, err=%s",
                this->ibis_obj.ConvertDirPathToStr(p_new_direct_route).c_str(),
                this->ibis_obj.GetLastError());*/
        p_bad_direct_route_info->reason = IBDIAG_BAD_DR_NODE_NODE_DESCRIPTION;
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, CHECK_CA_GUID_STR "Failed to get node description for direct route %s, err=%s\n",
                this->ibis_obj.ConvertDirPathToStr(p_new_direct_route).c_str(),
                this->ibis_obj.GetLastError());
        this->AddDupGUIDDetectError(p_new_direct_route, p_new_node_info->NodeGUID,
                p_new_node_info->NodeType, p_new_direct_route,
                true, false, "try to compare node description");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
    }

    string desc((char *)curr_node_desc.Byte);
    trim_last_whitespaces(desc);

    node_desc =
        p_prev_neighbor_port->p_remotePort->p_node->getOrigNodeDescription();

    if (node_desc != desc) {
        this->SetLastError(CHECK_CA_GUID_STR "duplicated node GUID was found (reached CA with already known node GUID but different node description: "   \
                "old node description=%s, new node description=%s)", node_desc.c_str(), (char *)curr_node_desc.Byte);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DUPLICATED_GUID);
    }

    //Stage 2.3 - if reached CA with same node guid but different entry point - in this case we can't
    //do nothing ==> therefore we will assume everything is good
    if (p_prev_neighbor_port->p_remotePort->num != p_new_node_info->LocalPortNum) {
        string err_str = "reached CA with already known node / system GUID / node description "  \
                "but different entry point ==> nothing to do, assuming no duplicated GUID\n";
        /*this->AddDupGUIDDetectError(p_new_direct_route, p_new_node_info->NodeGUID,
                p_new_node_info->NodeType, p_new_direct_route,
                false, false, err_str);*/
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, CHECK_CA_GUID_STR "%s", err_str.c_str());
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
    }

    //Stage 2.4 - check previous neighbor port
    if (prev_neighbor_port_index != p_new_direct_route->path.BYTE[p_new_direct_route->length - 1]) {
        this->SetLastError(CHECK_CA_GUID_STR "duplicated node GUID was found (reached CA with already known node GUID but through previous different port index: "  \
                "old previous port index=%u, new previous port index=%u)",
                prev_neighbor_port_index, p_new_direct_route->path.BYTE[p_new_direct_route->length - 1]);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DUPLICATED_GUID);
    }
    //Stage 2.5 - send node info to previous node and check previous node guid
    SMP_NodeInfo curr_node_info;
    direct_route_t previous_new_direct_route = *p_new_direct_route;
    --previous_new_direct_route.length;
    if (this->ibis_obj.SMPNodeInfoMadGetByDirect(&previous_new_direct_route, &curr_node_info)) {
        /*this->SetLastError(CHECK_CA_GUID_STR "Failed to get node info for direct route %s, err=%s",
                this->ibis_obj.ConvertDirPathToStr(&previous_new_direct_route).c_str(),
                this->ibis_obj.GetLastError());*/
        p_bad_direct_route_info->reason = IBDIAG_BAD_DR_NODE_NODE_INFO;
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, CHECK_CA_GUID_STR "Failed to get node info for direct route %s, err=%s\n",
                this->ibis_obj.ConvertDirPathToStr(&previous_new_direct_route).c_str(),
                this->ibis_obj.GetLastError());
        this->AddDupGUIDDetectError(p_new_direct_route, p_new_node_info->NodeGUID,
                p_new_node_info->NodeType, &previous_new_direct_route,
                true, false, "send node info to previous node and check previous node guid");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
    }
    if (p_prev_neighbor_node->guid_get() != curr_node_info.NodeGUID) {
        this->SetLastError(CHECK_CA_GUID_STR "duplicated node GUID was found (reached CA with already known node GUID but through different neighbor: " \
                "old previous neighbor GUID=" U64H_FMT ", new previous neighbor GUID=" U64H_FMT ")",
                p_prev_neighbor_node->guid_get(), curr_node_info.NodeGUID);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DUPLICATED_GUID);
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


// Returns: SUCCESS_CODE / ERR_CODE_DB_ERR / ERR_CODE_DUPLICATED_GUID / ERR_CODE_FABRIC_ERROR / ERR_CODE_EXCEEDS_MAX_HOPS
int IBDiag::CheckIfSameSWDevice(IN direct_route_t *p_new_direct_route,
        IN direct_route_t *p_old_direct_route,
        IN struct SMP_NodeInfo *p_new_node_info,
        OUT IbdiagBadDirectRoute_t *p_bad_direct_route_info)
{
    IBDIAG_ENTER;
    //retrieve reverse direct route of the old route
    direct_route_t rev_old_direct_route;
    int rc = this->GetReverseDirectRoute(&rev_old_direct_route, p_old_direct_route);
    if (rc)
        IBDIAG_RETURN(rc);

    //if user map file (node guid to name) was given,
    //refer to original node description, instead of one provided by user
    string node_desc = string("");

    //Stage 1 - we reached SW with same guid as the SW we started to run from
    if (!rev_old_direct_route.length) {
        //Stage 1.1 - try to compare node description
        struct SMP_NodeDesc curr_node_desc;
        if (this->ibis_obj.SMPNodeDescMadGetByDirect(p_new_direct_route, &curr_node_desc)) {
            /*this->SetLastError(CHECK_SW_GUID_STR "Failed to get node description for direct route %s, err=%s",
                    this->ibis_obj.ConvertDirPathToStr(p_new_direct_route).c_str(),
                    this->ibis_obj.GetLastError());*/
            p_bad_direct_route_info->reason = IBDIAG_BAD_DR_NODE_NODE_DESCRIPTION;
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, CHECK_SW_GUID_STR "Failed to get node description for direct route %s, err=%s\n",
                    this->ibis_obj.ConvertDirPathToStr(p_new_direct_route).c_str(),
                    this->ibis_obj.GetLastError());
            this->AddDupGUIDDetectError(p_new_direct_route, p_new_node_info->NodeGUID,
                    p_new_node_info->NodeType, p_new_direct_route,
                    true, false, "try to compare node description");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
        }
        
        string desc((char *)curr_node_desc.Byte);
        trim_last_whitespaces(desc);
        node_desc = this->root_node->getOrigNodeDescription();

        if (node_desc != desc) {
            this->SetLastError(CHECK_SW_GUID_STR "duplicated node GUID was found (reached SW with same node GUID as starting discovery from but different node description: "   \
                    "old node description=%s, new node description=%s)", node_desc.c_str(), (char *)curr_node_desc.Byte);
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DUPLICATED_GUID);
        }

        //Stage 1.2 - nothing more to do ==> therefore we will assume everything is good
        string err_str = "reached SW with same node GUID + same node description as starting discovery from "   \
                "==> nothing to do, assuming no duplicated GUID\n";
        this->AddDupGUIDDetectError(p_new_direct_route, p_new_node_info->NodeGUID,
                p_new_node_info->NodeType, p_new_direct_route,
                false, false, err_str);
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, CHECK_SW_GUID_STR "%s", err_str.c_str());
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
    }

    //Stage 2 - we reached SW with known guid (but not guid of SW we started to run from)
    //get previous node with this node guid
    IBNode *p_old_node = this->GetNodeByDirectRoute(p_old_direct_route);
    if (!p_old_node) {
        this->SetLastError(CHECK_SW_GUID_STR "DB error - can't find node for direct route = %s", this->ibis_obj.ConvertDirPathToStr(p_old_direct_route).c_str());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    //Stage 2.1 - try to compare node description
    struct SMP_NodeDesc curr_node_desc;
    if (this->ibis_obj.SMPNodeDescMadGetByDirect(p_new_direct_route, &curr_node_desc)) {
        /*this->SetLastError(CHECK_SW_GUID_STR "Failed to get node description for direct route %s, err=%s",
                this->ibis_obj.ConvertDirPathToStr(p_new_direct_route).c_str(),
                this->ibis_obj.GetLastError());*/
        p_bad_direct_route_info->reason = IBDIAG_BAD_DR_NODE_NODE_DESCRIPTION;
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, CHECK_SW_GUID_STR "Failed to get node description for direct route %s, err=%s\n",
                this->ibis_obj.ConvertDirPathToStr(p_new_direct_route).c_str(),
                this->ibis_obj.GetLastError());
        this->AddDupGUIDDetectError(p_new_direct_route, p_new_node_info->NodeGUID,
                p_new_node_info->NodeType, p_new_direct_route,
                true, false, "try to compare node description");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
    }

    string desc((char *)curr_node_desc.Byte);
    trim_last_whitespaces(desc);

    node_desc = p_old_node->getOrigNodeDescription();

    if (node_desc != desc) {
        this->SetLastError(CHECK_SW_GUID_STR "duplicated node GUID was found (reached SW with already known node GUID but different node description: "   \
                "old node description=%s, new node description=%s)", node_desc.c_str(), (char *)curr_node_desc.Byte);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DUPLICATED_GUID);
    }

    //concatenate new direct route with the reverse direct route (of the old direct route)
    direct_route_t total_rev_direct_route;
    rc = this->ConcatDirectRoutes(p_new_direct_route, &rev_old_direct_route, &total_rev_direct_route);
    if (rc) {
        if (rc == IBDIAG_ERR_CODE_EXCEEDS_MAX_HOPS) {
            this->AddDupGUIDDetectError(p_new_direct_route, p_new_node_info->NodeGUID,
                    p_new_node_info->NodeType, &rev_old_direct_route,
                    false, true, "concatenate new direct route with the reverse direct route");
        }
        IBDIAG_RETURN(rc);
    } else
    IBDIAG_LOG(TT_LOG_LEVEL_DISCOVER, CHECK_SW_GUID_STR "Reverse old direct route = %s\n"   \
            CHECK_SW_GUID_STR "Concat direct route = %s\n",
            this->ibis_obj.ConvertDirPathToStr(&rev_old_direct_route).c_str(),
            this->ibis_obj.ConvertDirPathToStr(&total_rev_direct_route).c_str());

    struct SMP_NodeInfo curr_node_info;
    struct SMP_PortInfo curr_port_info;
    phys_port_t port_num;
    for (int i = 1;  i <= rev_old_direct_route.length; ++i) {
        //Stage 2.2 - send port info to last port and check active
        total_rev_direct_route.length = (u_int8_t)(p_new_direct_route->length + i - 1);
        port_num = total_rev_direct_route.path.BYTE[total_rev_direct_route.length];
        IBDIAG_LOG(TT_LOG_LEVEL_DISCOVER, CHECK_SW_GUID_STR "Sending SMPPortInfoMad to DR=%s to port=%u\n",
                this->ibis_obj.ConvertDirPathToStr(&total_rev_direct_route).c_str(), port_num);
        if (this->ibis_obj.SMPPortInfoMadGetByDirect(&total_rev_direct_route, port_num, &curr_port_info)) {
            /*this->SetLastError(CHECK_SW_GUID_STR "Failed to get port info for direct route %s (port=%u), err=%s",
                    this->ibis_obj.ConvertDirPathToStr(&total_rev_direct_route).c_str(),
                    port_num,
                    this->ibis_obj.GetLastError());*/
            p_bad_direct_route_info->reason = IBDIAG_BAD_DR_PORT_PORT_INFO;
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, CHECK_SW_GUID_STR "Failed to get port info for direct route %s (port=%u), err=%s\n",
                    this->ibis_obj.ConvertDirPathToStr(&total_rev_direct_route).c_str(),
                    port_num,
                    this->ibis_obj.GetLastError());
            this->AddDupGUIDDetectError(p_new_direct_route, p_new_node_info->NodeGUID,
                    p_new_node_info->NodeType, &total_rev_direct_route,
                    true, false, "send port info to last port and check active");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
        }
        if (curr_port_info.PortPhyState != IB_PORT_PHYS_STATE_LINK_UP) {  //not active port - all ports should be active
            this->SetLastError(CHECK_SW_GUID_STR "duplicated node GUID was found (reached SW with already known node GUID but found not active port in reverse path algorithm)");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DUPLICATED_GUID);
        }

        //Stage 2.3 - send node info to DR and check entry port same as old DR
        ++total_rev_direct_route.length;
        IBDIAG_LOG(TT_LOG_LEVEL_DISCOVER, CHECK_SW_GUID_STR "Sending SMPNodeInfoMad to DR=%s\n",
                this->ibis_obj.ConvertDirPathToStr(&total_rev_direct_route).c_str());
        if (this->ibis_obj.SMPNodeInfoMadGetByDirect(&total_rev_direct_route, &curr_node_info)) {
            /*this->SetLastError(CHECK_SW_GUID_STR "Failed to get node information for direct route %s, err=%s",
                    this->ibis_obj.ConvertDirPathToStr(&total_rev_direct_route).c_str(),
                    this->ibis_obj.GetLastError());*/
            p_bad_direct_route_info->reason = IBDIAG_BAD_DR_NODE_NODE_INFO;
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, CHECK_SW_GUID_STR "Failed to get node information for direct route %s, err=%s\n",
                    this->ibis_obj.ConvertDirPathToStr(&total_rev_direct_route).c_str(),
                    this->ibis_obj.GetLastError());
            this->AddDupGUIDDetectError(p_new_direct_route, p_new_node_info->NodeGUID,
                    p_new_node_info->NodeType, &total_rev_direct_route,
                    true, false, "send node info to DR and check entry port same as old DR");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
        }
        IBDIAG_LOG(TT_LOG_LEVEL_DISCOVER, CHECK_SW_GUID_STR "Expected entry port=%u, Received entry port=%u\n",
                curr_node_info.LocalPortNum, p_old_direct_route->path.BYTE[p_old_direct_route->length - i]);
        if (curr_node_info.LocalPortNum != p_old_direct_route->path.BYTE[p_old_direct_route->length - i]) {
            this->SetLastError(CHECK_SW_GUID_STR "duplicated node GUID was found (reached SW with already known node GUID but entry ports are not match in reverse path algorithm)");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DUPLICATED_GUID);
        }

        //Stage 2.4 - check guids are the same as in old DR
        direct_route_t direct_route_to_curr_compared_node;
        direct_route_to_curr_compared_node = *p_old_direct_route;
        direct_route_to_curr_compared_node.length = (u_int8_t)(p_old_direct_route->length - i);
        //get compared node
        IBNode *curr_compared_node = this->GetNodeByDirectRoute(&direct_route_to_curr_compared_node);
        if (!curr_compared_node) {
            this->SetLastError(CHECK_SW_GUID_STR "DB error - can't find node for direct route = %s",
                    this->ibis_obj.ConvertDirPathToStr(&direct_route_to_curr_compared_node).c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }
        IBDIAG_LOG(TT_LOG_LEVEL_DISCOVER, CHECK_SW_GUID_STR "Checking expected node at DR=%s\n" \
                "Received GUID=" U64H_FMT " Compared GUID=" U64H_FMT "\n",
                this->ibis_obj.ConvertDirPathToStr(&direct_route_to_curr_compared_node).c_str(),
                curr_node_info.NodeGUID, curr_compared_node->guid_get());
        if (curr_node_info.NodeGUID != curr_compared_node->guid_get()) {
            this->SetLastError(CHECK_SW_GUID_STR "duplicated node GUID was found (reached SW with already known node GUID but compared nodes GUIDs not match in reverse path algorithm)");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DUPLICATED_GUID);
        }
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


// Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::IsDuplicatedGuids(IN direct_route_t *p_new_direct_route,
        IN struct SMP_NodeInfo *p_new_node_info,
        OUT bool *duplicated_node_guid,
        OUT bool *duplicated_port_guid,
        OUT bool *is_visited_node_already,
        OUT bool *is_visited_port_already,
        OUT direct_route_t **p_old_direct_route,
        OUT IbdiagBadDirectRoute_t *p_bad_direct_route_info)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    *duplicated_node_guid = false;
    *duplicated_port_guid = false;
    *is_visited_node_already = false;
    *is_visited_port_already = false;
    *p_old_direct_route = NULL;

    //check for duplicated node guid
    list_p_direct_route old_p_direct_routers_list = this->GetDirectRoutesByNodeGuid(p_new_node_info->NodeGUID);
    if (old_p_direct_routers_list.size() != 0) {    //already known node guid
        for (list_p_direct_route::iterator it = old_p_direct_routers_list.begin();
                it != old_p_direct_routers_list.end(); ++it) {  //check each old route
            direct_route_t *p_curr_old_direct_route = (*it);
            IBDIAG_LOG(TT_LOG_LEVEL_DISCOVER, "\nDuplicated node GUID check - GUID=" U64H_FMT "\n"   \
                    "New direct route = %s\n"   \
                    "Duplicated node GUID check - Old direct route = %s\n",
                    p_new_node_info->NodeGUID,
                    this->ibis_obj.ConvertDirPathToStr(p_new_direct_route).c_str(),
                    this->ibis_obj.ConvertDirPathToStr(p_curr_old_direct_route).c_str());

            IBNode *p_node=GetNodeByDirectRoute(p_curr_old_direct_route);

            if ( !p_node || p_node->toIgnore) {
                if (! p_node) {
                    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                               "Ignore invalid node, direct route = %s\n",
                               ibis_obj.ConvertDirPathToStr(p_curr_old_direct_route).c_str());
                } else {
                    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Ignore invalid node %s GUID="
                           U64H_FMT "direct route = %s\n" ,
                           p_node->name.c_str(),
                           p_new_node_info->NodeGUID,
                           ibis_obj.ConvertDirPathToStr(p_curr_old_direct_route).c_str());
                }
                *p_old_direct_route = p_curr_old_direct_route;
                rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
                continue;
            }

            if (p_new_node_info->NodeType == IB_SW_NODE) {  //switch
                rc = this->CheckIfSameSWDevice(p_new_direct_route, p_curr_old_direct_route, p_new_node_info, p_bad_direct_route_info);
            } else {    //hca device
                rc = this->CheckIfSameCADevice(p_new_direct_route, p_curr_old_direct_route, p_new_node_info, p_bad_direct_route_info);
            }

            if (rc == IBDIAG_SUCCESS_CODE) {       //not duplicated, we found that this node
                                            //is already visited by old DR
                *duplicated_node_guid = false;
                *p_old_direct_route = p_curr_old_direct_route;
                break;
            } else {                        //can be duplicated
                if ( (rc == IBDIAG_ERR_CODE_DUPLICATED_GUID) || \
                        (rc == IBDIAG_ERR_CODE_FABRIC_ERROR) || \
                        (rc == IBDIAG_ERR_CODE_EXCEEDS_MAX_HOPS) ) {
                    if (rc == IBDIAG_ERR_CODE_DUPLICATED_GUID) {   //duplicated for sure
                        *duplicated_node_guid = true;
                        this->ibdiag_discovery_status = DISCOVERY_DUPLICATED_GUIDS;
                    } else {                                //assuming not duplicated and already been
                                                            //visited by old DR. we don't break because
                                                            //there is a chance that it is duplicated
                        *p_old_direct_route = p_curr_old_direct_route;
                    }
                    continue;   //try to continue
                }
                IBDIAG_RETURN(rc);      //stop - real error
            }
        }
        *is_visited_node_already = !*duplicated_node_guid;
    }

    if (rc == IBDIAG_ERR_CODE_FABRIC_ERROR) {
        IBDIAG_RETURN(rc);
    }
    //check for duplicated port guid
    old_p_direct_routers_list = this->GetDirectRoutesByPortGuid(p_new_node_info->PortGUID);
    if (old_p_direct_routers_list.size() != 0) {    //already known port guid
        IBDIAG_LOG(TT_LOG_LEVEL_DISCOVER, CHECK_PORT_GUID_STR "GUID=" U64H_FMT "\n"  \
                "New direct route = %s\n",
                p_new_node_info->PortGUID,
                this->ibis_obj.ConvertDirPathToStr(p_new_direct_route).c_str());

        if (p_new_node_info->NodeType == IB_SW_NODE) {  //switch
            if (!(*is_visited_node_already)) {  //we reached new switch and not same sw from different port
                *duplicated_port_guid = true;
                this->ibdiag_discovery_status = DISCOVERY_DUPLICATED_GUIDS;
            }
        } else {    //hca
            *duplicated_port_guid = true;   //we reach each port only once during discovery
            this->SetLastError(CHECK_PORT_GUID_STR "duplicated port GUID was found");
            this->ibdiag_discovery_status = DISCOVERY_DUPLICATED_GUIDS;
        }
        *is_visited_port_already = !*duplicated_port_guid;
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


/****************************************************/
/*******************public methods*******************/
/****************************************************/
//Returns: SUCCESS_CODE / ERR_CODE_CHECK_FAILED / ERR_CODE_NO_MEM / ERR_CODE_DB_ERR
int IBDiag::CheckDuplicatedGUIDs(list_p_fabric_general_err& guids_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    if (!this->IsDuplicatedGuidsFound())
        IBDIAG_RETURN(rc);

    //collect node guids errors
    for (map_guid_list_p_direct_route::iterator nI = this->bfs_known_node_guids.begin();
            nI != this->bfs_known_node_guids.end(); ++nI) {
        if (nI->second.size() > 1) {
            for (list_p_direct_route::iterator it = nI->second.begin(); it != nI->second.end(); ++it) {
                IBNode *p_node = this->GetNodeByDirectRoute(*it);
                if (!p_node) {
                    this->SetLastError("DB error - failed to get node object for direct route=%s", this->ibis_obj.ConvertDirPathToStr(*it).c_str());
                    IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
                }
                FabricErrDuplicatedNodeGuid *p_curr_fabric_guid_err =
                        new FabricErrDuplicatedNodeGuid(p_node, this->ConvertDirPathToStr(*it), nI->first);
                if (!p_curr_fabric_guid_err) {
                    this->SetLastError("Failed to allocate FabricErrDuplicatedNodeGuid");
                    IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
                }
                guids_errors.push_back(p_curr_fabric_guid_err);
            }
            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
        }
    }
    //collect port guids errors
    for (map_guid_list_p_direct_route::iterator pI = this->bfs_known_port_guids.begin();
            pI != this->bfs_known_port_guids.end(); ++pI) {
        if (pI->second.size() > 1) {
            for (list_p_direct_route::iterator it = pI->second.begin(); it != pI->second.end(); ++it) {
                IBNode *p_node = this->GetNodeByDirectRoute(*it);
                if (!p_node) {
                    this->SetLastError("DB error - failed to get node object for direct route=%s", this->ibis_obj.ConvertDirPathToStr(*it).c_str());
                    IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
                }
                FabricErrDuplicatedPortGuid *p_curr_fabric_guid_err =
                        new FabricErrDuplicatedPortGuid(p_node, this->ConvertDirPathToStr(*it), pI->first);
                if (!p_curr_fabric_guid_err) {
                    this->SetLastError("Failed to allocate FabricErrDuplicatedPortGuid");
                    IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
                }
                guids_errors.push_back(p_curr_fabric_guid_err);
            }
            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
        }
    }
    IBDIAG_RETURN(rc);
}


// Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::PrintNodesDuplicatedGuids()
{
    IBDIAG_ENTER;
    for (map_guid_list_p_direct_route::iterator nI = this->bfs_known_node_guids.begin();
            nI != this->bfs_known_node_guids.end(); ++nI) {
        if (nI->second.size() > 1) {
            printf("\nNode GUID=" U64H_FMT " is duplicated in the following direct routes:\n", nI->first);
            for (list_p_direct_route::iterator it = nI->second.begin(); it != nI->second.end(); ++it) {
                IBNode *p_node = this->GetNodeByDirectRoute(*it);
                if (!p_node) {
                    this->SetLastError("DB error - failed to get node object for direct route=%s", this->ibis_obj.ConvertDirPathToStr(*it).c_str());
                    IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
                }
                printf("    Node = %s, DR = %s\n",
                        p_node->getName().c_str(), this->ibis_obj.ConvertDirPathToStr(*it).c_str());
            }
        }
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


// Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::PrintPortsDuplicatedGuids()
{
    IBDIAG_ENTER;
    for (map_guid_list_p_direct_route::iterator nI = this->bfs_known_port_guids.begin();
            nI != this->bfs_known_port_guids.end(); ++nI) {
        if (nI->second.size() > 1) {
            printf("\nPort GUID=" U64H_FMT " is duplicated in the following direct routes:\n", nI->first);
            for (list_p_direct_route::iterator it = nI->second.begin(); it != nI->second.end(); ++it) {
                IBNode *p_node = this->GetNodeByDirectRoute(*it);
                if (!p_node) {
                    this->SetLastError("DB error - failed to get node object for direct route=%s", this->ibis_obj.ConvertDirPathToStr(*it).c_str());
                    IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
                }
                printf("    Node = %s, DR = %s\n",
                        p_node->getName().c_str(), this->ibis_obj.ConvertDirPathToStr(*it).c_str());
            }
        }
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


void IBDiag::PrintDupGuidsDetectionErrors()
{
    IBDIAG_ENTER;
    printf("Duplicated GUIDs detection errors are:\n");
    for (list_string::iterator it = this->duplicated_guids_detection_errs.begin();
            it != this->duplicated_guids_detection_errs.end(); ++it)
        printf("%s\n", (*it).c_str());
    IBDIAG_RETURN_VOID;
}

