/*
 * Copyright (c) 2017-2018 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <infiniband/ibdiag/ibdiag_fabric.h>
#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h>

int IBDiagFabric::UpdateFabric(const string &csv_file)
{
    INFO_PRINT("Loading the fabric from file ... \n");

    int rc = IBDIAG_SUCCESS_CODE;
    CsvFileStream csv_file_stream(csv_file, m_csv_parser);

    SectionParser<NodeRecord>  node_section_parser;
    node_section_parser.Init("NODES");
    rc = m_csv_parser.ParseAndHandleSection <NodeRecord, IBDiagFabric>(csv_file_stream,
            node_section_parser, this, &IBDiagFabric::CreateNode);
    if (rc) {
        ERR_PRINT("Failed to parse NODES section in csv file\n");
        return rc;
    }

    SectionParser<PortRecord>  port_section_parser;
    port_section_parser.Init("PORTS");
    rc = m_csv_parser.ParseAndHandleSection <PortRecord, IBDiagFabric>(csv_file_stream,
            port_section_parser, this, &IBDiagFabric::CreatePort);
    if (rc) {
        ERR_PRINT("Failed to parse PORTS section in csv file\n");
        return rc;
    }

    SectionParser<SwitchRecord>  switch_section_parser;
    switch_section_parser.Init("SWITCHES");
    rc = m_csv_parser.ParseAndHandleSection <SwitchRecord, IBDiagFabric>(csv_file_stream,
            switch_section_parser, this, &IBDiagFabric::CreateSwitch);
    if (rc) {
        ERR_PRINT("Failed to parse SWITCHES section in csv file\n");
        return rc;
    }

    SectionParser<LinkRecord>  link_section_parser;
    link_section_parser.Init("LINKS");
    rc = m_csv_parser.ParseAndHandleSection <LinkRecord, IBDiagFabric>(csv_file_stream,
            link_section_parser, this, &IBDiagFabric::CreateLink);
    if (rc) {
        ERR_PRINT("Failed to parse LINKS section in csv file\n");
        return rc;
    }

    return rc;
}

int IBDiagFabric::CreateNode(const NodeRecord &nodeRecord)
{
    int rc = 0;

    IBNode* p_node = discovered_fabric.makeNode((IBNodeType)nodeRecord.node_type,
                                                (phys_port_t)nodeRecord.num_ports,
                                                nodeRecord.system_image_guid,
                                                nodeRecord.node_guid,
                                                nodeRecord.vendor_id,
                                                nodeRecord.device_id,
                                                nodeRecord.revision,
                                                nodeRecord.node_description);

    if (!p_node) {
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }

    //take care of progress bar
    ++nodes_found;
    if (nodeRecord.node_type == IB_CA_NODE)
       ++ca_found;
    else
       ++sw_found;
    ports_found += nodeRecord.num_ports;

    struct SMP_NodeInfo node_info;

    node_info.NumPorts = (phys_port_t)nodeRecord.num_ports;
    node_info.NodeType = nodeRecord.node_type;
    node_info.ClassVersion = nodeRecord.class_version;
    node_info.BaseVersion = nodeRecord.base_version;
    node_info.SystemImageGUID = nodeRecord.system_image_guid;
    node_info.NodeGUID = nodeRecord.node_guid;
    node_info.PortGUID = nodeRecord.port_guid;
    node_info.DeviceID = nodeRecord.device_id;
    node_info.PartitionCap = nodeRecord.partition_cap;
    node_info.revision = nodeRecord.revision;
    node_info.VendorID = nodeRecord.vendor_id;
    node_info.LocalPortNum = nodeRecord.local_port_num;

    rc = this->fabric_extended_info.addSMPNodeInfo(p_node, node_info);

    return rc;
}

int IBDiagFabric::CreatePort(const PortRecord &portRecord)
{
    int rc = 0;

    IBNode* p_node = discovered_fabric.NodeByGuid[portRecord.node_guid];
    IBLinkSpeed correct_speed = (IBLinkSpeed)(portRecord.link_speed_actv & 0xff);

    static u_int16_t base_lid; // static - to keep the value for all switch ports
    static u_int8_t lmc;
    static u_int32_t cap_mask;

    if (!(p_node->type == IB_SW_NODE && portRecord.port_num != 0)) {
        base_lid = portRecord.lid;
        lmc = portRecord.lmc;
        cap_mask = portRecord.cap_mask;
    }

    if ((cap_mask & IBIS_IB_PORT_CAP_HAS_EXT_SPEEDS) &&
            ((portRecord.link_speed_actv >> 8) & 0xff)) {
        /* port supports extended speeds and active extended speed
           is not zero ==> we have extended speed */
        correct_speed = extspeed2speed((u_int8_t)((portRecord.link_speed_actv >> 8) & 0xff));
    }

    IBPort *p_port = discovered_fabric.setNodePort(p_node,
                                                   portRecord.port_guid,
                                                   base_lid,
                                                   lmc,
                                                   portRecord.port_num,
                                                   (IBLinkWidth)portRecord.link_width_actv,
                                                   correct_speed,
                                                   (IBPortState)portRecord.port_state);

    if (!p_port)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);

    struct SMP_PortInfo port_info;

    port_info.MKey = portRecord.m_key;
    port_info.GIDPrfx = portRecord.gid_prefix;
    port_info.MSMLID = portRecord.msm_lid;
    port_info.LID = portRecord.lid;
    port_info.CapMsk = portRecord.cap_mask;
    port_info.M_KeyLeasePeriod = portRecord.m_key_lease_period;
    port_info.DiagCode = portRecord.diag_code;
    port_info.LinkWidthActv = portRecord.link_width_actv;
    port_info.LinkWidthSup = portRecord.link_width_sup;
    port_info.LinkWidthEn = portRecord.link_width_en;
    port_info.LocalPortNum = portRecord.local_port_num;
    port_info.LinkSpeedEn = (u_int8_t)(portRecord.link_speed_en & 0xff);
    port_info.LinkSpeedActv = (u_int8_t)(portRecord.link_speed_actv & 0xff);
    port_info.LMC = portRecord.lmc;
    port_info.MKeyProtBits = portRecord.m_key_prot_bits;
    port_info.LinkDownDefState = portRecord.link_down_def_state;
    port_info.PortPhyState = portRecord.port_phy_state;
    port_info.PortState = portRecord.port_state;
    port_info.LinkSpeedSup = (u_int8_t)(portRecord.link_speed_sup & 0xff);
    port_info.VLArbHighCap = portRecord.vl_arbit_High_Cap;
    port_info.VLHighLimit = portRecord.vl_high_limit;
    port_info.InitType = portRecord.init_type;
    port_info.VLCap = portRecord.vl_cap;
    port_info.MSMSL = portRecord.msm_sl;
    port_info.NMTU = portRecord.nmtu;
    port_info.FilterRawOutb = portRecord.filter_raw_outbound;
    port_info.FilterRawInb = portRecord.filter_raw_inbound;
    port_info.PartEnfOutb = portRecord.part_enf_outbound;
    port_info.PartEnfInb = portRecord.part_enf_inbound;
    port_info.OpVLs = portRecord.op_VLs;
    port_info.HoQLife = portRecord.hoq_life;
    port_info.VLStallCnt = portRecord.vl_stall_cnt;
    port_info.MTUCap = portRecord.mtu_cap;
    port_info.InitTypeReply = portRecord.init_type_reply;
    port_info.VLArbLowCap = portRecord.vl_arbit_low_cap;
    port_info.PKeyViolations = portRecord.pkey_violations;
    port_info.MKeyViolations = portRecord.mkey_violations;
    port_info.SubnTmo = portRecord.subn_time_out;
    port_info.ClientReregister = portRecord.client_reregister;
    port_info.GUIDCap = portRecord.guid_cap;
    port_info.QKeyViolations = portRecord.qkey_violations;
    port_info.MaxCreditHint = portRecord.max_credit_hint;
    port_info.OverrunErrs = portRecord.overrun_errs;
    port_info.LocalPhyError = portRecord.local_phy_error;
    port_info.RespTimeValue = portRecord.resp_time_value;
    port_info.LinkRoundTripLatency = portRecord.link_round_trip_latency;
    port_info.LinkSpeedExtEn = (u_int8_t)((portRecord.link_speed_en >> 8) & 0xff);
    port_info.LinkSpeedExtActv = (u_int8_t)((portRecord.link_speed_actv >> 8) & 0xff);
    port_info.LinkSpeedExtSup = (u_int8_t)((portRecord.link_speed_sup >> 8) & 0xff);

    if (portRecord.cap_mask_2 != "N/A")
        port_info.CapMsk2 = (u_int16_t)strtoul(portRecord.cap_mask_2.c_str(), NULL, 10);

    rc = this->fabric_extended_info.addSMPPortInfo(p_port, port_info);

    return rc;
}

int IBDiagFabric::CreateSwitch(const SwitchRecord &switchRecord)
{
    int rc = 0;

    IBNode* p_node = discovered_fabric.NodeByGuid[switchRecord.node_guid];

    struct SMP_SwitchInfo switch_info;

    switch_info.LinearFDBCap = switchRecord.linear_FDB_cap;
    switch_info.RandomFDBCap = switchRecord.random_FDB_cap;
    switch_info.MCastFDBCap = switchRecord.mcast_FDB_cap;
    switch_info.LinearFDBTop = switchRecord.linear_FDB_top;
    switch_info.DefPort = switchRecord.def_port;
    switch_info.DefMCastPriPort = switchRecord.def_mcast_pri_port;
    switch_info.DefMCastNotPriPort = switchRecord.def_mcast_not_pri_port;
    switch_info.LifeTimeValue = switchRecord.life_time_value;
    switch_info.PortStateChange = switchRecord.port_state_change;
    switch_info.OptimizedSLVLMapping = switchRecord.optimized_SLVL_mapping;
    switch_info.LidsPerPort = switchRecord.lids_per_port;
    switch_info.PartEnfCap = switchRecord.part_enf_cap;
    switch_info.InbEnfCap = switchRecord.inb_enf_cap;
    switch_info.OutbEnfCap = switchRecord.outb_enf_cap;
    switch_info.FilterRawInbCap = switchRecord.filter_raw_inb_cap;
    switch_info.FilterRawOutbCap = switchRecord.filter_raw_outb_cap;
    switch_info.ENP0 = switchRecord.en_port0;
    switch_info.MCastFDBTop = switchRecord.mcast_FDB_top;

    rc = fabric_extended_info.addSMPSwitchInfo(p_node, switch_info);

    IBDIAG_RETURN(rc);
}

int IBDiagFabric::CreateLink(const LinkRecord &linkRecord)
{
    int rc = 0;

    discovered_fabric.getNodeByGuid(linkRecord.node_guid1)->
            getPort(linkRecord.port_num1)->p_remotePort =
            discovered_fabric.getNodeByGuid(linkRecord.node_guid2)->
            getPort(linkRecord.port_num2);

    discovered_fabric.getNodeByGuid(linkRecord.node_guid2)->
            getPort(linkRecord.port_num2)->p_remotePort =
            discovered_fabric.getNodeByGuid(linkRecord.node_guid1)->
            getPort(linkRecord.port_num1);

    return rc;
}


/*******************************
 *       SECTION RECORDS       *
 *******************************/

int NodeRecord::Init(vector < ParseFieldInfo <class NodeRecord> > &parse_section_info)
{
    int rc = 0;

    parse_section_info.push_back(
            ParseFieldInfo <class NodeRecord>("NodeDesc", &NodeRecord::SetNodeDescription));

    parse_section_info.push_back(
            ParseFieldInfo <class NodeRecord>("NumPorts", &NodeRecord::SetNumPorts));

    parse_section_info.push_back(
            ParseFieldInfo <class NodeRecord>("NodeType", &NodeRecord::SetNodeType));

    parse_section_info.push_back(
            ParseFieldInfo <class NodeRecord>("ClassVersion", &NodeRecord::SetClassVersion));

    parse_section_info.push_back(
            ParseFieldInfo <class NodeRecord>("BaseVersion", &NodeRecord::SetBaseVersion));

    parse_section_info.push_back(
            ParseFieldInfo <class NodeRecord>("SystemImageGUID", &NodeRecord::SetSystemImageGUID));

    parse_section_info.push_back(
            ParseFieldInfo <class NodeRecord>("NodeGUID", &NodeRecord::SetNodeGUID));

    parse_section_info.push_back(
            ParseFieldInfo <class NodeRecord>("PortGUID", &NodeRecord::SetPortGUID));

    parse_section_info.push_back(
            ParseFieldInfo <class NodeRecord>("DeviceID", &NodeRecord::SetDeviceID));

    parse_section_info.push_back(
            ParseFieldInfo <class NodeRecord>("PartitionCap", &NodeRecord::SetPartitionCap));

    parse_section_info.push_back(
            ParseFieldInfo <class NodeRecord>("revision", &NodeRecord::SetRevision));

    parse_section_info.push_back(
            ParseFieldInfo <class NodeRecord>("VendorID", &NodeRecord::SetVendorID));

    parse_section_info.push_back(
            ParseFieldInfo <class NodeRecord>("LocalPortNum", &NodeRecord::SetLocalPortNum));

   return(rc);
}

int PortRecord::Init(vector < ParseFieldInfo <class PortRecord> > &parse_section_info)
{
    int rc = 0;

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("NodeGuid", &PortRecord::SetNodeGuid));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("PortGuid", &PortRecord::SetPortGuid));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("PortNum", &PortRecord::SetPortNum));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("MKey", &PortRecord::SetMKey));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("GIDPrfx", &PortRecord::SetGIDPrfx));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("MSMLID", &PortRecord::SetMSMLID));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("LID", &PortRecord::SetLid));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("CapMsk", &PortRecord::SetCapMsk));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("M_KeyLeasePeriod", &PortRecord::SetM_KeyLeasePeriod));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("DiagCode", &PortRecord::SetDiagCode));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("LinkWidthActv", &PortRecord::SetLinkWidthActv));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("LinkWidthSup", &PortRecord::SetLinkWidthSup));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("LinkWidthEn", &PortRecord::SetLinkWidthEn));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("LocalPortNum", &PortRecord::SetLocalPortNum));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("LinkSpeedEn", &PortRecord::SetLinkSpeedEn));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("LinkSpeedActv", &PortRecord::SetLinkSpeedActv));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("LMC", &PortRecord::SetLMC));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("MKeyProtBits", &PortRecord::SetMKeyProtBits));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("LinkDownDefState", &PortRecord::SetLinkDownDefState));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("PortPhyState", &PortRecord::SetPortPhyState));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("PortState", &PortRecord::SetPortState));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("LinkSpeedSup", &PortRecord::SetLinkSpeedSup));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("VLArbHighCap", &PortRecord::SetVLArbHighCap));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("VLHighLimit", &PortRecord::SetVLHighLimit));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("InitType", &PortRecord::SetInitType));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("VLCap", &PortRecord::SetVLCap));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("MSMSL", &PortRecord::SetMSMSL));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("NMTU", &PortRecord::SetNMTU));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("FilterRawOutb", &PortRecord::SetFilterRawOutb));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("FilterRawInb", &PortRecord::SetFilterRawInb));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("PartEnfOutb", &PortRecord::SetPartEnfOutb));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("PartEnfInb", &PortRecord::SetPartEnfInb));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("OpVLs", &PortRecord::SetOpVLs));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("HoQLife", &PortRecord::SetHoQLife));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("VLStallCnt", &PortRecord::SetVLStallCnt));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("MTUCap", &PortRecord::SetMTUCap));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("InitTypeReply", &PortRecord::SetInitTypeReply));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("VLArbLowCap", &PortRecord::SetVLArbLowCap));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("PKeyViolations", &PortRecord::SetPKeyViolations));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("MKeyViolations", &PortRecord::SetMKeyViolations));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("SubnTmo", &PortRecord::SetSubnTmo));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("ClientReregister", &PortRecord::SetClientReregister));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("GUIDCap", &PortRecord::SetGUIDCap));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("QKeyViolations", &PortRecord::SetQKeyViolations));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("MaxCreditHint", &PortRecord::SetMaxCreditHint));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("OverrunErrs", &PortRecord::SetOverrunErrs));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("LocalPhyError", &PortRecord::SetLocalPhyError));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("RespTimeValue", &PortRecord::SetRespTimeValue));

    parse_section_info.push_back(ParseFieldInfo <class PortRecord>
            ("LinkRoundTripLatency", &PortRecord::SetLinkRoundTripLatency));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("CapMsk2", &PortRecord::SetCapMsk2));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("FECActv", &PortRecord::SetFECActv));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("RetransActv", &PortRecord::SetRetransActv));

   return(rc);
}

int SwitchRecord::Init(vector < ParseFieldInfo <class SwitchRecord> > &parse_section_info)
{
    int rc = 0;

    parse_section_info.push_back(
                ParseFieldInfo <class SwitchRecord>("NodeGUID", &SwitchRecord::SetNodeGuid));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("LinearFDBCap", &SwitchRecord::SetLinearFDBCap));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("RandomFDBCap", &SwitchRecord::SetRandomFDBCap));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("MCastFDBCap", &SwitchRecord::SetMCastFDBCap));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("LinearFDBTop", &SwitchRecord::SetLinearFDBTop));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("DefPort", &SwitchRecord::SetDefPort));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("DefMCastPriPort", &SwitchRecord::SetDefMCastPriPort));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("DefMCastNotPriPort", &SwitchRecord::SetDefMCastNotPriPort));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("LifeTimeValue", &SwitchRecord::SetLifeTimeValue));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("PortStateChange", &SwitchRecord::SetPortStateChange));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("OptimizedSLVLMapping", &SwitchRecord::SetOptimizedSLVLMapping));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("LidsPerPort", &SwitchRecord::SetLidsPerPort));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("PartEnfCap", &SwitchRecord::SetPartEnfCap));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("InbEnfCap", &SwitchRecord::SetInbEnfCap));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("OutbEnfCap", &SwitchRecord::SetOutbEnfCap));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("FilterRawInbCap", &SwitchRecord::SetFilterRawInbCap));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("FilterRawOutbCap", &SwitchRecord::SetFilterRawOutbCap));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("ENP0", &SwitchRecord::SetENP0));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("MCastFDBTop", &SwitchRecord::SetMCastFDBTop));

   return(rc);
}

int LinkRecord::Init(vector < ParseFieldInfo <class LinkRecord> > &parse_section_info)
{
    int rc = 0;

    parse_section_info.push_back(
            ParseFieldInfo <class LinkRecord>("NodeGuid1", &LinkRecord::SetNodeGuid1));

   parse_section_info.push_back(
            ParseFieldInfo <class LinkRecord>("PortNum1", &LinkRecord::SetPortNum1));

   parse_section_info.push_back(
            ParseFieldInfo <class LinkRecord>("NodeGuid2", &LinkRecord::SetNodeGuid2));

   parse_section_info.push_back(
            ParseFieldInfo <class LinkRecord>("PortNum2", &LinkRecord::SetPortNum2));

   return rc;
}
