/*
 *  Copyright (c) 2004-2019 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <infiniband/ibdiag/ibdiag_fabric_errs.h>


/*****************************************************/
/*****************************************************/
/*****************************************************/
void DumpCSVFabricErrorListTable(list_p_fabric_general_err& list_errors, CSVOut &csv_out,
                                 string name, EnFabricErrLevel_t type)
{
    IBDIAG_ENTER;
    if (list_errors.empty())
        IBDIAG_RETURN_VOID;

    for (unsigned i = 0; i < name.size(); ++i) {
        if (name[i] == ' ') {
            name[i] = '_';
            continue;
        }
        if ((name[i] >= 'a') && (name[i] <= 'z')) {
            name[i] = (char)((name[i] - 'a') + 'A');
        }
    }

    if (type == EN_FABRIC_ERR_WARNING)
        csv_out.DumpStart(string("WARNINGS_" + name).c_str());
    else
        csv_out.DumpStart(string("ERRORS_" + name).c_str());

    stringstream sstream;
    sstream << "Scope," << "NodeGUID," << "PortGUID," << "PortNumber," << "EventName," << "Summary" << endl;
    csv_out.WriteLine(sstream.str());

    for (list_p_fabric_general_err::iterator it = list_errors.begin(); it != list_errors.end(); ++it) {
        sstream.str("");
        sstream << (*it)->GetCSVErrorLine() << endl;
        csv_out.WriteLine(sstream.str());
    }

    if (type == EN_FABRIC_ERR_WARNING)
        csv_out.DumpEnd(string("WARNINGS_" + name).c_str());
    else
        csv_out.DumpEnd(string("ERRORS_" + name).c_str());

    IBDIAG_RETURN_VOID;
}


void CleanFabricErrorsList(list_p_fabric_general_err& list_errors)
{
    IBDIAG_ENTER;
    for (list_p_fabric_general_err::iterator it = list_errors.begin(); it != list_errors.end(); ++it)
        if (*it)
            delete *it;
    list_errors.clear();
    IBDIAG_RETURN_VOID;
}


string DescToCsvDesc(string desc)
{
    IBDIAG_ENTER

    if (desc == "")
        IBDIAG_RETURN("NA");

    string result = desc;
    size_t idx = result.find(',', 0);
    while (idx != string::npos) {
        result[idx] = '-';
        idx = result.find(',', idx + 1);
    }

    IBDIAG_RETURN(result);
}


/*****************************************************/
/*****************************************************/
/*****************************************************/
string FabricErrDuplicatedNodeGuid::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[1024];
    sprintf(buffer, STR_FMT "," U64H_FMT "," U8H_FMT "," U8H_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->duplicated_guid,
            0x0,            //NA
            0x0,            //NA
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());
    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

FabricErrDuplicatedNodeGuid::FabricErrDuplicatedNodeGuid(IBNode *p_node, string direct_route_to_node_str, u_int64_t duplicated_guid) :
        FabricErrGuid(p_node, direct_route_to_node_str, duplicated_guid)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_CLUSTER);
    this->err_desc.assign(FER_DUPLICATED_NODE_GUID);
    char buffer[1024];
    sprintf(buffer, "Node GUID = " U64H_FMT " is duplicated at: ", this->duplicated_guid);
    this->description.assign(buffer);
    this->description += "Node = ";
    this->description += this->p_node->getName();
    if (this->p_node->orig_description.length()) {
        this->description += " (";
        this->description += this->p_node->orig_description;
        this->description += ")";
    }
    this->description += ", DR = ";
    this->description += this->direct_route_to_node_str;
    IBDIAG_RETURN_VOID;
}

string FabricErrDuplicatedPortGuid::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[1024];
    memset(buffer, 0, sizeof(buffer));
    sprintf(buffer, STR_FMT "," U64H_FMT "," U64H_FMT "," U8H_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->p_node->guid_get(),
            this->duplicated_guid,
            0x0,            //NA
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());
    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

FabricErrDuplicatedPortGuid::FabricErrDuplicatedPortGuid(IBNode *p_node, string direct_route_to_node_str, u_int64_t duplicated_guid) :
        FabricErrGuid(p_node, direct_route_to_node_str, duplicated_guid)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_CLUSTER);
    this->err_desc.assign(FER_DUPLICATED_PORT_GUID);
    char buffer[1024];
    sprintf(buffer, "Port GUID = " U64H_FMT " is duplicated at: ", this->duplicated_guid);
    this->description.assign(buffer);
    this->description += "Node = ";
    this->description += this->p_node->getName();
    this->description += ", DR = ";
    this->description += this->direct_route_to_node_str;
    IBDIAG_RETURN_VOID;
}


/*****************************************************/
/*****************************************************/
/*****************************************************/
string FabricErrNode::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[1024];
    sprintf(buffer, STR_FMT "," U64H_FMT "," U8H_FMT "," U8H_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->p_node->guid_get(),
            0x0,            //NA
            0x0,            //NA
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());
    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrNode::GetErrorLine()
{
    IBDIAG_ENTER;
    string line;
    line = this->p_node->getName();
    line += " - ";
    line += this->description;
    IBDIAG_RETURN(line);
}


FabricErrNodeNotRespond::FabricErrNodeNotRespond(IBNode *p_node, string desc) : FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_NODE_NOT_RESPOND);
    this->description = "No response for MAD";
    if (desc != "") {
        this->description += " ";
        this->description += desc;
    }
    IBDIAG_RETURN_VOID;
}


FabricErrNodeWrongConfig::FabricErrNodeWrongConfig(IBNode *p_node, string desc) : FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_NODE_WRONG_CONFIG);
    this->description = "Node is configured with wrong value";
    if (desc != "") {
        this->description += " ";
        this->description += desc;
    }
    IBDIAG_RETURN_VOID;
}


FabricErrNodeDuplicatedNodeDesc::FabricErrNodeDuplicatedNodeDesc(IBNode *p_node) : FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_CLUSTER);
    this->err_desc.assign(FER_NODE_DUPLICATED_NODE_DESC);
    char buffer[1024];
    sprintf(buffer, "Node with GUID=" U64H_FMT " is configured with duplicated node description - %s",
            this->p_node->guid_get(),
            this->p_node->description.c_str());
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}


FabricErrNodeMlnxCountersPageVer::FabricErrNodeMlnxCountersPageVer(
        IBNode *p_node,
        unsigned int page,
        unsigned int curr_ver,
        unsigned int latest_ver) :
        FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_MLNX_CNTRS_WRONG_PAGE_VER);
    char buffer[1024];
    sprintf(buffer,
            "Node has Mlnx Counters Page" U32D_FMT " version " U32D_FMT   \
            ", while ibutils supports version " U32D_FMT,
            page, curr_ver, latest_ver);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}


FabricErrNodeWrongFWVer::FabricErrNodeWrongFWVer(IBNode *p_node, string desc) : FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_NODE_WRONG_FW_VER);
    this->description = desc;
    IBDIAG_RETURN_VOID;
}


FabricErrNodeNotSupportCap::FabricErrNodeNotSupportCap(IBNode *p_node, string desc) : FabricErrNode(p_node)
{
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_NODE_NOT_SUPPORT_CAP);
    this->description = desc;
    IBDIAG_RETURN_VOID;

}

FabricErrNodeInvalidLid::FabricErrNodeInvalidLid(IBNode *p_node, phys_port_t port, lid_t lid, uint8_t lmc) : FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_CLUSTER);
    this->err_desc.assign(FER_NODE_INVALID_LID);
    char buffer[1024];
    sprintf(buffer, "Configured with invalid lid=%u lmc=%u on port %u", lid, lmc, port);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID
}

FabricErrSmpGmpFwMismatch::FabricErrSmpGmpFwMismatch(IBNode *p_node,
                              fw_version_obj_t &smp_fw,
                              fw_version_obj_t &gmp_fw) : FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_NODE_SMP_GMP_FW_MISMATCH);
    char buffer[1024];
    sprintf(buffer, "Mismatch in reported firmware version. "
                    "SMP: %u.%u.%u, GMP: %u, %u, %u",
                    smp_fw.major, smp_fw.minor, smp_fw.sub_minor,
                    gmp_fw.major, gmp_fw.minor, gmp_fw.sub_minor);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID
}

FabricErrSmpGmpCapMaskExist::FabricErrSmpGmpCapMaskExist(
    IBNode *p_node, bool is_smp, capability_mask_t &mask) : FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_NODE_SMP_GMP_CAP_MASK_EXIST);
    char buffer[1024];
    stringstream mask_str;

    mask_str << mask;
    sprintf(buffer, "A %s capability mask already exists. Overriden with %s. ",
            is_smp ? "SMP" : "GMP", mask_str.str().c_str());
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID
}
/******************************************************/
/*****************************************************/
/*****************************************************/
string FabricErrSM::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[1024];
    if (this->p_sm_obj)
        sprintf(buffer, STR_FMT "," U64H_FMT "," U64H_FMT "," U32D_FMT "," STR_FMT "," "\"" STR_FMT "\"",
                this->scope.c_str(),
                this->p_sm_obj->p_port->p_node->guid_get(),
                this->p_sm_obj->p_port->guid_get(),
                this->p_sm_obj->p_port->num,
                this->err_desc.c_str(),
                DescToCsvDesc(this->description).c_str());
    else
        sprintf(buffer, STR_FMT "," U8H_FMT "," U8H_FMT "," U8H_FMT "," STR_FMT "," "\"" STR_FMT "\"",
                this->scope.c_str(),
                0x0,            //NA
                0x0,            //NA
                0x0,            //NA
                this->err_desc.c_str(),
                DescToCsvDesc(this->description).c_str());
    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrSM::GetErrorLine()
{
    IBDIAG_ENTER;
    string line;
    if (this->p_sm_obj) {
        line = this->p_sm_obj->p_port->getName();
        line += " - ";
    }
    line += this->description;
    IBDIAG_RETURN(line);
}


FabricErrSMNotFound::FabricErrSMNotFound(sm_info_obj_t *p_sm_obj) : FabricErrSM(p_sm_obj)
{
    IBDIAG_ENTER;
    this->p_sm_obj = NULL;
    this->scope.assign(SCOPE_CLUSTER);
    this->err_desc.assign(FER_SM_NOT_FOUND);
    this->description = "Not found master subnet manager in fabric";
    IBDIAG_RETURN_VOID;
}


FabricErrSMUnknownState::FabricErrSMUnknownState(sm_info_obj_t *p_sm_obj) : FabricErrSM(p_sm_obj)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_CLUSTER);
    this->err_desc.assign(FER_SM_UNKNOWN_STATE);
    char buffer[1024];
    sprintf(buffer, "Running subnet manager with unknown state=%u", this->p_sm_obj->smp_sm_info.SmState);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}


FabricErrSMManyExists::FabricErrSMManyExists(sm_info_obj_t *p_sm_obj) : FabricErrSM(p_sm_obj)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_CLUSTER);
    this->err_desc.assign(FER_SM_MANY_EXISTS);
    this->description = "Running duplicated master subnet manager";
    IBDIAG_RETURN_VOID;
}


FabricErrSMNotCorrect::FabricErrSMNotCorrect(sm_info_obj_t *p_sm_obj) : FabricErrSM(p_sm_obj)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_CLUSTER);
    this->err_desc.assign(FER_SM_NOT_CORRECT);
    char buffer[1024];
    sprintf(buffer, "Master subnet manager that is running is not the correct one, subnet manager with priority=%u, GUID=" U64H_FMT " should be the master",
            this->p_sm_obj->smp_sm_info.Priority,
            this->p_sm_obj->smp_sm_info.GUID);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}


/******************************************************/
/******************************************************/
/******************************************************/
string FabricErrPM::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[1024];
    sprintf(buffer, STR_FMT "," U64H_FMT "," U64H_FMT "," U32D_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->p_port->p_node->guid_get(),
            this->p_port->guid_get(),
            this->p_port->num,
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());
    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrPM::GetErrorLine()
{
    IBDIAG_ENTER;
    string line;
    line = this->p_port->getName();
    line += " - ";
    line += this->description;
    IBDIAG_RETURN(line);
}


FabricErrPMCounterOverflow::FabricErrPMCounterOverflow(IBPort *p_port, string counter_name,
        u_int64_t overflow_value) : FabricErrPM(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_PM_COUNTER_OVERFLOW);
    char buff1[512];
    sprintf(buff1, U64D_FMT, overflow_value);
    char buff2[1024];
    sprintf(buff2, "%-35s : %-10s (overflow)", counter_name.c_str(), buff1);
    this->description.assign(buff2);
    IBDIAG_RETURN_VOID
}

FabricErrPMBaseCalcCounterOverflow::FabricErrPMBaseCalcCounterOverflow(
        IBPort *p_port, string counter_name) : FabricErrPM(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_PM_BASE_COUNTER_OVERFLOW);
    char buff1[1024];
    sprintf(buff1, "Base PM counters of %s counter are overflown,"
                   "  please run ibdiagnet -PC to reset.", counter_name.c_str());
    this->description.assign(buff1);
    IBDIAG_RETURN_VOID
}

FabricErrPMCounterExceedThreshold::FabricErrPMCounterExceedThreshold(IBPort *p_port, string counter_name,
        u_int64_t expected_value, u_int64_t actual_value) : FabricErrPM(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_PM_COUNTER_EXCEED_THRESH);
    char buff1[512];
    sprintf(buff1, U64D_FMT, actual_value);
    char buff2[1024];
    sprintf(buff2, "%-35s : %-10s (threshold=" U64D_FMT ")", counter_name.c_str(), buff1, expected_value);
    this->description.assign(buff2);
    IBDIAG_RETURN_VOID
}


FabricErrPMCounterNotSupported::FabricErrPMCounterNotSupported(IBPort *p_port, string counter_name) :
        FabricErrPM(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_PM_COUNTER_NOT_SUPPORTED);
    char buff[1024];
    sprintf(buff, "%-35s : %-10s (Not Supported)", counter_name.c_str(), "");
    this->description.assign(buff);
    IBDIAG_RETURN_VOID
}


FabricErrPMErrCounterIncreased::FabricErrPMErrCounterIncreased(IBPort *p_port, string counter_name,
        u_int64_t expected_value, u_int64_t actual_value) : FabricErrPM(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_PM_COUNTER_INCREASED);
    char buffer[1024];
    sprintf(buffer, "\"%s\" increased during the run "  \
            "(difference value=" U64D_FMT ",difference allowed threshold=" U64D_FMT ")",
            counter_name.c_str(),
            actual_value, expected_value);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID
}


/******************************************************/
/******************************************************/
/******************************************************/
string FabricErrPMCountersAll::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string res = "";
    if (this->csv_err_line != "") {
        res += this->csv_err_line;
    }
    IBDIAG_RETURN(res);
}

string FabricErrPMCountersAll::GetErrorLine()
{
    IBDIAG_ENTER;
    char buff[1024];
    string res = "";
    if (this->err_line != "") {
        sprintf(buff, "lid=" U16H_FMT " dev=" U32D_FMT " %s\n",
                this->p_port->base_lid,
                this->p_port->p_node->devId,
                this->p_port->getName().c_str());
        res += buff;
        sprintf(buff, "%-35s : %-10s\n",
                "Performance Monitor counter", "Value");
        res += "        ";
        res += buff;
        res += this->err_line;
    }
    IBDIAG_RETURN(res);
}


FabricErrPMCountersAll::FabricErrPMCountersAll(IBPort *p_port, list_p_pm_err& pm_errors) :
        FabricErrGeneral(), p_port(p_port), err_line(""), csv_err_line("")
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);                     //dummy one
    this->err_desc.assign("PM_COUNTERS_ALL");           //dummy one
    for (list_p_pm_err::iterator it = pm_errors.begin(); it != pm_errors.end(); ++it) {
        if (it != pm_errors.begin()) {
            this->err_line += "\n";
            this->csv_err_line += "\n";
        }
        this->err_line += "        ";
        this->err_line += (*it)->GetErrorLine();
        this->csv_err_line += (*it)->GetCSVErrorLine();
    }
    IBDIAG_RETURN_VOID;
}


/******************************************************/
/******************************************************/
/******************************************************/
string FabricErrBER::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[1024];
    sprintf(buffer, STR_FMT "," U64H_FMT "," U64H_FMT "," U32D_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->p_port->p_node->guid_get(),
            this->p_port->guid_get(),
            this->p_port->num,
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());
    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrBER::GetErrorLine()
{
    IBDIAG_ENTER;
    string line;
    line = this->p_port->getName();
    line += " - ";
    line += this->description;
    IBDIAG_RETURN(line);
}

FabricErrFwBERExceedThreshold::FabricErrFwBERExceedThreshold(IBPort *p_port,
        double thresh, double value) : FabricErrBER(p_port)
{
    IBDIAG_ENTER;

    scope.assign(SCOPE_PORT);
    err_desc.assign(FER_BER_EXCEED_THRESH);

    char buffer[1024];
    sprintf(buffer, "BER exceeds threshold - BER value = %e / threshold = %e",
            value,
            thresh);
    description.assign(buffer);

    IBDIAG_RETURN_VOID
}

FabricErrBERExceedThreshold::FabricErrBERExceedThreshold(IBPort *p_port,
        u_int64_t expected_value, long double actual_value) : FabricErrBER(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_BER_EXCEED_THRESH);
    char buffer[1024];
    if (expected_value == OVERFLOW_VAL_64_BIT)
        sprintf(buffer, "BER exceeds the threshold in port = %s"    \
                "(BER value=%Le, threshold=%e)",
                this->p_port->getName().c_str(),
                (1.0 / actual_value),
                0.0);
    else
        sprintf(buffer, "BER exceeds the threshold in port = %s"    \
                "(BER value=%Le, threshold=%e)",
                this->p_port->getName().c_str(),
                (1.0 / actual_value),
                (1.0 / (double)expected_value));
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID
}


FabricErrEffBERExceedThreshold::FabricErrEffBERExceedThreshold(IBPort *p_port,
        u_int64_t expected_value, long double actual_value) : FabricErrBER(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_BER_EXCEED_THRESH);
    char buffer[1024];
    if (expected_value == OVERFLOW_VAL_64_BIT)
        sprintf(buffer, "Effective BER exceeds the threshold in port = %s"    \
                "(BER value=%Le, threshold=%e)",
                this->p_port->getName().c_str(),
                (1.0 / actual_value),
                0.0);
    else
        sprintf(buffer, "Effective BER exceeds the threshold in port = %s"    \
                "(BER value=%Le, threshold=%e)",
                this->p_port->getName().c_str(),
                (1.0 / actual_value),
                (1.0 / (double)expected_value));
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID
}


FabricErrBERNoRcvData::FabricErrBERNoRcvData(IBPort *p_port) : FabricErrBER(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_BER_NO_RCV_DATA);
    this->description = "No received data - can't calculate BER";
    IBDIAG_RETURN_VOID
}


FabricErrBERIsZero::FabricErrBERIsZero(IBPort *p_port) : FabricErrBER(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_BER_IS_ZERO);
    this->description = "BER value is ZERO";
    IBDIAG_RETURN_VOID
}

FabricErrEffBERIsZero::FabricErrEffBERIsZero(IBPort *p_port) : FabricErrBER(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_BER_IS_ZERO);
    this->description = "Effective BER value is ZERO";
    IBDIAG_RETURN_VOID
}

/******************************************************/
/******************************************************/
/******************************************************/
string FabricErrPort::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[1024];
    sprintf(buffer, STR_FMT "," U64H_FMT "," U64H_FMT "," U32D_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->p_port->p_node->guid_get(),
            this->p_port->guid_get(),
            this->p_port->num,
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());
    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrPort::GetErrorLine()
{
    IBDIAG_ENTER;
    string line;
    line = this->p_port->getName();
    line += " - ";
    line += this->description;
    IBDIAG_RETURN(line);
}


FabricErrPortZeroLid::FabricErrPortZeroLid(IBPort *p_port) : FabricErrPort(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_CLUSTER);
    this->err_desc.assign(FER_PORT_ZERO_LID);
    this->description = "Configured with ZERO lid";
    IBDIAG_RETURN_VOID
}


FabricErrPortDuplicatedLid::FabricErrPortDuplicatedLid(IBPort *p_port) : FabricErrPort(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_CLUSTER);
    this->err_desc.assign(FER_PORT_DUPLICATED_LID);
    char buffer[1024];
    sprintf(buffer, "Configured with duplicated lid=%u", p_port->base_lid);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID
}

FabricErrPortNotRespond::FabricErrPortNotRespond(IBPort *p_port, string desc) : FabricErrPort(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_PORT_NOT_RESPOND);
    this->description = "No response for MAD";
    if (desc != "") {
        this->description += " ";
        this->description += desc;
    }
    IBDIAG_RETURN_VOID
}


FabricErrPortWrongConfig::FabricErrPortWrongConfig(IBPort *p_port, string desc) : FabricErrPort(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_PORT_WRONG_CONFIG);
    this->description = "Port is configured with wrong value";
    if (desc != "") {
        this->description += " ";
        this->description += desc;
    }
    IBDIAG_RETURN_VOID
}


FabricErrPortNotSupportCap::FabricErrPortNotSupportCap(IBPort *p_port, string desc) : FabricErrPort(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_PORT_NOT_SUPPORT_CAP);
    this->description = desc;
    IBDIAG_RETURN_VOID
}

FabricErrPortInvalidValue::FabricErrPortInvalidValue(IBPort *p_port, string desc) : FabricErrPort(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_PORT_INVALID_VALUE);
    this->description = desc;
    IBDIAG_RETURN_VOID
}

/******************************************************/
/******************************************************/
/******************************************************/
string FabricErrDR::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line = "";
    char buffer[1024];
    sprintf(buffer, STR_FMT "," U8H_FMT "," U8H_FMT "," U8H_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            0x0,            //NA
            0x0,            //NA
            0x0,            //NA
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());
    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

FabricErrDR::FabricErrDR(string direct_route_str) :
        FabricErrGeneral(), direct_route_str(direct_route_str)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_CLUSTER);
    this->err_desc.assign(FER_DIRECT_ROUTE);
    this->description = "Bad link was found. ";
    this->description += this->direct_route_str;
    IBDIAG_RETURN_VOID;
}


/******************************************************/
/******************************************************/
/******************************************************/
string FabricErrLink::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[2096];
    sprintf(buffer, STR_FMT "," U64H_FMT "," U64H_FMT "," U32D_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->p_port1->p_node->guid_get(),
            this->p_port1->guid_get(),
            this->p_port1->num,
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());
    csv_line.assign(buffer);
    csv_line += "\n";
    sprintf(buffer, STR_FMT "," U64H_FMT "," U64H_FMT "," U32D_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->p_port2->p_node->guid_get(),
            this->p_port2->guid_get(),
            this->p_port2->num,
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());
    csv_line += string(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrLink::GetErrorLine()
{
    IBDIAG_ENTER;
    string line;
    line = "Link: ";
    line += this->p_port1->getName();
    line += "<-->";
    line += this->p_port2->getName();
    line += " - ";
    line += this->description;
    IBDIAG_RETURN(line);
}


FabricErrLinkLogicalStateWrong::FabricErrLinkLogicalStateWrong(IBPort *p_port1, IBPort *p_port2) :
        FabricErrLink(p_port1, p_port2)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_LINK_LOGICAL_STATE_WRONG);
    char buffer[1024];
    sprintf(buffer, "Logical state is different in connected ports (port=%s state is %s and remote port=%s state is %s)",
            this->p_port1->getName().c_str(), portstate2char(this->p_port1->get_internal_state()),
            this->p_port2->getName().c_str(), portstate2char(this->p_port2->get_internal_state()));
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID
}


FabricErrLinkLogicalStateNotActive::FabricErrLinkLogicalStateNotActive(IBPort *p_port1, IBPort *p_port2) :
        FabricErrLink(p_port1, p_port2)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_LINK_LOGICAL_STATE_NOT_ACTIVE);
    char buffer[1024];
    sprintf(buffer, "Link logical state is %s", portstate2char(this->p_port1->get_internal_state()));
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID
}


FabricErrLinkDifferentSpeed::FabricErrLinkDifferentSpeed(IBPort *p_port1, IBPort *p_port2) :
        FabricErrLink(p_port1, p_port2)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_LINK_LOGICAL_DIFFERENT_SPEED);
    char buffer[1024];
    sprintf(buffer, "Speed is different in connected ports (port=%s speed is %s and remote port=%s speed is %s)",
            this->p_port1->getName().c_str(), speed2char(this->p_port1->get_internal_speed()),
            this->p_port2->getName().c_str(), speed2char(this->p_port2->get_internal_speed()));
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID
}


FabricErrLinkUnexpectedSpeed::FabricErrLinkUnexpectedSpeed(IBPort *p_port1, IBPort *p_port2, string desc) :
        FabricErrLink(p_port1, p_port2)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_LINK_LOGICAL_UNEXPECTED_SPEED);
    char buffer[1024];
    sprintf(buffer, "Unexpected actual link speed %s", speed2char(p_port1->get_internal_speed()));
    this->description.assign(buffer);
    if (desc != "") {
        this->description += " ";
        this->description += desc;
    }
    IBDIAG_RETURN_VOID
}


FabricErrLinkDifferentWidth::FabricErrLinkDifferentWidth(IBPort *p_port1, IBPort *p_port2) :
        FabricErrLink(p_port1, p_port2)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_LINK_LOGICAL_DIFFERENT_WIDTH);
    char buffer[1024];
    sprintf(buffer, "Width is different in connected ports (port=%s width is %s and remote port=%s width is %s)",
            this->p_port1->getName().c_str(), width2char(this->p_port1->get_internal_width()),
            this->p_port2->getName().c_str(), width2char(this->p_port2->get_internal_width()));
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID
}


FabricErrLinkUnexpectedWidth::FabricErrLinkUnexpectedWidth(IBPort *p_port1, IBPort *p_port2, string desc) :
        FabricErrLink(p_port1, p_port2)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_LINK_LOGICAL_UNEXPECTED_WIDTH);
    char buffer[1024];
    sprintf(buffer, "Unexpected width, actual link width is %s", width2char(p_port1->get_internal_width()));
    this->description.assign(buffer);
    if (desc != "") {
        this->description += " ";
        this->description += desc;
    }
    IBDIAG_RETURN_VOID
}


FabricErrLinkAutonegError::FabricErrLinkAutonegError(IBPort *p_port1, IBPort *p_port2, string desc) :
        FabricErrLink(p_port1, p_port2)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_LINK_LOGICAL_AUTONEG_ERR);
    char buffer[1024];
    sprintf(buffer, "Autoneg should fail on this link");
    this->description.assign(buffer);
    if (desc != "") {
        this->description += " ";
        this->description += desc;
    }
    IBDIAG_RETURN_VOID
}

/******************************************************/
/******************************************************/
/******************************************************/
FabricErrPKeyMismatch::FabricErrPKeyMismatch(IBPort *p_port1, IBPort *p_port2,
        string port1_pkey_str, string port2_pkey_str) :
	FabricErrGeneral(), p_port1(p_port1), p_port2(p_port2)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_PKEY_MISMATCH);
    this->description = "Mismatching pkeys between ";

    this->description += this->p_port1->getName().c_str();
    if (port1_pkey_str != "") {
        this->description += " {";
        this->description += port1_pkey_str;
        this->description += "}";
    }

    this->description += " <--> ";

    this->description += this->p_port2->getName().c_str();
    if (port2_pkey_str != "") {
        this->description += " {";
        this->description += port2_pkey_str;
        this->description += "}";
    }

    IBDIAG_RETURN_VOID;
}

string FabricErrPKeyMismatch::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[2096];
    sprintf(buffer, STR_FMT "," U64H_FMT "," U64H_FMT "," U32D_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->p_port1->p_node->guid_get(),
            this->p_port1->guid_get(),
            this->p_port1->num,
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());
    csv_line.assign(buffer);
    csv_line += "\n";
    sprintf(buffer, STR_FMT "," U64H_FMT "," U64H_FMT "," U32D_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->p_port2->p_node->guid_get(),
            this->p_port2->guid_get(),
            this->p_port2->num,
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());
    csv_line += string(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrPKeyMismatch::GetErrorLine()
{
    IBDIAG_ENTER;
    string line;
    line = this->description;
    IBDIAG_RETURN(line);
}


/******************************************************/
/******************************************************/
/******************************************************/
FabricErrAGUID::FabricErrAGUID(IBPort *p_port1, string owner_name, u_int64_t guid, string guid_type):
        FabricErrGeneral(),
        p_port(p_port1), guid_owner_name(owner_name),
        duplicated_guid(guid), guid_type(guid_type)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_AGUID);
    char buffer[1024];
    sprintf(buffer, "Alias GUID " U64H_FMT " on port %s, used on %s as %s",
            this->duplicated_guid,
            this->p_port->getName().c_str(),
            this->guid_owner_name.c_str(),
            this->guid_type.c_str());
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID
}

string FabricErrAGUID::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[1024];
    sprintf(buffer, STR_FMT "," U64H_FMT "," U8H_FMT "," U8H_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->duplicated_guid,
            0x0,            //NA
            0x0,            //NA
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());
    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrAGUID::GetErrorLine()
{
    IBDIAG_ENTER;
    string line;
    line = this->p_port->getName();
    line += " - ";
    line += this->description;
    IBDIAG_RETURN(line);
}

FabricErrAGUIDInvalidFirstEntry::FabricErrAGUIDInvalidFirstEntry(IBPort *port, u_int64_t first_alias_guid):
    FabricErrGeneral(),
    p_port(port),
    guid_zero_index(first_alias_guid)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_AGUID);
    char buffer[1024];
    sprintf(buffer, "Invalid guid ( " U64H_FMT " ) at the entry zero in guid info table of port %s. "
            "It must be equal to the port guid ( " U64H_FMT  " ).",
            this->guid_zero_index,
            this->p_port->getName().c_str(),
            this->p_port->guid_get());
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID
}

string FabricErrAGUIDInvalidFirstEntry::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[1024];

    sprintf(buffer, STR_FMT "," U64H_FMT "," U8H_FMT "," U8H_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->guid_zero_index,
            0x0,            //NA
            0x0,            //NA
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());

    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrAGUIDInvalidFirstEntry::GetErrorLine()
{
    IBDIAG_ENTER;
    string line;
    line = this->p_port->getName();
    line += " - ";
    line += this->description;
    IBDIAG_RETURN(line);
}

/******************************************************/
/******************************************************/
/******************************************************/
FabricErrPortInfoFail::FabricErrPortInfoFail(IBNode *p_node,
                                             unsigned int    port_num,
                                             const char *error_desc) :
	FabricErrGeneral(), m_p_node(p_node),m_port_num(port_num)
{
    IBDIAG_ENTER;
    char port_num_buff[3];
    snprintf(port_num_buff,2,"%u",m_port_num);
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_PORT_INFO_FAILED);
    this->description = "Port info failed on ";

    this->description += m_p_node->name.c_str();
    this->description += "  port number:";
    this->description += port_num_buff;
    this->description += "  with error:";
    this->description += error_desc;

    IBDIAG_RETURN_VOID;
}

string FabricErrPortInfoFail::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[2096];
    sprintf(buffer, STR_FMT "," U64H_FMT "," U32D_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            scope.c_str(),
            m_p_node->guid_get(),
            m_port_num,
            err_desc.c_str(),
            DescToCsvDesc(description).c_str());
    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrPortInfoFail::GetErrorLine() //TODO TAMIR change to const string &
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(description);
}

/******************************************************/
/******************************************************/
/******************************************************/
FabricErrDiscovery::FabricErrDiscovery(IBNode *p_node,
                                       uint8_t max_hops) :
	FabricErrGeneral(), p_node(p_node), max_hops(max_hops)
{
    IBDIAG_ENTER;
    char hop_num_buff[32];
    sprintf(hop_num_buff, "%u", max_hops);
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_DISCOVERY_REACHED_MAX_HOP);
    this->description = "Discovery failed on ";
    this->description += p_node->name.c_str();
    this->description += "  when reached hop number:";
    this->description += hop_num_buff;
    IBDIAG_RETURN_VOID;
}

string FabricErrDiscovery::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[2096];
    sprintf(buffer, STR_FMT "," U64H_FMT "," U32D_FMT "," "\"" STR_FMT "\"",
            scope.c_str(),
            p_node->guid_get(),
            max_hops,
            DescToCsvDesc(description).c_str());
    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

FabricErrVPort::FabricErrVPort(IBVPort *p_port1, string owner_name,
                               u_int64_t guid, string guid_type):
    FabricErrGeneral(),
    p_port(p_port1), guid_owner_name(owner_name),
    duplicated_guid(guid), guid_type(guid_type)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_AGUID);
    char buffer[1024];
    sprintf(buffer, "VPort GUID " U64H_FMT " on port %s, used on %s as %s",
            this->duplicated_guid,
            this->p_port->getName().c_str(),
            this->guid_owner_name.c_str(),
            this->guid_type.c_str());
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID
}

string FabricErrVPort::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[1024];
    sprintf(buffer, STR_FMT "," U64H_FMT "," U8H_FMT "," U8H_FMT "," STR_FMT
            "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->duplicated_guid,
            0x0,            //NA
            0x0,            //NA
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());
    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrDiscovery::GetErrorLine()
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(description);
}

string FabricErrVPort::GetErrorLine()
{
    IBDIAG_ENTER;
    string line;
    line = this->p_port->getName();
    line += " - ";
    line += this->description;
    IBDIAG_RETURN(line);
}

FabricErrVPortGUIDInvalidFirstEntry::FabricErrVPortGUIDInvalidFirstEntry(
        IBPort *phys_port, IBVPort *port, u_int64_t first_alias_guid):
        FabricErrGeneral(),
        p_port(port),
        p_phys_port(phys_port),
        guid_at_zero_index(first_alias_guid)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_AGUID);
    char buffer[1024];
    sprintf(buffer, "Invalid guid ( " U64H_FMT " ) at the entry"
            " zero in guid info table of port %s. "
            "It must be equal to the port guid ( " U64H_FMT  " ).",
            this->guid_at_zero_index,
            this->p_port->getName().c_str(),
            this->p_phys_port->guid_get());
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID
}

string FabricErrVPortGUIDInvalidFirstEntry::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[1024];

    sprintf(buffer, STR_FMT "," U64H_FMT "," U8H_FMT "," U8H_FMT "," STR_FMT
            "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->guid_at_zero_index,
            0x0,            //NA
            0x0,            //NA
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());

    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrVPortGUIDInvalidFirstEntry::GetErrorLine()
{
    IBDIAG_ENTER;
    string line;
    line = this->p_port->getName();
    line += " - ";
    line += this->description;
    IBDIAG_RETURN(line);
}

FabricErrVPortIvalidTopIndex::FabricErrVPortIvalidTopIndex(IBPort *port, uint16_t cap,
        uint16_t top):
        FabricErrGeneral(),
        p_port(port),
        cap_idx(cap),
        top_idx(top)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_VIRT_INFO_INVALID_TOP);
    char buffer[1024];
    sprintf(buffer, "on port %s, top index %d can't be greater than "
            "capability index %d",
            this->p_port->getName().c_str(),
            this->top_idx,
            this->cap_idx);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID
}

string FabricErrVPortIvalidTopIndex::GetCSVErrorLine()
{
   IBDIAG_ENTER;
    string csv_line;
    char buffer[1024];

    sprintf(buffer, STR_FMT "," U64H_FMT "," U64H_FMT "," U8H_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->p_port->p_node->guid_get(),
            this->p_port->guid_get(),
            this->p_port->num,
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());

    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrVPortIvalidTopIndex::GetErrorLine()
{
    IBDIAG_ENTER;
    string line;
    line = this->p_port->getName();
    line += " - ";
    line += this->description;
    IBDIAG_RETURN(line);
}

/********************************************/
/********************************************/
/********************************************/
string FabricErrVPortInvalid::GetCSVErrorLine()
{
   IBDIAG_ENTER;
    string csv_line;
    char buffer[1024];

    sprintf(buffer, STR_FMT "," U64H_FMT "," U64H_FMT "," U8H_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->p_port->p_node->guid_get(),
            this->p_port->guid_get(),
            this->p_port->num,
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());

    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrVPortInvalid::GetErrorLine()
{
    IBDIAG_ENTER;
    string line;
    line = this->p_port->getName();
    line += " - ";
    line += this->description;
    IBDIAG_RETURN(line);
}

FabricErrVLidZero::FabricErrVLidZero(IBPort *port, IBVPort *vport):
        FabricErrVPortInvalid(port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_VIRT_INFO_INVALID_VLID);
    char buffer[1024];
    sprintf(buffer, "On vport %s, lid_required is 1 but vlid is 0",
            vport->getName().c_str());
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID
}


FabricErrInvalidIndexForVLid::FabricErrInvalidIndexForVLid(
        IBPort *port,
        IBVPort *vport,
        u_int16_t lid_by_vport_idx):
        FabricErrVPortInvalid(port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_VIRT_INFO_INVALID_VLID);
    char buffer[1024];
    sprintf(buffer, "Invalid index num %d for vport %s, vport by index not found",
            lid_by_vport_idx,
            vport->getName().c_str());
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID
}

FabricErrVlidForVlidByIndexIsZero::FabricErrVlidForVlidByIndexIsZero(
        IBPort *port, IBVPort *vport, IBVPort *vport_by_index,
        u_int16_t lid_by_vport_idx):
        FabricErrVPortInvalid(port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_VIRT_INFO_INVALID_VLID);
    char buffer[1024];
    sprintf(buffer, "Found vlid 0 in vport %s." \
                    "Found by vport %s with lid_required=0 and index num %d",
            vport_by_index->getName().c_str(),
            vport->getName().c_str(),
            lid_by_vport_idx);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID
}

/******************************************************/
/******************************************************/
/******************************************************/
SharpErrEdgeNodeNotFound::SharpErrEdgeNodeNotFound(
        IBNode *p_node, lid_t rlid) :
        FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_SHARP_EDGE_NODE_NOT_FOUND);
    char buffer[1024];
    sprintf(buffer,
            "Node has no Edge remote lid: %d",
            rlid);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

SharpErrDisconnectedTreeNode::SharpErrDisconnectedTreeNode(
        IBNode *p_node, u_int16_t tree_id) :
        FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_SHARP_DISCONNECTED_TREE_NODE);
    char buffer[1024];
    sprintf(buffer,
            "In Node %s Tree Node is Disconnected for treeID: %d",
            p_node->getName().c_str(), tree_id);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

SharpErrNodeTreeIDNotMatchGetRespondTreeID::SharpErrNodeTreeIDNotMatchGetRespondTreeID(
        IBNode *p_node, u_int16_t tree_id) :
        FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_SHARP_TREE_ID_NOT_MATCH);
    char buffer[1024];
    sprintf(buffer,
            "In Node %s treeID not match get respond for treeID: %d",
            p_node->getName().c_str(), tree_id);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

SharpErrRootTreeNodeAlreadyExistsForTreeID::SharpErrRootTreeNodeAlreadyExistsForTreeID(
        IBNode *p_node, u_int16_t tree_id) :
        FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_SHARP_ROOT_ALREADY_EXISTS);
    char buffer[1024];
    sprintf(buffer,
            "In Node %s found root tree (parent qpn 0) which is already exists for treeID: %d",
            p_node->getName().c_str(), tree_id);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

SharpErrDuplicatedQPNForAggNode::SharpErrDuplicatedQPNForAggNode(
        IBNode *p_node, u_int16_t tree_id, u_int16_t dup_tree_id,
        u_int32_t qpn) :
        FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_SHARP_DUPLICATE_QPN_FOR_AN);
    char buffer[1024];
    sprintf(buffer,
            "In Node %s Duplicated QPN %d for TreeID: %d and TreeID: %d",
            p_node->getName().c_str(), qpn, tree_id, dup_tree_id);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

SharpErrQPNotActive::SharpErrQPNotActive(
        IBNode *p_node, u_int32_t qpn, u_int8_t qp_state) :
        FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_SHARP_QP_NOT_ACTIVE);
    char buffer[1024];
    sprintf(buffer,
            "In Node %s QPN %d state is not active. QP State: %d",
            p_node->getName().c_str(), qpn, qp_state);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

SharpErrRQPNotValid::SharpErrRQPNotValid(
        IBNode *p_node, u_int32_t qpn, u_int32_t rqpn) :
        FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_SHARP_RQP_NOT_VALID);
    char buffer[1024];
    sprintf(buffer,
            "In Node %s parent RQPN %d (as read from the parent QPCConfig)" \
            "is not the peer qpn. Peer QP: %d",
            p_node->getName().c_str(), rqpn, qpn);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}
