/*
 *  Copyright (c) 2004-2018 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <stdarg.h>
#include <typeinfo>
#include <algorithm>
#include <infiniband/ibdiag/ibdiag_ibdm_extended_info.h>


/****************************************************/
static unsigned int get_max(unsigned int num)
{
    IBDIAG_ENTER;
    unsigned r = 0;     // r will be lg(num)

    while (num >>= 1)   // unroll for more speed...
        r++;

    IBDIAG_RETURN(1 << r);
}

IBLinkSpeed CalcFinalSpeed(u_int32_t speed1, u_int32_t speed2)
{
    IBDIAG_ENTER;

    u_int32_t common_speed = speed1 & speed2;

    if (!common_speed) {
        IBDIAG_RETURN(IB_UNKNOWN_LINK_SPEED);
    }

    IBLinkSpeed res;
    //extended speeds at second byte
    u_int32_t ext_speed  = common_speed & 0xff00;
    //mellanox extended speeds at third byte
    u_int32_t mlnx_speed = common_speed & 0xff0000;

    //look first at greater speeds: mlnx speeds and extended speeds
    if (mlnx_speed || ext_speed) {
        u_int32_t ext_speed_max =
            !ext_speed ? 0 : (u_int32_t)get_max(ext_speed);
        u_int32_t mlnx_speed_max =
            !mlnx_speed ? 0 : (u_int32_t)get_max(mlnx_speed);

        //usualy, extended speeds > mlnx speeds, except EDR20 and FDR
        //next time the new speed is added consider calling function speed2bits
        //on ext_speed_max, mlnx_speed_max, compare the results to select max
        if (!ext_speed_max ||
            ((IBLinkSpeed)mlnx_speed_max == IB_LINK_SPEED_EDR_20 && //EDR20
             (IBLinkSpeed)ext_speed_max == IB_LINK_SPEED_14))       //FDR
            res = (IBLinkSpeed)mlnx_speed_max;
        else
            res = (IBLinkSpeed)ext_speed_max;
    } else {
        //regular speeds at first byte
        u_int32_t reg_speed  = common_speed & 0xff;
        u_int32_t reg_speed_max =
            !reg_speed ? 0 :  (u_int32_t)get_max(reg_speed);
        res = (IBLinkSpeed)reg_speed_max;
    }

    IBDIAG_RETURN(res);
}

//the following method can be changed like speed calculation
IBLinkWidth CalcFinalWidth(u_int8_t width1, u_int8_t width2)
{
    IBDIAG_ENTER;

    u_int8_t intersection = (u_int8_t)(width1 & width2);

    if (intersection & IB_LINK_WIDTH_12X) IBDIAG_RETURN(IB_LINK_WIDTH_12X);
    if (intersection & IB_LINK_WIDTH_8X) IBDIAG_RETURN(IB_LINK_WIDTH_8X);
    if (intersection & IB_LINK_WIDTH_4X) IBDIAG_RETURN(IB_LINK_WIDTH_4X);
    if (intersection & IB_LINK_WIDTH_2X) IBDIAG_RETURN(IB_LINK_WIDTH_2X);
    if (intersection & IB_LINK_WIDTH_1X) IBDIAG_RETURN(IB_LINK_WIDTH_1X);

    IBDIAG_RETURN(IB_UNKNOWN_LINK_WIDTH);
}

// In InfiniBand the link rate is baud rate, so 1 kb/s is 1000 bits/sec
// 1 Gbits/sec = 1000 ^ 3 bits per sec
u_int64_t CalcLinkRate(IBLinkWidth link_width, IBLinkSpeed link_speed)
{
    IBDIAG_ENTER;

    u_int64_t width_val;
    switch (link_width) {
    case IB_LINK_WIDTH_1X: {
        width_val = 1;
        break;
    }
    case IB_LINK_WIDTH_4X: {
        width_val = 4;
        break;
    }
    case IB_LINK_WIDTH_8X: {
        width_val = 8;
        break;
    }
    case IB_LINK_WIDTH_12X: {
        width_val = 12;
        break;
    }
    case IB_LINK_WIDTH_2X: {
        width_val = 2;
        break;
    }
    default:
        IBDIAG_RETURN(0);
    }

    u_int64_t speed_val;
    switch (link_speed) {
    case IB_LINK_SPEED_2_5: {
        speed_val = 0x9502F900ULL;      //2.5 * 1000^3
        break;
    }
    case IB_LINK_SPEED_5: {
        speed_val = 0x12A05F200ULL;     //5 * 1000^3
        break;
    }
    case IB_LINK_SPEED_10: {
        speed_val = 0x2540BE400ULL;     //10 * 1000^3
        break;
    }
    case IB_LINK_SPEED_14: {
        speed_val = 0x342770C00ULL;     //14 * 1000^3
        break;
    }
    case IB_LINK_SPEED_25: {
        speed_val = 0x5D21DBA00ULL;     //25 * 1000^3
        break;
    }
    case IB_LINK_SPEED_50: {
        speed_val = 0xBA43B7400ULL;     //50 * 1000^3
        break;
    }
    case IB_LINK_SPEED_FDR_10: {
        //speed_val = 0x300000000ULL;     //12 (12 in 8/10 encoding is like 10 in 64/66 encoding) * 1024^3
        speed_val = 0x2540BE400ULL;     //10 * 1000^3
        break;
    }
    case IB_LINK_SPEED_EDR_20: {
        speed_val = 0x4A817C800ULL;     //20 * 1000^3
        break;
    }

    default:
        IBDIAG_RETURN(0);
    }

    IBDIAG_RETURN(width_val * speed_val);
}

// get the mlnx extended speeds | extended speeds | regular speeds
//for enabled supported and active
void GetTotalSpeeds(IN SMP_PortInfo *p_port_info,
                    IN SMP_MlnxExtPortInfo *p_mlnx_ext_port_info,
                    IN u_int32_t cap_mask,
                    OUT u_int32_t *sup_speed,
                    OUT u_int32_t *en_speed,
                    OUT u_int32_t *actv_speed)
{
    u_int32_t l_sup_speed, l_en_speed, l_actv_speed;

    l_sup_speed = l_en_speed = l_actv_speed = 0;

    if (p_port_info) {
        l_sup_speed    = p_port_info->LinkSpeedSup;
        l_en_speed     = p_port_info->LinkSpeedEn;
        l_actv_speed   = p_port_info->LinkSpeedActv;

        //check port's extended speed if needed
        if (cap_mask & IBIS_IB_PORT_CAP_HAS_EXT_SPEEDS) {
            l_sup_speed  |= (p_port_info->LinkSpeedExtSup << 8);
            l_en_speed   |= (p_port_info->LinkSpeedExtEn << 8);
            l_actv_speed |= (p_port_info->LinkSpeedExtActv << 8);
        }
    }

    //check port's mlnx extended speed if needed
    if (p_mlnx_ext_port_info) {
        l_sup_speed  |= (p_mlnx_ext_port_info->LinkSpeedSupported << 16);
        l_en_speed   |= (p_mlnx_ext_port_info->LinkSpeedEnabled << 16);
        l_actv_speed |= (p_mlnx_ext_port_info->LinkSpeedActive << 16);
    }

    if (sup_speed)
        *sup_speed  = l_sup_speed;
    if (en_speed)
        *en_speed   = l_en_speed;
    if (actv_speed)
        *actv_speed = l_actv_speed;
}

/****************************************************/
/******************private methods*******************/
/****************************************************/
void IBDMExtendedInfo::SetLastError(const char *fmt, ...)
{
    IBDIAG_ENTER;
    char buff[IBDIAG_ERR_SIZE];
    va_list args;

    memset(buff, 0, sizeof(buff));
    va_start(args, fmt);
    vsprintf(buff, fmt, args);
    va_end(args);
    this->last_error.assign(buff);
    IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "-E- %s\n", this->last_error.c_str());
    IBDIAG_RETURN_VOID;
}


template <class OBJ_VEC_TYPE, class OBJ_TYPE>
void IBDMExtendedInfo::addPtrToVec(OBJ_VEC_TYPE& vector_obj, OBJ_TYPE *p_obj)
{
    IBDIAG_ENTER;
    if ((vector_obj.size() > p_obj->createIndex + 1) &&
            vector_obj[p_obj->createIndex]) //if exist already
        IBDIAG_RETURN_VOID;
    if (vector_obj.empty() || (vector_obj.size() < p_obj->createIndex + 1))
        for (int i = (int)vector_obj.size(); i < (int)p_obj->createIndex + 1; ++i)
            vector_obj.push_back(NULL);
    vector_obj[p_obj->createIndex] = p_obj;
    IBDIAG_RETURN_VOID;
}

template <class OBJ_VEC_TYPE, class OBJ_TYPE>
OBJ_TYPE * IBDMExtendedInfo::getPtrFromVec(OBJ_VEC_TYPE& vector_obj, u_int32_t idx)
{
    IBDIAG_ENTER;
    //not exists
    if (idx < 0 || (vector_obj.size() < idx + 1))
             IBDIAG_RETURN(NULL);
    IBDIAG_RETURN(vector_obj[idx]);
}

template <class OBJ_VEC_TYPE, class OBJ_TYPE, class DATA_VEC_TYPE, class DATA_TYPE>
int IBDMExtendedInfo::addDataToVec(OBJ_VEC_TYPE& vector_obj,
        OBJ_TYPE *p_obj,
        DATA_VEC_TYPE& vector_data,
        DATA_TYPE& data)
{
    IBDIAG_ENTER;
    if (!p_obj)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    //if exist already
    if ((vector_data.size() >= p_obj->createIndex + 1) &&
            vector_data[p_obj->createIndex])
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    //add
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Adding %s for (%s, index=%u)\n",
            typeid(data).name(), p_obj->getName().c_str(), p_obj->createIndex);
    if (vector_data.empty() || (vector_data.size() < p_obj->createIndex + 1))
        for (int i = (int)vector_data.size(); i < (int)p_obj->createIndex + 1; ++i)
            vector_data.push_back(NULL);
    DATA_TYPE *p_curr_data = new DATA_TYPE;
    if (!p_curr_data) {
        this->SetLastError("Failed to allocate %s", typeid(data).name());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
    }
    *p_curr_data = data;
    vector_data[p_obj->createIndex] = p_curr_data;
    this->addPtrToVec(vector_obj, p_obj);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

template <class OBJ_VEC_TYPE, class OBJ_TYPE, class DATA_VEC_TYPE, class DATA_TYPE>
int IBDMExtendedInfo::addDataToVecInVec(OBJ_VEC_TYPE& vector_obj,
        OBJ_TYPE *p_obj,
        DATA_VEC_TYPE& vec_of_vectors,
        u_int32_t data_idx,
        DATA_TYPE& data)
{
    IBDIAG_ENTER;
    if (!p_obj)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    //if exist already
    if ((vec_of_vectors.size() >= p_obj->createIndex + 1) &&
            (vec_of_vectors[p_obj->createIndex].size() >= data_idx + 1))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    //add
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Adding %s for (%s, index=%u, vec_idx=%u)\n",
            typeid(data).name(), p_obj->getName().c_str(), p_obj->createIndex, data_idx);
    if (vec_of_vectors.empty() || (vec_of_vectors.size() < p_obj->createIndex + 1))
        vec_of_vectors.resize(p_obj->createIndex + 1);

    if (vec_of_vectors[p_obj->createIndex].empty() ||
            (vec_of_vectors[p_obj->createIndex].size() < data_idx + 1))
        for (int i = (int)vec_of_vectors[p_obj->createIndex].size(); i < (int)data_idx + 1; ++i)
            vec_of_vectors[p_obj->createIndex].push_back(NULL);
    DATA_TYPE *p_curr_data = new DATA_TYPE;
    if (!p_curr_data) {
        this->SetLastError("Failed to allocate %s", typeid(data).name());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
    }
    *p_curr_data = data;
    vec_of_vectors[p_obj->createIndex][data_idx] = p_curr_data;
    this->addPtrToVec(vector_obj, p_obj);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

template <class OBJ_VEC_TYPE, class OBJ_TYPE>
OBJ_TYPE * IBDMExtendedInfo::getPtrFromVecInVec(OBJ_VEC_TYPE& vec_of_vectors,
        u_int32_t idx1,
        u_int32_t idx2)
{
    IBDIAG_ENTER;
    //not exists
    if (idx1 < 0 || (vec_of_vectors.size() < idx1 + 1))
        IBDIAG_RETURN(NULL);
    if (idx2 < 0 || (vec_of_vectors[idx1].size() < idx2 + 1))
        IBDIAG_RETURN(NULL);
    IBDIAG_RETURN(vec_of_vectors[idx1][idx2]);
}


/****************************************************/
/*******************public methods*******************/
/****************************************************/
void IBDMExtendedInfo::CleanUpInternalDB()
{
    IBDIAG_ENTER;
    for (list_p_sm_info_obj::iterator it = this->sm_info_obj_list.begin();
            it != this->sm_info_obj_list.end();
            ++it)
        if (*it)
            delete (*it);
    for (vector_p_smp_node_info::iterator it = this->smp_node_info_vector.begin();
            it < this->smp_node_info_vector.end(); ++it)
        if (*it)
            delete (*it);
    for (vector_p_smp_virtual_info::iterator it = this->smp_virtual_info_vector.begin();
            it < this->smp_virtual_info_vector.end(); ++it)
        if (*it)
            delete (*it);
    for (vector_p_smp_vnode_info::iterator it = this->smp_vnode_info_vector.begin();
            it < this->smp_vnode_info_vector.end(); ++it)
        if (*it)
            delete (*it);
    for (vector_p_smp_switch_info::iterator it = this->smp_switch_info_vector.begin();
            it < this->smp_switch_info_vector.end(); ++it)
        if (*it)
            delete (*it);
    for (vector_p_smp_vport_info::iterator it = this->smp_vport_info_vector.begin();
            it < this->smp_vport_info_vector.end(); ++it)
        if (*it)
            delete (*it);
    for (vector_p_smp_port_info::iterator it = this->smp_port_info_vector.begin();
            it < this->smp_port_info_vector.end(); ++it)
        if (*it)
            delete (*it);
    for (vector_p_smp_port_info_ext::iterator it = this->smp_port_info_ext_vector.begin();
            it < this->smp_port_info_ext_vector.end(); ++it)
        if (*it)
            delete (*it);
    for (vector_p_smp_mlnx_ext_port_info::iterator it = this->smp_mlnx_ext_port_info_vector.begin();
            it < this->smp_mlnx_ext_port_info_vector.end(); ++it)
        if (*it)
            delete (*it);
    for (vector_p_vs_general_info::iterator it = this->vs_general_info_vector.begin();
            it < this->vs_general_info_vector.end(); ++it)
        if (*it)
            delete (*it);
    for (vector_p_pm_cap_mask::iterator it = this->pm_cap_mask_vector.begin();
                it < this->pm_cap_mask_vector.end(); ++it)
        if (*it)
            delete (*it);
    for (vector_p_pm_info_obj::iterator it = this->pm_info_obj_vector.begin();
            it < this->pm_info_obj_vector.end(); ++it) {
        if (*it) {
            if ((*it)->p_port_counters) {
                delete (*it)->p_port_counters;
            }
            if ((*it)->p_extended_port_counters) {
                delete (*it)->p_extended_port_counters;
            }
            if ((*it)->p_port_ext_speeds_counters) {
                delete (*it)->p_port_ext_speeds_counters;
            }
            if ((*it)->p_port_ext_speeds_rsfec_counters) {
                delete (*it)->p_port_ext_speeds_rsfec_counters;
            }
            if ((*it)->p_port_llr_statistics) {
                delete (*it)->p_port_llr_statistics;
            }
            if ((*it)->p_port_calc_counters) {
                delete (*it)->p_port_calc_counters;
            }
            delete (*it);
        }
    }

    for (vector_v_smp_pkey_tbl::iterator it = this->smp_pkey_tbl_v_vector.begin();
            it != this->smp_pkey_tbl_v_vector.end();
            ++it) {
        for (vector_p_smp_pkey_tbl::iterator it2 = (*it).begin(); it2 != (*it).end(); ++it2)
            delete (*it2);
        (*it).clear();
    }
    for (vector_v_smp_pkey_tbl::iterator it = this->smp_vport_pkey_tbl_v_vector.begin();
            it != this->smp_vport_pkey_tbl_v_vector.end();
            ++it) {
        for (vector_p_smp_pkey_tbl::iterator it2 = (*it).begin(); it2 != (*it).end(); ++it2)
            delete (*it2);
        (*it).clear();
    }

    for (vector_v_smp_guid_tbl::iterator it = this->smp_guid_tbl_v_vector.begin();
            it != this->smp_guid_tbl_v_vector.end();
            ++it) {
        for (vector_p_smp_guid_tbl::iterator it2 = (*it).begin(); it2 != (*it).end(); ++it2)
            delete (*it2);
        (*it).clear();
    }
    for (vector_v_smp_vport_state::iterator it = this->smp_vport_state_vector.begin();
            it < this->smp_vport_state_vector.end(); ++it) {
        for (vector_p_smp_vport_state::iterator it2 = (*it).begin(); it2 != (*it).end(); ++it2)
            delete (*it2);
        (*it).clear();
    }


    //make sure all vectors / lists are empty
    this->nodes_vector.clear();
    this->ports_vector.clear();

    this->sm_info_obj_list.clear();

    this->smp_node_info_vector.clear();
    this->smp_virtual_info_vector.clear();
    this->smp_vnode_info_vector.clear();
    this->smp_switch_info_vector.clear();
    this->smp_port_info_vector.clear();
    this->smp_vport_info_vector.clear();
    this->smp_port_info_ext_vector.clear();
    this->smp_mlnx_ext_port_info_vector.clear();
    this->vs_general_info_vector.clear();
    this->pm_cap_mask_vector.clear();
    this->pm_info_obj_vector.clear();

    this->smp_pkey_tbl_v_vector.clear();
    this->smp_guid_tbl_v_vector.clear();
    this->smp_vport_state_vector.clear();
    IBDIAG_RETURN_VOID;
}


IBDMExtendedInfo::IBDMExtendedInfo()
{
    IBDIAG_ENTER;
    //do nothing
    IBDIAG_RETURN_VOID;
}


IBDMExtendedInfo::~IBDMExtendedInfo()
{
    IBDIAG_ENTER;
    this->CleanUpInternalDB();
    IBDIAG_RETURN_VOID;
}


const char* IBDMExtendedInfo::GetLastError()
{
    IBDIAG_ENTER;
    if (this->last_error != "")
        IBDIAG_RETURN(this->last_error.c_str());
    IBDIAG_RETURN("Unknown");
}


/**************/
IBNode * IBDMExtendedInfo::getNodePtr(u_int32_t node_index)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVec<vector_p_node, IBNode>(this->nodes_vector, node_index)));
}

IBPort * IBDMExtendedInfo::getPortPtr(u_int32_t port_index)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVec<vector_p_port, IBPort>(this->ports_vector, port_index)));
}

IBVNode * IBDMExtendedInfo::getVNodePtr(u_int32_t vnode_index)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVec<vector_p_vnode, IBVNode>(this->vnodes_vector, vnode_index)));
}

IBVPort * IBDMExtendedInfo::getVPortPtr(u_int32_t vport_index)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVec<vector_p_vport, IBVPort>(this->vports_vector, vport_index)));
}


/**************/
//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDMExtendedInfo::addSMPNodeInfo(IBNode *p_node, struct SMP_NodeInfo &smpNodeInfo)
{
    IBDIAG_ENTER;

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "addSMPNodeInfo node GUID " U64H_FMT"\n",
            smpNodeInfo.NodeGUID);

    IBDIAG_RETURN(this->addDataToVec(this->nodes_vector,
            p_node,
            this->smp_node_info_vector,
            smpNodeInfo));
}

struct SMP_NodeInfo * IBDMExtendedInfo::getSMPNodeInfo(u_int32_t node_index)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVec<vector_p_smp_node_info, struct SMP_NodeInfo>(this->smp_node_info_vector, node_index)));
}


/**************/
//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDMExtendedInfo::addSMPSwitchInfo(IBNode *p_node, struct SMP_SwitchInfo &smpSwitchInfo)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(this->addDataToVec(this->nodes_vector,
            p_node,
            this->smp_switch_info_vector,
            smpSwitchInfo));
}

struct SMP_SwitchInfo * IBDMExtendedInfo::getSMPSwitchInfo(u_int32_t node_index)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVec<vector_p_smp_switch_info, struct SMP_SwitchInfo>(this->smp_switch_info_vector, node_index)));
}


/**************/
//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDMExtendedInfo::addSMPPortInfo(IBPort *p_port, struct SMP_PortInfo &smpPortInfo)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(this->addDataToVec(this->ports_vector,
            p_port,
            this->smp_port_info_vector,
            smpPortInfo));
}

struct SMP_PortInfo * IBDMExtendedInfo::getSMPPortInfo(u_int32_t port_index)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVec<vector_p_smp_port_info, struct SMP_PortInfo>(this->smp_port_info_vector, port_index)));
}


/**************/
//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int
IBDMExtendedInfo::addSMPPortInfoExtended(
    IBPort *p_port, struct SMP_PortInfoExtended &smpPortInfoExt)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(this->addDataToVec(this->ports_vector,
            p_port,
            this->smp_port_info_ext_vector,
            smpPortInfoExt));
}

struct SMP_PortInfoExtended *
IBDMExtendedInfo::getSMPPortInfoExtended(u_int32_t port_index)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(
        (this->getPtrFromVec<vector_p_smp_port_info_ext,
                             struct SMP_PortInfoExtended>(
                                 this->smp_port_info_ext_vector, port_index)));
}


/**************/
//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDMExtendedInfo::addSMPMlnxExtPortInfo(IBPort *p_port, struct SMP_MlnxExtPortInfo &smpMlnxExtPortInfo)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(this->addDataToVec(this->ports_vector,
            p_port,
            this->smp_mlnx_ext_port_info_vector,
            smpMlnxExtPortInfo));
}

struct SMP_MlnxExtPortInfo * IBDMExtendedInfo::getSMPMlnxExtPortInfo(u_int32_t port_index)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVec<vector_p_smp_mlnx_ext_port_info, struct SMP_MlnxExtPortInfo>(this->smp_mlnx_ext_port_info_vector, port_index)));
}

//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDMExtendedInfo::addMlnxCntrsObject(IBPort *p_port)
{
    IBDIAG_ENTER;

    if (this->vs_mlnx_cntrs_obj_vector.empty() ||
        (this->vs_mlnx_cntrs_obj_vector.size() < p_port->createIndex + 1))
       for (int i = (int)this->vs_mlnx_cntrs_obj_vector.size();
            i < (int)p_port->createIndex + 1; ++i)
            this->vs_mlnx_cntrs_obj_vector.push_back(NULL);

    if (!this->vs_mlnx_cntrs_obj_vector[p_port->createIndex]) {
        this->vs_mlnx_cntrs_obj_vector[p_port->createIndex] =
            new vs_mlnx_cntrs_obj_t;
        if (!this->vs_mlnx_cntrs_obj_vector[p_port->createIndex]) {
            this->SetLastError("Failed to allocate vs_mlnx_cntrs_obj_t");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
        }
        CLEAR_STRUCT(*(this->vs_mlnx_cntrs_obj_vector[p_port->createIndex]));
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

struct VS_DiagnosticData * IBDMExtendedInfo::getVSDiagnosticCountersPage0(
                                                    u_int32_t port_index)
{
    IBDIAG_ENTER;
    //not exists
    if (port_index < 0 ||
        (this->vs_mlnx_cntrs_obj_vector.size() < port_index + 1))
        IBDIAG_RETURN(NULL);
    if (!this->vs_mlnx_cntrs_obj_vector[port_index])
        IBDIAG_RETURN(NULL);
    IBDIAG_RETURN(this->vs_mlnx_cntrs_obj_vector[port_index]->p_mlnx_cntrs_p0);
}

/**************/
//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDMExtendedInfo::addVSDiagnosticCountersPage0(
    IBPort *p_port,
    struct VS_DiagnosticData &vs_mlnx_cntrs_p0)
{
    //if exist already
    if ((this->vs_mlnx_cntrs_obj_vector.size() >= p_port->createIndex + 1) &&
        this->vs_mlnx_cntrs_obj_vector[p_port->createIndex] &&
        this->vs_mlnx_cntrs_obj_vector[p_port->createIndex]->p_mlnx_cntrs_p0)
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    //add
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Adding %s for (%s, index=%u)\n",
               typeid(vs_mlnx_cntrs_p0).name(), p_port->getName().c_str(),
               p_port->createIndex);
    int rc = addMlnxCntrsObject(p_port);
    if (rc)
        IBDIAG_RETURN(rc);

    struct VS_DiagnosticData *p_vs_mlnx_cntrs_page0 =
        new struct VS_DiagnosticData;
    if (!p_vs_mlnx_cntrs_page0) {
        this->SetLastError("Failed to allocate %s",
                           typeid(vs_mlnx_cntrs_p0).name());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
    }
    *p_vs_mlnx_cntrs_page0 = vs_mlnx_cntrs_p0;
    this->vs_mlnx_cntrs_obj_vector[p_port->createIndex]->p_mlnx_cntrs_p0 =
        p_vs_mlnx_cntrs_page0;
    this->addPtrToVec(this->ports_vector, p_port);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

struct VS_DiagnosticData * IBDMExtendedInfo::getVSDiagnosticCountersPage1(
                                                    u_int32_t port_index)
{
    IBDIAG_ENTER;
    //not exists
    if (port_index < 0 ||
        (this->vs_mlnx_cntrs_obj_vector.size() < port_index + 1))
        IBDIAG_RETURN(NULL);
    if (!this->vs_mlnx_cntrs_obj_vector[port_index])
        IBDIAG_RETURN(NULL);
    IBDIAG_RETURN(this->vs_mlnx_cntrs_obj_vector[port_index]->p_mlnx_cntrs_p1);
}

/**************/
//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDMExtendedInfo::addVSDiagnosticCountersPage1(IBPort *p_port,
                                struct VS_DiagnosticData &vs_mlnx_cntrs_p1)
{
    //if exist already
    if ((this->vs_mlnx_cntrs_obj_vector.size() >= p_port->createIndex + 1) &&
        this->vs_mlnx_cntrs_obj_vector[p_port->createIndex] &&
        this->vs_mlnx_cntrs_obj_vector[p_port->createIndex]->p_mlnx_cntrs_p1)
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    //add
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Adding %s for (%s, index=%u)\n",
               typeid(vs_mlnx_cntrs_p1).name(), p_port->getName().c_str(),
               p_port->createIndex);
    int rc = addMlnxCntrsObject(p_port);
    if (rc)
        IBDIAG_RETURN(rc);

    struct VS_DiagnosticData *p_vs_mlnx_cntrs_page1 =
        new struct VS_DiagnosticData;
    if (!p_vs_mlnx_cntrs_page1) {
        this->SetLastError("Failed to allocate %s",
                           typeid(vs_mlnx_cntrs_p1).name());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
    }
    *p_vs_mlnx_cntrs_page1 = vs_mlnx_cntrs_p1;
    this->vs_mlnx_cntrs_obj_vector[p_port->createIndex]->p_mlnx_cntrs_p1 =
        p_vs_mlnx_cntrs_page1;
    this->addPtrToVec(this->ports_vector, p_port);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

struct VS_DiagnosticData * IBDMExtendedInfo::getVSDiagnosticCountersPage255(u_int32_t port_index)
{
    IBDIAG_ENTER;
    //not exists
    if (port_index < 0 || (this->vs_mlnx_cntrs_obj_vector.size() < port_index + 1))
        IBDIAG_RETURN(NULL);
    if (!this->vs_mlnx_cntrs_obj_vector[port_index])
        IBDIAG_RETURN(NULL);
    IBDIAG_RETURN(this->vs_mlnx_cntrs_obj_vector[port_index]->p_mlnx_cntrs_p255);
}

int IBDMExtendedInfo::addVSDiagnosticCountersPage255(IBPort *p_port,
                                struct VS_DiagnosticData &vs_mlnx_cntrs_p255)
{
    //if exist already
    if (this->vs_mlnx_cntrs_obj_vector.size() >= p_port->createIndex + 1 &&
            this->vs_mlnx_cntrs_obj_vector[p_port->createIndex] &&
            this->vs_mlnx_cntrs_obj_vector[p_port->createIndex]->p_mlnx_cntrs_p255)
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    //add
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Adding %s for (%s, index=%u)\n",
               typeid(vs_mlnx_cntrs_p255).name(), p_port->getName().c_str(),
               p_port->createIndex);
    int rc = addMlnxCntrsObject(p_port);
    if (rc)
        IBDIAG_RETURN(rc);

    struct VS_DiagnosticData *p_vs_mlnx_cntrs_page255 = new struct VS_DiagnosticData;
    if (!p_vs_mlnx_cntrs_page255) {
        this->SetLastError("Failed to allocate %s", typeid(vs_mlnx_cntrs_p255).name());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
    }

    *p_vs_mlnx_cntrs_page255 = vs_mlnx_cntrs_p255;
    this->vs_mlnx_cntrs_obj_vector[p_port->createIndex]->p_mlnx_cntrs_p255 =
            p_vs_mlnx_cntrs_page255;

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

/**************/
//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDMExtendedInfo::addVSGeneralInfo(IBNode *p_node, struct VendorSpec_GeneralInfo &vsGeneralInfo)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(this->addDataToVec(this->nodes_vector,
            p_node,
            this->vs_general_info_vector,
            vsGeneralInfo));
}

struct VendorSpec_GeneralInfo * IBDMExtendedInfo::getVSGeneralInfo(u_int32_t node_index)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVec<vector_p_vs_general_info, struct VendorSpec_GeneralInfo>(this->vs_general_info_vector, node_index)));
}


/**************/
//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDMExtendedInfo::addPMCapMask(IBNode *p_node, u_int16_t pm_cap_mask)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(this->addDataToVec(this->nodes_vector,
            p_node,
            this->pm_cap_mask_vector,
            pm_cap_mask));
}

u_int16_t * IBDMExtendedInfo::getPMCapMask(u_int32_t node_index)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVec<vector_p_pm_cap_mask, u_int16_t>(this->pm_cap_mask_vector, node_index)));
}

//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDMExtendedInfo::addPMOptionMask(IBNode *p_node, struct PortSampleControlOptionMask &pm_option_mask)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(this->addDataToVec(this->nodes_vector,
            p_node,
            this->pm_option_mask_vector,
            pm_option_mask));
}

struct PortSampleControlOptionMask * IBDMExtendedInfo::getPMOptionMask(u_int32_t node_index)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVec<vector_p_pm_option_mask,
                   struct PortSampleControlOptionMask>(this->pm_option_mask_vector, node_index)));
}


//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDMExtendedInfo::addPMObjectInfo(IBPort *p_port)
{
    IBDIAG_ENTER;

    if (this->pm_info_obj_vector.empty() || (this->pm_info_obj_vector.size() < p_port->createIndex + 1))
       for (int i = (int)this->pm_info_obj_vector.size(); i < (int)p_port->createIndex + 1; ++i)
            this->pm_info_obj_vector.push_back(NULL);

    if (!this->pm_info_obj_vector[p_port->createIndex]) {
        this->pm_info_obj_vector[p_port->createIndex] = new pm_info_obj_t;
        if (!this->pm_info_obj_vector[p_port->createIndex]) {
            this->SetLastError("Failed to allocate pm_info_obj_t");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
        }
        CLEAR_STRUCT(*(this->pm_info_obj_vector[p_port->createIndex]));
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

/**************/
//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDMExtendedInfo::addPMPortCounters(IBPort *p_port, struct PM_PortCounters &pmPortCounters)
{
    IBDIAG_ENTER;
    if (!p_port)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    //if exist already
    if ((this->pm_info_obj_vector.size() >= p_port->createIndex + 1) &&
            this->pm_info_obj_vector[p_port->createIndex] &&
            this->pm_info_obj_vector[p_port->createIndex]->p_port_counters)
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    //add
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Adding %s for (%s, index=%u)\n",
            typeid(pmPortCounters).name(), p_port->getName().c_str(), p_port->createIndex);

    int rc = addPMObjectInfo(p_port);
    if (rc)
        IBDIAG_RETURN(rc);
    /*
    if (this->pm_info_obj_vector.empty() || (this->pm_info_obj_vector.size() < p_port->createIndex + 1))
       for (u_int32_t i = this->pm_info_obj_vector.size(); i < p_port->createIndex + 1; ++i)
            this->pm_info_obj_vector.push_back(NULL);

    if (!this->pm_info_obj_vector[p_port->createIndex]) {
        this->pm_info_obj_vector[p_port->createIndex] = new pm_info_obj_t;
        if (!this->pm_info_obj_vector[p_port->createIndex]) {
            this->SetLastError("Failed to allocate pm_info_obj_t");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
        }
        CLEAR_STRUCT(*(this->pm_info_obj_vector[p_port->createIndex]));
    }
    */

    struct PM_PortCounters *p_curr_port_counters = new struct PM_PortCounters;
    if (!p_curr_port_counters) {
        this->SetLastError("Failed to allocate %s", typeid(pmPortCounters).name());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
    }
    *p_curr_port_counters = pmPortCounters;
    this->pm_info_obj_vector[p_port->createIndex]->p_port_counters = p_curr_port_counters;
    this->addPtrToVec(this->ports_vector, p_port);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

struct PM_PortCounters * IBDMExtendedInfo::getPMPortCounters(u_int32_t port_index)
{
    IBDIAG_ENTER;
    //not exists
    if (port_index < 0 || (this->pm_info_obj_vector.size() < port_index + 1))
        IBDIAG_RETURN(NULL);
    if (!this->pm_info_obj_vector[port_index])
        IBDIAG_RETURN(NULL);
    IBDIAG_RETURN(this->pm_info_obj_vector[port_index]->p_port_counters);
}


/**************/
//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDMExtendedInfo::addPMPortCountersExtended(IBPort *p_port, struct PM_PortCountersExtended &pmPortCountersExtended)
{
    IBDIAG_ENTER;
    if (!p_port)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    //if exist already
    if ((this->pm_info_obj_vector.size() >= p_port->createIndex + 1) &&
            this->pm_info_obj_vector[p_port->createIndex] &&
            this->pm_info_obj_vector[p_port->createIndex]->p_extended_port_counters)
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    //add
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Adding %s for (%s, index=%u)\n",
            typeid(pmPortCountersExtended).name(), p_port->getName().c_str(), p_port->createIndex);
    int rc = addPMObjectInfo(p_port);
    if (rc)
        IBDIAG_RETURN(rc);
    /*
    if (this->pm_info_obj_vector.empty() || (this->pm_info_obj_vector.size() < p_port->createIndex + 1))
       for (u_int32_t i = this->pm_info_obj_vector.size(); i < p_port->createIndex + 1; ++i)
            this->pm_info_obj_vector.push_back(NULL);

    if (!this->pm_info_obj_vector[p_port->createIndex]) {
        this->pm_info_obj_vector[p_port->createIndex] = new pm_info_obj_t;
        if (!this->pm_info_obj_vector[p_port->createIndex]) {
            this->SetLastError("Failed to allocate pm_info_obj_t");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
        }
        CLEAR_STRUCT(*(this->pm_info_obj_vector[p_port->createIndex]));
    }
    */

    struct PM_PortCountersExtended *p_curr_port_counters_extended = new struct PM_PortCountersExtended;
    if (!p_curr_port_counters_extended) {
        this->SetLastError("Failed to allocate %s", typeid(pmPortCountersExtended).name());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
    }
    *p_curr_port_counters_extended = pmPortCountersExtended;
    this->pm_info_obj_vector[p_port->createIndex]->p_extended_port_counters = p_curr_port_counters_extended;
    this->addPtrToVec(this->ports_vector, p_port);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

struct PM_PortCountersExtended * IBDMExtendedInfo::getPMPortCountersExtended(u_int32_t port_index)
{
    IBDIAG_ENTER;
    //not exists
    if (port_index < 0 || (this->pm_info_obj_vector.size() < port_index + 1))
        IBDIAG_RETURN(NULL);
    if (!this->pm_info_obj_vector[port_index])
        IBDIAG_RETURN(NULL);
    IBDIAG_RETURN(this->pm_info_obj_vector[port_index]->p_extended_port_counters);
}

/**************/
//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDMExtendedInfo::addPMPortExtSpeedsCounters(IBPort *p_port, struct PM_PortExtendedSpeedsCounters &pmPortExtendedSpeedsCounters)
{
    IBDIAG_ENTER;
    if (!p_port)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    //if exist already
    if ((this->pm_info_obj_vector.size() >= p_port->createIndex + 1) &&
            this->pm_info_obj_vector[p_port->createIndex] &&
            this->pm_info_obj_vector[p_port->createIndex]->p_port_ext_speeds_counters)
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    //add
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Adding %s for (%s, index=%u)\n",
            typeid(pmPortExtendedSpeedsCounters).name(), p_port->getName().c_str(), p_port->createIndex);

    int rc = addPMObjectInfo(p_port);
    if (rc)
        IBDIAG_RETURN(rc);
    /*
    if (this->pm_info_obj_vector.empty() || (this->pm_info_obj_vector.size() < p_port->createIndex + 1))
       for (u_int32_t i = this->pm_info_obj_vector.size(); i < p_port->createIndex + 1; ++i)
            this->pm_info_obj_vector.push_back(NULL);

    if (!this->pm_info_obj_vector[p_port->createIndex]) {
        this->pm_info_obj_vector[p_port->createIndex] = new pm_info_obj_t;
        if (!this->pm_info_obj_vector[p_port->createIndex]) {
            this->SetLastError("Failed to allocate pm_info_obj_t");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
        }
        CLEAR_STRUCT(*(this->pm_info_obj_vector[p_port->createIndex]));
    }
    */

    struct PM_PortExtendedSpeedsCounters *p_port_ext_speeds_counters = new struct PM_PortExtendedSpeedsCounters;
    if (!p_port_ext_speeds_counters) {
        this->SetLastError("Failed to allocate %s", typeid(pmPortExtendedSpeedsCounters).name());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
    }
    *p_port_ext_speeds_counters = pmPortExtendedSpeedsCounters;
    this->pm_info_obj_vector[p_port->createIndex]->p_port_ext_speeds_counters = p_port_ext_speeds_counters;
    this->addPtrToVec(this->ports_vector, p_port);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

struct PM_PortExtendedSpeedsCounters * IBDMExtendedInfo::getPMPortExtSpeedsCounters(u_int32_t port_index)
{
    IBDIAG_ENTER;
    //not exists
    if (port_index < 0 || (this->pm_info_obj_vector.size() < port_index + 1))
        IBDIAG_RETURN(NULL);
    if (!this->pm_info_obj_vector[port_index])
        IBDIAG_RETURN(NULL);
    IBDIAG_RETURN(this->pm_info_obj_vector[port_index]->p_port_ext_speeds_counters);
}

/**************/
//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDMExtendedInfo::addPMPortExtSpeedsRSFECCounters(IBPort *p_port,
    struct PM_PortExtendedSpeedsRSFECCounters &pmPortExtSpeedsRSFECCounters)
{
    IBDIAG_ENTER;
    if (!p_port)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    //if exist already
    if ((this->pm_info_obj_vector.size() >= p_port->createIndex + 1) &&
        this->pm_info_obj_vector[p_port->createIndex] &&
        this->pm_info_obj_vector[p_port->createIndex]->
                                            p_port_ext_speeds_counters)
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    //add
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Adding %s for (%s, index=%u)\n",
               typeid(pmPortExtSpeedsRSFECCounters).name(),
               p_port->getName().c_str(), p_port->createIndex);

    int rc = addPMObjectInfo(p_port);
    if (rc)
        IBDIAG_RETURN(rc);

    struct PM_PortExtendedSpeedsRSFECCounters
        *p_port_ext_speeds_rsfec_counters =
        new struct PM_PortExtendedSpeedsRSFECCounters;
    if (!p_port_ext_speeds_rsfec_counters) {
        this->SetLastError("Failed to allocate %s",
                           typeid(pmPortExtSpeedsRSFECCounters).name());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
    }
    *p_port_ext_speeds_rsfec_counters = pmPortExtSpeedsRSFECCounters;
    this->pm_info_obj_vector[p_port->createIndex]->
        p_port_ext_speeds_rsfec_counters = p_port_ext_speeds_rsfec_counters;
    this->addPtrToVec(this->ports_vector, p_port);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

}

struct PM_PortExtendedSpeedsRSFECCounters *
IBDMExtendedInfo::getPMPortExtSpeedsRSFECCounters(u_int32_t port_index)
{
    IBDIAG_ENTER;
    //not exists
    if (port_index < 0 || (this->pm_info_obj_vector.size() < port_index + 1))
        IBDIAG_RETURN(NULL);
    if (!this->pm_info_obj_vector[port_index])
        IBDIAG_RETURN(NULL);
    IBDIAG_RETURN(this->pm_info_obj_vector[port_index]->
                  p_port_ext_speeds_rsfec_counters);
}

/**************/
//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDMExtendedInfo::addVSPortLLRStatistics(IBPort *p_port, struct VendorSpec_PortLLRStatistics &vsPortLLRCounters)
{
    IBDIAG_ENTER;
    if (!p_port)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    //if exist already
    if ((this->pm_info_obj_vector.size() >= p_port->createIndex + 1) &&
            this->pm_info_obj_vector[p_port->createIndex] &&
            this->pm_info_obj_vector[p_port->createIndex]->p_port_llr_statistics)
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    //add
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Adding %s for (%s, index=%u)\n",
            typeid(vsPortLLRCounters).name(), p_port->getName().c_str(), p_port->createIndex);

    int rc = addPMObjectInfo(p_port);
    if (rc)
        IBDIAG_RETURN(rc);

    struct VendorSpec_PortLLRStatistics *p_port_llr_statistics = new struct VendorSpec_PortLLRStatistics;
    if (!p_port_llr_statistics) {
        this->SetLastError("Failed to allocate %s", typeid(vsPortLLRCounters).name());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
    }
    *p_port_llr_statistics = vsPortLLRCounters;
    this->pm_info_obj_vector[p_port->createIndex]->p_port_llr_statistics = p_port_llr_statistics;
    this->addPtrToVec(this->ports_vector, p_port);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

struct VendorSpec_PortLLRStatistics * IBDMExtendedInfo::getVSPortLLRStatistics(u_int32_t port_index)
{
    IBDIAG_ENTER;
    //not exists
    if (port_index < 0 || (this->pm_info_obj_vector.size() < port_index + 1))
        IBDIAG_RETURN(NULL);
    if (!this->pm_info_obj_vector[port_index])
        IBDIAG_RETURN(NULL);
    IBDIAG_RETURN(this->pm_info_obj_vector[port_index]->p_port_llr_statistics);
}

// PortRcvErrorDetails and PortXmitDiscardDetails

int IBDMExtendedInfo::addPMPortRcvErrorDetails(IBPort *p_port,
        struct PM_PortRcvErrorDetails &pmPortRcvErrorDetails)
{
    IBDIAG_ENTER;
    if (!p_port)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    //if exist already
    if ((this->pm_info_obj_vector.size() >= p_port->createIndex + 1) &&
         this->pm_info_obj_vector[p_port->createIndex] &&
         this->pm_info_obj_vector[p_port->createIndex]->p_port_rcv_error_details)
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    //add
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Adding %s for (%s, index=%u)\n",
               typeid(pmPortRcvErrorDetails).name(), p_port->getName().c_str(), p_port->createIndex);

    int rc = addPMObjectInfo(p_port);
    if (rc)
        IBDIAG_RETURN(rc);

    struct PM_PortRcvErrorDetails *p_curr_port_rcv_error_details = new struct PM_PortRcvErrorDetails;
    if (!p_curr_port_rcv_error_details)
    {
        this->SetLastError("Failed to allocate %s", typeid(pmPortRcvErrorDetails).name());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
    }

    *p_curr_port_rcv_error_details = pmPortRcvErrorDetails;
    this->pm_info_obj_vector[p_port->createIndex]->p_port_rcv_error_details =
            p_curr_port_rcv_error_details;
    this->addPtrToVec(this->ports_vector, p_port);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

PM_PortRcvErrorDetails* IBDMExtendedInfo::getPMPortRcvErrorDetails(u_int32_t port_index)
{
    IBDIAG_ENTER;
    // if not exists
    if (port_index < 0 || (this->pm_info_obj_vector.size() < port_index + 1))
        IBDIAG_RETURN(NULL);
    if (!this->pm_info_obj_vector[port_index])
        IBDIAG_RETURN(NULL);
    IBDIAG_RETURN(this->pm_info_obj_vector[port_index]->p_port_rcv_error_details);
}

int IBDMExtendedInfo::addPMPortXmitDiscardDetails(IBPort *p_port,
        struct PM_PortXmitDiscardDetails &pmPortXmitDiscardDetails)
{
    IBDIAG_ENTER;
    if (!p_port)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    //if exist already
    if ((this->pm_info_obj_vector.size() >= p_port->createIndex + 1) &&
         this->pm_info_obj_vector[p_port->createIndex] &&
         this->pm_info_obj_vector[p_port->createIndex]->p_port_xmit_discard_details)
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    //add
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Adding %s for (%s, index=%u)\n",
               typeid(pmPortXmitDiscardDetails).name(), p_port->getName().c_str(),
               p_port->createIndex);

    int rc = addPMObjectInfo(p_port);
    if (rc)
        IBDIAG_RETURN(rc);

    struct PM_PortXmitDiscardDetails *p_curr_port_xmit_discard_details =
            new struct PM_PortXmitDiscardDetails;
    if (!p_curr_port_xmit_discard_details)
    {
        this->SetLastError("Failed to allocate %s", typeid(pmPortXmitDiscardDetails).name());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
    }

    *p_curr_port_xmit_discard_details = pmPortXmitDiscardDetails;
    this->pm_info_obj_vector[p_port->createIndex]->p_port_xmit_discard_details =
            p_curr_port_xmit_discard_details;
    this->addPtrToVec(this->ports_vector, p_port);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

PM_PortXmitDiscardDetails* IBDMExtendedInfo::getPMPortXmitDiscardDetails(u_int32_t port_index)
{
    IBDIAG_ENTER;
    // if not exists
    if (port_index < 0 || (this->pm_info_obj_vector.size() < port_index + 1))
        IBDIAG_RETURN(NULL);
    if (!this->pm_info_obj_vector[port_index])
        IBDIAG_RETURN(NULL);
    IBDIAG_RETURN(this->pm_info_obj_vector[port_index]->p_port_xmit_discard_details);
}

/**************/
//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDMExtendedInfo::addPMPortCalculatedCounters(IBPort *p_port, struct PM_PortCalcCounters &pmPortCalcCounters)
{
    IBDIAG_ENTER;
    if (!p_port)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    //if exist already
    if ((this->pm_info_obj_vector.size() >= p_port->createIndex + 1) &&
            this->pm_info_obj_vector[p_port->createIndex] &&
            this->pm_info_obj_vector[p_port->createIndex]->p_port_calc_counters)
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    //add
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Adding %s for (%s, index=%u)\n",
            typeid(pmPortCalcCounters).name(), p_port->getName().c_str(), p_port->createIndex);

    int rc = addPMObjectInfo(p_port);
    if (rc)
        IBDIAG_RETURN(rc);

    struct PM_PortCalcCounters *p_port_calc_counters = new struct PM_PortCalcCounters;
    if (!p_port_calc_counters) {
        this->SetLastError("Failed to allocate %s", typeid(pmPortCalcCounters).name());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
    }
    *p_port_calc_counters = pmPortCalcCounters;
    this->pm_info_obj_vector[p_port->createIndex]->p_port_calc_counters = p_port_calc_counters;
    this->addPtrToVec(this->ports_vector, p_port);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

struct PM_PortCalcCounters * IBDMExtendedInfo::getPMPortCalcCounters(u_int32_t port_index)
{
    IBDIAG_ENTER;
    //not exists
    if (port_index < 0 || (this->pm_info_obj_vector.size() < port_index + 1))
        IBDIAG_RETURN(NULL);
    if (!this->pm_info_obj_vector[port_index])
        IBDIAG_RETURN(NULL);
    IBDIAG_RETURN(this->pm_info_obj_vector[port_index]->p_port_calc_counters);
}


/**************/
//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDMExtendedInfo::addSMPSMInfoObj(IBPort *p_port, struct SMP_SMInfo &smpSMInfo)
{
    IBDIAG_ENTER;
    if (!p_port)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);
    sm_info_obj_t *p_curr_sm_info = new sm_info_obj_t;
    if (!p_curr_sm_info) {
        this->SetLastError("Failed to allocate sm_info_t");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
    }
    p_curr_sm_info->smp_sm_info = smpSMInfo;
    p_curr_sm_info->p_port = p_port;
    this->sm_info_obj_list.push_back(p_curr_sm_info);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


/**************/
//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDMExtendedInfo::addSMPPKeyTable(IBPort *p_port, struct SMP_PKeyTable &smpPKeyTable, u_int32_t block_idx)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(this->addDataToVecInVec(this->ports_vector,
            p_port,
            this->smp_pkey_tbl_v_vector,
            block_idx,
            smpPKeyTable));
}

struct SMP_PKeyTable * IBDMExtendedInfo::getSMPPKeyTable(u_int32_t port_index, u_int32_t block_idx)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVecInVec<vector_v_smp_pkey_tbl, struct SMP_PKeyTable>(this->smp_pkey_tbl_v_vector, port_index, block_idx)));
}


/**************/
//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDMExtendedInfo::addSMPGUIDInfo(IBPort *p_port, struct SMP_GUIDInfo &smpGUIDInfo, u_int32_t block_idx)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(this->addDataToVecInVec(this->ports_vector,
            p_port,
            this->smp_guid_tbl_v_vector,
            block_idx,
            smpGUIDInfo));
}

struct SMP_GUIDInfo * IBDMExtendedInfo::getSMPGUIDInfo(u_int32_t port_index, u_int32_t block_idx)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVecInVec<vector_v_smp_guid_tbl, struct SMP_GUIDInfo>(this->smp_guid_tbl_v_vector, port_index, block_idx)));
}

int IBDMExtendedInfo::applySubCluster()
{
    IBDIAG_ENTER;
    uint8_t i; //port index

    for (vector_p_node::iterator nI = nodes_vector.begin();
            nI != nodes_vector.end(); nI++) {
        IBNode *p_curr_node = *nI;
        //nodes can be removed from vector by previous call to ApplySubCluster
        if (!p_curr_node)
            continue;

        i = 0;
        if (p_curr_node->type == IB_CA_NODE)
            i = 1;

        //delete node and/or unrelevant ports to the path
        //delete (mark as NULL) this node's ports
        for (;i <= p_curr_node->numPorts; i++) {
            IBPort *p_curr_port = p_curr_node->getPort(i);
            //ports can be removed from vector by previous call
            //to ApplySubCluster
            if(!p_curr_port)
                continue;
            //node/port isn't in subcluster ,need to removed this node
            if (!p_curr_node->getInSubFabric() ||
                !p_curr_port->getInSubFabric()) {
                //disconnect of the link that involves this port is done
                //in the Fabric.cpp applySubCluster
                if (p_curr_port->createIndex < ports_vector.size())
                    ports_vector[p_curr_port->createIndex] = NULL;
            }
        }
        //if node isn't in sub_node, need to mark it as NULL in vector
        if(!p_curr_node->getInSubFabric())
            *nI = NULL;
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


int IBDMExtendedInfo::addSMPVirtualizationInfo(
        IBPort *p_port,
        struct SMP_VirtualizationInfo &virtual_info)
{
    IBDIAG_ENTER;

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "addSMPVirtualizationInfo port GUID "
               U64H_FMT"\n",
               p_port->guid_get());

    IBDIAG_RETURN(this->addDataToVec(this->ports_vector,
            p_port,
            this->smp_virtual_info_vector,
            virtual_info));
}

struct SMP_VirtualizationInfo * IBDMExtendedInfo::getSMPVirtualizationInfo(
                                                    u_int32_t port_index)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((
            this->getPtrFromVec<vector_p_smp_virtual_info,
                                struct SMP_VirtualizationInfo>(
                                            this->smp_virtual_info_vector,
                                            port_index)));
}



/**************/
//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDMExtendedInfo::addSMPVPortState(IBPort *p_port,
                                       struct SMP_VPortState &smpVPortState,
                                       u_int32_t block_idx)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(this->addDataToVecInVec(this->ports_vector,
            p_port,
            this->smp_vport_state_vector,
            block_idx,
            smpVPortState));
}

struct SMP_VPortState * IBDMExtendedInfo::getSMPVPortState(u_int32_t port_index,
                                                           u_int32_t block_idx)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((
            this->getPtrFromVecInVec<vector_v_smp_vport_state,
                                     struct SMP_VPortState>(
                                                this->smp_vport_state_vector,
                                                port_index, block_idx)));
}

uint8_t IBDMExtendedInfo::getSMPVPortStateVectorSize(u_int32_t port_index)
{
   IBDIAG_ENTER;
   uint8_t vec_size = 0;
   if (this->smp_vport_state_vector.size() > port_index)
      vec_size = (uint8_t)(this->smp_vport_state_vector[port_index].size());
   IBDIAG_RETURN(vec_size);
}

/**************/
//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDMExtendedInfo::addSMPVPortInfo(IBVPort *p_vport,
                                      struct SMP_VPortInfo &smpVPortInfo)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(this->addDataToVec(this->vports_vector,
            p_vport,
            this->smp_vport_info_vector,
            smpVPortInfo));
}

struct SMP_VPortInfo * IBDMExtendedInfo::getSMPVPortInfo(u_int32_t port_index)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((
            this->getPtrFromVec<vector_p_smp_vport_info,
                                struct SMP_VPortInfo>(
                                    this->smp_vport_info_vector,
                                    port_index)));
}

/**************/
//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDMExtendedInfo::addSMPVNodeInfo(IBVNode *p_vnode,
                                      struct SMP_VNodeInfo &smpVNodeInfo)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(this->addDataToVec(this->vnodes_vector,
            p_vnode,
            this->smp_vnode_info_vector,
            smpVNodeInfo));
}

struct SMP_VNodeInfo * IBDMExtendedInfo::getSMPVNodeInfo(u_int32_t vnode_index)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((
            this->getPtrFromVec<vector_p_smp_vnode_info,
                                struct SMP_VNodeInfo>(
                                    this->smp_vnode_info_vector,
                                    vnode_index)));
}

/**************/
//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDMExtendedInfo::addSMPVPortPKeyTable(IBVPort *p_vport, struct SMP_PKeyTable &smpVPortPKeyTable, u_int16_t block_idx)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(this->addDataToVecInVec(this->vports_vector,
            p_vport,
            this->smp_vport_pkey_tbl_v_vector,
            block_idx,
            smpVPortPKeyTable));
}

struct SMP_PKeyTable * IBDMExtendedInfo::getSMPVPortPKeyTable(u_int32_t vport_index, u_int16_t block_idx)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVecInVec<vector_v_smp_pkey_tbl,
                   struct SMP_PKeyTable>(this->smp_vport_pkey_tbl_v_vector, vport_index, block_idx)));
}

/**************/
//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDMExtendedInfo::addSMPQosConfigSL(IBPort *p_port, SMP_QosConfigSL &qos_config_sl)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(this->addDataToVec(this->ports_vector,
                                     p_port,
                                     this->smp_qos_config_sl_vector,
                                     qos_config_sl));
}

struct SMP_QosConfigSL * IBDMExtendedInfo::getSMPQosConfigSL(u_int32_t port_index)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVec
                   <vector_p_smp_qos_config_sl, struct SMP_QosConfigSL>(
                   this->smp_qos_config_sl_vector, port_index)));
}

/**************/
//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDMExtendedInfo::addSMPVPortQosConfigSL(IBVPort *p_vport, SMP_QosConfigSL &qos_config_sl)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(this->addDataToVec(this->vports_vector,
                                     p_vport,
                                     this->smp_vport_qos_config_sl_vector,
                                     qos_config_sl));
}

struct SMP_QosConfigSL * IBDMExtendedInfo::getSMPVPortQosConfigSL(u_int32_t port_index)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVec
                   <vector_p_smp_qos_config_sl, struct SMP_QosConfigSL>(
                   this->smp_vport_qos_config_sl_vector, port_index)));
}

/**************/
//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDMExtendedInfo::addSMPTempSensing(IBNode *p_node, SMP_TempSensing &p_temp_sense)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(this->addDataToVec(this->nodes_vector,
            p_node,
            this->smp_temp_sensing_vector,
            p_temp_sense));
}

struct SMP_TempSensing * IBDMExtendedInfo::getSMPTempSensing(u_int32_t node_index)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVec<vector_p_smp_temp_sensing,
                                       struct SMP_TempSensing>(
                                           this->smp_temp_sensing_vector,
                                           node_index)));
}

int IBDMExtendedInfo::addSMPRouterInfo(IBNode *p_node,
                                       struct SMP_RouterInfo &smpRouterInfo)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(this->addDataToVec(this->nodes_vector,
            p_node,
            this->smp_router_info_vector,
            smpRouterInfo));
}

struct SMP_RouterInfo * IBDMExtendedInfo::getSMPRouterInfo(u_int32_t node_index)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVec<vector_p_smp_router_info,
                                       struct SMP_RouterInfo>(
                                           this->smp_router_info_vector,
                                           node_index)));
}

int IBDMExtendedInfo::addSMPAdjSiteLocalSubnTbl(
        IBNode *p_node,
        struct SMP_AdjSiteLocalSubnTbl &smpRouterTbl,
        u_int8_t block_idx) {
    IBDIAG_ENTER;
    IBDIAG_RETURN(this->addDataToVecInVec(this->nodes_vector,
            p_node,
            this->smp_adj_router_tbl_v_vector,
            block_idx,
            smpRouterTbl));
}

struct SMP_AdjSiteLocalSubnTbl *
IBDMExtendedInfo::getSMPAdjSiteLocalSubnTbl(u_int32_t node_index,
                                            u_int8_t block_idx)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVecInVec<vector_v_smp_adj_router_tbl,
                                            struct SMP_AdjSiteLocalSubnTbl>(
                                                this->smp_adj_router_tbl_v_vector,
                                                node_index, block_idx)));
}

int IBDMExtendedInfo::addSMPNextHopTbl(
        IBNode *p_node,
        struct SMP_NextHopTbl &smpRouterTbl,
        u_int32_t block_idx) {
    IBDIAG_ENTER;
    IBDIAG_RETURN(this->addDataToVecInVec(this->nodes_vector,
                                          p_node,
                                          this->smp_nh_router_tbl_v_vector,
                                          block_idx,
                                          smpRouterTbl));
}

struct SMP_NextHopTbl *
IBDMExtendedInfo::getSMPNextHopTbl(u_int32_t node_index,
                                   u_int32_t block_idx)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVecInVec<vector_v_smp_next_hop_router_tbl,
                                            struct SMP_NextHopTbl>(
                                                this->smp_nh_router_tbl_v_vector,
                                                node_index, block_idx)));
}
