/*
 *  Copyright (c) 2004-2019 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include "infiniband/ibdiag/ibdiag.h"
#include "infiniband/ibdiag/ibdiag_clbck.h"

extern IBDiagClbck ibDiagClbck;


/****************************************************/
typedef map < u_int16_t, u_int8_t > map_pkey_membership;
typedef map < IBPort *, u_int8_t > map_p_port_membership;
typedef map < u_int16_t, map_p_port_membership > map_pkey_p_port_membership;
typedef vector < struct P_Key_Block_Element > vec_pkey_block_elem;

/****************************************************/
void readPortPartitionTableToMap(IBDMExtendedInfo *extended_info,
        get_pkey_table_func_t get_pkey_tbl,
        u_int32_t createIndex, u_int16_t cap,
        map_pkey_membership &pkeys_to_mem)
{
    IBDIAG_ENTER;
    struct SMP_PKeyTable *p_pkey_tbl;
    int num_of_blocks;
    int tbl_size;
    u_int16_t pkey;

    num_of_blocks = (cap + IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS - 1) /
            IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS;
    tbl_size = IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS;

    for (u_int16_t i = 0; i < num_of_blocks; ++i) {
        p_pkey_tbl = (extended_info->*get_pkey_tbl)(createIndex, i);
        if (!p_pkey_tbl)
            continue;

        if ((i + 1) * IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS  > cap)
            tbl_size = cap % IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS;

        for (int j = 0; j < tbl_size; ++j) {
            pkey = p_pkey_tbl->PKey_Entry[j].P_KeyBase;
            if (pkey == 0)
                continue;

            if (p_pkey_tbl->PKey_Entry[j].Membership_Type) {
                pkeys_to_mem.insert(pair<u_int16_t, u_int8_t>(pkey, (u_int8_t)1));
            } else {
                pkeys_to_mem.insert(pair<u_int16_t, u_int8_t>(pkey, (u_int8_t)0));
            }
        }
    }
    IBDIAG_RETURN_VOID;
}

static void readPortPartitionTableToVec(IBDMExtendedInfo *extended_info,
        IBPort * p_port, u_int16_t cap,
        vec_pkey_block_elem &vec_pkeys)
{
    IBDIAG_ENTER;
    struct SMP_PKeyTable *p_pkey_tbl;
    int num_of_blocks;
    int tbl_size;

    vec_pkeys.clear();
    vec_pkeys.resize(cap);
    for (u_int32_t i = 0; i < cap; ++i) {
        vec_pkeys[i].P_KeyBase = 0;
        vec_pkeys[i].Membership_Type = 0;
    }
    num_of_blocks = (cap + IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS - 1) /
            IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS;
    tbl_size = IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS;

    for (int i = 0; i < num_of_blocks; ++i) {
        p_pkey_tbl = extended_info->getSMPPKeyTable(p_port->createIndex, i);
        if (!p_pkey_tbl)
            continue;

        if ((i + 1) * IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS  > cap)
            tbl_size = cap % IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS;

        for (int j = 0; j < tbl_size; ++j) {
            vec_pkeys[i * IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS + j].P_KeyBase =
                    p_pkey_tbl->PKey_Entry[j].P_KeyBase;
            vec_pkeys[i * IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS + j].Membership_Type =
                    p_pkey_tbl->PKey_Entry[j].Membership_Type;
        }
    }
    IBDIAG_RETURN_VOID;
}


static int comparePartitionKeys(map_pkey_membership &pkeys1,
        map_pkey_membership &pkeys2)
{
    IBDIAG_ENTER;
    map_pkey_membership::iterator it1, it2;
    for (it1 = pkeys1.begin(); it1 != pkeys1.end(); ++it1) {
        it2 = pkeys2.find((*it1).first);
        if (it2 == pkeys2.end() || (*it1) != (*it2))
            IBDIAG_RETURN(1);
    }
    for (it2 = pkeys2.begin(); it2 != pkeys2.end(); ++it2) {
        it1 = pkeys1.find((*it2).first);
        if (it1 == pkeys1.end() || (*it1) != (*it2))
            IBDIAG_RETURN(1);
    }
    IBDIAG_RETURN(0);
}


static string getPartitionKeysStr(map_pkey_membership &pkeys)
{
    IBDIAG_ENTER;
    map_pkey_membership::iterator it;
    string mem_desc, str = "[Pkey-MemType]: ";
    char buff [1024];
    for (it = pkeys.begin(); it != pkeys.end(); ++it) {
        if (it != pkeys.begin())
            str += ";";
        switch ((*it).second) {
        case 0:
            mem_desc = "limited";
            break;
        case 1:
            mem_desc = "Full";
            break;
        default:
            sprintf(buff, "Unknown(" U8H_FMT ")", (*it).second);
            mem_desc.assign(buff);
            break;
        }
        sprintf(buff, U8H_FMT"-%s", (*it).first, mem_desc.c_str());
        str += buff;
    }
    IBDIAG_RETURN(str);
}


/****************************************************/
/******************private methods*******************/
/****************************************************/
void IBDiag::DumpPartitionKeys(ofstream &sout)
{
    IBDIAG_ENTER;
    char line[2096];
    u_int16_t partition_cap;
    map_pkey_p_port_membership pkey_to_p_port_mem;

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node)   //node not exists
            continue;

        // We do not dump switch physical ports pkey tables
        if (p_curr_node->type == IB_SW_NODE)
            continue;

        SMP_NodeInfo *p_node_info =
                this->fabric_extended_info.getSMPNodeInfo(p_curr_node->createIndex);
        if (!p_node_info)
            continue;
        partition_cap = p_node_info->PartitionCap;
        // use unsigned int index to prevent overwrap
        for (unsigned int i = 1; i <= p_curr_node->numPorts; ++i) {
            IBPort * p_curr_port = p_curr_node->getPort((phys_port_t)i);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            map_pkey_membership pkey_to_mem;
            readPortPartitionTableToMap(&this->fabric_extended_info,
                    (get_pkey_table_func_t)(&IBDMExtendedInfo::getSMPPKeyTable),
                    p_curr_port->createIndex, partition_cap, pkey_to_mem);

            // add this port pkeys to map of pkey->port->membership
            for (map_pkey_membership::iterator it_pm = pkey_to_mem.begin();
                    it_pm != pkey_to_mem.end();
                    ++it_pm) {
                u_int16_t pkey = it_pm->first;
                u_int8_t mem = it_pm->second;

                map_p_port_membership p_port_to_mem;
                map_pkey_p_port_membership::iterator it_ppm = pkey_to_p_port_mem.find(pkey);
                if (it_ppm == pkey_to_p_port_mem.end()) {
                    p_port_to_mem.clear();
                    pkey_to_p_port_mem.insert(pair < u_int16_t, map_p_port_membership > (pkey, p_port_to_mem));
                }

                it_ppm = pkey_to_p_port_mem.find(pkey);
                map_p_port_membership::iterator it_pm1 = it_ppm->second.find(p_curr_port);
                if (it_pm1 == it_ppm->second.end()) {
                    it_ppm->second.insert(pair< IBPort*, u_int8_t >(p_curr_port, mem));
                }
            }
        }
    }

    // finally dump all
    for (map_pkey_p_port_membership::iterator it_ppm = pkey_to_p_port_mem.begin();
            it_ppm != pkey_to_p_port_mem.end(); ++it_ppm) {
        memset(line, 0, sizeof(line));
        sprintf(line, "GROUP PKey:" U16H_FMT " Hosts:%lu",
            it_ppm->first,
            (long unsigned int)it_ppm->second.size());
        sout << line << endl;

        for (map_p_port_membership::iterator it_pm = it_ppm->second.begin();
                it_pm != it_ppm->second.end(); ++it_pm) {
            string membership_str = "Limited";
            if (it_pm->second)
                membership_str = "Full";

            sprintf(line, "%s %s lid=" U16H_FMT " guid=" U64H_FMT " dev=%d",
                membership_str.c_str(),
                it_pm->first->getName().c_str(),
                it_pm->first->base_lid,
                it_pm->first->guid_get(),
                it_pm->first->p_node->devId);
            sout << "  " << line << endl;
        }
        sout << endl;
    }
    IBDIAG_RETURN_VOID;
}


/****************************************************/
/*******************public methods*******************/
/****************************************************/
//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::BuildPartitionKeysDB(list_p_fabric_general_err &pkey_errors,
        progress_func_nodes_t progress_func)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    u_int16_t partition_cap;
    struct SMP_PKeyTable pkey_table;

    ibDiagClbck.Set(this, &fabric_extended_info, &pkey_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = IBDiagSMPPkeyTableGetClbck;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    progress_bar_nodes_t progress_bar_nodes;
    CLEAR_STRUCT(progress_bar_nodes);

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node)   //node not exists
            continue;

        p_curr_node->appData1.val = 0;

        ++progress_bar_nodes.nodes_found;
        if (p_curr_node->type == IB_SW_NODE)
            ++progress_bar_nodes.sw_found;
        else
            ++progress_bar_nodes.ca_found;
        if (progress_func)
            progress_func(&progress_bar_nodes, &this->discover_progress_bar_nodes);

        if (p_curr_node->type == IB_SW_NODE) {
            SMP_SwitchInfo *p_switch_info =
                    this->fabric_extended_info.getSMPSwitchInfo(p_curr_node->createIndex);
            if (!p_switch_info)
                continue;
            partition_cap = p_switch_info->PartEnfCap;
        } else {
            SMP_NodeInfo *p_node_info =
                    this->fabric_extended_info.getSMPNodeInfo(p_curr_node->createIndex);
            if (!p_node_info)
                continue;
            partition_cap = p_node_info->PartitionCap;
        }
        int num_of_blocks =
                (partition_cap + IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS - 1) /
                IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS;

        for (unsigned int i = 1; i <= p_curr_node->numPorts; ++i) {
            phys_port_t curr_port = (phys_port_t)i;
            IBPort *p_curr_port = p_curr_node->getPort(curr_port);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            //get direct route for this port
            direct_route_t *p_curr_direct_route =
                    this->GetDirectRouteByPortGuid(p_curr_port->guid_get());
            if (!p_curr_direct_route) {
                this->SetLastError("DB error - can't find direct route to node=%s, port=%u",
                                   p_curr_node->getName().c_str(), (u_int8_t)p_curr_port->num);
                rc = IBDIAG_ERR_CODE_DB_ERR;
                goto exit;
            }

            for (u_int16_t j = 0; j < num_of_blocks; ++j) {
                clbck_data.m_data1 = p_curr_port;
                clbck_data.m_data2 = (void *)(uintptr_t)j;
                this->ibis_obj.SMPPKeyTableGetByDirect(p_curr_direct_route,
                        curr_port, j, &pkey_table, &clbck_data);
                /*
                if (this->ibis_obj.SMPPkeyTableGetByLid(lid,
                        i, j, &pkey_table)) {
                    FabricErrNodeNotRespond *p_curr_fabric_pkey_err =
                            new FabricErrNodeNotRespond(p_curr_node, "SMPPkeyTableGetByLid");
                    if (!p_curr_fabric_pkey_err) {
                        this->SetLastError("Failed to allocate FabricErrNodeNotRespond");
                        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
                    }
                    pkey_errors.push_back(p_curr_fabric_pkey_err);
                    rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
                    goto next_node;
                }

                int rc2 = this->fabric_extended_info.addSMPPKeyTable(p_curr_port, pkey_table, j);
                if (rc2) {
                    this->SetLastError("Failed to add SMPPKeyTable for port=%s, err=%s",
                            p_curr_port->getName().c_str(), this->fabric_extended_info.GetLastError());
                    IBDIAG_RETURN(rc2);
                }
                */
                if (ibDiagClbck.GetState())
                    goto exit;
                if (p_curr_node->appData1.val != 0)
                    goto next_node;
            }    //block loop
        } //port loop
next_node:
        continue;
    }  //node loop
exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!pkey_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_NO_MEM / ERR_CODE_CHECK_FAILED
int IBDiag::CheckPartitionKeys(list_p_fabric_general_err& pkey_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    map_pkey_membership node_pkeys_map;
    map_pkey_membership switch_pkeys_map;

    //mark all ports as not checked
    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port)                   //port not exists
            continue;
        p_curr_port->counter1 = 0;
    }

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node)   //node not exists
            continue;

        // not checking switch-to-switch pkeys - TODO
        if (p_curr_node->type == IB_SW_NODE)
            continue;

        struct SMP_NodeInfo *p_curr_node_info = NULL;
        struct SMP_SwitchInfo *p_remote_switch_info = NULL;
        struct SMP_PortInfo *p_remote_port_info = NULL;

        for (unsigned int i = 1; i <= p_curr_node->numPorts; ++i) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            if (!p_curr_port)
                continue;

            IBPort *p_remote_port = p_curr_port->p_remotePort;
            if (!p_remote_port)
                continue;

            if ((p_curr_port->counter1) || (p_curr_port->p_remotePort->counter1))   //already checked
                continue;
            //mark as checked
            ++p_curr_port->counter1;
            ++p_curr_port->p_remotePort->counter1;

            IBNode *p_remote_switch = p_remote_port->p_node;
            if (!p_remote_switch || p_remote_switch->type != IB_SW_NODE)
                continue;

            p_curr_node_info =
                    this->fabric_extended_info.getSMPNodeInfo(p_curr_node->createIndex);
            if (!p_curr_node_info)
                continue;
            p_remote_switch_info =
                    this->fabric_extended_info.getSMPSwitchInfo(p_remote_switch->createIndex);
            if (!p_remote_switch_info)
                continue;
            p_remote_port_info =
                    this->fabric_extended_info.getSMPPortInfo(p_remote_port->createIndex);
            if (!p_remote_port_info)
                continue;

            if (p_remote_switch->type == IB_SW_NODE &&
                    !(p_remote_port_info->PartEnfInb || p_remote_port_info->PartEnfOutb)) {
                continue;
            }

            u_int16_t node_partition_cap = p_curr_node_info->PartitionCap;
            u_int16_t switch_partition_cap = p_remote_switch_info->PartEnfCap;
            if (!switch_partition_cap)
                continue;

            node_pkeys_map.clear();
            readPortPartitionTableToMap(&this->fabric_extended_info,
                    (get_pkey_table_func_t)(&IBDMExtendedInfo::getSMPPKeyTable),
                    p_curr_port->createIndex, node_partition_cap, node_pkeys_map);

            // add all vports pkeys of this port to node_pkeys_map
            map_vportnum_vport::iterator VPortsIt = p_curr_port->VPorts.begin();

            for ( ; VPortsIt != p_curr_port->VPorts.end(); ++VPortsIt) {
                IBVPort *cur_vport = VPortsIt->second;

                if (!cur_vport) {
                    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "-E- Unexpected null vport in the vports map\n");
                    continue;
                }

                SMP_VNodeInfo *cur_vnode_info =
                        this->fabric_extended_info.getSMPVNodeInfo(cur_vport->m_p_vnode->createIndex);

                readPortPartitionTableToMap(&this->fabric_extended_info,
                        (get_pkey_table_func_t)(&IBDMExtendedInfo::getSMPVPortPKeyTable),
                        cur_vport->createIndex, cur_vnode_info->vpartition_cap, node_pkeys_map);
            }

            switch_pkeys_map.clear();
            readPortPartitionTableToMap(&this->fabric_extended_info,
                    (get_pkey_table_func_t)(&IBDMExtendedInfo::getSMPPKeyTable),
                    p_remote_port->createIndex, switch_partition_cap, switch_pkeys_map);
            if (comparePartitionKeys(switch_pkeys_map, node_pkeys_map)) {
                string pkey_str1 = getPartitionKeysStr(node_pkeys_map);
                string pkey_str2 = getPartitionKeysStr(switch_pkeys_map);
                FabricErrPKeyMismatch *p_curr_fabric_pkey_err =
                        new FabricErrPKeyMismatch(p_curr_port, p_remote_port,
                                pkey_str1, pkey_str2);
                if (!p_curr_fabric_pkey_err) {
                    this->SetLastError("Failed to allocate FabricErrPKeyMismatch");
                    IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
                }
                pkey_errors.push_back(p_curr_fabric_pkey_err);
                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
                continue;
            }
        }
    }
    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR
int IBDiag::WritePKeyFile(const char *file_name)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);
    ofstream sout;
    int rc = this->OpenFile(file_name, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);
    this->DumpPartitionKeys(sout);
    sout.close();
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::DumpPartitionKeysCSVTable(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    char line[2096];
    u_int16_t partition_cap;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    csv_out.DumpStart(SECTION_PKEY);

    stringstream sstream;
    sstream << "NodeGUID,PortGUID,LocalPortNum,BlockNum,Index,PKey,Membership" << endl;
    csv_out.WriteLine(sstream.str());

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node)   //node not exists
            continue;

        // We do not dump switch physical ports pkey tables
        if (p_curr_node->type == IB_SW_NODE)
            continue;

        SMP_NodeInfo *p_node_info =
                this->fabric_extended_info.getSMPNodeInfo(p_curr_node->createIndex);
        if (!p_node_info)
            continue;
        partition_cap = p_node_info->PartitionCap;
        for (unsigned int i = 1; i <= p_curr_node->numPorts; ++i) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            vec_pkey_block_elem vec_pkeys;
            readPortPartitionTableToVec(&this->fabric_extended_info,
                    p_curr_port, partition_cap, vec_pkeys);

            memset(line, 0, sizeof(line));
            sstream.str("");
            for (u_int32_t idx = 0; idx < vec_pkeys.size(); ++idx) {
                u_int8_t membership = vec_pkeys[idx].Membership_Type;
                u_int16_t pkey = vec_pkeys[idx].P_KeyBase;
                if (!pkey)
                    continue;

                sprintf(line, U64H_FMT ","	// node guid
                        U64H_FMT "," 	// port guid
                        U32D_FMT "," 	// port num
                        U32D_FMT "," 	// block num
                        U32D_FMT "," 	// index
                        U16H_FMT ",%d", // pkey, membership
                        p_curr_node->guid_get(),
                        p_curr_port->guid_get(),
                        i,
                        idx / 32,
                        idx % 32,
                        pkey,
                        membership);
                sstream << line << endl;
                csv_out.WriteLine(sstream.str());
            }
        }
    }

    csv_out.DumpEnd(SECTION_PKEY);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

