/*
 *  Copyright (c) 2004-2019 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include "infiniband/ibdiag/ibdiag.h"
#include "infiniband/ibdiag/ibdiag_clbck.h"
#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h>
#include <math.h>

extern IBDiagClbck ibDiagClbck;


#define IS_TO_CHECK_EXT_SPEEDS_COUNTERS(check_counters_bitset,            \
                                        node_type)                        \
((((check_counters_bitset & CHECK_EXT_SPEEDS_COUNTERS_ON_SW) && node_type == IB_SW_NODE) ||                  \
  ((check_counters_bitset & CHECK_EXT_SPEEDS_COUNTERS_ON_ALL))) ? true : false)

#define IS_TO_PRINT_LLR_COUNTERS(check_counters_bitset) \
((check_counters_bitset & PRINT_LLR_COUNTERS) ? true : false)


#define IS_SUPPORT_XMIT_WAIT(cap_mask) ((cap_mask) & 0x1000)
#define IS_SUPPORT_EXTENDED_COUNTERS(cap_mask) ((cap_mask) & 0x0400)
#define IS_SUPPORT_EXTENDED_COUNTERS_NO_IETF(cap_mask) ((cap_mask) & 0x0200)
#define IS_SUPPORT_EXTENDED_SPEEDS_COUNTERS(cap_mask) ((cap_mask) & IBIS_IB_PORT_CAP_HAS_EXT_SPEEDS)
//check that Performance Management ClassPortInfo:CapabilityMask has IsRSFECCountersSupported
#define IS_SUPPORT_EXT_SPEEDS_RSFEC_COUNTERS(cap_mask) ((cap_mask) & 0x4000)

#define MAX_LANE_NUMBER 12

/* Defines the source of counter in the Counter DB (below) */
typedef enum counter_mad_src
{
    PM_PORT_CNT_MAD = 0,
    PM_PORT_CNT_EXT_MAD,
    PM_PORT_EXT_SPEEDS_CNTS_MAD,
    PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD,
    //most of the llr shouldn't be checked
    VS_PORT_LLR_CNTS_MAD_HIDDEN,
    PM_PORT_CALCULATED_COUNTER,
    //for llr counters that are checked (max_retransmission_rate)
    VS_PORT_LLR_CNTS_MAD
} counter_mad_src_t;

/****************************************************/
/* global counters array */
typedef struct pm_counter {
    string name;
    u_int64_t overflow_value;
    u_int8_t real_size;             /* in bytes */
    u_int8_t struct_offset;         /* in bytes */
    u_int16_t cap_mask_bits;        /* which cap mask supports this counter */
    u_int8_t diff_check_threshold;  /* threshold between 2 samples */
    int16_t threshold;            /* counter threshold (currently only for retransmission rate) */
    counter_mad_src_t counter_src;  /* source of counter (in which struct to find it) */
    bool subject_to_threshold_keyword_all; /* is the counter sublect to keyword 'all' in -PC */
} pm_counter_t;


#define PM_COUNTERS_ARR_SIZE 91
static pm_counter_t pm_counters_arr[PM_COUNTERS_ARR_SIZE] = {
        /* 16 bits */
        {"symbol_error_counter",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortCounters, SymbolErrorCounter),
                0x0, 1, -1, PM_PORT_CNT_MAD, true},
        {"port_rcv_remote_physical_errors",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortCounters, PortRcvRemotePhysicalErrors),
                0x0, 1, -1, PM_PORT_CNT_MAD, true},
        {"port_rcv_errors",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortCounters, PortRcvErrors),
                0x0, 1, -1, PM_PORT_CNT_MAD, true},
        {"port_xmit_discard",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortCounters, PortXmitDiscards),
                0x0, 10, -1, PM_PORT_CNT_MAD, true},
        {"port_rcv_switch_relay_errors",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortCounters, PortRcvSwitchRelayErrors),
                0x0, 1, -1, PM_PORT_CNT_MAD, true},
        {"vl15_dropped",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortCounters, VL15Dropped),
                0x0, 10, -1, PM_PORT_CNT_MAD, true},

        /* 8 bits */
        {"link_error_recovery_counter",
                OVERFLOW_VAL_8_BIT, 1, (u_int8_t)OFFSET_OF(struct PM_PortCounters, LinkErrorRecoveryCounter),
                0x0, 1, -1, PM_PORT_CNT_MAD, true},
        {"link_down_counter",
                OVERFLOW_VAL_8_BIT, 1, (u_int8_t)OFFSET_OF(struct PM_PortCounters, LinkDownedCounter),
                0x0, 1, -1, PM_PORT_CNT_MAD, true},
        {"port_xmit_constraint_errors",
                OVERFLOW_VAL_8_BIT, 1, (u_int8_t)OFFSET_OF(struct PM_PortCounters, PortXmitConstraintErrors),
                0x0, 10, -1, PM_PORT_CNT_MAD, true},
        {"port_rcv_constraint_errors",
                OVERFLOW_VAL_8_BIT, 1, (u_int8_t)OFFSET_OF(struct PM_PortCounters, PortRcvConstraintErrors),
                0x0, 10, -1, PM_PORT_CNT_MAD, true},

        /* 4 bits - in struct the size is 1 byte */
        {"local_link_integrity_errors",
                OVERFLOW_VAL_4_BIT, 1, (u_int8_t)OFFSET_OF(struct PM_PortCounters, LocalLinkIntegrityErrors),
                0x0, 10, -1, PM_PORT_CNT_MAD, true},
        {"excessive_buffer_errors",
                OVERFLOW_VAL_4_BIT, 1, (u_int8_t)OFFSET_OF(struct PM_PortCounters, ExcessiveBufferOverrunErrors),
                0x0, 10, -1, PM_PORT_CNT_MAD, true},

        /* 32 bits */
        {"port_xmit_data",
                OVERFLOW_VAL_NONE, 4, (u_int8_t)OFFSET_OF(struct PM_PortCounters, PortXmitData),
                0x0, 0, -1, PM_PORT_CNT_MAD, false},
        {"port_rcv_data",
                OVERFLOW_VAL_NONE, 4, (u_int8_t)OFFSET_OF(struct PM_PortCounters, PortRcvData),
                0x0, 0, -1, PM_PORT_CNT_MAD, false},
        {"port_xmit_pkts",
                OVERFLOW_VAL_NONE, 4, (u_int8_t)OFFSET_OF(struct PM_PortCounters, PortXmitPkts),
                0x0, 0, -1, PM_PORT_CNT_MAD, false},
        {"port_rcv_pkts",
                OVERFLOW_VAL_NONE, 4, (u_int8_t)OFFSET_OF(struct PM_PortCounters, PortRcvPkts),
                0x0, 0, -1, PM_PORT_CNT_MAD, false},
        {"port_xmit_wait",
                OVERFLOW_VAL_NONE, 4, (u_int8_t)OFFSET_OF(struct PM_PortCounters, PortXmitWait),
                0x1000, 0, -1, PM_PORT_CNT_MAD, false},

        /* 64 bits */
        {"port_xmit_data_extended",
                OVERFLOW_VAL_NONE, 8, (u_int8_t)OFFSET_OF(struct PM_PortCountersExtended, PortXmitData),
                0x0400 | 0x0200, 0, -1, PM_PORT_CNT_EXT_MAD, false},
        {"port_rcv_data_extended",
                OVERFLOW_VAL_NONE, 8, (u_int8_t)OFFSET_OF(struct PM_PortCountersExtended, PortRcvData),
                0x0400 | 0x0200, 0, -1, PM_PORT_CNT_EXT_MAD, false},
        {"port_xmit_pkts_extended",
                OVERFLOW_VAL_NONE, 8, (u_int8_t)OFFSET_OF(struct PM_PortCountersExtended, PortXmitPkts),
                0x0400 | 0x0200, 0, -1, PM_PORT_CNT_EXT_MAD, false},
        {"port_rcv_pkts_extended",
                OVERFLOW_VAL_NONE, 8, (u_int8_t)OFFSET_OF(struct PM_PortCountersExtended, PortRcvPkts),
                0x0400 | 0x0200, 0, -1, PM_PORT_CNT_EXT_MAD, false},
        {"port_unicast_xmit_pkts",
                OVERFLOW_VAL_NONE, 8, (u_int8_t)OFFSET_OF(struct PM_PortCountersExtended, PortUniCastXmitPkts),
                0x0200, 0, -1, PM_PORT_CNT_EXT_MAD, false},
        {"port_unicast_rcv_pkts",
                OVERFLOW_VAL_NONE, 8, (u_int8_t)OFFSET_OF(struct PM_PortCountersExtended, PortUniCastRcvPkts),
                0x0200, 0, -1, PM_PORT_CNT_EXT_MAD, false},
        {"port_multicast_xmit_pkts",
                OVERFLOW_VAL_NONE, 8, (u_int8_t)OFFSET_OF(struct PM_PortCountersExtended, PortMultiCastXmitPkts),
                0x0200, 0, -1, PM_PORT_CNT_EXT_MAD, false},
        {"port_multicast_rcv_pkts",
                OVERFLOW_VAL_NONE, 8, (u_int8_t)OFFSET_OF(struct PM_PortCountersExtended, PortMultiCastRcvPkts),
                0x0200, 0, -1, PM_PORT_CNT_EXT_MAD, false},

        /* Extended speeds counters */
        {"sync_header_err_cnt",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, SyncHeaderErrorCounter),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"unknown_block_cnt",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, UnknownBlockCounter),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"error_detection_counter_lane0",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, ErrorDetectionCounterLane[0]),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"error_detection_counter_lane1",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, ErrorDetectionCounterLane[1]),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"error_detection_counter_lane2",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, ErrorDetectionCounterLane[2]),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"error_detection_counter_lane3",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, ErrorDetectionCounterLane[3]),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"error_detection_counter_lane4",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, ErrorDetectionCounterLane[4]),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"error_detection_counter_lane5",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, ErrorDetectionCounterLane[5]),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"error_detection_counter_lane6",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, ErrorDetectionCounterLane[6]),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"error_detection_counter_lane7",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, ErrorDetectionCounterLane[7]),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"error_detection_counter_lane8",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, ErrorDetectionCounterLane[8]),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"error_detection_counter_lane9",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, ErrorDetectionCounterLane[9]),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"error_detection_counter_lane10",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, ErrorDetectionCounterLane[10]),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"error_detection_counter_lane11",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, ErrorDetectionCounterLane[11]),
                0, 1, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"fec_correctable_block_counter_lane0",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECCorrectableBlockCountrLane[0]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"fec_correctable_block_counter_lane1",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECCorrectableBlockCountrLane[1]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"fec_correctable_block_counter_lane2",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECCorrectableBlockCountrLane[2]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"fec_correctable_block_counter_lane3",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECCorrectableBlockCountrLane[3]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"fec_correctable_block_counter_lane4",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECCorrectableBlockCountrLane[4]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"fec_correctable_block_counter_lane5",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECCorrectableBlockCountrLane[5]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"fec_correctable_block_counter_lane6",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECCorrectableBlockCountrLane[6]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"fec_correctable_block_counter_lane7",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECCorrectableBlockCountrLane[7]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"fec_correctable_block_counter_lane8",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECCorrectableBlockCountrLane[8]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"fec_correctable_block_counter_lane9",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECCorrectableBlockCountrLane[9]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"fec_correctable_block_counter_lane10",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECCorrectableBlockCountrLane[10]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"fec_correctable_block_counter_lane11",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECCorrectableBlockCountrLane[11]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"fec_uncorrectable_block_counter_lane0",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECUncorrectableBlockCounterLane[0]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"fec_uncorrectable_block_counter_lane1",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECUncorrectableBlockCounterLane[1]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"fec_uncorrectable_block_counter_lane2",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECUncorrectableBlockCounterLane[2]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"fec_uncorrectable_block_counter_lane3",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECUncorrectableBlockCounterLane[3]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"fec_uncorrectable_block_counter_lane4",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECUncorrectableBlockCounterLane[4]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"fec_uncorrectable_block_counter_lane5",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECUncorrectableBlockCounterLane[5]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"fec_uncorrectable_block_counter_lane6",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECUncorrectableBlockCounterLane[6]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"fec_uncorrectable_block_counter_lane7",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECUncorrectableBlockCounterLane[7]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"fec_uncorrectable_block_counter_lane8",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECUncorrectableBlockCounterLane[8]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"fec_uncorrectable_block_counter_lane9",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECUncorrectableBlockCounterLane[9]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"fec_uncorrectable_block_counter_lane10",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECUncorrectableBlockCounterLane[10]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},
        {"fec_uncorrectable_block_counter_lane11",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsCounters, FECUncorrectableBlockCounterLane[11]),
                0, 0, -1, PM_PORT_EXT_SPEEDS_CNTS_MAD, true},

        /* LLR counters */
        /* For VS_PORT_LLR_CNTS_MAD_HIDDEN counters we'll not report overflow, threshold and threshold-diff */
        {"port_rcv_cells",
                OVERFLOW_VAL_64_BIT, 8, (u_int8_t)OFFSET_OF(struct VendorSpec_PortLLRStatistics, PortRcvCells),
                0, 0, -1, VS_PORT_LLR_CNTS_MAD_HIDDEN, false},
        {"port_rcv_cell_for_retry",
                OVERFLOW_VAL_64_BIT, 8, (u_int8_t)OFFSET_OF(struct VendorSpec_PortLLRStatistics, PortRcvCellForRetry),
                0, 1, -1, VS_PORT_LLR_CNTS_MAD_HIDDEN, true},
        {"port_rcv_retry",
                OVERFLOW_VAL_64_BIT, 8, (u_int8_t)OFFSET_OF(struct VendorSpec_PortLLRStatistics, PortRcvRetry),
                0, 1, -1, VS_PORT_LLR_CNTS_MAD_HIDDEN, true},
        {"port_xmit_cells",
                OVERFLOW_VAL_64_BIT, 8, (u_int8_t)OFFSET_OF(struct VendorSpec_PortLLRStatistics, PortXmitCells),
                0, 0, -1, VS_PORT_LLR_CNTS_MAD_HIDDEN, false},
        {"port_xmit_retry_cells",
                OVERFLOW_VAL_64_BIT, 8, (u_int8_t)OFFSET_OF(struct VendorSpec_PortLLRStatistics, PortXmitRetryCells),
                0, 1, -1, VS_PORT_LLR_CNTS_MAD_HIDDEN, true},
        {"port_xmit_retry",
                OVERFLOW_VAL_64_BIT, 8, (u_int8_t)OFFSET_OF(struct VendorSpec_PortLLRStatistics, PortXmitRetry),
                0, 1, -1, VS_PORT_LLR_CNTS_MAD_HIDDEN, true},
        {"port_symbol_error",
                OVERFLOW_VAL_64_BIT, 8, (u_int8_t)OFFSET_OF(struct VendorSpec_PortLLRStatistics, PortSymbolError),
                0, 1, -1, VS_PORT_LLR_CNTS_MAD_HIDDEN, true},
        {"port_error_detection_counter_lane0",
                OVERFLOW_VAL_64_BIT, 8, (u_int8_t)OFFSET_OF(struct VendorSpec_PortLLRStatistics, PortErrorDetectionCounterLane0),
                0, 1, -1, VS_PORT_LLR_CNTS_MAD_HIDDEN, true},
        {"port_error_detection_counter_lane1",
                OVERFLOW_VAL_64_BIT, 8, (u_int8_t)OFFSET_OF(struct VendorSpec_PortLLRStatistics, PortErrorDetectionCounterLane1),
                0, 1, -1, VS_PORT_LLR_CNTS_MAD_HIDDEN, true},
        {"port_error_detection_counter_lane2",
                OVERFLOW_VAL_64_BIT, 8, (u_int8_t)OFFSET_OF(struct VendorSpec_PortLLRStatistics, PortErrorDetectionCounterLane2),
                0, 1, -1, VS_PORT_LLR_CNTS_MAD_HIDDEN, true},
        {"port_error_detection_counter_lane3",
                OVERFLOW_VAL_64_BIT, 8, (u_int8_t)OFFSET_OF(struct VendorSpec_PortLLRStatistics, PortErrorDetectionCounterLane3),
                0, 1, -1, VS_PORT_LLR_CNTS_MAD_HIDDEN, true},
        {"max_retransmission_rate",
                OVERFLOW_VAL_16_BIT, 2, (u_int8_t)OFFSET_OF(struct VendorSpec_PortLLRStatistics, MaxRetransmissionRate),
                0, 1, 500, VS_PORT_LLR_CNTS_MAD, false},
        /* Calculated Counters */
        {"retransmission_per_sec",
                OVERFLOW_VAL_64_BIT, 8, (u_int8_t)OFFSET_OF(struct PM_PortCalcCounters, RetransmissionPerSec),
                0, 1, 500, PM_PORT_CALCULATED_COUNTER, false},

        /* Extended Speeds counters when RS-FEC is active */
        {"fec_corrected_symbol_counter_lane0",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, FECCorrectedSymbolCounterLane[0]),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, true},
        {"fec_corrected_symbol_counter_lane1",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, FECCorrectedSymbolCounterLane[1]),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, true},
        {"fec_corrected_symbol_counter_lane2",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, FECCorrectedSymbolCounterLane[2]),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, true},
        {"fec_corrected_symbol_counter_lane3",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, FECCorrectedSymbolCounterLane[3]),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, true},
        {"fec_corrected_symbol_counter_lane4",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, FECCorrectedSymbolCounterLane[4]),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, true},
        {"fec_corrected_symbol_counter_lane5",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, FECCorrectedSymbolCounterLane[5]),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, true},
        {"fec_corrected_symbol_counter_lane6",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, FECCorrectedSymbolCounterLane[6]),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, true},
        {"fec_corrected_symbol_counter_lane7",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, FECCorrectedSymbolCounterLane[7]),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, true},
        {"fec_corrected_symbol_counter_lane8",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, FECCorrectedSymbolCounterLane[8]),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, true},
        {"fec_corrected_symbol_counter_lane9",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, FECCorrectedSymbolCounterLane[9]),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, true},
        {"fec_corrected_symbol_counter_lane10",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, FECCorrectedSymbolCounterLane[10]),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, true},
        {"fec_corrected_symbol_counter_lane11",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, FECCorrectedSymbolCounterLane[11]),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, true},
        {"port_fec_correctable_block_counter",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, PortFECCorrectableBlockCounter),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, true},
        {"port_fec_uncorrectable_block_counter",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, PortFECUncorrectableBlockCounter),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, true},
        {"port_fec_corrected_symbol_counter",
                OVERFLOW_VAL_32_BIT, 4, (u_int8_t)OFFSET_OF(struct PM_PortExtendedSpeedsRSFECCounters, PortFECCorrectedSymbolCounter),
                0x4000, 0, -1, PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD, true},
};

/****************************************************/
static long double CalcBERFireCodeFEC(struct PM_PortExtendedSpeedsCounters *p_counters1,
                                      struct PM_PortExtendedSpeedsCounters *p_counters2)
{
    IBDIAG_ENTER;
    long double ber = 0.0;
    long double sum_crr_cntr1= 0.0, sum_uncrr_cntr1 = 0.0;
    long double sum_crr_cntr2= 0.0, sum_uncrr_cntr2 = 0.0;

    for (uint8_t  i = 0; i < MAX_LANE_NUMBER; i++) {
        sum_uncrr_cntr1 += p_counters1->FECUncorrectableBlockCounterLane[i];
        sum_crr_cntr1 += p_counters1->FECCorrectableBlockCountrLane[i];
    }

    for (uint8_t  i = 0; i < MAX_LANE_NUMBER; i++) {
        sum_uncrr_cntr2 += p_counters2->FECUncorrectableBlockCounterLane[i];
        sum_crr_cntr2 += p_counters2->FECCorrectableBlockCountrLane[i];
    }

    ber = (sum_crr_cntr2 - sum_crr_cntr1) +
          K_FCODE *
          (sum_uncrr_cntr2 - sum_uncrr_cntr1);

    IBDIAG_RETURN(ber);
}

static long double EDPLSum(struct PM_PortExtendedSpeedsCounters *p_counters1,
                           struct PM_PortExtendedSpeedsCounters *p_counters2)
{
    IBDIAG_ENTER;
    //sum of edpls over lanes/time
    long double ber = 0.0;
    for (int i = 0; i < MAX_LANE_NUMBER; i++)
        ber += p_counters2->ErrorDetectionCounterLane[i];

    for (int i = 0; i < MAX_LANE_NUMBER; i++)
        ber -= p_counters1->ErrorDetectionCounterLane[i];

    IBDIAG_RETURN(ber);
}

static bool is_counter_supported(u_int16_t node_cap_mask, u_int16_t counter_cap_mask)
{
    IBDIAG_ENTER;
    if (counter_cap_mask == 0)              //supported
        IBDIAG_RETURN(true);
    if (node_cap_mask & counter_cap_mask)
        IBDIAG_RETURN(true);
    IBDIAG_RETURN(false);
}

//Returns: SUCCESS_CODE / IBDIAG_ERR_CODE_CHECK_FAILED
int counter_name_2_index(string name, int &index) {
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    static map < string, int > name2index;
    map < string, int >::iterator it;

    // first time the function is called
    if ( !name2index.size() ) {
        for ( int i = 0; i < PM_COUNTERS_ARR_SIZE; i++ ) {
            name2index[pm_counters_arr[i].name] = i;
        }
    }

    // check if the given name is a valid counter name
    it = name2index.find(name);

    if ( it == name2index.end() ) {
        rc = IBDIAG_ERR_CODE_CHECK_FAILED;
        index = -1;
    } else {
        index = it->second;
    }
    IBDIAG_RETURN(rc);
}

int get_value(void* buff, u_int8_t size, u_int64_t &value)
{
    IBDIAG_ENTER;
    int rc = 0;
    switch (size) {
        case 1:
            value = *(u_int8_t *)(buff);
            break;
        case 2:
            value = *(u_int16_t *)(buff);
            break;
        case 4:
            value = *(u_int32_t *)(buff);
            break;
        case 8:
            value = *(u_int64_t *)(buff);
            break;
        default:
            rc = 1;
            value = 0;
    }

    IBDIAG_RETURN(rc);
}

/****************************************************/
/******************private methods*******************/
/****************************************************/
void IBDiag::DumpPortExtendedSpeedsCounters(ofstream &sout,
                                            bool en_per_lane_cnts,
                                            IBPort *p_curr_port,
                                            u_int32_t port_idx)
{
    IBDIAG_ENTER;
    struct PM_PortExtendedSpeedsCounters *p_ext_speeds_cnts = NULL;
    struct PM_PortExtendedSpeedsRSFECCounters *p_ext_speeds_rsfec_cnts = NULL;
    char curr_counters_line[2096] = {'\0'};
    u_int32_t error_detection_cnt_total = 0;
    u_int16_t sync_header_error_counter = 0;
    u_int16_t unknown_block_counter = 0;
    u_int64_t correctable_block_cnt_total = 0;
    u_int64_t uncorrectable_block_cnt_total = 0;
    u_int64_t corrected_symbol_cnt_total = 0;

    IBLinkWidth p_curr_port_width = p_curr_port->get_common_width();
    int lane_number = LinkWidthToLane(p_curr_port_width);

    p_ext_speeds_cnts = this->fabric_extended_info.
        getPMPortExtSpeedsCounters(port_idx);
    p_ext_speeds_rsfec_cnts =
        this->fabric_extended_info.getPMPortExtSpeedsRSFECCounters(port_idx);

    // Counters not available
    if (p_ext_speeds_cnts) {
        for (int i = 0; i < lane_number; i++) {
            error_detection_cnt_total +=
                p_ext_speeds_cnts->ErrorDetectionCounterLane[i];
            correctable_block_cnt_total +=
                p_ext_speeds_cnts->FECCorrectableBlockCountrLane[i];
            uncorrectable_block_cnt_total +=
                p_ext_speeds_cnts->FECUncorrectableBlockCounterLane[i];
        }
        sync_header_error_counter =
            p_ext_speeds_cnts->SyncHeaderErrorCounter;
        unknown_block_counter = p_ext_speeds_cnts->UnknownBlockCounter;
    } else if (p_ext_speeds_rsfec_cnts) {
        for (int i = 0; i < lane_number; i++)
            corrected_symbol_cnt_total +=
                p_ext_speeds_rsfec_cnts->FECCorrectedSymbolCounterLane[i];
        sync_header_error_counter =
            p_ext_speeds_rsfec_cnts->SyncHeaderErrorCounter;
        unknown_block_counter =
            p_ext_speeds_rsfec_cnts->UnknownBlockCounter;
    }

    //common counters
    if (p_ext_speeds_cnts || p_ext_speeds_rsfec_cnts) {
        sprintf(curr_counters_line,
                "sync_header_error_counter=" U32H_FMT "\n" \
                "unknown_block_counter=" U32H_FMT "\n",
                sync_header_error_counter,
                unknown_block_counter);
        sout << curr_counters_line;
    } else {
        sprintf(curr_counters_line,
                "sync_header_error_counter=NA\n" \
                "unknown_block_counter=NA\n");
        sout << curr_counters_line;
    }


    //RS-FEC mode
    if (isRSFEC(p_curr_port->get_fec_mode())) {
        if (p_ext_speeds_rsfec_cnts) {
            sprintf(curr_counters_line,
                    "fec_corrected_symbol_counter_total=" U64H_FMT "\n",
                    corrected_symbol_cnt_total);
            sout << curr_counters_line;
            if (en_per_lane_cnts) {
                for (int i = 0; i < lane_number; i++) {
                    sprintf(curr_counters_line, "%s[%d]=" U32H_FMT "\n",
                            "fec_corrected_symbol_counter_lane", i,
                            p_ext_speeds_rsfec_cnts->
                            FECCorrectedSymbolCounterLane[i]);
                    sout << curr_counters_line;
                }
            }
            sprintf(curr_counters_line,
                    "port_fec_correctable_block_counter=" U32H_FMT "\n"\
                    "port_fec_uncorrectable_block_counter=" U32H_FMT "\n"\
                    "port_fec_corrected_symbol_counter=" U32H_FMT "\n",
                    p_ext_speeds_rsfec_cnts->PortFECCorrectableBlockCounter,
                    p_ext_speeds_rsfec_cnts->PortFECUncorrectableBlockCounter,
                    p_ext_speeds_rsfec_cnts->PortFECCorrectedSymbolCounter);
            sout << curr_counters_line;
        } else {
            sprintf(curr_counters_line,
                    "fec_corrected_symbol_counter_total=NA\n");
            sout << curr_counters_line;
            if (en_per_lane_cnts) {
                for (int i = 0; i < lane_number; i++) {
                    sprintf(curr_counters_line, "%s[%d]=NA\n",
                            "fec_corrected_symbol_counter_lane", i);
                    sout << curr_counters_line;
                }
            }
            sprintf(curr_counters_line,
                    "port_fec_correctable_block_counter=NA\n"\
                    "port_fec_uncorrectable_block_counter=NA\n"\
                    "port_fec_corrected_symbol_counter=NA\n");
            sout << curr_counters_line;
        }
    } else { //not RS-FEC
        if (p_ext_speeds_cnts) {
            sprintf(curr_counters_line,
                    "error_detection_counter_total=" U32H_FMT "\n",
                    error_detection_cnt_total);
            sout << curr_counters_line;
            //when fec is "No FEC", these counters are irrelevant
            if (p_curr_port->get_fec_mode() != IB_FEC_NO_FEC) {
                sprintf(curr_counters_line,
                        "fec_correctable_block_counter_total=" U64H_FMT "\n" \
                        "fec_uncorrectable_block_counter_total=" U64H_FMT "\n",
                        correctable_block_cnt_total,
                        uncorrectable_block_cnt_total);
                sout << curr_counters_line;
            }
            if (en_per_lane_cnts) {
                for (int i = 0; i < lane_number; i++) {
                    sprintf(
                        curr_counters_line,
                        "%s[%d]=" U32H_FMT "\n",
                        "error_detection_counter_lane", i,
                        p_ext_speeds_cnts->ErrorDetectionCounterLane[i]);
                    sout << curr_counters_line;
                }
                //when fec is "No FEC", these counters are irrelevant
                if (p_curr_port->get_fec_mode() != IB_FEC_NO_FEC) {
                    for (int i = 0; i < lane_number; i++) {
                        sprintf(
                            curr_counters_line,
                            "%s[%d]=" U32H_FMT "\n",
                            "fec_correctable_block_counter_lane", i,
                            p_ext_speeds_cnts->FECCorrectableBlockCountrLane[i]);
                        sout << curr_counters_line;
                    }
                    for (int i = 0; i < lane_number; i++) {
                        sprintf(
                            curr_counters_line, "%s[%d]=" U32H_FMT "\n",
                            "fec_uncorrectable_block_counter_lane", i,
                            p_ext_speeds_cnts->
                            FECUncorrectableBlockCounterLane[i]);
                        sout << curr_counters_line;
                    }
                }
            }
        } else {
            sout << "error_detection_counter_total=NA\n";

            //when fec is "No FEC", these counters are irrelevant
            if (p_curr_port->get_fec_mode() != IB_FEC_NO_FEC)
                sout << "fec_correctable_block_counter_total=NA\n" \
                        "fec_uncorrectable_block_counter_total=NA\n";
            if (en_per_lane_cnts) {
                for (int i = 0; i < lane_number; i++) {
                    sprintf(curr_counters_line, "%s[%d]=NA\n",
                            "error_detection_counter_lane", i);
                    sout << curr_counters_line;
                }
                //when fec is "No FEC", these counters are irrelevant
                if (p_curr_port->get_fec_mode() != IB_FEC_NO_FEC) {
                    for (int i = 0; i < lane_number; i++) {
                        sprintf(curr_counters_line, "%s[%d]=NA\n",
                                "fec_correctable_block_counter_lane", i);
                        sout << curr_counters_line;
                    }
                    for (int i = 0; i < lane_number; i++) {
                        sprintf(curr_counters_line, "%s[%d]=NA\n",
                                "fec_uncorrectable_block_counter_lane", i);
                        sout << curr_counters_line;
                    }
                }
            }
        }
    }

    IBDIAG_RETURN_VOID;
}

void IBDiag::DumpPortCounters(ofstream &sout,
                              u_int32_t check_counters_bitset,
                              bool en_per_lane_cnts)
{
    IBDIAG_ENTER;
    struct PM_PortCounters *p_curr_port_counters = NULL;
    struct PM_PortCountersExtended *p_curr_port_counters_extended = NULL;
    struct VendorSpec_PortLLRStatistics *p_curr_port_llr_statistics = NULL;
    struct PM_PortCalcCounters *p_port_calc_counters = NULL;
    struct PM_PortRcvErrorDetails *p_port_rcv_error_details = NULL;
    struct PM_PortXmitDiscardDetails *p_port_xmit_discard_details = NULL;

    char curr_counters_line[2096];
    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port)               //port not exist
            continue;

        p_curr_port_counters = this->fabric_extended_info.getPMPortCounters(i);
        if (!p_curr_port_counters)      //info not exist
            continue;

        memset(curr_counters_line, 0, sizeof(curr_counters_line));
        sprintf(curr_counters_line, "Port=" U32D_FMT    \
                " Lid=" U16H_FMT    \
                " GUID=" U64H_FMT   \
                " Device=" U32D_FMT \
                " Port Name=" STR_FMT,
                p_curr_port->num,
                p_curr_port->base_lid,
                p_curr_port->guid_get(),
                p_curr_port->p_node->devId,
                p_curr_port->getName().c_str());
        sout << "-------------------------------------------------------" << endl;
        sout << curr_counters_line << endl;
        sout << "-------------------------------------------------------" << endl;
        memset(curr_counters_line, 0, sizeof(curr_counters_line));
        sprintf(curr_counters_line,
                "link_down_counter=" U32H_FMT "\n"  \
                "link_error_recovery_counter=" U32H_FMT "\n"  \
                "symbol_error_counter=" U32H_FMT "\n"  \
                "port_rcv_remote_physical_errors=" U32H_FMT "\n"  \
                "port_rcv_errors=" U32H_FMT "\n"  \
                "port_xmit_discard=" U32H_FMT "\n"  \
                "port_rcv_switch_relay_errors=" U32H_FMT "\n"  \
                "excessive_buffer_errors=" U32H_FMT "\n"  \
                "local_link_integrity_errors=" U32H_FMT "\n"  \
                "port_rcv_constraint_errors=" U32H_FMT "\n"  \
                "port_xmit_constraint_errors=" U32H_FMT "\n"  \
                "vl15_dropped=" U32H_FMT "\n"  \
                "port_xmit_data=" U32H_FMT "\n"  \
                "port_rcv_data=" U32H_FMT "\n"  \
                "port_xmit_pkts=" U32H_FMT "\n"  \
                "port_rcv_pkts=" U32H_FMT "\n"  \
                "port_xmit_wait=" U32H_FMT "\n",
                p_curr_port_counters->LinkDownedCounter,
                p_curr_port_counters->LinkErrorRecoveryCounter,
                p_curr_port_counters->SymbolErrorCounter,
                p_curr_port_counters->PortRcvRemotePhysicalErrors,
                p_curr_port_counters->PortRcvErrors,
                p_curr_port_counters->PortXmitDiscards,
                p_curr_port_counters->PortRcvSwitchRelayErrors,
                p_curr_port_counters->ExcessiveBufferOverrunErrors,
                p_curr_port_counters->LocalLinkIntegrityErrors,
                p_curr_port_counters->PortRcvConstraintErrors,
                p_curr_port_counters->PortXmitConstraintErrors,
                p_curr_port_counters->VL15Dropped,
                p_curr_port_counters->PortXmitData,
                p_curr_port_counters->PortRcvData,
                p_curr_port_counters->PortXmitPkts,
                p_curr_port_counters->PortRcvPkts,
                p_curr_port_counters->PortXmitWait);
        sout << curr_counters_line;

        p_curr_port_counters_extended = this->fabric_extended_info.getPMPortCountersExtended(i);
        if (!p_curr_port_counters_extended) {       //info not exist
            sprintf(curr_counters_line,
                    "port_xmit_data_extended=NA\n" \
                    "port_rcv_data_extended=NA\n"  \
                    "port_xmit_pkts_extended=NA\n" \
                    "port_rcv_pkts_extended=NA\n"  \
                    "port_unicast_xmit_pkts=NA\n"  \
                    "port_unicast_rcv_pkts=NA\n"   \
                    "port_multicast_xmit_pkts=NA\n"    \
                    "port_multicast_rcv_pkts=NA\n");
        } else {
            sprintf(curr_counters_line,
                    "port_xmit_data_extended=" U64H_FMT "\n"  \
                    "port_rcv_data_extended=" U64H_FMT "\n"  \
                    "port_xmit_pkts_extended=" U64H_FMT "\n"  \
                    "port_rcv_pkts_extended=" U64H_FMT "\n"  \
                    "port_unicast_xmit_pkts=" U64H_FMT "\n"  \
                    "port_unicast_rcv_pkts=" U64H_FMT "\n"  \
                    "port_multicast_xmit_pkts=" U64H_FMT "\n"  \
                    "port_multicast_rcv_pkts=" U64H_FMT "\n",
                    p_curr_port_counters_extended->PortXmitData,
                    p_curr_port_counters_extended->PortRcvData,
                    p_curr_port_counters_extended->PortXmitPkts,
                    p_curr_port_counters_extended->PortRcvPkts,
                    p_curr_port_counters_extended->PortUniCastXmitPkts,
                    p_curr_port_counters_extended->PortUniCastRcvPkts,
                    p_curr_port_counters_extended->PortMultiCastXmitPkts,
                    p_curr_port_counters_extended->PortMultiCastRcvPkts);
        }
        sout << curr_counters_line;

        // extended speeds counter
        if (IS_TO_CHECK_EXT_SPEEDS_COUNTERS(check_counters_bitset,
                                            p_curr_port->p_node->type)) {
            this->DumpPortExtendedSpeedsCounters(sout, en_per_lane_cnts,
                                                 p_curr_port, i);
        }


        p_port_calc_counters = this->fabric_extended_info.getPMPortCalcCounters(i);
        if ( !p_port_calc_counters ) {
            sprintf(curr_counters_line, "retransmission_per_sec=NA\n");
            sout << curr_counters_line;
        } else {
            sprintf(curr_counters_line, "retransmission_per_sec=" U64H_FMT "\n",
                    p_port_calc_counters->RetransmissionPerSec);
            sout << curr_counters_line;
        }

        p_curr_port_llr_statistics = this->fabric_extended_info.getVSPortLLRStatistics(i);

        if (!this->capability_module.IsSupportedGMPCapability(
            p_curr_port->p_node, EnGMPCapIsMaxRetransmissionRateSupported) ||
            !p_curr_port_llr_statistics)
            sout << "max_retransmission_rate=NA" << endl;
        else {
            sprintf(curr_counters_line, "max_retransmission_rate=" U16H_FMT "\n",
                    p_curr_port_llr_statistics->MaxRetransmissionRate);
            sout << curr_counters_line;
        }

        // LLR counter
        if (IS_TO_PRINT_LLR_COUNTERS(check_counters_bitset)) {
            // Counters not available
            if (!p_curr_port_llr_statistics) {
                sprintf(curr_counters_line,
                        "port_rcv_cells=NA\n" \
                        "port_rcv_cell_for_retry=NA\n" \
                        "port_rcv_retry=NA\n" \
                        "port_xmit_cells=NA\n" \
                        "port_xmit_retry_cells=NA\n" \
                        "port_xmit_retry=NA\n"
                        "port_symbol_error=NA\n"
                        "port_error_detection_counter_lane0=NA\n"
                        "port_error_detection_counter_lane1=NA\n"
                        "port_error_detection_counter_lane2=NA\n"
                        "port_error_detection_counter_lane3=NA\n");
                sout << curr_counters_line;
            }

            // Counters available
            else {

                sprintf(curr_counters_line,
                        "port_rcv_cells=" U64H_FMT "\n" \
                        "port_rcv_cell_for_retry=" U64H_FMT "\n" \
                        "port_rcv_retry=" U64H_FMT "\n" \
                        "port_xmit_cells=" U64H_FMT "\n" \
                        "port_xmit_retry_cells=" U64H_FMT "\n" \
                        "port_xmit_retry=" U64H_FMT "\n" \
                        "port_symbol_error=" U64H_FMT "\n" \
                        "port_error_detection_counter_lane0="  U64H_FMT "\n" \
                        "port_error_detection_counter_lane1="  U64H_FMT "\n" \
                        "port_error_detection_counter_lane2="  U64H_FMT "\n" \
                        "port_error_detection_counter_lane3="  U64H_FMT "\n",
                        p_curr_port_llr_statistics->PortRcvCells,
                        p_curr_port_llr_statistics->PortRcvCellForRetry,
                        p_curr_port_llr_statistics->PortRcvRetry,
                        p_curr_port_llr_statistics->PortXmitCells,
                        p_curr_port_llr_statistics->PortXmitRetryCells,
                        p_curr_port_llr_statistics->PortXmitRetry,
                        p_curr_port_llr_statistics->PortSymbolError,
                        p_curr_port_llr_statistics->PortErrorDetectionCounterLane0,
                        p_curr_port_llr_statistics->PortErrorDetectionCounterLane1,
                        p_curr_port_llr_statistics->PortErrorDetectionCounterLane2,
                        p_curr_port_llr_statistics->PortErrorDetectionCounterLane3);
                sout << curr_counters_line;
            }
        }

        // PortRcvErrorDetails and PortXmitDiscardDetails
        p_port_rcv_error_details = this->fabric_extended_info.getPMPortRcvErrorDetails(i);
        p_port_xmit_discard_details = this->fabric_extended_info.getPMPortXmitDiscardDetails(i);

        struct PortSampleControlOptionMask * p_option_mask;
        p_option_mask = this->fabric_extended_info.getPMOptionMask(p_curr_port->p_node->createIndex);

        if (!p_port_rcv_error_details) {
            sprintf(curr_counters_line,
                    "port_local_physical_errors=NA\n" \
                    "port_malformed_packet_errors=NA\n" \
                    "port_buffer_overrun_errors=NA\n" \
                    "port_dlid_mapping_errors=NA\n" \
                    "port_vl_mapping_errors=NA\n" \
                    "port_looping_errors=NA\n" );
            sout << curr_counters_line;
        } else {
            (p_option_mask->PortRcvErrorDetails_PortLocalPhysicalErrors)?
                sprintf(curr_counters_line, "port_local_physical_errors=" U16H_FMT "\n",
                        p_port_rcv_error_details->PortLocalPhysicalErrors):
                sprintf(curr_counters_line, "port_local_physical_errors=NA\n");
            sout << curr_counters_line;

            (p_option_mask->PortRcvErrorDetails_PortMalformedPacketErrors)?
                sprintf(curr_counters_line, "port_malformed_packet_errors=" U16H_FMT "\n",
                        p_port_rcv_error_details->PortMalformedPacketErrors):
                sprintf(curr_counters_line, "port_malformed_packet_errors=NA\n");
            sout << curr_counters_line;

            (p_option_mask->PortRcvErrorDetails_PortBufferOverrunErrors)?
                sprintf(curr_counters_line, "port_buffer_overrun_errors=" U16H_FMT "\n",
                        p_port_rcv_error_details->PortBufferOverrunErrors):
                sprintf(curr_counters_line, "port_buffer_overrun_errors=NA\n");
            sout << curr_counters_line;

            (p_option_mask->PortRcvErrorsDetails_PortDLIDMappingErrors)?
                sprintf(curr_counters_line, "port_dlid_mapping_errors=" U16H_FMT "\n",
                        p_port_rcv_error_details->PortDLIDMappingErrors):
                sprintf(curr_counters_line, "port_dlid_mapping_errors=NA\n");
            sout << curr_counters_line;

            (p_option_mask->PortRcvErrorsDetails_PortVlMappingErrors)?
                sprintf(curr_counters_line, "port_vl_mapping_errors=" U16H_FMT "\n",
                        p_port_rcv_error_details->PortVLMappingErrors):
                sprintf(curr_counters_line, "port_vl_mapping_errors=NA\n");
            sout << curr_counters_line;

            (p_option_mask->PortRcvErrorDetails_PortLoopingErrors)?
                sprintf(curr_counters_line, "port_looping_errors=" U16H_FMT "\n",
                        p_port_rcv_error_details->PortLoopingErrors):
                sprintf(curr_counters_line, "port_looping_errors=NA\n");
            sout << curr_counters_line;
        }

        if (!p_port_xmit_discard_details) {
            sprintf(curr_counters_line,
                    "port_inactive_discards=NA\n" \
                    "port_neighbor_mtu_discards=NA\n" \
                    "port_sw_lifetime_limit_discards=NA\n" \
                    "port_sw_hoq_lifetime_limit_discards=NA\n");
            sout << curr_counters_line;
        } else {
            (p_option_mask->PortXmitDiscardDetails_PortInactiveDiscards)?
                sprintf(curr_counters_line, "port_inactive_discards=" U16H_FMT "\n",
                        p_port_xmit_discard_details->PortInactiveDiscards):
                sprintf(curr_counters_line, "port_inactive_discards=NA\n");
            sout << curr_counters_line;

            (p_option_mask->PortXmitDiscardsDetails_PortNeighborMTUDiscards)?
                sprintf(curr_counters_line, "port_neighbor_mtu_discards=" U16H_FMT "\n",
                        p_port_xmit_discard_details->PortNeighborMTUDiscards):
                sprintf(curr_counters_line, "port_neighbor_mtu_discards=NA\n");
            sout << curr_counters_line;

            (p_option_mask->PortXmitDiscardDetails_PortSwLifetimeLimitDiscards)?
                sprintf(curr_counters_line, "port_sw_lifetime_limit_discards=" U16H_FMT "\n",
                        p_port_xmit_discard_details->PortSwLifetimeLimitDiscards):
                sprintf(curr_counters_line, "port_sw_lifetime_limit_discards=NA\n");
            sout << curr_counters_line;

            (p_option_mask->PortXmitDiscardDetails_PortSwHOQLifeLimitDiscards)?
                sprintf(curr_counters_line, "port_sw_hoq_lifetime_limit_discards=" U16H_FMT "\n",
                        p_port_xmit_discard_details->PortSwHOQLifetimeLimitDiscards):
                sprintf(curr_counters_line, "port_sw_hoq_lifetime_limit_discards=NA\n");
            sout << curr_counters_line;
        }

        sout << endl;
    }
    IBDIAG_RETURN_VOID;
}

/****************************************************/
/*******************public methods*******************/
/****************************************************/
void IBDiag::CopyPMInfoObjVector(OUT vector_p_pm_info_obj& new_pm_obj_info_vector)
{
    IBDIAG_ENTER;
    vector_p_pm_info_obj& pm_info_obj_vector =
            this->fabric_extended_info.getPMInfoObjVectorRef();

    //copy vector
    for (vector_p_pm_info_obj::iterator it = pm_info_obj_vector.begin();
            it != pm_info_obj_vector.end(); ++it)
        new_pm_obj_info_vector.push_back(*it);

    //clean vector - to need to delete because pointers are valid in new vector
    pm_info_obj_vector.clear();
    IBDIAG_RETURN_VOID;
}


void IBDiag::CleanPMInfoObjVector(IN vector_p_pm_info_obj& curr_pm_obj_info_vector)
{
    IBDIAG_ENTER;

    //clean vector
    for (vector_p_pm_info_obj::iterator it = curr_pm_obj_info_vector.begin();
            it != curr_pm_obj_info_vector.end(); ++it) {
        if (*it) {
            if ((*it)->p_port_counters)
                delete (*it)->p_port_counters;
            if ((*it)->p_extended_port_counters)
                delete (*it)->p_extended_port_counters;
            if ((*it)->p_port_ext_speeds_counters)
                delete (*it)->p_port_ext_speeds_counters;
            if ((*it)->p_port_llr_statistics)
                delete (*it)->p_port_llr_statistics;
            if ((*it)->p_port_calc_counters)
                delete (*it)->p_port_calc_counters;
            if ((*it)->p_port_ext_speeds_rsfec_counters)
                delete (*it)->p_port_ext_speeds_rsfec_counters;
            if ((*it)->p_port_rcv_error_details)
                delete (*it)->p_port_rcv_error_details;
            if ((*it)->p_port_xmit_discard_details)
                delete (*it)->p_port_xmit_discard_details;
            delete (*it);
        }
    }

    curr_pm_obj_info_vector.clear();

    IBDIAG_RETURN_VOID;
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / IBDIAG_ERR_CODE_FABRIC_ERROR /
int IBDiag::BuildClassPortInfoDB(list_p_fabric_general_err& pm_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    static bool buildDB = true;
    if(!buildDB) //run only once
        IBDIAG_RETURN(rc);
    buildDB = false;

    progress_bar_nodes_t progress_bar_nodes;
    struct IB_ClassPortInfo class_port_info;
    CLEAR_STRUCT(progress_bar_nodes);

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end();
            ++nI) {
        IBNode *p_node = (*nI).second;
        if (!p_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
        }

        u_int32_t startPort = 0;
        u_int32_t numPorts = 0;
        ++progress_bar_nodes.nodes_found;
        if (p_node->type == IB_SW_NODE) {
            startPort = 0;
            numPorts = 1;
            ++progress_bar_nodes.sw_found;
        } else {
            startPort = 1;
            numPorts = p_node->numPorts + 1;
            ++progress_bar_nodes.ca_found;
        }

        progress_bar_retrieve_from_nodes(&progress_bar_nodes,
            &this->discover_progress_bar_nodes, "PMClassPortInfo");

        // go over all base lids of this node:
        for (u_int32_t i = startPort; i < numPorts; ++i) {
            IBPort *p_port = p_node->getPort((phys_port_t)i);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_port || (i != 0 && !p_port->is_data_worthy()))
                continue;

            u_int16_t *p_cap_mask =
                this->fabric_extended_info.getPMCapMask(p_node->createIndex);

            //send class port info if cap mask not exist
            if (!p_cap_mask) {
                clbck_data.m_handle_data_func = IBDiagPMCapMaskClbck;
                clbck_data.m_data1 = p_node;
                this->ibis_obj.PMClassPortInfoGet(
                    p_port->base_lid, &class_port_info, &clbck_data);
            }
            break;
        }
        if(ibDiagClbck.GetState())
            goto exit;
    }
exit:
    this->ibis_obj.MadRecAll();

    if (rc) {
        if (last_error.empty())
            this->SetLastError("BuildClassPortInfoDB Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "BuildClassPortInfoDB Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!pm_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    SCREEN_PRINT("\n");
    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
//This may send class port info MAD if cap mask not exist.
int IBDiag::ReadCapMask(IBNode *p_node,
                        IBPort *p_port,
                        u_int16_t &cap_mask,
                        u_int32_t &port_info_cap_mask)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    u_int16_t *p_cap_mask =
        this->fabric_extended_info.getPMCapMask(p_node->createIndex);

    //send class port info if cap mask not exist
    if (!p_cap_mask) {
        IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
    }
    cap_mask = *p_cap_mask;
    rc = ReadPortInfoCapMask(p_node, p_port, port_info_cap_mask);
    IBDIAG_RETURN(rc);
}

int IBDiag::ReadPortInfoCapMask(IBNode *p_node,
                                IBPort *p_port,
                                u_int32_t &port_info_cap_mask,
                                u_int16_t *p_port_info_cap_mask2)
{
    IBDIAG_ENTER;

    IBPort *p_tmp_port = p_port;

    if (p_port->p_node->type == IB_SW_NODE) {
        p_tmp_port = p_port->p_node->getPort(0);
        if (!p_tmp_port) {
            this->SetLastError("DB error - can not found manage port for switch=%s\n",
                    p_port->p_node->getName().c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }
    }

    struct SMP_PortInfo *p_port_info =
        this->fabric_extended_info.getSMPPortInfo(p_tmp_port->createIndex);

    if (!p_port_info) {
        this->SetLastError("DB error - can not found port info for port=%s\n",
                        p_tmp_port->getName().c_str());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }
    port_info_cap_mask = p_port_info->CapMsk;

    if (p_port_info_cap_mask2)
        *p_port_info_cap_mask2 = p_port_info->CapMsk2;

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

u_int8_t IBDiag::PMIsOptionalAttrSupported(IBNode * p_node, int attr_id)
{
    IBDIAG_ENTER;
    u_int8_t rc = 0;
    struct PortSampleControlOptionMask * p_option_mask;
    p_option_mask = this->fabric_extended_info.getPMOptionMask(p_node->createIndex);
    if (!p_option_mask) {
        IBDIAG_LOG(TT_LOG_LEVEL_ERROR,
                   "Error DB - Can not found PortSampleControlOptionMask for " \
                   "node=%s\n",
                   p_node->name.c_str());
        IBDIAG_RETURN(rc);
    }

    switch (attr_id) {
        case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_RCV_DATA_VL:
        case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_XMIT_DATA_VL:
        case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_RCV_PKT_VL:
        case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_XMIT_PKT_VL:
            rc = this->capability_module.IsSupportedGMPCapability(p_node,
                    EnGMPCapIsPerVLCountersSupported);
            break;
        case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_RCV_DATA_VL_EXT:
        case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_XMIT_DATA_VL_EXT:
        case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_RCV_PKT_VL_EXT:
        case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_XMIT_PKT_VL_EXT:
        case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_XMIT_WAIT_VL_EXT:
            rc = this->capability_module.IsSupportedGMPCapability(p_node,
                    EnGMPCapIsPerVLExtendedSupported);
            break;
        case IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_XMIT_DATA_SL:
            rc = p_option_mask->PortXmitDataSL_n;
            break;
        case IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_RCV_DATA_SL:
            rc = p_option_mask->PortRcvDataSL_n;
            break;
        case IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_XMIT_DATA_SL_EXT:
            rc= p_option_mask->PortXmitDataSLExt_n;
            break;
        case IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_RCV_DATA_SL_EXT:
            rc = p_option_mask->PortRcvDataSLExt_n;
            break;
        case IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_RCV_ERROR_DETAILS:
            rc = p_option_mask->PortRcvErrorDetails_PortLocalPhysicalErrors ||
                 p_option_mask->PortRcvErrorDetails_PortMalformedPacketErrors ||
                 p_option_mask->PortRcvErrorDetails_PortBufferOverrunErrors ||
                 p_option_mask->PortRcvErrorsDetails_PortDLIDMappingErrors ||
                 p_option_mask->PortRcvErrorsDetails_PortVlMappingErrors ||
                 p_option_mask->PortRcvErrorDetails_PortLoopingErrors;
            break;
        case IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_XMIT_DISCARD_DETAILS:
            rc = p_option_mask->PortXmitDiscardDetails_PortInactiveDiscards ||
                 p_option_mask->PortXmitDiscardsDetails_PortNeighborMTUDiscards ||
                 p_option_mask->PortXmitDiscardDetails_PortSwLifetimeLimitDiscards ||
                 p_option_mask->PortXmitDiscardDetails_PortSwHOQLifeLimitDiscards;
            break;
        default:
            break;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::BuildPortOptionMaskDB(list_p_fabric_general_err& pm_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    static bool buildDB = true;
    if(!buildDB) //run only once
        IBDIAG_RETURN(rc);
    buildDB = false;

    rc = BuildClassPortInfoDB(pm_errors);
    if (rc)
        IBDIAG_RETURN(rc);

    progress_bar_nodes_t progress_bar_nodes;
    struct PM_PortSamplesControl sample_control;
    CLEAR_STRUCT(progress_bar_nodes);

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end();
            ++nI) {
        IBNode *p_node = (*nI).second;
        if (!p_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s",
                               (*nI).first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
        }

        ++progress_bar_nodes.nodes_found;
        if (p_node->type == IB_SW_NODE) {
            ++progress_bar_nodes.sw_found;
        } else {
            ++progress_bar_nodes.ca_found;
        }
        progress_bar_retrieve_from_nodes(&progress_bar_nodes,
            &this->discover_progress_bar_nodes, "PMPortSampleControl");

        // we try to get PortSampleControl only from 1 port although
        // it should be sent per port by spec because no different
        // in the 'option mask' field is expected.
        u_int32_t startPort = 1;
        u_int32_t numPorts = p_node->numPorts + 1;

        for (u_int32_t i = startPort; i < numPorts; ++i) {
            IBPort *p_port = p_node->getPort((phys_port_t)i);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_port || !p_port->is_data_worthy())
                continue;

            struct PortSampleControlOptionMask *p_option_mask =
                this->fabric_extended_info.getPMOptionMask(p_node->createIndex);

            //send Port Sample Control if option_mask not exist
            if (!p_option_mask) {
                clbck_data.m_handle_data_func = IBDiagPMOptionMaskClbck;
                clbck_data.m_data1 = p_node;
                this->ibis_obj.PMPortSampleControlGet(
                    p_port->base_lid, (phys_port_t)i, &sample_control, &clbck_data);
            }
            break;
        }
        if(ibDiagClbck.GetState())
            goto exit;
    }
exit:
    this->ibis_obj.MadRecAll();

    if (rc) {
        if (last_error.empty())
            this->SetLastError("BuildOptionMaskDB Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "BuildOptionMaskDB Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!pm_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    SCREEN_PRINT("\n");
    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::BuildPortCountersDB(list_p_fabric_general_err& pm_errors,
                                progress_func_nodes_t progress_func,
                                u_int32_t check_counters_bitset)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ResetAppData();
    ibDiagClbck.Set(this, &fabric_extended_info, &pm_errors);

    int rc = BuildPortOptionMaskDB(pm_errors);  //include 'BuildClassPortInfoDB'
    //don't exit, if only some of the nodes failed to retrieve ClassPortInfo
    if (rc && rc != IBDIAG_ERR_CODE_FABRIC_ERROR) {
        IBDIAG_RETURN(rc);
    }
    int rc2;
    progress_bar_nodes_t progress_bar_nodes;
    CLEAR_STRUCT(progress_bar_nodes);

    // go over all nodes:
    u_int32_t port_info_cap_mask = 0;
    u_int16_t curr_pm_cap_mask = 0;
    bool to_read_cap = true;
    struct PM_PortCounters curr_port_counters;
    struct PM_PortCountersExtended curr_port_counters_extended;
    struct PM_PortExtendedSpeedsCounters curr_port_ext_speeds_counters;
    struct PM_PortExtendedSpeedsRSFECCounters    curr_port_ext_speeds_RSFEC_counters;
    struct VendorSpec_PortLLRStatistics curr_port_llr_statistics;
    struct PM_PortRcvErrorDetails curr_port_rcv_error_details;
    struct PM_PortXmitDiscardDetails curr_port_xmit_discard_details;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end();
            ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        //get cap mask for this node
        to_read_cap = true;

        ++progress_bar_nodes.nodes_found;
        if (p_curr_node->type == IB_SW_NODE)
            ++progress_bar_nodes.sw_found;
        else
            ++progress_bar_nodes.ca_found;
        if (progress_func)
            progress_func(&progress_bar_nodes, &this->discover_progress_bar_nodes);

        //check LLR support by DeviceID
        struct SMP_NodeInfo * p_curr_node_info =
            this->fabric_extended_info.getSMPNodeInfo(p_curr_node->createIndex);
        if (!p_curr_node_info) {
            this->SetLastError("DB error - failed to get SMPNodeInfo for node: %s", p_curr_node->name.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }
        if (!this->capability_module.IsSupportedGMPCapability(
            p_curr_node, EnGMPCapIsPortLLRStatisticsSupported)) {
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "node %s: DeviceID %u (0x%x) Does not support LLR, skipping \n",
                       p_curr_node->name.c_str(),
                       p_curr_node_info->DeviceID,
                       p_curr_node_info->DeviceID);

            p_curr_node->appData1.val |= NOT_SUPPORT_LLR_COUNTERS;
        }

        //check PM_PortRcvErrorDetails and PM_PortXmitDiscardDetails support by DeviceID
        if (!PMIsOptionalAttrSupported(p_curr_node,
                                       IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_RCV_ERROR_DETAILS)) {
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                       "node %s: DeviceID %u (0x%x) Does not support PortRcvErrorDetails,"
                       " skipping \n",
                       p_curr_node->name.c_str(),
                       p_curr_node_info->DeviceID,
                       p_curr_node_info->DeviceID);
            p_curr_node->appData1.val |= NOT_SUPPORT_PORT_RCV_ERROR_DETAILS;
        }

        if (!PMIsOptionalAttrSupported(p_curr_node,
            IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_XMIT_DISCARD_DETAILS)) {
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                       "node %s: DeviceID %u (0x%x) Does not support PortXmitDiscardDetails,"
                       " skipping \n",
                       p_curr_node->name.c_str(),
                       p_curr_node_info->DeviceID,
                       p_curr_node_info->DeviceID);
            p_curr_node->appData1.val |= NOT_SUPPORT_PORT_XMIT_DISCARD_DETAILS;
        }


        // go over all ports of this node:
        for (u_int32_t i = 1; i < (unsigned)p_curr_node->numPorts + 1; ++i) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            //for switch read cap only on port 0, for others read on every port
            if (to_read_cap) {
                rc2 = ReadCapMask(p_curr_node, p_curr_port,
                                  curr_pm_cap_mask, port_info_cap_mask);
                if (rc2) {
                    rc = rc2;
                    if (rc == IBDIAG_ERR_CODE_FABRIC_ERROR) {
                        IBDIAG_LOG(TT_LOG_LEVEL_MAD, "node:%s Failed to read cap mask - skip to next node.\n",
                                   (*nI).first.c_str());
                        goto next_node;     //not reasonable to continue
                    }
                    goto exit;
                }
                IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "node:%s pm_cap_mask:0x%x port_info_cap_mask:0x%x\n",
                           (*nI).first.c_str(), curr_pm_cap_mask, port_info_cap_mask);
                if (p_curr_node->type == IB_SW_NODE)
                    to_read_cap = false;
            }

            //send port counters mad get for this port
            clbck_data.m_handle_data_func = IBDiagPMPortCountersGetClbck;
            clbck_data.m_data1 = p_curr_port;
            rc2 = this->ibis_obj.PMPortCountersGet(p_curr_port->base_lid,
                p_curr_port->num, &curr_port_counters, &clbck_data);
            if (rc2)
                continue;  //we will try to proceed without query for extended counters

            //check if xmit wait is supported
            if (GetSpecialCAPortType(p_curr_node) == NotSpecial && // skip special nodes
                !(p_curr_node->appData1.val & NOT_SUPPORT_XMIT_WAIT)) {
                if (IS_SUPPORT_XMIT_WAIT(curr_pm_cap_mask) == 0) {
                    FabricErrNodeNotSupportCap *p_curr_fabric_pm_err =
                            new FabricErrNodeNotSupportCap(p_curr_node,
                                    "This device does not support xmit wait counter capability");
                    if (!p_curr_fabric_pm_err) {
                        this->SetLastError("Failed to allocate FabricErrPortNotRespond");
                        rc = IBDIAG_ERR_CODE_NO_MEM;
                        goto exit;
                    }
                    pm_errors.push_back(p_curr_fabric_pm_err);
                    rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

                    //mark we already checked for support
                    p_curr_node->appData1.val |= NOT_SUPPORT_XMIT_WAIT;
                }
            }

            //check if counters extended are supported
            if (!IS_SUPPORT_EXT_PORT_FAILED(p_curr_node->appData1.val)) {
                if ( IS_SUPPORT_EXTENDED_COUNTERS(curr_pm_cap_mask) ||
                     IS_SUPPORT_EXTENDED_COUNTERS_NO_IETF(curr_pm_cap_mask)) {

                    clbck_data.m_handle_data_func = IBDiagPMPortCountersExtendedGetClbck;
                    this->ibis_obj.PMPortCountersExtendedGet(p_curr_port->base_lid,
                        p_curr_port->num, &curr_port_counters_extended, &clbck_data);
                } else {

                    FabricErrNodeNotSupportCap *p_curr_fabric_pm_err =
                            new FabricErrNodeNotSupportCap(p_curr_node,
                                    "This device does not support extended port counters capability");
                    if (!p_curr_fabric_pm_err) {
                        this->SetLastError("Failed to allocate FabricErrPortNotRespond");
                        rc = IBDIAG_ERR_CODE_NO_MEM;
                        goto exit;
                    }
                    pm_errors.push_back(p_curr_fabric_pm_err);
                    rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

                    //mark we already checked for support
                    p_curr_node->appData1.val |= NOT_SUPPORT_EXT_PORT_COUNTERS;
                }
            }

            //check if extended speed counters supported
            if(IS_TO_CHECK_EXT_SPEEDS_COUNTERS(check_counters_bitset, p_curr_node->type) &&
               !IS_SUPPORT_EXT_SPEEDS_FAILED(p_curr_node->appData1.val)) {

                if(IS_SUPPORT_EXTENDED_SPEEDS_COUNTERS(port_info_cap_mask)) {
                    //if RS_FEC is active
                    if (isRSFEC(p_curr_port->get_fec_mode())) {
                        if (!IS_SUPPORT_EXT_SPEEDS_RSFEC_FAILED(
                                p_curr_node->appData1.val)) {
                            if (IS_SUPPORT_EXT_SPEEDS_RSFEC_COUNTERS(
                                curr_pm_cap_mask)) {
                                clbck_data.m_handle_data_func =
                                    IBDiagPMPortExtendedSpeedsRSFECGetClbck;
                                this->ibis_obj.
                                    PMPortExtendedSpeedsRSFECCountersGet(
                                    p_curr_port->base_lid,
                                    p_curr_port->num,
                                    &curr_port_ext_speeds_RSFEC_counters,
                                    &clbck_data);
                            } else {
                                FabricErrNodeNotSupportCap
                                    *p_curr_fabric_pm_err =
                                        new FabricErrNodeNotSupportCap(
                                            p_curr_node, "This device does not "
                                            "support port extended speeds RSFEC "
                                            "counters capability");
                                if (!p_curr_fabric_pm_err) {
                                    this->SetLastError("Failed to allocate "
                                                       "FabricErrPortNotRespond");
                                    rc = IBDIAG_ERR_CODE_NO_MEM;
                                    goto exit;
                                }
                                pm_errors.push_back(p_curr_fabric_pm_err);
                                rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

                                //mark we already checked for support
                                p_curr_node->appData1.val |=
                                    NOT_SUPPORT_EXT_SPEEDS_RSFEC_COUNTERS;
                            }
                        }
                    } else {//not RS_FEC
                        clbck_data.m_handle_data_func = IBDiagPMPortExtendedSpeedsGetClbck;
                        this->ibis_obj.PMPortExtendedSpeedsCountersGet(p_curr_port->base_lid,
                                p_curr_port->num, &curr_port_ext_speeds_counters, &clbck_data);
                    }

                } else {

                    FabricErrNodeNotSupportCap *p_curr_fabric_pm_err =
                            new FabricErrNodeNotSupportCap(p_curr_node,
                                    "This device does not support port extended speeds counters capability");
                    if (!p_curr_fabric_pm_err) {
                        this->SetLastError("Failed to allocate FabricErrPortNotRespond");
                        rc = IBDIAG_ERR_CODE_NO_MEM;
                        goto exit;
                    }
                    pm_errors.push_back(p_curr_fabric_pm_err);
                    rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

                    //mark we already checked for support
                    p_curr_node->appData1.val |= NOT_SUPPORT_EXT_SPEEDS_COUNTERS;
                }
            } //ext_speeds_counters

            if( !IS_SUPPORT_LLR_FAILED(p_curr_node->appData1.val)) {

                clbck_data.m_handle_data_func = IBDiagVSPortLLRStatisticsGetClbck;
                this->ibis_obj.VSPortLLRStatisticsGet(p_curr_port->base_lid,
                        p_curr_port->num, &curr_port_llr_statistics,
                        IS_TO_PRINT_LLR_COUNTERS(check_counters_bitset), &clbck_data);

            } //LLR

            // PortRcvErrorDetails and PortXmitDiscardDetails
            if (!IS_SUPPORT_PORT_RCV_ERROR_DETAILS(p_curr_node->appData1.val)) {
                clbck_data.m_handle_data_func = IBDiagPMPortRcvErrorDetailsGetClbck;
                this->ibis_obj.PMPortRcvErrorDetailsGet(p_curr_port->base_lid,
                                                        p_curr_port->num,
                                                        &curr_port_rcv_error_details,
                                                        &clbck_data);
            }

            if (!IS_SUPPORT_PORT_XMIT_DISCARD_DETAILS(p_curr_node->appData1.val)) {
                clbck_data.m_handle_data_func = IBDiagPMPortXmitDiscardDetailsGetClbck;
                this->ibis_obj.PMPortXmitDiscardDetailsGet(p_curr_port->base_lid,
                                                           p_curr_port->num,
                                                           &curr_port_xmit_discard_details,
                                                           &clbck_data);
            }

        }   //current node ports loop

        if (ibDiagClbck.GetState())
            goto exit;

next_node:
        continue;
    }   //node loop
exit:
    this->ibis_obj.MadRecAll();

    if (rc) {
        if (last_error.empty())
            this->SetLastError("BuildPortCountersDB Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "BuildPortCountersDB Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!pm_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

bool IBDiag::HandleSpecialPorts(CountersPerSLVL * cntrs_per_slvl,
                                SMP_MlnxExtPortInfo *p_curr_mepi,
                                IBPort *p_curr_port,
                                int &rc,
                                list_p_fabric_general_err &cntrs_per_slvl_errors)
{
    stringstream ss;
    ss << "This special port does not support PM "
       << cntrs_per_slvl->GetCntrHeader()
       << " MAD. type = "
       << +p_curr_mepi->SpecialPortType;

    FabricErrPortNotSupportCap *p_curr_fabric_port_err =
        new FabricErrPortNotSupportCap(p_curr_port, ss.str());

    if (!p_curr_fabric_port_err) {
        SetLastError("Failed to allocate FabricErrPortNotSupportCap");
        rc = IBDIAG_ERR_CODE_NO_MEM;
        IBDIAG_RETURN(false);
    }

    p_curr_fabric_port_err->SetLevel(EN_FABRIC_ERR_WARNING);
    cntrs_per_slvl_errors.push_back(p_curr_fabric_port_err);

    IBDIAG_RETURN(true);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::BuildDBOrResetSLVLCntrs(
        list_p_fabric_general_err &cntrs_per_slvl_errors,
        progress_func_ports_t progress_func,
        bool is_reset_cntr,
        CountersPerSLVL * cntrs_per_slvl)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;

    ibDiagClbck.Set(this, &fabric_extended_info, &cntrs_per_slvl_errors,
                    NULL, &capability_module);

    int rc_psc = BuildPortOptionMaskDB(cntrs_per_slvl_errors);
    //don't exit, if only some of the nodes failed to retrieve PortSampleControl
    if (rc_psc && rc_psc != IBDIAG_ERR_CODE_FABRIC_ERROR) {
        IBDIAG_RETURN(rc_psc);
    }

    progress_bar_ports_t progress_bar_ports;
    CLEAR_STRUCT(progress_bar_ports);

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = IBDiagGSIPerSLVLGetClbck;

    struct PM_PortRcvXmitCntrsSlVl pm_port_rcvxmit_data_slvl;

    for (phys_port_t curr_port_idx = 1;
         curr_port_idx < this->discovered_fabric.GetMaxNodePorts() + 1;
         ++curr_port_idx) {

        for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
                nI != this->discovered_fabric.NodeByName.end(); ++nI) {
            IBNode *p_curr_node = (*nI).second;
            if (!p_curr_node) {
                this->SetLastError(
                        "DB error - found null node in NodeByName map for key = %s",
                        (*nI).first.c_str());
                rc = IBDIAG_ERR_CODE_DB_ERR;
                goto exit;
            }
            //skip if this node has less ports
            if (p_curr_node->numPorts < curr_port_idx)
                continue;

            //deals with progress bar
            ++progress_bar_ports.ports_found;
            if (progress_func)
                progress_func(&progress_bar_ports, &this->discover_progress_bar_ports);

            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)curr_port_idx);
            if (!p_curr_port)                       //port not exist
                continue;
            //Skip ports that are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port->is_data_worthy())
                continue;

            //appData1 is for marking "supported"/"not supported" attributes
            if (PM_PER_SLVL_ATTR_IS_NOT_SUPPORT_CAPABILTY(p_curr_node,
                                                          cntrs_per_slvl->GetAttrCapBit()))
                continue;

            //appData2 is for marking "support checked"/"supported not checked" attributes
            if (PM_PER_SLVL_ATTR_IS_SUPPORT_NOT_CHECKED_CAPABILTY(p_curr_node,
                                                         cntrs_per_slvl->GetAttrCapBit())) {
                //mark "supported checked"/"supported not checked" appdata (appData2)
                p_curr_node->appData2.val |= cntrs_per_slvl->GetAttrCapBit();
                if (!PMIsOptionalAttrSupported(p_curr_node, cntrs_per_slvl->GetAttrId())) {
                    //mark "supported"/"not supported" appdata (appData1)
                    p_curr_node->appData1.val |= cntrs_per_slvl->GetAttrCapBit();
                    FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                        new FabricErrNodeNotSupportCap(p_curr_node,
                                "This device does not support PM " +
                                cntrs_per_slvl->GetCntrHeader() + " capability");
                    if (!p_curr_fabric_node_err) {
                        this->SetLastError("Failed to allocate FabricErrNodeNotSupportCap");
                        rc = IBDIAG_ERR_CODE_DB_ERR;
                        goto exit;
                    }
                    p_curr_fabric_node_err->SetLevel(EN_FABRIC_ERR_WARNING);
                    cntrs_per_slvl_errors.push_back(p_curr_fabric_node_err);
                    continue;
                }
            }

            // special ports
            if (p_curr_node->type == IB_SW_NODE) {
                if (!this->capability_module.IsSupportedSMPCapability(
                            p_curr_node, EnSMPCapIsSpecialPortsMarkingSupported)) {

                    // if not checked yet for this node
                    if (!(p_curr_node->appData2.val & NOT_SUPPORT_SPECIAL_PORTS_MARKING_CHECKED)) {

                        p_curr_node->appData2.val |= NOT_SUPPORT_SPECIAL_PORTS_MARKING_CHECKED;

                        FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                            new FabricErrNodeNotSupportCap(p_curr_node,
                                "This device does not support Special Ports Marking capability");
                        if (!p_curr_fabric_node_err) {
                            this->SetLastError("Failed to allocate FabricErrNodeNotSupportCap");
                            rc = IBDIAG_ERR_CODE_NO_MEM;
                            goto exit;
                        }
                        p_curr_fabric_node_err->SetLevel(EN_FABRIC_ERR_WARNING);
                        cntrs_per_slvl_errors.push_back(p_curr_fabric_node_err);
                    }

                } else { // Special Ports Marking Supported

                    struct SMP_MlnxExtPortInfo *p_curr_mepi =
                            fabric_extended_info.getSMPMlnxExtPortInfo(p_curr_port->createIndex);
                    if (!p_curr_mepi) {
                        IBDIAG_LOG(TT_LOG_LEVEL_ERROR,
                                   "Error DB - Can not found MlnxExtPortInfo for " \
                                   "node=%s, port=%u\n",
                                   p_curr_node->name.c_str(), curr_port_idx);

                        rc = IBDIAG_ERR_CODE_DB_ERR;
                        IBDIAG_RETURN(rc);
                    }

                    if (p_curr_mepi->IsSpecialPort) {

                        switch (cntrs_per_slvl->GetAttrId()) {
                            case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_RCV_DATA_VL:
                            case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_XMIT_DATA_VL:
                            case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_RCV_PKT_VL:
                            case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_XMIT_PKT_VL:

                                if(!(p_curr_mepi->SpecialPortCapabilityMask &
                                        (1 << EnSPCapIsPerVLCountersSupported))) {
                                    if (!HandleSpecialPorts(cntrs_per_slvl, p_curr_mepi,
                                            p_curr_port, rc, cntrs_per_slvl_errors))
                                        goto exit;
                                    continue;
                                }
                                break;

                            case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_RCV_DATA_VL_EXT:
                            case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_XMIT_DATA_VL_EXT:
                            case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_RCV_PKT_VL_EXT:
                            case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_XMIT_PKT_VL_EXT:

                                if(!(p_curr_mepi->SpecialPortCapabilityMask &
                                        (1 << EnSPCapIsPerVLExtendedSupported))) {
                                    if (!HandleSpecialPorts(cntrs_per_slvl, p_curr_mepi,
                                            p_curr_port, rc, cntrs_per_slvl_errors))
                                       goto exit;
                                   continue;
                                }
                                break;

                            case IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_XMIT_WAIT_VL_EXT:

                                if(!(p_curr_mepi->SpecialPortCapabilityMask &
                                        (1 << EnSPCapIsPortXmitWaitVLExtendedSupported))) {
                                    if (!HandleSpecialPorts(cntrs_per_slvl, p_curr_mepi,
                                            p_curr_port, rc, cntrs_per_slvl_errors))
                                       goto exit;
                                    continue;
                                }
                        }
                    }
                }
            } // special ports

            /*
             * Send the MAD
             */
            clbck_data.m_data1 = p_curr_port;
            clbck_data.m_data2 = cntrs_per_slvl;
            clbck_data.m_data3 = (void *)(uintptr_t)is_reset_cntr;

            if (!cntrs_per_slvl->IsVSClass())
                this->ibis_obj.PMPerSLVLCounters(
                        is_reset_cntr,
                        p_curr_port->base_lid,
                        curr_port_idx,
                        cntrs_per_slvl->GetAttrId(),
                        &pm_port_rcvxmit_data_slvl,
                        &clbck_data);
            else
                this->ibis_obj.VSPerVLCounters(
                        is_reset_cntr,
                        p_curr_port->base_lid,
                        curr_port_idx,
                        cntrs_per_slvl->GetAttrId(),
                        &pm_port_rcvxmit_data_slvl,
                        &clbck_data);

            if(ibDiagClbck.GetState())
                goto exit;
        }
    }
exit:
    this->ibis_obj.MadRecAll();

    if (rc) {
        if (last_error.empty())
            this->SetLastError("Retrieve of VS PerSLVLCounters Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "Retrieve of VS PerSLVLCounters Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!cntrs_per_slvl_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::DumpPortCountersCSVTable(CSVOut &csv_out, u_int32_t check_counters_bitset)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    csv_out.DumpStart(SECTION_PM_INFO);

    stringstream sstream;
    sstream << "NodeGUID," << "PortGUID," "PortNumber,"
            << "LinkDownedCounter," << "LinkErrorRecoveryCounter," << "SymbolErrorCounter,"
            << "PortRcvRemotePhysicalErrors," << "PortRcvErrors," << "PortXmitDiscards,"
            << "PortRcvSwitchRelayErrors," << "ExcessiveBufferOverrunErrors," << "LocalLinkIntegrityErrors,"
            << "PortRcvConstraintErrors," << "PortXmitConstraintErrors," << "VL15Dropped,"
            << "PortXmitData," << "PortRcvData,"
            << "PortXmitPkts," << "PortRcvPkts,"
            << "PortXmitWait,"
            << "PortXmitDataExtended," << "PortRcvDataExtended,"
            << "PortXmitPktsExtended," << "PortRcvPktsExtended,"
            << "PortUniCastXmitPkts," << "PortUniCastRcvPkts,"
            << "PortMultiCastXmitPkts," << "PortMultiCastRcvPkts";

    if (IS_TO_CHECK_EXT_SPEEDS_COUNTERS(check_counters_bitset, IB_SW_NODE)) {
        const char * port_extended_speeds_counters_names[] = {
            "ErrorDetectionCounterLane",
            "FECCorrectableBlockCounterLane",
            "FECUncorrectableBlockCounterLane",
            "FECCorrectedSymbolCounterLane",
        };

        sstream << ",SyncHeaderErrorCounter,UnknownBlockCounter";

        size_t max_ext_speeds_counters = sizeof(port_extended_speeds_counters_names) / sizeof (char *);
        for (size_t i=0 ; i < max_ext_speeds_counters ; i++) {
            for (u_int16_t j=0 ; j < MAX_LANE_NUMBER ; j++) {
                sstream << "," << port_extended_speeds_counters_names[i] << "[" << j << "]";
            }
        }

        sstream <<
            ",PortFECCorrectableBlockCounter,"\
            "PortFECUncorrectableBlockCounter,PortFECCorrectedSymbolCounter";
    }

    // calculated counter
    sstream << ",retransmission_per_sec, max_retransmission_rate";

    // PortRcvErrorDetails counters
    sstream << ",PortLocalPhysicalErrors,PortMalformedPacketErrors"
         << ",PortBufferOverrunErrors,PortDLIDMappingErrors"
         << ",PortVLMappingErrors,PortLoopingErrors";

    // PortXmitDiscardDetails counters
    sstream << ",PortInactiveDiscards,PortNeighborMTUDiscards"
         << ",PortSwLifetimeLimitDiscards,PortSwHOQLifetimeLimitDiscards" << endl;
    csv_out.WriteLine(sstream.str());

    struct PM_PortCounters *p_curr_port_counters = NULL;
    struct PM_PortCountersExtended *p_curr_port_counters_extended = NULL;
    struct PM_PortExtendedSpeedsCounters *p_curr_port_ext_speeds_counters = NULL;
    struct PM_PortExtendedSpeedsRSFECCounters
        *p_curr_port_ext_speeds_rsfec_counters = NULL;
    struct PM_PortCalcCounters *p_port_calc_counters = NULL;
    struct PM_PortRcvErrorDetails *p_port_rcv_error_details = NULL;
    struct PM_PortXmitDiscardDetails *p_port_xmit_discard_details = NULL;

    char curr_counters_line[1024];

    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port)               //port not exist
            continue;
        p_curr_port_counters = this->fabric_extended_info.getPMPortCounters(i);
        if (!p_curr_port_counters)       //info not exist
            continue;

        memset(curr_counters_line, 0, sizeof(curr_counters_line));
        sstream.str("");
        sprintf(curr_counters_line, U64H_FMT "," U64H_FMT "," U32D_FMT ","   \
                U32D_FMT "," U32D_FMT "," U32D_FMT ","  \
                U32D_FMT "," U32D_FMT "," U32D_FMT ","  \
                U32D_FMT "," U32D_FMT "," U32D_FMT ","  \
                U32D_FMT "," U32D_FMT "," U32D_FMT ","  \
                U32D_FMT "," U32D_FMT "," U32D_FMT ","  \
                U32D_FMT "," U32D_FMT ",",
                p_curr_port->p_node->guid_get(),
                p_curr_port->guid_get(),
                p_curr_port->num,
                p_curr_port_counters->LinkDownedCounter,
                p_curr_port_counters->LinkErrorRecoveryCounter,
                p_curr_port_counters->SymbolErrorCounter,
                p_curr_port_counters->PortRcvRemotePhysicalErrors,
                p_curr_port_counters->PortRcvErrors,
                p_curr_port_counters->PortXmitDiscards,
                p_curr_port_counters->PortRcvSwitchRelayErrors,
                p_curr_port_counters->ExcessiveBufferOverrunErrors,
                p_curr_port_counters->LocalLinkIntegrityErrors,
                p_curr_port_counters->PortRcvConstraintErrors,
                p_curr_port_counters->PortXmitConstraintErrors,
                p_curr_port_counters->VL15Dropped,
                p_curr_port_counters->PortXmitData,
                p_curr_port_counters->PortRcvData,
                p_curr_port_counters->PortXmitPkts,
                p_curr_port_counters->PortRcvPkts,
                p_curr_port_counters->PortXmitWait);
        sstream << curr_counters_line;

        p_curr_port_counters_extended = this->fabric_extended_info.getPMPortCountersExtended(i);
        if (!p_curr_port_counters_extended) {       //info not exist
            sprintf(curr_counters_line, "0xfffffffffffffffe," \
                    "0xfffffffffffffffe," \
                    "0xfffffffffffffffe," \
                    "0xfffffffffffffffe," \
                    "0xfffffffffffffffe," \
                    "0xfffffffffffffffe," \
                    "0xfffffffffffffffe," \
                    "0xfffffffffffffffe");
            sstream << curr_counters_line;
        } else {

            memset(curr_counters_line, 0, sizeof(curr_counters_line));
            sprintf(curr_counters_line,
                    U64H_FMT "," U64H_FMT ","    \
                    U64H_FMT "," U64H_FMT ","    \
                    U64H_FMT "," U64H_FMT ","    \
                    U64H_FMT "," U64H_FMT,
                    p_curr_port_counters_extended->PortXmitData,
                    p_curr_port_counters_extended->PortRcvData,
                    p_curr_port_counters_extended->PortXmitPkts,
                    p_curr_port_counters_extended->PortRcvPkts,
                    p_curr_port_counters_extended->PortUniCastXmitPkts,
                    p_curr_port_counters_extended->PortUniCastRcvPkts,
                    p_curr_port_counters_extended->PortMultiCastXmitPkts,
                    p_curr_port_counters_extended->PortMultiCastRcvPkts);
            sstream << curr_counters_line;
        }

        if (IS_TO_CHECK_EXT_SPEEDS_COUNTERS(check_counters_bitset, IB_SW_NODE)) {
            p_curr_port_ext_speeds_counters =
                this->fabric_extended_info.getPMPortExtSpeedsCounters(i);
            p_curr_port_ext_speeds_rsfec_counters =
                this->fabric_extended_info.getPMPortExtSpeedsRSFECCounters(i);

            //first print counters common to both layouts:
            //regular extended speed counters and rsfec extended speed counters
            if (!p_curr_port_ext_speeds_counters &&
                !p_curr_port_ext_speeds_rsfec_counters) {       //info not exist
                sprintf(curr_counters_line, ",-1,-1");
            } else {
                u_int16_t sync_header_error_counter = 0;
                u_int16_t unknown_block_counter = 0;

                if (p_curr_port_ext_speeds_counters) {
                    sync_header_error_counter =
                    p_curr_port_ext_speeds_counters->SyncHeaderErrorCounter;
                    unknown_block_counter =
                        p_curr_port_ext_speeds_counters->UnknownBlockCounter;
                } else {
                    sync_header_error_counter =
                    p_curr_port_ext_speeds_rsfec_counters->
                        SyncHeaderErrorCounter;
                    unknown_block_counter =
                    p_curr_port_ext_speeds_rsfec_counters->UnknownBlockCounter;
                }
                sprintf(curr_counters_line, ",%u,%u",
                        sync_header_error_counter,
                        unknown_block_counter);
            }
            sstream << curr_counters_line;

            //EDPL
            if (!p_curr_port_ext_speeds_counters)
                sprintf(curr_counters_line,
                        ",-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1");
            else
                //it's sure not RS-FEC, if we have this data
                sprintf(
                curr_counters_line,
                ",%u,%u,%u,%u,%u,%u,%u,%u,%u,%u,%u,%u",
                p_curr_port_ext_speeds_counters->ErrorDetectionCounterLane[0],
                p_curr_port_ext_speeds_counters->ErrorDetectionCounterLane[1],
                p_curr_port_ext_speeds_counters->ErrorDetectionCounterLane[2],
                p_curr_port_ext_speeds_counters->ErrorDetectionCounterLane[3],
                p_curr_port_ext_speeds_counters->ErrorDetectionCounterLane[4],
                p_curr_port_ext_speeds_counters->ErrorDetectionCounterLane[5],
                p_curr_port_ext_speeds_counters->ErrorDetectionCounterLane[6],
                p_curr_port_ext_speeds_counters->ErrorDetectionCounterLane[7],
                p_curr_port_ext_speeds_counters->ErrorDetectionCounterLane[8],
                p_curr_port_ext_speeds_counters->ErrorDetectionCounterLane[9],
                p_curr_port_ext_speeds_counters->ErrorDetectionCounterLane[10],
                p_curr_port_ext_speeds_counters->ErrorDetectionCounterLane[11]
                );
            sstream << curr_counters_line;

            //in case of "No FEC" mode those counters are NA
            if (p_curr_port_ext_speeds_counters &&
                p_curr_port->get_fec_mode() != IB_FEC_NO_FEC)
                sprintf(
                curr_counters_line,
                ",%u,%u,%u,%u,%u,%u,%u,%u,%u,%u,%u,%u" \
                ",%u,%u,%u,%u,%u,%u,%u,%u,%u,%u,%u,%u",
                p_curr_port_ext_speeds_counters->FECCorrectableBlockCountrLane[0],
                p_curr_port_ext_speeds_counters->FECCorrectableBlockCountrLane[1],
                p_curr_port_ext_speeds_counters->FECCorrectableBlockCountrLane[2],
                p_curr_port_ext_speeds_counters->FECCorrectableBlockCountrLane[3],
                p_curr_port_ext_speeds_counters->FECCorrectableBlockCountrLane[4],
                p_curr_port_ext_speeds_counters->FECCorrectableBlockCountrLane[5],
                p_curr_port_ext_speeds_counters->FECCorrectableBlockCountrLane[6],
                p_curr_port_ext_speeds_counters->FECCorrectableBlockCountrLane[7],
                p_curr_port_ext_speeds_counters->FECCorrectableBlockCountrLane[8],
                p_curr_port_ext_speeds_counters->FECCorrectableBlockCountrLane[9],
                p_curr_port_ext_speeds_counters->FECCorrectableBlockCountrLane[10],
                p_curr_port_ext_speeds_counters->FECCorrectableBlockCountrLane[11],
                p_curr_port_ext_speeds_counters->FECUncorrectableBlockCounterLane[0],
                p_curr_port_ext_speeds_counters->FECUncorrectableBlockCounterLane[1],
                p_curr_port_ext_speeds_counters->FECUncorrectableBlockCounterLane[2],
                p_curr_port_ext_speeds_counters->FECUncorrectableBlockCounterLane[3],
                p_curr_port_ext_speeds_counters->FECUncorrectableBlockCounterLane[4],
                p_curr_port_ext_speeds_counters->FECUncorrectableBlockCounterLane[5],
                p_curr_port_ext_speeds_counters->FECUncorrectableBlockCounterLane[6],
                p_curr_port_ext_speeds_counters->FECUncorrectableBlockCounterLane[7],
                p_curr_port_ext_speeds_counters->FECUncorrectableBlockCounterLane[8],
                p_curr_port_ext_speeds_counters->FECUncorrectableBlockCounterLane[9],
                p_curr_port_ext_speeds_counters->FECUncorrectableBlockCounterLane[10],
                p_curr_port_ext_speeds_counters->FECUncorrectableBlockCounterLane[11]);
            else
                sprintf(curr_counters_line,
                        ",-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1" \
                        ",-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1");
            sstream << curr_counters_line;

            if (p_curr_port_ext_speeds_rsfec_counters)
                sprintf(
                curr_counters_line,
                ",%u,%u,%u,%u,%u,%u,%u,%u,%u,%u,%u,%u"\
                ",%u,%u,%u",
                p_curr_port_ext_speeds_rsfec_counters->FECCorrectedSymbolCounterLane[0],
                p_curr_port_ext_speeds_rsfec_counters->FECCorrectedSymbolCounterLane[1],
                p_curr_port_ext_speeds_rsfec_counters->FECCorrectedSymbolCounterLane[2],
                p_curr_port_ext_speeds_rsfec_counters->FECCorrectedSymbolCounterLane[3],
                p_curr_port_ext_speeds_rsfec_counters->FECCorrectedSymbolCounterLane[4],
                p_curr_port_ext_speeds_rsfec_counters->FECCorrectedSymbolCounterLane[5],
                p_curr_port_ext_speeds_rsfec_counters->FECCorrectedSymbolCounterLane[6],
                p_curr_port_ext_speeds_rsfec_counters->FECCorrectedSymbolCounterLane[7],
                p_curr_port_ext_speeds_rsfec_counters->FECCorrectedSymbolCounterLane[8],
                p_curr_port_ext_speeds_rsfec_counters->FECCorrectedSymbolCounterLane[9],
                p_curr_port_ext_speeds_rsfec_counters->FECCorrectedSymbolCounterLane[10],
                p_curr_port_ext_speeds_rsfec_counters->FECCorrectedSymbolCounterLane[11],
                p_curr_port_ext_speeds_rsfec_counters->PortFECCorrectableBlockCounter,
                p_curr_port_ext_speeds_rsfec_counters->PortFECUncorrectableBlockCounter,
                p_curr_port_ext_speeds_rsfec_counters->PortFECCorrectedSymbolCounter);
            else
                sprintf(curr_counters_line,
                        ",-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1"\
                        ",-1,-1,-1");
            sstream << curr_counters_line;
        }

        // calculated counters
        p_port_calc_counters = this->fabric_extended_info.getPMPortCalcCounters(i);

        if (!p_port_calc_counters) {
            sstream << ",0xfffffffffffffffe";
        } else {
            memset(curr_counters_line, 0, sizeof(curr_counters_line));
            sprintf(curr_counters_line,
                    "," U64H_FMT,
                    p_port_calc_counters->RetransmissionPerSec);
            sstream << curr_counters_line;
        }

        struct VendorSpec_PortLLRStatistics *p_curr_port_llr_statistics =
            this->fabric_extended_info.getVSPortLLRStatistics(i);
        if (!p_curr_port_llr_statistics ||
            !this->capability_module.IsSupportedGMPCapability(
                p_curr_port->p_node,
                EnGMPCapIsMaxRetransmissionRateSupported))
            sstream << ",-1";
        else {
            memset(curr_counters_line, 0, sizeof(curr_counters_line));
            sprintf(curr_counters_line, "," U16H_FMT,
                    p_curr_port_llr_statistics->MaxRetransmissionRate);
            sstream << curr_counters_line;
        }

        // PortRcvErrorDetails and PortXmitDiscardDetails
        p_port_rcv_error_details = this->fabric_extended_info.getPMPortRcvErrorDetails(i);
        p_port_xmit_discard_details = this->fabric_extended_info.getPMPortXmitDiscardDetails(i);

        struct PortSampleControlOptionMask * p_option_mask;
        p_option_mask = this->fabric_extended_info.getPMOptionMask(p_curr_port->p_node->createIndex);

        if (!p_port_rcv_error_details) {
            sprintf(curr_counters_line, ",-1,-1,-1,-1,-1,-1");
        } else {
            sprintf(curr_counters_line, ",%d,%d,%d,%d,%d,%d",
                    (p_option_mask->PortRcvErrorDetails_PortLocalPhysicalErrors)?
                        p_port_rcv_error_details->PortLocalPhysicalErrors : -1,
                    (p_option_mask->PortRcvErrorDetails_PortMalformedPacketErrors)?
                        p_port_rcv_error_details->PortMalformedPacketErrors : -1,
                    (p_option_mask->PortRcvErrorDetails_PortBufferOverrunErrors)?
                        p_port_rcv_error_details->PortBufferOverrunErrors : -1,
                    (p_option_mask->PortRcvErrorsDetails_PortDLIDMappingErrors)?
                        p_port_rcv_error_details->PortDLIDMappingErrors : -1,
                    (p_option_mask->PortRcvErrorsDetails_PortVlMappingErrors)?
                        p_port_rcv_error_details->PortVLMappingErrors : -1,
                    (p_option_mask->PortRcvErrorDetails_PortLoopingErrors)?
                        p_port_rcv_error_details->PortLoopingErrors : -1);
        }
        sstream << curr_counters_line;

        if (!p_port_xmit_discard_details) {
            sprintf(curr_counters_line, ",-1,-1,-1,-1");
        } else {
            sprintf(curr_counters_line, ",%d,%d,%d,%d",
                    (p_option_mask->PortXmitDiscardDetails_PortInactiveDiscards)?
                        p_port_xmit_discard_details->PortInactiveDiscards : -1,
                    (p_option_mask->PortXmitDiscardsDetails_PortNeighborMTUDiscards)?
                        p_port_xmit_discard_details->PortNeighborMTUDiscards : -1,
                    (p_option_mask->PortXmitDiscardDetails_PortSwLifetimeLimitDiscards)?
                        p_port_xmit_discard_details->PortSwLifetimeLimitDiscards : -1,
                    (p_option_mask->PortXmitDiscardDetails_PortSwHOQLifeLimitDiscards)?
                        p_port_xmit_discard_details->PortSwHOQLifetimeLimitDiscards : -1);
        }
        sstream << curr_counters_line;

        sstream << endl;
        csv_out.WriteLine(sstream.str());
    }

    csv_out.DumpEnd(SECTION_PM_INFO);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::ResetPortCounters(list_p_fabric_general_err& ports_errors,
                              progress_func_nodes_t progress_func,
                              u_int32_t check_counters_bitset)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ResetAppData();
    ibDiagClbck.Set(this, &fabric_extended_info, &ports_errors);

    int rc = BuildClassPortInfoDB(ports_errors);
    //don't exit, if only some of the nodes failed to retrieve ClassPortInfo
    if (rc && rc != IBDIAG_ERR_CODE_FABRIC_ERROR) {
        IBDIAG_RETURN(rc);
    }

    progress_bar_nodes_t progress_bar_nodes;
    CLEAR_STRUCT(progress_bar_nodes);

    u_int32_t port_info_cap_mask = 0;
    u_int16_t curr_pm_cap_mask = 0;
    bool to_read_cap = true;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;

    // go over all nodes:
    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end();
            ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        ++progress_bar_nodes.nodes_found;
        if (p_curr_node->type == IB_SW_NODE)
            ++progress_bar_nodes.sw_found;
        else
            ++progress_bar_nodes.ca_found;
        if (progress_func)
            progress_func(&progress_bar_nodes, &this->discover_progress_bar_nodes);

        //read capability for new node
        to_read_cap = true;

        //check LLR support by DeviceID
        struct SMP_NodeInfo * p_curr_node_info =
            this->fabric_extended_info.getSMPNodeInfo(p_curr_node->createIndex);
        if (!p_curr_node_info) {
            this->SetLastError("DB error - failed to get SMPNodeInfo for node: %s", p_curr_node->name.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }
        if (!this->capability_module.IsSupportedGMPCapability(
            p_curr_node, EnGMPCapIsPortLLRStatisticsSupported)) {
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "node %s: DeviceID %u (0x%x) Does not support LLR, skipping \n",
                       p_curr_node->name.c_str(),
                       p_curr_node_info->DeviceID,
                       p_curr_node_info->DeviceID);

            p_curr_node->appData1.val |= NOT_SUPPORT_LLR_COUNTERS;
        }

        // go over all ports of this node:
        for (u_int32_t i = 1; i <= p_curr_node->numPorts; ++i) { //start from port 1, no need to reset port 0
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            //for switch read cap only on port 0, for others read on every port
            if (to_read_cap) {
                rc = ReadCapMask(p_curr_node, p_curr_port,
                                 curr_pm_cap_mask, port_info_cap_mask);
                if (rc) {
                    if (rc == IBDIAG_ERR_CODE_FABRIC_ERROR) {
                        break;     //not reasonable to continue, goto next_node
                    }
                    goto exit;
                }
                if (p_curr_node->type == IB_SW_NODE)
                    to_read_cap = false;
            }

            //send port counters mad set for this port
            clbck_data.m_handle_data_func = IBDiagPMPortCountersClearClbck;
            clbck_data.m_data1 = p_curr_port;
            int rc2 = this->ibis_obj.PMPortCountersClear(p_curr_port->base_lid,
                                                         p_curr_port->num,
                                                         &clbck_data);

            if (rc2)
                continue;  //we will try to proceed without query for extended counters

            //check if counters extended are supported
            if (!IS_SUPPORT_EXT_PORT_FAILED(p_curr_node->appData1.val)) {
                if ( IS_SUPPORT_EXTENDED_COUNTERS(curr_pm_cap_mask) ||
                     IS_SUPPORT_EXTENDED_COUNTERS_NO_IETF(curr_pm_cap_mask)) {

                    clbck_data.m_handle_data_func = IBDiagPMPortCountersExtendedClearClbck;
                    this->ibis_obj.PMPortCountersExtendedClear(p_curr_port->base_lid,
                        p_curr_port->num, &clbck_data);
                } else {
                    FabricErrNodeNotSupportCap *p_curr_fabric_pm_err =
                            new FabricErrNodeNotSupportCap(p_curr_node,
                                    "This device does not support extended port counters capability");
                    if (!p_curr_fabric_pm_err) {
                        this->SetLastError("Failed to allocate FabricErrPortNotRespond");
                        rc = IBDIAG_ERR_CODE_NO_MEM;
                        goto exit;
                    }
                    ports_errors.push_back(p_curr_fabric_pm_err);
                    rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

                    //mark we already checked for support
                    p_curr_node->appData1.val |= NOT_SUPPORT_EXT_PORT_COUNTERS;
                }
            }

            //check if extended speed counters supported
            if(IS_TO_CHECK_EXT_SPEEDS_COUNTERS(check_counters_bitset, p_curr_node->type) &&
               !IS_SUPPORT_EXT_SPEEDS_FAILED(p_curr_node->appData1.val)) {
                if(IS_SUPPORT_EXTENDED_SPEEDS_COUNTERS(port_info_cap_mask)) {
                    //if RS_FEC is active
                    if (isRSFEC(p_curr_port->get_fec_mode())) {
                        if (!IS_SUPPORT_EXT_SPEEDS_RSFEC_FAILED(
                            p_curr_node->appData1.val)) {
                            if (IS_SUPPORT_EXT_SPEEDS_RSFEC_COUNTERS(
                                curr_pm_cap_mask)) {
                                clbck_data.m_handle_data_func =
                                    IBDiagPMPortExtendedSpeedsRSFECClearClbck;
                                this->ibis_obj.
                                    PMPortExtendedSpeedsRSFECCountersClear(
                                        p_curr_port->base_lid, p_curr_port->num,
                                        &clbck_data);
                            } else {
                                FabricErrNodeNotSupportCap *p_curr_fabric_pm_err =
                                        new FabricErrNodeNotSupportCap(
                                            p_curr_node, "This device does not "
                                            "support port extended speeds RSFEC "
                                            "counters capability");
                                if (!p_curr_fabric_pm_err) {
                                    this->SetLastError("Failed to allocate "
                                                       "FabricErrPortNotRespond");
                                    rc = IBDIAG_ERR_CODE_NO_MEM;
                                    goto exit;
                                }
                                ports_errors.push_back(p_curr_fabric_pm_err);
                                rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

                                //mark we already checked for support
                                p_curr_node->appData1.val |=
                                    NOT_SUPPORT_EXT_SPEEDS_RSFEC_COUNTERS;
                            }
                        }
                    } else { //no RS_FEC
                        clbck_data.m_handle_data_func =
                            IBDiagPMPortExtendedSpeedsClearClbck;
                        this->ibis_obj.PMPortExtendedSpeedsCountersClear(
                                        p_curr_port->base_lid,
                                        p_curr_port->num, &clbck_data);
                    }
                } else {

                    FabricErrNodeNotSupportCap *p_curr_fabric_pm_err =
                            new FabricErrNodeNotSupportCap(p_curr_node,
                                    "This device does not support port extended speeds counters capability");
                    if (!p_curr_fabric_pm_err) {
                        this->SetLastError("Failed to allocate FabricErrPortNotRespond");
                        rc = IBDIAG_ERR_CODE_NO_MEM;
                        goto exit;
                    }
                    ports_errors.push_back(p_curr_fabric_pm_err);
                    rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

                    //mark we already checked for support
                    p_curr_node->appData1.val |= NOT_SUPPORT_EXT_SPEEDS_COUNTERS;
                }

            } //ext_speeds_counters

            //always clear LLR counters and symbol error counter
            if(!IS_SUPPORT_LLR_FAILED(p_curr_node->appData1.val)) {
                clbck_data.m_data2 = (void*)IS_TO_PRINT_LLR_COUNTERS(check_counters_bitset);
                clbck_data.m_handle_data_func =
                    IBDiagVSPortLLRStatisticsClearClbck;
                this->ibis_obj.VSPortLLRStatisticsClear(
                        p_curr_port->base_lid,
                        p_curr_port->num,
                        true,
                        &clbck_data);
            }

            // PortRcvErrorDetails and PortXmitDiscardDetails

            if (!IS_SUPPORT_PORT_RCV_ERROR_DETAILS(p_curr_node->appData1.val)) {
               clbck_data.m_handle_data_func = IBDiagPMPortRcvErrorDetailsClearClbck;
               this->ibis_obj.PMPortRcvErrorDetailsClear(p_curr_port->base_lid,
                                                         p_curr_port->num,
                                                         &clbck_data);
            }

            if (!IS_SUPPORT_PORT_XMIT_DISCARD_DETAILS(p_curr_node->appData1.val)) {
                clbck_data.m_handle_data_func = IBDiagPMPortXmitDiscardDetailsClearClbck;
                this->ibis_obj.PMPortXmitDiscardDetailsClear(p_curr_port->base_lid,
                                                             p_curr_port->num,
                                                             &clbck_data);
            }

        } // port loop

        if (ibDiagClbck.GetState())
            goto exit;
    } //node loop
exit:
    this->ibis_obj.MadRecAll();

    if (rc) {
        if (last_error.empty())
            this->SetLastError("ResetPortCounters Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "ResetPortCounters Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!ports_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / IBDIAG_ERR_CODE_NO_MEM / ERR_CODE_INCORRECT_ARGS / IBDIAG_ERR_CODE_CHECK_FAILED
int IBDiag::CalcCounters(vector_p_pm_info_obj& prev_pm_info_obj_vec,
                         double diff_time_between_samples,
                         list_p_fabric_general_err& pm_errors) {
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    int rc2 = IBDIAG_SUCCESS_CODE;
    int llr_rcv_retry_index = -1;
    u_int64_t value1 = 0, value2 = 0;
    struct PM_PortCalcCounters curr_port_calc_counters;
    double rtr = 0.0;

    rc = counter_name_2_index("port_rcv_retry", llr_rcv_retry_index);
    if ( rc )
        IBDIAG_RETURN(rc);

    for ( u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i ) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if ( !p_curr_port )           //port not exists
            continue;
        if ( prev_pm_info_obj_vec.size() < i + 1 )    //info not exists
            continue;

        if ( !prev_pm_info_obj_vec[i] )               //info not exists
            continue;

        //llr_counters
        struct VendorSpec_PortLLRStatistics *p_curr_port_llr_statistics1 =
        prev_pm_info_obj_vec[i]->p_port_llr_statistics;
        struct VendorSpec_PortLLRStatistics *p_curr_port_llr_statistics2 =
        this->fabric_extended_info.getVSPortLLRStatistics(i);

        if (!p_curr_port_llr_statistics1 || !p_curr_port_llr_statistics2) {
            continue;
        }

        // check for overflow of the base counters
        value1 = p_curr_port_llr_statistics1->PortRcvRetry;
        value2 = p_curr_port_llr_statistics2->PortRcvRetry;

        // check for base counters overflow
        if ( (pm_counters_arr[llr_rcv_retry_index].overflow_value != OVERFLOW_VAL_NONE) &&
             (value1 == pm_counters_arr[llr_rcv_retry_index].overflow_value ||
              value2 == pm_counters_arr[llr_rcv_retry_index].overflow_value ||
              value1 > value2) ) {
            FabricErrPMBaseCalcCounterOverflow *p_curr_fabric_pm_err =
            new FabricErrPMBaseCalcCounterOverflow(p_curr_port, "retransmission_per_sec");
            if ( !p_curr_fabric_pm_err ) {
                this->SetLastError("Failed to allocate FabricErrPMCounterOverflow");
                IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
            }
            pm_errors.push_back(p_curr_fabric_pm_err);
            // don't store, i.e retransmission_per_sec is NA
            IBDIAG_LOG(TT_LOG_LEVEL_INFO, "Calc rtr for (%s). value1=" U64H_FMT ", value2=" U64H_FMT ", "
                       "retransmission_per_sec=NA (base is overflown), diff_time=%f\n",
                       p_curr_port->getName().c_str(), value1, value2,
                       diff_time_between_samples);
            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
        } else {
            // calculate retransmission rate  and store it

            //sanity check
            if (diff_time_between_samples == 0) {
                IBDIAG_LOG(TT_LOG_LEVEL_INFO,
                           "Can't calculate the retransmission_per_sec counter, as the diff time is zero");
                IBDIAG_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
            }
            rtr = (double)(value2 - value1)/ diff_time_between_samples;
            curr_port_calc_counters.RetransmissionPerSec = (u_int64_t)rtr;

            IBDIAG_LOG(TT_LOG_LEVEL_INFO, "Calc rtr for (%s). value1=" U64H_FMT ", value2=" U64H_FMT ", "
                       "retransmission_per_sec=" U64H_FMT ", diff_time=%f\n",
                       p_curr_port->getName().c_str(), value1, value2,
                       curr_port_calc_counters.RetransmissionPerSec, diff_time_between_samples);

            rc2 = this->fabric_extended_info.addPMPortCalculatedCounters(p_curr_port, curr_port_calc_counters);
            if ( rc2 ) {
                this->SetLastError("Failed to add PM_PortCalcCounters for port=%s, err=%s",
                                   p_curr_port->getName().c_str(), this->fabric_extended_info.GetLastError());
                IBDIAG_RETURN(rc2);
            }
        }
    }

    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / IBDIAG_ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_CHECK_FAILED / ERR_CODE_INCORRECT_ARGS
int IBDiag::CheckAllPMValues(list_p_fabric_general_err& pm_errors,
                             map_str_uint64& counters_to_threshold_map,
                             u_int32_t check_counters_bitset)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    u_int64_t all_errs_counters_threshold = 0;

    /*
     * Step 1 -
     *  Check if need to check all counters according to default threshold.
     *  If all is specified than all errors counters (not including extended)
     *  need to be checked.
     */
    bool need_to_check_all_errs_counters = (counters_to_threshold_map.find("all") != counters_to_threshold_map.end());
    if (need_to_check_all_errs_counters) {
        all_errs_counters_threshold = counters_to_threshold_map["all"];
    }

    for (int i = 0; i < PM_COUNTERS_ARR_SIZE; ++i) {
        //if not given specific threshold already
        if (counters_to_threshold_map.find(pm_counters_arr[i].name) == counters_to_threshold_map.end()) {
            // if not subject to keyword 'all' in -PC parameter, but has its own default threshold,
            // then use its own threshold even when 'all' is used
            if (!pm_counters_arr[i].subject_to_threshold_keyword_all) {
                if (pm_counters_arr[i].threshold != -1)
                    counters_to_threshold_map[pm_counters_arr[i].name] = pm_counters_arr[i].threshold;

            } else {
                // check if 'all' was given
                if (need_to_check_all_errs_counters)
                    counters_to_threshold_map[pm_counters_arr[i].name] = all_errs_counters_threshold;
            }
        }
    }

    /*
     * Step 2 -
     *  Building array of theshold by counters index.
     *  This will take O(n) but in the next steps we will be able to access
     *  the counters threshold by O(1).
     */
    u_int64_t *p_counters_thresh_by_idx[PM_COUNTERS_ARR_SIZE];
    for (int i = 0; i < PM_COUNTERS_ARR_SIZE; ++i)
        p_counters_thresh_by_idx[i] = NULL;
    for (map_str_uint64::iterator it = counters_to_threshold_map.begin();
            it != counters_to_threshold_map.end(); ++it) {
        if ((*it).first == "all")           //we already took care of that
            continue;
        for (int i = 0; i < PM_COUNTERS_ARR_SIZE; ++i) {
            if ((*it).first == pm_counters_arr[i].name) {
                p_counters_thresh_by_idx[i] = &((*it).second);
                break;
            }
            if (i == PM_COUNTERS_ARR_SIZE) {
                this->SetLastError("Unknown counter name %s was given as arg", it->first.c_str());
                IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);
            }
        }
    }

    /*
     * Step 3 -
     *  Go over all ports
     */
    for (u_int32_t curr_port_idx = 0;
            curr_port_idx < this->fabric_extended_info.getPortsVectorSize();
            ++curr_port_idx) {
        /*
         * Step 3.1 -
         *  Get all needed data for checks
         */
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(curr_port_idx);
        //Skip ports that don't exist or are down or have undefined state
        //or are not in the scope of the fabric.
        if (!p_curr_port || !p_curr_port->is_data_worthy())
            continue;

        IBNode *p_curr_node = p_curr_port->p_node;
        if (!p_curr_node) {
            this->SetLastError("DB error - found port with null node = %s", p_curr_port->getName().c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }
        u_int16_t *pm_cap_mask = this->fabric_extended_info.getPMCapMask(p_curr_node->createIndex);
        if (!pm_cap_mask)           //already reported to user no reason to continue here
            continue;

        // Extended Speeds Capability
        //bool* tmp_ext_speeds_cap;
        //bool ext_speeds_cap;
        //tmp_ext_speeds_cap = this->fabric_extended_info.getPMPortExtSpeedsCap(p_curr_node->createIndex);
        //ext_speeds_cap = tmp_ext_speeds_cap && *tmp_ext_speeds_cap;

        struct PM_PortCounters *p_curr_port_counters =
                this->fabric_extended_info.getPMPortCounters(curr_port_idx);
        struct PM_PortCountersExtended *p_curr_port_counters_extended =
                this->fabric_extended_info.getPMPortCountersExtended(curr_port_idx);
        struct PM_PortExtendedSpeedsCounters *p_curr_port_ext_speeds_cnts =
                this->fabric_extended_info.getPMPortExtSpeedsCounters(curr_port_idx);
        struct PM_PortExtendedSpeedsRSFECCounters
            *p_curr_port_ext_speeds_rsfec_cnts =
            this->fabric_extended_info.getPMPortExtSpeedsRSFECCounters(
                                        curr_port_idx);
        struct VendorSpec_PortLLRStatistics *p_curr_port_llr_statistics =
                this->fabric_extended_info.getVSPortLLRStatistics(curr_port_idx);
        struct PM_PortCalcCounters *p_curr_port_calc_counters =
            this->fabric_extended_info.getPMPortCalcCounters(curr_port_idx);

        /*
         * Step 3.2 -
         *  Go over all counters
         */
        u_int64_t value;
        u_int64_t thresh;
        list_p_pm_err curr_port_errors;
        for (int curr_counter_idx = 0; curr_counter_idx < PM_COUNTERS_ARR_SIZE; ++curr_counter_idx) {
            // Skip untested counters
            if( pm_counters_arr[curr_counter_idx].counter_src == PM_PORT_EXT_SPEEDS_CNTS_MAD &&
                (!IS_TO_CHECK_EXT_SPEEDS_COUNTERS(check_counters_bitset, p_curr_node->type)))
                continue;
            // Skip LLR counters - we don't check them, only report the value
            if(pm_counters_arr[curr_counter_idx].counter_src ==
               VS_PORT_LLR_CNTS_MAD_HIDDEN)
                continue;

            /*
             * Step 3.2.1 -
             *  skip unsupported counters
             */
            if (((pm_counters_arr[curr_counter_idx].counter_src == PM_PORT_CNT_EXT_MAD ||
                  pm_counters_arr[curr_counter_idx].counter_src ==
                    PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD) &&
                !is_counter_supported(*pm_cap_mask, pm_counters_arr[curr_counter_idx].cap_mask_bits)) ||
                (pm_counters_arr[curr_counter_idx].counter_src == PM_PORT_EXT_SPEEDS_CNTS_MAD &&
                 !p_curr_port_ext_speeds_cnts)) {
                continue;           //go to next counter
            }

            /*
             * Step 3.2.2 -
             *  Get counter value - regular / extended
             */
            value = 0;
            switch (pm_counters_arr[curr_counter_idx].counter_src) {
                case PM_PORT_CNT_MAD:
                {
                    if (!p_curr_port_counters)          //info not exist
                        continue;
                    rc = get_value((u_int8_t *)p_curr_port_counters + pm_counters_arr[curr_counter_idx].struct_offset,
                            pm_counters_arr[curr_counter_idx].real_size, value);
                    break;
                }
                case PM_PORT_CNT_EXT_MAD:
                {
                    if (!p_curr_port_counters_extended)
                        continue;                       //info not exist
                    rc = get_value((u_int8_t *)p_curr_port_counters_extended + pm_counters_arr[curr_counter_idx].struct_offset,
                                   pm_counters_arr[curr_counter_idx].real_size, value);
                    break;
                }
                case PM_PORT_EXT_SPEEDS_CNTS_MAD:
                {
                    if (!p_curr_port_ext_speeds_cnts)
                        continue;
                    rc = get_value((u_int8_t*)p_curr_port_ext_speeds_cnts + pm_counters_arr[curr_counter_idx].struct_offset,
                            pm_counters_arr[curr_counter_idx].real_size, value);
                    break;
                }
                case PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD:
                {
                    if (!p_curr_port_ext_speeds_rsfec_cnts)
                        continue;
                    rc = get_value((u_int8_t*)p_curr_port_ext_speeds_rsfec_cnts + pm_counters_arr[curr_counter_idx].struct_offset,
                            pm_counters_arr[curr_counter_idx].real_size, value);
                    break;
                }
                case VS_PORT_LLR_CNTS_MAD:
                {
                    if (!p_curr_port_llr_statistics)
                        continue;
                    rc = get_value((u_int8_t*)p_curr_port_llr_statistics + pm_counters_arr[curr_counter_idx].struct_offset,
                            pm_counters_arr[curr_counter_idx].real_size, value);
                    break;
                }
                case PM_PORT_CALCULATED_COUNTER:
                {
                    if (!p_curr_port_calc_counters) {
                        continue;
                    }
                    rc = get_value((u_int8_t*)p_curr_port_calc_counters + pm_counters_arr[curr_counter_idx].struct_offset,
                            pm_counters_arr[curr_counter_idx].real_size, value);
                    break;
                }
                default:
                    continue;

                if (rc) {
                    this->SetLastError("Invalid pm counter size: %s. Counter real_size is %d",
                                       pm_counters_arr[curr_counter_idx].name.c_str(),
                                       pm_counters_arr[curr_counter_idx].real_size);
                    IBDIAG_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
                }
            }

            /*
             * Step 3.2.3 -
             *  Check if overflowed (except for calculated counters) or exceeds threshold value
             */
            if (pm_counters_arr[curr_counter_idx].counter_src != PM_PORT_CALCULATED_COUNTER &&
                (pm_counters_arr[curr_counter_idx].overflow_value != OVERFLOW_VAL_NONE) &&
                (value == pm_counters_arr[curr_counter_idx].overflow_value)) {
                FabricErrPMCounterOverflow *p_curr_fabric_pm_err =
                        new FabricErrPMCounterOverflow(p_curr_port, pm_counters_arr[curr_counter_idx].name,
                                pm_counters_arr[curr_counter_idx].overflow_value);
                if (!p_curr_fabric_pm_err) {
                    this->SetLastError("Failed to allocate FabricErrPMCounterOverflow");
                    IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
                }
                curr_port_errors.push_back(p_curr_fabric_pm_err);
            } else if (p_counters_thresh_by_idx[curr_counter_idx]) {
                thresh = *(p_counters_thresh_by_idx[curr_counter_idx]);
                if (value > thresh) {
                    FabricErrPMCounterExceedThreshold *p_curr_fabric_pm_err =
                            new FabricErrPMCounterExceedThreshold(p_curr_port,
                                    pm_counters_arr[curr_counter_idx].name,
                                    thresh, value);
                    if (!p_curr_fabric_pm_err) {
                        this->SetLastError("Failed to allocate FabricErrPMCounterExceedThreshold");
                        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
                    }
                    curr_port_errors.push_back(p_curr_fabric_pm_err);
                }
            }
        }   //counters loop

        /*
         * Step 3.3 -
         *  Merge all counters errors
         */
        if (!curr_port_errors.empty()) {
            FabricErrPMCountersAll *p_curr_fabric_pm_err =
                    new FabricErrPMCountersAll(p_curr_port, curr_port_errors);
            if (!p_curr_fabric_pm_err) {
                this->SetLastError("Failed to allocate FabricErrPMCountersAll");
                IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
            }
            pm_errors.push_back(p_curr_fabric_pm_err);
            rc = IBDIAG_ERR_CODE_CHECK_FAILED;

            for (list_p_pm_err::iterator it = curr_port_errors.begin();
                    it != curr_port_errors.end(); ++it)
                delete((*it));
            curr_port_errors.clear();
        }
    }   //ports loop

    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_NO_MEM / ERR_CODE_CHECK_FAILED
int IBDiag::CheckCountersDiff(vector_p_pm_info_obj& prev_pm_info_obj_vec,
        list_p_fabric_general_err& pm_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    u_int64_t value1, value2, diff_value;
    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port)           //port not exists
            continue;
        if (prev_pm_info_obj_vec.size() < i + 1)    //info not exists
            continue;

        if (!prev_pm_info_obj_vec[i])               //info not exists
            continue;

        //regular counters
        struct PM_PortCounters *p_curr_port_counters1 = prev_pm_info_obj_vec[i]->p_port_counters;
        if (!p_curr_port_counters1)                 //info not exists
            continue;
        struct PM_PortCounters *p_curr_port_counters2 = this->fabric_extended_info.getPMPortCounters(i);
        if (!p_curr_port_counters2)                 //info not exists
            continue;

        //extended_port_counters
        struct PM_PortCountersExtended *p_curr_extended_port_counters1 =
            prev_pm_info_obj_vec[i]->p_extended_port_counters;
        struct PM_PortCountersExtended *p_curr_extended_port_counters2 =
            this->fabric_extended_info.getPMPortCountersExtended(i);

        //ext_speeds_counters
        struct PM_PortExtendedSpeedsCounters *p_curr_port_ext_speeds_counters1 =
            prev_pm_info_obj_vec[i]->p_port_ext_speeds_counters;
        struct PM_PortExtendedSpeedsCounters *p_curr_port_ext_speeds_counters2 =
            this->fabric_extended_info.getPMPortExtSpeedsCounters(i);

        //ext_speeds_counters for rsfec
        struct PM_PortExtendedSpeedsRSFECCounters
            *p_curr_port_ext_speeds_rsfec_counters1 =
            prev_pm_info_obj_vec[i]->p_port_ext_speeds_rsfec_counters;
        struct PM_PortExtendedSpeedsRSFECCounters
            *p_curr_port_ext_speeds_rsfec_counters2 =
            this->fabric_extended_info.getPMPortExtSpeedsRSFECCounters(i);

        //for max retransmission rate 
        // For all counters except max retransmission rate,
        // counter_src = VS_PORT_LLR_CNTS_MAD_HIDDEN 
        struct VendorSpec_PortLLRStatistics *p_curr_port_llr_statistics1 =
            prev_pm_info_obj_vec[i]->p_port_llr_statistics;
        struct VendorSpec_PortLLRStatistics *p_curr_port_llr_statistics2 =
            this->fabric_extended_info.getVSPortLLRStatistics(i);


        u_int8_t * p_curr_counters1 = NULL;
        u_int8_t * p_curr_counters2 = NULL;

        for (int j = 0; j < PM_COUNTERS_ARR_SIZE; ++j) {
            if (!pm_counters_arr[j].diff_check_threshold)   //checking only error counters
                continue;

            switch(pm_counters_arr[j].counter_src) {
                case PM_PORT_CNT_MAD:
                    p_curr_counters1 = (u_int8_t *)p_curr_port_counters1;
                    p_curr_counters2 = (u_int8_t *)p_curr_port_counters2;
                    break;
                case PM_PORT_CNT_EXT_MAD:
                    p_curr_counters1 = (u_int8_t *)p_curr_extended_port_counters1;
                    p_curr_counters2 = (u_int8_t *)p_curr_extended_port_counters2;
                    break;
                case PM_PORT_EXT_SPEEDS_CNTS_MAD:
                    p_curr_counters1 = (u_int8_t *)p_curr_port_ext_speeds_counters1;
                    p_curr_counters2 = (u_int8_t *)p_curr_port_ext_speeds_counters2;
                    break;
                case PM_PORT_EXT_SPEEDS_RSFEC_CNTS_MAD:
                    p_curr_counters1 =
                        (u_int8_t *)p_curr_port_ext_speeds_rsfec_counters1;
                    p_curr_counters2 =
                        (u_int8_t *)p_curr_port_ext_speeds_rsfec_counters2;
                    break;
                case VS_PORT_LLR_CNTS_MAD:
                    p_curr_counters1 =
                        (u_int8_t *)p_curr_port_llr_statistics1;
                    p_curr_counters2 =
                        (u_int8_t *)p_curr_port_llr_statistics2;
                    break;

                default:
                    continue;
            }

            if (!p_curr_counters1 || !p_curr_counters2)                 //info not exists
                continue;

            value1 = value2 = 0;
            rc  = get_value(p_curr_counters1 + pm_counters_arr[j].struct_offset,
                    pm_counters_arr[j].real_size, value1);
            rc |= get_value(p_curr_counters2 + pm_counters_arr[j].struct_offset,
                    pm_counters_arr[j].real_size, value2);
            if (rc) {
                this->SetLastError("Invalid pm counter size: %s. Counter real_size is %d",
                                   pm_counters_arr[j].real_size,
                                   pm_counters_arr[j].name.c_str());
                IBDIAG_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
            }

            diff_value = value2 - value1;

            if (diff_value >= pm_counters_arr[j].diff_check_threshold) {
                FabricErrPMErrCounterIncreased *p_curr_fabric_pm_err =
                        new FabricErrPMErrCounterIncreased(p_curr_port,
                                pm_counters_arr[j].name,
                                pm_counters_arr[j].diff_check_threshold,
                                diff_value);
                if (!p_curr_fabric_pm_err) {
                    this->SetLastError("Failed to allocate FabricErrPMErrCounterIncreased");
                    IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
                }
                pm_errors.push_back(p_curr_fabric_pm_err);
                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            }
        }
    }
    IBDIAG_RETURN(rc);
}

/*
 * time - in seconds units
 * symbol errors - delta of symbol error counter in this time slot
 * ber - to return calculated ber
 */
int IBDiag::CalcBER(IBPort *p_curr_port,
                    double time,
                    u_int64_t symbol_error,
                    long double &reciprocal_ber)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    long double ber = 0;
    IBLinkSpeed port_speed = p_curr_port->get_common_speed();

    if (!symbol_error) {
        reciprocal_ber = 0.0;
        IBDIAG_RETURN(rc);
    }

    //calculate BER
    /* BER = errors / data_bits ==>
     * diff_errors / delta(t) * link rate
     */
    long double link_rate =
        (long double)(CalcLinkRate(p_curr_port->get_common_width(),
                                   port_speed));

    reciprocal_ber =
        (time * link_rate) / symbol_error;

    SMP_MlnxExtPortInfo *p_curr_mlnx_ext_port_info =
    this->fabric_extended_info.getSMPMlnxExtPortInfo(p_curr_port->createIndex);

    if (p_curr_mlnx_ext_port_info &&
            (this->GetLLRActiveCellSize() ||
            (p_curr_mlnx_ext_port_info->RetransMode ==
             RETRANS_LLR_ACTIVE_CELL_64 ||
             p_curr_mlnx_ext_port_info->RetransMode ==
             RETRANS_LLR_ACTIVE_CELL_128))) {
        // if reciprocal_ber=0 need to avoid dividing by zero
        if (!reciprocal_ber) {
            reciprocal_ber = 0.0;
            IBDIAG_RETURN(rc);
        }

        unsigned int cell_size_bits = llr_active_cell2size(
                p_curr_mlnx_ext_port_info->RetransMode) * 8;

        //check that we have a valid active cell size
        if (!cell_size_bits)
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        // for llr, ber = symbol error/time
        ber = 1/reciprocal_ber;

        long double p2_errors = (1.0 -
                                 (powl((1.0 - ber), cell_size_bits) +
                                  cell_size_bits *
                                  ber *
                                  powl((1.0 - ber), (cell_size_bits - 1))));
        long double effective_ber = p2_errors * powl(2, -16.0);
        reciprocal_ber = 1/effective_ber;
    }

    IBDIAG_RETURN(rc);
}


//The formulas here are as in Raw BER in phy diag
//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::CalcPhyTest(vector_p_pm_info_obj& prev_pm_info_obj_vec,
                        double sec_between_samples,
                        CSVOut &csv_out)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    long double ber = 0.0;
    long double k_rsfec = 0.0;
    char curr_counters_line[256];
    IBFECMode port_fec = IB_FEC_NA;
    stringstream sstream;

    csv_out.DumpStart(SECTION_PHY_TEST);

    sstream << "NodeGUID,PortGUID,PortNumber,Value" << endl;
    csv_out.WriteLine(sstream.str());

    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize();
            ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port)           //port not exists
            continue;
        if (prev_pm_info_obj_vec.size() < i + 1)    //info not exists
            break;

        if (!prev_pm_info_obj_vec[i])               //info not exists
            continue;

        struct PM_PortCounters *p_curr_port_counters1 =
        prev_pm_info_obj_vec[i]->p_port_counters;
        if (!p_curr_port_counters1) {                 //info not exists
            rc = IBDIAG_ERR_CODE_DB_ERR;
            continue;
        }

        struct PM_PortCounters *p_curr_port_counters2 =
            this->fabric_extended_info.getPMPortCounters(i);
        if (!p_curr_port_counters2)        //info not exists
            continue;

        port_fec = p_curr_port->get_fec_mode();
        ber = 0.0;

        if (isRSFEC(port_fec)) {
            struct PM_PortExtendedSpeedsRSFECCounters *p_ext_speeds_rsfec_cnts1 =
                prev_pm_info_obj_vec[i]->p_port_ext_speeds_rsfec_counters;
            if (!p_ext_speeds_rsfec_cnts1)    //info not exists
                continue;

            struct PM_PortExtendedSpeedsRSFECCounters *p_ext_speeds_rsfec_cnts2 =
                this->fabric_extended_info.getPMPortExtSpeedsRSFECCounters(i);
            if (!p_ext_speeds_rsfec_cnts2)    //info not exists
                continue;
            SMP_MlnxExtPortInfo *p_curr_mlnx_ext_port_info =
                this->fabric_extended_info.getSMPMlnxExtPortInfo(p_curr_port->createIndex);
            if (!p_curr_mlnx_ext_port_info)
                continue;

            k_rsfec = p_curr_mlnx_ext_port_info->ActiveRSFECParity / 2 + 1;
            //calculate k_rsfec
            ber = (p_ext_speeds_rsfec_cnts2->PortFECCorrectedSymbolCounter -
                   p_ext_speeds_rsfec_cnts1->PortFECCorrectedSymbolCounter) +
                  k_rsfec *
                  (p_ext_speeds_rsfec_cnts2->PortFECUncorrectableBlockCounter -
                   p_ext_speeds_rsfec_cnts1->PortFECUncorrectableBlockCounter);
        } else {
            struct PM_PortExtendedSpeedsCounters *p_curr_port_ext_speeds_counters1 =
                prev_pm_info_obj_vec[i]->p_port_ext_speeds_counters;
            if (!p_curr_port_ext_speeds_counters1)   //info not exists
                continue;

            struct PM_PortExtendedSpeedsCounters *p_curr_port_ext_speeds_counters2 =
                this->fabric_extended_info.getPMPortExtSpeedsCounters(i);
            if (!p_curr_port_ext_speeds_counters2)   //info not exists
                continue;
            //undefined fec port
            if (port_fec == IB_FEC_NA)
                ber = max(CalcBERFireCodeFEC(p_curr_port_ext_speeds_counters1,
                                             p_curr_port_ext_speeds_counters2),
                          EDPLSum(p_curr_port_ext_speeds_counters1,
                                  p_curr_port_ext_speeds_counters2));
            else if (port_fec == IB_FEC_NO_FEC) {
                //up to QDR (8/10)
                if (p_curr_port->get_common_speed() < IB_LINK_SPEED_14)
                    //symbol error/time
                    ber = p_curr_port_counters2->SymbolErrorCounter -
                          p_curr_port_counters1->SymbolErrorCounter;
                else //64/66
                    //sum of edpls over lanes/time
                    ber = EDPLSum(p_curr_port_ext_speeds_counters1,
                                  p_curr_port_ext_speeds_counters2);
            } else if (port_fec == IB_FEC_FIRECODE_FEC)
                ber = CalcBERFireCodeFEC(p_curr_port_ext_speeds_counters1,
                                         p_curr_port_ext_speeds_counters2);
        }
        long double link_rate =
            (long double)(CalcLinkRate(p_curr_port->get_common_width(),
                        p_curr_port->get_common_speed()));
        long double data_rate =
            sec_between_samples * link_rate;

        if (data_rate)
            ber /= data_rate;
        else {
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto Exit;
        }

        memset(curr_counters_line, 0, sizeof(curr_counters_line));
        sstream.str("");
        sprintf(curr_counters_line, U64H_FMT "," U64H_FMT "," U32D_FMT ","   \
                "%Le",
                p_curr_port->p_node->guid_get(),
                p_curr_port->guid_get(),
                p_curr_port->num,
                ber);
        sstream << curr_counters_line << endl;
        csv_out.WriteLine(sstream.str());
    }

Exit:
    csv_out.DumpEnd(SECTION_PHY_TEST);
    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_NO_MEM / ERR_CODE_CHECK_FAILED
int IBDiag::CalcBERErrors(vector_p_pm_info_obj& prev_pm_info_obj_vec,
                          u_int64_t ber_threshold_reciprocal_val,
                          double sec_between_samples,
                          list_p_fabric_general_err& ber_errors,
                          CSVOut &csv_out)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    long double ber_value_reciprocal_val = 0.0;
    char curr_counters_line[256];
    stringstream sstream;

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "BER threshold=" U64D_FMT "\n",
               ber_threshold_reciprocal_val);

    csv_out.DumpStart(SECTION_BER_TEST);

    sstream << "NodeGUID,PortGUID,PortNumber,Value" << endl;
    csv_out.WriteLine(sstream.str());

    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize();
         ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port)           //port not exists
            continue;
        if (prev_pm_info_obj_vec.size() < i + 1)    //info not exists
            continue;

        if (!prev_pm_info_obj_vec[i])               //info not exists
            continue;

        struct PM_PortCounters *p_curr_port_counters1 =
        prev_pm_info_obj_vec[i]->p_port_counters;
        if (!p_curr_port_counters1) {                 //info not exists
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto Exit;
        }
        struct PM_PortCounters *p_curr_port_counters2 =
        this->fabric_extended_info.getPMPortCounters(i);
        if (!p_curr_port_counters2) {                 //info not exists
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto Exit;
        }

        //collect all errors values
        u_int64_t diff_errs_value = 0;
        diff_errs_value = p_curr_port_counters2->SymbolErrorCounter -
            p_curr_port_counters1->SymbolErrorCounter;

        rc = this->CalcBER(p_curr_port,
                           sec_between_samples,
                           diff_errs_value,
                           ber_value_reciprocal_val);
        memset(curr_counters_line, 0, sizeof(curr_counters_line));
        sstream.str("");
        sprintf(curr_counters_line, U64H_FMT "," U64H_FMT "," U32D_FMT ","   \
                 "%Le",
                p_curr_port->p_node->guid_get(),
                p_curr_port->guid_get(),
                p_curr_port->num,
                ber_value_reciprocal_val == 0 ?
                    0.0 : 1.0/ber_value_reciprocal_val);
        sstream << curr_counters_line << endl;
        csv_out.WriteLine(sstream.str());

        //info doesn't exist
        if (rc == IBDIAG_ERR_CODE_DB_ERR) {
            rc = IBDIAG_SUCCESS_CODE;
            continue;
        }

        if (rc == IBDIAG_SUCCESS_CODE && ber_value_reciprocal_val == 0) {
            if (ber_threshold_reciprocal_val == OVERFLOW_VAL_64_BIT) {    //need to report anyway
                FabricErrBERIsZero *p_curr_fabric_pm_err =
                        new FabricErrBERIsZero(p_curr_port);
                if (!p_curr_fabric_pm_err) {
                    this->SetLastError("Failed to allocate FabricErrBERIsZero");
                    rc = IBDIAG_ERR_CODE_NO_MEM;
                    goto Exit;
                }
                ber_errors.push_back(p_curr_fabric_pm_err);
                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            }
            continue;
        }

        if ( (ber_value_reciprocal_val < ber_threshold_reciprocal_val) ||
                (ber_threshold_reciprocal_val == OVERFLOW_VAL_64_BIT) ) {
            FabricErrBERExceedThreshold *p_curr_fabric_pm_err =
                    new FabricErrBERExceedThreshold(p_curr_port,
                            ber_threshold_reciprocal_val,
                            ber_value_reciprocal_val);
            if (!p_curr_fabric_pm_err) {
                this->SetLastError("Failed to allocate FabricErrBERExceedThreshold");
                rc = IBDIAG_ERR_CODE_NO_MEM;
                goto Exit;
            }
            ber_errors.push_back(p_curr_fabric_pm_err);
            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
        }
    }

Exit:
    csv_out.DumpEnd(SECTION_BER_TEST);
    IBDIAG_RETURN(rc);
}


list_string IBDiag::GetListOFPMNames()
{
    IBDIAG_ENTER;
    list_string names_list;
    for (int i = 0; i < PM_COUNTERS_ARR_SIZE; ++i)
        names_list.push_back(pm_counters_arr[i].name);
    names_list.push_back("all");
    IBDIAG_RETURN(names_list);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR
int IBDiag::WritePMFile(const char *file_name,
                        u_int32_t check_counters_bitset,
                        bool en_per_lane_cnts)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile(file_name, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);
    this->DumpPortCounters(sout,
                           check_counters_bitset,
                           en_per_lane_cnts);
    sout.close();
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::DumpPerSLVLPortCntrsCSVTable(CSVOut &csv_out, vec_slvl_cntrs  &slvl_cntrs_vec)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    CountersPerSLVL * p_cntrs_per_slvl;

    for (vec_slvl_cntrs::iterator it = slvl_cntrs_vec.begin();
         it != slvl_cntrs_vec.end(); it++) {

        p_cntrs_per_slvl = *it;

        csv_out.DumpStart(p_cntrs_per_slvl->GetCSVSectionHeader().c_str());

        p_cntrs_per_slvl->DumpSLVLCntrsHeader(csv_out);
        p_cntrs_per_slvl->DumpSLVLCntrsData(csv_out, this->fabric_extended_info);

        csv_out.DumpEnd(p_cntrs_per_slvl->GetCSVSectionHeader().c_str());

    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


void CountersPerSLVL::DumpSLVLCntrsHeader(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    string cntrssize_str;
    if (!this->m_is_ext_cntrs)
        cntrssize_str = "32";
    else
        cntrssize_str = "64";

    stringstream sstream;
    sstream << "PortName, LID, GUID";
    for (u_int32_t cnt = 0; cnt < IB_NUM_SL; cnt++)
        sstream << "," << this->m_header << cntrssize_str << "[" <<  cnt << "]";

    sstream << endl;
    csv_out.WriteLine(sstream.str());

    IBDIAG_RETURN_VOID;
}

void CountersPerSLVL::DumpSLVLCntrsData(CSVOut &csv_out, IBDMExtendedInfo &fabric_extended_info)
{
    IBDIAG_ENTER;
    char line[1024];
    u_int8_t operationalVLs = 0;
    stringstream sstream;

    for (set_port_data_update_t::iterator it = m_set_port_data_update.begin();
            it != m_set_port_data_update.end(); it++) {

        IBPort * p_port = (*it).first;
        const struct PM_PortRcvXmitCntrsSlVl * p_per_slvl_ctrs = &((*it).second);
        memset(line, 0, sizeof(line));
        sstream.str("");
        sprintf(line, "%s,%d," U64H_FMT,
                      p_port->getName().c_str(),
                      p_port->base_lid,
                      p_port->guid_get());
        sstream << line;

        SMP_PortInfo *curr_port_info = fabric_extended_info.getSMPPortInfo(p_port->createIndex);
        if (!curr_port_info) {
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "DB error - found connected port=%s"
                       " without SMPPortInfo",
                       p_port->getName().c_str());
            IBDIAG_RETURN_VOID;
        }

        operationalVLs = get_operational_vl_num(curr_port_info->OpVLs);

        union IBDiagSLVLCntrsData ibdiag_slvl_cntrs_data;
        CLEAR_STRUCT(ibdiag_slvl_cntrs_data);

        this->Unpack(ibdiag_slvl_cntrs_data, p_per_slvl_ctrs->Data.RawData_8x132_array.data);

        if (!this->m_is_ext_cntrs)
            this->Dump(ibdiag_slvl_cntrs_data.data32, arrsize(ibdiag_slvl_cntrs_data.data32),
                       operationalVLs, sstream);
        else
            this->Dump(ibdiag_slvl_cntrs_data.data64, arrsize(ibdiag_slvl_cntrs_data.data32),
                       operationalVLs, sstream);

        csv_out.WriteLine(sstream.str());
    }

    IBDIAG_RETURN_VOID;
}

void CountersPerSLVL::Dump(u_int32_t data[], size_t arrsize, u_int8_t operationalVLs, stringstream &sstream)
{
    IBDIAG_ENTER;
    for(u_int32_t i = 0; i < arrsize; i++)
        if (this->m_is_per_vl && i > operationalVLs)
            sstream << ",NA";
        else
            sstream << "," << data[i];

    sstream << endl;
    IBDIAG_RETURN_VOID;
}

void CountersPerSLVL::Dump(struct uint64bit data[], size_t arrsize, u_int8_t operationalVLs, stringstream &sstream)
{
    IBDIAG_ENTER;
    for(u_int32_t i = 0; i < arrsize; i++)
        if (this->m_is_per_vl && i > operationalVLs)
            sstream << ",NA";
        else {
            u_int64_t cntr64;
            cntr64 = data[i].High;
            cntr64 = cntr64 << 32;
            cntr64 |= 0xffffffff & data[i].Low;
            sstream << "," << cntr64;
        }

    sstream << endl;
    IBDIAG_RETURN_VOID;
}
