/*
 *  Copyright (c) 2004-2018 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <iomanip>

#include "infiniband/ibdm/SubnMgt.h"
#include "infiniband/ibdm/CredLoops.h"
#include "infiniband/ibdiag/ibdiag.h"
#include "infiniband/ibdiag/ibdiag_clbck.h"

extern IBDiagClbck ibDiagClbck;


bool AdditionalRoutingData::dump_rn_data = false;
u_int16_t AdditionalRoutingData::sw_supports_rn_count = 0;
u_int16_t AdditionalRoutingData::max_sub_group_direction_block = 0;
u_int8_t AdditionalRoutingData::max_direction_block = 0;
u_int8_t AdditionalRoutingData::max_plft = 0;
u_int16_t AdditionalRoutingData::max_string_block = 0;
u_int8_t AdditionalRoutingData::max_ports_block = 0;
bool AdditionalRoutingData::dump_rn_counters = false;
bool AdditionalRoutingData::clear_rn_counters = false;
u_int8_t AdditionalRoutingData::max_num_ports = 0;
bool AdditionalRoutingData::dump_full_ar = false;


/****************************************************/
/******************private methods*******************/
/****************************************************/
//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::DumpUCFDBSInfo(ofstream &sout)
{
    IBDIAG_ENTER;
    char curr_uc_fdb_line[2096];
    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        // check if not a switch
        if (p_curr_node->type != IB_SW_NODE)
            continue;

        memset(curr_uc_fdb_line, 0, sizeof(curr_uc_fdb_line));
        sprintf(curr_uc_fdb_line, "osm_ucast_mgr_dump_ucast_routes: Switch " U64H_FMT,
                p_curr_node->guid_get());
        sout << curr_uc_fdb_line << endl;

        u_int8_t maxPLFT =  p_curr_node->getMaxPLFT();
        for (u_int8_t pLFT = 0; pLFT <= maxPLFT ; ++pLFT) {
            sout << "PLFT_NUM: " << (int)pLFT << endl;
            sprintf(curr_uc_fdb_line, "LID    : Port : Hops : Optimal");
            sout << curr_uc_fdb_line << endl;

            ////sould use getLFDBTop, currently not set if not collecting PLFT
            u_int16_t lfdbTop = (u_int16_t)p_curr_node->LFT[pLFT].size();

            if (p_curr_node->isPLFTEnabled())
                lfdbTop = p_curr_node->getLFDBTop(pLFT);

            for (unsigned int i = 1; i <= lfdbTop; ++i) {
                u_int8_t port_for_curr_lid = p_curr_node->getLFTPortForLid((lid_t)i, pLFT);

                memset(curr_uc_fdb_line, 0, sizeof(curr_uc_fdb_line));
                if (port_for_curr_lid == IB_LFT_UNASSIGNED)
                    sprintf(curr_uc_fdb_line, U16H_FMT " : " STR_FMT, i, "UNREACHABLE");
                else
                    sprintf(curr_uc_fdb_line, U16H_FMT " : %03u  : 00   : yes", i, port_for_curr_lid);
                sout << curr_uc_fdb_line << endl;
            }
            sout << endl;
        }
        sout << endl;
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::DumpMCFDBSInfo(ofstream &sout)
{
    IBDIAG_ENTER;
    char curr_mc_fdb_line[2096];
    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        //check if hca
        if (p_curr_node->type == IB_CA_NODE)
            continue;

        memset(curr_mc_fdb_line, 0, sizeof(curr_mc_fdb_line));
        sprintf(curr_mc_fdb_line, "Switch " U64H_FMT "\n"   \
                "LID    : Out Port(s)",
                p_curr_node->guid_get());
        sout << curr_mc_fdb_line << endl;
        for (unsigned int i = 0; i <= p_curr_node->MFT.size(); ++i) {
            list_phys_ports ports_for_curr_lid = p_curr_node->getMFTPortsForMLid((lid_t)(i + 0xc000));
            if (ports_for_curr_lid.empty() == true)
                continue;

            memset(curr_mc_fdb_line, 0, sizeof(curr_mc_fdb_line));
            sprintf(curr_mc_fdb_line, U16H_FMT " : ", i + 0xc000);
            sout << curr_mc_fdb_line;
            for (list_phys_ports::iterator it = ports_for_curr_lid.begin();
                    it != ports_for_curr_lid.end(); ++it) {
                memset(curr_mc_fdb_line, 0, sizeof(curr_mc_fdb_line));
                sprintf(curr_mc_fdb_line, "0x%03x ", *it);
                sout << curr_mc_fdb_line;
            }
            sout << endl;
        }
        sout << endl;
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


int IBDiag::ReportNonUpDownCa2CaPaths(IBFabric *p_fabric, list_pnode rootNodes, string& output)
{
    IBDIAG_ENTER;
    map_pnode_rank nodesRank;
    if (SubnRankFabricNodesByRootNodes(p_fabric, rootNodes, nodesRank)) {
        output += "-E- fail to rank the fabric by the given root nodes.\n";
        IBDIAG_RETURN(1);
    }
    IBDIAG_RETURN(SubnReportNonUpDownCa2CaPaths(p_fabric, nodesRank));
}


/****************************************************/
/*******************public methods*******************/
/****************************************************/
//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::RetrieveUCFDBSInfo(list_p_fabric_general_err& retrieve_errors, progress_func_nodes_t progress_func)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    progress_bar_nodes_t progress_bar_nodes;
    CLEAR_STRUCT(progress_bar_nodes);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = IBDiagSMPLinearForwardingTableGetClbck;

    // go over all nodes:
    struct SMP_SwitchInfo *p_curr_switch_info;
    struct SMP_LinearForwardingTable curr_linear_forwarding_table;
    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node)   //node not exist
            continue;

        p_curr_node->appData1.val = 0;

        ++progress_bar_nodes.nodes_found;
        if (p_curr_node->type == IB_SW_NODE)
            ++progress_bar_nodes.sw_found;
        else
            ++progress_bar_nodes.ca_found;
        if (progress_func)
            progress_func(&progress_bar_nodes, &this->discover_progress_bar_nodes);

        //check if hca
        if (p_curr_node->type == IB_CA_NODE ||
            p_curr_node->type == IB_RTR_NODE)
            continue;

        if (p_curr_node->isAREnable())
            continue;

        //get switch info for this node
        p_curr_switch_info = this->fabric_extended_info.getSMPSwitchInfo(p_curr_node->createIndex);
        if (!p_curr_switch_info)   //no switch info - didn't build db or no response for this node
            continue;

        //get direct route for this node
        direct_route_t *p_curr_direct_route = this->GetDirectRouteByNodeGuid(p_curr_node->guid_get());
        if (!p_curr_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s", p_curr_node->getName().c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        //check LinearFDBTop value
        if (p_curr_switch_info->LinearFDBTop > (0xc000 - 1)) {
            FabricErrNodeWrongConfig *p_curr_fabric_node_err =
                    new FabricErrNodeWrongConfig(p_curr_node, "LinearFDBTop exceeds 0xc000");
            if (!p_curr_fabric_node_err) {
                this->SetLastError("Failed to allocate FabricErrNodeWrongConfig");
                rc = IBDIAG_ERR_CODE_NO_MEM;
                goto exit;
            }
            retrieve_errors.push_back(p_curr_fabric_node_err);
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
            continue;   //try next switch
        }

        //resize LFT table
        p_curr_node->resizeLFT((u_int16_t)(p_curr_switch_info->LinearFDBTop+1));

        //get ucfdbs info
        u_int16_t num_of_lid_to_port_blocks = (u_int16_t)((p_curr_switch_info->LinearFDBTop + IBIS_IB_MAD_SMP_LFT_NUM_BLOCKS) / IBIS_IB_MAD_SMP_LFT_NUM_BLOCKS);
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Switch=%s has LinearFDBTop=%u ==> Blocks=%u\n",
                p_curr_node->getName().c_str(), p_curr_switch_info->LinearFDBTop, num_of_lid_to_port_blocks);

        for (u_int16_t curr_block = 0; curr_block < num_of_lid_to_port_blocks; ++curr_block) {
            clbck_data.m_data1 = p_curr_node;
            clbck_data.m_data2 = (void *)(uintptr_t)curr_block;
            this->ibis_obj.SMPLinearForwardingTableGetByDirect(p_curr_direct_route, curr_block,
                    &curr_linear_forwarding_table, &clbck_data);

            if (ibDiagClbck.GetState())
                goto exit;
            if (p_curr_node->appData1.val != 0)
                break; //goto next_node;
        }
    }
exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::RetrieveMCFDBSInfo(list_p_fabric_general_err& retrieve_errors, progress_func_nodes_t progress_func)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = IBDiagSMPMulticastForwardingTableGetClbck;

    int rc = IBDIAG_SUCCESS_CODE;
    progress_bar_nodes_t progress_bar_nodes;
    CLEAR_STRUCT(progress_bar_nodes);

    // go over all nodes:
    struct SMP_SwitchInfo *p_curr_switch_info;
    struct SMP_MulticastForwardingTable curr_multicast_forwarding_table;

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node)   //node not exist
            continue;

        p_curr_node->appData1.val = 0;

        ++progress_bar_nodes.nodes_found;
        if (p_curr_node->type == IB_SW_NODE)
            ++progress_bar_nodes.sw_found;
        else
            ++progress_bar_nodes.ca_found;
        if (progress_func)
            progress_func(&progress_bar_nodes, &this->discover_progress_bar_nodes);

        //check if hca or router
        if (p_curr_node->type == IB_CA_NODE || p_curr_node->type == IB_RTR_NODE)
            continue;

        //nodes with num of ports > IBDIAG_MAX_SUPPORTED_NODE_PORTS not supported in our code
        if (p_curr_node->numPorts > IBDIAG_MAX_SUPPORTED_NODE_PORTS) {
            FabricErrNodeWrongConfig *p_curr_fabric_node_err =
                    new FabricErrNodeWrongConfig(p_curr_node,
                            "number of ports exceeds maximum supported, can not support fetch of mcfdbs");
            if (!p_curr_fabric_node_err) {
                this->SetLastError("Failed to allocate FabricErrNodeWrongConfig");
                rc = IBDIAG_ERR_CODE_NO_MEM;
                goto exit;
            }
            retrieve_errors.push_back(p_curr_fabric_node_err);
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
            continue;   //try next node
        }

        //get switch info for this node
        p_curr_switch_info = this->fabric_extended_info.getSMPSwitchInfo(p_curr_node->createIndex);
        if (!p_curr_switch_info)   //no switch info - didn't build db or no response for this node
            continue;

        //get direct route for this node
        direct_route_t *p_curr_direct_route = this->GetDirectRouteByNodeGuid(p_curr_node->guid_get());
        if (!p_curr_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s", p_curr_node->getName().c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        //check MCastFDBCap value
        if (p_curr_switch_info->MCastFDBCap > (0xffff - 0xc000 + 1)) {
            FabricErrNodeWrongConfig *p_curr_fabric_node_err =
                    new FabricErrNodeWrongConfig(p_curr_node, "MCastFDBCap exceeds range");
            if (!p_curr_fabric_node_err) {
                this->SetLastError("Failed to allocate FabricErrNodeWrongConfig");
                rc = IBDIAG_ERR_CODE_NO_MEM;
                goto exit;
            }
            retrieve_errors.push_back(p_curr_fabric_node_err);
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
            continue;   //try next switch
        }

        //get mcfdbs info
        u_int16_t max_mcast_fdb = 0;

        if (p_curr_switch_info->MCastFDBTop == 0 ) {        
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Switch=%s has MCastFDBTop=%u - use CastFDBCap:%u\n",
                       p_curr_node->getName().c_str(), 
                       p_curr_switch_info->MCastFDBTop,
                       p_curr_switch_info->MCastFDBCap);
            max_mcast_fdb =  p_curr_switch_info->MCastFDBCap;
        }
        else if (p_curr_switch_info->MCastFDBTop < 0xc000 ) {
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Switch=%s has MCastFDBTop=%u - no multicast configuration\n",
                    p_curr_node->getName().c_str(), p_curr_switch_info->MCastFDBTop);
            continue;
        } else {
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Switch=%s has MCastFDBTop=%u\n",
                       p_curr_node->getName().c_str(), 
                       p_curr_switch_info->MCastFDBTop);
            max_mcast_fdb = (u_int16_t)(p_curr_switch_info->MCastFDBTop - 0xc000 + 1);
        }

        lid_t num_of_lid_to_port_blocks = (lid_t)((max_mcast_fdb +
                                                   IBIS_IB_MAD_SMP_MFT_NUM_BLOCKS - 1) / IBIS_IB_MAD_SMP_MFT_NUM_BLOCKS);
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Switch=%s has max_mcast_fdb=%u ==> Blocks=%u\n",
                p_curr_node->getName().c_str(), max_mcast_fdb, num_of_lid_to_port_blocks);

        for (u_int16_t curr_block = 0; curr_block < num_of_lid_to_port_blocks; ++curr_block) {
            for (u_int8_t curr_port_group = 0;
                    curr_port_group < ((p_curr_node->numPorts + IBIS_IB_MAD_SMP_MFT_PORT_MASK_SIZE - 1) / IBIS_IB_MAD_SMP_MFT_PORT_MASK_SIZE);
                    ++curr_port_group) {

                clbck_data.m_data1 = p_curr_node;
                clbck_data.m_data2 = (void *)(uintptr_t)curr_block;
                clbck_data.m_data3 = (void *)(uintptr_t)curr_port_group;

                this->ibis_obj.SMPMulticastForwardingTableGetByDirect(p_curr_direct_route,
                        curr_port_group, curr_block, &curr_multicast_forwarding_table, &clbck_data);

            }   //port groups loop

            if (ibDiagClbck.GetState())
                goto exit;
            if (p_curr_node->appData1.val != 0)
                break; //not try to get all the rest of the tables
        }   //blocks loop

    }   //nodes loop
exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR / ERR_CODE_DB_ERR
int IBDiag::WriteUCFDBSFile(const char *file_name)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile(file_name, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);
    rc = this->DumpUCFDBSInfo(sout);
    sout.close();
    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR / ERR_CODE_DB_ERR
int IBDiag::WriteMCFDBSFile(const char *file_name)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile(file_name, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);
    rc = this->DumpMCFDBSInfo(sout);
    sout.close();
    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::RetrieveExtendedSwitchInfo(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = IBDiagSMPExtendedSwitchInfoGetClbck;

    int rc = IBDIAG_SUCCESS_CODE;

    // go over all nodes:
    struct ib_extended_switch_info ext_sw_info;

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node)   //node not exist
            continue;

        //check if hca
        if (p_curr_node->type == IB_CA_NODE)
            continue;

        if (!this->capability_module.IsSupportedSMPCapability(
            p_curr_node, EnSMPCapIsExtendedSwitchInfoSupported))
            continue;

        //get direct route for this node
        direct_route_t *p_curr_direct_route = this->GetDirectRouteByNodeGuid(p_curr_node->guid_get());
        if (!p_curr_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s", p_curr_node->getName().c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        clbck_data.m_data1 = p_curr_node;

        this->ibis_obj.SMPExtendedSwitchInfoMadGetSetByDirect(
            p_curr_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            &ext_sw_info, &clbck_data);

        if (ibDiagClbck.GetState())
            break;
    }   //nodes loop

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

/////////
/// PLFT
/////////

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::RetrievePLFTData(list_p_fabric_general_err& retrieve_errors,
                             unsigned int &supportedDev)
{
    IBDIAG_ENTER;
    supportedDev = 0;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    //progress_bar_nodes_t progress_bar_nodes;
    //CLEAR_STRUCT(progress_bar_nodes);

    direct_route_list directRouteList;

    rc = GetSwitchesDirectRouteList(directRouteList);
    if (rc)
        IBDIAG_RETURN(rc);

    rc = RetrievePLFTInfo(retrieve_errors, directRouteList);
    if (rc)
        IBDIAG_RETURN(rc);

    if (directRouteList.empty())
        IBDIAG_RETURN(rc);

    supportedDev = (unsigned int)directRouteList.size();
    GetDiscoverFabricPtr()->setPLFTEnabled();

    rc = RetrievePLFTMapping(retrieve_errors, directRouteList);
    if (rc)
        IBDIAG_RETURN(rc);

    rc = RetrievePLFTTop(retrieve_errors, directRouteList);

    IBDIAG_RETURN(rc);

}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::RetrievePLFTInfo(list_p_fabric_general_err& retrieve_errors, direct_route_list &directRouteList)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = IBDiagSMPPLFTInfoGetClbck;

    // go over all nodes:
    struct ib_private_lft_info plft_info;
    IBNode *p_curr_node = NULL;
    direct_route_t *p_curr_direct_route = NULL;
    direct_route_list_iter iter = directRouteList.begin();

    for ( ; iter != directRouteList.end() ; ++iter) {

        p_curr_node = iter->first;
        p_curr_direct_route = iter->second;


        if (!this->capability_module.IsSupportedSMPCapability(
            p_curr_node, EnSMPCapIsPrivateLinearForwardingSupported) ||
            !this->capability_module.IsSupportedSMPCapability(
            p_curr_node, EnSMPCapIsPortSLToPrivateLFTMapSupported))
            continue;

        clbck_data.m_data1 = p_curr_node;
        this->ibis_obj.SMPPLFTInfoMadGetSetByDirect(p_curr_direct_route,
                IBIS_IB_MAD_METHOD_GET,
                &plft_info, &clbck_data);

        if (ibDiagClbck.GetState())
            break;
    }

    this->ibis_obj.MadRecAll();

    rc = ibDiagClbck.GetState();
    if (rc)
        this->SetLastError(ibDiagClbck.GetLastError());
    else if (!retrieve_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    if (rc == IBDIAG_SUCCESS_CODE) {
        struct SMP_SwitchInfo *p_curr_switch_info = NULL;
        //remove nodes that do not support PLFT
        iter = directRouteList.begin();
        direct_route_list_iter prevIter = directRouteList.begin();
        for (;iter != directRouteList.end(); ) {
            prevIter = iter;
            ++iter;
            p_curr_node = prevIter->first;
            if (!p_curr_node->isPLFTEnabled()) {
                //get switch info for this node
                p_curr_switch_info =
                    this->fabric_extended_info.getSMPSwitchInfo(
                        p_curr_node->createIndex);

                if (!p_curr_switch_info)   //no switch info - didn't build db or no response for this node
                    p_curr_node->setLFDBTop(0, 0);
                else
                   p_curr_node->setLFDBTop(0, p_curr_switch_info->LinearFDBTop);

                directRouteList.erase(prevIter);
            }
        }
    }

    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::RetrievePLFTMapping(list_p_fabric_general_err& retrieve_errors, direct_route_list &directRouteList)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = IBDiagSMPPortSLToPrivateLFTMapGetClbck;

    // go over all nodes:
    struct ib_port_sl_to_private_lft_map plft_map;
    IBNode *p_curr_node = NULL;
    direct_route_t *p_curr_direct_route = NULL;
    direct_route_list_iter iter = directRouteList.begin();
    for ( ; iter != directRouteList.end() ; ++iter) {

        p_curr_node = iter->first;
        p_curr_direct_route = iter->second;

        p_curr_node->appData1.val = 0;

        //get ucfdbs info
        u_int8_t num_of_port_blocks = (u_int8_t)((p_curr_node->numPorts +
                                                  IBIS_IB_MAD_PLFT_MAP_PORT_BLOCK_SIZE) /
                                                 IBIS_IB_MAD_PLFT_MAP_PORT_BLOCK_SIZE);
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Switch=%s has %u ports ==> Blocks=%u\n",
                p_curr_node->getName().c_str(), p_curr_node->numPorts, num_of_port_blocks);

        for (u_int8_t curr_block = 0; curr_block < num_of_port_blocks; ++curr_block) {
            clbck_data.m_data1 = p_curr_node;
            clbck_data.m_data2 = (void *)(uintptr_t)curr_block;
            this->ibis_obj.SMPPortSLToPrivateLFTMapGetSetByDirect(
                p_curr_direct_route,
                IBIS_IB_MAD_METHOD_GET,
                curr_block,
                &plft_map, &clbck_data);

            if (ibDiagClbck.GetState())
                goto exit;
            if (p_curr_node->appData1.val != 0)
                break; //goto next_node;
        }
    }
exit:
    this->ibis_obj.MadRecAll();

    rc = ibDiagClbck.GetState();
    if (rc)
        this->SetLastError(ibDiagClbck.GetLastError());
    else if (!retrieve_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::RetrievePLFTTop(list_p_fabric_general_err& retrieve_errors, direct_route_list &directRouteList)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = IBDiagSMPPrivateLFTTopGetClbck;

    // go over all nodes:
    struct ib_private_lft_map plft_map;
    CLEAR_STRUCT(plft_map);
    plft_map.LFT_TopEn = 1;

    IBNode *p_curr_node = NULL;
    direct_route_t *p_curr_direct_route = NULL;
    direct_route_list_iter iter = directRouteList.begin();
    for ( ; iter != directRouteList.end() ; ++iter) {

        p_curr_node = iter->first;
        p_curr_direct_route = iter->second;

        p_curr_node->appData1.val = 0;

        for (u_int8_t pLFT = 0; pLFT <= p_curr_node->getMaxPLFT(); ++pLFT) {
            clbck_data.m_data1 = p_curr_node;
            clbck_data.m_data2 = (void *)(uintptr_t)pLFT;

            this->ibis_obj.SMPPLFTMapMadGetSetByDirect(
                p_curr_direct_route,
                IBIS_IB_MAD_METHOD_GET,
                pLFT,
                &plft_map, &clbck_data);

            if (ibDiagClbck.GetState())
                goto exit;
            if (p_curr_node->appData1.val != 0)
                break; //goto next_node;
        }
    }
exit:
    this->ibis_obj.MadRecAll();

    rc = ibDiagClbck.GetState();
    if (rc)
        this->SetLastError(ibDiagClbck.GetLastError());
    else if (!retrieve_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAG_RETURN(rc);
}

//////
// AR
//////

int IBDiag::GetSwitchesDirectRouteList(direct_route_list &directRouteList)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node)   //node not exist
            continue;

        //check if hca
        if (p_curr_node->type == IB_CA_NODE)
            continue;

        //get direct route for this node
        direct_route_t *p_curr_direct_route = this->GetDirectRouteByNodeGuid(p_curr_node->guid_get());
        if (!p_curr_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s", p_curr_node->getName().c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        directRouteList.push_back(make_pair(p_curr_node, p_curr_direct_route));
        p_curr_node->appData1.val = 0;
        p_curr_node->appData2.val = 0;

    }

    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::RetrieveARData(list_p_fabric_general_err& retrieve_errors, unsigned int &supportedDev,
                           AdditionalRoutingDataMap *p_routing_data_map)
{
    IBDIAG_ENTER;
    supportedDev = 0;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    //progress_bar_nodes_t progress_bar_nodes;
    //CLEAR_STRUCT(progress_bar_nodes);

    direct_route_list directRouteList;

    rc = GetSwitchesDirectRouteList(directRouteList);
    if (rc)
        IBDIAG_RETURN(rc);

    rc = RetrieveARInfo(retrieve_errors, directRouteList, p_routing_data_map);
    if (rc)
        IBDIAG_RETURN(rc);

    if (directRouteList.empty())
        IBDIAG_RETURN(rc);

    supportedDev = (unsigned int)directRouteList.size();

    rc = RetrieveARGroupTable(retrieve_errors, directRouteList);
    if (rc)
        IBDIAG_RETURN(rc);

    rc = RetrieveARLinearForwardingTable(retrieve_errors, directRouteList);

    //Set FABRIC_ERROR only after all available data is collected
    if (!rc && !retrieve_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAG_RETURN(rc);

}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::RetrieveARInfo(list_p_fabric_general_err& retrieve_errors,
                           direct_route_list &directRouteList,
                           AdditionalRoutingDataMap *p_routing_data_map)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    //progress_bar_nodes_t progress_bar_nodes;
    //CLEAR_STRUCT(progress_bar_nodes);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = IBDiagSMPARInfoGetClbck;

    // go over all nodes:
    struct adaptive_routing_info ar_info;

    IBNode *p_curr_node = NULL;
    direct_route_t *p_curr_direct_route = NULL;
    direct_route_list_iter iter = directRouteList.begin();
    for ( ; iter != directRouteList.end() ; ++iter) {

        p_curr_node = iter->first;
        p_curr_direct_route = iter->second;

        if (!this->capability_module.IsSupportedSMPCapability(
            p_curr_node, EnSMPCapIsAdaptiveRoutingRev1Supported))
            continue;

        clbck_data.m_data1 = p_curr_node;
        clbck_data.m_data2 = p_routing_data_map;
        clbck_data.m_data3 = p_curr_direct_route;

        this->ibis_obj.SMPARInfoGetSetByDirect(
            p_curr_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            false, //get_cap
            &ar_info, &clbck_data);

        if (ibDiagClbck.GetState())
            break;
    }

    this->ibis_obj.MadRecAll();

    rc = ibDiagClbck.GetState();
    if (rc)
        this->SetLastError(ibDiagClbck.GetLastError());
    //else if (!retrieve_errors.empty())
    //    rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    if (rc == IBDIAG_SUCCESS_CODE) {
        //remove nodes that do not support AR
        iter = directRouteList.begin();
        direct_route_list_iter prevIter = directRouteList.begin();
        for (;iter != directRouteList.end(); ) {
            prevIter = iter;
            ++iter;
            if (!prevIter->first->isAREnable())
                directRouteList.erase(prevIter);
        }
    }

    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::RetrieveARGroupTable(list_p_fabric_general_err& retrieve_errors,
                                 direct_route_list & directRouteList)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    //progress_bar_nodes_t progress_bar_nodes;
    //CLEAR_STRUCT(progress_bar_nodes);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = IBDiagSMPARGroupTableGetClbck;

    direct_route_list routeList = directRouteList;

    // go over all nodes:
    struct ib_ar_group_table ar_group_table;

    IBNode *p_curr_node = NULL;
    direct_route_t *p_curr_direct_route = NULL;
    bool isSend = false;

    while (!routeList.empty()) {
        direct_route_list_iter iter = routeList.begin();
        direct_route_list_iter prevIter = routeList.begin();
        isSend = false;
        for ( ; iter != routeList.end() ; ) {

            p_curr_node = iter->first;
            p_curr_direct_route = iter->second;
            prevIter = iter;
            ++iter;

            if (p_curr_node->appData2.val == RETRIEVE_STAGE_REC_WAIT)
                continue;
            else if (p_curr_node->appData2.val == RETRIEVE_STAGE_REC_DONE) {
                routeList.erase(prevIter);
                continue;
            }

            clbck_data.m_data1 = p_curr_node;
            p_curr_node->appData2.val = RETRIEVE_STAGE_REC_WAIT;

            this->ibis_obj.SMPARGroupTableGetSetByDirect(
                p_curr_direct_route,
                IBIS_IB_MAD_METHOD_GET,
                (u_int16_t)p_curr_node->appData1.val, //group_block
                0, //group_table
                0, //pLFT - global group
                &ar_group_table, &clbck_data);

            isSend = true;
        }

        if(!isSend && !routeList.empty())
            this->ibis_obj.MadRecAll();

        if (ibDiagClbck.GetState())
            break;
    }

    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        //else if (!retrieve_errors.empty())
        //    rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::RetrieveARLinearForwardingTable(
    list_p_fabric_general_err& retrieve_errors,
    direct_route_list & directRouteList)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    //progress_bar_nodes_t progress_bar_nodes;
    //CLEAR_STRUCT(progress_bar_nodes);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = IBDiagSMPARLinearForwardingTableGetClbck;

    // go over all nodes:
    struct ib_ar_linear_forwarding_table_sx ar_linear_forwarding_table;

    IBNode *p_curr_node = NULL;
    direct_route_t *p_curr_direct_route = NULL;
    for (u_int8_t pLFT = 0 ; pLFT < MAX_PLFT_NUM; ++pLFT ) {
        direct_route_list_iter iter = directRouteList.begin();
        for ( ; iter != directRouteList.end() ; ++iter) {

            p_curr_node = iter->first;
            p_curr_direct_route = iter->second;

            if (p_curr_node->getMaxPLFT() < pLFT)
                continue;

            if (pLFT == 0)
                p_curr_node->appData1.val = 0;

            u_int16_t lfDBTop = p_curr_node->getLFDBTop(pLFT);

            p_curr_node->resizeLFT((u_int16_t)(lfDBTop+1), pLFT);
            p_curr_node->resizeARLFT((u_int16_t)(lfDBTop+1), pLFT);

            u_int16_t num_of_lid_to_port_blocks =
                (u_int16_t)((lfDBTop + IBIS_IB_MAD_SMP_AR_LFT_NUM_BLOCKS_SX) /
                                  IBIS_IB_MAD_SMP_AR_LFT_NUM_BLOCKS_SX);
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                       "Switch=%s pLFT:%d has LinearFDBTop=%u ==> Blocks=%u\n",
                       p_curr_node->getName().c_str(), (int)pLFT, lfDBTop,
                       num_of_lid_to_port_blocks);

            for (u_int16_t curr_block = 0; curr_block < num_of_lid_to_port_blocks; ++curr_block) {
                clbck_data.m_data1 = p_curr_node;
                clbck_data.m_data2 = (void *)(uintptr_t)curr_block;
                clbck_data.m_data3 = (void *)(uintptr_t)pLFT;

                this->ibis_obj.SMPARLinearForwardingTableGetSetByDirect(
                    p_curr_direct_route,
                    IBIS_IB_MAD_METHOD_GET,
                    curr_block,
                    pLFT,
                    &ar_linear_forwarding_table, &clbck_data);

                if (ibDiagClbck.GetState())
                    break;
                if (p_curr_node->appData1.val != 0)
                    break; //goto next_node;
            }

            if (ibDiagClbck.GetState())
                break;
        }

        if (ibDiagClbck.GetState())
            break;
    }

    this->ibis_obj.MadRecAll();

    rc = ibDiagClbck.GetState();
    if (rc)
        this->SetLastError(ibDiagClbck.GetLastError());
    //else if (!retrieve_errors.empty())
    //    rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAG_RETURN(rc);
}

//////
// RN
//////

int IBDiag::AddRNDataMapEntry(AdditionalRoutingDataMap *p_routing_data_map,
                              IBNode *p_node,
                              direct_route_t *p_direct_route,
                              adaptive_routing_info *p_ar_info)
{
    if (AdditionalRoutingData::dump_full_ar ||
        ((AdditionalRoutingData::dump_rn_data ||
          AdditionalRoutingData::dump_rn_counters ||
          AdditionalRoutingData::clear_rn_counters) &&
         (p_ar_info->is_arn_sup || p_ar_info->is_frn_sup))) {

        pair <IBNode *, AdditionalRoutingData> entry;
        entry.first = p_node;

        entry.second.p_node = p_node;
        entry.second.p_direct_route = p_direct_route;

        memcpy(&(entry.second.ar_info), p_ar_info, sizeof(*p_ar_info));

        AdditionalRoutingDataMapInsertRes res = p_routing_data_map->insert(entry);
        if (!res.second)
            //If an equivalent key already existed. Should not happen
            IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

        AdditionalRoutingData & routing_data = res.first->second;

        //appData3.ptr is used by ARLinearForwardingTableGetClbck,
        //to update AdditionalRoutingData.ar_lft_table_vec
        //and by IBDiagSMPARGroupTableGetClbck
        //to update AdditionalRoutingData.group_table_vec
        //if dump_full_ar is true
        p_node->appData3.ptr = &routing_data;


        if (p_ar_info->is_arn_sup || p_ar_info->is_frn_sup)
            routing_data.sw_supports_rn_count++;

        if (AdditionalRoutingData::dump_rn_data &&
            (p_ar_info->is_arn_sup ||
             p_ar_info->is_frn_sup)) {

            //RNSubGroupDirectionTable
            routing_data.top_sub_group_direction_block =
                (u_int16_t) (((p_ar_info->group_top + 1) *
                 (p_ar_info->sub_grps_active + 1) - 1) /
                 IBIS_IB_MAD_SMP_RN_SUB_GROUP_DIRECTION_BLOCK_SIZE);

            routing_data.sub_group_direction_table_vec.resize(
                routing_data.top_sub_group_direction_block + 1);

            routing_data.max_sub_group_direction_block = max(
                routing_data.max_sub_group_direction_block,
                routing_data.top_sub_group_direction_block);

            //RNGenStringTable
            routing_data.top_direction_block =
                (u_int8_t)(p_ar_info->direction_num_sup /
                 IBIS_IB_MAD_SMP_RN_GEN_STRING_TABLE_BLOCK_SIZE);

            u_int8_t maxPLFT = p_node->getMaxPLFT();

            routing_data.gen_string_table_vec.resize(maxPLFT + 1);

            for (u_int8_t pLFT = 0 ; pLFT <= maxPLFT ; ++pLFT)
                routing_data.gen_string_table_vec[pLFT].resize(
                routing_data.top_direction_block + 1);

            routing_data.max_direction_block =
                max(routing_data.max_direction_block, routing_data.top_direction_block);

            routing_data.max_plft = max(routing_data.max_plft, maxPLFT);

            //RNRcvString
            routing_data.top_string_block =
                (u_int16_t)((0x1 << (p_ar_info->string_width_cap)) - 1)/
                IBIS_IB_MAD_SMP_RN_RCV_STRING_BLOCK_SIZE;

            routing_data.rcv_string_vec.resize(
                routing_data.top_string_block + 1);

            routing_data.max_string_block =
                max(routing_data.max_string_block, routing_data.top_string_block);

            //RNXmitPortMask
            routing_data.top_ports_block =
                (u_int8_t)(p_node->numPorts /
                 IBIS_IB_MAD_SMP_RN_XMIT_PORT_MASK_BLOCK_SIZE);

            routing_data.xmit_port_mask_vec.resize(
                routing_data.top_ports_block + 1);

            routing_data.max_ports_block =
                max(routing_data.max_ports_block, routing_data.top_ports_block);
        }

        if ((AdditionalRoutingData::dump_rn_counters ||
             AdditionalRoutingData::clear_rn_counters) &&
            (p_ar_info->is_arn_sup || p_ar_info->is_frn_sup)) {

            routing_data.max_num_ports = max(routing_data.max_num_ports,
                                             (u_int8_t)p_node->numPorts);

            //VS_PortRNCounters
            if (routing_data.dump_rn_counters)
                routing_data.port_rn_counters_vec.resize(p_node->numPorts + 1);
        }

        if (AdditionalRoutingData::dump_full_ar) {

            routing_data.top_group_table_block = 0;
            routing_data.top_ar_lft_table_block = 0;

            if (p_ar_info->group_top){
                u_int16_t top_block =
                    (u_int16_t)(((p_ar_info->group_top + 1) *
                                 (p_ar_info->sub_grps_active + 1)) /
                                IBIS_IB_MAD_SMP_AR_GROUP_TABLE_BLOCK_SIZE);

                routing_data.group_table_vec.resize(top_block + 1);
            }
        }
    }


    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::RetrieveRNData(list_p_fabric_general_err& retrieve_errors,
                           AdditionalRoutingDataMap *p_routing_data_map)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    int rc = IBDIAG_SUCCESS_CODE;
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;

    AdditionalRoutingDataMapIter routing_data_iter = p_routing_data_map->begin();
    AdditionalRoutingData *p_routing_data = NULL;

    //RNSubGroupDirectionTable
    struct rn_sub_group_direction_tbl sub_group_direction_table;
    CLEAR_STRUCT(sub_group_direction_table);

    clbck_data.m_handle_data_func = IBDiagSMPRNSubGroupDirectionTableGetClbck;

    for (u_int16_t block_num = 0 ;
         block_num <= AdditionalRoutingData::max_sub_group_direction_block;
         ++block_num) {

        clbck_data.m_data2 = (void *)(uintptr_t)block_num;

        for (routing_data_iter = p_routing_data_map->begin();
             routing_data_iter != p_routing_data_map->end();
             ++routing_data_iter){

            p_routing_data = &(routing_data_iter->second);

            if (!p_routing_data->isRNSupported())
                continue; //continue to next sw

            if (p_routing_data->top_sub_group_direction_block < block_num)
                continue; //continue to next sw

            clbck_data.m_data1 = p_routing_data;

            this->ibis_obj.SMPRNSubGroupDirectionTableGetSetByDirect(
                p_routing_data->p_direct_route,
                IBIS_IB_MAD_METHOD_GET,
                block_num,
                &sub_group_direction_table, &clbck_data);

            if (ibDiagClbck.GetState())
                break;
        }

        if (ibDiagClbck.GetState())
            break;
    }

    //RNGenStringTable
    struct rn_gen_string_tbl gen_string_table;
    CLEAR_STRUCT(gen_string_table);

    clbck_data.m_handle_data_func = IBDiagSMPRNGenStringTableGetClbck;

    for (u_int8_t pLFT = 0 ; pLFT <= AdditionalRoutingData::max_plft;
         ++pLFT ) {

        clbck_data.m_data3 = (void *)(uintptr_t)pLFT;

        for (u_int8_t block_num = 0 ;
             block_num <= AdditionalRoutingData::max_direction_block;
             ++block_num) {

            clbck_data.m_data2 = (void *)(uintptr_t)block_num;

            for (routing_data_iter = p_routing_data_map->begin();
                 routing_data_iter != p_routing_data_map->end();
                 ++routing_data_iter){

                p_routing_data = &(routing_data_iter->second);

                if (!p_routing_data->isRNSupported())
                    continue; //continue to next sw

                if (p_routing_data->top_direction_block < block_num ||
                    p_routing_data->p_node->getMaxPLFT() < pLFT)
                    continue; //continue to next sw

                clbck_data.m_data1 = p_routing_data;

                this->ibis_obj.SMPRNGenStringTableGetSetByDirect(
                    p_routing_data->p_direct_route,
                    IBIS_IB_MAD_METHOD_GET,
                    block_num,
                    pLFT,
                    &gen_string_table, &clbck_data);

                if (ibDiagClbck.GetState())
                    break;
            }//routing_data_iter
        } //block_num
    }//pLFT

    //RNGenBySubGroupPriority
    struct rn_gen_by_sub_group_prio gen_by_sg_priority;
    CLEAR_STRUCT(gen_by_sg_priority);

    clbck_data.m_handle_data_func = IBDiagSMPRNGenBySubGroupPriorityGetClbck;


    for (routing_data_iter = p_routing_data_map->begin();
         routing_data_iter != p_routing_data_map->end();
         ++routing_data_iter){

        p_routing_data = &(routing_data_iter->second);

        if (!p_routing_data->isRNSupported())
            continue; //continue to next sw

        clbck_data.m_data1 = p_routing_data;

        this->ibis_obj.SMPRNGenBySubGroupPriorityMadGetSetByDirect(
            p_routing_data->p_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            &gen_by_sg_priority, &clbck_data);

        if (ibDiagClbck.GetState())
            break;
    }


    //RNRcvString
    struct rn_rcv_string rcv_string;
    CLEAR_STRUCT(rcv_string);

    clbck_data.m_handle_data_func = IBDiagSMPRNRcvStringGetClbck;

    for (u_int16_t block_num = 0 ;
         block_num <= AdditionalRoutingData::max_string_block;
         ++block_num) {

        clbck_data.m_data2 = (void *)(uintptr_t)block_num;

        for (routing_data_iter = p_routing_data_map->begin() ;
             routing_data_iter != p_routing_data_map->end() ;
             ++routing_data_iter){

                p_routing_data = &(routing_data_iter->second);

                if (!p_routing_data->isRNSupported())
                    continue; //continue to next sw

                if (p_routing_data->top_string_block < block_num)
                    continue; //continue to next sw

                clbck_data.m_data1 = p_routing_data;

                this->ibis_obj.SMPRNRcvStringGetSetByDirect(
                    p_routing_data->p_direct_route,
                    IBIS_IB_MAD_METHOD_GET,
                    block_num,//string_block,
                    &rcv_string, &clbck_data);

                if (ibDiagClbck.GetState())
                    break;
        }
    }

    //RNXmitPortMask
    struct rn_xmit_port_mask xmit_port_mask;
    CLEAR_STRUCT(xmit_port_mask);

    clbck_data.m_handle_data_func = IBDiagSMPRNXmitPortMaskGetClbck;

    for (u_int8_t block_num = 0 ;
         block_num <= AdditionalRoutingData::max_ports_block;
         ++block_num) {

        clbck_data.m_data2 = (void *)(uintptr_t)block_num;

        for (routing_data_iter = p_routing_data_map->begin();
             routing_data_iter != p_routing_data_map->end();
             ++routing_data_iter){

                p_routing_data = &(routing_data_iter->second);

                if (!p_routing_data->isRNSupported())
                    continue; //continue to next sw

                if (p_routing_data->top_ports_block < block_num)
                    continue; //continue to next sw

                clbck_data.m_data1 = p_routing_data;

                this->ibis_obj.SMPRNXmitPortMaskGetSetByDirect(
                    p_routing_data->p_direct_route,
                    IBIS_IB_MAD_METHOD_GET,
                    block_num, //ports_block,
                    &xmit_port_mask, &clbck_data);

                if (ibDiagClbck.GetState())
                    break;
        }
    }

    this->ibis_obj.MadRecAll();

    rc = ibDiagClbck.GetState();
    if (rc)
        this->SetLastError(ibDiagClbck.GetLastError());
    else if (!retrieve_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAG_RETURN(rc);
}

int IBDiag::CheckSL2VLTables(string& output)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    output = "";
    ibdmClearInternalLog();

    //CA <--> CA connectivity
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "CheckSL2VLTables Start \n");
    SubnMgtCheckSL2VLTables(&(this->discovered_fabric));
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "CheckSL2VLTables End \n");

    char *buffer = ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    output += buffer;
    free(buffer);

    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::ClearAndGetRNCounters(list_p_fabric_general_err& retrieve_errors,
                                  AdditionalRoutingDataMap *p_routing_data_map)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    int rc = IBDIAG_SUCCESS_CODE;
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;

    AdditionalRoutingDataMapIter routing_data_iter = p_routing_data_map->begin();
    AdditionalRoutingData *p_routing_data = NULL;

    //VS_PortRNCounters
    if (AdditionalRoutingData::clear_rn_counters){

        clbck_data.m_handle_data_func = IBDiagVSPortRNCountersClearClbck;

        for (u_int8_t port_num = 1 ;
             port_num <= AdditionalRoutingData::max_num_ports;
             ++port_num) {

            clbck_data.m_data2 = (void *)(uintptr_t)port_num;

            for (routing_data_iter = p_routing_data_map->begin();
                 routing_data_iter != p_routing_data_map->end();
                 ++routing_data_iter){

                p_routing_data = &(routing_data_iter->second);

                if (!p_routing_data->isRNSupported())
                    continue; //continue to next sw

                if (p_routing_data->p_node->numPorts < port_num)
                    continue; //continue to next sw

                clbck_data.m_data1 = p_routing_data;

                this->ibis_obj.VSPortRNCountersClear(
                    p_routing_data->p_node->getPort(0)->base_lid,
                    port_num,
                    &clbck_data);

                if (ibDiagClbck.GetState())
                    break;
            }

            if (ibDiagClbck.GetState())
                break;
        }

        this->ibis_obj.MadRecAll();
    }

    if (AdditionalRoutingData::dump_rn_counters){

        clbck_data.m_handle_data_func = IBDiagVSPortRNCountersGetClbck;
        struct port_rn_counters rn_counters;

        for (u_int8_t port_num = 1 ;
             port_num <= AdditionalRoutingData::max_num_ports;
             ++port_num) {

            clbck_data.m_data2 = (void *)(uintptr_t)port_num;

            for (routing_data_iter = p_routing_data_map->begin();
                 routing_data_iter != p_routing_data_map->end();
                 ++routing_data_iter){

                p_routing_data = &(routing_data_iter->second);

                if (!p_routing_data->isRNSupported())
                    continue; //continue to next sw

                if (p_routing_data->p_node->numPorts < port_num)
                    continue; //continue to next sw

                clbck_data.m_data1 = p_routing_data;

                this->ibis_obj.VSPortRNCountersGet(
                    p_routing_data->p_node->getPort(0)->base_lid,
                    port_num,
                    &rn_counters,
                    &clbck_data);

                if (ibDiagClbck.GetState())
                    break;
            }

            if (ibDiagClbck.GetState())
                break;
        }

        this->ibis_obj.MadRecAll();
    }

    rc = ibDiagClbck.GetState();
    if (rc)
        this->SetLastError(ibDiagClbck.GetLastError());
    else if (!retrieve_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_IBDM_ERR / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::ReportFabricQualities(string& output, const char *outDir)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    output = "";
    ibdmClearInternalLog();

    //CA <--> CA histogram
    SubnMgtVerifyAllCaToCaRoutes(&(this->discovered_fabric), outDir);
    //LID --> Port histogram
    SubnMgtCheckFabricMCGrps(&(this->discovered_fabric));
    char *buffer = ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    output += buffer;
    free(buffer);
    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_IBDM_ERR / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::ReportCreditLoops(string& output, bool is_fat_tree, bool checkAR)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    output = "";
    ibdmClearInternalLog();

    list_pnode root_nodes;
    char *buffer;
    if (is_fat_tree) {
        // Calculate the minhop table for the switches
        SubnMgtCalcMinHopTables(&(this->discovered_fabric));

        // Analyze the fabric to find its root nodes using statistical methods
        // on the profiles of min hops to CAs
        root_nodes = SubnMgtFindRootNodesByMinHop(&(this->discovered_fabric));

        buffer = ibdmGetAndClearInternalLog();
        if (!buffer) {
            this->SetLastError("Failed to allocate buffer for ibdm output");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
        }
        output += buffer;
        free(buffer);
    }

    if (root_nodes.empty() == false) {      //fat tree
        char buffer2[64];
        memset(buffer2, 0, sizeof(buffer2));
        sprintf(buffer2, "\n-I- Found %u Roots:\n", (unsigned int)root_nodes.size());
        output += buffer2;
        for (list_p_node::iterator it = root_nodes.begin(); it != root_nodes.end(); ++it) {
            output += "    ";
            output += (*it)->getName();
            output += "\n";
        }
        this->ReportNonUpDownCa2CaPaths(&(this->discovered_fabric), root_nodes, output);
    } else {                                //generic - not fat tree
        CrdLoopAnalyze(&(this->discovered_fabric), checkAR);
    }

    buffer = ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    output += buffer;
    free(buffer);
    IBDIAG_RETURN(rc);
}

/****************** AR Routing Tests **********************************/

//Returns: SUCCESS_CODE / ERR_CODE_IBDM_ERR / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiag::ReportFabricARConnectivity(string& output)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    output = "";
    ibdmClearInternalLog();

    //CA <--> CA connectivity
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "SubnMgtVerifyAllARCaToCaRoutes Start \n");
    SubnMgtVerifyAllARCaToCaRoutes(&(this->discovered_fabric));
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "SubnMgtVerifyAllARCaToCaRoutes End \n");

    char *buffer = ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    output += buffer;
    free(buffer);
    IBDIAG_RETURN(rc);
}

/****************************************************/

int IBDiag::ParseSLVLFile(const string & file_name, string& output)
{
    IBDIAG_ENTER;

    ibdmClearInternalLog();

    int rc = discovered_fabric.parseSLVLFile(file_name);

    char *buffer= ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    output += buffer;
    free(buffer);

    if (rc)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR
int IBDiag::WriteSLVLFile(const char *file_name,
                          list_p_fabric_general_err& retrieve_errors,
                          progress_func_nodes_t progress_func)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile(file_name, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);
    this->DumpSLVLFile(sout, retrieve_errors, progress_func);
    sout.close();
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

#define IS_SUPPORT_SL_MAPPING(cap_mask) ((cap_mask) & 0x40)

bool IBDiag::HandleUnsupportedSLMapping(ofstream &sout,
                                        IBNode *p_node,
                                        phys_port_t port)
{
    IBPort *p_port = NULL;

    if (p_node->type == IB_SW_NODE)
        p_port = p_node->getPort(0);
    else
        p_port = p_node->getPort(port);

    if (!p_port)
        return false;

    struct SMP_PortInfo *p_port_info =
        this->fabric_extended_info.getSMPPortInfo(p_port->createIndex);

    if (!p_port_info)
        return false;

    if (IS_SUPPORT_SL_MAPPING(p_port_info->CapMsk))
        return false;

    char sLine[1024];
    if (p_node->type == IB_SW_NODE) {
        for (unsigned int curr_out_port = 1 ;
             curr_out_port <= p_node->numPorts ; ++curr_out_port) {

            for (unsigned int curr_in_port = 0 ;
                 curr_in_port <= p_node->numPorts ; ++curr_in_port) {

                if(curr_out_port == curr_in_port)
                    continue;

                sprintf(sLine, U64H_FMT " %u %u"
                    " 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00\n",
                    p_node->guid_get(), curr_in_port, curr_out_port);

                sout << sLine;

            }

        }
    } else {
        sprintf(sLine, U64H_FMT " 0 %u"
            " 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00\n",
            p_node->guid_get(),
            port);

        sout << sLine;
    }

    return true;

}

int IBDiag::ReadCASLVL(ofstream &sout,
                       clbck_data_t &clbck_data,
                       SMP_SLToVLMappingTable &slvl_mapping,
                       IBNode *p_node)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    IBPort *p_port=NULL;

    for (phys_port_t port = 1 ; port <= p_node->numPorts ; port++) {
        p_port = p_node->getPort(port);
        if (p_port == NULL || p_port->p_remotePort == NULL)
            continue;

        if (HandleUnsupportedSLMapping(sout, p_node, port))
            continue;

        direct_route_t *p_curr_direct_route = this->GetDirectRouteByPortGuid(p_port->guid_get());
        if (!p_curr_direct_route) {
            this->SetLastError("DB error - can't find direct route to port=%s", p_port->getName().c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        clbck_data.m_data1 = p_node;
        clbck_data.m_data2 = (void *)(uintptr_t)0;
        clbck_data.m_data3 = (void *)(uintptr_t)port;

        this->ibis_obj.SMPSLToVLMappingTableGetByDirect(p_curr_direct_route,
                (phys_port_t)0, (phys_port_t)0, &slvl_mapping, &clbck_data);
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::DumpSLVLFile(ofstream &sout,
                         list_p_fabric_general_err& retrieve_errors,
                         progress_func_nodes_t progress_func)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    progress_bar_nodes_t progress_bar_nodes;
    CLEAR_STRUCT(progress_bar_nodes);
    struct SMP_SLToVLMappingTable slvl_mapping;

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors, &sout);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = IBDiagSMPSLToVLMappingTableGetClbck;

    u_int32_t nodesNumber = this->fabric_extended_info.getNodesVectorSize();
    for (u_int32_t i = 0; i < nodesNumber; ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node)   //node not exist
            continue;

        if (progress_func) {
            ++progress_bar_nodes.nodes_found;
            if (p_curr_node->type == IB_SW_NODE)
                ++progress_bar_nodes.sw_found;
            else
                ++progress_bar_nodes.ca_found;

            progress_func(&progress_bar_nodes, &this->discover_progress_bar_nodes);
        }



        if (p_curr_node->type != IB_SW_NODE) {
            rc = ReadCASLVL(sout, clbck_data, slvl_mapping, p_curr_node);
            if (rc || ibDiagClbck.GetState())
                goto exit;
            continue;
        }

        if (HandleUnsupportedSLMapping(sout, p_curr_node, 0))
            continue;

        //get direct route for this node
        direct_route_t *p_curr_direct_route = this->GetDirectRouteByNodeGuid(p_curr_node->guid_get());
        if (!p_curr_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s", p_curr_node->getName().c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        //char sLine[1024];
        // use unsigned int index to prevent overwrap
        for (unsigned int curr_out_port = 1 ;
             curr_out_port <= p_curr_node->numPorts ; ++curr_out_port) {

            for (unsigned int curr_in_port = 0 ;
                 curr_in_port <= p_curr_node->numPorts ; ++curr_in_port) {

                if(curr_out_port == curr_in_port)
                    continue;

                clbck_data.m_data1 = p_curr_node;
                clbck_data.m_data2 = (void *)(uintptr_t)curr_in_port;
                clbck_data.m_data3 = (void *)(uintptr_t)curr_out_port;

                this->ibis_obj.SMPSLToVLMappingTableGetByDirect(p_curr_direct_route,
                        (phys_port_t)curr_out_port, (phys_port_t)curr_in_port, &slvl_mapping, &clbck_data);
            }

            if (ibDiagClbck.GetState())
                goto exit;
        }

    }
exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR / ERR_CODE_DB_ERR
int IBDiag::WritePLFTFile(const char *file_name)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile(file_name, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);
    rc = this->DumpPLFTInfo(sout);
    sout.close();
    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::DumpPLFTInfo(ofstream &sout)
{
    IBDIAG_ENTER;
    char curr_plft_line[2096];
    IBNode *p_curr_node = NULL;
    IBPort *p_port = NULL;

    sout << "File version: 1" << endl;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {
        p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        //check if hca
        if (p_curr_node->type == IB_CA_NODE ||
            !p_curr_node->isPLFTEnabled())
            continue;

        sprintf(curr_plft_line, "dump_plft: Switch " U64H_FMT ,
                p_curr_node->guid_get());
        sout << curr_plft_line << endl;


        for (phys_port_t port = 0; port <= p_curr_node->numPorts ; ++port) {


            if (port !=0 ) {
                p_port = p_curr_node->getPort(port);
                //Skip ports that don't exist or are down or have undefined
                //state or are not in the scope of the fabric.
                if ( !p_port || !p_port->is_data_worthy())
                    continue;
            }

            p_curr_node->getPLFTMapping(port, curr_plft_line);
            sout << "rq: " << (int)port << " sl-plfft: "  << curr_plft_line <<endl;
        }
        sout << endl;
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR / ERR_CODE_DB_ERR
int IBDiag::WriteARFile(const char *file_name)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile(file_name, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);
    rc = this->DumpARInfo(sout);
    sout.close();
    IBDIAG_RETURN(rc);
}

int IBDiag::DumpARInfo(ofstream &sout)
{
    IBDIAG_ENTER;
    char curr_ar_line[2096];

    sout << "File version: 2" << endl;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        //check if hca
        if (p_curr_node->type == IB_CA_NODE ||
            !p_curr_node->isAREnable())
            continue;

        sprintf(curr_ar_line, "dump_ar: Switch " U64H_FMT , p_curr_node->guid_get());

        sout << curr_ar_line ;
        sout << " fr_en: " << (p_curr_node->isFREnabled() ? "1" : "0");
        p_curr_node->getARActiveCfg(curr_ar_line);
        sout << " en_sl: " << curr_ar_line << endl;
        sout << "Groups Definition:" << endl;
        u_int16_t maxARGroupNumber = p_curr_node->getMaxARGroupNumber();

        for (u_int16_t groupNumber = 0;
              groupNumber <= maxARGroupNumber ; ++groupNumber) {
            p_curr_node->getARGroupCfg(groupNumber, curr_ar_line);
            sout << groupNumber << ": "  << curr_ar_line <<endl;
        }

        sout << "LFT Definition:" << endl;
        u_int8_t maxPLFT =  p_curr_node->getMaxPLFT();
        for (u_int8_t pLFT = 0; pLFT <= maxPLFT ; ++pLFT) {
            sout << "PLFT_NUM: " << (int)pLFT << endl;
            sout << "LID    : static port : AR group" << endl;

            u_int16_t lfdbTop = p_curr_node->getLFDBTop(pLFT);

            for (lid_t dlid = 1; dlid <= lfdbTop; ++dlid) {
                u_int8_t static_port = p_curr_node->getLFTPortForLid(dlid, pLFT);
                if (static_port == IB_LFT_UNASSIGNED)
                    sprintf(curr_ar_line, U16H_FMT " :  %s  :  %s" , dlid,
                            "UNREACHABLE" , "UNASSIGNED");
                else {
                    u_int16_t ports_group = p_curr_node->
                        getARLFTPortGroupForLid(dlid, pLFT);

                    if (ports_group == IB_AR_LFT_UNASSIGNED)
                        sprintf(curr_ar_line, U16H_FMT " :  %03u  :  %s" , dlid,
                                static_port , "UNASSIGNED");
                    else
                        sprintf(curr_ar_line, U16H_FMT " :  %03u  :  %03u" , dlid,
                                static_port , ports_group);
                }
                sout << curr_ar_line << endl;
            }
            sout << endl;
        }
        sout << endl;
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR / ERR_CODE_DB_ERR
int IBDiag::WriteFARFile(AdditionalRoutingDataMap *p_routing_data_map,
                         const char *file_name)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile(file_name, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);
    rc = this->DumpFARInfo(p_routing_data_map, sout);
    sout.close();
    IBDIAG_RETURN(rc);
}

void IBDiag::DumpPortsBitset(u_int64_t pgSubBlockElement,
                             phys_port_t portOffset,
                             ofstream &sout)
{
    for (phys_port_t port = 0 ; port < 64 ; ++port) {
        if ( pgSubBlockElement & ((u_int64_t)1 << port) )
            sout << (int)(port+portOffset) << ",";
    }
}

void IBDiag::DumpPortgroup(ib_portgroup_block_element &pgBlockElement,
                           ofstream &sout)
{
    if (pgBlockElement.SubGroup_0)
        DumpPortsBitset(pgBlockElement.SubGroup_0,
                      (phys_port_t)0, sout);
    if (pgBlockElement.SubGroup_1)
        DumpPortsBitset(pgBlockElement.SubGroup_1,
                     (phys_port_t)(1*64), sout);
    if (pgBlockElement.SubGroup_2)
        DumpPortsBitset(pgBlockElement.SubGroup_2,
                     (phys_port_t)(2*64), sout);
    if (pgBlockElement.SubGroup_3)
        DumpPortsBitset(pgBlockElement.SubGroup_3,
                     (phys_port_t)(3*64), sout);
}

int IBDiag::DumpFARInfo(AdditionalRoutingDataMap *p_routing_data_map,
                        ofstream &sout)
{
    IBDIAG_ENTER;
    char curr_ar_line[2096];

    sout << "File version: 2" << endl;

    for (AdditionalRoutingDataMap::iterator rnI = p_routing_data_map->begin();
            rnI != p_routing_data_map->end(); ++rnI) {
        AdditionalRoutingData *p_routing_data = &rnI->second;

        IBNode *p_curr_node = p_routing_data->p_node;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in AdditionalRoutingDataMap");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        //check if hca
        if (p_curr_node->type == IB_CA_NODE ||
            !p_curr_node->isAREnable())
            continue;

        sprintf(curr_ar_line, "dump_ar: Switch " U64H_FMT , p_curr_node->guid_get());

        sout << curr_ar_line ;
        sout << " fr_en: " << (p_curr_node->isFREnabled() ? "1" : "0");
        p_curr_node->getARActiveCfg(curr_ar_line);
        sout << " en_sl: " << curr_ar_line << endl << endl;

        int transpDisable = (int)p_curr_node->getTranspDisable();
        sout << "ByTransportDisable: (0x" << hex << transpDisable << "). ";
        if (transpDisable & 1)
            sout << "UD, ";
        if (transpDisable & 2)
            sout << "RC, ";
        if (transpDisable & 4)
            sout << "UC, ";
        if (transpDisable & 8)
            sout << "DCT";
        sout << dec << endl << endl; // return sout to decimal

        sout << "Groups Definition:" << endl << endl;
        u_int16_t maxARGroupNumber = p_curr_node->getMaxARGroupNumber();
        u_int16_t groupNumber = 0;
        u_int16_t subGroupNumber = 0;

        sout << std::setw(10) << std::left << "Group" <<
            std::setw(15) << std::left << "Sub Group" <<
            "Ports" << endl;
        sout << "-------------------------------------" << endl;

        //top_group_table_block might be 0 if no data was collected
        u_int16_t total_bloks =
            min((u_int16_t)(p_routing_data->top_group_table_block + 1),
                (u_int16_t) p_routing_data->group_table_vec.size());

        for (u_int16_t blockNumber = 0;
              blockNumber < total_bloks;
              ++blockNumber) {
            for (u_int8_t blockEntry = 0;
                 blockEntry < IBIS_IB_MAD_SMP_AR_GROUP_TABLE_BLOCK_SIZE ;
                 blockEntry++) {

                sout << std::setw(10) << std::left << groupNumber <<
                    std::setw(15) << std::left << subGroupNumber;

                DumpPortgroup(p_routing_data->group_table_vec[blockNumber].
                              Group[blockEntry], sout);

                sout << endl;

                subGroupNumber++;
                if(subGroupNumber > p_routing_data->ar_info.sub_grps_active) {
                    subGroupNumber = 0;
                    groupNumber++;
                }
                if (groupNumber > maxARGroupNumber)
                    break;
            }
            if (groupNumber > maxARGroupNumber)
                break;
        }

        sout << "\nLFT Definition:" << endl;
        u_int8_t maxPLFT =  p_curr_node->getMaxPLFT();
        for (u_int8_t pLFT = 0; pLFT <= maxPLFT ; ++pLFT) {
            sout << "\nPLFT_NUM: " << (int)pLFT << endl << endl;
            sout << std::setw(10) << std::left << "LID" <<
                std::setw(15) << std::left << "Static Port" <<
                std::setw(15) << std::left << "Lid State" <<
                "Group" << endl;
            sout << "-----------------------------------------------" << endl;

            u_int16_t lfdbTop = p_curr_node->getLFDBTop(pLFT);

            total_bloks =
                min((u_int16_t)(p_routing_data->top_ar_lft_table_block + 1),
                    (u_int16_t) p_routing_data->ar_lft_table_vec[pLFT].size());

            lid_t dlid = 0;
            for (u_int16_t blockNumber = 0;
                  blockNumber < total_bloks;
                  ++blockNumber) {
                for (u_int8_t blockEntry = 0;
                     blockEntry < IBIS_IB_MAD_SMP_AR_LFT_NUM_BLOCKS_SX ;
                     blockEntry++) {

                    ib_ar_lft_block_element_sx &lft_entry =
                        p_routing_data->ar_lft_table_vec[pLFT][blockNumber].
                        LidEntry[blockEntry];

                    sprintf(curr_ar_line, U16H_FMT, dlid);
                    sout << std::setw(10) << std::left << curr_ar_line <<
                        std::setw(15) << std::left;
                    if (lft_entry.DefaultPort == IB_LFT_UNASSIGNED)
                        sout << "UNREACHABLE";
                    else
                        sout << (int)lft_entry.DefaultPort;

                    sout << std::setw(15) << std::left;

                    switch(lft_entry.LidState) {
                        case AR_IB_LID_STATE_BOUNDED:
                            sout << "Bounded";
                            break;
                        case AR_IB_LID_STATE_FREE:
                            sout << "Free";
                            break;
                        case AR_IB_LID_STATE_STATIC:
                            sout << "Static";
                            break;
                        default:
                            sout << "UNKNOWN";
                            break;
                    }

                    if (lft_entry.LidState == AR_IB_LID_STATE_STATIC &&
                        !p_routing_data->ar_info.fr_enabled)
                        sout << "UNASSIGNED" << endl;
                    else
                        sout << lft_entry.GroupNumber << endl;

                    dlid++;
                    if(dlid > lfdbTop)
                        break;
                }
                if(dlid > lfdbTop)
                    break;
            }
            sout << endl;
        }// pLFT
        sout << endl;
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR / ERR_CODE_DB_ERR
int IBDiag::WriteRNFile(AdditionalRoutingDataMap *p_routing_data_map,
                        const char *file_name)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile(file_name, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);
    rc = this->DumpRNInfo(p_routing_data_map, sout);
    sout.close();
    IBDIAG_RETURN(rc);
}

const char *IBDiag::RNDecisionToStr(u_int8_t decision){
    switch (decision) {
        case 0:
            return "Discard";
        case 1:
            return "Consume ARN";
        case 2:
            return "Consume ARN/FRN";
        case 3:
            return "Pass-on";
    }

    return "Invalid";
}

#define RN_XMIT_PORT_MASK_GENERATE_ARN 0X1
#define RN_XMIT_PORT_MASK_GENERATE_FRN 0X2
#define RN_XMIT_PORT_MASK_PASS_ON_RN   0X4

#define TAB_DEF std::setw(20) << std::left

int IBDiag::DumpRNInfo(AdditionalRoutingDataMap *p_routing_data_map,
                       ofstream &sout)
{
    IBDIAG_ENTER;
    char curr_ar_line[2096];

    set <u_int8_t> directionSet;
    set <u_int8_t>::iterator directionSetIter;
    set <u_int16_t> stringSet;

    u_int16_t maxSubGroups = (u_int16_t)
        ((AdditionalRoutingData::max_sub_group_direction_block + 1) *
         IBIS_IB_MAD_SMP_RN_SUB_GROUP_DIRECTION_BLOCK_SIZE);

    u_int8_t maxDirections = (u_int8_t)
        ((AdditionalRoutingData::max_direction_block + 1) *
         IBIS_IB_MAD_SMP_RN_GEN_STRING_TABLE_BLOCK_SIZE);

    u_int16_t maxStrings = (u_int16_t)
        ((AdditionalRoutingData::max_string_block + 1) *
         IBIS_IB_MAD_SMP_RN_RCV_STRING_BLOCK_SIZE);

    vector <u_int16_t> groupsHistogram(maxSubGroups, 0);
    vector <u_int16_t> sgHistogram(maxSubGroups, 0);
    vector <u_int16_t> directionHistogram(maxDirections, 0);
    vector <u_int16_t> stringHistogram(maxStrings, 0);

    sout << "File version: 1" << endl;

    for (AdditionalRoutingDataMap::iterator rnI = p_routing_data_map->begin();
            rnI != p_routing_data_map->end(); ++rnI) {
        AdditionalRoutingData *p_routing_data = &rnI->second;

        if (!p_routing_data->isRNSupported())
            continue; //continue to next sw

        directionSet.clear();
        stringSet.clear();

        sprintf(curr_ar_line, "\n\ndump_rn: Switch " U64H_FMT ,
                p_routing_data->p_node->guid_get());
        sout << curr_ar_line ;

        sout << "\n\nRN Sub Group Direction Table:" << endl;
        sout <<     "#============================\n" << endl;

        sout << TAB_DEF << "Group" <<
                TAB_DEF << "Sub Group" <<
                "Direction\n";
        sout << "#------------------------------------------------\n";

        u_int16_t blockSize =
            IBIS_IB_MAD_SMP_RN_SUB_GROUP_DIRECTION_BLOCK_SIZE;

        unsigned group_num,sub_group_num;
        u_int16_t total_sg_num = 0;


        for (u_int16_t block = 0;
             block <= p_routing_data->top_sub_group_direction_block ; ++block) {

            for (unsigned blockEntry = 0;
                 blockEntry < blockSize; ++blockEntry) {

                group_num =  (blockEntry + block * blockSize) /
                    (p_routing_data->ar_info.sub_grps_active+1);
                sub_group_num = (blockEntry + block * blockSize) %
                    (p_routing_data->ar_info.sub_grps_active+1);

                if (group_num > p_routing_data->ar_info.group_top)
                    break;

                u_int8_t direction =
                    p_routing_data->sub_group_direction_table_vec[block].
                    direction_of_sub_group[blockEntry];

                sout << TAB_DEF << group_num <<
                        TAB_DEF << sub_group_num <<
                        (int)direction << endl;

                total_sg_num++;
                directionSet.insert(direction);
                directionHistogram[direction]++;
            }
        }
        sgHistogram[total_sg_num]++;

        sout << "\n\nRN Gen String Table:" << endl;
        sout <<     "#===================" << endl;

        blockSize =
            IBIS_IB_MAD_SMP_RN_GEN_STRING_TABLE_BLOCK_SIZE;

        unsigned direction;

        for (u_int8_t pLFT = 0;
             pLFT <= p_routing_data->max_plft ; ++pLFT) {

            sout << "\nPLFT_NUM: " << (int)pLFT << endl;
            sout << TAB_DEF << "Direction" <<
                    "Generated String\n";
            sout << "#-----------------------------------\n";

            for (directionSetIter = directionSet.begin();
                 directionSetIter != directionSet.end();
                 ++directionSetIter) {

                direction = *directionSetIter;

                u_int16_t block = (u_int16_t)(direction/blockSize);
                u_int16_t blockEntry = (u_int16_t)(direction%blockSize);

                u_int16_t genString =
                    p_routing_data->gen_string_table_vec[pLFT][block].
                    element[blockEntry].rn_gen_string;

                sout << TAB_DEF << direction <<
                    TAB_DEF << (int)genString << endl;

                stringSet.insert(genString);
                stringHistogram[genString]++;
            }

        }

        sout << "\n\nRN Gen By Sub Group Priority:" << endl;
        sout <<     "#============================\n" << endl;

        sout << TAB_DEF << "SG Priority" <<
            TAB_DEF << "gen_arn" << "gen_frn\n";
        sout << "#----------------------------------------------\n";

        for (unsigned sgPriority = 0;
             sgPriority <= p_routing_data->ar_info.sub_grps_active;
             ++sgPriority) {

            sout << TAB_DEF << sgPriority << TAB_DEF <<
                (int)p_routing_data->gen_by_sub_group_priority.
                element[sgPriority].gen_arn << TAB_DEF <<
                (int)p_routing_data->gen_by_sub_group_priority.
                element[sgPriority].gen_frn << endl;
        }

        sout << "\n\nRN Rcv String:" << endl;
        sout <<     "#=============\n" << endl;

        sout << TAB_DEF << "String" <<
            TAB_DEF << "Decision" <<
            TAB_DEF << "pLFT" <<
            "String to string\n";
        sout <<
            "#--------------------------------------------" <<
            "-------------------------------\n";

        blockSize = IBIS_IB_MAD_SMP_RN_RCV_STRING_BLOCK_SIZE;

        u_int32_t maxString =
            (0x1 << (p_routing_data->ar_info.string_width_cap)) - 1;

        unsigned stringNum;

        for (u_int16_t block = 0;
             block <= p_routing_data->top_string_block ; ++block) {

            for (unsigned blockEntry = 0;
                 blockEntry < blockSize; ++blockEntry) {

                stringNum =  (blockEntry + block * blockSize);

                if (stringNum > maxString)
                    break;

                sout << TAB_DEF << stringNum <<
                    TAB_DEF <<
                    RNDecisionToStr(p_routing_data->rcv_string_vec[block].
                                    element[blockEntry].decision) <<
                    TAB_DEF <<
                    (int)p_routing_data->rcv_string_vec[block].
                    element[blockEntry].plft_id <<
                    TAB_DEF <<
                    (int)p_routing_data->rcv_string_vec[block].
                    element[blockEntry].string2string << endl;
            }
        }

        sout << "\n\nRN Xmit Port Mask:" << endl;
        sout <<     "#=================\n" << endl;

        sout << TAB_DEF << "Port" << TAB_DEF << "Generate ARN" <<
            TAB_DEF << "Generate FRN" << "Pass on RN\n";
        sout << "#-----------------------------------------" <<
            "----------------------------\n";


        blockSize = IBIS_IB_MAD_SMP_RN_XMIT_PORT_MASK_BLOCK_SIZE;

        unsigned portNum;

        for (u_int16_t block = 0;
             block <= p_routing_data->top_string_block ; ++block) {

            for (unsigned blockEntry = 0;
                 blockEntry < blockSize; ++blockEntry) {

                portNum =  (blockEntry + block * blockSize);

                if (portNum > p_routing_data->p_node->numPorts)
                    break;

                u_int8_t element =
                    p_routing_data->xmit_port_mask_vec[block].
                        element[blockEntry];

                sout << TAB_DEF << portNum << TAB_DEF <<
                    (bool)(element & RN_XMIT_PORT_MASK_GENERATE_ARN) <<
                    TAB_DEF <<
                    (bool)(element & RN_XMIT_PORT_MASK_GENERATE_FRN) <<
                    TAB_DEF <<
                    (bool)(element & RN_XMIT_PORT_MASK_PASS_ON_RN) << endl;
            }
        }

        sout << "\n\nStatistics:" << endl;
        sout <<     "#==========\n" << endl;

        //Statistics
        u_int16_t grpNum = p_routing_data->p_node->getNonEmptyARGroupNumber();
        groupsHistogram[grpNum]++;

        sprintf(curr_ar_line, "Groups:%u    Sub Groups:%u"
                "    Directions:%u    Strings:%u\n\n",
                grpNum, total_sg_num,
                (int)directionSet.size(), (int)stringSet.size());
        sout << curr_ar_line ;

        sout << "#*************************************************************************************\n";

    }
    sout << "#*************************************************************************************\n";

    sout << "\nHistograms:" << endl;
    sout <<   "#==========\n\n" << endl;

    sout << TAB_DEF << "Groups Number"<< "SW Number\n";
    sout << "#----------------------------\n";

    int size = (int)groupsHistogram.size();
    for (int i = 0 ; i < size ; ++i) {
        if (groupsHistogram[i])
            sout << TAB_DEF << i << TAB_DEF << (int)groupsHistogram[i] << endl;
    }

    sout << "\n" << TAB_DEF << "SG Number"<< "SW Number\n";
    sout << "#----------------------------\n";

    size = (int)sgHistogram.size();
    for (int i = 0 ; i < size ; ++i) {
        if (sgHistogram[i])
            sout << TAB_DEF << i << TAB_DEF << (int)sgHistogram[i] << endl;
    }

    sout << "\n" << TAB_DEF << "Direction"<<
        "SG Number\n";
    sout << "#----------------------------\n";

    size = (int)directionHistogram.size();
    for (int i = 0 ; i < size ; ++i) {
        if (directionHistogram[i])
            sout << TAB_DEF << i <<
            TAB_DEF << (int)directionHistogram[i] << endl;
    }

    sout << "\n" << TAB_DEF << "Gen String"<<
        "SW Number\n";
    sout << "#----------------------------\n";

    size = (int)stringHistogram.size();
    for (int i = 0 ; i < size ; ++i) {
        if (stringHistogram[i])
            sout << TAB_DEF << i <<
            TAB_DEF << (int)stringHistogram[i] << endl;
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR / ERR_CODE_DB_ERR
int IBDiag::WriteRNCountersFile(AdditionalRoutingDataMap *p_routing_data_map,
                                const char *file_name)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile(file_name, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);
    rc = this->DumpRNCountersInfo(p_routing_data_map, sout);
    sout.close();
    IBDIAG_RETURN(rc);
}

int IBDiag::DumpRNCountersInfo(AdditionalRoutingDataMap *p_routing_data_map,
                               ofstream &sout)
{
    IBDIAG_ENTER;
    char curr_ar_line[2096];

    u_int64_t max_port_rcv_rn_pkt = 0;
    u_int64_t max_port_xmit_rn_pkt = 0;
    u_int64_t max_port_rcv_rn_error = 0;
    u_int64_t max_sw_relay_rn_error = 0;

    sout << "File version: 1" << endl;

    for (AdditionalRoutingDataMap::iterator rnI = p_routing_data_map->begin();
            rnI != p_routing_data_map->end(); ++rnI) {
        AdditionalRoutingData *p_routing_data = &rnI->second;

        if (!p_routing_data->isRNSupported())
            continue; //continue to next sw

        sprintf(curr_ar_line, "\n\ndump_rnc: Switch " U64H_FMT ,
                p_routing_data->p_node->guid_get());
        sout << curr_ar_line << endl << endl;

        sout << TAB_DEF << "Port" <<
            TAB_DEF << "Rcv RN Pkt" <<
            TAB_DEF << "Xmit RN Pkt" <<
            TAB_DEF << "Rcv RN Error" <<
            "Rcv SW Relay RN Error" << endl;
        sout << "#------------------------------------------------" <<
            "-----------------------------------------------------\n";

        for (u_int8_t port = 1;
             port <= p_routing_data->p_node->numPorts ; ++port) {

            sout << TAB_DEF << (int)port <<
                TAB_DEF <<
                p_routing_data->port_rn_counters_vec[port].port_rcv_rn_pkt <<
                TAB_DEF <<
                p_routing_data->port_rn_counters_vec[port].port_xmit_rn_pkt <<
                TAB_DEF <<
                p_routing_data->port_rn_counters_vec[port].port_rcv_rn_error <<
                p_routing_data->port_rn_counters_vec[port].
                    port_rcv_switch_relay_rn_error << endl;

            max_port_rcv_rn_pkt =
                max(max_port_rcv_rn_pkt,
                    p_routing_data->port_rn_counters_vec[port].port_rcv_rn_pkt);

            max_port_xmit_rn_pkt =
                max(max_port_xmit_rn_pkt,
                    p_routing_data->port_rn_counters_vec[port].port_xmit_rn_pkt);

            max_port_rcv_rn_error =
                max(max_port_rcv_rn_error,
                    p_routing_data->port_rn_counters_vec[port].port_rcv_rn_error);

            max_sw_relay_rn_error =
                max(max_sw_relay_rn_error,
                    p_routing_data->port_rn_counters_vec[port].
                    port_rcv_switch_relay_rn_error);

        }

        sout << "\n#*************************************************************************************\n";

    }
    sout << "#*************************************************************************************\n";

    sout << "\nMax Values:" << endl;
    sout <<   "#==========\n\n" << endl;

    sout << "Max Rcv RN Pkt: " <<
        max_port_rcv_rn_pkt <<
        "   Max Xmit RN Pkt: " <<
        max_port_xmit_rn_pkt <<
        "   Max Rcv RN Error: " <<
        max_port_rcv_rn_error <<
        "   Max Rcv SW Relay RN Error: " <<
        max_sw_relay_rn_error;

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR / ERR_CODE_DB_ERR
int IBDiag::WriteVL2VLFile(const char *file_name)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ofstream sout;
    int rc = this->OpenFile(file_name, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);
    rc = this->DumpVL2VLInfo(sout);
    sout.close();
    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::DumpVL2VLInfo(ofstream &sout)
{
    IBDIAG_ENTER;
    char curr_sw_line[1024];
    char curr_vl2vl_line[1024];
    IBNode *p_curr_node = NULL;

    sout << "File version: 1" << endl;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {
        p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        //check if hca
        if (p_curr_node->type == IB_CA_NODE)
            continue;

        p_curr_node->getSL2VLCfg(curr_vl2vl_line);
        if (curr_vl2vl_line[0] != '\0') {
            sprintf(curr_sw_line, "dump_vl2vl: Switch " U64H_FMT " ",
                    p_curr_node->guid_get());
            sout << curr_sw_line << curr_vl2vl_line << endl;
        }
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}
