/*
 * © Copyright (c) 2004-2019 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include "infiniband/ibdiag/ibdiag.h"
#include <infiniband/ibdiag/ibdiag_types.h>
#include "infiniband/ibdiag/ibdiag_clbck.h"

extern IBDiagClbck ibDiagClbck;

/****************************************************/
typedef map < u_int16_t, u_int8_t > map_pkey_membership;
typedef map < IBPort *, u_int8_t > map_p_port_membership;
typedef map < u_int16_t, map_p_port_membership > map_pkey_p_port_membership;
typedef vector < struct P_Key_Block_Element > vec_pkey_block_elem;
typedef map < IBVPort *, u_int8_t > map_p_vport_membership;
typedef map < u_int16_t, map_p_vport_membership > map_pkey_p_vport_membership;
typedef pair < u_int64_t, IBVPort * > pair_vguid_vport;

/****************************************************/

void readPortPartitionTableToMap(IBDMExtendedInfo *extended_info,
        get_pkey_table_func_t get_pkey_tbl,
        u_int32_t createIndex, u_int16_t cap,
        map_pkey_membership &pkeys_to_mem);

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::CheckAndSetVPortLid(list_p_fabric_general_err &vport_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    for (u_int32_t i = 0; i < this->fabric_extended_info.getVPortsVectorSize();
         ++i) {
        IBVPort *p_vport = this->fabric_extended_info.getVPortPtr(i);
        if (!p_vport || p_vport->get_vlid())
            continue;
        SMP_VPortInfo *p_vport_info =
            this->fabric_extended_info.getSMPVPortInfo(p_vport->createIndex);
        if (!p_vport_info)
            continue;
        //if lid_required then above get_vlid() should have return non zero.
        if (p_vport_info->lid_required) {
            FabricErrVLidZero *p_error =
                new FabricErrVLidZero(p_vport->getIBPortPtr(), p_vport);
            if (!p_error) {
                this->SetLastError("Failed to allocate"  \
                        " FabricErrVLidZero");
                IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
            }
            vport_errors.push_back(p_error);
            continue;
        }
        else {
            //from spec:
            ////If LIDRequired is set to 0, the index of the VPort
            ////(that shall have LIDRequired set to 1) this VPort is attached to.
            IBVPort * p_attached_vport =
                p_vport->getIBPortPtr()->getVPort(p_vport_info->lid_by_vport_idx);
            if (!p_attached_vport) {
                //add error invalid index for vlid
                FabricErrInvalidIndexForVLid *p_error =
                    new FabricErrInvalidIndexForVLid(
                            p_vport->getIBPortPtr(),
                            p_vport,
                            p_vport_info->lid_by_vport_idx);
                if (!p_error) {
                    this->SetLastError("Failed to allocate"  \
                            " FabricErrInvalidIndexForVLid");
                    IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
                }
                vport_errors.push_back(p_error);
                continue;
            } else if (p_attached_vport->get_vlid() == 0) {
                //add error remove lid by index is 0
                FabricErrVlidForVlidByIndexIsZero *p_error =
                    new FabricErrVlidForVlidByIndexIsZero(
                            p_vport->getIBPortPtr(),
                            p_vport,
                            p_attached_vport,
                            p_vport_info->lid_by_vport_idx);
                if (!p_error) {
                    this->SetLastError("Failed to allocate"  \
                            " FabricErrVlidForVlidByIndexIsZero");
                    IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
                }
                vport_errors.push_back(p_error);
                continue;
            }

            p_vport->set_vlid(p_attached_vport->get_vlid());
        }
    }

    IBDIAG_RETURN(rc);
}
//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::BuildVirtualizationDB(list_p_fabric_general_err &vport_errors,
                         progress_func_nodes_t progress_func)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    progress_bar_nodes_t progress_bar_nodes;
    CLEAR_STRUCT(progress_bar_nodes);

    ibDiagClbck.Set(this, &fabric_extended_info, &vport_errors, NULL);
    //go over all the nodes and discover the virtual ports
    for (map_str_pnode::iterator nI =
            this->discovered_fabric.NodeByName.begin();
         nI != this->discovered_fabric.NodeByName.end();
         ++nI) {
        IBNode *p_curr_node = nI->second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map"  \
                               " for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }
        ++progress_bar_nodes.nodes_found;
        if (p_curr_node->type == IB_SW_NODE)
            ++progress_bar_nodes.sw_found;
        else
            ++progress_bar_nodes.ca_found;
        if (progress_func)
            progress_func(&progress_bar_nodes,
                          &this->discover_progress_bar_nodes);
        //vports exist only in HCAs
        if (p_curr_node->type != IB_CA_NODE)
            continue;

        if (GetSpecialCAPortType(p_curr_node) != NotSpecial) // skip special nodes
            continue;

        if (!this->capability_module.IsSupportedSMPCapability(
                                        p_curr_node,
                                        EnSMPCapIsVirtualizationSupported))
            continue;

        rc = BuildVirtualizationInfoDB(p_curr_node);
        if(rc != IBDIAG_SUCCESS_CODE ) {
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,"Error during get on Virtualization" \
                       " Info on node %s"  \
                       " skipping sending VPortState and VPortInfo\n",
                       p_curr_node->name.c_str());
            continue;
        }

        // second stage to get only the not down vport.
        rc = BuildVPortStateDB(p_curr_node);
        if(rc != IBDIAG_SUCCESS_CODE) {
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,"Error during get on VPort" \
                       " State on node %s"  \
                       " skipping sending VPortInfo\n",
                       p_curr_node->name.c_str());
            continue;
        }

        // third stage to get the VPort Info only for not down vports
        rc = BuildVPortInfoDB(p_curr_node);
        if(rc != IBDIAG_SUCCESS_CODE) {
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,"Error during get on VPort" \
                       " Info on node %s skipping to next node\n",
                       p_curr_node->name.c_str());
            continue;
        }

        // fourth stage to get the VNode Info
        rc = BuildVNodeInfoDB(p_curr_node);
        if(rc != IBDIAG_SUCCESS_CODE) {
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,"Error during get on VNode" \
                       " Info on node %s skipping to next node\n",
                       p_curr_node->name.c_str());
            continue;
        }

        // fifth stage to get the VNode Description
        rc = BuildVNodeDescriptionDB(p_curr_node);
        if(rc != IBDIAG_SUCCESS_CODE) {
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,"Error during get on VNode" \
                       " Description on node %s skipping to next node\n",
                       p_curr_node->name.c_str());
            continue;
        }

        // sixth stage to get the p_key tables
        rc = BuildVPortPKeyTableDB(p_curr_node);
        if(rc != IBDIAG_SUCCESS_CODE) {
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,"Error during get on VPort" \
                       " PKey Table on node %s skipping to next node\n",
                       p_curr_node->name.c_str());
            continue;
        }
    }

    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::BuildVirtualizationInfoDB(IBNode *p_node)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = IBDiagSMPVirtualizationInfoGetClbck;

    for (unsigned int i = 1 ; i <=  p_node->numPorts ; i++) {
        IBPort *p_curr_port = p_node->getPort((phys_port_t)i);
        if (!p_curr_port || !p_curr_port->is_data_worthy()) //port not exist
            continue;

        clbck_data.m_data1 = p_curr_port;
        struct SMP_VirtualizationInfo virtual_info;
        this->ibis_obj.SMPVirtualizationInfoMadGetByLid(p_curr_port->base_lid,
                                                        &virtual_info,
                                                        &clbck_data);
        if (ibDiagClbck.GetState()) {
            goto exit;
        }
    }

exit:
    this->ibis_obj.MadRecAll();

    if (rc) {
        if (last_error.empty())
            this->SetLastError("Retrieve of VS VirtualizationInfo Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR,
                       "Retrieve of VS VirtualizationInfo Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
    }

    IBDIAG_RETURN(rc);
}


//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::BuildVPortStateDB(IBNode *p_node)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = IBDiagSMPVPortStateGetClbck;

    for (unsigned int i = 1 ; i <=  p_node->numPorts ; i++) {
        IBPort *p_curr_port = p_node->getPort((phys_port_t)i);
        if (!p_curr_port || !p_curr_port->is_data_worthy()) //port not exist
            continue;

        struct SMP_VirtualizationInfo *virtual_info =
            this->fabric_extended_info.getSMPVirtualizationInfo(
                                            p_curr_port->createIndex);
        if(!virtual_info || !virtual_info->virtualization_enable)
            continue;

        clbck_data.m_data1 = p_curr_port;
        u_int16_t vport_state_blocks =
            virtual_info->vport_index_top/VPORT_STATE_BLOCK_SIZE;
        for (u_int16_t j = 0; j <= vport_state_blocks; j++) {
            struct SMP_VPortState vport_state_block;
            //block index for the vport_state vector
            clbck_data.m_data2 = (void *)(uintptr_t)j;
            this->ibis_obj.SMPVPortStateMadGetByLid(p_curr_port->base_lid,
                                                    j,
                                                    &vport_state_block,
                                                    &clbck_data);
            if (ibDiagClbck.GetState())
                goto exit;
        }//vport blocks loop
    }//ports loop

exit:
    this->ibis_obj.MadRecAll();

    if (rc) {
        if (last_error.empty())
            this->SetLastError("Retrieve of VS VPortStateInfo Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR,
                       "Retrieve of VS VPortStateInfo Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
    }
    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::BuildVPortInfoDB(IBNode *p_node)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    uint8_t blocknum = 0;
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = IBDiagSMPVPortInfoGetClbck;

    for (unsigned int i = 1 ; i <=  p_node->numPorts ; i++) {
        // collect node's ports' vportinfo for the vports, that aren't down
        IBPort *p_curr_port = p_node->getPort((phys_port_t)i);
        if (!p_curr_port || !p_curr_port->is_data_worthy()) //port not exist
            continue;

        struct SMP_VirtualizationInfo *virtual_info =
            this->fabric_extended_info.getSMPVirtualizationInfo(
                                            p_curr_port->createIndex);
        if(!virtual_info || !virtual_info->virtualization_enable)
            continue;

        clbck_data.m_data1 = p_curr_port;
        struct SMP_VPortState *vport_state_block = NULL;
        for (virtual_port_t vPrtIdx = 0;
                vPrtIdx <= virtual_info->vport_index_top;
                vPrtIdx++) {
            // only when we need new state_block (0, 128, 256...)
            if ((vPrtIdx % VPORT_STATE_BLOCK_SIZE) == 0) {
                blocknum = (uint8_t) (vPrtIdx / VPORT_STATE_BLOCK_SIZE);
                vport_state_block = this->fabric_extended_info.getSMPVPortState(
                        p_curr_port->createIndex,
                        blocknum);
            }

            if (!vport_state_block)
                continue;

            uint8_t vp_state = vport_state_block->vport_state[
                vPrtIdx % VPORT_STATE_BLOCK_SIZE];
            if (IsPortStateDown((IBPortState)(vp_state)))
                continue;
            clbck_data.m_data2 = (void *)(uintptr_t)vPrtIdx;
            struct SMP_VPortInfo vport_info;
            this->ibis_obj.SMPVPortInfoMadGetByLid(p_curr_port->base_lid,
                    vPrtIdx,//vport number
                    &vport_info,
                    &clbck_data);
            if (ibDiagClbck.GetState())
                goto exit;
        }//vports loop
    }//ports loop

exit:
    this->ibis_obj.MadRecAll();

    if (rc) {
        if (last_error.empty())
            this->SetLastError("Retrieve of VS VPortInfo Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR,
                       "Retrieve of VS VPortInfo Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
    }

    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::BuildVNodeInfoDB(IBNode *p_node)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = IBDiagSMPVNodeInfoGetClbck;
    for (unsigned int i = 1 ; i <=  p_node->numPorts ; i++) {
        IBPort *p_curr_port = p_node->getPort((phys_port_t)i);
        if (!p_curr_port || !p_curr_port->is_data_worthy()) //port not exist
            continue;

        clbck_data.m_data1 = p_curr_port;
        map_vportnum_vport vports = p_curr_port->VPorts;
        for (map_vportnum_vport::iterator vpI = vports.begin();
                vpI != vports.end(); ++vpI) {
            IBVPort *p_vport = vpI->second;
            if(!p_vport)
                continue;

            clbck_data.m_data2 = p_vport;
            SMP_VNodeInfo vnode_info;
            this->ibis_obj.SMPVNodeInfoMadGetByLid(p_curr_port->base_lid,
                    p_vport->getVPortNum(),
                    &vnode_info,
                    &clbck_data);

            if (ibDiagClbck.GetState())
                goto exit;
        }
    }

exit:
    this->ibis_obj.MadRecAll();

    if (rc) {
        if (last_error.empty())
            this->SetLastError("Retrieve of VS VNodeInfo Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR,
                       "Retrieve of VS VNodeInfo Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::BuildVNodeDescriptionDB(IBNode *p_node)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = IBDiagSMPVNodeDescriptionGetClbck;

    map_guid_pvnode vnode_guids = this->discovered_fabric.VNodeByGuid;
    for (map_guid_pvnode::iterator vnI = vnode_guids.begin();
            vnI != vnode_guids.end(); ++vnI) {
        IBVNode *p_vnode = vnI->second;
        if (!p_vnode) //vnode not exist
            continue;

        clbck_data.m_data2 = p_vnode;
        map_vportnum_vport vports = p_vnode->VPorts;
        for (map_vportnum_vport::iterator vpI = vports.begin();
                vpI != vports.end(); ++vpI) {
            IBVPort *p_vport = vpI->second;
            if(!p_vport) //vport not exist
                continue;

            IBPort *p_port = p_vport->getIBPortPtr();
            clbck_data.m_data1 = p_port;
            SMP_NodeDesc vnode_description;
            this->ibis_obj.SMPVNodeDescriptionMadGetByLid(p_port->base_lid,
                    p_vport->getVPortNum(),
                    &vnode_description,
                    &clbck_data);

            if (ibDiagClbck.GetState())
                goto exit;
            break;
        }
    }

exit:
    this->ibis_obj.MadRecAll();

    if (rc) {
        if (last_error.empty())
            this->SetLastError("Retrieve of VS VNodeDescription Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR,
                       "Retrieve of VS VNodeDescription Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
    }

    IBDIAG_RETURN(rc);
}



//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::BuildVPortPKeyTableDB(IBNode *p_node)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    u_int16_t partition_cap;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = IBDiagSMPVPortPKeyGetClbck;

    for (unsigned int i = 1 ; i <=  p_node->numPorts ; i++) {
        IBPort *p_curr_port = p_node->getPort((phys_port_t)i);
        if (!p_curr_port || !p_curr_port->is_data_worthy()) //port not exist
            continue;

        clbck_data.m_data1 = p_curr_port;
        struct SMP_VirtualizationInfo *virtual_info =
            this->fabric_extended_info.getSMPVirtualizationInfo(
                                            p_curr_port->createIndex);
        if(!virtual_info || !virtual_info->virtualization_enable)
            continue;

        map_vportnum_vport vports = p_curr_port->VPorts;
        for (map_vportnum_vport::iterator vpI = vports.begin();
                vpI != vports.end(); ++vpI) {
            IBVPort *p_vport = vpI->second;

            if(!p_vport)
                continue;
            IBVNode *p_vnode = p_vport->getVNodePtr();
            if(!p_vnode)
                continue;
            struct SMP_VNodeInfo *vnode_info =
                this->fabric_extended_info.getSMPVNodeInfo(
                    p_vnode->createIndex);

            partition_cap = vnode_info->vpartition_cap;
            int num_of_blocks =
                (partition_cap + IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS - 1) /
                IBIS_IB_MAD_SMP_PKEY_TABLE_NUM_BLOCKS;
            clbck_data.m_data2 = p_vport;

            //get direct route for this port
            direct_route_t *p_curr_direct_route =
                    this->GetDirectRouteByPortGuid(p_curr_port->guid_get());
            if (!p_curr_direct_route) {
                this->SetLastError("DB error - can't find direct route to node=%s, port=%u",
                                   p_node->getName().c_str(), (u_int8_t)p_curr_port->num);
                rc = IBDIAG_ERR_CODE_DB_ERR;
                goto exit;
            }

            for (u_int16_t j = 0; j < num_of_blocks; ++j) {

                SMP_PKeyTable vport_pkey_tbl;
                //block index for the vport_state vector
                clbck_data.m_data3 = (void *)(uintptr_t)j;
                this->ibis_obj.SMPVPortPKeyTblMadGetByDirect(p_curr_direct_route,
                                                             p_vport->getVPortNum(),
                                                             j,
                                                             &vport_pkey_tbl,
                                                             &clbck_data);
                if (ibDiagClbck.GetState())
                    goto exit;
            }
        }//vport loop
    }//ports loop

exit:
    this->ibis_obj.MadRecAll();

    if (rc) {
        if (last_error.empty())
            this->SetLastError("Retrieve of VS VPortPkeyTable Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR,
                       "Retrieve of VS VPortPkeyTable Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
    }
    IBDIAG_RETURN(rc);
}

void IBDiag::DumpVPorts(ofstream &sout)
{
    IBDIAG_ENTER;
    char line[2096];
    vec_guids vec_guids;

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize();
            ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node)   //node not exists
            continue;
        // vports relevant only for HCAs
        if (p_curr_node->type == IB_SW_NODE)
            continue;

        // use unsigned int index to prevent overwrap
        for (unsigned int i = 1; i <= p_curr_node->numPorts; i++) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            if (!p_curr_port)
                continue;

            SMP_VirtualizationInfo *p_virtual_info =
                    this->fabric_extended_info.getSMPVirtualizationInfo(
                                                    p_curr_port->createIndex);
            if (!p_virtual_info || p_curr_port->VPorts.size() == 0)
                continue;

            memset(line, 0, sizeof(line));
            sprintf(line, "Port Name=%s, LID=%d, GUID=" U64H_FMT
                    ", Index Cap=%d, Index Top=%d",
                    p_curr_port->getName().c_str(), p_curr_port->base_lid,
                    p_curr_port->guid_get(),
                    p_virtual_info->vport_cap, p_virtual_info->vport_index_top);
            sout << line << endl;

            map_vportnum_vport vports = p_curr_port->VPorts;
            for (map_vportnum_vport::iterator vpI = vports.begin();
                    vpI != vports.end(); ++vpI) {
                IBVPort *p_vport = vpI->second;
                if(!p_vport)
                    continue;
                IBVNode *p_vnode = p_vport->getVNodePtr();
                if(!p_vnode)
                    continue;
                sprintf(line, "\tVPort%d: Guid=" U64H_FMT ", VLid=%d, State=%s,"
                        " VNode Guid=" U64H_FMT ", VNode Description=%s",
                        p_vport->getVPortNum(),
                        p_vport->guid_get(),
                        p_vport->get_vlid(),
                        portstate2char(p_vport->get_state()),
                        p_vnode->guid_get(),
                        p_vnode->getDescription().c_str());
                sout << line << endl;
            }
            sout << endl;
        }
    }
    IBDIAG_RETURN_VOID;
}

int IBDiag::DumpCSVVPortsTable(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    csv_out.DumpStart(SECTION_VPORTS);

    stringstream sstream;
    sstream << "NodeGuid," << "PortGUID," << "PortNum," << "VPortIndex,"
         << "VPortGuid," << "VPortLid," << "VCapMask," << "VGuidCap,"
         << "VPortClientReg," << "VPortState," << "QKEYViolations,"
         << "PKEYViolations," << "VPortProfile"
         << endl;
    csv_out.WriteLine(sstream.str());

    for (u_int32_t i = 0; i < this->fabric_extended_info.getVPortsVectorSize();
         ++i) {
        IBVPort *p_vport = this->fabric_extended_info.getVPortPtr(i);
        if (!p_vport)
            continue;
        SMP_VPortInfo *p_vport_info =
            this->fabric_extended_info.getSMPVPortInfo(p_vport->createIndex);
        if (!p_vport_info)
            continue;

        char curr_node_line[1024];
        memset(curr_node_line, 0, sizeof(curr_node_line));
        sstream.str("");
        sprintf(curr_node_line, U64H_FMT ","  U64H_FMT "," U32D_FMT "," \
                U32D_FMT "," U64H_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT \
                "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U64H_FMT,
                p_vport->getIBPortPtr()->p_node->guid_get(),
                p_vport->getIBPortPtr()->guid_get(),
                p_vport->getIBPortPtr()->num,
                p_vport->getVPortNum(),
                p_vport->guid_get(),
                p_vport->get_vlid(),
                p_vport_info->vport_capability_mask,
                p_vport_info->vguid_cap,
                p_vport_info->vport_client_reregister,
                p_vport_info->vport_state,
                p_vport_info->qkey_violations,
                p_vport_info->pkey_violations,
                p_vport_info->vport_profile);
        sstream << curr_node_line << endl;
        csv_out.WriteLine(sstream.str());
    }
    csv_out.DumpEnd(SECTION_VPORTS);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::DumpCSVVNodesTable(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    csv_out.DumpStart(SECTION_VNODES);

    stringstream sstream;
    sstream << "NodeGuid," << "PortGUID," << "PortNum," << "VPortIndex,"
         << "VNodeDesc,"<< "VNumberOfPorts," << "VLocalPortNum,"
         << "VPartitionCap," << "VNodeGuid"
         << endl;
    csv_out.WriteLine(sstream.str());

    for (u_int32_t i = 0; i < this->fabric_extended_info.getVNodesVectorSize();
         ++i) {
        IBVNode *p_vnode = this->fabric_extended_info.getVNodePtr(i);
        if (!p_vnode)
            continue;
        SMP_VNodeInfo *p_vnode_info =
            this->fabric_extended_info.getSMPVNodeInfo(p_vnode->createIndex);
        if (!p_vnode_info)
            continue;
        IBVPort *p_vport = NULL;
        for (map_vportnum_vport::iterator vpI = p_vnode->VPorts.begin();
                vpI != p_vnode->VPorts.end(); ++vpI) {
            p_vport = vpI->second;
            if(p_vport)
                break;
        }

        if (!p_vport)
            continue;
        char curr_node_line[1024];
        memset(curr_node_line, 0, sizeof(curr_node_line));
        sstream.str("");
        sprintf(curr_node_line, U64H_FMT ","  U64H_FMT "," U32D_FMT "," \
                U32D_FMT "," STR_FMT "," U32D_FMT ","  U32D_FMT "," \
                U32D_FMT "," U64H_FMT,
                //p_curr_node->getName().c_str(),
                p_vport->getIBPortPtr()->p_node->guid_get(),
                p_vport->getIBPortPtr()->guid_get(),
                p_vport->getIBPortPtr()->num,
                p_vport->getVPortNum(),
                p_vnode->getDescription().c_str(),
                p_vnode_info->vnum_ports,
                p_vnode_info->vlocal_port_num,
                p_vnode_info->vpartition_cap,
                p_vnode->guid_get());
        sstream << curr_node_line << endl;
        csv_out.WriteLine(sstream.str());
    }
    csv_out.DumpEnd(SECTION_VNODES);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR
int IBDiag::WriteVPortsFile(const char *file_name)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);
    ofstream sout;
    int rc = this->OpenFile(file_name, sout,
                            false, //no append
                            true); //add_header
    if (rc)
        IBDIAG_RETURN(rc);
    this->DumpVPorts(sout);
    sout.close();
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void IBDiag::DumpVPortsPKeyTbl(ofstream &sout)
{
    IBDIAG_ENTER;
    char line[2096];
    u_int16_t partition_cap;
    map_pkey_p_vport_membership pkey_to_p_vport_mem;

    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node)   //node not exists
            continue;

        // We do not dump switch physical ports pkey tables
        if (p_curr_node->type == IB_SW_NODE)
            continue;

        // use unsigned int index to prevent overwrap
        for (unsigned int i = 1; i <= p_curr_node->numPorts; ++i) {
            IBPort * p_curr_port = p_curr_node->getPort((phys_port_t)i);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            map_vportnum_vport vports = p_curr_port->VPorts;
            for (map_vportnum_vport::iterator vpI = vports.begin();
                    vpI != vports.end(); ++vpI) {
                IBVPort *p_vport = vpI->second;
                if(!p_vport)
                    continue;
                IBVNode *p_vnode = p_vport->getVNodePtr();
                if(!p_vnode)
                    continue;
                SMP_VNodeInfo *p_vnode_info =
                    this->fabric_extended_info.getSMPVNodeInfo(p_vnode->createIndex);
                if (!p_vnode_info)
                    continue;
                partition_cap = p_vnode_info->vpartition_cap;
                map_pkey_membership pkey_to_mem;
                readPortPartitionTableToMap(&this->fabric_extended_info,
                        (get_pkey_table_func_t)(&IBDMExtendedInfo::getSMPVPortPKeyTable),
                        p_vport->createIndex, partition_cap, pkey_to_mem);

                // add this port pkeys to map of pkey->port->membership
                for (map_pkey_membership::iterator it_pm = pkey_to_mem.begin();
                        it_pm != pkey_to_mem.end();
                        ++it_pm) {
                    u_int16_t pkey = it_pm->first;
                    u_int8_t mem = it_pm->second;

                    map_p_vport_membership p_port_to_mem;
                    map_pkey_p_vport_membership::iterator it_ppm = pkey_to_p_vport_mem.find(pkey);
                    if (it_ppm == pkey_to_p_vport_mem.end()) {
                        p_port_to_mem.clear();
                        std::pair<map_pkey_p_vport_membership::iterator,bool> ret;
                        ret = pkey_to_p_vport_mem.insert(pair < u_int16_t, map_p_vport_membership > (pkey, p_port_to_mem));
                        // insert retruned false :element already existed
                        if (!ret.second)
                            continue;
                        // insert succeeded, get the it_ppm.
                        it_ppm = ret.first;
                    }

                    map_p_vport_membership::iterator it_pm1 = it_ppm->second.find(p_vport);
                    if (it_pm1 == it_ppm->second.end()) {
                        it_ppm->second.insert(pair< IBVPort*, u_int8_t >(p_vport, mem));
                    }
                }
            } //for vport
        }//for phys ports
    }//for nodes

    // finally dump all
    for (map_pkey_p_vport_membership::iterator it_ppm = pkey_to_p_vport_mem.begin();
            it_ppm != pkey_to_p_vport_mem.end(); ++it_ppm) {
        memset(line, 0, sizeof(line));
        sprintf(line, "GROUP VPKey:" U16H_FMT " Virtual Ports:%lu",
            it_ppm->first,
            (long unsigned int)it_ppm->second.size());
        sout << line << endl;

        for (map_p_vport_membership::iterator it_pm = it_ppm->second.begin();
                it_pm != it_ppm->second.end(); ++it_pm) {
//            string membership_str = "Limited";
//            if (it_pm->second)
//                membership_str = "Full";

            sprintf(line, "%s %s guid=" U64H_FMT ,
                (it_pm->second ? "Full": "Limited"),
                it_pm->first->getName().c_str(),
                it_pm->first->guid_get());
            sout << "  " << line << endl;
        }
        sout << endl;
    }
    IBDIAG_RETURN_VOID;
}

//Returns: SUCCESS_CODE / ERR_CODE_NO_MEM / ERR_CODE_CHECK_FAILED
//The port_aguids includes all the phys port guids and the alias guids.
//The port_aguids was filled in the AGuid stage, in case that AGuid stage was
//skipped, the duplicated vport guids check will not include ports alias guids.
int IBDiag::CheckVPortDuplicatedGuids(list_p_fabric_general_err& vguid_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    //init the port dictionary with the ports from the discovery
    //then add only alias guids to it
    map_guid_pport phys_port_guids = this->discovered_fabric.PortByGuid;
    map_guid_pnode::iterator guid2NodeEnd =
        this->discovered_fabric.NodeByGuid.end();
    map_guid_pnode::iterator sysguid2NodeEnd =
        this->discovered_fabric.NodeBySystemGuid.end();
    map_guid_pvport vports_guids;
    map_guid_pport port_guids_to_check;
    u_int64_t vport_guid, phys_guid;
    IBVPort *p_vport;
    IBPort *phys_port;

    //if the alias guids stage wasn't skipped, we check against the map
    //that includes the phys ports and alias guids.
    //else we check only against the phys port guids.
    if (!this->port_aguids.empty())
        port_guids_to_check = this->port_aguids;
    else
        port_guids_to_check = this->discovered_fabric.PortByGuid;

    // go over all the phys ports and looks into their vports and check duplicated guids.
    // first check duplicated guids against other vports, second check duplicated guids
    // against phys ports and their alias guids, third check against system guids,
    // fourth check against node guids.
    for (map_guid_pport::iterator pI = phys_port_guids.begin();
            pI != phys_port_guids.end(); ++pI) {
        phys_port = pI->second;
        if (!phys_port || phys_port->p_node->type != IB_CA_NODE)
            continue;
        map_vportnum_vport phys_port_vports = phys_port->VPorts;
        for (map_vportnum_vport::iterator vpI = phys_port_vports.begin();
                vpI != phys_port_vports.end(); ++vpI) {
            p_vport = vpI->second;
            if (!p_vport)
                continue;
            vport_guid = p_vport->guid_get();
            FabricErrVPort *p_error;
            map_guid_pvport::iterator findVPort =
                vports_guids.find(vport_guid);
            // check if we have duplicated vport guid
            if (findVPort != vports_guids.end()) {
                p_error = new FabricErrVPortGuidDuplicated(
                        p_vport,
                        findVPort->second,
                        vport_guid);
                if (!p_error) {
                    this->SetLastError("Failed to allocate"  \
                                       " FabricErrVPortGuidDuplicated");
                    IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
                }
                vguid_errors.push_back(p_error);
                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            } else
                //add this vport guid to maps, only if didn't exist before
                vports_guids.insert(pair_vguid_vport(vport_guid, p_vport));

            // check duplicate vport guid in port(including alias, but only if
            // alias guids stage wasn't skipped) guids.
            // special case for vport0, need only to check that guid is as the
            // pyhs port guid.
            if (p_vport->getVPortNum() == 0) {
                phys_guid = phys_port->guid_get();
                if (!phys_guid)
                    continue;
                if (vport_guid != phys_guid) {
                    FabricErrVPortGUIDInvalidFirstEntry *p_err =
                        new FabricErrVPortGUIDInvalidFirstEntry(phys_port,
                                                                p_vport,
                                                                vport_guid);
                    if (!p_err) {
                        this->SetLastError("Failed to allocate"  \
                                           " FabricErrVPortInvalidFirstEntry");
                        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
                    }
                    vguid_errors.push_back(p_err);
                    rc = IBDIAG_ERR_CODE_CHECK_FAILED;
                }
                // Only need to verify that vport0 guid = physical port guid.
                // So no need to check vport guid0 against node guid,
                // system guid, since the physical guids duplications are
                // checked during discovery.
                // And since physical port guid = alias guid 0, vport guid 0
                // check against alias guids is covered in alias duplicate guids
                // check.
                continue;
            } else {
                // check vport guids against the phys port alias guids.
                map_guid_pport::iterator findVGuid =
                    port_guids_to_check.find(vport_guid);
                if (findVGuid != port_guids_to_check.end()) {
                    p_error = new FabricErrVPortGuidPGUIDDuplicated(
                            p_vport,
                            findVGuid->second,
                            vport_guid);
                    if (!p_error) {
                        this->SetLastError("Failed to allocate "  \
                                           "FabricErrVPortGuidPGUIDDuplicated");
                        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
                    }
                    vguid_errors.push_back(p_error);
                    rc = IBDIAG_ERR_CODE_CHECK_FAILED;
                }
            }
            // check vport guids against the node system guids.
            map_guid_pnode::iterator findSys =
                this->discovered_fabric.NodeBySystemGuid.find(vport_guid);
            if (findSys != sysguid2NodeEnd) {
                p_error = new FabricErrVPortSysGuidDuplicated(
                        p_vport,
                        findSys->second->p_system,
                        vport_guid);
                if (!p_error) {
                    this->SetLastError("Failed to allocate "  \
                                        "FabricErrVPortSysGuidDuplicated");
                    IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
                }
                vguid_errors.push_back(p_error);
                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            }
            // check vport guids against the node guids.
            map_guid_pnode::iterator findNode =
                this->discovered_fabric.NodeByGuid.find(vport_guid);
            if (findNode != guid2NodeEnd) {
                p_error = new FabricErrVPortNodeGuidDuplicated(
                        p_vport,
                        findNode->second,
                        vport_guid);
                if (!p_error) {
                    this->SetLastError("Failed to allocate "  \
                                       "FabricErrVPortNodeGuidDuplicated");
                    IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
                }
                vguid_errors.push_back(p_error);
                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            }
        }//vports loop
    }//phys ports loop
    IBDIAG_RETURN(rc);
}



//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_IO_ERR
int IBDiag::WriteVPortsPKeyTblFile(const char *file_name)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);
    ofstream sout;
    int rc = this->OpenFile(file_name, sout,
                            false, //no append
                            true); //add_header
    if (rc)
        IBDIAG_RETURN(rc);
    this->DumpVPortsPKeyTbl(sout);
    sout.close();
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


