/*
 * © Copyright (c) 2004-2018 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include "infiniband/ibdiag/ibdiag.h"
#include "infiniband/ibdiag/ibdiag_clbck.h"
#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h>
#include <arpa/inet.h>

extern IBDiagClbck ibDiagClbck;

int AddTreeIDToQPNList(map_qpn_to_treeid &sharp_an_qpns_to_treeid,
                       u_int32_t qpn,
                       u_int16_t tree_index)
{
    IBDIAG_ENTER;

    map_qpn_to_treeid::iterator it = sharp_an_qpns_to_treeid.find(qpn);
    if (it == sharp_an_qpns_to_treeid.end())
        sharp_an_qpns_to_treeid.insert(pair<u_int32_t, u_int16_t >(qpn, tree_index));
    else
        IBDIAG_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

/********SharpTreeEdge***********/
SharpTreeEdge::SharpTreeEdge(u_int32_t qp_num, u_int8_t qp_idx):
                             m_remote_tree_node(NULL), m_qpn(qp_num),
                             m_child_idx(qp_idx), m_qpc_config()
{
    IBDIAG_ENTER;
    IBDIAG_RETURN_VOID;
}

/********SharpTreeNode***********/
SharpTreeNode::SharpTreeNode(SharpAggNode *aggNode, u_int16_t treeid):
                             m_tree_id(treeid), m_child_idx(0), m_agg_node(aggNode),
                             m_parent(NULL)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN_VOID;
}

int SharpTreeNode::AddSharpTreeEdge(SharpTreeEdge *p_sharp_tree_eage,
                                    u_int8_t db_index)
{
    IBDIAG_ENTER;

    if (this->m_children.empty() || ((u_int8_t)this->m_children.size() < db_index + 1))
        this->m_children.resize(db_index + 1, NULL);

    if (!this->m_children[db_index]) {
        this->m_children[db_index] = p_sharp_tree_eage;
    }
    //TODO add error for inserting in bad child index

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

SharpTreeEdge * SharpTreeNode::GetSharpTreeEdge(u_int8_t db_index)
{
    IBDIAG_ENTER;
    //not exists
    if ((u_int8_t)m_children.size() < db_index + 1)
        IBDIAG_RETURN(NULL);
    IBDIAG_RETURN(m_children[db_index]);
}


SharpTree::SharpTree(SharpTreeNode *root):
                     m_root(root), m_max_radix(0)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN_VOID;
}

/********SharpAggNode***********/
SharpAggNode::SharpAggNode(IBPort * port):
                           m_port(port), m_an_info(), m_perf_cntr()
{
    IBDIAG_ENTER;
    IBDIAG_RETURN_VOID;
}

SharpAggNode::~SharpAggNode()
{
    IBDIAG_ENTER;
    for (vector_p_sharp_tree_node::iterator it = this->m_trees.begin();
         it != this->m_trees.end();
         it++) {
       SharpTreeNode * p_sharp_tree_node = (*it);
       delete p_sharp_tree_node;
    }
    this->m_trees.clear();

    IBDIAG_RETURN_VOID;
}

void SharpAggNode::SetANInfo(struct AM_ANInfo *an_info)
{
    IBDIAG_ENTER;

    this->m_an_info = *an_info;
    this->m_trees.resize(this->m_an_info.tree_table_size, NULL);

    IBDIAG_RETURN_VOID;
}

int SharpAggNode::AddSharpTreeNode(SharpTreeNode *p_sharp_tree_node,
                                   u_int16_t tree_index)
{
    IBDIAG_ENTER;

    if (this->m_trees.empty() || ((u_int16_t)this->m_trees.size() < tree_index + 1))
        this->m_trees.resize(tree_index + 1, NULL);

    if (!this->m_trees[tree_index]) {
        this->m_trees[tree_index] = p_sharp_tree_node;
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

SharpTreeNode * SharpAggNode::GetSharpTreeNode(u_int16_t tree_index)
{
    IBDIAG_ENTER;
    //not exists
    if ((u_int16_t)this->m_trees.size() < tree_index + 1)
        IBDIAG_RETURN(NULL);
    IBDIAG_RETURN(this->m_trees[tree_index]);
}

/********SharpMngr***********/
SharpMngr::SharpMngr(IBDiag * ibdiag):
                     m_fabric_max_trees_idx(0), m_ibdiag(ibdiag)
{
    IBDIAG_ENTER;
    ibDiagClbck.SetSharpMngr(this);
    IBDIAG_RETURN_VOID;
}

int SharpMngr::AddTreeRoot(u_int16_t tree_id,
                            SharpTreeNode *p_sharp_tree_node)
{
    IBDIAG_ENTER;

    if (this->m_sharp_root_nodes.empty() || ((u_int16_t)this->m_sharp_root_nodes.size() < tree_id + 1))
        this->m_sharp_root_nodes.resize(tree_id + 1, NULL);

    if (!this->m_sharp_root_nodes[tree_id]) {
        //make SharpTreeRoot
        SharpTree * p_sharp_tree = new SharpTree(p_sharp_tree_node);
        this->m_sharp_root_nodes[tree_id] = p_sharp_tree;
    } else
        IBDIAG_RETURN(1);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

SharpTree * SharpMngr::GetTree(u_int16_t tree_id)
{
    IBDIAG_ENTER;

    //not exists
    if ((u_int16_t)m_sharp_root_nodes.size() < tree_id + 1)
        IBDIAG_RETURN(NULL);
    IBDIAG_RETURN(m_sharp_root_nodes[tree_id]);
}

int  SharpMngr::DiscoverSharpAggNodes(
        list_p_fabric_general_err &sharp_discovery_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    progress_bar_nodes_t progress_bar_nodes;
    struct IB_ClassPortInfo class_port_info;
    CLEAR_STRUCT(class_port_info);
    CLEAR_STRUCT(progress_bar_nodes);

    clbck_data_t clbck_data;
    CLEAR_STRUCT(clbck_data);
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = IBDiagAggMngrClassPortInfoClbck;

    for (map_str_pnode::iterator nI = m_ibdiag->GetDiscoverFabricPtr()->NodeByName.begin();
            nI != m_ibdiag->GetDiscoverFabricPtr()->NodeByName.end();
            ++nI) {
        IBNode *p_node = (*nI).second;
        if (!p_node) {
            m_ibdiag->SetLastError("DB error - found null node in NodeByName map"   \
                                   " for node = %s", p_node->getName().c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
        }

        // AM relevant only for HCAs, we ignore SWs
        if (p_node->type == IB_SW_NODE)
            continue;

        ++progress_bar_nodes.nodes_found;
        ++progress_bar_nodes.ca_found;
        progress_bar_retrieve_from_nodes(&progress_bar_nodes,
            m_ibdiag->GetDiscoverProgressBarNodesPtr(), "SHARPAggMngrClassPortInfo");

        // go over all base lids of this node:
        for (u_int32_t i = 1; i < (u_int8_t)(p_node->numPorts + 1); ++i) {
            IBPort *p_port = p_node->getPort((phys_port_t)i);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_port || (i != 0 && !p_port->is_data_worthy()))
                continue;

            clbck_data.m_data1 = p_node;
            m_ibdiag->GetIbisPtr()->AMClassPortInfoGet(
                p_port->base_lid, DEFAULT_SL, &class_port_info, &clbck_data);
            break;
        }
        if(ibDiagClbck.GetState())
            goto exit;
    }
exit:
    m_ibdiag->GetIbisPtr()->MadRecAll();

    if (rc) {
        if (m_ibdiag->IsLastErrorEmpty())
            m_ibdiag->SetLastError("DiscoverSharpAggNodes Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "DiscoverSharpAggNodes Failed. \n");
    }

    IBDIAG_RETURN(rc);
}

int SharpMngr::BuildTreeConfigDB(
        list_p_fabric_general_err &sharp_discovery_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    progress_bar_nodes_t progress_bar_nodes;
    struct AM_TreeConfig tree_config;
    CLEAR_STRUCT(tree_config);
    CLEAR_STRUCT(progress_bar_nodes);

    clbck_data_t clbck_data;
    CLEAR_STRUCT(clbck_data);
    clbck_data.m_p_obj = &ibDiagClbck;

    //for every TreeID send TreeConfig
    for (u_int16_t tree_id = 0; tree_id < MAX_TREE_ID; tree_id++) {
        for (list_sharp_an::iterator nI = this->m_sharp_an.begin();
             nI != this->m_sharp_an.end();
             ++nI) {

            SharpAggNode *p_sharp_aggnode = *nI;
            if (!p_sharp_aggnode) {
                m_ibdiag->SetLastError("DB error - found null Aggregation node in sharp_am_nodes");
                rc = IBDIAG_ERR_CODE_DB_ERR;
                goto exit;
            }

            if (!tree_id) {
                ++progress_bar_nodes.nodes_found;
                ++progress_bar_nodes.ca_found;

                progress_bar_retrieve_from_nodes(&progress_bar_nodes,
                    m_ibdiag->GetDiscoverProgressBarNodesPtr(), "BuildTreeConfigDB");
            }
            clbck_data.m_handle_data_func = IBDiagSharpMngrTreeConfigClbck;

            clbck_data.m_data1 = p_sharp_aggnode;

            clbck_data.m_data2 = (void *)(uintptr_t)tree_id;
            tree_config.tree_id = tree_id;
            tree_config.num_of_children = MAX_CHILD_IDX_IN_TREE_CONFIG_MAD;

            m_ibdiag->GetIbisPtr()->AMTreeConfigGet(
                p_sharp_aggnode->GetIBPort()->base_lid, DEFAULT_SL,
                DEFAULT_AM_KEY, DEFAULT_AM_CLASS_VERSION, &tree_config, &clbck_data);

            if(ibDiagClbck.GetState())
                goto exit;
        }
        if(ibDiagClbck.GetState())
           goto exit;
    }
exit:
    m_ibdiag->GetIbisPtr()->MadRecAll();

    if (rc) {
        if (m_ibdiag->IsLastErrorEmpty())
            m_ibdiag->SetLastError("BuildSHARPAggMngrTreeConfig Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "BuildSHARPAggMngrTreeConfig Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            m_ibdiag->SetLastError(ibDiagClbck.GetLastError());
        else if (ibDiagClbck.GetNumErrors() && !sharp_discovery_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int  SharpMngr::BuildQPCConfigDB(
        list_p_fabric_general_err &sharp_discovery_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    progress_bar_nodes_t progress_bar_nodes;
    struct AM_QPCConfig qpc_config;
    CLEAR_STRUCT(qpc_config);
    CLEAR_STRUCT(progress_bar_nodes);

    clbck_data_t clbck_data;
    CLEAR_STRUCT(clbck_data);
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = IBDiagSharpMngrQPCConfigClbck;

    for (u_int8_t tree_id = 0; tree_id <= this->m_fabric_max_trees_idx; tree_id++) {
        for (list_sharp_an::iterator nI = this->m_sharp_an.begin();
             nI != this->m_sharp_an.end();
             ++nI) {

            SharpAggNode *p_sharp_agg_node = *nI;
            if (!p_sharp_agg_node) {
                m_ibdiag->SetLastError("DB error - found null Aggregation node in sharp_am_nodes");
                rc = IBDIAG_ERR_CODE_DB_ERR;
                goto exit;
            }

            IBNode * p_node = p_sharp_agg_node->GetIBPort()->p_node;
            SharpTreeNode * p_sharp_tree_node = p_sharp_agg_node->GetSharpTreeNode(tree_id);
            if (!p_sharp_tree_node)
                continue;

            if (p_node->type == IB_SW_NODE)
                continue;

            if (!tree_id) {
                ++progress_bar_nodes.nodes_found;
                ++progress_bar_nodes.ca_found;
                progress_bar_retrieve_from_nodes(&progress_bar_nodes,
                    m_ibdiag->GetDiscoverProgressBarNodesPtr(), "SHARPAggMngrQPCConfig");
            }

            for (u_int32_t i = 1; i < (u_int8_t)(p_node->numPorts + 1); ++i) {
                IBPort *p_port = p_node->getPort((phys_port_t)i);
                //Skip ports that don't exist or are down or have undefined state
                //or are not in the scope of the fabric.
                if (!p_port || (i != 0 && !p_port->is_data_worthy()))
                    continue;

                clbck_data.m_data1 = p_sharp_agg_node;

                SharpTreeEdge * p_edge_to_parent = p_sharp_tree_node->GetSharpParentTreeEdge();
                if (p_edge_to_parent) {
                    clbck_data.m_data2 = p_edge_to_parent;
                    qpc_config.qpn = p_edge_to_parent->GetQpn();

                    m_ibdiag->GetIbisPtr()->AMQPCConfigGet(
                            p_port->base_lid, DEFAULT_SL, DEFAULT_AM_KEY,
                            DEFAULT_AM_CLASS_VERSION, &qpc_config, &clbck_data);

                    if(ibDiagClbck.GetState())
                        goto exit;
                }

                for (u_int8_t j = 0; j < p_sharp_tree_node->GetChildrenSize(); j++) {

                    SharpTreeEdge * p_sharp_tree_edge = p_sharp_tree_node->GetSharpTreeEdge(j);
                    if (!p_sharp_tree_edge) {
                        m_ibdiag->SetLastError("DB error - found null SharpTreeEdge in trees DB"    \
                                               " for child index = %d", j);
                        rc = IBDIAG_ERR_CODE_DB_ERR;
                        goto exit;
                    }

                    clbck_data.m_data2 = p_sharp_tree_edge;
                    qpc_config.qpn = p_sharp_tree_edge->GetQpn();

                    m_ibdiag->GetIbisPtr()->AMQPCConfigGet(p_port->base_lid, 0, 0,
                            DEFAULT_AM_CLASS_VERSION, &qpc_config, &clbck_data);

                    if(ibDiagClbck.GetState())
                        goto exit;
                }
                break;
            }
            if(ibDiagClbck.GetState())
                goto exit;
        }
    } // trees loop

exit:
    m_ibdiag->GetIbisPtr()->MadRecAll();

    if (rc) {
        if (m_ibdiag->IsLastErrorEmpty())
            m_ibdiag->SetLastError("BuildQPCConfigDB Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "BuildQPCConfigDB Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            m_ibdiag->SetLastError(ibDiagClbck.GetLastError());
        else if (ibDiagClbck.GetNumErrors() && !sharp_discovery_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int SharpMngr::BuildSharpConfigurationDB(
        list_p_fabric_general_err &sharp_discovery_errors,
        progress_func_nodes_t progress_func)
{

    if (!m_ibdiag->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    ibDiagClbck.Set(m_ibdiag, m_ibdiag->GetIBDMExtendedInfoPtr(), &sharp_discovery_errors);

    rc = DiscoverSharpAggNodes(sharp_discovery_errors);
    if (rc) {
        ERR_PRINT("Failed to build AM Nodes DB.\n");
        IBDIAG_RETURN(rc);
    }

    SCREEN_PRINT("\n");
    INFO_PRINT("Discovered %u Aggregation Nodes.\n",
                       (int)m_sharp_supported_nodes.size());

    for (vector_ibnodes::iterator nI = this->m_sharp_supported_nodes.begin();
        nI != this->m_sharp_supported_nodes.end();
        ++nI) {
        IBNode *p_node = *nI;
        if (!p_node) {
            m_ibdiag->SetLastError("DB error - found null node in NodeByName map"   \
                                   " for node = %s", p_node->getName().c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
        }
        // ibnodes is only HCAs
        for (u_int32_t i = 1; i < (u_int8_t)(p_node->numPorts + 1); ++i) {
            IBPort *p_port = p_node->getPort((phys_port_t)i);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_port || (i != 0 && !p_port->is_data_worthy()))
                continue;
            SharpAggNode *p_sharp_agg_node = new SharpAggNode(p_port);

            this->m_sharp_an.push_back(p_sharp_agg_node);
            m_lid_to_sharp_agg_node.insert(pair<lid_t, SharpAggNode *>(p_port->base_lid,
                                                             p_sharp_agg_node));
            break;
        }
    }

    rc = BuildANInfoDB(sharp_discovery_errors);
    if (rc) {
        ERR_PRINT("Failed to build AMInfo DB.\n");
        IBDIAG_RETURN(rc);
    }

    rc = BuildTreeConfigDB(sharp_discovery_errors);
    if (rc) {
        ERR_PRINT("Failed to build AM TreeConfig DB.\n");
        IBDIAG_RETURN(rc);
    }

    SCREEN_PRINT("\n");
    rc = BuildQPCConfigDB(sharp_discovery_errors);
    if (rc) {
        ERR_PRINT("Failed to build AM QPCConfig DB.\n");
        IBDIAG_RETURN(rc);
    }

    SCREEN_PRINT("\n");
    IBDIAG_RETURN(rc);
}

int  SharpMngr::BuildANInfoDB(
        list_p_fabric_general_err &sharp_discovery_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    progress_bar_nodes_t progress_bar_nodes;
    struct AM_ANInfo an_info;
    CLEAR_STRUCT(an_info);
    CLEAR_STRUCT(progress_bar_nodes);

    clbck_data_t clbck_data;
    CLEAR_STRUCT(clbck_data);
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = IBDiagSharpMngrANInfoClbck;

    for (list_sharp_an::iterator nI = this->m_sharp_an.begin();
         nI != this->m_sharp_an.end();
         ++nI) {

        //This cannot be NULL as it is locally created and inserted.
        SharpAggNode *p_sharp_aggnode = *nI;

        ++progress_bar_nodes.nodes_found;
        ++progress_bar_nodes.ca_found;
        progress_bar_retrieve_from_nodes(&progress_bar_nodes,
            m_ibdiag->GetDiscoverProgressBarNodesPtr(), "SHARPINInfo");

        IBPort *p_port = p_sharp_aggnode->GetIBPort();
        clbck_data.m_data1 = p_sharp_aggnode;

        m_ibdiag->GetIbisPtr()->AMANInfoGet(
                p_port->base_lid, DEFAULT_SL, DEFAULT_AM_KEY, DEFAULT_AM_CLASS_VERSION,
                &an_info, &clbck_data);

        if (ibDiagClbck.GetState())
            goto exit;
    }

exit:
    m_ibdiag->GetIbisPtr()->MadRecAll();

    if (rc) {
        if (m_ibdiag->IsLastErrorEmpty())
            m_ibdiag->SetLastError("BuildANInfoDB Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "BuildANInfoDB Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            m_ibdiag->SetLastError(ibDiagClbck.GetLastError());
        else if (ibDiagClbck.GetNumErrors() && !sharp_discovery_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int  SharpMngr::BuildPerformanceCountersDB(
        list_p_fabric_general_err &sharp_discovery_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    progress_bar_nodes_t progress_bar_nodes;
    struct AM_PerformanceCounters perf_cntr;
    CLEAR_STRUCT(perf_cntr);
    CLEAR_STRUCT(progress_bar_nodes);

    clbck_data_t clbck_data;
    CLEAR_STRUCT(clbck_data);
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = IBDiagSharpMngrPerfCountersClbck;

    for (list_sharp_an::iterator nI = this->m_sharp_an.begin();
         nI != this->m_sharp_an.end();
         ++nI) {

        SharpAggNode *p_sharp_aggnode = *nI;
        if (!p_sharp_aggnode) {
            m_ibdiag->SetLastError("DB error - found null Aggregation node in sharp_am_nodes");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        ++progress_bar_nodes.nodes_found;
        ++progress_bar_nodes.ca_found;
        progress_bar_retrieve_from_nodes(&progress_bar_nodes,
            m_ibdiag->GetDiscoverProgressBarNodesPtr(), "SHARPPerformanceCounterts");

        IBPort *p_port = p_sharp_aggnode->GetIBPort();
        clbck_data.m_data1 = p_sharp_aggnode;

        m_ibdiag->GetIbisPtr()->AMPerformanceCountersGet(
                p_port->base_lid, DEFAULT_SL, DEFAULT_AM_KEY, DEFAULT_AM_CLASS_VERSION,
                &perf_cntr, &clbck_data);

        if(ibDiagClbck.GetState())
            goto exit;
    }

exit:
    m_ibdiag->GetIbisPtr()->MadRecAll();

    if (rc) {
        if (m_ibdiag->IsLastErrorEmpty())
            m_ibdiag->SetLastError("BuildPerformanceCountersDB Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "BuildPerformanceCountersDB Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            m_ibdiag->SetLastError(ibDiagClbck.GetLastError());
        else if (ibDiagClbck.GetNumErrors() && !sharp_discovery_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int  SharpMngr::ResetPerformanceCounters(
        list_p_fabric_general_err &sharp_discovery_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    progress_bar_nodes_t progress_bar_nodes;
    struct AM_PerformanceCounters perf_cntr;
    CLEAR_STRUCT(perf_cntr);
    CLEAR_STRUCT(progress_bar_nodes);

    clbck_data_t clbck_data;
    CLEAR_STRUCT(clbck_data);
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = IBDiagSharpMngrResetPerfCountersClbck;

    for (list_sharp_an::iterator nI = this->m_sharp_an.begin();
         nI != this->m_sharp_an.end();
         ++nI) {

        SharpAggNode *p_sharp_aggnode = *nI;
        if (!p_sharp_aggnode) {
            m_ibdiag->SetLastError("DB error - found null Aggregation node in sharp_am_nodes");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        ++progress_bar_nodes.nodes_found;
        ++progress_bar_nodes.ca_found;
        progress_bar_retrieve_from_nodes(&progress_bar_nodes,
            m_ibdiag->GetDiscoverProgressBarNodesPtr(), "SHARPPerformanceCounterts");

        IBPort *p_port = p_sharp_aggnode->GetIBPort();
        clbck_data.m_data1 = p_sharp_aggnode;
        perf_cntr.counter_select = ALL_COUNTERS_SELECT;

        m_ibdiag->GetIbisPtr()->AMPerformanceCountersSet(
                p_port->base_lid, DEFAULT_SL, DEFAULT_AM_KEY, DEFAULT_AM_CLASS_VERSION,
                &perf_cntr, &clbck_data);

        if(ibDiagClbck.GetState())
            goto exit;
    }

exit:
    m_ibdiag->GetIbisPtr()->MadRecAll();

    if (rc) {
        if (m_ibdiag->IsLastErrorEmpty())
            m_ibdiag->SetLastError("ResetPerformanceCounters Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "ResetPerformanceCounters Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            m_ibdiag->SetLastError(ibDiagClbck.GetLastError());
        else if (!sharp_discovery_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}
int SharpMngr::ConnectTreeEdges(list_p_fabric_general_err &sharp_discovery_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    for (list_sharp_an::iterator nI = this->m_sharp_an.begin();
            nI != this->m_sharp_an.end();
            ++nI) {

        SharpAggNode *p_sharp_agg_node = *nI;
        if (!p_sharp_agg_node) {
            m_ibdiag->SetLastError("DB error - found null Aggregation node in sharp_am_nodes");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        IBNode * p_node = p_sharp_agg_node->GetIBPort()->p_node;

        for (u_int16_t tree_idx = 0; tree_idx < p_sharp_agg_node->GetTreesSize(); tree_idx++) {

            SharpTreeNode * p_sharp_tree_node = p_sharp_agg_node->GetSharpTreeNode(tree_idx);
            if (!p_sharp_tree_node)
                continue;

            SharpTree * p_root = GetTree(tree_idx);
            if (p_root &&
                (p_root->GetMaxRadix() < p_sharp_tree_node->GetChildrenSize()))
                p_root->SetMaxRadix(p_sharp_tree_node->GetChildrenSize());

            for (u_int8_t db_idx = 0;
                 db_idx < p_sharp_tree_node->GetChildrenSize();
                 db_idx++) {

                SharpTreeEdge * p_sharp_tree_edge = p_sharp_tree_node->GetSharpTreeEdge(db_idx);
                if (!p_sharp_tree_edge)
                    continue;

                map_lid_to_sharpagg_node::iterator it =
                    m_lid_to_sharp_agg_node.find(p_sharp_tree_edge->GetQPCConfig().rlid);
                if (it == m_lid_to_sharp_agg_node.end()) {
                   IBPort *p_port = m_ibdiag->GetDiscoverFabricPtr()->
                       getPortByLid(p_sharp_tree_edge->GetQPCConfig().rlid);
                   if (!p_port || p_port->p_node->type == IB_SW_NODE) {
                       SharpErrEdgeNodeNotFound *p_curr_fabric_node_err =
                           new SharpErrEdgeNodeNotFound(
                                   p_node, p_sharp_tree_edge->GetQPCConfig().rlid);
                       if (!p_curr_fabric_node_err) {
                           m_ibdiag->SetLastError("Failed to allocate"  \
                                   " SharpErrEdgeNodeNotFound");
                           rc = IBDIAG_ERR_CODE_NO_MEM;
                       }
                       p_curr_fabric_node_err->SetLevel(EN_FABRIC_ERR_WARNING);
                       sharp_discovery_errors.push_back(p_curr_fabric_node_err);
                   }
                   continue;
                }

                SharpAggNode * remote_p_sharp_agg_node = it->second;
                if (!remote_p_sharp_agg_node) {
                    m_ibdiag->SetLastError("DB error - found null SharpAggNode"    \
                            " for lid = %d", p_sharp_tree_edge->GetQPCConfig().rlid);
                    rc = IBDIAG_ERR_CODE_DB_ERR;
                    IBDIAG_RETURN(rc);
                }
                SharpTreeNode * remote_p_sharp_tree_node =
                    remote_p_sharp_agg_node->GetSharpTreeNode(tree_idx);
                if (remote_p_sharp_tree_node) {

                   p_sharp_tree_edge->SetRemoteTreeNode(remote_p_sharp_tree_node);
                   remote_p_sharp_tree_node->SetChildIdx(p_sharp_tree_edge->GetChildIdx());
                   if (remote_p_sharp_tree_node->GetSharpParentTreeEdge())
                       remote_p_sharp_tree_node->GetSharpParentTreeEdge()->
                           SetRemoteTreeNode(p_sharp_tree_node);
                }
                //TODO add warning on remote SharpTreeNode not found for T
            }
        }
    }

exit:
    IBDIAG_RETURN(rc);
}

/*
 1. checking disconnected node
 2. checking duplicate qpn in sharp node
 3. checking rqpn is valid

TODO: check loops in the tree for nodes on the same tree_id
 */
int SharpMngr::CheckSharpTrees(list_p_fabric_general_err &sharp_discovery_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    // checking disconnected nodes on the tree that are not root
    for (list_sharp_an::iterator nI = this->m_sharp_an.begin();
         nI != this->m_sharp_an.end();
         ++nI) {

        SharpAggNode *p_sharp_agg_node = *nI;
        if (!p_sharp_agg_node) {
            m_ibdiag->SetLastError("DB error - found null Aggregation node in sharp_am_nodes");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        IBNode * p_node = p_sharp_agg_node->GetIBPort()->p_node;
        map_qpn_to_treeid sharp_an_qpns_to_treeid;

        for (u_int16_t tree_idx = 0; tree_idx < p_sharp_agg_node->GetTreesSize(); tree_idx++) {

            SharpTreeNode * p_sharp_tree_node = p_sharp_agg_node->GetSharpTreeNode(tree_idx);
            if (!p_sharp_tree_node)
                continue;
            SharpTreeEdge * p_parent_tree_edge =
                p_sharp_tree_node->GetSharpParentTreeEdge();

            //if not root and remote parent is null
            if (p_parent_tree_edge &&
                p_parent_tree_edge->GetQpn() &&
                !p_parent_tree_edge->GetRemoteTreeNode()) {
                SharpErrDisconnectedTreeNode *p_curr_fabric_node_err =
                    new SharpErrDisconnectedTreeNode(
                            p_node, tree_idx);
                if (!p_curr_fabric_node_err) {
                    m_ibdiag->SetLastError("Failed to allocate"  \
                            " SharpErrDisconnectedTreeNode");
                    rc = IBDIAG_ERR_CODE_NO_MEM;
                }
                p_curr_fabric_node_err->SetLevel(EN_FABRIC_ERR_WARNING);
                sharp_discovery_errors.push_back(p_curr_fabric_node_err);
            }

            // Check QP state
            if (p_parent_tree_edge &&
                p_parent_tree_edge->GetQPCConfig().state != QP_STATE_ACTIVE) {
                SharpErrQPNotActive *p_curr_fabric_node_err =
                    new SharpErrQPNotActive(
                            p_node, p_parent_tree_edge->GetQpn(),
                            p_parent_tree_edge->GetQPCConfig().state);
                if (!p_curr_fabric_node_err) {
                    m_ibdiag->SetLastError("Failed to allocate"  \
                            " SharpErrQPNotActive");
                    rc = IBDIAG_ERR_CODE_NO_MEM;
                }
                p_curr_fabric_node_err->SetLevel(EN_FABRIC_ERR_WARNING);
                sharp_discovery_errors.push_back(p_curr_fabric_node_err);
            }

            //checking duplicate qpn in sharp node
            for (u_int8_t child_idx = 0;
                 child_idx < p_sharp_tree_node->GetChildrenSize();
                 child_idx++) {

                SharpTreeEdge * p_sharp_tree_edge =
                    p_sharp_tree_node->GetSharpTreeEdge(child_idx);
                if (!p_sharp_tree_edge)
                    continue;

                if (AddTreeIDToQPNList(sharp_an_qpns_to_treeid,
                                       p_sharp_tree_edge->GetQpn(),
                                       tree_idx)) {
                    SharpErrDuplicatedQPNForAggNode *p_curr_fabric_node_err =
                        new SharpErrDuplicatedQPNForAggNode(
                                p_node, tree_idx,
                                sharp_an_qpns_to_treeid[p_sharp_tree_edge->GetQpn()],
                                p_sharp_tree_edge->GetQpn());
                    if (!p_curr_fabric_node_err) {
                        m_ibdiag->SetLastError("Failed to allocate"  \
                                " SharpErrDuplicatedQPNForAggNode");
                        rc = IBDIAG_ERR_CODE_NO_MEM;
                    }
                    sharp_discovery_errors.push_back(p_curr_fabric_node_err);
                }

                // Check QP state
                if (p_sharp_tree_edge->GetQPCConfig().state != QP_STATE_ACTIVE) {
                    SharpErrQPNotActive *p_curr_fabric_node_err =
                        new SharpErrQPNotActive(
                                p_node, p_sharp_tree_edge->GetQpn(),
                                p_sharp_tree_edge->GetQPCConfig().state);
                    if (!p_curr_fabric_node_err) {
                        m_ibdiag->SetLastError("Failed to allocate"  \
                                " SharpErrQPNotActive");
                        rc = IBDIAG_ERR_CODE_NO_MEM;
                    }
                    sharp_discovery_errors.push_back(p_curr_fabric_node_err);
                }

                // checking rqpn is valid
                if (p_sharp_tree_edge->GetRemoteTreeNode() &&
                    p_sharp_tree_edge->GetRemoteTreeNode()->GetSharpParentTreeEdge()) {

                    u_int32_t rqpn = p_sharp_tree_edge->GetQPCConfig().rqpn;
                    u_int32_t peer_qpn =
                        p_sharp_tree_edge->GetRemoteTreeNode()->GetSharpParentTreeEdge()->GetQpn();
                    if (rqpn != peer_qpn) {
                        SharpErrRQPNotValid *p_curr_fabric_node_err =
                            new SharpErrRQPNotValid(
                                    p_node, rqpn, peer_qpn);
                        if (!p_curr_fabric_node_err) {
                            m_ibdiag->SetLastError("Failed to allocate"  \
                                    " SharpErrRQPNotValid");
                            rc = IBDIAG_ERR_CODE_NO_MEM;
                        }
                        sharp_discovery_errors.push_back(p_curr_fabric_node_err);
                    }

                    u_int32_t qpn = p_sharp_tree_edge->GetQpn();
                    u_int32_t peer_rqpn = p_sharp_tree_edge->
                        GetRemoteTreeNode()->GetSharpParentTreeEdge()->GetQPCConfig().rqpn;
                    if (peer_rqpn != qpn) {
                        SharpErrRQPNotValid *p_curr_fabric_node_err =
                            new SharpErrRQPNotValid(
                                    p_node, qpn, peer_rqpn);
                        if (!p_curr_fabric_node_err) {
                            m_ibdiag->SetLastError("Failed to allocate"  \
                                    " SharpErrRQPNotValid");
                            rc = IBDIAG_ERR_CODE_NO_MEM;
                        }
                        sharp_discovery_errors.push_back(p_curr_fabric_node_err);
                    }
                }
            }
        }

    }

exit:
    IBDIAG_RETURN(rc);
}

void SharpTreeNode::DumpTree(int indent_level, ofstream &sout)
{
    IBDIAG_ENTER;
    char curr_counters_line[256];
    string indent_str = "";
    int radix, parent_qpn = 0, remote_parent_qpn = 0;

    if (!this->m_agg_node || !this->m_agg_node->GetIBPort() || !this->m_agg_node->GetIBPort()->p_node)
        //TODO add error
        IBDIAG_RETURN_VOID;

    for (int indent = 0; indent < indent_level; indent++)
        indent_str.append("\t");

    IBPort *p_curr_port = this->m_agg_node->GetIBPort();
    IBNode *p_curr_node = p_curr_port->p_node;
    sout << indent_str;
    memset(curr_counters_line, 0, sizeof(curr_counters_line));
    sprintf(curr_counters_line,
            "(" U32D_FMT "), AN:" STR_FMT ", lid:" U32D_FMT ", port guid:" U64H_FMT
            ", Child index:" U32D_FMT,
            indent_level,
            p_curr_node->description.c_str(),
            this->m_agg_node->GetIBPort()->base_lid,
            p_curr_port->guid_get(),
            this->m_child_idx);
    sout << curr_counters_line;

    //TODO no need for p_parent_sharp_tree_edge pointer just use this->m_parent
    SharpTreeEdge * p_parent_sharp_tree_edge = this->m_parent;
    if (this->m_parent) {
        parent_qpn = p_parent_sharp_tree_edge->GetQPCConfig().qpn;
        remote_parent_qpn = p_parent_sharp_tree_edge->GetQPCConfig().rqpn;
    }

    radix = GetChildrenSize();
    sprintf(curr_counters_line,
            ", parent QPn:" U32H_FMT ", remote parent QPn:" U32H_FMT
            ", radix:" U32D_FMT,
            parent_qpn,
            remote_parent_qpn,
            radix);
    sout << curr_counters_line << endl;

    for (u_int8_t child_idx = 0; child_idx < this->GetChildrenSize(); ++child_idx) {
        SharpTreeEdge * p_sharp_tree_edge = this->GetSharpTreeEdge(child_idx);
        if (!p_sharp_tree_edge)
            continue;

        if (p_sharp_tree_edge->GetRemoteTreeNode())
            p_sharp_tree_edge->GetRemoteTreeNode()->DumpTree(indent_level+1, sout);
    }

    IBDIAG_RETURN_VOID;
}

int SharpMngr::WriteSharpFile(const char *file_name)
{
    IBDIAG_ENTER;

    ofstream sout;
    int rc = m_ibdiag->OpenFile(file_name, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);
    this->SharpMngrDumpAllTrees(sout);

    this->SharpMngrDumpAllQPs(sout);


    sout.close();
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int SharpMngr::WriteSharpPMFile(const char *file_name)
{
    IBDIAG_ENTER;

    ofstream sout;
    int rc = m_ibdiag->OpenFile(file_name, sout,
                                false, //no append
                                true); //add_header
    if (rc)
        IBDIAG_RETURN(rc);

    this->SharpMngrDumpCounters(sout);

    sout.close();
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void SharpMngr::SharpMngrDumpAllTrees(ofstream &sout)
{
    IBDIAG_ENTER;
    for (u_int16_t tree_idx = 0; tree_idx < this->m_sharp_root_nodes.size();
         ++tree_idx) {
        SharpTree * root_node = this->m_sharp_root_nodes[tree_idx];
        if (!root_node || !root_node->GetRoot())
            continue;

        sout << "TreeID:" << tree_idx << ", Max Radix:"
        << root_node->GetMaxRadix() << endl;
        root_node->GetRoot()->DumpTree(0, sout);
        sout << endl;
    }

    IBDIAG_RETURN_VOID;
}

void SharpMngr::DumpQPC(ofstream &sout, struct AM_QPCConfig *qpconfig)
{
    IBDIAG_ENTER;
    char curr_counters_line[256];
    char gid[INET6_ADDRSTRLEN];
    if (!qpconfig)
        IBDIAG_RETURN_VOID;

    memset(curr_counters_line, 0, sizeof(curr_counters_line));
    sprintf(curr_counters_line,
            "QPN:" U32H_FMT ", State:" U32D_FMT ", TS:" U32H_FMT
            ", G:" U32D_FMT ", SL:" U32D_FMT ", RLID:" U32D_FMT
            ", Traffic Class:" U32D_FMT ", Hop Limit:" U32D_FMT ", RGID:" STR_FMT
            ", RQ PSN:" U32D_FMT ", SQ PSN:" U32D_FMT ", PKey:" U32H_FMT
            ", RQPN:" U32H_FMT ", RNR Mode:" U32D_FMT ", RNR Retry Limit:" U32H_FMT
            ", Timeout Retry Limit:" U32D_FMT ", Local Ack Timeout:" U32D_FMT
            ,
            qpconfig->qpn,
            qpconfig->state,
            qpconfig->ts,
            qpconfig->g,
            qpconfig->sl,
            qpconfig->rlid,
            qpconfig->traffic_class,
            qpconfig->hop_limit,
            inet_ntop(AF_INET6, qpconfig->rgid, gid, (unsigned int)sizeof gid),
            qpconfig->rq_psn,
            qpconfig->sq_psn,
            qpconfig->pkey,
            qpconfig->rqpn,
            qpconfig->rnr_mode,
            qpconfig->rnr_retry_limit,
            qpconfig->timeout_retry_limit,
            qpconfig->local_ack_timeout
            );
    sout << curr_counters_line;

    IBDIAG_RETURN_VOID;
}

int SharpMngr::SharpMngrDumpAllQPs(ofstream &sout)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    char curr_counters_line[256];

    for (list_sharp_an::iterator nI = this->m_sharp_an.begin();
         nI != this->m_sharp_an.end();
         ++nI) {

        SharpAggNode *p_sharp_aggnode = *nI;
        if (!p_sharp_aggnode) {
            m_ibdiag->SetLastError("DB error - found null Aggregation node in sharp_am_nodes");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        IBNode * p_node = p_sharp_aggnode->GetIBPort()->p_node;

        map_qpn_to_qpc_config qpc_map;
        //for every TreeID send QPCConfig
        for (u_int8_t i = 0; i < p_sharp_aggnode->GetTreesSize(); i++) {

            SharpTreeNode * p_sharp_tree_node = p_sharp_aggnode->GetSharpTreeNode(i);
            if (!p_sharp_tree_node)
                continue;

            //TODO no need to use p_parent_sharp_tree_edge just use p_sharp_tree_node->GetSharpParentTreeEdge()
            SharpTreeEdge * p_parent_sharp_tree_edge = p_sharp_tree_node->GetSharpParentTreeEdge();
            if (p_parent_sharp_tree_edge) {

                p_parent_sharp_tree_edge->AddToQpcMap(qpc_map);
            }

            for (u_int8_t j = 0; j < p_sharp_tree_node->GetChildrenSize(); j++) {

                SharpTreeEdge * p_sharp_tree_edge = p_sharp_tree_node->GetSharpTreeEdge(j);
                if (!p_sharp_tree_edge) {
                    m_ibdiag->SetLastError("DB error - found null SharpTreeEdge in trees DB for child index = %d", j);
                    continue;
                }

                p_sharp_tree_edge->AddToQpcMap(qpc_map);
            }
        }

        memset(curr_counters_line, 0, sizeof(curr_counters_line));
        sprintf(curr_counters_line,
                "AN:" STR_FMT ", lid:" U32D_FMT ", node guid:" U64H_FMT,
                p_node->description.c_str(),
                p_sharp_aggnode->GetIBPort()->base_lid,
                p_node->guid_get());
        sout << endl << curr_counters_line << endl;

        for (map_qpn_to_qpc_config::iterator nI = qpc_map.begin();
             nI != qpc_map.end();
             ++nI) {

            u_int32_t qpn = (*nI).first;
            struct AM_QPCConfig * qpconfig = (*nI).second;
            if (!qpn || !qpconfig)
                continue;
            DumpQPC(sout, qpconfig);
            sout << endl;
        }
    }

exit:
    IBDIAG_RETURN(rc);
}

int SharpMngr::SharpMngrDumpCounters(ofstream &sout)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    char curr_counters_line[1024];

    for (list_sharp_an::iterator nI = this->m_sharp_an.begin();
         nI != this->m_sharp_an.end();
         ++nI) {

        SharpAggNode *p_sharp_aggnode = *nI;
        if (!p_sharp_aggnode) {
            m_ibdiag->SetLastError("DB error - found null Aggregation node in sharp_am_nodes");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        IBNode * p_node = p_sharp_aggnode->GetIBPort()->p_node;

        memset(curr_counters_line, 0, sizeof(curr_counters_line));
        sprintf(curr_counters_line,
                "AggNodeDesc=" STR_FMT " Lid=" U32D_FMT " GUID=" U64H_FMT,
                p_node->description.c_str(),
                p_sharp_aggnode->GetIBPort()->base_lid,
                p_node->guid_get());
        sout << "-------------------------------------------------------" << endl;
        sout << curr_counters_line << endl;
        sout << "-------------------------------------------------------" << endl;
        const AM_PerformanceCounters perf_cntr = p_sharp_aggnode->GetPerfCounters();
        memset(curr_counters_line, 0, sizeof(curr_counters_line));
        sprintf(curr_counters_line,
                "packet_sent=" U64H_FMT "\n"\
                "ack_packet_sent=" U64H_FMT "\n"\
                "retry_packet_sent=" U64H_FMT "\n"\
                "rnr_event=" U64H_FMT "\n"\
                "timeout_event=" U64H_FMT "\n"\
                "oos_nack_rcv=" U64H_FMT "\n"\
                "rnr_nack_rcv=" U64H_FMT "\n"\
                "packet_discard_transport=" U64H_FMT "\n"\
                "packet_discard_sharp=" U64H_FMT "\n",
                perf_cntr.packet_sent,
                perf_cntr.ack_packet_sent,
                perf_cntr.retry_packet_sent,
                perf_cntr.rnr_event,
                perf_cntr.timeout_event,
                perf_cntr.oos_nack_rcv,
                perf_cntr.rnr_nack_rcv,
                perf_cntr.packet_discard_transport,
                perf_cntr.packet_discard_sharp);
        sout << curr_counters_line << endl;
    }

exit:
    IBDIAG_RETURN(rc);
}

