/*
 *  Copyright (c) 2004-2019 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <string.h>
#include <unistd.h>
extern char *optarg;
extern int optind, opterr, optopt;
#include <getopt.h>
#include <errno.h>
#include <algorithm>
#include <limits.h>
#include <stdlib.h>

#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h>

#include "ibdiagapp.h"
#include "ibdiagnet_debug_wrap.h"

#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_version.h>

/****************************************************/
/* globals */
extern DebugWrap debug_wrap_opj;


/****************************************************/
/* some defines */
#define N_ELEMENTS(array) sizeof(array)/sizeof(array[0])

#define COMPILE_TIME_ASSERT(pred)            \
    switch(0){case 0:case pred:;}

/****************************************************/
/******************private methods*******************/
/****************************************************/
int IBDiagapp::ParseSkipList(const char *skip_list)
{
    IBDIAGNET_ENTER;
    char *pch1 = NULL;
    size_t i, skip_options_list_length = skip_options_list.size();
    u_int32_t skip_mask = 0;
    char *buff = new char[strlen(skip_list) + 1];
    if (!buff) {
        ERR_PRINT("Failed to allocate buff for skip list, exiting...\n");
        IBDIAGNET_RETURN(1);
    }
    strcpy(buff, skip_list);
    pch1 = strtok(buff, ",");
    while (pch1) {
        for (i = 0; i < skip_options_list_length; ++i)
            if (strcmp(pch1, skip_options_list[i].name.c_str()) == 0) {
                skip_mask |= skip_options_list[i].value;
                break;
            }
        if (i == skip_options_list_length) {            //unknown option
            delete [] buff;
            IBDIAGNET_RETURN(1);
        }
        pch1 = strtok(NULL, ",");
    }

    if (skip_mask & SKIP_DUP_GUIDS_VALUE) //skip duplicated guids detection
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->TurnOFFDUPGUIDsDetection();
    if (skip_mask & SKIP_DUP_NODE_DESC_VALUE)               //skip duplicated node description
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->TurnOFFDUPNodeDescDetection();
    if (skip_mask & SKIP_VS_CAP_SMP_VALUE)                 // skip vs SMP capability sub stage
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->TurnOFFVsCapSMP();
    if (skip_mask & SKIP_VS_CAP_GMP_VALUE)                 // skip vs GMP capability sub stage
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->TurnOFFVsCapGMP();
    if (skip_mask & SKIP_STAGE_LIDS_VALUE &&
        this->stages[STAGE_IDX_LIDS] != NULL)                  //skip lids stage
        ((LidsStage *)this->stages[STAGE_IDX_LIDS])->MarkStageAsSkipped();
    if (skip_mask & SKIP_STAGE_LINKS_VALUE &&
        this->stages[STAGE_IDX_LINKS] != NULL)                 //skip links stage
        ((LinksStage *)this->stages[STAGE_IDX_LINKS])->MarkStageAsSkipped();
    if (skip_mask & SKIP_STAGE_SM_VALUE &&
        this->stages[STAGE_IDX_SM] != NULL)                    //skip sm stage
        ((SMStage *)this->stages[STAGE_IDX_SM])->MarkStageAsSkipped();
    if (skip_mask & SKIP_STAGE_PM_VALUE &&
        this->stages[STAGE_IDX_PM] != NULL)                    //skip pm stage
        ((PMStage *)this->stages[STAGE_IDX_PM])->MarkStageAsSkipped();
    if (skip_mask & SKIP_STAGE_NODES_INFO_VALUE &&
        this->stages[STAGE_IDX_NODES_INFO] != NULL)            //skip nodes_info stage
        ((NodesInfoStage *)this->stages[STAGE_IDX_NODES_INFO])->MarkStageAsSkipped();
    if (skip_mask & SKIP_STAGE_SPEED_WIDTH_CHECK &&
        this->stages[STAGE_IDX_SPEED_WIDTH] != NULL)           //skip speed / width check stage
        ((Speed_WidthStage *)this->stages[STAGE_IDX_SPEED_WIDTH])->MarkStageAsSkipped();
    if (skip_mask & SKIP_STAGE_PKEY_CHECK &&
        this->stages[STAGE_IDX_PKEY] != NULL)                  //skip pkey stage
        ((PKeyStage *)this->stages[STAGE_IDX_PKEY])->MarkStageAsSkipped();
    if (skip_mask & SKIP_STAGE_AGUID_CHECK &&
        this->stages[STAGE_IDX_AGUID] != NULL)                 //skip aguid stage
        ((AGUIDStage *)this->stages[STAGE_IDX_AGUID])->MarkStageAsSkipped();
    if (skip_mask & SKIP_STAGE_TEMPERATURE_SENSING &&
        this->stages[STAGE_IDX_TEMPSENSING] != NULL)           //skip temp sensing stage
        ((TemperatureSensing *)this->stages[STAGE_IDX_TEMPSENSING])->MarkStageAsSkipped();
    if ((skip_mask & SKIP_STAGE_VIRTUALIZATION) &&
        this->stages[STAGE_IDX_VIRTUALIZATION] != NULL)        //skip virtualization stage
        ((VirtualizationStage *)this->stages[STAGE_IDX_VIRTUALIZATION])->MarkStageAsSkipped();

    delete [] buff;
    IBDIAGNET_RETURN(0);
}

int IBDiagapp::HandleMKeyParams()
{
    IBDIAGNET_ENTER;

    // parse mkey_val
    if (!mkey_val.empty()) {
        if (is_non_neg_num(mkey_val) || is_hex_num(mkey_val)) {
            ParseMKeyValue();
            if (!mkey_path.empty())
                WARN_PRINT("Ignore mkey_files path, mkey was given as constant value \n");
            goto exit;
        } else {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                           OPTION_MKEY, mkey_val.c_str());
            PRINT("    (Legal value: 64 bit mkey value)\n");
            IBDIAGNET_RETURN(1);
        }
    }

    // parse mkey_path
    if (ParseMKeyFiles())
        IBDIAGNET_RETURN(1);

exit:
    ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->SetMKeyManager(this->p_mkeymngr);
    IBDIAGNET_RETURN(0);
}

int IBDiagapp::ParseMKeyValue()
{
    IBDIAGNET_ENTER;

    uint64_t const_mkey = (u_int64_t)strtoull(mkey_val.c_str(), NULL, 0);
    LOG_PRINT("using constant mkey: " U64H_FMT "\n", const_mkey);

    p_mkeymngr = new FilesBasedMKeyManager();
    p_mkeymngr->setConstMKey(const_mkey);

    ibdiag_obj.GetIbisPtr()->SetMKeyManager(p_mkeymngr);
    IBDIAGNET_RETURN(0);
}

int IBDiagapp::ParseMKeyFiles()
{
    IBDIAGNET_ENTER;

    uint64_t root_guid = 0;
    local_port_t local_ports_array[IBIS_MAX_LOCAL_PORTS];
    uint32_t local_ports_num = 0, choice = 0;

    p_mkeymngr = new FilesBasedMKeyManager();
    string neighbors_file_path = mkey_path + "/" NEIGHBORS_FILE_NAME;
    string guid2mkey_file_path = mkey_path + "/" GUID2MKEY_FILE_NAME;
    int res = p_mkeymngr->parseAndCreateMKeyFabric(neighbors_file_path,
                                                       guid2mkey_file_path);

    char *buffer = ibdmGetAndClearInternalLog();
    if (buffer) {
        PRINT("%s", buffer);
        free(buffer);
    }

    if (res) {
        ERR_PRINT("Can't parse the neighbors and guid2mkey files \n");
        IBDIAGNET_RETURN(1);
    }

    if (ibdiag_obj.GetIbisPtr() &&
        ibdiag_obj.GetIbisPtr()->
        GetAllLocalPortGUIDs(local_ports_array, &local_ports_num))
        IBDIAGNET_RETURN(1);

    choice = local_ports_num;
    for (u_int32_t i = 0; i < local_ports_num; ++i) {
        if (local_ports_array[i].logical_state == IB_PORT_STATE_ACTIVE) {
            choice = i;
            break;
        }
    }

    // no active local port was found
    if (choice == local_ports_num) {
        ERR_PRINT("MKeyManager failed to find local active port \n");
        IBDIAGNET_RETURN(1);
    }

    root_guid = ADB2C_BE64_TO_CPU(local_ports_array[choice].guid);
    res = p_mkeymngr->setMkeyManagerFabricTreeRoot(root_guid);
    if (res) {
        ERR_PRINT("MKeyManager failed to set fabric tree root \n");
        IBDIAGNET_RETURN(1);
    }

    ibdiag_obj.GetIbisPtr()->SetMKeyManager(p_mkeymngr);
    IBDIAGNET_RETURN(0);
}

string ltrim(const string& s)
{
    size_t start = s.find_first_not_of(WHITESPACE);
    return (start == string::npos) ? "" : s.substr(start);
}

string rtrim(const string& s)
{
    size_t end = s.find_last_not_of(WHITESPACE);
    return (end == string::npos) ? "" : s.substr(0, end + 1);
}

string trim(const string& s)
{
    return rtrim(ltrim(s));
}

int IBDiagapp::ParseConfigFile()
{
    IBDIAGNET_ENTER;

    ifstream fs(config_file.c_str());
    if (fs.fail()) {

        // if the user set config file in cmd line we fail
        if (find(p_main_parser->args.begin(), p_main_parser->args.end(), OPTION_CONFIG)
                != p_main_parser->args.end()) {
            ERR_PRINT("Failed to open configuration file: %s\n", config_file.c_str());
            IBDIAGNET_RETURN(1);
        }

        LOG_PRINT("Failed to open default configuration file: %s\n", DEFAULT_CONFIG_FILE);
        IBDIAGNET_RETURN(0);
    }

    char line[1024] = {'\0'};
    unsigned int lineNum = 0;
    list_p_command_line_req p_requesters_list = this->p_main_parser->GetRequestersList();
    list_p_command_line_req::iterator it_req;
    vec_option_t::iterator it_opt;

    while (fs.good()) {
        fs.getline(line, 1024);
        lineNum++;

        // skip empty lines and comments
        if (!line[0] || line[0] == '#')
            continue;

        string line_str(line);
        size_t equal = line_str.find('=');

        // skip lines with no '='
        if (equal == string::npos)
            continue;

        string name = trim(line_str.substr(0, equal));
        string val = trim(line_str.substr(equal + 1, string::npos));

        if (val.empty()) {
            ERR_PRINT("Failed to parse configuration file: %s, invalid value for flag %s, "
                      "value is empty, line: %u\n",
                      config_file.c_str(), name.c_str(), lineNum);
            IBDIAGNET_RETURN(1);
        }

        bool is_bool_flag = false;
        bool is_flag_exist = false;
        for (it_req = p_requesters_list.begin(); it_req != p_requesters_list.end(); ++it_req) {
             vec_option_t options_list = (*it_req)->GetOptions();
             for (it_opt = options_list.begin(); it_opt != options_list.end(); ++it_opt) {
                  if (it_opt->option_name == name) {
                      is_flag_exist = true;
                      if (it_opt->attributes & IBIS_CMD_BOOLEAN_FLAG)
                          is_bool_flag = true;
                      break;
                  }
             }
             if (is_flag_exist)
                 break;
        }

        if (it_req == p_requesters_list.end()) {
            ERR_PRINT("Failed to parse configuration file: %s, flag %s doesn't exist, line: %u\n",
                      config_file.c_str(), name.c_str(), lineNum);
            IBDIAGNET_RETURN(1);
        }

        if (is_bool_flag) {
            if (!(this->IsValidBoolValue(val))) {
                ERR_PRINT("Failed to parse configuration file: %s, "
                          "invalid value - %s - for flag %s, valid values: "
                          OPTION_DEF_VAL_TRUE "/" OPTION_DEF_VAL_FALSE
                          ", line: %u\n", val.c_str(), config_file.c_str(),
                          name.c_str(), lineNum);
                IBDIAGNET_RETURN(1);
            }
        }

        // skip options from command line.
        if (find(p_main_parser->args.begin(), p_main_parser->args.end(), name)
                != p_main_parser->args.end())
            continue;

        for (it_req = p_requesters_list.begin(); it_req != p_requesters_list.end(); ++it_req) {
             int rc = (*it_req)->HandleOption(name, val);
             if (rc == IBDIAG_SUCCESS_CODE)
                 break;
             if (rc != IBDIAG_ERR_CODE_FABRIC_ERROR)
                 IBDIAGNET_RETURN(rc);
        }

        if (it_req == p_requesters_list.end()) {
            ERR_PRINT("Failed to parse configuration file: %s, invalid flag %s, "
                      "line: %u\n", config_file.c_str(), name.c_str(), lineNum);
            IBDIAGNET_RETURN(1);
        }
    }

    IBDIAGNET_RETURN(0);
}

int IBDiagapp::ParsePOption(const char *option_p_list,
        map_str_uint64& counters_to_threshold_map)
{
    IBDIAGNET_ENTER;
    list_string legal_pm_names = this->ibdiag_obj.GetListOFPMNames();
    bool is_legal_counter_name;
    string str = option_p_list;
    string sub_str, sub_str_name, sub_str_thresh;
    size_t i,j;

    do {
        i = str.find(',');
        sub_str = str.substr(0, i);
        str.erase(0, i + 1);
        j = sub_str.find('=');
        sub_str_name = sub_str.substr(0, j);
        sub_str_thresh = sub_str.substr(j + 1, string::npos);

        //check counter for legal name
        is_legal_counter_name = false;
        if (sub_str_name == "all") {
            counters_to_threshold_map[sub_str_name] =
                    (u_int64_t)strtoull(sub_str_thresh.c_str(), NULL, 0);
            is_legal_counter_name = true;
        }
        for (list_string::iterator it = legal_pm_names.begin(); it != legal_pm_names.end(); ++it) {
            if (sub_str_name == (*it)) {
                counters_to_threshold_map[sub_str_name] =
                        (u_int64_t)strtoull(sub_str_thresh.c_str(), NULL, 0);
                is_legal_counter_name = true;
                break;
            }
        }
        if (is_legal_counter_name == false)
            IBDIAGNET_RETURN(1);
    } while (i != string::npos);
    IBDIAGNET_RETURN(0);
}


int IBDiagapp::ParseSkipPluginList(const char *skip_plugin_list)
{
    IBDIAGNET_ENTER;

    char *pch1 = NULL;
    char *buff = new char[strlen(skip_plugin_list) + 1];
    if (!buff) {
        ERR_PRINT("Failed to allocate buff for skip list, exiting...\n");
        IBDIAGNET_RETURN(1);
    }
    strcpy(buff, skip_plugin_list);

    list_string plugins_names;
    string errors;
    if (get_plugins_names(plugins_names, errors)) {
        ERR_PRINT("%s, exiting...\n", errors.c_str());
        delete [] buff;
        IBDIAGNET_RETURN(1);
    }

    pch1 = strtok(buff, ",");
    while (pch1) {
        if (strcmp(pch1, "all") == 0) {
            this->not_to_load_plugins_list.push_front("all");   //we push to front and this
                                                                //will indicate to skip load of all
            delete [] buff;
            IBDIAGNET_RETURN(0);
        }

        list_string::iterator it;
        for (it = plugins_names.begin(); it != plugins_names.end(); ++it)
            if (strcmp(pch1, (*it).c_str()) == 0)
                break;
        if (it == plugins_names.end())  {            //unknown option
            delete [] buff;
            IBDIAGNET_RETURN(1);
        }
        this->not_to_load_plugins_list.push_back(pch1); //push only to back
        pch1 = strtok(NULL, ",");
    }

    delete [] buff;
    IBDIAGNET_RETURN(0);
}


void IBDiagapp::PrintSummary()
{
    IBDIAGNET_ENTER;

    HDR_PRINT("Summary\n");
    INFO_PRINT("%-25s %-10s %-10s %-10s\n", "Stage", "Warnings", "Errors", "Comment");
    for (int i = 0; i < NUM_OF_STAGES; ++i) {
        if(!this->stages[i])
            continue;
        if (!this->stages[i]->IsSkipped())  //HACK - don't want to see skip comments
            INFO_PRINT("%s\n", this->stages[i]->GetSummaryLine().c_str());
    }
    for (list_p_symbols_plugin::iterator it = this->loaded_plugins_list.begin();
                it != this->loaded_plugins_list.end(); ++it) {
        Plugin *p_plugin = (*it)->p_plugin_obj;
        if (!p_plugin->IsSkipped())         //HACK - don't want to see skip comments
            INFO_PRINT("%s\n", p_plugin->GetSummaryLine().c_str());
    }
    PRINT("\n");
    INFO_PRINT("You can find detailed errors/warnings in: %s\n", generate_file_name(this->base_path, LOG_FILE).c_str());
    PRINT("\n\n");
    for (int i = 0; i < NUM_OF_STAGES; ++i) {
        if(!this->stages[i])
            continue;
        if (this->stages[i]->IsFinished() && this->stages[i]->GetFilesLine() != "")
            PRINT("%s", this->stages[i]->GetFilesLine().c_str());
    }
    for (list_p_symbols_plugin::iterator it = this->loaded_plugins_list.begin();
            it != this->loaded_plugins_list.end(); ++it) {
        Plugin *p_plugin = (*it)->p_plugin_obj;
        if (p_plugin->IsFinished() && p_plugin->GetFilesLine() != "")
            PRINT("%s", p_plugin->GetFilesLine().c_str());
    }
    PRINT("\n");

    IBDIAGNET_RETURN_VOID;
}


//Returns: 0 - OK / 1 - error / 2 - OK but need to exit
int IBDiagapp::GetLocalGUIDFromUser(u_int64_t& guid)
{
    IBDIAGNET_ENTER;

    local_port_t local_ports_array[IBIS_MAX_LOCAL_PORTS];
    u_int32_t local_ports_num;
    u_int32_t choice;
    if (this->ibdiag_obj.GetAllLocalPortGUIDs(local_ports_array, &local_ports_num))
        IBDIAGNET_RETURN(1);

    while (1) {
        PRINT("\nChoose a local port number with which to bind:\n\n");
        for (u_int32_t i = 0; i < local_ports_num; ++i)
            PRINT("\t%u: GUID " U64H_FMT ", lid %u, state %s\n",
                    i + 1, ADB2C_BE64_TO_CPU(local_ports_array[i].guid),
                    local_ports_array[i].lid,
                    portstate2char((IBPortState)local_ports_array[i].logical_state));
        PRINT("\n\t0: Exit\n");
        PRINT("\nEnter choice (0-%u): ", local_ports_num);
        fflush(stdout);
        if (scanf("%u", &choice) <= 0) {
            char junk[128];
            if (scanf("%s", junk) <= 0)
                PRINT("\nError: Cannot scan!\n");
        } else if (choice == 0) {
            IBDIAGNET_RETURN(2);   //valid exit
        } else if (choice <= local_ports_num) {
            break;
        }
        PRINT("\nLame choice! Please try again.\n");
    }

    --choice;
    //return guid in host order
    guid = ADB2C_BE64_TO_CPU(local_ports_array[choice].guid);
    PRINT("\nChoice GUID=" U64H_FMT "\n", guid);
    IBDIAGNET_RETURN(0);
}

/********************************************************/
/*******************protected methods********************/
/********************************************************/
int IBDiagapp::CreateOutputPath(string value)
{
    //try to create the new directory
    if (mkdir(value.c_str(), S_IRWXU | S_IRWXG | S_IRWXO) < 0) {
        if (errno != EEXIST) {
            ERR_PRINT("Failed to set output path to %s, proceeding \
                       with default one...\n", value.c_str());
            IBDIAGNET_RETURN(3);
        }
    }
    this->base_path = value;
    this->base_path += SEPERATOR;

    //debug file
    debug_wrap_opj.ChangeDebugLog(this->base_path);

    //close log file
    destroy_log_file();

    //open log file again in new output path
    string log_file = generate_file_name(this->base_path, LOG_FILE);
    if (contruct_log_file(log_file.c_str())) {
        SCREEN_ERR_PRINT("-E- failed to create log file: %s\n",
                         log_file.c_str());
        IBDIAGNET_RETURN(3);
    }
    IBDIAGNET_RETURN(0);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiagapp::DumpStartInfoCSVTable(int argc, char **argv)
{
    IBDIAGNET_ENTER;
    char curr_line[1024];
    time_t t = time(NULL);
    struct tm tm = *localtime(&t);
    list_string plugins_names;
    string errors;
    stringstream sstream;

    csv_out.SetCommentPos();

    csv_out.DumpStart(SECTION_RUN_INFO);
    sstream << "IBDIAG_Version,IBDM_Version,IBIS_Version,Date,Args" << endl;
    csv_out.WriteLine(sstream.str());

    sstream.str("");
    sstream << get_ibdiag_version();
    memset(curr_line, 0, sizeof(curr_line));
    strftime (curr_line,sizeof(curr_line),"\"%Y %m %d %H %M %S %Z %z\"",&tm);
    sstream  << "," <<curr_line << ",\"";

    for (int i = 1; i < argc; ++i)
        sstream << " " << argv[i];

    sstream << "\"" << endl;
    csv_out.WriteLine(sstream.str());

    csv_out.DumpEnd(SECTION_RUN_INFO);

    sstream.str("");
    csv_out.DumpStart(SECTION_PLUGINS_INFO);
    sstream << "Plugin,Location" << endl;
    csv_out.WriteLine(sstream.str());
    if (get_plugins_names(plugins_names, errors)) {
        //no use to print error here
        IBDIAGNET_RETURN(1);
    }

    for (list_string::iterator it = plugins_names.begin(); it != plugins_names.end(); ++it) {
        sstream.str("");
        sstream << *it << "," << get_plugin_path_str(*it) << endl;
        csv_out.WriteLine(sstream.str());
    }

    csv_out.DumpEnd(SECTION_PLUGINS_INFO);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

/****************************************************/
/*******************public methods*******************/
/****************************************************/
IBDiagapp::IBDiagapp(const char * default_output_path) : CommandLineRequester("Main"),
        ibdiagnet_status(FAILED_TO_CONSTRUCT),
        p_main_parser(NULL),
        num_of_errors_to_screen(DEFAULT_NUM_ERRORS_TO_SCREEN),
        m_max_smps_on_wire(DEFAULT_SMPS_WINDOW),
        m_max_gmps_on_wire(DEFAULT_GMPS_WINDOW),
        base_path(default_output_path),
        p_mkeymngr(NULL),
        config_file(DEFAULT_CONFIG_FILE)
{
    IBDIAGNET_ENTER;

    IBDIAGNET_RETURN_VOID;
}



IBDiagapp::~IBDiagapp()
{
    IBDIAGNET_ENTER;

    //Make sure that there are no pending and on the wire MADs.
    ibdiag_obj.GetIbisPtr()->MadRecAll();

    if (this->ibdiagnet_status != FAILED_TO_CONSTRUCT) {
        this->ibdiagnet_status = FAILED_TO_CONSTRUCT;
        for (int i = 0; i < NUM_OF_STAGES; ++i)
            if (this->stages[i]) {
                delete this->stages[i];
                this->stages[i] = NULL;
            }
        delete this->p_main_parser;
        destroy_log_file();
    }
    if (p_mkeymngr)
        delete p_mkeymngr;

    this->UnloadPlugins();
    IBDIAGNET_RETURN_VOID;
}


//Returns: 0 - OK / 1 - parsing error / 2 - OK but need to exit / 3 - some other error
int IBDiagapp::HandleOption(string name, string value)
{
    IBDIAGNET_ENTER;

    /* printf("Parsing (requester=\"%s\") \"%s\" with val=\"%s\"\n",
            this->getName().c_str(), name.c_str(), value.c_str()); */

    bool bool_flag_val = true;

    if (value == OPTION_DEF_VAL_NULL)
        IBDIAGNET_RETURN(0);

    if (name == OPTION_DEVICE) {
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->SetDeviceName(value);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_PORT) {
        if (!is_non_neg_num(value)) {
            ERR_PRINT_ARGS("Illegal value for \'-%c\' option: %s\n",
                    OPTION_PORT_SHORT, value.c_str());
            PRINT("    Only non negative number is allowed\n");
            IBDIAGNET_RETURN(3);
        }
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->SetPortNum((u_int8_t)strtoul(value.c_str(), NULL, 0));
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_GUID) {
        u_int64_t given_guid = strtoull(value.c_str(), NULL, 0);
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->SetPortGUID(given_guid);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_SKIP) {
        if (this->ParseSkipList(value.c_str())) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                    OPTION_SKIP, value.c_str());
            PRINT("    (Legal value: ");
            size_t skip_options_list_length = skip_options_list.size();
            for (size_t i = 0; i < skip_options_list_length - 1; ++i) {
                PRINT("%s | ", skip_options_list[i].name.c_str())
            }
            PRINT("%s)\n", skip_options_list[skip_options_list_length - 1].name.c_str())
            IBDIAGNET_RETURN(3);
        }
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_SKIP_PLUGIN) {
        if (this->ParseSkipPluginList(value.c_str())) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                    OPTION_SKIP_PLUGIN, value.c_str());
            list_string plugins_names;
            string errors;
            if (get_plugins_names(plugins_names, errors)) {
                //no use to print error here
                IBDIAGNET_RETURN(3);
            }
            PRINT("    (Legal value: ");
            for (list_string::iterator it = plugins_names.begin(); it != plugins_names.end(); ++it) {
                if (it != plugins_names.begin())
                    PRINT(" | ");
                PRINT("%s", (*it).c_str());
            }
            PRINT(")\n");
            IBDIAGNET_RETURN(3);
        }
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_PC) {
        this->ParseBoolValue(value, bool_flag_val);
        ((PMStage *)this->stages[STAGE_IDX_PM])->SetResetCounters(bool_flag_val);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_MLNX_CNTRS) {
        this->ParseBoolValue(value, bool_flag_val);
        ((MlnxCntrsStage *)this->stages[STAGE_IDX_MLNX_CNTRS])->MarkStage(bool_flag_val);
        IBDIAGNET_RETURN(0);
     } else if (name == OPTION_MLNX_CNTRS_RESET) {
        this->ParseBoolValue(value, bool_flag_val);
        ((MlnxCntrsStage *)this->stages[STAGE_IDX_MLNX_CNTRS])->SetResetCounters(bool_flag_val);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_P) {
        map_str_uint64& counters_to_threshold_map =
                ((PMStage *)this->stages[STAGE_IDX_PM])->GetRefToCountersToThresholdMap();
        if (this->ParsePOption(value.c_str(), counters_to_threshold_map)) {
            ERR_PRINT_ARGS("Illegal value for \'-%c\' option: %s\n"  \
                    "    (Legal value: one or more \"<PM counter>=<Thresh Limit>\"\n",
                    OPTION_P_SHORT, value.c_str());
            PRINT("    Legal PM Counter names are:\n");
            list_string allowed_names_list = this->ibdiag_obj.GetListOFPMNames();
            for (list_string::iterator it = allowed_names_list.begin();
                    it != allowed_names_list.end(); ++it)
                PRINT("    %s\n", it->c_str());
            IBDIAGNET_RETURN(3);
        }
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_PMPAUSE) {
        if (!is_non_neg_num(value)) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                    OPTION_PMPAUSE, value.c_str());
            PRINT("    Only non negative number is allowed\n");
            IBDIAGNET_RETURN(3);
        }
        ((PMStage *)this->stages[STAGE_IDX_PM])->SetPauseTimeBetweenSamples(
            (u_int32_t)strtoul(value.c_str(), NULL, 0));
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_BER) {
        this->ParseBoolValue(value, bool_flag_val);
        ((PMStage *)this->stages[STAGE_IDX_PM])->SetBERTest(bool_flag_val);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_BERDATA) {
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_BERTHRESH) {
        if (!is_non_neg_num(value)) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                    OPTION_BERTHRESH, value.c_str());
            PRINT("    Only non negative number is allowed\n");
            IBDIAGNET_RETURN(3);
        }
        ((PMStage *)this->stages[STAGE_IDX_PM])->SetBERThreshold(strtoul(value.c_str(), NULL, 0));
        ibdiag_obj.SetBERThreshold(strtoul(value.c_str(), NULL, 0));
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_LLR_ACTIVE_CELL) {
        if (strcmp(value.c_str(), "64") && strcmp(value.c_str(), "128") &&
            strcmp(value.c_str(), "0")) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n" \
                    "    (Legal value: %s).\n",
                    OPTION_LLR_ACTIVE_CELL, value.c_str(), OPTION_LLR_ACTIVE_CELL_ARG_DESC);
            IBDIAGNET_RETURN(3);
        }
        ibdiag_obj.SetLLRActiveCell(value);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_EXT_SPEEDS) {
        ((PMStage *)this->stages[STAGE_IDX_PM])->SetCheckExtSpeedsCounters(value.c_str());
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_ALL_LANES) {
        this->ParseBoolValue(value, bool_flag_val);
        ((PMStage *)this->stages[STAGE_IDX_PM])->SetPerLaneCounters(bool_flag_val);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_LLR_INTERNAL_COUNTERS) {
        ((PMStage *)this->stages[STAGE_IDX_PM])->SetPrintLLRCounters();
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_LS) {
        if ((char2speed(value.c_str()) == IB_UNKNOWN_LINK_SPEED) &&
            (value != IB_DISABLE_LINK_SPEED_STR)) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n" \
                           "    (Legal value: %s).\n",
                           OPTION_LS, value.c_str(), OPTION_LS_ARG_DESC);
            IBDIAGNET_RETURN(3);
        }
        ((Speed_WidthStage *)this->stages[STAGE_IDX_SPEED_WIDTH])->SetExpectedLinkSpeed(value);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_LW) {
        if ((char2width(value.c_str()) == IB_UNKNOWN_LINK_WIDTH) &&
            (value != IB_DISABLE_LINK_WIDTH_STR)) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n" \
                    "    (Legal value: %s).\n",
                    OPTION_LW, value.c_str(), OPTION_LW_ARG_DESC);
            IBDIAGNET_RETURN(3);
        }
        ((Speed_WidthStage *)this->stages[STAGE_IDX_SPEED_WIDTH])->SetExpectedLinkWidth(value);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_SCRRENNUMERRS) {
        if (!is_non_neg_num(value)) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                    OPTION_SCRRENNUMERRS, value.c_str());
            PRINT("    Only non negative number is allowed\n");
            IBDIAGNET_RETURN(3);
        }
        this->num_of_errors_to_screen = (unsigned int)strtoul(value.c_str(), NULL, 0);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_SMP_WINDOW) {
        if (!is_non_neg_num(value)) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                    OPTION_SMP_WINDOW, value.c_str());
            PRINT("    Only non negative number is allowed\n");
            IBDIAGNET_RETURN(3);
        }
        this->m_max_smps_on_wire = (unsigned int)strtoul(value.c_str(), NULL, 0);
        if (m_max_smps_on_wire > MAX_SMPS_WINDOW) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s > %d\n",
                    OPTION_SMP_WINDOW, value.c_str(), MAX_SMPS_WINDOW);
            IBDIAGNET_RETURN(3);
        }
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_GMP_WINDOW) {
        if (!is_non_neg_num(value)) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                    OPTION_GMP_WINDOW, value.c_str());
            PRINT("    Only non negative number is allowed\n");
            IBDIAGNET_RETURN(3);
        }
        this->m_max_gmps_on_wire = (unsigned int)strtoul(value.c_str(), NULL, 0);
        if (m_max_gmps_on_wire > MAX_GMPS_WINDOW) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s > %d\n",
                    OPTION_GMP_WINDOW, value.c_str(), MAX_GMPS_WINDOW);
            IBDIAGNET_RETURN(3);
        }
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_MAXHOPS) {
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->SetMaxHops((u_int8_t)strtoul(value.c_str(), NULL, 0));
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_NO_MEPI) {
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->SetNoMEPI();
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_FEC_MODE) {
        this->ParseBoolValue(value, bool_flag_val);
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->SetFECModeDump(bool_flag_val);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_TOPO_MATCH_LOCAL_SYSTEM) {
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_VERSION) {
        INFO_PRINT("%s\n", IBDIAGNET_VERSION);
        INFO_PRINT("Dependencies: %s\n", get_ibdiag_version().c_str());
        IBDIAGNET_RETURN(2);           //exit with no error
    } else if (name == OPTION_HELP) {
        this->LoadPlugins();
        SCREEN_PRINT("%s\n", this->p_main_parser->GetUsage().c_str());
        IBDIAGNET_RETURN(2);           //exit with no error
    }  else if (name == OPTION_DEEP_HELP) {
        this->LoadPlugins();
        SCREEN_PRINT("%s\n", this->p_main_parser->GetUsage().c_str());
        IBDIAGNET_RETURN(2);           //exit with no error
    }   else if (name == OPTION_MLNX_HELP) {
        this->LoadPlugins();
        SCREEN_PRINT("%s\n", this->p_main_parser->GetUsage(true).c_str());
        IBDIAGNET_RETURN(2);           //exit with no error
    } else if (name == OPTION_TOPO_COMPARE_NAMES) {
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_CAPABILITY_WRITE) {
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->SetOutputCapabilityFile(value);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_FULL_CAPABILITY_WRITE) {
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->SetOutputFullCapabilityFile(value);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_CAPABILITY_READ) {
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->
            SetInputCapabilityFile(value);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_BACK_COMPAT_DB) {
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->
            SetShowPortsDataFormat(value);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_MADS_TIMEOUT) {
        if (!is_non_neg_num(value)) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                    OPTION_MADS_TIMEOUT, value.c_str());
            PRINT("    Only non negative number is allowed\n");
            IBDIAGNET_RETURN(3);
        }
        ibdiag_obj.GetIbisPtr()->SetTimeout((int)strtoul(value.c_str(), NULL, 0));
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_MADS_RETRIES) {
        if (!is_non_neg_num(value)) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                    OPTION_MADS_RETRIES, value.c_str());
            PRINT("    Only non negative number is allowed\n");
            IBDIAGNET_RETURN(3);
        }
        ibdiag_obj.GetIbisPtr()->SetNumOfRetries((int)strtoul(value.c_str(), NULL, 0));
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_DUMP_LST_WITH_LMC) {
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->SetDumpLSTWithLMC();
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_MKEY) {
        mkey_val = value;
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_MKEY_FILES) {
        mkey_path = value;
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_CONFIG) {
        config_file = value;
        IBDIAGNET_RETURN(0);

    // deprecated options
    } else if (name == OPTION_LLR_COUNTERS || name == OPTION_LLR_SYMBOL_ERROR) {
        SCREEN_ERR_PRINT("The --%s option is deprecated\n", name.c_str());
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_MAP) {
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->
            SetNodeNameMapFile(value);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_ROUTERS) {
        this->ParseBoolValue(value, bool_flag_val);
        ((RoutersStage *)this->stages[STAGE_IDX_ROUTERS])->
            MarkStage(bool_flag_val);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_PER_SLVL_CNTRS) {
        this->ParseBoolValue(value, bool_flag_val);
        ((PMStage *)this->stages[STAGE_IDX_PM])->SetPerSLVLCounters(bool_flag_val);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_QOS) {
        this->ParseBoolValue(value, bool_flag_val);
        ((QoSStage *)this->stages[STAGE_IDX_QOS])->
            MarkStage(bool_flag_val);
        IBDIAGNET_RETURN(0);
    }


    //if we reached here than unknown option
    IBDIAGNET_RETURN(1);
}


void IBDiagapp::LoadPlugins()
{
    IBDIAGNET_ENTER;

    /*
     * load all plugins
     */
    list_string plugins_names;
    string errors;
    if (get_plugins_names(plugins_names, errors)) {
        ERR_PRINT("%s\n", errors.c_str());
        IBDIAGNET_RETURN_VOID
    }

    symbols_plugin_t *p_curr_sym_plugin;
    PRINT("----------\n");
    PRINT("Load Plugins from:\n%s\n", get_plugins_paths_str().c_str());
    PRINT("%-45s %-10s %s\n", "Plugin Name", "Result", "Comment");
    for (list_string:: iterator it = plugins_names.begin(); it != plugins_names.end(); ++it) {
        PRINT("%-45s ", (*it).c_str());
        for (list_string::iterator it2 = this->not_to_load_plugins_list.begin();
                it2 != this->not_to_load_plugins_list.end(); ++it2) {
            if ((*it) == (*it2)) {
                PRINT("%-10s %s\n", "", "Skipped");
                goto next_plugin_load;
            }
        }
        p_curr_sym_plugin = load_plugin((*it),
                &this->ibdiag_obj,
                errors);
        if (!p_curr_sym_plugin) {
            PRINT("%-10s %s\n", "Failed", errors.c_str());
            continue;
        }
        if (this->p_main_parser->AddRequester(p_curr_sym_plugin->p_plugin_obj)) {
            unload_plugin(p_curr_sym_plugin);
            PRINT("%-10s Plugin options issue - %s\n", "Failed", this->p_main_parser->GetErrDesc());
            continue;
        }
        PRINT("%-10s %s\n", "Succeeded", "Plugin loaded");
        this->loaded_plugins_list.push_back(p_curr_sym_plugin);

next_plugin_load:
        continue;
    }
    plugins_names.clear();
    PRINT("\n");

    IBDIAGNET_RETURN_VOID;
}


void IBDiagapp::UnloadPlugins()
{
    IBDIAGNET_ENTER;
    for (list_p_symbols_plugin::iterator it = this->loaded_plugins_list.begin();
            it != this->loaded_plugins_list.end(); ++it) {
        unload_plugin(*it);
    }
    this->loaded_plugins_list.clear();
    this->not_to_load_plugins_list.clear();
    IBDIAGNET_RETURN_VOID;
}


bool IBDiagapp::isValidLoadFromFilePath()
{
    char *res;
    const char *input_path;
    const char *output_path;
    char input_fullpath[PATH_MAX + 1] = {0};
    char output_fullpath[PATH_MAX + 1] = {0};
    string input_file_path = ((DiscoveryFabricStage *)this->stages[STAGE_IDX_DISCOVERY])->
                              GetLoadFromFilePath();
    string output_dir_path = this->base_path;
    string output_file_path;
    int errsv;

    if (((DiscoveryFabricStage *)this->stages[STAGE_IDX_DISCOVERY])->IsLoadFromFile()) {
        input_path = input_file_path.c_str();
        output_path = output_dir_path.c_str();

        if(!(res = realpath(input_path, input_fullpath))) {
            errsv = errno;
            ERR_PRINT("CSV file input path %s: %s\n", input_path, strerror(errsv));
            return false;
        }

        if (!(res = realpath(output_path, output_fullpath))) {
            errsv = errno;
            ERR_PRINT("CSV directory output path %s: %s\n", output_path, strerror(errsv));
            return false;
        }

        output_file_path = output_fullpath + string(SEPERATOR) + string(DB_FILE);

        if (string(input_fullpath) == output_file_path) {
            ERR_PRINT("The CSV file path in the \"load from file\" flag (%s)\n"
                      "    is the same as the output CSV file path (%s" DB_FILE ")\n",
                      input_path, output_path);
            return false;
        }
    }

    return true;
}


int IBDiagapp::CreateConfigFile(string file_path)
{
    IBDIAGNET_ENTER;

    ofstream fs(file_path.c_str());
    if (fs.fail()) {
        ERR_PRINT("Failed to create/open configuration file: %s\n", file_path.c_str());
        IBDIAGNET_RETURN(1);
    }

    fs << "# " << IBDIAGNET_VERSION << endl;
    fs << endl;
    fs << endl;

    list_p_command_line_req p_requesters_list = this->p_main_parser->GetRequestersList();

    for (list_p_command_line_req::iterator it_req = p_requesters_list.begin();
         it_req != p_requesters_list.end(); ++it_req) {
         vec_option_t options_list = (*it_req)->GetOptions();
         for (vec_option_t::iterator it_opt = options_list.begin();
              it_opt != options_list.end(); ++it_opt) {
              if (it_opt->attributes & IBIS_CMD_NO_CONF_FILE)
                  continue;

              string opt_description = string("# ") + it_opt->description;

              size_t prev_found = 0;
              size_t found_new_line = opt_description.find('\n', prev_found);
              size_t found_space = opt_description.find(" ", prev_found);
              while (found_space != string::npos) {
                  if ((found_new_line != string::npos) && (found_new_line < found_space)) {
                      opt_description.insert(found_new_line + 1, "# ");
                      prev_found = found_new_line;
                  }

                  if ((found_space - prev_found) > OPT_DESC_CONFIG_LEN) {
                      prev_found = found_space + 3;
                      opt_description.insert(found_space + 1, "\n# ");
                      found_space += 3;
                  }
                  found_new_line = opt_description.find('\n', found_space + 1);
                  found_space = opt_description.find(" ", found_space + 1);
              }
              if (found_new_line != string::npos)
                  opt_description.insert(found_new_line + 1, "# ");

              fs << opt_description << endl;

              if (it_opt->attributes & IBIS_CMD_BOOLEAN_FLAG) {
                  fs << "# value type: Boolean" << endl;
              } else {
                  fs << "# value type: " << it_opt->option_value << endl;
              }

              fs << it_opt->option_name << " = " << it_opt->default_value_str << endl;
              fs << endl;
         }
    }

    SCREEN_PRINT("Congiguration file was created: %s\n", file_path.c_str());
    fs.close();

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}


//Returns: 0 - OK / 1 - error / 2 - OK but need to exit
int IBDiagapp::GetOpt(int argc, char **argv)
{
    IBDIAGNET_ENTER;

    int rc2, rc = 1;
    list_p_command_line_req ignored_req_list;
    string log_file;

    if (this->ibdiagnet_status != NEED_TO_GET_ARGS) {
        SCREEN_ERR_PRINT("Can not parse arguments from user\n");
        IBDIAGNET_RETURN(rc);
    }
    /*
     * add the requester to the parser
     */
    if (this->p_main_parser->AddRequester(this)) {
        SCREEN_ERR_PRINT("Failed add options to main parser, err=%s\n", this->p_main_parser->GetErrDesc());
        goto exit;
    }

    /*
     * finally parse all options
     */
    rc2 = this->p_main_parser->ParseOptions(argc, argv, true);
    if (rc2 == 1) {         //parse error and we need to dump usage
        goto usage_and_exit;
    } else if (rc2 == 2) {  //OK but need to exit
        rc = 2;
        goto exit;
    } else if (rc2 == 3) {  //error but not need to dump usage
        goto exit;
    }

    // try to construct a log file
    log_file = generate_file_name(this->base_path, LOG_FILE);
    if (contruct_log_file(log_file.c_str())) {
        SCREEN_ERR_PRINT("Failed to create log file: %s\n", log_file.c_str());
        IBDIAGNET_RETURN(1);
    }

    // add argv to log and screen
    PRINT("Running: ");
    for (int i = 0; i < argc; ++i)
        PRINT("%s ", argv[i]);
    PRINT("\n");

    /*
     * load all plugins
     */
    if ((this->not_to_load_plugins_list.empty() == true) ||
            this->not_to_load_plugins_list.front() != "all") {
        this->LoadPlugins();

        /*
         * parse options again with plugins
         */
        ignored_req_list.push_back(this);
        rc2 = this->p_main_parser->ParseOptions(argc, argv, false, &ignored_req_list);
        if (rc2 == 1) {         //parse error and we need to dump usage
            goto usage_and_exit;
        } else if (rc2 == 2) {  //OK but need to exit
            rc = 2;
            goto exit;
        } else if (rc2 == 3) {  //error but not need to dump usage
            goto exit;
        }
    } else {
        PRINT("----------\n");
        INFO_PRINT("Plugins load will be skipped\n");
        PRINT("\n");
    }

    // get options from configuration file.
    rc2 = this->ParseConfigFile();
    if (rc2 == 2) {
        rc = 2;
        goto exit;
    } else if (rc2) {
        IBDIAGNET_RETURN(3);
    }

    // handle mkey params
    if (!mkey_val.empty() || !mkey_path.empty()) {
        if (HandleMKeyParams()) {
            IBDIAGNET_RETURN(3);
        }
    }

    /*
     *  some stages need to validate params
     */
    ((ToplogtMatchStage *)this->stages[STAGE_IDX_TOPLOGY_MATCH])->ValidateTopoParams();

    /*
     * check if guid 0 was given,
     * if such case than need to present list of guid to the user
     */
    if (((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->IsNeedDisplayAllPortsGuids()) {
        u_int64_t chosen_guid = 0;
        rc2 = this->GetLocalGUIDFromUser(chosen_guid);
        if (rc2 == 1) {
            ERR_PRINT("Failed to display user GUIDs list to choose from, err=%s\n,", this->ibdiag_obj.GetLastError());
            goto exit;
        } else if (rc2 == 2) {
            rc = 2;
            goto exit;
        }
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->SetPortGUID(chosen_guid);
    }

    if(!this->isValidLoadFromFilePath()){
       rc =1;
       goto exit;
    }

    this->ibdiagnet_status = NOT_INITILIAZED;
    rc = 0;
exit:
    IBDIAGNET_RETURN(rc);
usage_and_exit:
    ERR_PRINT("Failed to parse options, err=%s\n\n", this->p_main_parser->GetErrDesc());
    SCREEN_PRINT("%s\n", this->p_main_parser->GetUsage().c_str());
    IBDIAGNET_RETURN(rc);
}

/* OK: return = 0 , ERROR: return != 0 */
int IBDiagapp::Init()
{
     IBDIAGNET_ENTER;

    /*
     * initialize main parser
     */
    this->p_main_parser = new CommandLineParser(this->GetToolName());
    if (!this->p_main_parser) {
        SCREEN_ERR_PRINT("Failed to allocate main parser\n");
        IBDIAGNET_RETURN(1);
    }

    //init ibdiag lib
    if (this->ibdiag_obj.Init()) {
        SCREEN_ERR_PRINT("Failed to initialize IBDiag lib\n");
        IBDIAGNET_RETURN(1);
    }

    m_lc_plugins_mgr.Init(&this->ibdiag_obj);

    for (uint8_t i = 0; i < NUM_OF_STAGES; ++i) {
        this->stages[i] = CreateStage(i);
    }

    /*
     * add all main options
     */
    char buff_def_val[1024] = {0};

    this->AddOptions(OPTION_DEVICE, OPTION_DEVICE_SHORT, OPTION_DEVICE_ARG_DESC, OPTION_DEVICE_DESC,
                     OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_PORT, OPTION_PORT_SHORT, OPTION_PORT_ARG_DESC, OPTION_PORT_DESC,
                     OPTION_PORT_DEF_VAL);
    this->AddOptions(OPTION_GUID, OPTION_GUID_SHORT, OPTION_GUID_ARG_DESC, OPTION_GUID_DESC,
                     OPTION_GUID_DEF_VAL_STR);

    skip_obj_t obj2 = {SKIP_VS_CAP_SMP_VALUE, "vs_cap_smp" };
    skip_options_list.push_back(obj2);
    skip_obj_t obj3 = {SKIP_VS_CAP_GMP_VALUE, "vs_cap_gmp"};
    skip_options_list.push_back(obj3);
    skip_obj_t obj4 = {SKIP_STAGE_LINKS_VALUE, "links"};
    skip_options_list.push_back(obj4);
    skip_obj_t obj5 = {SKIP_STAGE_PM_VALUE, "pm"};
    skip_options_list.push_back(obj5);
    skip_obj_t obj6 = {SKIP_STAGE_SPEED_WIDTH_CHECK, "speed_width_check"};
    skip_options_list.push_back(obj6);
    skip_obj_t obj7 = {SKIP_STAGE_TEMPERATURE_SENSING, "temp_sensing"};
    skip_options_list.push_back(obj7);
    skip_obj_t obj8 = {SKIP_STAGE_VIRTUALIZATION, "virt"};
    skip_options_list.push_back(obj8);
    skip_obj_t obj = {SKIP_ALL_VALUE, "all"};
    skip_options_list.push_back(obj);

    string skip_list_str = OPTION_SKIP_DESC;
    skip_list_str += " Applicable skip stages:\n(";
    size_t skip_options_list_length = skip_options_list.size();
    for (size_t i = 0; i < skip_options_list_length - 1; ++i) {
        skip_list_str += skip_options_list[i].name.c_str();
        skip_list_str += " | ";
    }
    skip_list_str += skip_options_list[skip_options_list_length - 1].name.c_str();
    skip_list_str += ").";
    this->AddOptions(OPTION_SKIP, OPTION_SKIP_SHORT, OPTION_SKIP_ARG_DESC, skip_list_str,
                     OPTION_DEF_VAL_NULL);

    list_string plugins_names;
    string errors;
    if (get_plugins_names(plugins_names, errors)) {
        SCREEN_ERR_PRINT("%s\n", errors.c_str());
        IBDIAGNET_RETURN(1);
    }

    if (!errors.empty())
        PRINT("%s\n", errors.c_str());

    string skip_plugin_list_str = OPTION_SKIP_PLUGIN_DESC;
    skip_plugin_list_str += " Applicable skip plugins:\n(";
    for (list_string::iterator it = plugins_names.begin(); it != plugins_names.end(); ++it) {
        if (it != plugins_names.begin())
            skip_plugin_list_str += " | ";
        skip_plugin_list_str += (*it).c_str();
    }
    skip_plugin_list_str += ").";
    this->AddOptions(OPTION_SKIP_PLUGIN, OPTION_SKIP_PLUGIN_SHORT, OPTION_SKIP_PLUGIN_ARG_DESC,
                     skip_plugin_list_str, OPTION_DEF_VAL_NULL);

    this->AddOptions(OPTION_PER_SLVL_CNTRS, OPTION_PER_SLVL_CNTRS_SHORT,
                     OPTION_PER_SLVL_CNTRS_ARG_DESC, OPTION_PER_SLVL_CNTRS_DESC,
                     OPTION_DEF_VAL_FALSE,IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_MLNX_CNTRS, OPTION_MLNX_CNTRS_SHORT, OPTION_MLNX_CNTRS_ARG_DESC,
                     OPTION_MLNX_CNTRS_DESC, OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_MLNX_CNTRS_RESET, OPTION_MLNX_CNTRS_RESET_SHORT,
                     OPTION_MLNX_CNTRS_RESET_ARG_DESC, OPTION_MLNX_CNTRS_RESET_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_PC, OPTION_PC_SHORT, OPTION_PC_ARG_DESC, OPTION_PC_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_P, OPTION_P_SHORT, OPTION_P_ARG_DESC, OPTION_P_DESC,
                     OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_PMPAUSE, OPTION_PMPAUSE_SHORT, OPTION_PMPAUSE_ARG_DESC,
                     OPTION_PMPAUSE_DESC, OPTION_PMPAUSE_DEF_VAL);
    this->AddOptions(OPTION_BER, OPTION_BER_SHORT, OPTION_BER_ARG_DESC, OPTION_BER_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);
    // deprecated ber_use_data option
    this->AddOptions(OPTION_BERDATA, OPTION_BERDATA_SHORT, OPTION_BERDATA_ARG_DESC,
                     OPTION_BERDATA_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_DEPRECATED | IBIS_CMD_HIDDEN | IBIS_CMD_BOOLEAN_FLAG);
    sprintf(buff_def_val, "%llu", DEFAULT_BER_THRESHOLD);
    this->AddOptions(OPTION_BERTHRESH, OPTION_BERTHRESH_SHORT, OPTION_BERTHRESH_ARG_DESC,
                     OPTION_BERTHRESH_DESC, string(buff_def_val));
    this->AddOptions(OPTION_LLR_ACTIVE_CELL, OPTION_LLR_ACTIVE_CELL_SHORT,
                     OPTION_LLR_ACTIVE_CELL_ARG_DESC, OPTION_LLR_ACTIVE_CELL_DESC,
                     OPTION_LLR_ACTIVE_CELL_DEF_VAL);
    this->AddOptions(OPTION_EXT_SPEEDS, OPTION_EXT_SPEEDS_SHORT, OPTION_EXT_SPEEDS_ARG_DESC,
                     OPTION_EXT_SPEEDS_DESC, OPTION_EXT_SPEEDS_DEF_VAL);
    this->AddOptions(OPTION_ALL_LANES, OPTION_ALL_LANES_SHORT, OPTION_ALL_LANES_ARG_DESC,
                     OPTION_ALL_LANES_DESC, OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_LLR_INTERNAL_COUNTERS, OPTION_LLR_INTERNAL_COUNTERS_SHORT,
                     OPTION_LLR_INTERNAL_COUNTERS_ARG_DESC, OPTION_LLR_INTERNAL_COUNTERS_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_HIDDEN | IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_LS, OPTION_LS_SHORT, OPTION_LS_ARG_DESC, OPTION_LS_DESC,
                     IB_DISABLE_LINK_SPEED_STR);
    this->AddOptions(OPTION_LW, OPTION_LW_SHORT, OPTION_LW_ARG_DESC, OPTION_LW_DESC,
                     IB_DISABLE_LINK_WIDTH_STR);
    this->AddOptions(OPTION_TOPO_MATCH_LOCAL_SYSTEM, OPTION_TOPO_MATCH_LOCAL_SYSTEM_SHORT,
                     OPTION_TOPO_MATCH_LOCAL_SYSTEM_ARG_DESC, OPTION_TOPO_MATCH_LOCAL_SYSTEM_DESC,
                     OPTION_DEF_VAL_NULL ,IBIS_CMD_DEPRECATED | IBIS_CMD_HIDDEN);
    this->AddOptions(OPTION_TOPO_COMPARE_NAMES, OPTION_TOPO_COMPARE_NAMES_SHORT,
                     OPTION_TOPO_COMPARE_NAMES_ARG_DESC, OPTION_TOPO_COMPARE_NAMES_DESC,
                     OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_DEPRECATED | IBIS_CMD_HIDDEN | IBIS_CMD_BOOLEAN_FLAG);
    // deprecated llr option
    this->AddOptions(OPTION_LLR_COUNTERS, OPTION_LLR_COUNTERS_SHORT, OPTION_LLR_COUNTERS_ARG_DESC,
                     OPTION_LLR_COUNTERS_DESC, OPTION_DEF_VAL_NULL,
                     IBIS_CMD_DEPRECATED | IBIS_CMD_HIDDEN);
    this->AddOptions(OPTION_LLR_SYMBOL_ERROR, OPTION_LLR_SYMBOL_ERROR_SHORT,
                     OPTION_LLR_SYMBOL_ERROR_ARG_DESC, OPTION_LLR_SYMBOL_ERROR_DESC,
                     OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_DEPRECATED | IBIS_CMD_HIDDEN | IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_SCRRENNUMERRS, OPTION_SCRRENNUMERRS_SHORT,
                     OPTION_SCRRENNUMERRS_ARG_DESC, OPTION_SCRRENNUMERRS_DESC,
                     OPTION_SCRRENNUMERRS_DEF_VAL);
    this->AddOptions(OPTION_SMP_WINDOW, OPTION_SMP_WINDOW_SHORT, OPTION_SMP_WINDOW_ARG_DESC,
                     OPTION_SMP_WINDOW_DESC, OPTION_SMP_WINDOW_DEF_VAL);
    this->AddOptions(OPTION_GMP_WINDOW, OPTION_GMP_WINDOW_SHORT, OPTION_GMP_WINDOW_ARG_DESC,
                     OPTION_GMP_WINDOW_DESC, OPTION_GMP_WINDOW_DEF_VAL);
    this->AddOptions(OPTION_MAXHOPS, OPTION_MAXHOPS_SHORT, OPTION_MAXHOPS_ARG_DESC,
                     OPTION_MAXHOPS_DESC, OPTION_MAXHOPS_DEF_VAL);
    this->AddOptions(OPTION_NO_MEPI, OPTION_NO_MEPI_SHORT, OPTION_NO_MEPI_ARG_DESC,
                     OPTION_NO_MEPI_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_HIDDEN | IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_DUMP_LST_WITH_LMC, OPTION_DUMP_LST_WITH_LMC_SHORT,
                     OPTION_DUMP_LST_WITH_LMC_ARG_DESC, OPTION_DUMP_LST_WITH_LMC_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_HIDDEN | IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_CAPABILITY_READ,
                     OPTION_CAPABILITY_READ_SHORT,
                     OPTION_CAPABILITY_READ_ARG_DESC,
                     OPTION_CAPABILITY_READ_DESC,
                     OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_CAPABILITY_WRITE, OPTION_CAPABILITY_WRITE_SHORT,
                     OPTION_CAPABILITY_WRITE_ARG_DESC, OPTION_CAPABILITY_WRITE_DESC,
                     OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_FULL_CAPABILITY_WRITE,
                     OPTION_FULL_CAPABILITY_WRITE_SHORT,
                     OPTION_FULL_CAPABILITY_WRITE_ARG_DESC,
                     OPTION_FULL_CAPABILITY_WRITE_DESC, OPTION_DEF_VAL_NULL,
                     IBIS_CMD_HIDDEN);
    this->AddOptions(OPTION_BACK_COMPAT_DB, OPTION_BACK_COMPAT_DB_SHORT,
                     OPTION_BACK_COMPAT_DB_ARG_DESC,
                     OPTION_BACK_COMPAT_DB_DESC, OPTION_BACK_COMPAT_DB_DEF_VAL);
    this->AddOptions(OPTION_VERSION, OPTION_VERSION_SHORT, OPTION_VERSION_ARG_DESC,
                     OPTION_VERSION_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_INFO_FLAG | IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_HELP, OPTION_HELP_SHORT, OPTION_HELP_ARG_DESC, OPTION_HELP_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_INFO_FLAG | IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_DEEP_HELP, OPTION_DEEP_HELP_SHORT, OPTION_DEEP_HELP_ARG_DESC,
                     OPTION_DEEP_HELP_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_DEPRECATED | IBIS_CMD_INFO_FLAG | IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_MLNX_HELP, OPTION_MLNX_HELP_SHORT, OPTION_MLNX_HELP_ARG_DESC,
                     OPTION_MLNX_HELP_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_INFO_FLAG | IBIS_CMD_HIDDEN | IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_MADS_TIMEOUT, OPTION_MADS_TIMEOUT_SHORT, OPTION_MADS_TIMEOUT_ARG_DESC,
                     OPTION_MADS_TIMEOUT_DESC, OPTION_MADS_TIMEOUT_DEF_VAL);
    this->AddOptions(OPTION_MADS_RETRIES, OPTION_MADS_RETRIES_SHORT, OPTION_MADS_RETRIES_ARG_DESC,
                     OPTION_MADS_RETRIES_DESC, OPTION_MADS_RETRIES_DEF_VAL);
    this->AddOptions(OPTION_MAP, OPTION_MAP_SHORT, OPTION_MAP_ARG_DESC, OPTION_MAP_DESC,
                     OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_MKEY, OPTION_MKEY_SHORT, OPTION_MKEY_ARG_DESC,
                     OPTION_MKEY_DESC, OPTION_MKEY_DEF_VAL);
    this->AddOptions(OPTION_MKEY_FILES, OPTION_MKEY_FILES_SHORT, OPTION_MKEY_FILES_ARG_DESC,
                     OPTION_MKEY_FILES_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_ROUTERS, OPTION_ROUTERS_SHORT, OPTION_ROUTERS_ARG_DESC,
                     OPTION_ROUTERS_DESC, OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_QOS, OPTION_QOS_SHORT, OPTION_QOS_ARG_DESC, OPTION_QOS_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_FEC_MODE, OPTION_FEC_MODE_SHORT, OPTION_FEC_MODE_ARG_DESC,
                     OPTION_FEC_MODE_DESC, OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_CONFIG, OPTION_CONFIG_SHORT, OPTION_CONFIG_ARG_DESC,
                     OPTION_CONFIG_DESC, OPTION_DEF_VAL_NULL, IBIS_CMD_NO_CONF_FILE);

    this->ibdiagnet_status = NEED_TO_GET_ARGS;

    IBDIAGNET_RETURN(0);
}

int IBDiagapp::PostInit()
{
    IBDIAGNET_ENTER;

    if (this->ibdiagnet_status == NEED_TO_GET_ARGS) {
        ERR_PRINT("Need to get args from user before initialize\n");
        IBDIAGNET_RETURN(1);
    }
    if (this->ibdiagnet_status == READY) {
        ERR_PRINT("Already initialized\n");
        IBDIAGNET_RETURN(1);
    }

    //open db_csv file
    string db_file_str = generate_file_name(this->base_path, DB_FILE);
    string err_message;
    int rc = csv_out.Init(db_file_str, err_message);
    if (rc) {
        IBDIAGNET_LOG(TT_LOG_LEVEL_ERROR, "%s", err_message.c_str());
        ERR_PRINT("Failed to open database file %s for writing\n", db_file_str.c_str());
        IBDIAGNET_RETURN(1);
    }

    /*
    this->sout_db_csv.open(db_file_str.c_str());
    if (this->sout_db_csv.fail()) {
        ERR_PRINT("Failed to open database file %s for writing\n", db_file_str.c_str());
        IBDIAGNET_RETURN(1);
    }
    */

    //set some stuff in all stages
    for (int i = 0; i < NUM_OF_STAGES; ++i) {
        if(!this->stages[i])
            continue;
        this->stages[i]->SetNumErrsToScreen(&this->num_of_errors_to_screen);
        this->stages[i]->SetBasePath(&this->base_path);
        this->stages[i]->SetCSVOut(&this->csv_out);
    }
    for (list_p_symbols_plugin::iterator it = this->loaded_plugins_list.begin();
            it != this->loaded_plugins_list.end(); ++it) {
        (*it)->p_plugin_obj->SetNumErrsToScreen(&this->num_of_errors_to_screen);
        (*it)->p_plugin_obj->SetBasePath(&this->base_path);
        (*it)->p_plugin_obj->SetCSVOut(&this->csv_out);
    }

    ibdiag_obj.GetIbisPtr()->SetMaxMadsOnWire((u_int16_t)m_max_gmps_on_wire,
                                              (u_int8_t)m_max_smps_on_wire);

    this->ibdiagnet_status = READY;
    IBDIAGNET_RETURN(0);
}


int IBDiagapp::RunAllStages()
{
    IBDIAGNET_ENTER;

    int rc = 1;
    if (this->ibdiagnet_status != READY) {
        ERR_PRINT("Not initialized ibdiagnet class\n");
        IBDIAGNET_RETURN(rc);
    }

    for (int i = 0; i < NUM_OF_STAGES; ++i) {
        if(!this->stages[i])
            continue;
        if (this->stages[i]->IsActive()) {
            if (this->stages[i]->Prepare())
                goto exit;
            if (this->stages[i]->RetrieveInfo())
                goto exit;
            if (this->stages[i]->RunCheck())
                goto exit;
            this->stages[i]->MarkStageAsFinished();
        }
    }

    for (list_p_symbols_plugin::iterator it = this->loaded_plugins_list.begin();
            it != this->loaded_plugins_list.end(); ++it) {
        try {
            Plugin *p_plugin = (*it)->p_plugin_obj;
            if (p_plugin->IsActive()) {
                if (p_plugin->Prepare())
                    goto exit;
                if (p_plugin->RetrieveInfo())
                    goto exit;
                if (p_plugin->RunCheck())
                    goto exit;
                p_plugin->MarkStageAsFinished();
            }
        }
        catch(...) {            // Handle all exceptions
            ERR_PRINT("Plugin %s raised an exception\n",
                      (*it)->library_file_name.c_str());
            goto exit;
        }
    }


    rc = 0;
exit:
    this->PrintSummary();
    PRINT("\n");
    if (rc)
        ERR_PRINT("A fatal error occurred, exiting...\n");
    IBDIAGNET_RETURN(rc);
}

int IBDiagapp::Run(int argc, char **argv)
{

    IBDIAGNET_ENTER;
    //assert to make sure number of elements in bad_direct_route_reasons equal
    //to IBDIAG_BAD_DR_MAX
    COMPILE_TIME_ASSERT(N_ELEMENTS(bad_direct_route_reasons) ==
        IBDIAG_BAD_DR_MAX);

    //init
    if (this->Init()) {
        SCREEN_ERR_PRINT("%s failed to initialize, exiting...\n", argv[0]);
        IBDIAGNET_RETURN(1);
    }

    //get args
    int rc = this->GetOpt(argc, argv);
    if (rc) {
        if (rc == 2)            //2 - OK but need to exit
            IBDIAGNET_RETURN(0);
        IBDIAGNET_RETURN(1);
    }

    //init
    if (this->PostInit()) {
        ERR_PRINT("%s failed to initialize, exiting...\n", argv[0]);
        IBDIAGNET_RETURN(1);
    }

    //Write START_INFO section into db_csv
    rc = DumpStartInfoCSVTable(argc, argv);
    if (rc)
        IBDIAGNET_RETURN(1);

    //run all
    if (this->RunAllStages())
        IBDIAGNET_RETURN(1);

    csv_out.DumpIndexTableCSV();
    IBDIAGNET_RETURN(0);

}

