/*
 * Copyright (c) 2004-2018 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include "ibdiagapp_lc_plugin.h"
#include "ibdiagnet_plugins.h"

extern "C" {
int32_t register_lc_plugin_object(const char *plugin_object_type,
                                  const PF_RegisterParams *params);
int32_t invoke_service(const char * serviceName, void * serviceParams);
}

//public methods
LCPlugin::LCPlugin(IBDiag              *p_ibdiag,
                   const string        &plugin_object_name,
                   IObjectLCPlugin     *x_object,
                   PF_DestroyFunc       plugin_object_destroy_func,
                   LCPluginMgr         *p_plugins_mgr) :
    Plugin(plugin_object_name, p_ibdiag), m_p_plugin_object(x_object),
    m_plugin_object_destroy_func(plugin_object_destroy_func),
    m_p_plugins_mgr(p_plugins_mgr), m_tid(0)
{
    /*
     * this stage is skipped by default
     */
    this->MarkStageAsSkipped();

    //use plugin_object_name, and not name, as name comes with "Plugin: " prefix
    m_plugin_act_option = plugin_object_name + OPTION_PLUGIN_ACT_SUFFIX;
    m_plugin_opt_option = plugin_object_name + OPTION_PLUGIN_OPT_SUFFIX;
}

LCPlugin::~LCPlugin()
{
    if (m_p_plugin_object)
        m_plugin_object_destroy_func(m_p_plugin_object);
}

void LCPlugin::Init()
{
    string plugin_act_option_desc = OPTION_PLUGIN_ACT_DESC_PREFIX + name;
    string plugin_opt_option_desc = OPTION_PLUGIN_OPT_DESC_PREFIX + name;
    string description = LC_PLUGIN_DESC;
    char *option_help = NULL, *plugin_desc = NULL;

    if (!m_p_plugin_object->InvokeService("GetOptionsHelp", &option_help) &&
        option_help) {
        char plugin_help[100] = {'\0'};

        //limit the help, in case plugin forgets to limit it
        strncpy(plugin_help, option_help, sizeof(plugin_help) -1);

        plugin_opt_option_desc += "\n" + string(plugin_help);
    }

    if (!m_p_plugin_object->InvokeService("GetPluginDescription",
                                          &plugin_desc) && plugin_desc) {
        char plugin_desc_str[100] = {'\0'};

        strncpy(plugin_desc_str, plugin_desc, sizeof(plugin_desc_str) - 1);
        description = plugin_desc_str;
    }

    this->AddOptions(m_plugin_act_option, OPTION_PLUGIN_ACT_SHORT,
                     OPTION_PLUGIN_ACT_ARG_DESC, plugin_act_option_desc, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(m_plugin_opt_option, OPTION_PLUGIN_OPT_SHORT,
                     OPTION_PLUGIN_OPT_ARG_DESC, plugin_opt_option_desc, OPTION_DEF_VAL_NULL);
    this->AddDescription(description);
}

int LCPlugin::HandleOption(string opt_name, string value)
{
    if (opt_name == m_plugin_act_option) {
        this->MarkStageAsActive();
        IBDIAGNET_RETURN(0);
    } else if (opt_name == m_plugin_opt_option) {
        string internal_params = value;
        if (!m_p_plugin_object->InvokeService(
            "HandleOption", (void*)(internal_params.c_str())))
                IBDIAGNET_RETURN(0);
    }
    //if we reached here than unknown option
    IBDIAGNET_RETURN(1);
}

int LCPlugin::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    int anyErr = 0;

    addPortsParams_t     params;
    CLEAR_STRUCT(params);

    //start tracking plugin's cout
    ibdmClearInternalLog();

    m_p_plugins_mgr->GetPortsPtr(&params.ports,
                              params.ports_num);
    if (m_p_plugin_object->InvokeService("AddPorts", &params)) {
        ERR_PRINT("Passing the discovered fabric to %s failed.\n", name.c_str());
        anyErr++;
    }
    IBDIAGNET_RETURN(anyErr);
}

int LCPlugin::RetrieveInfo() {
    IBDIAGNET_ENTER;

    m_tid = p_ibdiag->GetIbisPtr()->GetNewTID();
    int res = m_p_plugin_object->InvokeService("RetrieveInfo", &m_tid);
    p_ibdiag->GetIbisPtr()->SetMADsCounter(m_tid);
    if (res)
        IBDIAGNET_RETURN(res);
    IBDIAGNET_RETURN(0);
}

int LCPlugin::RunCheck() {
    IBDIAGNET_ENTER;

    runCheckParams_t params;
    u_int32_t num_lines = 0;
    const char* ptr = NULL;
    stringstream sstream;
    params.num_lines = &num_lines;
    params.results = &ptr;

    if (m_p_plugin_object->InvokeService("RunCheck", &params)) {
        ERR_PRINT("Failed to get results from %s \n", name.c_str());
        IBDIAGNET_RETURN(1);
    }

    this->p_csv_out->DumpStart(name.c_str());
    sstream << *(params.results) << endl;
    this->p_csv_out->WriteLine(sstream.str());
    this->p_csv_out->DumpEnd(name.c_str());

    //get and analyze all the errors that plugin printed
    char *buffer = ibdmGetAndClearInternalLog();
    if (!buffer) {
        ERR_PRINT("Failed to allocate buffer for %s output\n", name.c_str());
        IBDIAGNET_RETURN(1);
    } else {
        u_int32_t errors = num_errors;
        u_int32_t warnings = num_warnings;
        string construct_res(buffer);
        this->ParseIBDMOutput(construct_res);

        if (num_errors > errors) {
            ERR_PRINT("%s finished with errors.\n", name.c_str());
        } else if (num_warnings > warnings) {
            WARN_PRINT("%s finished with warnings.\n", name.c_str());
        }
        LOG_PRINT("%s", buffer);
        free(buffer);
    }
    PRINT("\n");
    IBDIAGNET_RETURN(0);
}

/*****************************************************************************/
/*              The LC (Loosely Coupled) Plugins Manager                     */
/*****************************************************************************/
LCPluginMgr *g_p_lc_plugins_mgr = NULL;

LCPluginMgr::LCPluginMgr() : m_p_ibdiag(NULL)
{
    m_version.major = IBDIAGNET_LC_PLUGIN_API_VERSION_MAJOR;
    m_version.minor = IBDIAGNET_LC_PLUGIN_API_VERSION_MINOR;
}

LCPluginMgr::~LCPluginMgr()
{
    map_plugin_registry_t::iterator it = m_plugins_registry_map.begin();
    map_plugin_registry_t::iterator it_end = m_plugins_registry_map.end();

    for(; it != it_end; it++)
        this->ReleasePlugin(it);
}

int LCPluginMgr::Init(IBDiag *p_ibdiag)
{
    IBDIAGNET_ENTER;
    if (!p_ibdiag)
        IBDIAGNET_RETURN(1);
    m_p_ibdiag = p_ibdiag;

    //update the global pointer to plugins manager
    g_p_lc_plugins_mgr = this;

    IBDIAGNET_RETURN(0);
}

//this function is a hack to store lc (loosely coupled) plugin in the list
//that ibdiagnet holds
void lc_plugin_destroy(Plugin *p)
{
    //do nothing, will release lc (loosely coupled) plugins in lc plugins mgr
    return;
}

int LCPluginMgr::LoadLCPlugin(IN string plugin_name,
                              IN PF_InitFunc init_plugin,
                              OUT symbols_plugin_t *p_res)
{
    PF_PlatformServices params;

    CLEAR_STRUCT(params);
    pl_list_names_t::iterator pl_name_it =
        m_plugin_names.insert(m_plugin_names.end(), plugin_name);
    params.version = this->m_version;
    params.registerObject = register_lc_plugin_object;
    params.invokeService = invoke_service;
    params.plugin_handle = (uintptr_t)&(*pl_name_it);


    //set the current plugin name we are about
    //to register its objects in the init function
    m_plugin_to_register = plugin_name;
    PF_ExitFunc exit_func = init_plugin(&params);
    if (exit_func) {
        //register plugin's exit function
        m_plugins_registry_map[plugin_name].exit_func = exit_func;
        //TBD: create all objects and not just the first
        map_object_str_register_data_t::iterator first_plugin_obj =
            m_plugins_registry_map[plugin_name].objects_data.begin();
        string object_type = first_plugin_obj->first;
        PF_ObjectParams object_params;
        object_params.objectType = object_type.c_str();
        object_params.platformServices = &params;
        IObjectLCPlugin *object =
            (IObjectLCPlugin *)first_plugin_obj->second.
                registry_params.createFunc(&object_params);

        if (!object) {
            //TBD print some error
            this->ReleasePlugin(plugin_name);
            IBDIAGNET_RETURN(1);
        }
        LCPlugin *p_plugin = new LCPlugin(
            this->m_p_ibdiag, object_type,
            object, first_plugin_obj->second.registry_params.destroyFunc,
            this); //pass the plugin the manager
        if (!p_plugin) {
            this->ReleasePlugin(plugin_name);
            //TBD print some error
            IBDIAGNET_RETURN(1);
        }
        //store the plugin in manager
        m_plugins_registry_map[plugin_name].objects_data[object_type].p_lc_plugin_object = p_plugin;
        p_res->p_plugin_obj = p_plugin;
        p_res->create_func = NULL;//already created LCPlugin
        p_res->destroy_func = &lc_plugin_destroy;
    } else
        //remove plugin's registered objects
        this->ReleasePlugin(plugin_name);
    IBDIAGNET_RETURN(0);
}

//release the current plugin, which is currently in init state
void LCPluginMgr::ReleasePlugin(string &plugin_name)
{
    map_plugin_registry_t::iterator it =
        m_plugins_registry_map.find(plugin_name);

    //found
    if (it != m_plugins_registry_map.end())
        this->ReleasePlugin(it);
}

void LCPluginMgr::ReleasePlugin(map_plugin_registry_t::iterator &plugin_it)
{
    lc_plugin_registry_data_t plugin_registry = plugin_it->second;
    map_object_str_register_data_t::iterator obj_it =
        plugin_registry.objects_data.begin();
    for (; obj_it != plugin_registry.objects_data.end(); obj_it++) {
        if (obj_it->second.p_lc_plugin_object) {
            //delete the stored lc (loosely coupled) plugin object
            delete obj_it->second.p_lc_plugin_object;
            obj_it->second.p_lc_plugin_object = NULL;
        }
    }
    plugin_registry.objects_data.clear();
    //call plugin's exit function
    plugin_registry.exit_func();
    //now we can remove the plugin's entry altogether
    m_plugins_registry_map.erase(plugin_it);
}

int LCPluginMgr::AcquireFabricPorts()
{
    IBFabric * p_discovered_fabric = m_p_ibdiag->GetDiscoverFabricPtr();

    if (!m_p_ibdiag->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    CapabilityModule *p_capability_module = m_p_ibdiag->GetCapabilityModulePtr();

    plPortObject_t port_object;
    capability_mask_t mask;

    //pass the fabric
    for (map_str_pnode::iterator nI = p_discovered_fabric->NodeByName.begin();
            nI != p_discovered_fabric->NodeByName.end();
            ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            //error in db
            continue;
        }
        // go over all ports of this node:
        for (u_int32_t i = 1; i < (unsigned)p_curr_node->numPorts + 1; ++i) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            if (!p_curr_port)                       //port not exist
                continue;
            CLEAR_STRUCT(port_object);
            CLEAR_STRUCT(mask.mask);

            port_object.portGuid = p_curr_port->guid_get();
            port_object.nodeGuid = p_curr_node->guid_get();
            //convert ibdm types to plugin types
            switch(p_curr_node->type) {
            case IB_CA_NODE:
                port_object.type = PL_IB_CA_NODE;
                break;
            case IB_SW_NODE:
                port_object.type = PL_IB_SW_NODE;
                break;
            default:
                port_object.type = PL_IB_UNKNOWN_NODE_TYPE;
                break;
            }
            port_object.num = p_curr_port->num;
            port_object.lid = p_curr_port->base_lid;
            //add the peer info
            if (p_curr_port->p_remotePort) {
                port_object.peerPortGuid = p_curr_port->p_remotePort->guid_get();
                port_object.peerNum = p_curr_port->p_remotePort->num;
            }
            if (p_capability_module->GetCapability(p_curr_node,
                                                   true,
                                                   mask) ==
                IBDIAG_SUCCESS_CODE)
                //TBD check sizes
                memcpy(port_object.gmpCapabilityMask,
                       mask.mask, sizeof(port_object.gmpCapabilityMask));

            CLEAR_STRUCT(mask.mask);
            if (p_capability_module->GetCapability(p_curr_node,
                                                   false,
                                                   mask) ==
                IBDIAG_SUCCESS_CODE)
                memcpy(port_object.smpCapabilityMask,
                       mask.mask, sizeof(port_object.smpCapabilityMask));
            string port_name = p_curr_port->getName();
            strncpy(port_object.name,
                    port_name.c_str(),
                    sizeof(port_object.name) - 1);

            m_fabric_ports.push_back(port_object);
        }
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int LCPluginMgr::GetPortsPtr(const plPortObject_t ** pp_fabric_ports,
                             u_int32_t &ports_num)
{
    if (!pp_fabric_ports)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    //check if it's the first time the function is called
    if (m_fabric_ports.empty()) {
        int rc = 0;
        if ((rc = this->AcquireFabricPorts())) {
            ports_num = 0;
            IBDIAGNET_RETURN(rc);
        }
    }
    ports_num = (u_int32_t)m_fabric_ports.size();
    *pp_fabric_ports = ports_num ? &m_fabric_ports[0] : NULL;
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int32_t LCPluginMgr::RegisterLCPluginObject(const char *plugin_object_type,
                                            const PF_RegisterParams *params)
{
    IBDIAGNET_ENTER;

    string str_plugin_object_type = plugin_object_type;
    if (!params) {
        ERR_PRINT("Registering %s failed due to empty registry parameters\n",
                  str_plugin_object_type.c_str());

        IBDIAGNET_RETURN(1);
    }
    //if our version < plugin's, can't serve it
    if (this->m_version.major < params->version.major ||
        (this->m_version.major == params->version.major &&
         this->m_version.minor < params->version.minor))
        //TBD print some error
        IBDIAGNET_RETURN(1);
    //register this data within ibdiagnet plugins
    plugin_object_registry_data_t plugin_object_data;
    CLEAR_STRUCT(plugin_object_data);
    plugin_object_data.registry_params = *params;
    //the  LCPlugin pointer will come later
    plugin_object_data.p_lc_plugin_object = NULL;

    lc_plugin_registry_data_t lc_plugin_data;
    lc_plugin_data.objects_data[str_plugin_object_type] = plugin_object_data;
    lc_plugin_data.exit_func = NULL; //will be registered later
    string register_plugin = m_plugin_to_register;
    if (params->plugin_handle)
        register_plugin = *((string*)(params->plugin_handle));
    //else TBD error

    m_plugins_registry_map[register_plugin] = lc_plugin_data;

    IBDIAGNET_RETURN(0);
}

int32_t LCPluginMgr::InvokeService(const char * serviceName,
                                   void * serviceParams)
{
    IBDIAGNET_ENTER;
    if (!serviceName || !m_p_ibdiag)
        return -1;
    if (!strcmp(serviceName, "madGetSet")) {
        madGetSetParams_t *params = (madGetSetParams_t*)serviceParams;
        IBDIAGNET_RETURN(m_p_ibdiag->GetIbisPtr()->MadGetSet(
            params->lid,
            params->d_qp,
            params->sl,
            params->qkey,
            params->mgmt_class,
            params->method,
            params->attribute_id,
            params->attribute_modifier,
            params->data_offset,
            params->p_class_data,
            params->p_attribute_data,
            params->pack_class_data_func,
            params->unpack_class_data_func,
            params->dump_class_data_func,
            params->pack_attribute_data_func,
            params->unpack_attribute_data_func,
            params->dump_attribute_data_func,
            params->p_clbck_data));
    } else if (!strcmp(serviceName, "MadRecAll")) {
        m_p_ibdiag->GetIbisPtr()->MadRecAll();
    }
    IBDIAGNET_RETURN(0);
}

/*****************************************************************************/
/*        The C functions for LC (loosely coupled) plugin api                */
/*****************************************************************************/
extern "C" {
int32_t register_lc_plugin_object(const char *plugin_object_type,
                                  const PF_RegisterParams *params)
{
    //delegate to plugin manager
    return g_p_lc_plugins_mgr->RegisterLCPluginObject(plugin_object_type,
                                                      params);
}

int32_t invoke_service(const char * serviceName, void * serviceParams)
{
    //delegate to plugin manager
    return g_p_lc_plugins_mgr->InvokeService(serviceName, serviceParams);
}
}
