/*
 * © Copyright (c) 2004-2018 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <string.h>
#include <unistd.h>
extern char *optarg;
extern int optind, opterr, optopt;
#include <getopt.h>
#include <errno.h>

#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h>

#include "ibdiagnet.h"
#include "ibdiagnet_debug_wrap.h"

#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_version.h>

/****************************************************/
/* globals */
/****************************************************/
DebugWrap debug_wrap_opj(DEFAULT_OUTPUT_PATH);


/****************************************************/
/*******************public methods*******************/
/****************************************************/

IBDiagnet::IBDiagnet() : IBDiagapp(DEFAULT_OUTPUT_PATH)
{

}

Stage* IBDiagnet::CreateStage(uint8_t stageIndex) {

    IBDIAGNET_ENTER;
    Stage *stage = NULL;
    switch (stageIndex) {
        case STAGE_IDX_DISCOVERY:
            //discovery stage
            stage = new DiscoveryFabricStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_DISCOVERY);
            break;
        case STAGE_IDX_LIDS:
            //lids stage
            stage = new LidsStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_LIDS);
            break;
        case STAGE_IDX_TOPLOGY_MATCH:
            //topology match stage
            stage = new ToplogtMatchStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_TOPLOGY_MATCH);
            break;
        case STAGE_IDX_LINKS:
            //links stage
            stage = new LinksStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_LINKS);
            break;
        case STAGE_IDX_SM:
            //sm stage
            stage = new SMStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_SM);
            break;
        case STAGE_IDX_PM:
            //pm stage
            stage = new PMStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_PM);
            break;
       case STAGE_IDX_ROUTING:
            //routing stage
            stage = new RoutingStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_ROUTING);
            break;
        case STAGE_IDX_NODES_INFO:
            //nodes info stage
            stage = new NodesInfoStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_NODES_INFO);
            break;
        case STAGE_IDX_SPEED_WIDTH:
            //speed / width stage
            stage = new Speed_WidthStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_SPEED_WIDTH);
            break;
        case STAGE_IDX_PKEY:
            //pkey stage
            stage = new PKeyStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_PKEY);
            break;
        case STAGE_IDX_AGUID:
            //aguid stage
            stage = new AGUIDStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_AGUID);
            break;
        case STAGE_IDX_MLNX_CNTRS:
            //Mellanox Counters stage
            stage = new MlnxCntrsStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n",
                          STAGE_NAME_MLNX_CNTRS);
            break;
        case STAGE_IDX_VIRTUALIZATION:
            //Virtualiztion stage
            stage = new VirtualizationStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n",
                          STAGE_NAME_VIRTUALIZATION);
            break;
        case STAGE_IDX_TEMPSENSING:
            //Temperature stage
            stage = new TemperatureSensing(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n",
                          STAGE_NAME_TEMPSENSING);
            break;
        case STAGE_IDX_ROUTERS:
            stage = new RoutersStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n",
                          STAGE_NAME_ROUTERS);
            break;
        case STAGE_IDX_SHARP:
           //Sharp stage
           stage = new SharpStage(&this->ibdiag_obj);
           if (!stage)
               ERR_PRINT("Failed to allocate %s stage\n",
                         STAGE_NAME_SHARP);
           break;
        case STAGE_IDX_QOS:
           //QoS stage
           stage = new QoSStage(&this->ibdiag_obj);
           if (!stage)
               ERR_PRINT("Failed to allocate %s stage\n",
                         STAGE_NAME_QOS);
           break;
        default:
            break;
    }

    IBDIAGNET_RETURN(stage);
}

IBDiagnet::~IBDiagnet()
{
    IBDIAGNET_ENTER;

    IBDIAGNET_RETURN_VOID;
}

//Returns: 0 - OK / 1 - parsing error / 2 - OK but need to exit / 3 - some other error
int IBDiagnet::HandleOption(string name, string value)
{
    IBDIAGNET_ENTER;

    if (value == OPTION_DEF_VAL_NULL)
        IBDIAGNET_RETURN(0);

    int res = IBDiagapp::HandleOption(name, value);

    /* if res is 1 so it's unknown stage to ibdiagapp, need to
     * check for ibdiagnet stages */

    bool bool_flag_val = true;

    if (res == 1) {
        if (name == OPTION_CREATE_CONFIG) {
            this->LoadPlugins();
            this->generated_config_file = value;
            if (this->CreateConfigFile(this->generated_config_file)) {
                IBDIAGNET_RETURN(3);
            }
            IBDIAGNET_RETURN(2);           //exit with no error
        } else if (name == OPTION_VLR) {
            ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->
                SetPSLDumpFile(value);
            ((RoutingStage *)this->stages[STAGE_IDX_ROUTING])->UseVLR();
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_SA_DUMP) {
            ((RoutingStage *)this->stages[STAGE_IDX_ROUTING])->
                SetSADumpFile(value);
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_ROUTING_OPT) {
            if (!this->stages[STAGE_IDX_ROUTING]->IsActive()){
                ERR_PRINT_ARGS("Use %s option only after %s option.\n",
                        OPTION_ROUTING_OPT, OPTION_ROUTING);
                IBDIAGNET_RETURN(3);
            }
            string opt_str = "," + value + ",";
            if (opt_str.find(",vs,") != std::string::npos) {
                ((RoutingStage *)this->stages[STAGE_IDX_ROUTING])->UseVSRoute();
                INFO_PRINT("Using \"vs\" (vendor specific) routing option "
                           "(deprecated, enabled by default).\n");
            }
            if (opt_str.find(",far,") != std::string::npos) {
                ((RoutingStage *)this->stages[STAGE_IDX_ROUTING])->DumpFullAr();
                INFO_PRINT("Using \"far\" (dump full ar lft) routing option "
                           "(deprecated, enabled by default).\n");
            }
            if (opt_str.find(",skip_vs,") != std::string::npos) {
                ((RoutingStage *)this->stages[STAGE_IDX_ROUTING])->SkipVSRoute();
                INFO_PRINT("Skipping \"vs\" (vendor specific) routing option.\n");
            }
            if (opt_str.find(",skip_far,") != std::string::npos) {
                ((RoutingStage *)this->stages[STAGE_IDX_ROUTING])->SkipDumpFullAr();
                INFO_PRINT("Skipping \"far\" (dump full arlft) routing option.\n");
            }
            if (opt_str.find(",mcast,") != std::string::npos) {
                ((RoutingStage *)this->stages[STAGE_IDX_ROUTING])->
                    CheckMcastCredLoop();
                INFO_PRINT("Using \"mcast\" (mcast credit loop check) routing \
                            option.\n");
            }
            if (opt_str.find(",check_sl,") != std::string::npos) {
                ((RoutingStage *)this->stages[STAGE_IDX_ROUTING])->
                    SetCheckSL();
                INFO_PRINT("Using \"check_sl\" routing option.\n");
            }
            if (opt_str.find(",sl=") != std::string::npos) {
                size_t offset = opt_str.find(",sl=") + strlen(",sl=");
                u_int8_t sl = (u_int8_t)strtoul(
                    opt_str.substr(offset).c_str(), NULL, 0);
                if (sl >= IB_NUM_SL) {
                    ERR_PRINT_ARGS("Illegal value for \'--%s\' option: sl=%d\n",
                            OPTION_ROUTING_OPT, sl);
                } else {
                    ((RoutingStage *)this->stages[STAGE_IDX_ROUTING])->
                        SetDefaultSL(sl);
                    INFO_PRINT("Using \"sl=%d\" routing option.\n", sl);

                }
            }
            if (opt_str.find(",rn,") != std::string::npos) {
                ((RoutingStage *)this->stages[STAGE_IDX_ROUTING])->DumpRNDada();
                INFO_PRINT("Using \"rn\" (routing notification) routing option.\n");
            }
            if (opt_str.find(",crnc,") != std::string::npos) {
                ((RoutingStage *)this->stages[STAGE_IDX_ROUTING])->ClearRNCounters();
                INFO_PRINT("Using \"crnc\" (clear rn counters) routing option.\n");
            }
            if (opt_str.find(",drnc,") != std::string::npos) {
                ((RoutingStage *)this->stages[STAGE_IDX_ROUTING])->DumpRNCounters();
                INFO_PRINT("Using \"drnc\" (dump rn counters) routing option.\n");
            }
            if (opt_str.find(",dump_only,") != std::string::npos) {
                ((RoutingStage *)this->stages[STAGE_IDX_ROUTING])->DumpOnly();
                INFO_PRINT("Using \"dump_only\" routing option.\n");
            }
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_ROUTING) {
            this->ParseBoolValue(value, bool_flag_val);
            ((RoutingStage *)this->stages[STAGE_IDX_ROUTING])->
                MarkStage(bool_flag_val);
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_FATTREE) {
            this->ParseBoolValue(value, bool_flag_val);
            ((RoutingStage *)this->stages[STAGE_IDX_ROUTING])->SetFatTree(bool_flag_val);
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_TOPO_WRITE) {
            ((ToplogtMatchStage *)this->stages[STAGE_IDX_TOPLOGY_MATCH])->
                SetOutputTopoFile(value);
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_TOPO_MATCH_FILE) {
            ((ToplogtMatchStage *)this->stages[STAGE_IDX_TOPLOGY_MATCH])->
                SetInputTopoFile(value);
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_TOPO_IBNL_DIR) {
            ((ToplogtMatchStage *)this->stages[STAGE_IDX_TOPLOGY_MATCH])->
                SetTopoIbnlDir(value);
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_OUTPATH) {
            IBDIAGNET_RETURN(this->CreateOutputPath(value));
        } else if (name == OPTION_SCOPE) {
            ((DiscoveryFabricStage *)this->stages[STAGE_IDX_DISCOVERY])->
                SetScopePortGuidFile(value);
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_EXCLUDE_SCOPE) {
            ((DiscoveryFabricStage *)this->stages[STAGE_IDX_DISCOVERY])->
                SetExcludeScopePortGuidFile(value);
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_SHARP) {
            this->ParseBoolValue(value, bool_flag_val);
            ((SharpStage *)this->stages[STAGE_IDX_SHARP])->MarkStage(bool_flag_val);
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_SHARP_OPT) {
            if (!this->stages[STAGE_IDX_SHARP]->IsActive()){
                ((SharpStage *)this->stages[STAGE_IDX_SHARP])->MarkStageAsActive();
            }
            string opt_str = "," + value + ",";
            if (opt_str.find(",csc,") != std::string::npos) {
                ((SharpStage *)this->stages[STAGE_IDX_SHARP])->SetClearCounters();
                INFO_PRINT("Using \"csc\" (clear sharp counters) sharp option.\n");
            }
            if (opt_str.find(",dsc,") != std::string::npos) {
                ((SharpStage *)this->stages[STAGE_IDX_SHARP])->SetDumpCounters();
                INFO_PRINT("Using \"dsc\" (dump sharp counters) sharp option.\n");
            }
            IBDIAGNET_RETURN(0);

        } else if (name == OPTION_LOAD_FROM_FILE) {
            ((DiscoveryFabricStage *)this->stages[STAGE_IDX_DISCOVERY])->SetLoadFromFilePath(value);
            IBDIAGNET_RETURN(0);
        }
    }

    IBDIAGNET_RETURN(res);
}

int IBDiagnet::Init()
{
    IBDIAGNET_ENTER;
    int res;

    skip_obj_t obj0 = {SKIP_DUP_GUIDS_VALUE, "dup_guids"};
    skip_options_list.push_back(obj0);
    skip_obj_t obj1 = {SKIP_DUP_NODE_DESC_VALUE, "dup_node_desc"};
    skip_options_list.push_back(obj1);
    skip_obj_t obj2 = {SKIP_STAGE_LIDS_VALUE, "lids"};
    skip_options_list.push_back(obj2);
    skip_obj_t obj3 = {SKIP_STAGE_SM_VALUE, "sm"};
    skip_options_list.push_back(obj3);
    skip_obj_t obj4 = {SKIP_STAGE_NODES_INFO_VALUE, "nodes_info"};
    skip_options_list.push_back(obj4);
    skip_obj_t obj5 = {SKIP_STAGE_PKEY_CHECK, "pkey"};
    skip_options_list.push_back(obj5);
    skip_obj_t obj6 = {SKIP_STAGE_AGUID_CHECK, "aguid"};
    skip_options_list.push_back(obj6);

    res = IBDiagapp::Init();
    if (res)
        IBDIAGNET_RETURN(res);

    this->AddDescription(IBDIAGNET_TOOL_DESC);
    this->AddOptions(OPTION_VLR, OPTION_VLR_SHORT, OPTION_VLR_ARG_DESC, OPTION_VLR_DESC,
                     OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_ROUTING, OPTION_ROUTING_SHORT, OPTION_ROUTING_ARG_DESC,
                     OPTION_ROUTING_DESC, OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_ROUTING_OPT, OPTION_ROUTING_OPT_SHORT, OPTION_ROUTING_OPT_ARG_DESC,
                     OPTION_ROUTING_OPT_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_SA_DUMP, OPTION_SA_DUMP_SHORT, OPTION_SA_DUMP_ARG_DESC,
                     OPTION_SA_DUMP_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_FATTREE, OPTION_FATTREE_SHORT, OPTION_FATTREE_ARG_DESC,
                     OPTION_FATTREE_DESC, OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_SCOPE, OPTION_SCOPE_SHORT, OPTION_SCOPE_ARG_DESC,
                     OPTION_SCOPE_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_EXCLUDE_SCOPE, OPTION_EXCLUDE_SCOPE_SHORT,
                     OPTION_EXCLUDE_SCOPE_ARG_DESC, OPTION_EXCLUDE_SCOPE_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_TOPO_WRITE, OPTION_TOPO_WRITE_SHORT, OPTION_TOPO_WRITE_ARG_DESC,
                     OPTION_TOPO_WRITE_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_TOPO_MATCH_FILE, OPTION_TOPO_MATCH_FILE_SHORT,
                     OPTION_TOPO_MATCH_FILE_ARG_DESC, OPTION_TOPO_MATCH_FILE_DESC,
                     OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_TOPO_IBNL_DIR, OPTION_TOPO_IBNL_DIR_SHORT,
                     OPTION_TOPO_IBNL_DIR_ARG_DESC, OPTION_TOPO_IBNL_DIR_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_OUTPATH, OPTION_OUTPATH_SHORT, OPTION_OUTPATH_ARG_DESC,
                     OPTION_OUTPATH_DESC, DEFAULT_OUTPUT_PATH);
    this->AddOptions(OPTION_SHARP, OPTION_SHARP_SHORT, OPTION_SHARP_ARG_DESC, OPTION_SHARP_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_SHARP_OPT, OPTION_SHARP_OPT_SHORT, OPTION_SHARP_OPT_ARG_DESC,
                     OPTION_SHARP_OPT_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_LOAD_FROM_FILE, OPTION_LOAD_FROM_FILE_SHORT,
                     OPTION_LOAD_FROM_FILE_ARG_DESC, OPTION_LOAD_FROM_FILE_DESC,
                     OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_CREATE_CONFIG, OPTION_CREATE_CONFIG_SHORT,
                     OPTION_CREATE_CONFIG_ARG_DESC, OPTION_CREATE_CONFIG_DESC, OPTION_DEF_VAL_NULL,
                     IBIS_CMD_NO_CONF_FILE);
    IBDIAGNET_RETURN(0);
}


#define N_ELEMENTS(array) sizeof(array)/sizeof(array[0])

#define COMPILE_TIME_ASSERT(pred)            \
    switch(0){case 0:case pred:;}

/****************************************************/
int main (int argc, char **argv)
{
    IBDiagnet ibdiagnet_obj;
    //Any created files will have mode
    //S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH (0666)
     umask(0);
    IBDIAGNET_RETURN(ibdiagnet_obj.Run(argc, argv));
}

