/*
 *  Copyright (c) 2004-2013 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <dlfcn.h>
#include <list>
#include <map>

#include "ibdiagapp_lc_plugin.h"
#include "ibdiagnet_plugins.h"


/******************************************************/
#if !defined (_WIN32) && !defined(_WIN64)   /* Linux */
    #define PLUGIN_SURFIX ".so"
    #define PLUGIN_SURFIX_LEN 3
#else   /* Windows */
    #define PLUGIN_SURFIX ".dll"
    #define PLUGIN_SURFIX_LEN 4
#endif

#define PLUGINS_ENV "IBDIAGNET_PLUGINS_PATH"

map_str_str plugins_db;


/******************************************************/
static list_str get_plugins_paths()
{
    IBDIAGNET_ENTER;

    list_str dirs;

    char *manual_dirs = getenv(PLUGINS_ENV);
    if (manual_dirs != NULL) {
        string delimiters(":, ");
        string str = string(manual_dirs);
        // Skip delimiters at beginning.
        string::size_type last_pos = str.find_first_not_of(delimiters, 0);
        // Find first "non-delimiter".
        string::size_type pos = str.find_first_of(delimiters, last_pos);

         while (string::npos != pos || string::npos != last_pos) {
             // Found a token, add it to the vector.
             string curr_dir = str.substr(last_pos, pos - last_pos);
             for (list_str::iterator it = dirs.begin();
                     it != dirs.end(); ++it)
                 if ((*it) == curr_dir)
                     goto next_dir;
             dirs.push_back(curr_dir);

next_dir:
             // Skip delimiters.  Note the "not_of"
             last_pos = str.find_first_not_of(delimiters, pos);
             // Find next "non-delimiter"
             pos = str.find_first_of(delimiters, last_pos);
         }
    }

#ifdef PLUGINS_DIR
    dirs.push_back(PLUGINS_DIR);
#endif

    IBDIAGNET_RETURN(dirs);
}


/******************************************************/
string get_plugins_paths_str()
{
    IBDIAGNET_ENTER;
    list_str dirs = get_plugins_paths();
    string res = "";
    for (list_str::iterator it = dirs.begin();
            it != dirs.end(); ++it) {
        res += (*it);
        res += "\n";
    }
    res += "(You can specify more paths to be looked in with \"";
    res += PLUGINS_ENV;
    res += "\" env variable)\n";
    IBDIAGNET_RETURN(res);
}

string get_plugin_path_str(string plugin_name)
{
    IBDIAGNET_ENTER;
    string plugin_path = "";
    map_str_str::iterator it = plugins_db.find(plugin_name);
    if (it != plugins_db.end())
        plugin_path = (*it).second;

    IBDIAGNET_RETURN(plugin_path);
}

int get_plugins_names(OUT list_str& names,
        OUT string& errors)
{
    IBDIAGNET_ENTER;
    errors = "";

    // create db in first time
    if (plugins_db.empty()) {
        // get plugins locations
        list_str dirs = get_plugins_paths();

        // get plugins files tupes
        list_str files_types;
        files_types.push_front(PLUGIN_SURFIX);

        //find all plugins
        list_str res;
        for (list_str::iterator it = dirs.begin(); it != dirs.end(); ++it) {
            res = get_dir_files((*it), files_types);

            for (list_str::iterator it2 = res.begin(); it2 != res.end(); ++it2) {
                string lib_name = (*it) + SEPERATOR + (*it2);
                string plugin_name = (*it2).erase((*it2).size() - PLUGIN_SURFIX_LEN);

                if ((plugins_db.find(plugin_name) != plugins_db.end()) &&
                        (plugins_db[plugin_name] != lib_name)) {

                    errors += "Detected different plugin libraries with the same name, using the first one:\n";
                    errors += "\t1. ";
                    errors += plugins_db[plugin_name];
                    errors += "\n";
                    errors += "\t2. ";
                    errors += lib_name;
                    errors += "\n";

                    continue;
                }

                plugins_db[plugin_name] = lib_name;
            }
            res.clear();
        }
        files_types.clear();
    }

    names.clear();
    for (map_str_str::iterator it = plugins_db.begin();
            it != plugins_db.end(); ++it)
        names.push_back((*it).first);

    IBDIAGNET_RETURN(0);
}

/*****************************************************************************/

symbols_plugin_t * load_plugin(IN string plugin_name,
        IN IBDiag *p_ibdiag,
        OUT string& errors)
{
    IBDIAGNET_ENTER;
    const char *dlsym_error;
    errors = "";
    PF_InitFunc init_plugin = NULL;

    symbols_plugin_t *p_res = new symbols_plugin_t;
    if (!p_res) {
        errors += "Failed to allocate symbols_plugin_t";
        IBDIAGNET_RETURN(NULL);
    }
    p_res->library_file_name = plugins_db[plugin_name];
    dlerror();          //clean previous errors

    /*
     * load the library
     */
    //printf("\n%s\n\n", p_res->library_file_name.c_str());
    p_res->p_library_obj = dlopen(p_res->library_file_name.c_str(), RTLD_LOCAL  | RTLD_LAZY);
    dlsym_error = dlerror();
    if (dlsym_error) {
        errors += "Failed to load library, err=\"";
        errors += dlsym_error;
        errors += "\"";
        goto error_delete;
    }

    /*
     * load the symbols
     */

    //try first if it's a lc (loosely coupled) plugin
    init_plugin = (PF_InitFunc)dlsym(p_res->p_library_obj, "PF_initPlugin");
    dlsym_error = dlerror();
    if (!dlsym_error) {
        //it's a new 'loosely coupled' plugin
        if (g_p_lc_plugins_mgr->LoadLCPlugin(plugin_name, init_plugin,
                                             p_res)) {
            errors += ("Failed to allocate loosely coupled plugin object"
                       " for plugin " + plugin_name);
            goto error_unload_lib;
        }
    } else {
        errors += "Failed to load symbol PF_initPlugin, err=\"";
        errors += dlsym_error;
        errors += "\"\n";

        //now check if it's an old type of plugin
        p_res->create_func = (create_plugin_t *)dlsym(p_res->p_library_obj, "create");
        dlsym_error = dlerror();
        if (dlsym_error) {
            errors += "Failed to load symbol create, err=\"";
            errors += dlsym_error;
            errors += "\"";
            goto error_unload_lib;
        }
        p_res->destroy_func = (destroy_plugin_t *)dlsym(p_res->p_library_obj, "destroy");
        dlsym_error = dlerror();
        if (dlsym_error) {
            errors += "Failed to load symbol destroy, err=\"";
            errors += dlsym_error;
            errors += "\"";
            goto error_unload_lib;
        }
        //we got here, then it's an old type of plugin, let's clean any error
        dlerror();
        /*
         * create an instance of the plugin
         */
        p_res->p_plugin_obj = p_res->create_func(p_ibdiag);
        if (!p_res->p_plugin_obj) {
            errors += "Failed to allocate plugin obj";
            goto error_unload_lib;
        }
    }
    /*
     * call init
     */
    p_res->p_plugin_obj->Init();
    IBDIAGNET_RETURN(p_res);


error_unload_lib:
    dlclose(p_res->p_library_obj);
error_delete:
    delete p_res;
    IBDIAGNET_RETURN(NULL);
}


void unload_plugin(IN symbols_plugin_t *p_plugin_symbols)
{
    IBDIAGNET_ENTER;

    /*
     * destroy the plugin instance
     */
    if (p_plugin_symbols->p_plugin_obj) {
        p_plugin_symbols->destroy_func(p_plugin_symbols->p_plugin_obj);
        p_plugin_symbols->p_library_obj = NULL;
    }

    /*
     * unload the library
     */
    if (p_plugin_symbols->p_library_obj) {
        dlclose(p_plugin_symbols->p_library_obj);
        p_plugin_symbols->p_library_obj = NULL;
    }

    /*
     * delete the plugin symbols object
     */
    delete p_plugin_symbols;
    p_plugin_symbols = NULL;

    IBDIAGNET_RETURN_VOID;
}

