/*
 *  Copyright (c) 2004-2019 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <stdio.h>
#include <math.h>
#include <signal.h>
#include <map>
#include <list>
#include <iostream>
#include <errno.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>

#include "infiniband/ibdm/TopoMatch.h"
#include "infiniband/ibdm/SysDef.h"
#include <infiniband/ibdm/CredLoops.h>

#include "ibdiagnet_stages.h"


/****************************************************/
/* globals */
static bool g_vs_cap_gmp_run = false;

/****************************************************/
int check_if_local_link_is_down(IN IBDiag *p_ibdiag,
        OUT bool& result)
{
    IBDIAGNET_ENTER;
    u_int8_t local_port_state;
    if (p_ibdiag->GetLocalPortState(local_port_state)) {
        ERR_PRINT("%s\n", p_ibdiag->GetLastError());
        IBDIAGNET_RETURN(1);
    }

    switch (local_port_state) {
    case IB_PORT_STATE_DOWN: {
        result = true;
        break;
    }
    default: {
        result = false;
        break;
    }
    }
    IBDIAGNET_RETURN(0);
}


/******************************************************/
/////
//DiscoveryPath public methods:
/////
DiscoveryPathStage::DiscoveryPathStage(IBDiag *p_ibdiag) :
        DiscoveryStage(p_ibdiag)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int DiscoveryPathStage::DiscoverFabric()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    stringstream ss;

    rc = p_ibdiag->DiscoverPath(NULL, this->max_hops,
                                this->src_lid, this->dest_lid, this->dr_path,
                                &ss);
    if (rc)
        IBDIAGNET_RETURN(rc);

    PRINT("%s\n",ss.str().c_str());

    IBDIAGNET_RETURN(0);
}

int DiscoveryPathStage::Prepare()
{
    IBDIAGNET_ENTER;
    DiscoveryStage::Prepare();
    if (!this->to_check_dup_guids) {        //skipped duplicated guids detection
        this->p_ibdiag->GetRefToCheckDupGuids() = false;
    }
    IBDIAGNET_RETURN(0);
}

/******************************************************/
/////
//DiscoveryFabric private methods:
/////
int DiscoveryFabricStage::PostDiscover()
{
    IBDIAGNET_ENTER;
    string parser_output;
    int rc;

    string scope_file = this->scope_port_guid_file;
    u_int32_t errors = this->num_errors;
    u_int32_t warnings = this->num_warnings;

    //work on guids that are in a scope
    if (!scope_file.empty() &&
        access(scope_file.c_str(), F_OK) != -1) {
        PRINT("\n");
        INFO_PRINT("Load Scope file (%s).\n",
                   scope_file.c_str());
        rc = p_ibdiag->ParseScopePortGuidsFile(
                            scope_file.c_str(),
                            parser_output,
                            true);

        LOG_PRINT("%s\n", parser_output.c_str());

        //if critical errors found, just exit
        if (rc) {
            ERR_PRINT("Failed to parse Scope file (%s).\n",
                      scope_file.c_str());
            ++this->num_errors;
            IBDIAGNET_RETURN(1);
        }
        //parse number of errors, warnings
        this->ParseIBDMOutput(parser_output);
        if (this->num_errors > errors) {
            ERR_PRINT("Parsing the scope file finished with errors.\n");
        } else if (this->num_warnings > warnings) {
            WARN_PRINT("Parsing the scope file finished with warnings.\n");
        }
    }
    string exclude_scope_file = this->exclude_scope_port_guid_file;
    //init for exclude scope
    parser_output.clear();
    errors = this->num_errors;
    warnings = this->num_warnings;

    //work on guids that are excluded from scope
    if (!exclude_scope_file.empty() &&
        access(exclude_scope_file.c_str(), F_OK) != -1) {
        PRINT("\n");
        INFO_PRINT("Load Exclude Scope file (%s).\n",
                   exclude_scope_file.c_str());
        rc = p_ibdiag->ParseScopePortGuidsFile(
                            exclude_scope_file.c_str(),
                            parser_output,
                            false); //not to include, but to exclude
        LOG_PRINT("%s\n", parser_output.c_str());

        //if critical errors found, just exit
        if (rc) {
            ERR_PRINT("Failed to parse Exclude Scope file (%s).\n",
                      exclude_scope_file.c_str());
            ++this->num_errors;
            IBDIAGNET_RETURN(1);
        }
        //parse number of errors, warnings
        this->ParseIBDMOutput(parser_output);
        if (this->num_errors > errors) {
            ERR_PRINT("Parsing the exclude scope file finished"
                      " with errors.\n");
        } else if (this->num_warnings > warnings) {
            WARN_PRINT("Parsing the exclude scope file finished"
                       " with warnings.\n");
        }

    }
    IBDIAGNET_RETURN(0);
}

/////
//DiscoveryFabric public methods:
/////
DiscoveryFabricStage::DiscoveryFabricStage(IBDiag *p_ibdiag) :
        DiscoveryStage(p_ibdiag)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int DiscoveryFabricStage::DiscoverFabric()
{

    IBDIAGNET_ENTER;

    int rc = 0;

    if (load_from_file) {
        rc = p_ibdiag->DiscoverFabricFromFile(load_from_file_path, progress_bar_discovery);
    } else {
        rc = p_ibdiag->DiscoverFabric(progress_bar_discovery, this->max_hops);
    }
    IBDIAGNET_RETURN(rc);
}

int DiscoveryFabricStage::Prepare()
{
    IBDIAGNET_ENTER;
    DiscoveryStage::Prepare();
    if (!this->to_check_dup_guids) {        //skipped duplicated guids detection
        WARN_PRINT("Advanced %s skipped (Only Basic detection will be performed)\n",
                CHECK_NAME_DUPLICATED_GUIDS);
        ++this->num_warnings;
        this->p_ibdiag->GetRefToCheckDupGuids() = false;
    }
    IBDIAGNET_RETURN(0);
}

/******************************************************/
/////
//Discovery private methods:
/////
void DiscoveryStage::PrintDuplicatedGuidsDetectErrors()
{
    IBDIAGNET_ENTER;
    if (!this->p_ibdiag->GetRefToDupGuidsDetectionErrors().size())
        IBDIAGNET_RETURN_VOID;
    LOG_PRINT("\n");
    LOG_WARN_PRINT("Duplicated GUIDs detection finished with some warnings:\n");
    for (list_string::iterator eI = this->p_ibdiag->GetRefToDupGuidsDetectionErrors().begin();
            eI != this->p_ibdiag->GetRefToDupGuidsDetectionErrors().end(); ++eI) {
        //will be printed to both screen + log
        LOG_WARN_PRINT("%s", (*eI).c_str());
        ++this->num_warnings;
    }
    LOG_PRINT("\n");
    IBDIAGNET_RETURN_VOID;
}

/////
//Discovery public methods:
/////
DiscoveryStage::DiscoveryStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_DISCOVERY, p_ibdiag),
        device_name(""), port_num(0), port_guid(0),
        max_hops(IBDIAG_MAX_HOPS), to_check_dup_guids(true), to_check_dup_node_desc(true),
        to_check_vs_cap_smp(true), to_check_vs_cap_gmp(true), to_dump_fec_mode(false),
        initiliazed_with_device(false), initiliazed_with_guid(false), no_mepi(false),
        output_capability_file(""), output_full_capability_file(""),
        input_capability_file(""), show_ports_data_db2(true), dump_lst_with_lmc(false),
        node_name_map_file(""), p_mkey_manager(NULL), load_from_file(false)
{
    IBDIAGNET_ENTER;
    //check if we have environment variable with path
    //to capability mask configuration file
    char *capability_file = getenv(CAPABILITY_FILE_ENV);

    if (capability_file)
        input_capability_file = capability_file;

    char *map_file = getenv(MAP_FILE_ENV);

    //check if we have environment variable with path
    //to node name map file
    if (map_file)
        node_name_map_file = map_file;

    IBDIAGNET_RETURN_VOID;
}

int DiscoveryStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    //init - set port
    if (this->initiliazed_with_device && initiliazed_with_guid) {
        WARN_PRINT("Local GUID were given together with Device name / Port number, initializing only by GUID\n");
        ++this->num_warnings;
    }
    if (no_mepi) 
        this->p_ibdiag->SetNoMEPI();

    if (!load_from_file) { // in load_from_file we set the device and port later
        if (initiliazed_with_guid) {    //init by guid
            if (this->p_ibdiag->SetPort(this->port_guid)) {
                ERR_PRINT("Failed to initialize\n");
                ++this->num_errors;
                IBDIAGNET_RETURN(1);
            }
        } else {    //init by device - this is also default
            if (this->p_ibdiag->SetPort(this->device_name.c_str(), this->port_num)) {
                ERR_PRINT("Failed to initialize\n");
                ++this->num_errors;
                IBDIAGNET_RETURN(1);
            }
        }
    }

    if (!this->to_check_dup_node_desc) {    //skipped duplicated node description detection
        WARN_PRINT("%s skipped\n", CHECK_NAME_DUPLICATED_NODE_DESC);
        ++this->num_warnings;
    }
    if (!this->to_check_vs_cap_smp) {    //skipped Vendor Specific SMP capability mask inquiry
        WARN_PRINT("%s skipped\n", CHECK_VS_CAP_SMP);
        ++this->num_warnings;
    }
    if (!this->to_check_vs_cap_gmp) {    //skipped Vendor Specific GMP capability mask inquiry
        WARN_PRINT("%s skipped\n", CHECK_VS_CAP_GMP);
        ++this->num_warnings;
    }

    //try to read Node Name Map file
    if (!this->node_name_map_file.empty() &&
        access(this->node_name_map_file.c_str(), F_OK) != -1) {
        string parser_output = string("");
        INFO_PRINT("Load node name map file (%s).\n",
                   node_name_map_file.c_str());
        int rc = p_ibdiag->ParseNodeNameMapFile(node_name_map_file.c_str(),
                                                parser_output);
        LOG_PRINT("%s\n", parser_output.c_str());

        if (rc) {
            ERR_PRINT("Failed to parse Node Name Map file (%s).\n",
                      node_name_map_file.c_str());
            ++this->num_errors;
            IBDIAGNET_RETURN(1);
        }

        u_int32_t errors = this->num_errors;
        u_int32_t warnings = this->num_warnings;
        //parse number of errors, warnings
        this->ParseIBDMOutput(parser_output);
        if (this->num_errors > errors) {
            ERR_PRINT("Parsing the Node Name Map file finished " \
                      "with errors.\n");
        } else if (this->num_warnings > warnings) {
            WARN_PRINT("Parsing the Node Name Map file " \
                       "finished with warnings.\n");
        }
    }


    IBDIAGNET_RETURN(0);
}

int DiscoveryStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;

    int rc2 = DiscoverFabric();
    SCREEN_PRINT("\n");

    int rc = this->AnalyzeCheckResults(p_ibdiag->GetErrors(),
                CHECK_FABRIC_DISCOVER,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                false);

    if (rc) {
        /*
        PRINT("\n");
        ERR_PRINT("Failed to discover fabric, err=%s\n",
                this->p_ibdiag->GetLastError());
        ++this->num_errors;
        */
        if (this->to_check_dup_guids)   //maybe we failed because duplicated GUIDs detection had errors
            this->PrintDuplicatedGuidsDetectErrors();

        IBDIAGNET_RETURN(1);
    }

    INFO_PRINT("Discovered %u nodes (%u Switches & %u CA-s).\n",
            (*this->p_ibdiag->GetDiscoverProgressBarNodesPtr()).nodes_found,
            (*this->p_ibdiag->GetDiscoverProgressBarNodesPtr()).sw_found,
            (*this->p_ibdiag->GetDiscoverProgressBarNodesPtr()).ca_found);

    if (this->p_mkey_manager) {
        if (p_ibdiag->GetIbisPtr() && p_ibdiag->GetIbisPtr()->GetMKeyManager())
        {
            if (!p_mkey_manager->getConstMKey()) {
                CreateLid2GuidDB();
                p_mkey_manager->setLidToNodeGuidMap(lid2guid_map);
            }
        }
        else
            ERR_PRINT("can't get MKeyManager Ptr\n");
    }
    //this is essential part of discovery, we can't continue without it
    //(example, scope failed)
    if (PostDiscover())
        IBDIAGNET_RETURN(1);

    string parser_output;

    if(!psl_dump_file.empty()) {
        parser_output = "";
        INFO_PRINT("Load PSL file (%s).\n", psl_dump_file.c_str());

        rc = p_ibdiag->ParsePSLFile(psl_dump_file, parser_output);
        LOG_PRINT("%s\n", parser_output.c_str());

        if (rc) {
            ERR_PRINT("Failed to parse PSL file (%s).\n", psl_dump_file.c_str());
            ++this->num_errors;
            IBDIAGNET_RETURN(1);
        }
    }

    //try to read capability mask configuration file
    if (!this->input_capability_file.empty() &&
        access(this->input_capability_file.c_str(), F_OK) != -1) {
        parser_output = "";

        INFO_PRINT("Load Capability Mask Configuration file (%s).\n",
                   this->input_capability_file.c_str());

        rc = p_ibdiag->ParseCapabilityMaskFile(
            this->input_capability_file.c_str(), parser_output);
        LOG_PRINT("%s\n", parser_output.c_str());

        if (rc) {
            ERR_PRINT("Failed to parse Capability mask configuration file (%s).\n",
                      this->input_capability_file.c_str());
            ++this->num_errors;
        }
    }

    //check if we need to write the capability masks file
    if (!this->output_capability_file.empty()) {
        parser_output = "";
        if (p_ibdiag->DumpCapabilityMaskFile(
            this->output_capability_file.c_str(), parser_output)) {
            WARN_PRINT("Writing Capability Masks file to %s failed\n", this->output_capability_file.c_str());
            ++this->num_errors;
        } else
            this->AddGeneratedFileName("Capability Masks file",  this->output_capability_file.c_str());
        LOG_PRINT("%s\n", parser_output.c_str());
    }

    parser_output = "";
    //initialize capability mask caches
    rc2 = p_ibdiag->BuildCapabilityCache(parser_output);
    if (rc2) {
        LOG_PRINT("%s\n", parser_output.c_str());
        ERR_PRINT("Failed to init capability mask inner cache.\n");
        ++this->num_errors;
    }
    SCREEN_PRINT("\n");

    //SMP and GMP substages to get capability masks
    //GMP
    if (this->to_check_vs_cap_gmp) {
        bool can_send_mads_by_lid = false;
        if (check_if_can_send_mads_by_lid(this->p_ibdiag, can_send_mads_by_lid) ||
            !can_send_mads_by_lid) {
            INFO_PRINT("%s skipped\n", CHECK_VS_CAP_GMP);
            PRINT("\n");
        } else {
            list_p_fabric_general_err vs_cap_gmp_errors;
            int rc2 = p_ibdiag->BuildVsCapGmpDB(vs_cap_gmp_errors,
                                                progress_bar_retrieve_nodes);
            SCREEN_PRINT("\n");
            rc = this->AnalyzeCheckResults(vs_cap_gmp_errors,
                    CHECK_VS_CAP_GMP,
                    rc2,
                    IBDIAG_ERR_CODE_FABRIC_ERROR,
                    this->num_errors,
                    this->num_warnings,
                    true);

            g_vs_cap_gmp_run = true;
        }
    }

    //SMP
    if (this->to_check_vs_cap_smp) {
        list_p_fabric_general_err vs_cap_smp_errors;
        rc2 = p_ibdiag->BuildVsCapSmpDB(vs_cap_smp_errors,
                                        progress_bar_retrieve_nodes);
        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(vs_cap_smp_errors,
                CHECK_VS_CAP_SMP,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
    }

    //now we have capability -> can run ExtendedPortInfo MADs
    //to determine extended mellanox speeds speed and LLR
    //note: even if we haven't ran the SMP and GMP sub stages,
    //we might have a capability from initial configuration
    list_p_fabric_general_err vs_ext_pi_errors;
    rc2 = p_ibdiag->BuildExtendedPortInfo(vs_ext_pi_errors,
                                          progress_bar_retrieve_nodes);
    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(vs_ext_pi_errors,
            CHECK_VS_EXTENDED_PORT_INFO,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    //collect port info extended info for FEC data
    list_p_fabric_general_err pi_ext_errors;
    rc2 = p_ibdiag->BuildPortInfoExtendedDB(pi_ext_errors,
                                            progress_bar_retrieve_nodes);
    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(pi_ext_errors,
            CHECK_PORT_INFO_EXTENDED,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    if (!load_from_file) {
        //bring switch info
        list_p_fabric_general_err switch_errors;
        rc2 = this->p_ibdiag->BuildSwitchInfoDB(switch_errors, progress_bar_retrieve_nodes);
        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(switch_errors,
                CHECK_NAME_SW_INFO_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings);
    }

    //here we have correct speeds after VS ExtendePortIfo MADs were sent
    //dump db file
    string file_str = generate_file_name(*this->p_base_path, DB_FILE);
    if (this->p_ibdiag->DumpInternalDBCSVTable(*this->p_csv_out,
                                               this->show_ports_data_db2,
                                               this->to_dump_fec_mode)) {
        ERR_PRINT("Writing discovery info to %s failed, err=%s\n",
                file_str.c_str(), this->p_ibdiag->GetLastError());
        ++this->num_errors;
        IBDIAGNET_RETURN(1);    /* exiting - there are tools
                                    that need to parse this file */
    }
    this->AddGeneratedFileName("ibdiagnet database file", file_str);
    //dump lst file
    file_str = generate_file_name(*this->p_base_path, LST_FILE);
    if (this->p_ibdiag->WriteLSTFile(file_str.c_str(),
                                     this->dump_lst_with_lmc)) {
        WARN_PRINT("Writing lst file to %s failed\n", file_str.c_str());
        ++this->num_errors;
    }
    this->AddGeneratedFileName("LST file", file_str);

    //network dump file
    file_str = generate_file_name(*this->p_base_path, NET_DUMP_FILE);
    if (this->p_ibdiag->WriteNetDumpFile(file_str.c_str())) {
        WARN_PRINT("Writing network dump file to %s failed\n", file_str.c_str());
        ++this->num_errors;
    }
    this->AddGeneratedFileName("Network dump file", file_str);

    //dump the capability mask configurations after we are done with both stages
    //using it as a hidden option for debug
    if (!this->output_full_capability_file.empty()) {
        string parser_output;
        if (p_ibdiag->DumpCapabilityMaskFile(
            this->output_full_capability_file.c_str(),
            parser_output))
            WARN_PRINT("Writing Guid to Capability Masks file to %s failed\n",
                       this->output_full_capability_file.c_str());

        LOG_PRINT("%s\n", parser_output.c_str());
        parser_output = "";

        if (p_ibdiag->DumpGuid2Mask(this->output_full_capability_file.c_str(),
                                    parser_output))
            WARN_PRINT("Writing Guid to Capability Masks file to %s failed\n",
                       this->output_full_capability_file.c_str());
        LOG_PRINT("%s\n", parser_output.c_str());
    }

    IBDIAGNET_RETURN(0);
}

int DiscoveryStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc;
    list_p_fabric_general_err nodes_errors;

    //print duplicated GUIDs detection error if not skipped
    if (this->to_check_dup_guids)
        this->PrintDuplicatedGuidsDetectErrors();

    //duplicated GUIDs detection - we can find duplicated GUID even if
    //skipped in some cases (like duplicated port GUID)
    list_p_fabric_general_err guids_errors;
    int rc2 = this->p_ibdiag->CheckDuplicatedGUIDs(guids_errors);
    rc = this->AnalyzeCheckResults(guids_errors,
            CHECK_NAME_DUPLICATED_GUIDS,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);
    if (rc2 == IBDIAG_ERR_CODE_CHECK_FAILED)    //abort and exit
        IBDIAGNET_RETURN(1);

    //if needed than check for duplicated node description
    if (this->to_check_dup_node_desc) {
        rc2 = this->p_ibdiag->CheckDuplicatedNodeDescription(nodes_errors);
        if (rc2 == IBDIAG_ERR_CODE_CHECK_FAILED)
            WARN_PRINT("Note: If you have unmanaged systems then duplication can occur\n");
        rc = this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_DUPLICATED_NODE_DESC,
                rc2,
                IBDIAG_ERR_CODE_CHECK_FAILED,
                this->num_errors,
                this->num_warnings,
                true);
    }

    //if the port is down than exit
    bool local_link_is_down;
    if (check_if_local_link_is_down(this->p_ibdiag, local_link_is_down))
        IBDIAGNET_RETURN(1);
    if (local_link_is_down) {
        ERR_PRINT("The Local link is in DOWN state\n");
        IBDIAGNET_RETURN(1);
    }

    IBDIAGNET_RETURN(rc);
}

void DiscoveryStage::CreateLid2GuidDB()
{
    IBDIAGNET_ENTER;
    uint64_t guid = 0;
    uint16_t lid = 0;
    IBPort * p_port = NULL;
    if (!p_ibdiag->GetDiscoverFabricPtr())
        IBDIAGNET_RETURN_VOID;

    for (map_guid_pport::iterator It = p_ibdiag->GetDiscoverFabricPtr()->PortByGuid.begin();
         It != p_ibdiag->GetDiscoverFabricPtr()->PortByGuid.end();
         It++) {
        guid = (*It).first;
        p_port = (*It).second;
        if (!p_port)
            continue;
        lid = p_port->base_lid;
        lid2guid_map.insert(pair<uint16_t, uint64_t>(lid, guid));
    }
    IBDIAGNET_RETURN_VOID;
}

/******************************************************/
LidsStage::LidsStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_LIDS, p_ibdiag)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int LidsStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int LidsStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    //dummy one
    IBDIAGNET_RETURN(0);
}

int LidsStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc;

    list_p_fabric_general_err lids_errors;
    int rc2 = this->p_ibdiag->CheckLids(lids_errors);
    if (rc2 == IBDIAG_ERR_CODE_CHECK_FAILED)
        found_lids_errors = true;
    rc = this->AnalyzeCheckResults(lids_errors,
            CHECK_NAME_LIDS,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);
    IBDIAGNET_RETURN(rc);
}


/******************************************************/
/////
//ToplogtMatchStage public methods:
/////
ToplogtMatchStage::ToplogtMatchStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_TOPLOGY_MATCH, p_ibdiag),
        spec_fabric(),
        output_topo_file(""), input_topo_file(""), topo_ibnl_dir("")
{
    IBDIAGNET_ENTER;
    this->MarkStageAsSkipped(); //this stage will be skipped by default
	//this->compare_nodes_names = 0;
    IBDIAGNET_RETURN_VOID;
}

int ToplogtMatchStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int ToplogtMatchStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    if (topo_ibnl_dir.empty()) {
        topo_ibnl_dir = generate_file_name(*this->p_base_path, IBNL_OUTPUT_DIR);
        INFO_PRINT("Topo ibnl dir is not given, use default: %s\n", topo_ibnl_dir.c_str());
    } 

    //create a topology file if needed
    if (!this->output_topo_file.empty()) {

        //construct Systems - creates IBSystem ports and check IBNL definitions
        ibdmClearInternalLog();

        list< string > dirs;
        dirs.push_back(topo_ibnl_dir);
        theSysDefsCollection()->parseSysDefsFromDirs(dirs);

        this->p_ibdiag->GetDiscoverFabricPtr()->constructSystems();
        char *buffer= ibdmGetAndClearInternalLog();
        if (!buffer) {
            ERR_PRINT("Failed to allocate buffer for ibdm output\n");
            IBDIAGNET_RETURN(1);
        } else {
            u_int32_t errors = num_errors;
            u_int32_t warnings = num_warnings;
            string construct_res(buffer);
            this->ParseIBDMOutput(construct_res);

            if (num_errors > errors) {
                ERR_PRINT("Preparing topology for discovered fabric"
                          " finished with errors.\n");
            } else if (num_warnings > warnings) {
                WARN_PRINT("Preparing topology for discovered fabric"
                           " finished with warnings.\n");
            }
            LOG_PRINT("%s", buffer);
            free(buffer);
        }

        //try to create the ibnl output directory
        if (mkdir(topo_ibnl_dir.c_str(), S_IRWXU | S_IRWXG | S_IRWXO) < 0) {
            if (errno != EEXIST) {
                ERR_PRINT("Failed to set ibnl output path to %s, aborting topology write...\n", topo_ibnl_dir.c_str());
                goto next_topo_step;
            }
        }

        //try to create topo file and ibnl files
        INFO_PRINT("Creating topology file for discovered fabric - %s\n",
                this->output_topo_file.c_str());
        INFO_PRINT("The topology file may use custom system definitions"
                   " written into directory: %s\n", topo_ibnl_dir.c_str());

        ibdmClearInternalLog();
        if (!this->p_ibdiag->GetDiscoverFabricPtr()->dumpTopology(
                this->output_topo_file.c_str(), topo_ibnl_dir.c_str()))
            this->AddGeneratedFileName("Topology file",  this->output_topo_file.c_str());
        buffer = ibdmGetAndClearInternalLog();
        if (!buffer) {
            ERR_PRINT("Failed to allocate buffer for ibdm output\n");
            IBDIAGNET_RETURN(1);
        }
        PRINT("%s", buffer);
        free(buffer);
    }

    if (!this->input_topo_file.empty()) {

         //parsing the topology file
        ibdmClearInternalLog();

        //reparse topo_ibnl_dir in case aditional ibnls were added
        list< string > dirs;
        dirs.push_back(topo_ibnl_dir);
        theSysDefsCollection()->parseSysDefsFromDirs(dirs);

        int rc = GetSpecFabricPtr()->parseTopology(input_topo_file);
        char *buffer = ibdmGetAndClearInternalLog();
        if (!buffer) {
            ERR_PRINT("Failed to allocate buffer for ibdm output\n");
            IBDIAGNET_RETURN(1);
        }
        PRINT("%s", buffer);
        free(buffer);
        if (rc)
            ERR_PRINT("Parse topology file: %s finished whith errors\n",
                      input_topo_file.c_str());
    }

next_topo_step:
    IBDIAGNET_RETURN(0);
}

int ToplogtMatchStage::RunCheck()
{
    IBDIAGNET_ENTER;
    if (this->input_topo_file.empty())
        IBDIAGNET_RETURN(0);

    INFO_PRINT("Topology Matching\n");
    //get root port
    IBPort *p_root_port = this->p_ibdiag->GetRootPort();
    if (!p_root_port) {
        ERR_PRINT("Topology Matching failed, err=%s\n", this->p_ibdiag->GetLastError());
        IBDIAGNET_RETURN(1);
    }

    //match discovered fabric with spec fabric
    char *messages;
    TopoMatchOptions opts;

    /*
    if (TopoMatchFabrics(&this->spec_fabric,        // The specification fabric
            this->p_ibdiag->GetDiscoverFabricPtr(), // The discovered fabric
            p_root_port->p_node->name.c_str(),
                                                    // The name of the node to be the anchor point - in the spec fabric,
            p_root_port->num,                       // The port number of the anchor port - in the spec fabric
            p_root_port->guid_get(),                // Guid of the anchor port - in the discovered fabric
            &messages,
			this->compare_nodes_names)) {
           */
    //opts.mapFileName = mapFileName;
    opts.matchCAsByName = true; // match non-SW by name
    opts.matchSWsByName = true; // match SW by name
    //opts.csvFileName = csvFileName;
    opts.outMapFileName = generate_file_name(*this->p_base_path, "topodiff.names");

    if (TopoMatchFabricsFromEdge(&this->spec_fabric,        // The specification fabric
            this->p_ibdiag->GetDiscoverFabricPtr(), // The discovered fabric
            opts, &messages)) {
        WARN_PRINT("Found mismatches between the topology defined in %s and the discovered fabric.\n",
                this->input_topo_file.c_str());
        if (messages) {
            string topo_result = messages;
            this->ParseIBDMOutput(topo_result);
            LOG_PRINT("%s\n", topo_result.c_str());
        }
    } else {
        INFO_PRINT("The topology defined in %s perfectly matches the discovered fabric\n",
                this->input_topo_file.c_str());
    }

    PRINT("\n");
    free(messages);
    IBDIAGNET_RETURN(0);
}


/******************************************************/
LinksStage::LinksStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_LINKS, p_ibdiag)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int LinksStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int LinksStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    //dummy one
    IBDIAGNET_RETURN(0);
}

int LinksStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc;

    list_p_fabric_general_err links_errors;
    int rc2 = this->p_ibdiag->CheckLinks(links_errors);
    if (rc2 == IBDIAG_ERR_CODE_CHECK_FAILED)
        found_links_errors = true;
    rc = this->AnalyzeCheckResults(links_errors,
            CHECK_NAME_LINKS,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);
    IBDIAGNET_RETURN(rc);
}


/******************************************************/
SMStage::SMStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_SM, p_ibdiag)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int SMStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int SMStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc;

    list_p_fabric_general_err sm_errors;
    int rc2 = this->p_ibdiag->BuildSMInfoDB(sm_errors);
    rc = this->AnalyzeCheckResults(sm_errors,
            CHECK_NAME_SM_INFO_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (!rc) {
        string file_str = generate_file_name(*this->p_base_path, DB_FILE);
        if (this->p_ibdiag->DumpSMInfoCSVTable(*this->p_csv_out)) {
            ERR_PRINT("Writing SM info to %s failed\n", file_str.c_str());
            ++this->num_errors;
        }
        file_str = generate_file_name(*this->p_base_path, SM_FILE);
        if (this->p_ibdiag->WriteSMFile(file_str.c_str())) {
            ERR_PRINT("Writing SM info to %s failed\n", file_str.c_str());
            ++this->num_errors;
        }
        this->AddGeneratedFileName("Subnet Manager file", file_str);
    }
    IBDIAGNET_RETURN(rc);
}

int SMStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc;

    list_p_fabric_general_err sm_errors;
    int rc2 = this->p_ibdiag->CheckSMInfo(sm_errors);
    rc = this->AnalyzeCheckResults(sm_errors,
            CHECK_NAME_SM_INFO,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);
    IBDIAGNET_RETURN(rc);
}


/******************************************************/
/////
//pm private methods:
/////
static bool pm_caught_signal = false;

#ifdef __WIN__
    int signals_array[] = {SIGINT};
#else
    int signals_array[] = {SIGINT, SIGQUIT, SIGABRT, SIGPIPE, SIGTERM};
#endif


void pm_stage_sig_handler(int signum)
{
    IBDIAGNET_ENTER;
    // no need to set handler again because we want the default handler
    // from here

    //mark we caught the signal
    pm_caught_signal = true;

    SCREEN_PRINT("\n");
    if (signum == SIGINT) {
        INFO_PRINT("Received User ^C (%s). Stopping\n", strsignal(signum));
    } else {
        INFO_PRINT("Received signal %d (%s). Stopping\n", signum, strsignal(signum));
    }
    IBDIAGNET_RETURN_VOID;
}

void PMStage::Sleep(u_int32_t seconds)
{
    IBDIAGNET_ENTER;
    u_int32_t seconds_left = seconds;
    INFO_PRINT("Going to sleep for %u seconds until next counters sample\n", seconds);
    while (!pm_caught_signal && seconds_left) {
        SCREEN_INFO_PRINT("Time left to sleep ... %u seconds.\r", seconds_left);
        fflush(stdout);
        sleep(1);
        --seconds_left;
    }
    SCREEN_PRINT("\n");
    IBDIAGNET_RETURN_VOID;
}

/////
//pm public methods:
/////
PMStage::PMStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_PM, p_ibdiag),
        to_reset_counters(false),
        to_calc_ber(false),
        m_check_counters_bitset(0x00),
        en_per_lane_cnts(false),
        en_per_slvl_cntrs(false),
        ber_threshold(DEFAULT_BER_THRESHOLD),
        pause_time_between_samples(DEFAULT_SECONDS_BETWEEN_COUNTERS_SAMPLES),
        can_send_mads_by_lid(false)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

PMStage::~PMStage()
{
    IBDIAGNET_ENTER;
    CountersPerSLVL * p_cntrs_per_slvl;

    this->p_ibdiag->CleanPMInfoObjVector(this->first_sample_pm_info_obj_vector);
    for (vec_slvl_cntrs::iterator it = m_slvl_cntrs_vec.begin();
         it != m_slvl_cntrs_vec.end(); it++) {
        p_cntrs_per_slvl = *it;
        if (p_cntrs_per_slvl)
            delete p_cntrs_per_slvl;
    }
}

int PMStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    if (check_if_can_send_mads_by_lid(this->p_ibdiag, this->can_send_mads_by_lid))
        IBDIAGNET_RETURN(1);
    IBDIAGNET_RETURN(0);
}

void PMStage::PreparePerSlVlCounters()
{

    IBDIAGNET_ENTER;
    //VS Class
    CountersPerSLVL * port_rcv_data_vl = new PortRcvDataVL();
    m_slvl_cntrs_vec.push_back(port_rcv_data_vl);

    CountersPerSLVL * port_xmit_data_vl = new PortXmitDataVL();
    m_slvl_cntrs_vec.push_back(port_xmit_data_vl);

    CountersPerSLVL * port_rcv_ext_data_vl = new PortRcvDataVLExt();
    m_slvl_cntrs_vec.push_back(port_rcv_ext_data_vl);

    CountersPerSLVL * port_xmit_ext_data_vl = new PortXmitDataVLExt();
    m_slvl_cntrs_vec.push_back(port_xmit_ext_data_vl);

    CountersPerSLVL * port_rcv_pkt_vl = new PortRcvPktVL();
    m_slvl_cntrs_vec.push_back(port_rcv_pkt_vl);

    CountersPerSLVL * port_xmit_pkt_vl = new PortXmitPktVL();
    m_slvl_cntrs_vec.push_back(port_xmit_pkt_vl);

    CountersPerSLVL * port_rcv_ext_pkt_vl = new PortRcvPktVLExt();
    m_slvl_cntrs_vec.push_back(port_rcv_ext_pkt_vl);

    CountersPerSLVL * port_xmit_ext_pkt_vl = new PortXmitPktVLExt();
    m_slvl_cntrs_vec.push_back(port_xmit_ext_pkt_vl);

    //PM Class
    CountersPerSLVL * port_xmit_data_sl = new PortXmitDataSL();
    m_slvl_cntrs_vec.push_back(port_xmit_data_sl);

    CountersPerSLVL * port_rcv_data_sl = new PortRcvDataSL();
    m_slvl_cntrs_vec.push_back(port_rcv_data_sl);

    CountersPerSLVL * port_xmit_data_sl_ext = new PortXmitDataSLExt();
    m_slvl_cntrs_vec.push_back(port_xmit_data_sl_ext);

    CountersPerSLVL * port_rcv_data_sl_ext = new PortRcvDataSLExt();
    m_slvl_cntrs_vec.push_back(port_rcv_data_sl_ext);

    IBDIAGNET_RETURN_VOID;
}


int PMStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    if (!this->can_send_mads_by_lid) {
        if (this->to_reset_counters)
            INFO_PRINT("%s skipped\n", CHECK_NAME_PM_RESET);
        INFO_PRINT("%s skipped\n", CHECK_NAME_PM_RETRIEVE);
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }

    list_p_fabric_general_err ports_errors;
    //if need to reset
    if (this->to_reset_counters) {
        int rc2  = this->p_ibdiag->ResetPortCounters(ports_errors,
                                                     progress_bar_reset_nodes,
                                                     m_check_counters_bitset);
        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(ports_errors,
                CHECK_NAME_PM_RESET,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
    }
    if (rc)
        IBDIAGNET_RETURN(rc);

    //retrieve all counters - first sample
    struct timeval start_measure;
    if ((rc = gettimeofday(&start_measure, NULL)) != 0) {
        IBDIAG_LOG(TT_LOG_LEVEL_INFO,
                   "Failed to get gettimeofday before first sample.");
        ERR_PRINT("Failed to retrieve first sample time\n");
        ++this->num_errors;
        IBDIAGNET_RETURN(1);
    }
    int rc2 = this->p_ibdiag->BuildPortCountersDB(ports_errors,
                                                  progress_bar_retrieve_nodes,
                                                  m_check_counters_bitset);
    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(ports_errors,
            CHECK_NAME_PM_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc)
        IBDIAGNET_RETURN(rc);

    if (this->pause_time_between_samples != 0) {   //need to take another sample
        //copy counters vector of first sample
        this->p_ibdiag->CopyPMInfoObjVector(this->first_sample_pm_info_obj_vector);

        //set a signal handler for kill signals
        for (unsigned int i = 0; i < (sizeof(signals_array) / sizeof(signals_array[0])); ++i) {
            signal(signals_array[i], pm_stage_sig_handler);
        }
        //pausing
        this->Sleep(this->pause_time_between_samples);
        //set default handlers for kill signals
        for (unsigned int i = 0; i < (sizeof(signals_array) / sizeof(signals_array[0])); ++i) {
            signal(signals_array[i], SIG_DFL);
        }
        PRINT("\n");

        //retrieve all counters - second sample
        struct timeval end_measure;
        if ((rc = gettimeofday(&end_measure, NULL)) != 0) {
            IBDIAG_LOG(TT_LOG_LEVEL_INFO,
                       "Failed to get gettimeofday before second sample.");
            ERR_PRINT("Failed to retrieve second sample time\n");
            ++this->num_errors;
            IBDIAGNET_RETURN(1);
        }

        int rc2 = this->p_ibdiag->BuildPortCountersDB(ports_errors,
                                                      progress_bar_retrieve_nodes,
                                                      m_check_counters_bitset);
        SCREEN_PRINT("\n");

        long long elapsed_mic_sec = (end_measure.tv_sec - start_measure.tv_sec)*1000000LL +
                                    (end_measure.tv_usec - start_measure.tv_usec);

        this->diff_time_between_samples = (double)elapsed_mic_sec / (double)1000000;

        rc = this->AnalyzeCheckResults(ports_errors,
                CHECK_NAME_PM_RETRIEVE2,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
        if (rc)
            IBDIAGNET_RETURN(rc);
    }

    if (this->en_per_slvl_cntrs) {
        this->PreparePerSlVlCounters();
        if (this->to_reset_counters) {
            for (vec_slvl_cntrs::iterator it = m_slvl_cntrs_vec.begin();
                    it != m_slvl_cntrs_vec.end(); it++) {
                CountersPerSLVL * cntrs_per_slvl = *it;
                int rc2 = this->p_ibdiag->BuildDBOrResetSLVLCntrs(
                        ports_errors,
                        progress_bar_retrieve_ports,
                        true,//clear slvl counters
                        cntrs_per_slvl);

                SCREEN_PRINT("\n");
                rc = this->AnalyzeCheckResults(ports_errors,
                        cntrs_per_slvl->GetCntrHeader(),
                        rc2,
                        IBDIAG_ERR_CODE_FABRIC_ERROR,
                        this->num_errors,
                        this->num_warnings);
                ports_errors.clear();
            }
        }

        for (vec_slvl_cntrs::iterator it = m_slvl_cntrs_vec.begin();
             it != m_slvl_cntrs_vec.end(); it++) {
            CountersPerSLVL * cntrs_per_slvl = *it;
            int rc2 = this->p_ibdiag->BuildDBOrResetSLVLCntrs(
                    ports_errors,
                    progress_bar_retrieve_ports,
                    false,//build slvl counters DB
                    cntrs_per_slvl);
            SCREEN_PRINT("\n");
            rc = this->AnalyzeCheckResults(ports_errors,
                    cntrs_per_slvl->GetCntrHeader(),
                    rc2,
                    IBDIAG_ERR_CODE_FABRIC_ERROR,
                    this->num_errors,
                    this->num_warnings);
            ports_errors.clear();
        }
    }

    IBDIAGNET_RETURN(rc);
}

int PMStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    if (!this->can_send_mads_by_lid) {
        INFO_PRINT("Ports counter Checks skipped\n");
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }
    list_p_fabric_general_err ports_errors;

    //calculate counters calculation
    int rc2 = this->p_ibdiag->CalcCounters(this->first_sample_pm_info_obj_vector,
                                           this->diff_time_between_samples,
                                           ports_errors);
    // Errors accumulated in the CalcCounters will be analyzed with CheckAllPMValues later
    //counters value check
    rc2 |= this->p_ibdiag->CheckAllPMValues(ports_errors,
                                            this->counters_to_threshold_map,
                                            m_check_counters_bitset);

    rc = this->AnalyzeCheckResults(ports_errors,
            CHECK_NAME_PM_VALUE,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);
    if (rc)
        IBDIAGNET_RETURN(rc);

    // now when we have the calculated counters, we can print to the files
    string file_str = generate_file_name(*this->p_base_path, DB_FILE);
    if (this->p_ibdiag->DumpPortCountersCSVTable(*this->p_csv_out, m_check_counters_bitset)) {
        ERR_PRINT("Writing ports counters to %s failed\n", file_str.c_str());
        ++this->num_errors;
    }

    if (this->en_per_slvl_cntrs) {
        if (this->p_ibdiag->DumpPerSLVLPortCntrsCSVTable(*this->p_csv_out,
                                                         this->m_slvl_cntrs_vec)) {
            ERR_PRINT("Write SLVL Counters PM DB to %s failed\n", file_str.c_str());
            ++this->num_errors;
        }
    }

    file_str = generate_file_name(*this->p_base_path, PM_FILE);
    if (this->p_ibdiag->WritePMFile(file_str.c_str(),
                                    m_check_counters_bitset,
                                    en_per_lane_cnts)) {
        ERR_PRINT("Writing ports counters to %s failed\n", file_str.c_str());
        ++this->num_errors;
    }
    this->AddGeneratedFileName("Ports Counters file", file_str);

    //abort all other checks - we don't have another sample
    if (this->pause_time_between_samples == 0) {
        INFO_PRINT("Ports counters Difference Check will be skipped - pause time is zero\n");
        IBDIAGNET_RETURN(rc);
    }

    //no increase check
    rc2 = this->p_ibdiag->CheckCountersDiff(this->first_sample_pm_info_obj_vector,
            ports_errors);

    rc = this->AnalyzeCheckResults(ports_errors,
            CHECK_NAME_PM_DIFF,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);

    if (rc)
        IBDIAGNET_RETURN(rc);

    //BER check - only if needed
    if (this->to_calc_ber) {
        rc2 = this->p_ibdiag->CalcBERErrors(this->first_sample_pm_info_obj_vector,
                (u_int64_t)(this->ber_threshold),
                this->diff_time_between_samples,
                ports_errors,
                *this->p_csv_out);
        rc = this->AnalyzeCheckResults(ports_errors,
                CHECK_NAME_BER,
                rc2,
                IBDIAG_ERR_CODE_CHECK_FAILED,
                this->num_errors,
                this->num_warnings);

        rc2 = this->p_ibdiag->CalcPhyTest(this->first_sample_pm_info_obj_vector,
                                          this->diff_time_between_samples,
                                          *this->p_csv_out);
    }
    IBDIAGNET_RETURN(rc);
}


/******************************************************/

/////
//RoutingStage public methods:
/////
RoutingStage::RoutingStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_ROUTING, p_ibdiag),
        retrieved_switch_info_ok(false), retrieved_uc_fdbs_ok(false),
        retrieved_mc_fdbs_ok(false), retrieved_ext_switch_info_ok(false),
        retrieved_plft_data_ok(false), retrieved_ar_data_ok(false),
        retrieved_rn_data_ok(false), ar_enabled(false),
        is_fat_tree(false), use_vlr(false),
        use_vs_route(true), dump_full_ar(true),
        dump_rn_data(false), dump_rn_counters(false), clear_rn_counters(false),
        check_mcast_credloop(false), check_sl(false), dump_only(false), default_sl(0)
{
    IBDIAGNET_ENTER;
    this->MarkStageAsSkipped(); //this stage will be skipped by default
    IBDIAGNET_RETURN_VOID;
}

int RoutingStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int RoutingStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc;
    int rc2;
    list_p_fabric_general_err nodes_errors;

    string parser_output;
    if(!sa_dump_file.empty() && check_mcast_credloop) {
        INFO_PRINT("Load SA Dump file (%s).\n", sa_dump_file.c_str());

        //To check vport and aguids in the sadump file,
        //need to call VirtualizationStage & AGuidStage before this function
        rc = p_ibdiag->ParseSADumpFile(sa_dump_file, parser_output);
        LOG_PRINT("%s\n", parser_output.c_str());

        if (rc) {
            ERR_PRINT("Failed to parse SA Dump file (%s).\n", sa_dump_file.c_str());
            ++this->num_errors;
            IBDIAGNET_RETURN(1);
        }
    }

    if (use_vs_route) {
        rc2  = this->p_ibdiag->RetrieveExtendedSwitchInfo(nodes_errors);
        if (rc2 == IBDIAG_SUCCESS_CODE)
            this->retrieved_ext_switch_info_ok = true;
        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_ROUTE_EXT_SW_INFO_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
        if (!rc) {
            string file_str = generate_file_name(*this->p_base_path, VL2VL_FILE);
            if (this->p_ibdiag->WriteVL2VLFile(file_str.c_str())) {
                ERR_PRINT("Writing VL2VL info to %s failed\n", file_str.c_str());
                ++this->num_errors;
            }
            this->AddGeneratedFileName("VL2VL file", file_str);
        }

        //bring PLFT Data
        unsigned int supportedDev;
        rc2  = this->p_ibdiag->RetrievePLFTData(nodes_errors, supportedDev);
        if (rc2 == IBDIAG_SUCCESS_CODE){
            this->retrieved_plft_data_ok = true;
            SCREEN_INFO_PRINT("PLFT is enabled on %d switches.\n",supportedDev);
        }
        rc = this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_ROUTE_PLFT_DATA_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);

        if (!rc) {
            string file_str = generate_file_name(*this->p_base_path, PLFT_FILE);
            if (this->p_ibdiag->WritePLFTFile(file_str.c_str())) {
                ERR_PRINT("Writing PLFT info to %s failed\n", file_str.c_str());
                ++this->num_errors;
            }
            this->AddGeneratedFileName("PLFT file", file_str);
        }

        //bring AR Data
        AdditionalRoutingDataMap routing_data_map;
        AdditionalRoutingDataMap * p_routing_data_map = NULL;
        if (dump_rn_data || dump_full_ar ||
            dump_rn_counters || clear_rn_counters) {
            p_routing_data_map = &routing_data_map;
            AdditionalRoutingData::dump_rn_data = dump_rn_data;
            AdditionalRoutingData::dump_rn_counters = dump_rn_counters;
            AdditionalRoutingData::clear_rn_counters = clear_rn_counters;
            AdditionalRoutingData::dump_full_ar = dump_full_ar;
        }

        rc2  = this->p_ibdiag->RetrieveARData(nodes_errors, supportedDev,
                                              p_routing_data_map);

        INFO_PRINT("Adaptive Routing is enabled on %d switches.\n",supportedDev);
        if (rc2 == IBDIAG_SUCCESS_CODE){
            this->retrieved_ar_data_ok = true;
            this->ar_enabled = (supportedDev > 0);
        }
        rc = this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_ROUTE_AR_DATA_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);

        if (!rc) {
            string file_str = generate_file_name(*this->p_base_path, AR_FILE);
            if (this->p_ibdiag->WriteARFile(file_str.c_str())) {
                ERR_PRINT("Writing AR info to %s failed\n", file_str.c_str());
                ++this->num_errors;
            }
            this->AddGeneratedFileName("AR file", file_str);
        }

        if (dump_full_ar) {
            string file_str = generate_file_name(*this->p_base_path, FAR_FILE);
            if (this->p_ibdiag->WriteFARFile(&routing_data_map, file_str.c_str())) {
                ERR_PRINT("Writing FAR to %s failed\n", file_str.c_str());
                ++this->num_errors;
            }
            this->AddGeneratedFileName("Full AR file", file_str);
        }

        if (dump_rn_data || dump_rn_counters || clear_rn_counters){
            INFO_PRINT("Routing Notification is supported on %d switches.\n",
                       AdditionalRoutingData::sw_supports_rn_count);

            if (AdditionalRoutingData::sw_supports_rn_count == 0)
                PRINT("\n");
        }

        if (dump_rn_data &&
            AdditionalRoutingData::sw_supports_rn_count != 0) {

            rc2  = this->p_ibdiag->RetrieveRNData(nodes_errors,
                                                  &routing_data_map);

            if (rc2 == IBDIAG_SUCCESS_CODE)
                this->retrieved_rn_data_ok = true;

            this->AnalyzeCheckResults(nodes_errors,
                    CHECK_NAME_ROUTE_RN_DATA_RETRIEVE,
                    rc2,
                    IBDIAG_ERR_CODE_FABRIC_ERROR,
                    this->num_errors,
                    this->num_warnings,
                    true);

            string file_str = generate_file_name(*this->p_base_path, RN_FILE);
            if (this->p_ibdiag->WriteRNFile(&routing_data_map, file_str.c_str())) {
                ERR_PRINT("Writing RN info to %s failed\n", file_str.c_str());
                ++this->num_errors;
            }
            this->AddGeneratedFileName("RN file", file_str);
        }
        if ((dump_rn_counters || clear_rn_counters) &&
            AdditionalRoutingData::sw_supports_rn_count != 0) {

            rc2  = this->p_ibdiag->ClearAndGetRNCounters(
                nodes_errors, &routing_data_map);

            this->AnalyzeCheckResults(nodes_errors,
                    CHECK_NAME_ROUTE_RN_COUNTERS_RETRIEVE,
                    rc2,
                    IBDIAG_ERR_CODE_FABRIC_ERROR,
                    this->num_errors,
                    this->num_warnings,
                    true);

            if (dump_rn_counters){
                string file_str = generate_file_name(*this->p_base_path, RNC_FILE);
                if (this->p_ibdiag->WriteRNCountersFile(&routing_data_map, file_str.c_str())) {
                    ERR_PRINT("Writing RN counters info to %s failed\n", file_str.c_str());
                    ++this->num_errors;
                }
                this->AddGeneratedFileName("RN counters file", file_str);
            }
        }
    }

    //bring unicast FDBs info
    rc2  = this->p_ibdiag->RetrieveUCFDBSInfo(nodes_errors, progress_bar_retrieve_nodes);
    if (rc2 == IBDIAG_SUCCESS_CODE)
        this->retrieved_uc_fdbs_ok = true;
    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(nodes_errors,
            CHECK_NAME_ROUTE_UCFDBS_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (!rc) {
        string file_str = generate_file_name(*this->p_base_path, UC_FDBS_FILE);
        if (this->p_ibdiag->WriteUCFDBSFile(file_str.c_str())) {
            ERR_PRINT("Writing unicast FDBS info to %s failed\n", file_str.c_str());
            ++this->num_errors;
        }
        this->AddGeneratedFileName("Unicast FDBS file", file_str);
    }
    if (!this->retrieved_uc_fdbs_ok)    //abort
        IBDIAGNET_RETURN(rc);

    //bring multicast FDBs info
    rc2  = this->p_ibdiag->RetrieveMCFDBSInfo(nodes_errors, progress_bar_retrieve_nodes);
    if (rc2 == IBDIAG_SUCCESS_CODE)
        this->retrieved_mc_fdbs_ok = true;
    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(nodes_errors,
            CHECK_NAME_ROUTE_MCASTDBS_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (!rc) {
        string file_str = generate_file_name(*this->p_base_path, MC_FDBS_FILE);
        if (this->p_ibdiag->WriteMCFDBSFile(file_str.c_str())) {
            ERR_PRINT("Writing multicast FDBS info to %s failed\n", file_str.c_str());
            ++this->num_errors;
        }
        this->AddGeneratedFileName("Multicast FDBS file", file_str);
    }

    //Write SLVL File
    if (use_vlr || use_vs_route || check_mcast_credloop || check_sl) {
        string file_str = generate_file_name(*this->p_base_path, SLVL_FILE);
        rc2  = this->p_ibdiag->WriteSLVLFile(file_str.c_str(),
                                             nodes_errors,
                                             progress_bar_retrieve_nodes);
        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_ROUTE_DUMP_SLVL_TABLE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
        if (!rc) {
            this->AddGeneratedFileName("SLVL Table file", file_str);

            INFO_PRINT("Load SLVL file.\n");

            string parser_output;
            rc = this->p_ibdiag->ParseSLVLFile(file_str, parser_output);
            this->ParseIBDMOutput(parser_output);
            LOG_PRINT("%s\n", parser_output.c_str());

            if (rc) {
                ERR_PRINT("Failed to parse SLVL file. \n");
                ++this->num_errors;
            }
        }
    }

    IBDIAGNET_RETURN(rc);
}

int RoutingStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    if (this->dump_only) {
        LOG_INFO_PRINT("Requested to skip Routing checks\n");
        IBDIAGNET_RETURN(rc);
    }

    if ((!this->retrieved_uc_fdbs_ok) || (!this->retrieved_mc_fdbs_ok)) {
        ERR_PRINT("Skipping Routing checks because not all information retrieved successfully\n");
        ++this->num_errors;
        IBDIAGNET_RETURN(rc);
    }

    this->p_ibdiag->SetDefaultSL(default_sl);

    if (use_vs_route) {

        if ((!this->retrieved_ext_switch_info_ok) ||
            (!this->retrieved_plft_data_ok) ||
            (!this->retrieved_ar_data_ok) ) {
            ERR_PRINT("Skipping Adaptive Routing checks because" \
                      " not all information retrieved successfully\n");
            ++this->num_errors;
            this->ar_enabled = false;
        }
    }

    LOG_PRINT("\n\n");
    string routing_output;

    if (check_sl) {
        LOG_PRINT("###################################\n");
        LOG_INFO_PRINT("SL2VL table check:\n");
        LOG_PRINT("###################################\n");
        this->p_ibdiag->CheckSL2VLTables(routing_output);
        this->ParseIBDMOutput(routing_output);
        LOG_PRINT("%s\n", routing_output.c_str());
    }

    LOG_PRINT("###################################\n");
    LOG_INFO_PRINT("Fabric Qualities Report:\n");
    LOG_PRINT("###################################\n");
    this->p_ibdiag->ReportFabricQualities(routing_output, this->p_base_path->c_str());
    this->ParseIBDMOutput(routing_output);
    LOG_PRINT("%s\n", routing_output.c_str());

    LOG_PRINT("###################################\n");
    LOG_INFO_PRINT("Credit Loops Report:\n");
    LOG_PRINT("###################################\n");
    if(check_mcast_credloop)
        CredLoopMode(0,1);

    this->p_ibdiag->ReportCreditLoops(routing_output, this->is_fat_tree, this->ar_enabled);
    this->ParseIBDMOutput(routing_output);
    LOG_PRINT("%s\n", routing_output.c_str());

    if (use_vs_route && this->ar_enabled) {

        LOG_PRINT("###################################\n");
        LOG_INFO_PRINT("Fabric AR Connectivity Report:\n");
        LOG_PRINT("###################################\n");
        this->p_ibdiag->ReportFabricARConnectivity(routing_output);
        this->ParseIBDMOutput(routing_output);
        LOG_PRINT("%s\n", routing_output.c_str());
    }

    LOG_PRINT("\n\n");

    IBDIAGNET_RETURN(rc);
}


/******************************************************/
NodesInfoStage::NodesInfoStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_NODES_INFO, p_ibdiag),
        can_send_mads_by_lid(false)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int NodesInfoStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    if (check_if_can_send_mads_by_lid(this->p_ibdiag, this->can_send_mads_by_lid))
            IBDIAGNET_RETURN(1);
    IBDIAGNET_RETURN(0);
}

int NodesInfoStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    if (!this->can_send_mads_by_lid) {
        INFO_PRINT("%s skipped\n", CHECK_NAME_NODES_INFO_RETRIEVE);
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }

    if (!g_vs_cap_gmp_run) {
        INFO_PRINT("%s wasn't ran. Skipping %s\n",
                   CHECK_VS_CAP_GMP, CHECK_NAME_NODES_INFO_RETRIEVE);
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }

    list_p_fabric_general_err nodes_errors;
    string file_str = generate_file_name(*this->p_base_path, DB_FILE);
    if (this->p_ibdiag->DumpNodesInfoCSVTable(*this->p_csv_out)) {
        ERR_PRINT("Writing Nodes Info to %s failed\n", file_str.c_str());
        ++this->num_errors;
    }
    file_str = generate_file_name(*this->p_base_path, NODES_INFO_FILE);
    if (this->p_ibdiag->WriteNodesInfoFile(file_str.c_str())) {
        ERR_PRINT("Writing Nodes Info to %s failed\n", file_str.c_str());
        ++this->num_errors;
    }
    this->AddGeneratedFileName("Nodes Information file", file_str);

    IBDIAGNET_RETURN(rc);
}

int NodesInfoStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    if (!this->can_send_mads_by_lid) {
        INFO_PRINT("%s skipped\n", CHECK_NAME_NODES_INFO_FW);
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }
    if (!g_vs_cap_gmp_run) {
        INFO_PRINT("%s wasn't ran. Skipping %s\n",
                   CHECK_VS_CAP_GMP, CHECK_NAME_NODES_INFO_FW);
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }

    list_p_fabric_general_err fw_errors;
    int rc2 = this->p_ibdiag->CheckFWVersion(fw_errors);
    rc = this->AnalyzeCheckResults(fw_errors,
            CHECK_NAME_NODES_INFO_FW,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);
    IBDIAGNET_RETURN(rc);
}


/******************************************************/
Speed_WidthStage::Speed_WidthStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_SPEED_WIDTH, p_ibdiag),
        expected_link_speed(""), expected_link_width("")
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int Speed_WidthStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int Speed_WidthStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    list_p_fabric_general_err links_errors;

    //check link speed
    if (this->expected_link_speed != "") {
        INFO_PRINT("Link Speed Check (Expected value given = %s)\n", this->expected_link_speed.c_str());
    } else {
        INFO_PRINT("Link Speed Check (Compare to supported link speed)\n");
    }
    int rc2 = this->p_ibdiag->CheckLinkSpeed(links_errors, this->expected_link_speed);
    rc = this->AnalyzeCheckResults(links_errors,
            CHECK_NAME_LINK_SPEED,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);

    //check link width
    if (this->expected_link_width != "") {
        INFO_PRINT("Link Width Check (Expected value given = %s)\n", this->expected_link_width.c_str());
    } else {
        INFO_PRINT("Link Width Check (Compare to supported link width)\n");
    }
    rc2 = this->p_ibdiag->CheckLinkWidth(links_errors, this->expected_link_width);
    rc = this->AnalyzeCheckResults(links_errors,
            CHECK_NAME_LINK_WIDTH,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);
    IBDIAGNET_RETURN(rc);
}


/******************************************************/
PKeyStage::PKeyStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_PKEY, p_ibdiag),
        can_send_mads_by_lid(false)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

PKeyStage::~PKeyStage()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int PKeyStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int PKeyStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    list_p_fabric_general_err pkey_errors;
    int rc2 = this->p_ibdiag->BuildPartitionKeysDB(pkey_errors, progress_bar_retrieve_nodes);
    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(pkey_errors,
            CHECK_NAME_PKEY_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (!rc) {
        string file_str = generate_file_name(*this->p_base_path, DB_FILE);
        if (this->p_ibdiag->DumpPartitionKeysCSVTable(*this->p_csv_out)) {
            ERR_PRINT("Writing partition key info to %s failed\n", file_str.c_str());
            ++this->num_errors;
        }
        file_str = generate_file_name(*this->p_base_path, PKEY_FILE);
        if (this->p_ibdiag->WritePKeyFile(file_str.c_str())) {
            ERR_PRINT("Writing partition keys info to %s failed\n", file_str.c_str());
            ++this->num_errors;
        }
        this->AddGeneratedFileName("Partition keys file", file_str);
    }

    IBDIAGNET_RETURN(rc);
}

int PKeyStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    list_p_fabric_general_err pkey_errors;
    int rc2 = this->p_ibdiag->CheckPartitionKeys(pkey_errors);
    rc = this->AnalyzeCheckResults(pkey_errors,
        STAGE_NAME_PKEY,
        rc2,
        IBDIAG_ERR_CODE_CHECK_FAILED,
        this->num_errors,
        this->num_warnings);
    IBDIAGNET_RETURN(rc);
}


/******************************************************/
AGUIDStage::AGUIDStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_AGUID, p_ibdiag),
        can_send_mads_by_lid(false)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

AGUIDStage::~AGUIDStage()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int AGUIDStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    if (check_if_can_send_mads_by_lid(this->p_ibdiag, this->can_send_mads_by_lid))
            IBDIAGNET_RETURN(1);
    IBDIAGNET_RETURN(0);
}

int AGUIDStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    if (!this->can_send_mads_by_lid) {
        INFO_PRINT("%s skipped\n", CHECK_NAME_AGUID_RETRIEVE);
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }

    list_p_fabric_general_err aguid_errors;
    int rc2 = this->p_ibdiag->BuildAliasGuidsDB(aguid_errors,
            progress_bar_retrieve_nodes);
    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(aguid_errors,
            CHECK_NAME_AGUID_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (!rc) {
        string file_str = generate_file_name(*this->p_base_path, DB_FILE);
        if (this->p_ibdiag->DumpAliasGUIDCSVTable(*this->p_csv_out)) {
            ERR_PRINT("Writing alias gudis info to %s failed\n", file_str.c_str());
            ++this->num_errors;
        }
        file_str = generate_file_name(*this->p_base_path, AGUID_FILE);
        if (this->p_ibdiag->WriteAliasGUIDFile(file_str.c_str())) {
            ERR_PRINT("Writing alias gudis info to %s failed\n", file_str.c_str());
            ++this->num_errors;
        }
        this->AddGeneratedFileName("Alias guids file", file_str);
    }
    IBDIAGNET_RETURN(rc);
}

int AGUIDStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    if (!this->can_send_mads_by_lid) {
        INFO_PRINT("%s skipped\n", CHECK_NAME_AGUID);
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }

    list_p_fabric_general_err aguid_errors;
    int rc2 = this->p_ibdiag->CheckDuplicatedAliasGuids(aguid_errors);
    rc = this->AnalyzeCheckResults(aguid_errors,
            STAGE_NAME_AGUID,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);
    IBDIAGNET_RETURN(rc);
}


/******************************************************/
MlnxCntrsStage::MlnxCntrsStage(IBDiag *p_ibdiag) :
    Stage(STAGE_NAME_MLNX_CNTRS, p_ibdiag),
    can_send_mads_by_lid(false), to_reset_counters(false)
{
    IBDIAGNET_ENTER;
    this->MarkStageAsSkipped(); //this stage will be skipped by default
    IBDIAGNET_RETURN_VOID;
}

MlnxCntrsStage::~MlnxCntrsStage()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int MlnxCntrsStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    if (check_if_can_send_mads_by_lid(this->p_ibdiag,
        this->can_send_mads_by_lid))
            IBDIAGNET_RETURN(1);
    IBDIAGNET_RETURN(0);
}

int MlnxCntrsStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    if (!this->can_send_mads_by_lid) {
        INFO_PRINT("%s skipped\n", CHECK_NAME_MLNX_CNTRS_RETRIEVE);
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }

    list_p_fabric_general_err mlnx_cntrs_errors;
    //if need to reset
    if (this->to_reset_counters) {
        int rc2  = this->p_ibdiag->ResetDiagnosticCounters(mlnx_cntrs_errors,
                                                     progress_bar_reset_nodes);
        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(mlnx_cntrs_errors,
                CHECK_NAME_MLNX_CNTRS_RESET,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
    }
    if (rc)
        IBDIAGNET_RETURN(rc);

    int rc2 = this->p_ibdiag->BuildVsDiagnosticCounters(
            mlnx_cntrs_errors,
            progress_bar_retrieve_nodes);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(mlnx_cntrs_errors,
            CHECK_NAME_MLNX_CNTRS_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc)
        IBDIAGNET_RETURN(rc);

    IBDIAGNET_RETURN(rc);
}

int MlnxCntrsStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    // now when we have the calculated counters, we can print to the files
    string file_str = generate_file_name(*this->p_base_path, DB_FILE);
    if (this->p_ibdiag->DumpDiagnosticCountersCSVTable(*this->p_csv_out)) {
        ERR_PRINT("Writing ports counters to %s failed\n", file_str.c_str());
        ++this->num_errors;
    }
    file_str = generate_file_name(*this->p_base_path, MLNX_CNTRS_FILE);
    if (this->p_ibdiag->WriteMlnxCntrsFile(file_str.c_str())) {
        ERR_PRINT("Writing Mellanox diagnostic counters to %s failed\n",
                  file_str.c_str());
        ++this->num_errors;
    }
    this->AddGeneratedFileName("Mellanox Counters file", file_str);
    IBDIAGNET_RETURN(rc);
}

//**********************************************
VirtualizationStage::VirtualizationStage(IBDiag *p_ibdiag) :
    Stage(STAGE_NAME_VIRTUALIZATION, p_ibdiag),
    can_send_mads_by_lid(false)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

VirtualizationStage::~VirtualizationStage()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int VirtualizationStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    if (check_if_can_send_mads_by_lid(this->p_ibdiag,
        this->can_send_mads_by_lid))
            IBDIAGNET_RETURN(1);
    IBDIAGNET_RETURN(0);
}

int VirtualizationStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    if (!this->can_send_mads_by_lid) {
        INFO_PRINT("%s skipped\n", CHECK_NAME_VIRTUALIZATION_RETRIEVE);
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }

    list_p_fabric_general_err vport_errors;
    int rc2 = p_ibdiag->BuildVirtualizationDB(vport_errors, progress_bar_retrieve_nodes);
    PRINT("\n");
    rc = this->AnalyzeCheckResults(vport_errors,
                                   STAGE_NAME_VIRTUALIZATION,
                                   rc2,
                                   IBDIAG_ERR_CODE_FABRIC_ERROR,
                                   this->num_errors,
                                   this->num_warnings,
                                   true);


    IBDIAGNET_RETURN(rc);
}

int VirtualizationStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    list_p_fabric_general_err vport_errors;
    this->p_ibdiag->CheckAndSetVPortLid(vport_errors);
    rc = this->AnalyzeCheckResults(vport_errors,
            CHECK_NAME_VIRTUAL_DISCOVERY_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);

    this->p_ibdiag->CheckVPortDuplicatedGuids(vport_errors);
    rc = this->AnalyzeCheckResults(vport_errors,
            CHECK_NAME_VIRTUAL_DISCOVERY_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);

    string file_str = generate_file_name(*this->p_base_path, DB_FILE);

    if (this->p_ibdiag->DumpCSVVNodesTable(*this->p_csv_out)) {
        ERR_PRINT("Writing VNodes to %s failed\n", file_str.c_str());
        ++this->num_errors;
    }
    if (this->p_ibdiag->DumpCSVVPortsTable(*this->p_csv_out)) {
        ERR_PRINT("Writing VPorts to %s failed\n", file_str.c_str());
        ++this->num_errors;
    }

    file_str = generate_file_name(*this->p_base_path, VPORTS_FILE);
    if (this->p_ibdiag->WriteVPortsFile(file_str.c_str())) {
        ERR_PRINT("Writing vports info to %s failed\n", file_str.c_str());
        ++this->num_errors;
    }
    this->AddGeneratedFileName("VPorts file", file_str);

    file_str = generate_file_name(*this->p_base_path, VPORTS_PKEY_FILE);
    if (this->p_ibdiag->WriteVPortsPKeyTblFile(file_str.c_str())) {
        ERR_PRINT("Writing vports pkey table to %s failed\n", file_str.c_str());
        ++this->num_errors;
    }
    this->AddGeneratedFileName("VPorts Pkey file", file_str);

    IBDIAGNET_RETURN(rc);
}

TemperatureSensing::TemperatureSensing (IBDiag *p_ibdiag) :
    Stage(STAGE_NAME_TEMPSENSING, p_ibdiag)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

TemperatureSensing::~TemperatureSensing()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int TemperatureSensing::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int TemperatureSensing::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    list_p_fabric_general_err temp_sensing_errors;

    int rc2 = this->p_ibdiag->BuildSMPTempSensing(
            temp_sensing_errors,
            progress_bar_retrieve_nodes);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(temp_sensing_errors,
            STAGE_NAME_TEMPSENSING,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc)
        IBDIAGNET_RETURN(rc);

    IBDIAGNET_RETURN(rc);
}

int TemperatureSensing::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    string file_str = generate_file_name(*this->p_base_path, DB_FILE);
    if (this->p_ibdiag->DumpTempSensingCSVTable(*this->p_csv_out)) {
        ERR_PRINT("Writing ports counters to %s failed\n", file_str.c_str());
        ++this->num_errors;
    }

    IBDIAGNET_RETURN(rc);
}

/*************************************************************************************************/
QoSStage::QoSStage (IBDiag *p_ibdiag) :
    Stage(STAGE_NAME_QOS, p_ibdiag)
{
    IBDIAGNET_ENTER;
    //mark this stage as skipped by default
    MarkStageAsSkipped();
    IBDIAGNET_RETURN_VOID;
}

QoSStage::~QoSStage()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int QoSStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int QoSStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;

    list_p_fabric_general_err qos_config_sl_errors;
    list_p_fabric_general_err vports_qos_config_sl_errors;
    int rc = this->p_ibdiag->BuildSMPQoSConfigSL(qos_config_sl_errors,
                                                 false,
                                                 progress_bar_retrieve_nodes);

    SCREEN_PRINT("\n");
    int rc2 = this->AnalyzeCheckResults(qos_config_sl_errors,
            CHECK_NAME_QOS_CONFIG_SL_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    rc = this->p_ibdiag->BuildSMPQoSConfigSL(vports_qos_config_sl_errors,
                                             true,
                                             progress_bar_retrieve_nodes);

    SCREEN_PRINT("\n");
    int rc3 = this->AnalyzeCheckResults(vports_qos_config_sl_errors,
            CHECK_NAME_VPORTS_QOS_CONFIG_SL_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    if (rc2)
        IBDIAGNET_RETURN(rc2);
    if (rc3)
        IBDIAGNET_RETURN(rc3);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int QoSStage::RunCheck()
{
    IBDIAGNET_ENTER;

    int rc1 = p_ibdiag->DumpQoSConfigSLCSVTable(*this->p_csv_out);
    if (rc1) {
        ERR_PRINT("Writing QoS config SL to csv file failed\n");
        ++this->num_errors;
    }

    int rc2 = p_ibdiag->DumpVPortsQoSConfigSLCSVTable(*this->p_csv_out);
    if (rc2) {
        ERR_PRINT("Writing VPorts QoS config SL to csv file failed\n");
        ++this->num_errors;
    }

    if (rc1)
        IBDIAGNET_RETURN(rc1);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}


RoutersStage::RoutersStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_ROUTERS, p_ibdiag)
{
    IBDIAGNET_ENTER;
    //mark this stage as skipped by default
    MarkStageAsSkipped();
    IBDIAGNET_RETURN_VOID;
}

RoutersStage::~RoutersStage()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int RoutersStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int RoutersStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    list_p_fabric_general_err errors;

    int rc2 = this->p_ibdiag->BuildRouterInfoDB(errors,
                                                progress_bar_retrieve_nodes);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(errors,
                                   STAGE_NAME_ROUTERS,
                                   rc2,
                                   IBDIAG_ERR_CODE_FABRIC_ERROR,
                                   this->num_errors,
                                   this->num_warnings);
    errors.clear();
    rc2 = this->p_ibdiag->BuildRouterTable(errors,
                                           progress_bar_retrieve_nodes);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(errors,
                                   STAGE_NAME_ROUTERS,
                                   rc2,
                                   IBDIAG_ERR_CODE_FABRIC_ERROR,
                                   this->num_errors,
                                   this->num_warnings);
    IBDIAGNET_RETURN(rc);
}

int RoutersStage::RunCheck()
{
    IBDIAGNET_ENTER;
    string file_str = generate_file_name(*this->p_base_path, DB_FILE);
    if (this->p_ibdiag->DumpRoutersInfoCSVTable(*this->p_csv_out)) {
        ERR_PRINT("Writing routers info to %s failed\n", file_str.c_str());
        ++this->num_errors;
    }

    if (this->p_ibdiag->DumpRoutersAdjSiteLocalSubnetCSVTable(
                            *this->p_csv_out)) {
        ERR_PRINT("Writing routers adjacent site local subnets table "
                  "to %s failed\n", file_str.c_str());
        ++this->num_errors;
    }

    if (this->p_ibdiag->DumpRoutersNextHopCSVTable(
                            *this->p_csv_out)) {
        ERR_PRINT("Writing routers next hop table "
                  "to %s failed\n", file_str.c_str());
        ++this->num_errors;
    }

    IBDIAGNET_RETURN(0);
}

SharpStage::SharpStage (IBDiag *p_ibdiag):
    Stage(STAGE_NAME_SHARP, p_ibdiag), can_send_mads_by_lid(false),
    dump_counters(false), clear_counters(false)
{
    IBDIAGNET_ENTER;
    this->MarkStageAsSkipped(); //this stage will be skipped by default

    IBDIAGNET_RETURN_VOID;
}

SharpStage::~SharpStage()
{
    IBDIAGNET_ENTER;
    delete p_sharp_mngr;
    IBDIAGNET_RETURN_VOID;
}

int SharpStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    if (check_if_can_send_mads_by_lid(this->p_ibdiag,
        this->can_send_mads_by_lid))
            IBDIAGNET_RETURN(1);

    p_sharp_mngr = new SharpMngr(p_ibdiag);
    IBDIAGNET_RETURN(0);
}

int SharpStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    if (!p_ibdiag->GetDiscoverFabricPtr() ||
        !p_ibdiag->GetIbisPtr())
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    list_p_fabric_general_err sharp_discovery_errors;

    int rc2 = p_sharp_mngr->BuildSharpConfigurationDB(sharp_discovery_errors,
                                         progress_bar_retrieve_nodes);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(sharp_discovery_errors,
            STAGE_NAME_SHARP,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc)
        IBDIAGNET_RETURN(rc);

    if (dump_counters) {
        rc2 = p_sharp_mngr->BuildPerformanceCountersDB(sharp_discovery_errors);
        if (rc) {
            ERR_PRINT("Failed to build AM PerfCounters DB.\n");
            IBDIAG_RETURN(rc);
        }

        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(sharp_discovery_errors,
                STAGE_NAME_SHARP,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
        if (rc)
            IBDIAGNET_RETURN(rc);
    }

    if (clear_counters) {
        rc2 = p_sharp_mngr->ResetPerformanceCounters(sharp_discovery_errors);
        if (rc) {
            ERR_PRINT("Failed to reset AM PerfCounters DB.\n");
            IBDIAG_RETURN(rc);
        }

        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(sharp_discovery_errors,
                STAGE_NAME_SHARP,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
        if (rc)
            IBDIAGNET_RETURN(rc);
    }

    IBDIAGNET_RETURN(rc);
}

int SharpStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    list_p_fabric_general_err sharp_discovery_errors;
    int rc2 = p_sharp_mngr->ConnectTreeEdges(sharp_discovery_errors);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(sharp_discovery_errors,
            STAGE_NAME_SHARP,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc)
        IBDIAGNET_RETURN(rc);


    rc2 = p_sharp_mngr->CheckSharpTrees(sharp_discovery_errors);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(sharp_discovery_errors,
            STAGE_NAME_SHARP,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc)
        IBDIAGNET_RETURN(rc);

    string file_str = generate_file_name(*this->p_base_path, SHARP_FILE);
    if (p_sharp_mngr->WriteSharpFile(file_str.c_str())) {
        ERR_PRINT("Write sharp DB to %s failed\n", file_str.c_str());
        ++this->num_errors;
    }
    this->AddGeneratedFileName("Sharp Data file", file_str);

    if (dump_counters) {
        file_str = generate_file_name(*this->p_base_path, SHARP_PM_FILE);
        if (p_sharp_mngr->WriteSharpPMFile(file_str.c_str())) {
            ERR_PRINT("Write sharp PM DB to %s failed\n", file_str.c_str());
            ++this->num_errors;
        }
        this->AddGeneratedFileName("Sharp PM Data file", file_str);
    }
    IBDIAGNET_RETURN(rc);
}
