/*
 *  Copyright (c) 2004-2019 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#ifndef _IBDIAGNET_STAGES_H_
#define _IBDIAGNET_STAGES_H_


#include <stdlib.h>

#include "infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h"
#include "infiniband/ibdiag/ibdiag.h"
#include "infiniband/ibdiag/sharp_mngr.h"
#include "ibdiagapp_types.h"


/******************************************************/
class DiscoveryStage: public Stage {
private:

protected:
    ////////////////////
    //members
    ////////////////////
    string device_name;
    u_int8_t port_num;
    u_int64_t port_guid;
    u_int8_t max_hops;
    bool to_check_dup_guids;
    bool to_check_dup_node_desc;
    bool to_check_vs_cap_smp;
    bool to_check_vs_cap_gmp;
    bool to_dump_fec_mode;
    bool initiliazed_with_device;
    bool initiliazed_with_guid;
    string psl_dump_file;
    bool no_mepi; // do not send SMPMlnxExtPortInfoMad
    string output_capability_file;
    string output_full_capability_file;
    string input_capability_file;
    bool show_ports_data_db2; //show all data on ports added in db2 format
                              //CapMsk2,FEC,Retransmission
    bool dump_lst_with_lmc; //write the LST file wite the ports lmc.
    string node_name_map_file;
    map<uint16_t, uint64_t> lid2guid_map;
    FilesBasedMKeyManager *p_mkey_manager;

    string load_from_file_path;
    bool load_from_file;

    ////////////////////
    //methods
    ////////////////////
    void PrintDuplicatedGuidsDetectErrors();
    virtual int PostDiscover() = 0;
public:
    ////////////////////
    //methods
    ////////////////////
    DiscoveryStage(IBDiag *p_ibdiag);
    virtual ~DiscoveryStage() {}

    inline void SetOutputCapabilityFile(const string &file_name) {
        this->output_capability_file = file_name;
    }
    inline void SetOutputFullCapabilityFile(const string &file_name) {
        this->output_full_capability_file = file_name;
    }
    inline void SetInputCapabilityFile(const string &file_name) {
        this->input_capability_file = file_name;
    }
    inline void SetNodeNameMapFile(const string &file_name) {
        this->node_name_map_file = file_name;
    }
    void CreateLid2GuidDB();

    inline void TurnOFFDUPGUIDsDetection() { this->to_check_dup_guids = false; }
    inline void TurnOFFDUPNodeDescDetection() { this->to_check_dup_node_desc = false; }
    inline void TurnOFFVsCapSMP() { this->to_check_vs_cap_smp = false; }
    inline void TurnOFFVsCapGMP() { this->to_check_vs_cap_gmp = false; }
    inline void SetFECModeDump(bool set_val) { this->to_dump_fec_mode = set_val; }
    inline void SetDeviceName(string device_name) {
        this->device_name = device_name;
        initiliazed_with_device = true;
    }
    inline void SetPortNum(u_int8_t port_num) {
        this->port_num = port_num;
        initiliazed_with_device = true;
    }
    inline void SetPortGUID(u_int64_t port_guid) {
        this->port_guid = port_guid;
        if (this->port_guid == OPTION_GUID_DEF_VAL)
            initiliazed_with_guid = false;
        else
            initiliazed_with_guid = true;
    }
    inline void SetMaxHops(u_int8_t max_hops) {
        this->max_hops = max_hops;
    }
    inline bool IsNeedDisplayAllPortsGuids() {
        if ((initiliazed_with_guid == true) && this->port_guid == 0) {
            return true;
        }
        return false;
    }
    inline void SetPSLDumpFile(const string &file_name) {
        this->psl_dump_file = file_name;
    }
    inline void SetNoMEPI() {
        this->no_mepi = true;
    }
    inline void SetDumpLSTWithLMC() {
        this->dump_lst_with_lmc = true;
    }

    //determine if to show new db format according to back_compat_db
    inline void SetShowPortsDataFormat(const string &format) {
        double version = atof(format.c_str());

        if (format == OPTION_BACK_COMPAT_DB_DEF_VAL)
            this->show_ports_data_db2 = true;
        else if (version < 2)
            this->show_ports_data_db2 = false;
    }

    inline void SetMKeyManager(FilesBasedMKeyManager *p_mkey_manager) {

        this->p_mkey_manager = p_mkey_manager;
    }

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
    virtual int DiscoverFabric() = 0;
};

/******************************************************/
class DiscoveryFabricStage: public DiscoveryStage {
private:
    string scope_port_guid_file;
    string exclude_scope_port_guid_file;

    virtual int PostDiscover();
public:
    ////////////////////
    //methods
    ////////////////////
    DiscoveryFabricStage(IBDiag *p_ibdiag);
    ~DiscoveryFabricStage() {}
    int Prepare();
    virtual int DiscoverFabric();

    inline void SetScopePortGuidFile(const string &file_name) {
        this->scope_port_guid_file = file_name;
    }

    inline void SetExcludeScopePortGuidFile(const string &file_name) {
        this->exclude_scope_port_guid_file = file_name;
    }

    inline void SetLoadFromFilePath(const string &file_name) {
        this->load_from_file_path = file_name;
        load_from_file = true;
    }

    inline string GetLoadFromFilePath() {
        return this->load_from_file_path;
    }

    inline bool IsLoadFromFile() {
        return load_from_file;
    }
};

/******************************************************/
class DiscoveryPathStage: public DiscoveryStage {
private:
    lid_t src_lid;
    lid_t dest_lid;
    direct_route_t dr_path;

    virtual int PostDiscover() { return 0; };
public:
    ////////////////////
    //methods
    ////////////////////
    DiscoveryPathStage(IBDiag *p_ibdiag);
    ~DiscoveryPathStage() {}
    int Prepare();
    virtual int DiscoverFabric();

    inline void SetDestLid(lid_t dest_lid) {
        this->dest_lid = dest_lid;
    }
    inline void SetSrcLid(lid_t src_lid) {
        this->src_lid = src_lid;
    }
    inline direct_route_t * GetDRPath() {
        return &(this->dr_path);
    }
    inline int SetDRPath(string value) {
        std::istringstream iss (value);
        string token;

        while(std::getline(iss, token, ',')) {
            this->dr_path.path.BYTE[this->dr_path.length++] =
                (u_int8_t)strtoul(token.c_str(), NULL, 0);
        }
        IBDIAGNET_RETURN(0);
    }
};

/******************************************************/
class LidsStage: public Stage {
private:
    ////////////////////
    //members
    ////////////////////

public:
    ////////////////////
    //methods
    ////////////////////
    LidsStage(IBDiag *p_ibdiag);
    ~LidsStage() {}

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};


/******************************************************/
class ToplogtMatchStage: public Stage {
private:
    ////////////////////
    //members
    ////////////////////
    IBFabric spec_fabric;

    string output_topo_file;
    string input_topo_file;
    string topo_ibnl_dir;

public:
    ////////////////////
    //methods
    ////////////////////
    ToplogtMatchStage(IBDiag *p_ibdiag);
    ~ToplogtMatchStage() {}

    inline IBFabric * GetSpecFabricPtr() {
        return &(this->spec_fabric);
    }

    inline void SetOutputTopoFile(const string &file_name) {
        this->output_topo_file = file_name;
        this->MarkStageAsActive();
    }
    inline void SetInputTopoFile(const string &file_name) {
        this->input_topo_file = file_name;
        this->MarkStageAsActive();
    }
    inline void SetTopoIbnlDir(const string &dir_name) {
        this->topo_ibnl_dir = dir_name;
    }    
    inline void ValidateTopoParams() {
        /*
        if ((this->input_topo_file == "") && (this->anchor_sys_name != "")) {
            WARN_PRINT("Local system name is specified, but topology is not given." \
                    "The former is ignored.\n");
            return;
        }
        */
        return;
    }

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};


/******************************************************/
class LinksStage: public Stage {
private:
    ////////////////////
    //members
    ////////////////////

public:
    ////////////////////
    //methods
    ////////////////////
    LinksStage(IBDiag *p_ibdiag);
    ~LinksStage() {}

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};


/******************************************************/
class SMStage: public Stage {
private:
    ////////////////////
    //members
    ////////////////////

public:
    ////////////////////
    //methods
    ////////////////////
    SMStage(IBDiag *p_ibdiag);
    ~SMStage() {}

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};


/******************************************************/
class PMStage: public Stage {
private:
    ////////////////////
    //members
    ////////////////////
    bool to_reset_counters;
    bool to_calc_ber;
    u_int32_t m_check_counters_bitset;
    //bool to_check_ext_speeds_counters;
    //bool to_check_ext_speeds_on_hca;
    bool en_per_lane_cnts;
    bool en_per_slvl_cntrs;
    u_int64_t ber_threshold;
    u_int32_t pause_time_between_samples;
    double diff_time_between_samples; //elapsed seconds between 2 samples

    map_str_uint64          counters_to_threshold_map;
    vector_p_pm_info_obj    first_sample_pm_info_obj_vector;
    vec_slvl_cntrs          m_slvl_cntrs_vec;

    bool can_send_mads_by_lid;

    ////////////////////
    //methods
    ////////////////////
    void Sleep(u_int32_t seconds);
    void PreparePerSlVlCounters();
public:
    ////////////////////
    //methods
    ////////////////////
    PMStage(IBDiag *p_ibdiag);
    ~PMStage();

    inline map_str_uint64& GetRefToCountersToThresholdMap() {
        return this->counters_to_threshold_map;
    }
    inline void SetResetCounters(bool set_val) {
        this->to_reset_counters = set_val;
    }
    inline void SetBERTest(bool set_val) {
        this->to_calc_ber = set_val;
    }
    inline void SetBERThreshold(u_int64_t ber_threshold) {
        if (ber_threshold == 0) {   /* ZERO is considered as overflow value ==> print all calculations */
            this->ber_threshold = OVERFLOW_VAL_64_BIT;
            return;
        }
        this->ber_threshold = ber_threshold;
    }

    inline void SetPauseTimeBetweenSamples(u_int32_t seconds) {
        this->pause_time_between_samples = seconds;
    }

    inline void SetCheckExtSpeedsCounters(const char * dev_type) {
        if (dev_type && strcmp(dev_type,"none")==0) {
            m_check_counters_bitset &= UNSET_EXT_SPEEDS_COUNTERS;
        } else {
            m_check_counters_bitset |= CHECK_EXT_SPEEDS_COUNTERS_ON_SW;
            if (dev_type && strcmp(dev_type,"all")==0) {
                m_check_counters_bitset |= CHECK_EXT_SPEEDS_COUNTERS_ON_ALL;
            }
        }
    }
    inline void SetPrintLLRCounters() {
        m_check_counters_bitset |= PRINT_LLR_COUNTERS;
    }
    inline void SetPerLaneCounters(bool set_val) {
        en_per_lane_cnts = set_val;
    }

    inline void SetPerSLVLCounters(bool set_val) {
        en_per_slvl_cntrs = set_val;
    }

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};


/******************************************************/
class RoutingStage: public Stage {
private:
    ////////////////////
    //members
    ////////////////////
    bool retrieved_switch_info_ok;
    bool retrieved_uc_fdbs_ok;
    bool retrieved_mc_fdbs_ok;
    bool retrieved_ext_switch_info_ok;
    bool retrieved_plft_data_ok;
    bool retrieved_ar_data_ok;
    bool retrieved_rn_data_ok;
    bool ar_enabled;
    bool is_fat_tree;
    bool use_vlr;
    bool use_vs_route;
    bool dump_full_ar;
    bool dump_rn_data;
    bool dump_rn_counters;
    bool clear_rn_counters;
    bool check_mcast_credloop;
    bool check_sl;
    bool dump_only;
    u_int8_t  default_sl;

    string sa_dump_file;

public:
    ////////////////////
    //methods
    ////////////////////
    RoutingStage(IBDiag *p_ibdiag);
    ~RoutingStage() {}

    inline void SetFatTree(bool set_val) {
        this->is_fat_tree = set_val;
    }
    inline void UseVLR() {
        this->use_vlr = true;
    }
    void UseVSRoute(){} // for deprecated option vs
    void DumpFullAr(){} // for deprecated option far
    void SkipVSRoute() {
        use_vs_route = false;
    }
    void SkipDumpFullAr() {
        dump_full_ar = false;
    }
    void DumpRNDada(){
        dump_rn_data = true;
    }
    void DumpRNCounters(){
        dump_rn_counters = true;
    }
    void ClearRNCounters(){
        clear_rn_counters = true;
    }
    void CheckMcastCredLoop() {
        check_mcast_credloop = true;
    }
    void SetCheckSL() {
        check_sl = true;
    }
    void SetDefaultSL(u_int8_t sl) {
        default_sl = sl;
    }
    void SetSADumpFile(const string &file_name) {
        this->sa_dump_file = file_name;
    }
    void DumpOnly(){
        dump_only = true;
    }

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};


/******************************************************/
class NodesInfoStage: public Stage {
private:
    ////////////////////
    //members
    ////////////////////
    bool can_send_mads_by_lid;

public:
    ////////////////////
    //methods
    ////////////////////
    NodesInfoStage(IBDiag *p_ibdiag);
    ~NodesInfoStage() {}

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};


/******************************************************/
class Speed_WidthStage: public Stage {
private:
    ////////////////////
    //members
    ////////////////////
    string expected_link_speed;
    string expected_link_width;

public:
    ////////////////////
    //methods
    ////////////////////
    Speed_WidthStage(IBDiag *p_ibdiag);
    ~Speed_WidthStage() {}

    inline void SetExpectedLinkSpeed(const string &expected_link_speed) {
	if (expected_link_speed == IB_DISABLE_LINK_SPEED_STR)
            this->expected_link_speed = "";
        else
            this->expected_link_speed = expected_link_speed;
    }
    inline void SetExpectedLinkWidth(const string &expected_link_width) {
        if (expected_link_width == IB_DISABLE_LINK_WIDTH_STR)
            this->expected_link_width = "";
        else
            this->expected_link_width = expected_link_width;
    }

    int Prepare();
    int RetrieveInfo() { return 0; }
    int RunCheck();
};


/******************************************************/
class PKeyStage: public Stage {
private:
    ////////////////////
    //members
    ////////////////////
    bool can_send_mads_by_lid;

public:
    ////////////////////
    //methods
    ////////////////////
    PKeyStage(IBDiag *p_ibdiag);
    ~PKeyStage();

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};


/******************************************************/
class AGUIDStage: public Stage {
private:
    ////////////////////
    //members
    ////////////////////
    bool can_send_mads_by_lid;

public:
    ////////////////////
    //methods
    ////////////////////
    AGUIDStage(IBDiag *p_ibdiag);
    ~AGUIDStage();

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};


/******************************************************/
class MlnxCntrsStage: public Stage {
private:
    ////////////////////
    //members
    ////////////////////
    bool can_send_mads_by_lid;
    bool to_reset_counters;
public:
    ////////////////////
    //methods
    ////////////////////
    MlnxCntrsStage(IBDiag *p_ibdiag);
    ~MlnxCntrsStage();

    inline void SetResetCounters(bool set_val) {
        this->to_reset_counters = set_val;
    }

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};


/******************************************************/
class VirtualizationStage: public Stage {
private:
    ////////////////////
    //members
    ////////////////////
    bool can_send_mads_by_lid;
public:
    ////////////////////
    //methods
    ////////////////////
    VirtualizationStage(IBDiag *p_ibdiag);
    ~VirtualizationStage();

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};

/******************************************************/
class TemperatureSensing: public Stage {
private:
    ////////////////////
    //members
    ////////////////////
public:
    ////////////////////
    //methods
    ////////////////////
    TemperatureSensing(IBDiag *p_ibdiag);
    ~TemperatureSensing();

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};

class RoutersStage : public Stage {
private:
public:
    RoutersStage(IBDiag *p_ibdiag);
    ~RoutersStage();

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};

/******************************************************/
class QoSStage : public Stage {
private:
public:
    QoSStage(IBDiag *p_ibdiag);
    ~QoSStage();

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};

/******************************************************/
class SharpStage: public Stage {
private:
    ////////////////////
    //members
    ////////////////////
    bool can_send_mads_by_lid;
    bool dump_counters;
    bool clear_counters;

    SharpMngr *p_sharp_mngr;
public:
    ////////////////////
    //methods
    ////////////////////
    SharpStage(IBDiag *p_ibdiag);
    ~SharpStage();

    void SetDumpCounters() { dump_counters = true; }
    void SetClearCounters() { clear_counters = true; }

    int Prepare();
    int RetrieveInfo();
    int RunCheck();
};

#endif          /* not defined _IBDIAGNET_STAGES_H_ */
