/*
 *  Copyright (c) 2004-2018 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <string.h>
#include <unistd.h>
extern char *optarg;
extern int optind, opterr, optopt;
#include <getopt.h>
#include <errno.h>
#include <infiniband/ibdm/Regexp.h>
#include <iostream>
#include <string>
#include <sstream>
#include <istream>
#include <algorithm>
#include <iterator>
#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h>


#include "ibdiagpath.h"

#include "ibdiagnet_debug_wrap.h"

#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_version.h>

/****************************************************/
/* globals */
/****************************************************/
DebugWrap debug_wrap_opj(DEFAULT_OUTPUT_PATH);


/****************************************************/
/*******************public methods*******************/
/****************************************************/

IBDiagpath::IBDiagpath() : IBDiagapp(DEFAULT_OUTPUT_PATH)
{

}

Stage* IBDiagpath::CreateStage(uint8_t stageIndex) {

    IBDIAGNET_ENTER;
    Stage *stage = NULL;
    switch (stageIndex) {
        case STAGE_IDX_DISCOVERY:
            //discovery stage
            stage = new DiscoveryPathStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_DISCOVERY);
            break;
        case STAGE_IDX_LINKS:
            //links stage
            stage = new LinksStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_LINKS);
            break;
        case STAGE_IDX_PM:
            //pm stage
            stage = new PMStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_PM);
            break;
        case STAGE_IDX_SPEED_WIDTH:
            //speed / width stage
            stage = new Speed_WidthStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_SPEED_WIDTH);
            break;
        case STAGE_IDX_MLNX_CNTRS:
            //Diagnostic Counters stage
            stage = new MlnxCntrsStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n", STAGE_NAME_MLNX_CNTRS);
            break;
        case STAGE_IDX_VIRTUALIZATION:
            //Virtualiztion stage
            stage = new VirtualizationStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n",
                          STAGE_NAME_VIRTUALIZATION);
            break;
        case STAGE_IDX_TEMPSENSING:
            //Temperature stage
            stage = new TemperatureSensing(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n",
                          STAGE_NAME_TEMPSENSING);
            break;
        case STAGE_IDX_ROUTERS:
            stage = new RoutersStage(&this->ibdiag_obj);
            if (!stage)
                ERR_PRINT("Failed to allocate %s stage\n",
                          STAGE_NAME_ROUTERS);
            break;
        case STAGE_IDX_QOS:
           //QoS stage
           stage = new QoSStage(&this->ibdiag_obj);
           if (!stage)
               ERR_PRINT("Failed to allocate %s stage\n",
                       STAGE_NAME_QOS);
           break;
        default:
            break;
    }

    IBDIAGNET_RETURN(stage);
}

IBDiagpath::~IBDiagpath()
{
    IBDIAGNET_ENTER;

    IBDIAGNET_RETURN_VOID;
}

//Returns: 0 - OK / 1 - parsing error / 2 - OK but need to exit / 3 - some other error
int IBDiagpath::HandleOption(string name, string value)
{
    IBDIAGNET_ENTER;
    int res = IBDiagapp::HandleOption(name, value);

    /* if res is 1 so it's unknown stage to ibdiagapp, need to
     * check for ibdiagpath stages */
    if (res == 1) {
        if (name == OPTION_SRC_LID) {
            lid_t srclid = (lid_t)strtoul(value.c_str(), NULL, 0);
            if (IsValidLid(srclid)) {
                ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                        OPTION_SRC_LID, value.c_str());
                IBDIAGNET_RETURN(1);
            }
            ((DiscoveryPathStage *)this->stages[STAGE_IDX_DISCOVERY])->
                SetSrcLid(srclid);
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_DEST_LID) {
            lid_t destlid = (lid_t)strtoul(value.c_str(), NULL, 0);
            if (IsValidLid(destlid)) {
                ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                        OPTION_DEST_LID, value.c_str());
                IBDIAGNET_RETURN(1);
            }
            ((DiscoveryPathStage *)this->stages[STAGE_IDX_DISCOVERY])->
                SetDestLid(destlid);
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_DR_PATH) {
            //0,1,8,5,3,1
            regExp drPath("(0(,([0-9])+)*)");
            rexMatch *p_rexRes = drPath.apply(value.c_str());
            if (!p_rexRes) {
                ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                        OPTION_DR_PATH, value.c_str());
                IBDIAGNET_RETURN(1);
            }
            ((DiscoveryPathStage *)this->stages[STAGE_IDX_DISCOVERY])->
                SetDRPath(value);
            IBDIAGNET_RETURN(0);
        } else if (name == OPTION_OUTPATH) {
            IBDIAGNET_RETURN(this->CreateOutputPath(value));
        }
    }

    IBDIAGNET_RETURN(res);
}

int IBDiagpath::Init()
{
    IBDIAGNET_ENTER;
    IBDiagapp::Init();

    this->AddDescription(IBDIAGPATH_TOOL_DESC);
    this->AddOptions(OPTION_SRC_LID, OPTION_SRC_LID_SHORT,
                     OPTION_SRC_LID_ARG_DESC, OPTION_SRC_LID_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_DEST_LID, OPTION_DEST_LID_SHORT,
                     OPTION_DEST_LID_ARG_DESC, OPTION_DEST_LID_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_DR_PATH, OPTION_DR_PATH_SHORT,
                     OPTION_DR_PATH_ARG_DESC, OPTION_DR_PATH_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_OUTPATH, OPTION_OUTPATH_SHORT,
                     OPTION_OUTPATH_ARG_DESC, OPTION_OUTPATH_DESC, DEFAULT_OUTPUT_PATH);
    ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->
        TurnOFFDUPGUIDsDetection();
    IBDIAGNET_RETURN(0);
}


#define N_ELEMENTS(array) sizeof(array)/sizeof(array[0])

#define COMPILE_TIME_ASSERT(pred)            \
    switch(0){case 0:case pred:;}

/****************************************************/
int main (int argc, char **argv)
{
    IBDiagpath ibdiagpath_obj;
    //Any created files will have mode
    //S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH (0666)
     umask(0);

    IBDIAGNET_RETURN(ibdiagpath_obj.Run(argc, argv));
}

