/*
 *  Copyright (c) 2004-2019 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <math.h>
#include <bitset>
#include <stdio.h>
#include <string.h>

#include "cable_diag.h"
#include "cable_diag_errs.h"

#define SET_PORT_NOT_SUPPORT_CAPABILTY(p_port) \
        if (p_port && (p_port->num > 0)) \
            p_port->counter2 = 1;

#define IS_PORT_NOT_SUPPORT_CAPABILTY(p_port) \
        ((p_port && (p_port->num > 0) && \
          (p_port->counter2 > 0)) ? true : false )


void EyeOpenGetDelegator(const clbck_data_t &clbck_data,
                         int rec_status,
                         void *p_attribute_data);

void CableInfoGetDelegator(const clbck_data_t &clbck_data,
                           int rec_status,
                           void *p_attribute_data);

static double mW_to_dBm(double mW)
{
    if (mW == 0)
        return -999.999;
    return 10 * log10(mW);
}

/****************************************************/
/***************** Class CableInfo ******************/
/****************************************************/

/*
 * private methods
 */
bool CableInfo::IsValidDateCode(char *p_date_code)
{
    for (int i = 0; i < SIZE_OF_DATE_CODE; i++)
        if (!isdigit(p_date_code[i]))
            return false;

    return true;
}

string CableInfo::ConvertSupportedSpeedToStr(u_int8_t supported_speed)
{
    IBDIAGNET_ENTER;
    string str = "";

    if (this->supported_speed & 0x1) {
        str += "SDR/";
    }
    if (this->supported_speed & 0x2) {
        str += "DDR/";
    }
    if (this->supported_speed & 0x4) {
        str += "QDR/";
    }
    if (this->supported_speed & 0x8) {
        str += "FDR/";
    }
    if (this->supported_speed & 0x10) {
        str += "EDR/";
    }
    if (this->supported_speed & 0x20) {
        str += "HDR/";
    }

    if (str == "") {
        str = "NA";
    } else {
        str.erase(str.size() - 1, 1);
    }
    IBDIAGNET_RETURN(str);
}

void CableInfo::SetCableTemperature(int8_t temp)
{
    IBDIAGNET_ENTER;

    if (this->cable_type == CABLE_TYPE_COPPER_UNEQUALIZED ||
            this->cable_type == CABLE_TYPE_COPPER_PASSIVE_EQUALIZED ||
            temp < CABLE_MIN_SUPPORT_TEMPERATURE ||
            temp > CABLE_MAX_SUPPORT_TEMPERATURE)
        IBDIAGNET_RETURN_VOID;

    char buff[24];
    sprintf(buff, "%dC", temp);
    this->temperature = buff;
    IBDIAGNET_RETURN_VOID;
}

string CableInfo::ConvertCableTypeToStr(u_int8_t cable_type)
{
    IBDIAGNET_ENTER;
    string str = "NA";
    if (this->cable_return_status != CABLEINFO_VS_STATUS_SUCCESS) {
        str = "NA - " + ConvertCableInfoVSStatusToStr(this->cable_return_status);
        IBDIAGNET_RETURN(str);
    }

    if (this->cable_type == 0x0) {
        str = "850 nm VCSEL";
    } else if (this->cable_type == 0x1) {
        str = "1310 nm VCSEL";
    } else if (this->cable_type == 0x2) {
        str = "1550 nm VCSEL";
    } else if (this->cable_type == 0x3) {
        str = "1310 nm FP";
    }  else if (this->cable_type == 0x4) {
        str = "1310 nm DFB";
    } else if (this->cable_type == 0x5) {
        str = "1550 nm DFB";
    }  else if (this->cable_type == 0x6) {
        str = "1310 nm EML";
    } else if (this->cable_type == 0x7) {
        str = "1550 nm EML";
    } else if (this->cable_type == 0x8) {
        str = "others";
    } else if (this->cable_type == 0x9) {
        str = "1490 nm DFB";
    } else if (this->cable_type == 0xa) {
        str = "Copper cable- unequalized";
    } else if (this->cable_type == 0xb) {
        str = "Copper cable- passive equalized";
    } else if (this->cable_type == 0xc) {
        str = "Copper cable- near and far end limiting active equalizers";
    } else if (this->cable_type == 0xd) {
        str = "Copper cable- far end limiting active equalizers";
    } else if (this->cable_type == 0xe) {
        str = "Copper cable- near end limiting active equalizers";
    } else if (this->cable_type == 0xf) {
        str = "Copper cable- linear active equalizers";
    } else if (this->cable_type == 0xff) {
        str = "NA";
    }

    IBDIAGNET_RETURN(str);
}

string CableInfo::ConvertCDREnableTxRxToStr(bool is_csv)
{
    IBDIAGNET_ENTER;
    string str;
    char buff[24];

    CLEAR_STRUCT(buff);
    if (IsModule() || IsActiveCable()) {
        if (is_csv) {
            sprintf(buff, "0x%x", this->cdr_control);
            str = buff;
        } else {
            //Tx
            if ((this->cdr_present & 0x2) != 0) {
                sprintf(buff, "0x%x ", this->cdr_control >> 4);
                str = buff;
            } else
                str = "N/A ";
            //Rx
            CLEAR_STRUCT(buff);
            if ((this->cdr_present & 0x1) != 0) {
                sprintf(buff, "0x%x", this->cdr_control & 0x0f);
                str += buff;
            } else
                str += "N/A";
        }
    } else  {
        if (is_csv)
            str = "\"NA\"";
        else
            str = "N/A N/A";
    }
    IBDIAGNET_RETURN(str);
}

string CableInfo::ConvertInputEqToStr(bool is_csv)
{
    IBDIAGNET_ENTER;
    string str;
    char buff[24];

    if (IsModule() || IsActiveCable()) {
        CLEAR_STRUCT(buff);
        if (!is_csv)
            sprintf(buff, "%u %u %u %u",
                    (this->input_eq >> 12) & 0xf,
                    (this->input_eq >> 8) & 0xf,
                    (this->input_eq >> 4) & 0xf,
                    this->input_eq & 0xf);
        else
            sprintf(buff, "%x%x%x%x",
                    (this->input_eq >> 12) & 0xf,
                    (this->input_eq >> 8) & 0xf,
                    (this->input_eq >> 4) & 0xf,
                    this->input_eq & 0xf);
        str = buff;
    } else {
        if (!is_csv)
            str = "N/A";
        else
            str = "\"NA\"";
    }
    IBDIAGNET_RETURN(str);
}

string CableInfo::ConvertOutputAmpToStr(bool is_csv)
{
    IBDIAGNET_ENTER;
    string str;
    char buff[24];

    if (IsModule() || IsActiveCable()) {
        CLEAR_STRUCT(buff);
        if (!is_csv)
            sprintf(buff, "%u %u %u %u",
                    (this->output_amp >> 12) & 0xf,
                    (this->output_amp >> 8) & 0xf,
                    (this->output_amp >> 4) & 0xf,
                    this->output_amp & 0xf);
        else
            sprintf(buff, "%x%x%x%x",
                    (this->output_amp >> 12) & 0xf,
                    (this->output_amp >> 8) & 0xf,
                    (this->output_amp >> 4) & 0xf,
                    this->output_amp & 0xf);
        str = buff;
    } else {
        if (!is_csv)
            str = "N/A";
        else
            str = "\"NA\"";
    }
    IBDIAGNET_RETURN(str);
}

string CableInfo::ConvertOutputEmpToStr(bool is_csv)
{
    IBDIAGNET_ENTER;
    string str;
    char buff[24];
    if (IsModule() || IsActiveCable()) {
        CLEAR_STRUCT(buff);
        if (!is_csv)
            sprintf(buff, "%u %u %u %u",
                    (this->output_emp >> 12) & 0xf,
                    (this->output_emp >> 8) & 0xf,
                    (this->output_emp >> 4) & 0xf,
                    this->output_emp & 0xf);
        else
            sprintf(buff, "%x%x%x%x",
                    (this->output_emp >> 12) & 0xf,
                    (this->output_emp >> 8) & 0xf,
                    (this->output_emp >> 4) & 0xf,
                    this->output_emp & 0xf);

        str = buff;
    } else {
        if (!is_csv)
            str = "N/A";
        else
            str = "\"NA\"";
    }
    IBDIAGNET_RETURN(str);
}

string CableInfo::ConvertAttenuationToStr(bool is_csv)
{
    IBDIAGNET_ENTER;
    string str;
    char buff[24];
    CLEAR_STRUCT(buff);

    if (IsPassiveCable()) {
        if (!is_csv)
            sprintf(buff, "%u %u %u %u",
                    this->attenuation_2_5g,
                    this->attenuation_5g,
                    this->attenuation_7g,
                    this->attenuation_12g);
        else
            sprintf(buff, "%u,%u,%u,%u",
                    this->attenuation_2_5g,
                    this->attenuation_5g,
                    this->attenuation_7g,
                    this->attenuation_12g);
        str = buff;
    } else {
        if (!is_csv)
            str = "N/A N/A N/A N/A";
        else
            str = "\"NA\",\"NA\",\"NA\",\"NA\"";
    }

    IBDIAGNET_RETURN(str);
}

string CableInfo::ConvertFwVersionToStr(bool is_csv)
{
    IBDIAGNET_ENTER;
    string str;

    if (IsMlnxMmf() || IsMlnxPsm()) {
        str = this->fw_version;
    } else  {
        if (!is_csv)
            str = "N/A";
        else
            str = "\"NA\"";
    }

    IBDIAGNET_RETURN(str);
}

string CableInfo::ConvertPageE9Addr128ToStr()
{
    IBDIAGNET_ENTER;
    string str;
    char buff[1024];
    CLEAR_STRUCT(buff);

    if (IsMlnxMmf() || IsMlnxPsm()) {

        sprintf(buff, "," U16H_FMT "," U16H_FMT "," U16H_FMT ","    \
                U16H_FMT "," U16H_FMT "," U16H_FMT ","    \
                U16H_FMT "," U16H_FMT "," U16H_FMT ","    \
                U16H_FMT "," U16H_FMT "," U16H_FMT ","    \
                U8H_FMT "," U8H_FMT "," U16H_FMT "," U16H_FMT ","   \
                U8H_FMT "," U8H_FMT "," U16H_FMT "," U16H_FMT ","   \
                U16H_FMT "," U16H_FMT,
	            cableinfo_E9_128_175.number_of_resets,
                cableinfo_E9_128_175.max_rssi_lane[0],
                cableinfo_E9_128_175.max_rssi_lane[1],
                cableinfo_E9_128_175.max_rssi_lane[2],
                cableinfo_E9_128_175.max_rssi_lane[3],
                cableinfo_E9_128_175.min_rssi_lane[0],
                cableinfo_E9_128_175.min_rssi_lane[1],
                cableinfo_E9_128_175.min_rssi_lane[2],
                cableinfo_E9_128_175.min_rssi_lane[3],
                cableinfo_E9_128_175.max_temp_module,
                cableinfo_E9_128_175.max_temp_vcsel,
                cableinfo_E9_128_175.min_temp_module,
                cableinfo_E9_128_175.min_temp_vd,
                cableinfo_E9_128_175.max_temp_vd,
                cableinfo_E9_128_175.min_temp_vcsel,
                cableinfo_E9_128_175.vcsel_temp_hist_0_40,
                cableinfo_E9_128_175.min_temp_tia,
                cableinfo_E9_128_175.max_temp_tia,
                cableinfo_E9_128_175.vcsel_temp_hist_55_70,
                cableinfo_E9_128_175.vcsel_temp_hist_40_55,
                cableinfo_E9_128_175.vcsel_temp_hist_85_100,
                cableinfo_E9_128_175.vcsel_temp_hist_70_85);
        str = buff;
    } else
        for(int i = 0; i < PageE9Addr128_FIELDS_SIZE; i++)
            str += ",\"NA\"";

    IBDIAGNET_RETURN(str);
}

string CableInfo::ConvertPageE9Addr176ToStr()
{
    IBDIAGNET_ENTER;
    string str;
    char buff[1024];
    CLEAR_STRUCT(buff);

    if (IsMlnxMmf() || IsMlnxPsm()) {

        sprintf(buff, "," U16H_FMT "," U16H_FMT "," U16H_FMT "," U16H_FMT ","   \
                U16H_FMT "," U16H_FMT "," U16H_FMT "," U16H_FMT "," \
                U16H_FMT "," U16H_FMT "," U16H_FMT "," U16H_FMT "," \
                U16H_FMT "," U16H_FMT "," U16H_FMT "," U16H_FMT,
                cableinfo_E9_176_211.bias_counter[0],
                cableinfo_E9_176_211.bias_counter[1],
                cableinfo_E9_176_211.bias_counter[2],
                cableinfo_E9_176_211.bias_counter[3],
                cableinfo_E9_176_211.mod_counter[0],
                cableinfo_E9_176_211.mod_counter[1],
                cableinfo_E9_176_211.mod_counter[2],
                cableinfo_E9_176_211.mod_counter[3],
                cableinfo_E9_176_211.min_vcc_tx,
                cableinfo_E9_176_211.max_vcc_tx,
                cableinfo_E9_176_211.min_vcc_rx,
                cableinfo_E9_176_211.max_vcc_rx,
                cableinfo_E9_176_211.min_vcc_bu,
                cableinfo_E9_176_211.max_vcc_bu,
                cableinfo_E9_176_211.min_vcc_sd,
                cableinfo_E9_176_211.max_vcc_sd);
        str = buff;
    } else
        for(int i = 0; i < PageE9Addr176_FIELDS_SIZE; i++)
            str += ",\"NA\"";

    IBDIAGNET_RETURN(str);
}

string CableInfo::ConvertDateCodeToStr()
{
    IBDIAGNET_ENTER;
    string str;

    if (this->date_code == CABLE_UNDEFINED)
        str = this->date_code;
    else
        str = this->date_code.substr(0, 2) + "-" + this->date_code.substr(2, 2) + "-" +
              this->date_code.substr(4, 2);

    IBDIAGNET_RETURN(str);
}

string CableInfo::ConvertRXOutputDisableToStr()
{
    IBDIAGNET_ENTER;
    char RX_output_disable[4];
    string str;

    CLEAR_STRUCT(RX_output_disable);
    sprintf(RX_output_disable, "0x%01x", (this->adaptive_eq_control >> 4) & 0xf);

    str = RX_output_disable;

    IBDIAGNET_RETURN(str);
}

string CableInfo::ConvertTXAdaptiveEqualizationEnableToStr()
{
    IBDIAGNET_ENTER;
    char TX_adaptive_equalization_enable[4];
    string str;

    CLEAR_STRUCT(TX_adaptive_equalization_enable);
    sprintf(TX_adaptive_equalization_enable, "0x%01x", this->adaptive_eq_control & 0xf);

    str = TX_adaptive_equalization_enable;

    IBDIAGNET_RETURN(str);
}

int CableInfo::SetCableInfo(u_int8_t not_supported_status,
                            u_int8_t *eeprom_payload,
                            u_int8_t eeprom_index,
                            u_int8_t page_num)
{
    IBDIAGNET_ENTER;
    string whitespaces =  " \t\f\v\n\r";
    size_t found;
    char buff[1024];
    int rc = IBDIAG_SUCCESS_CODE;
    if (not_supported_status) {
        if (this->cable_return_status == 0xff)
            this->cable_return_status = not_supported_status;
        IBDIAGNET_RETURN(rc);
    }

    this->cable_return_status = CABLEINFO_VS_STATUS_SUCCESS;

    if (page_num == EEPROM_CABLE_PAGE_FIRST_GET) {
        if (eeprom_index == EEPROM_CABLE_ADDR_FIRST_GET) { //cable_info_128_167
            CableInfo_Payload_Addr_128_167 cable_info_128_167;
            CableInfo_Payload_Addr_128_167_unpack(&cable_info_128_167, eeprom_payload);

            //eth_com_codes_10g_40g
            this->eth_com_codes_10g_40g = cable_info_128_167.SpecCompliance0;

            //identifier
            this->identifier = cable_info_128_167.Identifier;

            //connector_type
            this->connector = cable_info_128_167.Connector;

            //cable speed
            this->supported_speed = cable_info_128_167.ExtendedModuleCodes;

            //device_technology
            this->cable_type = (u_int8_t)(cable_info_128_167.DeviceTech >> 4);

            //vendor
            size_t s = sizeof(cable_info_128_167.VendorName);
            memcpy(buff, cable_info_128_167.VendorName, s);
            buff[s] = '\0';
            this->vendor = buff;
            found = this->vendor.find_last_not_of(whitespaces);
            if (found != string::npos)
                this->vendor.erase(found + 1);
            else
                this->vendor.clear();           //str is all whitespaces
            this->vendor = DescToCsvDesc(this->vendor);

            //OUI
            sprintf(buff, "0x%x", cable_info_128_167.VendorOUI);
            this->oui = buff;

            //length - there are so many cables that not according to spec so we try the follow approach
            this->lengthsmfiber = cable_info_128_167.LengthSMFiber;
            this->lengthom3     = cable_info_128_167.LengthOM3;
            this->lengthom2     = cable_info_128_167.LengthOM2;
            this->lengthom1     = cable_info_128_167.LengthOM1;
            this->lengthcopper  = cable_info_128_167.LengthCopperOrActive;

            sprintf(buff, "NA");
            this->length_str = buff;
           // IsModule: if (indication = Module) =>  length = NA
           if ( !IsModule() && (this->identifier == QSFP_ID_VALUE ||
                                this->identifier == QSFP_PLUS_ID_VALUE ||
                                this->identifier == QSFP28_ID_VALUE)) {
                //real value
                if (cable_info_128_167.LengthCopperOrActive != 0) {
                    if (cable_info_128_167.LengthCopperOrActive == 255) {
                        sprintf(buff, "greater than 254 m");
                        this->length_str = buff;
                    } else {
                        sprintf(buff, "%u m", cable_info_128_167.LengthCopperOrActive);
                        this->length_str = buff;
                    }
                } else {    //try to guess the length
                    if (cable_info_128_167.LengthOM1 != 0) {
                        sprintf(buff, "%u m", cable_info_128_167.LengthOM1);
                        this->length_str = buff;
                    } else if (cable_info_128_167.LengthOM2 != 0) {
                        sprintf(buff, "%u m", cable_info_128_167.LengthOM2);
                        this->length_str = buff;
                    }  else if (cable_info_128_167.LengthOM3 != 0) {
                        sprintf(buff, "%u m", cable_info_128_167.LengthOM3 * 2);   //mul by 2 because in units of 2 m.
                        this->length_str = buff;
                    } else if (cable_info_128_167.LengthSMFiber != 0) {
                        sprintf(buff, "%u km", cable_info_128_167.LengthSMFiber);
                        this->length_str = buff;
                    }
                }
            }

            this->cdr_present = cable_info_128_167.CdrPresent;
            this->transmitter_technology = cable_info_128_167.DeviceTech;

            if (cable_info_128_167.ExtPowerClass == 0)
                this->power_class = cable_info_128_167.PowerClass + 1;
            else
                this->power_class = cable_info_128_167.ExtPowerClass + 4;

        } else if (eeprom_index == EEPROM_CABLE_ADDR_SECOND_GET) { //cable_info_168_211

            CableInfo_Payload_Addr_168_211 cable_info_168_211;
            CableInfo_Payload_Addr_168_211_unpack(&cable_info_168_211, eeprom_payload);

            this->attenuation_2_5g = cable_info_168_211.Attenuation_2_5g;
            this->attenuation_5g = cable_info_168_211.Attenuation_5g;
            this->attenuation_7g = cable_info_168_211.Attenuation_7g;
            this->attenuation_12g = cable_info_168_211.Attenuation_12g;

            this->eth_com_codes_ext = cable_info_168_211.EthComCodesExt;
            this->qsfp_options = cable_info_168_211.QSFPOptions;
            //PN
            size_t s = sizeof(cable_info_168_211.VendorPN);
            memcpy(buff, cable_info_168_211.VendorPN, s);
            buff[s] = '\0';
            this->pn = buff;
            found = this->pn.find_last_not_of(whitespaces);
            if (found != string::npos)
                this->pn.erase(found + 1);
            else
                this->pn.clear();               //str is all whitespaces
            this->pn = DescToCsvDesc(this->pn);

            //SN
            s = sizeof(cable_info_168_211.VendorSN);
            memcpy(buff, cable_info_168_211.VendorSN, s);
            buff[s] = '\0';
            this->sn = buff;
            found = this->sn.find_last_not_of(whitespaces);
            if (found != string::npos)
                this->sn.erase(found + 1);
            else
                this->sn.clear();               //str is all whitespaces
            this->sn = DescToCsvDesc(this->sn);

            //Rev
            s = sizeof(cable_info_168_211.VendorRev);
            memcpy(buff, cable_info_168_211.VendorRev, s);
            buff[s] = '\0';
            this->rev = buff;
            found = this->rev.find_last_not_of(whitespaces);
            if (found != string::npos)
                this->rev.erase(found + 1);
            else
                this->rev.clear();              //str is all whitespaces
            this->rev = DescToCsvDesc(this->rev);
        } else if (eeprom_index == EEPROM_CABLE_ADDR_THIRD_GET) { //cable_info_0_47
            CableInfo_Payload_Addr_0_47 cable_info_0_47;
            CableInfo_Payload_Addr_0_47_unpack(&cable_info_0_47, eeprom_payload);

            this->SetCableTemperature((int8_t)(cable_info_0_47.DiagDeviceTemperature >> 8));
            this->diag_supply_voltage = cable_info_0_47.DiagSupplyVoltage;

            this->RX1Power = cable_info_0_47.DiagRX1Power / 10000.0;
            this->RX2Power = cable_info_0_47.DiagRX2Power / 10000.0;
            this->RX3Power = cable_info_0_47.DiagRX3Power / 10000.0;
            this->RX4Power = cable_info_0_47.DiagRX4Power / 10000.0;
            this->TX1Bias = cable_info_0_47.DiagTX1Bias / 500.0;
            this->TX2Bias = cable_info_0_47.DiagTX2Bias / 500.0;
            this->TX3Bias = cable_info_0_47.DiagTX3Bias / 500.0;
            this->CDR_TX_RX_loss_indicator = cable_info_0_47.CDRTXRXLossIndicator;
            this->adaptive_equalization_fault = cable_info_0_47.AdaptiveEqualizationFault;
            this->TX_RX_LOL_indicator = cable_info_0_47.TXRXLOLIndicator;
            this->temperature_alarm_and_Warning = cable_info_0_47.TemperatureAlarmandWarning;
            this->voltage_alarm_and_warning = cable_info_0_47.VoltageAlarmAndWarning;
            this->RX_power_alarm_and_warning = cable_info_0_47.RXPowerAlarmAndWarning;
            this->TX_power_alarm_and_warning = cable_info_0_47.TXPowerAlarmAndWarning;

            u_int8_t *p_TX_power_alarm_and_warning = (u_int8_t *)&this->TX_bias_alarm_and_warning;
            p_TX_power_alarm_and_warning[0] = cable_info_0_47.TXBiasAlarmAndWarningLow;
            p_TX_power_alarm_and_warning[1] = cable_info_0_47.TXBiasAlarmAndWarningHigh;

        } else if (eeprom_index == CABLE_ADDR_48) { //cable_info_48_79

            CableInfo_Payload_Addr_48_79 cable_info_48_79;
            CableInfo_Payload_Addr_48_79_unpack(&cable_info_48_79, eeprom_payload);

            this->TX4Bias = cable_info_48_79.DiagTX4Bias / 500.0;
            this->TX1Power = cable_info_48_79.DiagTX1Power / 10000.0;
            this->TX2Power = cable_info_48_79.DiagTX2Power / 10000.0;
            this->TX3Power = cable_info_48_79.DiagTX3Power / 10000.0;
            this->TX4Power = cable_info_48_79.DiagTX4Power / 10000.0;

        } else if (eeprom_index == CABLE_ADDR_80) { //cable_info_80_127
            CableInfo_Payload_Addr_80_127 cable_info_80_127;
            CableInfo_Payload_Addr_80_127_unpack(&cable_info_80_127, eeprom_payload);

            this->cdr_control = cable_info_80_127.CdrControl;

        } else if (eeprom_index == CABLE_ADDR_212) { //cable_info_212_255
            CableInfo_Payload_Addr_212_255 cable_info_212_255;
            CableInfo_Payload_Addr_212_255_unpack(&cable_info_212_255, eeprom_payload);

            //Date Code
            if (this->IsValidDateCode(cable_info_212_255.DateCode)) {
                memcpy(buff, cable_info_212_255.DateCode, SIZE_OF_DATE_CODE);
                buff[SIZE_OF_DATE_CODE] = '\0';
                this->date_code = buff;
            }

            memcpy(&this->lot, cable_info_212_255.DateCode + LOT_OFFSET, sizeof(this->lot));

            this->RXpower_type = cable_info_212_255.DDMIType & 8;

            this->mlnx_vendor_byte = cable_info_212_255.MlnxVendorByte;
            this->mlnx_revision = cable_info_212_255.MlnxRevision;
            this->mellanox_cap = cable_info_212_255.MellanoxCap;

            if (this->nominal_br_100 == 0xff)
                this->nominal_br = cable_info_212_255.NominalBr_250 / 4;
            else
                this->nominal_br = this->nominal_br_100 / 10;

            //SN
            size_t s = sizeof(cable_info_212_255.MlnxSFGSN);
            memcpy(buff, cable_info_212_255.MlnxSFGSN, s);
            buff[s] = '\0';
            this->mlnx_sfg_sn = buff;
            found = this->mlnx_sfg_sn.find_last_not_of(whitespaces);
            if (found != string::npos)
                this->mlnx_sfg_sn.erase(found + 1);
            else
                this->mlnx_sfg_sn.clear();               //str is all whitespaces
            this->mlnx_sfg_sn = DescToCsvDesc(this->mlnx_sfg_sn);

        } else
            rc = IBDIAG_ERR_CODE_INCORRECT_ARGS;

    } else if (page_num == CABLE_PAGE_03) {
        if (eeprom_index == CABLE_ADDR_212) {
            CableInfo_Payload_Page_3_Addr_212_255 cable_info_P_3_212_255;
            CableInfo_Payload_Page_3_Addr_212_255_unpack(
                    &cable_info_P_3_212_255, eeprom_payload);
            this->input_eq = cable_info_P_3_212_255.InputEq;
            this->output_amp = cable_info_P_3_212_255.OutputAmp;
            this->output_emp = cable_info_P_3_212_255.OutputEmp;
            this->TX_adaptive_equalization_freeze =
                  cable_info_P_3_212_255.TXAdaptiveEqualizationFreeze;
            this->adaptive_eq_control = cable_info_P_3_212_255.AdaptiveEqControl;
        }
    } else if (page_num == CABLE_PAGE_F0) {
        if (eeprom_index == CABLE_ADDR_128) {
            CableInfo_Payload_Page_F0_Addr_128_171 cable_info_P_F0_128_171;
            CableInfo_Payload_Page_F0_Addr_128_171_unpack(
                    &cable_info_P_F0_128_171, eeprom_payload);

            //FW Rev
            size_t s = sizeof(cable_info_P_F0_128_171.FWVersion);
            memcpy(buff, cable_info_P_F0_128_171.FWVersion, s);
            buff[s] = '\0';
            this->fw_version = buff;
            found = this->fw_version.find_last_not_of(whitespaces);
            if (found != string::npos)
                this->fw_version.erase(found + 1);
            else
                this->fw_version.clear();               //str is all whitespaces
            this->fw_version = DescToCsvDesc(this->fw_version);
        }
    } else if (page_num == CABLE_PAGE_E9) {
        if (eeprom_index == CABLE_ADDR_128) {
            CableInfo_Payload_Page_E9_Addr_128_175 cable_info_P_E9_128_175;
            CableInfo_Payload_Page_E9_Addr_128_175_unpack(
                    &cable_info_P_E9_128_175, eeprom_payload);

            this->cableinfo_E9_128_175 = cable_info_P_E9_128_175;
        } else if (eeprom_index == CABLE_ADDR_176) {
            CableInfo_Payload_Page_E9_Addr_176_211 cable_info_P_E9_176_211;
            CableInfo_Payload_Page_E9_Addr_176_211_unpack(
                    &cable_info_P_E9_176_211, eeprom_payload);

            this->cableinfo_E9_176_211 = cable_info_P_E9_176_211;
        }
    } else
        rc = IBDIAG_ERR_CODE_INCORRECT_ARGS;

    IBDIAGNET_RETURN(rc);
}

string CableInfo::c_str()
{
    IBDIAGNET_ENTER;
    char buff[2048];

    sprintf(buff, "Vendor: %s\nOUI: %s\nPN: %s\nSN: %s\nRev: %s\nLength: %s\nType: %s\nSupportedSpeed: %s\nTemperature: %s\n"
            "PowerClass: %u\nNominalBitrate: %u Gb/s\nCDREnableTxRx: %s\n"
            "InputEq: %s\nOutputAmp: %s\nOutputEmp: %s\nFW Version: %s\n"
            "Attenuation(2.5,5,7,12): %s\nRX power type: %s\n"
            "RX1 Power: %.3f mW, %.3f dBm\n"
            "RX2 Power: %.3f mW, %.3f dBm\n"
            "RX3 Power: %.3f mW, %.3f dBm\n"
            "RX4 Power: %.3f mW, %.3f dBm\n"
            "TX1 Bias: %.3f mA\n"
            "TX2 Bias: %.3f mA\n"
            "TX3 Bias: %.3f mA\n"
            "TX4 Bias: %.3f mA\n"
            "TX1 Power: %.3f mW, %.3f dBm\n"
            "TX2 Power: %.3f mW, %.3f dBm\n"
            "TX3 Power: %.3f mW, %.3f dBm\n"
            "TX4 Power: %.3f mW, %.3f dBm\n"
            "CDR Latched TX/RX Loss Indicator: " U8H_FMT "\n"
            "Latched Adaptive Equalization Fault: " U8H_FMT "\n"
            "Latched  TX / RX LOL Indicator: " U8H_FMT "\n"
            "Latched Temperature Alarm and Warning: " U8H_FMT "\n"
            "Latched Voltage Alarm and Warning: " U8H_FMT "\n"
            "RX Power Alarm and Warning: " U16H_FMT "\n"
            "TX Bias Alarm and Warning: " U16H_FMT "\n"
            "TX Power Alarm and Warning: " U16H_FMT "\n"
            "Supply Voltage Reporting: " U16H_FMT "\n"
            "Transmitter Technology: " U8H_FMT "\n"
            "Extended Specification Compliance Codes: " U8H_FMT "\n"
            "DateCode: " STR_FMT "\n"
            "Lot: " U16H_FMT "\n"
            "TX Adaptive Equalization Freeze: " U8H_FMT "\n"
            "RX Output Disable: " STR_FMT "\n"
            "TX Adaptive Equalization Enable: " STR_FMT "\n",
            this->vendor.c_str(),
            this->oui.c_str(),
            this->pn.c_str(),
            this->sn.c_str(),
            this->rev.c_str(),
            this->length_str.c_str(),
            this->ConvertCableTypeToStr(this->cable_type).c_str(),
            this->ConvertSupportedSpeedToStr(this->supported_speed).c_str(),
            this->temperature.c_str(),
            this->power_class,
            this->nominal_br,
            this->ConvertCDREnableTxRxToStr(false).c_str(),
            this->ConvertInputEqToStr(false).c_str(),
            this->ConvertOutputAmpToStr(false).c_str(),
            this->ConvertOutputEmpToStr(false).c_str(),
            this->ConvertFwVersionToStr(false).c_str(),
            this->ConvertAttenuationToStr(false).c_str(),
            this->RXpower_type? "Average Power" : "OMA",
            this->RX1Power, mW_to_dBm(this->RX1Power),
            this->RX2Power, mW_to_dBm(this->RX2Power),
            this->RX3Power, mW_to_dBm(this->RX3Power),
            this->RX4Power, mW_to_dBm(this->RX4Power),
            this->TX1Bias,
            this->TX2Bias,
            this->TX3Bias,
            this->TX4Bias,
            this->TX1Power, mW_to_dBm(this->TX1Power),
            this->TX2Power, mW_to_dBm(this->TX2Power),
            this->TX3Power, mW_to_dBm(this->TX3Power),
            this->TX4Power, mW_to_dBm(this->TX4Power),
            this->CDR_TX_RX_loss_indicator,
            this->adaptive_equalization_fault,
            this->TX_RX_LOL_indicator,
            this->temperature_alarm_and_Warning,
            this->voltage_alarm_and_warning,
            this->RX_power_alarm_and_warning,
            this->TX_bias_alarm_and_warning,
            this->TX_power_alarm_and_warning,
            this->diag_supply_voltage,
            this->transmitter_technology,
            this->eth_com_codes_ext,
            this->ConvertDateCodeToStr().c_str(),
            this->lot,
            this->TX_adaptive_equalization_freeze,
            this->ConvertRXOutputDisableToStr().c_str(),
            this->ConvertTXAdaptiveEqualizationEnableToStr().c_str());

    IBDIAGNET_RETURN(buff);
}


string CableInfo::hdr_str()
{
    IBDIAGNET_ENTER;
    string result = "NodeGuid,PortGuid,PortNum";
    result += ",Vendor,OUI,PN,SN,Rev,LengthSMFiber,LengthOM3,LengthOM2,LengthOM1,LengthCopperOrActive";
    result += ",Identifier,Connector,Type,SupportedSpeed";
    result += ",LengthDesc,TypeDesc,SupportedSpeedDesc,Temperature,PowerClass,NominalBitrate,CDREnableTxRx";
    result += ",InputEq,OutputAmp,OutputEmp,FWVersion,Attenuation2.5G,Attenuation5G,Attenuation7G,Attenuation12G";
    result += ",RXPowerType,RX1Power,RX2Power,RX3Power,RX4Power";
    result += ",TX1Bias,TX2Bias,TX3Bias,TX4Bias";
    result += ",TX1Power,TX2Power,TX3Power,TX4Power";
    result += ",RX1LatchedLossIndicator,RX2LatchedLossIndicator";
    result += ",RX3LatchedLossIndicator,RX4LatchedLossIndicator,TX1LatchedLossIndicator";
    result += ",TX2LatchedLossIndicator,TX3LatchedLossIndicator,TX4LatchedLossIndicator";
    result += ",TX1AdaptiveEqualizationFaultIndicator,TX2AdaptiveEqualizationFaultIndicator";
    result += ",TX3AdaptiveEqualizationFaultIndicator,TX4AdaptiveEqualizationFaultIndicator";
    result += ",RX1CDRLOL,RX2CDRLOL,RX3CDRLOL,RX4CDRLOL,TX1CDRLOL,TX2CDRLOL,TX3CDRLOL,TX4CDRLOL";
    result += ",HighTemperatureAlarm,LowTemperatureAlarm,HighTemperatureWarning";
    result += ",LowTemperatureWarning,InitializationFlagComplete,HighSupplyVoltageAlarm";
    result += ",LowSupplyVoltageAlarm,HighSupplyVoltageWarning,LowSupplyVoltageWarning";
    result += ",HighRX1PowerAlarm,LowRX1PowerAlarm,HighRX1PowerWarning,LowRX1PowerWarning";
    result += ",HighRX2PowerAlarm,LowRX2PowerAlarm,HighRX2PowerWarning,LowRX2PowerWarning";
    result += ",HighRX3PowerAlarm,LowRX3PowerAlarm,HighRX3PowerWarning,LowRX3PowerWarning";
    result += ",HighRX4PowerAlarm,LowRX4PowerAlarm,HighRX4PowerWarning,LowRX4PowerWarning";
    result += ",HighTX1BiasAlarm,LowTX1BiasAlarm,HighTX1BiasWarning,LowTX1BiasWarning";
    result += ",HighTX2BiasAlarm,LowTX2BiasAlarm,HighTX2BiasWarning,LowTX2BiasWarning";
    result += ",HighTX3BiasAlarm,LowTX3BiasAlarm,HighTX3BiasWarning,LowTX3BiasWarning";
    result += ",HighTX4BiasAlarm,LowTX4BiasAlarm,HighTX4BiasWarning,LowTX4BiasWarning";
    result += ",HighTX1PowerAlarm,LowTX1PowerAlarm,HighTX1PowerWarning,LowTX1PowerWarning";
    result += ",HighTX2PowerAlarm,LowTX2PowerAlarm,HighTX2PowerWarning,LowTX2PowerWarning";
    result += ",HighTX3PowerAlarm,LowTX3PowerAlarm,HighTX3PowerWarning,LowTX3PowerWarning";
    result += ",HighTX4PowerAlarm,LowTX4PowerAlarm,HighTX4PowerWarning,LowTX4PowerWarning";
    result += ",SupplyVoltageReporting,TransmitterTechnology,ActiveWavelengthControl";
    result += ",CooledTransmitterDevice,ActivePinDetector,TunableTransmitter";
    result += ",ExtendedSpecificationComplianceCodes,DateCode,Lot";
    result += ",TX1AdaptiveEqualizationFreeze,TX2AdaptiveEqualizationFreeze";
    result += ",TX3AdaptiveEqualizationFreeze,TX4AdaptiveEqualizationFreeze";
    result += ",RXOutputDisable,TXAdaptiveEqualizationEnable";
    IBDIAGNET_RETURN(result);
}


string CableInfo::csv_str()
{
    IBDIAGNET_ENTER;
    char buff[2048];

    //NodeGuid,PortGuid,PortNum,
    //Vendor,OUI,PN,SN,Rev,LengthSMFiber,LengthOM3,LengthOM2,LengthOM1,LengthCopperOrActive,
    //Identifier,Connector,Type,SupportedSpeed,
    //LengthDesc,TypeDesc,SupportedSpeedDesc
    sprintf(buff, U64H_FMT "," U64H_FMT "," U32D_FMT ","    \
            "\"" STR_FMT "\"," STR_FMT ",\"" STR_FMT "\",\"" STR_FMT "\",\"" STR_FMT "\","  \
            U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT ","    \
            U8H_FMT "," U8H_FMT "," U32D_FMT "," U32D_FMT ","   \
            "\"" STR_FMT "\",\"" STR_FMT "\",\"" STR_FMT "\"," STR_FMT ","  \
            U32D_FMT "," U32D_FMT "," STR_FMT "," STR_FMT "," STR_FMT   \
            "," STR_FMT ",\"" STR_FMT "\"," STR_FMT "," U32D_FMT "," \
            FLT_FMT "," FLT_FMT "," FLT_FMT "," FLT_FMT "," \
            FLT_FMT "," FLT_FMT "," FLT_FMT "," FLT_FMT "," \
            FLT_FMT "," FLT_FMT "," FLT_FMT "," FLT_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT \
            "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT \
            "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT \
            "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT \
            "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT \
            "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT \
            "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT \
            "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT \
            "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT \
            "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT \
            "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT \
            "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT \
            "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT \
            "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U16H_FMT ",0x%01x" \
            "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," STR_FMT \
            "," U16H_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," STR_FMT \
            "," STR_FMT,
            this->p_port->p_node->guid_get(), this->p_port->guid_get(), this->p_port->num,
            this->vendor.c_str(),
            this->oui.c_str(),
            this->pn.c_str(),
            this->sn.c_str(),
            this->rev.c_str(),
            this->lengthsmfiber,
            this->lengthom3,
            this->lengthom2,
            this->lengthom1,
            this->lengthcopper,
            this->identifier,
            this->connector,
            this->cable_type,
            this->supported_speed,
            this->length_str.c_str(),
            this->ConvertCableTypeToStr(this->cable_type).c_str(),
            this->ConvertSupportedSpeedToStr(this->supported_speed).c_str(),
            this->temperature.c_str(),
            this->power_class,
            this->nominal_br * 1000,
            this->ConvertCDREnableTxRxToStr(true).c_str(),
            this->ConvertInputEqToStr(true).c_str(),
            this->ConvertOutputAmpToStr(true).c_str(),
            this->ConvertOutputEmpToStr(true).c_str(),
            this->ConvertFwVersionToStr(true).c_str(),
            // Returns 3 fields
            this->ConvertAttenuationToStr(true).c_str(),
            this->RXpower_type,
            this->RX1Power,
            this->RX2Power,
            this->RX3Power,
            this->RX4Power,
            this->TX1Bias,
            this->TX2Bias,
            this->TX3Bias,
            this->TX4Bias,
            this->TX1Power,
            this->TX2Power,
            this->TX3Power,
            this->TX4Power,
            (u_int8_t)bitset<8>(this->CDR_TX_RX_loss_indicator)[0],
            (u_int8_t)bitset<8>(this->CDR_TX_RX_loss_indicator)[1],
            (u_int8_t)bitset<8>(this->CDR_TX_RX_loss_indicator)[2],
            (u_int8_t)bitset<8>(this->CDR_TX_RX_loss_indicator)[3],
            (u_int8_t)bitset<8>(this->CDR_TX_RX_loss_indicator)[4],
            (u_int8_t)bitset<8>(this->CDR_TX_RX_loss_indicator)[5],
            (u_int8_t)bitset<8>(this->CDR_TX_RX_loss_indicator)[6],
            (u_int8_t)bitset<8>(this->CDR_TX_RX_loss_indicator)[7],
            (u_int8_t)bitset<8>(this->adaptive_equalization_fault)[4],
            (u_int8_t)bitset<8>(this->adaptive_equalization_fault)[5],
            (u_int8_t)bitset<8>(this->adaptive_equalization_fault)[6],
            (u_int8_t)bitset<8>(this->adaptive_equalization_fault)[7],
            (u_int8_t)bitset<8>(this->TX_RX_LOL_indicator)[0],
            (u_int8_t)bitset<8>(this->TX_RX_LOL_indicator)[1],
            (u_int8_t)bitset<8>(this->TX_RX_LOL_indicator)[2],
            (u_int8_t)bitset<8>(this->TX_RX_LOL_indicator)[3],
            (u_int8_t)bitset<8>(this->TX_RX_LOL_indicator)[4],
            (u_int8_t)bitset<8>(this->TX_RX_LOL_indicator)[5],
            (u_int8_t)bitset<8>(this->TX_RX_LOL_indicator)[6],
            (u_int8_t)bitset<8>(this->TX_RX_LOL_indicator)[7],
            (u_int8_t)bitset<8>(this->temperature_alarm_and_Warning)[7],
            (u_int8_t)bitset<8>(this->temperature_alarm_and_Warning)[6],
            (u_int8_t)bitset<8>(this->temperature_alarm_and_Warning)[5],
            (u_int8_t)bitset<8>(this->temperature_alarm_and_Warning)[4],
            (u_int8_t)bitset<8>(this->temperature_alarm_and_Warning)[0],
            (u_int8_t)bitset<8>(this->voltage_alarm_and_warning)[7],
            (u_int8_t)bitset<8>(this->voltage_alarm_and_warning)[6],
            (u_int8_t)bitset<8>(this->voltage_alarm_and_warning)[5],
            (u_int8_t)bitset<8>(this->voltage_alarm_and_warning)[4],
            (u_int16_t)bitset<16>(this->RX_power_alarm_and_warning)[15],
            (u_int16_t)bitset<16>(this->RX_power_alarm_and_warning)[14],
            (u_int16_t)bitset<16>(this->RX_power_alarm_and_warning)[13],
            (u_int16_t)bitset<16>(this->RX_power_alarm_and_warning)[12],
            (u_int16_t)bitset<16>(this->RX_power_alarm_and_warning)[11],
            (u_int16_t)bitset<16>(this->RX_power_alarm_and_warning)[10],
            (u_int16_t)bitset<16>(this->RX_power_alarm_and_warning)[9],
            (u_int16_t)bitset<16>(this->RX_power_alarm_and_warning)[8],
            (u_int16_t)bitset<16>(this->RX_power_alarm_and_warning)[7],
            (u_int16_t)bitset<16>(this->RX_power_alarm_and_warning)[6],
            (u_int16_t)bitset<16>(this->RX_power_alarm_and_warning)[5],
            (u_int16_t)bitset<16>(this->RX_power_alarm_and_warning)[4],
            (u_int16_t)bitset<16>(this->RX_power_alarm_and_warning)[3],
            (u_int16_t)bitset<16>(this->RX_power_alarm_and_warning)[2],
            (u_int16_t)bitset<16>(this->RX_power_alarm_and_warning)[1],
            (u_int16_t)bitset<16>(this->RX_power_alarm_and_warning)[0],
            (u_int16_t)bitset<16>(this->TX_bias_alarm_and_warning)[15],
            (u_int16_t)bitset<16>(this->TX_bias_alarm_and_warning)[14],
            (u_int16_t)bitset<16>(this->TX_bias_alarm_and_warning)[13],
            (u_int16_t)bitset<16>(this->TX_bias_alarm_and_warning)[12],
            (u_int16_t)bitset<16>(this->TX_bias_alarm_and_warning)[11],
            (u_int16_t)bitset<16>(this->TX_bias_alarm_and_warning)[10],
            (u_int16_t)bitset<16>(this->TX_bias_alarm_and_warning)[9],
            (u_int16_t)bitset<16>(this->TX_bias_alarm_and_warning)[8],
            (u_int16_t)bitset<16>(this->TX_bias_alarm_and_warning)[7],
            (u_int16_t)bitset<16>(this->TX_bias_alarm_and_warning)[6],
            (u_int16_t)bitset<16>(this->TX_bias_alarm_and_warning)[5],
            (u_int16_t)bitset<16>(this->TX_bias_alarm_and_warning)[4],
            (u_int16_t)bitset<16>(this->TX_bias_alarm_and_warning)[3],
            (u_int16_t)bitset<16>(this->TX_bias_alarm_and_warning)[2],
            (u_int16_t)bitset<16>(this->TX_bias_alarm_and_warning)[1],
            (u_int16_t)bitset<16>(this->TX_bias_alarm_and_warning)[0],
            (u_int16_t)bitset<16>(this->TX_power_alarm_and_warning)[15],
            (u_int16_t)bitset<16>(this->TX_power_alarm_and_warning)[14],
            (u_int16_t)bitset<16>(this->TX_power_alarm_and_warning)[13],
            (u_int16_t)bitset<16>(this->TX_power_alarm_and_warning)[12],
            (u_int16_t)bitset<16>(this->TX_power_alarm_and_warning)[11],
            (u_int16_t)bitset<16>(this->TX_power_alarm_and_warning)[10],
            (u_int16_t)bitset<16>(this->TX_power_alarm_and_warning)[9],
            (u_int16_t)bitset<16>(this->TX_power_alarm_and_warning)[8],
            (u_int16_t)bitset<16>(this->TX_power_alarm_and_warning)[7],
            (u_int16_t)bitset<16>(this->TX_power_alarm_and_warning)[6],
            (u_int16_t)bitset<16>(this->TX_power_alarm_and_warning)[5],
            (u_int16_t)bitset<16>(this->TX_power_alarm_and_warning)[4],
            (u_int16_t)bitset<16>(this->TX_power_alarm_and_warning)[3],
            (u_int16_t)bitset<16>(this->TX_power_alarm_and_warning)[2],
            (u_int16_t)bitset<16>(this->TX_power_alarm_and_warning)[1],
            (u_int16_t)bitset<16>(this->TX_power_alarm_and_warning)[0],
            this->diag_supply_voltage,
            this->transmitter_technology >> 4,
            (u_int8_t)bitset<8>(this->transmitter_technology)[3],
            (u_int8_t)bitset<8>(this->transmitter_technology)[2],
            (u_int8_t)bitset<8>(this->transmitter_technology)[1],
            (u_int8_t)bitset<8>(this->transmitter_technology)[0],
            this->eth_com_codes_ext,
            this->ConvertDateCodeToStr().c_str(),
            this->lot,
            (u_int8_t)bitset<8>(this->TX_adaptive_equalization_freeze)[3],
            (u_int8_t)bitset<8>(this->TX_adaptive_equalization_freeze)[2],
            (u_int8_t)bitset<8>(this->TX_adaptive_equalization_freeze)[1],
            (u_int8_t)bitset<8>(this->TX_adaptive_equalization_freeze)[0],
            this->ConvertRXOutputDisableToStr().c_str(),
            this->ConvertTXAdaptiveEqualizationEnableToStr().c_str());

    IBDIAGNET_RETURN(buff);
}


string CableInfo::csv_str_db_1()
{
    IBDIAGNET_ENTER;
    char buff[1024];

    //NodeGuid,PortGuid,PortNum,
    sprintf(buff, U64H_FMT "," U64H_FMT "," U32D_FMT ","    \
            "\"" STR_FMT "\","   \
            U8H_FMT "," U8H_FMT "," U8H_FMT ","   \
            U16H_FMT "," U32H_FMT STR_FMT STR_FMT,
            this->p_port->p_node->guid_get(), this->p_port->guid_get(), this->p_port->num,
            this->mlnx_sfg_sn.c_str(),
            this->eth_com_codes_10g_40g,
            this->mlnx_revision,
            this->mlnx_vendor_byte,
            this->mellanox_cap,
            this->qsfp_options,
            this->ConvertPageE9Addr128ToStr().c_str(),
            this->ConvertPageE9Addr176ToStr().c_str());
    IBDIAGNET_RETURN(buff);
}

/****************************************************/
/***************** Class CableDiag ******************/
/****************************************************/
/*
 * private methods
 */

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int CableDiag::EyeOpenGetByLid(u_int16_t lid,
        u_int8_t port_num, u_int8_t group_lane,
        struct SMP_EyeOpen *p_smp_eye_open,
        const clbck_data_t *p_clbck_data)
{
    IBDIAGNET_ENTER;

    CLEAR_STRUCT(*p_smp_eye_open);

    /*
     * set attribute modifier with all values
     */
    struct SMP_MAD_AttrMod_EyeOpen attr_mod;
    CLEAR_STRUCT(attr_mod);
    attr_mod.PortSelect = port_num;
    attr_mod.GroupLane = group_lane;
    //the following code not working because pack is doing BE also
    //u_int32_t packed_attr_mod = 0;
    //SMP_MAD_AttrMod_EyeOpen_pack(&attr_mod, (u_int8_t *)&packed_attr_mod);
    u_int32_t attr_mod_field = attr_mod.GroupLane << 8 | attr_mod.PortSelect;

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMP_EYE_OPENER MAD by lid = %u port = %u\n",
             lid, (int)port_num);
    /*
     * send the mad by lid
     */
    int rc = this->p_ibis_obj->SMPMadGetSetByLid(lid,
            IBIS_IB_MAD_METHOD_GET,
            IB_ATTR_SMP_EYE_OPENER,
            attr_mod_field,         //packed_attr_mod,
            p_smp_eye_open,
            (const pack_data_func_t)SMP_EyeOpen_pack,
            (const unpack_data_func_t)SMP_EyeOpen_unpack,
            (const dump_data_func_t)SMP_EyeOpen_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int CableDiag::EyeOpenGetByDirect(direct_route_t *p_direct_route,
        u_int8_t port_num, u_int8_t group_lane,
        struct SMP_EyeOpen *p_smp_eye_open,
        const clbck_data_t *p_clbck_data)
{
    IBDIAGNET_ENTER;

    CLEAR_STRUCT(*p_smp_eye_open);

    /*
     * set attribute modifier with all values
     */
    struct SMP_MAD_AttrMod_EyeOpen attr_mod;
    CLEAR_STRUCT(attr_mod);
    attr_mod.PortSelect = port_num;
    attr_mod.GroupLane = group_lane;
    //the following code not working because pack is doing BE also
    //u_int32_t packed_attr_mod = 0;
    //SMP_MAD_AttrMod_EyeOpen_pack(&attr_mod, (u_int8_t *)&packed_attr_mod);
    u_int32_t attr_mod_field = attr_mod.GroupLane << 8 | attr_mod.PortSelect;

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMP_EYE_OPENER MAD by direct = %s port = %u\n",
             this->p_ibis_obj->ConvertDirPathToStr(p_direct_route).c_str(),
             (int)port_num);
    /*
     * send the mad by direct route
     */
    int rc = this->p_ibis_obj->SMPMadGetSetByDirect(p_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            IB_ATTR_SMP_EYE_OPENER,
            attr_mod_field,         //packed_attr_mod,
            p_smp_eye_open,
            (const pack_data_func_t)SMP_EyeOpen_pack,
            (const unpack_data_func_t)SMP_EyeOpen_unpack,
            (const dump_data_func_t)SMP_EyeOpen_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int CableDiag::CableInfoGetByLid(IN u_int16_t lid,
        IN u_int8_t port_num,
        IN u_int8_t addr, IN u_int8_t size, IN u_int8_t page_num,
        OUT struct SMP_CableInfo *p_smp_cable_info,
        OUT u_int8_t& vs_status,
        const clbck_data_t *p_clbck_data)
{
    IBDIAGNET_ENTER;

    CLEAR_STRUCT(*p_smp_cable_info);
    vs_status = 0;

    p_smp_cable_info->address = addr;
    p_smp_cable_info->size = size;
    p_smp_cable_info->page_number = page_num;
    p_smp_cable_info->device_address = EEPROM_CABLE_DEVICE_ADDR;

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMP_CABLE_INFO MAD by lid = %u port = %u\n",
             lid, (int)port_num);
    /*
     * send the mad by lid
     */
    int rc = this->p_ibis_obj->SMPMadGetSetByLid(lid,
            IBIS_IB_MAD_METHOD_GET,
            IB_ATTR_SMP_CABLE_INFO,
            port_num,
            p_smp_cable_info,
            (const pack_data_func_t)SMP_CableInfo_pack,
            (const unpack_data_func_t)SMP_CableInfo_unpack,
            (const dump_data_func_t)SMP_CableInfo_dump,
            p_clbck_data);
    vs_status = (u_int8_t)((rc & 0x7f00) >> 8); //take the class specific status
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int CableDiag::CableInfoGetByDirect(IN direct_route_t *p_direct_route,
        IN u_int8_t port_num,
        IN u_int8_t addr, IN u_int8_t size, IN u_int8_t page_num,
        IN u_int32_t password,
        OUT struct SMP_CableInfo *p_smp_cable_info,
        OUT u_int8_t& vs_status,
        const clbck_data_t *p_clbck_data)
{
    IBDIAGNET_ENTER;

    CLEAR_STRUCT(*p_smp_cable_info);
    vs_status = 0;

    p_smp_cable_info->address = addr;
    p_smp_cable_info->size = size;
    p_smp_cable_info->page_number = page_num;
    p_smp_cable_info->device_address = EEPROM_CABLE_DEVICE_ADDR;
    if (password) {
        p_smp_cable_info->password = password;
        p_smp_cable_info->pw_v = 1;
    }

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMP_CABLE_INFO MAD by direct = %s port = %u\n",
             this->p_ibis_obj->ConvertDirPathToStr(p_direct_route).c_str(),
             (int)port_num);
    /*
     * send the mad by direct route
     */
    int rc = this->p_ibis_obj->SMPMadGetSetByDirect(p_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            IB_ATTR_SMP_CABLE_INFO,
            port_num,
            p_smp_cable_info,
            (const pack_data_func_t)SMP_CableInfo_pack,
            (const unpack_data_func_t)SMP_CableInfo_unpack,
            (const dump_data_func_t)SMP_CableInfo_dump,
            p_clbck_data);
    vs_status = (u_int8_t)((rc & 0x7f00) >> 8); //take the class specific status
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


//Returns: IBDIAGNET_SUCCESS_CODE / IBDIAGNET_ERR_CODE_DB_ERR
int CableDiag::MarkAllPortsNotVisited(OUT u_int32_t& max_ports_per_node)
{
    IBDIAGNET_ENTER;

    // mark all ports as not visited
    max_ports_per_node = 0;
    for (map_str_pnode::iterator nI = this->p_discovered_fabric->NodeByName.begin();
            nI != this->p_discovered_fabric->NodeByName.end(); ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s",
                               (*nI).first.c_str());
                               IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        p_curr_node->appData1.val = NOT_SUPPORT_CHECK; //node eye open capability
        p_curr_node->appData3.val = NOT_SUPPORT_CHECK; //node cable info capability
        if (p_curr_node->numPorts > max_ports_per_node)
            max_ports_per_node = p_curr_node->numPorts;

        // go over all ports of this node, not query manage port, start from port 1
        // use unsigned int index to prevent overwrap
        for (unsigned int port_num = 1; port_num <= p_curr_node->numPorts; ++port_num) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)port_num);
            if (!p_curr_port)                       //port not exist
                continue;
            p_curr_port->counter1 = NOT_VISITED;
            p_curr_port->counter2 = 0; //clear not support flag
        }
    }
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}


//Returns: IBDIAGNET_SUCCESS_CODE / IBDIAGNET_ERR_CODE_NO_MEM / IBDIAGNET_ERR_CODE_DB_ERR
int CableDiag::AddSmpEyeOpen(IBPort *p_port, IBPort *p_rem_port,
        SMP_EyeOpen& smp_eye_open_curr_port, u_int8_t group_lane)
{
    IBDIAGNET_ENTER;

    //guarantee a unique map between port and idx in array
    unsigned port_vec_idx = MAX(p_port->createIndex, p_rem_port->createIndex);
    phys_port_t port_num        = (phys_port_t)((p_port->createIndex == port_vec_idx) ? 1 : 0);
    phys_port_t rem_port_num    = (phys_port_t)((p_rem_port->createIndex == port_vec_idx) ? 1 : 0);

    if ( (this->cables_vector.empty()) || (this->cables_vector.size() <= port_vec_idx + 1) )
        for (size_t i = this->cables_vector.size(); i < port_vec_idx + 1; ++i)
            this->cables_vector.push_back(NULL);
    if (this->cables_vector[p_port->createIndex] != this->cables_vector[p_rem_port->createIndex]) {
        this->SetLastError("DB error - found ports with different cable data, %s and %s",
                p_port->getName().c_str(), p_rem_port->getName().c_str());
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    cable_data_t *p_curr_cable_data = this->cables_vector[port_vec_idx];
    if (p_curr_cable_data == NULL) {   //if not exist
        p_curr_cable_data = new cable_data_t;
        if (!p_curr_cable_data) {
            this->SetLastError("Failed to allocate cable_data_t");
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_NO_MEM);
        }
        CLEAR_STRUCT(*p_curr_cable_data);
        this->cables_vector[p_port->createIndex] =
                this->cables_vector[p_rem_port->createIndex]    = p_curr_cable_data;
        p_curr_cable_data->data_per_port[port_num].p_port       = p_port;
        p_curr_cable_data->data_per_port[rem_port_num].p_port   = p_rem_port;
    }

    //finally add the obj
    SMP_EyeOpen *p_curr_smp_eye_open = new SMP_EyeOpen;
    if (!p_curr_smp_eye_open) {
        this->SetLastError("Failed to allocate SMP_EyeOpen");
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_NO_MEM);
    }
    *p_curr_smp_eye_open = smp_eye_open_curr_port;
    p_curr_cable_data->data_per_port[port_num].eye_open[group_lane] = p_curr_smp_eye_open;

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: IBDIAGNET_SUCCESS_CODE / IBDIAGNET_ERR_CODE_NO_MEM / IBDIAGNET_ERR_CODE_DB_ERR
int CableDiag::GetSMPCableInfo(IBPort *p_port, IBPort *p_rem_port,
                               CableInfo *& p_cable_info)
{
    IBDIAGNET_ENTER;

    p_cable_info = NULL;

    //guarantee a unique map between port and idx in array
    u_int32_t rem_port_create_index = (p_rem_port == NULL ? 0 : p_rem_port->createIndex);
    unsigned port_vec_idx = MAX(p_port->createIndex, rem_port_create_index);
    phys_port_t port_num        = (phys_port_t)((p_port->createIndex == port_vec_idx) ? 1 : 0);
    phys_port_t rem_port_num    = (phys_port_t)((rem_port_create_index == port_vec_idx) ? 1 : 0);

    if ( (this->cables_vector.empty()) || (this->cables_vector.size() <= port_vec_idx + 1) )
        for (size_t i = this->cables_vector.size(); i < port_vec_idx + 1; ++i)
            this->cables_vector.push_back(NULL);
    if (p_rem_port && (this->cables_vector[p_port->createIndex] != this->cables_vector[rem_port_create_index])) {
        this->SetLastError("DB error - found ports with different cable data, %s and %s",
                p_port->getName().c_str(),p_rem_port->getName().c_str() );
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    cable_data_t *p_curr_cable_data = this->cables_vector[port_vec_idx];
    if (p_curr_cable_data == NULL) {   //if not exist
        p_curr_cable_data = new cable_data_t;
        if (!p_curr_cable_data) {
            this->SetLastError("Failed to allocate cable_data_t");
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_NO_MEM);
        }
        CLEAR_STRUCT(*p_curr_cable_data);
        if (p_rem_port) {
            this->cables_vector[p_port->createIndex] =
                    this->cables_vector[rem_port_create_index]      = p_curr_cable_data;
            p_curr_cable_data->data_per_port[port_num].p_port       = p_port;
            p_curr_cable_data->data_per_port[rem_port_num].p_port   = p_rem_port;
        } else {
            this->cables_vector[p_port->createIndex] = p_curr_cable_data;
            p_curr_cable_data->data_per_port[port_num].p_port       = p_port;
            p_curr_cable_data->data_per_port[0].p_port              = NULL;
        }
    }

    p_cable_info = p_curr_cable_data->data_per_port[port_num].cable_info;

    if (p_cable_info == NULL) {
        p_cable_info = new CableInfo(p_port);
        if (!p_cable_info) {
            this->SetLastError("Failed to allocate CableInfo");
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_NO_MEM);
        }
        p_curr_cable_data->data_per_port[port_num].cable_info = p_cable_info;
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}


void CableDiag::GetDirectRoute(IBNode *p_node,
                               IBPort *p_port,
                               direct_route_t **p_direct_route)
{
    IBDIAGNET_ENTER;

    *p_direct_route = NULL;

    //Get direct route for this node by port guid
    //We need to query HCAs by direct route to correct port.
    //For switches, it's the same direct route
    //by port guid or by node guid
    *p_direct_route = this->p_ibdiag->
        GetDirectRouteByPortGuid(p_port->guid_get());

    //maybe disconnected HCA port, and then we don't have direct route by port
    //try to get by node guid, some HCAs with old fw (ConnectX3) still
    //support query by such direct route
    if (!*p_direct_route && p_node->type != IB_SW_NODE)
        *p_direct_route = this->p_ibdiag->GetDirectRouteByNodeGuid(p_node->guid_get());

    IBDIAGNET_RETURN_VOID;
}


//Returns: IBDIAGNET_SUCCESS_CODE / IBDIAGNET_ERR_CODE_DB_ERR / IBDIAGNET_ERR_CODE_NO_MEM / IBDIAGNET_ERR_CODE_FABRIC_ERROR
int CableDiag::BuildEyeOpenDB(list_p_fabric_general_err &cable_errors, progress_func_ports_t progress_func)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    progress_bar_ports_t progress_bar_ports;
    CLEAR_STRUCT(progress_bar_ports);

    // mark all ports as not visited
    u_int32_t max_ports_per_node;
    int rc2 = this->MarkAllPortsNotVisited(max_ports_per_node);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    m_ClbckErrorState = IBDIAG_SUCCESS_CODE;
    p_cable_errors = &cable_errors;
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = this;

    // go over all ports
    struct SMP_EyeOpen curr_eye_open;
    u_int8_t max_group_lane_range = 1;
    for (u_int8_t i = 0; i < max_group_lane_range; ++i) {
        for (unsigned int curr_port_idx = 1; curr_port_idx <= max_ports_per_node; ++curr_port_idx) {
            for (map_str_pnode::iterator nI = this->p_discovered_fabric->NodeByName.begin();
                    nI != this->p_discovered_fabric->NodeByName.end(); ++nI) {
                IBNode *p_curr_node = (*nI).second;
                if (!p_curr_node) {
                    this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());
                    rc = IBDIAG_ERR_CODE_DB_ERR;
                    goto exit;
                }

                //skip if this node has less ports
                if (p_curr_node->numPorts < curr_port_idx)
                    continue;

                //deals with progress bar
                ++progress_bar_ports.ports_found;
                if (progress_func)
                    progress_func(&progress_bar_ports, this->p_ibdiag->GetDiscoverProgressBarPortsPtr());

                //skip if this node already marked as not supported
                if (p_curr_node->appData1.val == NOT_SUPPORT_CAPABILTY)
                    continue;
                //check if this node supports eye open capability
                if (p_curr_node->appData1.val == NOT_SUPPORT_CHECK) {
                    if (!this->p_capability_module->IsSupportedSMPCapability(
                        p_curr_node, EnSMPCapIsSMPEyeOpenSupported)) {
                        p_curr_node->appData1.val = NOT_SUPPORT_CAPABILTY;
                        FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                                new FabricErrNodeNotSupportCap(p_curr_node,
                                        "This device does not support eye open capability");
                        if (!p_curr_fabric_node_err) {
                            this->SetLastError("Failed to allocate FabricErrNodeNotSupportCap");
                            rc = IBDIAG_ERR_CODE_NO_MEM;
                            goto exit;
                        }
                        cable_errors.push_back(p_curr_fabric_node_err);
                        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
                        continue;
                    }
                    p_curr_node->appData1.val = SUPPORT_CAPABILTY;
                }

                IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)curr_port_idx);
                if (!p_curr_port)                       //port not exist
                    continue;
                //Skip ports that are down or have undefined state
                //or are not in the scope of the fabric.
                if (!p_curr_port->is_data_worthy())
                    continue;
                //skip if port not QDR
                if (p_curr_port->get_common_speed() != IB_LINK_SPEED_10)
                    continue;

                direct_route_t *p_curr_direct_route = NULL;
                this->GetDirectRoute(p_curr_node, p_curr_port,
                                     &p_curr_direct_route);
                if (!p_curr_direct_route) {
                    this->SetLastError("DB error - can't find direct "
                                       "route to node=%s (port guid: 0x%x)",
                                       p_curr_node->getName().c_str(),
                                       p_curr_port->guid_get());
                    rc = IBDIAG_ERR_CODE_DB_ERR;
                    goto exit;
                }

                //determine group lanes range
                u_int8_t group_lane_range = 0;
                switch (p_curr_port->get_common_width()) {
                case IB_LINK_WIDTH_1X: {
                    group_lane_range = 1;
                    break;
                }
                case IB_LINK_WIDTH_2X: {
                    group_lane_range = 1;
                    break;
                }
                case IB_LINK_WIDTH_4X: {
                    group_lane_range = 1;
                    break;
                }
                case IB_LINK_WIDTH_8X: {
                    group_lane_range = 2;
                    break;
                }
                case IB_LINK_WIDTH_12X: {
                    group_lane_range = 3;
                    break;
                }
                default: {
                    char buff[1024];
                    sprintf(buff, "Unknown active width=\"%s\"",
                            width2char(p_curr_port->get_common_width()));
                    FabricErrPortWrongConfig *p_curr_fabric_port_err =
                            new FabricErrPortWrongConfig(p_curr_port, buff);
                    if (!p_curr_fabric_port_err) {
                        this->SetLastError("Failed to allocate FabricErrPortWrongConfig");
                        rc = IBDIAG_ERR_CODE_NO_MEM;
                        goto exit;
                    }
                    cable_errors.push_back(p_curr_fabric_port_err);
                    rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
                    continue;   //next node
                }}

                if (i == 0) {
                    if (group_lane_range > max_group_lane_range)
                        max_group_lane_range = group_lane_range;
                } else if (i >= group_lane_range) {
                    continue;
                }

                //go over all group lanes
                //for (u_int8_t i = 0; i < group_lane_range; ++i) {
                clbck_data.m_handle_data_func = EyeOpenGetDelegator;
                clbck_data.m_data1 = p_curr_port;
                clbck_data.m_data2 = (void *)(uintptr_t)i;
                this->EyeOpenGetByDirect(p_curr_direct_route,
                        p_curr_port->num, i, &curr_eye_open,
                        &clbck_data);

                if (m_ClbckErrorState)
                    goto exit;

            }   //loop of nodes
        }   //loop of max ports
    }   //group lane loop [0,MAX_GROUP_LANE-1]

exit:
    this->p_ibis_obj->MadRecAll();

    if (m_ClbckErrorState)
        rc = m_ClbckErrorState;
    else if (!cable_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAGNET_RETURN(rc);
}


//Returns: IBDIAGNET_SUCCESS_CODE / IBDIAGNET_ERR_CODE_CHECK_FAILED / IBDIAGNET_ERR_CODE_NO_MEM
int CableDiag::CheckEyeBoundSum(list_p_fabric_general_err &cable_errors)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    //mark all cables as not visited
    for (vec_p_cable_data::iterator it = this->cables_vector.begin();
            it != this->cables_vector.end(); ++it) {
        if ((*it))
            (*it)->app_data = NOT_VISITED;
    }

    //go over all cables
    for (vec_p_cable_data::iterator it = this->cables_vector.begin();
            it != this->cables_vector.end(); ++it) {
        cable_data_t *p_curr_cable_data = (*it);

        //skip if cable info not exist
        if (!p_curr_cable_data)
            continue;
        //skip if visited by rem port already
        if (p_curr_cable_data->app_data == VISITED)
            continue;
        //mark as visited
        p_curr_cable_data->app_data = VISITED;

        //go over both ports of this cable
        for (int curr_port_num = 0; curr_port_num < 2; ++curr_port_num) {
            //go over all smp_eye_open info we have for this port
            for (int  curr_group_lane = 0; curr_group_lane < MAX_GROUP_LANES; ++curr_group_lane) {
                SMP_EyeOpen *curr_eye_open =
                        p_curr_cable_data->data_per_port[curr_port_num].eye_open[curr_group_lane];
                if (!curr_eye_open)         //if info not exists
                    continue;

                //go over all lanes
                u_int8_t bound_negval = 0, bound_posval = 0;
                u_int16_t eye_bound_sum = 0;

                for (int curr_lane = 0; curr_lane < 4; ++curr_lane) {
                    if (curr_lane == 0) {
                        bound_negval = curr_eye_open->NegativeBound_Lane0;
                        bound_posval = curr_eye_open->PositiveBound_Lane0;
                    } else if (curr_lane == 1) {
                        bound_negval = curr_eye_open->NegativeBound_Lane1;
                        bound_posval = curr_eye_open->PositiveBound_Lane1;
                    } else if (curr_lane == 2) {
                        bound_negval = curr_eye_open->NegativeBound_Lane2;
                        bound_posval = curr_eye_open->PositiveBound_Lane2;
                    } else {            //if (curr_lane == 3) {
                        bound_negval = curr_eye_open->NegativeBound_Lane3;
                        bound_posval = curr_eye_open->PositiveBound_Lane3;
                    }
                    bound_negval = (u_int8_t)(256 - bound_negval);
                    eye_bound_sum = (u_int16_t)(bound_negval + bound_posval);

                    if (this->eye_open_bound_min_thresh > eye_bound_sum) {
                        FabricErrEyeBoundBelowThresh *p_curr_fabric_port_err =
                                new FabricErrEyeBoundBelowThresh(p_curr_cable_data->data_per_port[curr_port_num].p_port,
                                        curr_group_lane + 1 + curr_lane,
                                        bound_negval, bound_posval,
                                        this->eye_open_bound_min_thresh);
                        if (!p_curr_fabric_port_err) {
                            this->SetLastError("Failed to allocate FabricErrEyeBoundBelowThresh");
                            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_NO_MEM);
                        }
                        cable_errors.push_back(p_curr_fabric_port_err);
                        rc = IBDIAG_ERR_CODE_CHECK_FAILED;
                    }
                    if (this->eye_open_bound_max_thresh < eye_bound_sum) {
                        FabricErrEyeBoundAboveThresh *p_curr_fabric_port_err =
                                new FabricErrEyeBoundAboveThresh(p_curr_cable_data->data_per_port[curr_port_num].p_port,
                                        curr_group_lane + 1 + curr_lane,
                                        bound_negval, bound_posval,
                                        this->eye_open_bound_max_thresh);
                        if (!p_curr_fabric_port_err) {
                            this->SetLastError("Failed to allocate FabricErrEyeBoundAboveThresh");
                            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_NO_MEM);
                        }
                        cable_errors.push_back(p_curr_fabric_port_err);
                        rc = IBDIAG_ERR_CODE_CHECK_FAILED;
                    }

                    if (p_curr_cable_data->data_per_port[curr_port_num].p_port->get_common_width() == IB_LINK_WIDTH_1X)
                        break;          //if only one lane than not go over all lanes
                }   //lanes loop [0,3]
            }   //group lane loop [0,MAX_GROUP_LANE-1]
        }   //cable ports loop [0,1]
    }   //cables vec loop

    IBDIAGNET_RETURN(rc);
}

//Returns: IBDIAGNET_SUCCESS_CODE / IBDIAGNET_ERR_CODE_DB_ERR / IBDIAGNET_ERR_CODE_NO_MEM / IBDIAGNET_ERR_CODE_FABRIC_ERROR
int CableDiag::BuildCableInfoDB(list_p_fabric_general_err &cable_errors,
                                progress_func_ports_add_msg_t progress_func,
                                u_int8_t phase, u_int32_t max_ports_per_node)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    progress_bar_ports_t progress_bar_ports;
    CLEAR_STRUCT(progress_bar_ports);

    m_ClbckErrorState = IBDIAG_SUCCESS_CODE;
    p_cable_errors = &cable_errors;
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = this;

    // go over all ports
    struct SMP_CableInfo cable_info_arr;

    for (list_p_cable_info_addr_t::iterator it = this->cable_info_addr_list_by_phase[phase].begin();
         it != this->cable_info_addr_list_by_phase[phase].end(); ++it) {
        cable_info_address_t * p_cable_info_address = *it;
        progress_bar_ports.ports_found = 0;
        SCREEN_PRINT("\n");
        for (unsigned int curr_port_idx = 1; curr_port_idx <= max_ports_per_node; ++curr_port_idx) {
            for (map_str_pnode::iterator nI = this->p_discovered_fabric->NodeByName.begin();
                    nI != this->p_discovered_fabric->NodeByName.end(); ++nI) {
                IBNode *p_curr_node = (*nI).second;
                if (!p_curr_node) {
                    this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());
                    rc = IBDIAG_ERR_CODE_DB_ERR;
                    goto exit;
                }

                //skip if this node has less ports
                if (p_curr_node->numPorts < curr_port_idx)
                    continue;

                IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)curr_port_idx);
                if (!p_curr_port)                       //port not exist
                    continue;

                if (p_ibdiag->GetSpecialCAPortType(p_curr_node) != NotSpecial || // CA node is special
                    p_ibdiag->GetSpecialPortType(p_curr_port) != NotSpecial)   // SW port is special
                    continue;

                //skip if this node already marked as not supported
                if (p_curr_node->appData1.val == NOT_SUPPORT_CAPABILTY)
                    continue;
                //check if this node supports cable info capability
                if (p_curr_node->appData1.val == NOT_SUPPORT_CHECK) {
                    if (!this->p_capability_module->IsSupportedSMPCapability(
                        p_curr_node, EnSMPCapIsCableInfoSupported)) {
                        p_curr_node->appData1.val = NOT_SUPPORT_CAPABILTY;
                        FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                                new FabricErrNodeNotSupportCap(p_curr_node,
                                        "This device does not support cable info capability");
                        if (!p_curr_fabric_node_err) {
                            this->SetLastError("Failed to allocate FabricErrNodeNotSupportCap");
                            rc = IBDIAG_ERR_CODE_DB_ERR;
                            goto exit;
                        }
                        cable_errors.push_back(p_curr_fabric_node_err);
                        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
                        continue;
                    }
                    p_curr_node->appData1.val = SUPPORT_CAPABILTY;
                }

                //Skip ports that are not in sub fabric.
                if (!p_curr_port->getInSubFabric())
                    continue;

                //Skip down ports if 'cable_info_disconnected' flag is off.
                if (p_curr_port->get_internal_state() <= IB_PORT_STATE_DOWN &&
                        !this->to_get_disconnected_port_info)
                    continue;

                //deals with progress bar
                char msg[128] = {0};
                sprintf(msg, "Page: %d, Address: %d", p_cable_info_address->page_num,
                        p_cable_info_address->address);
                ++progress_bar_ports.ports_found;
                if (progress_func)
                    progress_func(&progress_bar_ports,
                                  this->p_ibdiag->GetDiscoverProgressBarPortsPtr(), msg);

                CableInfo *p_cable_info = NULL;
                m_ClbckErrorState = this->GetSMPCableInfo(
                    p_curr_port, p_curr_port->p_remotePort, p_cable_info);
                if (m_ClbckErrorState)
                    goto exit;

                if (p_cable_info_address->func_get_condition &&
                    (!(p_cable_info_address->func_get_condition(p_cable_info))))
                        continue;

                if (p_cable_info_address->password) {
                    if (p_curr_node->appData3.val == NOT_SUPPORT_CAPABILTY)
                        continue;
                    //check if this node supports cable info capability
                    if (p_curr_node->appData3.val == NOT_SUPPORT_CHECK) {
                        if (!this->p_capability_module->IsSupportedSMPCapability(
                                    p_curr_node, EnSMPCapIsCableInfoPasswordSupported)) {
                            p_curr_node->appData3.val = NOT_SUPPORT_CAPABILTY;
                            continue;
                        }
                        p_curr_node->appData3.val = SUPPORT_CAPABILTY;
                    }
                }

                direct_route_t *p_curr_direct_route = NULL;
                this->GetDirectRoute(p_curr_node, p_curr_port,
                                     &p_curr_direct_route);
                if (!p_curr_direct_route) {
                    this->SetLastError("DB error - can't find direct route "
                                       "to node=%s (port guid: 0x%x)",
                                       p_curr_node->getName().c_str(),
                                       p_curr_port->guid_get());
                    rc = IBDIAG_ERR_CODE_DB_ERR;
                    goto exit;
                }


                clbck_data.m_handle_data_func = CableInfoGetDelegator;
                clbck_data.m_data1 = p_curr_port;
                clbck_data.m_data2 = (void*)(intptr_t)p_cable_info_address->address;
                clbck_data.m_data3 = (void*)(intptr_t)p_cable_info_address->page_num;

                u_int8_t vs_status;
                this->CableInfoGetByDirect(p_curr_direct_route,
                            p_curr_port->num,
                            p_cable_info_address->address,
                            p_cable_info_address->addr_size,
                            p_cable_info_address->page_num,
                            p_cable_info_address->password,
                            &cable_info_arr,
                            vs_status, &clbck_data);

                if (m_ClbckErrorState)
                    goto exit;

            }   //loop of nodes
        }   //loop of max ports
    } //loop on address
exit:
    this->p_ibis_obj->MadRecAll();

    if (m_ClbckErrorState)
        rc = m_ClbckErrorState;
    else if (!cable_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAGNET_RETURN(rc);
}


void CableDiag::DumpCSVCablesInfo(CSVOut &csv_out)
{
    IBDIAGNET_ENTER;
    stringstream sstream;

    //mark all cables as not visited
    for (vec_p_cable_data::iterator it = this->cables_vector.begin();
            it != this->cables_vector.end(); ++it) {
        if ((*it))
            (*it)->app_data = NOT_VISITED;
    }
    csv_out.DumpStart(SECTION_CABLE_INFO);

    sstream << CableInfo::hdr_str() << endl;
    csv_out.WriteLine(sstream.str());

    //go over all cables
    for (vec_p_cable_data::iterator it = this->cables_vector.begin();
            it != this->cables_vector.end(); ++it) {
        cable_data_t *p_curr_cable_data = (*it);

        //skip if cable info not exist
        if (!p_curr_cable_data)
            continue;
        //skip if visited by rem port already
        if (p_curr_cable_data->app_data == VISITED)
            continue;
        //mark as visited
        p_curr_cable_data->app_data = VISITED;

        //go over both ports of this cable
        for (int curr_port_num = 0; curr_port_num < 2; ++curr_port_num) {
            CableInfo *p_curr_cable_info =
                    p_curr_cable_data->data_per_port[curr_port_num].cable_info;
            //skip if not exist
            if (!p_curr_cable_info)
                continue;

            sstream.str("");
            sstream << p_curr_cable_info->csv_str() << endl;
            csv_out.WriteLine(sstream.str());
        }   //cable ports loop [0,1]
    }   //cables vec loop

    csv_out.DumpEnd(SECTION_CABLE_INFO);

    // SECTION_CABLE_INFO_DB_1
    for (vec_p_cable_data::iterator it = this->cables_vector.begin();
            it != this->cables_vector.end(); ++it) {
        if ((*it))
            (*it)->app_data = NOT_VISITED;
    }

    csv_out.DumpStart(SECTION_CABLE_INFO_DB_1);

    sstream.str("");
    sstream << "NodeGuid,PortGuid,PortNum";
    for (u_int32_t cnt = 0; cnt < SECTION_CABLE_INFO_DB_1_FIELDS_NUM; cnt++)
        sstream << ",field" <<  cnt;
    sstream << endl;
    csv_out.WriteLine(sstream.str());

    //go over all cables
    for (vec_p_cable_data::iterator it = this->cables_vector.begin();
            it != this->cables_vector.end(); ++it) {
        cable_data_t *p_curr_cable_data = (*it);

        //skip if cable info not exist
        if (!p_curr_cable_data)
            continue;
        //skip if visited by rem port already
        if (p_curr_cable_data->app_data == VISITED)
            continue;
        //mark as visited
        p_curr_cable_data->app_data = VISITED;

        //go over both ports of this cable
        for (int curr_port_num = 0; curr_port_num < 2; ++curr_port_num) {
            CableInfo *p_curr_cable_info =
                    p_curr_cable_data->data_per_port[curr_port_num].cable_info;
            //skip if not exist
            if (!p_curr_cable_info)
                continue;

            sstream.str("");
            sstream << p_curr_cable_info->csv_str_db_1() << endl;
            csv_out.WriteLine(sstream.str());
        }   //cable ports loop [0,1]
    }   //cables vec loop

    csv_out.DumpEnd(SECTION_CABLE_INFO_DB_1);
    IBDIAGNET_RETURN_VOID;
}


void CableDiag::DumpCSVEyeOpenInfo(CSVOut &csv_out)
{
    IBDIAGNET_ENTER;
    char buff[1024];
    stringstream sstream;

    //mark all cables as not visited
    for (vec_p_cable_data::iterator it = this->cables_vector.begin();
            it != this->cables_vector.end(); ++it) {
        if ((*it))
            (*it)->app_data = NOT_VISITED;
    }

    csv_out.DumpStart(SECTION_EYE_OPEN_INFO);

    sstream << "NodeGuid,PortGuid,PortNum,LaneNum,NegativeBound,PositiveBound" << endl;
    csv_out.WriteLine(sstream.str());

    //go over all cables
    for (vec_p_cable_data::iterator it = this->cables_vector.begin();
            it != this->cables_vector.end(); ++it) {
        cable_data_t *p_curr_cable_data = (*it);

        //skip if cable info not exist
        if (!p_curr_cable_data)
            continue;
        //skip if visited by rem port already
        if (p_curr_cable_data->app_data == VISITED)
            continue;
        //mark as visited
        p_curr_cable_data->app_data = VISITED;

        //go over both ports of this cable
        for (int curr_port_num = 0; curr_port_num < 2; ++curr_port_num) {
            //go over all smp_eye_open info we have for this port
            for (int  curr_group_lane = 0; curr_group_lane < MAX_GROUP_LANES; ++curr_group_lane) {
                SMP_EyeOpen *curr_eye_open =
                        p_curr_cable_data->data_per_port[curr_port_num].eye_open[curr_group_lane];
                if (!curr_eye_open)         //if info not exists
                    continue;

                //go over all lanes
                u_int8_t bound_negval = 0, bound_posval = 0;
                for (int curr_lane = 0; curr_lane < 4; ++curr_lane) {
                    if (curr_lane == 0) {
                        bound_negval = curr_eye_open->NegativeBound_Lane0;
                        bound_posval = curr_eye_open->PositiveBound_Lane0;
                    } else if (curr_lane == 1) {
                        bound_negval = curr_eye_open->NegativeBound_Lane1;
                        bound_posval = curr_eye_open->PositiveBound_Lane1;
                    } else if (curr_lane == 2) {
                        bound_negval = curr_eye_open->NegativeBound_Lane2;
                        bound_posval = curr_eye_open->PositiveBound_Lane2;
                    } else {            //if (curr_lane == 3) {
                        bound_negval = curr_eye_open->NegativeBound_Lane3;
                        bound_posval = curr_eye_open->PositiveBound_Lane3;
                    }
                    bound_negval = (u_int8_t)(256 - bound_negval);

                    sstream.str("");
                    sprintf(buff, U64H_FMT "," U64H_FMT "," U32D_FMT ","    \
                            U32D_FMT "," U32D_FMT "," U32D_FMT,
                            p_curr_cable_data->data_per_port[curr_port_num].p_port->p_node->guid_get(),
                            p_curr_cable_data->data_per_port[curr_port_num].p_port->guid_get(),
                            p_curr_cable_data->data_per_port[curr_port_num].p_port->num,
                            curr_group_lane + 1 + curr_lane,
                            bound_negval,
                            bound_posval);
                    sstream << buff << endl;
                    csv_out.WriteLine(sstream.str());
                    if (p_curr_cable_data->data_per_port[curr_port_num].p_port->get_common_width() == IB_LINK_WIDTH_1X)
                        break;          //if only one lane than not go over all lanes
                }   //lanes loop [0,3]
            }   //group lane loop [0,MAX_GROUP_LANE-1]
        }   //cable ports loop [0,1]
    }   //cables vec loop

    csv_out.DumpEnd(SECTION_EYE_OPEN_INFO);
    IBDIAGNET_RETURN_VOID;
}


void CableDiag::DumpCablesInfo(ofstream &sout)
{
    IBDIAGNET_ENTER;

    //mark all cables as not visited
    for (vec_p_cable_data::iterator it = this->cables_vector.begin();
            it != this->cables_vector.end(); ++it) {
        if ((*it))
            (*it)->app_data = NOT_VISITED;
    }

    //go over all cables
    char buff[1024];
    for (vec_p_cable_data::iterator it = this->cables_vector.begin();
            it != this->cables_vector.end(); ++it) {
        cable_data_t *p_curr_cable_data = (*it);

        //skip if cable info not exist
        if (!p_curr_cable_data)
            continue;
        //skip if visited by rem port already
        if (p_curr_cable_data->app_data == VISITED)
            continue;
        //mark as visited
        p_curr_cable_data->app_data = VISITED;

        //go over both ports of this cable
        for (int curr_port_num = 0; curr_port_num < 2; ++curr_port_num) {
            CableInfo *p_curr_cable_info =
                    p_curr_cable_data->data_per_port[curr_port_num].cable_info;
            if (!p_curr_cable_info)         //skip if not exist for this port
                continue;

            IBPort * p_port = p_curr_cable_data->data_per_port[curr_port_num].p_port;
            if (p_port == NULL)
                continue; //disconnected
            memset(buff, 0, sizeof(buff));
            sprintf(buff, "Port=" U32D_FMT    \
                    " Lid=" U16H_FMT    \
                    " GUID=" U64H_FMT   \
                    " Port Name=" STR_FMT,
                    p_port->num,
                    p_port->base_lid,
                    p_port->guid_get(),
                    p_port->getName().c_str());
            sout << "-------------------------------------------------------" << endl;
            sout << buff << endl;
            sout << "-------------------------------------------------------" << endl;
            sout << p_curr_cable_info->c_str() << endl << endl;
        }   //cable ports loop [0,1]
    }   //cables vec loop

    IBDIAGNET_RETURN_VOID;
}


void CableDiag::DumpEyeOpenInfo(ofstream &sout)
{
    IBDIAGNET_ENTER;

    //mark all cables as not visited
    for (vec_p_cable_data::iterator it = this->cables_vector.begin();
            it != this->cables_vector.end(); ++it) {
        if ((*it))
            (*it)->app_data = NOT_VISITED;
    }

    //go over all cables
    char buff[1024];
    for (vec_p_cable_data::iterator it = this->cables_vector.begin();
            it != this->cables_vector.end(); ++it) {
        cable_data_t *p_curr_cable_data = (*it);

        //skip if cable info not exist
        if (!p_curr_cable_data)
            continue;
        //skip if visited by rem port already
        if (p_curr_cable_data->app_data == VISITED)
            continue;
        //mark as visited
        p_curr_cable_data->app_data = VISITED;

        //go over both ports of this cable
        for (int curr_port_num = 0; curr_port_num < 2; ++curr_port_num) {
            //go over all smp_eye_open info we have for this port
            for (int  curr_group_lane = 0; curr_group_lane < MAX_GROUP_LANES; ++curr_group_lane) {
                SMP_EyeOpen *curr_eye_open =
                        p_curr_cable_data->data_per_port[curr_port_num].eye_open[curr_group_lane];
                if (!curr_eye_open)         //skip if not exists
                    continue;

                //go over all lanes
                for (int curr_lane = 0; curr_lane < 4; ++curr_lane) {
                    //node guid, port guid, port num, lane num
                    sprintf(buff, U64H_FMT "," U64H_FMT "," U32D_FMT "," U32D_FMT,
                            p_curr_cable_data->data_per_port[curr_port_num].p_port->p_node->guid_get(),
                            p_curr_cable_data->data_per_port[curr_port_num].p_port->guid_get(),
                            p_curr_cable_data->data_per_port[curr_port_num].p_port->num,
                            curr_group_lane + 1 + curr_lane);
                    sout << buff << ",";

                    /*Low/High bounds are accroding to +32 (zir x)
                    Low bound range is [-32,0]
                    High bound range is [0,31]

                    Low/High Amp are according to 0 (zir y).
                    Low Amp range is [-128,0]
                    High Amp range is [0,127]

                    Calibration point values are [-128,127]

                    for -9 the value in the field in the mad is 256-9=246
                    */

                    u_int8_t amp_high, amp_low, bound_pos, bound_neg, selected_opt_tx, selected_opt_rx, amp_cal;
                    int amp_cal_with_sign;
                    #define EYEOPENIG_GET(lane); do {   \
                        amp_high        = curr_eye_open->HighAmplitude_Lane##lane;  \
                        amp_low         = curr_eye_open->LowAmplitude_Lane##lane;   \
                        bound_pos       = curr_eye_open->PositiveBound_Lane##lane;  \
                        bound_neg       = curr_eye_open->NegativeBound_Lane##lane;  \
                        selected_opt_tx = curr_eye_open->SelectedOptionTX_Lane##lane;   \
                        selected_opt_rx = curr_eye_open->SelectedOptionRX_Lane##lane;   \
                        amp_cal         = curr_eye_open->AmpCalibration_Lane##lane; \
                    } while(0);

                    if (curr_lane == 0) {
                        EYEOPENIG_GET(0);
                    } else if (curr_lane == 1) {
                        EYEOPENIG_GET(1);
                    } else if (curr_lane == 2) {
                        EYEOPENIG_GET(2);
                    } else {            //if (curr_lane == 3) {
                        EYEOPENIG_GET(3);
                    }
                    bound_neg = (u_int8_t)(256 - bound_neg);
                    amp_low = (u_int8_t)(256 - amp_low);
                    amp_cal_with_sign = (amp_cal > 127) ? ((256 - amp_cal) * -1) : amp_cal;

                    //link width, autoneg state, high_amp_lane0, low_amp_lane0, positive_bound_lane0, negative_bound_lane0,
                    //selected_option_tx_lane0, selected_option_rx_lane0, amp_calibration_lane0,
                    sprintf(buff, U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT "," U32D_FMT ","    \
                            U32D_FMT "," U32D_FMT ",%d",
                            curr_eye_open->LinkWidthActive,
                            curr_eye_open->AutoNegoState,
                            amp_high, amp_low, bound_pos, bound_neg,
                            selected_opt_tx, selected_opt_rx, amp_cal_with_sign);
                    sout << buff << endl;

                    if (p_curr_cable_data->data_per_port[curr_port_num].p_port->get_common_width() == IB_LINK_WIDTH_1X)
                        break;          //if only one lane than not go over all lanes
                }   //lanes loop [0,3]
            }   //group lane loop [0,MAX_GROUP_LANE-1]
        }   //cable ports loop [0,1]
    }   //cables vec loop

    IBDIAGNET_RETURN_VOID;
}


//Returns: SUCCESS_CODE / ERR_CODE_IO_ERR
int CableDiag::WriteCableFile(const char *file_name)
{
    IBDIAGNET_ENTER;

    ofstream sout;
    string err_message;
    int rc = construct_secure_file(file_name, sout, false, err_message);
    if (rc) {
        if (err_message.empty())
            this->SetLastError("Failed to open file = %s for writing", file_name);
        else
            this->SetLastError(err_message.c_str());

        IBDIAGNET_RETURN(rc);
    }

    this->DumpCablesInfo(sout);

    sout.close();
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}


//Returns: SUCCESS_CODE / ERR_CODE_IO_ERR
int CableDiag::WriteEyeExpertFile(const char *file_name)
{
    IBDIAGNET_ENTER;

    ofstream sout;
    string err_message;
    int rc = construct_secure_file(file_name, sout, false, err_message);
    if (rc) {
        if (err_message.empty())
            this->SetLastError("Failed to open file = %s for writing", file_name);
        else
            this->SetLastError(err_message.c_str());

        IBDIAGNET_RETURN(rc);
    }

    /*
    sout.open(file_name);
    if (sout.fail()) {
        this->SetLastError("Failed to open file = %s for writing", file_name);
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_IO_ERR);
    }
    */
    sout << "# This database file was automatically generated by " << this->name << endl;
    sout << endl << endl;
    this->DumpEyeOpenInfo(sout);

    sout.close();
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}


/****************************************************/
/*
 * public methods
 */

CableDiag::CableDiag(IBDiag *p_ibdiag) : Plugin(PLUGIN_NAME, p_ibdiag),
        eye_open_bound_min_thresh(EYE_MIN_DEFAULT_THRESH),
        eye_open_bound_max_thresh(EYE_MAX_DEFAULT_THRESH),
        to_check_eye_bounds(false), to_get_eye_open_info(false), to_dump_eye_expert_file(false),
        to_get_cable_info(false), to_get_disconnected_port_info(false),
        m_ClbckErrorState(IBDIAG_SUCCESS_CODE), p_cable_errors(NULL)
{
    IBDIAGNET_ENTER;
    /*
     * this stage is skipped by default
     */
    this->MarkStageAsSkipped();

    /*
     * add all options
     */
    this->AddOptions(OPTION_CABLE_INFO_GET, OPTION_CABLE_INFO_GET_SHORT,
                     OPTION_CABLE_INFO_GET_ARG_DESC, OPTION_CABLE_INFO_GET_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_CABLE_INFO_DICONNECTED, OPTION_CABLE_INFO_DICONNECTED_SHORT,
                     OPTION_CABLE_INFO_DICONNECTED_ARG_DESC, OPTION_CABLE_INFO_DICONNECTED_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_EYE_OPEN_INFO_GET, OPTION_EYE_OPEN_INFO_GET_SHORT,
                     OPTION_EYE_OPEN_INFO_GET_ARG_DESC, OPTION_EYE_OPEN_INFO_GET_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_HIDDEN | IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_EYE_OPEN_INFO_CHECK, OPTION_EYE_OPEN_INFO_CHECK_SHORT,
                     OPTION_EYE_OPEN_INFO_CHECK_ARG_DESC, OPTION_EYE_OPEN_INFO_CHECK_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_HIDDEN | IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_EYE_EXPERT, OPTION_EYE_EXPERT_SHORT, OPTION_EYE_EXPERT_ARG_DESC,
                     OPTION_EYE_EXPERT_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_HIDDEN | IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_EYE_MIN_THRESH, OPTION_EYE_MIN_THRESH_SHORT,
                     OPTION_EYE_MIN_THRESH_ARG_DESC, OPTION_EYE_MIN_THRESH_DESC,
                     OPTION_EYE_MIN_THRESH_DEF_VAL, IBIS_CMD_HIDDEN);
    this->AddOptions(OPTION_EYE_MAX_THRESH, OPTION_EYE_MAX_THRESH_SHORT,
                     OPTION_EYE_MAX_THRESH_ARG_DESC, OPTION_EYE_MAX_THRESH_DESC,
                     OPTION_EYE_MAX_THRESH_DEF_VAL, IBIS_CMD_HIDDEN);
    this->AddDescription(PLUGIN_DESC);

    this->p_discovered_fabric = this->p_ibdiag->GetDiscoverFabricPtr();
    this->p_ibis_obj = this->p_ibdiag->GetIbisPtr();
    this->p_capability_module = this->p_ibdiag->GetCapabilityModulePtr();
    IBDIAGNET_RETURN_VOID;
}


CableDiag::~CableDiag()
{
    IBDIAGNET_ENTER;

    for (vec_p_cable_data::iterator it = this->cables_vector.begin();
            it != this->cables_vector.end(); ++it) {
        cable_data_t *p_curr_cable_data = *it;
        if (p_curr_cable_data) {
            //delete cable data from both ports
            for (int i = 0; i < 2; ++i) {
                if (p_curr_cable_data->data_per_port[i].p_port) {
                    this->cables_vector[p_curr_cable_data->data_per_port[i].p_port->createIndex] = NULL;
                }
                //delete cable info
                if (p_curr_cable_data->data_per_port[i].cable_info)
                    delete p_curr_cable_data->data_per_port[i].cable_info;
                //delete eye open
                for (int j = 0; j < MAX_GROUP_LANES; ++j)
                    if (p_curr_cable_data->data_per_port[i].eye_open[j])
                        delete p_curr_cable_data->data_per_port[i].eye_open[j];
            }
            delete p_curr_cable_data;
        }
    }

    for (int i = 0; i < MAX_PHASES; i++) {
       for (list_p_cable_info_addr_t::iterator it = cable_info_addr_list_by_phase[i].begin();
            it != cable_info_addr_list_by_phase[i].end();
            it++)
           delete (*it);

       cable_info_addr_list_by_phase[i].clear();
    }

    IBDIAGNET_RETURN_VOID;
}


int CableDiag::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}


int CableDiag::RetrieveInfo()
{
    IBDIAGNET_ENTER;

    int rc = 0;
    list_p_fabric_general_err cable_errors;
    int rc2;

    if (this->to_get_eye_open_info) {
        rc2 = this->BuildEyeOpenDB(cable_errors, progress_bar_retrieve_ports);
        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(cable_errors,
                CHECK_NAME_EYE_OPEN_INFO_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
        if (rc)
            IBDIAGNET_RETURN(rc);

        //dump eye open to db file
        string file_str = generate_file_name(*this->p_base_path, DB_FILE);
        this->DumpCSVEyeOpenInfo(*this->p_csv_out);

        //dump eye expert file if needed - we will hide this file under port attributes
        if (this->to_dump_eye_expert_file == true) {
            file_str = generate_file_name(*this->p_base_path, EYE_EXPERT_FILE);
            if (this->WriteEyeExpertFile(file_str.c_str())) {
                ERR_PRINT("Writing port attributes %s failed\n", file_str.c_str());
                ++this->num_errors;
            }
            this->AddGeneratedFileName("Port Attributes file", file_str);
        }
    }

    if (this->to_get_cable_info) {

        this->CreatePagesList();

        u_int32_t max_ports_per_node;
        // mark all ports as not visited
        int rc2 = this->MarkAllPortsNotVisited(max_ports_per_node);
        if (rc2)
            IBDIAGNET_RETURN(rc2);

        int rc_all_phases = 0;
        for (int phase = 0; phase < MAX_PHASES; phase++) {
            rc2 = this->BuildCableInfoDB(cable_errors, progress_bar_retrieve_ports_add_msg, phase,
                                         max_ports_per_node);

            if (rc2)
                rc_all_phases = rc2;
        }

        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(cable_errors,
                CHECK_NAME_CABLE_INFO_RETRIEVE,
                rc_all_phases,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
        if (rc)
            IBDIAGNET_RETURN(rc);

        //dump cables to db file
        string file_str = generate_file_name(*this->p_base_path, DB_FILE);
        this->DumpCSVCablesInfo(*this->p_csv_out);

        //dump cable file
        file_str = generate_file_name(*this->p_base_path, CABLES_FILE);
        if (this->WriteCableFile(file_str.c_str())) {
            ERR_PRINT("Writing cables info to %s failed\n", file_str.c_str());
            ++this->num_errors;
        }
        this->AddGeneratedFileName("Cables Information file", file_str);
    }

    IBDIAGNET_RETURN(rc);
}


int CableDiag::RunCheck()
{
    IBDIAGNET_ENTER;

    int rc = 0;
    list_p_fabric_general_err cable_errors;

    //check eye open threshold
    if ((this->to_get_eye_open_info) && (this->to_check_eye_bounds)) {
        int rc2 = this->CheckEyeBoundSum(cable_errors);
        rc = this->AnalyzeCheckResults(cable_errors,
                CHECK_NAME_EYE_OPEN_BOUND,
                rc2,
                IBDIAG_ERR_CODE_CHECK_FAILED,
                this->num_errors,
                this->num_warnings);
    }

    IBDIAGNET_RETURN(rc);
}


//Returns: 0 - OK / 1 - parsing error / 2 - OK but need to exit / 3 - some other error
int CableDiag::HandleOption(string name, string value)
{
    IBDIAGNET_ENTER;

    bool bool_flag_val = true;

    if (value == OPTION_DEF_VAL_NULL)
        IBDIAGNET_RETURN(0);

    if (name == OPTION_EYE_MIN_THRESH) {
        if (!is_non_neg_num(value)) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                    OPTION_EYE_MIN_THRESH, value.c_str());
            PRINT("    Only non negative number is allowed\n");
            IBDIAGNET_RETURN(3);
        }
        this->SetEyeOpenBoundMinThresh((u_int16_t)strtoul(value.c_str(), NULL, 0));
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_EYE_MAX_THRESH) {
        if (!is_non_neg_num(value)) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                    OPTION_EYE_MAX_THRESH, value.c_str());
            PRINT("    Only non negative number is allowed\n");
            IBDIAGNET_RETURN(3);
        }
        this->SetEyeOpenBoundMaxThresh((u_int16_t)strtoul(value.c_str(), NULL, 0));
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_EYE_OPEN_INFO_GET) {
        this->SetEyeOpenInfoGet();
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_CABLE_INFO_GET) {
        this->ParseBoolValue(value, bool_flag_val);
        this->SetCableInfoGet(bool_flag_val);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_EYE_OPEN_INFO_CHECK) {
        this->SetEyeOpenBoundCheck();
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_EYE_EXPERT) {
        this->SetDumpEyeExpertFile();
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_CABLE_INFO_DICONNECTED) {
        this->ParseBoolValue(value, bool_flag_val);
        this->SetQueryDisconnectedPorts(bool_flag_val);
        IBDIAGNET_RETURN(0);
    }

    //if we reached here than unknown option
    IBDIAGNET_RETURN(1);
}


// CableDiag Clbck

void EyeOpenGetDelegator(const clbck_data_t &clbck_data,
                     int rec_status,
                     void *p_attribute_data)
{
    reinterpret_cast<CableDiag*>(clbck_data.m_p_obj)->
        EyeOpenGetClbck(clbck_data, rec_status, p_attribute_data);
}

void CableDiag::EyeOpenGetClbck(const clbck_data_t &clbck_data,
                                int rec_status,
                                void *p_attribute_data)
{
    if (m_ClbckErrorState)
        return;

    rec_status &= 0x00ff;   //bits [8-15] are class specific - don't need them here
    IBPort *p_port = (IBPort *)clbck_data.m_data1;

    if (rec_status &&
        ((p_port->p_node->appData1.val == NOT_SUPPORT_CAPABILTY) ||
         IS_PORT_NOT_SUPPORT_CAPABILTY(p_port)))
        IBDIAGNET_RETURN_VOID;

    if (rec_status == IBIS_MAD_STATUS_UNSUP_METHOD_ATTR) {         //not supported for this device
        p_port->p_node->appData1.val = NOT_SUPPORT_CAPABILTY;
        FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                new FabricErrNodeNotSupportCap(p_port->p_node,
                        "The firmware of this device does not support eye open capability");
        if (!p_curr_fabric_node_err) {
            this->SetLastError("Failed to allocate FabricErrNodeNotSupportCap");
            m_ClbckErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            p_cable_errors->push_back(p_curr_fabric_node_err);
        IBDIAGNET_RETURN_VOID;
    } else if (rec_status) {           //failed to get response
        SET_PORT_NOT_SUPPORT_CAPABILTY(p_port);
        FabricErrPortNotRespond *p_curr_fabric_port_err =
                new FabricErrPortNotRespond(p_port, "SMPEyeOpen");
        if (!p_curr_fabric_port_err) {
            this->SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ClbckErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            p_cable_errors->push_back(p_curr_fabric_port_err);
        IBDIAGNET_RETURN_VOID;
    }

    /* check AutoNegoStatus:
     *  0 - success
     *  1 - Speed Not QDR
     *  2 - Auto Negotiation in progress
     *  3 - Link is down
     *  ==>
     *      if 1 or 3 than we have bug in tool or FW,
     *      value 2 can occur.
     */
    struct SMP_EyeOpen *p_eye_open =
        (struct SMP_EyeOpen *)p_attribute_data;
    u_int8_t group_lane = (u_int8_t)(intptr_t)clbck_data.m_data2;

    if (p_eye_open->AutoNegoState) {
        SET_PORT_NOT_SUPPORT_CAPABILTY(p_port);

        FabricErrGeneral *p_curr_cable_err = NULL;
        if (p_eye_open->AutoNegoState == AUTONEGO_VALUE_IN_PROGRESS) {
            p_curr_cable_err =
                    new FabricErrEyeOpenInfoRetrieveAutonegInProgress(p_port);
        } else {
            p_curr_cable_err =
                    new FabricErrEyeOpenInfoRetrieveGeneral(p_port, p_eye_open->AutoNegoState);
        }
        if (!p_curr_cable_err) {
            this->SetLastError("Failed to allocate FabricErrGeneral");
            m_ClbckErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            p_cable_errors->push_back(p_curr_cable_err);
        IBDIAGNET_RETURN_VOID;
    }

    //everything is ok
    m_ClbckErrorState = this->AddSmpEyeOpen(p_port, p_port->p_remotePort, *p_eye_open, group_lane);
    IBDIAGNET_RETURN_VOID;
}

void CableInfoGetDelegator(const clbck_data_t &clbck_data,
                     int rec_status,
                     void *p_attribute_data)
{
    reinterpret_cast<CableDiag*>(clbck_data.m_p_obj)->
        CableInfoGetClbck(clbck_data, rec_status, p_attribute_data);
}

void CableDiag::CableInfoGetClbck(const clbck_data_t &clbck_data,
                                 int rec_status,
                                 void *p_attribute_data)
{
    if (m_ClbckErrorState)
        return;

    u_int8_t vs_status = (u_int8_t)((rec_status & 0x7f00) >> 8); //take the class specific status
    rec_status &= 0x00ff;   //bits [8-15] are class specific - don't need them here
    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    u_int8_t eeprom_index = (u_int8_t)(intptr_t)clbck_data.m_data2;
    u_int8_t page_num = (u_int8_t)(intptr_t)clbck_data.m_data3;

    u_int8_t not_supported_status = 0;

    if (rec_status &&
        ((p_port->p_node->appData1.val == NOT_SUPPORT_CAPABILTY) ||
         IS_PORT_NOT_SUPPORT_CAPABILTY(p_port)))
        IBDIAGNET_RETURN_VOID;

    if (rec_status == IBIS_MAD_STATUS_UNSUP_METHOD_ATTR) {         //not supported for this device
        p_port->p_node->appData1.val = NOT_SUPPORT_CAPABILTY;
        FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                new FabricErrNodeNotSupportCap(p_port->p_node,
                        "The firmware of this device does not support cable info capability");
        if (!p_curr_fabric_node_err) {
            this->SetLastError("Failed to allocate FabricErrNodeNotSupportCap");
            m_ClbckErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            p_cable_errors->push_back(p_curr_fabric_node_err);
        IBDIAGNET_RETURN_VOID;
    } else if (rec_status == IBIS_MAD_STATUS_INVALID_FIELD) {
        /* check vs_status:
         * only bad QSFP / No eeprom are allowed.
         * ==>
         * All others are bug in tool or FW
         */
        FabricErrGeneral *p_curr_cable_err = NULL;
        SET_PORT_NOT_SUPPORT_CAPABILTY(p_port);

        switch (vs_status) {
            case CABLEINFO_VS_STATUS_PORT_NOT_QSFP: {
                not_supported_status = vs_status;
                goto set_cable_info;
            }
            case CABLEINFO_VS_STATUS_BAD_QSFP_CABLE: {
                //no reasonable to proceed with this node so we will mark it as not supported
                //always we will get this error with this noe because there is a bad cable
                //connected to this node
                p_port->p_node->appData1.val = NOT_SUPPORT_CAPABILTY;
                p_curr_cable_err =
                        new FabricErrCableInfoRetrieveBadQSFPFound(p_port->p_node);
                break;
            }
            case CABLEINFO_VS_STATUS_NO_EEPROM: {
                p_curr_cable_err =
                        new FabricErrCableInfoRetrieveNoEEprom(p_port);
                break;
            }
            default: {
                p_curr_cable_err =
                        new FabricErrCableInfoRetrieveGeneral(p_port,
                                                              eeprom_index,
                                                              page_num,
                                                              vs_status);
                break;
            }
        }
        if (!p_curr_cable_err) {
            this->SetLastError("Failed to allocate FabricErrGeneral");
            m_ClbckErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            p_cable_errors->push_back(p_curr_cable_err);
        IBDIAGNET_RETURN_VOID;
    } else if (rec_status) {           //failed to get response
        SET_PORT_NOT_SUPPORT_CAPABILTY(p_port);
        FabricErrPortNotRespond *p_curr_fabric_port_err =
                new FabricErrPortNotRespond(p_port, "SMPCableInfo");
        if (!p_curr_fabric_port_err) {
            this->SetLastError("Failed to allocate FabricErrPortNotRespond");
            m_ClbckErrorState = IBDIAG_ERR_CODE_NO_MEM;
        } else
            p_cable_errors->push_back(p_curr_fabric_port_err);
        IBDIAGNET_RETURN_VOID;
    }

set_cable_info:
    struct SMP_CableInfo *p_SMPCableInfo =
        (struct SMP_CableInfo *)p_attribute_data;

    CableInfo *p_cable_info = NULL;
    m_ClbckErrorState = this->GetSMPCableInfo(
            p_port, p_port->p_remotePort, p_cable_info);
    if (m_ClbckErrorState)
        IBDIAGNET_RETURN_VOID;

    m_ClbckErrorState = p_cable_info->SetCableInfo(not_supported_status,
                                                   p_SMPCableInfo->Byte,
                                                   eeprom_index, page_num);
    if (m_ClbckErrorState)
        this->SetLastError("SetCableInfo Failed");

    IBDIAGNET_RETURN_VOID;
}

bool CableDiag::IsActiveCableActiveModule(CableInfo *p_cable_info)
{
    IBDIAGNET_ENTER;
    if (p_cable_info->IsModule() || p_cable_info->IsActiveCable())
        IBDIAGNET_RETURN(true);

    IBDIAGNET_RETURN(false);
}

bool CableDiag::IsMlnxMMFMlnxPSM(CableInfo *p_cable_info)
{
    IBDIAGNET_ENTER;
    if (p_cable_info->IsMlnxMmf() || p_cable_info->IsMlnxPsm())
        IBDIAGNET_RETURN(true);

    IBDIAGNET_RETURN(false);
}

void CableDiag::CreatePagesList()
{
    IBDIAGNET_ENTER;
    cable_info_addr_list_by_phase[0].push_back(new struct cable_info_address(
                EEPROM_CABLE_PAGE_FIRST_GET,
                EEPROM_CABLE_ADDR_FIRST_GET,
                EEPROM_CABLE_SIZE_FIRST_GET,
                0,
                NULL));

    cable_info_addr_list_by_phase[0].push_back(new struct cable_info_address(
                EEPROM_CABLE_PAGE_FIRST_GET,
                EEPROM_CABLE_ADDR_SECOND_GET,
                EEPROM_CABLE_SIZE_SECOND_GET,
                0,
                NULL));

    cable_info_addr_list_by_phase[0].push_back(new struct cable_info_address(
                EEPROM_CABLE_PAGE_FIRST_GET,
                CABLE_ADDR_48,
                CABLE_SIZE_32,
                0,
                NULL));

    cable_info_addr_list_by_phase[1].push_back(new struct cable_info_address(
                EEPROM_CABLE_PAGE_FIRST_GET,
                EEPROM_CABLE_ADDR_THIRD_GET,
                EEPROM_CABLE_SIZE_THIRD_GET,
                0,
                &IsActiveCableActiveModule));

    cable_info_addr_list_by_phase[1].push_back(new struct cable_info_address(
                EEPROM_CABLE_PAGE_FIRST_GET,
                CABLE_ADDR_80,
                CABLE_SIZE_48,
                0,
                &IsActiveCableActiveModule));

    cable_info_addr_list_by_phase[1].push_back(new struct cable_info_address(
                EEPROM_CABLE_PAGE_FIRST_GET,
                CABLE_ADDR_212,
                CABLE_SIZE_43,
                0,
                NULL));

    cable_info_addr_list_by_phase[1].push_back(new struct cable_info_address(
                CABLE_PAGE_03,
                CABLE_ADDR_212,
                CABLE_SIZE_43,
                0,
                &IsActiveCableActiveModule));

    cable_info_addr_list_by_phase[2].push_back(new struct cable_info_address(
                CABLE_PAGE_F0,
                CABLE_ADDR_128,
                CABLE_SIZE_44,
                CABLE_PAGE_F0_PASSWOED,
                &IsMlnxMMFMlnxPSM));

    cable_info_addr_list_by_phase[2].push_back(new struct cable_info_address(
                CABLE_PAGE_E9,
                CABLE_ADDR_128,
                CABLE_SIZE_48,
                CABLE_PAGE_E9_PASSWOED,
                &IsMlnxMMFMlnxPSM));

    cable_info_addr_list_by_phase[2].push_back(new struct cable_info_address(
                CABLE_PAGE_E9,
                CABLE_ADDR_176,
                CABLE_SIZE_36,
                CABLE_PAGE_E9_PASSWOED,
                &IsMlnxMMFMlnxPSM));

    IBDIAGNET_RETURN_VOID;
}
