/*
 *  Copyright (c) 2004-2013 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include "cable_diag_errs.h"


/****************************************************/
string ConvertCableInfoVSStatusToStr(IN u_int8_t vs_status)
{
    IBDIAGNET_ENTER;
    string result;

    switch (vs_status) {
    case CABLEINFO_VS_STATUS_NO_EEPROM: {
        result = "No eeprom in connected cable";
        break;
    }
    case CABLEINFO_VS_STATUS_INVALID_PORT: {
        result = "Invalid port number";
        break;
    }
    case CABLEINFO_VS_STATUS_INVALID_PAGE_NUM: {
        result = "Invalid page number (>15)";
        break;
    }
    case CABLEINFO_VS_STATUS_INVALID_ADDR_OR_SIZE: {
        result = "Address / size  is invalid. ((addr+size)>255) or size>48 or size==0";
        break;
    }
    case CABLEINFO_VS_STATUS_INVALID_SLAVE_ADDR: {
        result = "Slave address is invalid. (slv_addr != 0x50)";
        break;
    }
    //we don't suppose to get the following errors,
    //we check them before getting here
    case CABLEINFO_VS_STATUS_CABLE_NOT_CONNECTED: {
        result = "Cable not connected";
        break;
    }
    case CABLEINFO_VS_STATUS_PORT_NOT_QSFP: {
        result = "Port type is not QSFP (SFP/CX4 or internal)";
        break;
    }
    case CABLEINFO_VS_STATUS_BAD_QSFP_CABLE: {
        result = "Bad QSFP cable (cable ignore ModSel# signal)";
        break;
    }
    case CABLEINFO_VS_STATUS_SUCCESS: {
        result = "Success";
        break;
    }
    default: {
        result = "UNKNOWN";
        break;
    }
    }
    IBDIAGNET_RETURN(result);
};


string ConvertAutonegValueToStr(IN u_int8_t autoneg_val)
{
    IBDIAGNET_ENTER;
    string result;

    switch (autoneg_val) {
    case AUTONEGO_VALUE_IN_PROGRESS: {
        result = "Auto negotiation is in progress";
        break;
    }
    //we don't suppose to get the following errors,
    //we check them before getting here
    case AUTONEGO_VALUE_NOT_QDR: {
        result = "Port is not QDR";
        break;
    }
    case AUTONEGO_VALUE_LINK_DOWN: {
        result = "Link is down";
        break;
    }
    case AUTONEGO_VALUE_SUCCESS: {
        result = "Success";
        break;
    }
    default: {
        result = "UNKNOWN";
        break;
    }
    }
    IBDIAGNET_RETURN(result);
}


/****************************************************/
FabricErrEyeOpenInfoRetrieveAutonegInProgress::FabricErrEyeOpenInfoRetrieveAutonegInProgress(
        IBPort *p_port) : FabricErrPort(p_port)
{
    IBDIAGNET_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_EYE_OPEN_INFO_AUTONEG);
    this->description = "Failed to get eye open information";
    this->description += ": ";
    this->description += "Auto negotiation is in progress";
    IBDIAGNET_RETURN_VOID
}


FabricErrEyeOpenInfoRetrieveGeneral::FabricErrEyeOpenInfoRetrieveGeneral(IBPort *p_port,
        u_int8_t autoneg_val) : FabricErrPort(p_port)
{
    IBDIAGNET_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_EYE_OPEN_INFO_ERR_GENERAL);
    this->description = "Failed to get eye open information";
    this->description += ": ";
    this->description += ConvertAutonegValueToStr(autoneg_val);
    IBDIAGNET_RETURN_VOID;
}


/****************************************************/
FabricErrEyeBoundBelowThresh::FabricErrEyeBoundBelowThresh(IBPort *p_port, int lane_num,
        u_int16_t negative_bound, u_int16_t positive_bound, u_int16_t threshold) :
        FabricErrPort(p_port), lane_num(lane_num),
        negative_bound(negative_bound), positive_bound(positive_bound), sum_bound_threshold(threshold)
{
    IBDIAGNET_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_EYE_BOUND_BELOW_THRESH);
    char buffer[1024];
    sprintf(buffer, "Lane=%u :: Negative bound=%u, Positive bound=%u ==> "    \
            "Bound sum=%u is too low for threshold=%u",
            this->lane_num,
            this->negative_bound, this->positive_bound,
            this->negative_bound + this->positive_bound,
            this->sum_bound_threshold);
    this->description.assign(buffer);
    IBDIAGNET_RETURN_VOID
}

FabricErrEyeBoundAboveThresh::FabricErrEyeBoundAboveThresh(IBPort *p_port, int lane_num,
        u_int16_t negative_bound, u_int16_t positive_bound, u_int16_t threshold) :
        FabricErrPort(p_port), lane_num(lane_num),
        negative_bound(negative_bound), positive_bound(positive_bound), sum_bound_threshold(threshold)
{
    IBDIAGNET_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_EYE_BOUND_ABOVE_THRESH);
    char buffer[1024];
    sprintf(buffer, "Lane=%u :: Negative bound=%u, Positive bound=%u ==> "    \
            "Bound sum=%u is too high for threshold=%u",
            this->lane_num,
            this->negative_bound, this->positive_bound,
            this->negative_bound + this->positive_bound,
            this->sum_bound_threshold);
    this->description.assign(buffer);
    IBDIAGNET_RETURN_VOID
}


/****************************************************/
FabricErrCableInfoRetrieveBadQSFPFound::FabricErrCableInfoRetrieveBadQSFPFound(IBNode *p_node) :
        FabricErrNode(p_node)
{
    IBDIAGNET_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_CABLE_INFO_BAD_QSFP);
    this->description = "Failed to get cable information";
    this->description += ": ";
    this->description += "Bad QSFP cable (cable ignore ModSel# signal) is connected to this node";
    IBDIAGNET_RETURN_VOID;
}

FabricErrCableInfoRetrieveNoEEprom::FabricErrCableInfoRetrieveNoEEprom(IBPort *p_port) :
        FabricErrPort(p_port)
{
    IBDIAGNET_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_CABLE_INFO_NO_EEPROM);
    this->description = "Failed to get cable information";
    this->description += ": ";
    this->description += "No eeprom in connected cable";
    IBDIAGNET_RETURN_VOID;
}

FabricErrCableInfoRetrieveGeneral::FabricErrCableInfoRetrieveGeneral(IBPort *p_port,
        u_int8_t eeprom_index,
        u_int8_t page_num,
        u_int8_t vs_status) : FabricErrPort(p_port)
{
    IBDIAGNET_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_CABLE_INFO_ERR_GENERAL);
    this->description = "Failed to get cable information";
    this->description += ": ";
    //disconnected HCA/RTR port, don't have the direct route to it
    //connectIB and newer fw implement
    //"MAD sent to Channel Adapters and Routers
    //impact only the port it was received on."
    //However older fw in the field (connectX3) still support
    //direct route to node + port num, and we may not see that error there.
    if (!p_port->p_remotePort && p_port->p_node->type != IB_SW_NODE) {
        this->description += "Disconnected CA port";
    } else {
        char buffer[1024];
        sprintf(buffer, "For page=%u address=%u, ",
                page_num,eeprom_index);
        this->description.assign(buffer);
        this->description += ConvertCableInfoVSStatusToStr(vs_status);
    }
    IBDIAGNET_RETURN_VOID;
}

