/*
 * Copyright (c) 2004-2019 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <typeinfo>
#include <math.h>
#include <assert.h>
#include "phy_diag.h"
#include "acc_reg.h"
#include "phy_diag_errs.h"

static inline bool IsSupportAccRegFailed(u_int64_t app_data,
                                         u_int64_t acc_reg_type,
                                         u_int64_t not_support_bit)
{
    return (app_data & (acc_reg_type | not_support_bit));
}

bool keycomp(AccRegKey *k1, AccRegKey *k2)
{
    return ((*k1) < (*k2));
};

Register::Register(uint32_t reg_id,
                   const unpack_data_func_t p_unpack_func,
                   string sn,
                   uint32_t fn,
                   uint64_t nsb,
                   string header,
                   PhyPluginSupportedNodesType support_nodes,
                   bool to_dump,
                   AccRegVia_t acc_reg_type)
{
    IBDIAGNET_ENTER;
    register_id = reg_id;
    section_name = sn;
    unpack_data_func = p_unpack_func;
    fields_num = fn;
    not_supported_bit = nsb;
    m_support_nodes = support_nodes;
    this->to_dump = to_dump;
    this->header = header;
    this->acc_reg_type = acc_reg_type;
    IBDIAGNET_RETURN_VOID;
}

bool Register::IsRegSupportNodeType(IBNodeType type)
{
    IBDIAGNET_ENTER;
    bool rc = IsPhyPluginSupportNodeType(m_support_nodes, type);
    IBDIAGNET_RETURN(rc);
}


void Register::DumpRegisterHeader(stringstream &sstream) {
    IBDIAGNET_ENTER;
    if (this->header.empty())
        for (u_int32_t cnt = 0; cnt < this->fields_num; cnt++)
            sstream << ",field" <<  cnt;
    else
        sstream << this->header;

    IBDIAGNET_RETURN_VOID;
}

void Register::GetGroupData(struct acc_reg_data& areg, u_int8_t& group, u_int16_t& indices)
{
    IBDIAGNET_ENTER;
    assert(true); // make sure that this function called only from
                  // derived registers that overrun this function.

    IBDIAGNET_RETURN_VOID;
}

int Register::AvailableSensors(struct acc_reg_data& areg, list_uint8& sensors)
{
    IBDIAGNET_ENTER;
    assert(true); // make sure that this function called only from
                  // derived classes that overrun this function.

    IBDIAGNET_RETURN(IBDIAG_ERR_CODE_ILLEGAL_FUNCTION_CALL);
}

int Register::SensorsBitsToList(u_int64_t sensors,
                                u_int8_t max_sensors,
                                list_uint8& sensors_list)
{
    IBDIAGNET_ENTER;
    u_int64_t mask = 1;

    for (u_int8_t i = 0; i < max_sensors; ++i, mask <<= 1) {
        if (mask & sensors)
            sensors_list.push_back(i);
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int Register::SensorsCountToList(u_int8_t sensor_count, list_uint8& sensors_list)
{
    IBDIAGNET_ENTER;

    for (u_int8_t i = 0; i < sensor_count; ++i)
        sensors_list.push_back(i);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int Register::BuildDB(AccRegHandler* p_handler,
                      list_p_fabric_general_err &phy_errors,
                      progress_func_nodes_t progress_func)
{
    IBDIAGNET_ENTER;
    assert(false); // make sure that this function called only from
                   // derived classes that overrun this function.

    IBDIAGNET_RETURN(IBDIAG_ERR_CODE_ILLEGAL_FUNCTION_CALL);
}

//*****************************************
//************* PPLL Register functions ***
//*****************************************
PPLLRegister::PPLLRegister():Register(ACCESS_REGISTER_ID_PPLL,
                                      (const unpack_data_func_t)ppll_reg_unpack,
                                      ACC_REG_PPLL_INTERNAL_SECTION_NAME,
                                      ACC_REG_PPLL16NM_FIELDS_NUM,
                                      NOT_SUPPORT_PPLL)
{
}

void PPLLRegister::PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg)
{
    IBDIAGNET_ENTER;
    acc_reg->register_id = this->register_id;
    struct ppll_reg ppll;
    CLEAR_STRUCT(ppll);
    ppll.pll_group = ((AccRegKeyGroup *)p_key)->group_num;
    ppll_reg_pack(&ppll, acc_reg->reg.data);
    IBDIAGNET_RETURN_VOID;
}

void PPLLRegister::DumpRegisterData(struct acc_reg_data areg, stringstream &sstream)
{
    IBDIAGNET_ENTER;
    char buff[1024] = {0};
    struct ppll_reg *p_ppll = &areg.ppll;
    memset(buff, 0, sizeof(buff));

    //register fields
    sprintf(buff, "version_%u," UH_FMT "," UH_FMT "," UH_FMT ",",
            p_ppll->version,
            p_ppll->num_pll_groups,
            p_ppll->pll_group,
            p_ppll->num_plls);

    sstream << buff;
    memset(buff, 0, sizeof(buff));

    if (p_ppll->version == 0) { // ppll_28nm

        struct pll_status_28nm status0;
        struct pll_status_28nm status1;
        struct pll_status_28nm status2;
        struct pll_status_28nm status3;

        pll_status_28nm_unpack(&status0, p_ppll->pll_status0.pll_status_data.data);
        pll_status_28nm_unpack(&status1, p_ppll->pll_status1.pll_status_data.data);
        pll_status_28nm_unpack(&status2, p_ppll->pll_status2.pll_status_data.data);
        pll_status_28nm_unpack(&status3, p_ppll->pll_status3.pll_status_data.data);

        sprintf(buff, UH_FMT "," UH_FMT ","
                UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
                UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
                UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
                UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
                UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
                UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
                UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT
                ",NA,NA,NA,NA,NA,NA,NA,NA,NA"  // add NAs to fill the missing
                ",NA,NA,NA,NA,NA,NA,NA,NA,NA"  // number of fields in compare to
                ",NA,NA,NA,NA,NA,NA,NA,NA,NA", // the new version. (68 - 41 = 27)
                p_ppll->ae,

                status0.lock_cal,
                status0.lock_status,
                status0.algo_f_ctrl,
                status0.analog_algo_num_var,
                status0.f_ctrl_measure,
                status0.analog_var,
                status0.high_var,
                status0.low_var,
                status0.mid_var,

                status1.lock_cal,
                status1.lock_status,
                status1.algo_f_ctrl,
                status1.analog_algo_num_var,
                status1.f_ctrl_measure,
                status1.analog_var,
                status1.high_var,
                status1.low_var,
                status1.mid_var,

                status2.lock_cal,
                status2.lock_status,
                status2.algo_f_ctrl,
                status2.analog_algo_num_var,
                status2.f_ctrl_measure,
                status2.analog_var,
                status2.high_var,
                status2.low_var,
                status2.mid_var,

                status3.lock_cal,
                status3.lock_status,
                status3.algo_f_ctrl,
                status3.analog_algo_num_var,
                status3.f_ctrl_measure,
                status3.analog_var,
                status3.high_var,
                status3.low_var,
                status3.mid_var);
    }

    if (p_ppll->version == 1) { // ppll_16nm

        struct pll_status_16nm status0;
        struct pll_status_16nm status1;
        struct pll_status_16nm status2;
        struct pll_status_16nm status3;

        pll_status_16nm_unpack(&status0, p_ppll->pll_status0.pll_status_data.data);
        pll_status_16nm_unpack(&status1, p_ppll->pll_status1.pll_status_data.data);
        pll_status_16nm_unpack(&status2, p_ppll->pll_status2.pll_status_data.data);
        pll_status_16nm_unpack(&status3, p_ppll->pll_status3.pll_status_data.data);

        sprintf(buff,
                UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
                UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
                UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
                UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
                UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
                UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
                UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
                UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
                UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
                UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
                UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
                UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
                UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT,
                status0.pll_speed,
                status0.clock_valid,
                status0.plllock_clk_val,
                status0.lock_clk_val_cause,
                status0.clear_cause,
                status0.ae,
                status0.lock_lost_counter,
                status0.lock_pulse,
                status0.pll_pwrup,
                status0.lock_status,
                status0.fctrl_measure,
                status0.mid_val,
                status0.low_var,
                status0.high_var,
                status0.analog_var,
                status0.lock,

                status1.pll_speed,
                status1.clock_valid,
                status1.plllock_clk_val,
                status1.lock_clk_val_cause,
                status1.clear_cause,
                status1.ae,
                status1.lock_lost_counter,
                status1.lock_pulse,
                status1.pll_pwrup,
                status1.lock_status,
                status1.fctrl_measure,
                status1.mid_val,
                status1.low_var,
                status1.high_var,
                status1.analog_var,
                status1.lock,

                status2.pll_speed,
                status2.clock_valid,
                status2.plllock_clk_val,
                status2.lock_clk_val_cause,
                status2.clear_cause,
                status2.ae,
                status2.lock_lost_counter,
                status2.lock_pulse,
                status2.pll_pwrup,
                status2.lock_status,
                status2.fctrl_measure,
                status2.mid_val,
                status2.low_var,
                status2.high_var,
                status2.analog_var,
                status2.lock,

                status3.pll_speed,
                status3.clock_valid,
                status3.plllock_clk_val,
                status3.lock_clk_val_cause,
                status3.clear_cause,
                status3.ae,
                status3.lock_lost_counter,
                status3.lock_pulse,
                status3.pll_pwrup,
                status3.lock_status,
                status3.fctrl_measure,
                status3.mid_val,
                status3.low_var,
                status3.high_var,
                status3.analog_var,
                status3.lock);
    }

    sstream << buff << endl;
    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//************* PTAS Register functions ***
//*****************************************
PTASRegister::PTASRegister():Register(ACCESS_REGISTER_ID_PTAS,
                                      (const unpack_data_func_t)ptas_reg_unpack,
                                      ACC_REG_PTAS_INTERNAL_SECTION_NAME,
                                      ACC_REG_PTAS_FIELDS_NUM,
                                      NOT_SUPPORT_PTAS)
{
}

void PTASRegister::PackData(AccRegKey *p_key,
                            struct SMP_AccessRegister *acc_reg)
{
    IBDIAGNET_ENTER;
    acc_reg->register_id = this->register_id;
    IBDIAGNET_RETURN_VOID;
}

void PTASRegister::DumpRegisterData(struct acc_reg_data areg, stringstream &sstream)
{
    IBDIAGNET_ENTER;
    char buff[1024] = {0};
    struct ptas_reg *p_ptas = &areg.ptas;
    memset(buff, 0, sizeof(buff));
    //register fields
    sprintf(buff,
            UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
            UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
            UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
            UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
            UH_FMT "," UH_FMT "," UH_FMT ,
            p_ptas->algorithm_options,
            p_ptas->repetitions_mode,
            p_ptas->num_of_repetitions,
            p_ptas->grade_version,
            p_ptas->height_grade_type,
            p_ptas->phase_grade_type,
            p_ptas->height_grade_weight,
            p_ptas->phase_grade_weight,
            p_ptas->gisim_measure_bits,
            p_ptas->adaptive_tap_measure_bits,
            p_ptas->ber_bath_high_error_threshold,
            p_ptas->ber_bath_mid_error_threshold,
            p_ptas->ber_bath_low_error_threshold,
            p_ptas->one_ratio_high_threshold,
            p_ptas->one_ratio_high_mid_threshold,
            p_ptas->one_ratio_low_mid_threshold,
            p_ptas->one_ratio_low_threshold,
            p_ptas->ndeo_error_threshold,
            p_ptas->mixer_offset_step_size,
            p_ptas->mix90_phase_for_voltage_bath,
            p_ptas->mixer_offset_start,
            p_ptas->mixer_offset_end,
            p_ptas->ber_test_time);
    sstream << buff << endl;
    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//********   SLRegister functions    ******
//*****************************************
SLRegister::SLRegister(uint32_t reg_id,
                       const unpack_data_func_t p_unpack_func,
                       string sn,
                       uint32_t fn,
                       uint64_t nsb,
                       map_akey_areg *mpein_map,
                       map_akey_areg *mpir_map) : Register(reg_id,
                                                           p_unpack_func,
                                                           sn,
                                                           fn,
                                                           nsb),
                                                  p_mpein_map(mpein_map),
                                                  p_mpir_map(mpir_map){}

/*
 * this register is for slrg slrp sltp registers in pci mode.
 * it goes over all valid MPIR queries and get the above registers.
 * p_handler - contain the MPIR map.
 * */
int SLRegister::BuildDB(AccRegHandler* p_handler,
                        list_p_fabric_general_err &phy_errors,
                        progress_func_nodes_t progress_func)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    clbck_data_t clbck_data;
    clbck_data.m_handle_data_func = SMPAccessRegisterHandlerGetDelegator;
    clbck_data.m_p_obj = p_handler;

    p_handler->SetHeader(PORT_LANE_HEADER);

    for(map_akey_areg::iterator mpir_iter = p_mpir_map->begin();
            mpir_iter != p_mpir_map->end(); mpir_iter++) {

        AccRegKeyDPN *p_dpnkey = (AccRegKeyDPN*)mpir_iter->first;
        mpir_reg mpir = mpir_iter->second.mpir;
        mpein_reg mpein = p_mpein_map->find(p_dpnkey)->second.mpein;

        IBNode *p_curr_node =
                p_handler->p_phy_diag->p_discovered_fabric->getNodeByGuid(p_dpnkey->node_guid);

        if (!p_curr_node) {
            p_handler->p_phy_diag->SetLastError("DB error - found null node in "
                    "NodeByName map for key = " U64H_FMT, p_dpnkey->node_guid);
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        direct_route_t *p_curr_direct_route = p_handler->p_phy_diag->p_ibdiag->
                GetDirectRouteByNodeGuid(p_curr_node->guid_get());
        if (!p_curr_direct_route) {
            p_handler->p_phy_diag->SetLastError("DB error - can't find direct route "
                                                "to node=%s (node guid: 0x%x)",
                                                p_curr_node->getName().c_str(),
                                                p_curr_node->guid_get());
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        IBPort *p_curr_port = NULL;
        for (u_int32_t port_num = 1; port_num < (unsigned)p_curr_node->numPorts + 1; ++port_num) {
            p_curr_port = p_curr_node->getPort((phys_port_t)port_num);

            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;
            break; // use the first valid port.
        }

        if (!p_curr_port) {
            p_handler->p_phy_diag->SetLastError("DB error - failed to find valid port for node %s",
                                                p_curr_node->getName().c_str());
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }


        for (u_int8_t lane = 0; lane < mpein.link_width_active; lane++) {
            AccRegKeyPortLane *p_plkey = new AccRegKeyPortLane(
                    p_curr_node->guid_get(),
                    p_curr_port->guid_get(),
                    mpir.local_port,
                    lane);

            clbck_data.m_data1 = p_curr_node;
            clbck_data.m_data2 = p_plkey;
            struct SMP_AccessRegister mad_areg;
            CLEAR_STRUCT(mad_areg.reg.data);

            PackData(p_plkey, &mad_areg);
            p_handler->p_phy_diag->SMPAccRegGetByDirect(p_curr_direct_route,
                                                        p_curr_port->num,
                                                        &mad_areg,
                                                        &clbck_data);

        } // lane loop
    } // mpir loop

    IBDIAGNET_RETURN(rc);
}

//*****************************************
//************* SLRP Register functions ***
//*****************************************
SLRPRegister::SLRPRegister(u_int8_t pnat, string sn, map_akey_areg *mpein_map, map_akey_areg *mpir_map):
    SLRegister(ACCESS_REGISTER_ID_SLRP,
               (const unpack_data_func_t)slrp_reg_unpack,
               sn,
               ACC_REG_SLRP16NM_FIELDS_NUM,
               NOT_SUPPORT_SLRP,
               mpein_map,
               mpir_map),
    m_pnat(pnat)
{
    if (pnat == ACC_REG_PNAT_OOB_PORT)
        m_support_nodes = HCA_ONLY;
}

void SLRPRegister::PackData(AccRegKey *p_key,
                            struct SMP_AccessRegister *acc_reg)
{
    IBDIAGNET_ENTER;
    acc_reg->register_id = this->register_id;
    struct slrp_reg slrp;
    CLEAR_STRUCT(slrp);
    slrp.local_port = ((AccRegKeyPortLane *)p_key)->port_num;
    slrp.lane = ((AccRegKeyPortLane *)p_key)->lane;
    slrp.pnat = m_pnat;
    slrp_reg_pack(&slrp, acc_reg->reg.data);
    IBDIAGNET_RETURN_VOID;
}

void SLRPRegister::DumpRegisterData(struct acc_reg_data areg, stringstream &sstream)
{
    IBDIAGNET_ENTER;
    struct slrp_reg *p_slrp = &areg.slrp;

    sstream << +p_slrp->status << ','
            << +p_slrp->version << ','
            << +p_slrp->local_port << ','
            << +p_slrp->pnat << ','
            << +p_slrp->lane << ',';

    if (p_slrp->version == 0 || p_slrp->version == 1) { // 40nm, 28nm

        slrp_28nm slrp_28;
        slrp_28nm_unpack(&slrp_28, p_slrp->page_data.slrp_data_set.data);

        sstream << +slrp_28.mix90phase         << ','
                << +slrp_28.dp90sel            << ','
                << +slrp_28.dp_sel             << ','
                << +slrp_28.ib_sel             << ','
                << +slrp_28.ffe_tap3           << ','
                << +slrp_28.ffe_tap2           << ','
                << +slrp_28.ffe_tap1           << ','
                << +slrp_28.ffe_tap0           << ','
                << +slrp_28.ffe_tap7           << ','
                << +slrp_28.ffe_tap6           << ','
                << +slrp_28.ffe_tap5           << ','
                << +slrp_28.ffe_tap4           << ','
                << +slrp_28.ffe_tap_en         << ','
                << +slrp_28.mixerbias_tap_amp  << ','
                << +slrp_28.ffe_tap8           << ','
                << +slrp_28.slicer_offset0     << ','
                << +slrp_28.ffe_tap_offset1    << ','
                << +slrp_28.ffe_tap_offset0    << ','
                << +slrp_28.mixer_offset1      << ','
                << +slrp_28.mixer_offset0      << ','
                << +slrp_28.mixerbgn_refn      << ','
                << +slrp_28.mixerbgn_refp      << ','
                << +slrp_28.mixerbgn_inn       << ','
                << +slrp_28.mixerbgn_inp       << ','
                << +slrp_28.mixer_offset_cm1   << ','
                << +slrp_28.lctrl_input        << ','
                << +slrp_28.slicer_gctrl       << ','
                << +slrp_28.ref_mixer_vreg     << ','
                << +slrp_28.sel_slicer_lctrl_l << ','
                << +slrp_28.sel_slicer_lctrl_h << ','
                << +slrp_28.slicer_offset_cm   << ','
                << +slrp_28.mixer_offset_cm0   << ','
                << +slrp_28.common_mode        << ','
                << "NA,NA,NA,NA,NA,NA,NA,NA,NA,NA"; // add NAs to fill the empty fields compare to 16nm (48 - 38)
    }

    if (p_slrp->version == 3) { // 16nm
        slrp_16nm slrp_16;
        slrp_16nm_unpack(&slrp_16, p_slrp->page_data.slrp_data_set.data);

        sstream << +slrp_16.dp_sel                         << ','
                << +slrp_16.sel_ref_out_mixer_biasgen_eo   << ','
                << +slrp_16.sel_ref_out_mixer_biasgen_down << ','
                << +slrp_16.sel_ref_out_mixer_biasgen_mid  << ','
                << +slrp_16.sel_ref_out_mixer_biasgen_up   << ','
                << +slrp_16.sel_ref_in_mixer_biasgen_eo    << ','
                << +slrp_16.sel_ref_in_mixer_biasgen_down  << ','
                << +slrp_16.sel_ref_in_mixer_biasgen_mid   << ','
                << +slrp_16.sel_ref_in_mixer_biasgen_up    << ','
                << +slrp_16.gctrl_bin_bgn1_p               << ','
                << +slrp_16.gctrl_bin_bgn1_n               << ','
                << +slrp_16.gctrl_bin_bgn0_p               << ','
                << +slrp_16.gctrl_bin_bgn0_n               << ','
                << +slrp_16.mixer_offset_eye_scan          << ','
                << +slrp_16.sel_amp_mixer_biasgen_eo       << ','
                << +slrp_16.sel_amp_mixer_biasgen_down     << ','
                << +slrp_16.sel_amp_mixer_biasgen_mid      << ','
                << +slrp_16.sel_amp_mixer_biasgen_up       << ','
                << +slrp_16.sel_bias_en_2                  << ','
                << +slrp_16.sel_bias_en_1                  << ','
                << +slrp_16.sel_bias_en_0                  << ','
                << +slrp_16.selgc_ref_op1_alev_ctle        << ','
                << +slrp_16.selgc_ref_op0_alev_ctle        << ','
                << +slrp_16.sel_bias2ctle                  << ','
                << +slrp_16.sel_digital_gctrl              << ','
                << +slrp_16.sel_gctrln_en_2                << ','
                << +slrp_16.sel_gctrln_en_1                << ','
                << +slrp_16.sel_gctrln_en_0                << ','
                << +slrp_16.sel_bias_en_6                  << ','
                << +slrp_16.sel_bias_en_5                  << ','
                << +slrp_16.sel_bias_en_4                  << ','
                << +slrp_16.sel_bias_en_3                  << ','
                << +slrp_16.sel_gctrlp_en_3                << ','
                << +slrp_16.sel_gctrlp_en_2                << ','
                << +slrp_16.sel_gctrlp_en_1                << ','
                << +slrp_16.sel_gctrlp_en_0                << ','
                << +slrp_16.sel_gctrln_en_6                << ','
                << +slrp_16.sel_gctrln_en_5                << ','
                << +slrp_16.sel_gctrln_en_4                << ','
                << +slrp_16.sel_gctrln_en_3                << ','
                << +slrp_16.sel_gctrlp_en_6                << ','
                << +slrp_16.sel_gctrlp_en_5                << ','
                << +slrp_16.sel_gctrlp_en_4;
    }
    sstream << endl;
    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//************* SLTP Register functions ***
//*****************************************
SLTPRegister::SLTPRegister(u_int8_t pnat, string sn, map_akey_areg *mpein_map, map_akey_areg *mpir_map):
    SLRegister(ACCESS_REGISTER_ID_SLTP,
               (const unpack_data_func_t)sltp_reg_unpack,
               sn,
               ACC_REG_SLTP_FIELDS_NUM,
               NOT_SUPPORT_SLTP,
               mpein_map,
               mpir_map),
    m_pnat(pnat)
{
    if (pnat == ACC_REG_PNAT_OOB_PORT)
        m_support_nodes = HCA_ONLY;
}

void SLTPRegister::PackData(AccRegKey *p_key,
                            struct SMP_AccessRegister *acc_reg)
{
    IBDIAGNET_ENTER;
    acc_reg->register_id = this->register_id;
    struct sltp_reg sltp;
    CLEAR_STRUCT(sltp);
    sltp.local_port = ((AccRegKeyPortLane *)p_key)->port_num;
    sltp.lane = ((AccRegKeyPortLane *)p_key)->lane;
    sltp.pnat = m_pnat;
    sltp_reg_pack(&sltp, acc_reg->reg.data);
    IBDIAGNET_RETURN_VOID;
}

void SLTPRegister::DumpRegisterData(struct acc_reg_data areg, stringstream &sstream)
{
    IBDIAGNET_ENTER;
    char buff[1024] = {0};
    struct sltp_reg *p_sltp = &areg.sltp;
    memset(buff, 0, sizeof(buff));
    //register fields
    sprintf(buff,
            UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
            UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ,
            p_sltp->status,
            p_sltp->version,
            p_sltp->polarity,
            p_sltp->ob_tap0,
            p_sltp->ob_tap1,
            p_sltp->ob_tap2,
            p_sltp->ob_preemp_mode,
            p_sltp->ob_reg,
            p_sltp->ob_bias);
    sstream << buff << endl;
    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//************* SLRG Register functions ***
//*****************************************
SLRGRegister::SLRGRegister(u_int8_t pnat, string sn, map_akey_areg *mpein_map, map_akey_areg *mpir_map):
    SLRegister(ACCESS_REGISTER_ID_SLRG,
               (const unpack_data_func_t)slrg_reg_unpack,
               sn,
               ACC_REG_SLRG28NM_FIELDS_NUM,
               NOT_SUPPORT_SLRG,
               mpein_map,
               mpir_map),
    m_pnat(pnat)
{
    if (pnat == ACC_REG_PNAT_OOB_PORT)
        m_support_nodes = HCA_ONLY;
}


void SLRGRegister::PackData(AccRegKey *p_key,
                            struct SMP_AccessRegister *acc_reg)
{
    IBDIAGNET_ENTER;
    acc_reg->register_id = this->register_id;
    struct slrg_reg slrg;
    CLEAR_STRUCT(slrg);
    slrg.local_port = ((AccRegKeyPortLane *)p_key)->port_num;
    slrg.lane = ((AccRegKeyPortLane *)p_key)->lane;
    slrg.pnat = m_pnat;
    slrg_reg_pack(&slrg, acc_reg->reg.data);
    IBDIAGNET_RETURN_VOID;
}

void SLRGRegister::DumpRegisterData(struct acc_reg_data areg, stringstream &sstream)
{
    IBDIAGNET_ENTER;
    struct slrg_reg *p_slrg = &areg.slrg;

    sstream << +p_slrg->status << ','
            << +p_slrg->version << ','
            << +p_slrg->local_port << ','
            << +p_slrg->pnat << ','
            << +p_slrg->lane << ',';


    if (p_slrg->version == 0 || p_slrg->version == 1) { // 40nm, 28nm

        slrg_28nm slrg_28;
        slrg_28nm_unpack(&slrg_28, p_slrg->page_data.slrg_data_set.data);

        sstream << +slrg_28.grade_lane_speed << ','
                << +slrg_28.grade_version << ','
                << +slrg_28.grade << ','
                << +slrg_28.height_grade_type << ','
                << +slrg_28.height_grade << ','
                << +slrg_28.height_dz << ','
                << +slrg_28.height_dv << ','
                << +slrg_28.height_sigma << ','
                << +slrg_28.phase_grade_type << ','
                << +slrg_28.phase_grade << ','
                << +slrg_28.phase_eo_pos << ','
                << +slrg_28.phase_eo_neg << ','
                << +slrg_28.ffe_set_tested << ','
                << +slrg_28.test_errors_per_lane;

    }

    if (p_slrg->version == 3) { // 16nm
        slrg_16nm slrg_16;
        slrg_16nm_unpack(&slrg_16, p_slrg->page_data.slrg_data_set.data);

        sstream << +slrg_16.grade_lane_speed << ','
                << +slrg_16.grade_version << ','
                << +slrg_16.grade << ','
                << +slrg_16.up_eye_grade << ','
                << +slrg_16.mid_eye_grade << ','
                << +slrg_16.dn_eye_grade << ','
                << "NA,NA,NA,NA,NA,NA,NA,NA"; // add NAs to fill the empty fields compare to 28nm (19 - 11)
    }
    sstream << endl;
    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//************* MPEIN Register functions ***
//*****************************************
MPEINRegister::MPEINRegister():Register(ACCESS_REGISTER_ID_MPEIN,
                                      (const unpack_data_func_t)mpein_reg_unpack,
                                      ACC_REG_MPEIN_INTERNAL_SECTION_NAME,
                                      ACC_REG_MPEIN_FIELDS_NUM,
                                      NOT_SUPPORT_MPEIN)
{
}


void MPEINRegister::PackData(AccRegKey *p_key,
                            struct SMP_AccessRegister *acc_reg)
{
    IBDIAGNET_ENTER;
    acc_reg->register_id = this->register_id;
    AccRegKeyDPN *p_dpnkey = (AccRegKeyDPN*)p_key;
    struct mpein_reg mpein;
    CLEAR_STRUCT(mpein);

    mpein.depth = p_dpnkey->depth;
    mpein.pcie_index = p_dpnkey->pci_idx;
    mpein.node = p_dpnkey->pci_node;

    mpein_reg_pack(&mpein, acc_reg->reg.data);
    IBDIAGNET_RETURN_VOID;
}

void MPEINRegister::DumpRegisterData(struct acc_reg_data areg, stringstream &sstream)
{
    IBDIAGNET_ENTER;
    char buff[1024] = {0};
    struct mpein_reg *p_mpein = &areg.mpein;
    memset(buff, 0, sizeof(buff));
    //register fields
    sprintf(buff,
            UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
            UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
            UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT,
            p_mpein->capability_mask,
            p_mpein->link_speed_enabled,
            p_mpein->link_width_enabled,
            p_mpein->link_speed_active,
            p_mpein->link_width_active,
            p_mpein->num_of_vfs,
            p_mpein->num_of_pfs,
            p_mpein->bdf0,
            p_mpein->lane_reversal,
            p_mpein->port_type,
            p_mpein->max_payload_size,
            p_mpein->max_read_request_size,
            p_mpein->port_state,
            p_mpein->device_status,
            p_mpein->receiver_detect_result);
    sstream << buff << endl;
    IBDIAGNET_RETURN_VOID;
}

void LoopPCIDepthDelegator(const clbck_data_t &clbck_data,
                           int rec_status,
                           void *p_attribute_data)
{
    reinterpret_cast<MPEINRegister*>
    (reinterpret_cast<AccRegHandler*>(clbck_data.m_p_obj)->GetPReg())->
    LoopPCIDepth(clbck_data, rec_status, p_attribute_data);
}

void LoopPCINodeDelegator(const clbck_data_t &clbck_data,
                          int rec_status,
                          void *p_attribute_data)
{
    reinterpret_cast<MPEINRegister*>
    (reinterpret_cast<AccRegHandler*>(clbck_data.m_p_obj)->GetPReg())->
    LoopPCINode(clbck_data, rec_status, p_attribute_data);
}


/*
 * Build DB in the following order:
 * For each CA
 *   For each PCI index
 *     For each PCI depth
 *       For each PCI node
*/
int MPEINRegister::BuildDB(AccRegHandler* p_handler,
                           list_p_fabric_general_err &phy_errors,
                           progress_func_nodes_t progress_func)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    // clbck_data.m_handle_data_func - SMPAccessRegisterHandlerGetDelegator
    // clbck_data.m_p_obj            - handler
    // clbck_data.m_data1            - node
    // clbck_data.m_data2            - key
    // clbck_data.m_data3            - port
    clbck_data_t clbck_data;
    clbck_data.m_handle_data_func = LoopPCIDepthDelegator;
    clbck_data.m_p_obj = p_handler;
    progress_bar_nodes_t progress_bar_nodes;
    CLEAR_STRUCT(progress_bar_nodes);

    p_handler->SetHeader(PCI_DPN_HEADER);

    for (map_str_psys::iterator sys_iter = p_handler->p_phy_diag->p_discovered_fabric->SystemByName.begin();
            sys_iter != p_handler->p_phy_diag->p_discovered_fabric->SystemByName.end(); sys_iter++) {

        IBSystem *p_curr_sys = sys_iter->second;
        if (!p_curr_sys) {
            p_handler->p_phy_diag->SetLastError("DB error - found null node in"  \
                    " SystemByName map for key = %s",
                    sys_iter->first.c_str());
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        IBNode *p_curr_node = p_curr_sys->NodeByName.begin()->second;
        if (!p_curr_node) {
            p_handler->p_phy_diag->SetLastError("DB error - found null node in"  \
                    " NodeByName map in IBSystem for key = %s",
                    p_curr_sys->NodeByName.begin()->first.c_str());
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (p_curr_node->type == IB_SW_NODE)
            continue;

        // skip ANs
        // under assumption that ANs are connected on port #1
        IBPort *p_port = p_curr_node->getPort(1);
        if (p_port && p_port->p_remotePort && p_port->p_remotePort->isSpecialPort())
            continue;

        ++progress_bar_nodes.nodes_found;
        if (p_curr_node->type == IB_SW_NODE)
            ++progress_bar_nodes.sw_found;
        else
            ++progress_bar_nodes.ca_found;
        if (progress_func)
            progress_func(&progress_bar_nodes,
                    p_handler->p_phy_diag->p_ibdiag->GetDiscoverProgressBarNodesPtr());

        if (!IsSupportAccRegFailed(p_curr_node->appData1.val,
                                   NOT_SUPPORT_SMP_ACCESS_REGISTER,
                                   GetNotSupportedBit())) {
            if (!p_handler->p_phy_diag->p_capability_module->IsSupportedSMPCapability(
                    p_curr_node, EnSMPCapIsAccessRegisterSupported)) {
                p_curr_node->appData1.val |= NOT_SUPPORT_SMP_ACCESS_REGISTER;
                FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                        new FabricErrNodeNotSupportCap(p_curr_node,
                                "This device does not support " \
                                "SMP access register MAD capability");
                if (!p_curr_fabric_node_err) {
                    p_handler->p_phy_diag->SetLastError("Failed to allocate " \
                                       "FabricErrNodeNotSupportCap");
                    IBDIAGNET_RETURN(IBDIAG_ERR_CODE_NO_MEM);
                }
                phy_errors.push_back(p_curr_fabric_node_err);
                rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
                continue;
            }
        } else
            //skip already reported node
            continue;

        direct_route_t *p_curr_direct_route = p_handler->p_phy_diag->p_ibdiag->
                GetDirectRouteByNodeGuid(p_curr_node->guid_get());
        if (!p_curr_direct_route) {
            p_handler->p_phy_diag->SetLastError("DB error - can't find direct route "
                                                "to node=%s (node guid: 0x%x)",
                                                p_curr_node->getName().c_str(),
                                                p_curr_node->guid_get());
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        clbck_data.m_data1 = p_curr_node;

        for (u_int32_t port_num = 1; port_num < (unsigned)p_curr_node->numPorts + 1; ++port_num) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)port_num);

            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            clbck_data.m_data3 = p_curr_port;

            for (u_int8_t idx = 0; idx < MAX_PCI_IDX; ++idx) {

                AccRegKeyDPN *p_dpnkey = new AccRegKeyDPN(
                        p_curr_node->guid_get(),
                        0,        // depth
                        idx,      // pci index
                        0);       // pci node

                clbck_data.m_data2 = p_dpnkey;

                struct SMP_AccessRegister mad_areg;
                CLEAR_STRUCT(mad_areg);
                PackData(p_dpnkey, &mad_areg);

                p_handler->p_phy_diag->SMPAccRegGetByDirect(p_curr_direct_route,
                                                            port_num,
                                                            &mad_areg,
                                                            &clbck_data);
            } // pci index loop
            break; // send only for one port per node
        } // port loop
    } // system loop

    IBDIAGNET_RETURN(rc);
}

int MPEINRegister::LoopPCIDepth(const clbck_data_t &clbck_data,
                                int rec_status,
                                void *p_attribute_data)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    AccRegHandler *p_handler = (AccRegHandler*)clbck_data.m_p_obj;
    IBNode *p_curr_node = (IBNode*)clbck_data.m_data1;
    AccRegKeyDPN *p_dpnkey_old = (AccRegKeyDPN *)clbck_data.m_data2;
    IBPort *p_curr_port = (IBPort*)clbck_data.m_data3;
    direct_route_t *p_curr_direct_route = p_handler->p_phy_diag->p_ibdiag->
                                          GetDirectRouteByNodeGuid(p_curr_node->guid_get());

    // invoke handler callback in order to handle return status and insert reply to the map
    rc = p_handler->SMPAccessRegisterHandlerGetClbck(clbck_data, rec_status, p_attribute_data);

    if (rc)
        IBDIAGNET_RETURN(rc);

    clbck_data_t clbck_data_new = clbck_data;
    clbck_data_new.m_handle_data_func = LoopPCINodeDelegator;

    SMP_AccessRegister *p_access_reg = (SMP_AccessRegister*)p_attribute_data;
    mpein_reg mpein;
    CLEAR_STRUCT(mpein);
    mpein_reg_unpack(&mpein, p_access_reg->reg.data);

    if (mpein.port_type == PCI_PORT_TYPE_US) {

        for (u_int8_t depth = 1; depth < MAX_PCI_DEPTH; ++depth) {

            AccRegKeyDPN *p_dpnkey = new AccRegKeyDPN(
                    p_curr_node->guid_get(),
                    depth,                   // depth
                    p_dpnkey_old->pci_idx,   // pci index
                    0);                      // pci node

            clbck_data_new.m_data2 = p_dpnkey;

            struct SMP_AccessRegister mad_areg;
            CLEAR_STRUCT(mad_areg);
            PackData(p_dpnkey, &mad_areg);

            p_handler->p_phy_diag->SMPAccRegGetByDirect(p_curr_direct_route,
                                                        p_curr_port->num,
                                                        &mad_areg,
                                                        &clbck_data_new);
        }
    }

    IBDIAGNET_RETURN(rc);
}

int MPEINRegister::LoopPCINode(const clbck_data_t &clbck_data,
                                int rec_status,
                                void *p_attribute_data)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    AccRegHandler *p_handler = (AccRegHandler*)clbck_data.m_p_obj;
    IBNode *p_curr_node = (IBNode*)clbck_data.m_data1;

    AccRegKeyDPN *p_dpnkey_old = (AccRegKeyDPN *)clbck_data.m_data2;
    IBPort *p_curr_port = (IBPort*)clbck_data.m_data3;
    direct_route_t *p_curr_direct_route = p_handler->p_phy_diag->p_ibdiag->
                                          GetDirectRouteByNodeGuid(p_curr_node->guid_get());

    rc = p_handler->SMPAccessRegisterHandlerGetClbck(clbck_data, rec_status, p_attribute_data);

    if (rc)
        IBDIAGNET_RETURN(rc);

    clbck_data_t clbck_data_new = clbck_data;
    clbck_data_new.m_handle_data_func = SMPAccessRegisterHandlerGetDelegator;

    SMP_AccessRegister *p_access_reg = (SMP_AccessRegister*)p_attribute_data;
    mpein_reg mpein;
    CLEAR_STRUCT(mpein);
    mpein_reg_unpack(&mpein, p_access_reg->reg.data);

    if (mpein.port_type == PCI_PORT_TYPE_DS) {

        for (u_int8_t node = 1; node < MAX_PCI_NODES; ++node) {

            AccRegKeyDPN *p_dpnkey = new AccRegKeyDPN(
                    p_curr_node->guid_get(),
                    p_dpnkey_old->depth,        // depth
                    p_dpnkey_old->pci_idx,      // pci index
                    node);                      // pci node

            clbck_data_new.m_data2 = p_dpnkey;

            struct SMP_AccessRegister mad_areg;
            CLEAR_STRUCT(mad_areg);
            PackData(p_dpnkey, &mad_areg);

            p_handler->p_phy_diag->SMPAccRegGetByDirect(p_curr_direct_route,
                                                        p_curr_port->num,
                                                        &mad_areg,
                                                        &clbck_data_new);
        }
    }

    IBDIAGNET_RETURN(rc);
}

bool MPEINRegister::IsValidPakcet(acc_reg_data reg_data)
{
    u_int8_t port_type = reg_data.mpein.port_type;

    if (port_type == PCI_PORT_TYPE_EP ||
        port_type == PCI_PORT_TYPE_US ||
        port_type == PCI_PORT_TYPE_DS)
        return true;

    return false;
}

//*****************************************
//***      MPIR Register functions      ***
//*****************************************
MPIRRegister::MPIRRegister(map_akey_areg &mpein_map) :
                                        Register(ACCESS_REGISTER_ID_MPIR,
                                                 (const unpack_data_func_t)mpir_reg_unpack,
                                                 ACC_REG_MPIR_INTERNAL_SECTION_NAME,
                                                 ACC_REG_MPIR_FIELDS_NUM,
                                                 NOT_SUPPORT_MPIR),
                                        mpein_map(mpein_map) // this register should be sent to
                                                             // all valid mpein registers.
{
}

void MPIRRegister::PackData(AccRegKey *p_key, SMP_AccessRegister *acc_reg)
{
    IBDIAGNET_ENTER;
    acc_reg->register_id = this->register_id;

    AccRegKeyDPN* p_dpnkey = (AccRegKeyDPN*)p_key;
    mpir_reg mpir;
    CLEAR_STRUCT(mpir);

    mpir.depth = p_dpnkey->depth;
    mpir.pcie_index = p_dpnkey->pci_idx;
    mpir.node = p_dpnkey->pci_node;

    mpir_reg_pack(&mpir, acc_reg->reg.data);
    IBDIAGNET_RETURN_VOID;
}

void MPIRRegister::DumpRegisterData(acc_reg_data areg, stringstream &sstream)
{
    IBDIAGNET_ENTER;
    mpir_reg *mpir = &areg.mpir;

    sstream << +mpir->host_buses << ','
            << +mpir->sdm << ','
            << +mpir->subordinate_bus << ','
            << +mpir->secondary_bus << ','
            << +mpir->device << ','
            << +mpir->bus << ','
            << +mpir->local_port << ','
            << +mpir->slot_number << ','
            << +mpir->num_con_devices << ','
            << +mpir->slot_cap << endl;

    IBDIAGNET_RETURN_VOID;
}

int MPIRRegister::BuildDB(AccRegHandler* p_handler,
                          list_p_fabric_general_err &phy_errors,
                          progress_func_nodes_t progress_func)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    clbck_data_t clbck_data;
    clbck_data.m_handle_data_func = SMPAccessRegisterHandlerGetDelegator;
    clbck_data.m_p_obj = p_handler;

    p_handler->SetHeader(PCI_DPN_HEADER);

    // go over all the mpein MADs and send mpir accordingly
    for (map_akey_areg::iterator mpein_iter = mpein_map.begin();
         mpein_iter != mpein_map.end(); mpein_iter++) {

        AccRegKeyDPN *p_dpnkey = (AccRegKeyDPN*)mpein_iter->first;

        IBNode *p_curr_node =
                p_handler->p_phy_diag->p_discovered_fabric->getNodeByGuid(p_dpnkey->node_guid);

        if (!p_curr_node) {
            p_handler->p_phy_diag->SetLastError("DB error - found null node in "
                                                "NodeByName map for key = " U64H_FMT,
                                                p_dpnkey->node_guid);
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        direct_route_t *p_curr_direct_route = p_handler->p_phy_diag->p_ibdiag->
                GetDirectRouteByNodeGuid(p_curr_node->guid_get());
        if (!p_curr_direct_route) {
            p_handler->p_phy_diag->SetLastError("DB error - can't find direct route "
                                                "to node=%s (node guid: 0x%x)",
                                                p_curr_node->getName().c_str(),
                                                p_curr_node->guid_get());
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        IBPort *p_curr_port = NULL;
        for (u_int32_t port_num = 1; port_num < (unsigned)p_curr_node->numPorts + 1; ++port_num) {
            p_curr_port = p_curr_node->getPort((phys_port_t)port_num);

            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;
            break; // use the first valid port.
        }

        if (!p_curr_port) {
            p_handler->p_phy_diag->SetLastError("DB error - failed to find valid port for node %s",
                                                p_curr_node->getName().c_str());
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        struct SMP_AccessRegister mad_areg;
        CLEAR_STRUCT(mad_areg);
        PackData(p_dpnkey, &mad_areg);
        AccRegKeyDPN *p_dpnkey_new = new AccRegKeyDPN(p_dpnkey->node_guid,
                                                      p_dpnkey->depth,
                                                      p_dpnkey->pci_idx,
                                                      p_dpnkey->pci_node);

        clbck_data.m_data1 = p_curr_node;
        clbck_data.m_data2 = p_dpnkey_new;

        p_handler->p_phy_diag->SMPAccRegGetByDirect(p_curr_direct_route,
                                                    p_curr_port->num,
                                                    &mad_areg,
                                                    &clbck_data);
    }

    IBDIAGNET_RETURN(rc);
}


//*****************************************
//***      MFCR Register functions      ***
//*****************************************
MFCRRegister::MFCRRegister() : Register(ACCESS_REGISTER_ID_MFCR,
                                        (const unpack_data_func_t)mfcr_reg_unpack,
                                        ACC_REG_MFCR_INTERNAL_SECTION_NAME,
                                        UNUSED_FIELDS_NUM,
                                        NOT_SUPPORT_MFCR,
                                        AVAILABLE_FANS,
                                        SW_ONLY,
                                        false)
{
}

void MFCRRegister::PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg)
{
    IBDIAGNET_ENTER;
    acc_reg->register_id = this->register_id;
    IBDIAGNET_RETURN_VOID;
}

void MFCRRegister::DumpRegisterData(struct acc_reg_data areg, stringstream& sstream)
{
    IBDIAGNET_ENTER;

    struct mfcr_reg *p_mfcr = &areg.mfcr;
    sstream << hex << p_mfcr->tacho_active << dec << endl;

    IBDIAGNET_RETURN_VOID;
}

int MFCRRegister::AvailableSensors(struct acc_reg_data& areg, list_uint8& sensors_list)
{
    IBDIAGNET_ENTER;
    struct mfcr_reg *p_mfcr = &areg.mfcr;
    int rc  = SensorsBitsToList(p_mfcr->tacho_active, MAX_FANS, sensors_list);
    IBDIAGNET_RETURN(rc);
}

//*****************************************
//***      MFSM Register functions      ***
//*****************************************
MFSMRegister::MFSMRegister() : Register(ACCESS_REGISTER_ID_MFSM,
                                        (const unpack_data_func_t)mfsm_reg_unpack,
                                        ACC_REG_MFSM_INTERNAL_SECTION_NAME,
                                        UNUSED_FIELDS_NUM,
                                        NOT_SUPPORT_MFSM,
                                        FANS_SPEED_HEADER,
                                        SW_ONLY)
{
}

void MFSMRegister::PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg)
{
    IBDIAGNET_ENTER;
    acc_reg->register_id = this->register_id;
    struct mfsm_reg mfsm;
    CLEAR_STRUCT(mfsm);
    mfsm.tacho = ((AccRegKeyNodeSensor *)p_key)->sensor_id;
    mfsm_reg_pack(&mfsm, acc_reg->reg.data);
    IBDIAGNET_RETURN_VOID;
}

void MFSMRegister::DumpRegisterData(struct acc_reg_data areg, stringstream& sstream)
{
    IBDIAGNET_ENTER;
    char buff[1024] = {0};
    struct mfsm_reg *p_mfsm = &areg.mfsm;

    sprintf(buff, U32D_FMT, p_mfsm->rpm);

    sstream << buff << endl;
    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//***      MFSL Register functions      ***
//*****************************************
MFSLRegister::MFSLRegister() : Register(ACCESS_REGISTER_ID_MFSL,
                                        (const unpack_data_func_t)mfsl_reg_unpack,
                                        ACC_REG_MFSL_INTERNAL_SECTION_NAME,
                                        UNUSED_FIELDS_NUM,
                                        NOT_SUPPORT_MFSL,
                                        FANS_THRESHOLDS_HEADER,
                                        SW_ONLY)
{
}

void MFSLRegister::PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg)
{
    IBDIAGNET_ENTER;

    acc_reg->register_id = this->register_id;
    struct mfsl_reg mfsl;
    CLEAR_STRUCT(mfsl);
    mfsl.tacho = ((AccRegKeyNodeSensor *)p_key)->sensor_id;
    mfsl_reg_pack(&mfsl, acc_reg->reg.data);

    IBDIAGNET_RETURN_VOID;
}

void MFSLRegister::DumpRegisterData(struct acc_reg_data areg, stringstream& sstream)
{
    IBDIAGNET_ENTER;
    char buff[1024] = {0};
    struct mfsl_reg *p_mfsl = &areg.mfsl;

    sprintf(buff, U32D_FMT "," U32D_FMT, p_mfsl->tach_min, p_mfsl->tach_max);

    sstream << buff << endl;
    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//***      FORE Register functions      ***
//*****************************************
FORERegister::FORERegister() : Register(ACCESS_REGISTER_ID_FORE,
                                        (const unpack_data_func_t)fore_reg_unpack,
                                        ACC_REG_FORE_INTERNAL_SECTION_NAME,
                                        UNUSED_FIELDS_NUM,
                                        NOT_SUPPORT_FORE,
                                        FANS_ALERT_HEADER,
                                        SW_ONLY)
{
}

void FORERegister::PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg)
{
    IBDIAGNET_ENTER;
    acc_reg->register_id = this->register_id;
    IBDIAGNET_RETURN_VOID;
}

void FORERegister::DumpRegisterData(struct acc_reg_data areg, stringstream& sstream)
{
    IBDIAGNET_ENTER;
    struct fore_reg *p_fore = &areg.fore;
    stringstream under_limit;
    stringstream over_limit;
    u_int16_t mask = 1;

    if (!p_fore->fan_under_limit && !p_fore->fan_over_limit) {
        sstream << "-1,-1" << endl;
        IBDIAGNET_RETURN_VOID;
    }

    for (int i = 0; i < MAX_FANS; ++i, mask <<= 1) {
        if (mask & p_fore->fan_under_limit)
            under_limit << i << "|";
        if (mask & p_fore->fan_over_limit)
            over_limit << i << "|";
    }

    string under;
    string over;

    if (!p_fore->fan_under_limit)
        under = "-1";
    else {
        under = under_limit.str();
        under = under.substr(0, under.size() - 1);

    }

    if (!p_fore->fan_over_limit)
        over = "-1";
    else {
        over = over_limit.str();
        over = over.substr(0, over.size() - 1);

    }

    sstream << under << "," << over << endl;
    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//***     MFNR Register functions       ***
//*****************************************
MFNRRegister::MFNRRegister() : Register(ACCESS_REGISTER_ID_MFNR,
                                        (const unpack_data_func_t)mfnr_reg_unpack,
                                        ACC_REG_MFNR_INTERNAL_SECTION_NAME,
                                        UNUSED_FIELDS_NUM,
                                        NOT_SUPPORT_MFNR,
                                        FANS_NUMBERS_HEADER,
                                        SW_ONLY)
{
}

void MFNRRegister::PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg)
{
    IBDIAGNET_ENTER;

    acc_reg->register_id = this->register_id;
    struct mfnr_reg mfnr;
    CLEAR_STRUCT(mfnr);
    mfnr.fan_index = ((AccRegKeyNodeSensor *)p_key)->sensor_id;
    mfnr_reg_pack(&mfnr, acc_reg->reg.data);

    IBDIAGNET_RETURN_VOID;
}

void MFNRRegister::DumpRegisterData(struct acc_reg_data areg, stringstream& sstream)
{
    IBDIAGNET_ENTER;
    struct mfnr_reg *p_mfnr = &areg.mfnr;
    sstream << p_mfnr->serial_number << endl;
    IBDIAGNET_RETURN_VOID;
}

//******************************************
//***      MTCAP Register functions      ***
//******************************************
MTCAPRegister::MTCAPRegister() : Register(ACCESS_REGISTER_ID_MTCAP,
                                        (const unpack_data_func_t)mtcap_reg_unpack,
                                        ACC_REG_MTCAP_INTERNAL_SECTION_NAME,
                                        UNUSED_FIELDS_NUM,
                                        NOT_SUPPORT_MTCAP,
                                        TEMPERATURE_SENSORS_COUNT,
                                        SW_ONLY,
                                        false)
{
}

void MTCAPRegister::PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg)
{
    IBDIAGNET_ENTER;
    acc_reg->register_id = this->register_id;
    IBDIAGNET_RETURN_VOID;
}

void MTCAPRegister::DumpRegisterData(struct acc_reg_data areg, stringstream& sstream)
{
    IBDIAGNET_ENTER;

    char buff[32] = {0};
    struct mtcap_reg *p_mtcap = &areg.mtcap;
    sprintf(buff, U32D_FMT, p_mtcap->sensor_count);
    sstream << buff << endl;

    IBDIAGNET_RETURN_VOID;
}

int MTCAPRegister::AvailableSensors(struct acc_reg_data& areg, list_uint8& sensors_list)
{
    IBDIAGNET_ENTER;
    struct mtcap_reg *p_mtcap = &areg.mtcap;
    IBDIAGNET_RETURN(SensorsCountToList(p_mtcap->sensor_count, sensors_list));
}

//*****************************************
//***      MTMP Register functions      ***
//*****************************************
MTMPRegister::MTMPRegister() : Register(ACCESS_REGISTER_ID_MTMP,
                                        (const unpack_data_func_t)mtmp_reg_unpack,
                                        ACC_REG_MTMP_INTERNAL_SECTION_NAME,
                                        UNUSED_FIELDS_NUM,
                                        NOT_SUPPORT_MTMP,
                                        TEMPERATURE_SENSORS_HEADER,
                                        SW_ONLY)
{
}

void MTMPRegister::PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg)
{
    IBDIAGNET_ENTER;

    acc_reg->register_id = this->register_id;
    struct mtmp_reg mtmp;
    CLEAR_STRUCT(mtmp);
    mtmp.sensor_index = ((AccRegKeyNodeSensor *)p_key)->sensor_id;
    mtmp_reg_pack(&mtmp, acc_reg->reg.data);

    IBDIAGNET_RETURN_VOID;
}

void MTMPRegister::DumpRegisterData(struct acc_reg_data areg, stringstream& sstream)
{
    IBDIAGNET_ENTER;
    char buff[1024] = {0};
    struct mtmp_reg *p_mtmp = &areg.mtmp;

    int16_t temperature = p_mtmp->temperature;
    int16_t max_temperature = p_mtmp->max_temperature;
    int16_t threshold_lo = p_mtmp->temperature_threshold_lo;
    int16_t threshold_hi = p_mtmp->temperature_threshold_hi;

    sprintf(buff, FLT_FMT "," FLT_FMT "," FLT_FMT "," FLT_FMT,
            (float)temperature / 8, (float)max_temperature / 8,
            (float)threshold_lo / 8, (float)threshold_hi / 8);

    sstream << p_mtmp->sensor_name_hi << p_mtmp->sensor_name_lo << "," << buff << endl;

    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//***      MTWE Register functions      ***
//*****************************************
MTWERegister::MTWERegister() : Register(ACCESS_REGISTER_ID_MTWE,
                                        (const unpack_data_func_t)mtwe_reg_unpack,
                                        ACC_REG_MTWE_INTERNAL_SECTION_NAME,
                                        UNUSED_FIELDS_NUM,
                                        NOT_SUPPORT_MTWE,
                                        TEMPERATURE_SENSORS_ALERT_HEADER,
                                        SW_ONLY)
{
}

void MTWERegister::PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg)
{
    IBDIAGNET_ENTER;
    acc_reg->register_id = this->register_id;
    IBDIAGNET_RETURN_VOID;
}

void MTWERegister::DumpRegisterData(struct acc_reg_data areg, stringstream& sstream)
{
    IBDIAGNET_ENTER;
    struct mtwe_reg *p_mtwe = &areg.mtwe;
    stringstream sensor_warning;
    const int num_of_warnings = 4;
    const int bits_per_warning = 32;
    u_int32_t mask = 1;
    u_int32_t sensor_warnings[num_of_warnings];
    string warning;

    if (!p_mtwe->sensor_warning0 && !p_mtwe->sensor_warning1 &&
        !p_mtwe->sensor_warning2 && !p_mtwe->sensor_warning3 ) {
        sstream << "-1" << endl;
        IBDIAGNET_RETURN_VOID;
    }

    sensor_warnings[0] = p_mtwe->sensor_warning0;
    sensor_warnings[1] = p_mtwe->sensor_warning1;
    sensor_warnings[2] = p_mtwe->sensor_warning2;
    sensor_warnings[3] = p_mtwe->sensor_warning3;

    for (int i = 0; i < num_of_warnings; ++i) {
        for (int j = 0; j < bits_per_warning; ++j, mask <<= 1) {
            if (mask & sensor_warnings[i])
                sensor_warning << (i * bits_per_warning + j) << "|";
        }
        mask = 1;
    }

    warning = sensor_warning.str();
    warning = warning.substr(0, warning.size() - 1);
    sstream << warning << endl;
    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//***     MVCAP Register functions     ***
//*****************************************
MVCAPRegister::MVCAPRegister() : Register(ACCESS_REGISTER_ID_MVCAP,
                                          (const unpack_data_func_t)mvcap_reg_unpack,
                                          ACC_REG_MVCAP_INTERNAL_SECTION_NAME,
                                          UNUSED_FIELDS_NUM,
                                          NOT_SUPPORT_MVCAP,
                                          AVAILABLE_POWER_SENSORS,
                                          SW_ONLY,
                                          false)
{
}

void MVCAPRegister::PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg)
{
    IBDIAGNET_ENTER;
    acc_reg->register_id = this->register_id;
    IBDIAGNET_RETURN_VOID;
}

void MVCAPRegister::DumpRegisterData(struct acc_reg_data areg, stringstream& sstream)
{
    IBDIAGNET_ENTER;

    struct mvcap_reg *p_mvcap = &areg.mvcap;
    sstream << hex << p_mvcap->sensor_map << dec << endl;

    IBDIAGNET_RETURN_VOID;
}

int MVCAPRegister::AvailableSensors(struct acc_reg_data& areg, list_uint8& sensors_list)
{
    IBDIAGNET_ENTER;
    struct mvcap_reg *p_mvcap = &areg.mvcap;
    int rc = SensorsBitsToList(p_mvcap->sensor_map, MAX_VOLTAGE_CURRENT_SENSORS, sensors_list);
    IBDIAGNET_RETURN(rc);
}

//*****************************************
//***      MVCR Register functions      ***
//*****************************************
MVCRRegister::MVCRRegister() : Register(ACCESS_REGISTER_ID_MVCR,
                                        (const unpack_data_func_t)mvcr_reg_unpack,
                                        ACC_REG_MVCR_INTERNAL_SECTION_NAME,
                                        UNUSED_FIELDS_NUM,
                                        NOT_SUPPORT_MVCR,
                                        POWER_SENSORS_HEADER,
                                        SW_ONLY)
{
}

void MVCRRegister::PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg)
{
    IBDIAGNET_ENTER;

    acc_reg->register_id = this->register_id;
    struct mvcr_reg mvcr;
    CLEAR_STRUCT(mvcr);
    mvcr.sensor_index = ((AccRegKeyNodeSensor *)p_key)->sensor_id;
    mvcr_reg_pack(&mvcr, acc_reg->reg.data);

    IBDIAGNET_RETURN_VOID;
}

void MVCRRegister::DumpRegisterData(struct acc_reg_data areg, stringstream& sstream)
{
    IBDIAGNET_ENTER;
    char buff[128] = {0};
    char name[9] = {0};
    struct mvcr_reg *p_mvcr = &areg.mvcr;

    memcpy(name, &p_mvcr->sensor_name, sizeof(u_int64_t));

    // todo change current from 'N/A' to real value when supported.
    sprintf(buff, FLT_FMT "," "N/A",
            p_mvcr->voltage_sensor_value / 100.0);

    sstream << name << "," << buff << endl;
    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//***      MSPS Register functions      ***
//*****************************************
MSPSRegister::MSPSRegister() : Register(ACCESS_REGISTER_ID_MSPS,
                                        (const unpack_data_func_t)msps_reg_unpack,
                                        ACC_REG_MSPS_INTERNAL_SECTION_NAME,
                                        UNUSED_FIELDS_NUM,
                                        NOT_SUPPORT_MSPS,
                                        POWER_SUPPLIES_HEADER,
                                        SW_ONLY,
                                        true,
                                        VIA_GMP)
{
}

static string DumpPSU(struct psu_status *psu)
{
    string psu_status;

    switch(psu->present) {
        case 0 :
            psu_status += ",N/A,N/A";
            break;
        case 1 :
            psu_status += ",Yes,No";
            break;
        case 2 :
            psu_status += ",No,No";
            break;
        case 5 :
            psu_status += ",Yes,Yes";
            break;
        case 6 :
            psu_status += ",No,Yes";
            break;
        default:
            psu_status += ",,";
    }

    switch(psu->ac_status) {
        case 0 :
            psu_status += ",N/A";
            break;
        case 1 :
            psu_status += ",Yes";
            break;
        case 2 :
            psu_status += ",No";
            break;
        default:
            psu_status += ",";
    }

    switch(psu->dc_status) {
        case 0 :
            psu_status += ",N/A";
            break;
        case 1 :
            psu_status += ",OK";
            break;
        case 2 :
            psu_status += ",Error";
            break;
        default:
            psu_status += ",";
    }

    switch(psu->alert) {
        case 0 :
            psu_status += ",N/A";
            break;
        case 1 :
            psu_status += ",No";
            break;
        case 2 :
            psu_status += ",Yes";
            break;
        default:
            psu_status += ",";
    }

    switch(psu->fan) {
        case 0 :
            psu_status += ",N/A";
            break;
        case 1 :
            psu_status += ",No Fan";
            break;
        case 2 :
            psu_status += ",OK";
            break;
        case 3 :
            psu_status += ",Error";
            break;
        default:
            psu_status += ",";
    }

    switch(psu->temp) {
        case 0 :
            psu_status += ",N/A";
            break;
        case 1 :
            psu_status += ",OK";
            break;
        case 2 :
            psu_status += ",Overtemp";
            break;
        default:
            psu_status += ",";
    }

    if (psu->snp) {
        psu_status += ',';
        psu_status += psu->serial_number;
    } else
        psu_status += ",N/A";

    return psu_status;
}

void MSPSRegister::DumpRegisterData(struct acc_reg_data areg, stringstream& sstream)
{
    IBDIAGNET_ENTER;

    struct msps_reg *p_msps = &areg.msps;
    string node_guid = sstream.str(); // get the node guid from the key

    sstream << "1" << DumpPSU(&p_msps->psu0) << endl
            << node_guid << "2" << DumpPSU(&p_msps->psu1) << endl;

    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//***      MSGI Register functions      ***
//*****************************************
MSGIRegister::MSGIRegister() : Register(ACCESS_REGISTER_ID_MSGI,
                                        (const unpack_data_func_t)msgi_reg_unpack,
                                        ACC_REG_MSGI_INTERNAL_SECTION_NAME,
                                        UNUSED_FIELDS_NUM,
                                        NOT_SUPPORT_MSGI,
                                        SYSTEM_GENERAL_INFORMATION_HEADER,
                                        SW_ONLY,
                                        true,
                                        VIA_GMP)
{
}

void MSGIRegister::DumpRegisterData(struct acc_reg_data areg, stringstream& sstream)
{
    IBDIAGNET_ENTER;

    struct msgi_reg *p_msgi = &areg.msgi;

    sstream << p_msgi->serial_number << ','
            << p_msgi->part_number << ','
            << p_msgi->revision << ','
            << '\"' << p_msgi->product_name << '\"' << endl;

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//***      SLRIP Register functions     ***
//*****************************************
SLRIPRegister::SLRIPRegister() : Register(ACCESS_REGISTER_ID_SLRIP,
                                         (const unpack_data_func_t)slrip_reg_unpack,
                                         ACC_REG_SLRIP_INTERNAL_SECTION_NAME,
                                         ACC_REG_SLRIP_FIELDS_NUM,
                                         NOT_SUPPORT_SLRIP)
{
}

void SLRIPRegister::PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg)
{
    IBDIAGNET_ENTER;
    acc_reg->register_id = this->register_id;

    struct slrip_reg slrip;
    CLEAR_STRUCT(slrip);
    slrip.local_port = ((AccRegKeyPortLane *)p_key)->port_num;
    slrip.lane = ((AccRegKeyPortLane *)p_key)->lane;
    slrip.pnat = ACC_REG_PNAT_IB_PORT;
    slrip.ib_sel = ((AccRegKeyPortLane *)p_key)->idx_in_lane;
    slrip_reg_pack(&slrip, acc_reg->reg.data);

    IBDIAGNET_RETURN_VOID;
}

void SLRIPRegister::DumpRegisterData(struct acc_reg_data areg, stringstream& sstream)
{
    IBDIAGNET_ENTER;

    struct slrip_reg *p_slrip = &areg.slrip;

    sstream << "ib_sel_" << +p_slrip->ib_sel << ','
            << +p_slrip->lane << ','
            << +p_slrip->pnat << ','
            << +p_slrip->local_port << ','
            << +p_slrip->version << ','
            << +p_slrip->status << ','

            << +p_slrip->ffe_tap0 << ','
            << +p_slrip->ffe_tap1 << ','
            << +p_slrip->ffe_tap2 << ','
            << +p_slrip->ffe_tap3 << ','
            << +p_slrip->ffe_tap4 << ','
            << +p_slrip->ffe_tap5 << ','
            << +p_slrip->ffe_tap6 << ','
            << +p_slrip->ffe_tap7 << ','
            << +p_slrip->ffe_tap8 << ','

            << p_slrip->ffe_tap_en<< ','
            << +p_slrip->sel_enc0 << ','
            << +p_slrip->sel_enc1 << ','
            << +p_slrip->sel_enc2 << ','
            << +p_slrip->mixer_offset0 << ','
            << +p_slrip->mixer_offset1 << ','
            << +p_slrip->saved0_sel_enc0 << ','
            << +p_slrip->saved0_mixer_offset0 << ','
            << +p_slrip->saved1_sel_enc0 << ','
            << +p_slrip->saved1_mixer_offset0 << endl;

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//***      SLREG Register functions     ***
//*****************************************
SLREGRegister::SLREGRegister() : Register(ACCESS_REGISTER_ID_SLREG,
                                         (const unpack_data_func_t)slreg_reg_unpack,
                                         ACC_REG_SLREG_INTERNAL_SECTION_NAME,
                                         ACC_REG_SLREG_FIELDS_NUM,
                                         NOT_SUPPORT_SLREG)
{
}

void SLREGRegister::PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg)
{
    IBDIAGNET_ENTER;
    acc_reg->register_id = this->register_id;

    struct slreg_reg slreg;
    CLEAR_STRUCT(slreg);
    slreg.local_port = ((AccRegKeyPortLane *)p_key)->port_num;
    slreg.lane = ((AccRegKeyPortLane *)p_key)->lane;
    slreg.pnat = ACC_REG_PNAT_IB_PORT;
    slreg.eidx = ((AccRegKeyPortLane *)p_key)->idx_in_lane;
    slreg_reg_pack(&slreg, acc_reg->reg.data);

    IBDIAGNET_RETURN_VOID;
}

void SLREGRegister::DumpRegisterData(struct acc_reg_data areg, stringstream& sstream)
{
    IBDIAGNET_ENTER;

    struct slreg_reg *p_slreg = &areg.slreg;

    sstream << "eidx_" << +p_slreg->eidx << ','
            << +p_slreg->lane << ','
            << +p_slreg->pnat << ','
            << +p_slreg->local_port << ','
            << +p_slreg->version << ','
            << +p_slreg->status << ','
            << +p_slreg->eye_grade << ","
            << +p_slreg->initial_error << ','
            << +p_slreg->post_error << ','
            << +p_slreg->initial_area << ','
            << +p_slreg->post_area << ','
            << +p_slreg->post_area_after_centering << ','
            << +p_slreg->gd_iter_num_error << ','
            << +p_slreg->gd_iter_num_area << ','
            << +p_slreg->initial_sel_enc_slave << ','
            << +p_slreg->post_sel_enc_slave << ','
            << +p_slreg->post_sel_enc_slave_centering << ','
            << +p_slreg->rect_size << ','
            << +p_slreg->cur_phase_ib << ','
            << +p_slreg->ffe_gain << ','
            << +p_slreg->init_all_gains_bad_value << ','
            << +p_slreg->update_single_gain_bad_value << ','
            << +p_slreg->monotonic_flow_height_inconsistent << ','
            << +p_slreg->monotonic_reached_high_limit << ','
            << +p_slreg->monotonic_reached_low_limit << ','
            << +p_slreg->area_zero_event_counter << ','
            << +p_slreg->sel_enc_tune_area_failure << endl;
    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//***      SLCCT Register functions     ***
//*****************************************
SLCCTRegister::SLCCTRegister() : Register(ACCESS_REGISTER_ID_SLCCT,
                                         (const unpack_data_func_t)slcct_reg_unpack,
                                         ACC_REG_SLCCT_INTERNAL_SECTION_NAME,
                                         ACC_REG_SLCCT_FIELDS_NUM,
                                         NOT_SUPPORT_SLCCT)
{
}

void SLCCTRegister::PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg)
{
    IBDIAGNET_ENTER;
    acc_reg->register_id = this->register_id;

    struct slcct_reg slcct;
    CLEAR_STRUCT(slcct);
    slcct.local_port = ((AccRegKeyPortLane *)p_key)->port_num;
    slcct.lane = ((AccRegKeyPortLane *)p_key)->lane;
    slcct.pnat = ACC_REG_PNAT_IB_PORT;
    slcct.conf_index = ((AccRegKeyPortLane *)p_key)->idx_in_lane? SLCCT_CTLE_ARR_SIZE : 0;
    slcct.num_of_idx = SLCCT_CTLE_ARR_SIZE;

    slcct_reg_pack(&slcct, acc_reg->reg.data);

    IBDIAGNET_RETURN_VOID;
}

void SLCCTRegister::DumpRegisterData(struct acc_reg_data areg, stringstream& sstream)
{
    IBDIAGNET_ENTER;

    struct slcct_reg *p_slcct = &areg.slcct;

    sstream << +p_slcct->pnat << ','
            << +p_slcct->version << ','
            << +p_slcct->status << ','
            << +p_slcct->conf_index << ','
            << +p_slcct->num_of_idx;

    for (int i = 0; i < p_slcct->num_of_idx; ++i) {
        sstream << ',' << +p_slcct->ctle_calib_set[i].bgn_offset_n0
                << ',' << +p_slcct->ctle_calib_set[i].bgn_offset_p0
                << ',' << +p_slcct->ctle_calib_set[i].bgn_offset_n1
                << ',' << +p_slcct->ctle_calib_set[i].bgn_offset_p1;
    }

    sstream << endl;

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//***      PPAMP Register functions     ***
//*****************************************
PPAMPRegister::PPAMPRegister() : Register(ACCESS_REGISTER_ID_PPAMP,
                                         (const unpack_data_func_t)ppamp_reg_unpack,
                                         ACC_REG_PPAMP_INTERNAL_SECTION_NAME,
                                         ACC_REG_PPAMP_FIELDS_NUM,
                                         NOT_SUPPORT_PPAMP)
{
}

void PPAMPRegister::PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg)
{
    IBDIAGNET_ENTER;

    acc_reg->register_id = this->register_id;

    struct ppamp_reg ppamp;
    CLEAR_STRUCT(ppamp);
    ppamp.opamp_group_type = ((AccRegKeyTypeGroup *)p_key)->group_type;
    ppamp.opamp_group = ((AccRegKeyTypeGroup *)p_key)->group_num;
    ppamp.start_index = ((AccRegKeyTypeGroup *)p_key)->start_index;
    ppamp.num_of_indices = ((AccRegKeyTypeGroup *)p_key)->num_of_indices;
    ppamp_reg_pack(&ppamp, acc_reg->reg.data);

    IBDIAGNET_RETURN_VOID;
}

void PPAMPRegister::DumpRegisterData(struct acc_reg_data areg, stringstream& sstream)
{
    IBDIAGNET_ENTER;

    struct ppamp_reg *p_ppamp = &areg.ppamp;

    sstream << +p_ppamp->max_opamp_group << ','
            << p_ppamp->max_num_of_indices;

    for (int i = 0; i < PPAMP_IDX_DATA_SIZE; ++i) {
        sstream << ',' << p_ppamp->index_data[i];
    }

    sstream << endl;

    IBDIAGNET_RETURN_VOID;
}

void PPAMPRegister::GetGroupData(struct acc_reg_data& areg, u_int8_t& group, u_int16_t& indices)
{
    IBDIAGNET_ENTER;

    group = areg.ppamp.max_opamp_group;
    indices = areg.ppamp.max_num_of_indices;

    IBDIAGNET_RETURN_VOID;
}


//**************************************************
//***   Access Register Handler Implementation   ***
//**************************************************
AccRegHandler::AccRegHandler(Register *p_register, class PhyDiag *p_pd,
                             string hh) : phy_errors(&p_pd->phy_errors),
                                          can_send_mads_by_lid(false),
                                          clbck_error_state(IBDIAG_SUCCESS_CODE),
                                          p_reg(p_register),
                                          handler_header(hh),
                                          data_map(keycomp),
                                          p_phy_diag(p_pd)
{
}

AccRegHandler::~AccRegHandler()
{
    IBDIAGNET_ENTER;

    if (p_reg)
        delete p_reg;

    IBDIAGNET_RETURN_VOID;
}

void SMPAccessRegisterHandlerGetDelegator(const clbck_data_t &clbck_data,
                                          int rec_status,
                                          void *p_attribute_data)
{
    reinterpret_cast<AccRegHandler*>(clbck_data.m_p_obj)->
        SMPAccessRegisterHandlerGetClbck(clbck_data, rec_status,
                                         p_attribute_data);
}

int AccRegHandler::SMPAccessRegisterHandlerGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAGNET_ENTER;
     if (clbck_error_state)
         IBDIAGNET_RETURN(1);
    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    IBNode *p_node = (IBNode *)clbck_data.m_data1;

    if (rec_status &&
        IsSupportAccRegFailed(p_node->appData1.val,
                              NOT_SUPPORT_SMP_ACCESS_REGISTER,
                              this->p_reg->GetNotSupportedBit()))
        IBDIAGNET_RETURN(1);

    if (rec_status == IBIS_MAD_STATUS_UNSUP_METHOD_ATTR) { //not supported for this device
        p_node->appData1.val |= NOT_SUPPORT_SMP_ACCESS_REGISTER;
        FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
            new FabricErrNodeNotSupportCap(p_node,
                                           "The firmware of this device " \
                                           "does not support " \
                                           "access register capability");
        if (!p_curr_fabric_node_err) {
            this->p_phy_diag->SetLastError("Failed to allocate" \
                                           " FabricErrNodeNotSupportCap");
            clbck_error_state = IBDIAG_ERR_CODE_NO_MEM;
        } else
            phy_errors->push_back(p_curr_fabric_node_err);
        IBDIAGNET_RETURN(1);
    } else if (rec_status) { //failed to get response
        p_node->appData1.val |= NOT_SUPPORT_SMP_ACCESS_REGISTER;
        FabricErrNodeNotRespond *p_curr_fabric_node_err =
            new FabricErrNodeNotRespond(p_node, "SMPAccessRegister");
        if (!p_curr_fabric_node_err) {
            this->p_phy_diag->SetLastError("Failed to allocate" \
                                           " FabricErrPortNotRespond");
            clbck_error_state = IBDIAG_ERR_CODE_NO_MEM;
        } else
            phy_errors->push_back(p_curr_fabric_node_err);
        IBDIAGNET_RETURN(1);
    }

    struct SMP_AccessRegister *p_access_reg =
        (struct SMP_AccessRegister *)p_attribute_data;

    //already reported that this node doesn't support
    if (p_access_reg->status &&
        IsSupportAccRegFailed(p_node->appData1.val,
                              NOT_SUPPORT_SMP_ACCESS_REGISTER,
                              this->p_reg->GetNotSupportedBit()))
        IBDIAGNET_RETURN(1);

    //check if the access register status, is it supported, etc.
    if (p_access_reg->status) {
        p_node->appData1.val |= this->p_reg->GetNotSupportedBit();
        FabricErrGeneral *p_curr_phy_err =
            new FabricNodeErrPhyRetrieveGeneral(p_node ,p_access_reg->status);
        if (!p_curr_phy_err) {
            this->p_phy_diag->SetLastError("Failed to allocate" \
                                           " FabricErrGeneral");
            clbck_error_state = IBDIAG_ERR_CODE_NO_MEM;
        } else
            phy_errors->push_back(p_curr_phy_err);
        IBDIAGNET_RETURN(1);
    }

    struct acc_reg_data register_unpack_data;
    CLEAR_STRUCT(register_unpack_data);
    this->p_reg->unpack_data_func(register_unpack_data.data, p_access_reg->reg.data);
    AccRegKey *p_key = (AccRegKey *)clbck_data.m_data2;

    if (!p_reg->IsValidPakcet(register_unpack_data)) {
        IBDIAGNET_RETURN(1);
    }

    // insert the data into a map
    std::pair<std::map<AccRegKey *, struct acc_reg_data>::iterator,bool> ret;
    ret = this->data_map.insert(pair <AccRegKey *,
                                struct acc_reg_data>(p_key,
                                                     register_unpack_data));

    if ((!ret.second) || clbck_error_state) {
        this->p_phy_diag->SetLastError("Failed to add %s data for node=%s," \
                " err=%s",
                (CHECK_NAME_RETRIEVE(this->
                                     GetPReg()->
                                     GetSectionName())).c_str(),
                p_node->getName().c_str(), this->p_phy_diag->GetLastError());

        IBDIAGNET_RETURN(1);
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void GMPAccessRegisterHandlerGetDelegator(const clbck_data_t &clbck_data,
                                          int rec_status,
                                          void *p_attribute_data)
{
    reinterpret_cast<AccRegHandler*>(clbck_data.m_p_obj)->
        GMPAccessRegisterHandlerGetClbck(clbck_data, rec_status, p_attribute_data);
}

void AccRegHandler::GMPAccessRegisterHandlerGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAGNET_ENTER;

    if (clbck_error_state)
         return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here
    IBNode *p_node = (IBNode *)clbck_data.m_data1;

    if (rec_status && IsSupportAccRegFailed(p_node->appData1.val,
                                            NOT_SUPPORT_GMP_ACCESS_REGISTER,
                                            this->p_reg->GetNotSupportedBit()))
        IBDIAGNET_RETURN_VOID;

    if (rec_status == IBIS_MAD_STATUS_UNSUP_METHOD_ATTR) { // not supported for this device
        p_node->appData1.val |= NOT_SUPPORT_GMP_ACCESS_REGISTER;

        FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
            new FabricErrNodeNotSupportCap(p_node,
                                           "The firmware of this device does not support " \
                                           "GMP access register capability");

        if (!p_curr_fabric_node_err) {
            this->p_phy_diag->SetLastError("Failed to allocate FabricErrNodeNotSupportCap");
            clbck_error_state = IBDIAG_ERR_CODE_NO_MEM;
        } else
            phy_errors->push_back(p_curr_fabric_node_err);

        IBDIAGNET_RETURN_VOID;

    } else if (rec_status == MAD_STATUS_UNSUP_REGISTER_GMP) {
        p_node->appData1.val |= this->p_reg->GetNotSupportedBit();

        char buff[256];
        sprintf(buff,
                "The firmware of this device does not support register ID: 0x%x",
                this->p_reg->GetRegisterID());

        FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
            new FabricErrNodeNotSupportCap(p_node, buff);

        if (!p_curr_fabric_node_err) {
            this->p_phy_diag->SetLastError("Failed to allocate FabricErrNodeNotSupportCap");
            clbck_error_state = IBDIAG_ERR_CODE_NO_MEM;
        } else
            phy_errors->push_back(p_curr_fabric_node_err);

        IBDIAGNET_RETURN_VOID;
    } else if (rec_status) { // failed to get response
        p_node->appData1.val |= NOT_SUPPORT_GMP_ACCESS_REGISTER;

        FabricErrNodeNotRespond *p_curr_fabric_node_err =
            new FabricErrNodeNotRespond(p_node, "GMPAccessRegister");

        if (!p_curr_fabric_node_err) {
            this->p_phy_diag->SetLastError("Failed to allocate" \
                                           " FabricErrPortNotRespond");
            clbck_error_state = IBDIAG_ERR_CODE_NO_MEM;
        } else
            phy_errors->push_back(p_curr_fabric_node_err);

        IBDIAGNET_RETURN_VOID;
    }

    struct GMP_AccessRegister *p_access_reg = (struct GMP_AccessRegister *)p_attribute_data;

    struct acc_reg_data register_unpack_data;
    CLEAR_STRUCT(register_unpack_data);
    this->p_reg->unpack_data_func(register_unpack_data.data, p_access_reg->reg.data);
    AccRegKey *p_key = (AccRegKey *)clbck_data.m_data2;

    // insert the data into the map
    pair<map<AccRegKey*, struct acc_reg_data>::iterator,bool> ret;
    ret = data_map.insert(pair <AccRegKey*, struct acc_reg_data>(p_key, register_unpack_data));

    if ((!ret.second) || clbck_error_state) {
        this->p_phy_diag->SetLastError("Failed to add %s data for node=%s, err=%s",
                    (CHECK_NAME_RETRIEVE(GetPReg()->GetSectionName())).c_str(),
                    p_node->getName().c_str(),
                    this->p_phy_diag->GetLastError());
    }

    IBDIAGNET_RETURN_VOID;
}

int AccRegNodeHandler::BuildDB(list_p_fabric_general_err &phy_errors,
                               progress_func_nodes_t progress_func)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->p_phy_diag->p_ibdiag->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    progress_bar_nodes_t progress_bar_nodes;
    CLEAR_STRUCT(progress_bar_nodes);

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = this;

    for (map_str_pnode::iterator nI =
            this->p_phy_diag->p_discovered_fabric->NodeByName.begin();
         nI != this->p_phy_diag->p_discovered_fabric->NodeByName.end();
         ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->p_phy_diag->SetLastError("DB error - found null node in"  \
                                           " NodeByName map for key = %s",
                                           (*nI).first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_reg->IsRegSupportNodeType(p_curr_node->type))
            continue;

        ++progress_bar_nodes.nodes_found;
        if (p_curr_node->type == IB_SW_NODE)
            ++progress_bar_nodes.sw_found;
        else
            ++progress_bar_nodes.ca_found;
        if (progress_func)
            progress_func(&progress_bar_nodes,
                          this->p_phy_diag->p_ibdiag->
                                GetDiscoverProgressBarNodesPtr());

        if (p_reg->GetRegisterType() == VIA_SMP) {

            if (!IsSupportAccRegFailed(p_curr_node->appData1.val,
                                       NOT_SUPPORT_SMP_ACCESS_REGISTER,
                                       p_reg->GetNotSupportedBit())) {
                if (!this->p_phy_diag->p_capability_module->IsSupportedSMPCapability(
                        p_curr_node, EnSMPCapIsAccessRegisterSupported))
                {
                    p_curr_node->appData1.val |= NOT_SUPPORT_SMP_ACCESS_REGISTER;
                    FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                            new FabricErrNodeNotSupportCap(p_curr_node,
                                    "This device does not support " \
                                    "SMP access register MAD capability");
                    if (!p_curr_fabric_node_err) {
                        this->p_phy_diag->SetLastError("Failed to allocate " \
                                           "FabricErrNodeNotSupportCap");
                        rc = IBDIAG_ERR_CODE_NO_MEM;
                        goto exit;
                    }
                    phy_errors.push_back(p_curr_fabric_node_err);
                    rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
                    continue;
                }
            } else
                //skip already reported node
                continue;

            clbck_data.m_data1 = p_curr_node;
            AccRegKeyNode *p_nkey = new AccRegKeyNode(p_curr_node->guid_get());
            clbck_data.m_data2 = p_nkey;
            clbck_data.m_handle_data_func = SMPAccessRegisterHandlerGetDelegator;

            direct_route_t *p_curr_direct_route = NULL;
            //Get direct route for by node guid, so we won't have 2 MADs
            //on the same device at the same time.
            p_curr_direct_route = this->p_phy_diag->p_ibdiag->GetDirectRouteByNodeGuid(
                p_curr_node->guid_get());
            if (!p_curr_direct_route) {
                this->p_phy_diag->SetLastError("DB error - can't find direct route " \
                                   "to node=%s (node guid: 0x%x)",
                                   p_curr_node->getName().c_str(),
                                   p_curr_node->guid_get());
                rc = IBDIAG_ERR_CODE_DB_ERR;
                goto exit;
            }

            struct SMP_AccessRegister mad_areg;
            CLEAR_STRUCT(mad_areg);

            p_reg->PackData(p_nkey, &mad_areg);
            this->p_phy_diag->SMPAccRegGetByDirect(p_curr_direct_route,
                    0, //port 0
                    &mad_areg,
                    &clbck_data);
        }

        if (p_reg->GetRegisterType() == VIA_GMP) {

            if (!IsSupportAccRegFailed(p_curr_node->appData1.val,
                                       NOT_SUPPORT_GMP_ACCESS_REGISTER,
                                       p_reg->GetNotSupportedBit())) {
                if (!this->p_phy_diag->p_capability_module->IsSupportedGMPCapability(
                        p_curr_node, EnGMPCapIsAccessRegisterSupported))
                {
                    p_curr_node->appData1.val |= NOT_SUPPORT_GMP_ACCESS_REGISTER;
                    FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                            new FabricErrNodeNotSupportCap(p_curr_node,
                                    "This device does not support " \
                                    "GMP access register MAD capability");
                    if (!p_curr_fabric_node_err) {
                        this->p_phy_diag->SetLastError("Failed to allocate " \
                                           "FabricErrNodeNotSupportCap");
                        rc = IBDIAG_ERR_CODE_NO_MEM;
                        goto exit;
                    }
                    phy_errors.push_back(p_curr_fabric_node_err);
                    rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
                    continue;
                }
            } else
                //skip already reported node
                continue;

            clbck_data.m_data1 = p_curr_node;
            AccRegKeyNode *p_nkey = new AccRegKeyNode(p_curr_node->guid_get());
            clbck_data.m_data2 = p_nkey;
            clbck_data.m_handle_data_func = GMPAccessRegisterHandlerGetDelegator;

            uint16_t lid;
            uint8_t lmc; // unused. just for calling the next function.
            p_curr_node->getLidAndLMC(0, lid, lmc);

            struct GMP_AccessRegister mad_areg;
            CLEAR_STRUCT(mad_areg);

            this->p_phy_diag->GMPAccRegGet(lid,
                    p_reg->GetRegisterID(),
                    &mad_areg,
                    &clbck_data);
        }

        if (clbck_error_state)
            goto exit;
    }

exit:
    this->p_phy_diag->p_ibis_obj->MadRecAll();
    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAGNET_RETURN(rc);
}

int AccRegPortHandler::BuildDB(list_p_fabric_general_err &phy_errors,
                               progress_func_nodes_t progress_func)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->p_phy_diag->p_ibdiag->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    progress_bar_nodes_t progress_bar_nodes;
    CLEAR_STRUCT(progress_bar_nodes);

    clbck_data_t clbck_data;
    clbck_data.m_handle_data_func = SMPAccessRegisterHandlerGetDelegator;
    clbck_data.m_p_obj = this;

    for (map_str_pnode::iterator nI =
            this->p_phy_diag->p_discovered_fabric->NodeByName.begin();
         nI != this->p_phy_diag->p_discovered_fabric->NodeByName.end();
         ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->p_phy_diag->SetLastError("DB error - found null node in"  \
                                           " NodeByName map for key = %s",
                                           (*nI).first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_reg->IsRegSupportNodeType(p_curr_node->type))
            continue;

        ++progress_bar_nodes.nodes_found;
        if (p_curr_node->type == IB_SW_NODE)
            ++progress_bar_nodes.sw_found;
        else
            ++progress_bar_nodes.ca_found;
        if (progress_func)
            progress_func(&progress_bar_nodes,
                          this->p_phy_diag->p_ibdiag->GetDiscoverProgressBarNodesPtr());

        if (!IsSupportAccRegFailed(p_curr_node->appData1.val,
                                   NOT_SUPPORT_SMP_ACCESS_REGISTER,
                                   p_reg->GetNotSupportedBit())) {
            if (!this->p_phy_diag->p_capability_module->IsSupportedSMPCapability(
                    p_curr_node, EnSMPCapIsAccessRegisterSupported))
            {
                p_curr_node->appData1.val |= NOT_SUPPORT_SMP_ACCESS_REGISTER;
                FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                        new FabricErrNodeNotSupportCap(p_curr_node,
                                "This device does not support " \
                                "SMP access register MAD capability");
                if (!p_curr_fabric_node_err) {
                    this->p_phy_diag->SetLastError("Failed to allocate " \
                                       "FabricErrNodeNotSupportCap");
                    rc = IBDIAG_ERR_CODE_NO_MEM;
                    goto exit;
                }
                phy_errors.push_back(p_curr_fabric_node_err);
                rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
                continue;
            }
        } else
            //skip already reported node
            continue;

        direct_route_t *p_curr_direct_route = NULL;
        //Get direct route for by node guid, so we won't have 2 MADs
        //on the same device at the same time.
        p_curr_direct_route = this->p_phy_diag->p_ibdiag->GetDirectRouteByNodeGuid(
            p_curr_node->guid_get());
        if (!p_curr_direct_route) {
            this->p_phy_diag->SetLastError("DB error - can't find direct route " \
                               "to node=%s (node guid: 0x%x)",
                               p_curr_node->getName().c_str(),
                               p_curr_node->guid_get());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        // go over all ports of this node:
        for (u_int32_t i = 1; i < (unsigned)p_curr_node->numPorts + 1; ++i) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            //port doesn't exist or down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            clbck_data.m_data1 = p_curr_node;

            AccRegKeyPort *p_plkey = new AccRegKeyPort(
                        p_curr_node->guid_get(),
                        p_curr_port->guid_get(),
                        i);

            clbck_data.m_data2 = p_plkey;
            struct SMP_AccessRegister mad_areg;
            CLEAR_STRUCT(mad_areg);
            mad_areg.register_id = p_reg->GetRegisterID();

            p_reg->PackData(p_plkey, &mad_areg);
            this->p_phy_diag->SMPAccRegGetByDirect(p_curr_direct_route,
                    i,
                    &mad_areg,
                    &clbck_data);

            if (clbck_error_state)
                goto exit;
        }
    }
exit:
    this->p_phy_diag->p_ibis_obj->MadRecAll();
    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAGNET_RETURN(rc);
}

int AccRegGroupHandler::BuildDB(list_p_fabric_general_err &phy_errors,
                                   progress_func_nodes_t progress_func)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    map_pnode_to_max_group_num_t pnode_to_max_group_num;

    if (!this->p_phy_diag->p_ibdiag->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    progress_bar_nodes_t progress_bar_nodes;
    CLEAR_STRUCT(progress_bar_nodes);

    clbck_data_t clbck_data;
    clbck_data.m_handle_data_func = SMPAccessRegisterHandlerGetDelegator;
    clbck_data.m_p_obj = this;

    for (map_str_pnode::iterator nI =
            this->p_phy_diag->p_discovered_fabric->NodeByName.begin();
         nI != this->p_phy_diag->p_discovered_fabric->NodeByName.end();
         ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->p_phy_diag->SetLastError("DB error - found null node in"  \
                                           " NodeByName map for key = %s",
                                           (*nI).first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_reg->IsRegSupportNodeType(p_curr_node->type))
            continue;

        ++progress_bar_nodes.nodes_found;
        if (p_curr_node->type == IB_SW_NODE)
            ++progress_bar_nodes.sw_found;
        else
            ++progress_bar_nodes.ca_found;
        if (progress_func)
            progress_func(&progress_bar_nodes,
                          this->p_phy_diag->p_ibdiag->GetDiscoverProgressBarNodesPtr());

        if (!IsSupportAccRegFailed(p_curr_node->appData1.val,
                                   NOT_SUPPORT_SMP_ACCESS_REGISTER,
                                   p_reg->GetNotSupportedBit())) {
            if (!this->p_phy_diag->p_capability_module->IsSupportedSMPCapability(
                    p_curr_node, EnSMPCapIsAccessRegisterSupported))
            {
                p_curr_node->appData1.val |= NOT_SUPPORT_SMP_ACCESS_REGISTER;
                FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                        new FabricErrNodeNotSupportCap(p_curr_node,
                                "This device does not support " \
                                "SMP access register MAD capability");
                if (!p_curr_fabric_node_err) {
                    this->p_phy_diag->SetLastError("Failed to allocate " \
                                       "FabricErrNodeNotSupportCap");
                    rc = IBDIAG_ERR_CODE_NO_MEM;
                    goto exit;
                }
                phy_errors.push_back(p_curr_fabric_node_err);
                rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
                continue;
            }
        } else
            //skip already reported node
            continue;

        direct_route_t *p_curr_direct_route = NULL;
        //Get direct route for by node guid, so we won't have 2 MADs
        //on the same device at the same time.
        p_curr_direct_route =
            this->p_phy_diag->p_ibdiag->GetDirectRouteByNodeGuid(
                   p_curr_node->guid_get());
        if (!p_curr_direct_route) {
            this->p_phy_diag->SetLastError("DB error - can't find direct route " \
                               "to node=%s (node guid: 0x%x)",
                               p_curr_node->getName().c_str(),
                               p_curr_node->guid_get());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        clbck_data.m_data1 = p_curr_node;
        //send group 0, to discover how many groups there are.
        u_int32_t gr = 0;
        AccRegKeyGroup *p_plkey = new AccRegKeyGroup(
                p_curr_node->guid_get(),
                gr);

        clbck_data.m_data2 = p_plkey;
        struct SMP_AccessRegister mad_areg;
        CLEAR_STRUCT(mad_areg);

        p_reg->PackData(p_plkey, &mad_areg);
        this->p_phy_diag->SMPAccRegGetByDirect(p_curr_direct_route,
                0,//port 0
                &mad_areg,
                &clbck_data);

        if (clbck_error_state)
            goto exit;
    }
    // need to get all the mads for group 0
    this->p_phy_diag->p_ibis_obj->MadRecAll();

    //go over all the nodes and send for the rest of the group other then 0
    for (map_akey_areg::iterator nI = this->data_map.begin();
            nI != this->data_map.end(); ++nI) {
        AccRegKey * p_aregkey = (*nI).first;
        if (!p_aregkey)
            this->p_phy_diag->SetLastError("DB error - found null key in data_map");
        map_guid_pnode::iterator ngI = this->p_phy_diag->p_discovered_fabric->NodeByGuid.find(((AccRegKeyGroup *)p_aregkey)->node_guid);
        if (ngI == this->p_phy_diag->p_discovered_fabric->NodeByGuid.end())
            goto exit;
        IBNode *p_curr_node = (*ngI).second;
        if (!p_curr_node) {
            this->p_phy_diag->SetLastError("DB error - found null node in"  \
                                           " NodeByGuid map for key = 0x%x",
                                           ((AccRegKeyGroup *)p_aregkey)->node_guid);
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        struct acc_reg_data areg = (*nI).second;
        struct ppll_reg *p_ppll = (struct ppll_reg *)areg.data;
        pnode_to_max_group_num.insert(pair < IBNode *, u_int8_t >(p_curr_node, p_ppll->num_plls));
    }

    for (map_pnode_to_max_group_num_t::iterator nI = pnode_to_max_group_num.begin();
            nI != pnode_to_max_group_num.end(); ++nI) {
        IBNode *p_curr_node = (*nI).first;
        direct_route_t *p_curr_direct_route = NULL;
        //Get direct route for by node guid, so we won't have 2 MADs
        //on the same device at the same time.
        p_curr_direct_route =
            this->p_phy_diag->p_ibdiag->GetDirectRouteByNodeGuid(
                   p_curr_node->guid_get());
        if (!p_curr_direct_route) {
            this->p_phy_diag->SetLastError("DB error - can't find direct route " \
                               "to node=%s (node guid: 0x%x)",
                               p_curr_node->getName().c_str(),
                               p_curr_node->guid_get());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        clbck_data.m_data1 = p_curr_node;
        u_int8_t max_gr = (*nI).second;
        //starting from group 1, we already sent group 0
        for (u_int8_t gr = 1; gr <= max_gr; gr++) {
            AccRegKeyGroup *p_plkey = new AccRegKeyGroup(
                    p_curr_node->guid_get(),
                    gr);

            clbck_data.m_data2 = p_plkey;
            struct SMP_AccessRegister mad_areg;
            CLEAR_STRUCT(mad_areg);
            p_reg->PackData(p_plkey, &mad_areg);
            this->p_phy_diag->SMPAccRegGetByDirect(p_curr_direct_route,
                    0,//port 0
                    &mad_areg,
                    &clbck_data);

            if (clbck_error_state)
                goto exit;
        }
    }

exit:
    this->p_phy_diag->p_ibis_obj->MadRecAll();
    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAGNET_RETURN(rc);
}

int AccRegLaneHandler::BuildDB(list_p_fabric_general_err &phy_errors,
                               progress_func_nodes_t progress_func)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->p_phy_diag->p_ibdiag->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    progress_bar_nodes_t progress_bar_nodes;
    CLEAR_STRUCT(progress_bar_nodes);

    clbck_data_t clbck_data;
    clbck_data.m_handle_data_func = SMPAccessRegisterHandlerGetDelegator;
    clbck_data.m_p_obj = this;

    for (map_str_pnode::iterator nI =
            this->p_phy_diag->p_discovered_fabric->NodeByName.begin();
         nI != this->p_phy_diag->p_discovered_fabric->NodeByName.end();
         ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->p_phy_diag->SetLastError("DB error - found null node in"  \
                                           " NodeByName map for key = %s",
                                           (*nI).first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_reg->IsRegSupportNodeType(p_curr_node->type))
            continue;

        ++progress_bar_nodes.nodes_found;
        if (p_curr_node->type == IB_SW_NODE)
            ++progress_bar_nodes.sw_found;
        else
            ++progress_bar_nodes.ca_found;
        if (progress_func)
            progress_func(&progress_bar_nodes,
                          this->p_phy_diag->p_ibdiag->GetDiscoverProgressBarNodesPtr());

        if (!IsSupportAccRegFailed(p_curr_node->appData1.val,
                                   NOT_SUPPORT_SMP_ACCESS_REGISTER,
                                   p_reg->GetNotSupportedBit())) {
            if (!this->p_phy_diag->p_capability_module->IsSupportedSMPCapability(
                    p_curr_node, EnSMPCapIsAccessRegisterSupported))
            {
                p_curr_node->appData1.val |= NOT_SUPPORT_SMP_ACCESS_REGISTER;
                FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                        new FabricErrNodeNotSupportCap(p_curr_node,
                                "This device does not support " \
                                "SMP access register MAD capability");
                if (!p_curr_fabric_node_err) {
                    this->p_phy_diag->SetLastError("Failed to allocate " \
                                       "FabricErrNodeNotSupportCap");
                    rc = IBDIAG_ERR_CODE_NO_MEM;
                    goto exit;
                }
                phy_errors.push_back(p_curr_fabric_node_err);
                rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
                continue;
            }
        } else
            //skip already reported node
            continue;

        direct_route_t *p_curr_direct_route = NULL;
        //Get direct route for by node guid, so we won't have 2 MADs
        //on the same device at the same time.
        p_curr_direct_route = this->p_phy_diag->p_ibdiag->GetDirectRouteByNodeGuid(
            p_curr_node->guid_get());
        if (!p_curr_direct_route) {
            this->p_phy_diag->SetLastError("DB error - can't find direct route " \
                               "to node=%s (node guid: 0x%x)",
                               p_curr_node->getName().c_str(),
                               p_curr_node->guid_get());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        // go over all ports of this node:
        for (u_int32_t i = 1; i < (unsigned)p_curr_node->numPorts + 1; ++i) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            //port doesn't exist or down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            clbck_data.m_data1 = p_curr_node;

            for (u_int32_t ln = 0; ln < this->m_lane_num; ln++) {
                AccRegKeyPortLane *p_plkey = new AccRegKeyPortLane(
                        p_curr_node->guid_get(),
                        p_curr_port->guid_get(),
                        i,
                        ln);

                clbck_data.m_data2 = p_plkey;
                struct SMP_AccessRegister mad_areg;
                CLEAR_STRUCT(mad_areg.reg.data);
                mad_areg.register_id = p_reg->GetRegisterID();

                p_reg->PackData(p_plkey, &mad_areg);
                this->p_phy_diag->SMPAccRegGetByDirect(p_curr_direct_route,
                        i,
                        &mad_areg,
                        &clbck_data);

                if (clbck_error_state)
                    goto exit;
            }
            break;
        }
    }
exit:
    this->p_phy_diag->p_ibis_obj->MadRecAll();
    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAGNET_RETURN(rc);
}

int AccRegPortLaneHandler::BuildDB(list_p_fabric_general_err &phy_errors,
                                   progress_func_nodes_t progress_func)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->p_phy_diag->p_ibdiag->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    progress_bar_nodes_t progress_bar_nodes;
    CLEAR_STRUCT(progress_bar_nodes);

    clbck_data_t clbck_data;
    clbck_data.m_handle_data_func = SMPAccessRegisterHandlerGetDelegator;
    clbck_data.m_p_obj = this;

    for (map_str_pnode::iterator nI =
            this->p_phy_diag->p_discovered_fabric->NodeByName.begin();
         nI != this->p_phy_diag->p_discovered_fabric->NodeByName.end();
         ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->p_phy_diag->SetLastError("DB error - found null node in"  \
                                           " NodeByName map for key = %s",
                                           (*nI).first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_reg->IsRegSupportNodeType(p_curr_node->type))
            continue;

        ++progress_bar_nodes.nodes_found;
        if (p_curr_node->type == IB_SW_NODE)
            ++progress_bar_nodes.sw_found;
        else
            ++progress_bar_nodes.ca_found;
        if (progress_func)
            progress_func(&progress_bar_nodes,
                          this->p_phy_diag->p_ibdiag->GetDiscoverProgressBarNodesPtr());

        if (!IsSupportAccRegFailed(p_curr_node->appData1.val,
                                   NOT_SUPPORT_SMP_ACCESS_REGISTER,
                                   p_reg->GetNotSupportedBit())) {
            if (!this->p_phy_diag->p_capability_module->IsSupportedSMPCapability(
                    p_curr_node, EnSMPCapIsAccessRegisterSupported))
            {
                p_curr_node->appData1.val |= NOT_SUPPORT_SMP_ACCESS_REGISTER;
                FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                        new FabricErrNodeNotSupportCap(p_curr_node,
                                "This device does not support " \
                                "SMP access register MAD capability");
                if (!p_curr_fabric_node_err) {
                    this->p_phy_diag->SetLastError("Failed to allocate " \
                                       "FabricErrNodeNotSupportCap");
                    rc = IBDIAG_ERR_CODE_NO_MEM;
                    goto exit;
                }
                phy_errors.push_back(p_curr_fabric_node_err);
                rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
                continue;
            }
        } else
            //skip already reported node
            continue;

        direct_route_t *p_curr_direct_route = NULL;
        //Get direct route for by node guid, so we won't have 2 MADs
        //on the same device at the same time.
        p_curr_direct_route = this->p_phy_diag->p_ibdiag->GetDirectRouteByNodeGuid(
            p_curr_node->guid_get());
        if (!p_curr_direct_route) {
            this->p_phy_diag->SetLastError("DB error - can't find direct route " \
                               "to node=%s (node guid: 0x%x)",
                               p_curr_node->getName().c_str(),
                               p_curr_node->guid_get());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        // go over all ports of this node:
        for (u_int32_t i = 1; i < (unsigned)p_curr_node->numPorts + 1; ++i) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            //port doesn't exist or down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            clbck_data.m_data1 = p_curr_node;

            for (u_int32_t ln = 0; ln < LANE_NUM; ln++) {
                for (u_int8_t lane_idx = 0; lane_idx < max_idx_in_lane; lane_idx++) {

                    struct SMP_AccessRegister mad_areg;
                    CLEAR_STRUCT(mad_areg);

                    AccRegKeyPortLane *p_plkey = new AccRegKeyPortLane(
                            p_curr_node->guid_get(),
                            p_curr_port->guid_get(),
                            i,
                            ln,
                            lane_idx);

                    clbck_data.m_data2 = p_plkey;
                    mad_areg.register_id = p_reg->GetRegisterID();
                    p_reg->PackData(p_plkey, &mad_areg);

                    this->p_phy_diag->SMPAccRegGetByDirect(p_curr_direct_route,
                            i,
                            &mad_areg,
                            &clbck_data);

                    if (clbck_error_state)
                        goto exit;
                }
            }
        }
    }

exit:
    this->p_phy_diag->p_ibis_obj->MadRecAll();
    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAGNET_RETURN(rc);
}

int AccRegNodeSensorHandler::BuildDB(list_p_fabric_general_err &phy_errors,
                                     progress_func_nodes_t progress_func)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->p_phy_diag->p_ibdiag->IsDiscoveryDone())
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    progress_bar_nodes_t progress_bar_nodes;
    CLEAR_STRUCT(progress_bar_nodes);

    clbck_data_t clbck_data;
    clbck_data.m_handle_data_func = SMPAccessRegisterHandlerGetDelegator;
    clbck_data.m_p_obj = this;

    for (map_akey_areg::iterator areg_iter = p_acc_reg_handler->data_map.begin();
            areg_iter != p_acc_reg_handler->data_map.end(); ++areg_iter) {

        AccRegKeyNode *p_nkey = (AccRegKeyNode*)areg_iter->first;
        IBNode *p_curr_node = p_phy_diag->p_discovered_fabric->NodeByGuid[p_nkey->node_guid];

        if (!p_curr_node) {
            this->p_phy_diag->SetLastError("DB error - found null node in" \
                                           " NodeByGuid map for key = " U64H_FMT,
                                           p_nkey->node_guid);
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_reg->IsRegSupportNodeType(p_curr_node->type))
            continue;

        ++progress_bar_nodes.nodes_found;
        if (p_curr_node->type == IB_SW_NODE)
            ++progress_bar_nodes.sw_found;
        else
            ++progress_bar_nodes.ca_found;
        if (progress_func)
            progress_func(&progress_bar_nodes,
                          this->p_phy_diag->p_ibdiag->GetDiscoverProgressBarNodesPtr());

        if (!IsSupportAccRegFailed(p_curr_node->appData1.val,
                                   NOT_SUPPORT_SMP_ACCESS_REGISTER,
                                   p_reg->GetNotSupportedBit())) {
            if (!this->p_phy_diag->p_capability_module->IsSupportedSMPCapability(
                    p_curr_node, EnSMPCapIsAccessRegisterSupported))
            {
                p_curr_node->appData1.val |= NOT_SUPPORT_SMP_ACCESS_REGISTER;
                FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                        new FabricErrNodeNotSupportCap(p_curr_node,
                                "This device does not support " \
                                "SMP access register MAD capability");
                if (!p_curr_fabric_node_err) {
                    this->p_phy_diag->SetLastError("Failed to allocate " \
                                       "FabricErrNodeNotSupportCap");
                    rc = IBDIAG_ERR_CODE_NO_MEM;
                    goto exit;
                }
                phy_errors.push_back(p_curr_fabric_node_err);
                rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
                continue;
            }
        } else
            //skip already reported node
            continue;

        direct_route_t *p_curr_direct_route = NULL;
        //Get direct route for by node guid, so we won't have 2 MADs
        //on the same device at the same time.
        p_curr_direct_route = this->p_phy_diag->p_ibdiag->GetDirectRouteByNodeGuid(
            p_curr_node->guid_get());
        if (!p_curr_direct_route) {
            this->p_phy_diag->SetLastError("DB error - can't find direct route " \
                               "to node=%s (node guid: 0x%x)",
                               p_curr_node->getName().c_str(),
                               p_curr_node->guid_get());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        clbck_data.m_data1 = p_curr_node;

        list_uint8 sensors;
        rc = p_acc_reg_handler->GetPReg()->AvailableSensors(areg_iter->second, sensors);

        if (rc)
            goto exit;

        // go over all sensors of this node:
        for (list_uint8::iterator sensors_iter = sensors.begin();
                sensors_iter != sensors.end(); ++sensors_iter) {

            AccRegKeyNodeSensor *p_ns_key =
                    new AccRegKeyNodeSensor(p_nkey->node_guid, *sensors_iter);
            clbck_data.m_data2 = p_ns_key;
            struct SMP_AccessRegister mad_areg;
            CLEAR_STRUCT(mad_areg);

            p_reg->PackData(p_ns_key, &mad_areg);
            this->p_phy_diag->SMPAccRegGetByDirect(p_curr_direct_route,
                                                   0, //port 0
                                                   &mad_areg,
                                                   &clbck_data);
            if (clbck_error_state)
                goto exit;
        }
    }

exit:
    this->p_phy_diag->p_ibis_obj->MadRecAll();
    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    IBDIAGNET_RETURN(rc);
}


int AccRegTypeGroupHandler::BuildDB(list_p_fabric_general_err &phy_errors,
                                     progress_func_nodes_t progress_func)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->p_phy_diag->p_ibdiag->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    progress_bar_nodes_t progress_bar_nodes;
    CLEAR_STRUCT(progress_bar_nodes);

    clbck_data_t clbck_data;
    clbck_data.m_handle_data_func = SMPAccessRegisterHandlerGetDelegator;
    clbck_data.m_p_obj = this;

    for (map_str_pnode::iterator nI = p_phy_diag->p_discovered_fabric->NodeByName.begin();
         nI != p_phy_diag->p_discovered_fabric->NodeByName.end(); ++nI) {

        IBNode *p_curr_node = nI->second;
        if (!p_curr_node) {
            this->p_phy_diag->SetLastError("DB error - found null node in"  \
                                           " NodeByName map for key = %s",
                                           nI->first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        ++progress_bar_nodes.nodes_found;
        if (p_curr_node->type == IB_SW_NODE)
            ++progress_bar_nodes.sw_found;
        else
            ++progress_bar_nodes.ca_found;
        if (progress_func)
            progress_func(&progress_bar_nodes,
                          this->p_phy_diag->p_ibdiag->GetDiscoverProgressBarNodesPtr());

        if (!IsSupportAccRegFailed(p_curr_node->appData1.val,
                                   NOT_SUPPORT_SMP_ACCESS_REGISTER,
                                   p_reg->GetNotSupportedBit())) {
            if (!this->p_phy_diag->p_capability_module->IsSupportedSMPCapability(
                    p_curr_node, EnSMPCapIsAccessRegisterSupported)) {

                p_curr_node->appData1.val |= NOT_SUPPORT_SMP_ACCESS_REGISTER;
                FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                        new FabricErrNodeNotSupportCap(p_curr_node,
                                "This device does not support " \
                                "SMP access register MAD capability");

                if (!p_curr_fabric_node_err) {
                    this->p_phy_diag->SetLastError("Failed to allocate " \
                                       "FabricErrNodeNotSupportCap");
                    rc = IBDIAG_ERR_CODE_NO_MEM;
                    goto exit;
                }

                phy_errors.push_back(p_curr_fabric_node_err);
                rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
                continue;
            }
        } else
            //skip already reported node
            continue;

        direct_route_t *p_curr_direct_route = NULL;
        p_curr_direct_route = p_phy_diag->p_ibdiag->
                GetDirectRouteByNodeGuid(p_curr_node->guid_get());
        if (!p_curr_direct_route) {
            this->p_phy_diag->SetLastError("DB error - can't find direct route " \
                                           "to node=%s (node guid: 0x%x)",
                                           p_curr_node->getName().c_str(),
                                           p_curr_node->guid_get());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        clbck_data.m_data1 = p_curr_node;

        for (int tp = 0; tp < max_type; tp++) {
            //send group 0, to discover how many groups there are.
            //Skip PCIe Group
            if(tp == 1)
                continue;
            AccRegKeyTypeGroup *p_gpkey = new AccRegKeyTypeGroup(
                    p_curr_node->guid_get(), tp, 0, 0, max_indices);

            clbck_data.m_data2 = p_gpkey;
            struct SMP_AccessRegister mad_areg;
            CLEAR_STRUCT(mad_areg.reg.data);

            p_reg->PackData(p_gpkey, &mad_areg);
            this->p_phy_diag->SMPAccRegGetByDirect(p_curr_direct_route,
                    0,//port 0
                    &mad_areg,
                    &clbck_data);

            if (clbck_error_state)
                goto exit;
        }
    }

    // need to get all the mads for group 0
    this->p_phy_diag->p_ibis_obj->MadRecAll();

    //go over all the nodes and send for the rest of the group other then 0
    for (map_akey_areg::iterator aregI = data_map.begin(); aregI != data_map.end(); ++aregI) {

        AccRegKeyTypeGroup* p_aregkey = (AccRegKeyTypeGroup*)aregI->first;
        if (!p_aregkey) {
            this->p_phy_diag->SetLastError("DB error - found null key in data_map");
            goto exit;
        }

        // make sure that we are on group 0
        if (p_aregkey->group_num != 0)
            continue;

        map_guid_pnode::iterator ngI = p_phy_diag->p_discovered_fabric->
            NodeByGuid.find(p_aregkey->node_guid);
        if (ngI == this->p_phy_diag->p_discovered_fabric->NodeByGuid.end()) {
            this->p_phy_diag->SetLastError("DB error - GUID " U64H_FMT
                                           "not found in NodeByGuid map",
                                           p_aregkey->node_guid);
            goto exit;
        }

        IBNode *p_curr_node = ngI->second;
        if (!p_curr_node) {
            this->p_phy_diag->SetLastError("DB error - found null node in"  \
                                           " NodeByGuid map for key = 0x%x",
                                           p_aregkey->node_guid);
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        direct_route_t *p_curr_direct_route = p_phy_diag->p_ibdiag->
                GetDirectRouteByNodeGuid(p_curr_node->guid_get());

        if (!p_curr_direct_route) {
            this->p_phy_diag->SetLastError("DB error - can't find direct route " \
                                           "to node=%s (node guid: 0x%x)",
                                           p_curr_node->getName().c_str(),
                                           p_curr_node->guid_get());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        clbck_data.m_data1 = p_curr_node;
        u_int8_t opamp_gr_type = p_aregkey->group_type;

        struct acc_reg_data areg = aregI->second;
        u_int8_t max_opamp_gr;
        u_int16_t max_num_idx;

        this->p_reg->GetGroupData(areg, max_opamp_gr, max_num_idx);

        //start from group 1, we already sent group 0
        for (u_int8_t gr = 1; gr <= max_opamp_gr; gr++) {
            for (u_int16_t start_idx = 0; start_idx < max_num_idx; start_idx += max_indices) {

                u_int16_t idx_to_access =(((max_num_idx - start_idx) < max_indices)?
                        (max_num_idx % max_indices) : max_indices);

                AccRegKeyTypeGroup *p_gpkey = new AccRegKeyTypeGroup(
                        p_curr_node->guid_get(),
                        opamp_gr_type,
                        gr,
                        start_idx,
                        idx_to_access);

                clbck_data.m_data2 = p_gpkey;
                struct SMP_AccessRegister mad_areg;
                CLEAR_STRUCT(mad_areg.reg.data);
                p_reg->PackData(p_gpkey, &mad_areg);

                this->p_phy_diag->SMPAccRegGetByDirect(p_curr_direct_route,
                        0,//port 0
                        &mad_areg,
                        &clbck_data);

                if (clbck_error_state)
                    goto exit;
            }
        }
    }

exit:
    this->p_phy_diag->p_ibis_obj->MadRecAll();
    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAGNET_RETURN(rc);
}


int AccRegSpecificHandler::BuildDB(list_p_fabric_general_err &phy_errors,
                                   progress_func_nodes_t progress_func)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->p_phy_diag->p_ibdiag->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    rc = this->p_reg->BuildDB(this, phy_errors, progress_func);

    this->p_phy_diag->p_ibis_obj->MadRecAll();
    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAGNET_RETURN(rc);
}


void AccRegHandler::DumpCSV(CSVOut &csv_out)
{
    IBDIAGNET_ENTER;
    stringstream sstream;

    csv_out.DumpStart(p_reg->GetSectionName().c_str());

    sstream << handler_header;
    p_reg->DumpRegisterHeader(sstream);
    sstream << endl;
    csv_out.WriteLine(sstream.str());

    for (map_akey_areg::iterator nI = this->data_map.begin();
         nI != this->data_map.end(); ++nI) {
        AccRegKey * p_aregkey = (*nI).first;
        if (!p_aregkey)
            this->p_phy_diag->SetLastError("DB error - found null key in data_map");

        sstream.str("");
        p_aregkey->DumpKeyData(sstream);
        struct acc_reg_data areg = (*nI).second;
        p_reg->DumpRegisterData(areg, sstream);
        csv_out.WriteBuf(sstream.str());
    }
    csv_out.DumpEnd(p_reg->GetSectionName().c_str());
    IBDIAGNET_RETURN_VOID;
}
