/*
 * Copyright (c) 2004-2019 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#ifndef _ACC_REGISTER_H_
#define _ACC_REGISTER_H_

#include "acc_reg_key.h"
#include "phy_diag_types.h"

struct acc_reg_data {
    //This union should help access the data without pointer casting
    //and this fixes the strict aliasing rule error/warning.
    union {
        uint8_t data[160];
        struct slrg_reg slrg;
        struct slrp_reg slrp;
        struct sltp_reg sltp;
        struct ptas_reg ptas;
        struct ppll_reg ppll;
        struct mpein_reg mpein;
        struct mpir_reg mpir;

        // fans registers
        struct mfcr_reg mfcr;
        struct mfsm_reg mfsm;
        struct mfsl_reg mfsl;
        struct fore_reg fore;
        struct mfnr_reg mfnr;
        // temperature registers
        struct mtcap_reg mtcap;
        struct mtmp_reg mtmp;
        struct mtwe_reg mtwe;
        // current/voltage register
        struct mvcap_reg mvcap;
        struct mvcr_reg mvcr;
        // power supplies register
        struct msps_reg msps;
        // general information register
        struct msgi_reg msgi;

        struct slrip_reg slrip;
        struct slreg_reg slreg;
        struct slcct_reg slcct;
        struct ppamp_reg ppamp;
    };
};

typedef map < AccRegKey *, acc_reg_data, bool (*)(AccRegKey *, AccRegKey *)> map_akey_areg;
typedef map < IBNode *, u_int8_t > map_pnode_to_max_group_num_t;
class AccRegHandler;

bool keycomp(AccRegKey *k1, AccRegKey *k2);

void SMPAccessRegisterHandlerGetDelegator(const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data);

//Register class defines common members to all registers.
//Each register needs to have its dedicated class, that inherits Register
//and defines its own PackData and DumpRegisterData.
class Register {

protected:
    uint32_t    register_id;
    uint32_t    fields_num;
    //That is a bit to set in case register isn't supported
    //and to avoid sending the register more than once to the same node.
    uint64_t    not_supported_bit;
    string      section_name;
    string      header;
    PhyPluginSupportedNodesType m_support_nodes;
    bool        to_dump;
    AccRegVia_t acc_reg_type;

public:
    Register(uint32_t reg_id,
             const unpack_data_func_t p_unpack_func,
             string sn,
             uint32_t fn,
             uint64_t nsb,
             string header = "",
             PhyPluginSupportedNodesType support_nodes = SUPPORT_ALL,
             bool to_dump = true,
             AccRegVia_t acc_reg_type = VIA_SMP);

    virtual ~Register(){}

    unpack_data_func_t unpack_data_func;
    virtual void PackData(AccRegKey *p_key,
                          struct SMP_AccessRegister *acc_reg) = 0;
    uint64_t GetNotSupportedBit() {
        return not_supported_bit;
    }
    string GetSectionName() {
        return section_name;
    }
    uint32_t GetRegisterID() {
        return register_id;
    }
    bool ToDump() {
        return to_dump;
    }
    AccRegVia_t GetRegisterType() {
        return acc_reg_type;
    }
    inline bool IsRegSupportNodeType(IBNodeType type);

    void DumpRegisterHeader(stringstream &sstream);

    virtual void DumpRegisterData(struct acc_reg_data, stringstream &sstream) = 0;
    virtual void GetGroupData(struct acc_reg_data& areg, u_int8_t& group, u_int16_t& indices);
    virtual int AvailableSensors(struct acc_reg_data& areg, list_uint8& sensors);
    static int SensorsBitsToList(u_int64_t sensors, u_int8_t max_sensors, list_uint8& sensors_list);
    static int SensorsCountToList(u_int8_t sensor_count, list_uint8& sensors_list);

    virtual int BuildDB(AccRegHandler* p_handler,
                        list_p_fabric_general_err &phy_errors,
                        progress_func_nodes_t progress_func);

    virtual bool IsValidPakcet(acc_reg_data reg_data) { return true; }
};


class SLRegister : public Register {
private:
    map_akey_areg *p_mpein_map; // for PCI mode only (pnat = 3)
    map_akey_areg *p_mpir_map;  // for PCI mode only (pnat = 3)

public:
    SLRegister(uint32_t reg_id,
               const unpack_data_func_t p_unpack_func,
               string sn,
               uint32_t fn,
               uint64_t nsb,
               map_akey_areg *mpein_map,
               map_akey_areg *mpir_map);
    virtual ~SLRegister(){}

    virtual void PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg) = 0;
    virtual void DumpRegisterData(struct acc_reg_data, stringstream &sstream) = 0;

    virtual int BuildDB(AccRegHandler* p_handler,
                        list_p_fabric_general_err &phy_errors,
                        progress_func_nodes_t progress_func);
};

class SLTPRegister : public SLRegister {
private:
        u_int8_t    m_pnat;

public:
        SLTPRegister(u_int8_t pnat = ACC_REG_PNAT_IB_PORT,
                     string sn = ACC_REG_SLTP_INTERNAL_SECTION_NAME,
                     map_akey_areg *mpein_map = NULL,
                     map_akey_areg *mpir_map = NULL);
        virtual ~SLTPRegister(){}

        virtual void PackData(AccRegKey *p_key,
                              struct SMP_AccessRegister *acc_reg);

        virtual void DumpRegisterData(struct acc_reg_data areg, stringstream &sstream);
};

class SLRGRegister : public SLRegister {
private:
        u_int8_t    m_pnat;

public:
        SLRGRegister(u_int8_t pnat = ACC_REG_PNAT_IB_PORT,
                     string sn = ACC_REG_SLRG_INTERNAL_SECTION_NAME,
                     map_akey_areg *mpein_map = NULL,
                     map_akey_areg *mpir_map = NULL);
        virtual ~SLRGRegister(){}

        virtual void PackData(AccRegKey *p_key,
                              struct SMP_AccessRegister *acc_reg);

        virtual void DumpRegisterData(struct acc_reg_data areg, stringstream &sstream);
};

class SLRPRegister : public SLRegister {
private:
        u_int8_t    m_pnat;

public:
        SLRPRegister(u_int8_t pnat = ACC_REG_PNAT_IB_PORT,
                     string sn = ACC_REG_SLRP_INTERNAL_SECTION_NAME,
                     map_akey_areg *mpein_map = NULL,
                     map_akey_areg *mpir_map = NULL);
        virtual ~SLRPRegister(){}

        virtual void PackData(AccRegKey *p_key,
                              struct SMP_AccessRegister *acc_reg);

        virtual void DumpRegisterData(struct acc_reg_data areg, stringstream &sstream);
};

class PTASRegister : public Register {
public:

        PTASRegister();
        virtual ~PTASRegister(){}

        virtual void PackData(AccRegKey *p_key,
                              struct SMP_AccessRegister *acc_reg);

        virtual void DumpRegisterData(struct acc_reg_data areg, stringstream &sstream);
};

class PPLLRegister : public Register {
public:

        PPLLRegister();
        virtual ~PPLLRegister(){}

        virtual void PackData(AccRegKey *p_key,
                              struct SMP_AccessRegister *acc_reg);

        virtual void DumpRegisterData(struct acc_reg_data areg, stringstream &sstream);
};

class MPEINRegister : public Register {
public:
        MPEINRegister();
        virtual ~MPEINRegister(){}

        virtual void PackData(AccRegKey *p_key,
                struct SMP_AccessRegister *acc_reg);

        virtual void DumpRegisterData(struct acc_reg_data areg, stringstream &sstream);

        virtual int BuildDB(AccRegHandler* p_handler,
                            list_p_fabric_general_err &phy_errors,
                            progress_func_nodes_t progress_func);

        virtual bool IsValidPakcet(acc_reg_data reg_data);

        int LoopPCIDepth(const clbck_data_t &clbck_data,
                         int rec_status,
                         void *p_attribute_data);

        int LoopPCINode(const clbck_data_t &clbck_data,
                        int rec_status,
                        void *p_attribute_data);
};

class MPIRRegister : public Register {

    map_akey_areg &mpein_map;

public:
    MPIRRegister(map_akey_areg &mpein_map);
    virtual ~MPIRRegister(){}

    virtual void PackData(AccRegKey *p_key, SMP_AccessRegister *acc_reg);

    virtual void DumpRegisterData(acc_reg_data areg, stringstream &sstream);

    virtual int BuildDB(AccRegHandler* p_handler, // handler containing the source map of mpein
                        list_p_fabric_general_err &phy_errors,
                        progress_func_nodes_t progress_func);
};


// fans registers
class MFCRRegister : public Register {
public:
    MFCRRegister();
    virtual ~MFCRRegister() { }

    virtual void PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg);
    virtual void DumpRegisterData(struct acc_reg_data areg, stringstream &sstream);
    virtual int AvailableSensors(struct acc_reg_data& areg, list_uint8& sensors_list);
};

class MFSMRegister : public Register {
public:
    MFSMRegister();
    virtual ~MFSMRegister() { }

    virtual void PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg);
    virtual void DumpRegisterData(struct acc_reg_data areg, stringstream &sstream);
};

class MFSLRegister : public Register {
public:
    MFSLRegister();
    virtual ~MFSLRegister() { }

    virtual void PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg);
    virtual void DumpRegisterData(struct acc_reg_data areg, stringstream &sstream);
};

class FORERegister : public Register {
public:
    FORERegister();
    virtual ~FORERegister() { }

    virtual void PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg);
    virtual void DumpRegisterData(struct acc_reg_data areg, stringstream &sstream);
};

class MFNRRegister : public Register {
public:
    MFNRRegister();
    virtual ~MFNRRegister() { }

    virtual void PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg);
    virtual void DumpRegisterData(struct acc_reg_data areg, stringstream &sstream);
};

// temperature registers
class MTCAPRegister : public Register {
public:
    MTCAPRegister();
    virtual ~MTCAPRegister() { }

    virtual void PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg);
    virtual void DumpRegisterData(struct acc_reg_data areg, stringstream &sstream);
    virtual int AvailableSensors(struct acc_reg_data& areg, list_uint8& sensors_list);
};

class MTMPRegister : public Register {
public:
    MTMPRegister();
    virtual ~MTMPRegister() { }

    virtual void PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg);
    virtual void DumpRegisterData(struct acc_reg_data areg, stringstream &sstream);
};

class MTWERegister : public Register {
public:
    MTWERegister();
    virtual ~MTWERegister() { }

    virtual void PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg);
    virtual void DumpRegisterData(struct acc_reg_data areg, stringstream &sstream);
};

// current/voltage registers
class MVCAPRegister : public Register {
public:
    MVCAPRegister();
    virtual ~MVCAPRegister() { }

    virtual void PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg);
    virtual void DumpRegisterData(struct acc_reg_data areg, stringstream& sstream);
    virtual int AvailableSensors(struct acc_reg_data& areg, list_uint8& sensors_list);
};

class MVCRRegister : public Register {
public:
    MVCRRegister();
    virtual ~MVCRRegister() { }

    virtual void PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg);
    virtual void DumpRegisterData(struct acc_reg_data areg, stringstream &sstream);
};

// power supplies register
class MSPSRegister : public Register {
public:
    MSPSRegister();
    virtual ~MSPSRegister() { }

    virtual void PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg) { }
    virtual void DumpRegisterData(struct acc_reg_data areg, stringstream &sstream);
};

// general information register
class MSGIRegister : public Register {
public:
    MSGIRegister();
    virtual ~MSGIRegister() { }

    virtual void PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg) { }
    virtual void DumpRegisterData(struct acc_reg_data areg, stringstream &sstream);
};

/*
 * Registers for HDR bringup
 * */

class SLRIPRegister : public Register {
public:
    SLRIPRegister();
    virtual ~SLRIPRegister() { }

    virtual void PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg);
    virtual void DumpRegisterData(struct acc_reg_data areg, stringstream &sstream);
};

class SLREGRegister : public Register {
public:
    SLREGRegister();
    virtual ~SLREGRegister() { }

    virtual void PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg);
    virtual void DumpRegisterData(struct acc_reg_data areg, stringstream &sstream);
};

class SLCCTRegister : public Register {
public:
    SLCCTRegister();
    virtual ~SLCCTRegister() { }

    virtual void PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg);
    virtual void DumpRegisterData(struct acc_reg_data areg, stringstream &sstream);
};

class PPAMPRegister : public Register {
public:
    PPAMPRegister();
    virtual ~PPAMPRegister() { }

    virtual void PackData(AccRegKey *p_key, struct SMP_AccessRegister *acc_reg);
    virtual void DumpRegisterData(struct acc_reg_data areg, stringstream &sstream);
    virtual void GetGroupData(struct acc_reg_data& areg, u_int8_t& group, u_int16_t& indices);
};


// AccRegHandler defines a generic way of reading registers' data across the fabric.
// There are 4 types of access to register:
// by port
// by port + lane
// by node
// by pll group
class AccRegHandler {
protected:
    list_p_fabric_general_err   *phy_errors;
    bool                        can_send_mads_by_lid;
    int                         clbck_error_state;
    Register                    *p_reg;
    string                      handler_header;

public:
    map_akey_areg               data_map;
    class PhyDiag               *p_phy_diag;

    AccRegHandler(Register *p_register, class PhyDiag *p_pd,
                  string hh);
    virtual ~AccRegHandler();

    virtual int BuildDB(list_p_fabric_general_err &phy_errors,
                        progress_func_nodes_t progress_func) = 0;

    int SMPAccessRegisterHandlerGetClbck(const clbck_data_t &clbck_data,
                                         int rec_status,
                                         void *p_attribute_data);

    void GMPAccessRegisterHandlerGetClbck(const clbck_data_t &clbck_data,
                                          int rec_status,
                                          void *p_attribute_data);
    void DumpCSV(CSVOut &csv_out);

    Register * GetPReg() {
        return p_reg;
    }

    bool ToDumpRegister() {
        return p_reg->ToDump();
    }

    void SetHeader(string header) {
        handler_header = header;
    }
};

class AccRegLaneHandler : public AccRegHandler {
    u_int32_t m_lane_num;
public:
    AccRegLaneHandler(Register *p_reg, class PhyDiag *p_pd, u_int32_t lane_num):
        AccRegHandler(p_reg, p_pd, PORT_LANE_HEADER),
        m_lane_num(lane_num){}
    virtual ~AccRegLaneHandler(){}

    virtual int BuildDB(list_p_fabric_general_err &phy_errors,
                        progress_func_nodes_t progress_func);
};

class AccRegPortLaneHandler : public AccRegHandler {
    u_int8_t max_idx_in_lane;
public:
    AccRegPortLaneHandler(Register *p_reg, class PhyDiag *p_pd, u_int8_t max_idx = 1):
        AccRegHandler(p_reg, p_pd, PORT_LANE_HEADER), max_idx_in_lane(max_idx){}

    virtual ~AccRegPortLaneHandler(){}

    virtual int BuildDB(list_p_fabric_general_err &phy_errors,
                        progress_func_nodes_t progress_func);
};

class AccRegNodeHandler : public AccRegHandler {
public:
    AccRegNodeHandler(Register *p_reg, class PhyDiag *p_pd):
        AccRegHandler(p_reg, p_pd, NODE_HEADER){}
    virtual ~AccRegNodeHandler(){}

    virtual int BuildDB(list_p_fabric_general_err &phy_errors,
                        progress_func_nodes_t progress_func);
};

class AccRegPortHandler : public AccRegHandler {
public:
    AccRegPortHandler(Register *p_reg, class PhyDiag *p_pd):
        AccRegHandler(p_reg, p_pd, PORT_HEADER){}
    virtual ~AccRegPortHandler(){}

    virtual int BuildDB(list_p_fabric_general_err &phy_errors,
                        progress_func_nodes_t progress_func);
};

class AccRegGroupHandler : public AccRegHandler {
private:
    uint32_t group_max_idx;
public:
    AccRegGroupHandler(Register *p_reg, class PhyDiag *p_pd):
        AccRegHandler(p_reg, p_pd, GROUP_HEADER){}
    virtual ~AccRegGroupHandler(){}

    virtual int BuildDB(list_p_fabric_general_err &phy_errors,
                        progress_func_nodes_t progress_func);
};

class AccRegNodeSensorHandler : public AccRegHandler {
private:
    AccRegHandler *p_acc_reg_handler; // pointer for the capability register

public:
    AccRegNodeSensorHandler(Register *p_reg, class PhyDiag *p_pd, AccRegHandler *p_acc_reg):
        AccRegHandler(p_reg, p_pd, NODE_SENSOR_HEADER),
        p_acc_reg_handler(p_acc_reg){}
    virtual ~AccRegNodeSensorHandler(){}

    virtual int BuildDB(list_p_fabric_general_err &phy_errors,
                        progress_func_nodes_t progress_func);
};

class AccRegTypeGroupHandler : public AccRegHandler {
private:

    uint8_t max_type;
    uint16_t max_indices;

public:
    AccRegTypeGroupHandler(Register *p_reg, class PhyDiag *p_pd,
                           uint8_t types, uint16_t indices):
        AccRegHandler(p_reg, p_pd, TYPE_GROUP_HEADER), max_type(types), max_indices(indices){}
    virtual ~AccRegTypeGroupHandler(){}

    virtual int BuildDB(list_p_fabric_general_err &phy_errors,
                        progress_func_nodes_t progress_func);
};


class AccRegSpecificHandler : public AccRegHandler {
public:
    AccRegSpecificHandler(Register *p_reg, class PhyDiag *p_pd):
        AccRegHandler(p_reg, p_pd, PORT_HEADER){}
    virtual ~AccRegSpecificHandler(){}

    virtual int BuildDB(list_p_fabric_general_err &phy_errors,
                        progress_func_nodes_t progress_func);
};


#endif //_ACC_REGISTER_H_

