 /*
 * Copyright (c) 2004-2018 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include "diagnostic_data.h"


DiagnosticDataInfo::DiagnosticDataInfo(int page_id, int support_version,
                               int num_fields, int not_supported_bit, u_int32_t dd_type,
                               string header, u_int8_t is_per_node,
                               PhyPluginSupportedNodesType support_nodes):
                               m_page_id(page_id),
                               m_support_version(support_version),
                               m_num_fields(num_fields),
                               m_not_supported_bit(not_supported_bit),
                               m_dd_type(dd_type),
                               m_header(header),
                               m_is_per_node(is_per_node),
                               m_support_nodes(support_nodes)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

void DiagnosticDataInfo::DumpDiagnosticDataHeaderStart(CSVOut &csv_out) const
{
    IBDIAGNET_ENTER;
    stringstream sstream;
    csv_out.DumpStart(this->GetSectionHeader().c_str());

    sstream << "NodeGuid,PortGuid,PortNum,Version";

    for (u_int32_t cnt = 0; cnt < m_num_fields; cnt++)
        sstream << ",field" <<  cnt;
    sstream << endl;
    csv_out.WriteLine(sstream.str());

    IBDIAGNET_RETURN_VOID;
}

void DiagnosticDataInfo::DumpDiagnosticDataHeaderEnd(CSVOut &csv_out) const
{
    IBDIAGNET_ENTER;

    csv_out.DumpEnd(this->GetSectionHeader().c_str());

    IBDIAGNET_RETURN_VOID;
}

DiagnosticDataPhysLayerCntrs::DiagnosticDataPhysLayerCntrs():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_PHYS_LAYER_CNTRS_PAGE,
                       DIAGNOSTIC_DATA_PHYS_LAYER_CNTRS_VERSION,
                       DIAGNOSTIC_DATA_PHYS_LAYER_CNTRS_NUM_FIELDS,
                       NOT_SUPPORT_DIAGNOSTIC_DATA_PHYS_CNTRS,
                       DD_PHY_TYPE,
                       SECTION_PHY_CNTRS,
                       false)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

void DiagnosticDataPhysLayerCntrs::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd) const
{
    IBDIAGNET_ENTER;
    char buff[256];
    //unpack physical layer counters page
    struct DDPhysCounters phy_cntrs;
    DDPhysCounters_unpack(&phy_cntrs, (uint8_t *)&(dd.data_set));

    u_int32_t size = sizeof(phy_cntrs.counter) / 8;

    memset(buff, 0, sizeof(buff));

    for (u_int32_t cnt = 0; cnt < size; cnt++) {
        sprintf(buff, U64H_FMT ",",
                phy_cntrs.counter[cnt]);
        sstream << buff;
    }

    sprintf(buff, U32H_FMT "," U32H_FMT,
            phy_cntrs.counter24,
            phy_cntrs.counter25);

    sstream << buff;

    IBDIAGNET_RETURN_VOID;
}

DiagnosticDataPCIECntrs::DiagnosticDataPCIECntrs():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_PCIE_CNTRS_PAGE,
                       DIAGNOSTIC_DATA_PCIE_CNTRS_VERSION,
                       DIAGNOSTIC_DATA_PCIE_CNTRS_NUM_FIELDS,
                       NOT_SUPPORT_DIAGNOSTIC_DATA_PCIE_CNTRS,
                       DD_PCI_TYPE,
                       SECTION_PCI_CNTRS,
                       true)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

void DiagnosticDataPCIECntrs::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd) const
{
    IBDIAGNET_ENTER;
    char buff[256];
    //unpack physical layer counters page
    struct DDPCIPerfCounters p_pci_cntrs;
    DDPCIPerfCounters_unpack(&p_pci_cntrs, (uint8_t *)&(dd.data_set));

    memset(buff, 0, sizeof(buff));

    sprintf(buff,
            U32H_FMT "," U32H_FMT "," U32H_FMT  "," \
            U32H_FMT  "," U32H_FMT  "," U32H_FMT  ","   \
            U32H_FMT  "," U32H_FMT  "," U32H_FMT "," U32H_FMT,
            p_pci_cntrs.life_time_counter_high,
            p_pci_cntrs.life_time_counter_low,
            p_pci_cntrs.rx_errors,
            p_pci_cntrs.tx_errors,
            p_pci_cntrs.l0_to_recovery_eieos,
            p_pci_cntrs.l0_to_recovery_ts,
            p_pci_cntrs.l0_to_recovery_framing,
            p_pci_cntrs.l0_to_recovery_retrain,
            p_pci_cntrs.crc_error_dllp,
            p_pci_cntrs.crc_error_tlp);
    sstream << buff;

    IBDIAGNET_RETURN_VOID;
}

void DiagnosticDataPCIECntrs::DumpDiagnosticDataHeaderStart(CSVOut &csv_out) const
{
    IBDIAGNET_ENTER;
    stringstream sstream;
    csv_out.DumpStart(this->GetSectionHeader().c_str());

    sstream << "NodeGuid,PCIIndex,Depth,PCINode,Version";

    for (u_int32_t cnt = 0; cnt < m_num_fields; cnt++)
        sstream << ",field" <<  cnt;
    sstream << endl;
    csv_out.WriteLine(sstream.str());

    IBDIAGNET_RETURN_VOID;
}

// PDDR Troubleshooting Info
DiagnosticDataTroubleshootingInfo::DiagnosticDataTroubleshootingInfo():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_TROUBLESHOOTING_INFO_PAGE,
                       DIAGNOSTIC_DATA_TROUBLESHOOTING_INFO_VERSION,
                       DIAGNOSTIC_DATA_TROUBLESHOOTING_INFO_NUM_FIELDS,
                       NOT_SUPPORT_DD_TROUBLESHOOTING_INFO,
                       DD_PHY_TYPE,
                       SECTION_TROUBLESHOOTING_INFO,
                       false)
{
}

void DiagnosticDataTroubleshootingInfo::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd) const
{
    IBDIAGNET_ENTER;
    struct DDTroubleShootingInfo troubleshooting_info;
    DDTroubleShootingInfo_unpack(&troubleshooting_info, (uint8_t *)&(dd.data_set));

    sstream << troubleshooting_info.status_opcode << ','
            << '\"' << troubleshooting_info.status_message << '\"';

    IBDIAGNET_RETURN_VOID;
}

// PDDR Operation Info
DiagnosticDataOperationInfo::DiagnosticDataOperationInfo():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_OPERATION_INFO_PAGE,
                       DIAGNOSTIC_DATA_OPERATION_INFO_VERSION,
                       DIAGNOSTIC_DATA_OPERATION_INFO_NUM_FIELDS,
                       NOT_SUPPORT_DD_OPERATION_INFO,
                       DD_PHY_TYPE,
                       SECTION_OPERATION_INFO,
                       false)
{
}

void DiagnosticDataOperationInfo::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd) const
{
    IBDIAGNET_ENTER;
    struct DDOperationInfo operation_info;
    DDOperationInfo_unpack(&operation_info, (uint8_t *)&(dd.data_set));

    sstream << +operation_info.proto_active << ','
            << +operation_info.neg_mode_active << ','
            << +operation_info.phy_mngr_fsm_state << ','
            << +operation_info.ib_phy_fsm_state << ','
            << operation_info.phy_manager_link_width_enabled << ','
            << operation_info.phy_manager_link_proto_enabled << ','
            << operation_info.core_to_phy_link_width_enabled << ','
            << operation_info.core_to_phy_link_proto_enabled << ','
            << operation_info.cable_proto_cap << ','
            << operation_info.link_width_active << ','
            << operation_info.link_speed_active << ','
            << +operation_info.retran_mode_active << ','
            << +operation_info.retran_mode_request << ','
            << +operation_info.loopback_mode << ','
            << operation_info.fec_mode_active << ','
            << operation_info.fec_mode_request << ','
            << operation_info.profile_fec_in_use;

    IBDIAGNET_RETURN_VOID;
}

// PDDR Phy Info
DiagnosticDataPhyInfo::DiagnosticDataPhyInfo():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_PHY_INFO_PAGE,
                       DIAGNOSTIC_DATA_PHY_INFO_VERSION,
                       DIAGNOSTIC_DATA_PHY_INFO_NUM_FIELDS,
                       NOT_SUPPORT_DD_PHY_INFO,
                       DD_PHY_TYPE,
                       SECTION_PHY_INFO,
                       false)
{
}

void DiagnosticDataPhyInfo::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd) const
{
    IBDIAGNET_ENTER;
    struct DDPhyInfo phy_info;
    DDPhyInfo_unpack(&phy_info, (uint8_t *)&(dd.data_set));

    sstream << +phy_info.remote_device_type << ','
            << +phy_info.port_notifications << ','
            << phy_info.num_of_negotiation_attempts << ','
            << +phy_info.ib_revision << ','
            << +phy_info.lp_ib_revision << ','
            << +phy_info.hw_link_phy_state << ','
            << phy_info.phy_manger_disable_mask << ','
            << phy_info.pcs_phy_state << ','
            << phy_info.lp_proto_enabled << ','
            << phy_info.lp_fec_mode_support << ','
            << phy_info.lp_fec_mode_request << ','
            << phy_info.ib_last_link_down_reason << ','
            << phy_info.speed_deg_db << ','
            << phy_info.degrade_grade_lane_0 << ','
            << phy_info.degrade_grade_lane_1 << ','
            << phy_info.degrade_grade_lane_2 << ','
            << phy_info.degrade_grade_lane_3 << ','
            << phy_info.ib_phy_fsm_state_trace << ','
            << +phy_info.irisc_status << ','
            << phy_info.stamping_reason;

    IBDIAGNET_RETURN_VOID;
}

// PDDR Module Info
DiagnosticDataModuleInfo::DiagnosticDataModuleInfo():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_MODULE_INFO_PAGE,
                       DIAGNOSTIC_DATA_MODULE_INFO_VERSION,
                       DIAGNOSTIC_DATA_MODULE_INFO_NUM_FIELDS,
                       NOT_SUPPORT_DD_MODULE_INFO,
                       DD_PHY_TYPE,
                       SECTION_MODULE_INFO,
                       false)
{
}

void DiagnosticDataModuleInfo::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd) const
{
    IBDIAGNET_ENTER;
    struct DDModuleInfo module_info;
    DDModuleInfo_unpack(&module_info, (uint8_t *)&(dd.data_set));

    uint8_t fw_hi = (uint8_t)(module_info.fw_version >> 24);
    uint8_t fw_mid = (uint8_t)((module_info.fw_version >> 16) & 0x00ff);
    uint16_t fw_lo = (uint16_t)(module_info.fw_version & 0x0000ffff);

    sstream << +module_info.cable_technology << ','
            << +module_info.cable_breakout << ','
            << +module_info.cable_type << ','
            << +module_info.cable_vendor << ','
            << +module_info.cable_length << ','
            << +module_info.cable_identifier << ','
            << +module_info.cable_power_class << ','
            << +module_info.cable_rx_amp << ','
            << +module_info.cable_rx_emphasis << ','
            << +module_info.cable_tx_equalization << ','
            << +module_info.cable_attenuation_12g << ','
            << +module_info.cable_attenuation_7g << ','
            << +module_info.cable_attenuation_5g << ','
            << +module_info.rx_cdr_cap << ','
            << +module_info.tx_cdr_cap << ','
            << +module_info.rx_cdr_state << ','
            << +module_info.tx_cdr_state << ','
            << '\"' << module_info.vendor_name << '\"' << ','
            << '\"' << module_info.vendor_pn << '\"' << ','
            << '\"' << module_info.vendor_rev << '\"' << ','
            << +fw_hi << "." << +fw_mid  << "." << fw_lo << ','
            << '\"' << module_info.vendor_sn << '\"' << ','
            << module_info.temperature << ','
            << module_info.voltage << ','
            << module_info.rx_power_lane_0 << ','
            << module_info.rx_power_lane_1 << ','
            << module_info.rx_power_lane_2 << ','
            << module_info.rx_power_lane_3 << ','
            << module_info.tx_power_lane_0 << ','
            << module_info.tx_power_lane_1 << ','
            << module_info.tx_power_lane_2 << ','
            << module_info.tx_power_lane_3 << ','
            << module_info.tx_bias_lane_0 << ','
            << module_info.tx_bias_lane_1 << ','
            << module_info.tx_bias_lane_2 << ','
            << module_info.tx_bias_lane_3 << ','
            << module_info.temperature_high_th << ','
            << module_info.temperature_low_th << ','
            << module_info.voltage_high_th << ','
            << module_info.voltage_low_th << ','
            << module_info.rx_power_high_th << ','
            << module_info.rx_power_low_th << ','
            << module_info.tx_power_high_th << ','
            << module_info.tx_power_low_th << ','
            << module_info.tx_bias_high_th << ','
            << module_info.tx_bias_low_th << ','
            << module_info.wavelength;

    IBDIAGNET_RETURN_VOID;
}

// PDDR Link Down Info
DiagnosticDataLinkDownInfo::DiagnosticDataLinkDownInfo():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_LINK_DOWN_INFO_PAGE,
                       DIAGNOSTIC_DATA_LINK_DOWN_INFO_VERSION,
                       DIAGNOSTIC_DATA_LINK_DOWN_INFO_NUM_FIELDS,
                       NOT_SUPPORT_DD_LINK_DOWN_INFO,
                       DD_PHY_TYPE,
                       SECTION_LINK_DOWN_INFO,
                       false)
{
}

void DiagnosticDataLinkDownInfo::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd) const
{
    IBDIAGNET_ENTER;
    struct DDLinkDownInfo link_down_info;
    DDLinkDownInfo_unpack(&link_down_info, (uint8_t *)&(dd.data_set));

    sstream << +link_down_info.down_blame << ','
            << +link_down_info.local_reason_opcode << ','
            << +link_down_info.remote_reason_opcode << ','
            << +link_down_info.e2e_reason_opcode << ','
            << +link_down_info.last_ber_mantissa << ','
            << +link_down_info.last_ber_exp << ','
            << +link_down_info.ber_mantissa << ','
            << +link_down_info.ber_exp << ','
            << +link_down_info.min_ber_mantissa << ','
            << +link_down_info.min_ber_exp << ','
            << +link_down_info.max_ber_mantissa << ','
            << +link_down_info.max_ber_exp << ','
            << link_down_info.num_of_ber_alarams;

    IBDIAGNET_RETURN_VOID;
}


// PPCNT pages
// PPCNT RS FEC Histograms
DiagnosticDataRSHistograms::DiagnosticDataRSHistograms():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_RS_FEC_HISTOGRAM_PAGE,
                       DIAGNOSTIC_DATA_RS_FEC_HISTOGRAM_VERSION,
                       DIAGNOSTIC_DATA_RS_FEC_HISTOGRAM_NUM_FIELDS,
                       NOT_SUPPORT_DD_RS_FEC_HISTOGRAM,
                       DD_PHY_TYPE,
                       SECTION_RS_FEC_HISTOGRAM,
                       false)
{
}

void DiagnosticDataRSHistograms::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd) const
{
    IBDIAGNET_ENTER;
    struct DD_RS_Histograms rs_histograms;
    DD_RS_Histograms_unpack(&rs_histograms, (uint8_t *)&(dd.data_set));

    sstream << rs_histograms.hist[0];
    for (int i = 1; i < DIAGNOSTIC_DATA_RS_FEC_HISTOGRAM_NUM_FIELDS; ++i) {
        sstream << ',' << rs_histograms.hist[i];
    }

    IBDIAGNET_RETURN_VOID;
}

// PPCNT PLR Counters
DiagnosticDataPLRCounters::DiagnosticDataPLRCounters():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_PLR_COUNTERS_PAGE,
                       DIAGNOSTIC_DATA_PLR_COUNTERS_VERSION,
                       DIAGNOSTIC_DATA_PLR_COUNTERS_NUM_FIELDS,
                       NOT_SUPPORT_DD_PLR_COUNTERS,
                       DD_PHY_TYPE,
                       SECTION_PLR_COUNTERS,
                       false)
{
}

void DiagnosticDataPLRCounters::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd) const
{
    IBDIAGNET_ENTER;
    struct DD_PLR_Counters plr_counters;
    DD_PLR_Counters_unpack(&plr_counters, (uint8_t *)&(dd.data_set));

    sstream << plr_counters.plr_rcv_codes << ','
            << plr_counters.plr_rcv_code_err << ','
            << plr_counters.plr_rcv_uncorrectable_code << ','
            << plr_counters.plr_xmit_codes << ','
            << plr_counters.plr_xmit_retry_codes << ','
            << plr_counters.plr_xmit_retry_events << ','
            << plr_counters.plr_sync_events << ','
            << plr_counters.hi_retransmission_rate;

    IBDIAGNET_RETURN_VOID;
}


// PPCNT Phy Statistics
DiagnosticDataPhyStatistics::DiagnosticDataPhyStatistics():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_PHY_LAYER_STATISTICS_PAGE,
                       DIAGNOSTIC_DATA_PHY_LAYER_STATISTICS_VERSION,
                       DIAGNOSTIC_DATA_PHY_LAYER_STATISTICS_NUM_FIELDS,
                       NOT_SUPPORT_DD_PHY_LAYER_STATISTICS,
                       DD_PHY_TYPE,
                       SECTION_PHY_LAYER_STATISTICS,
                       false)
{
}

void DiagnosticDataPhyStatistics::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd) const
{
    IBDIAGNET_ENTER;
    struct DD_PhyStatisticalCounters phy_statistics;
    DD_PhyStatisticalCounters_unpack(&phy_statistics, (uint8_t *)&(dd.data_set));

    sstream << phy_statistics.time_since_last_clear << ','
            << phy_statistics.phy_received_bits << ','
            << phy_statistics.phy_symbol_errors << ','
            << phy_statistics.phy_corrected_bits << ','
            << phy_statistics.phy_raw_errors_lane0<< ','
            << phy_statistics.phy_raw_errors_lane1 << ','
            << phy_statistics.phy_raw_errors_lane2 << ','
            << phy_statistics.phy_raw_errors_lane3 << ','
            << +phy_statistics.raw_ber_coef << ','
            << +phy_statistics.raw_ber_magnitude << ','
            << +phy_statistics.effective_ber_coef << ','
            << +phy_statistics.effective_ber_magnitude;

    IBDIAGNET_RETURN_VOID;
}

// PPCNT ZL FEC Counters
DiagnosticDataZLFECCounters::DiagnosticDataZLFECCounters():
    DiagnosticDataInfo(DIAGNOSTIC_DATA_ZL_FEC_COUNTERS_PAGE,
                       DIAGNOSTIC_DATA_ZL_FEC_COUNTERS_VERSION,
                       DIAGNOSTIC_DATA_ZL_FEC_COUNTERS_NUM_FIELDS,
                       NOT_SUPPORT_DD_ZL_FEC_COUNTERS,
                       DD_PHY_TYPE,
                       SECTION_ZL_FEC_COUNTERS,
                       false)
{
}

void DiagnosticDataZLFECCounters::DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd) const
{
    IBDIAGNET_ENTER;
    struct DD_ZL_FEC_Counters zl_fec_counters;
    DD_ZL_FEC_Counters_unpack(&zl_fec_counters, (uint8_t *)&(dd.data_set));

    sstream << zl_fec_counters.rewind_counter << ','
            << zl_fec_counters.data_rewined << ','
            << zl_fec_counters.pkt_rewined << ','
            << zl_fec_counters.total_rcv_data << ','
            << zl_fec_counters.total_rcv_pkt<< ','
            << zl_fec_counters.stay_in_slow;

    IBDIAGNET_RETURN_VOID;
}
