 /*
 * Copyright (c) 2004-2018 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#ifndef _DIAGNOSTIC_DATA_H_
#define _DIAGNOSTIC_DATA_H_

#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h>
#include "../internal_packets/internal_packets_layouts.h"
#include "phy_diag_types.h"

#define DIAGNOSTIC_DATA_PHYS_LAYER_CNTRS_PAGE       254
#define DIAGNOSTIC_DATA_PHYS_LAYER_CNTRS_VERSION     1
#define DIAGNOSTIC_DATA_PHYS_LAYER_CNTRS_NUM_FIELDS  26

#define DIAGNOSTIC_DATA_PCIE_CNTRS_PAGE           2
#define DIAGNOSTIC_DATA_PCIE_CNTRS_VERSION        1
#define DIAGNOSTIC_DATA_PCIE_CNTRS_NUM_FIELDS     10

// PDDR pages
#define DIAGNOSTIC_DATA_TROUBLESHOOTING_INFO_PAGE        253
#define DIAGNOSTIC_DATA_TROUBLESHOOTING_INFO_VERSION     1
#define DIAGNOSTIC_DATA_TROUBLESHOOTING_INFO_NUM_FIELDS  2

#define DIAGNOSTIC_DATA_OPERATION_INFO_PAGE        252
#define DIAGNOSTIC_DATA_OPERATION_INFO_VERSION     1
#define DIAGNOSTIC_DATA_OPERATION_INFO_NUM_FIELDS  17

#define DIAGNOSTIC_DATA_PHY_INFO_PAGE        251
#define DIAGNOSTIC_DATA_PHY_INFO_VERSION     1
#define DIAGNOSTIC_DATA_PHY_INFO_NUM_FIELDS  20

#define DIAGNOSTIC_DATA_MODULE_INFO_PAGE        250
#define DIAGNOSTIC_DATA_MODULE_INFO_VERSION     1
#define DIAGNOSTIC_DATA_MODULE_INFO_NUM_FIELDS  47

#define DIAGNOSTIC_DATA_LINK_DOWN_INFO_PAGE        248
#define DIAGNOSTIC_DATA_LINK_DOWN_INFO_VERSION     1
#define DIAGNOSTIC_DATA_LINK_DOWN_INFO_NUM_FIELDS  13

// PPCNT pages
#define DIAGNOSTIC_DATA_RS_FEC_HISTOGRAM_PAGE        247
#define DIAGNOSTIC_DATA_RS_FEC_HISTOGRAM_VERSION     1
#define DIAGNOSTIC_DATA_RS_FEC_HISTOGRAM_NUM_FIELDS  16

#define DIAGNOSTIC_DATA_PLR_COUNTERS_PAGE        246
#define DIAGNOSTIC_DATA_PLR_COUNTERS_VERSION     1
#define DIAGNOSTIC_DATA_PLR_COUNTERS_NUM_FIELDS  8

#define DIAGNOSTIC_DATA_PHY_LAYER_STATISTICS_PAGE        245
#define DIAGNOSTIC_DATA_PHY_LAYER_STATISTICS_VERSION     1
#define DIAGNOSTIC_DATA_PHY_LAYER_STATISTICS_NUM_FIELDS  12

#define DIAGNOSTIC_DATA_ZL_FEC_COUNTERS_PAGE        244
#define DIAGNOSTIC_DATA_ZL_FEC_COUNTERS_VERSION     1
#define DIAGNOSTIC_DATA_ZL_FEC_COUNTERS_NUM_FIELDS  6


/******************************************************/
/******************************************************/
/******************************************************/
class DiagnosticDataInfo {
private:


protected:
    u_int32_t         m_page_id;
    u_int32_t         m_support_version;
    u_int32_t         m_num_fields;
    u_int32_t         m_not_supported_bit;
    u_int32_t         m_dd_type;
    string            m_header;
    u_int8_t          m_is_per_node;
    PhyPluginSupportedNodesType m_support_nodes;

public:
    DiagnosticDataInfo(int page_id, int support_version,
                       int num_fields, int not_supported_bit,
                       u_int32_t dd_type, string header, u_int8_t is_per_node,
                       PhyPluginSupportedNodesType support_nodes = SUPPORT_ALL);

    virtual ~DiagnosticDataInfo() {}

    string GetSectionHeader() const { return m_header; }
    inline int GetPageId() const { return m_page_id; }
    inline int GetSupportedVersion() const { return m_support_version; }
    inline int GetNotSupportedBit() const { return m_not_supported_bit; }
    inline u_int32_t GetDDType() const { return m_dd_type; }
    inline u_int8_t IsPerNode() const { return m_is_per_node; }

    inline bool IsDDPageSupportNodeType(IBNodeType type);

    virtual void DumpDiagnosticDataHeaderStart(CSVOut &csv_out) const;
    void DumpDiagnosticDataHeaderEnd(CSVOut &csv_out) const;

    virtual void DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd) const = 0;
};

bool DiagnosticDataInfo::IsDDPageSupportNodeType(IBNodeType type)
{
    IBDIAGNET_ENTER;
    bool rc = IsPhyPluginSupportNodeType(m_support_nodes, type);
    IBDIAGNET_RETURN(rc);
}

class DiagnosticDataPhysLayerCntrs : public DiagnosticDataInfo {
public:
    DiagnosticDataPhysLayerCntrs();
    virtual ~DiagnosticDataPhysLayerCntrs() {};

    virtual void DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd) const;
};

class DiagnosticDataPCIECntrs : public DiagnosticDataInfo {
public:
    DiagnosticDataPCIECntrs();
    virtual ~DiagnosticDataPCIECntrs() {};

    virtual void DumpDiagnosticDataHeaderStart(CSVOut &csv_out) const;
    virtual void DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd) const;
};

// PDDR Troubleshooting Info
class DiagnosticDataTroubleshootingInfo : public DiagnosticDataInfo {
public:
    DiagnosticDataTroubleshootingInfo();
    virtual ~DiagnosticDataTroubleshootingInfo() {};

    virtual void DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd) const;
};
// PDDR Operation Info
class DiagnosticDataOperationInfo : public DiagnosticDataInfo {
public:
    DiagnosticDataOperationInfo();
    virtual ~DiagnosticDataOperationInfo() {};

    virtual void DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd) const;
};
// PDDR Phy Info
class DiagnosticDataPhyInfo : public DiagnosticDataInfo {
public:
    DiagnosticDataPhyInfo();
    virtual ~DiagnosticDataPhyInfo() {};

    virtual void DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd) const;
};
// PDDR Module Info
class DiagnosticDataModuleInfo : public DiagnosticDataInfo {
public:
    DiagnosticDataModuleInfo();
    virtual ~DiagnosticDataModuleInfo() {};

    virtual void DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd) const;
};
// PDDR Link Down Info
class DiagnosticDataLinkDownInfo : public DiagnosticDataInfo {
public:
    DiagnosticDataLinkDownInfo();
    virtual ~DiagnosticDataLinkDownInfo() {};

    virtual void DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd) const;
};


// PPCNT pages
class DiagnosticDataRSHistograms : public DiagnosticDataInfo {
public:
    DiagnosticDataRSHistograms();
    virtual ~DiagnosticDataRSHistograms() {};

    virtual void DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd) const;
};

class DiagnosticDataPLRCounters : public DiagnosticDataInfo {
public:
    DiagnosticDataPLRCounters();
    virtual ~DiagnosticDataPLRCounters() {};

    virtual void DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd) const;
};

class DiagnosticDataPhyStatistics : public DiagnosticDataInfo {
public:
    DiagnosticDataPhyStatistics();
    virtual ~DiagnosticDataPhyStatistics() {};

    virtual void DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd) const;
};

class DiagnosticDataZLFECCounters : public DiagnosticDataInfo {
public:
    DiagnosticDataZLFECCounters();
    virtual ~DiagnosticDataZLFECCounters() {};

    virtual void DumpDiagnosticData(stringstream &sstream, VS_DiagnosticData &dd) const;
};

#endif // _DIAGNOSTIC_DATA_H_
