/*
 * Copyright (c) 2004-2019 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <sstream>
#include <typeinfo>
#include <math.h>
#include "phy_diag.h"
#include "phy_diag_errs.h"

static inline bool IsSupportAccRegFailed(u_int64_t app_data,
                                         u_int64_t not_support_bit)
{
    return (app_data & (NOT_SUPPORT_SMP_ACCESS_REGISTER | \
                        not_support_bit));
}

bool IsPhyPluginSupportNodeType(PhyPluginSupportedNodesType supported_nodes, IBNodeType type)
{
    IBDIAGNET_ENTER;
    bool rc = true;
    switch(supported_nodes) {
        case HCA_ONLY:
            if (type != IB_CA_NODE) rc = false;
            break;
        case SW_ONLY:
            if (type != IB_SW_NODE) rc = false;
            break;
        default:
            rc = true;
    }

    IBDIAGNET_RETURN(rc);
}

/****************************************************/
/*
 * public methods
 */
PhyDiag::PhyDiag(IBDiag *p_ibdiag) : Plugin(PLUGIN_NAME, p_ibdiag),
        can_send_mads_by_lid(false),
        to_get_phy_info(false), to_get_ppamp(false), to_dump_cap_reg(false),
        to_reset_counters(false), to_get_pci_info(false), to_reset_pci_counters(false),
        clbck_error_state(IBDIAG_SUCCESS_CODE), ber_threshold_err(BER_THRESHOLD_ERROR),
        ber_threshold_warn(BER_THRESHOLD_WARNING)
{
    IBDIAGNET_ENTER;
    /*
     * this stage is skipped by default
     */
    this->MarkStageAsSkipped();

    /*
     * add all options
     */
    stringstream def_val_sstream;
    string def_val_string;

    this->AddOptions(OPTION_PHY_INFO_GET, OPTION_PHY_INFO_GET_SHORT,
                     OPTION_PHY_INFO_GET_ARG_DESC, OPTION_PHY_INFO_GET_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_PPAMP_GET, OPTION_PPAMP_GET_SHORT,
                     OPTION_PPAMP_GET_ARG_DESC, OPTION_PPAMP_GET_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_SHOW_CAP_REG, OPTION_SHOW_CAP_REG_SHORT,
                     OPTION_SHOW_CAP_REG_ARG_DESC, OPTION_SHOW_CAP_REG_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_PHY_INFO_RESET, OPTION_PHY_INFO_RESET_SHORT,
                     OPTION_PHY_INFO_RESET_ARG_DESC,
                     OPTION_PHY_INFO_RESET_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_PCI_INFO_GET, OPTION_PCI_INFO_GET_SHORT,
                     OPTION_PCI_INFO_GET_ARG_DESC, OPTION_PCI_INFO_GET_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_PCI_INFO_RESET, OPTION_PCI_INFO_RESET_SHORT,
                     OPTION_PCI_INFO_RESET_ARG_DESC,
                     OPTION_PCI_INFO_RESET_DESC, OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);

    def_val_sstream << BER_THRESHOLD_ERROR;
    def_val_string = def_val_sstream.str();

    this->AddOptions(OPTION_BER_THRESHOLD_ERROR,
                     OPTION_BER_THRESHOLD_ERROR_SHORT,
                     OPTION_BER_THRESHOLD_ERROR_ARG_DESC,
                     OPTION_BER_THRESHOLD_ERROR_DESC,
                     def_val_string);

    def_val_sstream.str("");
    def_val_sstream << BER_THRESHOLD_WARNING;
    def_val_string = def_val_sstream.str();
    this->AddOptions(OPTION_BER_THRESHOLD_WARNING,
                     OPTION_BER_THRESHOLD_WARNING_SHORT,
                     OPTION_BER_THRESHOLD_WARNING_ARG_DESC,
                     OPTION_BER_THRESHOLD_WARNING_DESC,
                     def_val_string);

    this->AddDescription(PLUGIN_DESC);

    this->p_discovered_fabric = this->p_ibdiag->GetDiscoverFabricPtr();
    this->p_ibis_obj = this->p_ibdiag->GetIbisPtr();
    this->p_ibdm_extended_info = this->p_ibdiag->GetIBDMExtendedInfoPtr();
    this->p_capability_module = this->p_ibdiag->GetCapabilityModulePtr();

    phy_errors.clear();

    IBDIAGNET_RETURN_VOID;
}

PhyDiag::~PhyDiag()
{
    IBDIAGNET_ENTER;

    for (vec_diagnostic_data_t::iterator it = diagnostic_data_vec.begin();
         it != diagnostic_data_vec.end(); it++) {
        delete (*it);
    }

    for (vec_pareg::iterator it = reg_handlers_vec.begin();
            it != reg_handlers_vec.end(); it++) {
        delete (*it);
    }

    for (vec_pareg::iterator it = pci_reg_handlers_vec.begin();
            it != pci_reg_handlers_vec.end(); it++) {
        delete (*it);
    }

    IBDIAGNET_RETURN_VOID;
}

int PhyDiag::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    if (check_if_can_send_mads_by_lid(this->p_ibdiag,
                                      this->can_send_mads_by_lid) ||
        !this->can_send_mads_by_lid) {
        INFO_PRINT("%s skipped\n",
                    CHECK_NAME_PHY_CNTRS_RETRIEVE);
        PRINT("\n");
    }

    //reset the app of the nodes, to track supported capabilities
    //related to this plugin
    this->p_ibdiag->ResetAppData(true);
    IBDIAGNET_RETURN(0);
}

int PhyDiag::RetrieveInfo()
{
    IBDIAGNET_ENTER;

    int rc = 0;
    int rc2;

    if (this->to_get_phy_info) {
        if (this->can_send_mads_by_lid) {

            // PPCNT Phy counters
            DiagnosticDataInfo * p_dd_phys_layer_cntrs = new DiagnosticDataPhysLayerCntrs();
            this->diagnostic_data_vec.push_back(p_dd_phys_layer_cntrs);

            // PDDR Troubleshooting Info
            DiagnosticDataInfo * p_dd_troubleshooting_info = new DiagnosticDataTroubleshootingInfo();
            this->diagnostic_data_vec.push_back(p_dd_troubleshooting_info);

            // PDDR Operation Info
            DiagnosticDataInfo * p_dd_operation_info = new DiagnosticDataOperationInfo();
            this->diagnostic_data_vec.push_back(p_dd_operation_info);

            // PDDR Phy Info
            DiagnosticDataInfo * p_dd_phy_info = new DiagnosticDataPhyInfo();
            this->diagnostic_data_vec.push_back(p_dd_phy_info);

            // PDDR Module Info
            DiagnosticDataInfo * p_dd_module_info = new DiagnosticDataModuleInfo();
            this->diagnostic_data_vec.push_back(p_dd_module_info);

            // PDDR Link Down Info
            DiagnosticDataInfo * p_dd_link_down_info = new DiagnosticDataLinkDownInfo();
            this->diagnostic_data_vec.push_back(p_dd_link_down_info);

            // PPCNT RS Histograms
            DiagnosticDataInfo * p_dd_rs_histograms = new DiagnosticDataRSHistograms();
            this->diagnostic_data_vec.push_back(p_dd_rs_histograms);

            // PPCNT PLR Counters
            DiagnosticDataInfo * p_dd_plr_counters = new DiagnosticDataPLRCounters();
            this->diagnostic_data_vec.push_back(p_dd_plr_counters);

            // PPCNT Phy Statistics
            DiagnosticDataInfo * p_dd_phy_statistics = new DiagnosticDataPhyStatistics();
            this->diagnostic_data_vec.push_back(p_dd_phy_statistics);

            // PPCNT ZL FEC Counters
            DiagnosticDataInfo * p_dd_zl_fec_counters = new DiagnosticDataZLFECCounters();
            this->diagnostic_data_vec.push_back(p_dd_zl_fec_counters);

            if (this->to_reset_counters) {
                int rc2  = this->ResetPhyCounters(this->phy_errors,
                                                  progress_bar_reset_ports,
                                                  DD_PHY_TYPE);
                SCREEN_PRINT("\n");
                rc = this->AnalyzeCheckResults(this->phy_errors,
                                               CHECK_NAME_PHY_CNTRS_RESET((string)SECTION_PHY_CNTRS),
                                               rc2,
                                               IBDIAG_ERR_CODE_FABRIC_ERROR,
                                               this->num_errors,
                                               this->num_warnings,
                                               true);
            }
            if (rc)
                IBDIAGNET_RETURN(rc);

            for (u_int32_t dd_idx = 0; dd_idx < this->diagnostic_data_vec.size(); ++dd_idx) {

                DiagnosticDataInfo * p_dd = this->diagnostic_data_vec[dd_idx];
                rc2 = this->BuildPhyCountersDB(this->phy_errors,
                                               progress_bar_retrieve_nodes,
                                               DD_PHY_TYPE,
                                               dd_idx);
                SCREEN_PRINT("\n");
                rc = this->AnalyzeCheckResults(this->phy_errors,
                                               CHECK_NAME_RETRIEVE(p_dd->GetSectionHeader()),
                                               rc2,
                                               IBDIAG_ERR_CODE_FABRIC_ERROR,
                                               this->num_errors,
                                               this->num_warnings,
                                               true);
                if (rc)
                    IBDIAGNET_RETURN(rc);
            }
        }


        Register *p_reg_slrg = new SLRGRegister();
        AccRegHandler *p_areg_slrg = new AccRegPortLaneHandler(p_reg_slrg, this);

        this->reg_handlers_vec.push_back(p_areg_slrg);

        Register *p_reg_slrp = new SLRPRegister();
        AccRegHandler *p_areg_slrp = new AccRegPortLaneHandler(p_reg_slrp, this);

        this->reg_handlers_vec.push_back(p_areg_slrp);

        Register *p_reg_sltp = new SLTPRegister();
        AccRegHandler *p_areg_sltp = new AccRegPortLaneHandler(p_reg_sltp, this);

        this->reg_handlers_vec.push_back(p_areg_sltp);

        Register *p_reg_ptas = new PTASRegister();
        AccRegHandler *p_areg_ptas = new AccRegNodeHandler(p_reg_ptas, this);

        this->reg_handlers_vec.push_back(p_areg_ptas);
        Register *p_reg_ppll = new PPLLRegister();
        AccRegHandler *p_areg_ppll = new AccRegGroupHandler(p_reg_ppll, this);

        this->reg_handlers_vec.push_back(p_areg_ppll);

        // fans registers
        Register *p_reg_mfcr = new MFCRRegister();
        AccRegHandler *p_areg_mfcr = new AccRegNodeHandler(p_reg_mfcr, this);
        this->reg_handlers_vec.push_back(p_areg_mfcr);

        Register *p_reg_mfsm = new MFSMRegister();
        AccRegHandler *p_areg_mfsm = new AccRegNodeSensorHandler(p_reg_mfsm, this, p_areg_mfcr);
        this->reg_handlers_vec.push_back(p_areg_mfsm);

        Register *p_reg_mfsl = new MFSLRegister();
        AccRegHandler *p_areg_mfsl = new AccRegNodeSensorHandler(p_reg_mfsl, this, p_areg_mfcr);
        this->reg_handlers_vec.push_back(p_areg_mfsl);

        Register *p_reg_fore = new FORERegister();
        AccRegHandler *p_areg_fore = new AccRegNodeHandler(p_reg_fore, this);
        this->reg_handlers_vec.push_back(p_areg_fore);

        Register *p_reg_mfnr = new MFNRRegister();
        AccRegHandler *p_areg_mfnr = new AccRegNodeSensorHandler(p_reg_mfnr, this, p_areg_mfcr);
        this->reg_handlers_vec.push_back(p_areg_mfnr);

        // temperature registers
        Register *p_reg_mtcap = new MTCAPRegister();
        AccRegHandler *p_areg_mtcap = new AccRegNodeHandler(p_reg_mtcap, this);
        this->reg_handlers_vec.push_back(p_areg_mtcap);

        Register *p_reg_mtmp = new MTMPRegister();
        AccRegHandler *p_areg_mtmp = new AccRegNodeSensorHandler(p_reg_mtmp, this, p_areg_mtcap);
        this->reg_handlers_vec.push_back(p_areg_mtmp);

        Register *p_reg_mtwe = new MTWERegister();
        AccRegHandler *p_areg_mtwe = new AccRegNodeHandler(p_reg_mtwe, this);
        this->reg_handlers_vec.push_back(p_areg_mtwe);

        // current/voltage register
        Register *p_reg_mvcap = new MVCAPRegister();
        AccRegHandler *p_areg_mvcap = new AccRegNodeHandler(p_reg_mvcap, this);
        this->reg_handlers_vec.push_back(p_areg_mvcap);

        Register *p_reg_mvcr = new MVCRRegister();
        AccRegHandler *p_areg_mvcr = new AccRegNodeSensorHandler(p_reg_mvcr, this, p_areg_mvcap);
        this->reg_handlers_vec.push_back(p_areg_mvcr);

        // power supplies register
        Register *p_reg_msps = new MSPSRegister();
        AccRegHandler *p_areg_msps = new AccRegNodeHandler(p_reg_msps, this);
        this->reg_handlers_vec.push_back(p_areg_msps);

        // System General Information register
        Register *p_reg_msgi = new MSGIRegister();
        AccRegHandler *p_areg_msgi = new AccRegNodeHandler(p_reg_msgi, this);
        this->reg_handlers_vec.push_back(p_areg_msgi);

        // NEW HDR registers
        //slrip
        Register *p_reg_slrip = new SLRIPRegister();
        AccRegHandler *p_areg_slrip = new AccRegPortLaneHandler(p_reg_slrip, this, SLRIP_IB_SEL_NUM);
        this->reg_handlers_vec.push_back(p_areg_slrip);
        //slreg
        Register *p_reg_slreg = new SLREGRegister();
        AccRegHandler *p_areg_slreg = new AccRegPortLaneHandler(p_reg_slreg, this, SLREG_EIDX_NUM);
        this->reg_handlers_vec.push_back(p_areg_slreg);
        //slcct
        Register *p_reg_slcct = new SLCCTRegister();
        AccRegHandler *p_areg_slcct = new AccRegPortLaneHandler(p_reg_slcct, this, SLCCT_CYCLES_NUM);
        this->reg_handlers_vec.push_back(p_areg_slcct);

        if (this->to_get_ppamp) {
            Register *p_reg_ppamp = new PPAMPRegister();
            AccRegHandler *p_areg_ppamp = new AccRegTypeGroupHandler(p_reg_ppamp, this,
                    MAX_OPAMP_GROUP_TYPE, PPAMP_IDX_DATA_SIZE);
            this->reg_handlers_vec.push_back(p_areg_ppamp);
        }


        for (uint32_t i = 0; i < this->reg_handlers_vec.size(); ++i) {
            AccRegHandler *p_areg_handler = this->reg_handlers_vec[i];
            rc2 = p_areg_handler->BuildDB(this->phy_errors,progress_bar_retrieve_nodes);

            SCREEN_PRINT("\n");
            rc = this->AnalyzeCheckResults(this->phy_errors,
                                           CHECK_NAME_RETRIEVE(p_areg_handler->
                                                               GetPReg()->
                                                               GetSectionName()),
                                           rc2,
                                           IBDIAG_ERR_CODE_FABRIC_ERROR,
                                           this->num_errors,
                                           this->num_warnings,
                                           true);
        }
    }
    //TODO chech if rc !+0 , if yes return

    if (this->to_get_pci_info) {

        Register *p_reg_mpein = new MPEINRegister();
        AccRegHandler *p_areg_mpein = new AccRegSpecificHandler(p_reg_mpein, this);
        this->pci_reg_handlers_vec.push_back(p_areg_mpein);

        Register *p_reg_mpir = new MPIRRegister(p_areg_mpein->data_map);
        AccRegHandler *p_areg_mpir = new AccRegSpecificHandler(p_reg_mpir, this);
        this->pci_reg_handlers_vec.push_back(p_areg_mpir);

        Register *p_reg_slrg = new SLRGRegister(ACC_REG_PNAT_OOB_PORT,
                                                ACC_REG_SLRG_PCI_INTERNAL_SECTION_NAME,
                                                &p_areg_mpein->data_map,
                                                &p_areg_mpir->data_map);
        AccRegHandler *p_areg_slrg = new AccRegSpecificHandler(p_reg_slrg, this);
        this->pci_reg_handlers_vec.push_back(p_areg_slrg);

        Register *p_reg_slrp = new SLRPRegister(ACC_REG_PNAT_OOB_PORT,
                                                ACC_REG_SLRP_PCI_INTERNAL_SECTION_NAME,
                                                &p_areg_mpein->data_map,
                                                &p_areg_mpir->data_map);
        AccRegHandler *p_areg_slrp = new AccRegSpecificHandler(p_reg_slrp, this);
        this->pci_reg_handlers_vec.push_back(p_areg_slrp);

        Register *p_reg_sltp = new SLTPRegister(ACC_REG_PNAT_OOB_PORT,
                                                ACC_REG_SLTP_PCI_INTERNAL_SECTION_NAME,
                                                &p_areg_mpein->data_map,
                                                &p_areg_mpir->data_map);
        AccRegHandler *p_areg_sltp = new AccRegSpecificHandler(p_reg_sltp, this);
        this->pci_reg_handlers_vec.push_back(p_areg_sltp);

        for (uint32_t i = 0; i < this->pci_reg_handlers_vec.size(); ++i) {
            AccRegHandler *p_areg_handler = this->pci_reg_handlers_vec[i];
            rc2 = p_areg_handler->BuildDB(this->phy_errors,progress_bar_retrieve_nodes);

            SCREEN_PRINT("\n");
            rc = this->AnalyzeCheckResults(this->phy_errors,
                                           CHECK_NAME_RETRIEVE(p_areg_handler->
                                                               GetPReg()->
                                                               GetSectionName()),
                                           rc2,
                                           IBDIAG_ERR_CODE_FABRIC_ERROR,
                                           this->num_errors,
                                           this->num_warnings,
                                           true);
        }

        if (this->can_send_mads_by_lid) {
            DiagnosticDataInfo * p_dd_pcime = new DiagnosticDataPCIECntrs();
            this->diagnostic_data_vec.push_back(p_dd_pcime);

            if (this->to_reset_pci_counters) {
                int rc2  = this->ResetPCICounters(this->phy_errors,
                                                  progress_bar_reset_ports,
                                                  DD_PCI_TYPE,
                                                  &p_areg_mpein->data_map);
                SCREEN_PRINT("\n");
                rc = this->AnalyzeCheckResults(this->phy_errors,
                                               CHECK_NAME_PHY_CNTRS_RESET((string)SECTION_PCI_CNTRS),
                                               rc2,
                                               IBDIAG_ERR_CODE_FABRIC_ERROR,
                                               this->num_errors,
                                               this->num_warnings,
                                               true);
            }
            if (rc)
                IBDIAGNET_RETURN(rc);

            for (u_int32_t dd_idx = 0; dd_idx < this->diagnostic_data_vec.size(); ++dd_idx) {

                DiagnosticDataInfo * p_dd = this->diagnostic_data_vec[dd_idx];
                rc2 = this->BuildPCICountersDB(this->phy_errors,
                                               progress_bar_retrieve_nodes,
                                               DD_PCI_TYPE,
                                               dd_idx,
                                               &p_areg_mpein->data_map);

                SCREEN_PRINT("\n");
                rc = this->AnalyzeCheckResults(this->phy_errors,
                                               CHECK_NAME_RETRIEVE(p_dd->GetSectionHeader()),
                                               rc2,
                                               IBDIAG_ERR_CODE_FABRIC_ERROR,
                                               this->num_errors,
                                               this->num_warnings,
                                               true);
                if (rc)
                    IBDIAGNET_RETURN(rc);
            }
        }
    }

    IBDIAGNET_RETURN(rc);
}

//Returns: 0 - OK / 1 - parsing error / 2 - OK but need to exit / 3 - some other error
int PhyDiag::HandleOption(string name, string value)
{
    IBDIAGNET_ENTER;

    bool bool_flag_val = true;

    if (value == OPTION_DEF_VAL_NULL)
        IBDIAGNET_RETURN(0);

    if (name == OPTION_PHY_INFO_GET) {
        this->ParseBoolValue(value, bool_flag_val);
        this->SetPhyInfoGet(bool_flag_val);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_PPAMP_GET) {
        this->ParseBoolValue(value, bool_flag_val);
        this->SetPPAMPGet(bool_flag_val);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_SHOW_CAP_REG) {
        this->ParseBoolValue(value, bool_flag_val);
        this->SetCapRegShow(bool_flag_val);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_PHY_INFO_RESET) {
        this->ParseBoolValue(value, bool_flag_val);
        this->SetPhyInfoGet(bool_flag_val);
        this->SetPhyInfoClear(bool_flag_val);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_PCI_INFO_GET) {
        this->ParseBoolValue(value, bool_flag_val);
        this->SetPciInfoGet(bool_flag_val);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_PCI_INFO_RESET) {
        this->ParseBoolValue(value, bool_flag_val);
        this->SetPciInfoClear(bool_flag_val);
        this->SetPciInfoGet(bool_flag_val);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_BER_THRESHOLD_ERROR) {
        ber_threshold_err = strtod(value.c_str(), NULL);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_BER_THRESHOLD_WARNING) {
        ber_threshold_warn = strtod(value.c_str(), NULL);
        IBDIAGNET_RETURN(0);
    }

    //if we reached here than unknown option
    IBDIAGNET_RETURN(1);
}

int PhyDiag::HandleSpecialPorts(IBNode *p_curr_node,
                                IBPort *p_curr_port,
                                u_int32_t i)
{
    IBDIAGNET_ENTER;

    if (!this->p_capability_module->IsSupportedSMPCapability(
                p_curr_node, EnSMPCapIsSpecialPortsMarkingSupported)) {

        // if not checked yet for this node
        if (!(p_curr_node->appData2.val & NOT_SUPPORT_SPECIAL_PORTS_MARKING_CHECKED)) {

            p_curr_node->appData2.val |= NOT_SUPPORT_SPECIAL_PORTS_MARKING_CHECKED;

            FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                new FabricErrNodeNotSupportCap(p_curr_node,
                    "This device does not support Special Ports Marking capability");

            if (!p_curr_fabric_node_err) {
                this->SetLastError("Failed to allocate FabricErrNodeNotSupportCap");
                IBDIAGNET_RETURN(IBDIAG_ERR_CODE_NO_MEM);
            }

            p_curr_fabric_node_err->SetLevel(EN_FABRIC_ERR_WARNING);
            phy_errors.push_back(p_curr_fabric_node_err);
        }

    } else if (p_curr_port->special_port_sup_diag_data) { // Special Ports Marking Supported

        struct SMP_MlnxExtPortInfo *p_curr_mepi = this->p_ibdm_extended_info->
                getSMPMlnxExtPortInfo(p_curr_port->createIndex);
        if (!p_curr_mepi) {
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR,
                       "Error DB - Can't find MlnxExtPortInfo for " \
                       "node=%s, port=%u\n",
                       p_curr_node->name.c_str(), i);
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (p_curr_mepi->IsSpecialPort &&
                !(p_curr_mepi->SpecialPortCapabilityMask &
                        (1 << EnSPCapIsDiagnosticDataSupported))) {

                p_curr_port->special_port_sup_diag_data = false;

                stringstream ss;
                ss << "This special port does not support DiagnosticData MAD."
                   << " type = "
                   << +p_curr_mepi->SpecialPortType;

                FabricErrPortNotSupportCap *p_curr_fabric_port_err =
                    new FabricErrPortNotSupportCap(p_curr_port, ss.str());

                if (!p_curr_fabric_port_err) {
                    SetLastError("Failed to allocate FabricErrPortNotSupportCap");
                    IBDIAGNET_RETURN(IBDIAG_ERR_CODE_NO_MEM);
                }

                p_curr_fabric_port_err->SetLevel(EN_FABRIC_ERR_WARNING);
                phy_errors.push_back(p_curr_fabric_port_err);
                IBDIAGNET_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
        }
    } else { // this special port does not support DiagnosticData
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

/*
 * private methods
 */
//Returns: IBDIAGNET_SUCCESS_CODE / IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS /
//IBDIAGNET_ERR_CODE_DB_ERR / IBDIAGNET_ERR_CODE_NO_MEM /
//IBDIAGNET_ERR_CODE_FABRIC_ERROR
int PhyDiag::ResetPhyCounters(list_p_fabric_general_err &phy_errors,
                              progress_func_ports_t progress_func,
                              u_int32_t dd_type)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->p_ibdiag->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    progress_bar_ports_t progress_bar_ports;
    CLEAR_STRUCT(progress_bar_ports);

    clbck_data_t clbck_data;
    clbck_data.m_handle_data_func = PhyCountersResetDelegator;
    clbck_data.m_p_obj = this;

    struct VS_DiagnosticData p_cntrs;
    for (u_int32_t dd_idx = 0; dd_idx < this->diagnostic_data_vec.size();
         ++dd_idx) {
        DiagnosticDataInfo * p_dd = this->diagnostic_data_vec[dd_idx];

        if (p_dd->GetDDType() != dd_type)
            continue;

        clbck_data.m_data1 = (void *)(uintptr_t)dd_idx;
        clbck_data.m_data2 = p_dd;

        for (map_str_pnode::iterator nI =
             this->p_discovered_fabric->NodeByName.begin();
             nI != this->p_discovered_fabric->NodeByName.end();
             ++nI) {
            IBNode *p_curr_node = (*nI).second;
            if (!p_curr_node) {
                this->SetLastError("DB error - found null node in NodeByName map"  \
                                   " for key = %s", (*nI).first.c_str());
                rc = IBDIAG_ERR_CODE_DB_ERR;
                goto exit;
            }

            if (!IsSupportDiagnosticDataCountersFailed(p_curr_node->appData1.val,
                                                      p_dd->GetNotSupportedBit())) {
                if (!this->p_capability_module->IsSupportedGMPCapability(
                        p_curr_node, EnGMPCapIsDiagnosticDataSupported)) {
                    p_curr_node->appData1.val |= NOT_SUPPORT_DIAGNOSTIC_DATA;
                    FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                            new FabricErrNodeNotSupportCap(p_curr_node,
                                    "This device does not support diagnostic data MAD capability");
                    if (!p_curr_fabric_node_err) {
                        this->SetLastError("Failed to allocate FabricErrNodeNotSupportCap");
                        rc = IBDIAG_ERR_CODE_NO_MEM;
                        goto exit;
                    }
                    phy_errors.push_back(p_curr_fabric_node_err);
                    rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
                    continue;
                }
            } else
                continue;

            // go over all ports of this node:
            for (u_int32_t i = 1; i < (unsigned)p_curr_node->numPorts + 1; ++i) {
                IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
                //Skip ports that don't exist or are down or have undefined state
                //or are not in the scope of the fabric.
                if (!p_curr_port || !p_curr_port->is_data_worthy())
                    continue;

                // special ports
                if (p_curr_node->type == IB_SW_NODE) {
                    rc = HandleSpecialPorts(p_curr_node, p_curr_port, i);
                    switch (rc) {
                        case IBDIAG_ERR_CODE_NO_MEM: goto exit;
                        case IBDIAG_ERR_CODE_DB_ERR: IBDIAGNET_RETURN(rc);
                        case IBDIAG_ERR_CODE_FABRIC_ERROR: continue;
                    }
                }

                //deals with progress bar
                ++progress_bar_ports.ports_found;
                if (progress_func)
                    progress_func(&progress_bar_ports,
                            this->p_ibdiag->GetDiscoverProgressBarPortsPtr());

                clbck_data.m_data1 = p_curr_port;
                this->p_ibis_obj->VSDiagnosticDataPageClear(
                    p_curr_port->base_lid,
                    p_dd->IsPerNode() ? 0 : p_curr_port->num,
                    p_dd->GetPageId(),
                    &p_cntrs,
                    &clbck_data);

                if (clbck_error_state)
                    goto exit;
                if (p_dd->IsPerNode())
                    break;
            }
        }
    }
exit:
    this->p_ibis_obj->MadRecAll();

    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAGNET_RETURN(rc);
}

//Returns: IBDIAGNET_SUCCESS_CODE / IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS /
//IBDIAGNET_ERR_CODE_DB_ERR / IBDIAGNET_ERR_CODE_NO_MEM /
//IBDIAGNET_ERR_CODE_FABRIC_ERROR
int PhyDiag::ResetPCICounters(list_p_fabric_general_err &phy_errors,
                              progress_func_ports_t progress_func,
                              u_int32_t dd_type,
                              map_akey_areg *data_map)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->p_ibdiag->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    progress_bar_ports_t progress_bar_ports;
    CLEAR_STRUCT(progress_bar_ports);

    clbck_data_t clbck_data;
    clbck_data.m_handle_data_func = PhyCountersResetDelegator;
    clbck_data.m_p_obj = this;

    struct VS_DiagnosticData pci_cntrs;
    for (u_int32_t dd_idx = 0; dd_idx < this->diagnostic_data_vec.size();
         ++dd_idx) {
         DiagnosticDataInfo * p_dd = this->diagnostic_data_vec[dd_idx];

         if (p_dd->GetDDType() != dd_type)
             continue;

         clbck_data.m_data1 = (void *)(uintptr_t)dd_idx;

         for (map_akey_areg::iterator iter = data_map->begin();
              iter != data_map->end(); ++iter) {
              AccRegKeyDPN *p_acc_reg_key_dpn = (AccRegKeyDPN *)iter->first;
              if (!p_acc_reg_key_dpn)
                  continue;

              IBNode *p_curr_node = this->p_discovered_fabric->
                                    getNodeByGuid(p_acc_reg_key_dpn->node_guid);

              if (!p_curr_node)
                  continue;

              if (!IsSupportDiagnosticDataCountersFailed(p_curr_node->appData1.val,
                                                        p_dd->GetNotSupportedBit())) {
                  if (!this->p_capability_module->IsSupportedGMPCapability(
                          p_curr_node, EnGMPCapIsDiagnosticDataSupported)) {
                      p_curr_node->appData1.val |= NOT_SUPPORT_DIAGNOSTIC_DATA;
                      FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                              new FabricErrNodeNotSupportCap(p_curr_node,
                                      "This device does not support diagnostic data MAD capability");
                      if (!p_curr_fabric_node_err) {
                          this->SetLastError("Failed to allocate FabricErrNodeNotSupportCap");
                          rc = IBDIAG_ERR_CODE_NO_MEM;
                          goto exit;
                      }
                      phy_errors.push_back(p_curr_fabric_node_err);
                      rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
                      continue;
                  }
              } else
                 continue;

              IBPort *p_curr_port = NULL;

              // go over all ports of this node:
              for (u_int32_t i = 1; i < (unsigned)p_curr_node->numPorts + 1; ++i) {
                  p_curr_port = p_curr_node->getPort((phys_port_t)i);
                  //Skip ports that don't exist or are down or have undefined state
                  //or are not in the scope of the fabric.
                  if (!p_curr_port || !p_curr_port->is_data_worthy())
                      continue;

                  break;
              }

              if (!p_curr_port)
                  continue;

              //deals with progress bar
              ++progress_bar_ports.ports_found;
              if (progress_func)
                  progress_func(&progress_bar_ports,
                          this->p_ibdiag->GetDiscoverProgressBarPortsPtr());

              clbck_data.m_data1 = p_curr_port;
              clbck_data.m_data2 = p_acc_reg_key_dpn;
              u_int8_t page_number = (u_int8_t)p_dd->GetPageId();
              u_int32_t attribute_modifier = 0;
              attribute_modifier = ((p_acc_reg_key_dpn->depth & 0x3f) << 24) | (page_number << 16)
                                    | (p_acc_reg_key_dpn->pci_node << 8) | p_acc_reg_key_dpn->pci_idx;

              this->p_ibis_obj->VSDiagnosticDataPageClear_AM(p_curr_port->base_lid,
                                                             attribute_modifier,
                                                             &pci_cntrs,
                                                             &clbck_data);

              if (clbck_error_state)
                  goto exit;
        }
    }
exit:
    this->p_ibis_obj->MadRecAll();

    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAGNET_RETURN(rc);
}

//Returns: IBDIAGNET_SUCCESS_CODE / IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS /
//IBDIAGNET_ERR_CODE_DB_ERR / IBDIAGNET_ERR_CODE_NO_MEM /
//IBDIAGNET_ERR_CODE_FABRIC_ERROR
int PhyDiag::BuildPhyCountersDB(list_p_fabric_general_err &phy_errors,
                                progress_func_nodes_t progress_func,
                                u_int32_t dd_type, u_int32_t dd_idx)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->p_ibdiag->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    progress_bar_nodes_t progress_bar_nodes;
    CLEAR_STRUCT(progress_bar_nodes);

    clbck_data_t clbck_data;
    clbck_data.m_handle_data_func = PhyCountersGetDelegator;
    clbck_data.m_p_obj = this;


    DiagnosticDataInfo * p_dd = this->diagnostic_data_vec[dd_idx];

    if (p_dd->GetDDType() != dd_type)
        IBDIAGNET_RETURN(rc);

    clbck_data.m_data1 = (void *)(uintptr_t)dd_idx;
    clbck_data.m_data2 = p_dd;

    for (map_str_pnode::iterator nI = this->p_discovered_fabric->NodeByName.begin();
            nI != this->p_discovered_fabric->NodeByName.end(); ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_dd->IsDDPageSupportNodeType(p_curr_node->type))
            continue;

        ++progress_bar_nodes.nodes_found;
        if (p_curr_node->type == IB_SW_NODE)
            ++progress_bar_nodes.sw_found;
        else
            ++progress_bar_nodes.ca_found;
        if (progress_func)
            progress_func(&progress_bar_nodes,
                          this->p_ibdiag->GetDiscoverProgressBarNodesPtr());

        if (!IsSupportDiagnosticDataCountersFailed(p_curr_node->appData1.val,
                                                  p_dd->GetNotSupportedBit())) {
            if (!this->p_capability_module->IsSupportedGMPCapability(
                    p_curr_node, EnGMPCapIsDiagnosticDataSupported)) {
                p_curr_node->appData1.val |= NOT_SUPPORT_DIAGNOSTIC_DATA;
                FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                        new FabricErrNodeNotSupportCap(p_curr_node,
                                "This device does not support diagnostic data MAD capability");
                if (!p_curr_fabric_node_err) {
                    this->SetLastError("Failed to allocate FabricErrNodeNotSupportCap");
                    rc = IBDIAG_ERR_CODE_NO_MEM;
                    goto exit;
                }
                phy_errors.push_back(p_curr_fabric_node_err);
                rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
                continue;
            }
        } else
            continue;


        // go over all ports of this node:
        for (u_int32_t i = 1; i < (unsigned)p_curr_node->numPorts + 1; ++i) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            if (!p_curr_port)                       //port not exist
                continue;
            //Skip ports that are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port->is_data_worthy())
                continue;

            // special ports
            if (p_curr_node->type == IB_SW_NODE) {
                rc = HandleSpecialPorts(p_curr_node, p_curr_port, i);
                switch (rc) {
                    case IBDIAG_ERR_CODE_NO_MEM: goto exit;
                    case IBDIAG_ERR_CODE_DB_ERR: IBDIAGNET_RETURN(rc);
                    case IBDIAG_ERR_CODE_FABRIC_ERROR: continue;
                }
            }

            clbck_data.m_data3 = p_curr_port;
            struct VS_DiagnosticData phys_cntrs;

            //get phys' layer counters
            this->p_ibis_obj->VSDiagnosticDataGet(p_curr_port->base_lid,
                                                  p_dd->IsPerNode() ? 0 : p_curr_port->num,
                                                  p_dd->GetPageId(),
                                                  &phys_cntrs,
                                                  &clbck_data);
            if (clbck_error_state)
                goto exit;
            if (p_dd->IsPerNode())
                break;
        }
    }

exit:
    this->p_ibis_obj->MadRecAll();

    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAGNET_RETURN(rc);
}


//Returns: IBDIAGNET_SUCCESS_CODE / IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS /
//IBDIAGNET_ERR_CODE_DB_ERR / IBDIAGNET_ERR_CODE_NO_MEM /
//IBDIAGNET_ERR_CODE_FABRIC_ERROR
int PhyDiag::BuildPCICountersDB(list_p_fabric_general_err &phy_errors,
                                progress_func_nodes_t progress_func,
                                u_int32_t dd_type,
                                u_int32_t dd_idx,
                                map_akey_areg *data_map)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->p_ibdiag->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    progress_bar_nodes_t progress_bar_nodes;
    CLEAR_STRUCT(progress_bar_nodes);

    clbck_data_t clbck_data;
    clbck_data.m_handle_data_func = PCICountersGetDelegator;
    clbck_data.m_p_obj = this;

    DiagnosticDataInfo * p_dd = this->diagnostic_data_vec[dd_idx];

    if (p_dd->GetDDType() != dd_type)
        IBDIAGNET_RETURN(rc);

    clbck_data.m_data1 = (void *)(uintptr_t)dd_idx;

    for (map_akey_areg::iterator iter = data_map->begin();
            iter != data_map->end(); ++iter) {
         AccRegKeyDPN *p_acc_reg_key_dpn = (AccRegKeyDPN *)iter->first;
         if (!p_acc_reg_key_dpn)
             continue;

         IBNode *p_curr_node = this->p_discovered_fabric->
                               getNodeByGuid(p_acc_reg_key_dpn->node_guid);
         if (!p_curr_node)
             continue;

         ++progress_bar_nodes.nodes_found;
         if (p_curr_node->type == IB_SW_NODE)
             ++progress_bar_nodes.sw_found;
         else
             ++progress_bar_nodes.ca_found;
         if (progress_func)
             progress_func(&progress_bar_nodes,
                           this->p_ibdiag->GetDiscoverProgressBarNodesPtr());

         if (!IsSupportDiagnosticDataCountersFailed(p_curr_node->appData1.val,
                                                   p_dd->GetNotSupportedBit())) {
             if (!this->p_capability_module->IsSupportedGMPCapability(
                     p_curr_node, EnGMPCapIsDiagnosticDataSupported)) {
                 p_curr_node->appData1.val |= NOT_SUPPORT_DIAGNOSTIC_DATA;
                 FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                         new FabricErrNodeNotSupportCap(p_curr_node,
                                 "This device does not support diagnostic data MAD capability");
                 if (!p_curr_fabric_node_err) {
                     this->SetLastError("Failed to allocate FabricErrNodeNotSupportCap");
                     rc = IBDIAG_ERR_CODE_NO_MEM;
                     goto exit;
                 }
                 phy_errors.push_back(p_curr_fabric_node_err);
                 rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
                 continue;
             }
         } else
             continue;

        IBPort *p_curr_port = NULL;

        //get the first valid active port of this node:
        for (u_int32_t i = 1; i < (unsigned)p_curr_node->numPorts + 1; ++i) {
            p_curr_port = p_curr_node->getPort((phys_port_t)i);
            if (!p_curr_port)                       //port not exist
                continue;
            //Skip ports that are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port->is_data_worthy())
                continue;

            break;
         }

         if (!p_curr_port)
             continue;

         clbck_data.m_data2 = p_acc_reg_key_dpn;
         clbck_data.m_data3 = p_curr_port;
         struct VS_DiagnosticData pci_cntrs;
         u_int8_t page_number = (u_int8_t)p_dd->GetPageId();
         u_int32_t attribute_modifier = 0;
         attribute_modifier = ((p_acc_reg_key_dpn->depth & 0x3f) << 24) | (page_number << 16)
                               | (p_acc_reg_key_dpn->pci_node << 8) | p_acc_reg_key_dpn->pci_idx;

         //get phys' layer counters
         this->p_ibis_obj->VSDiagnosticDataGet_AM(p_curr_port->base_lid,
                                                  attribute_modifier,
                                                  &pci_cntrs,
                                                  &clbck_data);
         if (clbck_error_state)
             goto exit;
    }

exit:
    this->p_ibis_obj->MadRecAll();

    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAGNET_RETURN(rc);
}

void PhyCountersResetDelegator(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<PhyDiag*>(clbck_data.m_p_obj)->
        PhyCountersResetClbck(clbck_data, rec_status, p_attribute_data);
}

void PhyDiag::PhyCountersResetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAGNET_ENTER;
    if (clbck_error_state || !p_ibdiag)
        return;

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    IBPort *p_port = (IBPort *)clbck_data.m_data1;
    if (rec_status) {
        p_port->p_node->appData1.val |= NOT_SUPPORT_DIAGNOSTIC_DATA;
        FabricErrPortNotRespond *p_fabric_pm_err =
                new FabricErrPortNotRespond(p_port, "VSDiagnosticCountersClear");
        if (!p_fabric_pm_err) {
            SetLastError("Failed to allocate FabricErrPortNotRespond");
            clbck_error_state = IBDIAG_ERR_CODE_NO_MEM;
        } else
            this->phy_errors.push_back(p_fabric_pm_err);
    }
    IBDIAGNET_RETURN_VOID;
}

void PhyCountersGetDelegator(const clbck_data_t &clbck_data,
                             int rec_status,
                             void *p_attribute_data)
{
    reinterpret_cast<PhyDiag*>(clbck_data.m_p_obj)->
        PhyCountersGetClbck(clbck_data, rec_status, p_attribute_data);
}

void PhyDiag::PhyCountersGetClbck(const clbck_data_t &clbck_data,
                                  int rec_status,
                                  void *p_attribute_data)
{
    if (clbck_error_state || !p_ibdiag)
        return;

    rec_status &= 0x00ff;   //bits [8-15] are class specific - don't need them here
    IBPort *p_port = (IBPort *)clbck_data.m_data3;
    DiagnosticDataInfo *p_dd = (DiagnosticDataInfo *)clbck_data.m_data2;

    if (rec_status &&
        IsSupportDiagnosticDataCountersFailed(p_port->p_node->appData1.val,
                                              p_dd->GetNotSupportedBit()))
        IBDIAGNET_RETURN_VOID;

    if (rec_status == IBIS_MAD_STATUS_UNSUP_METHOD_ATTR) {         //not supported for this device
        p_port->p_node->appData1.val |= NOT_SUPPORT_DIAGNOSTIC_DATA;
        FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                new FabricErrNodeNotSupportCap(p_port->p_node,
                                               "The firmware of this device " \
                                               "does not support " \
                                               "VSDiagnosticData");
        if (!p_curr_fabric_node_err) {
            this->SetLastError("Failed to allocate FabricErrNodeNotSupportCap");
            clbck_error_state = IBDIAG_ERR_CODE_NO_MEM;
        } else
            this->phy_errors.push_back(p_curr_fabric_node_err);
        IBDIAGNET_RETURN_VOID;
    } else if (rec_status) {           //failed to get response
        p_port->p_node->appData1.val |= NOT_SUPPORT_DIAGNOSTIC_DATA;
        FabricErrPortNotRespond *p_curr_fabric_port_err =
                new FabricErrPortNotRespond(p_port, "VSDiagnosticDataGet");

        if (!p_curr_fabric_port_err) {
            this->SetLastError("Failed to allocate FabricErrPortNotRespond");
            clbck_error_state = IBDIAG_ERR_CODE_NO_MEM;
        } else
            this->phy_errors.push_back(p_curr_fabric_port_err);
        IBDIAGNET_RETURN_VOID;
    }

    struct VS_DiagnosticData *p_dd_struct =
        (struct VS_DiagnosticData *)p_attribute_data;

    //check if the requested page in DiagnosticDataInfo is unsupported
    if (p_dd_struct->CurrentRevision == 0 ||
        p_dd_struct->BackwardRevision > p_dd->GetSupportedVersion() ||
        p_dd->GetSupportedVersion() > p_dd_struct->CurrentRevision) {
        p_port->p_node->appData1.val |= p_dd->GetNotSupportedBit();
        string error_desc =
            "The firmware of this device does not support VSDiagnosticData" +
            p_dd->GetSectionHeader() +
            "Get";
        FabricErrNodeNotSupportCap *p_curr_phy_err =
                new FabricErrNodeNotSupportCap(p_port->p_node,
                                                error_desc);
        if (!p_curr_phy_err) {
            this->SetLastError("Failed to allocate FabricErrNodeNotSupportCap");
            clbck_error_state = IBDIAG_ERR_CODE_NO_MEM;
        } else
            this->phy_errors.push_back(p_curr_phy_err);
        IBDIAGNET_RETURN_VOID;
    }

    u_int32_t dd_idx = (u_int32_t)(intptr_t)clbck_data.m_data1;

    int rc = this->addPhysLayerPortCounters(p_port, *p_dd_struct, dd_idx);
    if (rc)
        //already set last error
        clbck_error_state = rc;

    IBDIAGNET_RETURN_VOID;
}

void PCICountersGetDelegator(const clbck_data_t &clbck_data,
                             int rec_status,
                             void *p_attribute_data)
{
    reinterpret_cast<PhyDiag*>(clbck_data.m_p_obj)->
        PCICountersGetClbck(clbck_data, rec_status, p_attribute_data);
}

void PhyDiag::PCICountersGetClbck(const clbck_data_t &clbck_data,
                                  int rec_status,
                                  void *p_attribute_data)
{
    if (clbck_error_state || !p_ibdiag)
        return;

    rec_status &= 0x00ff;   //bits [8-15] are class specific - don't need them here
    IBPort *p_port = (IBPort *)clbck_data.m_data3;
    u_int32_t dd_idx = (u_int32_t)(intptr_t)clbck_data.m_data1;
    DiagnosticDataInfo *p_dd = this->diagnostic_data_vec[dd_idx];
    AccRegKey *p_acc_reg_key = (AccRegKey *)clbck_data.m_data2;

    if (rec_status &&
        IsSupportDiagnosticDataCountersFailed(p_port->p_node->appData1.val,
                                              p_dd->GetNotSupportedBit()))
        IBDIAGNET_RETURN_VOID;

    if (rec_status == IBIS_MAD_STATUS_UNSUP_METHOD_ATTR) {         //not supported for this device
        p_port->p_node->appData1.val |= NOT_SUPPORT_DIAGNOSTIC_DATA;
        FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                new FabricErrNodeNotSupportCap(p_port->p_node,
                                               "The firmware of this device " \
                                               "does not support " \
                                               "VSDiagnosticData");
        if (!p_curr_fabric_node_err) {
            this->SetLastError("Failed to allocate FabricErrNodeNotSupportCap");
            clbck_error_state = IBDIAG_ERR_CODE_NO_MEM;
        } else
            this->phy_errors.push_back(p_curr_fabric_node_err);
        IBDIAGNET_RETURN_VOID;
    } else if (rec_status) {           //failed to get response
        p_port->p_node->appData1.val |= NOT_SUPPORT_DIAGNOSTIC_DATA;
        FabricErrPortNotRespond *p_curr_fabric_port_err =
                new FabricErrPortNotRespond(p_port, "VSDiagnosticData");

        if (!p_curr_fabric_port_err) {
            this->SetLastError("Failed to allocate FabricErrPortNotRespond");
            clbck_error_state = IBDIAG_ERR_CODE_NO_MEM;
        } else
            this->phy_errors.push_back(p_curr_fabric_port_err);
        IBDIAGNET_RETURN_VOID;
    }

    struct VS_DiagnosticData *p_dd_struct =
        (struct VS_DiagnosticData *)p_attribute_data;

    //check if the requested page in DiagnosticDataInfo is unsupported
    if (p_dd_struct->CurrentRevision == 0 ||
        p_dd_struct->BackwardRevision > p_dd->GetSupportedVersion() ||
        p_dd->GetSupportedVersion() > p_dd_struct->CurrentRevision) {
        p_port->p_node->appData1.val |= p_dd->GetNotSupportedBit();
        string error_desc =
            "The firmware of this device does not support VSDiagnosticData" +
            p_dd->GetSectionHeader() +
            "Get";
        FabricErrNodeNotSupportCap *p_curr_phy_err =
                new FabricErrNodeNotSupportCap(p_port->p_node,
                                                error_desc);
        if (!p_curr_phy_err) {
            this->SetLastError("Failed to allocate FabricErrNodeNotSupportCap");
            clbck_error_state = IBDIAG_ERR_CODE_NO_MEM;
        } else
            this->phy_errors.push_back(p_curr_phy_err);
        IBDIAGNET_RETURN_VOID;
    }

    int rc = this->addPhysLayerPCICounters(p_acc_reg_key, *p_dd_struct, dd_idx);
    if (rc)
        //already set last error
        clbck_error_state = rc;

    IBDIAGNET_RETURN_VOID;
}

IBNode* PhyDiag::getNodePtr(u_int32_t node_index)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVec<vector_p_node, IBNode>(
                            this->nodes_vector, node_index)));
}

IBPort* PhyDiag::getPortPtr(u_int32_t port_index)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVec<vector_p_port, IBPort>(
                            this->ports_vector, port_index)));
}

template <class OBJ_VEC_TYPE, class OBJ_TYPE>
void PhyDiag::addPtrToVec(OBJ_VEC_TYPE& vector_obj, OBJ_TYPE *p_obj)
{
    IBDIAG_ENTER;
    if ((vector_obj.size() > p_obj->createIndex + 1) &&
            vector_obj[p_obj->createIndex]) //if exist already
        IBDIAG_RETURN_VOID;
    if (vector_obj.empty() || (vector_obj.size() < p_obj->createIndex + 1))
        for (int i = (int)vector_obj.size(); i < (int)p_obj->createIndex + 1; ++i)
            vector_obj.push_back(NULL);
    vector_obj[p_obj->createIndex] = p_obj;
    IBDIAG_RETURN_VOID;
}

template <class OBJ_VEC_TYPE, class OBJ_TYPE, class DATA_VEC_TYPE, class DATA_TYPE>
int PhyDiag::addDataToVec(OBJ_VEC_TYPE& vector_obj,
        OBJ_TYPE *p_obj,
        DATA_VEC_TYPE& vector_data,
        DATA_TYPE& data)
{
    IBDIAG_ENTER;
    if (!p_obj)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    //if exist already
    if ((vector_data.size() >= p_obj->createIndex + 1) &&
            vector_data[p_obj->createIndex])
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    //add
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Adding %s for (%s, index=%u)\n",
            typeid(data).name(), p_obj->getName().c_str(), p_obj->createIndex);
    if (vector_data.empty() || (vector_data.size() < p_obj->createIndex + 1))
        for (int i = (int)vector_data.size(); i < (int)p_obj->createIndex + 1; ++i)
            vector_data.push_back(NULL);
    DATA_TYPE *p_curr_data = new DATA_TYPE;
    if (!p_curr_data) {
        this->SetLastError("Failed to allocate %s", typeid(data).name());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
    }
    *p_curr_data = data;
    vector_data[p_obj->createIndex] = p_curr_data;
    this->addPtrToVec(vector_obj, p_obj);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

template <class OBJ_VEC_TYPE, class OBJ_TYPE>
OBJ_TYPE * PhyDiag::getPtrFromVec(OBJ_VEC_TYPE& vector_obj, u_int32_t idx)
{
    IBDIAG_ENTER;
    //not exists
    if (idx < 0 || (vector_obj.size() < idx + 1))
        IBDIAG_RETURN(NULL);
    IBDIAG_RETURN(vector_obj[idx]);
}

template <class OBJ_VEC_TYPE, class OBJ_TYPE, class DATA_VEC_TYPE, class DATA_TYPE>
int PhyDiag::addDataToVecInVec(OBJ_VEC_TYPE& vector_obj,
        OBJ_TYPE *p_obj,
        DATA_VEC_TYPE& vec_of_vectors,
        u_int32_t data_idx,
        DATA_TYPE& data)
{
    IBDIAG_ENTER;
    if (!p_obj)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    //if exist already
    if ((vec_of_vectors.size() >= p_obj->createIndex + 1) &&
            (vec_of_vectors[p_obj->createIndex].size() >= data_idx + 1))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    //add
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Adding %s for (%s, index=%u, vec_idx=%u)\n",
            typeid(data).name(), p_obj->getName().c_str(), p_obj->createIndex, data_idx);
    if (vec_of_vectors.empty() || (vec_of_vectors.size() < p_obj->createIndex + 1))
        vec_of_vectors.resize(p_obj->createIndex + 1);

    if (vec_of_vectors[p_obj->createIndex].empty() ||
            (vec_of_vectors[p_obj->createIndex].size() < data_idx + 1))
        for (int i = (int)vec_of_vectors[p_obj->createIndex].size(); i < (int)data_idx + 1; ++i)
            vec_of_vectors[p_obj->createIndex].push_back(NULL);
    DATA_TYPE *p_curr_data = new DATA_TYPE;
    if (!p_curr_data) {
        this->SetLastError("Failed to allocate %s", typeid(data).name());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
    }
    *p_curr_data = data;
    vec_of_vectors[p_obj->createIndex][data_idx] = p_curr_data;
    this->addPtrToVec(vector_obj, p_obj);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

template <class OBJ_VEC_TYPE, class OBJ_TYPE>
OBJ_TYPE * PhyDiag::getPtrFromVecInVec(OBJ_VEC_TYPE& vec_of_vectors,
        u_int32_t idx1,
        u_int32_t idx2)
{
    IBDIAG_ENTER;
    //not exists
    if (idx1 < 0 || (vec_of_vectors.size() < idx1 + 1))
        IBDIAG_RETURN(NULL);
    if (idx2 < 0 || (vec_of_vectors[idx1].size() < idx2 + 1))
        IBDIAG_RETURN(NULL);
    IBDIAG_RETURN(vec_of_vectors[idx1][idx2]);
}

int PhyDiag::addDataToMapInVec(AccRegKey *p_acc_reg_key,
        vec_map_akey_diagnostic_data_t& vec_of_maps,
        u_int32_t data_idx,
        struct VS_DiagnosticData& data)
{
    IBDIAG_ENTER;

    stringstream sstream;
    sstream.str("");
    string key_data_str;

    if (!p_acc_reg_key)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    p_acc_reg_key->DumpKeyData(sstream);
    key_data_str = sstream.str();

    //add
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Adding %s for (Acc Reg Key: %s, vec_idx=%u)\n",
            typeid(data).name(), key_data_str.c_str(), data_idx);

    if (vec_of_maps.empty() ||
            (vec_of_maps.size() < data_idx + 1))
        for (int i = (int)vec_of_maps.size(); i < (int)data_idx + 1; ++i)
            vec_of_maps.push_back(NULL);

    if (!vec_of_maps[data_idx])
        vec_of_maps[data_idx] = new map_akey_diagnostic_data_t(keycomp);

    struct VS_DiagnosticData *p_curr_data = new struct VS_DiagnosticData;
    if (!p_curr_data) {
        this->SetLastError("Failed to allocate %s", typeid(data).name());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
    }
    *p_curr_data = data;

    std::pair<std::map<AccRegKey *, struct VS_DiagnosticData *>::iterator,bool> ret;
    ret = vec_of_maps[data_idx]->insert(pair <AccRegKey *,
                                struct VS_DiagnosticData *>(p_acc_reg_key,
                                                            p_curr_data));

    if (!ret.second) {
        this->SetLastError("%s already exist in map for (Acc Reg Key: %s, vec_idx=%u)\n",
                           typeid(data).name(), key_data_str.c_str(), data_idx);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

/**************/
//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int PhyDiag::addPhysLayerPortCounters(
    IBPort *p_port,
    struct VS_DiagnosticData &physCounters,
    u_int32_t dd_idx)
{
    IBDIAG_RETURN(this->addDataToVecInVec(this->ports_vector,
                p_port,
                this->port_cidx_and_dd_idx_to_dd_page,
                dd_idx,
                physCounters));
}

struct VS_DiagnosticData *
PhyDiag::getPhysLayerPortCounters(u_int32_t port_index, u_int32_t dd_idx)
{

 IBDIAG_ENTER;
 IBDIAG_RETURN((this->getPtrFromVecInVec<vec_vec_diagnostic_data_t, struct VS_DiagnosticData>
             (this->port_cidx_and_dd_idx_to_dd_page, port_index, dd_idx)));
}

int PhyDiag::addPhysLayerPCICounters(
    AccRegKey *p_acc_reg_key,
    struct VS_DiagnosticData &pciCounters,
    u_int32_t dd_idx)
{
    IBDIAG_RETURN(this->addDataToMapInVec(p_acc_reg_key,
                this->dd_idx_and_akey_to_dd_page,
                dd_idx,
                pciCounters));
}

int PhyDiag::addBER(IBPort *p_port, long double ber)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(this->addDataToVec(this->ports_vector, p_port,
                                     this->ber_vec,
                                     ber));
}

long double* PhyDiag::getBER(u_int32_t port_index)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVec<vec_long_double_t,
                                       long double>(
                                         this->ber_vec, port_index)));
}

int PhyDiag::addEffBER(IBPort *p_port, long double ber)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(this->addDataToVec(this->ports_vector, p_port,
                                     this->eff_ber_vec,
                                     ber));
}

long double* PhyDiag::getEffBER(u_int32_t port_index)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVec<vec_long_double_t,
                                       long double>(
                                         this->eff_ber_vec, port_index)));
}

void PhyDiag::DumpCSVPhyCounters(CSVOut &csv_out, u_int32_t dd_type)
{
    IBDIAGNET_ENTER;

    struct VS_DiagnosticData *p_curr_data = NULL;
    char buff[1024];
    stringstream sstream;

    for (u_int32_t dd_idx = 0; dd_idx < this->diagnostic_data_vec.size();
         ++dd_idx) {
        DiagnosticDataInfo * p_dd = this->diagnostic_data_vec[dd_idx];

        if (p_dd->GetDDType() != dd_type)
            continue;

        p_dd->DumpDiagnosticDataHeaderStart(csv_out);

        for (map_str_pnode::iterator nI = this->p_discovered_fabric->NodeByName.begin();
                nI != this->p_discovered_fabric->NodeByName.end(); ++nI) {
            IBNode *p_curr_node = (*nI).second;
            // go over all ports of this node:
            for (u_int32_t i = 1; i < (unsigned)p_curr_node->numPorts + 1; ++i) {
                IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
                if (!p_curr_port || !p_curr_port->is_data_worthy())
                    continue;


                p_curr_data = getPhysLayerPortCounters(p_curr_port->createIndex, dd_idx);
                if (!p_curr_data)
                    continue;

                memset(buff, 0, sizeof(buff));
                sstream.str("");
                sprintf(buff,
                        U64H_FMT "," U64H_FMT "," U32D_FMT "," U32D_FMT ",",
                        p_curr_port->p_node->guid_get(), p_curr_port->guid_get(),
                        p_curr_port->num, p_curr_data->CurrentRevision);
                sstream << buff;

                p_dd->DumpDiagnosticData(sstream, *p_curr_data);

                sstream << endl;
                csv_out.WriteLine(sstream.str());
                if (p_dd->IsPerNode())
                    break;
            }
        }

        p_dd->DumpDiagnosticDataHeaderEnd(csv_out);
    }
    IBDIAGNET_RETURN_VOID;
}


bool PhyDiag::isSupportFwBER(IBPort *p_port)
{
    IBDIAG_ENTER;

    DiagnosticDataInfo *p_dd;
    VS_DiagnosticData *p_curr_data = NULL;
    u_int32_t dd_idx = 0;

    for (dd_idx = 0; dd_idx < diagnostic_data_vec.size(); ++dd_idx) {
        p_dd = diagnostic_data_vec[dd_idx];
        if (p_dd->GetPageId() == DIAGNOSTIC_DATA_PHY_LAYER_STATISTICS_PAGE)
            break;
    }

    p_curr_data = getPhysLayerPortCounters(p_port->createIndex, dd_idx);

    if (p_curr_data)
        IBDIAGNET_RETURN(true);

    IBDIAGNET_RETURN(false);
}


void PhyDiag::DumpCSVPCICounters(CSVOut &csv_out, u_int32_t dd_type)
{
    IBDIAGNET_ENTER;

    AccRegKey *p_curr_acc_reg_key = NULL;
    struct VS_DiagnosticData *p_curr_data = NULL;
    char buff[1024];
    stringstream sstream;
    stringstream key_data_sstream;
    string key_data_str;

    for (u_int32_t dd_idx = 0; dd_idx < this->diagnostic_data_vec.size();
         ++dd_idx) {
         DiagnosticDataInfo *p_dd = this->diagnostic_data_vec[dd_idx];

         if (p_dd->GetDDType() != dd_type)
            continue;

         p_dd->DumpDiagnosticDataHeaderStart(csv_out);

         if (!dd_idx_and_akey_to_dd_page.empty() && dd_idx_and_akey_to_dd_page[dd_idx]) {
             map_akey_diagnostic_data_t *p_map_akey_dd = this->dd_idx_and_akey_to_dd_page[dd_idx];

             for (map_akey_diagnostic_data_t::iterator iter = p_map_akey_dd->begin();
                  iter != p_map_akey_dd->end(); ++iter) {
                 p_curr_acc_reg_key = iter->first;
                 if (!p_curr_acc_reg_key)
                     continue;
                  p_curr_data = iter->second;
                  if (!p_curr_data)
                      continue;

                  memset(buff, 0, sizeof(buff));
                  sstream.str("");
                  key_data_sstream.str("");
                  p_curr_acc_reg_key->DumpKeyData(key_data_sstream);
                  key_data_str = key_data_sstream.str();
                  sprintf(buff, STR_FMT U32D_FMT ",", key_data_str.c_str(),
                          p_curr_data->CurrentRevision);
                  sstream << buff;

                  p_dd->DumpDiagnosticData(sstream, *p_curr_data);

                  sstream << endl;
                  csv_out.WriteLine(sstream.str());
             }
         }

         p_dd->DumpDiagnosticDataHeaderEnd(csv_out);
    }

    IBDIAGNET_RETURN_VOID;
}


int PhyDiag::DumpNetDumpExt()
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    u_int32_t dd_idx = 0;
    DiagnosticDataInfo *p_dd;
    VS_DiagnosticData *p_curr_data = NULL;
    DD_PhyStatisticalCounters phy_statistics;
    string file_str = *p_base_path + NET_DUMP_EXT_FILE;
    ofstream sout;
    char port_line[900] = {'\0'};

    rc = p_ibdiag->OpenFile(file_str.c_str(), sout, false, true);
    if (rc){
        ERR_PRINT("Failed to open file: " NET_DUMP_EXT_FILE);
        IBDIAGNET_RETURN(rc);
    }

    sprintf(port_line,
            "%-2s : %-4s : %-3s : %-18s : "
            "%-12s : %-4s : %-7s : %-7s : "
            "%-7s : %-12s : %-19s : %-6s : "
            "%-15s : %-15s : %-10s : %-13s : %s",

            "Ty", "#",  "#IB", "GUID",
            "LID", "Sta", "PhysSta", "LWA",
            "LSA", "Conn LID (#)", "FEC mode", "RTR",
            "Raw BER", "Effective BER", "Symbol Err", "Effective Err", "Node Desc");
    sout << port_line << endl;

    for (; dd_idx < diagnostic_data_vec.size(); ++dd_idx) {
        p_dd = diagnostic_data_vec[dd_idx];
        if (p_dd->GetPageId() == DIAGNOSTIC_DATA_PHY_LAYER_STATISTICS_PAGE)
            break;
    }

    if (dd_idx == diagnostic_data_vec.size()) {
        ERR_PRINT("Can't find PHY_LAYER_STATISTICS_PAGE");
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    if (ber_threshold_err < ber_threshold_warn)
        WARN_PRINT("ber_threshold_err is lower than ber_threshold_warn")

    map_str_pnode::iterator nodesI = p_discovered_fabric->NodeByName.begin();
    for ( ; nodesI != p_discovered_fabric->NodeByName.end(); ++nodesI) {
        IBNode *p_curr_node = (*nodesI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName " \
                               "map for key = %s", (*nodesI).first.c_str());
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        // go over all ports of this node, start from port 1, no need to dump port 0
        for (u_int32_t port_i = 1; port_i < (unsigned)p_curr_node->numPorts + 1; ++port_i) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)port_i);
            if (!p_curr_port || !p_curr_port->getInSubFabric())
                continue;

            if (!p_curr_port->isValid())
                continue;

            SMP_PortInfo * p_curr_port_info = p_ibdiag->GetIBDMExtendedInfoPtr()->
                    getSMPPortInfo(p_curr_port->createIndex);
            if (!p_curr_port_info)
                continue;

            SMP_MlnxExtPortInfo *p_mepi = p_ibdiag->GetIBDMExtendedInfoPtr()->
                    getSMPMlnxExtPortInfo(p_curr_port->createIndex);

            PM_PortCounters *p_curr_port_counters =
                    p_ibdiag->GetIBDMExtendedInfoPtr()->getPMPortCounters(p_curr_port->createIndex);

            p_curr_data = getPhysLayerPortCounters(p_curr_port->createIndex, dd_idx);
            if (p_curr_data)
                DD_PhyStatisticalCounters_unpack(&phy_statistics,
                                                 (uint8_t *)&(p_curr_data->data_set));

            char lid_port[15] = {'\0'};
            if (p_curr_port->p_remotePort)
                sprintf(lid_port,
                        "%-5d (%s)",
                        p_curr_port->p_remotePort->base_lid,
                        p_curr_port->p_remotePort->numAsString().c_str());

            double raw_ber = phy_statistics.raw_ber_coef *
                    (1 / pow(10, phy_statistics.raw_ber_magnitude));
            double eff_ber = phy_statistics.effective_ber_coef *
                    (1 / pow(10, phy_statistics.effective_ber_magnitude));

            // set Errors / Warnings for bad links
            if (p_curr_data) {
                if (eff_ber > ber_threshold_err) {

                    FabricErrFwBERExceedThreshold *p_curr_fabric_ber_err =
                            new FabricErrFwBERExceedThreshold(p_curr_port, ber_threshold_err, eff_ber);
                    if (!p_curr_fabric_ber_err) {
                        this->SetLastError("Failed to allocate FabricErrFwBERExceedThreshold");
                        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_NO_MEM);
                    }
                    phy_errors.push_back(p_curr_fabric_ber_err);

                } else if (eff_ber > ber_threshold_warn) {

                    FabricErrFwBERExceedThreshold *p_curr_fabric_ber_err =
                            new FabricErrFwBERExceedThreshold(p_curr_port, ber_threshold_warn, eff_ber);
                    if (!p_curr_fabric_ber_err) {
                        this->SetLastError("Failed to allocate FabricErrFwBERExceedThreshold");
                        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_NO_MEM);
                    }
                    p_curr_fabric_ber_err->SetLevel(EN_FABRIC_ERR_WARNING);
                    phy_errors.push_back(p_curr_fabric_ber_err);
                }
            }

            stringstream raw_ber_str;
            stringstream eff_ber_str;
            stringstream effective_errors;
            stringstream symbol_errors;
            raw_ber_str << scientific << raw_ber;
            eff_ber_str << scientific << eff_ber;
            effective_errors << phy_statistics.phy_symbol_errors;
            if (p_curr_port_counters)
                symbol_errors << p_curr_port_counters->SymbolErrorCounter;

            CLEAR_STRUCT(port_line);
            sprintf(port_line,
                    "%-2s : %-4s : %-3u : " U64H_FMT " : "
                    "%-4u (0x%-3x) : %-4s : %-7s : %-7s : "
                    "%-7s : %-12s : %-19s : %-6s : "
                    "%-15s : %-15s : %-10s : %-13s : \"%s\"\n",

                    nodetype2char(p_curr_node->type),
                    p_curr_port->numAsString().c_str(),
                    p_curr_port->num,
                    p_curr_port->guid_get(),
                    p_curr_port->base_lid,
                    p_curr_port->base_lid,
                    portstate2char(p_curr_port->get_internal_state()),
                    portphysstate2char((IBPortPhysState)p_curr_port_info->PortPhyState),
                    p_curr_port->get_internal_state() == IB_PORT_STATE_DOWN ?
                            NOT_AVAILABLE : width2char(p_curr_port->get_internal_width()),
                    p_curr_port->get_internal_state() == IB_PORT_STATE_DOWN ?
                            NOT_AVAILABLE : speed2char(p_curr_port->get_internal_speed()),
                    lid_port,
                    fec2char(p_curr_port->get_fec_mode()),
                    !p_mepi ? RETRANS_MODE_NA :
                            retransmission2char((EnRetransmissionMode)p_mepi->RetransMode),
                    p_curr_data? raw_ber_str.str().c_str() : "",
                    p_curr_data? eff_ber_str.str().c_str() : "",
                    p_curr_port_counters ? symbol_errors.str().c_str() : "",
                    p_curr_data? effective_errors.str().c_str() : "",
                    p_curr_node->description.c_str());
            sout << port_line;
        }
    }

    sout.close();
    AddGeneratedFileName("Network dump ext. file", file_str);

    IBDIAGNET_RETURN(rc);
}

void PhyDiag::DumpCSVRawBER(CSVOut &csv_out)
{
    IBDIAGNET_ENTER;

    char buff[1024];
    stringstream sstream;

    csv_out.DumpStart(SECTION_RAW_BER);
    sstream << "NodeGuid,PortGuid,PortNum,field0" << endl;
    csv_out.WriteLine(sstream.str());

    for (u_int32_t i = 0; i < this->getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->getPortPtr(i);
        if (!p_curr_port)               //port not exist
            continue;

        if (isSupportFwBER(p_curr_port))
            continue;

        long double *p_ber = getBER(p_curr_port->createIndex);

        //info doesn't exist
        if (!p_ber)
            continue;

        //show exponent of the raw ber
        //in case of zero ber, the exponent is infinite, show 255 instead
        double ber_exponent = (*p_ber == 0) ?
                              ZERO_RAW_BER_EXPONENT :
                              -1.0 * log10(*p_ber);
        memset(buff, 0, sizeof(buff));
        sstream.str("");
        //NodeGuid,PortGuid,PortNum,raw ber exponent
        sprintf(buff,
                U64H_FMT "," U64H_FMT "," U32D_FMT ",%f",
                p_curr_port->p_node->guid_get(), p_curr_port->guid_get(),
                p_curr_port->num, ber_exponent);
        sstream << buff << endl;
        csv_out.WriteLine(sstream.str());
    }

    csv_out.DumpEnd(SECTION_RAW_BER);
    IBDIAGNET_RETURN_VOID;
}

void PhyDiag::DumpCSVEffectiveBER(CSVOut &csv_out)
{
    IBDIAGNET_ENTER;
    SMP_MlnxExtPortInfo* p_ext_port_info = NULL;
    char buff[1024];
    stringstream sstream;

    csv_out.DumpStart(SECTION_EFF_BER);
    sstream << "NodeGuid,PortGuid,PortNum,BER,FECMode,Retransmission" << endl;
    csv_out.WriteLine(sstream.str());

    for (u_int32_t i = 0; i < this->getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->getPortPtr(i);
        if (!p_curr_port)               //port not exist
            continue;

        if (isSupportFwBER(p_curr_port))
            continue;

        long double *p_reciprocal_ber = getEffBER(p_curr_port->createIndex);

        //info doesn't exist
        if (!p_reciprocal_ber)
            continue;
        p_ext_port_info =
            this->p_ibdm_extended_info->getSMPMlnxExtPortInfo(
                    p_curr_port->createIndex);
        if (!p_ext_port_info)
            continue;

        memset(buff, 0, sizeof(buff));
        sstream.str("");
        //NodeGuid,PortGuid,PortNum, effective reciprocal_ber
        sprintf(buff,
                U64H_FMT "," U64H_FMT "," U32D_FMT ",%Le," STR_FMT "," STR_FMT,
                p_curr_port->p_node->guid_get(), p_curr_port->guid_get(),
                p_curr_port->num,
                (*p_reciprocal_ber) == 0 ? 0.0 : 1.0/(*p_reciprocal_ber),
                fec2char(p_curr_port->get_fec_mode()),
                retransmission2char((EnRetransmissionMode)p_ext_port_info->
                                                          RetransMode));
        sstream << buff << endl;
        csv_out.WriteLine(sstream.str());
    }

    csv_out.DumpEnd(SECTION_EFF_BER);
    IBDIAGNET_RETURN_VOID;
}

void PhyDiag::DumpCSVSLRGExternalInfo(CSVOut &csv_out, AccRegHandler *p_areg_handler)
{
    IBDIAGNET_ENTER;
    bool any_data_per_port = false;
    char buff1[512] = {0}, buff2[512] = {0};
    char *p_in_line = NULL;
    u_int8_t gradeID = 0;
    u_int32_t grade = 0;
    int print_res = 0;
    stringstream sstream;

    csv_out.DumpStart(ACC_REG_SLRG_EXTERNAL_SECTION_NAME);
    //NodeGuid,PortGuid,PortNum,GradeID,Lane0Grade,Lane1Grade,Lane2Grade,Lane3Grade
    sstream << "NodeGuid,PortGuid,PortNum,GradeID";
    for (u_int8_t cnt = 0; cnt < LANE_NUM; cnt++)
        sstream << ",Lane" << (unsigned int)cnt <<  "Grade";
    sstream << endl;
    csv_out.WriteLine(sstream.str());

     for (map_akey_areg::iterator nI = p_areg_handler->data_map.begin();
         nI != p_areg_handler->data_map.end();) {

        any_data_per_port = false;
        gradeID = 0;
        p_in_line = buff2;
        print_res = 0;

        AccRegKey * p_aregkey = (*nI).first;
        if (!p_aregkey) {
            this->SetLastError("DB error - found null key in data_map");
            ++nI;
            continue;
        }

        if (((AccRegKeyPortLane *)p_aregkey)->lane != 0) {
            ++nI;
            continue;
        }

        for (u_int8_t ln = 0; ln < LANE_NUM; ln++) {
            p_aregkey = (*nI).first;
            if (nI == p_areg_handler->data_map.end() ||
                    ((AccRegKeyPortLane *)p_aregkey)->lane != ln) {
                any_data_per_port = false;
                break;
            }

            struct slrg_reg slrg = (slrg_reg&)(*nI).second;

            if (slrg.version == 0 || slrg.version == 1) { // 40nm, 28nm
                slrg_28nm slrg_28;
                slrg_28nm_unpack(&slrg_28, slrg.page_data.slrg_data_set.data);
                //GradeID=grade_version is the same for all lanes so take any available
                gradeID = slrg_28.grade_version;
                grade = slrg_28.grade;
            }

            if (slrg.version == 3) { // 16nm
                slrg_16nm slrg_16;
                slrg_16nm_unpack(&slrg_16, slrg.page_data.slrg_data_set.data);
                gradeID = slrg_16.grade_version;
                grade = slrg_16.grade;
            }

            any_data_per_port = true;

            //Lane0Grade,Lane1Grade,Lane2Grade,Lane3Grade
            print_res = sprintf(p_in_line, "," U32D_FMT, grade);
            if (print_res > 0)
                p_in_line += print_res;

            //increasing the map iterator
            ++nI;
        }
        if (any_data_per_port) {
            //NodeGuid,PortGuid,PortNum,GradeID,Lane0Grade,Lane1Grade,Lane2Grade,Lane3Grade
            sprintf(buff1,
                    U64H_FMT "," U64H_FMT "," U32D_FMT "," UH_FMT,
                    ((AccRegKeyPortLane *)p_aregkey)->node_guid,
                    ((AccRegKeyPortLane *)p_aregkey)->port_guid,
                    ((AccRegKeyPortLane *)p_aregkey)->port_num,
                    gradeID);
            sstream.str("");
            sstream << buff1 << buff2 << endl;
            csv_out.WriteLine(sstream.str());
        }
    }

    csv_out.DumpEnd(ACC_REG_SLRG_EXTERNAL_SECTION_NAME);
    IBDIAGNET_RETURN_VOID;
}


/*****************************************************************************/
static long double CalcBERFireCodeFEC(u_int64_t *p_counters)
{
    IBDIAGNET_ENTER;

    long double ber = 0.0;
    //(sum of FECCorrectableBlocks +
    // Kfcode*FECUncorrectableBlocks over all lanes)/time
    //FECCorrectableBlocks lane0 = fc_fec_corrected_blocks_lane0
    //FECUncorrectableBlocks lane0 = fc_fec_uncorrectable_blocks_lane0
    ber = p_counters[EN_FC_FEC_CORRECTED_BLOCKS_LANE0] +
          p_counters[EN_FC_FEC_CORRECTED_BLOCKS_LANE1] +
          p_counters[EN_FC_FEC_CORRECTED_BLOCKS_LANE2] +
          p_counters[EN_FC_FEC_CORRECTED_BLOCKS_LANE3] +
          K_FCODE *
          (p_counters[EN_FC_FEC_UNCORRECTABLE_BLOCKS_LANE0] +
           p_counters[EN_FC_FEC_UNCORRECTABLE_BLOCKS_LANE1] +
           p_counters[EN_FC_FEC_UNCORRECTABLE_BLOCKS_LANE2] +
           p_counters[EN_FC_FEC_UNCORRECTABLE_BLOCKS_LANE3]);
    IBDIAGNET_RETURN(ber);
}

static long double EDPLSum(u_int64_t *p_counters)
{
    IBDIAGNET_ENTER;
    //sum of edpls over lanes/time
    IBDIAGNET_RETURN(p_counters[EN_EDPL_LANE0] +
                     p_counters[EN_EDPL_LANE1] +
                     p_counters[EN_EDPL_LANE2] +
                     p_counters[EN_EDPL_LANE3]);
}

int PhyDiag::CalcRawBER()
{
    IBDIAGNET_ENTER;
    //calculate RAW BER
    long double ber = 0.0;
    struct VS_DiagnosticData *p_curr_data = NULL;
    u_int64_t                *p_curr_counters = NULL;
    long double k_rsfec = 0.0;

    for (u_int32_t i = 0; i < this->getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->getPortPtr(i);
        if (!p_curr_port)               //port not exist
            continue;

        if (isSupportFwBER(p_curr_port))
            continue;

        p_curr_data = this->getPhysLayerPortCounters(p_curr_port->createIndex, 0);
        if (!p_curr_data)       //info not exist
            continue;
        p_curr_counters =
            ((struct DDPhysCounters*)&(p_curr_data->data_set.gen_set))->counter;

        IBFECMode port_fec = p_curr_port->get_fec_mode();
        ber = 0.0;

        //undefined fec port
        if (port_fec == IB_FEC_NA)
            ber = max(CalcBERFireCodeFEC(p_curr_counters),
                      EDPLSum(p_curr_counters));
        else if (port_fec == IB_FEC_NO_FEC) {
            //up to QDR (8/10)
            if (p_curr_port->get_common_speed() < IB_LINK_SPEED_14)
                //symbol error/time
                ber = p_curr_counters[EN_SYMBOL_ERRORS];
            else //64/66
                //sum of edpls over lanes/time
                ber = EDPLSum(p_curr_counters);
        } else if (port_fec == IB_FEC_FIRECODE_FEC)
            //(sum of FECCorrectableBlocks +
            // Kfcode*FECUncorrectableBlocks over all lanes)/time
            //FECCorrectableBlocks lane0 = fc_fec_corrected_blocks_lane0
            //FECUncorrectableBlocks lane0 = fc_fec_uncorrectable_blocks_lane0
            ber = CalcBERFireCodeFEC(p_curr_counters);
        else if (isRSFEC(port_fec)) {
            //PortFECCorrectedSymbolsCounter + Krsfec*PortFECUncorrectableBlocks
            //PortFECCorrectedSymbolsCounter = rs_fec_corrected_symbols_total
            //PortFECUncorrectableBlocks = rs_fec_uncorrectable_blocks
            SMP_MlnxExtPortInfo* p_ext_port_info =
                this->p_ibdm_extended_info->getSMPMlnxExtPortInfo(
                    p_curr_port->createIndex);
            if (!p_ext_port_info)
                continue;
            k_rsfec = p_ext_port_info->ActiveRSFECParity / 2 + 1;
            //calculate k_rsfec
            ber = p_curr_counters[EN_RS_FEC_CORRECTED_SYMBOLS_TOTAL] +
                  k_rsfec * p_curr_counters[EN_RS_FEC_UNCORRECTABLE_BLOCKS];
        }

        long double link_rate =
            (long double)(CalcLinkRate(p_curr_port->get_common_width(),
                                       p_curr_port->get_common_speed()));
        long double data_rate =
            //EN_TIME_SINCE_LAST_CLEAR is in msec
            (p_curr_counters[EN_TIME_SINCE_LAST_CLEAR] / 1000) * link_rate;

        if (data_rate) {
            ber /= data_rate;
            this->addBER(p_curr_port, ber);
        }
    }
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int PhyDiag::CalcEffBER(u_int64_t ber_threshold_reciprocal_val,
                        list_p_fabric_general_err& ber_errors)
{
    IBDIAGNET_ENTER;
    int rc = 0;
    long double reciprocal_ber = 0.0;
    struct VS_DiagnosticData *p_curr_data = NULL;
    u_int64_t                *p_curr_counters = NULL;

    for (u_int32_t i = 0; i < this->getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->getPortPtr(i);
        if (!p_curr_port)               //port not exist
            continue;

        if (isSupportFwBER(p_curr_port))
            continue;

        p_curr_data = this->getPhysLayerPortCounters(p_curr_port->createIndex, 0);
        if (!p_curr_data)       //info not exist
            continue;
        p_curr_counters =
            ((struct DDPhysCounters*)&(p_curr_data->data_set.gen_set))->counter;

        reciprocal_ber = 0.0;
        rc = this->p_ibdiag->CalcBER(p_curr_port,
                                     p_curr_counters[EN_TIME_SINCE_LAST_CLEAR]/1000,
                                     p_curr_counters[EN_SYMBOL_ERRORS],
                                     reciprocal_ber);

        if (rc == IBDIAG_ERR_CODE_DB_ERR) {
            rc = IBDIAG_SUCCESS_CODE;
            continue;
        }

        if (rc == IBDIAG_SUCCESS_CODE && reciprocal_ber == 0) {
            if (ber_threshold_reciprocal_val == OVERFLOW_VAL_64_BIT) {    //need to report anyway
                FabricErrEffBERIsZero *p_curr_fabric_pm_err =
                        new FabricErrEffBERIsZero(p_curr_port);
                if (!p_curr_fabric_pm_err) {
                    this->SetLastError("Failed to allocate FabricErrEffBERIsZero");
                    IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
                }
                ber_errors.push_back(p_curr_fabric_pm_err);
                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            }
        } else if ((reciprocal_ber < (ber_threshold_reciprocal_val)) ||
                (ber_threshold_reciprocal_val == OVERFLOW_VAL_64_BIT) ) {
            FabricErrEffBERExceedThreshold *p_curr_fabric_pm_err =
                    new FabricErrEffBERExceedThreshold(p_curr_port,
                            ber_threshold_reciprocal_val,
                            reciprocal_ber);
            if (!p_curr_fabric_pm_err) {
                this->SetLastError("Failed to allocate FabricErrEffBERExceedThreshold");
                IBDIAG_RETURN(IBDIAG_ERR_CODE_NO_MEM);
            }
            ber_errors.push_back(p_curr_fabric_pm_err);
            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
        }

        this->addEffBER(p_curr_port, reciprocal_ber);
    }
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int PhyDiag::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    int rc2 = 0;

    //Dump first external data
    if (!this->can_send_mads_by_lid) {
        INFO_PRINT("%s skipped\n", CHECK_NAME_PHY_CNTRS_CHECK);
        PRINT("\n");
    } else if (this->to_get_phy_info) {
        rc2 = CalcEffBER(this->p_ibdiag->GetBERThreshold(), this->phy_errors);

        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(this->phy_errors,
                                       CHECK_NAME_EFF_BER,
                                       rc2,
                                       IBDIAG_ERR_CODE_CHECK_FAILED,
                                       this->num_errors,
                                       this->num_warnings);
        if (rc)
            IBDIAGNET_RETURN(rc);

        rc = CalcRawBER();
        //dump phy to db file
        this->DumpCSVPhyCounters(*this->p_csv_out, DD_PHY_TYPE);
        this->DumpCSVRawBER(*this->p_csv_out);
        this->DumpCSVEffectiveBER(*this->p_csv_out);

        rc2 = DumpNetDumpExt();
        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(this->phy_errors,
                                       CHECK_NAME_FW_BER,
                                       rc2,
                                       IBDIAG_ERR_CODE_CHECK_FAILED,
                                       this->num_errors,
                                       this->num_warnings);
        if (rc)
            IBDIAGNET_RETURN(rc);
    }

    //first, dump the internal registers
    for (uint32_t i = 0; i < this->reg_handlers_vec.size(); ++i) {
        AccRegHandler *p_areg_handler = this->reg_handlers_vec[i];
        if (!p_areg_handler->ToDumpRegister())
            if (!to_dump_cap_reg)
                continue;
        p_areg_handler->DumpCSV(*this->p_csv_out);
    }

    // second, dump the external registers
    // TODO: change to virtual functions empty DumpCSVExternalInfo in the Register Class.
    // SLRG will overrun this function
    for (uint32_t i = 0; i < this->reg_handlers_vec.size(); ++i) {
        AccRegHandler *p_areg_handler = this->reg_handlers_vec[i];
        if (p_areg_handler->GetPReg()->GetRegisterID() == ACCESS_REGISTER_ID_SLRG)
            this->DumpCSVSLRGExternalInfo(*this->p_csv_out, p_areg_handler);
    }

    if (this->to_get_pci_info) {

        if (!this->can_send_mads_by_lid) {
            INFO_PRINT("%s skipped\n", CHECK_NAME_PHY_CNTRS_CHECK);
            PRINT("\n");
        } else {
            this->DumpCSVPCICounters(*this->p_csv_out, DD_PCI_TYPE);
        }

        for (uint32_t i = 0; i < this->pci_reg_handlers_vec.size(); ++i) {
            AccRegHandler *p_areg_handler = this->pci_reg_handlers_vec[i];
            p_areg_handler->DumpCSV(*this->p_csv_out);
        }
    }

    IBDIAGNET_RETURN(rc);
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 * IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int PhyDiag::SMPAccRegGetByDirect(IN direct_route_t *p_direct_route,
                               IN u_int8_t port_num,
                               OUT struct SMP_AccessRegister *p_smp_acc_reg,
                               const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    p_smp_acc_reg->type_op = EMAD_TLV_TYPE_OPERATION;
    p_smp_acc_reg->len_op = EMAD_TLV_TYPE_OPERATION_LEN;
    p_smp_acc_reg->class_op = EMAD_CLASS_ACC_REG;
    p_smp_acc_reg->type_reg = EMAD_TLV_TYPE_REG;
    p_smp_acc_reg->method = EMAD_OP_TLV_METHOD_QUERY;
    p_smp_acc_reg->dr = 0x1;

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending ACC_REG MAD by direct = %s port = %u\n",
    this->p_ibis_obj->ConvertDirPathToStr(p_direct_route).c_str(),
    (int)port_num);
    /*
     * send the mad by direct route
     */
    int rc = this->p_ibis_obj->SMPMadGetSetByDirect(
        p_direct_route,
        IBIS_IB_MAD_METHOD_SET, //access register has to be set
        IB_ATTR_SMP_ACCESS_REG,
        port_num,
        p_smp_acc_reg,
        (const pack_data_func_t)SMP_AccessRegister_pack,
        (const unpack_data_func_t)SMP_AccessRegister_unpack,
        (const dump_data_func_t)SMP_AccessRegister_dump,
        p_clbck_data);

    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

int PhyDiag::GMPAccRegGet(u_int16_t lid,
                          u_int32_t register_id,
                          OUT struct GMP_AccessRegister *p_gmp_acc_reg,
                          const clbck_data_t *p_clbck_data)
{
    IBDIAGNET_ENTER;

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending ACC_REG MAD by lid = " U16H_FMT "\n", lid);

    int rc = this->p_ibis_obj->VSMadGetSet(lid,
        IBIS_IB_MAD_METHOD_GET,
        IB_ATTR_GMP_ACCESS_REG,
        register_id,
        p_gmp_acc_reg,
        (const pack_data_func_t)GMP_AccessRegister_pack,
        (const unpack_data_func_t)GMP_AccessRegister_unpack,
        (const dump_data_func_t)GMP_AccessRegister_dump,
        p_clbck_data);

    IBDIAGNET_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}


Plugin * create(IBDiag *p_ibdiag)
{
    return(new PhyDiag(p_ibdiag));
}

void destroy(Plugin *p)
{
    delete (PhyDiag *)p;
    return;
}
