/*
 * Copyright (c) 2004-2018 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */



#include <infiniband/ibdm/ARTraceRoute.h>

bool g_useSLVLPortGroup = true;

ARTraceRoutePath ARTraceRouteInfo::sm_ARTraceRoutePath;
//ARTraceRouteNodesInfoMap ARTraceRouteNodeInfo::sm_ARTraceRouteNodesInfo;

// Trace a route from slid to dlid by LFT
int ARTraceRouteByLFT(IBFabric *p_fabric,
                      lid_t sLid , lid_t dLid,
                      ARTraceRouteInfo *&pArRouteInfo)
{
    IBPort *p_port = p_fabric->getPortByLid(sLid);
    IBNode *p_node;
    pArRouteInfo = NULL;

    // make sure:
    if ( ! p_port ) {
        cout << "-E- Provided source:" << sLid
        << " lid is not mapped to a port!" << endl;
        return(1);
    }

    sl_vl_t slvl;
    slvl.SL = p_port->p_node->getPSLForLid(dLid);
    if ( slvl.SL == IB_SLT_UNASSIGNED ) {
        cout << "-E- Failed to get SL for node:" << p_port->p_node->name
                 << " dlid:" <<  dLid << endl;
        return 1;
    }
    slvl.VL = 0;
    //for in port 0 use SL even if VL2VL is set (for SW)
    slvl.VL = p_port->p_node->getVL(0, p_port->num, slvl);

    if ( p_port->p_node->type != IB_SW_NODE ) {

        if (slvl.VL == IB_SLT_UNASSIGNED) {
            cout << "-E- Failed to get VL for node:" << p_port->p_node->name
                << " inPort:0 outPort:"<< (int)p_port->num
                << " SL:" << (int)slvl.SL <<   endl;
            return(1);
        }
        if (slvl.VL == IB_DROP_VL) {
            cout << "-E- Dead end at:" << p_port->p_node->name
                << " Drop VL inPort:0 outPort:"<< (int)p_port->num
                << " SL:" << (int)slvl.SL <<   endl;
            return 1;
        }
    }

    if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE ) {
        cout << "--------------------------- TRACE PATH BY FDB -----------------------------" << endl;
        cout << "-V- Tracing from lid:" << sLid
            << " SL/VL:" << (int)slvl.SL << "/" << (int)slvl.VL
            << " to lid:" << dLid << endl;
    }

    // if the port is not a switch - go to the next switch:
    if ( p_port->p_node->type != IB_SW_NODE ) {
        // try the next one:
        if ( !p_port->p_remotePort ) {
            cout << "-E- Provided starting point is not connected !"
            << "lid:" << sLid << endl;
            return 1;
        }
        p_port = p_port->p_remotePort;
        p_node = p_port->p_node;
        if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
            cout << "-V- Arrived at Node:" << p_node->name
            << " Port:" << (unsigned int)p_port->num
            << " SL/VL:" << (int)slvl.SL << "/" << (int)slvl.VL <<endl;
    } else {
        // it is a switch :
        p_node = p_port->p_node;
    }


    // verify we are finally of a switch:
    if ( p_node->type != IB_SW_NODE ) {
        cout << "-E- Provided starting point is not connected to a switch !"
        << "lid:" << sLid << endl;
        return 1;
    }

    ARTraceRouteInfo * p_currInfo =
        ARTraceRouteNodeInfo::getRouteInfo(p_port, slvl, dLid);

    if (p_currInfo == NULL)
        return -1;

    pArRouteInfo = p_currInfo;

    p_currInfo->setCurrentAttributes(p_port->num);
    ARTraceRouteInfo::pathPushFront(p_currInfo);
    int pathSize = 1;

    ARTraceRouteInfo *p_childInfo = NULL;
    // traverse:
    while ( !ARTraceRouteInfo::isPathEmpty() ) {
        p_currInfo = ARTraceRouteInfo::pathFront();

        if (p_childInfo != NULL) {
            //finish discovery of child in path update parent
            p_currInfo->updateRouteStatistics(p_childInfo);
            p_childInfo = NULL;
        }

        phys_port_t out_port = p_currInfo->getNextPort();

        // if route discovery was ended from this roure info
        if ( out_port== IB_LFT_UNASSIGNED ) {
            p_childInfo = p_currInfo;
            p_currInfo->pathPopFront();
            pathSize--;

            if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
                p_currInfo->dumpRouteStatistics();
            //collect statistics, routes per hops
            continue;
        }

        // if the port number is 0 we must have reached the target
        // (and the target is a switch - currently unused)
        if ( out_port == 0 ) {
            uint16_t base_lid = 0;
            // get lid of any port of this node
            // use unsigned int index to prevent overwrap
            for ( unsigned int portNum = 0;
                (base_lid == 0) && (portNum <= p_node->numPorts); portNum++ ) {
                IBPort *p_port = p_node->getPort((phys_port_t)portNum);
                if ( p_port ) {
                    base_lid = p_port->base_lid;
                    break;
                }
            }
            if ( base_lid == 0 ) {
                cout << "-E- Fail to find node:" << p_node->name
                    << " base lid?" << endl;
                p_currInfo->addCount(AR_TRACE_ROUTE_DEAD_END);
                continue;
            }

            lid_t lidStep = (lid_t)(1 << p_port->lmc);
            if ( (base_lid > dLid) || (base_lid + lidStep - 1 < dLid) ) {
                cout << "-E- Dead end at port 0 of node:" << p_node->name << endl;
                p_currInfo->addCount(AR_TRACE_ROUTE_DEAD_END);
                continue;
            }
            p_currInfo->addGoodPath(0);
            continue;
        }

        // calc next node:
        ARTraceRouteInfo*  p_nextInfo =
            p_currInfo->getNextARTraceRouteInfo(out_port);
        if ( p_nextInfo == NULL )
            continue;
        ARTraceRouteInfo::pathPushFront(p_nextInfo);
        pathSize++;

        if ( pathSize > 256 ) {
            cout << "-E- Aborting after 256 hops." << endl;
            while ( !ARTraceRouteInfo::isPathEmpty() )
                ARTraceRouteInfo::pathPopFront();
            break;
        }
    }

    if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
        cout << "---------------------------------------------------------------------------\n" << endl;

    if ( !ARTraceRouteInfo::isPathEmpty() ) {
        ARTraceRouteInfo::clearPath();
        return -1;
    }

    return (p_currInfo->isErrorInPath() ? -1 : 0);
}

void ARTraceRouteInfo::pathPushFront(ARTraceRouteInfo *p_arInfo) {
    sm_ARTraceRoutePath.push_front(p_arInfo);
    p_arInfo->m_pNodeInfo->visit();
}

void ARTraceRouteInfo::pathPopFront() {
    sm_ARTraceRoutePath.front()->m_currInPort = IB_LFT_UNASSIGNED;
    sm_ARTraceRoutePath.front()->m_currOutPort = IB_LFT_UNASSIGNED;
    sm_ARTraceRoutePath.front()->m_pNodeInfo->leave();
    sm_ARTraceRoutePath.pop_front();
}

phys_port_t ARTraceRouteInfo::getNextPort()
{
    if (m_portsListIter == m_portsList.end())
        return getSkippedPort();

    if (m_incIter) {
        ++m_portsListIter;
        if (m_portsListIter == m_portsList.end())
            return getSkippedPort();
    } else
        m_incIter = true;

    return *m_portsListIter;
}

/*
phys_port_t ARTraceRouteInfo::getCurrentPort()
{
    if (m_portsListIter == m_portsList.end())
        return IB_LFT_UNASSIGNED;

    return *m_portsListIter;
} 
*/ 


void ARTraceRouteInfo::updateRouteStatistics(
    ARTraceRouteInfo *p_childInfo)
{
    for (int i = 0 ; i < AR_TRACE_ROUTE_END; ++i)
        m_routeStatistics[i] += p_childInfo->m_routeStatistics[i];

    m_errorInPath |= p_childInfo->m_errorInPath;

    m_minHops = min(m_minHops, p_childInfo->m_minHops + 1);
    m_maxHops = max(m_maxHops, p_childInfo->m_maxHops + 1);
}

void ARTraceRouteInfo::addGoodPath(unsigned int hops)
{
    m_routeStatistics[AR_TRACE_ROUTE_GOOD_PATH]++;
    m_minHops = min(m_minHops, hops);
    m_maxHops = max(m_maxHops, hops);
}

void ARTraceRouteInfo::dumpRouteStatistics()
{
    cout << "-V- Found total of "
        << m_routeStatistics[AR_TRACE_ROUTE_GOOD_PATH]
        << " pathes";

    if (m_routeStatistics[AR_TRACE_ROUTE_GOOD_PATH])
        cout << " with [" << m_minHops << ".." << m_maxHops << "] hops.";

    if (m_pNodeInfo)
        cout << " From Node:" << m_pNodeInfo->getNode()->name
            << " PLFT:" << (int)m_pLFT
            << " in Port:" << (int)m_inSLVLPortGroup
            << " to DLID:" << m_dLid << endl;
    else
        cout << endl;

    if (m_errorInPath) {
        cout << "-E- Found total of " ;

            if (m_routeStatistics[AR_TRACE_ROUTE_DEAD_END] &&
                m_routeStatistics[AR_TRACE_ROUTE_LOOP])

                cout << m_routeStatistics[AR_TRACE_ROUTE_DEAD_END]
                    << " dead end errors and "
                    << m_routeStatistics[AR_TRACE_ROUTE_LOOP]
                    << " loops errors." << endl;
            else if (m_routeStatistics[AR_TRACE_ROUTE_DEAD_END])
                cout << m_routeStatistics[AR_TRACE_ROUTE_DEAD_END]
                    << " dead end errors." << endl;
            else
                cout << m_routeStatistics[AR_TRACE_ROUTE_LOOP]
                    << " loops errors." << endl;
    }
}

ARTraceRouteInfo* ARTraceRouteInfo::findNextARTraceRouteInfo(
    phys_port_t out_port,lid_t dLid, bool &reachedDest) const
{

    reachedDest = false;

    IBNode *p_node = m_pNodeInfo->getNode();
    //HW prevents loopback on the same port
    if (g_useSLVLPortGroup)
    {
        if (p_node->getSLVLPortGroup(out_port) == m_inSLVLPortGroup)
            return NULL;
    } else if (out_port == m_inSLVLPortGroup) {
        return NULL;
    }

    // get the port on the other side
    IBPort *p_port = p_node->getPort(out_port);
    if ( ! (p_port &&
            p_port->p_remotePort) ) {
        return NULL;
    }

    IBPort *p_remotePort = p_port->p_remotePort;
    // check if we are done:
    lid_t lidStep = (lid_t)(1 << p_remotePort->lmc);
    bool done = ((p_remotePort->base_lid <= dLid) &&
            (p_remotePort->base_lid+lidStep - 1 >= dLid));

    if (done) {
        reachedDest = true;
        return NULL;
    }

    if (p_remotePort->p_node->type != IB_SW_NODE) {
        cout << "-E- Invalid route to lid:" << p_remotePort->base_lid
             << "instead of:" << dLid << endl;
        return NULL;
    }

    sl_vl_t oSLVL;
    p_node->getSLVL(m_inSLVLPortGroup, out_port, getInSLVL(), oSLVL);

    if (oSLVL.VL == IB_SLT_UNASSIGNED ||
        oSLVL.VL == IB_DROP_VL) {
        return NULL;
    }

    ARTraceRouteInfo * p_nextRouteInfo =
        ARTraceRouteNodeInfo::findRouteInfo(p_remotePort, oSLVL);

    return p_nextRouteInfo;
}

ARTraceRouteInfo* ARTraceRouteInfo::getNextARTraceRouteInfo(
    phys_port_t out_port)
{

    //HW prevents loopback on the same port
    if (m_currInPort==out_port) {
        m_skippedOutPort = m_currInPort;
        return NULL;
    }

    if (out_port == m_skippedOutPort)
        m_skippedOutPort = IB_LFT_UNASSIGNED;

    m_currOutPort = out_port;

    // get the port on the other side
    IBNode *p_node = m_pNodeInfo->getNode();
    IBPort *p_port = p_node->getPort(out_port);

    //if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
    //    cout << "-V- Calculate Going out on port:" << (unsigned int)out_port << endl;

    if ( ! (p_port &&
            p_port->p_remotePort &&
            p_port->p_remotePort->p_node) ) {
        cout << "-E- Dead end to lid:" << m_dLid << " at:"
            << p_node->name << " through port:"
            << (unsigned int)out_port << endl;
        addCount(AR_TRACE_ROUTE_DEAD_END);
        return NULL;
    }

    IBPort *p_remotePort = p_port->p_remotePort;
    IBNode *p_remoteNode = p_remotePort->p_node;
    // check if we are done:
    lid_t lidStep = (lid_t)(1 << p_remotePort->lmc);
    bool done = ((p_remotePort->base_lid <= m_dLid) &&
            (p_remotePort->base_lid+lidStep - 1 >= m_dLid));

    if (done) {
        if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
            cout << "-V- Going out from Node: " << p_node->name
                << " pLFT:" << (int)m_pLFT
                <<  " on port:" << (unsigned int)out_port
                << " Arrived at destination Node:" << p_remoteNode ->name
                << " DLID:" << m_dLid << endl;

    } else { // check if the dlid is a vlid on the phys port we reached.

        if (p_remotePort->p_node->type == IB_CA_NODE) {
            map_vportnum_vport::iterator vports_iter = p_remotePort->VPorts.begin();
            for ( ; vports_iter != p_remotePort->VPorts.end(); vports_iter++) {

                if (vports_iter->second->get_vlid() == m_dLid) {
                    // we found the destination vport lid.

                    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)

                        cout << "-V- Going out from Node: " << p_node->name
                             << " pLFT:" << (int)m_pLFT
                             <<  " on port:" << (unsigned int)out_port
                             << " Arrived at destination Node:" << p_remoteNode ->name
                             << " Port:" << p_remotePort->num
                             << " vPort: " << vports_iter->second->getVPortNum()
                             << " Virtual DLID:" << m_dLid << endl;

                    done = true;
                    break;
                }
            }
        }
    }

    if (done) {
        addGoodPath(1);
        return NULL;
    }

    if (p_remotePort->p_node->type != IB_SW_NODE) {
        cout << "-E- Invalid route to lid:" << p_remotePort->base_lid
             << "instead of:" << m_dLid << endl;
        addCount(AR_TRACE_ROUTE_DEAD_END);
        return NULL;
    }

    sl_vl_t oSLVL;
    p_node->getSLVL(m_currInPort, out_port, getInSLVL(), oSLVL);

    if (oSLVL.VL == IB_SLT_UNASSIGNED) {
        cout << "-E- Failed to get VL for node:" << p_port->p_node->name
            << " inPort:" <<  (int)m_currInPort
            << " outPort:"<< (int)out_port
            << " SL:" << (int)oSLVL.SL <<   endl;
        addCount(AR_TRACE_ROUTE_DEAD_END);
        return NULL;
    }
    if (oSLVL.VL == IB_DROP_VL) {
        cout << "-E- Dead end at:" << p_port->p_node->name
            << " Drop VL inPort:" <<  (int)m_currInPort
            << " outPort:"<< (int)out_port
            << " SL:" << (int)oSLVL.SL <<   endl;

        addCount(AR_TRACE_ROUTE_DEAD_END);
        return NULL;
    }

    if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
        cout << "-V- Going out from Node: " << p_node->name
            << " pLFT:" << (int)m_pLFT
            << " on port:" << (unsigned int)out_port
            << " SL/VL:" << (int)oSLVL.SL << "/" << (int)oSLVL.VL
            << " Arrived at Node:" << p_remoteNode ->name
            << " DLID:" << m_dLid << endl;

    ARTraceRouteInfo * p_nextRouteInfo =
        ARTraceRouteNodeInfo::getRouteInfo(p_remotePort, oSLVL, m_dLid);

    if(isLoopInRoute(p_nextRouteInfo)) {
        addCount(AR_TRACE_ROUTE_LOOP);
        return NULL;
    }

    p_nextRouteInfo->setCurrentAttributes(p_remotePort->num);
    return p_nextRouteInfo;
}

bool ARTraceRouteInfo::isLoopInRoute(ARTraceRouteInfo *p_routeInfo)
{

    if (p_routeInfo->m_currInPort == IB_LFT_UNASSIGNED)
        return false;

    cout << "-E- Find loop in path on Node:" <<
        p_routeInfo->m_pNodeInfo->getNode()->name
        << " to DLID:" << p_routeInfo->getDLID() << endl;

    ARTraceRoutePath::reverse_iterator rit = sm_ARTraceRoutePath.rbegin();
    for(;rit != sm_ARTraceRoutePath.rend(); ++rit)
        if(*rit == p_routeInfo)
            break;

    if (rit == sm_ARTraceRoutePath.rend()) {
        cout << "-E- Failed to find info in path:" << endl;
        return true;
    }

    IBNode *p_node = NULL;
    IBPort *p_port = NULL;
    IBPort *p_remotePort = NULL;
    phys_port_t currentPort;

    for (;rit != sm_ARTraceRoutePath.rend(); ++rit ) {
        p_node = (*rit)->m_pNodeInfo->getNode();
        currentPort =  (*rit)->getCurrentPort();
        p_port = p_node->getPort(currentPort);
        if (p_port == NULL) {
            cout << "-E- Failed to get node: " << p_node->name 
                << " port: " << (int) currentPort << endl;
            return true;
        }
        p_remotePort = p_port->p_remotePort;

        cout << "  - From Node:" << p_node->name
                << " pLFT:" << (unsigned int) (*rit)->getPLFT()
                << " port:" << (unsigned int) currentPort
                << " to Node:" << p_remotePort->p_node->name
                << " port:" << (unsigned int) p_remotePort->num << endl;
    }

    return true;
}

bool ARTraceRouteInfo::convertDestLid(lid_t dLid)
{
    if (m_useAR) {
       if ( m_pNodeInfo->getNode()->getARLFTPortGroupForLid(dLid,m_pLFT) !=
            m_arLFTPortGroup ) {
           /*
           cout << "-D- convertDestLid cache old LFTPortGroup:" << m_arLFTPortGroup
                << " new LFTPortGroup:"
                << m_pNodeInfo->getNode()->getARLFTPortGroupForLid(dLid,m_pLFT)
                << endl;
                */
           return false;
       }
    } else if(m_outStaticPort != m_pNodeInfo->getNode()->getLFTPortForLid(dLid, m_pLFT)) {
        //in case AR is disabled
        if (m_errorInPath) {
            /*
            cout << "-D- convertDestLid cache old StaticPort:" << (int)m_outStaticPort
                 << " new StaticPort:"
                 << (int)m_pNodeInfo->getNode()->getLFTPortForLid(dLid, m_pLFT)
                 << " m_errorInPath:" << m_errorInPath
                 << endl;
                 */

            return false;
        }

        bool reachedDest = false;
        ARTraceRouteInfo *nextRouteInfo = findNextARTraceRouteInfo(
            m_pNodeInfo->getNode()->getLFTPortForLid(dLid, m_pLFT), dLid, reachedDest);

        if ( reachedDest ){
            if (m_minHops != 1){
                /*
                cout << "-D- convertDestLid cache old StaticPort:" << (int)m_outStaticPort
                     << " new StaticPort:"
                     << (int)m_pNodeInfo->getNode()->getLFTPortForLid(dLid, m_pLFT)
                     << " m_minHops:" << m_minHops
                     << endl;
                     */
                return false;
            }
        } else if (nextRouteInfo) {
            if (((nextRouteInfo->m_minHops + 1) != m_minHops) ||
                ((nextRouteInfo->m_maxHops + 1) != m_maxHops)){
                /*
                cout << "-D- convertDestLid cache old StaticPort:" << (int)m_outStaticPort
                     << " new StaticPort:"
                     << (int)m_pNodeInfo->getNode()->getLFTPortForLid(dLid, m_pLFT)
                     << " old m_minHops:" << m_minHops
                     << " new m_minHops:" << nextRouteInfo->m_minHops
                     << " old m_maxHops:" << m_maxHops
                     << " new m_maxHops:" << nextRouteInfo->m_maxHops
                     << endl;
                     */
                return false;
            }
            if ((getGoodPathCount() != nextRouteInfo->getGoodPathCount()) ||
                nextRouteInfo->m_errorInPath){
                /*
                cout << "-D- convertDestLid cache old StaticPort:" << (int)m_outStaticPort
                     << " new StaticPort:"
                     << (int)m_pNodeInfo->getNode()->getLFTPortForLid(dLid, m_pLFT)
                     << " old GoodPath:" << getGoodPathCount()
                     << " new GoodPath:" << nextRouteInfo->getGoodPathCount()
                     << " new errorInPath:" << nextRouteInfo->m_errorInPath
                     << endl;
                     */
                return false;
            }
        } else {
            /*
            cout << "-D- convertDestLid findNextARTraceRouteInfo out port:"
                 << (int)m_pNodeInfo->getNode()->getLFTPortForLid(dLid, m_pLFT)
                 << endl;
                 */
            return false;
        }
    }

    m_dLid = dLid;
    return true;
}

void ARTraceRouteInfo::set(
    sl_vl_t inSLVL,
    u_int8_t inPort, u_int8_t inSLVLPortGroup,
    u_int8_t pLFT, lid_t dLid,
    ARTraceRouteNodeInfo *pNodeInfo)
{
    IBNode *p_node = pNodeInfo->getNode();
    bool useAR = false;
    if (p_node->isAREnable() &&
        p_node->isARActive(inSLVL))
        useAR = 1;

    //reset old data
    m_inSLVL = inSLVL;
    m_errorInPath = false;
    m_minHops = 0xFFFF;
    m_maxHops = 0;
    m_currInPort = IB_LFT_UNASSIGNED;
    m_currOutPort = IB_LFT_UNASSIGNED;
    m_skippedOutPort = IB_LFT_UNASSIGNED;
    m_incIter = false;
    clearRouteStatistics();

    m_inSLVLPortGroup = inSLVLPortGroup;
    m_pNodeInfo = pNodeInfo;
    m_pLFT = pLFT;
    m_dLid = dLid;
    m_useAR = useAR;

    m_arLFTPortGroup = IB_AR_LFT_UNASSIGNED;
    if(m_useAR)
        m_arLFTPortGroup =
            p_node->getARLFTPortGroupForLid(dLid,pLFT);

    m_outStaticPort = p_node->getLFTPortForLid(dLid, pLFT);

    p_node->getLFTPortListForLid(m_outStaticPort,
                                 m_arLFTPortGroup, m_portsList);

    if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE ) {
        char line[1024];
        p_node->getARGroupCfg(m_arLFTPortGroup, line);
        cout << "-V- set RouteInfo for Node: " << p_node->name
            << " SL/VL:" << (int)inSLVL.SL << "/" << (int)inSLVL.VL
            << " pLFT:" << (int)m_pLFT
            << " AR:" << (useAR ? "enabled" : "disabled")
            << " static out port:" << (int) m_outStaticPort
            << " group:" << m_arLFTPortGroup
            << " group members:"
            << line << endl;
    }

    m_portsListIter = m_portsList.begin();

    if (m_portsListIter == m_portsList.end()) {
        cout << "-E- Dead end to lid:" << m_dLid << " at:"
            << p_node->name << " pLFT:" << (int)pLFT << endl;
        addCount(AR_TRACE_ROUTE_DEAD_END);
    }  else if (*m_portsListIter == inPort &&
                    m_portsList.size() == 1) {
        cout << "-E- Dead end (loopback) to lid:" << m_dLid << " at:"
            << p_node->name << " pLFT:" << (int)pLFT << endl;
    }
}

ARTraceRouteInfo* ARTraceRouteNodeInfo::findInfo(
        IBPort *p_port, sl_vl_t inSLVL)
{
    IBNode *p_node = p_port->p_node;

    u_int8_t pLFT = p_node->getPLFTMapping(p_port->num, inSLVL);

    u_int8_t sl2vlPortGroup = p_port->num;
    if (g_useSLVLPortGroup)
        sl2vlPortGroup = p_node->getSLVLPortGroup(p_port->num);

    if (m_routeInfoCollection[inSLVL.VL].size() <= inSLVL.SL ||
        m_routeInfoCollection[inSLVL.VL][inSLVL.SL].size() <= sl2vlPortGroup ||
        m_routeInfoCollection[inSLVL.VL][inSLVL.SL][sl2vlPortGroup].size() <=
        pLFT)
        return NULL;

    return &m_routeInfoCollection[inSLVL.VL][inSLVL.SL][sl2vlPortGroup][pLFT];
}

ARTraceRouteInfo* ARTraceRouteNodeInfo::getInfo(
        IBPort *p_port, sl_vl_t inSLVL, lid_t dLid)
{
    IBNode *p_node = p_port->p_node;

    u_int8_t maxPLFT = p_node->getMaxPLFT();
    u_int8_t pLFT = p_node->getPLFTMapping(p_port->num, inSLVL);

    if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
        cout << "-V- AR pLFT MAP"
            << " Port:" << (int)p_port->num
            << " and SL/VL:" << (int)inSLVL.SL << "/" << (int)inSLVL.VL
            << " to pLFT:" << (int)pLFT
            << " on Node: " << p_node->name << endl;

    u_int8_t maxSL2VLPortGroup = p_node->numPorts;
    u_int8_t sl2vlPortGroup = p_port->num;
    if (g_useSLVLPortGroup)
        sl2vlPortGroup = p_node->getSLVLPortGroup(p_port->num);

    if (m_routeInfoCollection[inSLVL.VL].size() <=
        IBNode::getMaxSL() )
        m_routeInfoCollection[inSLVL.VL].
            resize(IBNode::getMaxSL() + 1);

    if (m_routeInfoCollection[inSLVL.VL][inSLVL.SL].size() <=
        maxSL2VLPortGroup )
        m_routeInfoCollection[inSLVL.VL][inSLVL.SL].
            resize(maxSL2VLPortGroup + 1);

    if (m_routeInfoCollection[inSLVL.VL][inSLVL.SL][sl2vlPortGroup].size() <=
        maxPLFT )
        m_routeInfoCollection[inSLVL.VL][inSLVL.SL][sl2vlPortGroup].
            resize(maxPLFT + 1);

    ARTraceRouteInfo &info =
        m_routeInfoCollection[inSLVL.VL][inSLVL.SL][sl2vlPortGroup][pLFT];

    if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
        cout << "-D-"
            << " sl2vlPortGroup:" << (int)sl2vlPortGroup
            << " pLFT:" << (int)pLFT
            << " returns:" << (info.isSet()?"old":"new")
            << " routeInfo:" << &info << endl;

    if (!info.isSet()) { //if new info
        info.set(inSLVL, p_port->num, sl2vlPortGroup, pLFT, dLid, this);
        m_usedRouteInfo.push_back(&info);
    }

    return &info;
}


//allocate neaded memory
int ARTraceRouteNodeInfo::prepare(IBFabric *p_fabric)
{
  // go over all nodes in the fabric
  for( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
       nI != p_fabric->NodeByName.end(); nI++) {
    IBNode *p_node = (*nI).second;
    if (p_node->type == IB_CA_NODE)
      continue;

    // allocate ARTraceRouteNodeInfo
    p_node->appData1.ptr = new ARTraceRouteNodeInfo(p_node);
    if (p_node->appData1.ptr == NULL) {
        cout << "-E- Failed to allocate memory for ARTraceRoute" << endl;
        return 1;
    }
  }
  return 0;
}

void ARTraceRouteNodeInfo::cleanup(IBFabric *p_fabric)
{
  // go over all nodes in the fabric
  for( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
       nI != p_fabric->NodeByName.end(); nI++) {
    IBNode *p_node = (*nI).second;
    if (p_node->type == IB_CA_NODE)
      continue;

    delete (ARTraceRouteNodeInfo*)p_node->appData1.ptr;
    p_node->appData1.ptr = NULL;
  }
}

void ARTraceRouteNodeInfo::clearDB(IBFabric *p_fabric)
{
  // go over all nodes in the fabric
  for( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
       nI != p_fabric->NodeByName.end(); nI++) {
    IBNode *p_node = (*nI).second;
    if (p_node->type == IB_CA_NODE)
      continue;

    ARTraceRouteNodeInfo *p_nodeInfo =
        (ARTraceRouteNodeInfo*)p_node->appData1.ptr;

    p_nodeInfo->visitCount = 0;

    UsedRouteInfo::iterator iter=p_nodeInfo->m_usedRouteInfo.begin();
    for ( ; iter != p_nodeInfo->m_usedRouteInfo.end() ; ++iter) {
        (*iter)->clear();
    }
    p_nodeInfo->m_usedRouteInfo.clear();
  }
}

void ARTraceRouteNodeInfo::checkDB(IBFabric *p_fabric, lid_t dlid)
{
  // go over all nodes in the fabric
  for( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
       nI != p_fabric->NodeByName.end(); nI++) {
    IBNode *p_node = (*nI).second;
    if (p_node->type == IB_CA_NODE)
      continue;

    ARTraceRouteNodeInfo *p_nodeInfo =
        (ARTraceRouteNodeInfo*)p_node->appData1.ptr;

    p_nodeInfo->visitCount = 0;

    UsedRouteInfo::iterator iter=p_nodeInfo->m_usedRouteInfo.begin();
    for ( ; iter != p_nodeInfo->m_usedRouteInfo.end() ; ++iter) {
        if(!(*iter)->convertDestLid(dlid)){
            //cout << "-D- clear cache" << endl;
            clearDB(p_fabric);
            return;
        }
    }
  }
}
