/*
 * Copyright (c) 2004-2018 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <iomanip>
#include <fstream>
#include <sstream>
#include <map>

#include <infiniband/ibdm/Congestion.h>
#include <infiniband/ibdm/SubnMgt.h>
#include <infiniband/ibdm/TraceRoute.h>


using namespace std;

/*
 * This file provides an API for analyzing congestion on the fabric
 * Link over subscription statistics are being provided as follows:
 *
 * CongInit(p_fabric) - initialize the data structures
 * CongZero(p_fabric) - zero out the paths per link counter
 * CongTrackPath(p_fabric, srcLid, dstLid) - trace one path
 * CongReport(p_fabric, stream &out) - report histogram
 * CongDump(p_fabric, fileName) - dump out the paths
 * CongCleanup(p_fabric) - cleanup
 *
 * The implemented algorithm is based on availablity of the FDBs
 */


//////////////////////////////////////////////////////////////////////////////
// the main data structure we keep is the list if paths going out
// on every out port of every node.

class flowData {
public:
  lid_t src;
  lid_t dst;
  double GuessBW;
  double ResBW;
  IBPort *minRateOutPort;
  map< IBPort *, double > portsBW;
  flowData(IBPort *p, lid_t s, lid_t d, double b) {
    src = s; dst = d; 
    GuessBW = b; ResBW = 0; 
    portsBW[p] = b;
    minRateOutPort = p;
  };
};

typedef pair< flowData*, phys_port_t > flow_in_port_pair;
typedef list< flow_in_port_pair > list_flow_in_port_pairs;
typedef map< IBPort *, list_flow_in_port_pairs, less<IBPort * > > map_pport_paths;
typedef map< IBPort *, int , less<IBPort * > > map_pport_int;
typedef map< int, float > map_int_float;

// calculate port rate in GBps
//it's speed in (Gbits * encoding)/8 to get Bytes
//so for 8/10 encoding, it's speed * 8/10 * 1/8 = speed/10
//for 64/66 encoding, it's speed * 64/66 * 1/8 = speed/8.25
//64/66 speeds are:
//FDR10 10.3125  Gbps  -> 10.3125/8.25  = 10/8
//FDR   14.0625  Gbps  -> 14.0625/8.25  = 13.63636364/8
//EDR20 20.625   Gbps  -> 20.625/8.25   = 20/8
//EDR   25.78125 Gbps  -> 25.78125/8.25 = 25/8
//HDR   53.125   Gbps  -> 53.125/8.25   = 53.125/8.25
double
getPortRateGBps(IBPort *p_port)
{
  double rate_GBps = 0;
  switch (p_port->get_common_speed()) {
  case IB_UNKNOWN_LINK_SPEED:
    cout << "-W- Ignoring unknown speed for port:" << p_port->getName() << endl;
    rate_GBps = 0;
    break;
  case IB_LINK_SPEED_2_5:    //SDR
    rate_GBps = 2.500/10.0;
    break;
  case IB_LINK_SPEED_5:      //DDR
    rate_GBps = 5.000/10.0;
    break;
  case IB_LINK_SPEED_10:     //QDR
    rate_GBps = 10.000/10.0;
    break;
  case IB_LINK_SPEED_14:     //FDR
    rate_GBps = 13.63636364/8.0;
    break;
  case IB_LINK_SPEED_25:     //EDR
    rate_GBps = 25.000/8.0;
    break;
  case IB_LINK_SPEED_50:     //HDR
    rate_GBps = 53.125/8.25;
    break;
  case IB_LINK_SPEED_FDR_10: //FDR10
    rate_GBps = 10.000/8.0;
    break;
  case IB_LINK_SPEED_EDR_20: //EDR20
    rate_GBps = 20.000/8.0;
    break;
  default:
    cout << "-E- Unknown link speed??? " << p_port->get_common_speed() << endl;
    exit(1);
    break;
  }
  
  switch (p_port->get_common_width()) {
  case IB_UNKNOWN_LINK_WIDTH:
    cout << "-W- Ignoring unknown speed for port:" << p_port->getName() << endl;
    rate_GBps = 0;
    break;
  case IB_LINK_WIDTH_1X:
    break;
  case IB_LINK_WIDTH_4X:
    rate_GBps *= 4.0;
    break;
  case IB_LINK_WIDTH_8X:
    rate_GBps *= 8.0;
    break;
  case IB_LINK_WIDTH_12X:
    rate_GBps *= 12.0;
    break;
  case IB_LINK_WIDTH_2X:
    rate_GBps *= 2.0;
    break;
  }
  return(rate_GBps);
}

// for each fabric we keep:
class CongFabricData {
private:
public:
  map_pport_paths portPaths;
  map_pport_int   portNumPaths;
  vector< flowData* > stageFlows;
  long            numPaths;
  int             stageWorstCase;
  int             worstWorstCase;
  list<int>       stageWorstCases;
  vec_int         numPathsHist;
  IBPort          *p_worstPort;
  int             maxRank;
  vec_int         flowBWHist;
  vec_int         stageWorstFlowBWHist;
  list<double>    stageAggBW;
  double          maxStageAggBW;
  double          minFlowBW;
  double          maxHostBW;
  bool            calcBW;
  bool            calcBWWithCC;
  CongFabricData() {
    stageWorstCase = 0;
    worstWorstCase = 0;
    p_worstPort = NULL;
    numPaths = 0;
    maxRank = 0;
    maxHostBW = 0.0;
    maxStageAggBW = 0.0;
    flowBWHist.resize(11,0);
    stageWorstFlowBWHist.resize(11,0);
    calcBW = false;
    calcBWWithCC = false;
  };
  
  inline unsigned int getBWHistBin(double bw) {
    // HACK we always use 10 bins
    return((unsigned int)(10.0*bw/maxHostBW));
  }
};

typedef std::map< IBFabric*, CongFabricData, std::less< IBFabric *> > map_pfabric_cong;

map_pfabric_cong CongFabrics;


//////////////////////////////////////////////////////////////////////////////
// Initialize the congestion tracking for the fabric.
int
CongInit(IBFabric *p_fabric, bool ccMode)
{
    if (CongFabrics.find(p_fabric) != CongFabrics.end()) {
        cout << "-E- Congestion Tracker already initialized" << endl;
        return(1);
    }

    CongFabrics[p_fabric] = CongFabricData();
    CongFabrics[p_fabric].calcBWWithCC = ccMode;
    map_pport_paths &portPaths = CongFabrics[p_fabric].portPaths;

    // init the vector for each connected port of the fabric
    // also obtain the max BW of the fabric
    IBNode *p_node;
    IBPort *p_port;
    for( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
            nI != p_fabric->NodeByName.end();
            nI++) {
        p_node = (*nI).second;
        if (p_node->rank > CongFabrics[p_fabric].maxRank)
                CongFabrics[p_fabric].maxRank = p_node->rank;

        // use unsigned int index to prevent overwrap
        for(unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
            p_port = p_node->getPort((phys_port_t)pn);
            if (!p_port || ! p_port->p_remotePort)
                continue;
            portPaths[p_port] = list_flow_in_port_pairs();

	    double BW = getPortRateGBps(p_port);
	    if (BW > CongFabrics[p_fabric].maxHostBW) {
	      CongFabrics[p_fabric].maxHostBW = BW;
	    }
        }
    }
    return(0);
}


// Cleanup...
int
CongCleanup(IBFabric *p_fabric)
{
    // get the reference to the actual data structure:
    map_pfabric_cong::iterator cI = CongFabrics.find(p_fabric);
    if (cI == CongFabrics.end()) {
        cout << "-E- Congestion Tracker not previously initialized" << endl;
        return(1);
    }
    // remove the entry should delete the entire structure:
    CongFabrics.erase(cI);
    return(0);
}

struct lessFlow
{
  bool operator()(const flowData* s1, const flowData* s2) const
  {
    return ( (s1->GuessBW < s2->GuessBW) ||
	     ( (s1->GuessBW == s2->GuessBW) && 
	       ( (s1->src < s2->src) ||
		 ( (s1->src == s2->src) && (s1->dst < s2->dst) ))));
  }
};

typedef set< flowData*, lessFlow > set_flows;

// update flow bandwidth if needed keeping the min guess updated 
// return 1 if some update was done
int
updateFlowBW(flowData* pFlow, double bw, IBPort *p_port, set_flows &GuessBW)
{
  double prevGuessBW = pFlow->GuessBW;
  bool updated = false;
  double minBW;

  if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) 
    cout << "-V- Checking new bw: " << bw << " for flow:" << pFlow->src << "," << pFlow->dst
	 << " after change on:" << p_port->getName() << endl;

  // we need to update this flow
  pFlow->portsBW[p_port] = bw;
  if ((prevGuessBW <= 0) || (bw <= prevGuessBW)) {
    minBW = bw;
    pFlow->minRateOutPort = p_port;
    updated = true;
  } else {
    minBW = bw;
    // find the new smallest flow
    map< IBPort *, double >::iterator pbwI;
    for (pbwI = pFlow->portsBW.begin(); pbwI != pFlow->portsBW.end(); pbwI++) {
      if ((*pbwI).second < minBW) {
	minBW = (*pbwI).second;
	pFlow->minRateOutPort = (*pbwI).first;
      }
    }
    if (minBW != prevGuessBW) {
      updated = true;
    }
  }

  // need to delete and re-insert the new min bw
  if (updated) {					
    if (prevGuessBW > 0) {
      set_flows::iterator gbwI = GuessBW.find(pFlow);
      if (gbwI == GuessBW.end()) {
	cout << "-E- BUG! Could not find the previously allocated BW of flow src:"
	     << pFlow->src << " dst:" << pFlow->dst << endl;
	exit(1);
      }
      GuessBW.erase(gbwI);
    }
    pFlow->GuessBW = minBW;
    GuessBW.insert(pFlow);
  
  
    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {
      cout << "-V- Updated Guess BW:" <<  pFlow->GuessBW
	   << " was: " << prevGuessBW << " for flow: " 
	   << pFlow->src << "," << pFlow->dst << " after change on:"
	   << p_port->getName() << endl;
    }
    return(1);
  }
  return(0);
}

// update the guess of BW for all flows going through the given output switch port
int
CongCalcOutPortFlowGuessBW(IBPort *p_port, set_flows &GuessBW, CongFabricData &congData)
{
  // now inspect all flows through the port split them into groups 
  // for CC mode - just one group
  // for Regular mode - by in port
  map_pport_paths::iterator pI = congData.portPaths.find(p_port);
  if (pI == congData.portPaths.end()) {
    cout << "-E- BUG: could not find port paths for port:" << p_port->getName() << endl;
    return(1);
  }

  map< phys_port_t, set< flowData*> > flowGroups;
  map< phys_port_t, double > prevAllocBWPerGroup;
  
  list_flow_in_port_pairs::iterator lI;
  phys_port_t grpPort;
  for (lI = (*pI).second.begin(); lI != (*pI).second.end(); lI++) {
    flowData *pFlow = (*lI).first;
    phys_port_t inPortNum = (*lI).second;
    double prevResBW = pFlow->ResBW;
    if (congData.calcBWWithCC) {
      // we have one group only - the out port
      grpPort = p_port->num;
    } else {
      // Parking-Lot so need to know the in port
      grpPort = inPortNum;
    }
	
    if (prevResBW > 0.0) {
      // accumulate previously assigned BW for ports on this group
      if (prevAllocBWPerGroup.find(grpPort) == prevAllocBWPerGroup.end()) {
	prevAllocBWPerGroup[grpPort] = prevResBW;
      } else {
	prevAllocBWPerGroup[grpPort] += prevResBW;
      }
    } else {
      // this port is still not assigned a BW so remember it
      flowGroups[grpPort].insert(pFlow);
    }
  } // all flows on the port
 
  // nothing to do if there are no Guess flows on this out port
  if (flowGroups.empty()) 
    return(0);
 
  if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) 
    cout << "-V- Found " << flowGroups.size() << " in ports (with guess flows) sending flows through:"
	 << p_port->getName() << endl;
  
  double portBW = getPortRateGBps(p_port);

  // if there are groups that are fully assigned (do not have guess BW)
  // remove the total BW from the port
  map< phys_port_t, double >::iterator aI;
  for (aI = prevAllocBWPerGroup.begin(); aI != prevAllocBWPerGroup.end(); aI++) 
    if (flowGroups.find((*aI).first) == flowGroups.end()) 
      portBW -= (*aI).second;
  
  // now we need to split the total port BW to each group
  double grpTotalBW = (double)portBW / (double)flowGroups.size();
  
  // loop on each group marking smaller flows
  map< phys_port_t, set< flowData*> >::iterator gI;
  for (gI = flowGroups.begin(); gI != flowGroups.end(); gI++) {
    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) 
      cout << "-V- Handling " << (*gI).second.size() << " flows through in-port:"
	   << (unsigned int)(*gI).first << endl;

    // if there are previously assigned ports remove their BW from the group bw
    double thisGroupBW = grpTotalBW;
    aI = prevAllocBWPerGroup.find((*gI).first);
    if (aI != prevAllocBWPerGroup.end()) {
      thisGroupBW -= (*aI).second;
    }
    double grpFlowBW = (double)thisGroupBW / (double)(*gI).second.size();
    set< flowData*>::iterator fI;
    for (fI = (*gI).second.begin(); fI != (*gI).second.end(); fI++) {
      flowData *pFlow = (*fI);

      // was this flow already assigned?
      if (pFlow->ResBW > 0.0) continue;
      
      updateFlowBW(pFlow, grpFlowBW, p_port, GuessBW);
    }
  }
  return(0);
}

// calculate the minimal guess by inspecting each output port and 
int
CongCalcGuessBW(IBFabric *p_fabric, set_flows &GuessBW, CongFabricData &congData)
{
  for (map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
       nI != p_fabric->NodeByName.end();
       nI++) {
    IBNode *p_node = (*nI).second;
    if (p_node->type != IB_SW_NODE) continue;

    for (unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
      IBPort *p_port = p_node->getPort((phys_port_t)pn);
      if (!p_port || !p_port->p_remotePort) continue;

      if (CongCalcOutPortFlowGuessBW(p_port, GuessBW, congData))
	return(1);
    }
  }
  return(0);
}

// given a switch node, inPort and a flow return the
// list of other flows going through the given in port
// return the number of found flows 
int
getSwitchFlowsThroughInPort(CongFabricData &congData, 
			    IBNode *pNode, 
			    phys_port_t inPortNum, 
			    flowData *pFlow, 
			    list< flowData*> &inPortFlows)
{
  unsigned int numFlows = 0;
  // go over all the ports of the switch and collect the 
  // other flows on that port that match the inPort num
  for (unsigned int pn = 1; pn <= pNode->numPorts; pn++) {
    IBPort *pPort = pNode->getPort((phys_port_t)pn);
    if (!pPort || !pPort->p_remotePort) continue;

    map_pport_paths::iterator pI = congData.portPaths.find(pPort);
    if (pI == congData.portPaths.end()) continue;
    
    // go over all flows on that port collecting ones from given 
    // inPortNum and not the given flow
    list_flow_in_port_pairs::iterator lI;
    for (lI = (*pI).second.begin(); lI != (*pI).second.end(); lI++) {
      flowData *pFlow1 = (*lI).first;
      phys_port_t inPNum = (*lI).second;

      if ((pFlow1 != pFlow) && (inPNum == inPortNum)) {
	inPortFlows.push_front(pFlow1);
	numFlows++;
      }
    }
  }
  return(numFlows);
}

// given a flow that was assigned a ResBW value go update all
// afected flows GuessBW
// This include recalculating GuessBW on downstream and upstream switches
// and also forcing in no CC case the GuessBW of all flows sharing
// same input port on upstream switches
int
CongUpdateAffectedFlows(IBFabric *p_fabric, 
			CongFabricData &congData, 
			flowData *pFlow, 
			set_flows &GuessBW)
{
  bool beforeMinRatePort = true;
  double bw = pFlow->ResBW;

  // traverse the flow path from source to destination
  IBPort *p_port = p_fabric->getPortByLid(pFlow->src);
  IBPort *p_dstPort = p_fabric->getPortByLid(pFlow->dst);
  while (p_port && p_port->p_remotePort) {
    if (p_port->p_remotePort == p_dstPort) 
      return(0);

    IBNode *p_node = p_port->p_remotePort->p_node;
    if (p_node->type != IB_SW_NODE) {
      cout << "-E- Updating flow src:" << pFlow->src << " dst:" << pFlow->dst
           << " remote node to port:" << p_port->getName() << " is not a SWITCH node? "
           << "Is this fabric just 2 hosts connected back to back?" << endl;
      return(1);
    }
    phys_port_t inPortNum = p_port->p_remotePort->num;
    phys_port_t outPortNum = p_node->getLFTPortForLid(pFlow->dst);
    p_port = p_node->getPort(outPortNum);
    if (!p_port || ! p_port->p_remotePort) {
      cout << "-E- No out port for flow src:" << pFlow->src << " dst:"
	   << pFlow->dst << " on switch:" << p_node->name
           << " port:" << (unsigned int)outPortNum << ". Is the FDB correct?" << endl;
      return(1);
    }

    // update this out port anyway
    if (CongCalcOutPortFlowGuessBW(p_port, GuessBW, congData))
      return(1);
    
    // now if we need to (not CC mode) and (not downstream) update other flows 
    // sharing the same in port
    if (!congData.calcBWWithCC && beforeMinRatePort) {
    
      // get all flows on that switch that go through that input port
      list< flowData*> inPortFlows;
      set<phys_port_t> affectedOutPortNums;
      
      getSwitchFlowsThroughInPort(congData, p_node, inPortNum, pFlow, inPortFlows);
      list< flowData*>::iterator fI;
      for (fI = inPortFlows.begin(); fI != inPortFlows.end(); fI++) {
	// update the GuessBW on these flows
	flowData *pAffectedFlow = (*fI);

	// was this flow already assigned?
	if (pAffectedFlow->ResBW > 0.0) continue;

	// note: we are using a p_port of the flow that forces this port
	// by its sharing of input buffer (HOL blocking),
	// so we cannot by mistake clean it based on the normal flows
	if (updateFlowBW(pAffectedFlow, bw, p_port, GuessBW)) {
	  // collect the affected output port nums
	  phys_port_t affectedPortNum = p_node->getLFTPortForLid(pAffectedFlow->dst);
	  if (affectedPortNum != outPortNum) 
	    affectedOutPortNums.insert(affectedPortNum);
	} // if  update 
      } // all affected flows
      
      // update the affected out ports
      set<phys_port_t>::iterator apI;
      for (apI =  affectedOutPortNums.begin(); apI !=  affectedOutPortNums.end(); apI++) {
	IBPort *affPort = p_node->getPort(*apI);
	if (!affPort) continue;
	if (CongCalcOutPortFlowGuessBW(affPort, GuessBW, congData))
	  return(1);
      }
    } else {
      // CC mode or after the the forcing port
      if (CongCalcOutPortFlowGuessBW(p_port, GuessBW, congData))
	return(1);
      
    } // need to handle affected flows on same input port

    // we need to use the node of the switch instead of the port.
    // the reason is that for HOL blocking we store the port of the
    // output creating the cong - not the FDB out of the flow
    if (p_port->p_node == pFlow->minRateOutPort->p_node) {
      beforeMinRatePort = false;
    }
    // Note p_port was previously advanced

  } // while ports on the path

  cout << "-E- Did not reach dst port for flow src:"
       << pFlow->src << " dst:" << pFlow->dst << endl;
  return(1);
}

// analyze a single stage BW per flow
int
CongCalcBW(IBFabric *p_fabric, bool dump, ostream &dumpF)
{
  // get the reference to the actual data structure:
  map_pfabric_cong::iterator cI = CongFabrics.find(p_fabric);
  if (cI == CongFabrics.end()) {
    cout << "-E- Congestion Tracker not previously initialized." << endl;
    return(1);
  }
  
  CongFabricData &congData = (*cI).second;
  set_flows GuessBW;
  congData.calcBW = true;

  // init all flows as being guessed
  bool anyDup = false;
  for (size_t f = 0; f < congData.stageFlows.size(); f++) {
    if (GuessBW.find(congData.stageFlows[f]) != GuessBW.end()) {
      cout << "-E- Duplicated flow:" << congData.stageFlows[f]->src
	   << " to:" << congData.stageFlows[f]->dst << endl;
      anyDup = true;
    } else {
      GuessBW.insert(congData.stageFlows[f]);
    }
  }
  if (anyDup) {
    exit(1);
  }
  
  // loop until no more flows to guess
  while (!GuessBW.empty()) {
    // calculate GuessBW by inspecting every switch outport flows
    CongCalcGuessBW(p_fabric, GuessBW, congData);
    
    // select the lowest GuessBW flow and remove from un-known
    set_flows::iterator minGuessBwFlowI = GuessBW.begin();
    flowData *pFlow = (*minGuessBwFlowI);
    GuessBW.erase(minGuessBwFlowI);

    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) 
      cout << "-V- Applying ResBW: " << pFlow->GuessBW 
	   << " to flow: " << pFlow->src << "," << pFlow->dst << endl;

    // set the flow BW to that value
    pFlow->ResBW = pFlow->GuessBW;

    // set affected flows sharing the flow upstream input buffers
    CongUpdateAffectedFlows(p_fabric, congData, pFlow, GuessBW);
  }
  
  // report
  unsigned int minBin = 0;
  double aggBW = 0.0;
  for (unsigned int i = 0; i < congData.stageFlows.size(); i++) {
    double bw = congData.stageFlows[i]->ResBW;
    aggBW+= bw;
    unsigned int bin = congData.getBWHistBin(bw);
    congData.flowBWHist[bin]++;
    if (!i || (bin && (bin < minBin))) {
      minBin = bin;
    }
    if (dump) {
      lid_t src = congData.stageFlows[i]->src;
      lid_t dst = congData.stageFlows[i]->dst;
      IBPort *srcPort = p_fabric->getPortByLid(src);
      IBPort *dstPort = p_fabric->getPortByLid(dst);
      dumpF << srcPort->getName() << ", " << src << ", " << dstPort->getName() << ", " << dst 
	    << ", " << fixed << setprecision(2) << bw << ", " << aggBW << endl;
    }
  }
  congData.stageAggBW.push_back(aggBW);
  if (aggBW > congData.maxStageAggBW)
    congData.maxStageAggBW = aggBW;
  congData.stageWorstFlowBWHist[minBin]++;
  for (unsigned int i = 0; i < congData.stageFlows.size(); i++) {
    delete congData.stageFlows[i];
  }
  congData.stageFlows.clear();

  return(0);
}

// analyze a single stage. We propogate a floating point fraction of
// link BW with each src,dst pair such that cong apear only on first
// link a contention run through
int
CongZero(IBFabric *p_fabric)
{
  map_int_float dst_frac;
  IBPort *stageWorstPort = NULL;
  
  // get the reference to the actual data structure:
  map_pfabric_cong::iterator cI = CongFabrics.find(p_fabric);
  if (cI == CongFabrics.end()) {
    cout << "-E- Congestion Tracker not previously initialized." << endl;
    return(1);
  }
  
  CongFabricData &congData = (*cI).second;
  congData.stageWorstCase = 0;
  
  for (map_pport_paths::iterator pI = congData.portPaths.begin();
       pI != congData.portPaths.end();
       pI++) {						
    IBPort *p_port = (*pI).first;
    int numPaths = (int)(*pI).second.size();
    congData.portNumPaths[p_port] = numPaths;
    
    if (congData.numPathsHist.size() <= (size_t)numPaths)
      congData.numPathsHist.resize(numPaths+1, 0);
    congData.numPathsHist[numPaths]++;
    
    if (congData.stageWorstCase < numPaths) {
      congData.stageWorstCase = numPaths;
      stageWorstPort = p_port;
    }
    
    // clear the list for the future ...
    (*pI).second.clear();
  }

  congData.stageWorstCases.push_back(congData.stageWorstCase);
  if (congData.stageWorstCase > congData.worstWorstCase) {
    congData.worstWorstCase = congData.stageWorstCase;
    congData.p_worstPort = stageWorstPort;
  }

  congData.portNumPaths.clear();
  return (0);
}

// Track a single path
int
CongTrackPath(IBFabric *p_fabric, lid_t srcLid, lid_t dstLid, double ratio)
{
    // get the reference to the actual data structure:
    map_pfabric_cong::iterator cI = CongFabrics.find(p_fabric);
    if (cI == CongFabrics.end()) {
        cout << "-E- Congestion Tracker not previously initialized" << endl;
        return(1);
    }

    CongFabricData &congData = (*cI).second;

    // find the source and destination ports:
    IBPort *p_srcPort = p_fabric->getPortByLid(srcLid);
    if (! p_srcPort) {
        cout << "-E- Fail to find port by source LID:" << srcLid << endl;
        return(1);
    }

    IBPort *p_dstPort = p_fabric->getPortByLid(dstLid);
    if (! p_dstPort) {
        cout << "-E- Fail to find port by destination LID:" << dstLid << endl;
        return(1);
    }

    double srcRate = getPortRateGBps(p_srcPort);
    flowData *pFlow = new flowData(p_srcPort, srcLid, dstLid, srcRate*ratio);
    congData.stageFlows.push_back(pFlow);
    IBNode *p_node;
    IBPort *p_port = p_srcPort;
    IBPort *p_remotePort = NULL;
    phys_port_t remPortNum = 0;
    int hopCnt = 0;

    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {
        cout << "-V-----------------------------------------------------" << endl;
        cout << "-V- Tracing from lid:" << srcLid << " to lid:"
                << dstLid << endl;
    }

    // if the port is not a switch - go to the next switch:
    if (p_port->p_node->type != IB_SW_NODE) {
        // try the next one:
        if (!p_port->p_remotePort) {
            cout << "-E- Provided starting point is not connected !"
                    << "lid:" << srcLid << endl;
            return 1;
        }

        p_remotePort = p_port->p_remotePort;
        p_node = p_remotePort->p_node;
	remPortNum = p_remotePort->num;
        hopCnt++;
        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            cout << "-V- Arrived at Node:" << p_node->name
		 << " Port:" << (unsigned int)p_port->p_remotePort->num << endl;
    } else {
        // it is a switch :
        p_node = p_port->p_node;
    }

    // verify we are finally on a switch:
    if (p_node->type != IB_SW_NODE) {
        cout << "-E- Provided starting point is not connected to a switch !"
                << "lid:" << srcLid << endl;
        return 1;
    }

    // traverse:
    int done = 0;
    while (!done) {
      // calc next node:
      phys_port_t pn = p_node->getLFTPortForLid(dstLid);
      if (pn == IB_LFT_UNASSIGNED) {
	cout << "-E- Unassigned LFT for lid:" << dstLid
	     << " Dead end at:" << p_node->name << endl;
	return 1;
      }
      
      // if the port number is 0 we must have reached the target node.
      // simply try see that p_remotePort of last step == p_dstPort
      if (pn == 0) {
	if (p_dstPort != p_remotePort) {
	  cout << "-E- Dead end at port 0 of node:" << p_node->name << endl;
	  return 1;
	}
	return 0;
      }
      
      // get the port on the other side
      p_port = p_node->getPort(pn);
      
      // we need to store this info for marking later
      list_flow_in_port_pairs &lst = congData.portPaths[p_port];
      lst.push_front(flow_in_port_pair(pFlow, remPortNum));

      if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
	cout << "-V- Going out on port:" << (unsigned int)pn << endl;
      
      if (! (p_port &&
	     p_port->p_remotePort &&
	     p_port->p_remotePort->p_node)) {
	cout << "-E- Dead end at:" << p_node->name << endl;
	return 1;
      }
      
      if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
	cout << "-V- Arrived at Node:" << p_port->p_remotePort->p_node->name
	     << " Port:" << (unsigned int)p_port->p_remotePort->num << endl;
      
      p_remotePort = p_port->p_remotePort;
      remPortNum = p_remotePort->num;
		     
      // check if we are done:
      done = (p_remotePort == p_dstPort);
      
      p_node = p_remotePort->p_node;
      if (hopCnt++ > 256) {
	cout << "-E- Aborting after 256 hops - loop in LFT?" << endl;
	return 1;
      }
    }
    
    congData.numPaths++;
    return(0);
}


// Report the worst usage and a histogram of link usage
int
CongReport(IBFabric *p_fabric, ostream &out)
{
  out << "-I- Final Congestion Report" << endl;
  int worstWorstPath = 0;
  map<int, int, less<int> > stageWorstCaseHist;
  
  // get the reference to the actual data structure:
  map_pfabric_cong::iterator cI = CongFabrics.find(p_fabric);
  if (cI == CongFabrics.end()) {
    cout << "-E- Congestion Tracker not previously initialized" << endl;
    return(1);
  }
  
  CongFabricData &congData = (*cI).second;
  // collect the histogram of stage worst paths
  for(list<int>::iterator lI = congData.stageWorstCases.begin();
      lI != congData.stageWorstCases.end();
      lI++) {
    stageWorstCaseHist[*lI]++;
    if (worstWorstPath < *lI) worstWorstPath = (*lI);
  }
  
  out << "---------------------------------------------------------------------------\n" << endl;
  out << "-I- Traced total:" << congData.numPaths << " paths" << endl;
  
  out << "-I- Worst link over subscrition:" << worstWorstPath
      << " port:" << congData.p_worstPort->getName() << endl;
  
  out << "---------------------- TOTAL CONGESTION HISTOGRAM ------------------------" << endl;
  out << "Describes distribution of oversubscription of paths per port." << endl;
  out << "NUM-PATHS NUM-OUT-PORTS" << endl;
  for (unsigned int b = 0; b < congData.numPathsHist.size(); ++b)
    if (congData.numPathsHist[b])
      out << setw(4) << b << "   " << congData.numPathsHist[b] << endl;
  out << "---------------------------------------------------------------------------\n" << endl;
  
  out << "---------------------- STAGE CONGESTION HISTOGRAM ------------------------" << endl;
  out << "Describes distribution of worst oversubscription of paths per stage." << endl;
  out << "WORST-CONG NUM-STAGES" << endl;
  for (map<int, int, less<int> >::iterator bI = stageWorstCaseHist.begin();
       bI != stageWorstCaseHist.end();
       bI++)
    out << setw(4) << (*bI).first << "   " << (*bI).second << endl;
  out << "---------------------------------------------------------------------------\n" << endl;

  if (congData.calcBW) {
    out << "---------------------- FLOW BW HISTOGRAM ------------------------" << endl;
    out << "Describes distribution of flows BW [GB/sec] accross all stages." << endl;
    out << "BW-RANGE NUM-FLOWS" << endl;
    double deltaBW = congData.maxHostBW/10.0;
    for (unsigned int b = 0; b < congData.flowBWHist.size(); ++b)
      out << fixed << setprecision(2) << setw(3) << b*deltaBW << "," << (b+1)*deltaBW << "   " 
	  << congData.flowBWHist[b] << endl;
    out << "---------------------------------------------------------------------------\n" << endl;
    
    out << "---------------------- STAGE WORST FLOW BW ------------------------" << endl;
    out << "Describes distribution of worst flow BW [GB/sec] per stage." << endl;
    out << "BW-RANGE NUM-STAGES" << endl;
    for (unsigned int b = 0; b < congData.flowBWHist.size(); ++b)
      out << fixed << setprecision(2) << setw(3) << b*deltaBW << "," << (b+1)*deltaBW << "   "
	  << congData.stageWorstFlowBWHist[b] << endl;
    out << "---------------------------------------------------------------------------\n" << endl;

    out << "---------------------- STAGE AGGREGATE FLOW BW ------------------------" << endl;
    out << "Describes distribution of Total Flows BW [GB/sec] per stage." << endl;
    out << "BW-RANGE NUM-STAGES" << endl;
    list<double>::iterator aI;
    vec_int stageAggregateBWHist(11,0);
    for ( aI = congData.stageAggBW.begin(); aI != congData.stageAggBW.end(); aI++) {
      unsigned int bin = (unsigned int)(10.0*(*aI)/congData.maxStageAggBW);
      stageAggregateBWHist[bin]++;
    }
    double deltaAggBW = congData.maxStageAggBW/10.0;
    for (unsigned int b = 0; b < stageAggregateBWHist.size(); ++b)
      out << fixed << setprecision(2) << setw(10) << b*deltaAggBW << "," 
	  << setw(10) << (b+1)*deltaAggBW << "   "
	  << stageAggregateBWHist[b] << endl;
    out << "---------------------------------------------------------------------------\n" << endl;

  }
  return(0);
}

// Dump out all link usages and details if available into
// the given ostream
int
CongDump(IBFabric *p_fabric, ostream &out)
{
  // get the reference to the actual data structure:
  map_pfabric_cong::iterator cI = CongFabrics.find(p_fabric);
  if (cI == CongFabrics.end()) {
    cout << "-E- Congestion Tracker not previously initialized" << endl;
    return(1);
  }

  CongFabricData &congData = (*cI).second;
  
  // go over all ports and dump out their paths...
  for (map_pport_paths::iterator pI = congData.portPaths.begin();
       pI != congData.portPaths.end();
       pI++) {
    // NOTE: we can not use here the congData.portNumPaths as it is set by
    // by CongZero which clears the list of paths...
    int numPaths = (int)(*pI).second.size();
    if (numPaths > 1) {
      out << "PORT:" << (*pI).first->getName() << " NUM:" << numPaths << endl;
      for (list_flow_in_port_pairs::iterator lI = (*pI).second.begin();
	   lI != (*pI).second.end();
	   lI++) {
	flowData *pFlow = (*lI).first;
	out << pFlow->src << " " << pFlow->dst << endl;
      }
    }
  }
  return(0);
}

// Dump out all link usages and details if available into
// the given ostream
int
CongDumpStage(ostream &f, 
	      IBFabric *fabric, 
	      unsigned int stage,
	      map< lid_t, unsigned int> &RankByLID)
{
  // get the reference to the actual data structure:
  map_pfabric_cong::iterator cI = CongFabrics.find(fabric);
  if (cI == CongFabrics.end()) {
    cout << "-E- Congestion Tracker not previously initialized" << endl;
    return(1);
  }

  CongFabricData &congData = (*cI).second;
  vector<unsigned int> numPairsHist(1,0);

  // first do the histogram
  for (map_pport_paths::iterator pI = congData.portPaths.begin();
       pI != congData.portPaths.end();
       pI++) {						
    unsigned int numPaths = (int)(*pI).second.size();
    if (numPairsHist.size() <= numPaths) 
      numPairsHist.resize(numPaths+1,0);
    numPairsHist[numPaths]++;
    if (numPaths > 1) {
      f << "PORT:" << (*pI).first->getName() << " NUM:" << numPaths << endl;
      for (list_flow_in_port_pairs::iterator lI = (*pI).second.begin();
	   lI != (*pI).second.end();
	   lI++) {
	flowData *pFlow = (*lI).first;
	IBPort *sPort = fabric->getPortByLid(pFlow->src);
	IBPort *dPort = fabric->getPortByLid(pFlow->dst);
	unsigned int sRank = RankByLID[pFlow->src];
	unsigned int dRank = RankByLID[pFlow->dst];
	
	f << sPort->getName() << " (LID/Rank " << pFlow->src << "/" << sRank << ") "
	  << dPort->getName() << " (LID/Rank " << pFlow->dst << "/" << dRank << ")" << endl;
      }
    }
  }

  f << "STAGE: " << stage << " NUM-PATHS:NUM-PORTS";
  for (unsigned int numPaths = 0; numPaths < numPairsHist.size(); numPaths++) {
    if (numPairsHist[numPaths])
      f << " " << numPaths << ":" << numPairsHist[numPaths];
  }
  f << endl;
  return(0);
}
