/*
 * Copyright (c) 2004-2019 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <set>
#include <algorithm>
#include <iomanip>

#include <infiniband/ibdm/CredLoops.h>
#include <infiniband/ibdm/TraceRoute.h>

#define MAX_ERROR_COUNT 1000

/*
 * Build a graph linked to the SW devices for input to output
 * links. We use the appData1 available on the nodes to ref the tables.
 * Based on this graph provide analysis on loops available on the
 * fabric
 *
 */
//////////////////////////////////////////////////////////////////////////////
//
// |------|  p1=p_port            |------| p2=p_portNext
// |      |                       |      |
// |    p1|>--------------------->|    p2|----------------
// |      |  \   channels         |      |   \   channels
// |------|   \--|=====|          |------|    \--|=====|
//               | VL0 |-|      next chan        | VL0 |
//               |-----| |     |========|        |-----|
//               | VL1 |  \    | P0-VL0 |   ---->| VL1 |
//               |-----|   \   |--------|  /     |-----|
//               | ... |    \->| P0-VL1 |-/      | ... |
//               |=====|  sl2vl|--------|        |=====|
//                             | P0-... |
//                             |========|
//                             | P1-VL0 |
//                             |--------|
//                             | P1-VL1 |
//                             |--------|
//                             | P1-... |
//                             |========|
//                             | ...... |
//                             |========|
//                             | Pn-VL0 |
//                             |========|
//
//////////////////////////////////////////////////////////////////////////////
using namespace std;

using namespace std;
using std::hex;
using std::setfill;
using std::setw;

//////////////////////////////////////////////////////////////////////////////
// We keep global flags to control how the check is being done:

// If non zero will consider all switch to switch paths too
static int CrdLoopIncludeUcastSwitchPaths = 0;

// If non zero consider multicast paths too
static int CrdLoopIncludeMcastPaths = 0;

// Map each MLID to a list of SL's that may be used for this MGRP traffic
// If no entry is set then we will assume all traffic on SL=0
#define map_mlid_sl_list  map<lid_t, list< uint8_t > >
static map_mlid_sl_list mlidSLs;

void CrdLoopCleanChannelsDfsState(IBFabric *p_fabric);

//////////////////////////////////////////////////////////////////////////////
// Apply DFS on a dependency graph
// Channels may have 3 states:
// Closed    - if was visited and is OK
// Open      - if it is on the stack
// Untouched - initial state
int CrdLoopDFS(CrdRoute/*VChannel* */&ch, /*list<VChannel *>*/list_crd_route &loop) {
  // Already been there
  if (ch.m_pvch->getFlag() == Closed)
    return 0;
  // Credit loop
  if (ch.m_pvch->getFlag() == Open) {
    loop.push_front(ch);
    return 1;
  }
  // Mark as open
  ch.m_pvch->setFlag(Open);
  // Make recursive steps
  for (int i=0; i<ch.m_pvch->getDependSize();i++) {
    CrdRoute &next = ch.m_pvch->getDependency(i);
    if (next.m_pvch) {
      if (CrdLoopDFS(next, loop)) {
        loop.push_front(ch);
        return 1;
      }
    }
  }
  // Mark as closed
  ch.m_pvch->setFlag(Closed);
  return 0;
}

//////////////////////////////////////////////////////////////////////////////

// Given a port a dlid and possible SL's
// traverse FDBs backwards to SRCs and return the LID of the source
// if SLID,DLID uses one of the provided SLs
lid_t
CredLoopDFSBwdToValidSrc(IBPort *port, lid_t dlid, set<uint8_t> &SLs)
{
  // stop on anything not a switch or on switch if we are in ALL mode.
  if ((port->p_node->type != IB_SW_NODE)|| CrdLoopIncludeUcastSwitchPaths) {
    uint8_t sl = port->p_node->getPSLForLid(dlid);
    if (SLs.find(sl) != SLs.end()) {
      return(port->base_lid);
    }
  }

  // it is a switch - go back through all ports not towards the DLID and
  // pointing to this switch for that dlid
  IBNode *node = port->p_node;
  // use unsigned int index to prevent overwrap
  for (unsigned int pn = 1; pn <= node->numPorts; pn++) {
    IBPort *iPort = node->getPort((phys_port_t)pn);
    if (!iPort || !iPort->p_remotePort) continue;
    IBPort *oPort = iPort->p_remotePort;
    IBNode *oNode = oPort->p_node;
    if ((oNode->type != IB_SW_NODE) ||
        (oNode->getLFTPortForLid(dlid) == oPort->num)) {
      lid_t slid = CredLoopDFSBwdToValidSrc(oPort, dlid, SLs);
      if (slid) return(slid);
    }
  }
  return(0);
}

//////////////////////////////////////////////////////////////////////////////

// Justify a Channel Dependency on a VL
/*
int CrdLoopJustifyDependence(VChannel *fromCh, VChannel *toCh,
                            lid_t &srcLid, lid_t &dstLid)
{
  IBPort *fromPort = fromCh->pPort;
  IBPort *toPort = toCh->pPort;
  srcLid = 0;
  dstLid = 0;

  // First we try unicast: look for all DLIDs flowing through the
  // first port and then through the second port. Also look for what SL's
  // would map to the given VL on the first port. Then try to route from
  // all SLIDs to the DLID and look for the first port on the path

  set<lid_t> DLIDs;
  IBNode *fromNode = fromPort->p_node;
  IBNode *toNode = toPort->p_node;
  phys_port_t fromPortNum = fromPort->num;
  phys_port_t toPortNum = toPort->num;
  phys_port_t inPortNum = fromPort->p_remotePort->num;
  for (lid_t l = 1; l <= fromNode->p_fabric->maxLid; l++) {
    // we may start with a CA but if not the out port for the dlid must match
    if ((fromNode->type == IB_SW_NODE) &&
        (fromNode->getLFTPortForLid(l) != fromPortNum))
      continue;

    // but if we go to a CA it is clearly an error
    if (toNode->type != IB_SW_NODE) {
      cout << "-E- Could not trace a credit loop into a CA !" << endl;
      return(1);
    }

    if (toNode->getLFTPortForLid(l) == toPortNum) {
      DLIDs.insert(l);
    }
  }

  // What SL's are satifying
  // toCh->vl == toNode->SL2VL(fromPort->remPort->num, toPort->num, sl)
  set<uint8_t> SLs;
  sl_vl_t slvl;
  for (uint8_t sl = 0; sl < 15; sl++) {
      slvl.SL = sl;
      slvl.VL = (u_int8_t)fromCh->vl;

    if (toNode->getVL(inPortNum,toPortNum,slvl) == toCh->vl) {
      SLs.insert(sl);
    }
  }

  // foreach DLID traverse back towards a source and then check if the
  // source is satisfying the SL mapping
  for (set<lid_t>::iterator dI = DLIDs.begin();
       dI != DLIDs.end(); dI++) {
    lid_t dlid = (*dI);
    lid_t slid = CredLoopDFSBwdToValidSrc(fromCh->pPort, dlid, SLs);
    if (slid) {
      srcLid = slid;
      dstLid = dlid;
      return(0);
    }
  }

  // No ucast routing found. Look for mcast.
  for (lid_t i = 0; i < toNode->MFT.size(); i++) {
    lid_t mlid = (lid_t)(i+0xc000);

    // is the MCG routing matching the 2 vcs?
    bool portsMatch = false;
    if ( (fromNode->MFT.size() > i) && (toNode->MFT.size() > i)) {
      uint64_t fromPortMask = fromNode->MFT[i];
      uint64_t toPortMask = toNode->MFT[i];
      if ( (fromPortMask & ((uint64_t)1 << fromPortNum)) &&
           (toPortMask & ((uint64_t)1 << toPortNum)) )
        portsMatch = true;
    }
    if (!portsMatch) continue;

    // lookup SL's of the MCG and see if matches one of the path SLs
    bool slMatch = false;
    map_mlid_sl_list::const_iterator mlidI = mlidSLs.find(mlid);
    if (mlidI != mlidSLs.end()) {
      for (list<uint8_t>::const_iterator slI = (*mlidI).second.begin();
           !slMatch && (slI != (*mlidI).second.end()); slI++) {
        if (SLs.find(*slI) == SLs.end()) {
          slMatch = true;
        }
      }
    } else {
      slMatch = true;
    }
    if (!slMatch) continue;

    // we got here if both ports match and sl match
    srcLid = mlid;
    dstLid = mlid;
    return(0);
  }

  return(1);
}
*/

//////////////////////////////////////////////////////////////////////////////

// Go over CA's apply DFS on the dependency graphs starting from CA's port

int CrdLoopFindLoops(IBFabric* p_fabric)
{
    static bool clean_channels_dfs_state = false;

    if (clean_channels_dfs_state) {
        //should be called if CrdLoopFindLoops was called before.
        CrdLoopCleanChannelsDfsState(p_fabric);
    }
    clean_channels_dfs_state = true;
    // go over all CA ports in the fabric
    for (lid_t i = p_fabric->minLid; i <= p_fabric->maxLid; i++) {
        IBPort *p_Port = p_fabric->getPortByLid(i);

        // HACK : can it be that there is a switch port 0 that starts a loop
        // that is not connected to any HCA path?
        if (!p_Port || (p_Port->p_node->type == IB_SW_NODE)) continue;

        // Go over all CA's channels and find untouched one
        for (uint8_t j=0;j < p_fabric->getNumVLs(); j++) {
            dfs_t state = p_Port->channels[j]->getFlag();
            if (state == Open) {
                cout << "-E- open channel outside of DFS" << endl;
                return 1;
            }
            // Already processed, continue
            if (state == Closed)
                continue;

            // Found starting point
            /*list<VChannel *> loop;*/
            list_crd_route loop;
            CrdRoute crdRoute(p_Port->channels[j],0,0);
            if (CrdLoopDFS(crdRoute, loop)) {
                cout << "-E- Credit loop found on the following path:" << endl;
                list_crd_route::iterator lI = loop.begin();
                list_crd_route::iterator pI = lI++;
                for ( ;lI != loop.end(); pI = lI, ++lI) {
                    cout << "    from port:" << pI->m_pvch->pPort->getName()
                    << " VL: " << pI->m_pvch->vl
                    << "  to port:" << lI->m_pvch->pPort->getName()
                    << " VL: " << lI->m_pvch->vl;
                    if (lI->m_slid != 0)
                        cout << " on path from lid: 0x" << setw(4) << setfill('0')
                                << hex << lI->m_slid << " to lid: 0x"  << setw(4)
                                << setfill('0')<< hex << lI->m_dlid;
                    else
                        cout << " on path to multicast lid: 0x" << setw(4) << setfill('0')
                        << hex << lI->m_dlid;

                    cout << dec << endl;
                }
                return 1;
            } //CrdLoopDFS
        }//SL loop
    }// lid loop
    return 0;
}

//////////////////////////////////////////////////////////////////////////////

// Trace a route from slid to dlid by AR LFT
// Add dependency edges recursively
int
CrdLoopMarkRouteByARLFT (IBFabric *p_fabric,
                         lid_t sLid , lid_t dLid, sl_vl_t slvl, IBPort *p_port,
                         int hopCnt, int &addedEdges,
                         list_lid &lids, IBNode *p_DstSwNode,
                         list_delayed_route_t &delayedRouteInfo)
{
    IBNode *p_node;
    IBPort *p_portNext;
    phys_port_t outPortNum = 0, inputPortNum = 0;
    phys_port_t sl2vlPortGroup;
    bool done;
    list_phys_ports portsList;
    int rc = 0;
    sl_vl_t nextSlvl;
    nextSlvl.SL = slvl.SL;
    stringstream sstream;
    string offset_str;

    if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE ) {
        for (int i = 0 ; i < hopCnt ; ++i)
            sstream << "  ";

        offset_str = sstream.str();

        if (lids.empty())
            cout << offset_str << "-I- DLID: 0x";
        else
            cout << offset_str << "-I- Group with DLID: 0x";

        cout << setw(4) << setfill('0')
            << hex << dLid << dec
            << " arrived at:"  << p_port->getName()
            << " SL: " << (int)slvl.SL
            << " VL: " << (int)slvl.VL << endl;
    }

    // check if we are done:
    done = ((p_port->p_remotePort->base_lid <= dLid) &&
            (p_port->p_remotePort->base_lid+((lid_t)1<<p_port->p_remotePort->lmc) - 1 >= dLid));

    if (p_DstSwNode == p_port->p_remotePort->p_node) {
        if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
            cout << offset_str
                << "-I- arrived at peer of switch connected to dest HCA." << endl;
        done = true;
    }

    if ( !done ) {

        if ( hopCnt > 256 ) {
            cout << "-E- Aborting after 256 hops - loop in LFT?" << endl;
            return 1;
        }
        if ( slvl.VL == IB_SLT_UNASSIGNED ) {
            cout << "-E- VL to destination is unassigned!"
                << " on out port:"<< p_port->getName()
                << "slid: " << sLid << " dlid:" << dLid << endl;
            return 1;
        }
        if (slvl.VL == IB_DROP_VL) {
            cout << "-E- Dead end at:" << p_port->p_node->name
            << " Drop VL from slid: " << sLid << " to dlid:" << dLid << endl;
            return 1;
        }
        //Verify that the vl is within the valid range
        if (slvl.VL >= IB_NUM_VL) {
            cout << "-E- Dead end at:" << p_port->p_node->name
            << " VL >= " << IB_NUM_VL << " from slid: " << sLid
            << " to dlid:" << dLid << endl;
            return 1;
        }

        // Get the node on the remote side
        p_node = p_port->p_remotePort->p_node;
        if (p_node->type != IB_SW_NODE) {
            cout << "-E- Dead end at:" << p_port->p_node->name
            << " Invalid LFT from slid: " << sLid << " to dlid:" << dLid << endl;
            return 1;
        }

        // Get remote port's number
        inputPortNum = p_port->p_remotePort->num;

        sl2vlPortGroup = p_node->getSLVLPortGroup(inputPortNum);
        u_int8_t pLFT = p_node->getPLFTMapping(inputPortNum, slvl);
        bool useAR = p_node->isARActive(slvl);

        u_int8_t isLidsGroup = 0;
        //Check if routing eaqual for all dest lids in the list
        if (!lids.empty()) {
            u_int16_t arLFTPortGroup = IB_AR_LFT_UNASSIGNED;
            phys_port_t portNum = IB_LFT_UNASSIGNED;

            isLidsGroup = 1;
            bool fail = false;

            if (useAR) {
                arLFTPortGroup = p_node->getARLFTPortGroupForLid(dLid,pLFT);

                for (list_lid::iterator itr = lids.begin();
                     itr != lids.end(); ++itr) {
                    if (p_node->getARLFTPortGroupForLid(*itr,pLFT) !=
                        arLFTPortGroup) {
                        fail = true;

                        if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
                            cout << offset_str
                                << "-I- ARLFTPortGroup: " << arLFTPortGroup
                                << " for DLID: 0x"
                                << setw(4) << setfill('0') << hex << dLid << dec
                                << " is not equal ARLFTPortGroup: "
                                << p_node->getARLFTPortGroupForLid(*itr,pLFT)
                                << " for DLID: 0x"
                                << setw(4) << setfill('0') << hex << *itr << dec
                                << " on node: " << p_node->getName()
                                << endl;

                        break;
                    }
                }
            }

            if (!useAR || arLFTPortGroup == IB_AR_LFT_UNASSIGNED) {
                portNum = p_node->getLFTPortForLid(dLid, pLFT);

                for (list_lid::iterator itr = lids.begin();
                     itr != lids.end(); ++itr) {
                    if (p_node->getLFTPortForLid(*itr,pLFT) !=
                        portNum) {
                        fail = true;

                        if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
                            cout << offset_str
                                << "-I- LFTPort: " << (int)portNum
                                << " for DLID: 0x"
                                << setw(4) << setfill('0') << hex << dLid << dec
                                << " is not equal LFTPort: "
                                << (int)p_node->getLFTPortForLid(*itr,pLFT)
                                << " for DLID: 0x"
                                << setw(4) << setfill('0') << hex << *itr << dec
                                << " on node: " << p_node->getName()
                                << endl;

                        break;
                    }
                }
            }

            if (fail) {
                delayedRouteInfo.push_back(
                    DelayedRouteInfo(slvl, p_port, hopCnt));
                if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
                    cout << offset_str
                        << "-I- The route to dlids group is not equal, continue later."
                        << endl;

                return 0;
            }
        }

        //check cache
        phys_port_t delayedOutPort = CrdLoopNodeInfo::updateCache(
            p_node, slvl, isLidsGroup, pLFT, sl2vlPortGroup, dLid);

        //if (delayedOutPort != 0) && (outPortNum != delayedOutPort) no need to
        //continue checking the next hop
        if ( delayedOutPort != 0 && FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
                cout << offset_str << "-I- Route already in cache"
                    << " node: " << p_node->getName()
                    << " SL: " << (int)slvl.SL
                    << " VL: " << (int)slvl.VL
                    << (isLidsGroup>0 ? " Group containing DLID: 0x"
                                      : " DLID: 0x")
                    << setw(4) << setfill('0') << hex << dLid << dec
                    << " plft: " << (int)pLFT
                    << " sl2vlPortGroup: " << (int)sl2vlPortGroup 
                    << " delayedOutPort: " << (int) delayedOutPort << endl;

        if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
            cout << offset_str << "-I- Add to cache"
                << " node: " << p_node->getName()
                << " SL: " << (int)slvl.SL
                << " VL: " << (int)slvl.VL
                << (isLidsGroup>0 ? " Group containing DLID: 0x"
                                  : " DLID: 0x")
                << setw(4) << setfill('0') << hex << dLid << dec
                << " plft: " << (int)pLFT
                << " sl2vlPortGroup: " << (int)sl2vlPortGroup << endl;

        // Get number of ports on the remote side
        phys_port_t numPorts = p_node->numPorts;
        // Init vchannel's number of possible dependencies
        p_port->channels[slvl.VL]->setDependSize((numPorts+1)*p_fabric->getNumVLs());

        // Get port num of the next hop
        //p_node->getLFTPortListForLid(dLid, inputPortNum, slvl, portsList);
        p_node->getLFTPortListForLid(dLid, pLFT, useAR, portsList);
        if (portsList.empty()) {
            cout << "-E- Unassigned LFT for lid:" << dLid
                << " Dead end at:" << p_node->name << endl;
            return 1;
        }

        list_phys_ports::iterator iter = portsList.begin();
        for ( ;  iter != portsList.end() ; ++iter) {
            outPortNum = *iter;

            if (inputPortNum == outPortNum) {
                if (delayedOutPort == 0)
                    CrdLoopNodeInfo::updateDelayedOutPort(
                        p_node, slvl, isLidsGroup, pLFT, sl2vlPortGroup, outPortNum);
                continue;
            }

            // Get VL of the next hop
            uint8_t nextVL = p_node->getVL(inputPortNum,outPortNum,slvl);
            if ( nextVL == IB_SLT_UNASSIGNED ) {
                cout << "-E- Unassigned SL2VL entry, node:" << p_node->name
                << ", iport:" << (unsigned int)inputPortNum << ", oport:"
                << (unsigned int)outPortNum << ", SL:" << (unsigned int)slvl.SL << endl;

                return 1;
            }
            if (nextVL == IB_DROP_VL) {
                cout << "-W- Drop VL on node:" << p_node->name
                     << ", iport:" << (unsigned int)inputPortNum << ", oport:"
                     << (unsigned int)outPortNum << ", SL:" << (unsigned int)slvl.SL << endl;
                return 1;
            }

            // get the next port on the other side
            p_portNext = p_node->getPort(outPortNum);

            if ( ! (p_portNext &&
                    p_portNext->p_remotePort &&
                    p_portNext->p_remotePort->p_node) ) {
                cout << "-E- Dead end at:" << p_node->name << endl;
                return 1;
            }
            // Now add an edge
            int set_res = p_port->channels[slvl.VL]->setDependency(
                outPortNum*p_fabric->getNumVLs()+nextVL,p_portNext->channels[nextVL],
                sLid, dLid, slvl.SL);

            if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE ) {
                cout << offset_str << "-I- DLID: 0x" << setw(4) << setfill('0')
                    << hex << dLid << dec;
                if (set_res == 0)
                    cout << " adds edge";
                else if (set_res == 1)
                    cout << " edge already exist";
                else
                    cout << " the rest of the route already exists starting";

                cout << " from:"  << p_port->getName()
                    << " VL: " << (int)slvl.VL
                    << " to: " << p_portNext->getName()
                    << " VL: " << (int)nextVL << endl;
            }

            if (set_res == 0){
               addedEdges++;
            } else if (set_res == 2)//already walked though this pass
                continue;

            // Advance
            if (delayedOutPort == 0 || outPortNum == delayedOutPort) {
                nextSlvl.VL = nextVL;

                if (outPortNum == delayedOutPort)
                    CrdLoopNodeInfo::updateDelayedOutPort(
                        p_node, slvl, isLidsGroup, pLFT, sl2vlPortGroup,
                        (phys_port_t)IB_LFT_UNASSIGNED);


                rc |= CrdLoopMarkRouteByARLFT (p_fabric, sLid, dLid, nextSlvl,
                                               p_portNext, hopCnt+1, addedEdges,
                                               lids, p_DstSwNode,
                                               delayedRouteInfo);
            }
        } //AR ports loop
    }

    return 0;
}

//////////////////////////////////////////////////////////////////////////////

// Trace a route from slid to dlid by AR LFT
// Add dependency edges
int
CrdLoopMarkRouteByARLFT (IBFabric *p_fabric,
                         lid_t sLid , lid_t dLid, int &addedEdges,
                         list_lid &lids, IBNode *p_DstSwNode)
{

    IBPort *p_port = p_fabric->getPortByLid(sLid);
    IBNode *p_node;
    list_lid emptyLids;

    int rc = 0;
    //IBPort *p_portNext;
    //phys_port_t outPortNum = 0, inputPortNum = 0;
    //int hopCnt = 0;
    //bool done;
    list_phys_ports portsList;
    list_delayed_route_t delayedRouteInfo;

    // make sure:
    if ( !p_port ) {
        cout << "-E- Provided source:" << sLid
        << " lid is not mapped to a port!" << endl;
        return(1);
    }

    if ( !p_port->p_remotePort ) {
        cout << "-E- Provided starting point is not connected !"
        << "lid:" << sLid << endl;
        return 1;
    }

    // Retrieve the relevant SL
    sl_vl_t slvl; // SL, VL;
    slvl.SL = p_port->p_node->getPSLForLid(dLid);
    if ( slvl.SL == IB_SLT_UNASSIGNED ) {
        cout << "-E- SL to destination is unassigned!"
            << " on out port:"<< p_port->getName()
            << "slid: " << sLid << " dlid:" << dLid << endl;
        return 1;
    }

    //for in port 0 use SL even if VL2VL is set (for SW)
    slvl.VL = p_port->p_node->getVL(0, p_port->num, slvl);
    //SL, VL validation will be done later

    // If started on a switch, need to use the correct output
    // port, not the first one found by getPortByLid
    if ( p_port->p_node->type == IB_SW_NODE ) {
        p_node = p_port->p_node;
        p_node->getLFTPortListForLid(dLid, 0, slvl, portsList);

        if (portsList.empty()) {
            cout << "-E- Unassigned LFT for lid:" << dLid
            << " Dead end at:" << p_port->p_node->name << endl;
            return 1;
        }

        list_phys_ports::iterator iter = portsList.begin();
        for ( ;  iter != portsList.end() ; ++iter) {
            p_port = p_node->getPort(*iter);
            if (p_port == NULL ||
                p_port->p_remotePort == NULL) {
                cout << "-E- Provided starting point is not connected !"
                << "lid:" << sLid << endl;
                continue;
            }

            //set the correct VL
            slvl.VL = p_port->p_node->getVL(0, p_port->num, slvl);
            rc |= CrdLoopMarkRouteByARLFT (
                p_fabric, sLid, dLid, slvl, p_port, 0, addedEdges,
                lids, p_DstSwNode, delayedRouteInfo);
        }

    } else {
        //Check if SL eaqual for all dest lids in the list
        if (!lids.empty()) {
            bool fail = false;

            for (list_lid::iterator itr = lids.begin();
                 itr != lids.end(); ++itr) {
                if (p_port->p_node->getPSLForLid(*itr) !=
                    slvl.SL) {
                    fail = true;
                    break;
                }
            }

            if (fail) {
                if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
                    cout << "-I- different SLs from SLID:" << (int)sLid
                        << " to HCAs connected to same SW including DLID:"
                        << (int)dLid
                        << " compute route to each DLID separately." << endl;

                for (list_lid::iterator itr = lids.begin();
                     itr != lids.end(); ++itr) {
                    rc |= CrdLoopMarkRouteByARLFT (p_fabric, sLid, *itr, slvl,
                                                   p_port, 0,
                                                   addedEdges, emptyLids, p_DstSwNode,
                                                   delayedRouteInfo);
                }
                return rc;
            }
        }

        rc =  CrdLoopMarkRouteByARLFT (p_fabric, sLid, dLid, slvl, p_port, 0,
                                       addedEdges, lids, p_DstSwNode,
                                       delayedRouteInfo);
    }

    if (!delayedRouteInfo.empty()) {
        list_delayed_route_t emptyRouteInfo;

        if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
            cout << "-I- on rotefrom SLID: " << (int)sLid
                << " to HCAs connected to same SW including DLID: "
                << (int)dLid
                << " compute " << (int) delayedRouteInfo.size()
                << " delayed routes." << endl;

        for (list_lid::iterator lidItr = lids.begin();
             lidItr != lids.end(); ++lidItr) {

            for (list_delayed_route_t::iterator itr = delayedRouteInfo.begin();
                 itr != delayedRouteInfo.end(); ++itr) {


                rc |= CrdLoopMarkRouteByARLFT (p_fabric, sLid, *lidItr,
                                               itr->m_slvl, itr->m_pPort,
                                               itr->m_hopCnt, addedEdges,
                                               emptyLids, p_DstSwNode,
                                               emptyRouteInfo);
            }

        }
    }
    return rc;
}


//////////////////////////////////////////////////////////////////////////////

// Trace a route from slid to dlid by LFT
// Add dependency edges
int
CrdLoopMarkRouteByLFT (IBFabric *p_fabric,
		       lid_t sLid , lid_t dLid)
{

  IBPort *p_port = p_fabric->getPortByLid(sLid);
  IBNode *p_node;
  IBPort *p_portNext;
  phys_port_t outPortNum = 0, inputPortNum = 0;
  int hopCnt = 0;
  bool done;

  // make sure:
  if (!p_port) {
    cout << "-E- Provided source:" << sLid
         << " lid is not mapped to a port!" << endl;
    return(1);
  }


  // Retrieve the relevant SL
  sl_vl_t slvl;
  slvl.SL = p_port->p_node->getPSLForLid(dLid);
  slvl.VL = 0; //just initialization will be set later
  if ( slvl.SL == IB_SLT_UNASSIGNED ) {
      cout << "-E- SL to destination is unassigned!"
          << " on out port:"<< p_port->getName()
          << "slid: " << sLid << " dlid:" << dLid << endl;
      return 1;
  }

  /*
  cout << "-D- SL:" << (int)slvl.SL << " at:" << p_port->p_node->name
      << " from slid: " << sLid << " to dlid:" << dLid << endl;
      */

  // If started on a switch, need to use the correct output
  // port, not the first one found by getPortByLid
  if (p_port->p_node->type == IB_SW_NODE) {
    //phys_port_t outPortNum = p_port->p_node->getLFTPortForLid(dLid);
    phys_port_t outPortNum = p_port->p_node->getLFTPortForLid(dLid, 0, slvl);

    if (outPortNum == IB_LFT_UNASSIGNED) {
      cout << "-E- Unassigned LFT for lid:" << dLid
           << " Dead end at:" << p_port->p_node->name << endl;
      return 1;
    }
    p_port = p_port->p_node->getPort(outPortNum);
  }
  slvl.VL = p_port->p_node->getVL(0, p_port->num, slvl);
  /*
  cout << "-D- VL:" << (int)slvl.VL
      << " on out port:"<< p_port->getName()
      << " from slid: " << sLid << " dlid:" << dLid << endl;
      */

  if (!p_port->p_remotePort) {
    cout << "-E- Provided starting point is not connected !"
         << "lid:" << sLid << endl;
    return 1;
  }

  if ( slvl.VL == IB_SLT_UNASSIGNED ) {
      cout << "-E- VL to destination is unassigned!"
          << " on out port:"<< p_port->getName()
          << "slid: " << sLid << " dlid:" << dLid << endl;
      return 1;
  }
  if (slvl.VL == IB_DROP_VL) {
      cout << "-E- Dead end at:" << p_port->p_node->name
      << " Drop VL from slid: " << sLid << " to dlid:" << dLid << endl;
      return 1;
  }

  // check if we are done:
  done = ((p_port->p_remotePort->base_lid <= dLid) &&
          (p_port->p_remotePort->base_lid+((lid_t)1<<p_port->p_remotePort->lmc) - 1 >= dLid));
  while (!done) {

    // Get the node on the remote side
    p_node = p_port->p_remotePort->p_node;
    // Get remote port's number
    inputPortNum = p_port->p_remotePort->num;
    // Get number of ports on the remote side
    phys_port_t numPorts = p_node->numPorts;
    // Init vchannel's number of possible dependencies
    p_port->channels[slvl.VL]->setDependSize((numPorts+1)*p_fabric->getNumVLs());

    // Get port num of the next hop
    outPortNum = p_node->getLFTPortForLid(dLid, inputPortNum, slvl);
    if (outPortNum == IB_LFT_UNASSIGNED) {
      cout << "-E- Unassigned LFT for lid:" << dLid << " Dead end at:" << p_node->name << endl;
      return 1;
    }

    // Get VL of the next hop
    uint8_t nextVL = p_node->getVL(inputPortNum,outPortNum,slvl);
    if (nextVL == IB_SLT_UNASSIGNED) {
      cout << "-E- Unassigned SL2VL entry, node:" << p_node->name
           << ", iport:" << (unsigned int)inputPortNum << ", oport:"
           << (unsigned int)outPortNum << ", SL:" << (unsigned int)slvl.SL << endl;

      return 1;
    }
    if (nextVL == IB_DROP_VL) {
        cout << "-W- Drop VL on node:" << p_node->name
             << ", iport:" << (unsigned int)inputPortNum << ", oport:"
             << (unsigned int)outPortNum << ", SL:" << (unsigned int)slvl.SL << endl;
        return 1;
    }
    /*
    cout << "-D- next VL:" << (int)slvl.VL << " on node:" << p_node->name
             << ", iport:" << (unsigned int)inputPortNum << ", oport:"
             << (unsigned int)outPortNum << ", SL:" << (unsigned int)slvl.SL << endl;
             */

    // get the next port on the other side
    p_portNext = p_node->getPort(outPortNum);

    if (! (p_portNext &&
           p_portNext->p_remotePort &&
           p_portNext->p_remotePort->p_node)) {
      cout << "-E- Dead end at:" << p_node->name << endl;
      return 1;
    }
    // Now add an edge
    p_port->channels[slvl.VL]->setDependency(
        outPortNum*p_fabric->getNumVLs()+nextVL,p_portNext->channels[nextVL],
        sLid, dLid);

    if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
        cout << "-I- DLID: 0x" << setw(4) << setfill('0')
            << hex << dLid << dec
            << " adds edge from:"  << p_port->getName()
            << " VL: " << (int)slvl.VL
            << " to: " << p_portNext->getName()
            << " VL: " << (int)nextVL << endl;

    // Advance
    p_port = p_portNext;
    slvl.VL = nextVL;
    if (hopCnt++ > 256) {
      cout << "-E- Aborting after 256 hops - loop in LFT?" << endl;
      return 1;
    }
    //Check if done
    done = ((p_port->p_remotePort->base_lid <= dLid) &&
	    (p_port->p_remotePort->base_lid+((lid_t)1<<p_port->p_remotePort->lmc) - 1 >= dLid));
  }

  return 0;
}

/////////////////////////////////////////////////////////////////////////////

// Go over all CA to CA paths and connect dependant vchannel by an edge

int
CrdLoopConnectUcastDepend(IBFabric* p_fabric, bool checkAR = false)
{
  int anyError = 0;
  unsigned int numPaths = 0;
  bool max_err_exceeded = false;
  int addedEdges = 0;
  list_lid all_lids;
  list_lid regular_lids;

  // go over all ports in the fabric
  //for ( j = p_fabric->minLid; j <= p_fabric->maxLid; j++ ) {
  for ( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
      nI != p_fabric->NodeByName.end(); nI++ ) {

    IBNode *p_DstSwNode = (*nI).second;
    //go over all CA dest lid ordered by connected SW
    if ( p_DstSwNode->type != IB_SW_NODE )
        continue;

    all_lids.clear();
    regular_lids.clear();

    for ( phys_port_t pn = 1; pn <= p_DstSwNode->numPorts; pn++ ) {
        IBPort *p_port = p_DstSwNode->getPort((phys_port_t)pn);
        if ( !p_port || !p_port->p_remotePort ||
             !p_port->p_remotePort->p_node ||
             p_port->p_remotePort->p_node->type == IB_SW_NODE)
            continue;

        if (!IsValidUcastLid(p_port->p_remotePort->base_lid)) {
            cout << "-W- Skip credit loop check toward node: " << p_port->p_remotePort->p_node->name
                 << " port: " << (unsigned int)p_port->p_remotePort->num
                 << " with invalid LID " << p_port->p_remotePort->base_lid
                 << endl;
            continue;
        }

        if (!p_port->isSpecialPort())
            regular_lids.push_back(p_port->p_remotePort->base_lid);

        all_lids.push_back(p_port->p_remotePort->base_lid);
        //p_dstPort = p_port->p_remotePort;
    }

    if (!CrdLoopIncludeUcastSwitchPaths &&
        all_lids.empty()) continue;

    // go over all the rest of the ports:
    for ( lid_t sLid = p_fabric->minLid; sLid <= p_fabric->maxLid; sLid++) {
      IBPort *p_srcPort = p_fabric->getPortByLid(sLid);
      if (!p_srcPort || p_srcPort->base_lid != sLid ||
          !p_srcPort->p_remotePort)
          continue;

      // Avoid tracing to itself or under same SW
      if (p_srcPort->p_node == p_DstSwNode ||
          ((p_srcPort->p_node->type != IB_SW_NODE) &&
           (p_srcPort->p_remotePort->p_node == p_DstSwNode)))
          continue;

      if (!CrdLoopIncludeUcastSwitchPaths &&
          (p_srcPort->p_node->type == IB_SW_NODE)) continue;

      if (anyError >= MAX_ERROR_COUNT){
          max_err_exceeded = true;
          break;
      }

      if (!p_srcPort->p_node->isValidPSLInfo()) {
          cout << "-E- " << p_srcPort->p_node->name
               << " does not have path SL information."
               << endl;
          anyError++;
          continue;
      }

      list_lid &lids = all_lids;

      // AN to AN communication doesn't contribute to credit loops, since AN communicates only
      // to AN on a one hop neighbor switch.
      // AN is special port.
      // Thus make sure the checked paths won't be between special ports (both src and dest).
      if (p_srcPort->p_remotePort->isSpecialPort())
          lids = regular_lids;

      // Trace the path but record the input to output ports used.
      if (checkAR) {

          if (!lids.empty()) {
              lid_t dlid = lids.front();

              int lidsNum = (int)lids.size();
              numPaths += (int)lids.size();

              if (lidsNum == 1)
                  lids.clear();

              if (CrdLoopMarkRouteByARLFT(p_fabric, sLid, dlid, addedEdges, lids, p_DstSwNode)) {
                 cout << "-E- Error in mark route from:"
                    << p_srcPort->p_node->name << " SLID:" << (unsigned int)sLid
                    << " to HCA group size:" << lidsNum
                    << " connected to:" << p_DstSwNode->name << endl;
                 anyError++;
              }
          }

          if (CrdLoopIncludeUcastSwitchPaths) {
              lid_t dlid;
              uint8_t lmc;
              lids.clear();

              if (!p_DstSwNode->getLidAndLMC((phys_port_t)0, dlid, lmc)){
                  numPaths++;
                  if (CrdLoopMarkRouteByARLFT(p_fabric, sLid, dlid, addedEdges, lids, NULL)) {
                     cout << "-E- Error in mark route from:"
                        << p_srcPort->p_node->name << " SLID:" << (unsigned int)sLid
                        << " to SW:" << p_DstSwNode->name << " DLID:" << (unsigned int)dlid
                        << endl;
                     anyError++;
                  }
              }
          }

      } else {
          lid_t dlid;

          if (CrdLoopIncludeUcastSwitchPaths) {
              uint8_t lmc;
              if (!p_DstSwNode->getLidAndLMC((phys_port_t)0, dlid, lmc) && IsValidUcastLid(dlid))
                  lids.push_back(dlid);
          }

          for (list_lid::iterator itr = lids.begin();
               itr != lids.end() ; ++itr) {
              dlid = *itr;

              numPaths++;

              if (CrdLoopMarkRouteByLFT(p_fabric, sLid, dlid)) {
                  cout << "-E- Error in mark route from:"
                     << p_srcPort->p_node->name << " SLID:" << (unsigned int)sLid
                     << " to DLID:" << (unsigned int)dlid
                     << endl;
                  anyError++;
              }
          }

      }

      if (anyError >= MAX_ERROR_COUNT){
          max_err_exceeded = true;
          break;
      }

    } // all targets

  } // all sources

  cout << "-I- Traced " << numPaths << " unicast paths" << endl;
  if (checkAR)
      cout << "-I- AR added " << addedEdges << " edges" << endl;    

  if (anyError) {

      if (max_err_exceeded)
        cout << "-E- Maximum error coun exceed. "
             << "Fail to traverse "
             << anyError << " CA to CA paths" << endl;

      else
          cout << "-E- Fail to traverse:"
               << anyError << " CA to CA paths" << endl;

      return 1;
  }

  return 0;
}

/////////////////////////////////////////////////////////////////////////////

// Go over all Multicast Groups on all switches and add the dependecies
// they create to the dependency graph
// Return number of errors it found
int
CrdLoopConnectMcastDepend(IBFabric* p_fabric)
{
  // We support providing an SL list to each MLID by provining a special file
  // HACK: we can ignore connectivity check of the MCG and treat every switch
  // separately. The connectivity analysis can be run independently if loops
  // found.

  // HACK: the algorithm assumes constant SL2VL which is not port dependant!
  //       otherwise it should have been propagating traffic from each CA and
  //       SW on the MGRP such that the out VL of the previous port is known...
  //
  // Algorithm:
  // Foreach switch
  //   Create empty port-to-port P2P(sl,in,out) mask matrix for each SL
  //   Foreach MLID
  //     Foreach SL of MLID
  //        Copy the MFT(MLID) port mask to the matrix
  //   Foreach SL
  //     Lookup VL by SL - this is where the hack comes in handy
  //     Foreach in-port
  //       Foreach out-port
  //         Create the dependency edge (port-driving-in-port, VL, out-port, VL)

  int nErrs = 0;
  int addedEdges = 0;

  // foreach Switch
  for (map_str_pnode::const_iterator nI = p_fabric->NodeByName.begin();
       nI != p_fabric->NodeByName.end(); nI++) {
    IBNode *p_node = (*nI).second;

    // we only do MFT on switches
    if (p_node->type != IB_SW_NODE) continue;

    // allocate the array of dependencies:
    uint8_t sl_in_out_dep[16][p_node->numPorts+1][p_node->numPorts+1];
    memset(sl_in_out_dep, 0, sizeof(uint8_t)*16*(p_node->numPorts+1)*(p_node->numPorts+1));

    // foreach MLID
    for (lid_t mftI = 0; mftI < p_node->MFT.size(); mftI++) {
      list<uint8_t> sls;

      // lookup SL's
      map_mlid_sl_list::const_iterator mlidI = mlidSLs.find((lid_t)(mftI+0xc000));
      if (mlidI != mlidSLs.end()) {
        sls = (*mlidI).second;
      } else {
        sls.push_back(0);
      }

      // now go over each SL at a time
      sl_vl_t slvl;
      for (list<uint8_t>::const_iterator lI = sls.begin();
           lI != sls.end();
           lI++) {
        slvl.SL = (*lI); //!!!SLVL
        slvl.VL = (uint8_t)(slvl.SL%p_fabric->numVLs); //!!!SLVL
        // check all ports of the MFT
        PortsBitset &port_mask = p_node->MFT[mftI];
        for (unsigned int i = 1; i <= p_node->numPorts; i++) {
          phys_port_t inPortNum = (phys_port_t)i;
          // we only care about port that are part of the MCG
          if (port_mask.test(inPortNum)) {
            for (unsigned int j = 1; j <= p_node->numPorts; j++) {
              phys_port_t outPortNum = (phys_port_t)j;
              if (inPortNum != outPortNum) {
                if (port_mask.test(outPortNum)) {
                  // if we already marked it skip
                  if (sl_in_out_dep[slvl.SL][inPortNum][outPortNum] == 1) continue;
                  sl_in_out_dep[slvl.SL][inPortNum][outPortNum] = 1;

                  IBPort *p_outPort = p_node->getPort(outPortNum);
                  if (! p_outPort) {
                    cout << "-E- Switch:" << p_node->name << " port:" << (unsigned int)outPortNum
                         << " is included in some MFT but is not connnected" << endl;
                    nErrs++;
                    continue;
                  }
                  IBPort *p_inPort = p_node->getPort(inPortNum);
                  if (! p_inPort) {
                    cout << "-E- Switch:" << p_node->name << " port:" << (unsigned int)inPortNum
                         << " is included in some MFT but is not connnected" << endl;
                    nErrs++;
                    continue;
                  }

                  IBPort *p_drvPort = p_inPort->p_remotePort;
                  if (! p_drvPort) {
                    cout << "-E- Switch:" << p_node->name << " port:" << (unsigned int)inPortNum
                         << " is included in some MFT but has no remote port." << endl;
                    nErrs++;
                    continue;
                  }

                  // we ignore MFT loops with starting port on CA?
                  if (p_drvPort->p_node->type != IB_SW_NODE) continue;

                  // ignore the MFT dependency if the other side is a switch
                  // and it does not forward this MFT through this port
                  if (p_drvPort->p_node->type == IB_SW_NODE) {
                    if (p_drvPort->p_node->MFT.size() <= mftI) continue;
                    if (!p_drvPort->p_node->MFT[mftI].test(p_drvPort->num))
                      continue;
                  }

                  // Lookup VL by SL
                  uint8_t vl = p_node->getVL(inPortNum,outPortNum,slvl); //!!!SLVL
                  if (vl == IB_SLT_UNASSIGNED) {
                    cout << "-E- Unassigned SL2VL entry, node:" << p_node->name
                         << ", iport:" << (unsigned int)inPortNum << ", oport:"
                         << (unsigned int)outPortNum << ", SL:" << (unsigned int)slvl.SL << endl;
                    continue;
                  }
                  if (vl == IB_DROP_VL) {
                      cout << "-W- Drop VL on node:" << p_node->name
                           << ", iport:" << (unsigned int)inPortNum << ", oport:"
                           << (unsigned int)outPortNum << ", SL:" << (unsigned int)slvl.SL << endl;
                      continue;
                  }

                  // Create the dependency edge (port-driving-in-port, VL, out-port, VL)
                  // Init vchannel's number of possible dependencies
                  p_drvPort->channels[vl]->setDependSize((p_node->numPorts+1)*p_fabric->getNumVLs());

                  // HACK: we assume the same VL was used entering to this node
                  if (!p_drvPort->channels[vl]->getDependencyPvch(outPortNum*p_fabric->getNumVLs()+vl)) {
                    p_drvPort->channels[vl]->setDependency(outPortNum*p_fabric->getNumVLs()+vl,
                                                           p_outPort->channels[vl], 0, (lid_t)(mftI + 0xc000));

                    if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
                        cout << "-I- MLID: 0x" << setw(4) << setfill('0')
                            << hex << mftI + 0xc000 << dec
                            << " adds edge from:"  << p_drvPort->getName()
                            << " VL: " << (int)vl
                            << " to: " << p_outPort->getName()
                            << " VL: " << (int)vl << endl;
                    addedEdges++;
                  }
                }
              }
            }
          }
        }
      }
    }
  }

  // Ref from LFT code:
  // outPortNum is the FBD port
  // get the next port on the other side
  // p_portNext = p_node->getPort(outPortNum);
  // Now add an edge
  // p_port->channels[VL]->setDependency(outPortNum*p_fabric->getNumVLs()+nextVL,
  //                 p_portNext->channels[nextVL]);
  cout << "-I- MFT added " << addedEdges << " edges to links dependency graph" << endl;
  return(nErrs);
}
/////////////////////////////////////////////////////////////////////////////
//recursive call used by AdvanceCrdLoopConnectMcastDepend
int
CrdLoopMarkRouteByMFT(IBFabric *p_fabric,
                      lid_t sLid , lid_t mlid, sl_vl_t slvl,
                      IBPort *p_port, int &addedEdges, int membersNum)
{
    IBNode *p_node;
    IBPort *p_portNext;
    phys_port_t outPortNum = 0, inputPortNum = 0;
    phys_port_t numPorts = 0;
    int rc = 0;
    sl_vl_t nextSlvl;
    nextSlvl.SL = slvl.SL;

    if (p_port->num != 0) { //not started at sw port 0
        // Get the node on the remote side
        p_node = p_port->p_remotePort->p_node;
        // Get remote port's number
        inputPortNum = p_port->p_remotePort->num;
        // Get number of ports on the remote side
        numPorts = p_node->numPorts;
        // Init vchannel's number of possible dependencies
        if (p_port->p_node->type == IB_SW_NODE)
            p_port->channels[slvl.VL]->setDependSize((numPorts+1)*p_fabric->getNumVLs());
    } else {
        // in case p_port->num == 0 this is SW group member
        // in the firs cycle we do not set Depend, only iterate on out ports
        // Get the node
        p_node = p_port->p_node;
        // Get SW number of ports
        numPorts = p_node->numPorts;
    }


    // Get port num of the next hop
    PortsBitset port_mask;
    size_t mft_size = p_node->MFT.size();
    if (0 <= (mlid - 0xc000) && mft_size > (lid_t)(mlid - 0xc000))
        port_mask = p_node->MFT[mlid - 0xc000];

    if (port_mask.none()) {
        if (membersNum > 1) {
            cout << "-E- Unassigned MFT for lid:" << mlid
                << " (0x" << hex << mlid << dec << ")"
                << " Dead end at:" << p_node->name << endl;
            return 1;
        } else {
            cout << "-I- Unassigned MFT for lid:" << mlid
                << " (0x" << hex << mlid << dec << ")"
                << " Ignore group with less than two members." << endl;
            return 0;
        }
    } else if (membersNum < 2) {
        cout << "-W- Assigned MFT for lid:" << mlid
            << " (0x" << hex << mlid << dec << ")"
            << " for group with less than two members."
            << " at:" << p_node->name << endl;
    }

    for (outPortNum = 1; outPortNum <= numPorts; outPortNum++) {
        if (!port_mask.test(outPortNum))
            continue;
        if (outPortNum == inputPortNum)
            continue;

        // Get VL of the next hop
        uint8_t nextVL = p_node->getVL(inputPortNum,outPortNum,slvl);
        if ( nextVL == IB_SLT_UNASSIGNED ) {
            cout << "-E- Unassigned SL2VL entry, node:" << p_node->name
            << ", iport:" << (unsigned int)inputPortNum << ", oport:"
            << (unsigned int)outPortNum << ", SL:" << (unsigned int)slvl.SL << endl;

            rc++;
            continue;
        }
        if (nextVL == IB_DROP_VL) {
            cout << "-W- Drop VL on node:" << p_node->name
                 << ", iport:" << (unsigned int)inputPortNum << ", oport:"
                 << (unsigned int)outPortNum << ", SL:" << (unsigned int)slvl.SL << endl;
            rc++;
            continue;
        }

        // get the next port on the other side
        p_portNext = p_node->getPort(outPortNum);

        if ( ! (p_portNext &&
                p_portNext->p_remotePort &&
                p_portNext->p_remotePort->p_node) ) {

            if (membersNum > 1) {
                cout << "-E- Dead end at:" << p_node->name
                    << "port:" << (int) outPortNum << endl;
                return 1;
            } else {
                cout << "-W- Dead end at:" << p_node->name
                    << " port:" << (int) outPortNum
                    << " Ignore group with less than two members." << endl;

                return 0;
            }
        }

        if (p_portNext->p_remotePort->p_node->type != IB_SW_NODE)
            continue;

        // Now add an edge
        if (inputPortNum != 0 &&
            p_port->p_node->type == IB_SW_NODE) {
            int set_res = p_port->channels[slvl.VL]->setDependency(
                outPortNum*p_fabric->getNumVLs()+nextVL,p_portNext->channels[nextVL],
                sLid, mlid, slvl.SL);
            if (set_res == 0){
                if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE )
                    cout << "-I- MLID: 0x" << setw(4) << setfill('0')
                        << hex << mlid << dec
                        << " adds edge from:"  << p_port->getName()
                        << " VL: " << (int)slvl.VL
                        << " to: " << p_portNext->getName()
                        << " VL: " << (int)nextVL << endl;
                addedEdges++;
            } else if (set_res == 2)//already walked though this pass
                continue;
        }

        // Advance
        nextSlvl.VL = nextVL;

        rc += CrdLoopMarkRouteByMFT (p_fabric, sLid, mlid, nextSlvl,
                                     p_portNext, addedEdges, membersNum);
    } //AR ports loop

    return 0;
}


// Go over all Mlids, for each mlid go over on all members. from each member travel
// on the multicast path for each SL, and add the dependecies
// they create to the dependency graph
// Return number of errors it found
int
AdvanceCrdLoopConnectMcastDepend(IBFabric* p_fabric)
{
    int nErrs = 0;
    int addedEdges = 0;

    lid_t mlid = 0;
    sl_vl_t slvl;

    // foreach mlid
    for (map_mcast_groups::const_iterator gI = p_fabric->McastGroups.begin();
         gI != p_fabric->McastGroups.end(); gI++) {

        mlid = gI->first;
        int membersNum = (int)gI->second.m_members.size();

        map_mcast_members::const_iterator mI = gI->second.m_members.begin();
        for ( ; mI != gI->second.m_members.end(); mI++) {

            IBPort *p_port = mI->first;

            for ( set_uint8::const_iterator slI = mI->second.m_sls.begin() ;
                  slI != mI->second.m_sls.end(); slI++) {

                slvl.SL = (*slI);
                slvl.VL = slvl.SL;
                //for in port 0 use SL even if VL2VL is set (for SW)
                slvl.VL = p_port->p_node->getVL(0, p_port->num, slvl);
                //Handle switch out port 0
                if ( p_port->p_node->type == IB_SW_NODE )
                    p_port = p_port->p_node->getPort(0);

                if ( slvl.VL == IB_SLT_UNASSIGNED ) {
                    cout << "-E- VL to destination is unassigned!"
                        << " on out port:"<< p_port->getName()
                        << "slid: " << p_port->base_lid
                        << " mlid:" << mlid
                        << " (0x" << hex << mlid << dec << ")"
                        << endl;
                    nErrs++;
                    continue;
                }
                if (slvl.VL == IB_DROP_VL) {
                    cout << "-E- Dead end at:" << p_port->p_node->name
                        << " Drop VL from slid: " << p_port->base_lid
                        << " to mlid:" << mlid
                        << " (0x" << hex << mlid << dec << ")"
                        << endl;
                    nErrs++;
                    continue;
                }

                nErrs += CrdLoopMarkRouteByMFT(p_fabric,
                                               p_port->base_lid,
                                               mlid,
                                               slvl,
                                               p_port,
                                               addedEdges,
                                               membersNum);
            }
        }
    }


    cout << "-I- MFT added " << addedEdges << " edges to links dependency graph" << endl;
    return(nErrs);
}


//////////////////////////////////////////////////////////////////////////////


// Prepare the data model
int
CrdLoopPrepare(IBFabric *p_fabric, bool checkAR) {
  // go over all nodes in the fabric
  for( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
       nI != p_fabric->NodeByName.end();
            nI++) {
    IBNode *p_node = (*nI).second;
    int nL = p_fabric->getNumVLs();

    // Go over all node's ports
    for (unsigned int k = 0; k <= p_node->numPorts; k++) {
      IBPort* p_Port = p_node->getPort((phys_port_t)k);
      if (!p_Port) continue;
      // Init virtual channel array
      p_Port->channels.resize(nL, NULL);
      for (int j=0;j<nL;j++)
        p_Port->channels[j] = new VChannel(p_Port, j);
    }
  }

  if (checkAR)
      return CrdLoopNodeInfo::prepare(p_fabric);

  return 0;
}

// Cleanup the data model
int
CrdLoopCleanup(IBFabric *p_fabric, bool checkAR) {
  for( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
       nI != p_fabric->NodeByName.end();
            nI++) {
    IBNode *p_node = (*nI).second;
    int nL = p_fabric->getNumVLs();

    // Go over all node's ports
    for (unsigned int k = 0; k <= p_node->numPorts; k++) {
      IBPort* p_Port = p_node->getPort((phys_port_t)k);
      if (!p_Port) continue;
      for (int j=0;j<nL;j++)
        if (p_Port->channels[j]) {
          delete p_Port->channels[j];
	  p_Port->channels[j] = NULL;
	} else {
	  cout << "EZ: double free on lid:" << dec << p_Port->base_lid
	       << " pn: " << k << " channel:" << j << endl;
	}
    }
  }

  if (checkAR)
      CrdLoopNodeInfo::cleanup(p_fabric);

  return(0);
}

void
CrdLoopCleanChannelsDfsState(IBFabric *p_fabric) {
    for (map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
        nI != p_fabric->NodeByName.end();
        nI++) {

        IBNode *p_node = (*nI).second;
        int nL = p_fabric->getNumVLs();

        // Go over all node's ports
        for (unsigned int k = 0; k <= p_node->numPorts; k++) {
            IBPort* p_Port = p_node->getPort((phys_port_t)k);
            if (!p_Port) continue;

            for (int j = 0; j < nL; j++) {
                if (p_Port->channels[j]) {
                    p_Port->channels[j]->setFlag(Untouched);
                }
            }
        }
    }
}


//////////////////////////////////////////////////////////////////////////////

// Top Level Subroutine:
int
CrdLoopAnalyze(IBFabric *p_fabric, bool checkAR) {
  int res=0;

  cout << "-I- Analyzing Fabric for Credit Loops "
       << (int)p_fabric->getNumSLs() <<" SLs, "
       << (int)p_fabric->getNumVLs() << " VLs used." << endl;

  do { // pretend loop

      // Init data structures
      if (CrdLoopPrepare(p_fabric, checkAR)) {
        cout << "-E- Fail to prepare data structures." << endl;
        res = 1;
        break;
      }
      // Create the dependencies for unicast traffic
      if (CrdLoopConnectUcastDepend(p_fabric)) {
        cout << "-E- Fail to build dependency graphs." << endl;
        res = 1;
        break;
      }
      // Do multicast if require
      if (CrdLoopIncludeMcastPaths) {
          if (p_fabric->McastGroups.empty()) {
              cout << "-W- It is recommended to use SA dump file"
                   << " with multicast credit loop check." << endl;
              res = CrdLoopConnectMcastDepend(p_fabric);
          } else
              res = AdvanceCrdLoopConnectMcastDepend(p_fabric);

          if ( res ) {
              cout << "-E- Fail to build multicast dependency graphs." << endl;
              res = 1;
              break;
          }
      }

      // Find the loops if exist
      res = CrdLoopFindLoops(p_fabric);
      if (!res)
        cout << "-I- no credit loops found" << endl;
      else
        cout << "-E- credit loops in routing"<<endl;

      if (checkAR) {

          cout << "-I- Analyzing Fabric for Credit Loops using AR. "
               << (int)p_fabric->getNumSLs() <<" SLs, "
               << (int)p_fabric->getNumVLs() << " VLs used." << endl;

          // Create the dependencies for unicast traffic
          if (CrdLoopConnectUcastDepend(p_fabric, true)) {
            cout << "-E- Fail to build dependency graphs." << endl;
            res = 1;
            break;
          }

          // Find the loops if exist
          res = CrdLoopFindLoops(p_fabric);
          if (!res)
            cout << "-I- no credit loops found in Adaptive Routing" << endl;
          else
            cout << "-E- credit loops in Adaptive Routing"<<endl;

      }
  } while(false);

  // cleanup:
  CrdLoopCleanup(p_fabric, checkAR);

  return res;
}

//////////////////////////////////////////////////////////////////////////////
int
CredLoopMode(int include_switch_to_switch_paths, int include_multicast) {
  CrdLoopIncludeUcastSwitchPaths = include_switch_to_switch_paths;
  CrdLoopIncludeMcastPaths = include_multicast;
  return 0;
}

//allocate neaded memory
int CrdLoopNodeInfo::prepare(IBFabric *p_fabric)
{
    CrdLoopNodeInfo *pInfo = NULL;

    // go over all nodes in the fabric
    for ( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
        nI != p_fabric->NodeByName.end(); nI++ ) {
        IBNode *p_node = (*nI).second;
        if ( p_node->type == IB_CA_NODE )
            continue;

        // allocate CrdLoopNodeInfo
        pInfo = new CrdLoopNodeInfo();
        p_node->appData1.ptr = pInfo;
        if ( pInfo == NULL ) {
            cout << "-E- Failed to allocate memory for CrdLoopNodeInfo" << endl;
            return 1;
        }

        pInfo->m_pNode = p_node;
    }
    return 0;
}

void CrdLoopNodeInfo::cleanup(IBFabric *p_fabric)
{

    // go over all nodes in the fabric
    for ( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
        nI != p_fabric->NodeByName.end(); nI++ ) {
        IBNode *p_node = (*nI).second;
        if ( p_node->type == IB_CA_NODE )
            continue;

        delete (CrdLoopNodeInfo*)p_node->appData1.ptr;
        p_node->appData1.ptr = NULL;
    }
}

phys_port_t CrdLoopNodeInfo::updateCache(const sl_vl_t &slvl, u_int8_t isLidsGroup,
        u_int8_t pLFT, phys_port_t sl2vlPortGroup, lid_t dLid)
{
    if (m_nodeInfo[slvl.VL][slvl.SL][isLidsGroup][pLFT].size() <= m_pNode->numPorts)
        m_nodeInfo[slvl.VL][slvl.SL][isLidsGroup][pLFT].resize(m_pNode->numPorts + 1);

    if (sl2vlPortGroup > m_pNode->numPorts) {
        cout << "-E- Assert - Invalid sl2vlPortGroup: " << +sl2vlPortGroup << " > num ports: "
             << +m_pNode->numPorts << endl;
        return 0;
    }

    if (m_nodeInfo[slvl.VL][slvl.SL][isLidsGroup][pLFT][sl2vlPortGroup].m_dlid == dLid)
        return m_nodeInfo[slvl.VL][slvl.SL][isLidsGroup][pLFT][sl2vlPortGroup].
        m_delayedOutPort;

    m_nodeInfo[slvl.VL][slvl.SL][isLidsGroup][pLFT][sl2vlPortGroup].
        m_dlid = dLid;
    m_nodeInfo[slvl.VL][slvl.SL][isLidsGroup][pLFT][sl2vlPortGroup].
        m_delayedOutPort = (phys_port_t)IB_LFT_UNASSIGNED;

    return 0;
}


