/*
 * Copyright (c) 2012-2013 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <iomanip>
#include <fstream>
#include <sstream>
#include <queue>
#include <algorithm>
#include <dirent.h>
#include <infiniband/ibdm/NetSplit.h>
#include <infiniband/ibdm/SubnMgt.h>
#include <infiniband/ibdm/TraceRoute.h>
#include <infiniband/ibdm/Regexp.h>

using namespace std;

#define map_port_dist_head_pair map<IBPort *, pair<unsigned int, IBPort*> >

// Given a head node BFS to all hosts marking closest head
int
bfsFromHeadThroughClosest(IBPort *head,
                          map_port_dist_head_pair &minDistHead)
{
  if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
    cout << "-V- bfsFromHeadThroughClosest from: "
         << head->getName() << endl;
  map<IBPort *, unsigned int> portDistance;
  queue<IBPort *> Q;
  Q.push(head);
  portDistance[head] = 0;
  while (!Q.empty()) {
    IBPort *port = Q.front();
    Q.pop();
    unsigned int dist = portDistance[port];

    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
      cout << "-V- " << dist << " bfsFromHeadThroughClosest visiting: "
           << port->getName() << endl;

    if (port->p_node->type == IB_SW_NODE) {
      // try all neighbors
      // use unsigned int index to prevent overwrap
      for (unsigned int pn = 1; pn <= port->p_node->numPorts; pn++) {
        IBPort *pPort = port->p_node->getPort((phys_port_t)pn);
        if (!pPort || !pPort->p_remotePort) continue;
        IBNode *pRemNode = pPort->p_remotePort->p_node;
        IBPort *pRemPort;
        if (pRemNode->type == IB_SW_NODE) {
          pRemPort = pRemNode->getPort(0);
        } else {
          pRemPort = pPort->p_remotePort;
        }

        // Q only if not visited
        if (portDistance.find(pRemPort) == portDistance.end()) {
          portDistance[pRemPort] = dist+1;
          Q.push(pRemPort);
        }
      }
    } else {
      map_port_dist_head_pair::iterator mdI = minDistHead.find(port);
      if ( (mdI == minDistHead.end()) || ((*mdI).second.first > dist) ) {
        minDistHead[port] = pair< unsigned int, IBPort *>(dist, head);
      }
      // only for the first port progress to other side
      if ((dist == 0) && port->p_remotePort) {
        IBPort *pRemPort = port->p_remotePort;
        // if remote is a switch use its base port
        if (pRemPort->p_node->type == IB_SW_NODE) {
          pRemPort = pRemPort->p_node->getPort(0);
        }
        if (portDistance.find(pRemPort) == portDistance.end()) {
          portDistance[pRemPort] = dist+1;
          Q.push(pRemPort);
        }
      }
    }
  }
  return(0);
}

// Given a fabric and a list of head ports split the hosts into groups
// by the minimal distance head node
int NetSplitGroupHostsByHeads(IBFabric &fabric, list<IBPort*> headPorts,
                              map_port_set_ports &portGroups)
{
  // BFS from each head marking distance to each of the
  // visited nodes. Only go through if closer than any other
  map_port_dist_head_pair minDistHead;

  list<IBPort*>::iterator lI;
  for (lI = headPorts.begin(); lI != headPorts.end(); lI++) {
    bfsFromHeadThroughClosest(*lI, minDistHead);
  }

  // now populate by head
  map_port_dist_head_pair::iterator dhI;
  for (dhI = minDistHead.begin(); dhI != minDistHead.end(); dhI++) {
    IBPort *port, *head;
    port = (*dhI).first;
    head = (*dhI).second.second;
    portGroups[head].insert(port);
  }

  cout << "-I- Grouped " << minDistHead.size() << " end-ports into "
       << portGroups.size() << " groups" << endl;
  return(0);
}

// Given a fabric with previosuly assigned MinHop tables and
// target node distance split all nodes by their min-hop distances
// by the minimal distance head node
int NetSplitGroupHostsByMinHop(IBFabric &fabric, unsigned int maxDist,
                               map_port_set_ports &portGroups)
{
  // visit every leaf switch and group the hosts if hop <= grpDist - 1
  // error if a host is already marked in a group but now meets
  // some other group (will fail DF).
  map<IBPort *, IBPort *> portToHead;
  set<IBNode *> visited;

  map_str_pnode::iterator nI;
  for( nI = fabric.NodeByName.begin();
       nI != fabric.NodeByName.end();
       nI++) {
    IBNode *node = (*nI).second;
    if (node->type == IB_SW_NODE) continue;

    for (unsigned int pn = 1; pn <= node->numPorts; pn++) {
      IBPort *port = node->getPort((phys_port_t)pn);
      if (!port || !port->p_remotePort) continue;
      IBNode *swNode = port->p_remotePort->p_node;
      if (swNode->type != IB_SW_NODE) continue;
      if (visited.find(swNode) != visited.end()) continue;
      visited.insert(swNode);

      // we may only have one group of this switch
      IBPort *groupPort = NULL;

      // OK so it is a new leaf switch
      for (lid_t lid = 1; lid <= fabric.maxLid; lid++) {
        IBPort* dPort = fabric.getPortByLid(lid);
        if (!dPort || dPort->p_node->type == IB_SW_NODE) continue;

        uint8_t minHops = swNode->getHops(NULL, lid);
        // the CA may be connected to the switch then it has minHops = 1
        // otherwise if the minHops < maxDist - 1 it also interests us
        if (minHops >= maxDist) continue;

        // now do we have a new group or
        if (portToHead.find(dPort) != portToHead.end()) {
          // exists must match the switch group
          if (groupPort) {
            if (portToHead[dPort] != groupPort) {
              cout << "-E- Missmatching group: " << groupPort->getName()
                   << " when inspecting CA: " << dPort->getName()
                   << " on leaf switch: " << swNode->getName() << endl;
              return(1);
            }
          } else {
            groupPort = portToHead[dPort];
          }
        } else {
          // no group was set yet to the port
          if (!groupPort) {
            groupPort = dPort;
          }
          portToHead[dPort] = groupPort;
        }
      } // all lids on the min-hop
    } // all CA ports
  } // all nodes

  // populate the groups
  map<IBPort*, IBPort*>::iterator mI;
  for (mI = portToHead.begin(); mI != portToHead.end(); mI++) {
    portGroups[(*mI).second].insert((*mI).first);
  }
  cout << "-I- Grouped " << portToHead.size() << " end-ports into "
       << portGroups.size() << " groups" << endl;
  return(0);
}

// write port groups into a file
int
NetSplitDumpGroupsFile(map_port_set_ports &portGroups,
                       string groupsFileName)
{
  ofstream gFile;
  string err_message;
  int rc = IBFabric::OpenFile(groupsFileName.c_str(), gFile, false, err_message);
  if (rc) {
      cout << "-E- " << err_message << endl;
      exit(1);
  }

  /*
  if (!gFile.good()) {
    cout << "-E- Could not open groups file:" << groupsFileName
         << " for write" << endl;
    exit(1);
  }
  */

  map_port_set_ports::iterator gI;
  for (gI = portGroups.begin(); gI != portGroups.end(); gI++) {
    IBPort *head = (*gI).first;
    if ((*gI).second.size() == 0) {
      cout << "-W- Skipping empty group: " << head->getName() << endl;
      continue;
    }

    IBPort *firstPort = (*(*gI).second.begin());
    gFile << "GROUP: " << firstPort->p_node->getName()
          << "/P" << (unsigned int)firstPort->num << endl;

    set< IBPort* >::iterator sI;
    for (sI = (*gI).second.begin(); sI != (*gI).second.end(); sI++) {
      IBPort *port = (*sI);
      gFile << port->p_node->getName() << "/P"
            << (unsigned int)port->num << endl;
    }
  }
  gFile.close();
  return(0);
}


// parse groups
// GROUP: name
// <Node>/P<n> | GUID
int
NetSplitParseGroupsFile(IBFabric &fabric, string fileName,
                        map_port_set_ports &portGroups)
{
  ifstream grpsFile;
  grpsFile.open(fileName.c_str(), ifstream::in);
  if (!grpsFile.is_open()) {
    cout << "-F- Given head-ports file could not be opened!" << endl;
    exit(1);
  }

  regExp grpLine("^\\s*GROUP:\\s*(\\S+)\\s*$");
  regExp nameLine("^\\s*(\\S+)/P([0-9]+)\\s*$");
  regExp guidLine("^\\s*0x([0-9a-zA-Z]+){16}\\s*$");

  char line[256];
  uint64_t guid;
  IBPort *head = NULL;
  cout << "-I- Obtaining groups from file:" << fileName << endl;
  while (grpsFile.getline(line,256) && grpsFile.good()) {
    rexMatch *p_rexRes;

    p_rexRes = grpLine.apply(line);
    if (p_rexRes) {
      if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
        cout << "-V- NetSplitParseGroupsFile new group: "
             << p_rexRes->field(1) << endl;
      head = NULL;
      delete p_rexRes;
      continue;
    }

    p_rexRes = nameLine.apply(line);
    if (p_rexRes) {
      string name = p_rexRes->field(1);
      IBNode *pNode = fabric.getNode(name);
      if (!pNode) {
        cout << "-E- Fail to find node:" << name << endl;
      } else {
        if (pNode->type == IB_SW_NODE) {
          cout << "-E- Cannot handle switches in groups node:" << name << endl;
        } else {
          phys_port_t pn = (phys_port_t)atoi(p_rexRes->field(2).c_str());
          IBPort *pPort = pNode->getPort(pn);
          if (!pPort) {
            cout << "-E- Fail to find node: " << name
                 << " port: " << (unsigned int)pn << endl;
          } else {
            if (!head) {
              head = pPort;
            }
            portGroups[head].insert(pPort);
          }
        }
      }
      delete p_rexRes;
      continue;
    }

    // try as GUID
    p_rexRes = nameLine.apply(line);
    if (p_rexRes) {
      guid = strtoull(line, NULL, 0);
      IBPort *pPort = fabric.getPortByGuid(guid);
      if (!pPort) {
        cout << "-E- Fail to find port by GUID:" << line << endl;
      } else {
        if (pPort->p_node->type == IB_SW_NODE) {
          cout << "-E- Cannot handle switch ports in groups:"
               << pPort->getName() << endl;
        } else {
          if (!head) {
            head = pPort;
          }
          portGroups[head].insert(pPort);
        }
      }
      delete p_rexRes;
      continue;
    }
    cout << "-E- Fail to parse line:" << line << endl;
  }
  unsigned int minSize = 0;
  unsigned int maxSize = 0;
  for (map_port_set_ports::iterator gI = portGroups.begin();
       gI != portGroups.end(); gI++) {
    unsigned int s = (unsigned int)(*gI).second.size();
    if (!maxSize) {
      maxSize = minSize = s;
    } else {
      if (s < minSize) minSize = s;
      if (s > maxSize) maxSize = s;
    }
  }
  cout << "-I- Defined " << portGroups.size() << " groups with sizes from: "
       << minSize << " to: " << maxSize << endl;
  grpsFile.close();
  return(portGroups.size() == 0);
}

// mark which links are part of which partition
int NetSplitMarkMinHopLinks(IBPort *headPort,                    // IN
                            set< IBPort* > &portGroup,           // IN
                            map<IBNode*, IBPort*> &switchToHead, // OUT
                            map<IBPort*, IBPort*> &portToHead)   // OUT
{
  queue<IBNode*> Q;
  set<IBNode*> visited;
  set<IBPort*> visitedPorts;

  // algorithm start BFS from all nodes in the group
  // follow the min hops to all other nodes in the group
  set<lid_t> groupLIDs;
  for (set< IBPort* >::iterator gI = portGroup.begin();
       gI != portGroup.end(); gI++) {
    IBPort *port = (*gI);
    for (lid_t lid = port->base_lid;
         lid < port->base_lid+(lid_t)(1<<port->lmc); lid++) {
      groupLIDs.insert(lid);
    }
    if (port->p_remotePort &&
        port->p_remotePort->p_node->type == IB_SW_NODE) {
      if (visited.find(port->p_remotePort->p_node) == visited.end()) {
        Q.push(port->p_remotePort->p_node);
        visited.insert(port->p_remotePort->p_node);
      }
      visitedPorts.insert(port->p_remotePort);
    }
  }

  // collect all switches visited into switchToHead
  while (!Q.empty()) {
    IBNode *node = Q.front();
    Q.pop();

    if (node->type != IB_SW_NODE) continue;

    // go over all the LIDs in the group and push the other side node
    // if not already there
    for ( set<lid_t>::iterator lI = groupLIDs.begin();
          lI != groupLIDs.end(); lI++) {
      lid_t lid = (*lI);
      int minHops = node->getHops(NULL, lid);
      if (minHops == IB_HOP_UNASSIGNED) {
        cout << "-W- Found - un-assigned hops for node:"
             << node->name << " to lid:" << lid << ")" << endl;
      } else {
        for (unsigned int pn = 1; pn <= node->numPorts; pn++) {
          IBPort *port = node->getPort((phys_port_t)pn);
          if (port && (node->getHops(port, lid) == minHops)) {
            if (visited.find(port->p_remotePort->p_node) == visited.end()) {
              Q.push(port->p_remotePort->p_node);
              visited.insert(port->p_remotePort->p_node);
            }
            visitedPorts.insert(port);
            visitedPorts.insert(port->p_remotePort);
          }
        }
      }
    }
  }

  // now move the visited marking to the outputs but inspect for duplicants.
  // is it allowed for a switch to be part in two group? NO!
  unsigned int numSwitches = 0;
  for (set<IBNode*>::iterator nI = visited.begin();
       nI != visited.end(); nI++) {
    IBNode *node = *nI;
    if (node->type != IB_SW_NODE) continue;

    numSwitches++;
    map<IBNode*, IBPort*>::iterator stohI = switchToHead.find(node);
    if (stohI != switchToHead.end()) {
      cout << "-E- switch: " << (*nI)->getName() << " part of group: "
           << (*stohI).second->getName() << " but now found on group: "
           << headPort->getName() << endl;
      exit(1);
    } else {
      switchToHead[node] = headPort;
    }
  }

  for (set<IBPort *>::iterator pI = visitedPorts.begin();
       pI != visitedPorts.end(); pI++) {
    portToHead[*pI] = headPort;
  }
  cout << "-I- MinHop links for:" << headPort->getName() << " include "
       << numSwitches << " switches and " << visitedPorts.size()
       << " links" << endl;
  return(0);
}

char fileNamify(char c) {
  if (c == '/')
    return('_');
  return(c);
}

// find a DR path to a node
int
NetSplitGetMinHopDRToPort(IBPort* from, IBPort* to, list<unsigned int> &dr)
{
  // get DST LID
  lid_t dLid = to->base_lid;
  IBPort *port = from;
  while (1) {
    if (port->p_node->type != IB_SW_NODE) {
      if (port == to) {
        return(0);
      } else if (port != from) {
        cout << "-E- BUG: got to a different end-port then requested." << endl;
        return(1);
      }
      dr.push_front(port->num);
      port = port->p_remotePort;
    } else {
      IBNode *node = port->p_node;

      if (to->p_node == node) {
        return(0);
      }

      // it is a switch find the first min hop port
      uint8_t minHops = node->getHops(NULL, dLid);
      if (minHops == IB_HOP_UNASSIGNED) {
        cout << "-W- Found - un-assigned hops for node:"
             << node->name << " to lid:" << dLid << ")" << endl;
        return(1);
      } else {
        IBPort *found = NULL;
        for (unsigned int pn = 1; !found && (pn <= node->numPorts); pn++) {
          IBPort *oPort = node->getPort((phys_port_t)pn);
          if (oPort && (node->getHops(oPort, dLid) == minHops)) {
            dr.push_front(pn);
            found = oPort->p_remotePort;
          }
        }
        if (!found) {
          cout << "-E- Got to a dead end going from: " << from->getName()
               << " to: " << to->getName() << " at: "
               << node->getName() << endl;
          return(1);
        }
        port = found;
      }
    }
  }
  return(1);
}

// write out the join/split scripts
int
NetSplitWriteSplitScripts(IBFabric &fabric,                      // IN
                          map_port_set_ports &portGroups,        // IN
                          map<IBNode*, IBPort*> &switchToHead,   // IN
                          map<IBPort*, IBPort*> &portToHead,     // IN
                          string outputDirName)                  // ~OUT
{
  // check if the out dir exists
  DIR *outDir = opendir(outputDirName.c_str());
  if (!outDir) {
    cout << "-E- Could not open output directory: " << outputDirName
         << endl;
    exit(1);
  }
  closedir(outDir);
  outDir = NULL;

  // convert the switch to head to head switches...
  map<IBPort*, set<IBNode*> > groupHeadSwitches;
  map<IBNode*, IBPort*>::iterator swI;
  for (swI = switchToHead.begin(); swI != switchToHead.end(); swI++) {
    groupHeadSwitches[(*swI).second].insert((*swI).first);
  }

  // for each of the groups find the edge of the group by
  // going through all group switches looking for non used ports
  map_port_set_ports groupEdgePorts;
  map_port_set_ports::iterator gI;
  for (gI = portGroups.begin(); gI != portGroups.end(); gI++) {
    IBPort *headPort = (*gI).first;
    // go over all switches of this group
    set <IBNode*> *grpSwitches = &groupHeadSwitches[headPort];
    set <IBNode*>::iterator gswI;
    for (gswI = (*grpSwitches).begin(); gswI != (*grpSwitches).end(); gswI++) {
      // for each sw find all ports that do not part in the MinHop
      IBNode *node = (*gswI);
      for (unsigned int pn = 1; pn <= node->numPorts; pn++) {
        IBPort *port = node->getPort((phys_port_t)pn);
        if (!port || ! port->p_remotePort ) continue;
        if (port->p_remotePort->p_node->type != IB_SW_NODE) continue;
        if (portToHead[port] != headPort) {
          groupEdgePorts[headPort].insert(port);
        }
      }
    }
  }

  // now format each group on/off file
  unsigned int numFiles = 0;
  for (gI = groupEdgePorts.begin(); gI != groupEdgePorts.end(); gI++) {
    IBPort *headPort = (*gI).first;
    string name = headPort->p_node->getName();
    // replace / by _
    transform(name.begin(), name.end(), name.begin(), fileNamify);
    string fileName = outputDirName + string("/") + name + string(".sh");
    ofstream f;
    string err_message;
    int rc = IBFabric::OpenFile(fileName.c_str(), f, false, err_message);
    if (rc) {
        cout << "-E- " << err_message << endl;
        continue;
    }

    /*
    if (! f.good()) {
      cout << "-E- Could not open output file: " << fileName << endl;
      continue;
    }
    */

    // use the node name without the Uxx as system name
    // HACK: work around the system names are not very well handled in LST
    string sysName = headPort->p_node->name;
    size_t upos = sysName.rfind("/U");
    if (upos != string::npos)
      sysName = sysName.substr(0,upos);

    f << "#!/bin/bash" << endl;
    f << "######################################################" << endl;
    f << "#\n# This code was automatically created by ibnetsplit\n" << endl;
    f << "######################################################" << endl;
    f << "# Change the below setting if required:" << endl;
    f << "# To be run from HCA matching:" << headPort->p_node->getName()
      << endl;
    f << "C=mlx4_0" << endl;
    f << "# To be run through port:" << endl;
    f << "P=" << (unsigned int)headPort->num << endl;
    f << "######################################################\n" << endl;
    f << "if test `hostname` != \"" << sysName << "\"; then" << endl;
    f << "   echo this script must be run from: " << sysName << endl;
    f << "   exit 1\nfi" << endl;
    f << "if test x$1 != xenable -a x$1 != xdisable; then " << endl;
    f << "   echo usage: $0 disable\\|enable" << endl;
    f << "   exit 1\nfi" << endl;
    f << "A=$1" << endl;

    set<IBPort*>::iterator pI;
    map<IBPort*, string> DRPerPort;
    for (pI = ((*gI).second).begin();
         pI != (*gI).second.end();
         pI++) {
      IBPort *port = *pI;
      f << "# dest: " << port->getName() << endl;

      list<unsigned int> dr;
      if (NetSplitGetMinHopDRToPort(headPort, port, dr)) {
        cout << "-E- Could not get a Directed Route path from: "
             << headPort->getName() << " to: " << port->getName() << endl;
        exit(1);
      }

      ostringstream DR;
      DR << "0";
      list<unsigned int>::iterator lI;
      for (lI = dr.begin(); lI != dr.end(); lI++) {
        DR << "," << (*lI);
      }
      DRPerPort[port] = DR.str();
      f << "DR='" << DR.str() <<"'"<< endl;
      f << "N=" << (unsigned int)(*pI)->num << endl;
      f << "PN='" << port->getName() << "'" << endl;
      f << "S=`smpquery PortInfo -C $C -P $P -D $DR $N | grep PhysLinkState | sed 's/.*\\.//'`\n"
	"if test $A = enable; then\n"
	"   if test x$S != xLinkUp ; then\n"
	"      echo Enabling Port: $PN\n"
	"      ibportstate -C $C -P $P -D $DR $N enable > /dev/null\n"
	"   else \n"
	"      echo Port $PN already enabled\n"
	"   fi\n"
	"else\n"
	"   if test x$S != xDisabled ; then\n"
	"      echo Disabling Port $PN\n"
	"      ibportstate -C $C -P $P -D $DR $N disable  > /dev/null\n"
	"   else\n"
	"      echo Port $PN already disabled\n"
        "   fi\n"
	"fi\n" << endl;
    }

    map<IBPort*, string>::const_iterator mI;

    // the code below waits to get link up from those being turned on
    f << "if test $A = enable; then\n" << endl;
    f << "   echo 'Waiting for ports to become active (max 60sec)'\n" << endl;
    // declare a variable for each port to track it
    unsigned i = 0;
    for (mI = DRPerPort.begin(); mI != DRPerPort.end(); mI++) {
      i++;
      f << "   En" << i << "=0" << endl;
    }

    f <<
      "   for i in `seq 1 10`; do\n"
      "      LEFT=0\n" << endl;
    i = 0;
    for (mI = DRPerPort.begin(); mI != DRPerPort.end(); mI++) {
      i++;
      IBPort *port = (*mI).first;
      f << "      # checking " << port->getName() << endl;
      f << "      if test $En" << i << " = 0; then" << endl;
      f << "         PN='" << port->getName() << "'" << endl;
      f << "         DR='" << (*mI).second <<"'"<< endl;
      f << "         N=" << (unsigned int)port->num << endl;
      f << "         S=`smpquery PortInfo -C $C -P $P -D $DR $N | grep PhysLinkState | sed 's/.*\\.//'`" << endl;
      f << "         if test x$S = xLinkUp; then" << endl;
      f << "            echo Port $PN Up" << endl;
      f << "            En" << i << "=1 " << endl;
      f << "         else" << endl;
      f << "            LEFT=1" << endl;
      f << "         fi\n" << endl;
      f << "      fi\n" << endl;
    }
    f << "      if test $LEFT = 0; then echo All Up; exit 0; fi\n" << endl;
    f << "      sleep 6\n" << endl;
    f << "   done\n" << endl;
    f << "   echo 'Gave up waiting !!!'\n" << endl;
    i = 0;
    for (mI = DRPerPort.begin(); mI != DRPerPort.end(); mI++) {
      i++;
      IBPort *port = (*mI).first;
      f << "   if test $En" << i << " != 0; then" << endl;
      f << "      echo Port '" << port->getName() << "' is still down" << endl;
      f << "   fi\n" << endl;
    }
    f << "   exit 1" << endl;
    f << "fi" << endl;
    f.close();
    numFiles++;
  }
  cout << "-I- Wrote " << numFiles << " scripts into:"
       << outputDirName << endl;
  return(0);
}
