/*
 * Copyright (c) 2004-2018 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


/*
 * Subnet Utilities:
 *
 * The file holds a set of utilities to be run on the subnet to mimic OpenSM
 * initialization and analyze the results:
 *
 * Assign Lids: SubnMgtAssignLids
 * Init min hop tables: SubnMgtCalcMinHopTables
 * Perform Enhanced LMC aware routing: SubnMgtOsmEnhancedRoute
 * Perform standard routing: SubnMgtOsmRoute
 * Verify all CA to CA routes: SubnMgtVerifyAllCaToCaRoutes
 */


#include <set>
#include <algorithm>
#include <iomanip>

#include <infiniband/ibdm/SubnMgt.h>
#include <infiniband/ibdm/TraceRoute.h>
#include <infiniband/ibdm/ARTraceRoute.h>
#include <infiniband/ibdm/Regexp.h>


///////////////////////////////////////////////////////////////////////////////
// Assign lids given the start NodePort
// NOTE: does not support a different LMC for SW and CA

// separate for clarity
static int
AssignNodeLids(IBPort *p_port, lid_t lid,  uint8_t lmc, lid_t numLidsPerPort)
{
    IBNode *p_node = p_port->p_node;
    IBFabric* p_fabric = p_node->p_fabric;
    if ( p_node->type != IB_SW_NODE ) {
        p_port->base_lid = lid;
        p_port->lmc = lmc;
        for ( lid_t l = lid ; l < lid + numLidsPerPort; l++ )
            p_fabric->setLidPort(l, p_port);
    } else {
        // go over all ports of the node:
        // use unsigned int index to prevent overwrap
        for ( unsigned int pn = 0; pn <= p_node->numPorts; pn++ ) {
            IBPort *p_port = p_node->getPort((phys_port_t)pn);
            if ( p_port ) {
                p_port->base_lid = lid;
                p_port->lmc = lmc;
                for ( lid_t l = lid ; l < lid + numLidsPerPort; l++ )
                    p_fabric->setLidPort(l, p_port);
            }
        }
    }
    return(0);
}

int
SubnMgtAssignLids (IBPort *p_smNodePort, uint8_t lmc)
{
  list<IBPort *> thisStepPorts;
  list<IBPort *> nextStepNodePorts;
  set<IBPort *> visited;
  phys_port_t pn;
  IBFabric *p_fabric;
  IBPort *p_port;
  IBNode *p_node;
  lid_t numLidsPerPort = GET_LID_STEP(lmc);

  if (!p_smNodePort ||
      !p_smNodePort->p_node ||
      !p_smNodePort->p_node->p_fabric) {
    cout << "-E- Assign LIDs called with invalid object " << endl;
    return 1;
  }

  if ((p_smNodePort->p_node->type == IB_SW_NODE) &&
      (p_smNodePort->num != 0)) {
    cout << "-E- SM Port is on switch thus must be port 0!" << endl;
    return 1;
  }

  p_fabric = p_smNodePort->p_node->p_fabric;

  lid_t lid = (lid_t)(1<<lmc);
  int step = 0;
  p_fabric->minLid = lid;

  thisStepPorts.push_back(p_smNodePort);
  // NOTE: visited is marked when inserting to next step
  visited.insert(p_smNodePort);

  // BFS Style ...
  while (thisStepPorts.size() > 0) {
    nextStepNodePorts.clear();
    step++;

    // go over all this step ports
    while (! thisStepPorts.empty()) {
      p_port = thisStepPorts.front();
      thisStepPorts.pop_front();

      // get the node
      p_node = p_port->p_node;

      AssignNodeLids(p_port, lid, lmc, numLidsPerPort);
      // do not forget to increment the lids
      lid = (lid_t)(lid + numLidsPerPort);

      // based on the node type we do the recursion and assignment
      phys_port_t fPortNum, tPortNum;
      if (p_node->type != IB_SW_NODE) {
	// only consider this port
	fPortNum = p_port->num;
	tPortNum = p_port->num;
      } else {
	// for a switch consider all ports
	fPortNum = 1;
	tPortNum = p_node->numPorts;
      }

      // go over all ports of the node:
      for (pn = fPortNum; pn <= tPortNum; pn++) {
	IBPort *p_nPort = p_node->getPort(pn);
	if (!p_nPort || !p_nPort->p_remotePort) continue;

	// if we have a remote port that is not visited
	IBPort *p_remPort = p_nPort->p_remotePort;
	IBNode *p_remNode = p_remPort->p_node;
	IBPort *p_repRemNodePort = p_remPort;

	// The below assumes a switch node always have port 0
	if (p_remNode->type == IB_SW_NODE) {
	  p_repRemNodePort = p_remNode->getPort(0);
	  if (!p_repRemNodePort) {
	    cout << "SubnMgtAssignLids: BUG no port 0 for switch: " << p_remNode->name << endl;
	    exit(1);
	  }
	}
	// if the port representing the remote node was not visited and not included in
	if (visited.find(p_repRemNodePort) == visited.end()) {
	  visited.insert(p_repRemNodePort);
	  nextStepNodePorts.push_back(p_repRemNodePort);
	}
      }
    }
    thisStepPorts = nextStepNodePorts;
  }

  p_fabric->maxLid = (lid_t)(lid - 1);
  p_fabric->caLmc = lmc;
  p_fabric->swLmc = lmc;
  cout << "-I- Assigned " << p_fabric->maxLid - p_fabric->minLid + 1
       << " LIDs (lmc=" << (unsigned int)lmc
       << ") in " << step << " steps" << endl;
  return 0;
}


///////////////////////////////////////////////////////////////////////////////
// OpenSM style relaxation algorithm:
// First step is to mark your own or neighbors
// Loop updating from neighbors hops untill no update
int
SubnMgtCalcMinHopTables (IBFabric *p_fabric)
{
    IBNode *p_node;
    map_str_pnode::iterator nI;
    lid_t base_lid;
    lid_t lid;
    uint8_t lmc;

    // first step
    // - update self on switches:
    // go over all nodes in the fabric
    for( nI = p_fabric->NodeByName.begin();
	 nI != p_fabric->NodeByName.end(); nI++) {
        p_node = (*nI).second;

        // we should not assign hops for non SW nodes:
        if (p_node->type == IB_SW_NODE) {
          p_node->getLidAndLMC(0, base_lid, lmc);
          for (lid = base_lid; lid < base_lid + ((lid_t)1<<lmc); lid++)
            p_node->setHops(NULL,lid,0); // assign all ports value
	}
    }

    // now do the hosts
    for( nI = p_fabric->NodeByName.begin();
	 nI != p_fabric->NodeByName.end(); nI++) {
      p_node = (*nI).second;
      if (p_node->type == IB_SW_NODE) continue;
      // a non switch node might have connections on both ports
      // and also we just want to update the switch on the other side
      int conPorts = 0;
      for (unsigned int i = 1; i <= p_node->numPorts; i++) {
          IBPort *p_port = p_node->getPort((phys_port_t)i);
	if (p_port &&
	    p_port->p_remotePort &&
	    p_port->p_remotePort->p_node->type == IB_SW_NODE) {
	  p_node->getLidAndLMC(p_port->num, base_lid, lmc);
	  conPorts++;
	  // update the switch:
	  if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
	    cout << "-V- Setting switch:" << p_port->p_remotePort->p_node->name
		 << " as neighbor of CA port: " << p_port->getName() << " lid: " << base_lid << endl;

	  for (lid = base_lid; lid < base_lid + ((lid_t)1<<lmc); lid++) {
	    p_port->p_remotePort->p_node->setHops(p_port->p_remotePort,lid,1);
	  }
	}
      }
      if (!conPorts)
	cout << "-W- CA with no connected ports:" << p_node->name << endl;
    }

    // second - loop until nothing to update:
    int anyUpdate;
    int loop = 0;
    do {
        loop++;
        anyUpdate = 0;

        // go over all switch nodes:
        for( nI = p_fabric->NodeByName.begin();
	     nI != p_fabric->NodeByName.end(); nI++) {
            p_node = (*nI).second;

            // we should not assign hops for non SW nodes:
            if (p_node->type != IB_SW_NODE)
                continue;

            // go over all lids (base) on this switch:
	    for (lid_t lid = 1; lid <= p_fabric->maxLid; lid++) {
                // go over all connected ports
            for (unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
                IBPort *p_port = p_node->getPort((phys_port_t)pn);

		  // do we have a port on the other side ? is it a SW ?
		  if (p_port &&
		      p_port->p_remotePort &&
		      (p_port->p_remotePort->p_node->type == IB_SW_NODE)) {
		    // the min we have for this lid is:
		    int minHops = p_node->getHops(p_port, lid);

		    // we need to update the local port hops only they will
		    // be made smaller by this step. I.e. the remote port
		    // hops value + 1 is < hops
		    uint8_t remNodeHops =
		      p_port->p_remotePort->p_node->getHops(NULL, lid);

		    if (remNodeHops + 1 < minHops) {
		      // need to update:
		      p_node->setHops(p_port, lid, (uint8_t)(remNodeHops + 1));
		      anyUpdate++;
		    }
		  }
                }
            }
        }
        // cout << "-I- Propagated:" << anyUpdate << " updates" << endl;
    } while (anyUpdate);
    cout << "-I- Init Min Hops Tables in:" << loop << " steps" << endl;

    // simply check that one can reach to all lids from all switches:
    // also we build a historgram of the number of ports one can use
    // to get to any lid.
    vec_int numPathsHist(50,0);
    int anyUnAssigend = 0;
    for( nI = p_fabric->NodeByName.begin();
            nI != p_fabric->NodeByName.end();
            nI++) {
        p_node = (*nI).second;
        if (p_node->type == IB_CA_NODE)
            continue;

        // go over all the lids.
        for (lid_t i = 1; i <= p_fabric->maxLid; i++ ) {
            // skip lids that are not mapped to a port:
	    IBPort *port = p_fabric->getPortByLid(i);
	    if (! port || (port->base_lid != i))
	      continue;

            int minHops = p_node->getHops(NULL, i);
            if (minHops == IB_HOP_UNASSIGNED) {
                cout << "-W- Found - un-assigned hops for node:"
                        << p_node->name << " to lid:" << i << ")" << endl;
                anyUnAssigend++;
            } else {
                // count all ports that have min hops to that lid (only HCAs count)
                IBPort *p_targetPort = p_fabric->getPortByLid(i);
                if (p_targetPort && (p_targetPort->p_node->type != IB_SW_NODE)) {
                    int numMinHopPorts = 0;
                    for (unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
                        IBPort *p_port = p_node->getPort((phys_port_t)pn);
                        if (p_port && (p_node->getHops(p_port, i) == minHops))
                            numMinHopPorts++;
                    }
                    numPathsHist[numMinHopPorts]++;
                }
            }
        }
    }

    cout << "------------------ NUM ALTERNATE PORTS TO CA HISTOGRAM --------------------" << endl;
    cout << "Describes how many out ports on every switch have the same Min Hop to each " << endl;
    cout << "target CA. Or in other words how many alternate routes are possible at the " << endl;
    cout << "switch level. This is useful to show the symmetry of the cluster.\n" << endl;
    cout << "OUT-PORTS NUM-SW-LID-PAIRS" << endl;
    for (int b = 0; b < 50 ; b++)
        if (numPathsHist[b])
            cout << setw(4) << b << "   " << numPathsHist[b] << endl;
    cout << "---------------------------------------------------------------------------\n" << endl;

    if (anyUnAssigend) {
        cout << "-W- Found - un-reachable lids." << endl;
        return 1;
    }

    // report the worst case hops count found.
    int maxHops = 0;
    IBNode *p_worstHopNode  = NULL;
    IBPort *p_worstHopPort  = NULL;
    lid_t worstHopLid = 0;
    IBNode *p_caNode;
    IBPort *p_caPort;
    vec_int maxHopsHist(50,0);

    for( nI = p_fabric->NodeByName.begin();
            nI != p_fabric->NodeByName.end();
            nI++) {
        p_caNode = (*nI).second;
        if (p_caNode->type != IB_CA_NODE)
            continue;

        // find the switch connected to the HCA port
        for (unsigned int n = 1; n <= p_caNode->numPorts; n++) {
            p_caPort = p_caNode->getPort((phys_port_t)n);
            // ignore the port if does not exist or not connected
            if (!p_caPort || !p_caPort->p_remotePort)
                continue;

            p_node = p_caPort->p_remotePort->p_node;

            // go over all lids found and get the min hop
            for (lid_t i = 1; i <= p_fabric->maxLid; i++) {
                // please ignore non CA lids and ourselves
                IBPort *p_port = p_fabric->getPortByLid(i);
                if (!p_port ||
		    (p_port->base_lid != i) ||
		    (p_port->p_node->type == IB_SW_NODE) ||
		    (p_caPort == p_port))
		  continue;

		int minHops = p_node->getHops(NULL, i);
		if (IB_HOP_UNASSIGNED != minHops) {
		  maxHopsHist[minHops]++;
		  if (minHops > maxHops) {
		    p_worstHopPort = p_node->getFirstMinHopPort(i);
		    p_worstHopNode = p_node;
		    worstHopLid = i;
		    maxHops = minHops;
		  }
		} else {
		  cout << "-W- Found - un-assigned hop for node:"
		       << p_node->name << " to port:" << p_port->p_node->name
		       << "/" << (unsigned int)p_port->num
		       << " (lid:" << i << ")" << endl;
		}
	    }
        }
    }

    // print the histogram:
    cout << "---------------------- CA to CA : MIN HOP HISTOGRAM -----------------------" << endl;
    cout << "The number of CA pairs that are in each number of hops distance." << endl;
    cout << "The data is based on topology only - even before any routing is run.\n" << endl;
    cout << "HOPS NUM-CA-CA-PAIRS" << endl;
    for (int b = 0; b <= maxHops ; b++)
        if (maxHopsHist[b])
            cout << setw(3) << b+1 << "   " << maxHopsHist[b] << endl;
    cout << "---------------------------------------------------------------------------\n" << endl;

    if (p_worstHopNode) {
        IBPort *p_worst_hop_lid_port = p_fabric->getPortByLid(worstHopLid);
        if (!p_worst_hop_lid_port)
            cout << "-E- fail to find port for worst hop lid:" << worstHopLid << endl;
        else
            cout << "-I- Found worst min hops:" << maxHops + 1 << " at node:"
                 << p_worstHopNode->name << " to node:"
                 << p_worst_hop_lid_port->p_node->name << endl;

        // the worst hop node lid
        TraceRouteByMinHops(p_fabric, p_worstHopPort->base_lid, worstHopLid);
    }
    return(0);
}


///////////////////////////////////////////////////////////////////////////////
// Fill in the FDB tables in an Extended OpesnSM style routing
// which is switch based, uses number of routes per port
// profiling and treat LMC assigned lids sequentialy
// Rely on running the SubnMgtCalcMinHopTables beforehand
// We added the notion of selecting the other system or node
// if same hop and profile
int
SubnMgtOsmEnhancedRoute(IBFabric *p_fabric)
{
    IBNode *p_node;
    cout << "-I- Using Enhanced OpenSM Routing" << endl;

    // we want to collect port subscriptions statistics:
    vec_int subscHist(10000,0);

    // also track the selections used:
    int numSelByOtherSys = 0;
    int numSelByOtherNode = 0;
    int numSelByMinSubsc = 0;

    // go over all nodes in the fabric
    for( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
            nI != p_fabric->NodeByName.end();
            nI++) {
        p_node = (*nI).second;

        // if not a switch cont
        if (p_node->type != IB_SW_NODE)
            continue;

        // define port profiles
        vec_int portsSubscriptions(p_node->numPorts,0);

	// track previous port to know if on same port (in lmc)
	IBPort *pPrevPort = 0;

	// We track the Systems we already went through:
	set<IBSystem *> goThroughSystems;

	// we also track the nodes we went through:
	set <IBNode *> goThroughNodes;

	// go over all lids on this switch:
	for (lid_t lid = 1; lid <= p_fabric->maxLid; lid++) {
            int targetIsHCA;
            IBPort *pTargetPort = p_fabric->getPortByLid(lid);
            if (pTargetPort && (pTargetPort->p_node->type == IB_SW_NODE))
                targetIsHCA = 0;
            else
                targetIsHCA = 1;

            int minHop = p_node->getHops(NULL, lid);

	    // if the min is 0 just map to port 0
	    if (minHop == 0) {
	      p_node->setLFTPortForLid(lid, 0);
	      continue;
	    }

	    // if we change dest we clear the tracking
	    if (pTargetPort != pPrevPort) {
	      goThroughSystems.clear();
	      goThroughNodes.clear();
	    }
	    pPrevPort = pTargetPort;

	    // we are going to track the min profile under three
	    // possible cases:
	    int minSubsShared = 100000;
	    int minSubsDiffNodes = 100000;
	    int minSubsDiffSystems = 100000;
	    phys_port_t minPortNumShared = 0;
	    phys_port_t minPortNumDiffNodes = 0;
	    phys_port_t minPortNumDiffSystems = 0;
	    int minSubs;

	    if (minHop < 255) {
	      // look for the port with min profile
	      for (unsigned int i = 1; i <= p_node->numPorts; i++) {
              phys_port_t pn = (phys_port_t)i;
		IBPort *p_port = p_node->getPort(pn);
		if (! p_port)
		  continue;
		if (! p_port->p_remotePort)
		  continue;

		// the hops should match the min
		if (p_node->getHops(p_port,lid) == minHop) {
		  minSubs = portsSubscriptions[pn-1];
		  IBNode   *p_remNode = p_port->p_remotePort->p_node;
		  IBSystem *p_system = p_remNode->p_system;

		  if (goThroughSystems.find(p_system) == goThroughSystems.end()) {
		    if (minSubsDiffSystems > minSubs) {
		      minSubsDiffSystems = minSubs;
		      minPortNumDiffSystems = pn;
		    }
		  }
		  if (goThroughNodes.find(p_remNode) == goThroughNodes.end()) {
		    if (minSubsDiffNodes > minSubs) {
		      minSubsDiffNodes = minSubs;
		      minPortNumDiffNodes = pn;
		    }
		  }

		  if (minSubsShared > minSubs) {
		    minSubsShared = minSubs;
		    minPortNumShared = pn;
		  }
		} // hop = min hops
	      } // all ports

	      // we always select the system over node over shared:
	      if (minPortNumDiffSystems) {
		minPortNumShared = minPortNumDiffSystems;
		numSelByOtherSys++;
	      } else if (minPortNumDiffNodes) {
		numSelByOtherNode++;
		minPortNumShared = minPortNumDiffNodes;
	      } else {
		numSelByMinSubsc++;
	      }

	      // so now we need to have the port number or error
	      if (!minPortNumShared) {
		cout << "-E- Cound not find min hop port!" << endl;
		return(1);
	      }

	      // Track used systems and nodes
	      IBPort *p_bestPort = p_node->getPort(minPortNumShared);
	      IBNode *p_remNode = p_bestPort->p_remotePort->p_node;
	      IBSystem *p_system = p_node->p_system;

	      goThroughSystems.insert(p_system);
	      goThroughNodes.insert(p_remNode);

	    } else {
	      // there is no path to that lid...
	      minPortNumShared = 255;
	    }

	    // track subscriptions:
	    if (targetIsHCA)
	      portsSubscriptions[minPortNumShared-1]++;

	    // assign the fdb table.
	    p_node->setLFTPortForLid(lid, minPortNumShared);
	} // all lids

        // we want to get some histogram of subsriptions.
        for (unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
            IBPort *p_port = p_node->getPort((phys_port_t)pn);
            if (p_port && p_port->p_remotePort) {
                if (portsSubscriptions[pn-1] == 0) {
                    cout << "-W- Unused port:" << p_port->getName() << endl;
                }
                subscHist[portsSubscriptions[pn-1]]++;
            }
        }
    } // all nodes

    cout << "-I- Enhanced selection by Sys:" << numSelByOtherSys
            << " Node:" << numSelByOtherNode
            << " Subscription:" << numSelByMinSubsc << endl;
   return(0);
}


///////////////////////////////////////////////////////////////////////////////
// Fill in the FDB tables in an OpesnSM style routing
// which is switch based, uses number of routes per port
// profiling and treat LMC assigned lids sequentialy
// Rely on running the SubnMgtCalcMinHopTables beforehand
int
SubnMgtOsmRoute(IBFabric *p_fabric)
{
  IBNode *p_node;
  cout << "-I- Using standard OpenSM Routing" << endl;

  // we want to collect port subscriptions statistics:
  vec_int subscHist(10000,0);

  // go over all nodes in the fabric
  for( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
       nI != p_fabric->NodeByName.end();
       nI++) {
    p_node = (*nI).second;

    // if not a switch cont
    if (p_node->type != IB_SW_NODE)
      continue;

    // define port profiles
    vec_int portsSubscriptions(p_node->numPorts,0);

    // go over all valid lid values (i.e. base lids )
    for (lid_t lid = 1; lid <= p_fabric->maxLid; lid++) {
      int targetIsHCA;
      IBPort *pTargetPort = p_fabric->getPortByLid(lid);
      if (pTargetPort && (pTargetPort->p_node->type == IB_SW_NODE))
        targetIsHCA = 0;
      else
        targetIsHCA = 1;

      // get the minimal hop count from this port:
      int minHop = p_node->getHops(NULL,lid);

      // if the min is 0 just map to port 0
      if (minHop == 0) {
        p_node->setLFTPortForLid(lid, 0);
        continue;
      }

      // initialize the min subsription to a huge number:
      int minSubsc = 100000;
      phys_port_t minSubsPortNum = 0;
#if 1
      // look for the port with min profile
      if (minHop != 255) {
        for (unsigned int i = 1; i <= p_node->numPorts; i++) {
            phys_port_t pn = (phys_port_t)i;
          IBPort *p_port = p_node->getPort(pn);

          if (! p_port)
            continue;

          // the hops should match the min
          if (p_node->getHops(p_port, lid) == minHop) {
            // Standard OpenSM Routing:
            // is it the lowest subscribed port:
            if (portsSubscriptions[pn-1] < minSubsc) {
              minSubsPortNum = pn;
              minSubsc = portsSubscriptions[pn-1];
            }
          }
        }
      } else {
	// it is possible to have "gaps" in LID assignment
        minSubsPortNum = 255;
      }
#else
      // do a random selection
      {
        vector<unsigned int> minHopOutPorts;
        for (unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
          IBPort *p_port = p_node->getPort((phys_port_t)pn);

          if (! p_port)
            continue;

          // the hops should match the min
          if (p_node->getHops(p_port, lid) == minHop) {
            minHopOutPorts.push_back(pn);
          }
        }
        double portRand = 1.0*minHopOutPorts.size()*rand()/RAND_MAX;
        unsigned int portIdx = int(portRand);
        minSubsPortNum = minHopOutPorts[portIdx];
      }
#endif
      // so now we need to have the port number or error
      if (!minSubsPortNum) {
        cout << "-E- Cound not find min hop port for lid:" << lid << " on node:" << p_node->name  << endl;
	p_node->repHopTable();
        return(1);
      }

      // track subscriptions only if target is not a switch:
      if (targetIsHCA)
        portsSubscriptions[minSubsPortNum-1]++;

      // assign the fdb table.
      p_node->setLFTPortForLid(lid, minSubsPortNum);
      if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
	cout << "-V- Setting switch:" << p_node->name
	     <<  " LFT(" << lid << ") = " << (unsigned int)minSubsPortNum << endl;
    } // all lids

    // we want to get some histogram of subsriptions.
    for (unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
      IBPort *p_port = p_node->getPort((phys_port_t)pn);
      if (p_port && p_port->p_remotePort) {
        if (portsSubscriptions[pn-1] == 0) {
          cout << "-W- Unused port:" << p_port->getName() << endl;
        }
        subscHist[portsSubscriptions[pn-1]]++;
      }
    }
  } // all nodes

  return(0);
}


///////////////////////////////////////////////////////////////////////////////
// Given a list of root nodes mark them with a zero rank
// Then BFS and rank min
// note we use the provided map of IBNode* to int for storing the rank
int
SubnRankFabricNodesByRootNodes(IBFabric *p_fabric,
        list_pnode rootNodes,
        map_pnode_rank &nodesRank)
{
    list_pnode curNodes, nextNodes;
    curNodes = rootNodes;
    uint8_t rank = 0;

    // rank by zero the starting nodes:
    for (list_pnode::iterator nI = rootNodes.begin();
            nI != rootNodes.end();
            nI++) {
        IBNode *p_node = (*nI);
        nodesRank[p_node] = 0;
        p_node->rank = 0;
    }

    // ok so now we BFS
    while (curNodes.size()) {
        nextNodes.clear();
        rank++;

        // go over cur step nodes
        for (list_pnode::iterator lI = curNodes.begin();
                lI != curNodes.end(); lI++) {
            IBNode *p_node = *lI;
            // go over all ports to find unvisited remote nodes
            for (unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
                IBPort *p_port = p_node->getPort((phys_port_t)pn);
                if (! p_port)
                    continue;
                // might have remote port
                if (p_port->p_remotePort) {
                    // was it visited?
                    IBNode *p_remNode = p_port->p_remotePort->p_node;
                    if (nodesRank.find(p_remNode) == nodesRank.end()) {
                        // add it
                        nextNodes.push_back(p_remNode);
                        // mark it:
                        nodesRank[p_remNode] = rank;
                        p_remNode->rank = (uint8_t)rank;
                    }
                }
            }
        }
        curNodes = nextNodes;
    }

    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
        cout << "-I- Max nodes rank=" << rank  << endl;
    return 0;
}


///////////////////////////////////////////////////////////////////////////////
// Given a regular expression for nodes mark them with a zero rank
// Then BFS and rank min
// note we use the provided map of IBNode* to int for storing the rank
int
SubnRankFabricNodesByRegexp(IBFabric *p_fabric,
        const char * nodeNameRex,
        map_pnode_rank &nodesRank)
{
    regExp nodeRex(nodeNameRex);
    rexMatch *p_rexRes;
    list_pnode rootNodes;

    // go over all nodes of the fabric;
    for (map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
            nI != p_fabric->NodeByName.end(); nI++) {
        // match rex ?
        p_rexRes = nodeRex.apply((*nI).first.c_str());
        if (p_rexRes) {
            cout  << "-I- Starting UpDown Routing from node:"
                    << (*nI).second->name << endl;
            rootNodes.push_back((*nI).second);
            delete p_rexRes;
        }
    }
    return SubnRankFabricNodesByRootNodes(p_fabric, rootNodes, nodesRank);
}


///////////////////////////////////////////////////////////////////////////////
// Clasify the routes for the same port to port
// Given two list of nodes provide the number of shared
// Systems and Shared nodes:
int
SubnFindPathCommonality(list_pnode *p_path1, list_pnode *p_path2,
                        list_phys_ports *p_portsList1,
                        list_phys_ports *p_portsList2,
                        int *commonSystems, int *commonNodes,
                        int *commonLinks)
{
    map_pnode_int nodesIntersection;
    map_pport_int portsIntersection;
    map_psystem_int systemIntersection;
    IBSystem *p_system;
    IBNode *p_node;
    IBPort *p_port;

    // we got it now:
    *commonNodes = 0;
    *commonSystems = 0;
    *commonLinks = 0;

    list_phys_ports::const_iterator portsI = p_portsList1->begin();
    int numPorts = (int)p_portsList1->size();
    int curr_port = 1;
    // Go over all nodes in first path add them to the map:
    for (list_pnode::const_iterator lI = p_path1->begin();
            lI != p_path1->end(); lI++) {
        p_node = *lI;
        nodesIntersection[p_node] = 1;
        p_system = p_node->p_system;
        systemIntersection[p_system] = 1;

        p_port = p_node->getPort(*portsI);
        if (p_port && curr_port < numPorts)
            portsIntersection[p_port] = 1;
        ++portsI;
        curr_port++;
    }

    // We only need to count the number of pre-existing nodes.
    // we do it once for each nodes.
    portsI = p_portsList2->begin();
    for (list_pnode::const_iterator lI = p_path2->begin();
            lI != p_path2->end(); lI++) {
        // we only count once for each node.
        p_node = *lI;
        map_pnode_int::iterator mI = nodesIntersection.find(p_node);
        if ( (mI != nodesIntersection.end()) &&
                ((*mI).second == 1) ) {
            (*commonNodes)++;
            // we increase it for next time
            (*mI).second++;
        }
        if (portsI != p_portsList2->end()) {
            p_port = p_node->getPort(*portsI);
            map_pport_int::iterator mpI = portsIntersection.find(p_port);
            if ( (mpI != portsIntersection.end()) &&
                    ((*mpI).second == 1) ) {
                (*commonLinks)++;
                // we increase it for next time
                (*mpI).second++;
            }
            ++portsI;
        }

        p_system = p_node->p_system;
        map_psystem_int::iterator sI = systemIntersection.find(p_system);
        if ( (sI != systemIntersection.end()) &&
                ((*sI).second == 1) ) {
            (*commonSystems)++;
            // we increase it for next time
            (*sI).second++;
        }
    }

    //cout << "P1:" << p_path1->size()
    //  << " P2:"<< p_path2->size()
    //  << " Intersaction:" << *commonNodes << "/" << *commonSystems<<endl;
    return(0);
}

#define MAX_COMMON_SYSTEMS_REPORT 10
#define MAX_COMMON_LINKS_REPORT   10

///////////////////////////////////////////////////////////////////////////////
// Verify point to point connectivity
int
SubnMgtVerifyAllCaToCaRoutes(IBFabric *p_fabric, const char *outDir)
{
  bool anyLmcUsed = false;
  bool anyCommonLinks = false;

  int anyError = 0, paths = 0;
  vec_int maxHopsHist(50,0);
  // we want to track common Nodes and Systems
  // based on the original path length
  int CommonNodes[50][16];
  int CommonSystems[50][16];
  int CommonLinks[50][16];

  unsigned int maxDepth = 0;
  int maxLinkSubscriptions = 0;
  int maxDlidPerOutPort = 0;
  int commonSystemsReport = 0;
  int commonLinksReport = 0;

  //FabricUtilsVerboseLevel = 0xff; //!!!

  // to track the actual paths going through each switch port
  // we need to have a map from switch node to a vector of count
  // per port.
  //int X1[10240];
  map_pnode_vec_int switchPathsPerOutPort;
  //int X2[10240];

  //memset(X1,0,sizeof(X1));
  //memset(X2,0,sizeof(X2));

  // track the number of dlids actually routed through each switch port.
  // to avoid memory scalability we do the path scanning with dest port
  // in the external loop. So we only need to look on the aggregated
  // vector per port at the end of all sources and sum up to the final results
  map_pnode_vec_int switchDLidsPerOutPort;

  cout << "-I- Verifying all CA to CA paths ... " << endl;
  // initialize the histograms:
  memset(CommonNodes, 0 , 50*16 * sizeof(int));
  memset(CommonSystems, 0 , 50*16 * sizeof(int));
  memset(CommonLinks, 0, 50*16 * sizeof(int));

  unsigned int hops, maxHops = 0;
  list_pnode path1, path2, *p_path;
  list_phys_ports portsList1, portsList2, *p_portsList;

  // NOTE: since it is OK to have just one of the DLIDs assigned to a single
  // port be available as path from any given source we do not complain
  // about dead paths unless all of them from sPort to dPort are not avaiable

  // go over all ports in the fabric
  for (lid_t dLid = p_fabric->minLid; dLid <= p_fabric->maxLid; dLid++) {

    IBPort *p_dstPort = p_fabric->getPortByLid(dLid);
    IBVPort *p_dstVPort = NULL;
    lid_t pathsPerDlid = 1;

    if (p_dstPort) {
      if (p_dstPort->base_lid != dLid || p_dstPort->p_node->type == IB_SW_NODE) continue;

      // go over all LMC values of the destination:
      pathsPerDlid = (lid_t)(1 << p_dstPort->lmc);
      if (pathsPerDlid > 1)
        anyLmcUsed = true;
    } else { // check if the dlid is a vport lid
        p_dstVPort = p_fabric->getVPortByLid(dLid);
        if (!p_dstVPort) continue;
    }

    // tracks if a path to current dlid was found per switch out port
    map_pnode_vec_int switchAnyPathsPerOutPort;

    // go over all the rest of the ports - but only once - so skip non base lids
    for (lid_t sLid = p_fabric->minLid; sLid <= p_fabric->maxLid; sLid++ ) {
      IBPort *p_srcPort = p_fabric->getPortByLid(sLid);
      if (!p_srcPort || p_srcPort->base_lid != sLid) continue;

      // Avoid tracing to itself
      if (sLid == dLid)
        continue;
      if (p_srcPort->p_node->type == IB_SW_NODE)
        continue;
      if (p_dstVPort && p_dstVPort->getIBPortPtr()->base_lid == sLid)
        continue;

      bool anyPathToDest = false;

      // go over all LMC combinations:
      for (lid_t l = 0; l < pathsPerDlid; l++) {
        paths++;

        // we track the path nodes in lists but need to know which one
        // to use:
        if (l == 0) {
          p_path = &path1;
          p_portsList = &portsList1;
        } else {
          p_path = &path2;
          p_portsList = &portsList2;
        }
        p_path->clear();
        p_portsList->clear();

        // now go and verify the path:
        if (!TraceRouteByLFT(p_fabric, sLid, lid_t(dLid + l), &hops,
                                              p_path, p_portsList, false)) {
          anyPathToDest = true;
          // track the hops histogram
          maxHopsHist[hops]++;
          if (hops > maxHops) maxHops = hops;

          // populate the number of the out ports along the path
          // but ignore the last element
          list_phys_ports::const_iterator pI = p_portsList->begin();
          for (list_pnode::const_iterator lI = p_path->begin();
               lI != p_path->end() && pI != p_portsList->end() ; lI++, pI++) {
            // init the ports count vector if needed
            IBNode *pNode = (*lI);
            if (switchPathsPerOutPort.find(pNode) == switchPathsPerOutPort.end()) {
              vec_int tmp(pNode->numPorts + 1,0);
              switchPathsPerOutPort[pNode] = tmp;
            }

            // init the marking of any path through the port if needed:
            if (switchAnyPathsPerOutPort.find(pNode) == switchAnyPathsPerOutPort.end()) {
              vec_int tmp(pNode->numPorts + 1,0);
              switchAnyPathsPerOutPort[pNode] = tmp;
            }

            list_pnode::const_iterator nlI = lI;
            nlI++;
            if (nlI != p_path->end()) {
              unsigned int outPort = *pI; //pNode->getLFTPortForLid((lid_t)(dLid + l));
              switchPathsPerOutPort[pNode][outPort]++;
              if (maxLinkSubscriptions < switchPathsPerOutPort[pNode][outPort])
                maxLinkSubscriptions = switchPathsPerOutPort[pNode][outPort];

              switchAnyPathsPerOutPort[pNode][outPort]++;
            }
          }

          // Analyze the path against the previous path:
          if (l != 0) {
            static int commonSystems, commonNodes, commonLinks;

            // we only care about paths longer then 2 hops since the two switches
            //  must be identical
            if (path1.size()) {
              SubnFindPathCommonality(&path1, &path2,
                                      &portsList1, &portsList2,
                                      &commonSystems, &commonNodes, &commonLinks);


              // track the max depth
              if (path1.size() > maxDepth)
                  maxDepth = (unsigned int)path1.size();

              if (maxDepth > 15) {
                cout << "-E- Found a path length > 15. Need to recompile with larger"
                     << " Histogram size!" << endl;
                exit(1);
              }

              if (!path1.size()) {
                cout << "-W- Zero size path1 ???" << endl;
                continue;
              }

              // store the statistics:
              CommonSystems[commonSystems][0]++;
              CommonSystems[commonSystems][path1.size()]++;
              CommonNodes[commonNodes][0]++;
              CommonNodes[commonNodes][path1.size()]++;
              if(commonLinks) {
                  CommonLinks[commonLinks][0]++;
                  CommonLinks[commonLinks][path1.size()-1]++;
                  anyCommonLinks = true;

                  if(commonLinksReport < MAX_COMMON_LINKS_REPORT) {
                    cout << "---- COMMON LINKS PATH ----- " << endl;
                    cout << "From:"
                         << p_srcPort->p_node->name << "/" << (int) p_srcPort->num
                         << " to:" << p_dstPort->p_node->name << "/" << (int) p_dstPort->num
                         << endl;
                    cout << "Path 1" << endl;
                    list_phys_ports::iterator portI = portsList1.begin();
                    for (list_pnode::iterator lI = path1.begin();
                         lI!= path1.end();++lI, ++portI)
                      cout << "." << (*lI)->name.c_str() <<
                        "/" << (int) *portI << endl;

                    cout << "Path 2" << endl;
                    portI = portsList2.begin();
                    for (list_pnode::iterator lI = path2.begin();
                         lI != path2.end();
                         ++lI, ++portI)
                      cout << "." << (*lI)->name.c_str() <<
                        "/" << (int) *portI << endl;

                    commonLinksReport++;
                  } else if (commonLinksReport == MAX_COMMON_LINKS_REPORT) {
                      cout << "... COMMON LINKS PATH TRUNCATED "<<
                          "(more than " << (int) MAX_COMMON_LINKS_REPORT <<
                          ")." << endl;
                      commonLinksReport++;
                  }


              }

              if (commonSystems > 5) {
                  if(commonSystemsReport < MAX_COMMON_SYSTEMS_REPORT) {
                    cout << "---- MORE THEN 5 COMMON SYSTEMS PATH ----- " << endl;
                    cout << "From:"
                         << p_srcPort->p_node->name << "/" << (int) p_srcPort->num
                         << " to:" << p_dstPort->p_node->name << "/" << (int) p_dstPort->num
                         << endl;
                    cout << "Path 1" << endl;
                    for (list_pnode::iterator lI = path1.begin();
                         lI!= path1.end();lI++)
                      cout << "." << (*lI)->name.c_str() << endl;

                    cout << "Path 2" << endl;
                    for (list_pnode::iterator lI = path2.begin();
                         lI != path2.end();
                         lI++ )
                      cout << "." << (*lI)->name.c_str() << endl;

                    commonSystemsReport++;
                  } else if (commonSystemsReport == MAX_COMMON_SYSTEMS_REPORT) {
                      cout << "... MORE THEN 5 COMMON SYSTEMS PATH TRUNCATED " <<
                          "(more than " << (int) MAX_COMMON_SYSTEMS_REPORT <<
                          ")." << endl;
                      commonSystemsReport++;
                  }

              }
            }
            // cleanup :
            // path2.clear();
          }

        } // route exists

      } // all lmc values

      if (!anyPathToDest) {
          if (anyError < 1000) {
            cout << "-E- Fail to find a path from:"
                 << p_srcPort->p_node->name << "/"
                 << (unsigned int)p_srcPort->num;
            if (p_dstPort) {
                cout << " to:" << p_dstPort->p_node->name << "/"
                     << (unsigned int)p_dstPort->num;
            } else { // dest is vport
                cout << " to " << p_dstVPort->getIBPortPtr()->p_node->name << "/"
                     << (unsigned int)p_dstVPort->getIBPortPtr()->num
                     << ", vnode: " << p_dstVPort->m_p_vnode->getDescription()
                     << ", vport: " << p_dstVPort->getVPortNum();
            }
            cout << endl;
          }
        anyError++;
      }

      //isPLFTEnabled check VL15 connectivity
      if (p_fabric->isPLFTEnabled()) {
          anyPathToDest = false;
          for (lid_t l = 0; l < pathsPerDlid; l++) {
              if(!TraceRouteByLFT(p_fabric, sLid, lid_t(dLid + l),
                                               NULL, NULL, NULL, true))
                  anyPathToDest = true;
          }
          if (!anyPathToDest) {
            cout << "-E- Fail to find a VL15 path from:"
                 << p_srcPort->p_node->name << "/"
                 << (unsigned int)p_srcPort->num;
            if (p_dstPort) {
                cout << " to:" << p_dstPort->p_node->name << "/"
                     << (unsigned int)p_dstPort->num;
            } else { // dest is vport
                cout << " to " << p_dstVPort->getIBPortPtr()->p_node->name << "/"
                     << (unsigned int)p_dstVPort->getIBPortPtr()->num
                     << ", vnode: " << p_dstVPort->m_p_vnode->getDescription()
                     << ", vport: " << p_dstVPort->getVPortNum();
            }
            cout << endl;
            anyError++;
          }
      }
    } // all src lids

    // cleanup the list of nodes
    //path1.clear();

    // add to dlid per port vector:
    for (map_pnode_vec_int::iterator nI = switchAnyPathsPerOutPort.begin();
         nI !=  switchAnyPathsPerOutPort.end();
         nI++) {
      IBNode *pNode = (*nI).first;
      for (unsigned int pn = 1; pn <= pNode->numPorts; pn++) {
        if (switchAnyPathsPerOutPort[pNode][pn]) {
          // init if required:
          if (switchDLidsPerOutPort.find(pNode) == switchDLidsPerOutPort.end()) {
            vec_int tmp(pNode->numPorts + 1,0);
            switchDLidsPerOutPort[pNode] = tmp;
          }

          switchDLidsPerOutPort[pNode][pn]++;
          if (switchDLidsPerOutPort[pNode][pn] > maxDlidPerOutPort)
            maxDlidPerOutPort = switchDLidsPerOutPort[pNode][pn];
        }
      }
    }

  } // all dlids

  // cleanup TraceRouteByLFT output lists
  // (they are also cleaned before every call to TraceRouteByLFT)
  path1.clear();
  path2.clear();
  portsList1.clear();
  portsList2.clear();

  cout << "---------------------- CA to CA : LFT ROUTE HOP HISTOGRAM -----------------" << endl;
  cout << "The number of CA pairs that are in each number of hops distance." << endl;
  cout << "This data is based on the result of the routing algorithm.\n" << endl;
  cout << "HOPS NUM-CA-CA-PAIRS" << endl;
  for (int b = 0; b <= (int)maxHops ; b++)
    if (maxHopsHist[b])
      cout << setw(3) << b+1 << "   " << maxHopsHist[b] << endl;
  cout << "---------------------------------------------------------------------------\n" << endl;

  if (anyLmcUsed) {

    if (anyCommonLinks) {
        cout << "------------------ LMC BASED ROUTING :COMMON LINKS HISTOGRAM -----------------" << endl;

        cout << "Describes the distribution of the number of common links between the " << endl;
        cout << "different LMC paths of all the CA to CA paths," << endl;
        cout << "not including the first and last links (from/to the CA).\n"
            << endl << endl;
        cout << "PATH HOPS|";
        unsigned int maxLinks = maxDepth - 1;

        for (unsigned int d = 1; d <= maxLinks; d++)
          cout <<  setw(6) << d << "|";
        cout << endl;

        for (unsigned int b = 0; b <= maxLinks ; b++)
          if (CommonLinks[b][0] != 0) {
            cout << "COMM="<< setw(4) << b << "|";
            for (unsigned int d = 1; d <= maxLinks; d++)
              cout << setw(6) << CommonLinks[b][d] << "|";
            cout << endl;
          }
        cout << "---------------------------------------------------------------------------\n" << endl;
    }


    cout << "------------------ LMC BASED ROUTING :COMMON NODES HISTOGRAM -----------------" << endl;

    cout << "Describes the distribution of the number of common nodes between the " << endl;
    cout << "different LMC paths of all the CA to CA paths.\n" << endl;
    cout << "COMMON-NODES NUM-CA-CA-PAIRS" << endl;
    cout << "PATH DPT|";
    for (unsigned int d = 1; d <= maxDepth; d++)
      cout <<  setw(6) << d << "|";
    cout << endl;

    for (unsigned int b = 0; b <= maxHops ; b++)
      if (CommonNodes[b][0] != 0) {
        cout << "COMM="<< setw(3) << b << "|";
        for (unsigned int d = 1; d <= maxDepth; d++)
          cout << setw(6) << CommonNodes[b][d] << "|";
        cout << endl;
      }
    cout << "---------------------------------------------------------------------------\n" << endl;

    cout << "---------------- LMC BASED ROTING :COMMON SYSTEMS HISTOGRAM ---------------" << endl;
    cout << "The distribution of the number of common systems between the " << endl;
    cout << "different LMC paths of all the CA to CA paths.\n" << endl;
    cout << "COMMON-SYSTEM NUM-CA-CA-PAIRS" << endl;
    cout << "PATH DPT|";
    for (unsigned int d = 1; d <= maxDepth; d++)
      cout <<  setw(6) << d << "|";
    cout << endl;
    for (unsigned int b = 0; b <= maxHops ; b++)
      if (CommonSystems[b][0] != 0) {
        cout << "COMM=" << setw(3) << b << "|";
        for (unsigned int d = 1; d <= maxDepth; d++)
          cout << setw(6) << CommonSystems[b][d] << "|";
        cout << endl;
      }
    cout << "---------------------------------------------------------------------------\n" << endl;
  }

  string outDirStr;
  if (outDir == NULL)
      outDirStr = string(SEPERATOR) + "tmp" + SEPERATOR;
  else
      outDirStr = string(outDir) + SEPERATOR;

  string fileName;
  int rc = 0;

  string err_message;
  // report the link over subscription histogram and dump out the
  // num paths per switch out port
  fileName = outDirStr + "ibdmchk.sw_out_port_num_paths";
  ofstream linkUsage;
  rc = IBFabric::OpenFile(fileName.c_str(), linkUsage, false, err_message);
  if (rc)
      cout << err_message << endl;
  else
      linkUsage << "# NUM-PATHS PORT-NAME " << endl;

  vec_int linkSubscriptionHist(maxLinkSubscriptions + 1,0);
  for (map_pnode_vec_int::iterator nI = switchPathsPerOutPort.begin();
       nI !=  switchPathsPerOutPort.end();
       nI++) {
    IBNode *p_node = (*nI).first;
    for (unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
      IBPort *p_port = p_node->getPort((phys_port_t)pn);
      if (p_port && p_port->p_remotePort &&
          (p_port->p_remotePort->p_node->type == IB_SW_NODE)) {
        if (!rc) linkUsage << setw(6) << ((*nI).second)[pn]  << " " << p_port->getName() << endl;
        linkSubscriptionHist[((*nI).second)[pn]]++;
      }
    }
  }
  if (!rc) linkUsage.close();

  cout << "---------- LFT CA to CA : SWITCH OUT PORT - NUM PATHS HISTOGRAM -----------" << endl;
  cout << "Number of actual paths going through each switch out port considering" << endl;
  cout << "all the CA to CA paths. Ports driving CAs are ignored (as they must" << endl;
  cout << "have = Nca - 1). If the fabric is routed correctly the histogram" << endl;
  cout << "should be narrow for all ports on same level of the tree." << endl;
  cout << "A detailed report is provided in /tmp/ibdmchk.sw_out_port_num_paths.\n" << endl;
  cout << "NUM-PATHS NUM-SWITCH-PORTS" << endl;
  for (int b = 0; b <= maxLinkSubscriptions; b++)
    if (linkSubscriptionHist[b])
      cout << setw(8) << b << "   " << linkSubscriptionHist[b] << endl;
  cout << "---------------------------------------------------------------------------\n" << endl;

  // now do the DLID per out port:
  fileName = outDirStr + "ibdmchk.sw_out_port_num_dlids";
  ofstream portDlidsUsage; //(outDirStr.c_str());
  rc = IBFabric::OpenFile(fileName.c_str(), portDlidsUsage, false, err_message);
  if (rc)
      cout << err_message << endl;
  else
       portDlidsUsage << "# NUM-DLIDS PORT-NAME " << endl;
  vec_int dlidsSubscriptionHist(maxDlidPerOutPort + 1,0);
  for (map_pnode_vec_int::iterator nI = switchDLidsPerOutPort.begin();
       nI != switchDLidsPerOutPort.end();
       nI++) {
    IBNode *pNode = (*nI).first;
    for (unsigned int pn = 1; pn <= pNode->numPorts; pn++) {
      IBPort *p_port = pNode->getPort((phys_port_t)pn);
      if (p_port && p_port->p_remotePort &&
          (p_port->p_remotePort->p_node->type == IB_SW_NODE)) {
        if (!rc) portDlidsUsage << setw(6) << ((*nI).second)[pn]  << " " << p_port->getName() << endl;
        dlidsSubscriptionHist[((*nI).second)[pn]]++;
      }
    }
  }
  if (!rc) portDlidsUsage.close();

  cout << "---------- LFT CA to CA : SWITCH OUT PORT - NUM DLIDS HISTOGRAM -----------" << endl;
  cout << "Number of actual Destination LIDs going through each switch out port considering" << endl;
  cout << "all the CA to CA paths. Ports driving CAs are ignored (as they must" << endl;
  cout << "have = Nca - 1). If the fabric is routed correctly the histogram" << endl;
  cout << "should be narrow for all ports on same level of the tree." << endl;
  cout << "A detailed report is provided in /tmp/ibdmchk.sw_out_port_num_dlids.\n" << endl;
  cout << "NUM-DLIDS NUM-SWITCH-PORTS" << endl;
  for (int b = 0; b <= maxDlidPerOutPort; b++)
    if (dlidsSubscriptionHist[b])
      cout << setw(8) << b << "   " << dlidsSubscriptionHist[b] << endl;
  cout << "---------------------------------------------------------------------------\n" << endl;

  if (anyError)
    cout << "-E- Found " << anyError << " missing paths"
         << " out of:" << paths << " paths" << endl;
  else
    cout << "-I- Scanned:" << paths << " CA to CA paths " << endl;

  cout << "---------------------------------------------------------------------------\n" << endl;
  return anyError;
}

///////////////////////////////////////////////////////////////////////////////
int
SubnMgtVerifyAllRoutes(IBFabric *p_fabric)
{
    int anyError = 0, paths = 0;

    cout << "-I- Verifying all paths ... " << endl;
    unsigned int hops, maxHops = 0;
    //list_pnode path;

    // go over all ports in the fabric
    for (lid_t sLid = p_fabric->minLid; sLid <= p_fabric->maxLid; sLid++) {
      IBPort *p_srcPort = p_fabric->getPortByLid(sLid);

      if (!p_srcPort || (p_srcPort->base_lid != sLid))
	continue;

      // go over all the rest of the ports:
      for (lid_t dLid = p_fabric->minLid; dLid <= p_fabric->maxLid; dLid++) {
	IBPort *p_dstPort = p_fabric->getPortByLid(dLid);

	if (!p_dstPort || p_dstPort->base_lid != dLid)
	  continue;

	// Avoid tracing to itself
	if (dLid == sLid)
	  continue;

	bool anyPathToDest = false;
	for (lid_t l = 0; l < ((lid_t)1<<p_dstPort->lmc); l++) {
	  paths++;

	  // now go and verify the path:
	  if (!TraceRouteByLFT(p_fabric, sLid, (lid_t)(dLid + l), &hops,
                           NULL, NULL, false)) {
	    if (hops > maxHops)
	      maxHops = hops;
	    anyPathToDest = true;
	  }
	  //path.clear();
	}
	if (!anyPathToDest) {
	  cout << "-E- Fail to find a path from:"
	       << p_srcPort->p_node->name << "/"
           << (unsigned int)p_srcPort->num
	       << " to:" << p_dstPort->p_node->name << "/"
           << (unsigned int)p_dstPort->num
	       << endl;
	  anyError++;
	}
    //isPLFTEnabled check VL15 connectivity
    if (p_fabric->isPLFTEnabled()) {
        anyPathToDest = false;
        lid_t pathsPerDlid = (lid_t)(1 << p_dstPort->lmc);
        for (lid_t l = 0; l < pathsPerDlid; l++) {
            if(!TraceRouteByLFT(p_fabric, sLid, lid_t(dLid + l),
                                             NULL, NULL, NULL, true))
                anyPathToDest = true;
        }
        if (!anyPathToDest) {
          cout << "-E- Fail to find a VL15 path from:"
               << p_srcPort->p_node->name << "/"
               << (unsigned int)p_srcPort->num
               << " to:" << p_dstPort->p_node->name << "/"
               << (unsigned int)p_dstPort->num
               << endl;
          anyError++;
        }
    }

      }
    }

    if (anyError)
        cout << "-E- Found " << anyError << " missing paths"
               << " out of:" << paths << " paths" << endl;
    else
        cout << "-I- Scanned:" << paths << " paths " << endl;

    cout << "---------------------------------------------------------------------------\n" << endl;
    return anyError;
}


///////////////////////////////////////////////////////////////////////////////
// Analyze the fabric to see if it we can recognize the root nodes
// Return the list of root nodes found.
// We limit our auto root recognition too symmetrical fat trees only.
// in this case every switch should either connect to level N-1 or N+1.
// We use this and check for it during the BFS.
typedef list <IBNode *> list_pnode;

list_pnode
SubnMgtFindTreeRootNodes(IBFabric *p_fabric)
{
    list_pnode nextNodes;
    list_pnode curNodes;
    list_pnode rootNodes;
    list_pnode emptyRes;
    IBNode *p_node;
    int rank = 0;
    map_pnode_rank nodeRankMap;

    cout << "-I- Automatically recognizing the tree root nodes ..."
            <<  endl;

    // find all non switch nodes and add them to the current step nodes
    for( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
            nI != p_fabric->NodeByName.end();
            nI++) {
        p_node = (*nI).second;
        if (p_node->type == IB_SW_NODE)
            continue;

        curNodes.push_back(p_node);
    }

    // BFS:
    while (! curNodes.empty()) {
        rank++;
        nextNodes.clear();

        // the last group should be our roots nodes.
        rootNodes = curNodes;

        if (0) {
            cout << "-V- Level:" << rank << " nodes:" << endl;
            int nInLine = 1;
            for (list_pnode::iterator nI = curNodes.begin();
                    nI != curNodes.end();
                    nI++) {
                if (nInLine == 16) {
                    nInLine = 0;
                    cout << endl;
                }
                nInLine++;
                cout << (*nI)->name << " ";
            }
        }

        // go over all this step ports
        while (! curNodes.empty()) {
            p_node = curNodes.front();
            curNodes.pop_front();

            // go over all ports of the node
            for (unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
                IBPort *p_port = p_node->getPort((phys_port_t)pn);

                // if there is a connection:
                if (p_port && p_port->p_remotePort) {
                    IBNode *p_remNode = p_port->p_remotePort->p_node;

                    // we ignore non SW nodes:
                    if (p_remNode->type != IB_SW_NODE)
                        continue;

                    // if already marked
                    map_pnode_rank::iterator nI = nodeRankMap.find(p_remNode);
                    if (nI != nodeRankMap.end()) {
                        int remNodeRank = (*nI).second;

                        // as we use the rank for signing visited status
                        // we might have the remote node be marked
                        // either with rank - 1 or rank + 1
                        if ((remNodeRank != rank + 1) && (remNodeRank != rank - 1)) {
                            cout << "-E- Given topology is not a pure levelized tree:" << endl;
                            cout << "    Node:" << p_remNode->name << " rank:" << remNodeRank
                                    << " accessed from node:" << p_node->name
                                    << " rank:" << rank << endl;
                            return emptyRes;
                        }
                    } else {
                        // first we mark the node as rank + 1
                        nodeRankMap[p_remNode] = (rank_t)(rank + 1);

                        // now we push it to the next level list:
                        nextNodes.push_back(p_remNode);
                    }
                }
            }
        }
        curNodes = nextNodes;
    }
    return rootNodes;
}


///////////////////////////////////////////////////////////////////////////////
// This routine is based on the min hop tables and
// the fact the statistics of the min hops changes for the
// root nodes:
typedef vector< int > vec_int;

list_pnode
SubnMgtFindRootNodesByMinHop(IBFabric *p_fabric)
{
    list_pnode rootNodes;
    unsigned int minHop;
    int numCas = 0;

    // go over all switch nodes and print the statistics:
    cout << "-I- Automatically recognizing the tree root nodes ..." <<  endl;

    for( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
            nI != p_fabric->NodeByName.end();
            nI++) {
        IBNode *p_node = (*nI).second;
        if (p_node->type != IB_SW_NODE) numCas++;
    }

    // find all non switch nodes and add them to the current step nodes
    for( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
            nI != p_fabric->NodeByName.end();
            nI++) {
        IBNode *p_node = (*nI).second;
        if (p_node->type != IB_SW_NODE)
            continue;

        // the min hop table should exist on the node:
        unsigned int maxHops = 0;

        // go over all nodes that are CA and calc the histogram of min hops to cas
        vec_int swToCaMinHopsHist(50, 0);

        // go over all ports in the fabric
        for (lid_t bLid = p_fabric->minLid; bLid <= p_fabric->maxLid; bLid++ ) {
            IBPort *p_port = p_fabric->getPortByLid(bLid);

            if (!p_port || (p_port->p_node->type == IB_SW_NODE))
                continue;

	    if (bLid != p_port->base_lid)
	      continue;

            // get the min hops to this port:
            minHop = p_node->getHops(NULL, bLid);

            swToCaMinHopsHist[minHop]++;

            // track the max:
            if (minHop > maxHops) maxHops = minHop;
        } // all lids

        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {
            // print out the hist for now:
            cout << " CA MIN HOP HISTOGRAM:" <<  p_node->name;
            for (unsigned int b = 0; b <= maxHops ; b++)
                cout << " " << setw(4) << swToCaMinHopsHist[b] ;
            cout << endl;
        }

        // we recognize spines by requiring one bar to be above 90% of the
        // number of CAs
        int numHopBarsOverThd1 = 0;
        int numHopBarsOverThd2 = 0;
        double thd1 = numCas * 0.9;
        double thd2 = numCas * 0.05;
        for (unsigned int b = 0; b <= maxHops ; b++) {
            if (swToCaMinHopsHist[b] > thd1) numHopBarsOverThd1++;
            if (swToCaMinHopsHist[b] > thd2) numHopBarsOverThd2++;
        }
        if ((numHopBarsOverThd1 == 1) && (numHopBarsOverThd2 == 1))
            rootNodes.push_back(p_node);
    } // all switches
    return rootNodes;
}

///////////////////////////////////////////////////////////////////////////////
// Find any routes that exist in the FDB's from CA to CA and do not adhare to
// the up/down rules. Report any crossing of the path.
int
SubnReportNonUpDownCa2CaPaths(IBFabric *p_fabric, map_pnode_rank &nodesRank)
{
  list_pnode path;
  list_phys_ports portsList;

  int anyError = 0, paths = 0, numBadPaths = 0;
  unsigned int hops;
  string firstChangeMsg;

  cout << "-I- Tracing all CA to CA paths for Credit Loops potential ..." << endl;

  // go over all ports in the fabric
  for (lid_t sLid = p_fabric->minLid; sLid <= p_fabric->maxLid; sLid++) {
    IBPort *p_srcPort = p_fabric->getPortByLid(sLid);

    // avoid too many errors...
    if (numBadPaths > 100)
      break;

    if (!p_srcPort || (p_srcPort->p_node->type == IB_SW_NODE))
      continue;

    if (sLid != p_srcPort->base_lid)
      continue;

    // go over all the rest of the ports:
    for (lid_t dLid = p_fabric->minLid; dLid <= p_fabric->maxLid; dLid++) {
      IBPort *p_dstPort = p_fabric->getPortByLid(dLid);
      if (!p_dstPort || (p_dstPort->p_node->type == IB_SW_NODE))
        continue;
      if (dLid != p_dstPort->base_lid)
        continue;

      // avoid too many errors...
      if (numBadPaths > 100)
        break;

      // Avoid tracing to itself
      if (dLid == sLid)
        continue;

      paths++;
      // go over all possible dLids
      bool anyPathToDest = false;
      for (lid_t pb = 0; pb < (lid_t)(1 << p_dstPort->lmc); pb++) {
        // now go and verify the path:
        if (!TraceRouteByLFT(p_fabric, sLid, (lid_t)(dLid+pb),
                                              &hops, &path, &portsList, false)) {
          int numChanges = 0;
          int prevGoingUp = 1, goingUp;
          int prevRank = 99, rank;
          IBNode *p_prevNode = NULL;
          anyPathToDest = true;

	  // Go through the path and check for up down transitions.
	  for (list_pnode::iterator lI = path.begin(); lI!= path.end(); lI++) {
	    IBNode *p_node = *lI;

	    // lookup the rank of the current node:
	    map_pnode_rank::iterator rI = nodesRank.find(p_node);
	    if (rI == nodesRank.end()) {
	      cout << "-E- Somehow we do not have rank for:" << p_node->name
		   << endl;
	      exit(1);
	    }

	    rank = (*rI).second;

	    // we are going up if
	    if (rank < prevRank)
	      goingUp = 1;
	    else
	      goingUp = 0;

	    // now look for dir change.
	    if (prevGoingUp != goingUp) {
	      // we need to report if not first change.
	      if (numChanges) {
		// header only once:
		if (numChanges == 1) {
		  cout << "-E- Potential Credit Loop on Path from:"
		       << p_srcPort->p_node->name << "/" << p_srcPort->num
		       << " to:" << p_dstPort->p_node->name
		       << "/" << p_dstPort->num
		       << endl;
		  cout << firstChangeMsg << endl;
		  numBadPaths++;
		}
		// now the point
		if (goingUp) {
		  cout << "  Going:Up ";
		} else {
		  cout << "  Going:Down ";
		}
		cout << "from:" << p_prevNode->name
		     << " to:" << p_node->name << endl;
	      } else {
		// first change so:
		firstChangeMsg = string("  Going:Down from:") +
		  p_prevNode->name + string(" to:") + p_node->name;
	      }
	      numChanges++;
	    }
	    prevRank = rank;
	    prevGoingUp = goingUp;
	    p_prevNode = p_node;
	  }
	}
        path.clear();
        portsList.clear();

      }

      if (!anyPathToDest) {
        cout << "-E- Fail to find a path from:"
             << p_srcPort->p_node->name << "/"
             << (unsigned int)p_srcPort->num
             << " to:" << p_dstPort->p_node->name << "/"
             << (unsigned int)p_dstPort->num
             << endl;
        anyError++;
      }
      //isPLFTEnabled check VL15 connectivity
      if (p_fabric->isPLFTEnabled()) {
          anyPathToDest = false;
          lid_t pathsPerDlid = (lid_t)(1 << p_dstPort->lmc);
          for (lid_t l = 0; l < pathsPerDlid; l++) {
              if(!TraceRouteByLFT(p_fabric, sLid, lid_t(dLid + l),
                                               NULL, NULL, NULL, true))
                  anyPathToDest = true;
          }
          if (!anyPathToDest) {
            cout << "-E- Fail to find a VL15 path from:"
                 << p_srcPort->p_node->name << "/"
                 << (unsigned int)p_srcPort->num
                 << " to:" << p_dstPort->p_node->name << "/"
                 << (unsigned int)p_dstPort->num
                 << endl;
            anyError++;
          }
      }

    }
  }

  if (numBadPaths) {
    if (numBadPaths > 100)
      cout << "-W- Stopped checking for loops after 100 errors" << endl;
    cout << "-E- Found:" << numBadPaths
         << " CA to CA paths that can cause credit loops." << endl;
  } else
    cout << "-I- No credit loops found in:" << paths
         << " CA to CA paths" << endl;
  cout << "---------------------------------------------------------------------------\n" << endl;
  return(numBadPaths + anyError);
}


///////////////////////////////////////////////////////////////////////////////
int
SubnMgtUpDnIsLegelStep(IBNode *curNode, IBNode *nxtNode)
{
    int curDir,nxtDir;
    //cout << "-V- B4 __IsLegelStep check" <<endl;
    curDir = (int) curNode->appData2.val;
    nxtDir = (int) nxtNode->appData2.val;
    //  cout << "-V- In  __IsLegelStep check , curDir="<<curDir<<" nxtDir="<<nxtDir<<"\n" << endl;
    if (curDir == 1 && nxtDir == 0)
        return 1;
    else
        return 0;
}


///////////////////////////////////////////////////////////////////////////////
// Do BFS and set Min Hops for the current port , based on previous defined ranking
int
SubnMgtUpDnBFSFromPort(lid_t lid,
        IBFabric *p_fabric ,
        map_pnode_rank &nodesRank )
{
    list_pnode      CurState , NextState ;
    IBPort          *tmpPort,*self;
    IBNode          *self_node;
    tmpPort = p_fabric->getPortByLid(lid);
    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {
        cout << "-V- BFS for lid="<<lid<<", type of port= " << tmpPort->p_node->type << endl;
        tmpPort->p_node->repHopTable();
    }
    if (tmpPort->p_node->type == IB_SW_NODE) {
        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            cout << "-V- This is a switch ..." << endl;
        self = tmpPort;
        self_node = tmpPort->p_node;

        // Update in MinHop Table port 0 with 1 hop
        //tmpPort = self_node->getPort(0);
        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            cout << "-V- tmpPort is : " << tmpPort << endl;
        if (self->base_lid == lid)
            self_node->setHops(tmpPort,lid,0);
        else
            self_node->setHops(tmpPort,lid,1);
        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            self_node->repHopTable();
    } else {
        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {
            cout << "-V- This is an HCA ..." << endl;
            cout << "-V- tmpPort :" << tmpPort << " remote Port : " << tmpPort->p_remotePort << endl;
        }
        self = tmpPort->p_remotePort;
        self_node = self->p_node;
        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            cout << "-V- After assignment of self " << self << " && self_node " << self_node << endl;
        // Check to see non switchable subnet
        if (self_node->type != IB_SW_NODE) {
            cout << "-W- This is a non switch subnet could not perform algorithm" << endl;
            return 1;
        }
        // Assign current lid (HCA) to the remote node
        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            self_node->repHopTable();
        // If we BFS through base_lid the same lid we can reach it by 0 hops
        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            cout << "-V- base lid : " <<self->base_lid << " BFS lid : " << lid << endl;
        self_node->setHops(self,lid,1);
        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            self_node->repHopTable();
    }

    // We use the current node appData2.val to store direction
    self->p_node->appData2.val = 0UL;
    // Push into list the Current Item
    CurState.push_back(self->p_node);
    // Iterate over all nodes in CurState list till empty
    //  cout << "-V- b4 while .." << endl;
    while (CurState.size()) {
        NextState.clear();
        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            cout << "-V- Iterating in while : size of CurState is : "
                  << CurState.size() << " size of NextState is : " << NextState.size() <<  endl;
        // go over cur step nodes
        for (list_pnode::iterator lI = CurState.begin();
                lI != CurState.end(); lI++) {
            IBNode *p_node = *lI;

            if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
                cout << "-V- Current switch handeled is : " << p_node->name << endl;

            // go over all ports to find unvisited remote nodes
            for (unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
                IBPort *p_port = p_node->getPort((phys_port_t)pn);

                // Only if current port is NULL or not connected skip it
                if (! p_port || ! p_port->p_remotePort)
                    continue;

                if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
                    cout << "-V- Handling port num : " << pn << " of node :" << p_node->name
                        << " p_type="<<p_node->type<<"\n"<<endl;

                IBPort *p_rem_port = p_node->getPort((phys_port_t)pn)->p_remotePort;
                IBNode *p_rem_node = p_rem_port->p_node;
                uint8_t minHop = 0;

                // Only if it is a switch then update its table
                if (p_rem_node->type == IB_SW_NODE) {
                    // First put UP / DOWN label to remote node
                    if (nodesRank[p_node] < nodesRank[p_rem_node])
                        // This is DOWN
                        p_rem_node->appData2.val = 1UL;
                    else
                        // This is UP
                        p_rem_node->appData2.val = 0UL;
                    // cout << "-V- b4 Validity check of UP/DOWN"<<endl;
                    // Check validity of direction
                    if (SubnMgtUpDnIsLegelStep(p_node,p_rem_node)) continue;
                    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
                        cout << "-V- Current step is legal ..." << endl;

                    // Update remote port MinHopTable according to Current MinHopTable
                    // Set minHop with the minimum hop count for this lid through current node
                    minHop = p_node->getHops(NULL,lid);

                    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {
                        cout << "-V- lid=" <<lid << " p_node="<<p_node<<"\n"<<endl;
                        cout << "-V- MinHopTable of current node at lid=" <<lid
                                << " port=0 is : " << minHop<<endl;
                        //cout << "-V- After minHop :"<<minHop<<"\n" <<endl;
                        cout << "-V- B4 Check of current_node(minHop)=" << minHop
                                << " and p_rem_node(minHop)="
                                << p_rem_node->getHops(p_rem_port, lid) << "\n"<<endl;
                    }

                    // we will use the remote node min hops to know if we
                    // already have visited this node
                    int remNodeHops = p_rem_node->getHops(NULL,lid);

                    // Check min hop count if better insert into NextState
                    // list && update the remote node Min Hop Table
                    if (minHop + 1 <= p_rem_node->getHops(p_rem_port, lid)) {
                        // Update MinHopTable of remote node and add it to NextState
                        p_rem_node->setHops(p_rem_port,lid,(uint8_t)(minHop + 1));

                        // only if the remote node did not have this hop we
                        // need to visit it next step
                        if (remNodeHops > minHop + 1) NextState.push_back(p_rem_node);

                        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {
                            cout << "-V- Updating MinHopTable of node : " << p_rem_node->name
                                    << " for lid :" <<  lid  << " Value is : " << minHop << "\n" << endl;
                            // Lets see the last update for the current node
                            p_rem_node->repHopTable();
                        }
                    }
                }
            }
        }
        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            cout << "-V- Start another iteration for NextState with : "
                  << NextState.size() << " elements" << endl;
        CurState = NextState;
    }

    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
        cout << "-V- End of Iterations ..." << endl;
    return(0);
}


///////////////////////////////////////////////////////////////////////////////
// Calc Min Hop Table using UP-DOWN algorithm
int
SubnMgtCalcUpDnMinHopTbls(IBFabric *p_fabric,
        map_pnode_rank &nodesRank)
{
    // go over all the lids and init their Min Hop Tables
    for (lid_t i = 1; i <= p_fabric->maxLid; i++) {
        IBPort *p_cur_port = p_fabric->getPortByLid(i);
        if (p_cur_port == NULL)
            continue;

        IBNode *p_cur_node = p_cur_port->p_node;

        if (p_cur_node->type == IB_SW_NODE) {
            // assign all ports of the current node/lid initial value
            p_cur_node->setHops(NULL,0,IB_HOP_UNASSIGNED);
        }
    }

    // Now do the UP-Down Min Hop propagation
    for (lid_t i = 1; i <= p_fabric->maxLid; i++)
      if (SubnMgtUpDnBFSFromPort(i,p_fabric,nodesRank))
	return (1);

    // dump the min hops
    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
        for( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
                nI != p_fabric->NodeByName.end();
                nI++) {
            IBNode *p_node = (*nI).second;
            if (p_node->type != IB_SW_NODE)
                continue;
            p_node->repHopTable();
        }
    return(0);
}


///////////////////////////////////////////////////////////////////////////////
// do up down given a node name regular expression to find the
// root nodes
int
SubnMgtCalcUpDnMinHopTblsByRootNodesRex(IBFabric *p_fabric,
                                        const char * rootNodesNameRex)
{
    map_pnode_rank nodesRank;

    // rank the fabric
    SubnRankFabricNodesByRegexp(p_fabric, rootNodesNameRex, nodesRank);

    // do the actual route
    SubnMgtCalcUpDnMinHopTbls( p_fabric, nodesRank );
    return(0);
}


///////////////////////////////////////////////////////////////////////////////
//
// Multicast Check Code
//
struct bfsEntry {
    IBNode *pNode;
    uint8_t inPort;
};

// return true if remote side is a switch and it points back to the given port
// by its MFT
bool isRemSwPortPointingBackByMFT( IBPort *pPort, uint16_t mlid)
{
    if ( !pPort || !pPort->p_remotePort ) return(false);
    IBPort *pRemPort = pPort->p_remotePort;
    IBNode *pRemNode = pRemPort->p_node;
    if ( pRemNode->type != IB_SW_NODE ) return(false);

    list_phys_ports mftPNs = pRemNode->getMFTPortsForMLid(mlid);
    return(find(mftPNs.begin(), mftPNs.end(), pRemPort->num) != mftPNs.end());
}

// look through all the remote switch of the given port if
// it has any port pointing back to the original port node
IBPort * getAnyPortPointingBackByMFT(IBPort *pPort, uint16_t mlid)
{
    if ( !pPort || !pPort->p_remotePort ) return(NULL);
    IBPort *pRemPort = pPort->p_remotePort;
    IBNode *pRemNode = pRemPort->p_node;
    if ( pRemNode->type != IB_SW_NODE ) return(NULL);

    list_phys_ports mftPNs = pRemNode->getMFTPortsForMLid(mlid);
    for ( list_phys_ports::const_iterator pnI = mftPNs.begin();
        pnI != mftPNs.end(); pnI++ ) {
        IBPort *pAnyRemPort = pRemPort->p_node->getPort(*pnI);
        if ( pAnyRemPort &&
             pAnyRemPort->p_remotePort &&
             pAnyRemPort->p_remotePort->p_node == pPort->p_node ) {
            return(pAnyRemPort);
        }
    }
    return(NULL);
}

// DFS Full Member MFTs
int dfsSendRecvMemMCG(IBNode *sw, IBPort *inPort, uint16_t mlid,
                  map<IBNode*,IBNode*> &swGroup,
                  bool onlyBiDir, int level)
{
    int anyErr = 0;
    char mlidStr[8];
    sprintf(mlidStr, "0x%04X", mlid);

    if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE ) {
        if ( inPort ) {
            cout << "-V- DFS FullMembers Visiting node:" << sw->name
            << " through port:" << (unsigned int)inPort->num
            << " for mlid:" << mlidStr << " level:" << level << endl;
        } else {
            cout << "-V- DFS FullMembers starting node:" << sw->name
            << " for mlid:" << mlidStr << endl;
        }
    }

    // get the list of MC group ports for this mlid of this node
    list_phys_ports portNums = sw->getMFTPortsForMLid(mlid);

    // go over all MC output ports of the current node ignoring the input
    list_phys_ports::iterator pnI;
    for ( pnI = portNums.begin(); pnI != portNums.end(); pnI++ ) {
        phys_port_t pn = (*pnI);

        IBPort *pPort = sw->getPort(pn);
        if ( ! pPort || ! pPort->p_remotePort ) continue;

        // ignore the port we got here through.
        if ( pPort == inPort ) continue;

        // get the remote node
        IBNode *pRemNode = pPort->p_remotePort->p_node;

        // we ignore remote HCAs
        if ( pRemNode->type != IB_SW_NODE ) continue;

        bool allowToProceed = true;

        // no need to go here if already visited
        map<IBNode*,IBNode*>::iterator vI = swGroup.find(pRemNode);
        if ( vI != swGroup.end() ) {
            // we only report loops on cases of !onlyBiDir
            if ( !onlyBiDir && ((*vI).second == swGroup[sw]) ) {
                cout << "-E- Found a loop on MLID:" << mlidStr
                << " got to node:" << pRemNode->name
                << " again through port:" << (unsigned int)pPort->p_remotePort->num
                << " connected to:" << sw->name << " port:"
                << (unsigned int)pn << endl;
                anyErr++;
            }
            allowToProceed = false;
        }

        // criteria for visiting is based on onlyBiDir
        if ( onlyBiDir ) {
            // if the remote node does not point back to this one
            // (i.e. the port bit is not set in the MFT do not go through ...
            if ( !isRemSwPortPointingBackByMFT(pPort, mlid) ) {
                cout << "-W- Found a non-bidir MC route on FullMember tree of MLID:"
                << mlidStr
                << "\n    From port:" << pPort->getName()
                << " to port:" << pPort->p_remotePort->getName()
                << " which does not point back." << endl;
                anyErr++;
                // do not go there
                allowToProceed = false;
            }
        }

        if ( allowToProceed ) {
            swGroup[pRemNode] = swGroup[sw];
            anyErr += dfsSendRecvMemMCG(pRemNode, pPort->p_remotePort, mlid,
                                    swGroup, onlyBiDir, level+1);
        }
    } // all ports
    return anyErr;
}

// DFS Sender Only Tree
int dfsSenderOnlyMCG(IBNode *sw, IBPort *inPort, uint16_t mlid,
                     map<IBNode*,IBNode*> &soGroup,
                     map<IBNode*, set<IBNode*> > &swPassingSenders,
                     map<IBNode*,IBNode*> &swGroup,
                     int level)
{
    int anyErr = 0;
    char mlidStr[8];
    sprintf(mlidStr, "0x%04X", mlid);

    if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE ) {
        if ( inPort ) {
            cout << "-V- DFS SenderOnlyMembers Visiting node:" << sw->name
            << " through port:" << (unsigned int)inPort->num
            << " for mlid:" << mlidStr << " level:" << level << endl;
        } else {
            cout << "-V- DFS SenderOnlyMembers starting node:" << sw->name
            << " for mlid:" << mlidStr << endl;
        }
    }

    // get the list of MC group ports for this mlid of this node
    list_phys_ports portNums = sw->getMFTPortsForMLid(mlid);
    if (!portNums.size()) {
        cout << "-E- Empty MFT. Got to dead end from host :" << sw->name
            << " for mlid:" << mlidStr << endl;
        anyErr++;
    } else if (portNums.size() > 1)
        cout << "-W- Sender passing node " << sw->name
             << " has more than one MFT forwarding port"
             << " for mlid:" << mlidStr << endl;

    // go over all MC output ports of the current node
    list_phys_ports::iterator pnI;
    for ( pnI = portNums.begin(); pnI != portNums.end(); pnI++ ) {
        phys_port_t pn = (*pnI);
        IBPort *pPort = sw->getPort(pn);
        if ( ! pPort || ! pPort->p_remotePort ) continue;

        // get the remote node
        IBNode *pRemNode = pPort->p_remotePort->p_node;

        //check for bi-direction link to inPort from MFT port
        if (pPort == inPort) {
            cout << "-E- Bi-Direction link for sender only from port: "
                 << pPort->getName() << "to port " << pRemNode->name << endl;
            anyErr++;
            continue;
        }

        // we ignore remote HCAs
        if ( pRemNode->type != IB_SW_NODE ) {
            cout << "-E- forward multicast package from sw "
                 << sw->name << " to non member HCA "
                 << pRemNode->name << " mlid: "
                 << mlidStr << endl;
            continue;
        }

        bool allowToProceed = true;

        // stop condition, if this remote node is a full member
        map<IBNode*,IBNode*>::iterator vI = swGroup.find(pRemNode);
        if ( vI != swGroup.end() ) {
            //we reached the first full member in dfs
            //insert the full member root the list of full members connecting
            //to this dfs branch.
            if (swPassingSenders[sw].find(swGroup[pRemNode]) !=
                swPassingSenders[sw].end()) {
                cout << "-E- duplicate multicast package from switch " << sw->name
                     << " in port " << (unsigned int)pn
                     <<" to mlid: " << mlidStr << endl;
                anyErr++;
            } else
                swPassingSenders[sw].insert(swGroup[pRemNode]);
            allowToProceed = false;
        } else {
            map<IBNode*,IBNode*>::iterator vI = soGroup.find(pRemNode);
            if (vI != soGroup.end()) {
                // if the remote switch has the same root as the sw we have a loop
                if (soGroup[sw] == soGroup[pRemNode]) {
                    cout << "-E- Found a loop on MLID:" << mlidStr
                        << " got to node:" << pRemNode->name
                        << " again through port:" << (unsigned int)pPort->p_remotePort->num
                        << " connected to:" << sw->name << " port:"
                        << (unsigned int)pn << endl;
                    anyErr++;
                } else {
                    //checking duplicated links to the FullMember
                    size_t current_size = swPassingSenders[sw].size();
                    size_t remote_size = swPassingSenders[pRemNode].size();
                    swPassingSenders[sw].insert(swPassingSenders[pRemNode].begin(),
                            swPassingSenders[pRemNode].end());
                    if ((current_size + remote_size) > swPassingSenders[sw].size()) {
                        cout << "-E- duplicate multicast package from host " << sw->name
                            << " in port " << (unsigned int)pn
                            <<" to mlid: " << mlidStr << endl;
                        anyErr++;
                    }
                }
                allowToProceed = false;
            }
        }

        if ( allowToProceed ) {
            soGroup[pRemNode] = soGroup[sw];
            anyErr += dfsSenderOnlyMCG(pRemNode, pPort->p_remotePort, mlid,
                                    soGroup, swPassingSenders, swGroup, level+1);

            //checking duplicated links to the FullMember
            size_t current_size = swPassingSenders[sw].size();
            size_t remote_size = swPassingSenders[pRemNode].size();
            swPassingSenders[sw].insert(swPassingSenders[pRemNode].begin(),
                                        swPassingSenders[pRemNode].end());
            if ((current_size + remote_size) > swPassingSenders[sw].size()) {
                cout << "-E- duplicate link from host " << sw->name
                     << " to host " << pRemNode->name
                     <<" on multicast group: " << mlidStr << endl;
                anyErr++;
            }
        }
    } // all ports
    return anyErr;
}

// Check multicast group given a list of Full Members Ports and a list
// of sender only ports
int
SubnMgtCheckMCGrpByMemPortLists(IBFabric *p_fabric,
                                uint16_t mlid,
                                list<IBPort *> &sendRecvMemPorts,
                                list<IBPort *> &sendOnlyPorts)
{
    char mlidStr[8];
    bool onlyBiDir = true;
    sprintf(mlidStr, "0x%04X", mlid);

    // the algorithm searches for loops, connection groups for both full and
    // send only semantics. Also report switches with garbadge entries

    // to overcome the missing port 0 in ibdm if listed port is of a switch it
    // means port 0

    //          | GROUP CAN BE | GROUP CAN'T BE
    //  ---------------------------------------
    //   SENDER | SENDER-ONLY- | SENDER-ONLY
    //          | FULL-MEMBER  |
    //          | (BIT 1000)   | (BIT 0100)
    //  ---------------------------------------
    // SEND/RECV| FULL-MEMBER  | NON-MEMBER
    //          | (BIT 0001)   | (BIT 0010)
    //
    // Data model:
    // VISITED[node] marked node visited in order to find loops
    // swGroup[node] = full member sw root.
    // soGroup[sw] = senderOnly sw root, from which we started path to the full member tree
    // swPassingSenders[sw] = list of, root full mems, reached for this sender (which comes from swGroup[first SW We found that is in the fullMember tree])
    //
    // SenderOnly Example:
    // (HCA1-SenderOnly) (HCA2-SenderOnly)
    //     |                   |
    //  [ SW-A ]            [ SW-B ]
    //      \               /
    //    [ SW-C ]    [ SW-D ]
    //        \         /
    //       [ SW-E - FullMemeberTreeRoot ]
    //         |
    //        [ ... ]
    //
    // After running the senderOnly algorithem we'll get this:
    // soGroup[SW-A] = SW-A
    // soGroup[SW-C] = SW-A
    // soGroup[SW-B] = SW-B
    // soGroup[SW-D] = SW-B
    //
    // swPassingSenders[SW-A] = < E >
    // swPassingSenders[SW-C] = < E >
    // swPassingSenders[SW-B] = < E >
    // swPassingSenders[SW-D] = < E >
    //
    // Algorithm:
    // NOTE: due to the requirement that full member connections
    //       will only utilize bi-dir links we need to search for
    //       loops with a simpler algorithm that is directional
    //       so loop detection is simpler and require a slight change
    //       in the dfs
    //
    // # Full Members
    // foreach full(send/recv) port
    //   if a CA port
    //     verify connected switch has entry to it or error
    //     get to the sw
    //
    //   if !swGroup[sw]
    //      swGroup[sw] = sw
    //      dfsSendRecvMemMCG(mlid, sw, swGroup, onlyBiDir) - traverse the switches tree
    //
    //   if !VISITED[sw]
    //       VISITED[sw] = sw
    //       dfsSendRecvMemMCG(mlid, sw, VISITED, !onlyBiDir)
    //
    //   if CA port swGroup[ca] = swGroup[sw]
    //
    // # Sender only
    // foreach sender port
    //   if a CA port
    //     get to the connected sw
    //
    //   swPassingSenders[port] = dfsSenderOnlyMCG(mlid, port, sw)
    //
    // # Summary
    // foreach switch in the fabric
    //  if has MFT for mlid
    //    if !swGroup[sw] or !SENDERS[sw]
    //      report as garbadge MFT
    //    if swGroup[sw] collect GROUP_MEMS[swGroup[sw]]
    // if GROUP_MEMS.size() > 1
    //   Report members in each group
    // foreach sender port
    //   if !SENDER_GROUPS[port]
    //     Error disconnected sender
    //   else
    //     verbose report groups GROUP_MEMS[GROUP[SENDER_GROUPS[port]]]

    // Data structures:
    map<IBNode*,IBNode*> swGroup;
    map<IBNode*,IBNode*> swVisited;
    map<IBNode*,IBNode*> soGroup;
    map<IBNode*, set<IBNode*> > swPassingSenders;
    map<IBPort*, set<IBNode*> > senderReachedGroups;
    map<IBNode*, set<IBNode*> > groupMembers;

    // Algorithm:
    int anyErr = 0;
    if (sendRecvMemPorts.size() == 1 && sendOnlyPorts.empty()) {
        list<IBPort *>::iterator mI = sendRecvMemPorts.begin();
        IBPort *pPort = *mI;
        cout << "-I- FullMember node:" << pPort->p_node->getName()
             << " is a single member for mlid " << mlidStr
             << endl;
        return(0);
    }

    // # Full Members
    for ( list<IBPort *>::iterator mI = sendRecvMemPorts.begin();
        mI != sendRecvMemPorts.end(); mI++ ) {
        IBPort *pPort = *mI;

        IBNode *startSw = NULL;
        if ( pPort->p_node->type != IB_SW_NODE ) {
            IBPort *pRemPort = pPort->p_remotePort;
            if ( !pRemPort ) {
                cout << "-E- FullMember port:" << pPort->getName()
                << " is not connected" << endl;
                anyErr++;
            } else {
                IBNode *pRemNode = pRemPort->p_node;

                // make sure there is a remote switch and it has MFT pointing here
                if ( !isRemSwPortPointingBackByMFT(pPort, mlid) ) {
                    cout << "-E- FullMember port:" << pPort->getName()
                    << " is not connected to MFT on attached switch:"
                    << pRemNode->name << endl;
                    anyErr++;
                }
                startSw = pRemNode;
            }
        } else {
            // starting from a switch
            startSw = pPort->p_node;
        }

        if ( !startSw ) continue;

        // checking for connection groups
        if ( swGroup.find(startSw) == swGroup.end() ) {
            swGroup[startSw] = startSw;
            anyErr += dfsSendRecvMemMCG(startSw, NULL, mlid, swGroup, onlyBiDir, 0);
        }

        if ( pPort->p_node->type != IB_SW_NODE ) {
            swGroup[pPort->p_node] = swGroup[startSw];
        }

        // checking for loops
        if ( swVisited.find(startSw) == swVisited.end() ) {
            swVisited[startSw] = startSw;
            anyErr += dfsSendRecvMemMCG(startSw, NULL, mlid, swVisited, !onlyBiDir, 0);
        }

    } // all full(send/recv) members

    // # Sender only
    // foreach sender port
    //   if a CA port
    //     get to the connected sw
    //
    //   SENDER_GROUPS[port] = dfsSenderOnlyMCG(mlid, port, sw)
    for ( list<IBPort *>::iterator mI = sendOnlyPorts.begin();
        mI != sendOnlyPorts.end(); mI++ ) {
        IBPort *pPort = *mI;

        IBNode *startSw = NULL;
        if ( pPort->p_node->type != IB_SW_NODE ) {
            IBPort *pRemPort = pPort->p_remotePort;
            if ( !pRemPort ) {
                cout << "-E- Send onlyMember port:" << pPort->getName()
                << " is not connected" << endl;
                anyErr++;
            } else {
                IBNode *pRemNode = pRemPort->p_node;
                if (isRemSwPortPointingBackByMFT(pPort, mlid) ) {
                    cout << "-E- SenderOnlyMember port:" << pPort->getName()
                        << " is connected Bi-Direct on attached switch:"
                        << pRemNode->name << endl;
                    anyErr++;
                }
                startSw = pRemNode;
            }
        } else {
            // starting from a switch
            startSw = pPort->p_node;
        }

        if ( !startSw ) continue;

        // if the sw is in the swgroup , meaning that sw is a fullMember
        // or part of the fullMember tree(send/recv non member)
        if (swGroup.find(startSw) != swGroup.end()) {
            // if the senderOnly port is also part of the fullMember tree
            // no need to dfs him.
            continue;
        }

        if (soGroup.find(startSw) == soGroup.end()) {
            soGroup[startSw] = startSw;
            dfsSenderOnlyMCG(startSw, NULL, mlid,
                             soGroup, swPassingSenders,
                             swGroup, 0);
            // checking for disconnect
            if (swPassingSenders[startSw].empty()) {
                cout << "-E- sender only node " << startSw->name
                    << " is disconnected from the multicast for mlid: "
                    << mlidStr << endl;
            }
        }

    } // all sendonly members

    // # Summary
    // foreach switch in the fabric
    //  if has MFT for mlid
    //    if !swGroup[sw] or !SENDERS[sw]
    //      report as garbadge MFT
    //    if swGroup[sw] collect GROUP_MEMS[swGroup[sw]]

    for ( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
        nI != p_fabric->NodeByName.end();
        nI++ ) {

        IBNode *p_node = (*nI).second;

        if ( swGroup.find(p_node) != swGroup.end() ) {
            groupMembers[swGroup[p_node]].insert(p_node);
        } else {
            //if node is a switch and not part of the senderOnly group
            if ( p_node->type == IB_SW_NODE &&
                 soGroup.find(p_node) == soGroup.end()) {

                // see if we have an MFT entry by the given lid:
                list_phys_ports portNums = p_node->getMFTPortsForMLid(mlid);
                if ( !portNums.empty() ) {
                    list_phys_ports::iterator fI =
                    find(portNums.begin(), portNums.end(), 0);
                    if ( portNums.end() != fI ) {
                        cout << "-E- Switch: " << p_node->name
                        << " is not connected to MLID:"
                        << mlidStr << endl;
                        anyErr++;
                    } else {
                        cout << "-W- Switch: " << p_node->name
                        << " has unconnected MFT entries for MLID:"
                        << mlidStr << endl;
                    }
                }
            }
        }
    }

    if ( groupMembers.size() > 1 ) {
        cout << "-E- Found FullMembers" << groupMembers.size()
        << " connection groups for MLID:"
        << mlidStr << endl;
        int numGrps = 0;
        for ( map<IBNode*, set<IBNode*> >::iterator gI = groupMembers.begin();
            gI != groupMembers.end(); gI++ ) {
            numGrps++;
            list <string> cas;
            list <string> sws;
            int numCAs = 0;
            int numSWs = 0;
            for ( set<IBNode*>::iterator mI = (*gI).second.begin();
                mI != (*gI).second.end(); mI++ ) {
                IBNode *pNode = *mI;
                if ( pNode->type == IB_SW_NODE ) {
                    numSWs++;
                    sws.push_back(pNode->name);
                } else {
                    numCAs++;
                    cas.push_back(pNode->name);
                }
            }
            cout << "    Group:" << numGrps << " has "<< numCAs << " CAs: "
            << compressNames(cas) << endl;
            cout << "    Group:" << numGrps << " has "<< numSWs << " SWs: "
            << compressNames(sws) << endl;
        }
        anyErr++;
    }

    return(0);
}

int
SubnMgtCheckMCGrp(IBFabric *p_fabric,
                  lid_t mlid,
                  class McastGroupInfo &mcast_info)
{
    int anyErr = 0;
    char mlidStr[8];
    list< IBNode *> groupSwitches;
    list< IBNode *> groupHCAs;
    list< IBPort *> groupSendRecvMemPorts;
    list< IBPort *> sendOnlyPorts;

    for (map_mcast_members::iterator mI = mcast_info.m_members.begin();
         mI != mcast_info.m_members.end();
         mI++) {
        class McastGroupMemberInfo &mcast_grp_mem = (*mI).second;
        IBPort * p_port = (*mI).first;

        if (mcast_grp_mem.is_sender_only)
            sendOnlyPorts.push_back(p_port);
        else {
            if (p_port->p_node->type == IB_SW_NODE) {
                //check that have port 0 in MFT (full memeber must have port 0
                list_phys_ports portNums = p_port->p_node->getMFTPortsForMLid(mlid);
                if ( !portNums.empty() ) {
                    list_phys_ports::iterator fI =
                        find(portNums.begin(), portNums.end(), 0);
                    if ( portNums.end() == fI ) {
                        cout << "-E- Node: " << p_port->p_node->name
                            << " is a full member and missing port 0 for MLID:"
                            << mlidStr << endl;
                        anyErr++;
                    }
                }
            }
            groupSendRecvMemPorts.push_back(p_port);
        }

        if (p_port->p_node->type == IB_SW_NODE)
            groupSwitches.push_back(p_port->p_node);
        else if (p_port->p_node->type == IB_CA_NODE)
            groupHCAs.push_back(p_port->p_node);
    }

    sprintf(mlidStr, "0x%04X", mlid);
    cout << "-I- Multicast Group:" << mlidStr << " has:" << groupSwitches.size()
    << " Switches and:" << groupHCAs.size() << " HCAs which includes: "<< groupSendRecvMemPorts.size()
    << " FullMember ports and:" << sendOnlyPorts.size() << " SenderOnly ports" << endl;

    if ( groupSwitches.empty() && groupHCAs.empty()) return(0);
    if (groupSendRecvMemPorts.empty()) return(0);
    anyErr = SubnMgtCheckMCGrpByMemPortLists(p_fabric, mlid,
                                             groupSendRecvMemPorts, sendOnlyPorts);
    return anyErr;
}

// Check a multicast group dealing only with full members
int
SubnMgtCheckMCGrp(IBFabric *p_fabric,
                  uint16_t mlid)
{
    list< IBNode *> groupSwitches;
    list< IBPort *> groupSendRecvMemPorts;
    list< IBPort *> sendOnlyPorts;
    int anyErr = 0;

    // find all switches that are part of this mcgrp.
    // fill in the groupSwitches, groupSwitchesWithHCAs and groupHCAs
    for ( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
        nI != p_fabric->NodeByName.end();
        nI++ ) {

        IBNode *p_node = (*nI).second;
        if ( p_node->type != IB_SW_NODE ) continue;

        // see if we have an MFT entry by the given lid:
        list_phys_ports portNums = p_node->getMFTPortsForMLid(mlid);

        if ( portNums.empty() ) continue;

        groupSwitches.push_back(p_node);

        // find all CAs connected to the group by following the links that
        // are marked in the MFT that connect to the group.
        for ( list_phys_ports::iterator lI = portNums.begin();
            lI != portNums.end();
            lI++ ) {
            IBPort *p_port = p_node->getPort(*lI);

            // handle member switch port 0
            if ( p_port && (*lI)==0 )
                groupSendRecvMemPorts.push_back(p_port);

            // we do not count disconnected ports
            if ( p_port && p_port->p_remotePort &&
                 (p_port->p_remotePort->p_node->type != IB_SW_NODE) ) {
                groupSendRecvMemPorts.push_back(p_port->p_remotePort);
            }
        }
    }

    char mlidStr[8];
    sprintf(mlidStr, "0x%04X", mlid);
    cout << "-I- Multicast Group:" << mlidStr << " has:" << groupSwitches.size()
    << " switches and:" << groupSendRecvMemPorts.size()
    << " FullMember ports" << endl;

    if ( ! groupSwitches.size() ) return(0);
    if ( ! groupSendRecvMemPorts.size() ) return(0);
    anyErr = SubnMgtCheckMCGrpByMemPortLists(p_fabric, mlid,
                                             groupSendRecvMemPorts, sendOnlyPorts);
    return anyErr;
}

///////////////////////////////////////////////////////////////////////////////
// Check all multicast groups :
// 1. all switches holding it are connected
// 2. No loops (i.e. a single BFS with no returns).
int
SubnMgtCheckFabricMCGrps(IBFabric *p_fabric)
{
    int anyErrs = 0;
    cout << "-I- Scanning all multicast groups for loops and connectivity..."
    << endl;
    if (p_fabric->McastGroups.empty()) {
        for ( set_uint16::iterator sI = p_fabric->mcGroups.begin();
                sI != p_fabric->mcGroups.end();
                sI++ )
            anyErrs += SubnMgtCheckMCGrp(p_fabric, *sI);
    } else {
        for (map_mcast_groups::iterator sI = p_fabric->McastGroups.begin();
                sI != p_fabric->McastGroups.end();
                sI++)
            anyErrs+= SubnMgtCheckMCGrp(p_fabric, (*sI).first, (*sI).second);
    }

    if ( anyErrs )
        cout << "-E- " << anyErrs << " multicast group checks failed" << endl;

    cout << "---------------------------------------------------------------------------\n" << endl;
    return anyErrs;
}


///////////////////////////////////////////////////////////////////////////////
struct upDnBfsEntry {
   IBNode *pNode;
   IBNode *pTurnNode;
   uint8_t inPort;
   int     down;
};

// Given a starting point switch for multicast traversal and an mlid
// make sure all traversals are following up/down rules.
// return 1 oif there is a violation
int
SubnReportNonUpDownMulticastGroupFromCaSwitch(IBFabric *p_fabric,
        IBNode *pSwitchNode,
        map_pnode_rank &nodesRank,
        uint16_t mlid)
{
    map< IBNode *, uint8_t, less < IBNode *> > visitedNodeFromPort;
    list< upDnBfsEntry > nodesQueue;
    upDnBfsEntry thisStep, nextStep;
    rank_t thisNodeRank, remNodeRank;
    int anyErr = 0;
    char mlidStr[8];
    sprintf(mlidStr, "0x%04X", mlid);

    thisStep.down = 0;
    thisStep.pNode = pSwitchNode;
    thisStep.inPort = 0; // start as we got in from port 0 we can go out
                        //  any port
    thisStep.pTurnNode = NULL;
    nodesQueue.push_back(thisStep);

    // we do the BFS through the queue
    while (nodesQueue.size()) {
        thisStep = nodesQueue.front();
        nodesQueue.pop_front();

        // we keep track of all visited nodes
        visitedNodeFromPort[thisStep.pNode] = thisStep.inPort;

        // get the list of MC group ports for this mlid of this node
        list_phys_ports portNums = thisStep.pNode->getMFTPortsForMLid(mlid);

        // lookup the rank of the current node:
        map_pnode_rank::iterator rI = nodesRank.find(thisStep.pNode);
        if (rI == nodesRank.end()) {
            cout << "-E- Somehow we do not have rank for:" << thisStep.pNode->name
                    << endl;
            exit(1);
        }
        thisNodeRank = (*rI).second;

        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            cout << "-V- Visiting node:" << thisStep.pNode->name
                  << " dir:" << (thisStep.down ? "DOWN" : "UP") << endl;

        // go over all MC output ports of the current node ignoring the input
        for (list_phys_ports::iterator pnI = portNums.begin();
                pnI != portNums.end(); pnI++) {
            phys_port_t pn = (*pnI);

            // ignore the port we got here through.
            if (pn == thisStep.inPort)
                continue;

            IBPort *pPort = thisStep.pNode->getPort(pn);
            if (! pPort || ! pPort->p_remotePort)
                continue;

            // get the remote node
            IBNode *pRemNode = pPort->p_remotePort->p_node;

            // we ignore remote HCAs
            if (pRemNode->type != IB_SW_NODE)
                continue;

            // if we already visited this node - it is a loop!
            map< IBNode *, uint8_t, less < IBNode *> >::iterator vI =
                    visitedNodeFromPort.find(pRemNode);

            if (vI != visitedNodeFromPort.end()) {
                int prevPort =  (*vI).second ;
                cout << "-E- Found a loop on MLID:" << mlidStr
                        << " got to node:" << pRemNode->name
                        << " through port:" << (unsigned int)pPort->p_remotePort->num
                        << " previoulsy visited through port:" << prevPort << endl;
                anyErr++;
                continue;
            }

            // lookup remote node rank:
            map_pnode_rank::iterator rI = nodesRank.find(pRemNode);
            if (rI == nodesRank.end()) {
                cout << "-E- Somehow we do not have rank for:" << pRemNode->name
                        << endl;
                exit(1);
            }
            remNodeRank = (*rI).second;

            // if we are goin up make sure we did not go down before.
            if (remNodeRank < thisNodeRank) {
                // we go up - was it a down?
                if (thisStep.down) {
                    cout << "-E- Non Up/Down on MLID:" << mlidStr
                            << " turning UP from:" << thisStep.pNode->name
                            << "/P" << (unsigned int)pn << "("<< thisNodeRank << ") to node:"
                            << pRemNode->name << "/P" << (unsigned int)pPort->p_remotePort->num
                            << "(" << remNodeRank << ") previoulsy turned down:"
                            << thisStep.pTurnNode->name << endl;
                    anyErr++;
                    continue;
                } else {
                    // we turned down here so track it:
                    nextStep.down = 0;
                    nextStep.pTurnNode = NULL;
                }
            } else {
                nextStep.pTurnNode = thisStep.pNode;
                nextStep.down = 1;
            }

            // push the node into next steps:
            nextStep.pNode = pRemNode;
            nextStep.inPort = pPort->p_remotePort->num;
            nodesQueue.push_back(nextStep);
        }
    }
    return anyErr;
}


///////////////////////////////////////////////////////////////////////////////
// Find any routes that exist in the FDB's from CA to CA and do not adhare to
// the up/down rules. Report any crossing of the path.
int
SubnReportNonUpDownMulticastGroupCa2CaPaths(IBFabric *p_fabric,
        map_pnode_rank &nodesRank,
        uint16_t mlid)
{
    int numBadPaths = 0, paths = 0;
    char mlidStr[8];
    sprintf(mlidStr, "0x%04X", mlid);

    cout << "-I- Tracing Multicast Group:" << mlidStr
            << " CA to CA paths for Credit Loops potential ..."
            << endl;

    // find all HCAs in the MGRP
    list< IBNode *> groupSwitchesConnToHCAs;

    // find all switches that are part of this mcgrp and connects to HCAs
    for (map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
            nI != p_fabric->NodeByName.end();
            nI++) {
        IBNode *p_node = (*nI).second;
        if (p_node->type != IB_SW_NODE)
            continue;

        // see if we have an MFT entry by the given lid:
        list_phys_ports portNums = p_node->getMFTPortsForMLid(mlid);

        if (portNums.empty())
            continue;

        // find all HCAs connected to the group by following the links that
        // are marked in the MFT that connect to the group.
        for (list_phys_ports::iterator lI = portNums.begin();
                lI != portNums.end();
                lI++) {
            IBPort *p_port = p_node->getPort(*lI);

            if (p_port && p_port->p_remotePort &&
                    p_port->p_remotePort->p_node->type != IB_SW_NODE) {
                groupSwitchesConnToHCAs.push_back(p_node);
                break;
            }
        }
    }

    cout << "-I- Multicast group:" << mlidStr << " has:"
            << groupSwitchesConnToHCAs.size()
            << " Switches connected to HCAs" << endl;

    // from each HCA traverse BFS through the switches.
    for (list_pnode::iterator lI = groupSwitchesConnToHCAs.begin();
            lI != groupSwitchesConnToHCAs.end();
            lI++) {
        // avoid too many errors reported:
        if (numBadPaths > 100)
            break;

        numBadPaths +=
                SubnReportNonUpDownMulticastGroupFromCaSwitch(
                        p_fabric, (*lI), nodesRank, mlid);
        paths++;
    }

    if (numBadPaths) {
        if (numBadPaths > 100)
            cout << "-W- Stopped checking multicast groups after 100 errors" << endl;
        cout << "-E- Found:" << numBadPaths
                << " Multicast:" << mlidStr
                << " CA to CA paths that can cause credit loops." << endl;
    } else
        cout << "-I- No credit loops found traversing:" << paths
               << " leaf switches for Multicast LID:" << mlidStr << endl;
    return 0;
}


///////////////////////////////////////////////////////////////////////////////
// Check all multicast groups :
// 1. all switches holding it are connected
// 2. No loops (i.e. a single BFS with no returns).
int
SubnMgtCheckFabricMCGrpsForCreditLoopPotential(IBFabric *p_fabric,
        map_pnode_rank &nodesRank)
{
    int anyErrs = 0;
    cout << "-I- Scanning all multicast groups for Credit Loops Potential ..." << endl;

    for (set_uint16::iterator sI = p_fabric->mcGroups.begin();
            sI != p_fabric->mcGroups.end();
            sI++)
        anyErrs +=
                SubnReportNonUpDownMulticastGroupCa2CaPaths(p_fabric, nodesRank, *sI);

    if (anyErrs)
        cout << "-E- " << anyErrs << " multicast groups failed" << endl;

    cout << "---------------------------------------------------------------------------\n" << endl;
    return anyErrs;
}


///////////////////////////////////////////////////////////////////////////////
// Fill in the FDB tables assuming the fabric is a fat tree.
// Also assumes the fabric is already ranked from roots downwards.
// We also assume MinHop tables are defined.
//
// LIMITATION: Currently only LMC=0 is supported
//
// The main idea behind this algorithm is that if we have enough
// root ports we can assign each port one destination LID.
// Then we propagate that selection by allowing routing to that LID
// only through that top root and downwards - unless closes at lower levels
//
// ALGORITHM:
// * Initialize a WORKMAP of all HCA LIDs - also count them
// * Calc total switch ports versus number of HCA lids.
// * If we do not have enough switch ports abort.
// * Loop on all rank=0 switches.
//   * For each port select one of the LIDs that it MinHop to and are
//     still in WORKMAP
//   * Traverse forward to that LID assigning LFT
//
// Traverse Forward to LID:
// * Given switch and target LID
// * Find the out-port to be used for going to that LID
//   (use # paths to select from many?)
// * Set FDB to that LID (actually done by the Backward traversal)
// * Recurse to the direction of the LID (need Min hop for that)
// * Perform Backward traversal through all ports connected to lower
//   level switches in-port = out-port
//
// Traverse Backward to LID:
// * Given current switch and LID, in-port
// * Set FDB to target LID to the "in-port"
// * Recurse through all ports connected to lower level switches
//   not including the in-port

static inline void
markPortUtilization(IBPort *p_port)
{
    p_port->counter1++;
}

// given source and destination nodes find the port with lowest
// utilization (subscriptions) and return its number
static phys_port_t
getLowestUtilzedPortFromTo(IBNode *p_fromNode, IBNode *p_toNode)
{
    unsigned int minUtil=0;
    phys_port_t minUtilPortNum=0;
    IBPort *p_port;

    for (unsigned int i = 1; i <= p_fromNode->numPorts; i++) {
        phys_port_t pn = (phys_port_t)i;
        p_port = p_fromNode->getPort(pn);

        if (! p_port)
            continue;
        if (! p_port->p_remotePort)
            continue;
        if (p_port->p_remotePort->p_node != p_toNode)
            continue;

        // the hops should match the min
        if ((minUtilPortNum == 0) || (p_port->counter1 < minUtil)) {
            minUtilPortNum = pn;
            minUtil = p_port->counter1;
        }
    }

    return( minUtilPortNum );
}

// given a node and a target LID find the port that has min hops
// to that LID and lowest utilization
static phys_port_t
getLowestUtilzedPortToLid(IBNode *p_node, lid_t dLid)
{
    unsigned int minUtil = 0;
    phys_port_t minUtilPortNum = 0;
    IBPort *p_port;

    // get the minimal hop count from this node:
    uint8_t minHop = p_node->getHops(NULL,dLid);

    for (unsigned int i = 1; i <= p_node->numPorts; i++) {
        phys_port_t pn = (phys_port_t)i;
        p_port = p_node->getPort(pn);

        if (! p_port)
            continue;
        if (! p_port->p_remotePort)
            continue;

        // the hops should match the min
        if (p_node->getHops(p_port, dLid) == minHop) {
            if ((minUtilPortNum == 0) || (p_port->counter1 < minUtil)) {
                minUtilPortNum = pn;
                minUtil = p_port->counter1;
            }
        }
    }
    return( minUtilPortNum );
}

int
SubnMgtFatTreeBwd(IBNode *p_node, uint16_t dLid, phys_port_t outPortNum)
{
    IBPort* p_port;

    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
        cout << "-V- SubnMgtFatTreeBwd from:" << p_node->name << " dlid:" << dLid
               << " out-port:" << outPortNum << endl;

    // Set FDB to target LID to the "in-port"
    p_node->setLFTPortForLid(dLid, outPortNum);

    // mark this port was utilized
    markPortUtilization(p_node->getPort(outPortNum));

    // get the remote node to avoid decending down through it
    p_port = p_node->getPort(outPortNum);
    IBNode *p_origRemNode = p_port->p_remotePort->p_node;

    // Recurse through all ports connected to lower level switches not
    // including the in-port
    for (unsigned int i = 1; i <= p_node->numPorts; i++) {
        phys_port_t pn = (phys_port_t)i;
        if (pn == outPortNum)
            continue;

        p_port = p_node->getPort(pn);
        if (!p_port || !p_port->p_remotePort)
            continue;

        IBNode *p_remNode = p_port->p_remotePort->p_node;
        // we might have several ports
        if (p_remNode == p_origRemNode)
            continue;
        if (p_remNode->type != IB_SW_NODE)
            continue;
        // avoid going up or sideways in the tree
        if (p_node->rank >= p_remNode->rank)
            continue;

        // avoid loops by inspecting the FDB value of the remote port
        if (p_remNode->getLFTPortForLid(dLid) != IB_LFT_UNASSIGNED)
            continue;

        // select the best port from the remote node to this one:
        phys_port_t remPortNum = getLowestUtilzedPortFromTo(p_remNode, p_node);
        SubnMgtFatTreeBwd(p_remNode, dLid, remPortNum);
    }
    return(0);
}


///////////////////////////////////////////////////////////////////////////////
int
SubnMgtFatTreeFwd(IBNode *p_node, uint16_t dLid)
{
    phys_port_t outPortNum = 0;
    IBPort *p_port;

    // Find the out-port to be used for going to that LID
    // (use # paths to select from many?)
    outPortNum = getLowestUtilzedPortToLid(p_node, dLid);

    if (!outPortNum) {
        cout << "-E- fail to find output port for switch:" << p_node->name
                << " to LID:" << dLid << endl;
        exit(1);
    }

    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
        cout << "-V- SubnMgtFatTreeFwd from:" << p_node->name << " dlid:" << dLid
        << " through port:" << outPortNum << endl;

    p_port = p_node->getPort(outPortNum);

    // Set FDB to that LID (actually done by the Backward traversal)

    // Recurse to the direction of the LID
    if (p_port->p_remotePort->p_node->type == IB_SW_NODE)
        SubnMgtFatTreeFwd(p_port->p_remotePort->p_node, dLid);

    // Perform Backward traversal through all ports connected to lower
    // level switches in-port = out-port
    SubnMgtFatTreeBwd(p_node, dLid, outPortNum);
   return(0);
}


///////////////////////////////////////////////////////////////////////////////
int
SubnMgtFatTreeRoute(IBFabric *p_fabric)
{
    IBNode *p_node;
    IBPort *p_port = NULL;

    cout << "-I- Using Fat Tree Routing" << endl;

    // HACK we currently do not support LMC > 0
    if (p_fabric->caLmc > 0) {
        cout << "-E- Fat Tree Router does not support LMC > 0 yet" << endl;
        return(1);
    }

    list<IBNode*> rootNodes;
    set<lid_t, less<lid_t> > unRoutedLids;

    int numHcaPorts = 0;
    int numRootPorts = 0;

    // get all root nodes
    for( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
            nI != p_fabric->NodeByName.end();
            nI++) {
        p_node = (*nI).second;

        // if not a switch just count
        if (p_node->type != IB_SW_NODE) {
            // count all ports that are connected
            for (unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
                p_port = p_node->getPort((phys_port_t)pn);
                if (p_port && p_port->p_remotePort) {
                    numHcaPorts++;
                    unRoutedLids.insert(p_port->base_lid);
                }
            }
        } else {
            // we only crae now about root switch ports
            if (p_node->rank == 0) {
                rootNodes.push_back(p_node);
                // count all ports that are connected
                for (unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
                    p_port = p_node->getPort((phys_port_t)pn);
                    if (p_port && p_port->p_remotePort)
                        numRootPorts++;
                }
            }
        }
    }

    // do we have enough root ports?
    if (numRootPorts < numHcaPorts) {
        cout << "-E- Can Route Fat-Tree - not enough root ports:"
                << numRootPorts << " < HCA ports:" << numHcaPorts << endl;
        return(1);
    }

    // Loop on all rank=0 switches.
    for (list<IBNode *>::iterator lI = rootNodes.begin();
            lI != rootNodes.end();
            lI++) {
        set<lid_t, less<lid_t> > switchAllocatedLids;

        p_node = *lI;

        // Foreach port select one of the LIDs that it MinHop to and are
        // still in WORKMAP
        for (unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
            p_port = p_node->getPort((phys_port_t)pn);
            if (! p_port || !p_port->p_remotePort)
                continue;

            // since most fat trees are fully connected we can simply try from the
            // head of the unRoutedLids...
            int found = 0;
            for ( set<lid_t, less<lid_t> >::iterator mI = unRoutedLids.begin();
                    !found && (mI != unRoutedLids.end());
                    mI++) {
                // is this dLid in our port min hop?
                uint16_t dLid = *mI;
                // is it in this port min hop?
                if (p_node->getHops(NULL, dLid) == p_node->getHops(p_port, dLid)) {
                    found = 1;
                    // remove from set
                    unRoutedLids.erase(mI);

                    // as it is important we handle lids in order we just collect
                    // them here and later use in order
                    switchAllocatedLids.insert(dLid);

                    // escape the for loop
                    break;
                }
            }
        } // all ports of switch

        // now handle all allocated lids of this switch:
        for ( set<lid_t, less<lid_t> >::iterator alI = switchAllocatedLids.begin();
                alI != switchAllocatedLids.end();
                alI++) {
            lid_t dLid = *alI;
            if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
                cout << "-V- Routing to LID:" << dLid << " through root port:"
                << p_port->getName() << endl;
            // Traverse forward to that LID assigning LFT
            SubnMgtFatTreeFwd(p_node, dLid);
        }
    } // all rank 0 switches

    // double check no more HCA LIDs to route...
    if (unRoutedLids.size()) {
        cout << "-E- " << unRoutedLids.size()
                   << " lids still not routed:" << endl;
        for( set<lid_t, less<lid_t> >::iterator sI = unRoutedLids.begin();
                sI != unRoutedLids.end();
                sI++)
            cout << "   " << *sI << endl;
        return(1);
    }
    return(0);
}


///////////////////////////////////////////////////////////////////////////////
// Recursivly DFS backwards through all switch ports
// (except for the port to the lids) and fill in HCAS reached
static int
dfsBackToCAByLftToDLIDs(IBNode *node,
        list<lid_t> &dstLids,
        phys_port_t dstPortNum,
        set<IBNode *> &visitedNodes,
        map<IBPort *, list<lid_t> > &HCAPortsLids)
{
    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {
        cout << "-V- Visiting " << node->name << " searching for lids:";
        for (list<lid_t>::const_iterator lI = dstLids.begin();
                lI != dstLids.end(); lI++) cout << *lI << ",";
        cout << endl;
    }

    if (node->type != IB_SW_NODE) {
        IBPort* port = node->getPort(dstPortNum);
        HCAPortsLids[port] = dstLids;
        return(0);
    }

    // first check which dst lid is still valid
    list<lid_t> subDstLids;
    for (list<lid_t>::const_iterator lI = dstLids.begin();
            lI != dstLids.end(); lI++) {
        lid_t lid = *lI;
        if (node->getLFTPortForLid(lid) == dstPortNum) {
            subDstLids.push_front(lid);
        }
    }

    // no paths to the dlids through the port
    if (subDstLids.size() == 0) {
        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            cout << "-V- Dead end" << endl;
        return(0);
    }

    // mark the node visited now
    visitedNodes.insert(node);

    // DFS through the other ports
    for (unsigned int i = 1; i <= node->numPorts; i++) {
        phys_port_t pn = (phys_port_t)i;
        if (pn != dstPortNum) {
            IBPort *port = node->getPort(pn);
            if (!port || ! port->p_remotePort) continue;
            // set remPort [IBPort_p_remotePort_get $port]
            // if {$remPort == ""} {continue}
            IBNode *p_remNode = port->p_remotePort->p_node;
            if (visitedNodes.find(p_remNode) == visitedNodes.end()) {
                phys_port_t remPortNum = port->p_remotePort->num;
                dfsBackToCAByLftToDLIDs(p_remNode, subDstLids, remPortNum,
                        visitedNodes, HCAPortsLids);
            }
        }
    }
    return(0);
}

// return 1 if LID is reachable by LFT from given port
/* UNUSED
static int
isThereLFTPathToLID(IBPort *p_port, lid_t lid)
{
    if (p_port->base_lid == lid)
        return(1);

    set<IBNode *> visitedNodes;
    visitedNodes.insert(p_port->p_node);
    while (p_port) {
        // get remote port
        IBPort *p_remPort = p_port->p_remotePort;
        if (!p_remPort)
            return(0);

        if (p_remPort->base_lid == lid)
            return(1);

        // if we already visited we are in a loop
        if (visitedNodes.find(p_remPort->p_node) != visitedNodes.end())
            return(0);

        visitedNodes.insert(p_remPort->p_node);

        // if it is a switch
        if (p_remPort->p_node->type == IB_SW_NODE) {
            phys_port_t pn = p_remPort->p_node->getLFTPortForLid(lid);
            p_port = p_remPort->p_node->getPort(pn);
        } else {
            p_port = NULL;
        }
    }
    return(0);
}
*/
// Obtain all the CA to CA port pairs going through the
// given port
/* UNUSED
int
SubnReportCA2CAPathsThroughSWPort(IBPort *p_port)
{
    // if not a switch port error
    if (p_port->p_node->type != IB_SW_NODE) {
        cout << "-E- Provided port:" << p_port->getName()
	           << " is not a switch port" << endl;
        return(1);
    }

    IBNode *node = p_port->p_node;

    // obtain the DLIDs on the given switch port
    list<lid_t> lidsThroughPort;
    for (lid_t i = 0; i < node->LFT[0].size(); i++) { //!!!PLFT
        if (node->getLFTEntry(i) == p_port->num) {
            // TODO: validate there is really a path to that node
            if (isThereLFTPathToLID(p_port, i)) { //!!!PLFT
                lidsThroughPort.push_front(i);
            } else {
                if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
                    cout << "-V- LID:" << i
                    << " pointed by LFT but is not reachable from:"
                    << p_port->getName() << endl;
            }
        }
    }
    if (lidsThroughPort.size() == 0) {
        cout << "-W- No paths through port:" << p_port->getName() << endl;
        return(1);
    }

    set<IBNode *> visitedNodes;
    map<IBPort *, list<lid_t> > HCAPortsLids;

    dfsBackToCAByLftToDLIDs(node, lidsThroughPort, p_port->num,
            visitedNodes, HCAPortsLids);
    IBFabric *fabric = node->p_fabric;
    if (HCAPortsLids.size()) {
        map<IBPort *, list<lid_t> >::const_iterator pI;
        for (pI = HCAPortsLids.begin(); pI != HCAPortsLids.end(); pI++) {
            IBPort *port = (*pI).first;
            list<lid_t>::const_iterator lI;
            cout << "From:" << port->getName() << " SLID:" << port->base_lid << endl;
            for (lI = (*pI).second.begin(); lI != (*pI).second.end(); lI++) {
                IBPort *dPort = fabric->getPortByLid(*lI);
                cout << "   To:" << dPort->getName() << " DLID:" << dPort->base_lid
                        << endl;
            }
        }
    }
    return(0);
}
*/

////////////////////////////////////////////////////////
int
SubnMgtVerifyAllARCaToCaRoutes(IBFabric *p_fabric)
{
    int anyError = 0, paths = 0;

    //FabricUtilsVerboseLevel = 0xff; //!!!

    cout << "-I- Verifying all CA to CA AR paths ... " << endl;
    // initialize the histograms:

    // NOTE: since it is OK to have just one of the DLIDs assigned to a single
    // port be available as path from any given source we do not complain
    // about dead paths unless all of them from sPort to dPort are not avaiable

    // go over all ports in the fabric
    ARTraceRouteInfo globalRouteInfo;
    //ARTraceRouteInfo lid2lidRouteInfo;
    ARTraceRouteInfo *pArRouteInfo = NULL;

    ARTraceRouteNodeInfo::prepare(p_fabric);

    //go over all CA dest lid ordered by connected SW
    for ( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
          nI != p_fabric->NodeByName.end(); nI++ ) {

        IBNode *p_node = (*nI).second;

        if ( p_node->type != IB_SW_NODE )
            continue;

        //clear cache for new base SW
        ARTraceRouteNodeInfo::clearDB(p_fabric);

        for ( phys_port_t pn = 1; pn <= p_node->numPorts; pn++ ) {
            IBPort *p_port = p_node->getPort((phys_port_t)pn);
            if ( !p_port || !p_port->p_remotePort ||
                 !p_port->p_remotePort->p_node ||
                 p_port->p_remotePort->p_node->type == IB_SW_NODE)
                continue;

            IBPort *p_dstPort = p_port->p_remotePort;
            lid_t dLid;
            uint8_t lmc;
            p_port->p_remotePort->p_node->getLidAndLMC(p_port->p_remotePort->num, dLid, lmc);



            // go over all LMC values of the destination:
            lid_t pathsPerDlid = (lid_t)(1 << p_dstPort->lmc);
            set< lid_t > all_lids;
            // go over all LMC combinations:
            for ( lid_t l = 0; l < pathsPerDlid; l++ )
                all_lids.insert(lid_t(dLid + l));

            // go over all vports:
            map_vportnum_vport::iterator vports_iter = p_dstPort->VPorts.begin();
            for ( ; vports_iter != p_dstPort->VPorts.end(); ++vports_iter) {
                IBVPort *p_vport =  vports_iter->second;
                if (!p_vport ||
                    !p_fabric->getVPortByLid(p_vport->get_vlid())||
                    p_vport->getVPortNum() == 0)
                    continue;
                all_lids.insert(p_vport->get_vlid());
            }

            set< lid_t >::iterator lids_iter = all_lids.begin();
            for (; lids_iter != all_lids.end(); ++lids_iter)
            {
                //check cache for new dLID, might clear DB
                ARTraceRouteNodeInfo::checkDB(p_fabric, *lids_iter);

                // go over all the rest of the ports - but only once - so skip non base lids
                for ( lid_t sLid = p_fabric->minLid; sLid <= p_fabric->maxLid; sLid++ ) {
                    IBPort *p_srcPort = p_fabric->getPortByLid(sLid);
                    if ( !p_srcPort || p_srcPort->base_lid != sLid ) continue;

                    // Avoid tracing to itself
                    if ( sLid == dLid )
                        continue;
                    if ( p_srcPort->p_node->type == IB_SW_NODE )
                        continue;

                    paths++;

                    // now go and verify the path:
                    ARTraceRouteByLFT(p_fabric, sLid, *lids_iter, pArRouteInfo);
                    if (pArRouteInfo) {
                        //ARTraceRouteByLFT returns RouteInfo of connected SW
                        //so we can use it as a child in updateRouteStatistics
                        globalRouteInfo.updateRouteStatistics(pArRouteInfo);
                        if ( pArRouteInfo->getGoodPathCount() == 0 )
                            anyError++;
                    } else
                        anyError++;

                } // all slids
            } // all_lids (lmc values + vlids)

            /*
            if ( lid2lidRouteInfo.getGoodPathCount() == 0 ) {
                cout << "-E- Fail to find a path from:"
                << p_srcPort->p_node->name << "/"
                << (unsigned int)p_srcPort->num
                << " to:" << p_dstPort->p_node->name << "/"
                << (unsigned int)p_dstPort->num
                << endl;
                anyError++;
            }
            lid2lidRouteInfo.clearRouteStatistics();
            */

        } // all ports connected to same SW
    } // all SW

    globalRouteInfo.dumpRouteStatistics();

    if ( anyError )
        cout << "-E- Found " << anyError << " missing paths"
        << " out of:" << paths << " paths" << endl;
    else
        cout << "-I- Scanned:" << paths << " CA to CA paths " << endl;

    cout << "---------------------------------------------------------------------------\n" << endl;

    ARTraceRouteNodeInfo::cleanup(p_fabric);

    return anyError;
}


int
SubnMgtCheckSL2VLTables(IBFabric *p_fabric)
{
    int anyError = 0;

    cout << "-I- Check all SL2VL Tables ... " << endl;

    //go over all Nodes
    for ( map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
          nI != p_fabric->NodeByName.end(); nI++ ) {

        IBNode *p_node = (*nI).second;

        anyError += p_node->checkSL2VLTable();
    }

    if(anyError == 0)
        cout << "-I- All SL2VL tables are OK." << endl;

    return anyError;
}
