/*
 * Copyright (c) 2004-2018 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


/*
 * Topology Matching
 *
 * The file holds a set of utilities to providea "topology matching"
 * functionality: comparing two fabrics based on some entry point.
 *
 * The comparison is of a "discovered" fabric to a "spec" fabric:
 * disc = The fabric de-facto as OpenSM see it.
 * spec = The fabric as specified in either a topology file or cables file.
 *
 * Through the matching algorithm:
 *  1. The specification fabric get guids annotations.
 *  2. Mismatches are being formatted into a text report.
 *
 * After Matching a Merger can be run to create the merged fabric.
 * This fabric only includes the discovered nodes but uses the
 * names of the specification fabric and systems from the specification
 * fabric.
 */


#include <iomanip>
#include <sstream>
#include <fstream>

#include <infiniband/ibdm/TopoMatch.h>
#include <infiniband/ibdm/Regexp.h>

// to accumulate number of error and warning messages during the topo match
typedef struct TopoResCounter {
    int errors;
    int warnings;

    TopoResCounter() : errors(0), warnings(0)
    {
    }
} TopoResCounter_t;


//////////////////////////////////////////////////////////////////////////////
// we use the node appData2.val as a "visited" marker for the
// topology matching algorithm by setting it's value to 1
static inline void
TopoMarkNodeAsMatchAlgoVisited(IBNode *p_node) {
    p_node->appData2.val = 1;
}

static inline int
TopoIsNodeMatchAlgoVisited(IBNode *p_node) {
    return (int)(p_node->appData2.val & 1);
}

// we use the node appData2.val as a "reported" marker for the
// topology matching algorithm by setting it's second bit (2) to 1
static inline void
TopoMarkNodeAsReported(IBNode *p_node) {
    p_node->appData2.val = (p_node->appData2.val | 2);
}

static inline int
TopoIsNodeMarkedAsReported(IBNode *p_node) {
    return (int)(p_node->appData2.val & 2);
}


//////////////////////////////////////////////////////////////////////////////
// Check for two ports to be matching:
// PortNumber must match,
// If RemotePort exists should:
// a. Same port number
// b. Same node guid if already defined
// c. Link width and speed (these are only added to diagnostics as warnings)
// d. Remote nodes have same number of ports...
// RETURN: 1 if match 0 otherwise.
static int
TopoMatchPorts(IBPort *p_sPort,
        IBPort *p_dPort,
        int doDiag,
        stringstream &diag)
{
    if (!p_sPort || !p_dPort) {
        return 0;
    }
    if (p_sPort->num != p_dPort->num) {
        if (doDiag)
            diag << "Port number mismatch found. The port:" << p_sPort->getName()
		 << " != discovered:" << (unsigned int)p_dPort->num << endl;
        return 0;
    }

    IBPort *p_dRemPort = p_dPort->p_remotePort;
    IBPort *p_sRemPort = p_sPort->p_remotePort;

    if (p_sRemPort && !p_dRemPort) {
        if (doDiag)
            diag << "Missing link from:" << p_sPort->getName()
            << " to:" << p_sRemPort->getName()
            << endl;
        return 0;
    } else if (!p_sRemPort && p_dRemPort) {
        if (doDiag)
            diag << "Extra link from:" << p_sPort->getName()
            << " to:" << p_dRemPort->getName()
            << endl;
        return 0;
    } else if (p_sRemPort) {
        // check for port numbers
        if (p_sRemPort->num != p_dRemPort->num) {
            // on HCA nodes we can try the other port ...
            if (p_dRemPort->p_node->type != IB_SW_NODE) {
                // we need the spec to define other port of the HCA as empty
                if (doDiag)
                    diag << "Probably switched CA ports on cable from:"
                    << p_sPort->getName()
			 << ". Expected port:" << (unsigned int)p_sRemPort->num
                    << " but got port:" <<  (unsigned int)p_dRemPort->num << endl;
            } else {
                if (doDiag)
                    diag << "Wrong port number on remote side of cable from:"
                    << p_sPort->getName()
                    << ". Expected port:" <<  (unsigned int)p_sRemPort->num
                    << " but got port:" <<  (unsigned int)p_dRemPort->num << endl;
                return 0;
            }
        }

        // check link width - since it does not affect the result
        // only in diagnostics mode:
        if (doDiag) {
            IBLinkWidth sPortWidth = p_sPort->get_common_width();
            IBLinkWidth dPortWidth = p_dPort->get_common_width();
            if ((sPortWidth != IB_UNKNOWN_LINK_WIDTH) &&
                (sPortWidth != dPortWidth))
                diag << "Wrong link width on:" << p_sPort->getName()
                << ". Expected:" << width2char(sPortWidth)
                << " got:" << width2char(dPortWidth) << endl;
            IBLinkSpeed sPortSpeed = p_sPort->get_common_speed();
            IBLinkSpeed dPortSpeed = p_dPort->get_common_speed();
            if ((sPortSpeed != IB_UNKNOWN_LINK_SPEED) &&
                (sPortSpeed != dPortSpeed))
                diag << "Wrong link speed on:" << p_sPort->getName()
                << ". Expected:" << speed2char(sPortSpeed)
                << " got:" << speed2char(dPortSpeed) << endl;
        }

        IBNode *p_dRemNode = p_dRemPort->p_node;
        IBNode *p_sRemNode = p_sRemPort->p_node;

        // check to see if the remote nodes are not already cross linked
        if (p_dRemNode->appData1.ptr) {
            if ((IBNode *)p_dRemNode->appData1.ptr != p_sRemNode) {
                IBPort *p_port =
                        ((IBNode *)p_dRemNode->appData1.ptr)->getPort(p_sRemPort->num);
                if (p_port) {
                    diag << "Link from port:" << p_sPort->getName()
                       << " should connect to port:" << p_sRemPort->getName()
                       << " but connects to (previously matched) port:"
                       << p_port->getName() << endl;
                } else {
                    diag << "Link from port:" << p_sPort->getName()
                       << " should connect to port:" << p_sRemPort->getName()
                       << " but connects to a port not supposed to be connected"
                       << " on (previously matched) node:"
                       << ((IBNode *)p_dRemNode->appData1.ptr)->name << endl;
                }
                return 0;
            }
        }

/*
        if (p_sRemNode->appData1.ptr) {
            if ((IBNode *)p_sRemNode->appData1.ptr != p_dRemNode) {
                diag << "Link from port:" << p_sPort->getName()
                << " should connect to node:" << p_sRemNode->name
                << " but connects to previously matched node:"
                << ((IBNode *)p_sRemNode->appData1.ptr)->name << endl;
                return 0;
            }
        }
*/

        // if spec node guid is defined verify same guid:
        if (p_sRemNode->guid_get() &&
                (p_sRemNode->guid_get() != p_dRemNode->guid_get())) {
            if (doDiag)
                diag << "Wrong node on cable from:" << p_sPort->getName()
                << ". Expected connection to node:"
                << guid2str(p_sRemNode->guid_get())
                << " but connects to:"
                << guid2str(p_dRemNode->guid_get()) << endl;
            return 0;
        }

        // same number of ports on nodes
        // Skip HCA - not all HCA ports should be connected
        if ((p_sRemNode->numPorts != p_dRemNode->numPorts) &&
            (p_dRemNode->type != IB_CA_NODE)) {
            if (doDiag)
                diag << "Other side of cable from:" << p_sPort->getName()
                << " difference in port count. Expected:"
                << (unsigned int)p_sRemNode->numPorts
                << " but got:"
                << (unsigned int)p_dRemNode->numPorts << endl;
            return 0;
        }
    }
    return 1;
}


// Consider node matching by inspecting all their ports.
// All ports but maxMiss should be matched to pass.
// RETURN: 1 if qualified as matching 0 otherwise.
static int
TopoQalifyNodesMatching(IBNode *p_sNode, IBNode *p_dNode)
{
    int maxMissed = 2;          // TODO : Maybe need to restrict by the number of ports?
    int numMissed = 0;
    IBPort *p_sPort;
    IBPort *p_dPort;
    stringstream tmpDiag;

    // we must make sure we did not assign a match to any of the nodes ...
    if ((p_sNode->appData1.ptr || p_dNode->appData1.ptr) &&
            (p_sNode->appData1.ptr != p_dNode->appData1.ptr))
        return 0;

    // we must have same number of ports ...
    if (p_dNode->numPorts != p_sNode->numPorts)
        return 0;

    // Try to match all ports of this node
    // use unsigned int index to prevent overwrap
    for (unsigned int pn = 1; pn <= p_dNode->numPorts; pn++) {
        p_dPort = p_dNode->getPort((phys_port_t)pn);
        p_sPort = p_sNode->getPort((phys_port_t)pn);

        // we might miss some ports:
        if (!p_dPort && !p_sPort) {
            continue;
        } else if (!p_dPort || !p_sPort) {
            if (p_dPort && p_dPort->p_remotePort) {
                tmpDiag << "Port:" << pn
                        << " exist only in discovered model." << endl;
                numMissed++;
            } else if (p_sPort && p_sPort->p_remotePort) {
                tmpDiag << "Port:" << pn
                        << " exist only in specification model." << endl;
                numMissed++;
            }
            continue;
        }

        // we do not care about cases where we miss a
        // discovered link:
        if (! p_dPort->p_remotePort)
            continue;

        // ports match so push remote side into BFS if ok
        if (!TopoMatchPorts(p_sPort, p_dPort, 1, tmpDiag)) {
            numMissed++;
            // if (numMissed > maxMissed) break;
        }
    }

    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {
        if (numMissed) {
            if (numMissed <= maxMissed) {
                cout << "-V- Qualified Nodes:" << p_sNode->name
                << " to:" << p_dNode->name << " with:" << numMissed
                << " mismatches!" <<endl;
            } else {
                cout << "-V- Disqualified Nodes:" << p_sNode->name
                << " to:" << p_dNode->name << " due to:" << numMissed
                << " mismatches!\n" << tmpDiag.str() << endl;
            }
        }
    }

    return numMissed <= maxMissed;
}


// Mark nodes as matching
static inline void
TopoMarkMatcedNodes(IBNode *p_node1, IBNode *p_node2, int &matchCounter)
{
    if (p_node1->appData1.ptr || p_node2->appData1.ptr) {
        if (p_node1->appData1.ptr ==  p_node2->appData1.ptr) {
            if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
                cout << "-V- Skipping previously Matched nodes:" << p_node1->name
                << " and:" <<  p_node2->name << endl;
        } else {
            if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
                cout << "-V- Requested to mark matching nodes:" << p_node1->name
                << " and:" <<  p_node2->name << " previously matched to others" << endl;
        }
    } else {
        // do the cross links on the spec and disc fabrics:
        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            cout << "-V- Matched Node:" << p_node1->name
            << " and:" <<  p_node2->name << endl;
        p_node1->appData1.ptr = p_node2;
        p_node2->appData1.ptr = p_node1;
        matchCounter++;
    }
}


// Perform a BFS on both fabrics and cross point between matching nodes.
static int
TopoDoBFSAndMatch(IBNode *p_sNodeAnchor,  // Starting node on the specification fabrric
        IBNode *p_dNodeAnchor,  // Starting node on the discovered fabrric
        int &numMatchedNodes,
        stringstream &diag)
{
    int dNumNodes, sNumNodes;
    IBNode *p_sNode, *p_dNode;
    IBNode *p_sRemNode, *p_dRemNode;
    IBFabric *p_dFabric, *p_sFabric;
    int status;

    // BFS through the matching ports only.
    // we keep track of the the discovered nodes only as it needs to be matched
    // already to get into this list and thus we have a pointer to the spec node.
    // To mark visited nodes we use the appData2.val
    list < IBNode * > bfsFifo;

    bfsFifo.push_back(p_dNodeAnchor);
    TopoMarkNodeAsMatchAlgoVisited(p_dNodeAnchor);

    // On discovered fabrics where the CA nodes are marked by name we
    // can start traversing from these nodes also - of they match any
    // spec system by name.
    p_dFabric = p_dNodeAnchor->p_fabric;
    p_sFabric = p_sNodeAnchor->p_fabric;
    if (p_dFabric->subnCANames) {
        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            cout << "-V- Matching nodes by name"  << endl;

        // we go through entire discovered fabric trying to find match by name
        for (map_str_pnode::iterator nI = p_dFabric->NodeByName.begin();
                nI != p_dFabric->NodeByName.end(); nI++) {
            IBNode *p_node1 = (*nI).second;

            // so we try to find a node by the same name on the spec fabric.
            map_str_pnode::iterator snI = p_sFabric->NodeByName.find((*nI).first);

            // no match
            if (snI == p_sFabric->NodeByName.end()) {
                if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
                    cout << "-V- No match for:" << (*nI).first << endl;
                continue;
            }
            IBNode *p_node2 = (*snI).second;

            // make sure they are not previously match somehome
            if (p_node1->appData1.ptr || p_node2->appData1.ptr)
                continue;

            // do not rush into matching - double check all the nodes ports ...
            int anyMismatch = 0;

            //not all HCA ports should be connected
            if ((p_node1->numPorts != p_node2->numPorts) &&
                (p_node1->type != IB_CA_NODE))
                continue;
            for (unsigned int i = 1; !anyMismatch && (i <= p_node1->numPorts);
                    i++) {
                phys_port_t pn = (phys_port_t)i;
                IBPort *p_dPort = p_node1->getPort(pn);
                IBPort *p_sPort = p_node2->getPort(pn);
                bool sConnected = (p_sPort && p_sPort->p_remotePort);
                bool dConnected = (p_dPort && p_dPort->p_remotePort);

                if (!sConnected && !dConnected)
                    continue;

                if (! TopoMatchPorts(p_sPort, p_dPort, 1, diag)) {
                    anyMismatch++;
                    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
                        cout << "-V- Matched node:" <<  (*nI).first
                        << " by name - but some ports are different." << endl;
                }

                if (0 && p_dPort->p_remotePort) {
                    IBNode *p_dRemNode = p_dPort->p_remotePort->p_node;
                    IBNode *p_sRemNode = p_sPort->p_remotePort->p_node;
                    if (! TopoQalifyNodesMatching(p_sRemNode, p_dRemNode)) {
                        anyMismatch++;
                        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
                            cout << "-V- Disqualified start nodes match:"
                            << p_node1->name << " by rem nodes" << endl;
                    }
                }
            }

            if (anyMismatch == 0) {

                if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
                    cout << "-V- Using name match for nodes:" << p_node1->name << endl;
                // set the cross pointers and visited flag:
                TopoMarkNodeAsMatchAlgoVisited(p_node1);

                TopoMarkMatcedNodes(p_node1, p_node2, numMatchedNodes);
                // If we do not start from the HCAs we avoid cases where the
                // wrong leaf is connected and the leaf is deep enough to hide
                // it
                // bfsFifo.push_back(p_node1);
            }
        }
    }

    // track any further unmatch:
    int anyUnmatch = 0;

    // BFS loop
    while (!bfsFifo.empty()) {
        p_dNode = bfsFifo.front();
        bfsFifo.pop_front();
        // just makeing sure
        if (! p_dNode->appData1.ptr) {
            cerr << "TopoMatchFabrics: Got to a BFS node with no previous match?"
                    << endl;
            abort();
        }

        p_sNode = (IBNode *)p_dNode->appData1.ptr;
        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            cout << "-V- Visiting Node:" << p_sNode->name << endl;

        // Try to match all ports of this node
        for (unsigned int i = 1; i <= p_dNode->numPorts; i++) {
            phys_port_t pn = (phys_port_t)i;
            IBPort *p_dPort = p_dNode->getPort(pn);
            IBPort *p_sPort = p_sNode->getPort(pn);

            if (!p_sPort && !p_dPort) {
                continue;
            } else if (!p_sPort) {
                diag << "Non existing Spec Port for: "
                        << p_dPort->getName() << endl;
                anyUnmatch++;
                continue;
            } else if (!p_dPort) {
                diag << "Non existing Discovered Port for: "
                        << p_sPort->getName() << endl;
                anyUnmatch++;
                continue;
            }

            // ports match so push remote side into BFS if ok
            if (TopoMatchPorts(p_sPort, p_dPort, 1, diag)) {
                // ports matched but is there a remote side?
                if (p_dPort->p_remotePort) {
                    p_dRemNode = p_dPort->p_remotePort->p_node;
                    p_sRemNode = p_sPort->p_remotePort->p_node;

                    // we need to qualify the match (looking one more step ahead)
                    if (TopoQalifyNodesMatching(p_sRemNode, p_dRemNode)) {
                        // propagate the matching of the nodes:
                        TopoMarkMatcedNodes(p_dRemNode, p_sRemNode, numMatchedNodes);

                        // if not already matched
                        if (!TopoIsNodeMatchAlgoVisited(p_dPort->p_remotePort->p_node)) {
                            bfsFifo.push_back(p_dPort->p_remotePort->p_node);
                            TopoMarkNodeAsMatchAlgoVisited(p_dPort->p_remotePort->p_node);
                        }
                    }
                }
            } else {
                anyUnmatch++;
            }
        }
    }

    status = anyUnmatch;
    sNumNodes = (int)p_sNodeAnchor->p_fabric->NodeByName.size();
    dNumNodes = (int)p_dNodeAnchor->p_fabric->NodeByName.size();
    if (numMatchedNodes != sNumNodes) {
        diag << "Some nodes are missing!" << endl;
        status = 1;
    }

    if (numMatchedNodes != dNumNodes) {
        diag << "Some extra or unmatched nodes were discovered." << endl;
        status = 1;
    }
    return status;
}


// Perform a BFS on both fabrics and cross point between matching nodes.
// Start the algorithm from node on the specification fabric and node on the discovered fabric
static int
TopoBFSAndMatchFromNodes(IBNode *p_sNode, // Starting port on the specification fabrric
        IBNode *p_dNode,
        stringstream &diag)
{
    int numMatchedNodes = 0;

    // Mark local and remote nodes as matching
    // this is an assumption because we need to start from some node
    TopoMarkMatcedNodes(p_sNode, p_dNode, numMatchedNodes);

    return TopoDoBFSAndMatch(p_sNode,
            p_dNode,
            numMatchedNodes,
            diag);
}


// Perform a BFS on both fabrics and cross point between matching nodes.
// Start the algorithm from port on the specification fabric and port on the discovered fabric
static int
TopoBFSAndMatchFromPorts(IBPort *p_sPort, // Starting port on the specification fabrric
        IBPort *p_dPort,
        stringstream &diag)
{
    int numMatchedNodes = 0;
    IBNode *p_sRemNode;
    IBNode *p_dRemNode;

    // Check first that we have a match from the anchor starting ports
    if (! TopoMatchPorts(p_sPort, p_dPort, 1, diag)) {
        diag << "Starting ports do not match. Did you switch the ports?" << endl;
        return 1;
    }

    // Mark the start node as visited:
    TopoMarkNodeAsMatchAlgoVisited(p_dPort->p_node);

    // if we do not have a remote port what do we really match?
    if (!p_dPort->p_remotePort) {
        diag << "No link connected to starting port. Nothing more to match."
                << endl;
        return 1;
    }

    p_dRemNode = p_dPort->p_remotePort->p_node;
    p_sRemNode = p_sPort->p_remotePort->p_node;

    // Mark local and remote nodes as matching
    TopoMarkMatcedNodes(p_dRemNode, p_sRemNode, numMatchedNodes);
    TopoMarkMatcedNodes(p_dPort->p_node, p_sPort->p_node, numMatchedNodes);

    return TopoDoBFSAndMatch(p_sRemNode,
            p_dRemNode,
            numMatchedNodes,
            diag);
}


//////////////////////////////////////////////////////////////////////////////
// Second step in matching - refine mismatched nodes:
// Based on the previous matching - produce a set of all spec nodes
// that are connected to other matched spec nodes.
// Then go over all these nodes and produce a "best fit" match for them:
// The nodes will be connected on the spec model to a set of other nodes.
// Go over all the node ports:
//  If connected to a matched node
//   Check to see what node is connected to the same port of the disc node
//   If it is un-matched - add it to the map of optential nodes adn
//      incr the match value count.
// Now observe the map of potential matching nodes:
// If there is a node that is  mapped by > half the number of ports
// mark it as matching.
// typedef map < IBNode *, int , less < IBNode * > >  map_pnode_int;

static int
TopoMatchNodeByAdjacentMatched(IBNode *p_sNode)
{
    int succeedMatch = 0;

    // This map will hold the number of matching ports for each candidate
    // discovered node
    map_pnode_int dNodeNumMatches;

    // we track the total number of connected ports of the spec node
    int numConnPorts = 0;

    // Go over all the node ports:
    for (unsigned int pn = 1; pn <= p_sNode->numPorts; pn++) {
        IBPort *p_sPort = p_sNode->getPort((phys_port_t)pn);

        //  If connected to a remote port
        if (! p_sPort || ! p_sPort->p_remotePort)
            continue;

        numConnPorts++;

        IBPort *p_sRemPort = p_sPort->p_remotePort;

        // But the remote spec node must be matched to be considered
        IBNode *p_dRemNode = (IBNode *)p_sRemPort->p_node->appData1.ptr;
        if (! p_dRemNode)
            continue;

        //   Check to see what node is connected to the same port of the disc node
        IBPort *p_dRemPort = p_dRemNode->getPort(p_sRemPort->num);

        // is it connected ?
        if (! p_dRemPort || ! p_dRemPort->p_remotePort)
            continue;

        IBNode *p_dCandidateNode = p_dRemPort->p_remotePort->p_node;

        // It must be un-matched
        if (p_dCandidateNode->appData1.ptr)
            continue;

        // OK so it is a valid candidate
        // - add it to the map of optential nodes adn
        //      incr the match value count.
        map_pnode_int::iterator cnI = dNodeNumMatches.find(p_dCandidateNode);
        if (cnI == dNodeNumMatches.end()) {
            dNodeNumMatches[p_dCandidateNode] = 1;
        } else {
            (*cnI).second++;
        }
    }

    // Now observe the map of potential matching nodes:
    // If there is a node that is  mapped by > half the number of ports
    // mark it as matching.
    for (map_pnode_int::iterator cnI = dNodeNumMatches.begin();
       cnI != dNodeNumMatches.end(); cnI++) {
        if ( (*cnI).second > numConnPorts / 2) {
            TopoMarkMatcedNodes(p_sNode, (*cnI).first, succeedMatch);
        }
    }
    return succeedMatch;
}


// Provide the list of un-matched spec nodes that
// are adjecant to a matched nodes
static list < IBNode *>
TopoGetAllSpecUnMatchedAdjacentToMatched(IBFabric *p_sFabric)
{
    list < IBNode *> adjNodes;
    for (map_str_pnode::iterator nI = p_sFabric->NodeByName.begin();
            nI != p_sFabric->NodeByName.end(); nI++) {
        IBNode *p_node = (*nI).second;

        // if the node is already matcehd skip it:
        if (p_node->appData1.ptr)
            continue;

        // go over all the node ports and check adjacent node.
        int done = 0;
        for (unsigned int pn = 1;
                (pn <= p_node->numPorts) && !done; pn++) {
            IBPort *p_port = p_node->getPort((phys_port_t)pn);
            // if there is a remote port
            if (p_port && p_port->p_remotePort) {
                // check if the other node is matched.
                if (p_port->p_remotePort->p_node->appData1.ptr) {
                    // add the node to the list and break the for
                    done = 1;
                    adjNodes.push_back(p_node);
                }
            }
        }
    }
    return adjNodes;
}


// Second Matching step
// Return the number of matched nodes by this step
static void
TopoMatchSpecNodesByAdjacentNode(IBFabric *p_sFabric)
{
    list < IBNode *> unMatchedWithAdjacentMatched;
    int numMatched = 0;

    unMatchedWithAdjacentMatched =
            TopoGetAllSpecUnMatchedAdjacentToMatched(p_sFabric);

    for( list < IBNode *>::iterator nI = unMatchedWithAdjacentMatched.begin();
            nI != unMatchedWithAdjacentMatched.end(); nI++)
        numMatched += TopoMatchNodeByAdjacentMatched(*nI);
}


//////////////////////////////////////////////////////////////////////////////
// Report mismatched nodes by examining entire system and looking for
// all system and then a single board (matching regexp).
// So the report will be optimized:
// If the entire system is missing - report only at this level
// If an entire board of the system missing - report only at this level
// Otherwise report at the specific node only.
static int
TopoReportMismatchedNode(IBNode *p_node,
        int reportMissing, // 0 = reportExtra ...
        stringstream &diag)
{
    int anyMismatch = 0;
    int MaxConnRep = 12;
    // we assume if a previous check was done we would have
    // never been called since the "reported" mark would have been set
    IBSystem *p_system = p_node->p_system;

    // we always mark the board of the node by examining all but the "UXXX"
    const char *p_lastSlash = rindex(p_node->name.c_str(), '/');
    char nodeBoardName[512];
    int  boardNameLength;
    if (!p_lastSlash) {
        strcpy(nodeBoardName, "NONE");
        boardNameLength = 0;
    } else {
        boardNameLength = (int)(p_lastSlash - p_node->name.c_str());
        strncpy(nodeBoardName, p_node->name.c_str(), boardNameLength);
        nodeBoardName[boardNameLength] = '\0';
    }

    int anyOfBoardMatch = 0;
    int anyOfSystemMatch = 0;
    list< IBNode *> boardNodes;

    // visit all nodes of the the specified system.
    for (map_str_pnode::iterator nI = p_system->NodeByName.begin();
            nI != p_system->NodeByName.end(); nI++) {
        IBNode *p_node = (*nI).second;

        // see if it of the same board:
        if (!strncmp(p_node->name.c_str(), nodeBoardName, boardNameLength)) {
            // on same board
            boardNodes.push_back(p_node);
            if (p_node->appData1.ptr) {
                // it was matched to discovered node!
                anyOfBoardMatch++;
                anyOfSystemMatch++;
                // since we only need one to match the board we do not need to
                // continue...
                break;
            }
        } else {
            if (p_node->appData1.ptr) {
                anyOfSystemMatch++;
            }
        }
    }

    // got the markings:
    if (! anyOfSystemMatch) {
        if (reportMissing) {
            diag << "Missing System:" << p_system->name << "("
                    << p_system->type << ")" << endl;

            // we limit the number of reported connections ...
            int numRep = 0;
            for (map_str_psysport::iterator spI = p_system->PortByName.begin();
                    (spI != p_system->PortByName.end()) && (numRep < MaxConnRep); spI++) {
                IBSysPort *p_sysPort = (*spI).second;
                if (p_sysPort->p_remoteSysPort) {
                    numRep++;
                    diag << "   Should be connected by cable from port: "
                            << (*spI).first
                            << "(" << p_sysPort->p_nodePort->p_node->name << "/P"
                            <<  (unsigned int)p_sysPort->p_nodePort->num << ")"
                            << " to:" << p_sysPort->p_remoteSysPort->p_system->name
                            << "/" << p_sysPort->p_remoteSysPort->name
                            << "(" << p_sysPort->p_remoteSysPort->p_nodePort->p_node->name
                            << "/P" <<  (unsigned int)p_sysPort->p_remoteSysPort->p_nodePort->num << ")"
                            << endl;
                }
            }
            if (numRep == MaxConnRep) diag << "   ..." << endl;
            diag << endl;
        } else {
            diag << "Extra System:" << p_system->name << endl;
            int numRep = 0;
            for (unsigned int pn = 1;
                    (pn <= p_node->numPorts) && (numRep < MaxConnRep); pn++) {
                IBPort *p_port = p_node->getPort((phys_port_t)pn);
                if (p_port && p_port->p_remotePort) {
                    numRep++;
                    if (p_port->p_remotePort->p_node->appData1.ptr) {
                        IBNode *p_sNode =
                                (IBNode *)p_port->p_remotePort->p_node->appData1.ptr;
                        // NOTE: we can not assume all ports were matched!
                        diag << "   Connected by cable from port: P" << pn
                                << " to:"
                                << p_sNode->name << "/P" << (unsigned int)p_port->p_remotePort->num
                                << endl;
                    } else {
                        diag << "   Connected by cable from port: P" << pn
                                << " to:" << p_port->p_remotePort->getName() << endl;
                    }
                }
            }
            if (numRep == MaxConnRep) diag << "   ..." << endl;
            diag << endl;
        }
        anyMismatch++;

        // also we do not need any more checks on this system
        for (map_str_pnode::iterator nI = p_system->NodeByName.begin();
                nI != p_system->NodeByName.end(); nI++) {
            IBNode *p_node = (*nI).second;
            TopoMarkNodeAsReported(p_node);
        }
    } else if (! anyOfBoardMatch) {
        if (reportMissing) {
            diag << "Missing System Board:" << nodeBoardName << endl;
            diag << endl;
        } else {
            diag << "Extra System Board:" << nodeBoardName << endl;
            diag << endl;
        }
        anyMismatch++;

        for (list< IBNode *>::iterator nI = boardNodes.begin();
                nI != boardNodes.end(); nI++) {
            TopoMarkNodeAsReported(*nI);
        }
    } else {
        if (reportMissing) {
            diag << "Missing Node:" << p_system->name
                    << "/" << p_node->name << endl;
            int numRep = 0;
            for (unsigned int pn = 1;
                    (pn <= p_node->numPorts) && (numRep < MaxConnRep); pn++) {
                IBPort *p_port = p_node->getPort((phys_port_t)pn);
                if (p_port && p_port->p_remotePort) {
                    numRep++;
                    diag << "   Should be connected by cable from port:"
                         << pn << " to:" << p_port->getName()
                         << endl;
                }
            }
            if (numRep == MaxConnRep) diag << "   ..." << endl;
            diag << endl;
        } else {
            diag << "Extra Node:" << p_system->name
                    << "/" << p_node->name << endl;
            int numRep = 0;
            for (unsigned int pn = 1;
                    (pn <= p_node->numPorts) && (numRep < MaxConnRep); pn++) {
                IBPort *p_port = p_node->getPort((phys_port_t)pn);
                if (p_port && p_port->p_remotePort) {
                    numRep++;
                    if (p_port->p_remotePort->p_node->appData1.ptr) {
                        IBNode *p_sNode =
                                (IBNode *)p_port->p_remotePort->p_node->appData1.ptr;
                        IBPort *p_tmpPort = p_sNode->getPort(p_port->p_remotePort->num);
                        if (p_tmpPort)
                            diag << "   Connected by cable from port:" << pn
                            << " to:"
                            << p_tmpPort->getName()
                            << endl;
                        else
                            diag << "   Connected by cable from port:" << pn
                            << " to:"
                            << p_sNode->name << "/P"
			    << (unsigned int)p_port->p_remotePort->num
                            << endl;
                    } else {
                        diag << "   Connected by cable from port:" << pn
                                << " to:" << p_port->p_remotePort->getName() << endl;
                    }
                }
            }
            if (numRep == MaxConnRep)
                diag << "   ..." << endl;
            diag << endl;
        }
        anyMismatch++;

        TopoMarkNodeAsReported(p_node);
    }
    return(anyMismatch);
}


// Given two matching nodes - report un-matching links.
// To avoid duplicated reporting:
// * We only report mismatching links to matched nodes.
// * We only report link to nodes with bigger pointer value
// RETURN number of mismatches
static int
TopoReportMatchedNodesUnMatchingLinks(IBNode *p_sNode,
        IBNode *p_dNode,
        stringstream &diag)
{
    int anyMismatch = 0;
    // mismatch is when port numbers differ
    // or the remote node is not the same.
    IBPort *p_sPort, *p_dRemPort;
    IBPort *p_dPort, *p_sRemPort;

    for (unsigned int i = 1; i <= p_dNode->numPorts; i++) {
        phys_port_t pn = (phys_port_t)i;
        p_dPort = p_dNode->getPort(pn);
        p_sPort = p_sNode->getPort(pn);
        if (p_dPort)
            p_dRemPort = p_dPort->p_remotePort;
        else
            p_dRemPort = NULL;

        if (p_sPort)
            p_sRemPort = p_sPort->p_remotePort;
        else
            p_sRemPort = NULL;

        // so we have four cases
        if (p_dRemPort && p_sRemPort) {
            // can we ignore this report:
            if (!p_dRemPort->p_node->appData1.ptr)
                continue;

            IBNode *p_actRemSNode = (IBNode *)p_dRemPort->p_node->appData1.ptr;

            // make sure the expected remote spec node is what we got.
            if (p_actRemSNode != p_sRemPort->p_node) {
                // clasify internal and external cables
                if (p_sNode->p_system != p_actRemSNode->p_system) {
                    IBPort *p_repPort = p_actRemSNode->getPort(p_dRemPort->num);
                    if (p_repPort)
                        diag << "Wrong cable connecting:" << p_sPort->getName()
                        << " to:" << p_repPort->getName()
                        << " instead of:" << p_sRemPort->getName() << endl;
                    else
                        diag << "Wrong cable connecting:" << p_sPort->getName()
                        << " to:" << p_actRemSNode->name << "/P"
			<< (unsigned int)p_dRemPort->num
                        << " instead of:" << p_sRemPort->getName() << endl;
                } else {
                    diag << "Wrong Internal Link connecting:" << p_sPort->getName()
                       << " to:" << p_actRemSNode->name << "/P"
                       << (unsigned int)p_dRemPort->num
                       << " instead of:" << p_sRemPort->getName() << endl;
                }
                anyMismatch++;
            } else {
                // make sure same port remote number
                if (p_dRemPort->num != p_sRemPort->num) {
                    if (p_sNode->p_system != p_actRemSNode->p_system) {
                        IBPort *p_repPort = p_actRemSNode->getPort(p_dRemPort->num);
                        if (p_repPort)
                            diag << "Wrong cable connecting:" << p_sPort->getName()
                            << " to:" << p_repPort->getName()
                            << " instead of:" << p_sRemPort->getName() << endl;
                        else
                            diag << "Wrong cable connecting:" << p_sPort->getName()
                            << " to:" << p_actRemSNode->name << "/P"
                            << (unsigned int)p_dRemPort->num
                            << " instead of:" << p_sRemPort->getName() << endl;
                    } else {
                        diag << "Wrong Internal Link connecting:" << p_sPort->getName()
                         << " to:" << p_actRemSNode->name << "/P"
                         <<  (unsigned int)p_dRemPort->num
                         << " instead of:" << p_sRemPort->getName() << endl;
                    }
                    anyMismatch++;
                }
            }
            // anyway make sure we got the width and speed.
            IBLinkWidth sPortWidth = p_sPort->get_common_width();
            IBLinkWidth dPortWidth = p_dPort->get_common_width();
            if ((sPortWidth != IB_UNKNOWN_LINK_WIDTH) &&
                (sPortWidth != dPortWidth)) {
                diag << "Wrong link width on:" << p_sPort->getName()
                     << ". Expected:" << width2char(sPortWidth)
                     << " got:" << width2char(dPortWidth) << endl;
                anyMismatch++;
            }
            IBLinkSpeed sPortSpeed = p_sPort->get_common_speed();
            IBLinkSpeed dPortSpeed = p_dPort->get_common_speed();
            if ((sPortSpeed != IB_UNKNOWN_LINK_SPEED) &&
                (sPortSpeed != dPortSpeed)) {
                diag << "Wrong link speed on:" << p_sPort->getName()
                     << ". Expected:" << speed2char(sPortSpeed)
                     << " got:" << speed2char(dPortSpeed) << endl;
                anyMismatch++;
            }
            // done with the case both spec and dicsovered links exists
        } else if (!p_dRemPort && p_sRemPort) {
            // We have a missing cable/link

            // can we ignore this report if not both spec nodes matched
            // or not up going ...
            if (!p_sRemPort->p_node->appData1.ptr)
                continue;

            // we can ignore cases where the other side of the link is
            // connected - since it will be reported later
            IBPort *p_actRemPort =
                    ((IBNode *)p_sRemPort->p_node->appData1.ptr)->getPort(p_sRemPort->num);
            if (p_actRemPort && p_actRemPort->p_remotePort)
                continue;

            // as the spec connections are reciprocal we do not want double the
            // report
            if (p_sPort > p_sRemPort) {
                // clasify the link as cable or internal
                if (p_sPort->p_sysPort || p_sRemPort->p_sysPort) {
                    diag << "Missing cable connecting:" << p_sPort->getName()
                       << " to:" <<  p_sRemPort->getName() << endl;
                } else {
                    diag << "Missing internal Link connecting:" << p_sPort->getName()
                       << " to:" << p_sRemPort->getName() << endl;
                }
            }
            anyMismatch++;
        } else if (p_dRemPort && !p_sRemPort) {
            // Only a discovered link exists:

            // can we ignore this report if not both discovered nodes matched
            // or not up going ...
            if (!p_dRemPort->p_node->appData1.ptr)
                continue;

            // we can ignore cases where the other side of the link was
            // supposed to be connected - since it will be reported later as wrong
            IBPort *p_actRemPort =
                    ((IBNode *)p_dRemPort->p_node->appData1.ptr)->getPort(p_dRemPort->num);
            if (p_actRemPort && p_actRemPort->p_remotePort)
                continue;

            // we also ignore it by comparing pointer values on discovered
            // fabric to avoid double reporting.
            if (p_dRemPort < p_dPort)
                continue;

            string portName;
            if (p_dRemPort->p_node->appData1.ptr) {
                IBPort *p_port =
                        ((IBNode *)p_dRemPort->p_node->appData1.ptr)->getPort(p_dRemPort->num);
                // We can not guarantee all ports match on matching nodes
                if (p_port) {
                    portName = p_port->getName();
                } else {
                    char buff[256];
                    sprintf(buff, "%s/P%d",
                            ((IBNode *)p_dRemPort->p_node->appData1.ptr)->name.c_str(),
                            p_dRemPort->num);

                    portName = string(buff);
                }
            } else {
                portName = p_dRemPort->getName();
            }

            string specPortName;
            if (p_sPort) {
                specPortName = p_sPort->getName();
            } else {
                char buf[16];
                sprintf(buf,"%d",pn);
                specPortName = p_sNode->name + "/" + buf;
            }

            // clasify the link as cable or internal
            if ((p_dNode->p_system != p_dRemPort->p_node->p_system) ||
                    (p_sPort && p_sPort->p_sysPort) ) {
                diag << "Extra cable connecting:" << specPortName
                        << " to:" << portName  << endl;
            } else {
                diag << "Extra internal Link connecting:" << specPortName
                        << " to:" << portName << endl;
            }
            anyMismatch++;
        }
    }
    if (anyMismatch)
        diag << endl;
  return(anyMismatch);
}

// Given two matching nodes - report un-matching node names
// RETURN number of mismatches
static int
TopoReportUnMatchingNames(IBNode *p_sNode,
		IBNode *p_dNode,
		stringstream &diag)
{
	size_t desc_first_space;

	// currently only supports CA matching
	if (p_sNode->type != IB_CA_NODE || p_dNode->type != IB_CA_NODE)
		return 0;

	if (!strcmp(p_sNode->getName().c_str(), p_dNode->getName().c_str()))
		return 0;

	// compare the first word from the discoverd node description,
	// to the name of the equivalent specified node,
	// needed to be replaced with a function
	desc_first_space = p_dNode->name.find(' ');
	if (desc_first_space > 0 &&
			!strcmp(p_sNode->getName().c_str(),
				p_dNode->name.substr(0, desc_first_space).c_str()))
		return 0;

	diag << "Matching nodes have different names:" << p_sNode->getName()
		<< " and:" << p_dNode->getName() << endl;
	diag << endl;

	return 1;
}

// Report Topology Mismatched
static int
TopoReportMismatches(IBNode *p_sNodeAnchor,            // Starting node on the specification fabrric
        IBNode *p_dNodeAnchor,
        stringstream &diag,
		int compareNodesNames)
{
    IBNode *p_sNode, *p_dNode;
    int anyMismatchedSpecNodes = 0;
    int anyMismatchedDiscNodes = 0;
    int anyMismatchedLinks = 0;
	int anyMismatchedNames = 0;

    // since we do not want to report errors created by previous node failure
    // to discover or match we will BFS from the start ports.

    // NOTE: we only put matched nodes in the fifo.
    list < IBNode * > bfsFifo;

    // If the starting nodes match:
    if (p_dNodeAnchor->appData1.ptr)
        bfsFifo.push_back(p_dNodeAnchor);

    // BFS through the fabric reporting
    while (! bfsFifo.empty()) {
        p_dNode = bfsFifo.front();
        bfsFifo.pop_front();
        // we got in here only with matching nodes:
        p_sNode = (IBNode *)p_dNode->appData1.ptr;
        if (!p_sNode) {
            cerr << "How did we get in BFS with mismatching nodes!" << endl;
            exit (1);
        }

        // Try to match all ports of the current nodes
        // we assume port counts match since the nodes match
        for (unsigned int i = 1; i <= p_dNode->numPorts; i++) {
            phys_port_t pn = (phys_port_t)i;
            IBPort *p_dPort = p_dNode->getPort(pn);
            IBPort *p_sPort = p_sNode->getPort(pn);

            if (! p_dPort || !p_sPort) continue;

            IBPort *p_dRemPort = p_dPort->p_remotePort;
            IBPort *p_sRemPort = p_sPort->p_remotePort;

            // we only do care about spec nodes for now:
            if (!p_sRemPort )
                continue;

            // if we did visit it by this reporting algorithm:
            if (TopoIsNodeMarkedAsReported(p_sRemPort->p_node)) continue;

            // if the remote spec node is not marked matched:
            if (!p_sRemPort->p_node->appData1.ptr) {
                if (TopoReportMismatchedNode(p_sRemPort->p_node, 1, diag))
                    anyMismatchedSpecNodes++;
            } else {
                // mark as visited.
                TopoMarkNodeAsReported(p_sRemPort->p_node);

                // ok so we matched the nodes - put them in for next step:
                if (p_dRemPort) {
                    if (p_dRemPort->p_node->appData1.ptr)
                        bfsFifo.push_back(p_dRemPort->p_node);
                }
            }
        }           // all ports
    }           // next nodes available...

    if (anyMismatchedSpecNodes)
        diag << endl;

    IBFabric *p_dFabric = p_dNodeAnchor->p_fabric;

    // now when we are done reporting missing - let us report extra nodes:
    // but we only want to count those who touch matching nodes BFS wise...
    // If the starting nodes match:
    if (p_dNodeAnchor->appData1.ptr) {
        bfsFifo.push_back(p_dNodeAnchor);
        TopoMarkNodeAsReported(p_dNodeAnchor);
    } else {
        diag << "Even starting nodes do not match! "
                << "Expected:" << p_sNodeAnchor->name
                << " but it is probably not connected correctly." <<  endl;
        anyMismatchedDiscNodes++;
    }

    // we track visited nodes in here by clearing their visited bits

    // BFS through the fabric reporting
    while (! bfsFifo.empty()) {
        p_dNode = bfsFifo.front();
        bfsFifo.pop_front();

        int anyPortMatch = 0;
        // Try to match all ports of the current nodes
        // we assume port counts match since the nodes match
        for (unsigned int pn = 1; pn <= p_dNode->numPorts; pn++) {
            IBPort *p_tmpPort = p_dNode->getPort((phys_port_t)pn);
            if (! p_tmpPort)
                continue;

            IBPort *p_dRemPort = p_tmpPort->p_remotePort;
            if (!p_dRemPort)
                continue;

            IBNode *p_node = p_dRemPort->p_node;

            // we only care about mismatched discovered nodes
            if (! p_node->appData1.ptr) {
                // if not already visited.
                if (! TopoIsNodeMarkedAsReported(p_node))
                    if (TopoReportMismatchedNode(p_node, 0, diag))
                        anyMismatchedDiscNodes++;
            } else {
                // if not already visited.
                if (! TopoIsNodeMarkedAsReported(p_node)){
                    // it is an OK node so we traverse through
                    bfsFifo.push_back(p_node);
                    TopoMarkNodeAsReported(p_node);
                }
                anyPortMatch++;
            }
        }           // all ports

        p_sNode = (IBNode *)p_dNode->appData1.ptr;
        if (!anyPortMatch) {
            diag << "None of node:" << p_sNode->name
                    << " remote nodes match the topology."
                    << " No further extra nodes reported" << endl;
        }

        //TopoMarkNodeAsReported(p_dNode);
    }           // next nodes available...

    if (anyMismatchedDiscNodes)
        diag << endl;

    // finaly we want to report any mismatched cables between matching nodes
    for (map_str_pnode::iterator nI = p_dFabric->NodeByName.begin();
            nI != p_dFabric->NodeByName.end(); nI++) {
        p_dNode = (*nI).second;
        p_sNode = (IBNode *)p_dNode->appData1.ptr;
        if ( p_sNode ) {
            // report any mismatched links on this node.
            if (TopoReportMatchedNodesUnMatchingLinks(p_sNode, p_dNode, diag))
                anyMismatchedLinks++;
        }
    }

    if (anyMismatchedLinks) {
        diag << endl;
    }

	// report any mismatched names of matching nodes
	if (p_dFabric->subnCANames && compareNodesNames) {
		for (map_str_pnode::iterator nI = p_dFabric->NodeByName.begin();
				nI != p_dFabric->NodeByName.end();
				nI++) {
			p_dNode = (*nI).second;
			p_sNode = (IBNode *)p_dNode->appData1.ptr;
			if (p_sNode && TopoReportUnMatchingNames(p_sNode, p_dNode, diag))
				anyMismatchedNames++;
		}
	}

    if (anyMismatchedNames)
        diag << endl;

    return anyMismatchedLinks +
    anyMismatchedDiscNodes +
    anyMismatchedSpecNodes +
	anyMismatchedNames;
}


//////////////////////////////////////////////////////////////////////////////
static void
TopoCleanUpBeforeMerge(IBFabric *p_sFabric, IBFabric *p_dFabric)
{
    IBNode *p_node;
    // since we only use the appData1 and appData2 on all node  ...
    for (map_str_pnode::iterator nI = p_sFabric->NodeByName.begin();
            nI != p_sFabric->NodeByName.end(); nI++) {
        p_node = (*nI).second;
        p_node->appData1.ptr = 0;
        p_node->appData2.val = 0;
    }

    for (map_str_pnode::iterator nI = p_dFabric->NodeByName.begin();
            nI != p_dFabric->NodeByName.end(); nI++) {
        p_node = (*nI).second;
        p_node->appData1.ptr = 0;
        p_node->appData2.val = 0;
    }
}


//////////////////////////////////////////////////////////////////////////////
// Return 0 if fabrics match 1 otherwise.
int
TopoMatchFabrics(IBFabric   *p_sFabric,         // The specification fabric
        IBFabric   *p_dFabric,                  // The discovered fabric
        const char *anchorNodeName,             // The name of the system to be the anchor point
        phys_port_t anchorPortNum,              // The port number of the anchor port
        uint64_t    anchorPortGuid,             // Guid of the anchor port
        char **messages,
		int checkNodeNames)						// Check that node's name match
{
    stringstream diag, tmpDiag;
    int status = 0;

    IBNode *p_sNode = NULL, *p_dNode = NULL;
    IBPort *p_sPort = NULL, *p_dPort = NULL;

    // get the anchor node on the spec fabric - by name:
    p_sNode = p_sFabric->getNode(anchorNodeName);
    if (!p_sNode) {
        diag << "Fail to find anchor node:"
                << anchorNodeName << " on the specification fabric." << endl;
        status = 1;
        goto Exit;
    }

    // get the anchor port on the discovered fabric - by guid.
    p_dPort = p_dFabric->getPortByGuid(anchorPortGuid);
    if (!p_dPort) {
        diag << "Fail to find anchor port guid:"
                << guid2str(anchorPortGuid)
                << " in the discovered fabric." << endl;
        status = 1;
        goto Exit;
    }

    // get the anchor node on the discovered fabric - by port obj:
    p_dNode = p_dPort->p_node;
    if (!p_dNode) {
        diag << "Fail to find anchor node:"
                << anchorNodeName << " on the discoverd fabric." << endl;
        status = 1;
        goto Exit;
    }

    // Not management port
    if (anchorPortNum != 0) {
        // get the anchor port on the spec fabric - by port num:
        // we can do it only if we run from non management port because port0 isn't a phyisical port
        p_sPort = p_sNode->getPort(anchorPortNum);
        if (!p_sPort) {
            diag << "Fail to find anchor port:"
                    << anchorNodeName <<  anchorPortNum
                    << " in the specification fabric." << endl;
            status = 1;
            goto Exit;
        }
    }

    // Cleanup the flags we use for tracking matching and progress
    TopoCleanUpBeforeMerge(p_sFabric, p_dFabric);

    // Not management port
    if (anchorPortNum != 0) {
        // Do a BFS matching nodes from each fabrics. When this leaves
        // the appData1.ptr will cross point. status is 0 if there is no
        // difference
        status = TopoBFSAndMatchFromPorts(p_sPort, p_dPort, tmpDiag);
    } else {
        // Do a BFS matching nodes from each fabrics. When this leaves
        // the appData1.ptr will cross point. status is 0 if there is no
        // difference
        status = TopoBFSAndMatchFromNodes(p_sNode, p_dNode, tmpDiag);
    }
    if (status) {
        diag << "Topology Matching First Phase Found Mismatches:\n"
                << tmpDiag.str() << endl;
    }

    // Do the second step in matching - rely on preexisting matched nodes
    // and try to map the unmatched.
    TopoMatchSpecNodesByAdjacentNode(p_sFabric);

    if (TopoReportMismatches(p_sNode, p_dNode, diag, checkNodeNames))
        status = 1;

 Exit:
      string msg(diag.str());
      int msgLen = (int)strlen(msg.c_str());
      if (msgLen) {
          *messages = (char *)malloc(msgLen + 1);
          memcpy(*messages, msg.c_str(), msgLen);
          (*messages)[msgLen] = '\0';
      } else {
          *messages = NULL;
      }
      return(status);
}


//////////////////////////////////////////////////////////////////////////////
// Given a disc node pointer create the merged system and node by inspecting
// the merged spec fabric too.
// The node will have all it's ports setup (no links) and the system and
// sysPorts as well
static IBNode *
TopoCopyNodeToMergedFabric(IBFabric *p_mFabric,
        IBNode   *p_dNode)
{
    IBNode    *p_sNode = (IBNode *)p_dNode->appData1.ptr;
    IBSystem  *p_system = 0;
    IBSysPort *p_sysPort = 0;
    IBNode    *p_node = 0;

    string nodeName = p_dNode->name;
    string sysName =  p_dNode->p_system->name;
    string sysType =  p_dNode->p_system->type;

    // if we have a matching spec node we use that instead:
    if (p_sNode) {
        nodeName = p_sNode->name;
        sysName =  p_sNode->p_system->name;
        sysType =  p_sNode->p_system->type;
    }

    // make sure the system exists:
    // NOTE - we can not use the makeSystem of the fabric since it
    // will instantiate the entire system.
    p_system = p_mFabric->getSystem(sysName);
    if (!p_system) {
        p_system = new IBSystem(sysName, p_mFabric, sysType);
    }

    // create the node:
    p_node = p_mFabric->getNode(nodeName);
    if (! p_node) {
        p_node =
                p_mFabric->makeNode(nodeName, p_system, p_dNode->type,
                              p_dNode->numPorts);

        // copy extra info:
        p_node->guid_set(p_dNode->guid_get());
        p_node->system_guid_set(p_dNode->system_guid_get());
        p_node->devId  = p_dNode->devId;
        p_node->revId  = p_dNode->revId;
        p_node->vendId = p_dNode->vendId;

        if (p_sNode)
            p_node->attributes = p_sNode->attributes;

        if (p_dNode && p_dNode->attributes.size()) {
            if (p_node->attributes.size()) {
                p_node->attributes += string(",") + p_dNode->attributes;
            } else {
                p_node->attributes = p_dNode->attributes;
            }
        }

	// for swicthes need to copy port 0
	phys_port_t startPort = 1;
	if (p_dNode->type == IB_SW_NODE)
	  startPort = 0;

        // create the node ports:
        for (unsigned int i = startPort; i <=  p_dNode->numPorts; i++) {
            phys_port_t pn = (phys_port_t)i;
            // we only care if the discovered port exists:
            IBPort *p_dPort = p_dNode->getPort(pn);
            if (! p_dPort)
                continue;

            // some fabrics might be generated with all ports
            // pre-allocated.
            IBPort *p_port = p_node->getPort(pn);
            if (!p_port) p_port = new IBPort(p_node, pn);

            // copy some vital data:
            p_port->guid_set(p_dPort->guid_get());
            p_port->base_lid = p_dPort->base_lid;
            p_port->lmc = p_dPort->lmc;
	    for (lid_t b = 0; b < (lid_t)(1<< p_port->lmc);b++)
	      p_mFabric->setLidPort((lid_t)(p_port->base_lid + b), p_port);
	    if (p_mFabric->maxLid < p_port->base_lid + (lid_t)(1<<p_port->lmc) - 1)
	      p_mFabric->maxLid =
		(lid_t)(p_port->base_lid + (lid_t)(1<<p_port->lmc) - 1);

            // if there is a matching spec port we use it and get the
            // sys port name etc
            IBPort *p_sPort = 0;
            if (p_sNode) p_sPort = p_sNode->getPort(pn);
            if (p_sPort) {
                // if we had a spec sysPort make it:
                if (p_sPort->p_sysPort) {
                    p_sysPort = new IBSysPort(p_sPort->p_sysPort->name, p_system);
                    p_sysPort->p_nodePort = p_port;
                    p_port->p_sysPort = p_sysPort;
                } else {
                    p_port->p_sysPort = 0;
                }
            } else {
                // the discovered fabric might have a sysport?
                if (p_dPort->p_sysPort) {
                    p_sysPort = new IBSysPort(p_dPort->p_sysPort->name, p_system);
                    p_sysPort->p_nodePort = p_port;
                    p_port->p_sysPort = p_sysPort;
                } else {
                    p_port->p_sysPort = 0;
                }
            }
        }
    }
    return p_node;
}


// Copy the link from the p_dPort to it's remote port to the
// given merged port.
static void
TopoCopyLinkToMergedFabric(IBFabric *p_mFabric,
        IBPort   *p_mPort,
        IBPort   *p_dPort)
{
    // make sure the remote node exists or copy it to the merged fabric
    IBNode *p_remNode =
            TopoCopyNodeToMergedFabric(p_mFabric, p_dPort->p_remotePort->p_node);

    // find the remote port on the merged fabric
    IBPort *p_remPort = p_remNode->getPort(p_dPort->p_remotePort->num);
    if (!p_remPort ) {
        cerr << "-F- No Remote port:" << (unsigned int)p_dPort->p_remotePort->num
                << " on node:" << p_remNode->name << endl;
        exit(1);
    }

    // we need to create sys ports connection if both ports are on sys ports
    if (p_remPort->p_sysPort && p_mPort->p_sysPort) {
        p_remPort->p_sysPort->connect(p_mPort->p_sysPort,
                p_dPort->get_common_width(),
                p_dPort->get_common_speed());
    } else {
/*
        // The following error messages are actually not required at
        // all. If there is a mismatch on an internal node this will happen.
        if (p_remPort->p_sysPort) {
            cout << "-E- Linking:" << p_mPort->getName() << " to:"
                    << p_remPort->p_node->name
                    << "/" << p_dPort->p_remotePort->num << endl;
            cout << "-F- BAD LINK since Remote System port:"
                    << p_remPort->p_sysPort->name << " exists, but no sys port for:"
                    << p_mPort->getName() << endl;
            // exit(1);
        } else if (p_mPort->p_sysPort) {
            cout << "-F- BAD LINK since local has System port:"
                    << p_mPort->p_sysPort->name << " exists, but no sys port for:"
                    << p_remPort->getName() << endl;
            // exit(1);
        } else
*/
        IBLinkSpeed dPortSpeed = p_dPort->get_common_speed();
        IBLinkWidth dPortWidth = p_dPort->get_common_width();
        p_mPort->set_internal_data(dPortSpeed, dPortWidth, IB_PORT_STATE_ACTIVE);
        p_remPort->set_internal_data(dPortSpeed, dPortWidth, IB_PORT_STATE_ACTIVE);
        p_mPort->connect(p_remPort);
    }
}


//////////////////////////////////////////////////////////////////////////////
// Build a merged fabric from a matched discovered and spec fabrics:
// * Every node from the discovered fabric must appear
// * We use matched nodes and system names.
int
TopoMergeDiscAndSpecFabrics(IBFabric  *p_sFabric,
        IBFabric  *p_dFabric,
        IBFabric  *p_mFabric)
{
    // go through all nodes of the discovered fabric.
    // copy all their systems, and links...
    p_dFabric->setLidPort(0, NULL);
    p_dFabric->minLid = 1;
    p_dFabric->maxLid = 0;

    for (map_str_pnode::iterator nI = p_dFabric->NodeByName.begin();
            nI != p_dFabric->NodeByName.end(); nI++) {
        IBNode *p_dNode = (*nI).second;

        // make sure the node exists as well as the system and ports and
        // sysports:
        IBNode *p_node = TopoCopyNodeToMergedFabric(p_mFabric, p_dNode);

        // go over all ports and connect accordingly:
        for (unsigned int i = 1; i <=  p_dNode->numPorts; i++) {
            phys_port_t pn = (phys_port_t)i;
            IBPort *p_dPort = p_dNode->getPort(pn);
            IBPort *p_port = p_node->getPort(pn);

            // we got a remote connection:
            if (p_port && p_dPort && p_dPort->p_remotePort) {
                TopoCopyLinkToMergedFabric(p_mFabric, p_port, p_dPort);
            }
        }
    }
    p_mFabric->minLid = p_dFabric->minLid;
    return 0;
}


//----------------------------------------------------------------------
//----------------------------------------------------------------------
// Another approach for matching fabric starting on the edges and
// moving into the fabric. Using NAMES to match.
//----------------------------------------------------------------------
//----------------------------------------------------------------------
static int
isGUIDBasedName(IBNode *p_node) {
  string g = guid2str(p_node->guid_get());
  string n = p_node->name;
  const char *gs = g.c_str()+2;
  const char *ns = n.c_str()+1;
  //cout << "G:" << gs << endl << " N:" << ns << endl;
  return (!strncmp(gs,ns,6));
}

// parse the name mapping file from SPEC node name to discovered
// node-name or NodeGUID then if found new matches mark them
static int
TopoMatchByMappingFile(IBFabric *p_sFabric,
                       IBFabric *p_dFabric,
                       string mapFileName,
		       map< IBNode* , TopoMatchedBy> &specNodeMatchReason,
                       stringstream &s)
{
  ifstream mapFile(mapFileName.c_str());

  if (!mapFile.good()) {
    cout << "-E- Cannot open mapping file: " << mapFileName << endl;
    return(1);
  }

  regExp mapGUIDLine("^\\s*(\\S+)\\s+(0x[0-9a-fA-F]+)\\s*$");
  regExp mapNameLine("^\\s*(\\S+)\\s+(\\S+)\\s*$");
  rexMatch *p_rexRes;
  int anyMiss = 0;
  int numMatchedNodes = 0;
  unsigned int lineNum = 0;
  char sLine[1024];
  while (mapFile.good()) {
    mapFile.getline(sLine, 1024);
    lineNum++;
    IBNode *sNode = NULL;
    IBNode *dNode = NULL;
    if ((p_rexRes = mapGUIDLine.apply(sLine))) {
      sNode = p_sFabric->getNode(p_rexRes->field(1));
      if (!sNode)
        cout << "-W- Could not find topo node: '" << p_rexRes->field(1) << "'" << endl;
      uint64_t dguid = strtoull(p_rexRes->field(2).c_str(), NULL, 16);
      dNode = p_dFabric->getNodeByGuid(dguid);
      if (!dNode)
        cout << "-W- Could not find lst node by GUID: '" << p_rexRes->field(2) << "'" <<  endl;
      delete p_rexRes;
    } else if ((p_rexRes = mapNameLine.apply(sLine))) {
      sNode = p_sFabric->getNode(p_rexRes->field(1));
      if (!sNode)
        cout << "-W- Could not find topo node: '" << p_rexRes->field(1) << "'" << endl;
      dNode = p_dFabric->getNode(p_rexRes->field(2));
      if (!dNode)
        cout << "-W- Could not find lst node: '" << p_rexRes->field(2) << "'" << endl;
      delete p_rexRes;
    } else if (strlen(sLine)) {
      cout << "-W- Ignoring ilegal line (" << lineNum << ") :" << sLine << endl;
      continue;
    }

    if (sNode && dNode) {
      if (dNode->appData1.ptr || sNode->appData1.ptr)
        continue;
      TopoMarkNodeAsMatchAlgoVisited(dNode);
      TopoMarkMatcedNodes(dNode, sNode, numMatchedNodes);
      specNodeMatchReason[sNode] = MATCH_BY_USER;
    } else {
      anyMiss++;
    }
  }

  s << "-I- Matched " << numMatchedNodes << " nodes by mapping file" << endl;
  return(numMatchedNodes);
}

// Scan through all nodes and try to match them by the name
// return the number of name matches found
static int
TopoMatchByName(IBFabric *p_sFabric, IBFabric *p_dFabric,
                bool matchCAsByName, bool matchSWsByName,
                map< IBNode* , TopoMatchedBy> &specNodeMatchReason,
                stringstream &s, set_nodes_ptr &unParsableNameNodes,
                TopoResCounter_t &resCounter)
{
  int numMatchedNodesByName = 0;
  list<IBNode*> matchingDiscHosts;
  // we go through entire discovered fabric trying to find match by name
  for (map_str_pnode::iterator nI = p_dFabric->NodeByName.begin();
       nI != p_dFabric->NodeByName.end();
       nI++) {
    IBNode *p_node1 = (*nI).second;

    // skip if name matching not desired
    if ( (!matchSWsByName && (p_node1->type == IB_SW_NODE)) ||
         (!matchCAsByName && (p_node1->type != IB_SW_NODE)) ) continue;

    // skip previously matched nodes
    if (p_node1->appData1.ptr) continue;

    // so we try to find a node by the same name on the spec fabric.
    map_str_pnode::iterator snI = p_sFabric->NodeByName.find((*nI).first);

    // no match
    if (snI == p_sFabric->NodeByName.end()) {
      // we do not require switches to match by name - but we do require CAs to
      if (p_node1->type != IB_SW_NODE) {
        if (isGUIDBasedName(p_node1)) {
          s << "-W- Discovered Host GUID:"
            << guid2str(p_node1->guid_get())
            << " have non-parsible NodeDescription: '"
            << p_node1->description << "'"
            << endl;
          // keep track of non name parse-able nodes
          unParsableNameNodes.insert(p_node1);
          resCounter.warnings ++;
        }
      }
    } else {
      IBNode *p_node2 = (*snI).second;
      // make sure they are not previously match somehow
      if (p_node2->appData1.ptr)
        continue;
      TopoMarkNodeAsMatchAlgoVisited(p_node1);
      TopoMarkMatcedNodes(p_node1, p_node2, numMatchedNodesByName);
      specNodeMatchReason[p_node2] = MATCH_BY_NAME;
      if (p_node1->type != IB_SW_NODE)
        matchingDiscHosts.push_back(p_node1);
    }
  }

  return(numMatchedNodesByName);
}

// Go over all un-matched nodes
//   Inspect all their ports PN
//     if remote ports node are pre-matched
//       get the spec node connected to that remote port PotentialNode
//       if that node is on same system ignore it if port-num is not PN
//   If all PotentialNode are the same - declare a match
//   Otherwise report all potential nodes
// return the number of new matches found
static int
TopoMatchInferByConnectedMatchedNodes(IBFabric *p_sFabric,
                                      IBFabric *p_dFabric,
				      map< IBNode* , TopoMatchedBy> &specNodeMatchReason,
                                      stringstream &s)
{
  int numMatched = 0;

  // we go through entire discovered fabric trying to find match by name
  for (map_str_pnode::iterator nI = p_dFabric->NodeByName.begin();
       nI != p_dFabric->NodeByName.end();
       nI++) {
    IBNode *p_dNode = (*nI).second;

    // skip matched nodes
    if (p_dNode->appData1.ptr) continue;

    map< const IBNode*, list< IBPort*> > candidateSNodesLocalDPorts;

    // go over all discovered node ports
    for (unsigned int i = 1; i <= p_dNode->numPorts; i++) {
        phys_port_t pn = (phys_port_t)i;
      // we do not care for non discovered ports
      IBPort *p_dPort = p_dNode->getPort(pn);
      if (!p_dPort)
        continue;

      IBPort *p_dRemPort = p_dPort->p_remotePort;
      if (!p_dRemPort)
        continue;

      // if the remote discovered node not already matched - ignore it
      IBNode *p_dRemNode = p_dRemPort->p_node;
      IBNode *p_sRemNode = (IBNode *)p_dRemNode->appData1.ptr;
      if (!p_sRemNode) continue;

      // get the candidate sNode that is behind that rem-snode port
      IBPort *p_sRemPort = p_sRemNode->getPort(p_dRemPort->num);
      if (!p_sRemPort) continue;
      IBPort *p_sCandidatePort = p_sRemPort->p_remotePort;
      if (!p_sCandidatePort) continue;
      const IBNode *p_sCandidateNode = p_sCandidatePort->p_node;

      // can it be an already matched node?
      if (p_sCandidateNode->appData1.ptr) continue;

      // we require nodes on same system to match the connected port num
      if (p_sCandidatePort->num != pn) continue;

      // OK track that potential node and the current port num
      candidateSNodesLocalDPorts[p_sCandidateNode].push_front(p_dPort);
    }

    // if we collected just ONE other potentially matching sNode good!
    unsigned int numCandidates = (unsigned int)candidateSNodesLocalDPorts.size();
    if (numCandidates == 1) {
      const IBNode *p_csNode = (*candidateSNodesLocalDPorts.begin()).first;
      IBNode *p_sNode = p_sFabric->getNode(p_csNode->name);
      TopoMarkNodeAsMatchAlgoVisited(p_dNode);
      TopoMarkMatcedNodes(p_dNode, p_sNode, numMatched);
      specNodeMatchReason[p_sNode] = MATCH_BY_CONN;
    }
  }

  if (numMatched) {
    s << "-I- Successfuly matched " << numMatched << " more nodes" << endl;
  }
  return(numMatched);
}

static int
TopoCheckUnMatchingNames(IBNode *p_sNode,
                         IBNode *p_dNode)
{
    // currently only supports CA matching
    if (p_sNode->type != IB_CA_NODE || p_dNode->type != IB_CA_NODE)
        return 0;

    if (!strcmp(p_sNode->getName().c_str(), p_dNode->getName().c_str()))
        return 0;

    return 1;
}

// go over entire fabric from edges and in and report
// each mismatch - do not propogate inwards through mismatches
static void
TopoMatchReportMatchStatus(IBFabric *p_sFabric, IBFabric *p_dFabric,
                           string csvFileName, stringstream &s,
                           set_nodes_ptr &unParsableNameNodes,
                           TopoResCounter_t &resCounter)
{
  ofstream csvFile;

  list< IBNode *> missingSpecNodes;   // spec nodes not matched
  list< IBPort *> missingSpecCable;   // spec ports with no disc connection
  list< IBPort *> extraDiscCables;    // a dPort connected on empty sPort
  list< IBPort *> badDiscCables;      // a port connecting to wrong! node
  set< IBPort *> halfSwapSpecCables; // remote port on same node but different port
  list< IBNode *> extraDiscNodes;      // un-matched nodes of the discovered fabric
  list< IBNode *> mismatchedNames;    // matched nodes with mismatched names
  list< IBPort *> badSpecWidth;
  list< IBPort *> badSpecSpeed;
  unsigned int numMatchPorts = 0;
  bool doCsv = false;
  IBNode *p_sNode;

  if (csvFileName.size()) {
    csvFile.open(csvFileName.c_str());
    if (!csvFile.good()) {
      cout << "-E- Could not write into:" << csvFileName << endl;
    } else {
      doCsv = true;
    }
  }

  // go over all spec nodes
  for (map_str_pnode::iterator nI = p_sFabric-> NodeByName.begin();
       nI != p_sFabric-> NodeByName.end(); nI++) {
    p_sNode = (*nI).second;

    if (!p_sNode->appData1.ptr) {
        bool isConnected = true;
        //Not all specification HCA nodes should be connected
        //for example: HCA_2 with only one port connected
        if (p_sNode->type == IB_CA_NODE) {
            isConnected = false;
            for (unsigned int pn = 1; pn <= p_sNode->numRealPorts(); pn++) {
              IBPort *p_sPort = p_sNode->getPort((phys_port_t)pn);
              isConnected = (p_sPort && p_sPort->p_remotePort);
              if (isConnected)
                  break;
            }
        }
        if (isConnected)
            missingSpecNodes.push_back(p_sNode);

        continue;
    }

    IBNode *p_dNode = (IBNode *)p_sNode->appData1.ptr;

    // go over all ports of the two nodes and compare destinations
    // only complain about remote matching nodes not matching on the disc or port not matching
    // note we assume numPorts must match on the devices
    for (unsigned int i = 1; i <= p_sNode->numRealPorts(); i++) {
        phys_port_t pn = (phys_port_t)i;
      IBPort *p_sPort = p_sNode->getPort(pn);
      IBPort *p_dPort = p_dNode->getPort(pn);

      bool sConnected = (p_sPort && p_sPort->p_remotePort);
      bool dConnected = (p_dPort && p_dPort->p_remotePort);

      if (!sConnected && !dConnected) {
        continue;
      } else if (sConnected && !dConnected) {
        // avoid double rep is risky may lead to missing report
        //if (p_sPort < p_sPort->p_remotePort)
        // If the spec remote node is not matching it is reported as missing.
        // so avoid reporting all its cables as missing too.
        if (p_sPort->p_remotePort->p_node &&
            p_sPort->p_remotePort->p_node->appData1.ptr)
          missingSpecCable.push_back(p_sPort);
      } else if (!sConnected && dConnected) {
        // avoid double rep is risky may lead to missing report
        //if (p_dPort < p_dPort->p_remotePort)
        // If the discovered remote node is not matching it is reported as extra.
        // report all its extra cables too.
          extraDiscCables.push_back(p_dPort);
      } else {
        // both ports exist - check they match
        IBNode *p_sRemNode = p_sPort->p_remotePort->p_node;
        IBNode *p_dRemNode = p_dPort->p_remotePort->p_node;

        // If the discovered remote node is not matching it is reported as extra.
        // report all its extra cables too.
        if (!p_dRemNode->appData1.ptr) {
          extraDiscCables.push_back(p_dPort);
          continue;
        }

        // If the spec remote node is not matching it is reported as missing.
        // so avoid reporting all its cables as missing too.
        if (!p_sRemNode->appData1.ptr)
          continue;

        IBNode *p_sActRemNode = (IBNode *)p_dRemNode->appData1.ptr;
        // so they must be same nodes...
        if (p_sActRemNode != p_sRemNode) {
          badDiscCables.push_back(p_dPort);
          continue;
        }

        // now we know they both point to same node - check the port
        // connection and properties
        if (p_sPort->p_remotePort->num != p_dPort->p_remotePort->num) {
          halfSwapSpecCables.insert(p_sPort);
          continue;
        }
        numMatchPorts++;

        // so it is a perfect match cable - check width/speed
        IBLinkWidth sPortWidth = p_sPort->get_common_width();
        if ((sPortWidth != IB_UNKNOWN_LINK_WIDTH) &&
            (sPortWidth != p_dPort->get_common_width())) {
          badSpecWidth.push_back(p_sPort);
        }

        IBLinkSpeed sPortSpeed = p_sPort->get_common_speed();
        if ((sPortSpeed != IB_UNKNOWN_LINK_SPEED) &&
            (sPortSpeed != p_dPort->get_common_speed())) {
          badSpecSpeed.push_back(p_sPort);
        }
      } //  both spec and disc ports exist
    } // all ports
  } // all spec nodes

  // go over all disc nodes
  for (map_str_pnode::iterator nI = p_dFabric-> NodeByName.begin();
       nI != p_dFabric-> NodeByName.end(); nI++) {
    IBNode *p_dNode = (*nI).second;

    if (!p_dNode->appData1.ptr) {
      extraDiscNodes.push_back(p_dNode);
    }
    // report any mismatched names of matching nodes
    if (p_dFabric->subnCANames) {
        p_sNode = (IBNode *)p_dNode->appData1.ptr;
        set_nodes_ptr::iterator nonParseNode = unParsableNameNodes.find(p_dNode);
        // we don't want to report un-matched names for non name parse-able nodes
        if (nonParseNode == unParsableNameNodes.end() &&
            p_sNode && TopoCheckUnMatchingNames(p_sNode, p_dNode))
            mismatchedNames.push_back(p_dNode);
    }
  }

  list< IBNode *>::iterator nI;
  list< IBPort *>::iterator pI;

  // now format the output messages
  if (missingSpecNodes.size()) {
    s << "-E- Total: " << missingSpecNodes.size() << " missing spec nodes" << endl;
    if (doCsv)
      csvFile << "#Missing-Spec-Nodes Node-NAME" << endl;

    for (nI = missingSpecNodes.begin(); nI != missingSpecNodes.end(); nI++) {
      s << "-E- Missing spec node:" << (*nI)->name << endl;
      if (doCsv)
        csvFile << "Missing-Spec-Nodes, " << (*nI)->name << endl;
      resCounter.errors ++;
    }
    s << "-------------------------------------------------------------------" << endl;
    if (doCsv)
      csvFile << endl;
  }

  // avoid reporting swapped cables
  list< IBPort *> missingCableNotSwapped;
  for (pI = missingSpecCable.begin(); pI != missingSpecCable.end();pI++) {
    IBPort *p_sPort = (*pI);
    if ((halfSwapSpecCables.find(p_sPort) == halfSwapSpecCables.end()) &&
        (halfSwapSpecCables.find(p_sPort->p_remotePort) == halfSwapSpecCables.end())) {
      missingCableNotSwapped.push_back(*pI);
    }
  }

  if (missingCableNotSwapped.size()) {
    s << "-E- Total: " << missingCableNotSwapped.size() << " missing spec cables" << endl;
    if (doCsv)
      csvFile << "# Missing-cables, from-port, missing-cable-to-port" << endl;
    for (pI = missingCableNotSwapped.begin(); pI != missingCableNotSwapped.end(); pI++) {
      IBPort *p_sPort = (*pI);
      s << "-E- Missing cable between " << p_sPort->getName()
        << " and " << p_sPort->p_remotePort->getName() << endl;
      if (doCsv)
        csvFile << "Missing-cables, " << p_sPort->getName()
                << ", " << p_sPort->p_remotePort->getName() << endl;
      resCounter.errors ++;
    }
    s << "-------------------------------------------------------------------" << endl;
    if (doCsv)
      csvFile << endl;
  }

  if (badDiscCables.size()) {
    s << "-E- Total: " << badDiscCables.size() << " miss-connected cables" << endl;
    if (doCsv)
      csvFile << "# Bad-Cable, from-port, connects-to-port, instead-of-port " << endl;
    for (pI = badDiscCables.begin(); pI != badDiscCables.end(); pI++) {
      IBPort *p_dPort = (*pI);
      IBNode *p_dNode = p_dPort->p_node;
      IBNode *p_sNode = (IBNode*)p_dNode->appData1.ptr;
      if (!p_sNode) {
        cerr << "BUG in badDiscCables print" << endl;
        exit(1);
      }
      IBPort *p_sPort = p_sNode->getPort(p_dPort->num);
      IBNode *p_sActRemNode = (IBNode *)p_dPort->p_remotePort->p_node->appData1.ptr;
      IBPort *p_sActRemPort = p_sActRemNode->getPort(p_dPort->p_remotePort->num);

      s << "-E- Miss-connected cable from:" << p_sPort->getName()
        << " to:" << p_sActRemPort->getName()
        << " instead of: " << p_sPort->p_remotePort->getName() << endl;
      if (doCsv)
        csvFile << "Bad-Cable, " << p_sPort->getName()
                << ", " << p_sActRemPort->getName()
                << ", " << p_sPort->p_remotePort->getName() << endl;
      resCounter.errors ++;
    }
    s << "-------------------------------------------------------------------" << endl;
    if (doCsv)
      csvFile << endl;
  }

  if (halfSwapSpecCables.size()) {
    s << "-E- Total: " << halfSwapSpecCables.size() << " cables connect to correct node but different port"
      << endl;
    if (doCsv)
      csvFile << "# Swapped-cables, from-port to-port instead-of-port" << endl;
    set<IBPort*>::iterator sI;
    for (sI = halfSwapSpecCables.begin(); sI != halfSwapSpecCables.end(); sI++) {
      IBPort *p_sPort = (*sI);
      IBNode *p_sNode = p_sPort->p_node;
      IBNode *p_dNode = (IBNode*)p_sNode->appData1.ptr;
      IBPort *p_dPort = p_dNode->getPort(p_sPort->num);

      IBNode *p_repNode = p_dPort->p_remotePort->p_node;
      IBPort *p_sRepPort = NULL;
      if (p_repNode->appData1.ptr) {
        p_repNode =
          (IBNode*)p_dPort->p_remotePort->p_node->appData1.ptr;
        p_sRepPort = p_repNode->getPort(p_dPort->p_remotePort->num);
      }

      if (p_sRepPort) {
        s << "-E- Swapped cable from:" << p_sPort->getName()
          << " to:" << p_sRepPort->getName()
          << " instead of: " << p_sPort->p_remotePort->getName() << endl;
        if (doCsv)
          csvFile << "Swapped-cables, " << p_sPort->getName()
                  << ", " << p_sRepPort->getName()
                  << ", " << p_sPort->p_remotePort->getName() << endl;
      } else {
        s << "-E- Swapped cable from:" << p_sPort->getName()
          << " to:" << p_repNode->name
          << "/P" << (unsigned int)p_dPort->p_remotePort->num
          << " instead of: " << p_sPort->p_remotePort->getName() << endl;
        if (doCsv)
          csvFile << "Swapped-cables, " << p_sPort->getName()
                  << ", " << p_repNode->name
                  << "/P" << (unsigned int)p_dPort->p_remotePort->num
                  << ", " << p_sPort->p_remotePort->getName() << endl;
      }
      resCounter.errors ++;
    }
    s << "-------------------------------------------------------------------" << endl;
    if (doCsv)
      csvFile << endl;
  }

  // avoid reporting swapped cables
  list< IBPort *> extraCablesNotSwapped;
  for (pI = extraDiscCables.begin(); pI != extraDiscCables.end();pI++) {
    IBPort *p_dPort = (*pI);
    IBNode *p_dNode = p_dPort->p_node;
    IBNode *p_sNode = (IBNode*)p_dNode->appData1.ptr;
    IBPort *p_sPort = p_sNode->getPort(p_dPort->num);
    IBNode *p_dRemNode = p_dPort->p_remotePort->p_node;
    IBNode *p_sRemNode = (IBNode*)p_dRemNode->appData1.ptr;
    IBPort *p_sRemPort = NULL;
    if (p_sRemNode)
      p_sRemPort = p_sRemNode->getPort( p_dPort->p_remotePort->num);

    if (halfSwapSpecCables.find(p_sPort) != halfSwapSpecCables.end())
      continue;

    if (p_sRemPort && (halfSwapSpecCables.find(p_sRemPort) != halfSwapSpecCables.end()))
      continue;

    extraCablesNotSwapped.push_back(*pI);
  }

  if (extraCablesNotSwapped.size()) {
    s << "-W- Total: " << extraCablesNotSwapped.size() << " extra cables discovered"
      << endl;
    if (doCsv)
      csvFile << "# Extra-Cable, from-port, to-port" << endl;
    for (pI = extraCablesNotSwapped.begin(); pI != extraCablesNotSwapped.end(); pI++) {
      IBPort *p_dPort = (*pI);
      IBNode *p_dNode = p_dPort->p_node;
      IBNode *p_sNode = (IBNode*)p_dNode->appData1.ptr;
      IBPort *p_sPort = p_sNode->getPort(p_dPort->num);

      // we want to use as much "match" info in the report:
      IBNode *p_repNode = p_dPort->p_remotePort->p_node;
      IBPort *p_sRepPort = NULL;
      if (p_repNode->appData1.ptr) {
        p_repNode =
          (IBNode*)p_dPort->p_remotePort->p_node->appData1.ptr;
        p_sRepPort = p_repNode->getPort(p_dPort->p_remotePort->num);
      }
      if (p_sPort) {
        s << "-W- Extra cable between " << p_sPort->getName()
            << " and ";
         if (doCsv)
          csvFile << "Extra-cable, " << p_sPort->getName() << ", ";

      } else {
        s << "-W- Extra cable between discovered "
          << p_dNode->getName() << "/P" << (unsigned int)p_dPort->num
          << " and ";
         if (doCsv)
          csvFile << "Extra-cable, " << p_dNode->getName()
                  << "/P" << (unsigned int)p_dPort->num << ", ";
      }
      if (p_sRepPort) {
        s << p_sRepPort->getName() << endl;
        if (doCsv)
          csvFile << p_sRepPort->getName() << endl;
      } else {
        s << " discovered " << p_repNode->name
          << "/P" << (unsigned int)p_dPort->p_remotePort->num << endl;
        if (doCsv)
          csvFile << p_repNode->name
                  << "/P" << (unsigned int)p_dPort->p_remotePort->num << endl;
      }

      resCounter.warnings ++;
    }
    s << "-------------------------------------------------------------------" << endl;
    if (doCsv)
      csvFile << endl;
  }

  if (extraDiscNodes.size()) {
    s << "-E- Total: " << extraDiscNodes.size() << " extra discovered nodes" << endl;
    if (doCsv)
      csvFile << "# Extra-Node, NodeName" << endl;
    for (nI = extraDiscNodes.begin(); nI != extraDiscNodes.end(); nI++) {
      s << "-E- Discovered extra node node:" << (*nI)->name << endl;
      if (doCsv)
        csvFile << "Extra-Node, " << (*nI)->name << endl;
      resCounter.errors ++;
    }
    s << "-------------------------------------------------------------------" << endl;
    if (doCsv)
      csvFile << endl;
  }

  if (badSpecWidth.size()) {
    s << "-W- Total: " << badSpecWidth.size() << " wrong width cables discovered" << endl;
    if (doCsv)
      csvFile << "# Bad-Width, port-name, expected-width, actual-width" << endl;
    for (pI = badSpecWidth.begin(); pI != badSpecWidth.end(); pI++) {
      IBPort *p_sPort = (*pI);
      IBNode *p_sNode = p_sPort->p_node;
      IBNode *p_dNode = (IBNode*)p_sNode->appData1.ptr;
      IBPort *p_dPort = p_dNode->getPort(p_sPort->num);
      s << "-W- Wrong link width on:" << p_sPort->getName()
        << ". Expected:" << width2char(p_sPort->get_common_width())
        << " got:" << width2char(p_dPort->get_common_width()) << endl;
      if (doCsv)
        csvFile << "Bad-Width, " << p_sPort->getName()
                << ", " << width2char(p_sPort->get_common_width())
                << ", " << width2char(p_dPort->get_common_width()) << endl;
      resCounter.warnings ++;
    }
    s << "-------------------------------------------------------------------" << endl;
    if (doCsv)
      csvFile << endl;
  }

  if (badSpecSpeed.size()) {
    s << "-W- Total: " << badSpecSpeed.size() << " wrong speed cables discovered" << endl;
    if (doCsv)
      csvFile << "# Bad-Speed, port-name, expected-speed, actual-speed" << endl;
    for (pI = badSpecSpeed.begin(); pI != badSpecSpeed.end(); pI++) {
      IBPort *p_sPort = (*pI);
      IBNode *p_sNode = p_sPort->p_node;
      IBNode *p_dNode = (IBNode*)p_sNode->appData1.ptr;
      IBPort *p_dPort = p_dNode->getPort(p_sPort->num);
      s << "-W- Wrong link speed on:" << p_sPort->getName()
        << ". Expected:" << speed2char(p_sPort->get_common_speed())
        << " got:" << speed2char(p_dPort->get_common_speed()) << endl;
      if (doCsv)
        csvFile << "Bad-Speed, " << p_sPort->getName()
                << ", " << speed2char(p_sPort->get_common_speed())
                << ", " << speed2char(p_dPort->get_common_speed()) << endl;
      resCounter.warnings ++;
    }
    s << "-------------------------------------------------------------------" << endl;
    if (doCsv)
      csvFile << endl;
  }

  if (mismatchedNames.size()) {
      s << "-W- Total: " << mismatchedNames.size() << " mismatched names discovered" << endl;
      if (doCsv)
        csvFile << "# Mismatched node names, expected , discovered" << endl;
      for (nI = mismatchedNames.begin(); nI != mismatchedNames.end(); nI++) {
        p_sNode = (IBNode*)(*nI)->appData1.ptr;
        s << "-W- Matching nodes have different names. Expected node name: " << p_sNode->getName() << ", Discovered node name: " << (*nI)->getName() << endl;
        if (doCsv)
          csvFile << "Mismatched node names, " << p_sNode->getName() << ", " << (*nI)->getName() << endl;
        resCounter.warnings ++;
      }
      s << "-------------------------------------------------------------------" << endl;
      if (doCsv)
        csvFile << endl;
  }

  if (doCsv)
    csvFile.close();

  s << "-I- Total: " << numMatchPorts << " fully matching ports found" << endl;
}

int
TopoMatchWriteMapFile(IBFabric *p_sFabric,      // The specification fabric
		      IBFabric *p_dFabric,      // The discovered fabric
		      map< IBNode* , TopoMatchedBy> &specNodeMatchReason,
		      string outMapFileName,
		      stringstream &s)
{
  ofstream csvFile;
  string err_message;
  int rc = IBFabric::OpenFile(outMapFileName.c_str(), csvFile, false, err_message);
  if (rc) {
      cout << "-E- " << err_message << endl;
      return rc;
  }

  csvFile << "# Match-Reason, Topo-Name, LST-NodeGUID, LST-Name/NodeDesc" << endl;

  // go over all spec nodes
  for (map_str_pnode::iterator nI = p_sFabric-> NodeByName.begin();
       nI != p_sFabric-> NodeByName.end(); nI++) {
    IBNode *p_sNode = (*nI).second;
    IBNode *p_dNode = (IBNode*)p_sNode->appData1.ptr;

    if (!p_dNode)
      continue;

    map< IBNode* , TopoMatchedBy>::const_iterator mI =
      specNodeMatchReason.find(p_sNode);
    if (mI == specNodeMatchReason.end()) {
      cerr << "BUG: Missing specNodeMatchReason for matched node:"
	   << p_sNode->name << endl;
      return(1);
    }

    csvFile << MatchByToStr((*mI).second) << ", "
	    << p_sNode->name << ", "
	    << guid2str(p_dNode->guid_get()) << ", "
	    << p_dNode->name << endl;
  }
  csvFile.close();
  s << "-I- Topo match map written to:" << outMapFileName << endl;
  return(0);
}

// Return 0 if fabrics match, number of warnings and errors otherwise.
// The detailed compare messages are returned in messages
int
TopoMatchFabricsFromEdge(IBFabric *p_sFabric,      // The specification fabric
                         IBFabric *p_dFabric,      // The discovered fabric
                         const TopoMatchOptions &opts,    // options
                         char **messages)
{
  stringstream diag, tmpDiag;
  unsigned int stage = 0;
  map< IBNode* , TopoMatchedBy> specNodeMatchReason;
  TopoResCounter_t resCounter;

  // Cleanup the flags we use for tracking matching and progress
  TopoCleanUpBeforeMerge(p_sFabric, p_dFabric);

  diag << "-I- Performing Topology Matching ..." << endl;
  diag << "-----------------------------------------------------------------"
       << endl;

  if (opts.mapFileName.size()) {
    diag << "-I- Stage " << ++stage
         << ": Pair nodes by a user-given mapping of topo-name and NodeDesc/NodeGUID ..."
         << endl;
    int numMatched =
      TopoMatchByMappingFile(p_sFabric, p_dFabric, opts.mapFileName,
			     specNodeMatchReason, diag);
    diag << "-I- Matched " << numMatched << " nodes by mapping file:"
	 << opts.mapFileName << endl;
    diag << "-------------------------------------------------------------------"
         << endl;
  }

  diag << "-I- Stage " << ++stage
       << ": Pair nodes by their topo-name and NodeDesc..." << endl;
  // to keep track of non name parse-able nodes
  set_nodes_ptr unParsableNameNodes;
  int numMatched =
    TopoMatchByName(p_sFabric, p_dFabric, opts.matchCAsByName,
                    opts.matchSWsByName, specNodeMatchReason, diag, unParsableNameNodes, resCounter);
  diag << "-I- Matched " << numMatched << " nodes by name." << endl;
  diag << "-------------------------------------------------------------------"
       << endl;

  unsigned int subStage = 0;
  ++stage;
  do {
    diag << "-I- Stage " << stage << "." << ++subStage
         << " Matching more nodes if connected to previously matched nodes ..."
	 << endl;
    numMatched =
      TopoMatchInferByConnectedMatchedNodes(p_sFabric, p_dFabric,
					    specNodeMatchReason, diag);
  } while (numMatched);

  // produce final matching info report
  diag << "-------------------------------------------------------------------"
       << endl;
  diag << "-I- Stage " << ++stage << ": Final reports ...." << endl;

  TopoMatchReportMatchStatus(p_sFabric, p_dFabric, opts.csvFileName, diag, unParsableNameNodes, resCounter);
  diag << "-----------------------------------------------------------------"
       << endl;

  // create mapping file
  if (opts.outMapFileName.size()) {
    TopoMatchWriteMapFile(p_sFabric, p_dFabric,
			  specNodeMatchReason,
			  opts.outMapFileName,
			  diag);
  }

  string msg(diag.str());
  int msgLen = (int)strlen(msg.c_str());
  if (msgLen) {
    *messages = (char *)malloc(msgLen + 1);
    memcpy(*messages, msg.c_str(), msgLen);
    (*messages)[msgLen] = '\0';
  } else {
    *messages = NULL;
  }

  return(resCounter.errors + resCounter.warnings);
}
