/*
 * Copyright (c) 2004-2019 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <iomanip>

#include <infiniband/ibdm/TraceRoute.h>

// Trace a direct route from the given SM node port
int TraceDRPathRoute(IBPort *p_smNodePort, list_phys_ports drPathPortNums)
{
    IBPort *p_port = p_smNodePort;
    IBNode *p_node;
    IBPort *p_remPort;
    IBNode *p_remNode;
    IBPort *p_nextPort;
    phys_port_t outPortNum;
    unsigned int hop = 0;
    list_phys_ports::iterator pI = drPathPortNums.begin();

    // check we haev at least two hops:
    if (drPathPortNums.size() < 2) {
        cout << "-E- We expect at least 2 hops to get out of the SM Node! (got:"
                << drPathPortNums.size() << ")" << endl;
        return(1);
    }

    // we require it to start with zero - marking the internal hop

    if (*pI != 0) {
        cout << "-E- Direct Route Ports list must start with a zero !" << endl;
        return(1);
    }

    // The second port in the we require it to start with zero - marking the internal hop
    pI++;
    if (*pI != p_port->num) {
        cout << "-E- Direct Route Ports second hop must match the SM port number !"
                << " (got:" << *pI << ")" << endl;
        return(1);
    }

    // we are at the port of the SM node so we are at hop 1.
    hop = 1;
    pI++; // points to next hop in the list

    // we traverse to null port or max number of hops reached.
    while (p_port && (hop < drPathPortNums.size())) {
        // get the node
        p_node = p_port->p_node;

        if (p_port->p_sysPort) {
            // is external port:
            cout << "[" << setw(3) <<hop << "] FROM Host:" << p_node->p_system->name.c_str()
                    << " Plug:" << p_port->p_sysPort->name.c_str() << endl;
            cout << "           Node:"  << p_node->name.c_str() << " Port:" << p_port->num << endl;
        } else {
            // internal so provide only internal data:
            cout << "[" << setw(3) << hop << "] FROM Node:"  << p_node->name.c_str() << " Port:"
		 << (unsigned int)p_port->num << endl;
        }

        // we calc next step port:
        p_nextPort = NULL;

        // start going to the other side:
        p_remPort = p_port->p_remotePort;
        if (p_remPort != NULL) {
            p_remNode = p_remPort->p_node;

            // The to section:
            if (p_remPort->p_sysPort) {
                // is external port:
                cout << "      TO   Host:" << p_remNode->p_system->name.c_str()
                        << " Plug:" << p_remPort->p_sysPort->name.c_str() << endl;
                cout << "           Node:"  << p_remNode->name.c_str() << " Port:" << p_remPort->num << endl;
            } else {
                // internal so provide only internal data:
                cout << "      TO   Node:"  << p_remNode->name.c_str() << " Port:" << p_remPort->num << endl;
            }

            if (pI != drPathPortNums.end()) {
                // calc the next port
                outPortNum = *pI;
                // we could get stuck on bad port num
                if (outPortNum > p_remNode->numPorts) {
                    cout << "-E- Bad port number:" << outPortNum << " hop:" << hop << endl;
                    return 1;
                }

                // port not connected
                if (p_remNode->getPort(outPortNum) == NULL) {
                    cout << "[" << setw(3) << hop << "] Broken Route: not connected port:" << outPortNum << endl;
                    return 1;
                }

                p_nextPort = p_remNode->getPort(outPortNum);
            }

            pI++;
            hop++;
            }
            p_port = p_nextPort;
    }
    return 0;
}

typedef set< IBNode *, less< IBNode * > > set_p_node;

// Trace the path between the lids based on min hop count only
int
TraceRouteByMinHops(IBFabric *p_fabric,
        lid_t slid , lid_t dlid)
{
    IBPort *p_port = p_fabric->getPortByLid(slid), *p_remPort, *p_nextPort;
    IBPort *p_dstPort = p_fabric->getPortByLid(dlid);
    IBNode *p_node, *p_remNode;
    unsigned int hop = 0;
    set_p_node visitedNodes;

    // make sure:
    if (! p_port) {
        cout << "-E- Provided source:" << slid
                << " lid is not mapped to a port!" << endl;
        return(1);
    }

    if (!p_dstPort) {
        cout << "-E- Provided destination:" << dlid
                << " lid is not mapped to a port!" << endl;
        return(1);
    }

    // find the physical port we out from:
    p_port = p_port->p_node->getFirstMinHopPort(dlid);
    if (!p_port) {
        cout << "-E- Fail to obtain minhop port for switch:" << slid << endl;
        return(1);
    }

    cout << "--------------------------- TRACE PATH BY MIN HOPS -------------------------" << endl;
    cout << "-I- Tracing by Min Hops from lid:" << slid
            << " to lid:" << dlid << endl;
    // we traverse to target dlid
    while (p_port) {
            // get the node
            p_node = p_port->p_node;

            if (p_port->p_sysPort) {
                // is external port:
                cout << "[" << setw(3) << hop << "] FROM Host:" << p_node->p_system->name.c_str()
                        << " Plug:" << p_port->p_sysPort->name.c_str() << endl;
                cout << "           Node:"  << p_node->name.c_str() << " Port:" << (unsigned int)p_port->num << endl;
            } else {
                // internal so provide only internal data:
                cout << "[" << setw(3) << hop << "] FROM Node:"  << p_node->name.c_str() << " Port:"
                        << (unsigned int)p_port->num << endl;
            }

            // we calc next step port:
            p_nextPort = NULL;

            // start going to the other side:
            p_remPort = p_port->p_remotePort;
            if (p_remPort != NULL) {
                p_remNode = p_remPort->p_node;

                // did we already visit this node?
                set_p_node::iterator sI = visitedNodes.find(p_remNode);
                if (sI != visitedNodes.end()) {
                    cout << "-E- Run into loop in min hop path at node:" << p_remNode->name << endl;
                    return 1;
                }
                visitedNodes.insert(p_remNode);

                // The to section:
                if (p_remPort->p_sysPort) {
                    // is external port:
                    cout << "      TO   Host:" << p_remNode->p_system->name.c_str()
                            << " Plug:" << p_remPort->p_sysPort->name.c_str() << endl;
                    cout << "           Node:"  << p_remNode->name.c_str() << " Port:" << (unsigned int)p_remPort->num << endl;
                } else {
                    // internal so provide only internal data:
                    cout << "      TO   Node:"  << p_remNode->name.c_str() << " Port:" << (unsigned int)p_remPort->num << endl;
                }

                // we need next port only if we are on a switch:
                if (p_remNode->type == IB_SW_NODE) {
                    p_nextPort = p_remNode->getFirstMinHopPort(dlid);
                } else {
		  // must have reached destination
		  if (p_remNode != p_dstPort->p_node) {
                    cout << "-E- got to node:" << p_remNode->name << " instead of:" << p_dstPort->p_node->name << endl;
                    return 1;
		  }
		}
                hop++;
            }
            p_port = p_nextPort;
    }
    cout << "---------------------------------------------------------------------------\n" << endl;
    return 0;
}

// Trace a route from slid to dlid by LFT
int TraceRouteByLFT(IBFabric *p_fabric,
        lid_t sLid , lid_t dLid,
        unsigned int *hops,
        list_pnode *p_nodesList,
        list_phys_ports *p_outPortsList,
        bool useVL15)
{
    IBPort *p_port = p_fabric->getPortByLid(sLid);
    IBNode *p_node;
    IBPort *p_remotePort;
    int hopCnt = 0;

    // make sure:
    if (! p_port) {
        cout << "-E- Provided source:" << sLid
                << " lid is not mapped to a port!" << endl;
        return(1);
    }

    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {
        cout << "--------------------------- TRACE PATH BY FDB -----------------------------" << endl;
        cout << "-V- Tracing from lid:" << sLid << " to lid:"
                << dLid << endl;
    }

    if (hops) *hops = 0;

    sl_vl_t slvl;
    if (useVL15) {
        slvl.SL = 0;
        slvl.VL = 15;
    } else {
        slvl.SL = p_port->p_node->getPSLForLid(dLid);
        slvl.VL = slvl.SL; //for in port 0 use SL even in VL2VL is set
        if ( slvl.SL == IB_SLT_UNASSIGNED ) {
            cout << "-E- Failed to get SL for node:" << p_port->p_node->name
                     << " dlid:" <<  dLid << endl;
            return 1;
        }
    }

    // if the port is not a switch - go to the next switch:
    if (p_port->p_node->type != IB_SW_NODE) {
        // try the next one:
        if (!p_port->p_remotePort) {
            cout << "-E- Provided starting point is not connected !"
                    << "lid:" << sLid << endl;
            return 1;
        }

        //Set VL for CA node
        if (!useVL15) {
            slvl.VL = p_port->p_node->getVL(0, p_port->num, slvl);
            if (slvl.VL == IB_SLT_UNASSIGNED) {
                cout << "-E- Failed to get VL for node:" << p_port->p_node->name
                    << " inPort:0 outPort:"<< (int)p_port->num
                    << " SL:" << (int)slvl.SL <<   endl;
                return(1);
            }
            if (slvl.VL == IB_DROP_VL) {
                cout << "-E- Dead end at:" << p_port->p_node->name
                    << " Drop VL inPort:0 outPort:"<< (int)p_port->num
                    << " SL:" << (int)slvl.SL <<   endl;
                return 1;
            }
        }

        p_node = p_port->p_remotePort->p_node;
        p_port = p_port->p_remotePort;

        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            cout << "-V- Arrived at Node:" << p_node->name
                  << " Port:" << (unsigned int)p_port->num << endl;
    } else {
        // it is a switch :
        p_node = p_port->p_node;
    }


    // verify we are finally of a switch:
    if (p_node->type != IB_SW_NODE) {
            cout << "-E- Provided starting point is not connected to a switch !"
                    << "lid:" << sLid << endl;
            return 1;
    }

    // traverse:
    int done = 0;
    while (!done) {

        // calc next node:
        phys_port_t pn = p_node->getLFTPortForLid(dLid, p_port->num, slvl);
        if (pn == IB_LFT_UNASSIGNED) {
            cout << "-E- Unassigned LFT for lid:" << dLid
                << " Dead end at:" << p_node->name
                << " PLFT:" << (int)p_node->getPLFTMapping(p_port->num, slvl);

            if ( FabricUtilsVerboseLevel & FABU_LOG_VERBOSE ) {
                char plft_line[1024];
                p_node->getPLFTMapping(p_port->num, plft_line);
                cout   << " sl/vl:" << (int)p_node->getUsedSLOrVL(slvl)
                    << " in port:" << (int)p_port->num
                    << " plft mapping:" << plft_line;
            }
            cout << endl;
            return 1;
        }

        // if the port number is 0 we must have reached the target
        if (pn == 0) {
            uint16_t base_lid = 0;
            // get lid of any port of this node
            // use unsigned int index to prevent overwrap
            for (unsigned int portNum = 0;
                    (base_lid == 0) && (portNum <= p_node->numPorts); portNum++) {
                IBPort *p_port = p_node->getPort((phys_port_t)portNum);
                if (p_port) {
                    base_lid = p_port->base_lid;
                    break;
                }
            }
            if (base_lid == 0) {
                cout << "-E- Fail to find node:" << p_node->name
                        << " base lid?" << endl;
                return 1;
            }

            lid_t lidStep = (lid_t)(1 << p_port->lmc);
            if ((base_lid > dLid) || (base_lid + lidStep - 1 < dLid)) {
                cout << "-E- Dead end at port 0 of node:" << p_node->name << endl;
                return 1;
            }
            return 0;
        }

        // insert the node pointer to the list
        if (p_nodesList && p_outPortsList) {
            p_nodesList->push_back(p_node);
            p_outPortsList->push_back(pn);
        }

        //set out VL to be the next VL
        if (!useVL15) {
            slvl.VL = p_node->getVL(p_port->num, pn, slvl);
            if (slvl.VL == IB_SLT_UNASSIGNED) {
                cout << "-E- Failed to get VL for node:" << p_port->p_node->name
                    << " inPort:" << (int)p_port->num
                    << " outPort:"<< (int)pn
                    << " SL:" << (int)slvl.SL <<   endl;
                return(1);
            }
            if (slvl.VL == IB_DROP_VL) {
                cout << "-E- Dead end at:" << p_port->p_node->name
                    << " Drop VL inPort:" << (int)p_port->num
                    << " outPort:"<< (int)pn
                    << " SL:" << (int)slvl.SL <<   endl;
                return 1;
            }
        }

        // get the port on the other side
        p_port = p_node->getPort(pn);

        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            cout << "-V- Going out on port:" << (unsigned int)pn << endl;

        if (! (p_port &&
                p_port->p_remotePort &&
                p_port->p_remotePort->p_node)) {
            cout << "-E- Dead end at:" << p_node->name << endl;
            return 1;
        }

        p_remotePort = p_port->p_remotePort;
        // check if we are done:
        lid_t lidStep = (lid_t)(1 << p_remotePort->lmc);
        done = ((p_remotePort->base_lid <= dLid) &&
                    (p_remotePort->base_lid+lidStep - 1 >= dLid));

        if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            cout << "-V- Arrived at Node: " << p_port->p_remotePort->p_node->name
		 << " Port: " << (unsigned int)p_port->p_remotePort->num
		 << " lid: " << p_remotePort->base_lid << " lmc: " << (unsigned int)p_remotePort->lmc
		 << endl;

        p_node = p_remotePort->p_node;
        p_port = p_remotePort;

        if (hops) (*hops)++;
        if (hopCnt++ > 256) {
            cout << "-E- Aborting after 256 hops - loop in LFT?" << endl;
            return 1;
        }

        // check if the dlid is a vlid on the phys port we reached.
        if (!done && p_node->type == IB_CA_NODE) {
            map_vportnum_vport::iterator vports_iter = p_port->VPorts.begin();
            for ( ; vports_iter != p_port->VPorts.end(); vports_iter++) {

                IBVPort *cur_vport = vports_iter->second;
                if (!cur_vport) {
                    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
                        cout << "-E- Unexpected null vport in the vports map" << endl;
                    continue;
                }

                if (cur_vport->get_vlid() == dLid) {
                    // we found the destination vport lid.
                    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
                        cout << "-V- Destination vlid found."
                             << " vPort: " << cur_vport->getVPortNum()
                             << " vlid: " << dLid
                             << endl;
                    return 0;
                }
            }
        }
    }

    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
        cout << "---------------------------------------------------------------------------\n" << endl;
    return 0;
}
