/*
 * Copyright (c) 2004-2013 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
 
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <regex.h>
#include <sstream>
#include <map>
#include <list>
#include <vector>
#include <iostream>

using namespace std;

// comparison strings by the number they rep
bool compareIntStr (string first, string second)
{
    unsigned int a, b;
    a = atoi(first.c_str());
    b = atoi(second.c_str());
    return(a < b);
}

// get a list of numbers and generate a nice consolidated sub list
string groupNumRanges(string prefix, string suffix,
                      list<string> &numStrs)
{

    ostringstream s;
    size_t n = numStrs.size();
    if ( n < 1 ) {
        return(prefix + suffix);
    } else if ( n == 1 ) {
        return(prefix + numStrs.front() + suffix);
    }

    s << prefix << "[";

    // convert the strings into numbers removing leading zeros
    numStrs.sort(compareIntStr);

    string startStr = numStrs.front();
    unsigned int start = atoi(startStr.c_str());
    string endStr = startStr;
    unsigned int end = start;
    for ( list<string>::iterator nI = numStrs.begin();
        nI != numStrs.end(); nI++ ) {

        if ( atoi((*nI).c_str()) > (int)end + 1 ) {
            // a space after the end
            if ( start == end ) {
                s << endStr << ",";
            } else {
                s << startStr << ".." << endStr << ",";
            }
            endStr = startStr = *nI;
            end = start = atoi(startStr.c_str());
        } else {
            endStr = *nI;
            end = atoi(endStr.c_str());
        }
    }

    if ( start == end ) {
        s << endStr;
    } else {
        s << startStr << ".." << endStr;
    }
    s << "]" << suffix;
    return(s.str());
}

// process every group by splitting it to pre num and post
// then look for matches and build next groups
// return true if more iterations are required
bool
groupingEngine (map<string, int> &groupFirstIdx)
{
    regex_t rex;
    regmatch_t matches[4];
    regcomp(&rex, "([^0-9]*)([0-9]+)(.*)", REG_EXTENDED);
    map<pair<string, string>, list<string> > newGroups;
    map<string, int> newGroupFirstIdx;

    list<string> res;
    for ( map<string, int>::iterator gI = groupFirstIdx.begin();
        gI != groupFirstIdx.end(); gI++ ) {
        string word = (*gI).first;
        int idx = (*gI).second;

        // try to find a number on the right of the idx:
        string prefix = word.substr(0, idx);
        string suffix = word.substr(idx);

        if ( regexec(&rex, suffix.c_str(), 4, matches, 0) ) {
            newGroupFirstIdx[word] = idx;
            continue;
        }

        prefix +=
        suffix.substr(matches[1].rm_so, matches[1].rm_eo - matches[1].rm_so);
        string num =
        suffix.substr(matches[2].rm_so, matches[2].rm_eo - matches[2].rm_so);
        string newSuffix =
        suffix.substr(matches[3].rm_so, matches[3].rm_eo - matches[3].rm_so);

        pair<string,string> key(prefix, newSuffix);
        newGroups[key].push_back(num);
    }

    // go over all new groups and see if we can collapse them:
    bool anyMore = false;
    map<pair<string, string>, list<string> >::iterator gI;
    for ( gI = newGroups.begin(); gI != newGroups.end(); gI++ ) {
        string prefix = (*gI).first.first;
        string suffix = (*gI).first.second;
        if ( suffix.size() ) anyMore = true;
        string newGroupStr = groupNumRanges(prefix, suffix, (*gI).second);
        newGroupFirstIdx[newGroupStr] = (int)(newGroupStr.size() - suffix.size());
    }
    groupFirstIdx.clear();
    groupFirstIdx = newGroupFirstIdx;
    regfree(&rex);
    return(anyMore);
}

// Algorithm:
// split the given words on the first number
// then group on the leading string and sort for continoues
string compressNames (list<string> &words)
{
    map<string, int> groupFirstIdx;

    // we need to prepare the first stage which is a list of words and
    // simply the index 0 for were to start the search for integers
    for ( list<string>::iterator wI = words.begin();
        wI != words.end(); wI++ ) {
        groupFirstIdx[*wI] = 0;
    }

    // now call the grouping engine until no refinement
    bool done;
    do {
        done = !groupingEngine(groupFirstIdx);
    } while ( !done );

    ostringstream s;
    for ( map<string, int>::iterator gI = groupFirstIdx.begin();
        gI != groupFirstIdx.end(); gI++ ) {
        s << (*gI).first << " ";
    }
    return(s.str());
}

#ifdef TEST_VECTORIZE
int main(int argc, char **argv)
{
    list<string> names;
    for ( unsigned int i = 1; i < argc; i++ ) {
        names.push_back(argv[i]);
    }
    cout << compressNames(names) << endl;
    return(0);
}

#endif // TEST_VECTORIZE
