/*
 * Copyright (c) 2004-2010 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


%{
    /*
     * header section
     */
    #pragma warning(disable:4242)
    #include <stdio.h>
    #include <stdlib.h>
    #include <string.h>
    
    #include <infiniband/ibdm/SysDef.h>
    #include <infiniband/ibdm/Fabric.h>
    
    #ifndef alloca
        #define alloca malloc
    #endif

    #define YYERROR_VERBOSE
    #define YYTYPE_INT16 int

    extern int yyerror(const char *msg);
    extern int ibnl_lex(void);
    extern int ibnl_lex_destroy(void);
%}

%union {
    IBNodeType tval;
    int        ival;
    char      *sval;
}

%token INT_NUM SYSTEM TOPSYSTEM NODE SUBSYSTEM NODETYPE NAME SPEED WIDTH LINE
%token CFG DASH END_DASH

%type <sval> NAME WIDTH SPEED
%type <ival> INT_NUM SYSTEM NODE SUBSYSTEM TOPSYSTEM CFG DASH END_DASH
%type <tval> NODETYPE

%start ibnl

%{
static int ibnlErr;
long lineNum;
static const char *gp_fileName;
static int gIsTopSystem = 0;
static list< char * > gSysNames;
static IBSystemsCollection *gp_sysColl = 0;
static IBSysDef *gp_curSysDef = 0;
static IBSysInst *gp_curInstDef = 0;

void
ibnlMakeSystem(list< char * > &sysNames)
{
#ifdef DEBUG
    printf("Making new system named:");
#endif
    gp_curSysDef = new IBSysDef(gp_fileName);

    for( list< char * >::iterator snI = sysNames.begin();
            snI != sysNames.end(); snI++) {
        char sname[1024];
        if (gIsTopSystem) {
            sprintf(sname, "%s", *snI);
        } else {
            sprintf(sname, "%s/%s", gp_fileName, *snI);
        }
        string sNameStr(sname);
        gp_sysColl->addSysDef(sNameStr, gp_curSysDef);
#ifdef DEBUG
        printf("%s ", sname);
#endif
    }
#ifdef DEBUG
    printf("\n");
#endif

    // cleanup for next systems.
    sysNames.erase(sysNames.begin(), sysNames.end());
}

void
ibnlMakeSubInstAttribute(char *hInst, char *attr, char *value)
{
#ifdef DEBUG
    printf("Making new sub instance attribute inst:%s %s=%s\n",
            hInst, attr, value);
#endif
    if (!gp_curSysDef) {
        printf("-E- How com e we got no system???\n");
        exit(3);
    }
    // append to existing attr or create new
    string hierInstName = string(hInst);
    string attrStr = string(attr);
    if (value)
        attrStr += "=" +  string(value);
    gp_curSysDef->setSubInstAttr(hierInstName, attrStr);
}

void
ibnlMakeNode(IBNodeType type, phys_port_t numPorts, char *devName, char* name, int numVirtualPorts)
{
#ifdef DEBUG
    printf(" Making Node:%s dev:%s ports:%d\n", name, devName, (int)numPorts);
#endif
    gp_curInstDef = new IBSysInst(name, devName, numPorts, type, numVirtualPorts);
    gp_curSysDef->addInst(gp_curInstDef);
}

void
ibnlMakeNodeToNodeConn(int fromPort,
                       const char *width, const char *speed,
                       const char *toNode, int toPort)
{
#ifdef DEBUG
    printf("  Connecting N-N port:%d to Node:%s/%d (w=%s,s=%s)\n",
            fromPort, toNode, toPort, width, speed);
#endif
    char buf1[8],buf2[8] ;
    sprintf(buf1, "%d", toPort);
    sprintf(buf2, "%d", fromPort);
    IBSysInstPort *p_port =
            new IBSysInstPort(buf2, toNode, buf1,
                    char2width(width), char2speed(speed));
        gp_curInstDef->addPort(p_port);
}

void
ibnlMakeNodeToPortConn(int fromPort, const char *width, const char *speed,
        const char *sysPortName)
{
#ifdef DEBUG
    printf("  System port:%s on port:%d (w=%s,s=%s)\n",
            sysPortName, fromPort, width, speed);
#endif
    char buf[8];
    sprintf(buf,"%d",fromPort);
    IBSysPortDef *p_sysPort =
            new IBSysPortDef(sysPortName, gp_curInstDef->getName(), buf,
                    char2width(width), char2speed(speed));
    gp_curSysDef->addSysPort(p_sysPort);
}

void
ibnlMakeSubsystem(char *masterName, char *instName)
{
#ifdef DEBUG
    printf(" Making SubSystem:%s of type:%s\n", instName, masterName);
#endif
    gp_curInstDef = new IBSysInst(instName, masterName);
    gp_curSysDef->addInst(gp_curInstDef);
}

void
ibnlRecordModification(char *subSystem, char *modifier)
{
#ifdef DEBUG
    printf("  Using modifier:%s on %s\n", modifier, subSystem);
#endif
    gp_curInstDef->addInstMod(subSystem, modifier);
}

void
ibnlMakeSubsystemToSubsystemConn(char *fromPort, const char *width, const char *speed,
        char *toSystem, char *toPort)
{
#ifdef DEBUG
    printf("  Connecting S-S port:%s to SubSys:%s/%s\n",
            fromPort, toSystem, toPort);
#endif
    IBSysInstPort *p_port =
            new IBSysInstPort(fromPort, toSystem, toPort,
                    char2width(width), char2speed(speed));
    gp_curInstDef->addPort(p_port);
}

void
ibnlMakeSubsystemToPortConn(const char *fromPort, const char *width, const char *speed,
        const char *toPort)
{
#ifdef DEBUG
    printf("  Connecting port:%s to SysPort:%s\n",
            fromPort, toPort);
#endif
    IBSysPortDef *p_sysPort =
            new IBSysPortDef(toPort, gp_curInstDef->getName(), fromPort,
                    char2width(width), char2speed(speed));
    gp_curSysDef->addSysPort(p_sysPort);
}

%}


%%
NL:
    LINE
    | NL LINE;

ONL:
    | NL;

ibnl: ONL systems topsystem;

systems:
    | systems system
    ;

sub_inst_attributes:
    | sub_inst_attributes sub_inst_attribute NL
    ;

sub_inst_attribute:
    NAME '=' NAME '=' NAME { ibnlMakeSubInstAttribute($1,$3,$5); }
    | NAME '=' NAME '=' INT_NUM {
        char buf[16];
        sprintf(buf, "%d", $5);
        ibnlMakeSubInstAttribute($1,$3,buf);
    }
    | NAME '=' NAME {ibnlMakeSubInstAttribute($1,$3,NULL); }
    ;

topsystem:
    TOPSYSTEM { gIsTopSystem = 1; }
    system_names { ibnlMakeSystem(gSysNames); }
    NL sub_inst_attributes
    insts
    ;

system:
    SYSTEM { gIsTopSystem = 0; }
    system_names { ibnlMakeSystem(gSysNames); } NL insts
    ;

system_names:
    system_name
    | system_names ',' system_name
    ;

system_name:
    NAME { gSysNames.push_back($1); }
    ;

insts:
    | insts node
    | insts subsystem
    ;

node:
    node_header NL node_connections
    ;

node_header:
      NODE NODETYPE INT_NUM             NAME NAME { ibnlMakeNode($2,(phys_port_t)$3   ,$4,$5, 0); }
    | NODE NODETYPE INT_NUM '+' INT_NUM NAME NAME { ibnlMakeNode($2,(phys_port_t)$3+$5,$6,$7, $5); }
    ;

node_connections:
    | node_connections node_connection NL
    ;

node_connection:
    node_to_node_link
    | node_to_port_link
    ;

node_to_node_link:
    INT_NUM DASH WIDTH DASH SPEED DASH END_DASH NAME INT_NUM {
        ibnlMakeNodeToNodeConn($1, $3, $5, $8, $9);
    }
    | INT_NUM DASH WIDTH DASH END_DASH NAME INT_NUM {
        ibnlMakeNodeToNodeConn($1, $3, "2.5", $6, $7);
    }
    | INT_NUM DASH SPEED DASH END_DASH NAME INT_NUM {
        ibnlMakeNodeToNodeConn($1, "4x", $3, $6, $7);
    }
    | INT_NUM DASH END_DASH NAME INT_NUM {
        ibnlMakeNodeToNodeConn($1, "4x", "2.5", $4, $5);
    }
    ;

node_to_port_link:
    INT_NUM DASH WIDTH DASH SPEED DASH END_DASH NAME {
        ibnlMakeNodeToPortConn($1, $3, $5, $8);
    }
    | INT_NUM DASH WIDTH DASH END_DASH NAME {
        ibnlMakeNodeToPortConn($1, $3, "2.5", $6);
    }
    | INT_NUM DASH SPEED DASH END_DASH NAME {
        ibnlMakeNodeToPortConn($1, "4x", $3, $6);
    }
    | INT_NUM DASH END_DASH NAME {
        ibnlMakeNodeToPortConn($1, "4x", "2.5", $4);
    }
    ;

subsystem:
    subsystem_header NL subsystem_connections
    | subsystem_header NL CFG insts_modifications NL subsystem_connections
    ;

subsystem_header:
    SUBSYSTEM NAME NAME { ibnlMakeSubsystem($2,$3); }
    ;

insts_modifications:
    | insts_modifications modification
    ;

modification:
    NAME '=' NAME { ibnlRecordModification($1,$3); }
    ;

subsystem_connections:
    | subsystem_connections subsystem_connection NL
    ;

subsystem_connection:
    subsystem_to_subsystem_link
    | subsystem_to_port_link
    ;

subsystem_to_subsystem_link:
    NAME DASH WIDTH DASH SPEED DASH END_DASH NAME NAME {
        ibnlMakeSubsystemToSubsystemConn($1, $3, $5, $8, $9);
    }
    | NAME DASH WIDTH DASH END_DASH NAME NAME {
        ibnlMakeSubsystemToSubsystemConn($1, $3, "2.5", $6, $7);
    }
    | NAME DASH SPEED DASH END_DASH NAME NAME {
        ibnlMakeSubsystemToSubsystemConn($1, "4x", $3, $6, $7);
    }
    | NAME DASH END_DASH NAME NAME {
        ibnlMakeSubsystemToSubsystemConn($1, "4x", "2.5", $4, $5);
    }
    ;

subsystem_to_port_link:
    NAME DASH WIDTH DASH SPEED DASH END_DASH NAME {
        ibnlMakeSubsystemToPortConn($1, $3, $5, $8);
    }
    | NAME DASH WIDTH DASH END_DASH NAME {
        ibnlMakeSubsystemToPortConn($1, $3, "2.5", $6);
    }
    | NAME DASH SPEED DASH END_DASH NAME {
        ibnlMakeSubsystemToPortConn($1, "4x", $3, $6);
    }
    | NAME DASH END_DASH NAME {
        ibnlMakeSubsystemToPortConn($1, "4x", "2.5", $4);
    }
    ;


%%
int yyerror(const char *msg)
{
    printf("-E-ibnlParse:%s at line:%ld\n", msg, lineNum);
    ibnlErr = 1;
    return 1;
}


/*
 * parse apollo route dump fil
 */
int ibnlParseSysDefs (IBSystemsCollection *p_sysColl, const char *fileName)
{
    extern FILE * ibnl_in;

    gp_sysColl = p_sysColl;
    gp_fileName = fileName;

    /*
     * open the file
     */
    ibnl_in = fopen(fileName,"r");
    if (!ibnl_in) {
        printf("-E- Fail to Open File:%s\n", fileName);
        return(1);
    }
    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
        printf("-I- Parsing:%s\n", fileName);

    ibnlErr = 0;
    lineNum = 1;

    /*
     * parse it
     */
    yyparse();

    fclose(ibnl_in);
    ibnl_lex_destroy();
    return(ibnlErr);
}
