/*
 * Copyright (c) 2004-2010 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <set>

#include <infiniband/ibdm/Fabric.h>
#include <infiniband/ibdm/ibsysapi/ibsysapi.h>


struct sysapi {
    IBFabric *fabric;
    int ibSysVerbose;
};


sysapi_handler_t
ibSysCreate()
{
    sysapi_handler_t ibSysHandler = new struct sysapi;
    if (!ibSysHandler)
        return NULL;
    ibSysHandler->fabric        = NULL;
    ibSysHandler->ibSysVerbose  = 0;
    return ibSysHandler;
}


void
ibSysDestroy(sysapi_handler_t ibSysHandler)
{
    if (ibSysHandler->fabric)
        delete ibSysHandler->fabric;
    delete ibSysHandler;
}


/* read the IBNL with specific configuration and prepare the local fabric */
int
ibSysInit(sysapi_handler_t ibSysHandler,
        int ibSysVerbose,
        char *sysType, char *cfg)
{
    ibSysHandler->ibSysVerbose = ibSysVerbose;

    if (ibSysHandler->fabric) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: already initialized\n");
        }
        return(1);
     }

    ibSysHandler->fabric = new IBFabric();
    if(!ibSysHandler->fabric)   {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: allocate memory for new fabric failed\n");
        }
        return(1);
    }
    IBSystem *p_system = ibSysHandler->fabric->makeSystem("SYS", sysType, cfg);
    if (!p_system) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: failed to make system:%s with cfg:%s\n",sysType, cfg);
        }
        return(1);
    }

    if (ibSysHandler->ibSysVerbose & IBSYS_INFO)
        printf("Info: initialized fabric with single system of type %s\n", sysType);
    return(0);
}


// DFS in the direction of the destination node adding paths as discovered
static int
dfsFromNodeToNode(sysapi_handler_t ibSysHandler,
        IBNode *curNode, IBNode *toNode,
        int *path, int hop, int *numFound, int numReq, int **drPaths,
        set< IBNode *, less< class IBNode *> > &visited,
        map<IBNode*, int > &esprance)
{
    if (ibSysHandler->ibSysVerbose & IBSYS_DEBUG) {
        printf("Debug: dfsFromNodeToNode at node %s hop:%d\n",
                curNode->name.c_str(), hop);
    }

    if (curNode == toNode) {
        for (int i = 0; i < hop; i++)
            drPaths[*numFound][i] = path[i];
        drPaths[*numFound][hop] = -1;
        if (ibSysHandler->ibSysVerbose & IBSYS_DEBUG) {
            printf("Debug: found %s at path #%d :", toNode->name.c_str(), *numFound);
            int i = 0;
            while (drPaths[*numFound][i] != -1)
                printf("%d,", drPaths[*numFound][i++]);
            printf("\n");
        }
        (*numFound)++;
        return(0);
    }

    /* sort the ports by the esprance of the next node */
    map< int, list< phys_port_t > > rankPorts;

    /* try all ports */
    map< int, list< phys_port_t > >::iterator rI;
    // use unsigned int index to prevent overwrap
    for (unsigned int i = 1; i <= curNode->numPorts; i++) {
        phys_port_t pn = (phys_port_t)i;
        IBPort *port = curNode->getPort(pn);
        if (port && port->p_remotePort) {
            IBNode *remNode = port->p_remotePort->p_node;
            int rank = esprance[remNode];
            if (rankPorts.find(rank) == rankPorts.end()) {
                list< phys_port_t > tmp;
                rankPorts[rank] = tmp;
            }

            rI = rankPorts.find(rank);
            (*rI).second.push_back(pn);
        }
    }

    /* now use the ranks and list of ports and dfs in order */
    int curRank = esprance[curNode];
    for (rI = rankPorts.begin(); rI != rankPorts.end(); rI++) {
        for (list< phys_port_t >::iterator lI = (*rI).second.begin();
                lI != (*rI).second.end(); lI++) {
            phys_port_t pn = *lI;
            IBPort *port = curNode->getPort(pn);
            IBNode *remNode = port->p_remotePort->p_node;
            // do not take secenic rouets - always go towards toNode
            int rank = esprance[remNode];
            if (rank >= curRank) continue;
            // only if not already visited
            if (visited.find(remNode) == visited.end()) {
                // we do want to find all paths to toNode ...
                if (remNode != toNode) visited.insert(remNode);
                path[hop] = pn;
                if (!dfsFromNodeToNode(ibSysHandler,
                        remNode, toNode, path, hop+1,
                        numFound, numReq, drPaths, visited, esprance))
                    if (*numFound == numReq) return(0);
            }
        }
    }

    /* we did not find all requested paths ... */
    return(1);
}


// Mark distance to destination node in "esprance"
static int
bfsFromToNodeMarkEsprance(sysapi_handler_t ibSysHandler,
        IBNode *toNode, map<IBNode*, int > &esprance)
{
    list< IBNode *> q;
    int rank;
    esprance[toNode] = 0;
    q.push_back(toNode);
    while (! q.empty()) {
        IBNode *node = q.front();
        q.pop_front();

        // go over all ports of the node and add to q only non visited
        for (unsigned int pn = 1; pn <= node->numPorts; pn++) {
            IBPort *port = node->getPort((phys_port_t)pn);
            if (!port || !port->p_remotePort) continue;
            rank = esprance[node];
            if (esprance.find(port->p_remotePort->p_node) == esprance.end()) {
                esprance[port->p_remotePort->p_node] = rank + 1;
                q.push_back(port->p_remotePort->p_node);
            }
        }
    }

    if (ibSysHandler->ibSysVerbose & IBSYS_DEBUG) {
        for ( map<IBNode*, int >::const_iterator eI = esprance.begin();
                eI != esprance.end(); eI++) {
            printf("Debug: Node:%s Esprance:%d\n", (*eI).first->name.c_str(),
                    (*eI).second);
        }
    }
    return(0);
}


/* query DR paths to a node  - paths are pre allocated */
/* return the number of filled paths in numPaths */
/* end of DR is a -1 */
int
ibSysGetDrPathsToNode(sysapi_handler_t ibSysHandler,
        char *fromNode, char *toNode,
        int *numPaths, int**drPaths)
{
    if (!ibSysHandler->fabric) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: fabric was not initialized.\n");
        }
        return(1);
    }

    if (!drPaths) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: No output path table provided\n");
        }
        return(1);
    }

    if (!numPaths) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: No number of paths provided\n");
        }
        return(1);
    }

    if (!fromNode) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: No source node provided\n");
        }
        return(1);
    }

    if (!toNode) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: No destination node provided\n");
        }
        return(1);
    }

    IBNode *sNode = ibSysHandler->fabric->getNode(fromNode);
    if (!sNode) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: failed to find from-node %s\n", fromNode);
        }
        return(1);
    }
    IBNode *dNode = ibSysHandler->fabric->getNode(toNode);
    if (!dNode) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: failed to find to-node %s\n", toNode);
        }
        return(1);
    }

    map<IBNode*, int > esprance;
    set< IBNode *, less< class IBNode *> > visited;

    int path[64];
    int hop = 0;
    path[0] = 0;
    path[1] = -1;

    /* will decrement as we go */
    int numFound = 0;

    bfsFromToNodeMarkEsprance(ibSysHandler,
            dNode, esprance);

    /* DFS to the Node from the local node */
    dfsFromNodeToNode(ibSysHandler,
            sNode, dNode, path, hop+1, &numFound, *numPaths,
            drPaths, visited, esprance);

    *numPaths = numFound;
    if (ibSysHandler->ibSysVerbose & IBSYS_INFO) {
        printf("Info: found %d paths from %s to %s\n",
                *numPaths, fromNode, toNode);
    }
    return(0);
}


/* query all node names in the system */
int
ibSysGetNodes(sysapi_handler_t ibSysHandler,
        int *numNodes, const char **nodeNames)
{
    int i;
    if (!numNodes) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: no num nodes provided.\n");
        }
        return(1);
    }

    if (!nodeNames) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: no nodeNames array provided.\n");
        }
        return(1);
    }

    if (!ibSysHandler->fabric) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: fabric was not initialized.\n");
        }
        return(1);
    }

    i = 0;
    for(map_str_pnode::const_iterator I = ibSysHandler->fabric->NodeByName.begin();
            (i < *numNodes) && (I != ibSysHandler->fabric->NodeByName.end()); I++) {
        nodeNames[i++]=(*I).first.c_str();
    }

    *numNodes = i;
    if (ibSysHandler->ibSysVerbose & IBSYS_INFO) {
        printf("Info: found %d nodes\n", i);
    }
    return(0);
}


/* query name of front panel port of a node port */
int
ibSysGetNodePortSysPort(sysapi_handler_t ibSysHandler,
        char *nodeName, phys_port_t portNum,
        const char **sysPortName)
{
    if (!ibSysHandler->fabric) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: fabric was not initialized.\n");
        }
        return(1);
    }

    if (!nodeName) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: No node name provided\n");
        }
        return(1);
    }

    if (!sysPortName) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: No destination port name provided\n");
        }
        return(1);
    }

    IBNode *node = ibSysHandler->fabric->getNode(nodeName);
    if (!node) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: failed to find node:%s\n", nodeName);
        }
        return(1);
    }

    IBPort *port = node->getPort(portNum);
    if (!port) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: no connection at node:%s port:%d\n",
                    nodeName, portNum);
        }
        *sysPortName = NULL;
        return(1);
    }

    if (!port->p_sysPort) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: no front pannel port at node:%s port:%d\n",
                    nodeName, portNum);
        }
        *sysPortName = NULL;
        return(1);
    }
    *sysPortName = port->p_sysPort->name.c_str();

    if (ibSysHandler->ibSysVerbose & IBSYS_INFO) {
        printf("Info: node %s port %d connects to system port %s\n",
                nodeName, portNum, *sysPortName);
    }
    return(0);
}


/* query node name and port given front pannel port name */
int
ibSysGetNodePortOnSysPort(sysapi_handler_t ibSysHandler,
        char *sysPortName,
        const char**nodeName, int *portNum)
{
    if (!ibSysHandler->fabric) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: fabric was not initialized.\n");
        }
        return(1);
    }

    if (!sysPortName) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: No front pannel port name provided\n");
        }
        return(1);
    }

    if (!nodeName) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: No node name provided\n");
        }
        return(1);
    }

    if (!portNum) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: No port number provided\n");
        }
        return(1);
    }

    IBSystem *system = ibSysHandler->fabric->getSystem("SYS");
    if (!system) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: failed to find system SYS\n");
        }
        return(1);
    }

    IBSysPort *sysPort = system->getSysPort(sysPortName);
    if (!sysPort) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: failed to find system front pannel port:%s\n",
                    sysPortName);
        }
        return(1);
    }

    if (!sysPort->p_nodePort) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: failed to find system front pannel port:%s node port?\n",
                    sysPortName);
        }
        return(1);
    }

    *nodeName = sysPort->p_nodePort->p_node->name.c_str();
    *portNum = sysPort->p_nodePort->num;

    if (ibSysHandler->ibSysVerbose & IBSYS_INFO) {
        printf("Info: system port %s connects to node %s port %d\n",
                sysPortName, *nodeName, *portNum);
    }
    return(0);
}


/* query other side of the port node port */
int
ibSysGetRemoteNodePort(sysapi_handler_t ibSysHandler,
        char *nodeName, phys_port_t portNum,
        const char **remNode, int *remPortNum)
{
    if (!ibSysHandler->fabric) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: fabric was not initialized.\n");
        }
        return(1);
    }

    if (!nodeName) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: No node name provided\n");
        }
        return(1);
    }

    if (!remNode) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: No remote node name provided\n");
        }
        return(1);
    }

    if (!remPortNum) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: No remote port number provided\n");
        }
        return(1);
    }

    IBNode *node = ibSysHandler->fabric->getNode(nodeName);
    if (!node) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: failed to find node:%s\n", nodeName);
        }
        return(1);
    }

    IBPort *port = node->getPort(portNum);
    if (!port || !port->p_remotePort) {
        if (ibSysHandler->ibSysVerbose & IBSYS_ERROR) {
            printf("Error: no connection at node:%s port:%d\n",
                        nodeName, portNum);
        }
        *remNode = NULL;
        return(1);
    }

    *remNode = port->p_remotePort->p_node->name.c_str();
    *remPortNum = port->p_remotePort->num;
    if (ibSysHandler->ibSysVerbose & IBSYS_INFO) {
        printf("Info: node %s port %d connects to node %s port %d\n",
                nodeName, portNum, *remNode, *remPortNum );
    }
    return(0);
}

