/*
 * Copyright (c) 2012-2013 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <getopt.h>
#include <fstream>

#include <infiniband/ibdm/Fabric.h>
#include <infiniband/ibdm/SubnMgt.h>
#include <infiniband/ibdm/Congestion.h>
#include <infiniband/ibdm/Regexp.h>
#include <map>
#include <list>

using namespace std;
const char usage[] = 
  "TOPO: ibcongest [-hdABC][-v lvl] -t <topo-file> [-n <sm-node>][-p <sm-port>]\n"
  "                [-r ftree|pgft|osm>] [-z roots-file] [-R rep-hsd-file] [-Q rep-bw-file] \n"
  "                [-O order-file] [-o order-file] {-a|-s <schedule>}\n" 
  "LST:  ibcongest [-hdABC][-v lvl] -l <lst> -f <fdbs> [-P <PLFT>] [-R rep-hsd-file] [-Q rep-bw-file]\n"
  "                [-O order-file] [-o order-file] {-a [-j <min-shift> -k <max-shift>]|-s <schedule>}\n";

void
show_usage() {
  cout << "Usage:" << endl;
  cout << usage;
}

void
show_help() {
  cout
    << "   STATIC CONGESTION ANALYSIS\n"
    << "\n"
    << "USAGE: \n"
    << "\n"
    << usage
    << "\n"
    << "DESCRIPTION:\n"
    << "\n"
    << "Calculate routing for a given topology (topo-mode) or use exstracted \n"
    << "lst/fdb files (lst-mode).\n"
    << "\n"
    << "Analyze congestion for a traffic schedule provided in a \"schedule-file\"\n"
    << "or use an automatically generated schedule of all-to-all-shift.\n"
    << "\n"
    << "In topo-mode the following options are used:\n"
    << "-t topo-file: mandatory. Topology file to be used.\n"
    << "-r ftree/pgft/osm: defines the routing engine to be used (osm).\n"
    << "-n sm-node: define the node name for the SM to run on (H-1/U1).\n"
    << "-p sm-port: define the port number of the sm-node the SM run from (1).\n"
    << "\n"
    << "In lst-mode the following optiontions are used:\n"
    << "-l lst-file: mandatoty. LST file obtained from OpenSM or ibdiagnet.\n"
    << "-f fdbs-file: mandatoty. FDBs file obtained from OpenSM or ibdiagnet.\n"
    << "-P PLFT: a file holding private LFT mapping on switch.\n"
    << "\n"
    << "OTHER ARGUMENTS:\n"
    << "-s schedule-file: a file defining the communication pattern. \n"
    << "   The format described in the below section SCHEDULE FILE.\n"
    << "-a : automatically generate an all-to-all schedule\n"
    << "-j min-shift -k max-shift : automatically generate a range of shifts\n"
    << "-z roots-file : a file holding names of switches which are the roots of the topology.\n"
    << "   This is mandatory if the -F flag is provided to report HSD while estimating flow fractions.\n"
    << "\n"
    << "OPTIONS:\n"
    << "-A : check all paths not just CA to CA\n"
    << "-B : Calculate and report actual bandwidth of each flow. Turned on also if -Q is provided.\n"
    << "-C : Assume Congestion Control. I.e. flow fairness and no back-pressure.\n"
    << "-F : Estimate flow fractions in HSD report. NOTE: this is inaccurate as it does not take\n"
    << "     effects of HoL blocking or round-robin arbitration into account\n"
    << "-O order-file: run simulated annealing and write out the result into\n"
    << "   the given file.\n"
    << "-o order-file: a file holding host name in each line, defining their ordering. \n"
    << "   Description of the order file is in section ORDER_FILE.\n"
    << "   Default order is by host name sort H-<N> by the N.\n"
    << "-m: ignore missing CA-to-CA paths\n"
    << "-R rep-hsd-file: report each stage over subscription into the given file name\n"
    << "-Q rep-bw-file: report each stage each flow BW (in csb format) into the given file name\n"
    << "-v lvl: verbose mode - prepare for being flooded... lvl is:\n"
    << "   1 - \n"
    << "   2 - order optimization\n"
    << "   3 - congestion analysis\n"
    << "   4 - fat tree routing\n"
    << "   5 - fat tree recognition\n"
    << "\n"
    << "-h: this message\n"
    << "\n"
    << "SCHEDULE FILE:\n"
    << "The schedule file describes a communication sequence. \n"
    << "There are 2 types of lines:\n"
    << "1. Stage line separates different stages of the schedule: \n"
    << "^STAGE(\\s+[0-9]+)?\n"
    << "\n"
    << "2. SRC DST line describes traffic in pairs\n"
    << "Optionally the destination may be given a specific 'path-bits'\n"
    << "that specify exact DLID from a range of DLIDs a destination may have.\n"
    << "Another option is the flow may be given a relative BW of the SRC node.\n"
    << "From the first rank to the second rank:\n"
    << "^\\s*[0-9]+\\s+[0-9]+(:([0-9]+))?(\\s+(0\\.[0-9]+))?\\s*\n"
    << "Or from a port name to a port name:\n"
    << "^\\s*(\\S+)/P([0-9]+)\\s+*(\\S+)/P([0-9]+)(:([0-9]+))?(\\s+(0\\.[0-9]+))?\\s*\n"
    << "For example: \n"
    << "A flow from node H-1 port 1 to H-100 port 1 that 50% of the BW of H-1\n"
    << "H-1/U1/P1 H-100/U1/P1:1 0.5\n"
    << "\n"
    << "\n"
    << "ORDER FILE:\n"
    << "The order file provides mapping between the rank (which is the line number)\n"
    << "and a host in the topology. There are two type of lines supported:\n"
    << "1. OpenSM order file style holding LID and host name\n"
    << "^\\s*(0x[0-9a-fA-F]+)\\s(\\s+)\\s+HCA-([0-9]+)\n"
    << "\n"
    << "2. Port name only: simply declare the order of host names to be used\n"
    << "^\\s*(\\s+)\\s+P?([0-9]+)\\s*$\n"
    << "\n"
    << "3. Dummy place holder\n"
    << "^DUMMY$\n"
    << "\n"
    << "DUMMY stands for an empty rank. Just a placeholder - not a real HCA.\n"
    << "Traffic from DUMMY or to DUMMY is not generated even if specified in \n"
    << "the schedule.\n"
    << endl;
}

// This program is doing a fabric init simulation or loads lst/fdbs
// Then runs some patterns and analyze congestion
bool bwDump = false;
ofstream bwDumpFile;
bool bwMode = false;

///////////////////////////////////////////////////////////////////////
//
// ROUTINES
//

// get a lid
lid_t
getPortLid(IBFabric *fabric, string nodeName, phys_port_t portNum, uint8_t bits)
{
  // find the node
  IBNode *pNode = fabric->getNode(nodeName);
  if (!pNode) {
    cout << "-E- could not find node:" << nodeName << endl;
    return(0);
  }
  
  // get the port
  IBPort *pPort = pNode->getPort(portNum);
  if (!pPort) {
    cout << "-E- could not find port:" << nodeName << "/" << portNum << endl;
    return(0);
  }
  
  // get the lid:
  lid_t baseLid = pPort->base_lid;
  return((lid_t)(baseLid + (lid_t)bits));
}

// Given names find the LIDs and do the trace using ibdmCongTrace
int 
traceRouteByNames(IBFabric *fabric, string srcNode, phys_port_t srcPort,
		  string dstNode, phys_port_t dstPort, uint8_t dstBits,
		  double ratio)
{
  lid_t sLid = getPortLid(fabric, srcNode, srcPort, 0);
  if (sLid == 0) {
    cout << "-E- could not find source LID" << endl;
    return(1);
  }

  lid_t dLid = getPortLid(fabric, dstNode, dstPort, dstBits);
  if (dLid == 0) {
    cout << "-E- could not find destination LID" << endl;
    return(1);
  }

  // trace it
  CongTrackPath(fabric, sLid, dLid, ratio);
  return(0);
}

// Given names find the LIDs and do the trace using ibdmCongTrace
int 
traceRouteByPorts(IBFabric *fabric, IBPort *srcPort, IBPort *dstPort, uint8_t dstBits, double ratio)
{
  lid_t sLid = (lid_t)(srcPort->base_lid);
  if (sLid == 0) {
    cout << "-E- could not find source LID" << endl;
    return(1);
  }

  lid_t dLid = (lid_t)(dstPort->base_lid + dstBits);
  if (dLid == 0) {
    cout << "-E- could not find destination LID" << endl;
    return(1);
  }

  // trace it
  CongTrackPath(fabric, sLid, dLid, ratio);
  return(0);
}

// we try to assign the lids in HCA name order.
// we assume we all HCAs are named H-<N>
int
assignLidsByTopology (IBFabric *fabric, string hostNameRex, phys_port_t pn) 
{
  regExp nodeRex(hostNameRex.c_str());
  rexMatch *p_rexRes;
  lid_t maxLid = 0;

  // try matching hosts
  map_str_pnode::iterator nI;
  for (nI = fabric->NodeByName.begin(); 
       nI != fabric->NodeByName.end(); nI++) {
    IBNode *pNode = (*nI).second;
    if (pNode->type == IB_SW_NODE) continue;
    IBPort *pPort = pNode->getPort(pn);
    if (!pPort) continue;

    p_rexRes = nodeRex.apply((*nI).first.c_str());
    if (p_rexRes) {
      lid_t hostNum = (lid_t)atoi(p_rexRes->field(1).c_str());
      pPort->base_lid = hostNum;
      fabric->setLidPort(hostNum, pPort);
      cout << "-I- Set port: " << pPort->getName() << " lid: "
	   << pPort->base_lid << endl;
      if (maxLid < hostNum) maxLid = hostNum;
      delete p_rexRes;
    } else {
      cout << "-W- non-switch node: '" << pNode->name 
	   << "' did not match regexp:" << hostNameRex << endl;
    }
  }

  // now the rest
  for (nI = fabric->NodeByName.begin(); 
       nI != fabric->NodeByName.end(); nI++) {
    IBNode *pNode = (*nI).second;
    
    p_rexRes = nodeRex.apply((*nI).first.c_str());
    if (!p_rexRes || (pNode->type != IB_SW_NODE)) {
      maxLid++;
      phys_port_t numPorts = pNode->numPorts;
      // use unsigned int index to prevent overwrap
      for (unsigned int pn = 1; pn <= numPorts; pn++) {
	IBPort *pPort = pNode->getPort((phys_port_t)pn);
	if (!pPort) continue;
	pPort->base_lid = maxLid;
	fabric->setLidPort(maxLid, pPort);
	cout << "-I- Set port: " << pPort->getName() << " lid: "
	     << pPort->base_lid << endl;
      }
    } else if (p_rexRes) {
      delete p_rexRes;
    }
  }
  fabric->maxLid = maxLid;
  return(0);
}

// clear counter 1 on all switch ports
int
clearSwitchPortCounters (IBFabric *fabric)
{
  map_str_pnode::iterator nI;
  for (nI = fabric->NodeByName.begin(); 
       nI != fabric->NodeByName.end(); nI++) {
    IBNode *pNode = (*nI).second;
    if (pNode->type == IB_SW_NODE) {
      phys_port_t numPorts = pNode->numPorts;
      for (unsigned int pn = 1; pn <= numPorts; pn++) {
	IBPort *pPort = pNode->getPort((phys_port_t)pn);
	if (pPort) {
	  pPort->counter1 = 0;
	}
      }
    }
  }
  return(0);
}

// generate order by sorting host names
// HACK: assume each HCA connects through just one port...
int
setOrderBySortingHostNames(IBFabric *fabric, phys_port_t pn,
			   map< unsigned int, IBPort *> &PortByRank,
			   map< lid_t, unsigned int> &RankByLID) 
{
  unsigned int rank = 0;
  // we use nodebyname and it is thus sorted by name...
  map_str_pnode::iterator nI;
  for (nI = fabric->NodeByName.begin(); 
       nI != fabric->NodeByName.end(); nI++) {
    IBNode *pNode = (*nI).second;
    if (pNode->type == IB_SW_NODE) continue;
    IBPort *pPort = pNode->getPort(pn);
    if (!pPort || !pPort->p_remotePort) continue;
    PortByRank[rank] = pPort;
    lid_t lid = pPort->base_lid;
    RankByLID[lid] = rank;
    rank++;
  }
  return(0);
}

// parse the order file sets PORT_BY_RANK
// return number of ranks defined
int
parseOrderFile (IBFabric *fabric, string fileName,
		map< unsigned int, IBPort *> &PortByRank,
		map< lid_t, unsigned int> &RankByLID) 
{
  regExp lidLine("^\\s*(0x[0-9a-fA-F]+)\\s(\\S+)\\s+HCA-([0-9]+)");
  regExp nameLine("^\\s*(\\S+)\\s+P?([0-9]+)\\s*$");
  regExp dummyLine("DUMMY");
  rexMatch *p_rexRes;
  ifstream f(fileName.c_str());
  if (f.good()) {
    cout << "-I- Parsing Order file:" << fileName << endl;
  } else {
    cout << "-E- Could not read Order file:"  << fileName << endl;
    exit(1);
  }
 
  unsigned int rank = 0;
  unsigned int hosts = 0;
  char sLine[1024];
  unsigned int lineNum = 0;
  while (f.good()) {
    f.getline(sLine, 1024);
    lineNum++;
    if ((p_rexRes = lidLine.apply(sLine))) {
      lid_t lid = (lid_t)strtoul(p_rexRes->field(1).c_str(),NULL,16);
      IBPort *pPort = fabric->getPortByLid(lid);
      if (!pPort) {
	cout << "-W- Fail to find port for LID: " << lid << endl;
	PortByRank[rank] = NULL;
      } else {
	PortByRank[rank] = pPort;
	RankByLID[lid] = rank;
	hosts++;
      }
      rank++;
      delete p_rexRes;
    } else if ((p_rexRes = dummyLine.apply(sLine))) {
      PortByRank[rank++] = NULL;
      delete p_rexRes;      
    } else if ((p_rexRes = nameLine.apply(sLine))) {
      string n = p_rexRes->field(1);
      phys_port_t pn = (phys_port_t)atoi(p_rexRes->field(2).c_str());
      IBNode *pNode = fabric->getNode(n);
      if (pNode) {
	IBPort *pPort = pNode->getPort(pn);
	if (pPort) {
	  PortByRank[rank] = pPort;
	  RankByLID[pPort->base_lid] = rank;
	  hosts++;
	} else {
	  cout << "-W- Fail to find port for node: " << n << " port: " 
	       << (unsigned int)pn << endl;
	  PortByRank[rank] = NULL;
	}
      } else {
	cout << "-W- Fail to find node: " << n << endl;
	PortByRank[rank] = NULL;
      }
      rank++;
      delete p_rexRes; 
    } else if (strlen(sLine)) {
      cout << "-W- ignoring line " << lineNum << " :'" << sLine << "'" << endl;
    }
  }
  f.close();
  cout << "-I- Order file: " << fileName << " defined: " 
       << hosts++ << " hosts" << endl;
  return(0);
}

// Automatically generate All-to-All schedule file
int 
genAllToAllSchedule(IBFabric *fabric, 
		    map< unsigned int, IBPort *> &PortByRank,
		    string outStagesFile)
{
  ofstream f(outStagesFile.c_str());
  if (!f.good()) {
    cout << "-E- Could not write schedule into:" << outStagesFile << endl;
    exit(1);
  }
  
  unsigned int N = (unsigned int)PortByRank.size();
  for (unsigned int s = 1; s < N; s++) {
    f << "STAGE " << s << endl;
    for (unsigned int i = 0; i < N; i++) {
      unsigned int j = (i + s) % N;
      f << i << " " << j << endl;
    }
  }
  f.close();
  return(0);
}

// Automatically generate All-to-All schedule file
int 
runAllToAllSchedule(IBFabric *fabric, 
		    string dumpFileName,
		    unsigned int startShift, unsigned int endShift,
		    map< unsigned int, IBPort *> &PortByRank,
		    map< lid_t, unsigned int> &RankByLID)
{
  ofstream dumpFile;
  bool dump = false;
  if (dumpFileName.size()) {
    dumpFile.open(dumpFileName.c_str());
    if (!dumpFile.good()) {
      cout << "-E- Could not open dump file: " << dumpFileName << endl;
      exit(1);
    }
    dump = true;
  }

  unsigned int N = (unsigned int)PortByRank.size();
  uint8_t dBits = 0;
  unsigned int stages = 0;
  if (!startShift) {
    startShift = 1;
  }
  
  if (endShift) {
    if (endShift >= N) {
      cout << "-E- Requested shift range end:" << endShift << " above N: " 
	   << N << " aborting" << endl;
      exit(1);
    }
    cout << "-I- Simulating shifts traffic from shift: " << startShift
	 << " to shift: " << endShift << endl; 
  } else {
    cout << "-I- Simulating all-to-all traffic for " << N << " nodes" << endl;
    endShift = N-1;
  }

  for (unsigned int s = startShift; s <= endShift; s++) {
    stages++;
    if (dump)
      dumpFile << "STAGE: " << stages << endl;
    
    for (unsigned int i = 0; i < N; i++) {
      unsigned int j = (i + s) % N;
      IBPort *srcPort = PortByRank[i];
      IBPort *dstPort = PortByRank[j];
      
      if (!srcPort || !dstPort) 
	continue;

      // do the actual work
      if (traceRouteByPorts(fabric, srcPort, dstPort, dBits, 1.0)) {
	cout << "-E- Error in tracing path: " << srcPort->getName()
	     << " to: " << dstPort->getName() << " bits: " << dBits << endl;
      }
    } // all nodes

    if (dump)
      CongDumpStage(dumpFile, fabric, stages, RankByLID);

    if (bwMode) {
      if (bwDump) {
	bwDumpFile << "STAGE: " << s << endl;
      }
      CongCalcBW(fabric, bwDump, bwDumpFile);
    }
    CongZero(fabric);
  }
  CongReport(fabric, cout);
 
  if (dump)
    dumpFile.close();
  
  return(0);
}

// parse the schedule file and perform the tracing accordingly
int
parseScheduleFile(IBFabric *fabric,
		  string schedFileName,
		  string dumpFileName,
		  map< unsigned int, IBPort *> &PortByRank,
		  map< lid_t, unsigned int> &RankByLID)
{
  ofstream dumpFile;
  bool dump = false;
  if (dumpFileName.size()) {
    dumpFile.open(dumpFileName.c_str());
    if (!dumpFile.good()) {
      cout << "-E- Could not open dump file: " << dumpFileName << endl;
      exit(1);
    }
    dump = true;
  }
  
  ifstream schedFile(schedFileName.c_str());
  if (!schedFile.good()) {
    cout << "-E- Fail to open communication schedule file: " << schedFileName << endl;
    exit(1);
  }

  cout << "-I- Parsing Schedule file:" << schedFileName << endl;
	 
  // parse the file defining the node groups by name and
  // perform the analysis
  unsigned int stages = 0;
  unsigned int numPairs = 0;
  char sLine[1024];
  regExp ranksSchedLine("^\\s*([0-9]+)\\s+([0-9]+)(:([0-9]+))?(\\s+(0\\.[0-9]+))?\\s*");
  regExp portsSchedLine("^\\s*(\\S+)/P([0-9]+)\\s+*(\\S+)/P([0-9]+)(:([0-9]+))?(\\s+(0\\.[0-9]+))?\\s*");
  regExp stageLine("^STAGE");
  regExp emptyLine("^\\s*(#|$)");
  rexMatch *p_rexRes;

  while (schedFile.good()) {
    schedFile.getline(sLine, 1024);
    
    if ((p_rexRes = ranksSchedLine.apply(sLine))) {
      unsigned int srcRank = atoi(p_rexRes->field(1).c_str());
      unsigned int dstRank = atoi(p_rexRes->field(2).c_str());
      uint8_t dBits   = 0;
      double rateRatio = 1.0;

      if (p_rexRes->field(4).size()) 
	dBits = (uint8_t)atoi(p_rexRes->field(4).c_str());
      
      if (p_rexRes->field(6).size()) 
	rateRatio = (double)atof(p_rexRes->field(6).c_str());

      delete p_rexRes;

      IBPort *srcPort = PortByRank[srcRank];
      IBPort *dstPort = PortByRank[dstRank];
      
      if (!srcPort || !dstPort) 
	continue;

      // do the actual work
      if (traceRouteByPorts(fabric, srcPort, dstPort, dBits, rateRatio)) {
	cout << "-E- Error in tracing path: " << srcPort->getName() 
	     << " to: " << dstPort->getName() << " bits: " << dBits 
	     << " ratio: " << rateRatio << endl;
      }
      numPairs++;
    } else if ((p_rexRes = portsSchedLine.apply(sLine))) {
      string srcNodeName = p_rexRes->field(1);
      phys_port_t srcPortNum = (phys_port_t)(atoi(p_rexRes->field(2).c_str()));
      string dstNodeName = p_rexRes->field(3);
      phys_port_t dstPortNum = (phys_port_t)(atoi(p_rexRes->field(4).c_str()));

      uint8_t dBits = 0;
      double rateRatio = 1.0;

      if (p_rexRes->field(6).size()) 
	dBits = (uint8_t)atoi(p_rexRes->field(6).c_str());
      
      if (p_rexRes->field(8).size()) 
	rateRatio = (double)atof(p_rexRes->field(8).c_str());

      if (rateRatio <= 0) {
	cout << "-E- Zero rate??? at line: " << sLine << endl;
	exit(1);
      }
      delete p_rexRes;

      IBNode *srcNode = fabric->getNode(srcNodeName);
      IBPort *srcPort = NULL;
      IBPort *dstPort = NULL;
      if (!srcNode) {
	cout << "-E- Could not find src node:" << srcNodeName << endl;
      } else {
	srcPort = srcNode->getPort(srcPortNum);
      }

      IBNode *dstNode = fabric->getNode(dstNodeName);
      if (!dstNode) {
	cout << "-E- Could not find dst node:" << dstNodeName << endl;
      } else {
	dstPort = dstNode->getPort(dstPortNum);
      }

      if (!srcPort || !dstPort) continue;

      // do the actual work
      if (traceRouteByPorts(fabric, srcPort, dstPort, dBits, rateRatio)) {
	cout << "-E- Error in tracing path: " << srcPort->getName()
	     << " to: " << dstPort->getName() << " bits: " << dBits 
	     << " ratio: " << rateRatio
	     << endl;
      } 
      numPairs++;
    } else if ((p_rexRes = stageLine.apply(sLine))) {
      // cleanup paths and prepare for next group
      if (stages > 0) {
	// NOTE: for the LINK level data to be valid we need to get it before
	// the ibdmCongClear actually clears the lists
	if (dump) {
	  CongDumpStage(dumpFile, fabric, stages, RankByLID);
	}
	
	// if {$found(repLink) > 0} {
	//   collectLinkPaths $fabric
	// }
	
	if (bwMode) {
	  if (bwDump) {
	    bwDumpFile << "STAGE:" << stages << endl;
	  }
	  CongCalcBW(fabric, bwDump, bwDumpFile);
	}
	CongZero(fabric);
      }
      
      stages++;
      cout << "STAGE:" << stages << "\r";
      cout.flush();
      delete p_rexRes;
    } else if (emptyLine.apply(sLine)) {
      delete p_rexRes;
    } else {
      cout << "-W- Ignoring line: " << sLine << endl;
    }
  }
  cout << "-I- Traced: " << numPairs << " pairs" << endl;
  schedFile.close();
  
  if (stages) {
    // last stage reportsls
    if (dump) {
      CongDumpStage(dumpFile, fabric, stages, RankByLID);
      dumpFile.close();
    }
    
    // we need to clear to get the histogram updated
    if (bwMode) {
      if (bwDump) {
	bwDumpFile << "STAGE:" << stages << endl;
      }
      CongCalcBW(fabric, bwDump, bwDumpFile);
    }
  
    CongZero(fabric);
    CongReport(fabric, cout);
  } else {
    cout << "-W- No stages define?" << endl;
  }
  return(0);
}

// return the roots as a list
int
parseRootsFile(IBFabric *fabric, string fileName, list<IBNode*> &roots)
{
  ifstream f(fileName.c_str());
  if (!f.good()) {
    cout << "-E- Fail to open roots file: " << fileName << endl;
    exit(1);
  }
  char sLine[1024];
  regExp guidLine("^\\s*(0x[0-9a-fA-F]+)");
  regExp nodeLine("^\\s*(\\S+)");
  rexMatch *p_rexRes;
  while (f.good()) {
    f.getline(sLine, 1024);
    if ((p_rexRes = guidLine.apply(sLine))) {
      uint64_t guid = strtoull(p_rexRes->field(1).c_str(), NULL, 16);
      IBNode *pNode = fabric->getNodeByGuid(guid);
      if (!pNode) {
	cout << "-W- Could not find root node by GUID: " 
	     << p_rexRes->field(1) << endl;
      } else {
	roots.push_front(pNode);
      }
      
      delete p_rexRes;
    } else if ((p_rexRes = nodeLine.apply(sLine))) {
      IBNode *pNode = fabric->getNode(p_rexRes->field(1));
      if (!pNode) {
	cout << "-W- Could not find root node: " 
	     << p_rexRes->field(1) << endl;
      } else {
	roots.push_front(pNode);
      }
      delete p_rexRes;
    }
  }
  return(0);
}

int
getRoots(IBFabric *fabric, string rootsFile, list<IBNode*> &roots) {
  if (rootsFile.size()) {
    parseRootsFile(fabric, rootsFile, roots);
  } else {
    roots = SubnMgtFindTreeRootNodes(fabric);
  }
  return(0);
}

/////////////////////////////////////////////////////////////////////////
//
// MAIN FLOW
//
int main(int argc, char**argv) 
{
  // defaults
  unsigned int verboseLevel = 0;
  string smNodeName("H-1/U1");
  phys_port_t smPortNum = 1;
  string routeType("osm");
  string outOrderFile;
  string ordFileName;
  string repStageHSDFile;
  string repStageBWFile;
  string schedFileName;
  string pLFTFileName;
  string fdbsFileName;
  string lstFileName;
  string rootsFileName;
  bool   ignoreMissingCAtoCA = false;
  string topoFile;
  string lstFile;
  bool   verifyAllPathsMode = false;
  bool   allToAllSchedule = false;
  bool   ccMode = false;
  unsigned int startShift = 0;
  unsigned int endShift = 0;
  int next_option;
  const char * const short_option = "hv:DQ:R:mt:n:p:r:o:as:l:f:P:O:z:ABCj:k:";
  /*
    In the array below, the 2nd parameter specified the number
    of arguments as follows:
    0: no arguments
    1: argument
    2: optional
  */
  
  const option long_option[] =
    {
      {  "verbose",   1,   NULL,   'v'},
      {  "help",      0,   NULL,   'h'},
      {  "rep-hsd",   1,   NULL,   'R'},
      {  "rep-bw",    1,   NULL,   'Q'},
      {  "all-paths", 0,   NULL,   'A'},
      {  "bw-mode"  , 0,   NULL,   'B'},
      {  "cong-ctrl", 0,   NULL,   'C'},
      {  "topo",      1,   NULL,   't'},
      {  "sm-node",   1,   NULL,   'n'},
      {  "router",    1,   NULL,   'r'},
      {  "roots",     1,   NULL,   'z'},
      {  "out-order", 1,   NULL,   'O'},
      {  "order-file",1,   NULL,   'o'},
      {  "all-to-all",0,   NULL,   'a'},
      {  "sched-file",1,   NULL,   's'},
      {  "lst",       1,   NULL,   'l'},
      {  "fdbs-file", 1,   NULL,   'f'},
      {  "PLFT",     1, NULL, 'P'},
      {  "min-shift", 1,   NULL,   'j'},
      {  "max-shift", 1,   NULL,   'k'},
      {  NULL,        0,	 NULL,	 0 }	/* Required at the end of the array */
    };
  
  cout << "--------------------------------------------------------------------------" << endl;
  
  do {
    next_option = getopt_long(argc, argv, short_option,long_option, NULL);
    switch (next_option) {
    case 'h':
      show_help();
      return 0;
      break;
      
    case 'v':
      /*
	Verbose Mode
      */
      verboseLevel = atoi(optarg);
      cout << "-I- Verbose Mode: " << verboseLevel << endl;
      break;
      
    case 'R':
      repStageHSDFile = optarg;
      cout << "-I- Dump Hot-Spot Details into: " << repStageHSDFile << endl;
      break;
    
    case 'Q':
      bwMode = true;
      repStageBWFile = optarg;
      cout << "-I- Dump BW Details into: " << repStageBWFile << endl;
      break;

    case 'D':
      break;
    
    case 't':
      topoFile = optarg;
      cout << "-I- Topology File: " << topoFile << endl;
      break;
    
    case 'A': 
      cout << "-I- Consider all paths (not just CA to CA)" << endl;
      verifyAllPathsMode = true;
      break;

    case 'B': 
      cout << "-I- Calculate Flows BW" << endl;
      bwMode = true;
      break;

    case 'C': 
      cout << "-I- Assume Congestion Control is ON" << endl;
      ccMode = true;
      break;
    
    case 'n':
      smNodeName = optarg;
      cout << "-I- SM Node Name:" << smNodeName << endl;
      break;
    
    case 'p':
      smPortNum = (phys_port_t)atoi(optarg);
      cout << "-I- SM Port Number:" << (unsigned int)smPortNum << endl;
      break;
    
    case 'r':
      routeType = optarg;
      if (routeType == "osm") {
	cout << "-I- Routing Mode: OSM" << endl;
      } else if (routeType == "ftree") {
	cout << "-I- Routing Mode: FTREE" << endl;
      } else {
	cout << "-E- Unsupported routing engine: " << optarg << endl;
	exit(1);
      }
      break;
    
    case 'o': 
      ordFileName = optarg;
      cout << "-I- Order File: " << ordFileName << endl;
      break;
    
    case 'j': 
      startShift = atoi(optarg);
      cout << "-I- Start Shift : " << startShift << endl;
      break;
      
    case 'k': 
      endShift = atoi(optarg);
      cout << "-I- End Shift : " << endShift << endl;
      break;
    
    case 'a':
      allToAllSchedule = true;
      cout << "-I- Running All-to-All schedule" << endl;
      break;
    
    case 's':
      schedFileName = optarg;
      cout << "-I- Schedule File: " << schedFileName << endl;
      break;

    case 'f':
      fdbsFileName = optarg;
      cout << "-I- FDBs File: " << fdbsFileName << endl;
      break;
    
    case 'P':
        pLFTFileName = optarg;
        cout << "-I- PLFT File: " << pLFTFileName << endl;
        break;

    case 'l':
      lstFileName = optarg;
      cout << "-I- Subnet-Lst File: " << lstFileName << endl;
      break;
    
    case 'm':
      ignoreMissingCAtoCA = true;
      cout << "-I- Ignore Missing CA to CA paths." << endl;
      break;
    
    case 'O':
      outOrderFile = optarg;
      cout << "-I- Output Order File:" << outOrderFile << endl;
      break;
    
    case 'z':
      rootsFileName = optarg;
      cout << "-I- Roots File:" << rootsFileName << endl;
      break;
    
    case -1:
      break; /* done with option */
    
    default: /* something wrong */
      show_usage();
      exit(1);
    }
  } while (next_option != -1);

  cout << "--------------------------------------------------------------------------" << endl;

  if (topoFile.size()) {
    cout << "-I- TOPO: " << topoFile << " SM: " << smNodeName << "/" << (unsigned int)smPortNum	
	 << " Routing: " << routeType << endl;
  } else {
    if (!(lstFileName.size() && fdbsFileName.size())) {
      cout << "-E- Missing -t or -l and -f Arguments." << endl;
      show_usage();
      exit(1);
    }
    cout << "-I- LST: " << lstFileName << " FDB: " << fdbsFileName << endl;
  }

  if (allToAllSchedule) {
    cout << "-I- Automatically generating all-to-all schedule" << endl;
  } else if (schedFileName.size()) {
    cout << "-I- Schedule file: " << schedFileName << endl;
  } else {
    cout << "-E- Missing either -a or -s flags" << endl;
    show_usage();
    exit(1);
  }

  if (ordFileName.size()) {
    cout << "-I- Order file: " << ordFileName << endl;
  }

  if (bwMode) {
    if (repStageBWFile.size()) {
      bwDumpFile.open(repStageBWFile.c_str());
      if (!bwDumpFile.good()) {
	cout << "-E- Could not open BW report file:" << repStageBWFile << endl;
	exit(1);
      }
      bwDump = true;
      bwDumpFile << "# SRC, SLID, DST, DLID, BW [GB/s], AGG-BW [GB/s]" << endl;
    }
  }

  cout << "--------------------------------------------------------------------------" << endl;

  ///////////////////////////////////////////////////////////////////////////////////////
  IBFabric fabric;
  uint8_t savedVerbosity;

  // Topology file mode requires LID assignment and routing 
  if (topoFile.size()) {
    if (fabric.parseTopology(topoFile)) {
      cout << "-E- Fail to parse topology file: " << topoFile << endl;
      exit(1);
    }
	 
    IBNode *smNode = fabric.getNode(smNodeName);
    if (!smNode) {
      cout << "-I- Fail to find SM node: " << smNodeName << endl;
      exit(1);
    }
	 
    IBPort *smPort = smNode->getPort(smPortNum);
    if (!smPort) {
      cout << "-E- Fail to find SM Port: " << smNodeName << "/" << smPortNum << endl;
      exit(1);
    }

    if (routeType == "ftree") {
      if (SubnMgtAssignLids(smPort, 0)) {
	cout << "-E- Fail to assign LIDs" << endl;
	exit(1);
      }
      
      list<IBNode*> roots;
      getRoots(&fabric, rootsFileName, roots);
      if (!roots.size()) {
	cout << "-E- Could not recognize root nodes! Please provide them manually" << endl;
	exit(1);
      }
      cout << "-I- Got " << roots.size() << " roots" << endl;
    
      map_pnode_rank nodesRank;
      if (SubnRankFabricNodesByRootNodes(&fabric, roots, nodesRank)) {
	cout << "-E- fail to rank fabric" << endl;
	exit(1);
      }
		  
      savedVerbosity = FabricUtilsVerboseLevel;
      if (verboseLevel > 4) {
	FabricUtilsVerboseLevel = 0xff;
      }
		  
      // match to tree:
      if (FatTreeAnalysis(&fabric)) {
	cout << "-E- Fail to perform fat tree routing" << endl;
	exit(1);
      }

      FabricUtilsVerboseLevel = savedVerbosity;
		  
    } else if (routeType == "osm") {
      if (SubnMgtAssignLids(smPort, 0)) {
	cout << "-E- Fail to assign LIDs" << endl;
	exit(1);
      }
    
      if (SubnMgtCalcMinHopTables(&fabric)) {
	cout << "-E- Fail to calculate MinHop tables" << endl;
	exit(1);
      }
    
      if (SubnMgtOsmRoute(&fabric)) {
	cout << "-E- Fail to Route" << endl;
	exit(1);
      }
    }

    // check the routing
    if (ignoreMissingCAtoCA) {
      cout << "-W- Ignore all CA to CA paths check" << endl;
    } else {
      if (verifyAllPathsMode) {
	if (SubnMgtVerifyAllRoutes(&fabric)) {
	  cout << "-E- Some missing paths" << endl;
	  exit(1);
	}				
      } else {
	if (SubnMgtVerifyAllCaToCaRoutes(&fabric)) {
	  cout << "-E- Some missing CA to CA paths" << endl;
	  exit(1);
	}
      }
    }
  } else {

    if (fabric.parseSubnetLinks(lstFileName)) {
      cout << "-E- Fail to parse lst: " << lstFile << endl;
      exit(1);
    }

    if (pLFTFileName.size() && fabric.parsePLFTFile(pLFTFileName)) {
        cout << "-E- Fail to parse PLFT: " << pLFTFileName << endl;
        exit(1);
    }

    if (fabric.parseFdbFile(fdbsFileName)) {
      cout << "-E- Fail to parse fdbs: " << fdbsFileName << endl;
      exit(1);
    }
  }

  // Optimize and generate order file
  if (outOrderFile.size()) {
    cout << "-E- The option to generate an optimized Order file not supported yet!" << endl;
    exit(1);
  }

  map< unsigned int, IBPort *> PortByRank;
  map< lid_t, unsigned int> RankByLID;

  if (ordFileName.size()) {
    if (parseOrderFile(&fabric, ordFileName, PortByRank, RankByLID)) {
      cout << "-E- could not parse the order file!" << endl;
      exit(1);
    }
  } else {
    if (setOrderBySortingHostNames(&fabric, 1, PortByRank, RankByLID)) {
      cout << "-E- could not generate order by sorting host names!" << endl;
      exit(1);
    }
  }

  if (CongInit(&fabric, ccMode)) {
    cout << "-E- Fail to init congestion analysis" << endl;
    exit(1);
  }

  savedVerbosity = FabricUtilsVerboseLevel;
  if (verboseLevel > 2) {
    FabricUtilsVerboseLevel = 0xff;
  }

  cout << "-I- Analyzing Congestion ..." << endl;
  cout.flush();

  if (schedFileName.size()) {
    parseScheduleFile(&fabric, schedFileName, repStageHSDFile, PortByRank, RankByLID);
  } else {
    runAllToAllSchedule(&fabric, repStageHSDFile, startShift, endShift, PortByRank, RankByLID);
    // char tmpSchedFile[128];
    // sprintf(tmpSchedFile, "all2all.sched.%d", getpid());
    // genAllToAllSchedule(&fabric, PortByRank, tmpSchedFile);
    // parseScheduleFile(&fabric, tmpSchedFile, repStageHSDFile, PortByRank, RankByLID);
    // unlink(tmpSchedFile);
  }

  FabricUtilsVerboseLevel = savedVerbosity;

  CongCleanup(&fabric);

  return(0);
}
