/*
 * Copyright (c) 2012 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <getopt.h>
#include <fstream>
#include <algorithm>
#include <functional>

#include <infiniband/ibdm/Fabric.h>
#include <infiniband/ibdm/SubnMgt.h>
#include <infiniband/ibdm/Congestion.h>
#include <infiniband/ibdm/Regexp.h>
#include <map>
#include <list>

using namespace std;

unsigned int verbose = 0;

const char usage[] =
  "ibgenperm [-h][-v][-t][-n nflow-per-link][-m lmc][-d min-dist] "
  "-l lst -f fdbs [-P PLFT] -i grp-regexp -s max-group-srcs\n";
void
show_usage() {
  cout << "Usage:" << endl;
  cout << usage;
}

void
show_help() {
  cout
    << "   GENERATE A PERMUTATION MEETING GIVEN OVER-SUBSCRIPTION\n"
    << "\n"
    << "USAGE: \n"
    << "\n"
    << usage
    << "\n"
    << "DESCRIPTION:\n"
    << "Given exstracted lst/fdb, split the nodes into groups using the provided \n"
    << "regular expression. Find a permutation (source destination pairs) \n"
    << "where the source and destination are from different group.\n"
    << "Such that the maximal number of flows through any link does not excceed the\n"
    << "nflow-per-link.\n"
    << "\n"
    << "ARGUMENTS:\n"
    << "-l <lst> : opensm-subnet.lst or ibdiagnet.lst file describing the network\n"
    << "-f <fdbs> : the opensm.fdbs or ibdiagnet.fdbs file\n"
    << "-P <PLFT> : a file holding private LFT mapping on switch.\n"
    << "-i group-regexp: a regular expression to extract the group from a host name.\n"
    << "   otherwise as regexp. A file should include <name>/P<port-num> or PortGUID.\n"
    << "-s max-grp-srcs: maximal number of srcs to take from a group.\n"
    << "\n"
    << "OPTIONS:\n"
    << "-d min-dist: accept only flows of minimal hop distance of min-dist (default 0).\n"
    << "-n nflow-per-link: The maximal number of flows per link (default 1).\n"
    << "-o out-file: output pairs file name (default pairs.txt)\n"
    << "-t : trust the FDBs and skip verification of all CA to CA paths\n"
    << "-m lmc : assume the network uses the given LMC value. I.e. each node is \n"
    << "   assigned 2^LMC consequative LIDs.\n"
    << "-v : verbose mode - prepare for being flooded... \n"
    << "-h: this message\n"
    << endl;
}

// This program is try to find pairing that does not produce
// contention

///////////////////////////////////////////////////////////////////////
//
// ROUTINES
//

// get ports by regexp or file
int
getPortGroups (IBFabric *fabric, string groupRexStr, map<string, set<IBPort*> >&GROUP_PORTS)
{
  regExp grpRex(groupRexStr.c_str());
  rexMatch *p_rexRes;

  unsigned int skipNode = 0;
  unsigned int grpsNode = 0;
  cout << "-I- definding nodes by regexp: " << groupRexStr << endl;
  map_str_pnode::iterator nI;
  for (nI = fabric->NodeByName.begin(); nI != fabric->NodeByName.end(); nI++) {
    IBNode *pNode = (*nI).second;
    if (pNode->type == IB_SW_NODE) continue;

    // HACK: we only use port 1
    IBPort *pPort = pNode->getPort(1);

    if (!pPort) continue;

    if ((p_rexRes = grpRex.apply(pNode->name.c_str()))) {
      string g = p_rexRes->field(1);
      GROUP_PORTS[g].insert(pPort);
      delete p_rexRes;
      grpsNode++;
    } else {
      skipNode++;
    }
  }
  map<string, set<IBPort*> >::iterator gI;
  for (gI = GROUP_PORTS.begin(); gI != GROUP_PORTS.end(); gI++) {
    cout << "-I- group: " << (*gI).first << " has: " << (*gI).second.size()
	 << " ports" << endl;
  }
  cout << "-I- Total of " << grpsNode << " nodes (skipped " << skipNode << ")" << endl;
  return(0);
}

int
clearSwitchPortCounters (IBFabric *fabric)
{

  map_str_pnode::iterator nI;
  for (nI = fabric->NodeByName.begin(); nI != fabric->NodeByName.end(); nI++) {
    IBNode *pNode = (*nI).second;
    if (pNode->type != IB_SW_NODE) continue;

    phys_port_t numPorts = pNode->numPorts;
    // use unsigned int index to prevent overwrap
    for (unsigned int pn = 1; pn <= numPorts; pn++) {
      IBPort *pPort = pNode->getPort((phys_port_t)pn);
      if (!pPort) continue;
      pPort->counter1 = 0;
    }
  }
  return(0);
}

// get a random order port number
phys_port_t
getRandPortOrder (phys_port_t nPorts)
{
  return((phys_port_t)(rand() % nPorts+1));
}

// randomize order for a given list
// instead use : random_shuffle ( myvector.begin(), myvector.end() );
//int
// randOrderList {in} {

// DFS towards the possible destination LIDs
// Filter out the possible detinations by the LFT port availability
// so:
// Calculate possible ports by load per port
// Go over possible dsts and lookup their ports.
// Descend these ports with reduced DSTS set
IBPort *
dfsAndGetFreeDstPath (IBPort *inPort, unsigned int hops,
		      set<lid_t> &dsts, map<lid_t, string> &dlidGrp,
		      unsigned int maxHsd, lid_t lmc, unsigned int minHops)
{
  IBPort *dstPort = NULL;
  if (verbose)
    cout << "-V- dfsAndGetFreeDstPath inPort:" << inPort->getName()
	 << " num-dsts:" << dsts.size() << endl;

  IBNode *node = inPort->p_node;

  // if we got to a non switch we are done
  if (node->type != IB_SW_NODE) {
    if (hops >= minHops) {
      lid_t dLid =  inPort->base_lid;
      for (lid_t pb = 0; pb < ((lid_t)1<<lmc); pb++) {
	dlidGrp.erase(dLid++);
      }
      return(inPort);
    } else {
      return(NULL);
    }
  }

  // calculate the possible ports
  phys_port_t numPorts = node->numPorts;

  multimap< unsigned int, IBPort *> portsByOrder;

  // order ports by their current count in increasing
  for (unsigned int pn = 1; pn <= numPorts; pn++) {
    IBPort *port = node->getPort((phys_port_t)pn);
    if (!port) continue;

    IBPort *remPort = port->p_remotePort;
    if (!remPort) continue;
    portsByOrder.insert(pair<unsigned int, IBPort *>(port->counter1, port));
  }

  multimap< unsigned int, IBPort *>::iterator pI;
  for (pI = portsByOrder.begin(); pI != portsByOrder.end(); pI++) {
    IBPort *port = (*pI).second;

    unsigned int cnt = port->counter1;
    // note that if already equal can't be used
    if (cnt >= maxHsd) continue;

    // filter DSTS if they go out our direction only
    set<lid_t> pnDsts;
    for (set<lid_t>::iterator dI = dsts.begin(); dI!=dsts.end(); dI++) {
      phys_port_t pn = node->getLFTPortForLid((*dI));
      if (pn == port->num) {
	pnDsts.insert(*dI);
      }
    }

    if (!pnDsts.empty()) {
      dstPort = dfsAndGetFreeDstPath(port->p_remotePort, hops+1,
				     pnDsts, dlidGrp, maxHsd, lmc, minHops);
      if (dstPort) {
	port->counter1++;
	return(dstPort);
      }
    }
  }
  return(NULL);
}

// traverse the network towards any unused DST and report when got there
// trough ports that are never overloaded
IBPort *
findDstForSrc (string grp, IBPort *srcPort,
	       map<lid_t, string> &dlidGrp, unsigned int maxHsd, lid_t lmc, unsigned int minHops)
{

  // get to the switch and then DFS from there
  IBPort *remPort = srcPort->p_remotePort;
  if (!remPort) {
    cout << "-E- No remote port for src: " << srcPort->getName() << endl;
    exit(1);
  }

  IBNode *remNode = remPort->p_node;
  if (remNode->type != IB_SW_NODE) {
    cout << "-E- Remote node is not a switch for src: " << srcPort->getName() << endl;
    exit(1);
  }

  // prepare possible dlids
  set<lid_t> dsts;
  map<lid_t, string>::iterator lI;
  for (lI = dlidGrp.begin(); lI != dlidGrp.end(); lI++) {
    if ((*lI).second != grp) {
      dsts.insert((*lI).first);
    }
  }
  IBPort *dstPort =
    dfsAndGetFreeDstPath(remPort, 1, dsts, dlidGrp, maxHsd, lmc, minHops);
  return(dstPort);
}

///////////////////////////////////////////////////////////////////////////
//
// MAIN FLOW
//

int
main (int argc, char** argv)
{
  string outFileName("ibgenperm.out");
  string lstFileName;
  string pLFTFileName;
  string fdbsFileName;
  unsigned int maxHSD = 1;
  lid_t lmc = 0;
  string groupRegExp;
  unsigned int maxGroupSrcs = 0;
  unsigned int minHops = 0;
  bool trustFdbs = false;
  int next_option;
  const char * const short_option = "hvto:n:l:f:P:i:m:s:d:";
  /*
    In the array below, the 2nd parameter specified the number
    of arguments as follows:
    0: no arguments
    1: argument
    2: optional
  */

  const option long_option[] =
    {
      {  "verbose",    0,   NULL,   'v'},
      {  "help",       0,   NULL,   'h'},
      {  "trust-fdbs", 0,   NULL,   't'},
      {  "max-hsd",    1,   NULL,   'n'},
      {  "lmc",        1,   NULL,   'm'},
      {  "lst",        1,   NULL,   'l'},
      {  "fdbs",       1,   NULL,   'f'},
      {  "PLFT",     1, NULL, 'P'},
      {  "grp-rex",    1,   NULL,   'i'},
      {  "max-grp-src",1,   NULL,   's'},
      {  "min-dist",   1,   NULL,   'd'},
      {  NULL,         0,   NULL,    0 }	/* Required at the end of the array */
    };

  cout << "--------------------------------------------------------------------------" << endl;

  do {
    next_option = getopt_long(argc, argv, short_option,long_option, NULL);
    switch (next_option) {
    case 'h':
      show_help();
      return 0;
      break;

    case 'v':
      /*
	Verbose Mode
      */
      verbose = 1;
      cout << "-I- Verbose Mode" << endl;
      break;

    case 't':
      trustFdbs = true;
      cout << "-I- Trust FDBs and skip CA to CA verification" << endl;
      break;

    case 'd':
      minHops = atoi(optarg);
      cout << "-I- Minimum path distance is: " << minHops << " hops" << endl;
      break;

    case 'n':
      maxHSD = atoi(optarg);
      cout << "-I- Max Number of Flows per Link: " << maxHSD << endl;
      break;

    case 'm':
      lmc = (lid_t)atoi(optarg);
      cout << "-I- LMC: " << lmc << endl;
      break;

    case 'l':
      lstFileName = optarg;
      cout << "-I- LST file:" <<  lstFileName << endl;
      break;

    case 'f':
      fdbsFileName = optarg;
      cout << "-I- FDBs file:" << fdbsFileName << endl;
      break;

    case 'P':
        pLFTFileName = optarg;
        cout << "-I- PLFT File: " << pLFTFileName << endl;
        break;

    case 'i':
      groupRegExp = optarg;
      cout << "-I- Groups Regular Expression:" << groupRegExp << endl;
      break;

    case 's':
      maxGroupSrcs = atoi(optarg);
      cout << "-I- Max sources from each group: " << maxGroupSrcs << endl;
      break;

    case -1:
      break; /* done with option */

    default: /* something wrong */
      show_usage();
      exit(1);
    }
  } while (next_option != -1);

  cout << "--------------------------------------------------------------------------" << endl;

  bool anyMissing = false;
  if (!fdbsFileName.size()) {
    cout << "-E- Missing -f FDBs file argument" << endl;
    anyMissing = true;
  }

  if (!lstFileName.size()) {
    cout << "-E- Missing -l LST file argument" << endl;
    anyMissing = true;
  }

  if (!groupRegExp.size()) {
    cout << "-E- Missing -i Regular Expression argument" << endl;
    anyMissing = true;
  }

  if (anyMissing) {
    show_usage();
    return(1);
  }

  ///////////////////////////////////////////////////////////////////
  IBFabric fabric;
  if (fabric.parseSubnetLinks(lstFileName)) {
    cout << "-E- Fail to parse lst: " << lstFileName << endl;
    exit(1);
  }

  map<string, set<IBPort*> > GROUP_PORTS;
  getPortGroups (&fabric,  groupRegExp, GROUP_PORTS);

  if (pLFTFileName.size() && fabric.parsePLFTFile(pLFTFileName)) {
      cout << "-E- Fail to parse PLFT: " << pLFTFileName << endl;
      exit(1);
  }

  if (fabric.parseFdbFile(fdbsFileName)) {
    cout << "-E- Fail to parse fdbs: " << fdbsFileName << endl;
    exit(1);
  }

  // provide the list of FDB routing for each root
  if (trustFdbs) {
    cout << "-W- Skip all CA to CA paths check" << endl;
  } else {
    if (SubnMgtVerifyAllCaToCaRoutes(&fabric)) {
      cout << "-E- Some missing CA to CA paths" << endl;
      exit(1);
    }
  }

  // initialize port counters
  clearSwitchPortCounters(&fabric);

  map<lid_t, string> dlidGrp;

  // go over all groups prepare group DLIDS
  map<string, set<IBPort*> >::iterator gI;
  for (gI = GROUP_PORTS.begin(); gI != GROUP_PORTS.end(); gI++) {
    set<IBPort*>::iterator pI;
    for (pI = (*gI).second.begin(); pI != (*gI).second.end(); pI++) {
      IBPort *port = (*pI);
      lid_t bLid = port->base_lid;
      if (!bLid) {
	cout <<  "-E- Zero LID for DST port: " << port->getName() << endl;
	exit(1);
      }

      for (lid_t pb = 0; pb < ((lid_t)1<<lmc); pb++) {
	dlidGrp[bLid++] = (*gI).first;
      }
    }
  }

  ofstream outFile(outFileName.c_str());
  if (!outFile.good()) {
    cout << "-E- cant write:" << outFileName << endl;
    exit(1);
  }

  outFile << "# This file was generated by ibgenperm" << endl;
  outFile << "# Args: $args" << endl;
  outFile << "########################################################################\n" << endl;
  outFile << "STAGE 1" << endl;

  // loop on all srcs and find first dst that is not used and
  // have a way to it with no too many HSD
  unsigned int numPairs = 0;
  unsigned int numFailed = 0;

  // start for each group
  for (gI = GROUP_PORTS.begin(); gI != GROUP_PORTS.end(); gI++) {
    string grp = (*gI).first;
    set<IBPort*>::iterator pI;
    vector<IBPort*> ports((*gI).second.size(), NULL);
    size_t idx = 0;
    for (pI = (*gI).second.begin(); pI != (*gI).second.end(); pI++) {
      ports[idx++] =  (*pI);
    }

    random_shuffle(ports.begin(), ports.end());
    unsigned int gFound = 0;
    for (vector<IBPort*>::iterator lI = ports.begin(); lI != ports.end(); lI++) {
      IBPort *srcPort = *lI;
      IBPort *dstPort =
	findDstForSrc(grp, srcPort, dlidGrp, maxHSD, lmc, minHops);
      if (dstPort)  {
	outFile << srcPort->getName() << " " << dstPort->getName() << endl;
	numPairs++;
	gFound++;
      } else {
	cout << "-W- Could not find dst for: " << srcPort->getName() << endl;
	numFailed++;
      }

      if (maxGroupSrcs && (gFound >= maxGroupSrcs)) {
	break;
      }
    }
    cout << "-I- Group: " << grp << " found: " << gFound << " pairs out of: "
	 << ports.size() << endl;
  }
  outFile.close();

  cout << "-I- Wrote " << numPairs << " pairs into: " << outFileName << endl;
  if (numFailed) {
    cout << "-W- Could not map " << numFailed << " sources" << endl;
  }

  return(0);
}
