/*
 * Copyright (c) 2012 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <unistd.h>
#include <getopt.h>
#include <fstream>

#include <infiniband/ibdm/Fabric.h>
#include <infiniband/ibdm/SubnMgt.h>
#include <infiniband/ibdm/NetSplit.h>
#include <infiniband/ibdm/Regexp.h>
#include <map>
#include <list>

using namespace std;
const char usage[] =
  "There are two steps that can be run also in one flow:\n"
  "  Group: ibnetsplit [-v][-h][-g grp-file] -s <.lst|.net|.topo> <-r head-ports|-d max-dist>\n"
  "  Split: ibnetsplit [-v][-h][-g grp-file] -s <.lst|.net|.topo> -o out-dir\n"
  "  Both:  ibnetsplit [-v][-h][-g grp-file] -s <.lst|.net|.topo> <-r head-ports|-d max-dist> -o out-dir\n";

void
show_usage() {
  cout << "Usage:" << endl;
  cout << usage;
}

void
show_help() {
  cout
    << " IB Net Splitter\n"
    << "-----------------\n"
    << "\n"
    << "USAGE:\n"
    << usage
    << "\n"
    << "DESCRIPTION:\n"
    << "The ibnetsplit utility provide means to automatically group hosts \n"
    << "and create scripts that can be run in order to split the network\n"
    << "into sub-networks each with one group hosts.\n"
    << "\n"
    << "Grouping:\n"
    << "The grouping is performed if the -r or -d options are provided.\n"
    << "If the -r is provided with a file containing group head ports,\n"
    << "the algorithm examines the hosts distance from the set of\n"
    << "node ports provided in the head-ports file (these are expected to\n"
    << "be the ports running standby SM's).\n"
    << "If the -d is provided with a maximum distance of the hosts in each\n"
    << "group, the algorithm partition the hosts by that distance.\n"
    << "NOTE: that some topologies may not be analyzed this way.\n\n"
    << "The results of the identified groups are printed into the file \n"
    << "defined by the -g option (default ibnetsplit.groups) and can be hand\n"
    << "manipulated. For groups where the head port is a switch, the group \n"
    << "file uses the FIRST host port as the port to run the isolation\n"
    << "script from.\n"
    << "\n"
    << "Splitting:\n"
    << "If the -o flag is included, this algorithm analyzes the MinHop table\n"
    << "of the topology and identifies the set of links and switches that\n"
    << "may potentially be used for routing each group ports.\n"
    << "The cross-switch links betwen switches of the group to other \n"
    << "switches are declared as split-links and the commands to turn them\n"
    << "off using Directed Routes from the original Group Head ports \n"
    << "are written into the out-dir provided by the -o flag.\n"
    << "\n"
    << "Both stages require a subnet definition file to be provided\n"
    << "by the -s flag. The supported formats for subnet definition are:\n"
    << "*.net - for ibnetdiscover \n"
    << "*.lst - for opensm-subnet.lst or ibiagnet.lst\n"
    << "*.topo - for a topology file\n"
    << "\n"
    << "HEAD PORTS FILE\n"
    << "This file is provided by the user and defines the ports by which\n"
    << "grouping of the other host ports is defined.\n"
    << "\n"
    << "Format:\n"
    << "Each line should contain either the name or the GUID of a single\n"
    << "port. For switches the port number shall be 0:\n"
    << "<node-name>/P<port-num>|<PGUID>\n"
    << "\n"
    << "GROUPS FILE\n"
    << "This file is generated by the program if the head-ports file is\n"
    << "provided to it. Alternatively it can be provided (or edited) by\n"
    << "the user if different grouping is desired. The generated script\n"
    << "for isolating or connecting the group should be run from the\n"
    << "first node in each group.\n"
    << "\n"
    << "Format:\n"
    << "Each line may be either:\n"
    << "\n"
    << "GROUP: <group name>\n"
    << "<node-name>/P<port-num>|<PGUID>\n\n"
    << "Author: Eitan Zahavi, Mellanox Technologies LTD.\n"
    << endl;
}

list <IBPort *>
parseHeadPortsFile( IBFabric *p_fabric, string fileName)
{
  ifstream namesFile;
  namesFile.open(fileName.c_str(), ifstream::in);
  if (!namesFile.is_open()) {
    cout << "-F- Given head-ports file could not be opened!" << endl;
    exit(1);
  }

  regExp nameLine("^(\\S+)/P([0-9]+)?$");
  regExp guidLine("^0x([0-9a-zA-Z]+){16}$");

  char line[256];
  uint64_t guid;
  list <IBPort *> roots;
  cout << "-I- Obtaining Head Ports from file:" << fileName << endl;
  while (namesFile.getline(line,256) && namesFile.good()) {
    rexMatch *p_rexRes;
    p_rexRes = nameLine.apply(line);
    if (p_rexRes) {
      string name = p_rexRes->field(1);
      IBNode *pNode = p_fabric->getNode(name);
      if (!pNode) {
        cout << "-E- Fail to find node:" << name << endl;
      } else {
        phys_port_t pn = (phys_port_t)atoi(p_rexRes->field(2).c_str());
        IBPort *pPort = pNode->getPort(pn);
        if (!pPort) {
          cout << "-E- Fail to find node: " << name << " port: " << pn << endl;
        } else {
          roots.push_back(pPort);
        }
      }
      delete p_rexRes;
      continue;
    }

    // try as GUID
    p_rexRes = nameLine.apply(line);
    if (p_rexRes) {
      guid = strtoull(line, NULL, 0);
      IBPort *pPort = p_fabric->getPortByGuid(guid);
      if (!pPort) {
        cout << "-E- Fail to find port by GUID:" << line << endl;
      } else {
        roots.push_back(pPort);
      }
      delete p_rexRes;
      continue;
    }

    cout << "-E- Fail to parse line:" << line << endl;
  }
  cout << "-I- Defined " << roots.size() << " head ports " << endl;
  namesFile.close();
  return roots;
}

// simply return first port that may run SM
IBPort *getFirstSMPort(IBFabric *p_fabric)
{
  IBPort *smPort;
  map_str_pnode::iterator nI;
  for( nI = p_fabric->NodeByName.begin();
       nI != p_fabric->NodeByName.end(); nI++) {
    IBNode *p_node = (*nI).second;
    if (p_node->type == IB_SW_NODE) {
      smPort = p_node->getPort(0);
      if (smPort)
	return(smPort);
    } else {
      // use unsigned int index to prevent overwrap
      for (unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
          smPort = p_node->getPort((phys_port_t)pn);
	if (smPort && smPort->p_remotePort)
	  return(smPort);
      }
    }
  }
  return(NULL);
}

int main (int argc, char **argv) {

  FabricUtilsVerboseLevel = FABU_LOG_ERROR;
  string subnetFileName = string("");
  string groupsFileName = string("ibnetsplit.groups");
  string headsFileName = string("");
  string outputDirName;
  unsigned int maxDist = 0;

  int next_option;
  const char * const short_option = "vhg:s:r:o:d:";
  //
  // In the array below, the 2nd parameter specified the number
  // of arguments as follows:
  // 0: no arguments
  // 1: argument
  // 2: optional
  //

  const option long_option[] = {
    {  "verbose",  1,   NULL,   'v'},
    {  "help",     1,   NULL,   'h'},
    {  "groups",   1,   NULL,   'g'},
    {  "subnet",   1,   NULL,   's'},
    {  "head",     1,   NULL,   'r'},
    {  "distance", 1,   NULL,   'd'},
    {  "out",      1,   NULL,   'o'},
    {  NULL,       0,   NULL,   0 }     /* Required at the end of the array */
  };

  do {
    next_option = getopt_long(argc, argv, short_option,long_option, NULL);
    switch(next_option) {
    case 'h':
      show_help();
      return 0;
      break;

    case 'v':
      FabricUtilsVerboseLevel |= FABU_LOG_VERBOSE;
      printf(" Verbose Mode\n");
      break;

    case 'g':
      groupsFileName = string(optarg);
      break;

    case 's':
      subnetFileName = string(optarg);
      break;

    case 'r':
      headsFileName = string(optarg);
      break;

    case 'd':
      maxDist = atoi(optarg);
      if (maxDist < 2) {
        cout << "-E- no logic in giving max-dist < 2 ..." << endl;
        exit(1);
      }
      break;

    case 'o':
      outputDirName = string(optarg);
      break;

    case -1:
      break; /* done with option */
    default: /* something wrong */
      show_usage();
      exit(1);
    }
  }
  while(next_option != -1);

  // check for mandatory
  if (! subnetFileName.size()) {
    cout << "-E- Missing -s parameter. The subnet must be provided." << endl;
    show_usage();
    return 1;
  }

  IBFabric fabric;

  if (maxDist && headsFileName.size()) {
    cout << "-E- Options -d and -r cannot be provided together." << endl;
    exit(1);
  }
  if (!maxDist && !headsFileName.size() && !outputDirName.size()) {
    cout << "-E- One of the options -d or -r or -o must be provided." << endl;
    exit(1);
  }

  // load the subnet whatever its type is
  // get the extension of the file name
  size_t pos = subnetFileName.rfind(".");
  if (pos == string::npos) {
    cout << "-E- No extension on subnet file:" << subnetFileName << endl;
    return 1;
  }

  string ext = subnetFileName.substr(pos+1);
  bool lidsAssigned = false;
  if (!ext.compare("topo")) {
    if (fabric.parseTopology(subnetFileName)) {
      cout << "-E- Fail to parse topology file:" << subnetFileName << endl;
      exit(1);
    }
  } else if (!ext.compare("lst")) {
    if (fabric.parseSubnetLinks(subnetFileName)) {
      cout << "-E- Fail to parse lst file:" << subnetFileName << endl;
      exit(1);
    }
    lidsAssigned = true;
  } else if (!ext.compare("net")) {
    if (fabric.parseIBNetDiscover(subnetFileName)) {
      cout << "-E- Fail to parse ibnetdiscover file:" << subnetFileName << endl;
      exit(1);
    }
    lidsAssigned = true;
  } else {
    cout << "-E- Do not know how to subnet file of type: " << ext << endl;
    return 1;
  }

  map_port_set_ports portGroups;

  if (!lidsAssigned) {
    // get the SM Port - any port use the first CA port from the network ...
    IBPort *p_smPort = getFirstSMPort(&fabric);
    cout << "-I- Assigning LIDs from: " << p_smPort->getName()
         << "  ... " << endl;
    if (SubnMgtAssignLids(p_smPort, 0)) {
      cout << "-E- Fail to assign LIDs." << endl;
      exit(1);
    }
  }

  cout << "-I- Calculating Min-Hop tables ... " << endl;
  if (SubnMgtCalcMinHopTables(&fabric)) {
    cout << "-E- Fail to update Min Hops Tables." << endl;
    exit(1);
  }

  // do the grouping
  if (headsFileName.size() > 0 || maxDist > 0) {
    if (headsFileName.size() > 0) {
      list <IBPort *> headPorts =
        parseHeadPortsFile(&fabric, headsFileName);

      cout << "-I- Grouping hosts by head ports ... " << endl;
      if (NetSplitGroupHostsByHeads(fabric, headPorts, portGroups)) {
        cout << "-E- Grouping by head ports failed. Aborting." << endl;
        return(1);
      }
    } else {
      if (NetSplitGroupHostsByMinHop(fabric,  maxDist, portGroups)) {
        cout << "-E- Grouping by min-hop failed. Aborting." << endl;
        return(1);
      }
    }

    if (NetSplitDumpGroupsFile(portGroups, groupsFileName)) {
      cout << "-E- Could not create groups file. Aborting." << endl;
      return(1);
    }
    cout << "-I- Wrote groups file: " << groupsFileName << endl;
  }

  if (outputDirName.size() > 0) {
    if (portGroups.size() == 0) {
      if (NetSplitParseGroupsFile(fabric, groupsFileName, portGroups)) {
        cout << "-E- Grouping was not obtained. Aborting." << endl;
        return(1);
      }
    }

    cout << "-I- Declaring group links by Min-Hop routing ... " << endl;
    map<IBPort*, IBPort*> portToHeadAssignment;
    map<IBNode*, IBPort*> switchToHeadAssignment;
    map_port_set_ports::iterator gI;
    for (gI = portGroups.begin(); gI != portGroups.end(); gI++) {
      NetSplitMarkMinHopLinks((*gI).first, (*gI).second,
                              switchToHeadAssignment, portToHeadAssignment);
    }

    cout << "-I- Formatting scripts ... " << endl;
    if (NetSplitWriteSplitScripts(fabric, portGroups,
                                  switchToHeadAssignment, portToHeadAssignment,
                                  outputDirName)) {
      cout << "-E- Fail to write splitting scripts. Aborting." << endl;
      return(1);
    }
  }

  return(0);
}
