/*
 * Copyright (c) 2017 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */



#include "ibis.h"

#define MAX_INPUT_STRING_SIZE   128
#define IBDIAGNET_INDEX_TABLE_FORMAT    "# INDEX_TABLE"
#define IBDIAGNET_INDEX_TABLE_OFFSET    "offset:"
#define IBDIAGNET_INDEX_TABLE_LINE      "line:"
#define IBDIAGNET_SECTION_START_PREFIX  "START_"
#define IBDIAGNET_SECTION_END_PREFIX    "END_"

using namespace std;

ibis_log_msg_function_t CsvParser::m_log_msg_function = ibis_log::ibis_log_msg_function;

ibis_log_msg_function_t CsvParser::GetLogMsgFunction()
{
    return m_log_msg_function;
}

CsvFileStream::CsvFileStream(string file_name, CsvParser &csv_parser): m_file_name(file_name)
{
    IBIS_LOG(TT_LOG_LEVEL_INFO,
             "-I- CsvFileStream opening file %s\n",
             m_file_name.c_str());

    m_cfs.exceptions(std::ifstream::failbit | std::ifstream::badbit);
    try {
        m_cfs.open(m_file_name.c_str());
    } catch (std::ifstream::failure &e) {
        IBIS_LOG(TT_LOG_LEVEL_ERROR,
                 "-E- Exception opening csv file %s: %s\n",
                 m_file_name.c_str(), e.what());
        throw;
    }

    if (!m_cfs.is_open()) {
        IBIS_LOG(TT_LOG_LEVEL_ERROR,
                "-E- CsvFileStream can't open csv file %s\n",
                m_file_name.c_str());
        throw std::ios_base::failure("CsvFileStream can't open csv file");
    }

    if (UpdateSectionOffsetTable(csv_parser)) {
        IBIS_LOG(TT_LOG_LEVEL_INFO,
                 "Failed to update offset table for file: %s - %s\n",
                 m_file_name.c_str(), strerror(errno));
    }
}

CsvFileStream::~CsvFileStream()
{
    m_cfs.close();
    m_section_name_to_offset.clear();
}

bool CsvFileStream::IsFileOpen()
{
    if (m_cfs.is_open()) {
        return true;
    }

    return false;
}

bool CsvParser::ValidateStringInput(const char *field_str)
{
    if (!field_str)
        return false;

    size_t input_size = strlen(field_str);
    if (!input_size)
        throw TypeParseError(TypeParseError::EMPTY_VALUE, field_str);
    if (input_size > MAX_INPUT_STRING_SIZE)
        throw TypeParseError(TypeParseError::TOO_LONG_VALUE, field_str);

    return true;
}

bool CsvParser::Parse(const char *field_str, string &field)
{
    IBIS_ENTER;
    bool rc = true;
    if (ValidateStringInput(field_str))
        field = field_str;
    IBIS_RETURN(rc);
}

bool CsvParser::Parse(const char *field_str, u_int64_t &field, u_int8_t base)
{
    IBIS_ENTER;
    bool rc = true;
    if (ValidateStringInput(field_str))
        field = (u_int64_t)strtoull(field_str, NULL, base);

    IBIS_RETURN(rc);
}

bool CsvParser::Parse(const char *field_str, u_int32_t &field, u_int8_t base)
{
    IBIS_ENTER;
    bool rc = true;
    if (ValidateStringInput(field_str))
        field = (u_int32_t)strtoul(field_str, NULL, base);;

    IBIS_RETURN(rc);
}

bool CsvParser::Parse(const char *field_str, u_int16_t &field, u_int8_t base)
{
    IBIS_ENTER;
    bool rc = true;
    if (ValidateStringInput(field_str))
        field = (u_int16_t)strtoul(field_str, NULL, base);

    IBIS_RETURN(rc);
}

bool CsvParser::Parse(const char *field_str, u_int8_t &field, u_int8_t base)
{
    IBIS_ENTER;
    bool rc = true;
    if (ValidateStringInput(field_str))
        field = (u_int8_t)strtoul(field_str, NULL, base);

    IBIS_RETURN(rc);
}

int CsvFileStream::UpdateSectionOffsetTable(CsvParser &csv_parser)
{
    map_str_to_offset_t l_section_name_to_offset;

    IBIS_ENTER;
    int rc = 0;
    if (!m_cfs.is_open()) {
        IBIS_LOG(TT_LOG_LEVEL_ERROR,
                 "-E- csv file %s is not open\n",
                 m_file_name.c_str());
        IBIS_RETURN(1);
    }

    char line_buff[LINE_BUFF_SIZE] = {0};
    vec_str_t line_tokens;
    m_section_name_to_offset.clear();
    int num_line = 0;
    bool found_idx_table = false;
     while (m_cfs.good()) {
         num_line++;

         try {
             m_cfs.getline(line_buff, LINE_BUFF_SIZE);
         } catch (std::ifstream::failure &e) {
             // If the file does not contain index table, we read the file till
             // the end.
             if (m_cfs.eof()) {
                 m_cfs.clear();
                 break;
             } else {
                 throw;
             }
         }

         // check if we need to jump to index table
         if (strstr(line_buff, IBDIAGNET_INDEX_TABLE_FORMAT) == line_buff) {

             char* tok;
             u_int64_t index_table_offset;
             u_int32_t index_table_line;

             tok = strtok(line_buff, " ");

             while (strcmp(tok, IBDIAGNET_INDEX_TABLE_OFFSET))
                 tok = strtok(NULL, " ");
             tok = strtok(NULL, " ");
             csv_parser.Parse(tok, index_table_offset);

             while (strcmp(tok, IBDIAGNET_INDEX_TABLE_LINE))
                 tok = strtok(NULL, " ");
             tok = strtok(NULL, " ");
             csv_parser.Parse(tok, index_table_line);

             m_cfs.seekg(index_table_offset);
             num_line = index_table_line;
             m_cfs.getline(line_buff, LINE_BUFF_SIZE);
         }

         if (!strcmp(line_buff, START_INDEX_TABLE)) {
             found_idx_table = true;
             while  ((rc = csv_parser.GetNextLineAndSplitIntoTokens(m_cfs, line_buff, line_tokens)) == 0) {
                 if (!strcmp(line_buff, END_INDEX_TABLE))
                     break;
                 num_line++;
                 if (line_tokens.size() != NUM_FIELDS_IN_INDEX_SECTION) {
                     IBIS_LOG(TT_LOG_LEVEL_ERROR,
                             "-W- CSV Parser: incorrect number of columns in INDEX_TABLE"
                             " section at line number %d. Continue to next line.\n",
                             num_line);
                     continue;
                 }

                 struct offset_info off_info;
                 csv_parser.Parse(line_tokens[1], off_info.offset, 10);
                 csv_parser.Parse(line_tokens[2], off_info.length, 10);
                 csv_parser.Parse(line_tokens[3], off_info.start_line, 10);
                 csv_parser.Parse(line_tokens[4], off_info.num_lines, 10);

                 m_section_name_to_offset.insert(
                         pair_string_offset_info_t(line_tokens[0], off_info));
             }

             if (rc) {
                 IBIS_LOG(TT_LOG_LEVEL_ERROR,
                         "-E- CSV Parser: Faild to parse line in"
                         " INDEX_TABLE section for csv file %s \n",
                         m_file_name.c_str());
                 IBIS_RETURN(rc);
             }

             //breaking the "while(f_csv.good())"
             break;
         }

        // Construct index table
        if (strstr(line_buff, IBDIAGNET_SECTION_START_PREFIX)) {

            string section_name (line_buff + strlen(IBDIAGNET_SECTION_START_PREFIX));
            string section_end_str = IBDIAGNET_SECTION_END_PREFIX + section_name;;

            uint64_t table_offset = m_cfs.tellg();
            struct offset_info offset_info = {};
            offset_info.offset = table_offset;
            offset_info.start_line = num_line + 1;

            // Read table definition - nothing to do
            m_cfs.getline(line_buff, LINE_BUFF_SIZE);
            num_line++;

            // Read table lines
            m_cfs.getline(line_buff, LINE_BUFF_SIZE);
            num_line++;
            while (strcmp(line_buff, section_end_str.c_str())) {
                    offset_info.num_lines++;
                    offset_info.length = (u_int64_t)m_cfs.tellg() - table_offset;

                    m_cfs.getline(line_buff, LINE_BUFF_SIZE);
                    num_line++;
            }

            l_section_name_to_offset.insert(pair_string_offset_info_t(section_name, offset_info));
         }
     }

     if (l_section_name_to_offset.size() && !found_idx_table) {
         m_section_name_to_offset.insert(l_section_name_to_offset.begin(), l_section_name_to_offset.end());
     } else if (!found_idx_table) {
         IBIS_LOG(TT_LOG_LEVEL_ERROR,
                 "-E- CSV Parser: Faild to find"
                 " INDEX_TABLE section for csv file %s \n",
                 m_file_name.c_str());
     }

    IBIS_RETURN(rc);
}

int CsvParser::GetNextLineAndSplitIntoTokens(istream& str,
                                             char *line_buff,
                                             vec_str_t &line_tokens)
{
    IBIS_ENTER;
    int length = 0;

    memset(line_buff, 0, LINE_BUFF_SIZE);
    str.getline(line_buff, LINE_BUFF_SIZE);

    length = strlen(line_buff);
    if (!length) {
        IBIS_LOG(TT_LOG_LEVEL_ERROR,
                 "-W- CSV Parser: Found empty line. Can't split into tokens.\n");
        IBIS_RETURN(0);
    }

    line_tokens.clear();

    bool in_apostrophes = false;
    bool whitespace_prefix = true;
    int start_of_field = 0;
    int end_of_field = 0;

    for (int i = 0; i <= length; i++) {

        // Handle end of field:
        // End of field is either when reached comma (outside of apostrophes) or
        // end of line (i == length).
        if (i == length ||
            (!in_apostrophes &&
             line_buff[i] == CSV_PARSING_CHARS_STRING[COMMA]/*','*/)) {

            line_buff[end_of_field] = '\0';

            // If field is not empty (start_of_field < end_of_field), push
            // pointer to field value string beginning, otherwise push NULL
            if (start_of_field < end_of_field) {
                line_tokens.push_back(&line_buff[start_of_field]);
            } else {
                line_tokens.push_back(NULL);
            }

            start_of_field = end_of_field = i + 1;
            in_apostrophes = false;
            whitespace_prefix = true;
            continue;
        }

        // Handle apostrophes:
        // All symbols, including spaces and commas are part of current field
        // until last apostrophes are part of the field.
        if (line_buff[i] == CSV_PARSING_CHARS_STRING[APOSTROPHES]/*'"'*/) {
            if (whitespace_prefix) {
                // Indicates that this are the first apostrophes after
                // whitespace prefix.
                start_of_field = end_of_field = i + 1;
                whitespace_prefix = false;
                in_apostrophes = true;
                continue;
            }

            if (in_apostrophes) {
                in_apostrophes = false;
            }

            end_of_field = i;
            continue;
        }

        if (isspace(line_buff[i])) {
            if (whitespace_prefix) {
                // Field starts at the next index or later
                start_of_field = end_of_field = i + 1;
            }
        } else {
            end_of_field = i + 1;
            whitespace_prefix = false;
        }
    }

    IBIS_RETURN(0);
}



