/*
 * Copyright (c) 2004-2019 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include "config.h"
#include <errno.h>
#include <stdarg.h>
#include <string.h>

#include <infiniband/umad.h>

#include "ibis.h"
#include "ibis_version.h"

#include <byteswap.h>

#ifndef be64toh
  #if __BYTE_ORDER == __LITTLE_ENDIAN
    #define be64toh(x) bswap_64(x)
  #elif __BYTE_ORDER == __BIG_ENDIAN
    #define be64toh(x) x
  #endif
#endif
#ifndef htobe64
  #define htobe64 be64toh
#endif

typedef typeof(((struct umad_port *)0)->port_guid) umad_guid_t;

/****************************************************/

const char * get_ibis_version()
{
    #ifdef IBIS_VERSION
        return IBIS_VERSION;
    #else
        return "unknown";
    #endif
}

// Default logging function
void ibis_log::ibis_log_msg_function(const char *file_name,
                                     unsigned line_num,
                                     const char *function_name,
                                     int level,
                                     const char *format, ...)
{
#ifdef DEBUG
    char ext_format[1024];
    va_list args;

    sprintf(ext_format, "(%s, %d, %s): %s",
            file_name, line_num, function_name, format);
    va_start(args, format);
    tt_vlog(TT_LOG_MODULE_IBIS, level, ext_format, args);
    va_end(args);
#endif
    return;
}

void ibis_log::ibis_log_mad_function(dump_data_func_t dump_func,
                                     void *mad_obj,
                                     bool msg_send_mad)
{
#ifdef DEBUG
    // NOTE: implementation is not thread safe
    if ((tt_is_module_verbosity_active(TT_LOG_MODULE_IBIS)) &&
            (tt_is_level_verbosity_active(TT_LOG_LEVEL_MAD))) {
        if (msg_send_mad)
            IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending the following MAD:\n");
        else
            IBIS_LOG(TT_LOG_LEVEL_MAD, "Received the following MAD:\n");
        dump_func(mad_obj, TT_OUT_PORT);
    }
#endif
    return;
}

ibis_log_msg_function_t Ibis::m_log_msg_function = ibis_log::ibis_log_msg_function;
ibis_log_mad_function_t Ibis::m_log_mad_function = ibis_log::ibis_log_mad_function;

int pending_mad_data_t::init() {
    if (!m_umad) {
        m_umad_size = (unsigned int)(IBIS_IB_MAD_SIZE + umad_size());
        m_umad = new u_int8_t[m_umad_size];
        if (!m_umad)
            return -1;
    }
    m_transaction_data = NULL;
    return 0;
}

/****************************************************/
/******************private methods*******************/
/****************************************************/
void Ibis::SetLastError(const char *fmt, ...)
{
    IBIS_ENTER;
    char buff[IBIS_ERR_SIZE];
    va_list args;


    memset(buff, 0, sizeof(buff));
    va_start(args, fmt);
    vsprintf(buff, fmt, args);
    va_end(args);
    this->last_error.assign(buff);
    IBIS_LOG(TT_LOG_LEVEL_ERROR, "-E- %s\n", this->last_error.c_str());
    IBIS_RETURN_VOID;
}

int Ibis::RegisterClassVersionToUmad(u_int8_t mgmt_class)
{
    IBIS_ENTER;
    //register all classes
    long methods_mask[IBIS_METHODS_MASK_LEN];

    CalculateMethodMaskByClass(mgmt_class, methods_mask);
    class_versions_vec_t &class_versions =
        class_versions_by_class[mgmt_class];

    if (class_versions.empty()) {
        this->SetLastError("Failed to register unsupported mgmt_class: %u", mgmt_class);
        IBIS_RETURN(1);
    }

    for (unsigned int i = 0; i < class_versions.size(); ++i) {
        if ((this->umad_agents_by_class[mgmt_class][class_versions[i]] =
                    umad_register(this->umad_port_id,
                        mgmt_class,
                        class_versions[i],
                        IBIS_IB_RMPP_TYPE_NONE,
                        methods_mask)) < 0) {
            this->SetLastError("Failed to register for mgmt_class: %u class version %u",
                               mgmt_class, class_versions[i]);
            IBIS_RETURN(1);
        } else {
            IBIS_LOG(TT_LOG_LEVEL_DEBUG,
                     "Register umad_agent: %u for mgmt_class: %u class version %u \n",
                     umad_agents_by_class[mgmt_class][class_versions[i]],
                     mgmt_class, class_versions[i]);
        }
    }

    IBIS_RETURN(0);
}


// Returns: 0 - success / 1 - error
int Ibis::Bind()
{
    IBIS_ENTER;
    //open umad port
    if (this->dev_name == "") {
        this->umad_port_id = umad_open_port(NULL, this->port_num);
        IBIS_LOG(TT_LOG_LEVEL_INFO, "Device name not provided, auto-detect is enabled\n");
    } else {
        this->umad_port_id = umad_open_port((char *)this->dev_name.c_str(), this->port_num);
    }
    if (this->umad_port_id < 0) {
        this->SetLastError("Failed to umad_open_port");
        IBIS_RETURN(1);
    }

    //alloc buffers for umad
    unsigned int mad_buffer_size = (unsigned int)(IBIS_IB_MAD_SIZE + umad_size());
    if (!(this->p_umad_buffer_send = umad_alloc(1, mad_buffer_size))) {
        this->SetLastError("Failed to umad_allocate buffer for send");
        IBIS_RETURN(1);
    }
    if (!(this->p_umad_buffer_recv = umad_alloc(1, mad_buffer_size))) {
        this->SetLastError("Failed to umad_allocate buffer for recv");
        IBIS_RETURN(1);
    }
    memset(this->p_umad_buffer_send, 0, mad_buffer_size);
    memset(this->p_umad_buffer_recv, 0, mad_buffer_size);


    if (RegisterClassVersionToUmad(IBIS_IB_CLASS_SMI))
        IBIS_RETURN(1);

    IBIS_LOG(TT_LOG_LEVEL_DEBUG, "Register SMI class done\n");

    if (RegisterClassVersionToUmad(IBIS_IB_CLASS_SMI_DIRECT))
        IBIS_RETURN(1);

    IBIS_LOG(TT_LOG_LEVEL_DEBUG, "Register SMI direct class done\n");

    if (RegisterClassVersionToUmad(IBIS_IB_CLASS_PERFORMANCE))
        IBIS_RETURN(1);

    IBIS_LOG(TT_LOG_LEVEL_DEBUG, "Register PERFORMANCE class done\n");

    if (RegisterClassVersionToUmad(IBIS_IB_CLASS_VENDOR_MELLANOX))
        IBIS_RETURN(1);

    IBIS_LOG(TT_LOG_LEVEL_DEBUG, "Register VENDOR_MELLANOX class done\n");

    if (RegisterClassVersionToUmad(IBIS_IB_CLASS_CC))
        IBIS_RETURN(1);

    IBIS_LOG(TT_LOG_LEVEL_DEBUG, "Register CC class done\n");

    if (RegisterClassVersionToUmad(IBIS_IB_CLASS_AM))
        IBIS_RETURN(1);

    IBIS_LOG(TT_LOG_LEVEL_DEBUG, "Register AM class done\n");

    //TODO - register other classes

    //set packets pointers
    this->p_pkt_send = (u_int8_t *)umad_get_mad(this->p_umad_buffer_send);
    this->p_pkt_recv = (u_int8_t *)umad_get_mad(this->p_umad_buffer_recv);

    IBIS_RETURN(0);
}

int Ibis::AddMethodToClass(uint8_t mgmt_class, uint8_t method)
{
    IBIS_ENTER;

    if (this->IsReady() == true) {
        this->SetLastError("Ibis already binded to port");
        IBIS_RETURN(IBIS_MAD_STATUS_GENERAL_ERR);
    }

    if (!method || (method & 0x80)) {
        this->SetLastError("Invalid method 0x%02x", mgmt_class, method);
        IBIS_RETURN(IBIS_MAD_STATUS_GENERAL_ERR);
    }

    replier_methods_list_by_class[mgmt_class].push_back(method);

    IBIS_RETURN(0);
}

void Ibis::CalculateMethodMaskByClass(u_int8_t mgmt_class,
        long (&method_mask)[IBIS_METHODS_MASK_LEN]) {
    long mask;
    long *addr = method_mask;
    memset(method_mask, 0, sizeof(method_mask));

    for (methods_list_t::iterator it = replier_methods_list_by_class[mgmt_class].begin();
            it != replier_methods_list_by_class[mgmt_class].end(); ++it) {
        mask = 1L << *it % (8 * sizeof(long));
        addr += *it / (8 * sizeof(long));
        *addr |= mask;
    }
}

// Returns: 0 - success / 1 - error
int Ibis::SendMad(u_int8_t mgmt_class, int umad_timeout, int umad_retries)
{
    IBIS_ENTER;

    u_int8_t class_version = *(u_int8_t *)(this->p_pkt_send + 2);
    int agent_id = GetAgentId(mgmt_class, class_version);
    if (agent_id == -1) {
        IBIS_LOG(TT_LOG_LEVEL_ERROR,
                 "No umad agent was registered for mgmt_class: %u version %u \n",
                 mgmt_class, class_version);
        this->SetLastError("Failed to send mad, Failed to find umad agent.");
        IBIS_RETURN(1);
    }

#ifdef DEBUG
    //print in wireshark format
    string mad_in_hex = "";
    char curr_mad_buffer[64];
    for (int i = 0; i < 4; ++i) {
        memset(&curr_mad_buffer, 0 , 64);
        sprintf(curr_mad_buffer, "0x%2.2x ", this->p_pkt_send[i]);
        mad_in_hex += curr_mad_buffer;
    }
    mad_in_hex += "\n";
    int j = 0;
    for (int i = 4; i < IBIS_IB_MAD_SIZE; ++i, ++j) {
        if (((j % 8) == 0) && ((j % 16) != 0)){
            mad_in_hex += "\t";
        }
        else if ((j % 16) == 0) {
            mad_in_hex += "\n";
        }
        memset(&curr_mad_buffer, 0 , 64);
        sprintf(curr_mad_buffer, "0x%2.2x ", this->p_pkt_send[i]);
        mad_in_hex += curr_mad_buffer;
    }
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending MAD Packet: %s\n", mad_in_hex.c_str());
    IBIS_LOG(TT_LOG_LEVEL_DEBUG,
             "Sending MAD umad_agent: %u for mgmt_class: %u class version %u \n",
             agent_id, mgmt_class, class_version);

#endif          /* DEBUG */

    if (umad_send(this->umad_port_id,
                    agent_id,
                    this->p_umad_buffer_send,
                    IBIS_IB_MAD_SIZE,
                    umad_timeout,
                    umad_retries) < 0) {
        this->SetLastError("Failed to send mad");
        IBIS_RETURN(1);
    }
    IBIS_RETURN(0);
}

void Ibis::DumpReceivedMAD()
{
#ifdef DEBUG
    IBIS_ENTER;

    //print in wireshark format
    string mad_in_hex = "";
    char curr_mad_buffer[64];
    for (int i = 0; i < 4; ++i) {
        memset(&curr_mad_buffer, 0 , 64);
        sprintf(curr_mad_buffer, "0x%2.2x ", this->p_pkt_recv[i]);
        mad_in_hex += curr_mad_buffer;
    }
    mad_in_hex += "\n";
    int j = 0;
    for (int i = 4; i < IBIS_IB_MAD_SIZE; ++i, ++j) {
        if (((j % 8) == 0) && ((j % 16) != 0)){
            mad_in_hex += "\t";
        }
        else if ((j % 16) == 0) {
            mad_in_hex += "\n";
        }
        memset(&curr_mad_buffer, 0 , 64);
        sprintf(curr_mad_buffer, "0x%2.2x ", this->p_pkt_recv[i]);
        mad_in_hex += curr_mad_buffer;
    }
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Receiving MAD Packet: %s\n", mad_in_hex.c_str());
    IBIS_RETURN_VOID;
#endif          /* DEBUG */
}

int Ibis::GetAgentId(u_int8_t mgmt_class, uint8_t class_version)
{
    IBIS_ENTER;

    if (class_version > IBIS_IB_MAX_CLASS_VERSION_SUPP) {
        IBIS_LOG(TT_LOG_LEVEL_ERROR,
            "GetAgentId: Invalid class version %u (mgmt_class: %u)\n",
            class_version, mgmt_class);

        IBIS_RETURN(-1);
    }

    IBIS_RETURN(this->umad_agents_by_class[mgmt_class][class_version]);
}

int Ibis::CheckValidAgentIdForClass(int recv_agent_id, u_int8_t mgmt_class,
                                    uint8_t class_version)
{
    IBIS_ENTER;
    int agent_id = GetAgentId(mgmt_class, class_version);

    if (recv_agent_id != agent_id) {
        this->SetLastError(
            "Recv_agent_id=%u != Expected umad_agent: %u for mgmt_class: %u class version %u",
            recv_agent_id,
            agent_id,
            mgmt_class, class_version);
        IBIS_RETURN(1);
    }

    IBIS_RETURN(0);
}

// Returns: 0 - success / 1 - error
int Ibis::RecvMad(u_int8_t mgmt_class, int umad_timeout)
{
    IBIS_ENTER;
    int length = IBIS_IB_MAD_SIZE;
    int recv_agent_id = -1;
    if ((recv_agent_id = umad_recv(this->umad_port_id,
            this->p_umad_buffer_recv,
            &length,
            umad_timeout)) < 0) {
        IBIS_LOG(TT_LOG_LEVEL_MAD, "Failed to receive mad, mgmt_class=" U8H_FMT "\n", mgmt_class);
        IBIS_RETURN(1);
    }

    u_int8_t class_version = *(u_int8_t *)(this->p_pkt_recv + 2);

    if (CheckValidAgentIdForClass(recv_agent_id, mgmt_class, class_version)) {
        IBIS_RETURN(1);
    }

#ifdef DEBUG
    //print in wireshark format
    DumpReceivedMAD();
#endif          /* DEBUG */
    IBIS_RETURN(0);
}

// Returns: 0 - success / 1 - error
int Ibis::RecvAsyncMad(int umad_timeout)
{
    IBIS_ENTER;
    int length = IBIS_IB_MAD_SIZE;
    int recv_agent_id = -1;
    if ((recv_agent_id = umad_recv(this->umad_port_id,
            this->p_umad_buffer_recv,
            &length,
            umad_timeout)) < 0) {
        IBIS_LOG(TT_LOG_LEVEL_MAD,"Failed to receive mad\n");
        IBIS_RETURN(1);
    }
    u_int8_t mgmt_class = *(u_int8_t *)(this->p_pkt_recv + 1); //mad_get_field(this->p_pkt_recv, 0, IB_MAD_MGMTCLASS_F);
    u_int8_t class_version = *(u_int8_t *)(this->p_pkt_recv + 2);
    if (CheckValidAgentIdForClass(recv_agent_id, mgmt_class, class_version)) {
        IBIS_RETURN(1);
    }

#ifdef DEBUG
    //print in wireshark format
    DumpReceivedMAD();
#endif          /* DEBUG */
    IBIS_RETURN(0);
}


int Ibis::ReceiveUnsolicitedMad(int timeout_ms)
{
    IBIS_ENTER;

    int length = IBIS_IB_MAD_SIZE;
    int agent_id = -1;
    u_int8_t mgmt_class;
    u_int8_t method;
    u_int16_t attribute_id;
    u_int8_t class_version;

    // TODO: handle MADs larger than 256 bytes
    if ((agent_id = umad_recv(this->umad_port_id,
            this->p_umad_buffer_recv,
            &length,
            timeout_ms)) < 0) {
        if (agent_id == -ETIMEDOUT) {
            IBIS_RETURN(0);
        } else {
            this->SetLastError("umad_recv ended with rc %d", agent_id);
            IBIS_RETURN(1);
        }
    }

    mgmt_class = *(u_int8_t *)(this->p_pkt_recv + 1);
    class_version = *(u_int8_t *)(this->p_pkt_recv + 2);
    method =  *(u_int8_t *)(this->p_pkt_recv + 3);
    attribute_id = ADB2C_BE16_TO_CPU(*(u_int16_t *)(this->p_pkt_recv + 16));

    if (CheckValidAgentIdForClass(agent_id, mgmt_class, class_version)) {
        IBIS_RETURN(1);
    }

#ifdef DEBUG
    DumpReceivedMAD();
#endif

    // Find handler for incoming MAD type
    attr_method_pair_to_handler_map_t::iterator handler_iter;
    pair<u_int16_t, u_int8_t> key = make_pair(attribute_id, method);

    handler_iter = m_mad_handlers_by_class[mgmt_class].find(key);
    if (handler_iter == m_mad_handlers_by_class[mgmt_class].end()) {
        this->SetLastError("No handler registered for management class 0x%02x, "
                           "attribute ID 0x%04x method 0x%02x",
                           mgmt_class, attribute_id, method);
        IBIS_RETURN(1);
    }

    mad_handler_t *p_handler = &handler_iter->second;

    ib_user_mad_t *p_mad = (ib_user_mad_t *)this->p_umad_buffer_recv;
    ib_address_t ib_address(ntohs(p_mad->addr.lid),
                            ntohl(p_mad->addr.qpn),
                            ntohl(p_mad->addr.qkey),
                            p_mad->addr.sl);

    // bit fields are represented by 8 bits
    u_int8_t attribute_data[IBIS_IB_MAD_SIZE * 8] = {};
    u_int8_t class_data[IBIS_IB_MAD_SIZE * 8] = {};

    p_handler->m_unpack_class_data_func(&class_data, this->p_pkt_recv);
    p_handler->m_unpack_attribute_data_func(&attribute_data,
                                            this->p_pkt_recv + p_handler->m_data_offset);

    p_handler->m_callback_func(&ib_address, &class_data, &attribute_data,
                               p_handler->m_context);

    IBIS_RETURN(0);
}

int Ibis::RegisterMadHandler(u_int8_t mgmt_class, u_int16_t attribute_id,
                             u_int8_t method,
                             u_int8_t data_offset,
                             unpack_data_func_t unpack_class_data_func,
                             dump_data_func_t   dump_class_data_func,
                             unpack_data_func_t unpack_attribute_data_func,
                             dump_data_func_t dump_attribute_data_func,
                             mad_handler_callback_func_t callback_func,
                             void *context)
{
    mad_handler_t mad_handler;

    mad_handler.m_data_offset = data_offset;
    mad_handler.m_unpack_class_data_func = unpack_class_data_func;
    mad_handler.m_dump_class_data_func = dump_class_data_func;
    mad_handler.m_unpack_attribute_data_func = unpack_attribute_data_func;
    mad_handler.m_dump_attribute_data_func = dump_attribute_data_func;
    mad_handler.m_callback_func = callback_func;
    mad_handler.m_context = context;

    pair<u_int16_t, u_int8_t> key = make_pair(attribute_id, method);
    attr_method_pair_to_handler_map_t::iterator handler_iter;

    handler_iter = m_mad_handlers_by_class[mgmt_class].find(key);
    if (handler_iter != m_mad_handlers_by_class[mgmt_class].end()) {
        this->SetLastError("Mad handler already registerd for management class "
                           "0x%02x, attribute ID 0x%04x, method 0x%02x",
                           mgmt_class, attribute_id, method);
        IBIS_RETURN(1);
    }

    // Register to class/method
    if (AddMethodToClass(mgmt_class, method)) {
        IBIS_RETURN(1);
    }

    m_mad_handlers_by_class[mgmt_class].insert(make_pair(key, mad_handler));

    IBIS_RETURN(0);
}


/*
 * Returns: mad status[bits: 0-15] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::DoRPC(u_int8_t mgmt_class)
{
    IBIS_ENTER;
    if (this->IsReady() == false) {
        this->SetLastError("Ibis setting port wasn't done");
        IBIS_RETURN(IBIS_MAD_STATUS_GENERAL_ERR);
    }

    u_int32_t trid_new;
    u_int32_t trid_org = *(u_int32_t *)(this->p_pkt_send + 12); //we will look at lower 32 bits
                                //of the tid field because umad lib take care of higher 32 bits,
                                //a pid of the process will be assigned there.

    //try to send + recv until status of response is ok
    for (int current_retry = 0; current_retry < this->retries; ++current_retry) {
        if (this->SendMad(mgmt_class, this->timeout, 0))
            IBIS_RETURN(IBIS_MAD_STATUS_SEND_FAILED);

        /* Use same timeout on receive side just in case send packet is lost somewhere */
        IBIS_LOG(TT_LOG_LEVEL_DEBUG, "Try number=%u to receive response for mad with TID=%u\n",
                    current_retry, trid_org);
        do {
            if (this->RecvMad(mgmt_class, this->timeout))
                IBIS_RETURN(IBIS_MAD_STATUS_RECV_FAILED);
            trid_new = *(u_int32_t *)(this->p_pkt_recv + 12);
            IBIS_LOG(TT_LOG_LEVEL_DEBUG, "(Original TID=%u, Received TID=%u)\n" , trid_org, trid_new);
        } while (trid_org != trid_new);         //make sure same tid as one that was sent


        int status = umad_status(this->p_umad_buffer_recv);
        IBIS_LOG(TT_LOG_LEVEL_DEBUG, "UMAD Status=" U16H_FMT "\n", status);
        if ((!status) || (status == ENOMEM))    {   /* done */
            //check if method response
            u_int8_t mad_method = *(u_int8_t *)(this->p_pkt_recv + 3);
            if (mad_method != IBIS_IB_MAD_METHOD_GET_RESPONSE) {
                IBIS_LOG(TT_LOG_LEVEL_MAD, "Failed to get response within the given time out");
                IBIS_RETURN(IBIS_MAD_STATUS_TIMEOUT);
            }
            //check if status success
            u_int16_t mad_status = ADB2C_BE16_TO_CPU(*(u_int16_t *)(this->p_pkt_recv + 4));
            IBIS_LOG(TT_LOG_LEVEL_DEBUG, "MAD Status=" U16H_FMT " description=%s",
                    mad_status, this->ConvertMadStatusToStr(mad_status).c_str());
            IBIS_RETURN(mad_status);
        }
        IBIS_LOG(TT_LOG_LEVEL_DEBUG, "Failed to receive response for mad with TID=%u - umad status=%u\n",
                    trid_org, status);
    }
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Failed to get response within the given time out");
    IBIS_RETURN(IBIS_MAD_STATUS_TIMEOUT);
}

/*
 * Returns: mad status[bits: 0-15] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::DoAsyncSend(u_int8_t mgmt_class)
{
    IBIS_ENTER;
    if (this->IsReady() == false) {
        this->SetLastError("Ibis setting port wasn't done");
        IBIS_RETURN(IBIS_MAD_STATUS_GENERAL_ERR);
    }

    //we will look at lower 32 bits of the tid field
    //because umad lib take care of higher 32 bits,
    //a pid of the process will be assigned there.
    IBIS_LOG(TT_LOG_LEVEL_DEBUG, "Send MAD with TID=%u\n",
             *(u_int32_t *)(this->p_pkt_send + 12));

    if (this->SendMad(mgmt_class, this->timeout, this->retries))
        IBIS_RETURN(IBIS_MAD_STATUS_SEND_FAILED);

    IBIS_RETURN(IBIS_MAD_STATUS_SUCCESS);
}

int Ibis::DoAsyncRec()
{
    IBIS_ENTER;
    //Multiply by 2 to make sure we will get timeout from driver
    if (this->RecvAsyncMad(this->timeout*this->retries * 2))
        IBIS_RETURN(IBIS_MAD_STATUS_RECV_FAILED);

    IBIS_LOG(TT_LOG_LEVEL_DEBUG, "Receive MAD with TID=%u\n",
             *(u_int32_t *)(this->p_pkt_recv + 12));

    int status = umad_status(this->p_umad_buffer_recv);
    IBIS_LOG(TT_LOG_LEVEL_DEBUG, "UMAD Status=" U16H_FMT "\n", status);
    if ((!status) || (status == ENOMEM))    {   /* done */
        //check if method response
        u_int8_t mad_method = *(u_int8_t *)(this->p_pkt_recv + 3);
        if (mad_method != IBIS_IB_MAD_METHOD_GET_RESPONSE) {
            IBIS_LOG(TT_LOG_LEVEL_MAD, "Failed to get response within the given time out\n");
            IBIS_RETURN(IBIS_MAD_STATUS_TIMEOUT);
        }
        //check if status success
        u_int16_t mad_status = ADB2C_BE16_TO_CPU(*(u_int16_t *)(this->p_pkt_recv + 4));
        IBIS_LOG(TT_LOG_LEVEL_DEBUG, "MAD Status=" U16H_FMT " description=%s",
                mad_status, this->ConvertMadStatusToStr(mad_status).c_str());
        IBIS_RETURN(mad_status);
    }

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Failed to get response within the given time out\n");
    IBIS_RETURN(IBIS_MAD_STATUS_TIMEOUT);
}

void Ibis::InvokeCallbackFunction(const clbck_data_t &clbck_data,
                                  int rec_status,
                                  void *p_attribute_data)
{
    suppressMadSending = true;
    clbck_data.m_handle_data_func(clbck_data, rec_status, p_attribute_data);
    suppressMadSending = false;
}

/****************************************************/
/*******************public methods*******************/
/****************************************************/
const char* Ibis::GetLastError()
{
    IBIS_ENTER;
    if (this->last_error != "")
        IBIS_RETURN(this->last_error.c_str());
    IBIS_RETURN("Unknown");
}

string Ibis::ConvertMadStatusToStr(u_int16_t status)
{
    IBIS_ENTER;

    string result = "";
    u_int16_t mad_status = status & 0x00ff;   //8-15 bits are class specific

    switch (mad_status) {
    case IBIS_MAD_STATUS_SUCCESS: {
        result = "Success";
        break;
    }
    case IBIS_MAD_STATUS_BUSY: {
        result = "Temporarily busy. MAD discarded. This is not an error.";
        break;
    }
    case IBIS_MAD_STATUS_REDIRECT: {
        result = "Redirection. This is not an error.";
        break;
    }
    case IBIS_MAD_STATUS_UNSUP_CLASS_VER: {
        result = "Bad version. Either the base version, or the class version, " \
                "or the combination of the two is not supported.";
        break;
    }
    case IBIS_MAD_STATUS_UNSUP_METHOD: {
        result = "The method specified is not supported.";
        break;
    }
    case IBIS_MAD_STATUS_UNSUP_METHOD_ATTR: {
        result = "The method/attribute combination is not supported.";
        break;
    }
    case IBIS_MAD_STATUS_INVALID_FIELD: {
        result = "One or more fields in the attribute "  \
                "or attribute modifier contain an invalid value. "  \
                "Invalid values include reserved values and values that "   \
                "exceed architecturally defined limits.";
        break;
    }
    default: {
        result = "UNKNOWN";
        break;
    }
    }

    IBIS_RETURN(result);
};


string Ibis::ConvertDirPathToStr(direct_route_t *p_curr_direct_route)
{
    IBIS_ENTER;
    if (!p_curr_direct_route)
        IBIS_RETURN("");
    char buffer[8];             //byte is 8 bit ==> 256 is max value ==> we need max 4 chars to store it
    string result = " [";
    for (int i = 0; i < p_curr_direct_route->length - 1; ++i) {
        sprintf(buffer, "%u,", p_curr_direct_route->path.BYTE[i]);
        result += buffer;
    }
    sprintf(buffer, "%u", p_curr_direct_route->path.BYTE[p_curr_direct_route->length - 1]);
    result += buffer;
    result += "]";
    IBIS_RETURN(result);
};


Ibis::Ibis():
        port_num(-1),
        ibis_status(NOT_INITILIAZED),
        p_umad_buffer_send(NULL), p_umad_buffer_recv(NULL), p_pkt_send(NULL), p_pkt_recv(NULL),
        mads_counter(0), p_mkeymngr(NULL), umad_port_id(-1),
        timeout(IBIS_TIMEOUT), retries(IBIS_RETRIES),usePSL(false),
        m_pending_gmps(0), m_pending_smps(0), m_max_gmps_on_wire(0), m_max_smps_on_wire(0),
        suppressMadSending(false)

{
    IBIS_ENTER;
    for (int mgmt_class = 0; mgmt_class < IBIS_IB_MAX_MAD_CLASSES; ++mgmt_class) {
        for (uint8_t version = 0; version <= IBIS_IB_MAX_CLASS_VERSION_SUPP; ++version) {
            this->umad_agents_by_class[mgmt_class][version] = -1;
        }
    }

    IBIS_RETURN_VOID;
}


Ibis::~Ibis()
{
    IBIS_ENTER;
    /* free umad buffers */
    if (this->p_umad_buffer_send)
        umad_free(this->p_umad_buffer_send);
    if (this->p_umad_buffer_recv)
        umad_free(this->p_umad_buffer_recv);

    for (transactions_map_t::iterator tr_iter = transactions_map.begin();
         tr_iter !=  transactions_map.end(); ++tr_iter)
        delete tr_iter->second;

    m_mads_on_node_map.clear();

    /* close port + unregister all classes */
    if (this->umad_port_id != -1) {
        for (int mgmt_class = 0; mgmt_class < IBIS_IB_MAX_MAD_CLASSES; ++mgmt_class) {
            for (uint8_t version = 0; version <= IBIS_IB_MAX_CLASS_VERSION_SUPP; ++version) {
                if (this->umad_agents_by_class[mgmt_class][version] != -1) {
                    IBIS_LOG(TT_LOG_LEVEL_DEBUG,
                             "Trying to unregister umad agent_id=%u, mgmt=%u version=%u\n",
                             this->umad_agents_by_class[mgmt_class][version],
                             mgmt_class, version);
                    if (umad_unregister(this->umad_port_id,
                                        this->umad_agents_by_class[mgmt_class][version])) {
                        this->SetLastError(
                            "Failed to unregister umad agent_id=%u, mgmt=%u version=%u",
                            this->umad_agents_by_class[mgmt_class][version],
                            mgmt_class, version);
                    }
                }
            }
        }
        if (umad_close_port(this->umad_port_id)) {
            this->SetLastError("Failed to close the umad port");
        }
    }

    /* close umad lib */
    if (umad_done()) {
        this->SetLastError("Failed to close UMAD library");
    }
    IBIS_RETURN_VOID;
}

void Ibis::InitClassVersionsDB()
{
    IBIS_ENTER;
    class_versions_by_class[IBIS_IB_CLASS_SMI].push_back(1);

    class_versions_by_class[IBIS_IB_CLASS_SMI_DIRECT].push_back(1);

    class_versions_by_class[IBIS_IB_CLASS_SA].push_back(2);

    class_versions_by_class[IBIS_IB_CLASS_PERFORMANCE].push_back(1);

    class_versions_by_class[IBIS_IB_CLASS_VENDOR_MELLANOX].push_back(1);

    class_versions_by_class[IBIS_IB_CLASS_DEVICE_MGMT].push_back(1);

    class_versions_by_class[IBIS_IB_CLASS_CC].push_back(1);

    class_versions_by_class[IBIS_IB_CLASS_BOARD_MGMT].push_back(1);

    class_versions_by_class[IBIS_IB_CLASS_AM].push_back(1);
    class_versions_by_class[IBIS_IB_CLASS_AM].push_back(2);

    IBIS_RETURN_VOID;
}


int Ibis::Init()
{
    IBIS_ENTER;
    if (this->IsInit() == false) {
        //TODO - fix umad_init to be call only once
        if (umad_init()) {
            this->SetLastError("Failed to initialize umad");
            IBIS_RETURN(1);
        }

#ifdef DEBUG_UMAD
        umad_debug(5);
#endif

        this->ibis_status = NOT_SET_PORT;
        IBIS_LOG(TT_LOG_LEVEL_INFO, "Ibis initialize done\n");
    }

    InitClassVersionsDB();
    IBIS_RETURN(0);
}


// Returns: 0 - success / 1 - error
int Ibis::SetPort(const char* device_name, phys_port_t port_num)
{
    IBIS_ENTER;
    if (this->IsInit() == false) {
        this->SetLastError("Ibis initialize wasn't done");
        IBIS_RETURN(1);
    }
    if (this->IsReady() == true) {
        this->SetLastError("Ibis set_port was already done");
        IBIS_RETURN(1);
    }

    //set device name + port number
    this->dev_name = (device_name) ? device_name : "";
    this->port_num = port_num;

    //bind
    int i_rc = this->Bind();
    this->ibis_status = READY;
    IBIS_RETURN(i_rc);
}


// Returns: 0 - success / 1 - error
int Ibis::SetPort(u_int64_t port_guid)
{
    IBIS_ENTER;
    if (this->IsInit() == false) {
        this->SetLastError("Ibis initialize wasn't done");
        IBIS_RETURN(1);
    }
    if (this->IsReady() == true) {
        this->SetLastError("Ibis set_port was already done");
        IBIS_RETURN(1);
    }

    char ca_names_array[IBIS_MAX_CAS][UMAD_CA_NAME_LEN];
    int ca_num = 0;
    umad_guid_t portguids_array[IBIS_MAX_PORTS_PER_CA + 1];
    int portguids_num = 0;
    umad_ca_t umad_ca;

    //if guid is zero than find the first active port
    if (!port_guid) {
        this->dev_name = "";
        this->port_num = 0;
        goto _found;
    }
    //get all local cas
    if ((ca_num = umad_get_cas_names(ca_names_array,
            IBIS_MAX_CAS)) < 0) {
        this->SetLastError("Failed to umad_get_cas_names");
        IBIS_RETURN(1);
    }
    //check for requested guid
    for (int ca_idx = 0; ca_idx < ca_num; ++ca_idx) {
        if ((portguids_num = umad_get_ca_portguids(ca_names_array[ca_idx],
                portguids_array,
                IBIS_MAX_PORTS_PER_CA + 1)) < 0) {
            this->SetLastError("Failed to umad_get_ca_portguids");
            IBIS_RETURN(1);
        }
        // use unsigned int index to prevent overwrap
        for (int port_num_idx = 0; port_num_idx < portguids_num; ++port_num_idx)
            if (port_guid == portguids_array[port_num_idx]) {
                this->dev_name = ca_names_array[ca_idx];
                this->port_num = (phys_port_t)port_num_idx;
                goto _found;
            }
    }
    /* no local CA owns this guid */
    this->SetLastError("Unable to find requested guid " U64H_FMT, port_guid);
    IBIS_RETURN(1);

_found:
    /* validate that node is an IB node type */
    char buff_dev_name[UMAD_CA_NAME_LEN];
    if (this->dev_name == "") {
        if (umad_get_ca(NULL, &umad_ca) < 0) {
            this->SetLastError("Failed to umad_get_ca");
            IBIS_RETURN(1);
        }
    } else {
        strcpy(buff_dev_name, this->dev_name.c_str());
        if (umad_get_ca(buff_dev_name, &umad_ca) < 0) {
            this->SetLastError("Failed to umad_get_ca");
            IBIS_RETURN(1);
        }
    }
    if (umad_ca.node_type < 1 || umad_ca.node_type > 3) {
        this->SetLastError("Type %d of node \'%s\' is not an IB node type\n",
                umad_ca.node_type, umad_ca.ca_name);
        umad_release_ca(&umad_ca);
        IBIS_RETURN(1);
    }
    umad_release_ca(&umad_ca);

    /* port found, try to bind */
    int i_rc = this->Bind();
    this->ibis_status = READY;
    IBIS_RETURN(i_rc);
}


// Returns: 0 - success / 1 - error
int Ibis::GetAllLocalPortGUIDs(OUT local_port_t local_ports_array[IBIS_MAX_LOCAL_PORTS],
        OUT u_int32_t *p_local_ports_num)
{
    IBIS_ENTER;
    if (this->IsInit() == false) {
        this->SetLastError("Ibis initialize wasn't done");
        IBIS_RETURN(1);
    }

    memset(local_ports_array, 0, IBIS_MAX_LOCAL_PORTS * sizeof(local_port_t));
    *p_local_ports_num = 0;
    char ca_names_array[IBIS_MAX_CAS][UMAD_CA_NAME_LEN];
    int ca_num = 0;
    umad_guid_t portguids_array[IBIS_MAX_PORTS_PER_CA + 1];
    int portguids_num = 0;
    umad_port_t umad_port;

    //get all local cas
    if ((ca_num = umad_get_cas_names(ca_names_array,
            IBIS_MAX_CAS)) < 0) {
        this->SetLastError("Failed to umad_get_cas_names");
        IBIS_RETURN(1);
    }
    //go over all cas
    for (int ca_idx = 0; ca_idx < ca_num; ++ca_idx) {
        if ((portguids_num = umad_get_ca_portguids(ca_names_array[ca_idx],
                portguids_array,
                IBIS_MAX_PORTS_PER_CA + 1)) < 0) {
            this->SetLastError("Failed to umad_get_ca_portguids");
            IBIS_RETURN(1);
        }
        //go over all ports for this ca
        for (int port_num_idx = 0; port_num_idx < portguids_num; ++port_num_idx) {
            if (!portguids_array[port_num_idx])
                continue;
            if (umad_get_port(ca_names_array[ca_idx], port_num_idx, &umad_port) < 0) {
                this->SetLastError("Failed to umad_get_port");
                IBIS_RETURN(1);
            }
            local_ports_array[*p_local_ports_num].guid = portguids_array[port_num_idx];
            local_ports_array[*p_local_ports_num].lid = (u_int16_t)umad_port.base_lid;
            local_ports_array[*p_local_ports_num].logical_state = (u_int8_t)umad_port.state;
            ++(*p_local_ports_num);
            umad_release_port(&umad_port);
        }
    }
    IBIS_RETURN(0);
}

// Returns: 0 - success / 1 - error
int Ibis::GetLocalPortProperties(OUT port_properties_t *p_port_properties)
{
    umad_port_t umad_port;
    char *ca_name;

    IBIS_ENTER;
    if (this->IsReady() == false) {
        this->SetLastError("Ibis setting port wasn't done");
        IBIS_RETURN(1);
    }

    ca_name = this->dev_name == "" ? NULL :
	      const_cast<char *>(this->dev_name.c_str());
    if (umad_get_port(ca_name, this->port_num, &umad_port)) {
        this->SetLastError("Failed to umad_get_port");
        IBIS_RETURN(1);
    }

    p_port_properties->base_lid = umad_port.base_lid;
    p_port_properties->sm_lid = umad_port.sm_lid;
    p_port_properties->subnet_prefix = be64toh(umad_port.gid_prefix);
    p_port_properties->port_guid = be64toh(umad_port.port_guid);

    umad_release_port(&umad_port);

    IBIS_RETURN(0);
}

// Returns: 0 - success / 1 - error
int Ibis::SetSendMadAddr(int d_lid, int d_qp, int sl, int qkey)
{
    IBIS_ENTER;
    if (this->IsReady() == false) {
        this->SetLastError("Ibis setting port wasn't done");
        IBIS_RETURN(1);
    }
    if (umad_set_addr(this->p_umad_buffer_send, d_lid, d_qp, sl, qkey) < 0) {
        this->SetLastError("Failed to set destination address, lid=%u", d_lid);
        IBIS_RETURN(1);
    }
    IBIS_RETURN(0);
}

int Ibis::SetPSLTable(const vector<uint8_t > &PSLTable)
{
    IBIS_ENTER;
    PSL = PSLTable;
    usePSL = true;
    IBIS_RETURN(0);
}

uint8_t
Ibis::getPSLForLid(u_int16_t lid)
{
    IBIS_ENTER;
    if (PSL.empty())
        IBIS_RETURN(usePSL ? (uint8_t)IB_SLT_UNASSIGNED : 0);
    if (PSL.size() < (unsigned)lid+1)
        return IB_SLT_UNASSIGNED;
    IBIS_RETURN(PSL[lid]);
}

// Set the PSL Table:
void
Ibis::setPSLForLid(lid_t lid, lid_t maxLid, uint8_t sl)
{
    unsigned int prev_size = (unsigned int)PSL.size();
    unsigned int new_size = (unsigned int)(maxLid + 1);
    if (prev_size < new_size) {
        PSL.resize(new_size);
        for(unsigned int i = prev_size; i<new_size; i++)
            PSL[i] = IB_SLT_UNASSIGNED;
    } 
    PSL[lid] = sl;
    usePSL = true;
}

