/*
 *  Copyright (c) 2004-2013 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <string.h>

#include "ibis.h"
#include "packets/packets_layouts.h"


/****************************************************/
/******************private methods*******************/
/****************************************************/
/*
 * Returns: mad status[bits: 0-15] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCMadGetSet(u_int16_t lid,
        u_int8_t sl,
        u_int8_t method,
        u_int16_t attribute_id,
        u_int32_t attribute_modifier,
        u_int64_t cc_key,
        void *p_cc_log_attribute_data,
        void *p_cc_mgt_attribute_data,
        const pack_data_func_t cc_pack_attribute_data_func,
        const unpack_data_func_t cc_unpack_attribute_data_func,
        const dump_data_func_t cc_dump_attribute_data_func,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    struct MAD_CongestionControl cc_mad;

    CLEAR_STRUCT(cc_mad);

    //build cc lid route header
    this->CommonMadHeaderBuild(&cc_mad.MAD_Header_Common,
            IBIS_IB_CLASS_CC,
            method,
            attribute_id,
            attribute_modifier);

    cc_mad.CC_Key = cc_key;

    /*
    if (p_cc_mgt_attribute_data) {
        IBIS_LOG(TT_LOG_LEVEL_MAD, "p_cc_mgt_attribute_data = %u\n", *(int*)p_cc_mgt_attribute_data);
    } else if (p_cc_log_attribute_data) {
        IBIS_LOG(TT_LOG_LEVEL_MAD, "p_cc_log_attribute_data = %u\n", *(int*)p_cc_log_attribute_data);
    }
    */
    //send the mad
    IBIS_RETURN(this->MadGetSet(lid, 1, sl, IBIS_IB_DEFAULT_QP1_QKEY,
            IBIS_IB_CLASS_CC,
            method, attribute_id, attribute_modifier,
            (p_cc_mgt_attribute_data) ? IBIS_IB_MGT_DATA_OFFSET_CC : IBIS_IB_LOG_DATA_OFFSET_CC,
            &cc_mad, 
            (p_cc_mgt_attribute_data) ? p_cc_mgt_attribute_data : p_cc_log_attribute_data,
            (const pack_data_func_t)MAD_CongestionControl_pack,
            (const unpack_data_func_t)MAD_CongestionControl_unpack,
            (const dump_data_func_t)MAD_CongestionControl_dump,
            (const pack_data_func_t)cc_pack_attribute_data_func,
            (const unpack_data_func_t)cc_unpack_attribute_data_func,
            (const dump_data_func_t)cc_dump_attribute_data_func,
            p_clbck_data));
}


/****************************************************/
/*******************public methods*******************/
/****************************************************/
/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCClassPortInfoGet(u_int16_t lid,
        u_int8_t sl,
        u_int64_t cc_key,
        struct IB_ClassPortInfo *p_ib_class_port_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_ib_class_port_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending IB_ClassPortInfo Get MAD lid = %u\n", lid);
    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CLASS_PORT_INFO,
            0,                        /* attribute modifier */
            cc_key,
            NULL,                     /* p_cc_log_attribute_data */
            p_ib_class_port_info,     /* p_cc_mgt_attribute_data */
            (const pack_data_func_t)IB_ClassPortInfo_pack,
            (const unpack_data_func_t)IB_ClassPortInfo_unpack,
            (const dump_data_func_t)IB_ClassPortInfo_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCClassPortInfoSet(u_int16_t lid,
        u_int8_t sl,
        u_int64_t cc_key,
        struct IB_ClassPortInfo *p_ib_class_port_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending IB_ClassPortInfo Get MAD lid = %u\n", lid);
    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_CLASS_PORT_INFO,
            0,                        /* attribute modifier */
            cc_key,
            NULL,                     /* p_cc_log_attribute_data */
            p_ib_class_port_info,     /* p_cc_mgt_attribute_data */
            (const pack_data_func_t)IB_ClassPortInfo_pack,
            (const unpack_data_func_t)IB_ClassPortInfo_unpack,
            (const dump_data_func_t)IB_ClassPortInfo_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCCongestionInfoGet(u_int16_t lid,
        u_int8_t sl,
        u_int64_t cc_key,
        struct CC_CongestionInfo *p_cc_congestion_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_congestion_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCCongestionInfo Get MAD lid = %u\n", lid);
    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CONGESTION_CONT_CONG_INFO,
            0,                        /* attribute modifier */
            cc_key,
            NULL,                     /* p_cc_log_attribute_data */
            p_cc_congestion_info,     /* p_cc_mgt_attribute_data */
            (const pack_data_func_t)CC_CongestionInfo_pack,
            (const unpack_data_func_t)CC_CongestionInfo_unpack,
            (const dump_data_func_t)CC_CongestionInfo_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCCongestionKeyInfoGet(u_int16_t lid,
        u_int8_t sl,
        u_int64_t cc_key,
        struct CC_CongestionKeyInfo *p_cc_congestion_key_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_congestion_key_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCCongestionKeyInfo Get MAD lid = %u\n", lid);
    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CONGESTION_CONT_CONG_KEY_INFO,
            0,                        /* attribute modifier */
            cc_key,
            NULL,                     /* p_cc_log_attribute_data */
            p_cc_congestion_key_info, /* p_cc_mgt_attribute_data */
            (const pack_data_func_t)CC_CongestionKeyInfo_pack,
            (const unpack_data_func_t)CC_CongestionKeyInfo_unpack,
            (const dump_data_func_t)CC_CongestionKeyInfo_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCCongestionKeyInfoSet(u_int16_t lid,
        u_int8_t sl,
        u_int64_t cc_key,
        struct CC_CongestionKeyInfo *p_cc_congestion_key_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCCongestionKeyInfo Set MAD lid = %u\n", lid);
    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_CONGESTION_CONT_CONG_KEY_INFO,
            0,                        /* attribute modifier */
            cc_key,
            NULL,                     /* p_cc_log_attribute_data */
            p_cc_congestion_key_info, /* p_cc_mgt_attribute_data */
            (const pack_data_func_t)CC_CongestionKeyInfo_pack,
            (const unpack_data_func_t)CC_CongestionKeyInfo_unpack,
            (const dump_data_func_t)CC_CongestionKeyInfo_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}
	

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCCongestionLogSwitchGet(u_int16_t lid,
        u_int8_t sl,
        u_int64_t cc_key,
        struct CC_CongestionLogSwitch *p_cc_congestion_log_sw,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_congestion_log_sw);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCCongestionLogSwitch Get MAD lid = %u\n", lid);
    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CONGESTION_CONT_CONG_LOG,
            0,                       /* attribute modifier */
            cc_key,
            p_cc_congestion_log_sw,  /* p_cc_log_attribute_data */
            NULL,                    /* p_cc_mgt_attribute_data */
            (const pack_data_func_t)CC_CongestionLogSwitch_pack,
            (const unpack_data_func_t)CC_CongestionLogSwitch_unpack,
            (const dump_data_func_t)CC_CongestionLogSwitch_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCCongestionLogCAGet(u_int16_t lid,
        u_int8_t sl,
        u_int64_t cc_key,
        struct CC_CongestionLogCA *p_cc_congestion_log_ca,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_congestion_log_ca);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCCongestionLogCA Get MAD lid = %u\n", lid);
    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CONGESTION_CONT_CONG_LOG,
            0,                       /* attribute modifier */
            cc_key,
            p_cc_congestion_log_ca,  /* p_cc_log_attribute_data */
            NULL,                    /* p_cc_mgt_attribute_data */
            (const pack_data_func_t)CC_CongestionLogCA_pack,
            (const unpack_data_func_t)CC_CongestionLogCA_unpack,
            (const dump_data_func_t)CC_CongestionLogCA_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCSwitchCongestionSettingGet(u_int16_t lid,
        u_int8_t sl,
        u_int64_t cc_key,
        struct CC_SwitchCongestionSetting *p_cc_sw_congestion_setting,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_sw_congestion_setting);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCSwitchCongestionSetting Get MAD lid = %u\n", lid);
    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CONGESTION_CONT_SW_CONG_SETTING,
            0,                          /* attribute modifier */
            cc_key,
            NULL,                       /* p_cc_log_attribute_data */
            p_cc_sw_congestion_setting, /* p_cc_mgt_attribute_data */
            (const pack_data_func_t)CC_SwitchCongestionSetting_pack,
            (const unpack_data_func_t)CC_SwitchCongestionSetting_unpack,
            (const dump_data_func_t)CC_SwitchCongestionSetting_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCSwitchCongestionSettingSet(u_int16_t lid,
        u_int8_t sl,
        u_int64_t cc_key,
        struct CC_SwitchCongestionSetting *p_cc_sw_congestion_setting,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCSwitchCongestionSetting Set MAD lid = %u\n", lid);
    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_CONGESTION_CONT_SW_CONG_SETTING,
            0,                          /* attribute modifier */
            cc_key,
            NULL,                       /* p_cc_log_attribute_data */
            p_cc_sw_congestion_setting, /* p_cc_mgt_attribute_data */
            (const pack_data_func_t)CC_SwitchCongestionSetting_pack,
            (const unpack_data_func_t)CC_SwitchCongestionSetting_unpack,
            (const dump_data_func_t)CC_SwitchCongestionSetting_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}
	

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCSwitchPortCongestionSettingGet(u_int16_t lid,
        u_int8_t sl,
        u_int64_t cc_key,
        u_int8_t block_idx,
        struct CC_SwitchPortCongestionSetting *p_cc_sw_port_congestion_setting,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_sw_port_congestion_setting);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCSwitchPortCongestionSetting Get MAD lid = %u\n", lid);
    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CONGESTION_CONT_SW_PORT_CONG_SETTING,
            block_idx,                       /* attribute modifier */
            cc_key,
            NULL,                            /* p_cc_log_attribute_data */
            p_cc_sw_port_congestion_setting, /* p_cc_mgt_attribute_data */
            (const pack_data_func_t)CC_SwitchPortCongestionSetting_pack,
            (const unpack_data_func_t)CC_SwitchPortCongestionSetting_unpack,
            (const dump_data_func_t)CC_SwitchPortCongestionSetting_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCSwitchPortCongestionSettingSet(u_int16_t lid,
        u_int8_t sl,
        u_int64_t cc_key,
        u_int8_t block_idx,
        struct CC_SwitchPortCongestionSetting *p_cc_sw_port_congestion_setting,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCSwitchPortCongestionSetting Set MAD lid = %u\n", lid);
    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_CONGESTION_CONT_SW_PORT_CONG_SETTING,
            block_idx,                       /* attribute modifier */
            cc_key,
            NULL,                            /* p_cc_log_attribute_data */
            p_cc_sw_port_congestion_setting, /* p_cc_mgt_attribute_data */
            (const pack_data_func_t)CC_SwitchPortCongestionSetting_pack,
            (const unpack_data_func_t)CC_SwitchPortCongestionSetting_unpack,
            (const dump_data_func_t)CC_SwitchPortCongestionSetting_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCCACongestionSettingGet(u_int16_t lid,
        u_int8_t sl,
        u_int64_t cc_key,
        struct CC_CACongestionSetting *p_cc_ca_congestion_setting,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_ca_congestion_setting);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCCACongestionSetting Get MAD lid = %u\n", lid);
    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CONGESTION_CONT_CA_CONG_SETTING,
            0,                          /* attribute modifier */
            cc_key,
            NULL,                       /* p_cc_log_attribute_data */
            p_cc_ca_congestion_setting, /* p_cc_mgt_attribute_data */
            (const pack_data_func_t)CC_CACongestionSetting_pack,
            (const unpack_data_func_t)CC_CACongestionSetting_unpack,
            (const dump_data_func_t)CC_CACongestionSetting_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCCACongestionSettingSet(u_int16_t lid,
        u_int8_t sl,
        u_int64_t cc_key,
        struct CC_CACongestionSetting *p_cc_ca_congestion_setting,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCCACongestionSetting Set MAD lid = %u\n", lid);
    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_CONGESTION_CONT_CA_CONG_SETTING,
            0,                          /* attribute modifier */
            cc_key,
            NULL,                       /* p_cc_log_attribute_data */
            p_cc_ca_congestion_setting, /* p_cc_mgt_attribute_data */
            (const pack_data_func_t)CC_CACongestionSetting_pack,
            (const unpack_data_func_t)CC_CACongestionSetting_unpack,
            (const dump_data_func_t)CC_CACongestionSetting_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCCongestionControlTableGet(u_int16_t lid,
        u_int8_t sl,
        u_int64_t cc_key,
        u_int8_t block_idx,
        struct CC_CongestionControlTable *p_cc_congestion_control_table,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_congestion_control_table);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCCongestionControlTable Get MAD lid = %u\n", lid);
    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CONGESTION_CONT_CC_TBL,
            block_idx,                     /* attribute modifier */
            cc_key,
            NULL,                          /* p_cc_log_attribute_data */
            p_cc_congestion_control_table, /* p_cc_mgt_attribute_data */
            (const pack_data_func_t)CC_CongestionControlTable_pack,
            (const unpack_data_func_t)CC_CongestionControlTable_unpack,
            (const dump_data_func_t)CC_CongestionControlTable_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCCongestionControlTableSet(u_int16_t lid,
        u_int8_t sl,
        u_int64_t cc_key,
        u_int8_t block_idx,
        struct CC_CongestionControlTable *p_cc_congestion_control_table,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCCongestionControlTable Set MAD lid = %u\n", lid);
    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_CONGESTION_CONT_CC_TBL,
            block_idx,                     /* attribute modifier */
            cc_key,
            NULL,                          /* p_cc_log_attribute_data */
            p_cc_congestion_control_table, /* p_cc_mgt_attribute_data */
            (const pack_data_func_t)CC_CongestionControlTable_pack,
            (const unpack_data_func_t)CC_CongestionControlTable_unpack,
            (const dump_data_func_t)CC_CongestionControlTable_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCTimeStampGet(u_int16_t lid,
        u_int8_t sl,
        u_int64_t cc_key,
        struct CC_TimeStamp *p_cc_time_stamp,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_time_stamp);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCTimeStamp Get MAD lid = %u\n", lid);
    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CONGESTION_CONT_TIME_STAMP,
            0,              /* attribute modifier */
            cc_key,
            NULL,            /* p_cc_log_attribute_data */
            p_cc_time_stamp, /* p_cc_mgt_attribute_data */
            (const pack_data_func_t)CC_TimeStamp_pack,
            (const unpack_data_func_t)CC_TimeStamp_unpack,
            (const dump_data_func_t)CC_TimeStamp_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

