/*
 * Copyright (c) 2004-2015 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <sys/time.h>

#include "ibis.h"
#include "packets/packets_layouts.h"

/****************************************************/
/******************private methods*******************/
/****************************************************/

/*
 * Checks if method is of a request MAD
 * Returns true if the method is of a request, otherwise false
 */
bool Ibis::IsRequestMad(uint8_t method)
{
    return (!(method & IBIS_IB_MAD_RESPONSE) &&
            method != IBIS_IB_MAD_METHOD_SEND &&
            method != IBIS_IB_MAD_METHOD_TRAP_REPRESS);
}

/*
 * Returns: mad status[bits: 0-15] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::MadGetSet(ib_address_t *p_ib_address,
        u_int8_t mgmt_class,
        u_int8_t method,
        u_int8_t data_offset,
        void *p_class_data,
        void *p_attribute_data,
        const pack_data_func_t pack_class_data_func,
        const unpack_data_func_t unpack_class_data_func,
        const dump_data_func_t dump_class_data_func,
        const pack_data_func_t pack_attribute_data_func,
        const unpack_data_func_t unpack_attribute_data_func,
        const dump_data_func_t dump_attribute_data_func,
        const clbck_data_t *p_clbck_data)
{
    u_int16_t lid;
    u_int8_t sl;
    u_int32_t d_qp;
    u_int32_t qkey;

    IBIS_ENTER;
    int rc = IBIS_MAD_STATUS_SUCCESS;

    bool is_smp = ((mgmt_class == IBIS_IB_CLASS_SMI_DIRECT) ||
                   (mgmt_class == IBIS_IB_CLASS_SMI));

    transaction_data_t * p_transaction_data = NULL;
    mads_on_node_map_t::iterator mads_iter;

    if (!p_ib_address) {
        rc = IBIS_MAD_STATUS_GENERAL_ERR;
        this->SetLastError("Invalid address was specified for the function");
        goto exit;
    }

    lid = p_ib_address->m_lid;
    sl = p_ib_address->m_sl;
    d_qp = p_ib_address->m_qp;
    qkey = p_ib_address->m_qkey;

    //pack_htons the mad - here data isn't valid (we will have to pack_htons it later)
    pack_class_data_func(p_class_data, this->p_pkt_send);

    if (p_clbck_data)
        IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending Synchronouse MAD\n", lid);
    else
        IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending Asynchronouse MAD\n", lid);

    this->GetLogMadFunction()(dump_attribute_data_func, p_attribute_data,
                              IBIS_LOG_MSG_SEND_MAD);

    //set attribute data
    pack_attribute_data_func(p_attribute_data, this->p_pkt_send + data_offset);

    if((sl == 0) &&
       (lid != 0xffff) && !is_smp ) {
        sl = getPSLForLid(lid);
        if (sl == IB_SLT_UNASSIGNED) {
            IBIS_LOG(TT_LOG_LEVEL_ERROR, "-E- faild fo find sl for dlid %u\n", lid);
            rc = IBIS_MAD_STATUS_GENERAL_ERR;
            goto exit;
        }
        IBIS_LOG(TT_LOG_LEVEL_MAD, "Set sl:%u for dlid %u\n", sl, lid);
    }

    //set destination address
    if (this->SetSendMadAddr(lid, d_qp, sl, qkey)) {
        rc = IBIS_MAD_STATUS_GENERAL_ERR;
        goto exit;
    }

    // non-request MADs
    if (!IsRequestMad(method)) {
        rc = SendMad(mgmt_class, 0, 0);
        goto exit;
    }

    //Synchronouse invocation
    if (p_clbck_data == NULL) {

        if (!transactions_map.empty()) {
            IBIS_LOG(TT_LOG_LEVEL_ERROR,
                     "-E- Synchronous MadGetSet call while "
                     "%u transactions pending. MadRecAll called.\n",
                     (int)transactions_map.size());
            MadRecAll();
        }

        //do rpc
        rc = this->DoRPC(mgmt_class);

        //first unpack_ntohs the data
        unpack_attribute_data_func(p_attribute_data, this->p_pkt_recv + data_offset);
        this->GetLogMadFunction()(dump_attribute_data_func, p_attribute_data,
                                  IBIS_LOG_MSG_RECEIVED_MAD);

        //second unpack_ntohs all the mad - the data section here is invalid (was ntohs already)
        unpack_class_data_func(p_class_data, this->p_pkt_recv);
        //this->GetLogMadFunction()(dump_class_data_func, p_class_data,
        //                          IBIS_LOG_MSG_RECEIVED_MAD);
        IBIS_RETURN(rc);
    }

    //Asynchronouse invocation

    //will send at most one mad on the wire per node
    //Set Node Addr
    node_addr_t node_addr;
    node_addr.m_lid = lid;
    node_addr.m_direct_route.length = 0;
    if (lid == 0xffff && is_smp) {
        MAD_SMP_Direct_Routed * p_smp_direct_routed_mad = (MAD_SMP_Direct_Routed *)p_class_data;
        memcpy(node_addr.m_direct_route.path.BYTE,
               p_smp_direct_routed_mad->InitPath.BYTE,
               sizeof(p_smp_direct_routed_mad->InitPath.BYTE));
        node_addr.m_direct_route.length =
            p_smp_direct_routed_mad->MAD_Header_SMP_Direct_Routed.HopCounter+1;
    }

    //search addr in m_mads_on_node_map
    mads_iter =  m_mads_on_node_map.find(node_addr);
    if (mads_iter == m_mads_on_node_map.end()) {
        //each address is insert only once per run
        list<pending_mad_data_t *> mads_list;
        m_mads_on_node_map[node_addr] = mads_list;

        mads_iter =  m_mads_on_node_map.find(node_addr);
        if (mads_iter == m_mads_on_node_map.end()) {
            this->SetLastError("Failed to insert entry to m_mads_on_node_map.");
            rc = IBIS_MAD_STATUS_GENERAL_ERR;
            goto exit;
        }
    }

    p_transaction_data = m_transaction_data_pool.alloc();

    if (!p_transaction_data) {
        this->SetLastError("Failed to create transaction_data.");
        rc = IBIS_MAD_STATUS_GENERAL_ERR;
        goto exit;
    }

    p_transaction_data->Set(data_offset,
                            unpack_class_data_func,
                            dump_class_data_func,
                            unpack_attribute_data_func,
                            dump_attribute_data_func,
                            is_smp,
                            *p_clbck_data,
                            &(mads_iter->second));

    if (!mads_iter->second.empty() || this->suppressMadSending) {
        // either there is a pending MAD for this node, or MAD sending is
        // suppressed.

        // create pending MAD data object
        pending_mad_data_t *p_mad_data = m_pending_mads_pool.alloc();
        if (!p_mad_data) {
            this->SetLastError("Failed to create pending_mad_data_t.");
            delete p_transaction_data;
            rc = IBIS_MAD_STATUS_GENERAL_ERR;
            goto exit;
        }
        memcpy(p_mad_data->m_umad, this->p_umad_buffer_send,
               p_mad_data->m_umad_size);
        p_mad_data->m_mgmt_class = mgmt_class;
        p_mad_data->m_transaction_data = p_transaction_data;

        if (this->suppressMadSending) {
            // postpone for future send
            mads_iter->second.push_back(NULL);
            mads_iter->second.push_back(p_mad_data);
            m_pending_nodes_transactions.push_back(p_transaction_data);
            IBIS_LOG(TT_LOG_LEVEL_DEBUG,
                     "MAD sending postponed for future send TID:%u data_ptr:%p \n",
                     *(u_int32_t *)(this->p_pkt_send + 12) /*TID*/,
                     &mads_iter->second);
        } else {
            // save mad and send after pending mad returns.
            mads_iter->second.push_back(p_mad_data);
            IBIS_LOG(TT_LOG_LEVEL_DEBUG,
                     "There is pending mad for this note TID:%u data_ptr:%p \n",
                     *(u_int32_t *)(this->p_pkt_send + 12) /*TID*/,
                     &mads_iter->second);
        }
    } else
        // MAD can be sent immediately
        AsyncSendAndRec(mgmt_class, p_transaction_data, NULL);

exit:
    if ((rc != IBIS_MAD_STATUS_SUCCESS) &&
        (p_clbck_data != NULL)){
        InvokeCallbackFunction(*p_clbck_data, rc, NULL);
    }
    IBIS_RETURN(rc);
}

/*
 * Returns: mad status[bits: 0-15] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::MadGetSet(u_int16_t lid,
                    u_int32_t d_qp,
                    u_int8_t sl,
                    u_int32_t qkey,
                    u_int8_t mgmt_class,
                    u_int8_t method,
                    u_int16_t attribute_id,
                    u_int32_t attribute_modifier,
                    u_int8_t data_offset,
                    void *p_class_data,
                    void *p_attribute_data,
                    const pack_data_func_t pack_class_data_func,
                    const unpack_data_func_t unpack_class_data_func,
                    const dump_data_func_t dump_class_data_func,
                    const pack_data_func_t pack_attribute_data_func,
                    const unpack_data_func_t unpack_attribute_data_func,
                    const dump_data_func_t dump_attribute_data_func,
                    const clbck_data_t *p_clbck_data)
{
    ib_address_t ib_address(lid, d_qp, qkey, sl);
    return MadGetSet(&ib_address,
                     mgmt_class,
                     method,
                     data_offset,
                     p_class_data,
                     p_attribute_data,
                     pack_class_data_func,
                     unpack_class_data_func,
                     dump_class_data_func,
                     pack_attribute_data_func,
                     unpack_attribute_data_func,
                     dump_attribute_data_func,
                     p_clbck_data);
}

//if pending_mad_data != NULL will receive mad regardless of mad_on_wire count
int Ibis::AsyncSendAndRec(u_int8_t mgmt_class,
                          transaction_data_t *p_transaction_data,
                          pending_mad_data_t *pending_mad_data)
{
    int rc = IBIS_MAD_STATUS_SUCCESS;
    while (true) {

        if (pending_mad_data) {
            //aditional mad is pending for the node
            //restore umad to send
            memcpy(this->p_umad_buffer_send, pending_mad_data->m_umad, pending_mad_data->m_umad_size);
            p_transaction_data = pending_mad_data->m_transaction_data;
            mgmt_class = pending_mad_data->m_mgmt_class;
        }

        u_int32_t trid = *(u_int32_t *)(this->p_pkt_send + 12); //we will look at lower 32 bits
                                    //of the tid field because umad lib take care of higher 32 bits,
                                    //a pid of the process will be assigned there.

        //insert to transactions_map
        transactions_map_t::iterator tr_iter =  transactions_map.find(trid);
        if (tr_iter != transactions_map.end()) {
            this->SetLastError("Unexpected transaction: %u. Currently used.", trid);
            IBIS_RETURN(IBIS_MAD_STATUS_GENERAL_ERR);
        }

        rc = this->DoAsyncSend(mgmt_class);
        if (rc) {
            InvokeCallbackFunction(p_transaction_data->m_clbck_data, rc, NULL);
            GetNextPendingData(p_transaction_data, pending_mad_data);
            delete p_transaction_data;

            if (pending_mad_data)
                continue;
            else
                IBIS_RETURN(rc);
        }

        transactions_map[trid] = p_transaction_data;
        //indicate mad on wire for node
        //If pending_mads is empty this is the first mad for this node
        //Else the mad is already at the head of the pending list and
        //     will be removed on AsyncRec
        if (p_transaction_data->m_pending_mads->empty())
            p_transaction_data->m_pending_mads->push_back(NULL);

        IBIS_LOG(TT_LOG_LEVEL_DEBUG,
                 "Send MAD with data_ptr:%p \n",
                 p_transaction_data->m_pending_mads);

        //enable receive of more than one MAD in order to andle previus
        //lost receive due to AsyncRec failure
        int recv_count = 0;

        if (p_transaction_data->m_is_smp) {
            m_pending_smps++;
            recv_count = ((m_pending_smps > m_max_smps_on_wire) ?
                          (m_pending_smps - m_max_smps_on_wire) : 0);
        } else {
            m_pending_gmps++;
            recv_count = ((m_pending_gmps > m_max_gmps_on_wire) ?
                          (m_pending_gmps - m_max_gmps_on_wire) : 0);
        }

        bool retry = false;
        while (recv_count > 0 || retry || pending_mad_data) {
            rc = AsyncRec(retry, pending_mad_data);
            if (rc == IBIS_MAD_STATUS_RECV_FAILED || // no use to Receive
                pending_mad_data ) {                 // nead to send before next receive
                break;
            }
            if (!retry)
                recv_count--; //real mad on the wire was received
        }

        while (!pending_mad_data && !m_pending_nodes_transactions.empty()) {
            // check if there are other pending transaction for other nodes
            p_transaction_data = m_pending_nodes_transactions.front();
            m_pending_nodes_transactions.pop_front();
            GetNextPendingData(p_transaction_data, pending_mad_data);
        }

        if (!pending_mad_data)
            break;
    }

    IBIS_RETURN(IBIS_MAD_STATUS_SUCCESS);

}

int Ibis::GetNextPendingData(transaction_data_t * p_transaction_data,
                             pending_mad_data_t *&next_pending_mad_data)
{
    IBIS_ENTER;

    next_pending_mad_data = NULL;

    if(p_transaction_data->m_pending_mads->empty()) {
        this->SetLastError("Unexpected empty pending_mads. list_ptr:%x",
                           p_transaction_data->m_pending_mads);
        IBIS_RETURN(IBIS_MAD_STATUS_GENERAL_ERR);
    } else {
        //remove current mad data from pending list
        pending_mad_data_t *data =
            p_transaction_data->m_pending_mads->front();
        p_transaction_data->m_pending_mads->pop_front();
        m_pending_mads_pool.free(data);

        //Check if aditional mad is pending for the node
        if(!p_transaction_data->m_pending_mads->empty())
            next_pending_mad_data =
                p_transaction_data->m_pending_mads->front();
    }
    IBIS_RETURN(IBIS_MAD_STATUS_SUCCESS);
}

int Ibis::AsyncRec(bool &retry,
                   pending_mad_data_t *&next_pending_mad_data)
{
    IBIS_ENTER;
    retry = false;
    next_pending_mad_data = NULL;

    int rc = DoAsyncRec();

    if (rc != IBIS_MAD_STATUS_RECV_FAILED) {
        //bit fields are represented by 8 bits
        u_int8_t attribute_data[IBIS_IB_MAD_SIZE*8] = {};
        u_int32_t trid_recv = *(u_int32_t *)(this->p_pkt_recv + 12);

        transactions_map_t::iterator tr_iter =  transactions_map.find(trid_recv);
        if (tr_iter == transactions_map.end()) {
            IBIS_LOG(TT_LOG_LEVEL_MAD, "Unknown received transaction: %u.", trid_recv);
            retry = true;
            IBIS_RETURN(IBIS_MAD_STATUS_GENERAL_ERR);
        }

        transaction_data_t *p_transaction_data = tr_iter->second;

        p_transaction_data->m_unpack_attribute_data_func(
            attribute_data, this->p_pkt_recv + p_transaction_data->m_data_offset);

        IBIS_LOG(TT_LOG_LEVEL_DEBUG,
                 "Rec MAD with data_ptr:%p \n",
                 p_transaction_data->m_pending_mads);
        this->GetLogMadFunction()(p_transaction_data->m_dump_attribute_data_func,
                                  attribute_data, IBIS_LOG_MSG_RECEIVED_MAD);
        //second unpack_ntohs all the mad - the data section here is invalid (was ntohs already)
        //unpack_class_data_func(p_class_data, this->p_pkt_recv);
        //this->GetLogMadFunction()(dump_class_data_func, p_class_data,
        //                          IBIS_LOG_MSG_RECEIVED_MAD);

        if (p_transaction_data->m_is_smp)
            m_pending_smps--;
        else
            m_pending_gmps--;

        InvokeCallbackFunction(p_transaction_data->m_clbck_data, rc, attribute_data);

        GetNextPendingData(p_transaction_data, next_pending_mad_data);
        /*
        if(p_transaction_data->m_pending_mads->empty())
            this->SetLastError("Unexpected empty pending_mads. list_ptr:%x",
                               p_transaction_data->m_pending_mads);
        else {
            //remove current mad data from pending list
            pending_mad_data_t *data =
                p_transaction_data->m_pending_mads->front();
            p_transaction_data->m_pending_mads->pop_front();
            m_pending_mads_pool.free(data);

            //Check if aditional mad is pending for the node
            if(!p_transaction_data->m_pending_mads->empty())
                next_pending_mad_data =
                    p_transaction_data->m_pending_mads->front();
        }
        */

        delete p_transaction_data;
        transactions_map.erase(tr_iter);

        //if timeout received, purge all pending mads on current node
        if (rc == IBIS_MAD_STATUS_TIMEOUT && next_pending_mad_data) {
            IBIS_LOG(TT_LOG_LEVEL_INFO,
                     "Purge all pending mads on node after receiving timeout.\n");
            MadRecTimeoutAll(next_pending_mad_data->m_transaction_data);
            next_pending_mad_data = NULL;
        }

    } else if (m_pending_smps || m_pending_gmps)
        IBIS_LOG(TT_LOG_LEVEL_ERROR, "-E- MAD receive failed.\n");

    IBIS_RETURN(rc);
}

/*
static int diff_msec(struct timeval *t1, struct timeval *t2)
{
	int dt;

	dt = (t2->tv_sec - t1->tv_sec)*1000 + (t2->tv_usec - t1->tv_usec)/1000;

	return dt;
}
*/

void Ibis::MadRecAll()
{
    IBIS_ENTER;
    int rc = IBIS_MAD_STATUS_SUCCESS;
    bool retry;
    pending_mad_data_t *pending_mad_data = NULL;
    transaction_data_t *p_transaction_data;
    int fail_count = 0;

    while (!transactions_map.empty()) {
        rc = AsyncRec(retry, pending_mad_data);

        // check if there are other pending transaction for other nodes
        if (!pending_mad_data && !m_pending_nodes_transactions.empty()) {
            p_transaction_data = m_pending_nodes_transactions.front();
            m_pending_nodes_transactions.pop_front();
            GetNextPendingData(p_transaction_data, pending_mad_data);
        }

        if (pending_mad_data) {
            AsyncSendAndRec(pending_mad_data->m_mgmt_class,
                            pending_mad_data->m_transaction_data,
                            pending_mad_data);
        } else if (rc == IBIS_MAD_STATUS_RECV_FAILED) {
            if (fail_count > 1) {
                //just to prevent endless loop if we lost a transaction
                this->SetLastError("Failed to receive all mads");
                MadRecTimeoutAll();
                break;
            }
            fail_count++;
        } else
            fail_count = 0;
    }

    if (m_pending_mads_pool.allocated()){
        this->SetLastError("Failed to send %d pending mads",
                           m_pending_mads_pool.allocated());

        TimeoutAllPendingMads();
    }

    m_pending_smps = 0;
    m_pending_gmps = 0;

    IBIS_RETURN_VOID;
}

void Ibis::MadRecTimeoutAll(transaction_data_t *p_transaction_data)
{
    IBIS_ENTER;
    pending_mad_data_t *next_pending_mad_data = NULL;


    while (true) {
        InvokeCallbackFunction(p_transaction_data->m_clbck_data,
                               IBIS_MAD_STATUS_TIMEOUT, NULL);

        GetNextPendingData(p_transaction_data, next_pending_mad_data);

        delete p_transaction_data;

        if (next_pending_mad_data)
            p_transaction_data = next_pending_mad_data->m_transaction_data;
        else
            break;
    }

    IBIS_RETURN_VOID;
}

void Ibis::MadRecTimeoutAll()
{
    IBIS_ENTER;

    for (transactions_map_t::iterator tr_iter = transactions_map.begin();
         tr_iter !=  transactions_map.end(); ++tr_iter)
        MadRecTimeoutAll(tr_iter->second);

    transactions_map.clear();

    for (transactions_queue_t::iterator q_iter = m_pending_nodes_transactions.begin();
         q_iter != m_pending_nodes_transactions.end(); ++q_iter)
        MadRecTimeoutAll(*q_iter);

    m_pending_nodes_transactions.clear();

    IBIS_RETURN_VOID;
}

void Ibis::TimeoutAllPendingMads()
{
    IBIS_ENTER;

    if (!transactions_map.empty())
        MadRecTimeoutAll();

    //clear all pending mads
    for (mads_on_node_map_t::iterator mads_iter = m_mads_on_node_map.begin();
         mads_iter !=  m_mads_on_node_map.end(); ++mads_iter) {


        pending_mads_on_node_t *pending_mads = &(mads_iter->second);
        for (pending_mads_on_node_t::iterator mads_on_node_iter = pending_mads->begin();
             mads_on_node_iter !=  pending_mads->end(); ++mads_on_node_iter) {

            pending_mad_data_t *p_pending_mad_data = *mads_on_node_iter;
            if (p_pending_mad_data == NULL)
                continue;

            transaction_data_t *p_transaction_data = p_pending_mad_data->m_transaction_data;
            InvokeCallbackFunction(p_transaction_data->m_clbck_data,
                                   IBIS_MAD_STATUS_TIMEOUT, NULL);

            delete p_transaction_data;
            m_pending_mads_pool.free(p_pending_mad_data);
        }
        pending_mads->clear();
    }

    IBIS_RETURN_VOID;
}

void Ibis::DumpMadData(dump_data_func_t dump_func, void *mad_obj, bool msg_send_mad)
{
    Ibis::m_log_mad_function(dump_func, mad_obj, msg_send_mad);
}


void Ibis::CommonMadHeaderBuild(struct MAD_Header_Common *mad_header,
                                u_int8_t mgmt_class,
                                u_int8_t method,
                                u_int16_t attribute_id,
                                u_int32_t attribute_modifier,
                                u_int8_t class_version)
{
    IBIS_ENTER;
    mad_header->BaseVersion           = IBIS_IB_BASE_VERSION;
    mad_header->MgmtClass             = mgmt_class;
    mad_header->ClassVersion          =
        class_version ? class_version : this->GetDefaultMgmtClassVersion(mgmt_class);
    mad_header->Method                = method;
    mad_header->Status                = 0;
    mad_header->ClassSpecific         = 0;
    mad_header->TID_Block_Element     = GetNewTID();
    mad_header->AttributeID           = attribute_id;
    mad_header->AttributeModifier     = attribute_modifier;
    IBIS_RETURN_VOID;
}


u_int8_t Ibis::GetDefaultMgmtClassVersion(u_int8_t mgmt_class)
{
    IBIS_ENTER;

    if (class_versions_by_class[mgmt_class].size() == 1)
        IBIS_RETURN(class_versions_by_class[mgmt_class][0]);

    IBIS_LOG(TT_LOG_LEVEL_ERROR,
             "No default mgmt class version for mgmt_class: %u\n",
             mgmt_class);
    throw std::invalid_argument("Invalid Management class number");
}


/****************************************************/
/*******************public methods*******************/
/****************************************************/
