/*
 *  Copyright (c) 2004-2013 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <string.h>

#include "ibis.h"
#include "packets/packets_layouts.h"


/****************************************************/
/******************private methods*******************/
/****************************************************/
/*
 * Returns: mad status[bits: 0-15] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::PMMadGetSet(u_int16_t lid,
        u_int8_t method,
        u_int16_t attribute_id,
        u_int32_t attribute_modifier,
        void *p_pm_attribute_data,
        const pack_data_func_t pm_pack_attribute_data_func,
        const unpack_data_func_t pm_unpack_attribute_data_func,
        const dump_data_func_t pm_dump_attribute_data_func,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    struct MAD_PerformanceManagement pm_mad;
    CLEAR_STRUCT(pm_mad);

    //build pm lid route header
    this->CommonMadHeaderBuild(&pm_mad.MAD_Header_Common,
            IBIS_IB_CLASS_PERFORMANCE,
            method,
            attribute_id,
            attribute_modifier);

    //send the mad
    IBIS_RETURN(this->MadGetSet(lid, 1, 0, IBIS_IB_DEFAULT_QP1_QKEY,
            IBIS_IB_CLASS_PERFORMANCE,
            method, attribute_id, attribute_modifier,
            IBIS_IB_DATA_OFFSET_PERFORMANCE,
            &pm_mad, p_pm_attribute_data,
            (const pack_data_func_t)MAD_PerformanceManagement_pack,
            (const unpack_data_func_t)MAD_PerformanceManagement_unpack,
            (const dump_data_func_t)MAD_PerformanceManagement_dump,
            (const pack_data_func_t)pm_pack_attribute_data_func,
            (const unpack_data_func_t)pm_unpack_attribute_data_func,
            (const dump_data_func_t)pm_dump_attribute_data_func,
            p_clbck_data));
}


/****************************************************/
/*******************public methods*******************/
/****************************************************/
/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::PMClassPortInfoGet(u_int16_t lid,
        struct IB_ClassPortInfo *p_ib_class_port_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_ib_class_port_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending IB_ClassPortInfo Get MAD lid = %u\n", lid);
    int rc = this->PMMadGetSet(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CLASS_PORT_INFO,
            0,
            p_ib_class_port_info,
            (const pack_data_func_t)IB_ClassPortInfo_pack,
            (const unpack_data_func_t)IB_ClassPortInfo_unpack,
            (const dump_data_func_t)IB_ClassPortInfo_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::PMPortCountersGet(u_int16_t lid,
        phys_port_t port_number,
        struct PM_PortCounters *p_port_counters,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_port_counters);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending PMPortCounters Get MAD lid = %u\n", lid);
    p_port_counters->PortSelect = port_number;
    int rc = this->PMMadGetSet(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_COUNTERS,
            0,
            p_port_counters,
            (const pack_data_func_t)PM_PortCounters_pack,
            (const unpack_data_func_t)PM_PortCounters_unpack,
            (const dump_data_func_t)PM_PortCounters_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::PMPortCountersSet(u_int16_t lid,
        struct PM_PortCounters *p_port_counters,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending PMPortCounters Set MAD lid = %u\n", lid);
    int rc = this->PMMadGetSet(lid,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_COUNTERS,
            0,
            p_port_counters,
            (const pack_data_func_t)PM_PortCounters_pack,
            (const unpack_data_func_t)PM_PortCounters_unpack,
            (const dump_data_func_t)PM_PortCounters_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::PMPortCountersClear(u_int16_t lid,
        phys_port_t port_number,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    struct PM_PortCounters p_port_counters;
    CLEAR_STRUCT(p_port_counters);

    //set mask to all counters
    memset(&p_port_counters.CounterSelect, 0xff, sizeof(p_port_counters.CounterSelect));
    p_port_counters.CounterSelect2.SetPortXmitWait = 0xff;  //clear also xmit_wait
                                                            //do it anyway

    //set port we want to  clear
    p_port_counters.PortSelect = port_number;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending PMPortCounters Set MAD (clear counters) lid = %u\n", lid);
    int rc = this->PMMadGetSet(lid,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_COUNTERS,
            0,
            &p_port_counters,
            (const pack_data_func_t)PM_PortCounters_pack,
            (const unpack_data_func_t)PM_PortCounters_unpack,
            (const dump_data_func_t)PM_PortCounters_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::PMPortCountersExtendedGet(u_int16_t lid,
        phys_port_t port_number,
        struct PM_PortCountersExtended *p_port_counters,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_port_counters);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending PMPortCountersExtended Get MAD lid = %u\n", lid);
    p_port_counters->PortSelect = port_number;
    int rc = this->PMMadGetSet(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_COUNTERS_EXT,
            0,
            p_port_counters,
            (const pack_data_func_t)PM_PortCountersExtended_pack,
            (const unpack_data_func_t)PM_PortCountersExtended_unpack,
            (const dump_data_func_t)PM_PortCountersExtended_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::PMPortCountersExtendedSet(u_int16_t lid,
        struct PM_PortCountersExtended *p_port_counters,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending PMPortCountersExtended Set MAD lid = %u\n", lid);
    int rc = this->PMMadGetSet(lid,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_COUNTERS_EXT,
            0,
            p_port_counters,
            (const pack_data_func_t)PM_PortCountersExtended_pack,
            (const unpack_data_func_t)PM_PortCountersExtended_unpack,
            (const dump_data_func_t)PM_PortCountersExtended_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::PMPortCountersExtendedClear(u_int16_t lid,
        phys_port_t port_number,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    struct PM_PortCountersExtended p_port_counters;
    CLEAR_STRUCT(p_port_counters);

    //set mask to all counters
    memset(&p_port_counters.CounterSelect, 0xff, sizeof(p_port_counters.CounterSelect));
    //set port we want to  clear
    p_port_counters.PortSelect = port_number;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending PMPortCountersExtended Set MAD (clear counters) lid = %u\n", lid);
    int rc = this->PMMadGetSet(lid,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_COUNTERS_EXT,
            0,
            &p_port_counters,
            (const pack_data_func_t)PM_PortCountersExtended_pack,
            (const unpack_data_func_t)PM_PortCountersExtended_unpack,
            (const dump_data_func_t)PM_PortCountersExtended_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::PMPortExtendedSpeedsCountersGet(u_int16_t lid,
        phys_port_t port_number,
        struct PM_PortExtendedSpeedsCounters *p_port_counters,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_port_counters);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending PMPortExtendedSpeedsCounters Get MAD lid = %u\n", lid);
    p_port_counters->PortSelect = port_number;
    int rc = this->PMMadGetSet(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_EXT_SPEEDS_COUNTERS,
            0,
            p_port_counters,
            (const pack_data_func_t)PM_PortExtendedSpeedsCounters_pack,
            (const unpack_data_func_t)PM_PortExtendedSpeedsCounters_unpack,
            (const dump_data_func_t)PM_PortExtendedSpeedsCounters_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
/* Note that writing (Set) anything other than zero into a counter
   results in undefined behavior.
int Ibis::PMPortExtendedSpeedsCountersSet(u_int16_t lid,
        struct PM_PortExtendedSpeedsCounters *p_port_counters)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending PMPortExtendedSpeedsCounters Set MAD lid = %u\n", lid);
    int rc = this->PMMadGetSet(lid,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_EXT_SPEEDS_COUNTERS,
            0,
            p_port_counters,
            (const pack_data_func_t)PM_PortExtendedSpeedsCounters_pack,
            (const unpack_data_func_t)PM_PortExtendedSpeedsCounters_unpack,
            (const dump_data_func_t)PM_PortExtendedSpeedsCounters_dump);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}
*/

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::PMPortExtendedSpeedsCountersClear(u_int16_t lid,
        phys_port_t port_number,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    struct PM_PortExtendedSpeedsCounters p_port_counters;
    CLEAR_STRUCT(p_port_counters);

    //set mask to all counters
    memset(&p_port_counters.CounterSelect, 0xff, sizeof(p_port_counters.CounterSelect));
    //set port we want to  clear
    p_port_counters.PortSelect = port_number;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending PMPortExtendedSpeedsCounters Set MAD (clear counters) lid = %u\n", lid);
    int rc = this->PMMadGetSet(lid,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_EXT_SPEEDS_COUNTERS,
            0,
            &p_port_counters,
            (const pack_data_func_t)PM_PortExtendedSpeedsCounters_pack,
            (const unpack_data_func_t)PM_PortExtendedSpeedsCounters_unpack,
            (const dump_data_func_t)PM_PortExtendedSpeedsCounters_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::PMPortExtendedSpeedsRSFECCountersGet(u_int16_t lid,
        phys_port_t port_number,
        struct PM_PortExtendedSpeedsRSFECCounters *p_port_counters,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_port_counters);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending PMPortExtendedSpeedsRSFECCounters Get MAD lid = %u\n", lid);
    p_port_counters->PortSelect = port_number;
    //same attribute as extended speeds counters MAD, different layout
    int rc = this->PMMadGetSet(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_EXT_SPEEDS_COUNTERS,
            0,
            p_port_counters,
            (const pack_data_func_t)PM_PortExtendedSpeedsRSFECCounters_pack,
            (const unpack_data_func_t)PM_PortExtendedSpeedsRSFECCounters_unpack,
            (const dump_data_func_t)PM_PortExtendedSpeedsRSFECCounters_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::PMPortExtendedSpeedsRSFECCountersClear(u_int16_t lid,
        phys_port_t port_number,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    struct PM_PortExtendedSpeedsCounters p_port_counters;
    CLEAR_STRUCT(p_port_counters);

    //set mask to all counters
    memset(&p_port_counters.CounterSelect, 0xff,
           sizeof(p_port_counters.CounterSelect));
    //set port we want to  clear
    p_port_counters.PortSelect = port_number;
    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending PMPortExtendedSpeedsRSFECCounters Set MAD "
             "(clear counters) lid = %u\n", lid);
    int rc = this->PMMadGetSet(lid,
           IBIS_IB_MAD_METHOD_SET,
           IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_EXT_SPEEDS_COUNTERS,
           0,
           &p_port_counters,
           (const pack_data_func_t)PM_PortExtendedSpeedsRSFECCounters_pack,
           (const unpack_data_func_t)PM_PortExtendedSpeedsRSFECCounters_unpack,
           (const dump_data_func_t)PM_PortExtendedSpeedsRSFECCounters_dump,
           p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::PMPortSampleControlGet(u_int16_t lid,
        phys_port_t port_number,
        struct PM_PortSamplesControl *p_sample_control,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_sample_control);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending PMPortSampleControl Get MAD lid = %u\n", lid);
    p_sample_control->PortSelect = port_number;
    int rc = this->PMMadGetSet(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_SAMPLES_CONTROL,
            0,
            p_sample_control,
            (const pack_data_func_t)PM_PortSamplesControl_pack,
            (const unpack_data_func_t)PM_PortSamplesControl_unpack,
            (const dump_data_func_t)PM_PortSamplesControl_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::PMPerSLVLCounters(bool is_reset_cntr,
        u_int16_t lid,
        phys_port_t port_number,
        u_int32_t attr_id,
        struct PM_PortRcvXmitCntrsSlVl *p_pm_port_rcvxmit_data_slvl,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_pm_port_rcvxmit_data_slvl);
    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending %s attribute %u Get MAD lid = %u port = %u\n",
             (is_reset_cntr ? "Clear" : "Get"),
             attr_id, lid, port_number);
    p_pm_port_rcvxmit_data_slvl->port_select = port_number;
    if (is_reset_cntr)
        p_pm_port_rcvxmit_data_slvl->counter_select = 0xffff;
    int rc = this->PMMadGetSet(lid,
                is_reset_cntr ? IBIS_IB_MAD_METHOD_SET : IBIS_IB_MAD_METHOD_GET,
                attr_id,
                0,
                p_pm_port_rcvxmit_data_slvl,
                (const pack_data_func_t)PM_PortRcvXmitCntrsSlVl_pack,
                (const unpack_data_func_t)PM_PortRcvXmitCntrsSlVl_unpack,
                (const dump_data_func_t)PM_PortRcvXmitCntrsSlVl_dump,
                p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

// PortRcvErrorDetails and PortXmitDiscardDetails

int Ibis::PMPortRcvErrorDetailsGet(u_int16_t lid,
        phys_port_t port_number,
        struct PM_PortRcvErrorDetails *p_pm_port_rcv_error_details,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    CLEAR_STRUCT(*p_pm_port_rcv_error_details);

    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending PMPortRcvErrorDetails Get MAD lid = %u port = %u\n", lid, port_number);

    p_pm_port_rcv_error_details->PortSelect = port_number;

    int rc = this->PMMadGetSet(lid,
                               IBIS_IB_MAD_METHOD_GET,
                               IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_RCV_ERROR_DETAILS,
                               0,
                               p_pm_port_rcv_error_details,
                               (const pack_data_func_t)PM_PortRcvErrorDetails_pack,
                               (const unpack_data_func_t)PM_PortRcvErrorDetails_unpack,
                               (const dump_data_func_t)PM_PortRcvErrorDetails_dump,
                               p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   // bits [8-15] are class specific - don't need them here
}

int Ibis::PMPortRcvErrorDetailsClear(u_int16_t lid,
        phys_port_t port_number,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    PM_PortRcvErrorDetails p_port_rcv_error_details;
    CLEAR_STRUCT(p_port_rcv_error_details);

    // set mask to all counters
    memset(&p_port_rcv_error_details.CounterSelect, 0xff,
           sizeof(p_port_rcv_error_details.CounterSelect));

    // set port we want to clear
    p_port_rcv_error_details.PortSelect = port_number;

    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending PMPortRcvErrorDetails Clear MAD lid = %u port = %u\n", lid, port_number);

    int rc = this->PMMadGetSet(lid,
                               IBIS_IB_MAD_METHOD_SET,
                               IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_RCV_ERROR_DETAILS,
                               0,
                               &p_port_rcv_error_details,
                               (const pack_data_func_t)PM_PortRcvErrorDetails_pack,
                               (const unpack_data_func_t)PM_PortRcvErrorDetails_unpack,
                               (const dump_data_func_t)PM_PortRcvErrorDetails_dump,
                               p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   // bits [8-15] are class specific - don't need them here
}

int Ibis::PMPortXmitDiscardDetailsGet(u_int16_t lid,
        phys_port_t port_number,
        struct PM_PortXmitDiscardDetails *p_pm_port_xmit_discard_details,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    CLEAR_STRUCT(*p_pm_port_xmit_discard_details);

    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending PMPortXmitDiscardDetails Get MAD lid = %u port = %u\n", lid, port_number);

    p_pm_port_xmit_discard_details->PortSelect = port_number;

    int rc = this->PMMadGetSet(lid,
                               IBIS_IB_MAD_METHOD_GET,
                               IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_XMIT_DISCARD_DETAILS,
                               0,
                               p_pm_port_xmit_discard_details,
                               (const pack_data_func_t)PM_PortXmitDiscardDetails_pack,
                               (const unpack_data_func_t)PM_PortXmitDiscardDetails_unpack,
                               (const dump_data_func_t)PM_PortXmitDiscardDetails_dump,
                               p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   // bits [8-15] are class specific - don't need them here
}

int Ibis::PMPortXmitDiscardDetailsClear(u_int16_t lid,
        phys_port_t port_number,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    PM_PortXmitDiscardDetails p_port_xmit_discard_details;
    CLEAR_STRUCT(p_port_xmit_discard_details);

    // set mask to all counters
    memset(&p_port_xmit_discard_details.CounterSelect, 0xff,
           sizeof(p_port_xmit_discard_details.CounterSelect));

    // set port we want to clear
    p_port_xmit_discard_details.PortSelect = port_number;

    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending PMPortXmitDiscardDetails Clear MAD lid = %u port = %u\n", lid, port_number);

    int rc = this->PMMadGetSet(lid,
                               IBIS_IB_MAD_METHOD_SET,
                               IBIS_IB_ATTR_PERF_MANAGEMENT_PORT_XMIT_DISCARD_DETAILS,
                               0,
                               &p_port_xmit_discard_details,
                               (const pack_data_func_t)PM_PortXmitDiscardDetails_pack,
                               (const unpack_data_func_t)PM_PortXmitDiscardDetails_unpack,
                               (const dump_data_func_t)PM_PortXmitDiscardDetails_dump,
                               p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   // bits [8-15] are class specific - don't need them here
}

