/*
 * © Copyright (c) 2004-2013 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <string.h>

#include "ibis.h"
#include "packets/packets_layouts.h"

/****************************************************/
/******************private methods*******************/
/****************************************************/
void Ibis::SMPHeaderDirectRoutedBuild(struct MAD_Header_SMP_Direct_Routed *smp_direct_route_header,
                                        u_int8_t method,
                                        u_int16_t attribute_id,
                                        u_int32_t attribute_modifier,
                                        u_int8_t direct_path_len)
{
    IBIS_ENTER;
    smp_direct_route_header->BaseVersion        = IBIS_IB_BASE_VERSION;
    smp_direct_route_header->MgmtClass          = IBIS_IB_CLASS_SMI_DIRECT;
    smp_direct_route_header->ClassVersion       = this->GetDefaultMgmtClassVersion(IBIS_IB_CLASS_SMI_DIRECT);
    smp_direct_route_header->Method             = method;
    smp_direct_route_header->Status             = 0;
    smp_direct_route_header->HopCounter         = direct_path_len - 1;
    smp_direct_route_header->HopPointer         = 0;
    smp_direct_route_header->D                  = 0;
    smp_direct_route_header->TID_Block_Element  = GetNewTID();
    smp_direct_route_header->AttributeID        = attribute_id;
    smp_direct_route_header->AttributeModifier  = attribute_modifier;
    IBIS_RETURN_VOID;
}


/*
 * Returns: mad status[bits: 0-15] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPMadGetSetByLid(u_int16_t lid,
        u_int8_t method,
        u_int16_t attribute_id,
        u_int32_t attribute_modifier,
        void *p_smp_attribute_data,
        const pack_data_func_t smp_pack_attribute_data_func,
        const unpack_data_func_t smp_unpack_attribute_data_func,
        const dump_data_func_t smp_dump_attribute_data_func,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    struct MAD_SMP_LID_Routed smp_lid_routed_mad;
    CLEAR_STRUCT(smp_lid_routed_mad);

    //build smp lid route header
    this->CommonMadHeaderBuild(&smp_lid_routed_mad.MAD_Header_Common,
            IBIS_IB_CLASS_SMI,
            method,
            attribute_id,
            attribute_modifier);

    smp_lid_routed_mad.M_Key = 0;
    MKeyManager * mkey_mngr = this->GetMKeyManager();
    if (mkey_mngr)
        smp_lid_routed_mad.M_Key = mkey_mngr->getMkeyByLid(lid);

    //send the mad
    IBIS_RETURN(this->MadGetSet(lid, 0, 0, 0,
            IBIS_IB_CLASS_SMI,
            method, attribute_id, attribute_modifier,
            IBIS_IB_DATA_OFFSET_SMP,
            &smp_lid_routed_mad, p_smp_attribute_data,
            (const pack_data_func_t)MAD_SMP_LID_Routed_pack,
            (const unpack_data_func_t)MAD_SMP_LID_Routed_unpack,
            (const dump_data_func_t)MAD_SMP_LID_Routed_dump,
            (const pack_data_func_t)smp_pack_attribute_data_func,
            (const unpack_data_func_t)smp_unpack_attribute_data_func,
            (const dump_data_func_t)smp_dump_attribute_data_func,
            p_clbck_data));
}


/*
 * Returns: mad status[bits: 0-15] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPMadGetSetByDirect(direct_route_t *p_direct_route,
        u_int8_t method,
        u_int16_t attribute_id,
        u_int32_t attribute_modifier,
        void *p_smp_attribute_data,
        const pack_data_func_t smp_pack_attribute_data_func,
        const unpack_data_func_t smp_unpack_attribute_data_func,
        const dump_data_func_t smp_dump_attribute_data_func,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    struct MAD_SMP_Direct_Routed smp_direct_routed_mad;
    CLEAR_STRUCT(smp_direct_routed_mad);

    //build smp direct route header
    this->SMPHeaderDirectRoutedBuild(&smp_direct_routed_mad.MAD_Header_SMP_Direct_Routed,
            method,
            attribute_id,
            attribute_modifier,
            p_direct_route->length);

    //build rest of fields
    smp_direct_routed_mad.M_Key = 0;
    MKeyManager * mkey_mngr = this->GetMKeyManager();
    if (mkey_mngr)
        smp_direct_routed_mad.M_Key =
            mkey_mngr->getMkeyFromDRPath(p_direct_route);

    smp_direct_routed_mad.DrSLID	= 0xffff;
    smp_direct_routed_mad.DrDLID	= 0xffff;
    memcpy(&smp_direct_routed_mad.InitPath, &p_direct_route->path, sizeof(p_direct_route->path));

    //send the mad
    IBIS_RETURN(this->MadGetSet(0xffff, 0, 0, 0,
            IBIS_IB_CLASS_SMI_DIRECT,
            method, attribute_id, attribute_modifier,
            IBIS_IB_DATA_OFFSET_SMP,
            &smp_direct_routed_mad, p_smp_attribute_data,
            (const pack_data_func_t)MAD_SMP_Direct_Routed_pack,
            (const unpack_data_func_t)MAD_SMP_Direct_Routed_unpack,
            (const dump_data_func_t)MAD_SMP_Direct_Routed_dump,
            (const pack_data_func_t)smp_pack_attribute_data_func,
            (const unpack_data_func_t)smp_unpack_attribute_data_func,
            (const dump_data_func_t)smp_dump_attribute_data_func,
            p_clbck_data));
}


/****************************************************/
/*******************public methods*******************/
/****************************************************/
/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPPortInfoMadGetByLid(u_int16_t lid,
        phys_port_t port_number,
        struct SMP_PortInfo *p_port_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_port_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPPortInfo MAD by lid = %u\n", lid);
    int rc = this->SMPMadGetSetByLid(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_PORT_INFO,
            port_number,
            p_port_info,
            (const pack_data_func_t)SMP_PortInfo_pack,
            (const unpack_data_func_t)SMP_PortInfo_unpack,
            (const dump_data_func_t)SMP_PortInfo_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPPortInfoExtMadGetByLid(
    u_int16_t lid,
    phys_port_t port_number,
    struct SMP_PortInfoExtended *p_port_info_ext,
    const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_port_info_ext);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPPortInfoExtended MAD by lid = %u\n", lid);
    int rc = this->SMPMadGetSetByLid(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_PORT_INFO_EXTENDED,
            port_number,
            p_port_info_ext,
            (const pack_data_func_t)SMP_PortInfoExtended_pack,
            (const unpack_data_func_t)SMP_PortInfoExtended_unpack,
            (const dump_data_func_t)SMP_PortInfoExtended_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPMlnxExtPortInfoMadGetByLid(u_int16_t lid,
        phys_port_t port_number,
        struct SMP_MlnxExtPortInfo *p_mlnx_ext_port_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_mlnx_ext_port_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPMlnxExtPortInfo MAD by lid = %u\n", lid);
    int rc = this->SMPMadGetSetByLid(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_MLNXEXTPORTINFO,
            port_number,
            p_mlnx_ext_port_info,
            (const pack_data_func_t)SMP_MlnxExtPortInfo_pack,
            (const unpack_data_func_t)SMP_MlnxExtPortInfo_unpack,
            (const dump_data_func_t)SMP_MlnxExtPortInfo_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPSwitchInfoMadGetByLid(u_int16_t lid,
        struct SMP_SwitchInfo *p_switch_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_switch_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPSwitchInfo MAD by lid = %u\n", lid);
    int rc = this->SMPMadGetSetByLid(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_SWITCH_INFO,
            0,
            p_switch_info,
            (const pack_data_func_t)SMP_SwitchInfo_pack,
            (const unpack_data_func_t)SMP_SwitchInfo_unpack,
            (const dump_data_func_t)SMP_SwitchInfo_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPNodeInfoMadGetByLid(u_int16_t lid,
        struct SMP_NodeInfo *p_node_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_node_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPNodeInfo MAD by lid = %u\n", lid);
    int rc = this->SMPMadGetSetByLid(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_NODE_INFO,
            0,
            p_node_info,
            (const pack_data_func_t)SMP_NodeInfo_pack,
            (const unpack_data_func_t)SMP_NodeInfo_unpack,
            (const dump_data_func_t)SMP_NodeInfo_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPNodeDescMadGetByLid(u_int16_t lid,
        struct SMP_NodeDesc *p_node_desc,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_node_desc);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPNodeDesc MAD by lid = %u\n", lid);
    int rc = this->SMPMadGetSetByLid(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_NODE_DESC,
            0,
            p_node_desc,
            (const pack_data_func_t)SMP_NodeDesc_pack,
            (const unpack_data_func_t)SMP_NodeDesc_unpack,
            (const dump_data_func_t)SMP_NodeDesc_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPSMInfoMadGetByLid(u_int16_t lid,
        struct SMP_SMInfo *p_sm_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_sm_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPSMInfo MAD by lid = %u\n", lid);
    int rc = this->SMPMadGetSetByLid(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_SMINFO,
            0,
            p_sm_info,
            (const pack_data_func_t)SMP_SMInfo_pack,
            (const unpack_data_func_t)SMP_SMInfo_unpack,
            (const dump_data_func_t)SMP_SMInfo_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPLinearForwardingTableGetByLid(u_int16_t lid,
        u_int32_t lid_to_port_block_num,
        struct SMP_LinearForwardingTable *p_linear_forwarding_table,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_linear_forwarding_table);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPLinearForwardingTable MAD by lid = %u\n", lid);
    int rc = this->SMPMadGetSetByLid(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_LINEARFORWTBL,
            lid_to_port_block_num,
            p_linear_forwarding_table,
            (const pack_data_func_t)SMP_LinearForwardingTable_pack,
            (const unpack_data_func_t)SMP_LinearForwardingTable_unpack,
            (const dump_data_func_t)SMP_LinearForwardingTable_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPMulticastForwardingTableGetByLid(u_int16_t lid,
        u_int8_t port_group,
        u_int32_t lid_to_port_block_num,
        struct SMP_MulticastForwardingTable *p_multicast_forwarding_table,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_multicast_forwarding_table);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPMulticastForwardingTable MAD by lid = %u\n", lid);
    u_int32_t attribute_modifier = lid_to_port_block_num + (port_group << 28);
    int rc = this->SMPMadGetSetByLid(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_MULTICASTFORWTBL,
            attribute_modifier,
            p_multicast_forwarding_table,
            (const pack_data_func_t)SMP_MulticastForwardingTable_pack,
            (const unpack_data_func_t)SMP_MulticastForwardingTable_unpack,
            (const dump_data_func_t)SMP_MulticastForwardingTable_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPPkeyTableGetByLid(u_int16_t lid,
        u_int16_t port_num,
        u_int16_t block_num,
        struct SMP_PKeyTable *p_pkey_table,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_pkey_table);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPPKeyTable MAD by lid = %u\n", lid);
    u_int32_t attribute_modifier = (port_num << 16) | block_num;
    int rc = this->SMPMadGetSetByLid(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_PKEY_TBL,
            attribute_modifier,
            p_pkey_table,
            (const pack_data_func_t)SMP_PKeyTable_pack,
            (const unpack_data_func_t)SMP_PKeyTable_unpack,
            (const dump_data_func_t)SMP_PKeyTable_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPGUIDInfoTableGetByLid(u_int16_t lid,
        u_int32_t block_num,
        struct SMP_GUIDInfo *p_guid_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_guid_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPGUIDInfo MAD by lid = %u\n", lid);
    int rc = this->SMPMadGetSetByLid(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_GUID_INFO,
            block_num,
            p_guid_info,
            (const pack_data_func_t)SMP_GUIDInfo_pack,
            (const unpack_data_func_t)SMP_GUIDInfo_unpack,
            (const dump_data_func_t)SMP_GUIDInfo_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPSLToVLMappingTableGetByLid(u_int16_t lid,
        phys_port_t out_port_number,
        phys_port_t in_port_number,
        struct SMP_SLToVLMappingTable *p_slvl_mapping,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_slvl_mapping);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPSLToVLMappingTable MAD by lid = %u"
             " out_port:%u in_port:%u \n",
             lid, out_port_number, in_port_number);

    u_int32_t attribute_modifier = out_port_number + (in_port_number << 8);

    int rc = this->SMPMadGetSetByLid(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_SLVL_TABLE,
            attribute_modifier,
            p_slvl_mapping,
            (const pack_data_func_t)SMP_SLToVLMappingTable_pack,
            (const unpack_data_func_t)SMP_SLToVLMappingTable_unpack,
            (const dump_data_func_t)SMP_SLToVLMappingTable_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPPLFTInfoMadGetSetByLid(u_int16_t lid,
        u_int8_t method,
        ib_private_lft_info *p_plft_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPPLFTInfoMadGetSetByDirect MAD by lid = %u"  \
            "method = %u\n",
            lid, method);
    /*
     * attr_mod must be 0
     */
    u_int32_t attr_mod = 0;

    int rc = this->SMPMadGetSetByLid(lid,
            method,
            IBIS_IB_ATTR_SMP_PLFT_INFO,
            attr_mod,
            p_plft_info,
            (const pack_data_func_t)ib_private_lft_info_pack,
            (const unpack_data_func_t)ib_private_lft_info_unpack,
            (const dump_data_func_t)ib_private_lft_info_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPPLFTDefMadGetSetByLid(u_int16_t lid,
        u_int8_t method,
        u_int8_t plft_block,
        struct ib_private_lft_def *p_plft_def,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending PLFTDef MAD by lid = %u"  \
            "method = %u, plft block = %u\n",
            lid, method, plft_block);
    /*
     * attr_mod[3-0] - block number of of 16 LFTs
     */
    u_int32_t attr_mod = (0x0000000f & plft_block);

    int rc = this->SMPMadGetSetByLid(lid,
            method,
            IBIS_IB_ATTR_SMP_PLFT_DEF,
            attr_mod,
            p_plft_def,
            (const pack_data_func_t)ib_private_lft_def_pack,
            (const unpack_data_func_t)ib_private_lft_def_unpack,
            (const dump_data_func_t)ib_private_lft_def_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPPLFTMapMadGetSetByLid(u_int16_t lid,
        u_int8_t method,
        u_int8_t pLFTID,
        struct ib_private_lft_map *p_plft_map,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending PLFTDef MAD by lid = %u"  \
            "method = %u, pLFTIDk = %u\n",
            lid, method, pLFTID);
    /*
     * attr_mod[7-0] - privateLFTID
     */
    u_int32_t attr_mod = (0x000000ff & pLFTID);

    int rc = this->SMPMadGetSetByLid(lid,
            method,
            IBIS_IB_ATTR_SMP_PLFT_MAP,
            attr_mod,
            p_plft_map,
            (const pack_data_func_t)ib_private_lft_map_pack,
            (const unpack_data_func_t)ib_private_lft_map_unpack,
            (const dump_data_func_t)ib_private_lft_map_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPPortSLToPrivateLFTMapGetSetByLid(u_int16_t lid,
        u_int8_t method,
        u_int8_t port_block,
        struct ib_port_sl_to_private_lft_map *p_plft_map,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending ortSLToPrivateLFTMap MAD by lid = %u"  \
            "method = %u, port block = %u\n",
            lid, method, port_block);
    /*
     * attr_mod[5-0] - port_block
     */
    u_int32_t attr_mod = (0x000000ff & port_block);

    int rc = this->SMPMadGetSetByLid(lid,
            method,
            IBIS_IB_ATTR_SMP_PORT_SL_TO_PLFT_MAP,
            attr_mod,
            p_plft_map,
            (const pack_data_func_t)ib_port_sl_to_private_lft_map_pack,
            (const unpack_data_func_t)ib_port_sl_to_private_lft_map_unpack,
            (const dump_data_func_t)ib_port_sl_to_private_lft_map_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

/****************************************************/
/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPARInfoGetSetByLid(u_int16_t lid,
        u_int8_t method,
        bool get_cap,
        struct adaptive_routing_info *p_ar_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    u_int32_t attr_mod = 0;
    if ((method == IBIS_IB_MAD_METHOD_GET) && get_cap) {
        attr_mod = (0x1 << 31);
        IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPARInfo GetCap MAD by lid = %u\n",
                lid);
    } else {
        IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPARInfo MAD by direct = %u, lid = %u\n",
            lid, method);
    }

    int rc = this->SMPMadGetSetByLid(lid,
            method,
            IBIS_IB_ATTR_SMP_AR_INFO,
            attr_mod,
            p_ar_info,
            (const pack_data_func_t)adaptive_routing_info_pack,
            (const unpack_data_func_t)adaptive_routing_info_unpack,
            (const dump_data_func_t)adaptive_routing_info_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

/****************************************************/
/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPARGroupTableGetSetByLid(u_int16_t lid,
        u_int8_t  method,
        u_int16_t group_block,
        u_int8_t  group_table,
        u_int8_t  pLFTID,
        struct ib_ar_group_table *p_ar_group_table,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPARGroupTable MAD by lid = %u, method = %u, "   \
            "group block = %u\n",
            lid, method, group_block);
    /*
     * attr_mod[6:0] - group block (block contains 2 groups)
     * attr_mod[19:16] - group table ==> in AR value is 0
     * attr_mod[31:24] - privateLFTID
     */
    u_int32_t attr_mod = 0x0000007f & group_block;
    attr_mod |= (group_table << 16);
    attr_mod |= (pLFTID << 24);

    int rc = this->SMPMadGetSetByLid(lid,
            method,
            IBIS_IB_ATTR_SMP_AR_GROUP_TABLE,
            attr_mod,
            p_ar_group_table,
            (const pack_data_func_t)ib_ar_group_table_pack,
            (const unpack_data_func_t)ib_ar_group_table_unpack,
            (const dump_data_func_t)ib_ar_group_table_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

/****************************************************/
/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPARLinearForwardingTableGetSetByLid(u_int16_t lid,
        u_int8_t method,
        u_int32_t lid_block,
        u_int8_t pLFTID,
        struct ib_ar_linear_forwarding_table_sx *p_ar_linear_forwarding_table,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPARLinearForwardingTable MAD by lid = %u, " \
            "method = %u, lid block = %u\n",
            lid, method, lid_block);
    /*
     * attr_mod[23:0] - lid block (block contains 32 lids)
     * attr_mod[31:24] - privateLFTID
     */
    u_int32_t attr_mod = (0x00ffffff & lid_block);
    attr_mod |= (pLFTID << 24);

    int rc = this->SMPMadGetSetByLid(lid,
            method,
            IBIS_IB_ATTR_SMP_AR_LINEAR_FORWARDING_TABLE_SX,
            attr_mod,
            p_ar_linear_forwarding_table,
            (const pack_data_func_t)ib_ar_linear_forwarding_table_sx_pack,
            (const unpack_data_func_t)ib_ar_linear_forwarding_table_sx_unpack,
            (const dump_data_func_t)ib_ar_linear_forwarding_table_sx_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPExtendedSwitchInfoMadGetSetByLid(u_int16_t lid,
        u_int8_t method,
        struct ib_extended_switch_info *p_ext_sw_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending ExtendedSwitchInfo MAD by lid = %u, " \
            "method = %u\n",
            lid, method);
    /*
     * attr_mod must be 0
     */
    u_int32_t attr_mod = 0;

    int rc = this->SMPMadGetSetByLid(lid,
            method,
            IBIS_IB_ATTR_SMP_EXTENDED_SWITCH_INFO,
            attr_mod,
            p_ext_sw_info,
            (const pack_data_func_t)ib_extended_switch_info_pack,
            (const unpack_data_func_t)ib_extended_switch_info_unpack,
            (const dump_data_func_t)ib_extended_switch_info_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

/****************************************************/

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPRNSubGroupDirectionTableGetSetByLid(u_int16_t lid,
        u_int8_t method,
        u_int16_t block_num,
        struct rn_sub_group_direction_tbl *p_sub_group_direction_table,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending RNSubGroupDirectionTable MAD by lid = %u, "  \
            "method = %u sub group block = %u\n",
            lid, method, block_num);
    /*
     * attr_mod[15:0] - block number
     */
    u_int32_t attr_mod = (0xff & block_num);

    int rc = this->SMPMadGetSetByLid(lid,
            method,
            IBIS_IB_ATTR_SMP_RN_SUB_GROUP_DIRECTION_TABLE,
            attr_mod,
            p_sub_group_direction_table,
            (const pack_data_func_t)rn_sub_group_direction_tbl_pack,
            (const unpack_data_func_t)rn_sub_group_direction_tbl_unpack,
            (const dump_data_func_t)rn_sub_group_direction_tbl_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPRNGenStringTableGetSetByLid(u_int16_t lid,
        u_int8_t method,
        u_int8_t direction_block,
        u_int8_t pLFTID,
        struct rn_gen_string_tbl *p_gen_string_table,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending RNGenStringTable MAD by lid = %u, "  \
            "method = %u direction block = %u pLFTID = %u\n",
            lid, method, direction_block, pLFTID);
    /*
     * attr_mod[10:8] - direction_block
     * attr_mod[7:0] - pLFTID
     */

    u_int32_t attr_mod = (0xf & pLFTID);
    attr_mod |= (0x7 & direction_block) << 8;

    int rc = this->SMPMadGetSetByLid(lid,
            method,
            IBIS_IB_ATTR_SMP_RN_GEN_STRING_TABLE,
            attr_mod,
            p_gen_string_table,
            (const pack_data_func_t)rn_gen_string_tbl_pack,
            (const unpack_data_func_t)rn_gen_string_tbl_unpack,
            (const dump_data_func_t)rn_gen_string_tbl_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

int Ibis::SMPRNGenBySubGroupPriorityMadGetSetByLid(u_int16_t lid,
        u_int8_t method,
        struct rn_gen_by_sub_group_prio *p_gen_by_sg_priority,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending RNGenBySubGroupPriority MAD by lid = %u, method = %u\n",
             lid, method);

    int rc = this->SMPMadGetSetByLid(lid,
            method,
            IBIS_IB_ATTR_SMP_RN_GEN_BY_SUB_GROUP_PRIORITY,
            0,
            p_gen_by_sg_priority,
            (const pack_data_func_t)rn_gen_by_sub_group_prio_pack,
            (const unpack_data_func_t)rn_gen_by_sub_group_prio_unpack,
            (const dump_data_func_t)rn_gen_by_sub_group_prio_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPRNRcvStringGetSetByLid(u_int16_t lid,
        u_int8_t method,
        u_int8_t string_block,
        struct rn_rcv_string *p_rcv_string,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending RNRcvString MAD by lid = %u, "  \
            "method = %u string block = %u\n",
            lid, method, string_block);
    /*
     * attr_mod[12:0] - block number
     */

    u_int32_t attr_mod = (0x1fff & string_block);

    int rc = this->SMPMadGetSetByLid(lid,
            method,
            IBIS_IB_ATTR_SMP_RN_RCV_STRING,
            attr_mod,
            p_rcv_string,
            (const pack_data_func_t)rn_rcv_string_pack,
            (const unpack_data_func_t)rn_rcv_string_unpack,
            (const dump_data_func_t)rn_rcv_string_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPRNXmitPortMaskGetSetByLid(u_int16_t lid,
        u_int8_t method,
        u_int8_t ports_block,
        struct rn_xmit_port_mask *p_xmit_port_mask,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending RNXmitPortMask MAD by lid = %u, "  \
            "method = %u ports block = %u\n",
            lid, method, ports_block);
    /*
     * attr_mod[0] - ports block
     */

    u_int32_t attr_mod = (0x1 & ports_block);

    int rc = this->SMPMadGetSetByLid(lid,
            method,
            IBIS_IB_ATTR_SMP_RN_XMIT_PORT_MASK,
            attr_mod,
            p_xmit_port_mask,
            (const pack_data_func_t)rn_xmit_port_mask_pack,
            (const unpack_data_func_t)rn_xmit_port_mask_unpack,
            (const dump_data_func_t)rn_xmit_port_mask_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPARGroupTableCopySetByLid(u_int16_t lid,
        u_int16_t group_to_copy,
        bool copy_direction,
        struct adaptive_routing_group_table_copy *p_group_table_copy,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending Set adaptive_routing_group_table_copy"
             " MAD by lid = %u, group to copy = %u copy direction = %s\n",
            lid, group_to_copy, (copy_direction?"yes":"no"));

    /*
     * attr_mod[11:0] - the AdaptiveGroup index to be copied.
     * attr_mod[12:12] - if set to 1, copies also the direction
     *                   associated with the AdaptiveRoutingSubGroups of that Group
     */
    u_int32_t attr_mod = (0x00000fff & group_to_copy);
    if (copy_direction)
        attr_mod |= (0x1 << 12);

    int rc = this->SMPMadGetSetByLid(lid,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_SMP_AR_GROUP_TABLE_COPY,
            attr_mod,
            p_group_table_copy,
            (const pack_data_func_t)adaptive_routing_group_table_copy_pack,
            (const unpack_data_func_t)adaptive_routing_group_table_copy_unpack,
            (const dump_data_func_t)adaptive_routing_group_table_copy_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPPortInfoMadGetByDirect(direct_route_t *p_direct_route,
        phys_port_t port_number,
        struct SMP_PortInfo *p_port_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_port_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPPortInfo MAD by direct = %s\n",
            this->ConvertDirPathToStr(p_direct_route).c_str());
    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_PORT_INFO,
            port_number,
            p_port_info,
            (const pack_data_func_t)SMP_PortInfo_pack,
            (const unpack_data_func_t)SMP_PortInfo_unpack,
            (const dump_data_func_t)SMP_PortInfo_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPPortInfoExtMadGetByDirect(
    direct_route_t *p_direct_route,
    phys_port_t port_number,
    struct SMP_PortInfoExtended *p_port_info_ext,
    const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_port_info_ext);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPPortInfoExtended MAD by direct = %s\n",
            this->ConvertDirPathToStr(p_direct_route).c_str());
    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_PORT_INFO_EXTENDED,
            port_number,
            p_port_info_ext,
            (const pack_data_func_t)SMP_PortInfoExtended_pack,
            (const unpack_data_func_t)SMP_PortInfoExtended_unpack,
            (const dump_data_func_t)SMP_PortInfoExtended_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPMlnxExtPortInfoMadGetByDirect(direct_route_t *p_direct_route,
        phys_port_t port_number,
        struct SMP_MlnxExtPortInfo *p_mlnx_ext_port_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_mlnx_ext_port_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPMlnxExtPortInfo MAD by direct = %s\n",
            this->ConvertDirPathToStr(p_direct_route).c_str());
    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_MLNXEXTPORTINFO,
            port_number,
            p_mlnx_ext_port_info,
            (const pack_data_func_t)SMP_MlnxExtPortInfo_pack,
            (const unpack_data_func_t)SMP_MlnxExtPortInfo_unpack,
            (const dump_data_func_t)SMP_MlnxExtPortInfo_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

//flags for smp vs general info mad's attribute modifier
#define SMP_VS_GENERAL_INFO_FW_INFO         0
#define SMP_VS_GENERAL_INFO_CAPABILITY_MASK 2

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPVSGeneralInfoFwInfoMadGetByDirect(direct_route_t *p_direct_route,
                                               struct FWInfo_Block_Element *p_general_info,
                                               const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_general_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending VSGeneralInfoFwInfo Get MAD by direct = %s\n",
             this->ConvertDirPathToStr(p_direct_route).c_str());

    u_int32_t attribute_modifier = 0;

    attribute_modifier |= (1 << SMP_VS_GENERAL_INFO_FW_INFO);

    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_VS_GENERAL_INFO,
            attribute_modifier,
            p_general_info,
            (const pack_data_func_t)FWInfo_Block_Element_pack,
            (const unpack_data_func_t)FWInfo_Block_Element_unpack,
            (const dump_data_func_t)FWInfo_Block_Element_dump,
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPVSGeneralInfoCapabilityMaskMadGetByDirect(direct_route_t *p_direct_route,
                                                       struct GeneralInfoCapabilityMask *p_general_info,
                                                       const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_general_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending VSGeneralInfoCapabilityMask Get MAD by direct = %s\n",
             this->ConvertDirPathToStr(p_direct_route).c_str());

    u_int32_t attribute_modifier = 0;

    attribute_modifier |= (1 << SMP_VS_GENERAL_INFO_CAPABILITY_MASK);

    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_VS_GENERAL_INFO,
            attribute_modifier,
            p_general_info,
            (const pack_data_func_t)GeneralInfoCapabilityMask_pack,
            (const unpack_data_func_t)GeneralInfoCapabilityMask_unpack,
            (const dump_data_func_t)GeneralInfoCapabilityMask_dump,
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPSwitchInfoMadGetByDirect(direct_route_t *p_direct_route,
        struct SMP_SwitchInfo *p_switch_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_switch_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPSwitchInfo MAD by direct = %s\n",
            this->ConvertDirPathToStr(p_direct_route).c_str());
    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_SWITCH_INFO,
            0,
            p_switch_info,
            (const pack_data_func_t)SMP_SwitchInfo_pack,
            (const unpack_data_func_t)SMP_SwitchInfo_unpack,
            (const dump_data_func_t)SMP_SwitchInfo_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPNodeInfoMadGetByDirect(direct_route_t *p_direct_route,
        struct SMP_NodeInfo *p_node_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_node_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPNodeInfo MAD by direct = %s\n",
            this->ConvertDirPathToStr(p_direct_route).c_str());
    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_NODE_INFO,
            0,
            p_node_info,
            (const pack_data_func_t)SMP_NodeInfo_pack,
            (const unpack_data_func_t)SMP_NodeInfo_unpack,
            (const dump_data_func_t)SMP_NodeInfo_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPNodeDescMadGetByDirect(direct_route_t *p_direct_route,
        struct SMP_NodeDesc *p_node_desc,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_node_desc);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPNodeDesc MAD by direct = %s\n",
            this->ConvertDirPathToStr(p_direct_route).c_str());
    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_NODE_DESC,
            0,
            p_node_desc,
            (const pack_data_func_t)SMP_NodeDesc_pack,
            (const unpack_data_func_t)SMP_NodeDesc_unpack,
            (const dump_data_func_t)SMP_NodeDesc_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPSMInfoMadGetByDirect(direct_route_t *p_direct_route,
        struct SMP_SMInfo *p_sm_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_sm_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPSMInfo MAD by direct = %s\n",
            this->ConvertDirPathToStr(p_direct_route).c_str());
    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_SMINFO,
            0,
            p_sm_info,
            (const pack_data_func_t)SMP_SMInfo_pack,
            (const unpack_data_func_t)SMP_SMInfo_unpack,
            (const dump_data_func_t)SMP_SMInfo_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPLinearForwardingTableGetByDirect(direct_route_t *p_direct_route,
        u_int32_t lid_to_port_block_num,
        struct SMP_LinearForwardingTable *p_linear_forwarding_table,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_linear_forwarding_table);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPLinearForwardingTable MAD by direct = %s\n",
            this->ConvertDirPathToStr(p_direct_route).c_str());
    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_LINEARFORWTBL,
            lid_to_port_block_num,
            p_linear_forwarding_table,
            (const pack_data_func_t)SMP_LinearForwardingTable_pack,
            (const unpack_data_func_t)SMP_LinearForwardingTable_unpack,
            (const dump_data_func_t)SMP_LinearForwardingTable_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPMulticastForwardingTableGetByDirect(direct_route_t *p_direct_route,
        u_int8_t port_group,
        u_int32_t lid_to_port_block_num,
        struct SMP_MulticastForwardingTable *p_multicast_forwarding_table,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_multicast_forwarding_table);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPMulticastForwardingTable MAD by direct = %s\n",
            this->ConvertDirPathToStr(p_direct_route).c_str());
    u_int32_t attribute_modifier = lid_to_port_block_num + (port_group << 28);
    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_MULTICASTFORWTBL,
            attribute_modifier,
            p_multicast_forwarding_table,
            (const pack_data_func_t)SMP_MulticastForwardingTable_pack,
            (const unpack_data_func_t)SMP_MulticastForwardingTable_unpack,
            (const dump_data_func_t)SMP_MulticastForwardingTable_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPPKeyTableGetByDirect(direct_route_t *p_direct_route,
        u_int16_t port_num,
        u_int16_t block_num,
        struct SMP_PKeyTable *p_pkey_table,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_pkey_table);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPPKeyTable MAD by direct = %s\n",
            this->ConvertDirPathToStr(p_direct_route).c_str());
    u_int32_t attribute_modifier = (port_num << 16) | block_num;
    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_PKEY_TBL,
            attribute_modifier,
            p_pkey_table,
            (const pack_data_func_t)SMP_PKeyTable_pack,
            (const unpack_data_func_t)SMP_PKeyTable_unpack,
            (const dump_data_func_t)SMP_PKeyTable_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPGUIDInfoTableGetByDirect(direct_route_t *p_direct_route,
        u_int32_t block_num,
        struct SMP_GUIDInfo *p_guid_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_guid_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPGUIDInfo MAD by direct = %s\n",
            this->ConvertDirPathToStr(p_direct_route).c_str());
    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_GUID_INFO,
            block_num,
            p_guid_info,
            (const pack_data_func_t)SMP_GUIDInfo_pack,
            (const unpack_data_func_t)SMP_GUIDInfo_unpack,
            (const dump_data_func_t)SMP_GUIDInfo_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPSLToVLMappingTableGetByDirect(direct_route_t *p_direct_route,
        phys_port_t out_port_number,
        phys_port_t in_port_number,
        struct SMP_SLToVLMappingTable *p_slvl_mapping,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_slvl_mapping);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPSLToVLMappingTable MAD by direct = %s"
             " out_port:%u in_port:%u \n",
             this->ConvertDirPathToStr(p_direct_route).c_str(),
             out_port_number, in_port_number);

    u_int32_t attribute_modifier = out_port_number + (in_port_number << 8);

    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_SLVL_TABLE,
            attribute_modifier,
            p_slvl_mapping,
            (const pack_data_func_t)SMP_SLToVLMappingTable_pack,
            (const unpack_data_func_t)SMP_SLToVLMappingTable_unpack,
            (const dump_data_func_t)SMP_SLToVLMappingTable_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPPLFTInfoMadGetSetByDirect(direct_route_t *p_direct_route,
        u_int8_t method,
        ib_private_lft_info *p_plft_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPPLFTInfoMadGetSetByDirect MAD by direct = %s, "  \
            "method = %u\n",
            this->ConvertDirPathToStr(p_direct_route).c_str(), method);
    /*
     * attr_mod must be 0
     */
    u_int32_t attr_mod = 0;

    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            method,
            IBIS_IB_ATTR_SMP_PLFT_INFO,
            attr_mod,
            p_plft_info,
            (const pack_data_func_t)ib_private_lft_info_pack,
            (const unpack_data_func_t)ib_private_lft_info_unpack,
            (const dump_data_func_t)ib_private_lft_info_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPPLFTDefMadGetSetByDirect(direct_route_t *p_direct_route,
        u_int8_t method,
        u_int8_t plft_block,
        struct ib_private_lft_def *p_plft_def,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending PLFTDef MAD by direct = %s, "  \
            "method = %u, plft block = %u\n",
            this->ConvertDirPathToStr(p_direct_route).c_str(), method, plft_block);
    /*
     * attr_mod[3-0] - block number of of 16 LFTs
     */
    u_int32_t attr_mod = (0x0000000f & plft_block);

    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            method,
            IBIS_IB_ATTR_SMP_PLFT_DEF,
            attr_mod,
            p_plft_def,
            (const pack_data_func_t)ib_private_lft_def_pack,
            (const unpack_data_func_t)ib_private_lft_def_unpack,
            (const dump_data_func_t)ib_private_lft_def_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPPLFTMapMadGetSetByDirect(direct_route_t *p_direct_route,
        u_int8_t method,
        u_int8_t pLFTID,
        struct ib_private_lft_map *p_plft_map,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending PLFTDef MAD by direct = %s, "  \
            "method = %u, pLFTIDk = %u\n",
            this->ConvertDirPathToStr(p_direct_route).c_str(), method, pLFTID);
    /*
     * attr_mod[7-0] - privateLFTID
     */
    u_int32_t attr_mod = (0x000000ff & pLFTID);

    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            method,
            IBIS_IB_ATTR_SMP_PLFT_MAP,
            attr_mod,
            p_plft_map,
            (const pack_data_func_t)ib_private_lft_map_pack,
            (const unpack_data_func_t)ib_private_lft_map_unpack,
            (const dump_data_func_t)ib_private_lft_map_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPPortSLToPrivateLFTMapGetSetByDirect(direct_route_t *p_direct_route,
        u_int8_t method,
        u_int8_t port_block,
        struct ib_port_sl_to_private_lft_map *p_plft_map,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending PortSLToPrivateLFTMap MAD by direct = %s, "  \
            "method = %u, port block = %u\n",
            this->ConvertDirPathToStr(p_direct_route).c_str(), method, port_block);
    /*
     * attr_mod[5-0] - port_block
     */
    u_int32_t attr_mod = (0x000000ff & port_block);

    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            method,
            IBIS_IB_ATTR_SMP_PORT_SL_TO_PLFT_MAP,
            attr_mod,
            p_plft_map,
            (const pack_data_func_t)ib_port_sl_to_private_lft_map_pack,
            (const unpack_data_func_t)ib_port_sl_to_private_lft_map_unpack,
            (const dump_data_func_t)ib_port_sl_to_private_lft_map_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPARInfoGetSetByDirect(direct_route_t *p_direct_route,
        u_int8_t method,
        bool get_cap,
        struct adaptive_routing_info *p_ar_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    u_int32_t attr_mod = 0;
    if ((method == IBIS_IB_MAD_METHOD_GET) && get_cap) {
        attr_mod = (0x1 << 31);
        IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPARInfo GetCap MAD by direct = %s\n",
                ConvertDirPathToStr(p_direct_route).c_str());
    } else {
        IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPARInfo MAD by direct = %s, method = %u\n",
            ConvertDirPathToStr(p_direct_route).c_str(), method);
    }

    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            method,
            IBIS_IB_ATTR_SMP_AR_INFO,
            attr_mod,
            p_ar_info,
            (const pack_data_func_t)adaptive_routing_info_pack,
            (const unpack_data_func_t)adaptive_routing_info_unpack,
            (const dump_data_func_t)adaptive_routing_info_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPARGroupTableGetSetByDirect(direct_route_t *p_direct_route,
        u_int8_t  method,
        u_int16_t group_block,
        u_int8_t  group_table,
        u_int8_t  pLFTID,
        struct ib_ar_group_table *p_ar_group_table,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPARGroupTable MAD by direct = %s, method = %u, "    \
            "group block = %u\n",
            this->ConvertDirPathToStr(p_direct_route).c_str(), method, group_block);
    /*
     * attr_mod[11:0] - group block (block contains 2 groups)
     * attr_mod[19:16] - group table
     * attr_mod[31:24] - privateLFTID
     */
    u_int32_t attr_mod = 0x00000fff & group_block;
    attr_mod |= (group_table << 16);
    attr_mod |= (pLFTID << 24);

    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            method,
            IBIS_IB_ATTR_SMP_AR_GROUP_TABLE,
            attr_mod,
            p_ar_group_table,
            (const pack_data_func_t)ib_ar_group_table_pack,
            (const unpack_data_func_t)ib_ar_group_table_unpack,
            (const dump_data_func_t)ib_ar_group_table_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPARLinearForwardingTableGetSetByDirect(direct_route_t *p_direct_route,
        u_int8_t method,
        u_int32_t lid_block,
        u_int8_t pLFTID,
        struct ib_ar_linear_forwarding_table_sx *p_ar_linear_forwarding_table,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPARLinearForwardingTable MAD by direct = %s, "  \
            "method = %u, lid block = %u\n",
            this->ConvertDirPathToStr(p_direct_route).c_str(), method, lid_block);
    /*
     * attr_mod[23:0] - lid block (block contains 32 lids)
     * attr_mod[31:24] - privateLFTID
     */
    u_int32_t attr_mod = (0x00ffffff & lid_block);
    attr_mod |= (pLFTID << 24);

    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            method,
            IBIS_IB_ATTR_SMP_AR_LINEAR_FORWARDING_TABLE_SX,
            attr_mod,
            p_ar_linear_forwarding_table,
            (const pack_data_func_t)ib_ar_linear_forwarding_table_sx_pack,
            (const unpack_data_func_t)ib_ar_linear_forwarding_table_sx_unpack,
            (const dump_data_func_t)ib_ar_linear_forwarding_table_sx_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPExtendedSwitchInfoMadGetSetByDirect(direct_route_t *p_direct_route,
        u_int8_t method,
        struct ib_extended_switch_info *p_ext_sw_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending ExtendedSwitchInfo MAD by direct = %s, "  \
            "method = %u\n",
            this->ConvertDirPathToStr(p_direct_route).c_str(), method);
    /*
     * attr_mod must be 0
     */
    u_int32_t attr_mod = 0;

    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            method,
            IBIS_IB_ATTR_SMP_EXTENDED_SWITCH_INFO,
            attr_mod,
            p_ext_sw_info,
            (const pack_data_func_t)ib_extended_switch_info_pack,
            (const unpack_data_func_t)ib_extended_switch_info_unpack,
            (const dump_data_func_t)ib_extended_switch_info_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPRNSubGroupDirectionTableGetSetByDirect(direct_route_t *p_direct_route,
        u_int8_t method,
        u_int16_t block_num,
        struct rn_sub_group_direction_tbl *p_sub_group_direction_table,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending RNSubGroupDirectionTable MAD by direct = %s, "  \
            "method = %u sub group block = %u\n",
            this->ConvertDirPathToStr(p_direct_route).c_str(), method, block_num);
    /*
     * attr_mod[15:0] - block number
     */
    u_int32_t attr_mod = (0xff & block_num);

    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            method,
            IBIS_IB_ATTR_SMP_RN_SUB_GROUP_DIRECTION_TABLE,
            attr_mod,
            p_sub_group_direction_table,
            (const pack_data_func_t)rn_sub_group_direction_tbl_pack,
            (const unpack_data_func_t)rn_sub_group_direction_tbl_unpack,
            (const dump_data_func_t)rn_sub_group_direction_tbl_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPRNGenStringTableGetSetByDirect(direct_route_t *p_direct_route,
        u_int8_t method,
        u_int8_t direction_block,
        u_int8_t pLFTID,
        struct rn_gen_string_tbl *p_gen_string_table,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending RNGenStringTable MAD by direct = %s, "  \
            "method = %u direction block = %u pLFTID = %u\n",
            this->ConvertDirPathToStr(p_direct_route).c_str(), method,
             direction_block, pLFTID);
    /*
     * attr_mod[?:8] - direction_block
     * attr_mod[7:0] - pLFTID
     */

    u_int32_t attr_mod = (0xf & pLFTID);
    attr_mod |= (0xf & direction_block) << 8; //!!!

    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            method,
            IBIS_IB_ATTR_SMP_RN_GEN_STRING_TABLE,
            attr_mod,
            p_gen_string_table,
            (const pack_data_func_t)rn_gen_string_tbl_pack,
            (const unpack_data_func_t)rn_gen_string_tbl_unpack,
            (const dump_data_func_t)rn_gen_string_tbl_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPRNGenBySubGroupPriorityMadGetSetByDirect(
        direct_route_t *p_direct_route,
        u_int8_t method,
        struct rn_gen_by_sub_group_prio *p_gen_by_sg_priority,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending RNGenBySubGroupPriority MAD by direct = %s, "  \
            "method = %u\n",
            this->ConvertDirPathToStr(p_direct_route).c_str(), method);

    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            method,
            IBIS_IB_ATTR_SMP_RN_GEN_BY_SUB_GROUP_PRIORITY,
            0,
            p_gen_by_sg_priority,
            (const pack_data_func_t)rn_gen_by_sub_group_prio_pack,
            (const unpack_data_func_t)rn_gen_by_sub_group_prio_unpack,
            (const dump_data_func_t)rn_gen_by_sub_group_prio_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPRNRcvStringGetSetByDirect(direct_route_t *p_direct_route,
        u_int8_t method,
        u_int16_t string_block,
        struct rn_rcv_string *p_rcv_string,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending RNRcvString MAD by direct = %s, "  \
            "method = %u string block = %u\n",
            this->ConvertDirPathToStr(p_direct_route).c_str(), method,
             string_block);
    /*
     * attr_mod[12:0] - block number
     */

    u_int32_t attr_mod = (0x1fff & string_block);

    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            method,
            IBIS_IB_ATTR_SMP_RN_RCV_STRING,
            attr_mod,
            p_rcv_string,
            (const pack_data_func_t)rn_rcv_string_pack,
            (const unpack_data_func_t)rn_rcv_string_unpack,
            (const dump_data_func_t)rn_rcv_string_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPRNXmitPortMaskGetSetByDirect(direct_route_t *p_direct_route,
        u_int8_t method,
        u_int8_t ports_block,
        struct rn_xmit_port_mask *p_xmit_port_mask,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending RNXmitPortMask MAD by direct = %s, "  \
             "method = %u ports block = %u\n",
             this->ConvertDirPathToStr(p_direct_route).c_str(), method,
             ports_block);
    /*
     * attr_mod[0] - ports block
     */

    u_int32_t attr_mod = (0x1 & ports_block);

    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            method,
            IBIS_IB_ATTR_SMP_RN_XMIT_PORT_MASK,
            attr_mod,
            p_xmit_port_mask,
            (const pack_data_func_t)rn_xmit_port_mask_pack,
            (const unpack_data_func_t)rn_xmit_port_mask_unpack,
            (const dump_data_func_t)rn_xmit_port_mask_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPARGroupTableCopySetByDirect(direct_route_t *p_direct_route,
        u_int16_t group_to_copy,
        bool copy_direction,
        struct adaptive_routing_group_table_copy *p_group_table_copy,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending Set adaptive_routing_group_table_copy"
            " MAD by direct = %s, group to copy = %u copy direction = %s\n",
            this->ConvertDirPathToStr(p_direct_route).c_str(),
            group_to_copy, (copy_direction?"yes":"no"));

    /*
     * attr_mod[11:0] - the AdaptiveGroup index to be copied.
     * attr_mod[12:12] - if set to 1, copies also the direction
     *                   associated with the AdaptiveRoutingSubGroups of that Group
     */
    u_int32_t attr_mod = (0x00000fff & group_to_copy);
    if (copy_direction)
        attr_mod |= (0x1 << 12);

    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_SMP_AR_GROUP_TABLE_COPY,
            attr_mod,
            p_group_table_copy,
            (const pack_data_func_t)adaptive_routing_group_table_copy_pack,
            (const unpack_data_func_t)adaptive_routing_group_table_copy_unpack,
            (const dump_data_func_t)adaptive_routing_group_table_copy_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

int Ibis::SMPVirtualizationInfoMadGetByLid(u_int16_t lid,
        struct SMP_VirtualizationInfo *p_virtual_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_virtual_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPVirtualizationInfo MAD by lid = %u\n", lid);
    int rc = this->SMPMadGetSetByLid(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_VIRTUALIZATION_INFO,
            0,
            p_virtual_info,
            (const pack_data_func_t)SMP_VirtualizationInfo_pack,
            (const unpack_data_func_t)SMP_VirtualizationInfo_unpack,
            (const dump_data_func_t)SMP_VirtualizationInfo_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::SMPVPortInfoMadGetByLid(u_int16_t lid,
        virtual_port_t vport_num,
        struct SMP_VPortInfo *p_vport_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_vport_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPVPortInfo MAD by lid = %u\n", lid);
    u_int32_t attribute_modifier = (vport_num << 16);
    int rc = this->SMPMadGetSetByLid(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_VPORT_INFO,
            attribute_modifier,
            p_vport_info,
            (const pack_data_func_t)SMP_VPortInfo_pack,
            (const unpack_data_func_t)SMP_VPortInfo_unpack,
            (const dump_data_func_t)SMP_VPortInfo_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::SMPVPortStateMadGetByLid(u_int16_t lid,
        u_int16_t block_num,
        struct SMP_VPortState *p_vport_state,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_vport_state);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPVPortState MAD by lid = %u\n", lid);
    u_int32_t attribute_modifier = block_num;
    int rc = this->SMPMadGetSetByLid(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_VPORT_STATE,
            attribute_modifier,
            p_vport_state,
            (const pack_data_func_t)SMP_VPortState_pack,
            (const unpack_data_func_t)SMP_VPortState_unpack,
            (const dump_data_func_t)SMP_VPortState_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::SMPVNodeInfoMadGetByLid(u_int16_t lid,
        virtual_port_t vport_num,
        struct SMP_VNodeInfo *p_vnode_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_vnode_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPVNodeInfo MAD by lid = %u\n", lid);
    u_int32_t attribute_modifier = (vport_num << 16);
    int rc = this->SMPMadGetSetByLid(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_VNODE_INFO,
            attribute_modifier,
            p_vnode_info,
            (const pack_data_func_t)SMP_VNodeInfo_pack,
            (const unpack_data_func_t)SMP_VNodeInfo_unpack,
            (const dump_data_func_t)SMP_VNodeInfo_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::SMPVNodeDescriptionMadGetByLid(u_int16_t lid,
        virtual_port_t vport_num,
        struct SMP_NodeDesc *p_vnode_description,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_vnode_description);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPVNodeDescription MAD by lid = %u\n", lid);
    u_int32_t attribute_modifier = (vport_num << 16);
    int rc = this->SMPMadGetSetByLid(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_VNODE_DESC,
            attribute_modifier,
            p_vnode_description,
            (const pack_data_func_t)SMP_NodeDesc_pack,
            (const unpack_data_func_t)SMP_NodeDesc_unpack,
            (const dump_data_func_t)SMP_NodeDesc_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::SMPVPortPKeyTblMadGetByLid(u_int16_t lid,
        virtual_port_t vport_num,
        u_int16_t block_num,
        struct SMP_PKeyTable *p_vport_pkey_tbl,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_vport_pkey_tbl);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending SMPVPortPKeyTbl MAD by lid = %u\n", lid);
    u_int32_t attribute_modifier = (vport_num << 16) | block_num;
    int rc = this->SMPMadGetSetByLid(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_VPORT_PKEY_TBL,
            attribute_modifier,
            p_vport_pkey_tbl,
            (const pack_data_func_t)SMP_PKeyTable_pack,
            (const unpack_data_func_t)SMP_PKeyTable_unpack,
            (const dump_data_func_t)SMP_PKeyTable_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::SMPVPortPKeyTblMadGetByDirect(direct_route_t *p_direct_route,
        virtual_port_t vport_num,
        u_int16_t block_num,
        struct SMP_PKeyTable *p_vport_pkey_tbl,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_vport_pkey_tbl);
    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending SMPVPortPKeyTbl MAD by direct = %s\n",
             this->ConvertDirPathToStr(p_direct_route).c_str());
    u_int32_t attribute_modifier = (vport_num << 16) | block_num;
    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_VPORT_PKEY_TBL,
            attribute_modifier,
            p_vport_pkey_tbl,
            (const pack_data_func_t)SMP_PKeyTable_pack,
            (const unpack_data_func_t)SMP_PKeyTable_unpack,
            (const dump_data_func_t)SMP_PKeyTable_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::SMPTempSensingDataGetByDirect(direct_route_t *p_direct_route,
        struct SMP_TempSensing *p_tempsens,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_tempsens);

    u_int32_t attribute_modifier = 0;

    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending SMPTemperatureSensing MAD by direct = %s\n",
             this->ConvertDirPathToStr(p_direct_route).c_str());

    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_TEMP_SENSING,
            attribute_modifier,
            p_tempsens,
            (const pack_data_func_t)SMP_TempSensing_pack,
            (const unpack_data_func_t)SMP_TempSensing_unpack,
            (const dump_data_func_t)SMP_TempSensing_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::SMPQosConfigSLGetByDirect(direct_route_t *p_direct_route,
        struct SMP_QosConfigSL *p_qos_config_sl,
        const clbck_data_t *p_clbck_data,
        phys_port_t port_num)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_qos_config_sl);
    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_QOS_CONFIG_SL,
            port_num,
            p_qos_config_sl,
            (const pack_data_func_t)SMP_QosConfigSL_pack,
            (const unpack_data_func_t)SMP_QosConfigSL_unpack,
            (const dump_data_func_t)SMP_QosConfigSL_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::SMPVPortQoSConfigSLGetByDirect(direct_route_t *p_direct_route,
        struct SMP_QosConfigSL *p_qos_config_sl,
        const clbck_data_t *p_clbck_data,
        virtual_port_t vport_num)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_qos_config_sl);

    u_int32_t attribute_modifier = IBIS_AM_VPORT_QOS_CONFIG_SL | vport_num;

    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_QOS_CONFIG_SL,
            attribute_modifier,
            p_qos_config_sl,
            (const pack_data_func_t)SMP_QosConfigSL_pack,
            (const unpack_data_func_t)SMP_QosConfigSL_unpack,
            (const dump_data_func_t)SMP_QosConfigSL_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::SMPRouterInfoGetByDirect(direct_route_t *p_direct_route,
        struct SMP_RouterInfo *p_router_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_router_info);
    int rc = this->SMPMadGetSetByDirect(p_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_SMP_ROUTER_INFO,
            0,
            p_router_info,
            (const pack_data_func_t)SMP_RouterInfo_pack,
            (const unpack_data_func_t)SMP_RouterInfo_unpack,
            (const dump_data_func_t)SMP_RouterInfo_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::SMPAdjRouterTableGetByDirect(direct_route_t *p_direct_route,
        u_int8_t block_num,
        struct SMP_AdjSiteLocalSubnTbl *p_router_tbl,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_router_tbl);
    //The attribute modifier BlockNumber AM[4:0].
    u_int32_t attribute_modifier = block_num & 0x1f;
    int rc = this->SMPMadGetSetByDirect(p_direct_route,
                    IBIS_IB_MAD_METHOD_GET,
                    IBIS_IB_ATTR_SMP_RTR_ADJ_SITE_LOCAL_SUBNETS_TABLE,
                    attribute_modifier,
                    p_router_tbl,
                    (const pack_data_func_t)SMP_AdjSiteLocalSubnTbl_pack,
                    (const unpack_data_func_t)SMP_AdjSiteLocalSubnTbl_unpack,
                    (const dump_data_func_t)SMP_AdjSiteLocalSubnTbl_dump,
                    p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::SMPNextHopRouterTableGetByDirect(direct_route_t *p_direct_route,
        u_int32_t block_num,
        struct SMP_NextHopTbl *p_router_tbl,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_router_tbl);
    //The attribute modifier BlockNumber AM[29:0].
    u_int32_t attribute_modifier = block_num & 0x3fffffff;
    int rc = this->SMPMadGetSetByDirect(p_direct_route,
                    IBIS_IB_MAD_METHOD_GET,
                    IBIS_IB_ATTR_SMP_RTR_NEXT_HOP_TABLE,
                    attribute_modifier,
                    p_router_tbl,
                    (const pack_data_func_t)SMP_NextHopTbl_pack,
                    (const unpack_data_func_t)SMP_NextHopTbl_unpack,
                    (const dump_data_func_t)SMP_NextHopTbl_dump,
                    p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}
