/*
 *  Copyright (c) 2004-2013 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <string.h>

#include "ibis.h"
#include "packets/packets_layouts.h"


/****************************************************/
/******************private methods*******************/
/****************************************************/
/*
 * Returns: mad status[bits: 0-15] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::VSMadGetSet(u_int16_t lid,
        u_int8_t method,
        u_int16_t attribute_id,
        u_int32_t attribute_modifier,
        void *p_vs_attribute_data,
        const pack_data_func_t vs_pack_attribute_data_func,
        const unpack_data_func_t vs_unpack_attribute_data_func,
        const dump_data_func_t vs_dump_attribute_data_func,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    struct MAD_VendorSpec vs_mad;
    CLEAR_STRUCT(vs_mad);

    //build vs lid route header
    this->CommonMadHeaderBuild(&vs_mad.MAD_Header_Common,
            IBIS_IB_CLASS_VENDOR_MELLANOX,
            method,
            attribute_id,
            attribute_modifier);

    vs_mad.V_Key = 0;   //TODO - new ibis feature

    //send the mad
    IBIS_RETURN(this->MadGetSet(lid, 1, 0, IBIS_IB_DEFAULT_QP1_QKEY,
            IBIS_IB_CLASS_VENDOR_MELLANOX,
            method, attribute_id, attribute_modifier,
            IBIS_IB_DATA_OFFSET_VENDOR_MELLANOX,
            &vs_mad, p_vs_attribute_data,
            (const pack_data_func_t)MAD_VendorSpec_pack,
            (const unpack_data_func_t)MAD_VendorSpec_unpack,
            (const dump_data_func_t)MAD_VendorSpec_dump,
            (const pack_data_func_t)vs_pack_attribute_data_func,
            (const unpack_data_func_t)vs_unpack_attribute_data_func,
            (const dump_data_func_t)vs_dump_attribute_data_func,
            p_clbck_data));
}


/****************************************************/
/*******************public methods*******************/
/****************************************************/
/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::VSGeneralInfoGet(u_int16_t lid,
        struct VendorSpec_GeneralInfo *p_general_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_general_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending VSGeneralInfo Get MAD lid = %u\n", lid);
    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_GENERAL_INFO,
            0,
            p_general_info,
            (const pack_data_func_t)VendorSpec_GeneralInfo_pack,
            (const unpack_data_func_t)VendorSpec_GeneralInfo_unpack,
            (const dump_data_func_t)VendorSpec_GeneralInfo_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

#define LLR_SYMBOL_ERRORS_FLAG_OFFSET  31

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::VSPortLLRStatisticsGet(u_int16_t lid,
        phys_port_t port_number,
        struct VendorSpec_PortLLRStatistics *p_port_llr_statistics,
        bool get_symbol_errors,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_port_llr_statistics);

    u_int32_t attribute_modifier = 0;

    p_port_llr_statistics->PortSelect = (u_int8_t)port_number;
    if (get_symbol_errors)
        attribute_modifier |= (1 << LLR_SYMBOL_ERRORS_FLAG_OFFSET);

    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending VSPortLLRStatistics Get MAD lid = %u port = %u\n",
             lid, port_number);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_LLR_STATISTICS,
            attribute_modifier,
            p_port_llr_statistics,
            (const pack_data_func_t)VendorSpec_PortLLRStatistics_pack,
            (const unpack_data_func_t)VendorSpec_PortLLRStatistics_unpack,
            (const dump_data_func_t)VendorSpec_PortLLRStatistics_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::VSPortLLRStatisticsClear(u_int16_t lid,
        phys_port_t port_number,
        bool clear_symbol_errors,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    u_int32_t attribute_modifier = 0;

    struct VendorSpec_PortLLRStatistics port_llr_statistics;
    CLEAR_STRUCT(port_llr_statistics);
    port_llr_statistics.PortSelect = port_number;
    if (clear_symbol_errors)
        attribute_modifier |= (1 << LLR_SYMBOL_ERRORS_FLAG_OFFSET);


    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending VSPortLLRStatistics Reset MAD lid = %u port = %u\n",
             lid, port_number);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_LLR_STATISTICS,
            attribute_modifier,
            &port_llr_statistics,
            (const pack_data_func_t)VendorSpec_PortLLRStatistics_pack,
            (const unpack_data_func_t)VendorSpec_PortLLRStatistics_unpack,
            (const dump_data_func_t)VendorSpec_PortLLRStatistics_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::VSPortLLRStatisticsSet(u_int16_t lid,
        phys_port_t port_number,
        struct VendorSpec_PortLLRStatistics *p_port_llr_statistics,
        bool set_symbol_errors,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_port_llr_statistics);

    u_int32_t attribute_modifier = 0;

    p_port_llr_statistics->PortSelect = (u_int8_t)port_number;
    if (set_symbol_errors)
        attribute_modifier |= (1 << LLR_SYMBOL_ERRORS_FLAG_OFFSET);

    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending VSPortLLRStatistics Set MAD lid = %u port = %u\n",
             lid, port_number);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_LLR_STATISTICS,
            attribute_modifier,
            p_port_llr_statistics,
            (const pack_data_func_t)VendorSpec_PortLLRStatistics_pack,
            (const unpack_data_func_t)VendorSpec_PortLLRStatistics_unpack,
            (const dump_data_func_t)VendorSpec_PortLLRStatistics_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED /
 * IBIS_MAD_STATUS_RECV_FAILED / IBIS_MAD_STATUS_TIMEOUT /
 * IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::VSDiagnosticDataPageClear(u_int16_t lid,
        phys_port_t port_number,
        u_int8_t page_number,
        struct VS_DiagnosticData *p_dc,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_dc);

    u_int32_t attribute_modifier = 0;

    attribute_modifier = port_number | (page_number << 16);

    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending VSDiagnosticCounters Set MAD lid = %u port = %u\n",
             lid, port_number);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_DIAGNOSTIC_COUNTERS,
            attribute_modifier,
            p_dc,
            (const pack_data_func_t)VS_DiagnosticData_pack,
            (const unpack_data_func_t)VS_DiagnosticData_unpack,
            (const dump_data_func_t)VS_DiagnosticData_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED /
 * IBIS_MAD_STATUS_RECV_FAILED / IBIS_MAD_STATUS_TIMEOUT /
 * IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::VSDiagnosticDataPageClear_AM(u_int16_t lid,
        u_int32_t attribute_modifier,
        struct VS_DiagnosticData *p_dc,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_dc);

    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending VSDiagnosticCounters Set MAD lid = %u\n",
             lid);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_DIAGNOSTIC_COUNTERS,
            attribute_modifier,
            p_dc,
            (const pack_data_func_t)VS_DiagnosticData_pack,
            (const unpack_data_func_t)VS_DiagnosticData_unpack,
            (const dump_data_func_t)VS_DiagnosticData_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::VSDiagnosticDataGet(u_int16_t lid,
        phys_port_t port_number,
        u_int8_t page_number,
        struct VS_DiagnosticData *p_dc,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_dc);

    u_int32_t attribute_modifier = 0;

    attribute_modifier = (page_number << 16) | port_number;

    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending VSDiagnosticCounters Get MAD lid = %u port = %u\n",
             lid, port_number);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_DIAGNOSTIC_COUNTERS,
            attribute_modifier,
            p_dc,
            (const pack_data_func_t)VS_DiagnosticData_pack,
            (const unpack_data_func_t)VS_DiagnosticData_unpack,
            (const dump_data_func_t)VS_DiagnosticData_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::VSDiagnosticDataGet_AM(u_int16_t lid,
        u_int32_t attribute_modifier,
        struct VS_DiagnosticData *p_dc,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_dc);

    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending VSDiagnosticCounters Get MAD lid = %u\n",
             lid);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_DIAGNOSTIC_COUNTERS,
            attribute_modifier,
            p_dc,
            (const pack_data_func_t)VS_DiagnosticData_pack,
            (const unpack_data_func_t)VS_DiagnosticData_unpack,
            (const dump_data_func_t)VS_DiagnosticData_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::VSPortRNCountersClear(u_int16_t lid,
        phys_port_t port_number,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    struct port_rn_counters rn_counters;
    CLEAR_STRUCT(rn_counters);

    rn_counters.counter_select = 0xF;
    rn_counters.port_select = port_number;

    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending VSPortRNCounters Clear MAD lid = %u port = %u\n",
             lid, port_number);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_RN_COUNTERS,
            0,
            &rn_counters,
            (const pack_data_func_t)port_rn_counters_pack,
            (const unpack_data_func_t)port_rn_counters_unpack,
            (const dump_data_func_t)port_rn_counters_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::VSPortRNCountersGet(u_int16_t lid,
        phys_port_t port_number,
        struct port_rn_counters *p_rn_counters,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_rn_counters);

    p_rn_counters->port_select = port_number;

    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending VSPortRNCounters Get MAD lid = %u port = %u\n",
             lid, port_number);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_RN_COUNTERS,
            0,
            p_rn_counters,
            (const pack_data_func_t)port_rn_counters_pack,
            (const unpack_data_func_t)port_rn_counters_unpack,
            (const dump_data_func_t)port_rn_counters_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::VSPerVLCounters(bool is_reset_cntr,
        u_int16_t lid,
        phys_port_t port_number,
        u_int32_t attr_id,
        struct PM_PortRcvXmitCntrsSlVl *p_pm_port_rcvxmit_data_slvl,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_pm_port_rcvxmit_data_slvl);

    if (is_reset_cntr)
        p_pm_port_rcvxmit_data_slvl->counter_select = 0xffff;
    p_pm_port_rcvxmit_data_slvl->port_select = port_number;
    IBIS_LOG(TT_LOG_LEVEL_MAD,
            "Sending %s attribute %u Get MAD lid = %u port = %u\n",
            (is_reset_cntr ? "Clear" : "Get"),
            attr_id, lid, port_number);

    int rc = this->VSMadGetSet(lid,
            is_reset_cntr ? IBIS_IB_MAD_METHOD_SET : IBIS_IB_MAD_METHOD_GET,
            attr_id,
            0,
            p_pm_port_rcvxmit_data_slvl,
            (const pack_data_func_t)PM_PortRcvXmitCntrsSlVl_pack,
            (const unpack_data_func_t)PM_PortRcvXmitCntrsSlVl_unpack,
            (const dump_data_func_t)PM_PortRcvXmitCntrsSlVl_dump,
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

