/*
 * © Copyright (c) 2004-2016 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include "ibis.h"

MkeyPort::MkeyPort(MkeyNode *pNodePtr):
                           m_guid(0), m_pPeerPort(NULL), m_pNode(pNodePtr)
{
    IBIS_ENTER;
    IBIS_RETURN_VOID;
}

// connect the port to another node port
int
MkeyPort::connect(MkeyPort *pOtherPort)
{
    // we can not simply override existing connections
    if (m_pPeerPort) {
        // we only do care if not the requested remote previously conn.
        if (m_pPeerPort != pOtherPort) {
            cout << "-E- Port " << m_pPeerPort->m_guid
                 << " already connected to " << m_pPeerPort->m_pPeerPort->m_guid
                 << " while connecting:" << pOtherPort->m_guid << endl;
          IBIS_RETURN(1);
        }
    }
    m_pPeerPort = pOtherPort;

    // Check the other side was not previously connected
    if (pOtherPort->m_pPeerPort) {
        if (pOtherPort->m_pPeerPort != this) {
            // it was connected to a wrong port so disconnect
            cout << "-E- Port " << pOtherPort->m_guid
                 << " already connected to:" << pOtherPort->m_pPeerPort->m_guid
                 << " while connecting:" << this->m_guid << endl;
          IBIS_RETURN(1);
        }
    }
    pOtherPort->m_pPeerPort = this;

    IBIS_RETURN(0);
}

MkeyNode::MkeyNode(uint64_t guid, uint64_t mkey, phys_port_t numPorts):
                          m_guid(guid), m_mkey(mkey), m_numPorts(numPorts)
{
    IBIS_ENTER;
    Ports.resize(numPorts+1, NULL);
    IBIS_RETURN_VOID;
}

MkeyNode::~MkeyNode()
{
    IBIS_ENTER;
    // use unsigned int index to prevent overwrap
    for (unsigned int i = 0; i < Ports.size(); i++) {
        MkeyPort *p_port = Ports[i];
        if (p_port)
            delete p_port;
    }
    Ports.clear();
    IBIS_RETURN_VOID;
}

MkeyPort *
MkeyNode::makePort(phys_port_t  portNum)
{
    IBIS_ENTER;
    if (Ports.size() <= portNum) {
        cout << "-E- makePort: invalid port number:" << portNum << endl;
        IBIS_RETURN(NULL);
    }

    if(!Ports[portNum])
        Ports[portNum] = new MkeyPort(this);

    IBIS_RETURN(Ports[portNum]);
}

MkeyNode * MkeyNode::getPeerNodeByPortNum(phys_port_t portNum)
{
    IBIS_ENTER;
    MkeyPort *p_port = NULL;
    if (Ports.empty() || (Ports.size() <= portNum))
        IBIS_RETURN(NULL);

    p_port = Ports[portNum];
    if (p_port && p_port->m_pPeerPort)
        IBIS_RETURN(p_port->m_pPeerPort->m_pNode)

    IBIS_RETURN(NULL);
}

FilesBasedMKeyManager::FilesBasedMKeyManager():m_constMkey(0), m_rootMkeyNode(NULL)
{
    IBIS_ENTER;
    IBIS_RETURN_VOID;
}

FilesBasedMKeyManager::~FilesBasedMKeyManager()
{
    IBIS_ENTER;
    MkeyNode *pMkeyNode = NULL;
    m_pguidToMkey.clear();
    for (nguid_to_MkeyNode_map_t::iterator It = m_nguidToMkeymngr.begin();
         It != m_nguidToMkeymngr.end(); It++) {
        pMkeyNode = (*It).second;
       if (pMkeyNode)
           delete pMkeyNode;
    }
    m_lidToMkey.clear();
    m_guidToNumOfPorts.clear();
    m_pguidPnToPguidPn.clear();
    IBIS_RETURN_VOID;
}

int FilesBasedMKeyManager::parseGuid2MKey(const string pguid2mkeyFilePath)

{
    IBIS_ENTER;
    unsigned int lineNum = 0, rc = 0;
    char sLine[1024] = {'\0'};
    // For every line in guid2mkey file, parse guid and his mkey.
    // Example :
    //  0x0002c90000000041 0x0000000000000001
    regExp pGuid2MkeyLine("\\s*(0x[0-9a-fA-F]+)\\s*(0x[0-9a-fA-F]+)\\s*");
    rexMatch *p_rexRes = NULL;

    ifstream fs(pguid2mkeyFilePath.c_str());

    if (fs.fail()) {
        cout << "-E- Failed to open file:" << pguid2mkeyFilePath << endl;
        IBIS_RETURN(1);
    }

    cout << "-I- Parsing file: " << pguid2mkeyFilePath << endl;

    //parse file
    while (fs.good()) {
        fs.getline(sLine, 1024);
        lineNum++;

        if (!sLine[0]) // skip empty lines
            continue;

        if ((p_rexRes = pGuid2MkeyLine.apply(sLine))) {

            uint64_t port_guid = (u_int64_t)strtoul(p_rexRes->field(1).c_str(),
                                                    NULL, 16);
            uint64_t mkey = (u_int64_t)strtoul(p_rexRes->field(2).c_str(),
                                               NULL, 16);
            if (!port_guid) {
                cout << "-E- Illegal port guid " <<  p_rexRes->field(1).c_str()
                     << ", at line: " << lineNum << endl;
                delete p_rexRes;
                p_rexRes = NULL;
                continue;
            }

            m_pguidToMkey.insert(std::pair<uint64_t, uint64_t>(port_guid,
                                                               mkey));
        } else
            cout << "-E- Unsupported syntax at line: " << lineNum << endl;

        if (p_rexRes) {
            delete p_rexRes;
            p_rexRes = NULL;
        }
    }

    fs.close();

    IBIS_RETURN(rc);
}


int FilesBasedMKeyManager::parseNeighbors(const string neighborsFilePath)
{
    IBIS_ENTER;
    unsigned int lineNum = 0, rc = 0;
    char sLine[1024] = {'\0'};
    // For every line in neighbors file, parse port guid and his peer port guid.
    // The line patern is: <port guid>:port_number <peer port guid>:port_number
    // Example :
    //  0x0002c90000000041:5 0x0002c90000000046:8
    regExp pGuidLine("\\s*(0x[0-9a-fA-F]+):([0-9]{1,3})\\s*(0x[0-9a-fA-F]+):([0-9]{1,3})\\s*");

    rexMatch *p_rexRes = NULL;
    pair_pguid_port_num_t pair_pguid_pn;
    pair_pguid_port_num_t peer_pair_pguid_pn;

    ifstream fs(neighborsFilePath.c_str());

    if (fs.fail()) {
        cout << "-E- Failed to open file:" << neighborsFilePath << endl;
        IBIS_RETURN(1);
    }

    cout << "-I- Parsing file: " << neighborsFilePath << endl;

    //parse file
    while (fs.good()) {
        fs.getline(sLine, 1024);
        lineNum++;

        if (!sLine[0]) // skip empty lines
            continue;

        if ((p_rexRes = pGuidLine.apply(sLine))) {

            uint64_t port_guid = (u_int64_t)strtoul(p_rexRes->field(1).c_str(),
                                                    NULL, 16);
            phys_port_t port_num = (u_int8_t)strtoul(p_rexRes->field(2).c_str(),
                                                     NULL, 10);
            pair_pguid_pn = pair_pguid_port_num_t(port_guid, port_num);

            uint64_t peer_port_guid =
                (u_int64_t)strtoul(p_rexRes->field(3).c_str(), NULL, 16);
            phys_port_t peer_port_num =
                (u_int8_t)strtoul(p_rexRes->field(4).c_str(), NULL, 10);
            peer_pair_pguid_pn =
                pair_pguid_port_num_t(peer_port_guid, peer_port_num);

            if (!port_guid) {
                cout << "-E- Illegal port guid " <<  p_rexRes->field(1).c_str()
                     << ", at line: " << lineNum << endl;
                delete p_rexRes;
                p_rexRes = NULL;
                continue;
            }

            if (!peer_port_guid) {
                cout << "-E- Illegal peer port guid " <<  p_rexRes->field(3).c_str()
                     << ", at line: " << lineNum << endl;
                delete p_rexRes;
                p_rexRes = NULL;
                continue;
            }

            if (peer_port_num == IBIS_INVALID_PORT_NUM) {
                cout << "-E- Invalid port num: " << IBIS_INVALID_PORT_NUM
                     << ", on guid: " << p_rexRes->field(3).c_str()
                     << ", at line: " << lineNum << endl;
                delete p_rexRes;
                p_rexRes = NULL;
                continue;
            }

            m_pguidPnToPguidPn.insert(make_pair(pair_pguid_pn,
                                                peer_pair_pguid_pn));

            pguid_to_port_num_map_t::iterator It =
                m_guidToNumOfPorts.find(port_guid);
            if (It == m_guidToNumOfPorts.end())
                m_guidToNumOfPorts.insert(pair_pguid_port_num_t(port_guid,
                                                                port_num));
            else if ((*It).second < port_num)
                (*It).second = port_num;

            pguid_to_port_num_map_t::iterator It1 =
                m_guidToNumOfPorts.find(peer_port_guid);
            if (It1 == m_guidToNumOfPorts.end())
                m_guidToNumOfPorts.insert(pair_pguid_port_num_t(peer_port_guid,
                                                                peer_port_num));
            else if ((*It1).second < peer_port_num)
                (*It1).second = peer_port_num;

        } else
            cout << "-E- Unsupported syntax at line: " << lineNum << endl;

        if (p_rexRes) {
            delete p_rexRes;
            p_rexRes = NULL;
        }
    }

    fs.close();

    IBIS_RETURN(rc);
}

uint64_t
FilesBasedMKeyManager::getMKeyByNodeGuid(uint64_t guid)
{
    IBIS_ENTER;
    uint64_t mkey = 0;
    pguid_to_mkey_map_t::iterator nI = m_pguidToMkey.find(guid);
    if (nI != m_pguidToMkey.end()) {
        mkey = (*nI).second;
    }
    IBIS_RETURN(mkey);
}

MkeyNode *
FilesBasedMKeyManager::makeMKeyNode(uint64_t nodeGuid)
{
    IBIS_ENTER;
    MkeyNode *p_node = NULL;
    uint64_t mkey = getMKeyByNodeGuid(nodeGuid);
    phys_port_t numPorts = 0;
    pguid_to_port_num_map_t::iterator It = m_guidToNumOfPorts.find(nodeGuid);
    numPorts = (*It).second;
    IBUTILS_LOG(TT_LOG_LEVEL_DEBUG,
                "makeMKeyNode numPorts:%d node guid:" U64H_FMT
                " \n",
                numPorts, nodeGuid);
    p_node = new MkeyNode(nodeGuid, mkey, numPorts);
    m_nguidToMkeymngr.insert(make_pair(nodeGuid, p_node));
    IBIS_RETURN(p_node);
}

MkeyPort *
FilesBasedMKeyManager::setMKeyNodePort(MkeyNode *pMkeyNode,
                             phys_port_t  portNum)
{
    IBIS_ENTER;
    // add port
    IBUTILS_LOG(TT_LOG_LEVEL_DEBUG,
               "setNodePort portNum:%d node guid:" U64H_FMT
               " \n",
               portNum, nodeGuid);

    if (!pMkeyNode)
        return NULL;
    MkeyPort *p_port = pMkeyNode->makePort(portNum);
    if (!p_port)
    {
        cout << "-E- failed to get port number: " << (unsigned int)portNum
                << " for node: " << pMkeyNode->m_guid << endl;
        return NULL;
    }

    IBIS_RETURN(p_port);
}

MkeyNode *
FilesBasedMKeyManager::getMKeyNodeByNodeGuid(uint64_t guid)
{
    IBIS_ENTER;
    MkeyNode *p_node = NULL;
    nguid_to_MkeyNode_map_t::iterator nI = m_nguidToMkeymngr.find(guid);
    if (nI != m_nguidToMkeymngr.end()) {
        p_node = (*nI).second;
    }
    IBIS_RETURN(p_node);
}

uint64_t
FilesBasedMKeyManager::getMkeyByLid(u_int16_t lid) const
{
    IBIS_ENTER;
    uint64_t mkey = 0;
    if (getConstMKey())
        IBIS_RETURN(getConstMKey());
    lid_to_mkey_map_t::const_iterator It = m_lidToMkey.find(lid);
    if (It != m_lidToMkey.end())
        mkey = (*It).second;
    IBIS_RETURN(mkey);
}

int FilesBasedMKeyManager::buildMkeyManagerFabricTree()
{
    IBIS_ENTER;
    int rc = 0;
    for (pguid_portNum_to_peer_pguid_portNum_map_t::iterator it =
            m_pguidPnToPguidPn.begin();
         it != m_pguidPnToPguidPn.end();
         ++it) {

        pair_pguid_port_num_t guid_port1= (*it).first;
        pair_pguid_port_num_t guid_port2= (*it).second;
        rc = addLink(guid_port1.first, guid_port1.second,
                     guid_port2.first, guid_port2.second);
        if (rc) {
            cout << "-E- failed to link node guid" << guid_port1.first
                 << " to node guid " << guid_port2.first << endl;
            goto Exit;
        }
    }
Exit:
    IBIS_RETURN(rc);
}

int FilesBasedMKeyManager::makeLinkBetweenPorts(MkeyPort *pPort1,
                                      MkeyPort *pPort2)
{
    IBIS_ENTER;
    if (!pPort1 || !pPort2)
        IBIS_RETURN(1);
    IBUTILS_LOG(TT_LOG_LEVEL_DEBUG,
               "make Link Between MKey Ports (node_guid;port_num): \n"
               "(0x%02x,%d)<->(0x%02x,%d)  \n"
               pPort1->p_node->guid, pPort1->num,
               pPort2->p_node->guid, pPort2->num);

    if(pPort1->connect(pPort2))
        IBIS_RETURN(1);

    IBIS_RETURN(0);
}


int FilesBasedMKeyManager::addLink(uint64_t nodeGuid1, phys_port_t portNum1,
                         uint64_t nodeGuid2, phys_port_t portNum2)
{
    IBIS_ENTER;
    MkeyNode  *pMkeyNode1 = NULL;
    MkeyNode  *pMkeyNode2 = NULL;
    MkeyPort  *pMkeyPort1 = NULL;
    MkeyPort  *pMkeyPort2 = NULL;

    //create the nodes
    pMkeyNode1 = getMKeyNodeByNodeGuid(nodeGuid1);
    if (!pMkeyNode1) {
        pMkeyNode1 = makeMKeyNode(nodeGuid1);
        if (!pMkeyNode1) {
            cout << "-E- failed to allocate new mkey node, guid=" << nodeGuid1 << endl;
            return 1;
        }

    }
    pMkeyNode2 = getMKeyNodeByNodeGuid(nodeGuid2);
    if (!pMkeyNode2) {
        pMkeyNode2 = makeMKeyNode(nodeGuid2);
        if (!pMkeyNode2) {
            cout << "-E- failed to allocate new mkey node, guid=" << nodeGuid2 << endl;
            return 1;
        }
    }

    //create the ports
    pMkeyPort1 = setMKeyNodePort(pMkeyNode1, portNum1);
    if (!pMkeyPort1) {
        cout << "-E- failed to allocate new port, guid=" << nodeGuid1 << endl;
        return 1;
    }
    pMkeyPort2 = setMKeyNodePort(pMkeyNode2, portNum2);
    if (!pMkeyPort2) {
        cout << "-E- failed to allocate new port, guid=" << nodeGuid2 << endl;
        return 1;
    }

    //create the link itself
    IBIS_RETURN(makeLinkBetweenPorts(pMkeyPort1, pMkeyPort2));
}

int FilesBasedMKeyManager::setMkeyManagerFabricTreeRoot(u_int64_t rootGuid)
{
    IBIS_ENTER;
    this->m_rootMkeyNode = getMKeyNodeByNodeGuid(rootGuid);
    if (!this->m_rootMkeyNode) {
        cout << "-E- FilesBasedMKeyManager failed to get root node from DB,"  \
                " guid=" << rootGuid << endl;
        IBIS_RETURN(1);
    }

    IBIS_RETURN(0);
}

uint64_t FilesBasedMKeyManager::getMkeyFromDRPath(const direct_route_t *pCurrDirectRoute) const
{
    IBIS_ENTER;
    uint64_t mkey = 0;
    if (getConstMKey())
        IBIS_RETURN(getConstMKey());

    MkeyNode * curr_node_mkey = m_rootMkeyNode;
    for (int i = 1; i < pCurrDirectRoute->length; i++) {
        phys_port_t port_num = pCurrDirectRoute->path.BYTE[i];
        if (!curr_node_mkey)
            IBIS_RETURN(mkey);
        if ((port_num <= 0) || (port_num > curr_node_mkey->m_numPorts)) {
            cout << "-E- FilesBasedMKeyManager invalid port num:" << port_num
                 << "at index:" << i << " from DR path." << endl;
            IBIS_RETURN(mkey);
        }
        curr_node_mkey = curr_node_mkey->getPeerNodeByPortNum(port_num);
    }

    if (curr_node_mkey)
        mkey = curr_node_mkey->getMKey();
    IBIS_RETURN(mkey);
}

void FilesBasedMKeyManager::setLidToNodeGuidMap(const lid_to_guid_map_t &lid2guid)
{
    IBIS_ENTER;
    uint64_t mkey = 0, guid = 0;
    lid_to_guid_map_t   lidToGuid = lid2guid;
    for (lid_to_guid_map_t::iterator it =
            lidToGuid.begin();
         it != lidToGuid.end();
         ++it) {
        guid = (*it).second;
        mkey = 0;
        pguid_to_mkey_map_t::iterator gI = m_pguidToMkey.find(guid);
        if (gI != m_pguidToMkey.end())
            mkey = (*gI).second;

        m_lidToMkey[(*it).first] = mkey;
    }
    IBIS_RETURN_VOID;
}

int FilesBasedMKeyManager::parseAndCreateMKeyFabric(string neighborsFilePath,
                                          string pguid2mkeyFilePath)
{
    IBIS_ENTER;

    int rc = parseNeighbors(neighborsFilePath);
    if (rc) {
        cout << "-E- Failed Parsing Neighbors file: " << neighborsFilePath << endl;
        goto Exit;
    }

    rc = parseGuid2MKey(pguid2mkeyFilePath);
    if (rc) {
        cout << "-E- Failed Parsing Guid2MKey file: " << pguid2mkeyFilePath << endl;
        goto Exit;
    }

    rc = buildMkeyManagerFabricTree();
    if (rc)
        cout << "-E- Failed Build MKey Manager" << endl;

Exit:
    IBIS_RETURN(rc);
}
