/*
 * © Copyright (c) 2004-2013 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "tool_trace.h"
#include "tool_trace_spinlock.h"

#if !defined(_WIN32) && !defined(_WIN64)
#include <sys/time.h>
#include <time.h>
#include <syslog.h>
#include <unistd.h>
#else           /* Windows */
#include <windows.h>
#endif          /* Linux */


/******************************************************/
/* new types */
#if !defined(_WIN32) && !defined(_WIN64)
typedef pid_t tt_pid_t;
static char *month_str[] = {
        "Jan",
        "Feb",
        "Mar",
        "Apr",
        "May",
        "Jun",
        "Jul",
        "Aug",
        "Sep",
        "Oct",
        "Nov",
        "Dec"
};
#else           /* Windows */
typedef DWORD tt_pid_t;
#endif          /* Linux */


typedef struct tt_log {
    FILE *out_port;                     /* log out port */
    char log_file_name[1024];           /* log file name */
    unsigned long count;                /* count the number of bytes in log */
    unsigned long max_size;             /* maximum log size */
    tt_spinlock_t lock;                 /* spinlock object */
    tt_log_module_t module;             /* module verbosity */
    tt_log_level_t level;               /* level verbosity */
    int  flush;                         /* indicate if we want to flush after every print to log */
    int accum_log_file;                 /* append to exist log file or not (open a new file) */
} tt_log_t;


/******************************************************/
/* globals */
static tt_log_t tt_log_obj;
static int log_exit_count = 0;
static int is_log_obj_ready = 0;

int construct_secure_file(const char *file_name, FILE ** file_handle)
{

    if (!file_handle)          //invalid file_handle ptr
        return 1;

    if (*file_handle)          //already open
        return 0;

#ifndef WIN32 //Fix security vulnerability
    srand((unsigned int)time(NULL));
    int r = rand();
    char tmp_path[1100]; // enough to hold name+all numbers up to 64-bits
    sprintf(tmp_path, "%s_%X",file_name, r);

    // if file exist and remove faild, return error
    if ((access(file_name, F_OK) != -1) && (remove(file_name) < 0))
        return 1;

    // if file exist and remove faild, return error
    if ((access(tmp_path, F_OK) != -1) && (remove(tmp_path) < 0))
        return 1;

    *file_handle = fopen(tmp_path, "w+");
    if (!*file_handle)
        return 1;

    if (rename(tmp_path, file_name)) {
        fclose(*file_handle);
        *file_handle = NULL;
        return 1;
    }
#else
    *file_handle = fopen(file_name, "w+");
    if (!*file_handle)
        return 1;
#endif
    return 0;
}

/******************************************************/
static int open_out_port(void)
{
    struct stat st;
    int i_err;

    /* open log file */
    if (tt_log_obj.accum_log_file)
        tt_log_obj.out_port = fopen(tt_log_obj.log_file_name, "a+");
    else
        construct_secure_file(tt_log_obj.log_file_name,
                              &tt_log_obj.out_port);
    if (!tt_log_obj.out_port) {
#if !defined(_WIN32) && !defined(_WIN64)    //TODO
        syslog(LOG_CRIT, "Cannot open file \'%s\' for %s: %s\n",
                tt_log_obj.log_file_name,
                tt_log_obj.accum_log_file ? "appending" : "writing",
                        strerror(errno));
#endif          /* Linux */
        fprintf(stderr, "Cannot open file \'%s\' for %s: %s\n",
                tt_log_obj.log_file_name,
                tt_log_obj.accum_log_file ? "appending" : "writing",
                        strerror(errno));
        return -1;
    }

    /* get the file bytes count */
#if !defined(_WIN32) && !defined(_WIN64)    //TODO
    i_err = fstat(fileno(tt_log_obj.out_port), &st);
#else           /* Windows */
    i_err = _fstat(_fileno(tt_log_obj.out_port), (struct _stat *)&st);
#endif          /* Linux */
    if (i_err != 0) {
#if !defined(_WIN32) && !defined(_WIN64)    //TODO
        syslog(LOG_CRIT, "Cannot query file \'%s\' for bytes size: %s\n",
                tt_log_obj.log_file_name, strerror(errno));
#endif          /* Linux */
        fprintf(stderr, "Cannot query file \'%s\' for bytes size: %s\n",
                tt_log_obj.log_file_name, strerror(errno));
        return -1;
    }
    tt_log_obj.count = (unsigned long)st.st_size;

#if !defined(_WIN32) && !defined(_WIN64)    //TODO
    syslog(LOG_NOTICE, "%s log file opened\n", tt_log_obj.log_file_name);
#endif          /* Linux */
    return 0;
}


/******************************************************/
TT_API int tt_log_construct(IN int flush,
        IN tt_log_module_t module_verbosity,
        IN tt_log_level_t level_verbosity,
        IN const char *log_file,
        IN unsigned long max_size,
        IN int accum_log_file)
{
    memset(&tt_log_obj, 0, sizeof(tt_log_obj));

    tt_log_obj.flush = flush;
    tt_log_obj.module = module_verbosity;
    tt_log_obj.level = level_verbosity;
    if (log_file != NULL)
        strcpy(tt_log_obj.log_file_name, log_file);
    tt_log_obj.max_size = max_size << 20;	/* convert size in MB to bytes */
    tt_log_obj.accum_log_file = accum_log_file;

    if (log_file == NULL ||
            !strcmp(log_file, "-") ||
            !strcmp(log_file, "stdout"))
        tt_log_obj.out_port = stdout;
    else if (!strcmp(log_file, "stderr"))
        tt_log_obj.out_port = stderr;
    else if (open_out_port())
        return -1;

    if (tt_spinlock_init(&tt_log_obj.lock))
        return -1;

    is_log_obj_ready = 1;
    return 0;
}


/******************************************************/
static void parse_verbosity_from_env_var(OUT tt_log_module_t *module_verbosity,
        OUT tt_log_level_t *level_verbosity)
{
    static tt_log_module_map_t map_moudle[TT_LOG_NUM_OF_MODULES] = {
            {"NONE", TT_LOG_MODULE_NONE},
            {"IBIS", TT_LOG_MODULE_IBIS},
            {"IBDIAG", TT_LOG_MODULE_IBDIAG},
            {"IBDM", TT_LOG_MODULE_IBDM},
            {"IBMGTSIM", TT_LOG_MODULE_IBMGTSIM},
            {"IBDIAGNET", TT_LOG_MODULE_IBDIAGNET},
            {"CCMGR", TT_LOG_MODULE_CCMGR},
            {"ARMGR", TT_LOG_MODULE_ARMGR},
            {"ALL", TT_LOG_MODULE_ALL}
    };
    char *tt_log_level_str;
    char *pch1, *pch2, *pch3;
    tt_log_module_t found_module_verb   = TT_LOG_MODULE_NONE;
    tt_log_level_t found_level_verb     = TT_LOG_LEVEL_NONE;
    int i;

    *module_verbosity   = TT_LOG_MODULE_NONE;
    *level_verbosity    = TT_LOG_LEVEL_NONE;
    tt_log_level_str = getenv(TT_LOG_LEVEL_ENV_VARIABLE_NAME);
    if ((!tt_log_level_str) || (!strcmp(tt_log_level_str, ""))) {    //env not exists
        *module_verbosity   = TT_LOG_DEFAULT_MODULE;
        *level_verbosity    = TT_LOG_DEFAULT_LEVEL;
        return;
    }
    if (!(pch1 = strtok(tt_log_level_str, ";")))
        goto error_parse;
    if (pch1[0] == '?')
        goto dump_help;
    pch2 = strtok(NULL, ";");
    if (strtok(NULL, ";"))
        goto error_parse;
    do {
        pch3 = strtok(pch1, "=");
        if (!pch3)
            goto error_parse;
        if (strcmp(pch3, "module") == 0) {
            if (found_module_verb) {
                printf("\"module\" is defined twice\n");
                goto error_parse;
            }
            found_module_verb = 1;
            while ((pch3 = strtok(NULL, ","))) {
                for (i = 0; i < TT_LOG_NUM_OF_MODULES; ++i) {
                    if (strcmp(pch3, map_moudle[i].module_name) == 0) {
                        *module_verbosity |= map_moudle[i].module;
                        break;
                    }
                }
                if (i == TT_LOG_NUM_OF_MODULES)
                    goto error_parse;
            }
        } else if (strcmp(pch3, "level") == 0) {
            if (found_level_verb) {
                printf("\"level\" is defined twice\n");
                goto error_parse;
            }
            found_level_verb = 1;
            while ((pch3 = strtok(NULL, ","))) {
                *level_verbosity |= (tt_log_level_t)strtoul(pch3, NULL, 0);
            }
        } else {
            goto error_parse;
        }
        pch1 = pch2;
        pch2 = NULL;
    } while (pch1);

    return;

dump_help:
#if !defined(_WIN32) && !defined(_WIN64)
    fprintf(stdout, "export %s=", TT_LOG_LEVEL_ENV_VARIABLE_NAME);
#else           /* Windows */
    fprintf(stdout, "set %s=", TT_LOG_LEVEL_ENV_VARIABLE_NAME);
#endif          /* Linux */
    fprintf(stdout, "\'level=val1,val2,...;module=str1,str2...\' / \'module=str1,str2...;level=val1,val2,...\'\n\n");
    fprintf(stdout, "Allowed values for level are [0x00-0xff]\n");
    fprintf(stdout, "Allowed strings for module are: ");
    for (i = 0; i < TT_LOG_NUM_OF_MODULES; ++i)
        fprintf(stdout, "%s ", map_moudle[i].module_name);
    fprintf(stdout, "\n\n");
    exit(1);

error_parse:
    *module_verbosity   = TT_LOG_DEFAULT_MODULE;
    *level_verbosity    = TT_LOG_DEFAULT_LEVEL;

#if !defined(_WIN32) && !defined(_WIN64)    //TODO
    syslog(LOG_WARNING,
            "-W- Failed to parse %s env variable, using default verbosity values\n",
                        TT_LOG_LEVEL_ENV_VARIABLE_NAME);
#endif          /* Linux */
    fprintf(stderr,
            "-W- Failed to parse %s env variable, using default verbosity values\n",
            TT_LOG_LEVEL_ENV_VARIABLE_NAME);
    return;
}


/******************************************************/
TT_API int tt_log_construct_v2(IN int flush,
        IN const char *log_file,
        IN unsigned long max_size,
        IN int accum_log_file)
{
    tt_log_module_t module_verbosity;
    tt_log_level_t level_verbosity;
    parse_verbosity_from_env_var(&module_verbosity, &level_verbosity);
    return tt_log_construct(flush, module_verbosity, level_verbosity,
            log_file, max_size, accum_log_file);
}


/******************************************************/
TT_API void tt_log_destroy(void)
{
    if (!is_log_obj_ready)
        return;

    tt_spinlock_destroy(&tt_log_obj.lock);
    if (tt_log_obj.out_port != stdout) {
        fclose(tt_log_obj.out_port);
        tt_log_obj.out_port = stdout;       //after close the log everything will be dump to screen
    }
}


/******************************************************/
#if !defined(_WIN32) && !defined(_WIN64)

static void truncate_log_file(void)
{
    int fd = fileno(tt_log_obj.out_port);
    if (ftruncate(fd, (off_t)0) < 0)
        fprintf(stderr, "truncate_log_file: cannot truncate: %s\n",
                strerror(errno));
    if (lseek(fd, (off_t)0, SEEK_SET) < 0)
        fprintf(stderr, "truncate_log_file: cannot rewind: %s\n",
                strerror(errno));
    tt_log_obj.count = 0;
}

#else           /* Windows */

static void truncate_log_file(void)
{
    fprintf(stderr,
            "truncate_log_file: cannot truncate on windows system (not supported yet)\n");
}

#endif          /* Linux */


/******************************************************/
#if !defined(_WIN32) && !defined(_WIN64)

static u_int64_t get_time_stamp(void)
{
    u_int64_t tstamp;
    struct timeval tv;

    timerclear(&tv);
    gettimeofday(&tv, NULL);

    /* Convert the time of day into a microsecond timestamp. */
    tstamp = ((u_int64_t) tv.tv_sec * 1000000) + (u_int64_t) tv.tv_usec;

    return (tstamp);
}

#endif          /* Linux */


/******************************************************/
static inline void __tt_vlog(IN tt_log_module_t module, IN tt_log_level_t verbosity,
                             IN const char *p_str, va_list args)
{
    char buffer[LOG_ENTRY_SIZE_MAX];
    int ret;
#if defined(_WIN32) || defined(_WIN64)
    tt_pid_t pid = GetCurrentThreadId();
    SYSTEMTIME st;
#else           /* Linux */
    tt_pid_t pid = getpid();
    time_t tim;
    struct tm result;
    u_int64_t time_usecs;
    u_int32_t usecs;
#endif          /* Windows */

    /* parse args to be printed */
    vsprintf(buffer, p_str, args);

    tt_spinlock_acquire(&tt_log_obj.lock);

#if !defined(_WIN32) && !defined(_WIN64)    //TODO
    /* this is a call to the syslog */
    if (verbosity & TT_LOG_LEVEL_SYS) {
        syslog(LOG_INFO, "%s\n", buffer);

        /* SYSLOG should go to stdout too */
        if (tt_log_obj.out_port != stdout) {
            printf("%s\n", buffer);
            fflush(stdout);
        }
    }
#endif          /* Linux */

    /* regular log to default out_port */
    if (tt_log_obj.max_size && (tt_log_obj.count > tt_log_obj.max_size)) {
        /* truncate here */
        truncate_log_file();
    }

#if defined(_WIN32) || defined(_WIN64)
    GetLocalTime(&st);
_retry:
    /* TODO - add month + day also here */
    ret =
            fprintf(tt_log_obj.out_port,
                    "[%02d:%02d:%02d:%03d][%04X][0x%02x 0x%02x] -> %s",
                    st.wHour, st.wMinute, st.wSecond, st.wMilliseconds,
                    pid, module, verbosity, buffer);
#else           /* Linux */
    time_usecs = get_time_stamp();
    tim = (time_t)(time_usecs / 1000000);
    usecs = (u_int32_t)(time_usecs % 1000000);
    localtime_r(&tim, &result);
_retry:
ret =
        fprintf(tt_log_obj.out_port,
                "[%s %02d %02d:%02d:%02d %06d][%04X][0x%02x 0x%02x] -> %s",
                (result.tm_mon < 12 ? month_str[result.tm_mon] : "???"), result.tm_mday,
                result.tm_hour, result.tm_min, result.tm_sec, usecs,
                pid, module, verbosity, buffer);
#endif          /* Windows */

    /*  flush log */
    if (ret > 0 &&
            (tt_log_obj.flush || (verbosity & (TT_LOG_LEVEL_ERROR | TT_LOG_LEVEL_SYS))) &&
            (fflush(tt_log_obj.out_port) < 0))
        ret = -1;

    if (ret >= 0) {
        log_exit_count = 0;
        tt_log_obj.count += (unsigned long)ret;
    } else if (log_exit_count < 3) {
        log_exit_count++;
        if (errno == ENOSPC && tt_log_obj.max_size) {
#ifdef DEBUG
            fprintf(stderr,
                    "tt_log: write failed: %s. Truncating log file.\n",
                    strerror(errno));
#endif          /* DEBUG */
            truncate_log_file();
            goto _retry;
        }
        fprintf(stderr, "tt_log: write failed: %s\n", strerror(errno));
    }

    tt_spinlock_release(&tt_log_obj.lock);
}


TT_API void tt_log(IN tt_log_module_t module, IN tt_log_level_t verbosity,
                   IN const char *p_str, ...)
{
    va_list args;

    /* Check if log exists */
    if (!is_log_obj_ready)
        return;
    /* Check module verbosity */
    if (!(module & tt_log_obj.module))
        return;
    /* Check level verbosity - if this is a call to syslog than always print it */
    if (!(verbosity & (TT_LOG_LEVEL_SYS | tt_log_obj.level)))
        return;

    va_start(args, p_str);
    __tt_vlog(module, verbosity, p_str, args);
    va_end(args);
}


TT_API void tt_vlog(IN tt_log_module_t module, IN tt_log_level_t verbosity,
                    IN const char *p_str, IN va_list args)
{
    /* Check if log exists */
    if (!is_log_obj_ready)
        return;
    /* Check module verbosity */
    if (!(module & tt_log_obj.module))
        return;
    /* Check level verbosity - if this is a call to syslog than always print it */
    if (!(verbosity & (TT_LOG_LEVEL_SYS | tt_log_obj.level)))
        return;

    __tt_vlog(module, verbosity, p_str, args);
}


/******************************************************/
TT_API FILE* tt_get_log_file_port(void)
{
    if (!is_log_obj_ready)
        return stdout;
    return tt_log_obj.out_port;
}


/******************************************************/
TT_API tt_log_module_t tt_log_get_module_verbosity(void)
{
    if (!is_log_obj_ready)
        return TT_LOG_MODULE_NONE;
    return tt_log_obj.module;
}

TT_API tt_log_level_t tt_log_get_level_verbosity(void)
{
    if (!is_log_obj_ready)
        return TT_LOG_LEVEL_NONE;
    return tt_log_obj.level;
}

TT_API void tt_log_set_module_verbosity(tt_log_module_t module)
{
    if (!is_log_obj_ready)
        return;
    tt_log_obj.module = module;
}

TT_API void tt_log_set_level_verbosity(tt_log_level_t level)
{
    if (!is_log_obj_ready)
        return;
    tt_log_obj.level = level;
}

TT_API int tt_is_module_verbosity_active(tt_log_module_t module)
{
    if (!is_log_obj_ready)
        return 0;
    return ((tt_log_obj.module & module) != 0);
}

TT_API int tt_is_level_verbosity_active(tt_log_level_t level)
{
    if (!is_log_obj_ready)
        return 0;
    return ((tt_log_obj.level & level) != 0);
}

