/*
 * Copyright (c) 2004-2019 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


/*
  IB Fabric Data Model
  This file holds implementation of the data model classes and methods

*/


#include <iomanip>
#include <sstream>
#include <algorithm>

#include <infiniband/ibdm/Fabric.h>
#include <infiniband/ibdm/SysDef.h>
#include <infiniband/ibdm/Regexp.h>

bool  IBNode::usePSL = false;
bool  IBNode::useSLVL = false;
u_int8_t IBNode::maxSL = 0;

#define MAX_INTERNAL_LOG_SIZE 1024*1024

//////////////////////////////////////////////////////////////////////////////
#include <infiniband/ibdm/ibdm_version.h>

const char * get_ibdm_version()
{
    #ifdef IBDM_VERSION
        return IBDM_VERSION;
    #else
        return "unknown";
    #endif
}


///////////////////////////////////////////////////////////////////////////////
// Track verbosity:
uint8_t FabricUtilsVerboseLevel = 0x1;


//////////////////////////////////////////////////////////////////////////////
//
// log functions and gloval variable
//
stringstream ibdmLog;
streambuf *ibdmCoutBuf;

void ibdmUseInternalLog()
{
    streambuf *psbuf = ibdmLog.rdbuf();
    ibdmCoutBuf = cout.rdbuf(psbuf);
}

void ibdmUseCoutLog()
{
    cout.rdbuf(ibdmCoutBuf);
}

void ibdmClearInternalLog()
{
    ibdmLog.str("");
}

char *ibdmGetAndClearInternalLog()
{
    char *res;
    ibdmUseCoutLog();
    int size = 0;
    try {
        ibdmLog.seekg(0, ios::end);
        size = (int)ibdmLog.tellg();
        ibdmLog.seekg(0, ios::beg);

        if (size > MAX_INTERNAL_LOG_SIZE) {
            ibdmLog.seekp(MAX_INTERNAL_LOG_SIZE);
            string truncate_str("... \n-E- Log to long - truncated. \n");
            ibdmLog << truncate_str;
            size = MAX_INTERNAL_LOG_SIZE + (int)truncate_str.length();
        }
        res = (char*)malloc(size + 1);
    } catch(std::bad_alloc& exc) {
        res = NULL;
    };

    if (!res)
        return NULL;

    ibdmLog.read(res, size);
    res[size] = '\0';

    ibdmClearInternalLog();
    ibdmUseInternalLog();
    return res;
}


//////////////////////////////////////////////////////////////////////////////
// static functions
static int
generateNodeAndSystemNames(IBNodeType type, uint64_t sysGuid, uint64_t nodeGuid,
			   string nodeDesc, bool generateNotFromDesc,
			   string &nodeName, string &sysName, string &sysType,
			   bool &mlx_nd_format, bool mapped_node_description = false)
{
    char nbuf[1024] = {'\0'};
    char sbuf[1000] = {'\0'};
    char tbuf[1024] = {'\0'};
    char desc[256] = {'\0'};
    regExp rexMF0("^MF0;(\\S+):([^/]+)/(\\S+)$");
    regExp rexHCA("^\\s*(\\S*)\\s+HCA-([0-9]+)");
    regExp rexmlx("^\\s*(\\S*)\\s+(mlx[0-9]+_[0-9]+)");
    string localhostStr("localhost");
    rexMatch *p_rexRes;
    mlx_nd_format = false;

    // if nothing complex matches...

    //if node description is user defined in the map file,
    //and it's not duplicate, then take it as the default name
    if (mapped_node_description && !generateNotFromDesc)
        sprintf(nbuf, "S%016" PRIx64 "/%s", sysGuid, nodeDesc.c_str());
    else
        sprintf(nbuf, "S%016" PRIx64 "/N%016" PRIx64, sysGuid, nodeGuid);

    sprintf(sbuf, "S%016" PRIx64, sysGuid);
    sprintf(tbuf, "SYS%016" PRIx64, sysGuid);

    if (nodeDesc.empty())
	goto done;

    memset(desc, 0, sizeof(desc));
    strcpy(desc, nodeDesc.c_str());

    // try HCA format
    if (type == IB_CA_NODE) {   //deal with hca
        p_rexRes = rexHCA.apply(desc);
        string hca_num = "U";
        if (!p_rexRes) {
            hca_num = "";
            p_rexRes = rexmlx.apply(desc);
            if (p_rexRes)
                mlx_nd_format = true;
        }
        if (p_rexRes) {
            if (!(generateNotFromDesc ||
                  p_rexRes->field(1).empty() ||
                  p_rexRes->field(1) == localhostStr)) {
                sprintf(sbuf, "%s", p_rexRes->field(1).c_str());
            }
            hca_num += p_rexRes->field(2);
            sprintf(nbuf, "%s/%s", sbuf, hca_num.c_str());
            sprintf(tbuf, "HCA");
            delete p_rexRes;
            goto done;
        }
    }

    if (generateNotFromDesc)
	goto done;

    // try MF0 format
    p_rexRes = rexMF0.apply(desc);
    if (p_rexRes) {
        sprintf(nbuf, "%s/%s", p_rexRes->field(1).c_str(), p_rexRes->field(3).c_str());
        sprintf(sbuf, "%s", p_rexRes->field(1).c_str());
        sprintf(tbuf, "%s", p_rexRes->field(2).c_str());
        delete p_rexRes;
        goto done;
    }

    // IPR
    if (type == IB_CA_NODE) {

        char host_name[256];
        char dev_name[256];

        if (sscanf(nodeDesc.c_str(), "Mellanox %s IO %s", dev_name, host_name) == 2) {
            sprintf(nbuf, "%s/IPR", host_name);
            sprintf(sbuf, "%s", host_name);
            sprintf(tbuf, "ISR%s", dev_name);
            goto done;
        }
    } else {
        // 4036, 4036E
        char sys_name[256];
        char sys_t[256];

        if (sscanf(nodeDesc.c_str(), "Mellanox 403%s # %s", sys_t, sys_name) == 2) {
            sprintf(nbuf, "%s/U1", sys_name);
            sprintf(sbuf, "%s", sys_name);
            sprintf(tbuf, "ISR403%s",sys_t);
            goto done;
        }
    }

done:
    nodeName.assign(nbuf);
    sysName.assign(sbuf);
    sysType.assign(tbuf);
    return(0);
}


//////////////////////////////////////////////////////////////////////////////
//
// CLASS IBPort:
//

// constructor
IBPort::IBPort(IBNode *p_nodePtr, phys_port_t number)
{
    p_node          = p_nodePtr;
    num             = number;
    p_sysPort       = NULL;
    p_remotePort    = NULL;
    base_lid        = IB_LID_UNASSIGNED;
    lmc             = 0;
    memset(&guid,0,sizeof(uint64_t));
    counter1        = 0;
    counter2        = 0;
    width           = IB_UNKNOWN_LINK_WIDTH;
    speed           = IB_UNKNOWN_LINK_SPEED;
    port_state      = IB_UNKNOWN_PORT_STATE;
    in_sub_fabric   = true;
    fec_mode        = IB_FEC_NA;
    special_port_type = NotSpecial;
    special_port_sup_diag_data = true;

    //set port index
    createIndex = ((p_nodePtr) && (p_nodePtr->p_fabric)) ? p_nodePtr->p_fabric->getPortIndex() : 0;
}

// destructor
IBPort::~IBPort()
{
    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {
        cout << "-I- Destructing Port:" << p_node->name
                << "/" << (unsigned int)num << endl;
    }

    for (map_vportnum_vport::iterator vpI = this->VPorts.begin();
            vpI != this->VPorts.end(); ++vpI) {
        IBVPort *p_vport = vpI->second;
        if(p_vport)
           delete p_vport;
    }

    // if was connected - remove the connection:
    if (p_remotePort)
        p_remotePort->p_remotePort = NULL;

    // if has a system port - delete it too.
    if (p_sysPort) {
        p_sysPort->p_nodePort = NULL;
        delete p_sysPort;
    }

    if(!channels.empty()) {
        size_t size = channels.size();
        for (size_t i = 0 ; i < size ; i++) {
            delete channels[i];
        }
        channels.clear();
    }
}


// if the node isn't in sub cluster then the port is also not in sub cluster
bool
IBPort::getInSubFabric() {
    if (!p_node->getInSubFabric())
        return false;
    else
        return this->in_sub_fabric;
}

void
IBPort::setInSubFabric(bool in_sub_fabric) {
    this->in_sub_fabric = in_sub_fabric;
}

// Set the guid in the fabric too:
void
IBPort::guid_set(uint64_t g)
{
    if (p_node && p_node->p_fabric) {
        p_node->p_fabric->PortByGuid[g] = this;
        guid = g;
    }
}

// Get the common sate on the link, i.e the minimum.
// In the normal fabric both ports on the link will have the same state,
// otherwise some port's fw is lying. Since we don't know which, we can assume
// the minimum.
// The ibdm algorithms (maybe topo match) rely on the ports on the link
// to have the same state.
// Also in case of different data, ibdiagnet will fail to construct lst file.
// Use this function whenever port state is needed.
// When you want to report an inconsistency use get_internal_state
IBPortState IBPort::get_common_state() {
    IBPortState state;

    if (p_remotePort)
        state = min(this->port_state, p_remotePort->port_state);
    else
        state = this->port_state;

    return state;
}

// Get the common width on the link, i.e the minimum.
// In the normal fabric both ports on the link will have the same width,
// otherwise some port's fw is lying. Since we don't know which, we can assume
// the minimum.
// The ibdm algorithms (maybe topo match) rely on the ports on the link
// to have the same width.
// Also in case of different data, ibdiagnet will fail to construct lst file.
// Use this function whenever port width is needed.
// When you want to report an inconsistency use get_internal_width
IBLinkWidth IBPort::get_common_width() {
    IBLinkWidth w;

    if (p_remotePort)
        w = min(this->width, p_remotePort->width);
    else
        w = this->width;
    return w;
}

// Get the common speed on the link, i.e the minimum.
// In the normal fabric both ports on the link will have the same speed,
// otherwise some port's fw is lying. Since we don't know which, we can assume
// the minimum.
// The ibdm algorithms (maybe topo match) rely on the ports on the link
// to have the same speed.
// Also in case of different data, ibdiagnet will fail to construct lst file.
// Use this function whenever port width is needed.
// When you want to report an inconsistency use get_internal_speed
IBLinkSpeed IBPort::get_common_speed() {
    IBLinkSpeed sp;

    if (p_remotePort)
        sp = min(this->speed, p_remotePort->speed);
    else
        sp = this->speed;
    return sp;
}

// Get the port name.
// If connects to a system port - use the system port name.
string
IBPort::getName()
{
    string name;
    if (p_sysPort && (!p_node || p_node->numPorts < 80)) {
        // we currently do not support topo-write of submodules flavor (x4 of x2)
        name = p_sysPort->p_system->name + string("/") + p_sysPort->name;
    } else {
        if (!p_node) {
            cerr << "Got a port with no node" << endl;
            abort();
        }

        name = p_node->getName() + string("/P") + numAsString();
    }
    return name;
}

string IBPort::numAsString()
{
    char buff[8];

    if (p_node->numPorts >= SPLIT_PORTS_SW_NUM_PORTS &&
            num != SPLIT_PORTS_SW_NUM_PORTS + 1) { // port 81 is not splitted

        if (port_state == IB_PORT_STATE_DOWN || width == IB_LINK_WIDTH_2X) {
            bool odd = num % 2;
            sprintf(buff,"%u/%u",
                    (odd ? num/2 + 1 : num/2),
                    (odd ? 1 : 2));
        } else { // regular cable (not split)
            sprintf(buff,"%u", num/2 + 1);
        }

    } else {
        sprintf(buff,"%u", num);
    }

    return string(buff);
}

bool IBPort::isValid()
{
    if (p_node->numPorts >= SPLIT_PORTS_SW_NUM_PORTS) {
        if (num && num % 2 == 0) {
            IBPort *odd_port = p_node->getPort((phys_port_t)(num - 1));
            if (odd_port->port_state != IB_PORT_STATE_DOWN &&
                odd_port->width != IB_LINK_WIDTH_2X)
                return false;
        }
    }

    return true;
}

// connect the port to another node port
void
IBPort::connect(IBPort *p_otherPort)
{
    // we can not simply override existing connections
    if (p_remotePort) {
        // we only do care if not the requested remote previously conn.
        if (p_remotePort != p_otherPort) {
            cout << "-W- Disconnecting: "
                    << p_remotePort->getName() << " previously connected to:"
                    << p_remotePort->getName()
                    << " while connecting:" << p_otherPort->getName() << endl;
            // the other side should be cleaned only if points here
            if (p_remotePort->p_remotePort == this) {
                p_remotePort->p_remotePort = NULL;
            }
        }
    }
    p_remotePort = p_otherPort;

    // Check the other side was not previously connected
    if (p_otherPort->p_remotePort) {
        if (p_otherPort->p_remotePort != this) {
            // it was connected to a wrong port so disconnect
            cout << "-W- Disconnecting: " << p_otherPort->getName()
                    << " previously connected to:"
                    << p_otherPort->p_remotePort->getName()
                    << " while connecting:" << this->getName() << endl;
            // the other side should be cleaned only if points here
            if (p_otherPort->p_remotePort->p_remotePort == p_otherPort) {
                p_otherPort->p_remotePort->p_remotePort = NULL;
            }
        }
    }
    p_otherPort->p_remotePort = this;
}

// disconnect two ports
int
IBPort::disconnect(int duringSysPortDisconnect)
{
    if (!p_remotePort) {
        cout << "-W- Trying to disconenct non connected port." << endl;
        return(1);
    }

    if (p_remotePort->p_remotePort != this) {
        cout << "-W- Remote port does not point back! Disconnecting self only."
                << endl;
        p_remotePort = NULL;
        return(1);
    }
    IBPort *p_remPort = p_remotePort;
    p_remotePort->p_remotePort = NULL;
    p_remotePort = NULL;
    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
        cout << "-I- Disconnected port:" << getName()
            << " from:" << p_remPort->getName() << endl;

    // might need to treat the sys port too - but mark it duringPortDisconenct
    if (p_sysPort && ! duringSysPortDisconnect)
        return(p_sysPort->disconnect(1));

    return(0);
}

bool IBPort::isSpecialPort()
{
    return (this->special_port_type != NotSpecial);
}

// Set the type of special port if it is
void IBPort::setSpecialPortType(IBSpecialPortType special_port_type)
{
    this->special_port_type = special_port_type;
}

// Get the type of special port if it is
IBSpecialPortType IBPort::getSpecialPortType()
{
    return this->special_port_type;
}

//////////////////////////////////////////////////////////////////////////////
//
// CLASS IBNode:
//

// Constructor:
IBNode::IBNode(string n,
        IBFabric *p_fab,
        IBSystem *p_sys,
        IBNodeType t, phys_port_t np) :
    pLFTEnabled(false), maxPLFT(0),
    arGroupTop(0), arEnableBySLMask(0), arSubGrpsActive(0), arMaxGroupNumber(0),
    frEnabled(false), toIgnore(false), numVirtPorts(0)
{
    name            = n;
    p_fabric        = p_fab;
    p_system        = p_sys;
    type            = t;
    numPorts        = np;
    guid            = 0;
    sys_guid        = 0;
    rank            = IBNODE_UNASSIGNED_RANK;
    attributes      = string("");
    description     = string("");
    orig_description = string("");
    appData1.ptr    = NULL;
    appData2.ptr    = NULL;

    devId	= 0;
    revId	= 0;
    vendId	= 0;

    in_sub_fabric = true;

    for (unsigned int i = 0; i <= numPorts; i++)
        Ports.push_back((IBPort *)NULL);

    if (t == IB_SW_NODE) {
      Ports[0] = new IBPort(this, 0);
    }

    pLFTTop.resize(MAX_PLFT_NUM);
    LFT.resize(MAX_PLFT_NUM);
    arLFT.resize(MAX_PLFT_NUM);

    arState.resize(MAX_PLFT_NUM);


    if(t== IB_SW_NODE){
        for(int i = 0; i< MAX_PLFT_NUM; i++){
            arState[i].resize(100, AR_IB_LID_STATE_LAST);
        }
    }


    p_system->NodeByName[name]  = this;
    p_fabric->NodeByName[name]  = this;

    //set node index
    createIndex = (p_fab) ? p_fab->getNodeIndex() : 0;
}

// Delete the node cleaning up all it's connections
IBNode::~IBNode()
{
    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {
        cout << "-I- Destructing Node:" << name << endl;
    }
    // delete all the node ports:
    phys_port_t p;
    // use unsigned int index to prevent overwrap
    for (unsigned int i = 0; i <= numPorts; i++) {
        p = (phys_port_t)i;
        IBPort *p_port = this->getPort(p);
        if (p_port) {
            delete p_port;
        }
    }

    // remove from the system NodesByName:
    if (p_system) {
        map_str_pnode::iterator nI = p_system->NodeByName.find(name);
        if (nI != p_system->NodeByName.end()) {
            p_system->NodeByName.erase(nI);
        }
    }

    // remove from the fabric NodesByName:
    if (p_fabric) {
        map_str_pnode::iterator nI = p_fabric->NodeByName.find(name);
        if (nI != p_fabric->NodeByName.end()) {
            p_fabric->NodeByName.erase(nI);
        }

        nI = p_fabric->FullNodeByName.find(name);
        if (nI != p_fabric->FullNodeByName.end()) {
            p_fabric->FullNodeByName.erase(nI);
        }
    }

    // remove from the fabric NodesByDesc:
    if (p_fabric) {
        map_str_list_pnode::iterator nI = p_fabric->NodeByDesc.find(description);
        if (nI != p_fabric->NodeByDesc.end()) {
            list_pnode::iterator nI2 = nI->second.begin();
            while (nI2 != nI->second.end()) {
                if ((*nI2) == this) {
                    nI->second.erase(nI2);
                    break;
                }
                ++nI2;
            }
            if (nI->second.empty())
                p_fabric->NodeByDesc.erase(nI);
        }
    }
}

void
IBNode::guid_set(uint64_t g)
{
    if (p_fabric) {
        p_fabric->NodeByGuid[g] = this;
        guid = g;
    }
}

void
IBNode::system_guid_set(uint64_t g)
{
    if (p_fabric) {
        p_fabric->NodeBySystemGuid[g] = this;
        sys_guid = g;
    }
}

int
IBNode::getLidAndLMC(phys_port_t portNum, lid_t &lid, uint8_t &lmc)
{
  if (type == IB_SW_NODE) {
    lid = Ports[0]->base_lid;
    lmc = Ports[0]->lmc;
  } else {
    if ((portNum >= Ports.size()) || !Ports[portNum]) {
      return(1);
    }
    lid = Ports[portNum]->base_lid;
    lmc = Ports[portNum]->lmc;
  }
  return(0);
}

// Set the min hop for the given port (* is all) lid pair
void
IBNode::setHops(IBPort *p_port, lid_t lid, uint8_t hops)
{
    if (MinHopsTable.empty()) {
        if (lid > p_fabric->maxLid) {
            cout << "-W- We got a bigget lid:" << lid
                    << " then maxLid:" << p_fabric->maxLid << endl;
            p_fabric->maxLid = lid;
        }

        // we allocate the complete table upfront
        MinHopsTable.resize(p_fabric->maxLid + 1);
        for(lid_t l = 0; l < p_fabric->maxLid + 1; l++) {
	  MinHopsTable[l].resize(numPorts + 1, IB_HOP_UNASSIGNED);
        }
    }

    // now do the job
    // if we were not passed a port do it for all!
    // only if lid is legal , otherwise the user meant ,just to init the
    // table with IB_HOP_UNASSIGNED
    if (lid != 0) {
        if (p_port == NULL) {
            // set it for all ports
            for(unsigned int i = 0; i <= numPorts; i++)
                MinHopsTable[lid][i] = hops;
        } else {
            MinHopsTable[lid][p_port->num] = hops;
        }
    } else {
        for (unsigned int l = 0; l < MinHopsTable.size(); l++)
            for(unsigned int i = 0; i <= numPorts; i++)
                MinHopsTable[l][i] = hops;
    }

    // keep track about the min hops per node:
    if (MinHopsTable[lid][0] > hops)
        MinHopsTable[lid][0] = hops;
} // method setHops

// Report Min Hop Table of the current Node
void
IBNode::repHopTable()
{
    cout << "-I- MinHopTable for Node:" << name << "\n"
            << "=========================\n" <<endl;
    if (MinHopsTable.empty()) {
        cout << "\tEmpty" << endl;
    } else {
        cout << "  " << setw(3) << "MIN" << " ";
        // Lid/Port header line
        for (unsigned int i = 1; i <= numPorts; ++i)
            cout << setw(2) << i << " " ;
        cout << endl;
        for (unsigned int i = 1; i <= (unsigned int)(3 * numPorts + 5); ++i)
            cout << "-";
        cout << endl;
        for (lid_t l = 1; l <= p_fabric->maxLid; ++l) {
            cout << setw(2) << l << "|";
            for (unsigned int i = 0; i <= numPorts; ++i) {
                int val=(int)MinHopsTable[l][i];
                if (val != 255)
                    cout << setw(2) << val << " " ;
                else
                    cout << setw(2) << "-" << " " ;
            }
            IBPort *p_port = p_fabric->getPortByLid(l);
            if (p_port)
	      cout << " " << p_port->getName();
            cout << endl;
        }
        cout << endl;
    }
}   // Method repHopTable

// Get the min number of hops defined for the given port or all
uint8_t
IBNode::getHops(IBPort *p_port, lid_t lid)
{
    // make sure it is initialized:
    if (MinHopsTable.empty() || (MinHopsTable.size() < (unsigned)lid+1))
        return IB_HOP_UNASSIGNED;
    if (MinHopsTable[lid].empty()) return IB_HOP_UNASSIGNED;
    if (p_port == NULL) return MinHopsTable[lid][0];
    return MinHopsTable[lid][p_port->num];
}

// Scan the node ports and find the first port with min hop to the lid
IBPort *
IBNode::getFirstMinHopPort(lid_t lid)
{
    // make sure it is a SW:
    if (type != IB_SW_NODE) {
        cout << "-E- Get best hop port must be run on SW nodes!" << endl;
        return NULL;
    }

    if (MinHopsTable.empty() || (MinHopsTable.size() < (unsigned)lid+1))
        return NULL;

    // the best hop is stored in port 0:
    int minHop = MinHopsTable[lid][0];
    for (unsigned int i = 1; i <=  numPorts; i++)
        if (MinHopsTable[lid][i] == minHop)
            return getPort((phys_port_t)i);
    return NULL;
}

// Set the Linear Forwarding Table:
void
IBNode::setLFTPortForLid(lid_t lid, phys_port_t portNum, u_int8_t pLFT)
{
    if (pLFT >= MAX_PLFT_NUM) {
        cout << "-E- setLFTPortForLid: Given pLFT:" << (int) pLFT
            << " is too high!" << endl;
        return;
    }

    unsigned int origSize =
        (unsigned int)(LFT[pLFT].empty() ? 0 : LFT[pLFT].size());
    // make sure the vector is init
    if (origSize < (unsigned)(lid+1))
        LFT[pLFT].resize(lid + 100, IB_LFT_UNASSIGNED);

    // now do the job
    LFT[pLFT][lid] = portNum;

    /*
    if (pLFT ==0 && lid == 40082 && portNum != 255 && (origSize < (unsigned)lid+1) ) {
        printf(" lid:%d pLFT:%d port:%d\n",(int)lid, (int)pLFT,(int)LFT[pLFT][lid]);
    }
    */

}

// Get the LFT for a given lid
phys_port_t
IBNode::getLFTPortForLid(lid_t lid, u_int8_t pLFT) const
{
    if (pLFT >= MAX_PLFT_NUM) {
        cout << "-E- getLFTPortForLid: Given pLFT:" << (int) pLFT
            << " is too high!" << endl;
        return IB_LFT_UNASSIGNED;
    }

    // make sure it is initialized:
    if (LFT[pLFT].empty() || (LFT[pLFT].size() < (unsigned)lid+1))
        return IB_LFT_UNASSIGNED;
    return ( LFT[pLFT][lid] );
}

// Set the Switch AR State Table (for the given lid)
void IBNode::setARstateForLid(lid_t lid, SMP_AR_LID_STATE state,
        u_int8_t pLFT) {

    if (pLFT >= MAX_PLFT_NUM) {
        cout << "-E- setARstateForLid: Given pLFT:" << (int) pLFT
                << " is too high!" << endl;
        return;
    }
    unsigned int origSize = (unsigned int) arState[pLFT].size();

    // make sure the vector is init
    if (origSize < (unsigned) (lid + 1))
        arState[pLFT].resize(lid + 100, AR_IB_LID_STATE_LAST);

    // now do the job
    arState[pLFT][lid] = state;
}

 //get the AR state for the given lid
SMP_AR_LID_STATE IBNode::getARstateForLid(lid_t lid, u_int8_t pLFT) const {

    if (pLFT >= MAX_PLFT_NUM) {
        cout << "-E- getARstateForLid: Given pLFT:" << (int) pLFT
                << " is too high!" << endl;
        return AR_IB_LID_STATE_LAST;
    }

    // make sure it is initialized:
    if (arState[pLFT].empty() || (arState[pLFT].size() < (unsigned) lid + 1))
        return AR_IB_LID_STATE_LAST;
    return (arState[pLFT][lid]);
}


// Resize the Linear Forwarding Table (can be resize with linear top value):
void
IBNode::resizeLFT(uint16_t newSize, u_int8_t pLFT)
{
    if (pLFT >= MAX_PLFT_NUM) {
        cout << "-E- resizeLFT: Given pLFT:" << (int) pLFT
            << " is too high!" << endl;
        return;
    }

    if (newSize >= 0xc000) {
        cout << "-E- resizeLFT : Given newSize:" << newSize
                << " is too high!" << endl;
        return;
    }
    LFT[pLFT].resize(newSize, IB_LFT_UNASSIGNED);
}

// Set the PSL Table:
void
IBNode::setPSLForLid(lid_t lid, lid_t maxLid, uint8_t sl)
{
    if (PSL.empty()) {
        PSL.resize(maxLid + 1);
        for(unsigned int i = 0; i<PSL.size(); i++)
            PSL[i] = IB_SLT_UNASSIGNED;
    }
    PSL[lid] = sl;
    usePSL = true;
    setMaxSL(sl);
}

// Get the PSL Table:
uint8_t
IBNode::getPSLForLid(lid_t lid)
{
    if (PSL.empty())
        return (usePSL ? (uint8_t)IB_SLT_UNASSIGNED :
                (uint8_t)p_fabric->GetDefaultSL());
    if (PSL.size() < (unsigned)lid+1)
        return IB_SLT_UNASSIGNED;
    return PSL[lid];
}

/*
SL2VL ACT Values:
1 = VL2VLOnly: All lookups use InVL instead of PacketSL
2 = LowVL2VL: For devices where IsMixedSLVLSupported is 1 - Lower 8 SLs [0 to 7] are lookup by InVL and the others by PacketSL
3 = HighVL2VL: For devices where IsMixedSLVLSupported is 1 - Higher 8 SLs [8 to 15] are lookup by InVL and the others by PacketSL
*/
#define SL2VL_ACT_VL2VL_ONLY 1
#define SL2VL_ACT_LOW_VL2VL 2
#define SL2VL_ACT_HIGH_VL2VL 3

//set replaceSLsByInVL

void IBNode::setSL2VLAct(u_int8_t slvlAct)
{

    if (!replaceSLsByInVL.empty()) {
        replaceSLsByInVL.resize(0);
    }

    if (slvlAct == 0) {
        return;
    }

    replaceSLsByInVL.resize(IB_NUM_SL, true);

    if (slvlAct == SL2VL_ACT_VL2VL_ONLY)
        return;

    u_int8_t start_idx = 0;

    if (slvlAct == SL2VL_ACT_LOW_VL2VL)
        start_idx = 8;

    for (u_int8_t offset = 0; offset < 8 ; offset++)
        replaceSLsByInVL[start_idx+offset] = false;
}

void IBNode::setVL2VL(u_int8_t sl)
{
    if (replaceSLsByInVL.empty())
        replaceSLsByInVL.resize(IB_NUM_SL, true);

    replaceSLsByInVL[sl] = true;
}

void IBNode::getSL2VLCfg(char *line) const
{
    if (line == NULL)
        return;

    line[0] = '\0';
    if (replaceSLsByInVL.empty())
        return;

    stringstream sstream;
    for (u_int8_t sl = 0; sl < IB_NUM_SL ; ++sl) {
        if (replaceSLsByInVL[sl]) {
            sstream << (int)sl <<", ";
        }
    }
    int size = sprintf(line, "%s", sstream.str().c_str());
    if (size > 2)
        line[size-2] = '\0';
}

u_int8_t IBNode::getUsedSLOrVL(sl_vl_t slvl) const
{
    if (replaceSLsByInVL.empty() ||
        !replaceSLsByInVL[slvl.SL])
        return slvl.SL;

    return slvl.VL;
}

// Set the SL2VL Table:
void
IBNode::setSLVL(phys_port_t iport,phys_port_t oport,uint8_t sl, uint8_t vl)
{
    if (iport > numPorts || oport > numPorts || sl >= IB_NUM_SL ) {
        cout << "-E- setSLVL: invalid input parameter!"
            << " iport:" << (int) iport
            << " oport:" << (int) oport
            << " sl:" << (int) sl << endl;
        return;
    }

    // Create an empty table
    if (SLVL.empty()) {
        SLVL.resize(numPorts + 1);
        for (unsigned int i = 0; i < SLVL.size(); ++i) {
            SLVL[i].resize(numPorts + 1);
            for (unsigned int j = 0; j < SLVL[i].size(); ++j) {
                SLVL[i][j].resize(IB_NUM_SL);
                for (unsigned int k = 0; k < SLVL[i][j].size(); ++k)
                    SLVL[i][j][k] = IB_SLT_UNASSIGNED;
            }
        }
    }
    SLVL[iport][oport][sl] = vl;
    useSLVL = true;
}

// Get the SLVL Table:
uint8_t
IBNode::getVL(phys_port_t iport, phys_port_t oport, sl_vl_t slvl) const
{
    // Identity mapping
    if (SLVL.empty())
        return (useSLVL ? (uint8_t)IB_SLT_UNASSIGNED : 
                (uint8_t)(slvl.SL%p_fabric->numVLs));

    u_int8_t nSLVL = (iport == 0 ? slvl.SL : getUsedSLOrVL(slvl));
    if (iport > numPorts || oport > numPorts || nSLVL >= IB_NUM_SL ) {
        cout << "-E- getVL: invalid input parameter!"
            << " iport:" << (int) iport
            << " oport:" << (int) oport
            << " sl/vl:" << (int) nSLVL << endl;
        return IB_SLT_UNASSIGNED;
    }

    return SLVL[iport][oport][nSLVL];
}

int
IBNode::checkSL2VLTable() const
{
    if (SLVL.empty()){
        cout << "-E- Node " << name << "does not have SL2VL table.";
        return 1;
    }

    int numErrors = 0;
    phys_port_t maxInPort = 0;
    if ( type == IB_SW_NODE )
        maxInPort = numPorts;

    vector<bool> activePorts(numPorts+1,false);
    activePorts[0] = true;

    //init activePorts
    for (phys_port_t port = 1; port <=  numPorts; ++port) {
        if(Ports.size() <= port || Ports[port] == NULL ||
           Ports[port]->p_remotePort == NULL)
            continue;
        activePorts[port] = true;
    }

    for (phys_port_t iport = 0; iport <=  maxInPort; ++iport) {

        if(activePorts[iport] == false)
            continue;

        for (phys_port_t oport = 1; oport <=  numPorts; ++oport) {

            if(oport == iport)
                continue;

            if(activePorts[oport] == false)
                continue;

            for (u_int8_t  SL = 0; SL <  IB_NUM_SL; ++SL) {
                if (replaceSLsByInVL.empty() ||
                    !replaceSLsByInVL[SL]){
                    if (SLVL[iport][oport][SL] >= IB_DROP_VL) {
                        cout << "-E- Node " << name <<
                            " Invalid VL:" << (int)SLVL[iport][oport][SL] <<
                            " For iport:" << (int)iport <<
                            " oport:" << (int)oport <<
                            " SL:" << (int)SL << endl;

                        numErrors++;
                    }
                }
            }//SL
        }//oport
    }//iport

    return numErrors;
}

void
IBNode::getSLVL(phys_port_t iport, phys_port_t oport,
                             const sl_vl_t &iSLVL, sl_vl_t &oSLVL)
{
    oSLVL.SL = iSLVL.SL;
    oSLVL.VL = getVL(iport, oport, iSLVL);
}

//group in ports with same out-sl-vl values
//check only neighbors ports
phys_port_t IBNode::getSLVLPortGroup(phys_port_t port) {
    if (slvlPortsGroups.empty())
        buildSLVLPortsGroups();
    return slvlPortsGroups[port];
}

void IBNode::buildSLVLPortsGroups()
{
    if (!slvlPortsGroups.empty())
        return;

    slvlPortsGroups.resize(numPorts+1,0);

    if (SLVL.empty()) {
        return;
    }

    phys_port_t portGroup = 0;

    slvlPortsGroups[0] = portGroup;
    u_int8_t portGroupsNum = 1;

    for (phys_port_t iPort = 1; iPort <= numPorts; ++iPort) {
        bool equalSLVL = true;
        for (phys_port_t oPort = 0; oPort <= numPorts; ++oPort) {
            for (u_int8_t slvl = 0; slvl < IB_NUM_SL; ++slvl) {
                if (iPort == oPort)
                    continue;
                if (portGroup == oPort)
                    //for SLVL[portGroup][portGroup] use SLVL[portGroup+1][portGroup]
                    equalSLVL = (SLVL[iPort][oPort][slvl] == SLVL[portGroup+1][oPort][slvl]);
                else
                    equalSLVL = (SLVL[iPort][oPort][slvl] == SLVL[portGroup][oPort][slvl]);

                if (!equalSLVL) {
                    /*
                    phys_port_t compInPort =
                        (portGroup == oPort ? (phys_port_t)(portGroup+1) : portGroup);
                    printf("Node: %s SLVL[%d][%d][%d]  = %d != SLVL[%d][%d][%d] = %d \n",
                           name.c_str(),
                           (int)iPort, (int)oPort, (int)slvl, (int)SLVL[iPort][oPort][slvl],
                           (int)compInPort, (int)oPort, (int)slvl, (int)SLVL[compInPort][oPort][slvl]);
                           */
                    break;
                }
            }//slvl
            if(!equalSLVL)
                break;
        }//oPort
        if (!equalSLVL) {
            portGroup = iPort;
            portGroupsNum++;
        }
        slvlPortsGroups[iPort] = portGroup;
    }//iPort
    //printf("Node: %s portGroupsNum:%d\n",name.c_str(),portGroupsNum);
}

// Set the Multicast FDB table
void
IBNode::setMFTPortForMLid(
        lid_t lid,
        phys_port_t portNum)
{
    if ((portNum > numPorts) || (portNum >= 64)) {
        cout << "-E- setMFTPortForMLid : Given port:" << (int)portNum
                << " is too high!" << endl;
        return;
    }

    // make sure the mlid is in range:
    if (lid < 0xc000) {
        cout << "-E- setMFTPortForMLid : Given lid:" << lid
                << " is out of range" << endl;
        return;
    }

    int idx = lid - 0xc000;

    // make sure we have enough vector:
    int prevSize = (int)MFT.size();
    if (prevSize <= idx)
        MFT.resize(idx + 10);

    MFT[idx].set(portNum);

    // we track all Groups:
    p_fabric->mcGroups.insert(lid);
}

void
IBNode::setMFTPortForMLid(lid_t lid, uint16_t portMask, uint8_t portGroup)
{
    if (portGroup >= 16) {
        cout << "-E- setMFTPortForMLid : Given portGroup:" << (int)portGroup
                    << " is out of range [0,16)!" << endl;
        return;
    }

    // make sure the mlid is in range:
    if (lid < 0xc000) {
        cout << "-E- setMFTPortForMLid : Given lid:" << lid
                << " is out of range" << endl;
        return;
    }

    int idx = lid - 0xc000;

    // make sure we have enough vector:
    int prevSize = (int)MFT.size();
    if (prevSize <= idx)
        MFT.resize(idx + 10);

    uint64_t curr_mask[4];
    MFT[idx].get_data(curr_mask[0], curr_mask[1], curr_mask[2], curr_mask[3]);

    uint8_t port_mask_block = (uint8_t)(portGroup / 4);
    uint8_t port_mask_offset = (uint8_t)(portGroup % 4);

    curr_mask[port_mask_block] |=
        (((uint64_t)portMask) << (16 * port_mask_offset));

    MFT[idx].set(curr_mask[0], curr_mask[1], curr_mask[2], curr_mask[3]);

    // we track all Groups:
    p_fabric->mcGroups.insert(lid);
}

// Get the list of ports for the given MLID from the MFT
list_phys_ports
IBNode::getMFTPortsForMLid(lid_t lid)
{
    list_phys_ports res;
    // make sure the mlid is in range:
    if (lid < 0xc000) {
        cout << "-E- getMFTPortsForMLid : Given lid:" << lid
                << " is out of range" << endl;
        return res;
    }

    lid_t idx = (lid_t)(lid - 0xc000);    //we check it > 0xc000 above
    if (MFT.size() <= idx)
        return res;

    for(unsigned int pn = 0; pn <= numPorts; ++pn)
        if (MFT[idx].test(pn))
            res.push_back((phys_port_t)pn);

    return res;
}

// create a new port by name if required to
IBPort *IBNode::makePort(phys_port_t num)
{
  if ((type == IB_SW_NODE) && (num == 0)) {   //switch
    if (!Ports[0])
      Ports[0] = new IBPort(this, 0);
    Ports[0]->lmc = p_fabric->swLmc;
    return Ports[0];
  }
  if ((num < 1) || (num > numPorts)) {
    cout << "-E- Given port number out of range: 1 < " << (int)num
	 << " < " << numPorts << endl;
    return NULL;
  }
  if (!Ports[num]) {
    Ports[num] = new IBPort(this, num);
    if (type == IB_SW_NODE) {
      Ports[num]->lmc = p_fabric->swLmc;
    } else {
      Ports[num]->lmc = p_fabric->caLmc;
    }
  }
  return Ports[num];
}

void IBNode::setPLFTEnabled()
{
    if (pLFTEnabled)
        return;
    pLFTEnabled = true;

    portSLToPLFTMap.resize(numPorts+1);
    for (unsigned int port = 0; port <=numPorts; ++port )
        portSLToPLFTMap[port].resize(IB_NUM_SL);
}

u_int8_t IBNode::getPLFTMapping(phys_port_t port, sl_vl_t slvl) const
{
    if (!pLFTEnabled)
        return 0;

    if (slvl.SL == IB_SLT_UNASSIGNED ||
        slvl.VL == IB_SLT_UNASSIGNED)
        return 0;

    //for VL15 use SL and ignore vl2vl configuration
    u_int8_t nSLVL = (slvl.VL == 15 ? slvl.SL : getUsedSLOrVL(slvl));
    return portSLToPLFTMap[port][nSLVL];
}

void IBNode::getPLFTMapping(phys_port_t port, char *plft_line) const
{
    if (plft_line == NULL)
        return;

    if (!pLFTEnabled ||
        portSLToPLFTMap.size() <= port ||
        portSLToPLFTMap[port].size() < IB_NUM_SL)
        sprintf(plft_line, "No PLFT Mapping available for port:%u", port);
    else
        sprintf(plft_line, "%u, %u, %u, %u, %u, %u, %u, %u, "\
                     "%u, %u, %u, %u, %u, %u, %u, %u",
                portSLToPLFTMap[port][0],
                portSLToPLFTMap[port][1],
                portSLToPLFTMap[port][2],
                portSLToPLFTMap[port][3],
                portSLToPLFTMap[port][4],
                portSLToPLFTMap[port][5],
                portSLToPLFTMap[port][6],
                portSLToPLFTMap[port][7],
                portSLToPLFTMap[port][8],
                portSLToPLFTMap[port][9],
                portSLToPLFTMap[port][10],
                portSLToPLFTMap[port][11],
                portSLToPLFTMap[port][12],
                portSLToPLFTMap[port][13],
                portSLToPLFTMap[port][14],
                portSLToPLFTMap[port][15]);
}

//AR

bool IBNode::isARActive(sl_vl_t slvl) const {
    if (!isAREnable())
        return false;

    u_int8_t nSLVL = getUsedSLOrVL(slvl);
    return (frEnabled || (((u_int16_t)1<< nSLVL) & arEnableBySLMask));
}

void IBNode::getARActiveCfg(char *line) const
{
    if (line == NULL)
        return;

    line[0] = '\0';
    if (!isAREnable())
        return;

    stringstream sstream;
    for (u_int8_t sl = 0; sl < IB_NUM_SL ; ++sl) {
        if (((u_int16_t)1<< sl) & arEnableBySLMask) {
            sstream << (int)sl <<", ";
        }
    }

    int size = sprintf(line, "%s", sstream.str().c_str());
    if (size > 2)
        line[size-2] = '\0';
}

void IBNode::getARGroupCfg(u_int16_t groupNumber, char *line) const
{
    if (line == NULL)
        return;

    line[0] = '\0';
    if (!isAREnable() || arPortGroups.empty() ||
        groupNumber > arMaxGroupNumber)
        return;

    stringstream sstream;
    list_phys_ports portsList = arPortGroups[groupNumber];
    list_phys_ports::const_iterator iter = portsList.begin();
    for ( ; iter != portsList.end() ; ++iter)
            sstream <<(int) *iter <<", ";

    int size = sprintf(line, "%s", sstream.str().c_str());
    if (size > 2)
        line[size-2] = '\0';
}

#define AR_PORT_GROUP_RESIZE_BY 100
void IBNode::setARPortGroup(u_int16_t groupNum, list_phys_ports portsList)
{
    u_int16_t origSize =
        (u_int16_t)(arPortGroups.empty() ? 0 : arPortGroups.size());
    // make sure the vector is init
    if (origSize < groupNum+1)
        arPortGroups.resize(groupNum + AR_PORT_GROUP_RESIZE_BY);

    // now do the job
    arPortGroups[groupNum].insert(arPortGroups[groupNum].end(),
                                  portsList.begin(), portsList.end());

    arMaxGroupNumber = max(arMaxGroupNumber, groupNum);
}

void
IBNode::setARLFTPortGroupForLid(lid_t lid, u_int16_t portGroup, u_int8_t pLFT)
{
    unsigned int origSize =
        (unsigned int)(arLFT[pLFT].empty() ? 0 : arLFT[pLFT].size());
    // make sure the vector is init
    if (origSize < (unsigned)lid+1)
        arLFT[pLFT].resize(lid + 100, IB_AR_LFT_UNASSIGNED);

    if (arMaxGroupNumber<portGroup || arPortGroups.empty())
        portGroup = IB_AR_LFT_UNASSIGNED;

    // now do the job
    arLFT[pLFT][lid] = portGroup;
}

u_int16_t
IBNode::getARLFTPortGroupForLid(lid_t lid, u_int8_t pLFT) const
{
    // make sure it is initialized:
    if (arLFT[pLFT].empty() || (arLFT[pLFT].size() < (unsigned)lid+1))
        return IB_AR_LFT_UNASSIGNED;
    return ( arLFT[pLFT][lid] );
}

//if no route return empty list
void
IBNode::getLFTPortListForLid(phys_port_t staticOutPortNum,
                             u_int16_t portGroup, list_phys_ports &portsList) const
{
    if (portGroup != IB_AR_LFT_UNASSIGNED) {
        portsList = arPortGroups[portGroup];
        if(!portsList.empty())
            return;
    }else if(!portsList.empty())
        portsList.clear();

    if (staticOutPortNum != IB_LFT_UNASSIGNED)
        portsList.push_back(staticOutPortNum);
}

void
IBNode::getLFTPortListForLid(lid_t lid, u_int8_t pLFT,
                             bool useAR, list_phys_ports &portsList) const
{
    phys_port_t portNum = IB_LFT_UNASSIGNED;
    if (!useAR)
        portNum = getLFTPortForLid(lid, pLFT);
    else {
        if (!arLFT[pLFT].empty() && (arLFT[pLFT].size() > (unsigned)lid)){
            u_int16_t portGroup = arLFT[pLFT][lid];
            if (portGroup != IB_AR_LFT_UNASSIGNED) {
                portsList = arPortGroups[portGroup];
                if (!portsList.empty())
                    return;
            }
        }
        //No AR for this lid
        portNum = getLFTPortForLid(lid, pLFT);
    }

    if(!portsList.empty())
        portsList.clear();

    if (portNum != IB_LFT_UNASSIGNED)
        portsList.push_back(portNum);
}

void
IBNode::getLFTPortListForLid(lid_t lid, phys_port_t inPort,
                             sl_vl_t slvl, list_phys_ports &portsList) const
{
    u_int8_t pLFT = getPLFTMapping(inPort, slvl);
    bool useAR = isARActive(slvl);
    getLFTPortListForLid(lid, pLFT, useAR, portsList);
}

// Resize the Linear Forwarding Table (can be resize with linear top value):
void
IBNode::resizeARLFT(uint16_t newSize, u_int8_t pLFT)
{
    if (newSize >= 0xc000) {
        cout << "-E- resizeARLFT : Given newSize:" << newSize
                << " is too high!" << endl;
        return;
    }
    arLFT[pLFT].resize(newSize, IB_AR_LFT_UNASSIGNED);
}

void IBNode::resizeARstate(uint16_t newSize, u_int8_t pLFT) {
    if (newSize >= 0xc000) {
        cout << "-E- resizeARLFT : Given newSize:" << newSize << " is too high!"
                << endl;
        return;
    }

    arState[pLFT].resize(newSize, AR_IB_LID_STATE_LAST);

}

void IBNode::setARSubGrp(u_int16_t group, u_int16_t subGroup, const list_phys_ports& portsList){

    if(arGroups.find(group) == arGroups.end()){
        arGroups[group] = ARgrp();
    }
    arGroups[group][subGroup] = portsList;

    //update ARPortGroup to include the subgroup.
    setARPortGroup(group, portsList);

}

//////////////////////////////////////////////////////////////////////////////
//
// CLASS IBSysPort:
//

// Constructor:
IBSysPort::IBSysPort(string n, class IBSystem *p_sys)
{
    p_system                    = p_sys;
    name                        = n;
    p_nodePort                  = NULL;
    p_remoteSysPort             = NULL;
    p_system->PortByName[name]  = this;
}

// Destructor
IBSysPort::~IBSysPort() {
    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {
        cout << "-I- Destructing SysPort:" << name << endl;
    }

    // if connected to another side remove the connection
    if (p_remoteSysPort)
        p_remoteSysPort->p_remoteSysPort = NULL;

    // remove from the map of the parent system
    if (p_system) {
        map_str_psysport::iterator pI = p_system->PortByName.find(name);
        if (pI != p_system->PortByName.end()) {
            p_system->PortByName.erase(pI);
        }
    }
}

// Connect two system ports. This will update both sides pointers
void
IBSysPort::connectPorts(IBSysPort *p_otherSysPort)
{
    // we can not simply override existing connections
    if (p_remoteSysPort) {
        // we only do care if not the requested remote previously conn.
        if (p_remoteSysPort != p_otherSysPort) {
            cout << "-W- Disconnecting system port: " << p_system->name << "-/"
                    << this->name << " previously connected to:"
                    << p_remoteSysPort->p_system->name << "-/"
                    << p_remoteSysPort->name
                    << " while connecting:" << p_otherSysPort->p_system->name
                    << "-/" << p_otherSysPort->name << endl;
            // the other side should be cleaned only if points here
            if (p_remoteSysPort->p_remoteSysPort == this) {
                p_remoteSysPort->p_remoteSysPort = NULL;
            }
        }
    }
    p_remoteSysPort = p_otherSysPort;

    // Check the other side was not previously connected
    if (p_otherSysPort->p_remoteSysPort) {
        if (p_otherSysPort->p_remoteSysPort != this) {
            // it was connected to a wrong port so disconnect
            cout << "-W- Disconnecting system port back: "
                    << p_otherSysPort->p_system->name << "-/"
                    << p_otherSysPort->name << " previously connected to:"
                    << p_otherSysPort->p_remoteSysPort->p_system->name << "-/"
                    << p_otherSysPort->p_remoteSysPort->name
                    << " while connecting:" << this->p_system->name
                    << "-/" << this->name << endl;
            // the other side should be cleaned only if points here
            if (p_otherSysPort->p_remoteSysPort->p_remoteSysPort == p_otherSysPort) {
                p_otherSysPort->p_remoteSysPort->p_remoteSysPort = NULL;
            }
        }
    }
    p_otherSysPort->p_remoteSysPort = this;
}

// Connect two ports. This will update both sides pointers of system ports and nodes ports
void
IBSysPort::connect (IBSysPort *p_otherSysPort,
                    IBLinkWidth width,
                    IBLinkSpeed speed,
                    IBPortState state)
{

    connectPorts(p_otherSysPort);
    // there should be a lower level port to connect too:
    if (p_remoteSysPort->p_nodePort && p_nodePort) {
        p_remoteSysPort->p_nodePort->set_internal_data(speed, width, IB_PORT_STATE_ACTIVE);
        p_nodePort->set_internal_data(speed, width, IB_PORT_STATE_ACTIVE);
        p_nodePort->connect(p_remoteSysPort->p_nodePort);
    } else
        cout << "-E- Connected sys ports but no nodes ports:"
            << p_system->name << "/" << name << " - "
            << p_remoteSysPort->p_system->name << "/"
            << p_remoteSysPort->name << endl;
} // Method connect

int
IBSysPort::disconnect(int duringPortDisconnect)
{
    if (!p_remoteSysPort) {
        cout << "-W- Trying to disconenct non connected system port." << endl;
        return(1);
    }

    if (p_remoteSysPort->p_remoteSysPort != this) {
        cout << "-W- Remote port does not point back! Disconnecting self only."
                << endl;
        p_remoteSysPort = NULL;
        return(1);
    }

    IBSysPort *p_remSysPort = p_remoteSysPort;
    p_remoteSysPort->p_remoteSysPort = NULL;
    p_remoteSysPort = NULL;
    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
        cout << "-I- Disconnected system port:" << name
            << " from:" << p_remSysPort->name << endl;

    // might need to treat the port too - but mark it duringPortDisconenct
    if (p_nodePort && ! duringPortDisconnect) return(p_nodePort->disconnect(1));
    return(0);
}


//////////////////////////////////////////////////////////////////////////////
//
// CLASS IBSystem:
//

// constructor:
IBSystem::IBSystem(const string &n, class IBFabric *p_fab, const string &t,
                   bool mlx_nd_format)
{
    if (p_fab->getSystem(n)) {
        cerr << "Can't deal with double allocation of same system!" << endl;
        abort();
    }
    name                        = n;
    type                        = t;
    p_fabric                    = p_fab;
    p_fabric->SystemByName[n]   = this;
    newDef                      = false;
    sys_mlx_nd_format           = mlx_nd_format;
    max_mlx4                    = MLX_NA;
    max_mlx5                    = MLX_NA;
}

// Destructor
IBSystem::~IBSystem()
{
    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
        cout << "-I- Destructing System:" << name << endl;

    // cleanup all allocated IBNodes:
    while (!NodeByName.empty()) {
        map_str_pnode::iterator pI = NodeByName.begin();
        // deleting a IBNode should cleanup the table
        IBNode *p_node = (*pI).second;
        if (p_node) {
            delete p_node;
        }
    }

    // cleanup all allocated sysPorts:
    // (should be empty after cleanup all allocated IBNodes)
    while (!PortByName.empty()) {
        map_str_psysport::iterator pI = PortByName.begin();
        // deleting a SysPort should cleanup the table
        IBSysPort *p_sysPort = (*pI).second;
        if (p_sysPort) {
            delete p_sysPort;
        }
    }

    // cleanup from parent fabric table of systems:
    if (p_fabric) {
        map_str_psys::iterator sI = p_fabric->SystemByName.find(name);
        if (sI != p_fabric->SystemByName.end())
            p_fabric->SystemByName.erase(sI);
    }
}

// make sure we got the port defined
IBSysPort *
IBSystem::makeSysPort(string pName)
{
    IBSysPort *p_port;
    map_str_psysport::iterator pI = PortByName.find(pName);
    if (pI == PortByName.end()) {
        p_port = new IBSysPort(pName, this);
        if (!p_port) return NULL;
        PortByName[pName] = p_port;
    } else {
        p_port = (*pI).second;
    }
    // connect the SysPort to the lower level nodes
    IBPort *p_nodePort = getSysPortNodePortByName(pName);
    if (! p_nodePort) return NULL;
    p_nodePort->p_sysPort = p_port;
    p_port->p_nodePort = p_nodePort;
    return p_port;
}

IBSysPort *
IBSystem::getSysPort(string name)
{
    IBSysPort *p_sysPort = NULL;
    map_str_psysport::iterator nI = PortByName.find(name);
    if (nI != PortByName.end()) {
        p_sysPort = (*nI).second;
    }
    return p_sysPort;
}

// Get a string with all the System Port Names (even if not connected)
list_str
IBSystem::getAllSysPortNames()
{
    list_str portNames;
    for (map_str_psysport::iterator pI = PortByName.begin();
            pI != PortByName.end();
            pI++) {
        portNames.push_back((*pI).first);
    }
    return portNames;
}

// get the node port for the given sys port by name
IBPort *
IBSystem::getSysPortNodePortByName(string sysPortName)
{
    map_str_psysport::iterator pI = PortByName.find(sysPortName);
    if (pI != PortByName.end()) {
        return ((*pI).second)->p_nodePort;
    }
    return NULL;
}

// Split the given cfg into a vector of board cfg codes
void
IBSystem::cfg2Vector(const string& cfg,
                    vector<string>& boardCfgs,
                    int numBoards)
{
    unsigned int i;
    int b = 0;
    unsigned int prevDelim = 0;
    const char *p_str = cfg.c_str();
    char bcfg[16];

    // skip leading spaces:
    for (i = 0;
            (i < strlen(p_str)) && ((p_str[i] == '\t') ||
                    (p_str[i] == ' '));
            i++);
    prevDelim = i;

    // scan each character:
    for (; (i < strlen(p_str)) && (b < numBoards); i++) {
        // either a delimiter or not:
        if (p_str[i] == ',') {
            strncpy(bcfg, p_str + prevDelim, i - prevDelim);
            bcfg[i - prevDelim] = '\0';
            boardCfgs.push_back(string(bcfg));
            prevDelim = i + 1;
            b++;
        }
    }

    if (prevDelim != i) {
        strncpy(bcfg, p_str + prevDelim, i - prevDelim);
        bcfg[i - prevDelim] = '\0';
        boardCfgs.push_back(string(bcfg));
        b++;
    }

    for (;b < numBoards; b++) {
        boardCfgs.push_back("");
    }
}

// Remove a system board
int
IBSystem::removeBoard(string boardName)
{
    list <IBNode *> matchedNodes;
    // we assume system name is followed by "/" by board to get the node name:
    string sysNodePrefix = name + string("/") + boardName + string("/");

    // go through all the system nodes.
    for (map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
            nI != p_fabric->NodeByName.end();
            nI++) {
        // if node name start matches the module name - we need to remove it.
        if (!strncmp((*nI).first.c_str(),
                sysNodePrefix.c_str(),
                strlen(sysNodePrefix.c_str()))) {
            matchedNodes.push_back((*nI).second);
        }
    }

    // Warn if no match:
    if (matchedNodes.empty()) {
        cout << "-W- removeBoard : Fail to find any node in:"
                << sysNodePrefix << " while removing:" << boardName << endl;
        return 1;
    }

    // go through the list of nodes and delete them
    list <IBNode *>::iterator lI = matchedNodes.begin();
    while (lI != matchedNodes.end()) {
        // cleanup the node from the fabric node by name:
        IBNode *p_node = *lI;
        // cout << "Removing node:" << p_node->name.c_str()  << endl;
        p_fabric->NodeByName.erase(p_node->name);
        delete p_node;
        matchedNodes.erase(lI);
        lI = matchedNodes.begin();
    }

    return 0;
}

void IBSystem::generateSysPortName(char *buf, IBNode *p_node, unsigned int  pn)
{
    /*
    // Open, it if you want the port name node node system to be 'P<port number>'
    // this syntax is similar to standard ibnls given in the shard directory (e.g ibnl for SX6036)
    if (this->NodeByName.size() == 1) {
        sprintf(buf,"P%u", pn);
        return;
     }
    */
    string nameWithoutSysName =
            p_node->name.substr(this->name.length()+1,
                                p_node->name.length() - this->name.length() - 1);

    sprintf(buf, "%s/P%d", nameWithoutSysName.c_str(), pn);
}

// this function is called for system that runs ndd daemon, that changes
// node description to be 'host name card nme' (ex: host mlx4_0).
// it fills in the missing nodes up to max index of cards, in order
// to dump to ibnl file a full system.
// so if we have a system with mlx4_1 card only, it will add mlx4_0 too.
int
IBSystem::CreateMissingSystemMlxNodes(map_str_pnode *pIBNLFabricNodes)
{
    string  cur_mlx = "mlx4_";
    if (this->max_mlx4 != MLX_NA) {
        for (int i = 0; i <= this->max_mlx4; i++) {
            stringstream cur_idx;
            cur_idx << i;
            map_str_pnode::iterator nI = NodeByName.find(name + "/" + cur_mlx + cur_idx.str());
            if (nI == NodeByName.end()) {
                pIBNLFabricNodes->insert(pair<string, IBNode*>(string(cur_mlx + cur_idx.str()), NULL));
            } else {
                (*pIBNLFabricNodes)[(*nI).first] = (*nI).second;
            }
        }
    }
    cur_mlx = "mlx5_";
    if (this->max_mlx5 != MLX_NA) {
        for (int i = 0; i <= this->max_mlx5; i++) {
            stringstream cur_idx;
            cur_idx << i;
            map_str_pnode::iterator nI = NodeByName.find(name + "/" + cur_mlx + cur_idx.str());
            if (nI == NodeByName.end()) {
                pIBNLFabricNodes->insert(pair<string, IBNode*>(string(cur_mlx + cur_idx.str()), NULL));
            } else {
                (*pIBNLFabricNodes)[(*nI).first] = (*nI).second;
            }
        }
    }
    return 0;
}


// Write out the system IBNL into the given directory
// and return the new system type
//
// We are facing here a "heuristic" approach for how one knows
// the number of system ports - since some may be added later.
//
// In the case of a single device system we can and should expose
// all device ports - simply as P<pn>.
//
// In the case of a combined system we can not tell. Adding new connections
// will require regenerating the IBNL (which is by SysGuid anyway)
int
IBSystem::dumpIBNL(const char *ibnlDir, string &sysType)
{
    sysType = this->type;
    string fileName = string(ibnlDir) + string("/") + sysType + string(".ibnl");
    ofstream ibnl;
    string err_message;
    int rc = IBFabric::OpenFile(fileName.c_str(), ibnl, false, err_message);

    if (rc) {
        cout << "-E- Failed to write IBNL into file: " << fileName << endl;
        return rc;
    }
    /*
    ibnl.open(fileName.c_str());
    if (ibnl.fail()) {
        cout << "-E- Failed to write IBNL into file: " << fileName << endl;
        return 1;
    }
    */

    map_str_pnode IBNLFabricNodes;
    if (this->sys_mlx_nd_format)
        CreateMissingSystemMlxNodes(&IBNLFabricNodes);
    else
        IBNLFabricNodes = NodeByName;

    ibnl << "TOPSYSTEM " << sysType << endl;

    // go over all nodes of the system:
    for (map_str_pnode::iterator nI = IBNLFabricNodes.begin();
            nI != IBNLFabricNodes.end(); nI++) {
        IBNode *p_node = (*nI).second;

        string nameWithoutSysName;

        if (p_node) {
            nameWithoutSysName = p_node->name.substr(name.length()+1,
                                    p_node->name.length() - name.length() - 1);
            if (p_node->type == IB_SW_NODE) {
                ibnl << "\nNODE SW " << (unsigned int)p_node->numPorts << " "
                    << "DEV" << p_node->devId << " " << nameWithoutSysName.c_str() << endl;
            } else if (p_node->type == IB_CA_NODE) {
                ibnl << "\nNODE CA " << (unsigned int)p_node->numPorts << " "
                    << "DEV" << p_node->devId << " " << nameWithoutSysName.c_str() << endl;
            } else if (p_node->type == IB_RTR_NODE) {
                ibnl << "\nNODE RTR " << (unsigned int)p_node->numPorts << " "
                    << "DEV" << p_node->devId << " " << nameWithoutSysName.c_str() << endl;
            }

            for (unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
                IBPort *p_port = p_node->getPort((phys_port_t)pn);
                if (! p_port) {
                    // invent a port ...
                    char buf[128];
                    generateSysPortName(buf, p_node, pn);
                    ibnl << "   " << pn << " -> " << buf << endl;
                    continue;
                }
                if (p_port->p_sysPort) {
                    ibnl << "   " << pn << " -> "
                        << p_port->p_sysPort->name << endl;
                } else if (p_port->p_remotePort) {
                    IBNode *p_remoteNode = p_port->p_remotePort->p_node;
                    string remoteNodeName = p_remoteNode->name.substr(
                            name.length()+1,
                            p_remoteNode->name.length() - name.length() - 1);

                    ibnl << "   "  << pn << " -"
                        << width2char(p_port->get_common_width())
                        << "-" << speed2char(p_port->get_common_speed()) << "G-> "
                        << remoteNodeName << " "
                        << (unsigned int)p_port->p_remotePort->num << endl;
                }
            }
        } else { //this is a 'missing' node we added for full dump to ibnl
            nameWithoutSysName = (*nI).first;
            //TBD change dev id according to mlx4_ or mlx5,
            //currently hardcoded CX4 (mlx5_)
            ibnl << "\nNODE CA 2 DEV4115 " << nameWithoutSysName.c_str() << endl;
            for (unsigned int pn = 1; pn <= 2; pn++)
                ibnl << "   " << pn << " -> " << nameWithoutSysName << "/P"<< pn <<endl;
        }
    }

    ibnl.close();
    return 0;
}


//////////////////////////////////////////////////////////////////////////////
//
// CLASS IBFabric:
//

// Destructor:
IBFabric::~IBFabric()
{
    /* cleanup all Systems and Nodes:
     * starting with nodes since they point back to their systems.
     * Make sure to delete FullNodeByName, if it exists as a result
     * of ApplySubCluster, since NodeByName doesn't hold then
     * all the allocated nodes.
     */
    map_str_pnode *pNodeByName = &NodeByName;

    if (!FullNodeByName.empty())
        pNodeByName = &FullNodeByName;

    while (!pNodeByName->empty()) {
        map_str_pnode::iterator nI = pNodeByName->begin();
        IBNode *p_node = (*nI).second;
        delete p_node;
    }

    map_guid_pvnode *pVNodeByName = &VNodeByGuid;
    while (!pVNodeByName->empty()) {
        map_guid_pvnode::iterator nI = pVNodeByName->begin();
        IBVNode *p_vnode = (*nI).second;
        if (p_vnode)
            delete p_vnode;
    }

    // now we do the systems
    while (!SystemByName.empty()) {
        map_str_psys::iterator sI = SystemByName.begin();
        // note this will cleanup the system from the table...
        IBSystem *p_sys = (*sI).second;
        delete p_sys;
    }
}

// make a new node if can not find it by name
IBNode *
IBFabric::makeNode(string n, IBSystem *p_sys,
                   IBNodeType type, phys_port_t numPorts)
{
    IBNode *p_node;
    map_str_pnode::iterator nI = NodeByName.find(n);
    if (nI == NodeByName.end()) {
        if (numPorts > IB_MAX_PHYS_NUM_PORTS) {
            cout << "-E- Node " << n << " has bad number of ports " << (unsigned int)numPorts << endl;
            return NULL;
        }
        p_node = new IBNode(n, this, p_sys, type, numPorts);
        if (numPorts > maxNodePorts)
            maxNodePorts = numPorts;
        NodeByName[n] = p_node;
        // if the node is uniq by name in the fabric it must be uniq in the sys
        p_sys->NodeByName[n] = p_node;
    } else {
        p_node = (*nI).second;
    }

    if (type == IB_SW_NODE)
      p_node->makePort(0);

    // if the fabric require all ports to be declared do it:
    if (defAllPorts)
        for (unsigned int i = 1; i <= numPorts; i++)
            p_node->makePort((phys_port_t)i);

    return p_node;
}

// Look for the node by its name
IBNode *
IBFabric::getNode(string name)
{
    IBNode *p_node = NULL;
    /*for (map_str_pnode::iterator nI = NodeByName.begin();
            nI != NodeByName.end(); ++nI)
        printf("\n%s\n", (*nI).second->name.c_str());*/
    map_str_pnode::iterator nI = NodeByName.find(name);
    if (nI != NodeByName.end()) {
        p_node = (*nI).second;
    }
    return p_node;
}

IBPort *
IBFabric::getPortByGuid(uint64_t guid)
{
    IBPort *p_port = NULL;
    map_guid_pport::iterator nI = PortByGuid.find(guid);
    if (nI != PortByGuid.end()) {
        p_port = (*nI).second;
    }
    return p_port;
}

IBPort *
IBFabric::getPortByGuid(uint64_t guid, bool get_vguid, bool get_aguid)
{
    IBPort *p_port = NULL;
    map_guid_pport::iterator nI = PortByGuid.find(guid);
    if (nI != PortByGuid.end()) {
        p_port = (*nI).second;
    } else {
        if (get_vguid) {
            map_guid_pvport::iterator vpI = VPortByGuid.find(guid);
            if (vpI != VPortByGuid.end()) {
                IBVPort *p_vport = (*vpI).second;
                if (p_vport)
                    p_port = p_vport->getIBPortPtr();
            }
        }
        if (p_port == NULL && get_aguid) {
            map_guid_pport::iterator agpI = PortByAGuid.find(guid);
            if (agpI != PortByAGuid.end())
                p_port = (*agpI).second;
        }
    }
    return p_port;
}

IBNode *
IBFabric::getNodeByGuid(uint64_t guid)
{
    IBNode *p_node = NULL;
    map_guid_pnode::iterator nI = NodeByGuid.find(guid);
    if (nI != NodeByGuid.end()) {
        p_node = (*nI).second;
    }
    return p_node;
}

IBVPort *
IBFabric::getVPortByGuid(uint64_t guid)
{
    IBVPort *p_vport = NULL;
    map_guid_pvport::iterator nI = VPortByGuid.find(guid);
    if (nI != VPortByGuid.end()) {
        p_vport = (*nI).second;
    }
    return p_vport;
}

// return the list of node pointers matching the required type
list_pnode *
IBFabric::getNodesByType(IBNodeType type)
{
    list_pnode *res = new list_pnode;
    for (map_str_pnode::iterator nI = NodeByName.begin(); nI != NodeByName.end(); nI++) {
        if ((type == IB_UNKNOWN_NODE_TYPE) || (type == ((*nI).second)->type)) {
            res->push_back(((*nI).second));
        }
    }
    return res;
}

// convert the given configuration string to modifiers list
// The syntax of the modifier string is comma sep board=modifier pairs
static int
cfgStrToModifiers(string cfg, map_str_str &mods)
{
    unsigned int i;
    unsigned int prevDelim = 0;
    const char *p_str = cfg.c_str();
    char bcfg[64];

    // skip leading spaces:
    for (i = 0; (i < strlen(p_str)) &&
                ((p_str[i] == '\t') || (p_str[i] == ' ')); i++);
    prevDelim = i;

    // scan each character:
    for (;i < strlen(p_str); i++) {
        // either a delimiter or not:
        if (p_str[i] == ',') {
            strncpy(bcfg, p_str + prevDelim, i - prevDelim);
            bcfg[i - prevDelim] = '\0';
            char *eqSign = strchr(bcfg, '=');
            if (eqSign) {
                eqSign[0] = '\0';
                string key = bcfg;
                string val = ++eqSign;
                mods[key] = val;
            } else {
                cout << "-E- Bad modifier syntax:" << bcfg
                        << "expected: board=modifier" << endl;
            }
            prevDelim = i + 1;
        }
    }

    if (prevDelim != i) {
        strncpy(bcfg, p_str + prevDelim, i - prevDelim);
        bcfg[i - prevDelim] = '\0';
        char *eqSign = strchr(bcfg, '=');
        if (eqSign) {
            eqSign[0] = '\0';
            string key = bcfg;
            string val = ++eqSign;
            mods[key] = val;
        } else {
            cout << "-E- Bad modifier syntax:" << bcfg
                    << "expected: board=modifier" << endl;
        }
    }
    return(0);
}

// crate a new generic system - basically an empty container for nodes...
IBSystem *
IBFabric::makeGenericSystem(const string &name, const string &sysType,
                            bool mlx_nd_format)
{
    IBSystem *p_sys;

    // make sure we do not previoulsy have this system defined.
    map_str_psys::iterator sI = SystemByName.find(name);
    if (sI == SystemByName.end()) {
        p_sys = new IBSystem(name,this,sysType, mlx_nd_format);
    } else {
        p_sys = (*sI).second;
    }

    return p_sys;
}

// crate a new system - the type must have a predefined sysdef
IBSystem *
IBFabric::makeSystem(string name, string type, string cfg)
{
    IBSystem *p_sys;

    // make sure we do not previoulsy have this system defined.
    map_str_psys::iterator sI = SystemByName.find(name);
    if (sI == SystemByName.end()) {
        // We base our system building on the system definitions:
        map_str_str mods;

        // convert the given configuration string to modifiers map
        cfgStrToModifiers(cfg, mods);

        p_sys = theSysDefsCollection()->makeSystem(this, name, type, mods);
        if (!p_sys) {
            cout << "-E- Fail to find System class:" << type
                    << endl;
            return NULL;
        }

        SystemByName[name] = p_sys;

        // if the fabric require all ports to be declared do it:
        if (defAllPorts) {
            list_str portNames = p_sys->getAllSysPortNames();
            for (list_str::const_iterator pnI = portNames.begin();
                    pnI != portNames.end();
                    pnI++) {
                p_sys->makeSysPort(*pnI);
            }
        }
    } else {
        p_sys = (*sI).second;
    }
    return p_sys;
}

// Look for the system by its name
IBSystem *
IBFabric::getSystem(string name)
{
    IBSystem *p_system = NULL;
    map_str_psys::iterator nI = SystemByName.find(name);
    if (nI != SystemByName.end()) {
        p_system = (*nI).second;
    }
    return p_system;
}

// Add a cable connection
int
IBFabric::addCable(string t1, string n1, string p1,
            string t2, string n2, string p2,
            IBLinkWidth width, IBLinkSpeed speed)
{
    // make sure the nodes exists:
    IBSystem *p_sys1 = makeSystem(n1,t1);
    IBSystem *p_sys2 = makeSystem(n2,t2);

    // check please:
    if (! (p_sys1 && p_sys2)) {
        cout << "-E- Fail to make either systems:" << n1 << " or:"
                << n2 << endl;
        return 1;
    }

    // check types
    if (p_sys1->type != t1) {
        cout << "-W- Provided System1 Type:" << t1
                << " does not match pre-existing system:" << n1
                << " type:" << p_sys1->type << endl;
    }

    if (p_sys2->type != t2) {
        cout << "-W- Provided System1 Type:" << t2
                << " does not match pre-existing system:" << n2
                << " type:" << p_sys2->type << endl;
    }

    // make sure the sys ports exists
    IBSysPort *p_port1 = p_sys1->makeSysPort(p1);
    IBSysPort *p_port2 = p_sys2->makeSysPort(p2);
    if (! (p_port1 && p_port2)) {
        if (p_port1 == NULL)
            cout << "-E- Fail to make port: " << p1
            << " in system: " << n1 << " of type: " << t1 << endl;

        if (p_port2 == NULL)
            cout << "-E- Fail to make port: " << p2
            << " in system: " << n2 << " of type: " << t2 << endl;

        return 1;
    }

    // make sure they are not previously connected otherwise
    if (p_port1->p_remoteSysPort && (p_port1->p_remoteSysPort != p_port2)) {
        cout << "-E- Port:"
                << p_port1->p_system->name << "/"
                << p_port1->name
                << " already connected to:"
                << p_port1->p_remoteSysPort->p_system->name << "/"
                << p_port1->p_remoteSysPort->name <<endl;
        return 1;
    }

    if (p_port2->p_remoteSysPort && (p_port2->p_remoteSysPort != p_port1)) {
        cout << "-E- Port:"
                << p_port2->p_system->name << "/"
                << p_port2->name
                << " already connected to:"
                << p_port2->p_remoteSysPort->p_system->name << "/"
                << p_port2->p_remoteSysPort->name << endl;
        return 1;
    }

    // connect them
    p_port1->connect(p_port2, width, speed);
    p_port2->connect(p_port1, width, speed);
    return 0;
}

// Parse the cabling definition file
int
IBFabric::parseCables(const string &fn)
{
    ifstream f(fn.c_str());
    char sLine[1024];
    string n1, t1, p1, n2, t2, p2;
    regExp cablingLine("[ \t]*([^ \t]+)[ \t]+([^ \t]+)[ \t]+([^ \t]+)[ \t]+([^ \t]+)[ \t]+([^ \t]+)[ \t]+([^ \t]+)[ \t]*");
    regExp ignoreLine("^[ \t]*(#|[ \t]*$)");
    rexMatch *p_rexRes;
    IBLinkSpeed speed = IB_UNKNOWN_LINK_SPEED;
    IBLinkWidth width = IB_UNKNOWN_LINK_WIDTH;

    if (! f) {
        cout << "-E- Fail to open file:" << fn.c_str() << endl;
        return 1;
    }

    cout << "-I- Parsing cabling definition:" << fn.c_str() << endl;

    int numCables = 0;
    int lineNum = 0;
    while (f.good()) {
        lineNum++;
        f.getline(sLine,1024);
        // <SysType1> <sysName1> <portName1> <SysType2> <sysName2> <portName2>
        p_rexRes = cablingLine.apply(sLine);

        if (p_rexRes) {
            t1 = p_rexRes->field(1);
            n1 = p_rexRes->field(2);
            p1 = p_rexRes->field(3);
            t2 = p_rexRes->field(4);
            n2 = p_rexRes->field(5);
            p2 = p_rexRes->field(6);
            if (addCable(t1, n1, p1, t2, n2, p2, width, speed)) {
                cout << "-E- Fail to make cable"
                        << " (line:" << lineNum << ")"
                        << endl;
                delete p_rexRes;
                return 1;
            }
            numCables++;
            delete p_rexRes;
            continue;
        }

        // check if leagel ignored line
        p_rexRes = ignoreLine.apply(sLine);
        if (p_rexRes) {
            delete p_rexRes;
        } else {
            cout << "-E- Bad syntax on line:" << sLine << endl;
        }
    }

    cout << "-I- Defined " << SystemByName.size() << "/" << NodeByName.size()
            << " systems/nodes " << endl;
    f.close();
    return 0;
}

// Parse the topology definition file
int IBFabric::parseTopology(const string &fn)
{
    // get the extension of the file name
    size_t pos = fn.rfind(".");
    string ext;

    // If file extension is "lst" or "ibnd" try to parse the file as
    // subnet-lst file or ibnetdiscover file. otherwise try to parse the
    // file as topo file.
    if (pos != string::npos) {
        ext = fn.substr(pos + 1);
        if (!ext.compare("lst")) {
            if (parseSubnetLinks(fn)) {
                cout << "-E- Fail to parse lst file:" << fn.c_str() << endl;
                return 1;
            }
            return 0;
        } else if (!ext.compare("ibnd")) {
            if (parseIBNetDiscover(fn)) {
                cout << "-E- Fail to parse ibnetdiscover file:" << fn.c_str() << endl;
                return 1;
            }
            return 0;
        }
    }

    // Use topo file file parser as default
    if (parseTopoFile(fn)) {
        cout << "-E- Fail to parse topology file:" << fn.c_str() << endl;
        return 1;
    }

    return 0;
}

// Parse topo file and build the fabric from it.
int IBFabric::parseTopoFile(const string &fn)
{
    ifstream f(fn.c_str());
    char sLine[1024];
    string n1 = string(""), t1, p1, n2, t2, p2, cfg = string("");
    regExp sysLine("^[ \t]*([^/ \t]+)[ \t]+([^/ \t]+)[ \t]*( CFG:(.*))?$");
    regExp sysModule("([^ \t,]+)(.*)");
    regExp portLine("^[ \t]+([^ \t]+)[ \t]+-((1|4|8|12)[xX]-)?((2.5|5|10|14|25|50|FDR10|EDR20)G-)?[>]"
            "[ \t]+([^ \t]+)[ \t]+([^ \t]+)[ \t]+([^ \t]+)[ \t]*$");
    regExp ignoreLine("^[ \t]*#");
    regExp emptyLine("^[ \t]*$");
    rexMatch *p_rexRes;
    IBSystem *p_system = NULL;
    IBLinkSpeed speed;
    IBLinkWidth width;

    if (! f.is_open()) {
        cout << "-E- Fail to open file:" << fn.c_str() << endl;
        return 1;
    }

    cout << "-I- Parsing topology definition:" << fn.c_str() << endl;

    // we need two passes since we only get system configuration
    // on a system definition line.
    int lineNum = 0;
    while (f.good()) {
        lineNum++;
        f.getline(sLine,1024);

        // check if legal ignored line
        p_rexRes = ignoreLine.apply(sLine);
        if (p_rexRes) {
            delete p_rexRes;
            continue;
        }

        // First look for system line:
        p_rexRes = sysLine.apply(sLine);
        if (p_rexRes) {
            t1 = p_rexRes->field(1);
            n1 = p_rexRes->field(2);
            cfg = p_rexRes->field(4);
            //printf("%s %s %s\n", t1.c_str(),n1.c_str(),cfg.c_str());
            p_system = makeSystem(n1,t1,cfg);

            // check please:
            if (! p_system) {
                cout << "-E- Fail to make system:" << n1
                        << " of type:" <<  t1
                        << " (line:" << lineNum << ")"
                        << endl;
                delete p_rexRes;
                return 1;
            }
            delete p_rexRes;
            continue;
        }
    }

    lineNum = 0;
    f.close();
    f.clear();
    f.open(fn.c_str());

    if (! f.is_open()) {
        cout << "-E- Fail to re open file:" << fn.c_str() << endl;
        return 1;
    }

    int numCables = 0;

    while (f.good()) {
        lineNum++;
        f.getline(sLine,1024);

        // check if legal ignored line
        p_rexRes = ignoreLine.apply(sLine);
        if (p_rexRes) {
            delete p_rexRes;
            continue;
        }

        // look for system line:
        p_rexRes = sysLine.apply(sLine);

        if (p_rexRes) {
            t1 = p_rexRes->field(1);
            n1 = p_rexRes->field(2);
            cfg = p_rexRes->field(4);
            p_system = makeSystem(n1,t1,cfg);

            // check please:
            if (! p_system) {
                cout << "-E- Fail to make system:" << n1
                        << " of type:" <<  t1
                        << " (line:" << lineNum << ")"
                        << endl;
                delete p_rexRes;
                return 1;
            }
            delete p_rexRes;
            continue;
        }

        // is it a port line:
        p_rexRes = portLine.apply(sLine);
        if (p_rexRes) {
            if (p_system) {
                p1 = p_rexRes->field(1);
                width = char2width((p_rexRes->field(3) + "x").c_str());
                speed = char2speed((p_rexRes->field(5)).c_str());

                t2 = p_rexRes->field(6);
                n2 = p_rexRes->field(7);
                p2 = p_rexRes->field(8);
                if (addCable(t1, n1, p1, t2, n2, p2, width, speed)) {
                    cout << "-E- Fail to make cable"
                            << " (line:" << lineNum << ")"
                            << endl;
                    delete p_rexRes;
                    return 1;
                }
                numCables++;
            } else {
                cout << "-E- Fail to make connection as local system "
                        << "not defined (line:" << lineNum << ")"
                        << endl;
                delete p_rexRes;
                return 1;
            }
            delete p_rexRes;
            continue;
        }

        // check if empty line - marking system end:
        p_rexRes = emptyLine.apply(sLine);
        if (p_rexRes) {
            p_system = NULL;
            delete p_rexRes;
            continue;
        }

        cout << "-W- Ignoring '" << sLine << "'"
                << " (line:" << lineNum << ")" << endl;
    }

    cout << "-I- Defined " << SystemByName.size() << "/"
            << NodeByName.size() << " systems/nodes " << endl;
    f.close();
    return 0;
}

// Add a link into the fabric - this will create nodes / ports and link between them
// by calling the forward methods makeNode + setNodePortsystem + makeLinkBetweenPorts
int
IBFabric::addLink(string type1, phys_port_t numPorts1,
        uint64_t sysGuid1, uint64_t nodeGuid1,  uint64_t portGuid1,
        int vend1, device_id_t devId1, int rev1, string desc1,
        lid_t lid1, uint8_t lmc1, phys_port_t portNum1,
        string type2, phys_port_t numPorts2,
        uint64_t sysGuid2, uint64_t nodeGuid2,  uint64_t portGuid2,
        int vend2, device_id_t devId2, int rev2, string desc2,
        lid_t lid2, uint8_t lmc2, phys_port_t portNum2,
        IBLinkWidth width, IBLinkSpeed speed, IBPortState portState)
{
    IBNode *p_node1 = NULL, *p_node2 = NULL;
    IBPort *p_port1 = NULL, *p_port2 = NULL;

    //create the nodes
    p_node1 = getNodeByGuid(nodeGuid1);
    if (!p_node1) {
        p_node1 = makeNode(char2nodetype(type1.c_str()), numPorts1, sysGuid1, nodeGuid1,
                vend1, devId1, rev1, desc1);
        if (!p_node1) {
            cout << "-E- failed to allocate new node, guid=" << nodeGuid1 << endl;
            return 1;
        }
    }
    p_node2 = getNodeByGuid(nodeGuid2);
    if (!p_node2) {
        p_node2 = makeNode(char2nodetype(type2.c_str()), numPorts2, sysGuid2, nodeGuid2,
                vend2, devId2, rev2, desc2);
        if (!p_node2) {
            cout << "-E- failed to allocate new node, guid=" << nodeGuid2 << endl;
            return 1;
        }
    }

    //create the ports
    p_port1 = setNodePort(p_node1, portGuid1, lid1, lmc1, portNum1,
            width, speed, portState);
    if (!p_port1) {
        cout << "-E- failed to allocate new port, guid=" << portGuid1 << endl;
        return 1;
    }
    p_port2 = setNodePort(p_node2, portGuid2, lid2, lmc2, portNum2,
            width, speed, portState);
    if (!p_port2) {
        cout << "-E- failed to allocate new port, guid=" << portGuid2 << endl;
        return 1;
    }

    //create the link itself
    return makeLinkBetweenPorts(p_port1, p_port2);
}

// Create a new node in the fabric if not exist already (by guid),
// or get a pointer for the existing node
// Also create the system if required.
// Return a pointer for the node if success or NULL otherwise
IBNode *
IBFabric::makeNode(IBNodeType type, phys_port_t numPorts,
                  uint64_t sysGuid, uint64_t nodeGuid,
                  int vend, device_id_t devId, int rev, string desc
                  ) {
    //remove white spaces from description
    string whitespaces =  " \t\f\v\n\r";

    int found = (int)desc.find_last_not_of(whitespaces);
    if (found != (int)string::npos)
        desc.erase(found + 1);

    string lDesc = desc;
    IBNode *p_node   = NULL;
    bool mlx_nd_format = false;
    // find out if we know this description
    string nodeName;
    string sysName;
    string sysType;
    map_guid_str::iterator it = this->NGuid2Name.find(nodeGuid);
    bool is_mapped_description = false;

    //user gave maping of node guid to name
    if (it != this->NGuid2Name.end()) {
        //use user given name as description too
        lDesc = it->second;
        is_mapped_description = true;
    }

    bool uniqueNodeDesc = this->NodeByDesc[lDesc].empty();
    generateNodeAndSystemNames(type, sysGuid, nodeGuid, lDesc,
                               !uniqueNodeDesc, nodeName, sysName, sysType,
                               mlx_nd_format, is_mapped_description);

    // make sure the system exists
    IBSystem *p_sys  = NULL;
    p_sys = makeGenericSystem(sysName, sysType, mlx_nd_format);
    if (!p_sys) {
      cout << "-E- failed to allocate new generic system: " << sysName << endl;
      return NULL;
    }

    // create the node
    p_node = makeNode(nodeName, p_sys, type, numPorts);
    if (!p_node) {
        cout << "-E- failed to allocate new node: " << nodeName << endl;
        return NULL;
    }

    if (is_mapped_description)
        //we overide the description, so save the original description
        p_node->orig_description = desc;

    //add the node to NodeByDesc map
    this->NodeByDesc[lDesc].push_back(p_node);

    // we want to use the host names if they are defined:
    if (lDesc.size()) {
        p_node->attributes = string("host=") + lDesc;
        p_node->description = lDesc;
    }

    // we know the guids so set them
    p_node->guid_set(nodeGuid);
    //let the node remember its system guid (case: server with 2 HCAs)
    p_node->system_guid_set(sysGuid);

    // copy some data...
    p_node->devId    = devId;
    p_node->revId    = rev;
    p_node->vendId   = vend;

    return p_node;
}

// Create a new port for the specific node if port not exist,
// or set the port with the given data.
// NOTE that for LMC > 0 we do not get the multiple lid
// assignments - just the base lid.
// so we need to assign them ourselves (for CAs) if we have LMC
// Return a pointer for the port if success or NULL otherwise.
IBPort *
IBFabric::setNodePort(IBNode * p_node, uint64_t portGuid,
                      lid_t lid, uint8_t lmc, phys_port_t  portNum,
                      IBLinkWidth width, IBLinkSpeed speed, IBPortState port_state
                      ) {
    // add port
    IBUTILS_LOG(TT_LOG_LEVEL_DEBUG,
               "setNodePort portNum:%d lid:%d guid:" U64H_FMT
               " state:%d ,to node name: %s \n",
               portNum, lid, portGuid, port_state,
               p_node->name.c_str());

    IBPort *p_port = p_node->makePort(portNum);
    if (!p_port)
    {
        cout << "-E- failed to get port number: " << (unsigned int)portNum
                << " for node: " << p_node->name << endl;
        return NULL;
    }

    p_port->guid_set(portGuid);

    // handle LMC :
    p_port->lmc = lmc;
    lid_t numLidsPerPort = (lid_t)((lid_t)1 << lmc);
    p_port->base_lid = lid;
    for (lid_t l = lid; l < lid + numLidsPerPort; l++)
        setLidPort(l, p_port);

    // copy some data...
    p_port->set_internal_speed(speed);
    p_port->set_internal_width(width);
    p_port->set_internal_state(port_state);

    // handle switch Port[0]
    if (p_node->type == IB_SW_NODE) {
      IBPort *port0 = p_node->makePort(0);
      port0->base_lid = lid;
      port0->lmc = p_port->lmc;
    }

    return p_port;
}

// Create a new link in the fabric between ports.
// create system ports for the specific ports if required.
// Return 0 if success, else otherwise.
int
IBFabric::makeLinkBetweenPorts(IBPort *p_port1,
        IBPort *p_port2)
{
    IBUTILS_LOG(TT_LOG_LEVEL_DEBUG,
               "make Link Between Ports (node_name;lid;port_num): \n"
               "(%s,%d,%d)<->(%s,%d,%d)  \n"
               p_port1->p_node->name.c_str(), p_port1->base_lid, p_port1->num,
               p_port2->p_node->name.c_str(), p_port2->base_lid, p_port2->num);

    if (p_port1->get_internal_speed() != p_port2->get_internal_speed())
        cout << "Connected ports with different speed" << endl;
    if (p_port1->get_internal_width() != p_port2->get_internal_width())
        cout << "Connected ports with different width" << endl;
    if (p_port1->get_internal_state() != p_port2->get_internal_state())
        cout << "Connected ports with different states" << endl;

    p_port1->connect(p_port2);
    return 0;
}

// Parse an OpenSM Subnet file and build the fabric accordingly
int
IBFabric::parseSubnetLinks(string fn)
{

  regExp lstLine("^\\s*"
		 "[{]\\s*(SW|CA)(-SM)?\\s+Ports:([0-9A-Fa-f]{2})\\s+"
		 "SystemGUID:([0-9A-Fa-f]{16})\\s+NodeGUID:([0-9A-Fa-f]{16})\\s+PortGUID:([0-9A-Fa-f]{16})\\s+"
		 "VenID:([0-9A-Fa-f]+)\\s+DevID:([0-9A-Fa-f]+)\\s+Rev:([0-9A-Fa-f]+)\\s+"
		 "[{]([^}]+)[}]\\s+LID:([0-9A-Fa-f]{4})(:([0-8]))?\\s+PN:([0-9A-Fa-f]{2})\\s*[}]\\s+"
		 "[{]\\s*(SW|CA)(-SM)?\\s+Ports:([0-9A-Fa-f]{2})\\s+"
		 "SystemGUID:([0-9A-Fa-f]{16})\\s+NodeGUID:([0-9A-Fa-f]{16})\\s+PortGUID:([0-9A-Fa-f]{16})\\s+"
		 "VenID:([0-9A-Fa-f]+)\\s+DevID:([0-9A-Fa-f]+)\\s+Rev:([0-9A-Fa-f]+)\\s+"
		 "[{]([^}]+)[}]\\s+LID:([0-9A-Fa-f]{4})(:([0-8]))?\\s+PN:([0-9A-Fa-f]{2})\\s*[}]\\s+"
		 "PHY=([0-9]+x)\\s+LOG=(ACT|ARM|INI)\\s+(SPD=(\\S+))\\s*$"
		 );
  rexMatch *p_rexRes;

  ifstream f(fn.c_str());

  char sLine[1024];
  int status = 0;

  if (! f) {
    cout << "-E- Fail to open file:" << fn.c_str() << endl;
    return 1;
  }

  cout << "-I- Parsing Subnet file:" << fn.c_str() << endl;

  int lineNum = 0;
  while (f.good()) {
    lineNum++;

    f.getline(sLine,1024);
    if (!strlen(sLine)) continue;

    p_rexRes = lstLine.apply(sLine);
    if (!p_rexRes) {
      cout << "-W- fail to parse LST line:" << lineNum << endl;
      status++;
    } else {
      unsigned int i = 1;
      string type1 = p_rexRes->field(i++);
      i++; // skip SM?
      phys_port_t numPorts1 = (phys_port_t)strtol(p_rexRes->field(i++).c_str(), NULL, 16);
      uint64_t sysGuid1 = strtoull(p_rexRes->field(i++).c_str(), NULL, 16);
      uint64_t nodeGuid1 = strtoull(p_rexRes->field(i++).c_str(), NULL, 16);
      uint64_t portGuid1 = strtoull(p_rexRes->field(i++).c_str(), NULL, 16);
      uint32_t vend1 = (uint32_t)strtoul(p_rexRes->field(i++).c_str(), NULL, 16);
      device_id_t devId1 = (device_id_t)strtol(p_rexRes->field(i++).c_str(), NULL, 16);
      uint32_t rev1 = (uint32_t)strtoul(p_rexRes->field(i++).c_str(), NULL, 16);
      string desc1 = p_rexRes->field(i++);
      lid_t lid1 = (lid_t)strtol(p_rexRes->field(i++).c_str(), NULL, 16);
      bool lmc1Defined = p_rexRes->field(i++).size() != 0;
      uint8_t lmc1 = swLmc;
      if (lmc1Defined) {
	lmc1 = (uint8_t)(strtol(p_rexRes->field(i++).c_str(), NULL, 16));
      } else {
	i++;
	if (type1.compare("SW"))
	  lmc1 = caLmc;
      }
      phys_port_t portNum1 = (phys_port_t)strtol(p_rexRes->field(i++).c_str(), NULL, 16);

      string type2 = p_rexRes->field(i++);
      i++; // skip SM?
      phys_port_t numPorts2 = (phys_port_t)strtol(p_rexRes->field(i++).c_str(), NULL, 16);
      uint64_t sysGuid2 = strtoull(p_rexRes->field(i++).c_str(), NULL, 16);
      uint64_t nodeGuid2 = strtoull(p_rexRes->field(i++).c_str(), NULL, 16);
      uint64_t portGuid2 = strtoull(p_rexRes->field(i++).c_str(), NULL, 16);
      uint32_t vend2 = (uint32_t)strtoul(p_rexRes->field(i++).c_str(), NULL, 16);
      device_id_t devId2 = (device_id_t)strtol(p_rexRes->field(i++).c_str(), NULL, 16);
      uint32_t rev2 = (uint32_t)strtoul(p_rexRes->field(i++).c_str(), NULL, 16);
      string desc2 = p_rexRes->field(i++);
      lid_t lid2 = (lid_t)strtol(p_rexRes->field(i++).c_str(), NULL, 16);
      bool lmc2Defined = p_rexRes->field(i++).size() != 0;
      uint8_t lmc2 = swLmc;
      if (lmc2Defined) {
	lmc2 = (uint8_t)(strtol(p_rexRes->field(i++).c_str(), NULL, 16));
      } else {
	i++;
	if (type2.compare("SW"))
	  lmc2 = caLmc;
      }
      phys_port_t portNum2 = (phys_port_t)strtol(p_rexRes->field(i++).c_str(), NULL, 16);

      IBLinkWidth width = char2width(p_rexRes->field(i++).c_str());
      IBPortState portState = char2portstate(p_rexRes->field(i++).c_str());
      IBLinkSpeed speed  = IB_UNKNOWN_LINK_SPEED;
      if (p_rexRes->field(i++).size()) {
	speed = char2speed(p_rexRes->field(i++).c_str());
      }

      if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
	cout << "-V- Adding Link\n from:"
	     << " type1:" << type1 << " numPorts1:" << (unsigned int)numPorts1
	     << hex << " sysGuid1:" << sysGuid1 << " nodeGuid1:" << nodeGuid1
	     << " portGuid1:" << portGuid1 << dec << " vend1:" << vend1
	     << " devId1:" << devId1 << " rev1:" << rev1 << " desc1:" << desc1
	     << " lid1:" << lid1 << " lmc1:" << (unsigned int)lmc1 << " portNum1:" << (unsigned int)portNum1
	     << " \n   to:"
	     << " type2:" << type2 << " numPorts2:" << (unsigned int)numPorts2
	     << hex << " sysGuid2:" << sysGuid2 << " nodeGuid2:" << nodeGuid2
	     << " portGuid2:" << portGuid2 << dec << " vend2:" << vend2
	     << " devId2:" << devId2 << " rev2:" << rev2 << " desc2:" << desc2
	     << " lid2:" << lid2 << " lmc2:" << (unsigned int)lmc2 << " portNum2:" << (unsigned int)portNum2
	     << " \n width:" << width << " speed:" << speed << " portState:" << portState << endl;

      if (addLink(type1, numPorts1, sysGuid1, nodeGuid1, portGuid1,
		  vend1, devId1, rev1, desc1, lid1, lmc1, portNum1,
		  type2, numPorts2, sysGuid2, nodeGuid2, portGuid2,
		  vend2, devId2, rev2, desc2, lid2, lmc2, portNum2,
		  width, speed, portState)) {
	cout << "-W- fail to create link for LST line:" << lineNum << endl;
	status++;
      }
      delete p_rexRes;
    }
  }
  cout << "-I- Defined " << SystemByName.size() << "/" << NodeByName.size()
       << " systems/nodes. Max LID is " << maxLid  << endl;
  f.close();
  return 0;
}

// Parse an OpenSM FDBs file and set the LFT table accordingly
int
IBFabric::parseFdbFile(string fn)
{
    ifstream f(fn.c_str());
    int switches = 0, fdbLines=0;
    char sLine[1024];
    // dump_ucast_routes: Switch 0x2c90000213700
    // 0x0001 : 006  : 01   : yes
    regExp switchLine("dump_ucast_routes: Switch 0x([0-9a-z]+)");
    regExp lidLine("0x([0-9a-zA-Z]+) : ([0-9]+)");
    regExp plftLine("PLFT_NUM:\\s*([0-9]+)");

    rexMatch *p_rexRes;

    if (! f) {
        cout << "-E- Fail to open file:" << fn.c_str() << endl;
        return 1;
    }

    cout << "-I- Parsing FDBs file:" << fn.c_str() << endl;

    IBNode *p_node = NULL;
    int anyErr = 0;
    u_int8_t pLFT = 0;

    while (f.good()) {
        f.getline(sLine,1024);

        p_rexRes = switchLine.apply(sLine);
        if (p_rexRes) {
            // Got a new switch - find the node:
            uint64_t guid;
            guid = strtoull(p_rexRes->field(1).c_str(), NULL, 16);
            p_node = getNodeByGuid(guid);
            if (!p_node) {
                cout << "-E- Fail to find node with guid:"
                        << p_rexRes->field(1) << endl;
                anyErr++;
            } else {
                switches++;
                pLFT = 0;
            }
            delete p_rexRes;
            continue;
        }

        if (!p_node)
            continue;

        p_rexRes = lidLine.apply(sLine);
        if (p_rexRes) {
            // Got a new lid port pair
            if (p_node) {
                lid_t lid = (lid_t)strtol((p_rexRes->field(1)).c_str(), NULL, 16);
                phys_port_t port = (phys_port_t)strtol((p_rexRes->field(2)).c_str(), NULL, 10);
                if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
                    cout << "-V- Setting FDB for:" << p_node->name
			 << " lid:" << lid << " port:" << (unsigned int)port << endl;

                p_node->setLFTPortForLid(lid, port, pLFT);
                fdbLines++;
            }
            delete p_rexRes;
            continue;
        }

        p_rexRes = plftLine.apply(sLine);
        if (p_rexRes) {
            pLFT = (u_int8_t)strtol(
                (p_rexRes->field(1)).c_str(), NULL, 10);

            if(pLFT > p_node->getMaxPLFT()) {
                cout << "-E- Invalig PLFT:" << (unsigned int)pLFT
                    << " for node with guid:0x"
                    << hex << p_node->guid_get() << dec << endl;
                anyErr++;
            }

            delete p_rexRes;
            continue;
        }

        // is it an ignore line ?
        //cout << "-W- Ignoring line:" << sLine << endl;
    }
    cout << "-I- Defined " << fdbLines << " fdb entries for:"
            << switches << " switches" << endl;
    f.close();
    return anyErr;
}

// Parse PSL file and set the SLT tables accordingly
int
IBFabric::parsePSLFile(const string& fn)
{
    ifstream f(fn.c_str());

    char sLine[1024];
    // 0x0002c90000000099 154 0
    // srcguid dlid sl
    regExp slLine("0x([0-9a-fA-F]+) ([0-9]+) ([0-9]+)");
    // Switch 0x0002c902004050f8, base LID 13, "MF0;switch-112090:MTS3600/U1"
    regExp osmLine1("^(Channel Adapter|Switch|Router) 0x([0-9a-f]+),");
    regExp osmLine2("^0x([0-9a-fA-F]+)[ \t]+:[ \t]+([0-9]+)");
    regExp formatLine("^\\s*version:\\s*([0-9]+)\\.([0-9]+)\\.([0-9]+)\\s*$");
    regExp engineLine("^\\s*routing_engine:\\s*(\\S+)\\s*$");
    regExp ignoreLine("^\\s*(#|$)");
    rexMatch *p_rexRes;

    if (f.fail()) {
        cout << "-E- Fail to open file:" << fn.c_str() << endl;
        return 1;
    }

    cout << "-I- Parsing SL file:" << fn.c_str() << endl;

    int anyErr = 0;
    // versioning may be important one day
    unsigned int vMaj, vMid, vMin;
    vMaj = vMin = vMid = 0;

    // we use enumeration for the engine type
    enum {UNKNOWN, T2Q} engine_type = UNKNOWN;
    bool beforeAnyData = true;
    unsigned int lineNum = 0;

    while (f.good()) {
        f.getline(sLine,1024);
        lineNum++;
	if ((p_rexRes = slLine.apply(sLine))) {
	  lid_t lid = (lid_t)strtoull(p_rexRes->field(2).c_str(), NULL, 10);
	  maxLid = (lid > maxLid) ? lid : maxLid;
	  beforeAnyData = false;
	  delete p_rexRes;
        } else if ((p_rexRes = osmLine2.apply(sLine))) {
	  lid_t lid = (lid_t)strtoull(p_rexRes->field(1).c_str(), NULL, 16);
	  maxLid = (lid > maxLid) ? lid : maxLid;
	  beforeAnyData = false;
	  delete p_rexRes;
        } else if ((p_rexRes = osmLine1.apply(sLine))) {
	  delete p_rexRes;
	} else if ((p_rexRes = formatLine.apply(sLine))) {
	  if (!beforeAnyData) {
	    cout << "-E- Format must be defined before any data line!" << endl;
	    anyErr++;
	  }
	  vMaj = (unsigned int)strtoul(p_rexRes->field(1).c_str(), NULL, 10);
	  vMid = (unsigned int)strtoul(p_rexRes->field(2).c_str(), NULL, 10);
	  vMin = (unsigned int)strtoul(p_rexRes->field(3).c_str(), NULL, 10);
	  delete p_rexRes;
	} else if  ((p_rexRes = engineLine.apply(sLine))) {
	  if (!beforeAnyData) {
	    cout << "-E- Routing engine must be defined before any data line!" << endl;
	    anyErr++;
	  }
	  if (p_rexRes->field(1) == "torus-2QoS") {
	    engine_type = T2Q;
	    cout << "-I- Routing Engine is: T2Q" << endl;
	  } else {
	    cout << "-E- Unsupported routing engine: " << p_rexRes->field(1) << endl;
	    anyErr++;
	  }
	  delete p_rexRes;
	} else if  ((p_rexRes = ignoreLine.apply(sLine))) {
	  delete p_rexRes;
	} else {
	  cout << "-E- Unsupported syntax at line: " << lineNum << endl;
	  anyErr++;
	}
    }
    f.close();

    // validate supported format
    if ((vMaj > 1) ||
	((vMaj == 1) && (vMid > 0)) ||
	((vMaj == 1) && (vMid == 0) && (vMin > 0))) {
      cout << "-E- File version:" << vMaj << "." << vMid << "." << vMin
	   << " > max-supported-version: 1.0.0" << endl;
      anyErr++;
    }

    if (anyErr) {
      return 1;
    }

    // Make second pass and build the tables
    f.open(fn.c_str(),ifstream::in);
    if (f.fail()) {
        cout << "-E- Fail to open file:" << fn.c_str() << endl;
        return 1;
    }

    uint64_t guid = 0;
    IBNode* p_node = NULL;
    set< IBNode *> srcNodes;

    lineNum = 0;
    while (f.good()) {
        f.getline(sLine,1024);
	lineNum++;
        if ((p_rexRes = slLine.apply(sLine))) {
            guid = strtoull(p_rexRes->field(1).c_str(), NULL, 16);
            lid_t lid = (lid_t)strtoull(p_rexRes->field(2).c_str(), NULL, 10);
            uint8_t sl = (uint8_t)strtoull(p_rexRes->field(3).c_str(), NULL, 10);

            p_node = getNodeByGuid(guid);
            if (!p_node) {
                cout << "-E- Fail to find node with guid:"
                        << p_rexRes->field(1).c_str() << endl;
                anyErr++;
            } else {
                // Update number of used SLs
                numSLs = sl+1 > numSLs ? (uint8_t)(sl+1):numSLs;
                // Insert table entry
                p_node->setPSLForLid(lid, maxLid, sl);
            }
            delete p_rexRes;
        } else if ((p_rexRes = osmLine1.apply(sLine))) {
            guid = strtoull(p_rexRes->field(2).c_str(), NULL, 16);
            IBPort *p_port =  getPortByGuid(guid);
            if (!p_port) {
                cout << "-E- Fail to find port with guid:"
                        <<  p_rexRes->field(2).c_str() << endl;
                anyErr++;
                guid = 0;
            } else {
                p_node = p_port->p_node;
		srcNodes.clear();
		srcNodes.insert(p_node);
		// for T2Q engine - find all connected hosts and add to srcs
		if (engine_type == T2Q) {
		  if (p_node->type != IB_SW_NODE) {
		    cout << "-W- Line: " << lineNum << " provided T2Q mode node is not a switch: " << p_node->name << endl;
		  } else {
		    for (unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
		      IBPort *p_port = p_node->getPort((phys_port_t)pn);
		      if (!p_port || !p_port->p_remotePort) continue;
		      if (p_port->p_remotePort->p_node->type != IB_SW_NODE) {
			srcNodes.insert(p_port->p_remotePort->p_node);
		      }
		    }
		  }
		}
            }
            delete p_rexRes;
        } else if ((p_rexRes = osmLine2.apply(sLine))) {
            if (guid != 0) {
                lid_t lid = (lid_t)strtoull(p_rexRes->field(1).c_str(), NULL, 16);
                uint8_t sl = (uint8_t)strtoull(p_rexRes->field(2).c_str(), NULL, 10);
                // Update number of used SLs
                numSLs = sl+1 > numSLs ? (uint8_t)(sl+1):numSLs;

		if (engine_type == T2Q) {
		  // need to first collect all destinations off that switch
		  IBPort *p_dPort = getPortByLid(lid);
		  if (!p_dPort) {
		    cout << "-E- could not find port by LID:" << lid << endl;
		  } else {
		    IBNode *p_dNode = p_dPort->p_node;
		    if (p_dNode->type != IB_SW_NODE) {
		      cout << "-W- Line: " << lineNum << " provided T2Q mode dlid that is not a switch: " << p_dNode->name << endl;
		    } else {
		      unsigned int numSL2VLAssigned = 0;
		      for (unsigned int pn = 1; pn <= p_dNode->numPorts; pn++) {
                  IBPort *p_port = p_dNode->getPort((phys_port_t)pn);
			if (!p_port || !p_port->p_remotePort) continue;
			if (p_port->p_remotePort->p_node->type != IB_SW_NODE) {
			  // make sure we cover ALL DLIDs of the destination
			  lid_t baseLid;
			  uint8_t lmc;
			  p_port->p_remotePort->p_node->getLidAndLMC(p_port->p_remotePort->num, baseLid, lmc);
			  for (lid_t dlid = baseLid; dlid < baseLid + (lid_t)(1 << lmc); dlid++) {
			    if (dlid > maxLid) {
			      cout << "-E- BUG: got dlid:" << dlid << " base:" << baseLid << " lmc:" << lmc
				   << " > maxLid:" << maxLid << endl;
			      abort();
			    }
			    // ok so mark all src nodes
			    for (set<IBNode*>::iterator sI = srcNodes.begin(); sI != srcNodes.end(); sI++) {
			      (*sI)->setPSLForLid(dlid, maxLid, sl);
			      numSL2VLAssigned++;
			    }
			  }
			}
		      }
		      if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
			cout << "-V- Assigned:" << numSL2VLAssigned << " SL2VL entries going from: "
			     << p_node->name << " to: " << p_dNode->name << endl;
		    } // p_dNode is a switch
		  } // p_dPort exists
		} else { // not T2Q
		  // Insert table entry
		  p_node->setPSLForLid(lid, maxLid, sl);
		}
            } else {
	      cout << "-E- Skipping line:" << lineNum << " since no previous device is defined" << endl;
	      anyErr++;
            }
            delete p_rexRes;
        } else if ((p_rexRes = formatLine.apply(sLine))) {
	  delete p_rexRes;
	} else if  ((p_rexRes = engineLine.apply(sLine))) {
	  delete p_rexRes;
	} else if  ((p_rexRes = ignoreLine.apply(sLine))) {
	  delete p_rexRes;
	}
    }
    cout << "-I- Defined "<< (int)numSLs << " SLs in use" <<endl;
    f.close();
    return anyErr;
}

// Parse SLVL file and set the SL2VL tables accordingly
int
IBFabric::parseSLVLFile(const string& fn)
{
    numVLs = 1;
    ifstream f(fn.c_str());

    char sLine[1024];
    // 0x0002c90000000201 5 1 0x01 0x23 0x45 0x67 0x89 0xab 0xcd 0xe7
    // guid iport oport 0x(sl0)(sl1) 0x(sl2)(sl3)...
    regExp slLine("^0x([0-9a-f]+) ([0-9]+) ([0-9]+) 0x([0-9a-f])([0-9a-f]) 0x([0-9a-f])([0-9a-f]) 0x([0-9a-f])([0-9a-f]) 0x([0-9a-f])([0-9a-f]) 0x([0-9a-f])([0-9a-f]) 0x([0-9a-f])([0-9a-f]) 0x([0-9a-f])([0-9a-f]) 0x([0-9a-f])([0-9a-f])");
    // Switch 0x0002c902004050f8, base LID 13, "MF0;switch-112090:MTS3600/U1"
    regExp osmLine1("^(Channel Adapter|Switch) 0x([0-9a-f]+),");

    // 14  11  : 0  1  2  3  4  5  6  7  8  9  10 11 12 13 14 7
    regExp osmLine2("^([0-9]+)[ \t]+([0-9]+)[ \t]+:[ \t]+([0-9]+)[ \t]+([0-9]+)[ \t]+([0-9]+)[ \t]+([0-9]+)[ \t]+([0-9]+)[ \t]+([0-9]+)[ \t]+([0-9]+)[ \t]+([0-9]+)[ \t]+([0-9]+)[ \t]+([0-9]+)[ \t]+([0-9]+)[ \t]+([0-9]+)[ \t]+([0-9]+)[ \t]+([0-9]+)[ \t]+([0-9]+)[ \t]+([0-9]+)[ \t]*$");
    rexMatch *p_rexRes;

    if (f.fail()) {
        cout << "-E- Fail to open file:" << fn.c_str() << endl;
        return 1;
    }

    cout << "-I- Parsing SLVL file:" << fn.c_str() << endl;

    int anyErr = 0;
    uint64_t guid = 0;
    IBNode* p_node = NULL;

    // Parse the file
    while (f.good()) {
        f.getline(sLine,1024);
        p_rexRes = slLine.apply(sLine);
        if (p_rexRes) {
            guid =
                    strtoull(p_rexRes->field(1).c_str(), NULL, 16);
            phys_port_t iport =
                    (phys_port_t)strtoull(p_rexRes->field(2).c_str(), NULL, 10);
            phys_port_t oport =
                    (phys_port_t)strtoull(p_rexRes->field(3).c_str(), NULL, 10);
            p_node = getNodeByGuid(guid);
            if (!p_node) {
                cout << "-E- Fail to find node with guid:"
                        << p_rexRes->field(1).c_str() << endl;
                anyErr++;
            } else {
                for (uint8_t i = 0; i < IB_NUM_SL; ++i) {
                    // Extract the VL value
                    uint8_t vl = (uint8_t)strtoull(p_rexRes->field(4+i).c_str(), NULL, 16);
                    if (vl < IB_DROP_VL)
                        numVLs = numVLs > vl+1 ? numVLs : (uint8_t)(vl+1);
                    // Set the table entry
                    p_node->setSLVL(iport, oport, i, vl);
                }
            }
            delete p_rexRes;
        } else if ((p_rexRes = osmLine1.apply(sLine))) {
            guid =
                    strtoull(p_rexRes->field(2).c_str(), NULL, 16);
            IBPort *p_port = getPortByGuid(guid);
            if (!p_port) {
                cout << "-E- Fail to find node with guid: 0x"
                        << p_rexRes->field(2).c_str() << endl;
                anyErr++;
                guid = 0;
            } else {
                p_node = p_port->p_node;
            }
            delete p_rexRes;
        } else if ((p_rexRes = osmLine2.apply(sLine))) {
            if (guid != 0) {
                phys_port_t iport =
                        (phys_port_t)strtoull(p_rexRes->field(1).c_str(), NULL, 10);
                phys_port_t oport =
                        (phys_port_t)strtoull(p_rexRes->field(2).c_str(), NULL, 10);

                for (uint8_t i = 0; i < IB_NUM_SL; ++i) {
                    // Extract the VL value
                    uint8_t vl = (uint8_t)strtoull(p_rexRes->field(3+i).c_str(), NULL, 10);
                    numVLs = (numVLs > vl+1) ? numVLs : (uint8_t)(vl+1);
                    // Set the table entry
                    p_node->setSLVL(iport, oport, i, vl);
                }
            } else {
                cout << "-E- Ignoring SL2VL line with no previous matching guid"
                        << endl;
            }
            delete p_rexRes;
        }
    }
    cout << "-I- Defined "<< (int)numVLs << " VLs in use" <<endl;
    f.close();
    return anyErr;
}

// Parse an OpenSM MCFDBs file and set the MFT table accordingly
int
IBFabric::parseMCFdbFile(string fn)
{
    ifstream f(fn.c_str());
    int switches = 0, fdbLines=0;
    char sLine[1024];
    // Switch 0x0002c9010bb90090
    // LID    : Out Port(s)
    // 0xC000 : 0x007
    // 0xC001 : 0x007
    regExp switchLine("Switch 0x([0-9a-z]+)");
    regExp lidLine("0x([0-9a-zA-Z]+) :(.*)");
    rexMatch *p_rexRes;

    if (! f) {
        cout << "-E- Fail to open file:" << fn.c_str() << endl;
        return 1;
    }

    cout << "-I- Parsing Multicast FDBs file:" << fn.c_str() << endl;

    IBNode *p_node = NULL;
    int anyErr = 0;

    while (f.good()) {
        f.getline(sLine,1024);

        p_rexRes = switchLine.apply(sLine);
        if (p_rexRes) {
            // Got a new switch - find the node:
            uint64_t guid;
            guid = strtoull(p_rexRes->field(1).c_str(), NULL, 16);
            p_node = getNodeByGuid(guid);
            if (!p_node) {
                cout << "-E- Fail to find switch: node:"
                        << p_rexRes->field(1) << endl;
                anyErr++;
            } else {
                switches++;
            }
            delete p_rexRes;
            continue;
        }

        p_rexRes = lidLine.apply(sLine);
        if (p_rexRes) {
            // Got a new lid port pair
            if (p_node) {
                lid_t lid = (lid_t)strtol((p_rexRes->field(1)).c_str(), NULL, 16);

                char buff[(p_rexRes->field(2)).size() + 1];
                strcpy(buff, p_rexRes->field(2).c_str());

                char *pPortChr = strtok(buff, " ");
                while (pPortChr) {
                    phys_port_t port = (phys_port_t)strtol(pPortChr, NULL, 16);
                    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
                        cout << "-V- Setting Multicast FDB for:" << p_node->name
                            << " lid:" << lid << " port:" << (unsigned int)port << endl;

                    p_node->setMFTPortForMLid(lid, port);
                    pPortChr = strtok(NULL, " ");
                    fdbLines++;
                }
            }
            delete p_rexRes;
            continue;
        }

        // is it an ignore line ?
        //cout << "-W- Ignoring line:" << sLine << endl;
    }
    cout << "-I- Defined " << fdbLines << " Multicast Fdb entries for:"
            << switches << " switches" << endl;
    f.close();
    return anyErr;
}

int
IBFabric::parseCommaSeperatedValues(
    const string &line, vector<u_int32_t> &vecRes)
{
    int ret_size = 0;

    //count values number
    size_t vla_count = std::count(line.begin(), line.end(), ',') +1;
    if (vecRes.size() < vla_count)
        vecRes.resize(vla_count, 0);

    std::string::size_type startIdx = 0;
    std::string::size_type endIdx = line.find(',');
    std::string::size_type length = line.length();

    for(; startIdx  <  length ; )  {
        if (endIdx == std::string::npos)
            endIdx = length;
        vecRes[ret_size] =
            (u_int32_t)strtol(line.substr(startIdx,endIdx-startIdx).c_str(),
            NULL,0);
        startIdx = ++endIdx;
        endIdx=line.find(',',startIdx);
        ret_size++;
    }

    return ret_size;
}

int
IBFabric::getFileVersion(ifstream &f, u_int16_t &fileVersion)
{
    char sLine[1024];
    regExp versionLine("File version:\\s*([0-9]+)");
    regExp emplyLine("^\\s*$");
    regExp ignoreLine("^\\s*(#|$)");
    rexMatch *p_rexRes;
    fileVersion = 0;

    while (f.good()) {
        f.getline(sLine,1024);
        if(sLine[0] == '\0' || sLine[0] == '#')
            continue;

        if ((p_rexRes = ignoreLine.apply(sLine))) {
            delete p_rexRes;
            continue;
        }

        if ((p_rexRes = emplyLine.apply(sLine))) {
            delete p_rexRes;
            continue;
        }

        if ((p_rexRes = versionLine.apply(sLine))) {
            fileVersion = (u_int8_t)strtol(
                (p_rexRes->field(1)).c_str(), NULL, 10);
            delete p_rexRes;
            return 0;
        }
        return 1;
    }

    return 1;
}

int
IBFabric::parseVL2VLFile(string fn)
{
    ifstream f(fn.c_str());
    int switches = 0;
    char sLine[1024];
    // dump_ucast_routes: Switch 0x2c90000213700 lhiefghfl, sdfkjhsdf
    // 0x0001 : 006  : 01   : yes
    regExp switchLine("dump_vl2vl: Switch 0x([0-9a-z]+)\\s+(.*)");
    rexMatch *p_rexRes;
    vector<u_int32_t> vecRes(IB_NUM_SL);

    if (! f) {
        cout << "-E- Fail to open file:" << fn.c_str() << endl;
        return 1;
    }

    cout << "-I- Parsing VL2VL file:" << fn.c_str() << endl;

    IBNode *p_node = NULL;
    int anyErr = 0;

    u_int16_t fileVersion;
    if (getFileVersion(f, fileVersion)) {
        cout << "-E- Fail to read file version from:" << fn << endl;
        return 1;
    }
    if (fileVersion != 1) {
        cout << "-E- Unsupported file version:" << fileVersion
            << "for " << fn << endl;
        return 1;
    }

    while (f.good()) {
        f.getline(sLine,1024);

        p_rexRes = switchLine.apply(sLine);
        if (p_rexRes) {
            // Got a new switch - find the node:
            uint64_t guid;
            guid = strtoull(p_rexRes->field(1).c_str(), NULL, 16);
            p_node = getNodeByGuid(guid);
            if (!p_node) {
                cout << "-E- Fail to find node with guid:"
                        << p_rexRes->field(1) << endl;
                anyErr++;
            } else {
                switches++;
                int sls = parseCommaSeperatedValues(p_rexRes->field(2), vecRes);
                if (sls > IB_NUM_SL) {
                    cout << "-E- invalid vl2vl line for node with guid:"
                            << p_rexRes->field(1) << endl;
                    anyErr++;
                } else {
                    for (int i = 0 ; i < sls ; ++i) {
                        if (vecRes[i] >= IB_NUM_SL) {
                            cout << "-E- invalid sl:" << vecRes[i]
                                << " in vl2vl line for node with guid:"
                                    << p_rexRes->field(1) << endl;
                            anyErr++;
                            break;
                        }
                        else
                            p_node->setVL2VL((u_int8_t)vecRes[i]);
                    }

                }
            }
            delete p_rexRes;
            continue;
        }
        // is it an ignore line ?
        //cout << "-W- Ignoring line:" << sLine << endl;
    }
    cout << "-I- Defined vl2vl on " << switches << " switches" << endl;
    f.close();
    return anyErr;
}




int IBFabric::parseFARFile(string fn) {
    ifstream f(fn.c_str());
    //copy
    int switches = 0;
    int fdbLines = 0;
    int grpLines = 0;
    char sLine[1024];
    char dbgLine[1024];
    bool parseGrp = false;
    bool parseLid = false;
    regExp switchLine(
            "dump_ar: Switch 0x([0-9a-z]+)\\s+(fr_en:\\s*([0-1])\\s+)?en_sl:\\s*(.*)");
    regExp grpDef("Groups Definition:");
    regExp grpLine("([0-9]+)\\s+([0-9]+)\\s+(([0-9]+,)+)"); //

    regExp lftDef("LFT Definition:");
    regExp plftLine("PLFT_NUM:\\s*([0-9]+)");
    regExp lidLine1(
            "0x([0-9a-zA-Z]+)\\s*(UNREACHABLE|[0-9]+)\\s*(Static|Free|Bounded)\\s*(UNASSIGNED|[0-9]+)");

    rexMatch *p_rexRes;

    /*
     dump_ar: Switch 0x0002c90000000029 fr_en: 0 en_sl: 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15

     Groups Definition:

     Group     Sub Group      Ports
     -------------------------------------
     0         0              1,2,3,

     LFT Definition:

     PLFT_NUM: 0

     LID       Static Port    Lid State      Group
     -----------------------------------------------
     0x0000    UNREACHABLE    Static         UNASSIGNED
     0x0001    6              Static         UNASSIGNED
     0x0002    1              Free           0
     *
     * */

    if (!f) {
        cout << "-E- Fail to open file:" << fn.c_str() << endl;
        return 1;
    }

    cout << "-I- Parsing FAR file:" << fn.c_str() << endl; //FAR

    IBNode *p_node = NULL;
    u_int8_t pLFT = 0;
    lid_t maxLid = 1024;
    u_int16_t maxGroup = 0;
    bool isGroupSet = false;
    int anyErr = 0;
    vector<u_int32_t> vecRes(IB_NUM_SL);
    vector<u_int32_t> grpRes(256);

    u_int16_t fileVersion;
    if (getFileVersion(f, fileVersion)) {
        cout << "-E- Fail to read file version from:" << fn << endl;
        return 1;
    }

    if (fileVersion > 2) {
        cout << "-E- Unsupported file version:" << fileVersion << "for " << fn
                << endl;
        return 1;
    }

    while (f.good()) {
        f.getline(sLine, 1024);
        if (sLine[0] == '\0' || sLine[0] == '#')
            continue;

        if (parseLid) {
            //case1:
            //IS PLFT LINE?
            p_rexRes = plftLine.apply(sLine);
            if (p_rexRes) {
                pLFT = (u_int8_t) strtol((p_rexRes->field(1)).c_str(), NULL,
                        10);

                if (pLFT > p_node->getMaxPLFT()) {
                    cout << "-E- Invalig PLFT:" << (unsigned int) pLFT
                            << " for node with guid:0x" << hex
                            << p_node->guid_get() << dec << endl;
                    anyErr++;
                }

                p_node->clearLFT(pLFT);
                p_node->resizeLFT((uint16_t) (maxLid + 1), pLFT);
                p_node->resizeARLFT((uint16_t) (maxLid + 1), pLFT);

                delete p_rexRes;

                //go down 3 lines:
                f.getline(sLine, 1024); //empty
                f.getline(sLine, 1024); //LID       Static Port    Lid State      Group
                f.getline(sLine, 1024); //-----------------------------------------------
                //next line will be lids..
                continue;
            }
            //case2:
            //LID LINE:

            p_rexRes = lidLine1.apply(sLine);
            if (p_rexRes) {

//            LID       Static Port    Lid State      Group
//            -----------------------------------------------
//            0x0000    UNREACHABLE    Static         UNASSIGNED
//            0x0001    3              Static         UNASSIGNED

                lid_t lid = (lid_t) strtol(p_rexRes->field(1).c_str(), NULL,
                        16);
                maxLid = max(maxLid, lid);

                if (strcmp(p_rexRes->field(2).c_str(), "UNREACHABLE") == 0) { //if port = UNREACHABLE_PORT;
                    continue; //its taken care by default
                }

                //DETERMINE STATIC-PORT
                phys_port_t port = (phys_port_t) strtol(
                        p_rexRes->field(2).c_str(), NULL, 10);

                p_node->setLFTPortForLid(lid, port, pLFT);

                SMP_AR_LID_STATE state;
                if (strcmp(p_rexRes->field(3).c_str(), "Static") == 0) {
                    state = AR_IB_LID_STATE_STATIC;
                } else if (strcmp(p_rexRes->field(3).c_str(), "Free") == 0) {
                    state = AR_IB_LID_STATE_FREE;
                } else if (strcmp(p_rexRes->field(3).c_str(), "Bounded") == 0) {
                    state = AR_IB_LID_STATE_BOUNDED;
                } else {
                    state = AR_IB_LID_STATE_LAST;
                    cout << "-E- Invalid State:" << p_rexRes->field(3).c_str()
                            << " for node with guid:0x" << hex
                            << p_node->guid_get() << dec << endl;
                    anyErr++;
                }

                p_node->setARstateForLid(lid, state, pLFT);

                //GET GROUP
                bool realGroup = false;
                bool grpSet = false;
                u_int16_t group = 0;
                if (strcmp(p_rexRes->field(4).c_str(), "UNASSIGNED") != 0) {
                    group = (u_int16_t) strtol(p_rexRes->field(4).c_str(), NULL,
                            10);
                    realGroup = true;
                }

                if(!isGroupSet){
                   ; //should not happen!  for debug
                }

                if (group > maxGroup) { //|| !isGroupSet
                    cout << "-E- Invalid group:" << (unsigned int) group
                            << " for node with guid:0x" << hex
                            << p_node->guid_get() << dec << " lid:" << lid
                            << endl;
                    anyErr++;
                } else {
                    if (realGroup) {
                        p_node->setARLFTPortGroupForLid(lid, group, pLFT);
                    }
                    grpSet = true;
                }
                delete (p_rexRes);
                fdbLines++;
                if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {
                    cout << "-V- Setting AR FDB for:" << p_node->name
                            << " PLFT:" << (int) pLFT << " lid:" << lid
                            << " static port:" << (unsigned int) port
                            << " state:" << (int) state;
                    if (grpSet) {
                        cout << " group:" << (unsigned int) group;
                        cout << endl;
                    }
                }
                continue;
            }

        } else if (parseGrp) {
//            Group     Sub Group      Ports
//            -------------------------------------
//            0         0              1,2,3,
            p_rexRes = grpLine.apply(sLine);
            if (p_rexRes) {
                u_int16_t group = (u_int16_t) strtol(p_rexRes->field(1).c_str(),
                NULL, 10);
                u_int16_t subGroup = (u_int16_t) strtol(
                        p_rexRes->field(2).c_str(), NULL, 10);
                list_phys_ports portsList;

                int ports = parseCommaSeperatedValues(
                        p_rexRes->field(3), grpRes);
                for (int i = 0; i < ports; ++i) {
                    if (grpRes[i] > p_node->numPorts) {
                        cout << "-E- invalid port:" << grpRes[i] << " in group:"
                                << group << " in sub group:" << subGroup
                                << " for node with guid:" << hex
                                << p_node->guid_get() << dec << endl;
                        anyErr++;
                        break;
                    } else
                        portsList.push_back((phys_port_t) grpRes[i]);
                }

                if (!portsList.empty()) {
                    p_node->setARSubGrp(group, subGroup, portsList);
                    grpLines++;

                    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {

                        p_node->getARGroupCfg(group, dbgLine);
                        cout << "-V- set group:" << group << " with sub group:"
                                << subGroup << " for node with guid:" << hex
                                << p_node->guid_get() << dec << " ports:"
                                << dbgLine << endl;

                    }
                }
                //

                maxGroup = max(maxGroup, group);
                isGroupSet = true;
                delete p_rexRes;
                continue;
            }
        } //parsegrp

        //if no group, no lid
        //IS SWITCH?
        //COPIED AS IS from parse AR, assumining the line is the same.
        p_rexRes = switchLine.apply(sLine);
        if (p_rexRes) {
            parseGrp = false;
            parseLid = false;
            maxGroup = 0;
            isGroupSet = false;

            // Got a new switch - find the node:
            uint64_t guid;
            guid = strtoull(p_rexRes->field(1).c_str(), NULL, 16);
            p_node = getNodeByGuid(guid);
            if (!p_node) {
                cout << "-E- Fail to find node with guid:" << p_rexRes->field(1)
                        << endl;
                anyErr++;
            } else {
                switches++;
                bool frEnabled = (bool) strtol((p_rexRes->field(3)).c_str(),
                NULL, 10);
                if (frEnabled)
                    p_node->setFREnabled();
                int sls = parseCommaSeperatedValues(p_rexRes->field(4), vecRes);
                if (sls > IB_NUM_SL) {
                    cout << "-E- invalid en_sl line for node with guid:"
                            << p_rexRes->field(1) << endl;
                    anyErr++;
                } else {
                    for (int i = 0; i < sls; ++i) {
                        if (vecRes[i] >= IB_NUM_SL) {
                            cout << "-E- invalid sl:" << vecRes[i]
                                    << " in en_sl line for node with guid:"
                                    << p_rexRes->field(1) << endl;
                            anyErr++;
                            break;
                        } else
                            p_node->setAREnabled((u_int8_t) vecRes[i]);
                    }
                }
            }
            delete p_rexRes;
            continue;
        }

        //if there is already a switch found
        if (p_node) {
            //SEARCH FOR GROUP DEF
            p_rexRes = grpDef.apply(sLine);
            if (p_rexRes) {
                //need to go down some lines to get to data:
                f.getline(sLine, 1024); //empty
                f.getline(sLine, 1024); //Group     Sub Group      Ports
                f.getline(sLine, 1024); //-------------------------------------
                parseGrp = true;
                parseLid = false;
                delete p_rexRes;
                continue;
            }
            //(OR) SEARCH FOR LFT DEF
            p_rexRes = lftDef.apply(sLine);
            if (p_rexRes) {
                parseGrp = false;
                parseLid = true;
                delete p_rexRes;
                continue;
            }

        }
    } //while f.good...

    cout << "-I- 	FAR Defined " << fdbLines << " fdb entries and " << grpLines
            << " group entries for:" << switches << " switches" << endl;
    f.close();
    return anyErr;
}






int
IBFabric::parseARFile(string fn)
{
    ifstream f(fn.c_str());
    int switches = 0;
    int fdbLines=0;
    int grpLines=0;
    char sLine[1024];
    char dbgLine[1024];
    bool parseGrp = false;
    bool parseLid = false;
    //dump_ar: Switch 0x0002c90000000001 en_sl: 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15
    //Groups Definition:
    //0: 3, 4
    //1: 3, 4
    //2: 4, 3
    //LFT Definition:
    //PLFT_NUM: 0
    //LID    : static port : AR group
    //0x0001 :  003  :  000
    //0x0002 :  004  :  001
    //0x0003 :  004  :  001
    regExp switchLine("dump_ar: Switch 0x([0-9a-z]+)\\s+(fr_en:\\s*([0-1])\\s+)?en_sl:\\s*(.*)");
    regExp grpDef("Groups Definition:");
    regExp grpLine("([0-9]+):\\s+(.*)");
    regExp lftDef("LFT Definition:");
    regExp plftLine("PLFT_NUM:\\s*([0-9]+)");
    regExp lidLine1("0x([0-9a-zA-Z]+)\\s*:(.*)");
    regExp lidLine2("\\s+([0-9]+)\\s*:(.*)");
    regExp lidLine3("\\s+([0-9]+)");

    rexMatch *p_rexRes;
    rexMatch *p_rexRes2;
    rexMatch *p_rexRes3;
    vector<u_int32_t> vecRes(IB_NUM_SL);
    vector<u_int32_t> grpRes(256);

    if (! f) {
        cout << "-E- Fail to open file:" << fn.c_str() << endl;
        return 1;
    }

    cout << "-I- Parsing AR file:" << fn.c_str() << endl;

    IBNode *p_node = NULL;
    u_int8_t pLFT = 0;
    lid_t maxLid = 1024;
    u_int16_t maxGroup = 0;
    bool isGroupSet = false;
    int anyErr = 0;

    char *iter = NULL;
    char *seperator = NULL;
    char *endptr = NULL;

    u_int16_t fileVersion;
    if (getFileVersion(f, fileVersion)) {
        cout << "-E- Fail to read file version from:" << fn << endl;
        return 1;
    }

    if (fileVersion > 2) {
        cout << "-E- Unsupported file version:" << fileVersion
            << "for " << fn << endl;
        return 1;
    }

    while (f.good()) {
        f.getline(sLine,1024);

        if(sLine[0] == '\0' || sLine[0] == '#')
            continue;

        if (parseLid) {
            //optimized parsing for lidLine1
            if (sLine[0] >= '0' && sLine[0] <= '9') {
                for ( iter = sLine ; (*iter != '\0' && *iter != ':') ; ++iter );
                if (*iter != ':')
                    continue; //invalid line
                seperator = iter++;
                for ( ; (*iter != '\0' && *iter != ':') ; ++iter );
                if (*iter == ':') {
                    lid_t lid = (lid_t)strtol(sLine, NULL, 16);
                    maxLid = max(maxLid, lid);

                    phys_port_t port = (phys_port_t)strtol(
                        seperator+1, &endptr, 10);
                    if (endptr == seperator+1)
                        continue;  //port unreachable

                    bool grpSet = false;
                    p_node->setLFTPortForLid(lid, port, pLFT);

                    seperator = iter;
                    u_int16_t group = 0;

                    group = (u_int16_t)strtol(
                            seperator+1, &endptr, 10);

                    if (endptr != seperator+1){

                        if (group > maxGroup || !isGroupSet) {
                            cout << "-E- Invalig group:" << (unsigned int)group
                                << " for node with guid:0x"
                                << hex << p_node->guid_get() << dec
                                << " lid:" << lid  << endl;
                            anyErr++;
                        } else {
                            p_node->setARLFTPortGroupForLid(lid, group, pLFT);
                            grpSet = true;
                        }
                    }
                    fdbLines++;

                    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {
                        cout << "-V- Setting AR FDB for:" << p_node->name
                            << " PLFT:" << (int) pLFT
                            << " lid:" << lid
                            << " static port:" << (unsigned int)port;
                        if (grpSet)
                            cout << " group:" << (unsigned int)group;
                        cout << endl;
                    }
                    continue;
                }
            }
            p_rexRes = lidLine1.apply(sLine);
            if (p_rexRes) {
                lid_t lid = (lid_t)strtol(p_rexRes->field(1).c_str(), NULL, 16);
                maxLid = max(maxLid, lid);

                p_rexRes2 = lidLine2.apply(p_rexRes->field(2).c_str());
                if (p_rexRes2) {
                    phys_port_t port = (phys_port_t)strtol(
                        p_rexRes2->field(1).c_str(), NULL, 10);
                    bool grpSet = false;

                    p_node->setLFTPortForLid(lid, port, pLFT);

                    u_int16_t group = 0;
                    p_rexRes3 = lidLine3.apply(p_rexRes2->field(2).c_str());
                    if (p_rexRes3) {
                        group = (u_int16_t)strtol(
                            p_rexRes3->field(1).c_str(), NULL, 10);

                        if (group > maxGroup || !isGroupSet) {
                            cout << "-E- Invalid group:" << (unsigned int)group
                                << " for node with guid:0x"
                                << hex << p_node->guid_get() << dec
                                << " lid:" << lid  << endl;
                            anyErr++;
                        } else {
                            p_node->setARLFTPortGroupForLid(lid, group, pLFT);
                            grpSet = true;
                        }

                        delete p_rexRes3;
                    }
                    fdbLines++;

                    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {
                        cout << "-V- Setting AR FDB for:" << p_node->name
                            << " PLFT:" << (int) pLFT
                            << " lid:" << lid
                            << " static port:" << (unsigned int)port;
                        if (grpSet)
                            cout << " group:" << (unsigned int)group;
                        cout << endl;
                    }
                    delete p_rexRes2;
                }
                delete p_rexRes;
                continue;
            }

            p_rexRes = plftLine.apply(sLine);
            if (p_rexRes) {
                pLFT = (u_int8_t)strtol(
                    (p_rexRes->field(1)).c_str(), NULL, 10);

                if(pLFT > p_node->getMaxPLFT()) {
                    cout << "-E- Invalig PLFT:" << (unsigned int)pLFT
                        << " for node with guid:0x"
                        << hex << p_node->guid_get() << dec << endl;
                    anyErr++;
                }

                p_node->clearLFT(pLFT);
                p_node->resizeLFT((uint16_t)(maxLid+1), pLFT);
                p_node->resizeARLFT((uint16_t)(maxLid+1), pLFT);

                delete p_rexRes;
                continue;
            }
        } else if (parseGrp) {
            p_rexRes = grpLine.apply(sLine);
            if (p_rexRes) {
                u_int16_t group = (u_int16_t)strtol(
                    p_rexRes->field(1).c_str(), NULL, 10);
                list_phys_ports portsList;

                int ports = parseCommaSeperatedValues(p_rexRes->field(2), grpRes);
                for (int i = 0 ; i < ports ; ++i) {
                    if (grpRes[i]  > p_node->numPorts) {
                        cout << "-E- invalid port:" << grpRes[i]
                            << " in group:" << group
                            << " for node with guid:"
                            << hex << p_node->guid_get() << dec << endl;
                        anyErr++;
                        break;
                    }
                    else
                        portsList.push_back((phys_port_t)grpRes[i]);
                }

                if (!portsList.empty()) {
                    p_node->setARPortGroup(group, portsList);
                    grpLines++;

                    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {

                        p_node->getARGroupCfg(group, dbgLine);
                        cout << "-V- set group:" << group
                            << " for node with guid:"
                            << hex << p_node->guid_get() << dec
                            << " ports:" << dbgLine << endl;
                    }
                }
                maxGroup = max(maxGroup, group);
                isGroupSet = true;

                delete p_rexRes;
                continue;
            }
        }

        p_rexRes = switchLine.apply(sLine);
        if (p_rexRes) {
            parseGrp = false;
            parseLid = false;
            maxGroup = 0;
            isGroupSet = false;

            // Got a new switch - find the node:
            uint64_t guid;
            guid = strtoull(p_rexRes->field(1).c_str(), NULL, 16);
            p_node = getNodeByGuid(guid);
            if (!p_node) {
                cout << "-E- Fail to find node with guid:"
                        << p_rexRes->field(1) << endl;
                anyErr++;
            } else {
                switches++;
                bool frEnabled = (bool)strtol((p_rexRes->field(3)).c_str(), NULL, 10);
                if (frEnabled)
                    p_node->setFREnabled();
                int sls = parseCommaSeperatedValues(p_rexRes->field(4), vecRes);
                if (sls > IB_NUM_SL) {
                    cout << "-E- invalid en_sl line for node with guid:"
                            << p_rexRes->field(1) << endl;
                    anyErr++;
                } else {
                    for (int i = 0 ; i < sls ; ++i) {
                        if (vecRes[i] >= IB_NUM_SL) {
                            cout << "-E- invalid sl:" << vecRes[i]
                                << " in en_sl line for node with guid:"
                                    << p_rexRes->field(1) << endl;
                            anyErr++;
                            break;
                        }
                        else
                            p_node->setAREnabled((u_int8_t)vecRes[i]);
                    }
                }
            }
            delete p_rexRes;
            continue;
        }

        if (p_node) {
            p_rexRes = grpDef.apply(sLine);
            if (p_rexRes) {
                parseGrp = true;
                parseLid = false;

                delete p_rexRes;
                continue;
            }

            p_rexRes = lftDef.apply(sLine);
            if (p_rexRes) {
                parseGrp = false;
                parseLid = true;

                delete p_rexRes;
                continue;
            }
        }

        // is it an ignore line ?
        //cout << "-W- Ignoring line:" << sLine << endl;
    }
    cout << "-I- AR Defined " << fdbLines << " fdb entries and "
        << grpLines << " group entries for:"
        << switches << " switches" << endl;
    f.close();
    return anyErr;
}

int
IBFabric::parsePLFTFile(string fn)
{
    ifstream f(fn.c_str());
    int switches = 0;
    int plftLines=0;
    char sLine[1024];
    //char dbgLine[1024];
    //dump_ar: Switch 0x0002c90000000001
    //rq: 0 sl-plfft: 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1
    //rq: 1 sl-plfft: 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1
    regExp switchLine("dump_plft: Switch 0x([0-9a-z]+)");
    regExp plftLine("rq: ([0-9]+) sl-plfft:(.*)");

    rexMatch *p_rexRes;
    vector<u_int32_t> vecRes(IB_NUM_SL);

    if (! f) {
        cout << "-E- Fail to open file:" << fn.c_str() << endl;
        return 1;
    }

    cout << "-I- Parsing PLFT file:" << fn.c_str() << endl;

    IBNode *p_node = NULL;
    int anyErr = 0;
    u_int16_t fileVersion;
    if (getFileVersion(f, fileVersion)) {
        cout << "-E- Fail to read file version from:" << fn << endl;
        return 1;
    }
    if (fileVersion != 1) {
        cout << "-E- Unsupported file version:" << fileVersion
            << "for " << fn << endl;
        return 1;
    }

    while (f.good()) {
        f.getline(sLine,1024);

        if(sLine[0] == '\0' || sLine[0] == '#')
            continue;

        p_rexRes = switchLine.apply(sLine);
        if (p_rexRes) {
            // Got a new switch - find the node:
            uint64_t guid;
            guid = strtoull(p_rexRes->field(1).c_str(), NULL, 16);
            p_node = getNodeByGuid(guid);
            if (!p_node) {
                cout << "-E- Fail to find node with guid:"
                        << p_rexRes->field(1) << endl;
                anyErr++;
            } else {
                switches++;
                p_node->setPLFTEnabled();
            }
            delete p_rexRes;
            continue;
        }

        if (!p_node)
            continue;

        p_rexRes = plftLine.apply(sLine);
        if (p_rexRes) {
            phys_port_t port = (phys_port_t)strtol(
                p_rexRes->field(1).c_str(), NULL, 10);

            int sls = parseCommaSeperatedValues(p_rexRes->field(2), vecRes);
            if (sls > IB_NUM_SL) {
                cout << "-E- invalid sl-plfft line for node with guid:"
                    << hex << p_node->guid_get() << dec << endl;
                anyErr++;
            } else {
                for (u_int8_t sl = 0 ; sl < sls ; ++sl) {
                    p_node->setPLFTMapping(port, sl, (u_int8_t)vecRes[sl]);
                    plftLines++;
                }
            }

            delete p_rexRes;
            continue;
        }

        // is it an ignore line ?
        //cout << "-W- Ignoring line:" << sLine << endl;
    }
    cout << "-I- PLFT Defined " << plftLines << " plft entries for:"
        << switches << " switches" << endl;
    f.close();
    return anyErr;
}

int
IBFabric::parseSADumpFile(string fn)
{
    ifstream f(fn.c_str());
    int groups = 0;
    int members = 0;
    char sLine[1024];
    char *iter;
    lid_t mlid = 0;
    u_int32_t sl_flow_hop = 0;
    u_int8_t sl = 0;
    u_int64_t guid = 0;
    McastGroupInfo tmpMcastGroupInfo;
    McastGroupInfo *p_groupInfo = NULL;
    McastGroupMemberInfo tmpMcastGroupMemberInfo;

    //MC Group 0xc002 : mgid=0xff12401bffff0000:0x0000000000000001 port_gid=0xfe80000000000000:0x0002c903007b78b0
    //                  qkey=0x00000b1b mlid=0xc002 mtu=0x84 tclass=0x00 pkey=0xffff rate=0x83 pkt_life=0x80
    //                  sl_flow_hop=0x00000000 scope_state=0x01 proxy_join=0x0

    //mcm_port: port_gid=0xfe80000000000000:0x0002c90300342002 scope_state=0x01 proxy_join=0x0

    if (! f) {
        cout << "-E- Fail to open file:" << fn.c_str() << endl;
        return 1;
    }

    cout << "-I- Parsing SA Dump file:" << fn.c_str() << endl;

    int anyErr = 0;

    while (f.good()) {
        f.getline(sLine,1024);

        if(sLine[0] == '\0' || sLine[0] == '#')
            continue;
        iter = sLine;

        //trim leading whitespace
        while(isspace(*iter)) iter++;

        if (iter-sLine > 100) continue;

        if (!strncmp(iter, "MC Group", 8)) {
            iter = strstr(iter, " mlid=");
            if (iter == NULL) {
                cout << "-E- failed to find mlid in MC Group line: "<< sLine << endl;
                anyErr++;
                continue;
            }
            iter+=6;//length of mlid=
            mlid = (lid_t)strtol(iter, NULL, 16);

            if (mlid < 0xc000) {
                cout << "-E- invalid mlid: " << hex << mlid << dec
                    << " in MC Group line: "<< sLine << endl;
                anyErr++;
                continue;
            }

            iter = strstr(iter, " sl_flow_hop=");
            if (iter == NULL) {
                cout << "-E- failed to find sl_flow_hop in MC Group line: "<< sLine << endl;
                mlid = 0;
                anyErr++;
                continue;
            }
            iter+=13;//length of " sl_flow_hop="
            sl_flow_hop = (u_int32_t)strtol(iter, NULL, 16);
            sl = (u_int8_t)(sl_flow_hop>>28);


            groups++;

            //inset
            McastGroupMapInsertRes res =
                McastGroups.insert(make_pair(mlid, tmpMcastGroupInfo));
            p_groupInfo = &res.first->second;

        } else if (!strncmp(iter, "mcm_port", 8)) {
            if (mlid==0) continue;

            iter = strstr(iter, " port_gid=");
            if (iter == NULL) {
                cout << "-E- failed to find port_gid in mcm_port line: "
                    << sLine << endl;
                anyErr++;
                continue;
            }
            iter+=10;//length of " port_gid="
            iter = strchr(iter, ':');
            if (iter == NULL) {
                cout << "-E- failed to parse port_gid in mcm_port line: "
                    << sLine << endl;
                anyErr++;
                continue;
            }
            iter++;
            guid = (u_int64_t)strtoull(iter, NULL, 0);
            iter = strstr(iter, " scope_state=");
            iter+=13;//length of " scope_state="
            bool sender_only = false;
            uint8_t state = (uint8_t)strtoul(iter, NULL, 16);
            uint8_t sender_only_bits =
                (IB_MCAST_STATE_SENDER_ONLY_FULL_MEMBER_BIT |
                 IB_MCAST_STATE_SENDER_ONLY_BIT);
            // state 8 bit field is consist of:
            //     Scope (4 bits) - MGID Address Scope
            //     JoinState (4 bits) - Join/Leave Status requested by the port.
            //                          Second bit saying that this port is
            //                          sender-only port.
            if (((state & 0x0f) | sender_only_bits) == sender_only_bits)
                sender_only = true;

            members++;

            IBPort *p_port = NULL;
            if (!(p_port = this->getPortByGuid(guid, true, true))) {
                cout << "-E- failed to find port guid: 0x" << std::hex
                     << guid << std::dec << " in fabric. If the guid is virtual "
                     << "or alias make sure these stages were not skipped. error"
                     << " in line: " << sLine << endl;
                anyErr++;
                continue;
            }

            //inset
            McastMemberMapInsertRes res =
                p_groupInfo->m_members.insert(
                    make_pair(p_port, tmpMcastGroupMemberInfo));
            res.first->second.m_sls.insert(sl);
            // if this p_port was first inserted
            if (res.second)
                res.first->second.is_sender_only = sender_only;
            else
                res.first->second.is_sender_only &= sender_only;


            /*
            cout << "-E- Insert mlid: " << mlid
                << " guid:" << guid
                << " sl:" << (int)sl << endl;
                */

        } else continue;
    }
    cout << "-I- SA DUMP : " << groups << " groups:"
        << members << " members" << endl;
    f.close();
    return anyErr;
}

int IBFabric::parseNodeNameMapFile(string fn)
{
    ifstream mapFile(fn.c_str());

    if (!mapFile.good()) {
        cout << "-E- Cannot open mapping file: " << fn << endl;
        return(1);
    }
    //# comment
    //<guid> "<name>"
    regExp mapGUIDLine("^\\s*(0x[0-9a-fA-F]+)\\s+\"(((\\S+)\\s*)+)\"\\s*$");
    regExp ignoreLine("^[ \t]*(#|[ \t]*$)");
    rexMatch *p_rexRes;
    unsigned int lineNum = 0;
    char sLine[1024] = {0};
    int anyErr = 0;

    while (mapFile.good()) {
        mapFile.getline(sLine, sizeof(sLine) - 1);
        lineNum++;
        if ((p_rexRes = mapGUIDLine.apply(sLine))) {
            uint64_t dguid = strtoull(p_rexRes->field(1).c_str(), NULL, 16);
            string newName = p_rexRes->field(2);
            std::pair<map_guid_str::iterator, bool> ret =
                NGuid2Name.insert(std::pair<uint64_t, string>(dguid, newName));
            //check if such guid already exist in the file (insert fails)
            if (ret.second == false) {
                ios::fmtflags saveflags(cout.flags());
                cout << "-W- Line " << lineNum
                     << ": Node guid 0x" << std::hex << dguid
                     << " already exists with name \"" << ret.first->second
                     << "\"" << endl;
                cout.flags(saveflags);
            }
            delete p_rexRes;
        } else if ((p_rexRes = ignoreLine.apply(sLine))) //legally ignored
            delete p_rexRes;
        else if (strlen(sLine)) {
            cout << "-W- Ignoring illegal line (" << lineNum << ") : "
                 << sLine << endl;
            anyErr ++;
            continue;
        }
    }

    mapFile.close();
    return(0);
}

// dump out the contents of the entire fabric
void
IBFabric::dump(ostream &sout)
{
    sout << "--------------- FABRIC DUMP ----------------------" << endl;
    // we start with all systems at top level:
    for (map_str_psys::iterator sI = SystemByName.begin();
            sI != SystemByName.end();
            sI++) {
        IBSystem *p_system = (*sI).second;
        sout << "\nSystem:" << p_system->name << " (" << p_system->type
                << ")" << endl;
        for (map_str_psysport::iterator pI = p_system->PortByName.begin();
                pI != p_system->PortByName.end();
                pI++) {
            IBSysPort *p_port = (*pI).second;
            IBLinkWidth width = IB_UNKNOWN_LINK_WIDTH;
            IBLinkSpeed speed = IB_UNKNOWN_LINK_SPEED;

            if (! p_port) continue;

            // node port
            sout << "  " << p_port->name;
            if (p_port->p_nodePort) {
                sout << " (" << p_port->p_nodePort->p_node->name << "/"
                        << (unsigned int)p_port->p_nodePort->num << ")";
                width = p_port->p_nodePort->get_common_width();
                speed = p_port->p_nodePort->get_common_speed();
            } else {
                sout << " (ERR: NO NODE PORT?)";
            }

            // remote sys port?
            if (p_port->p_remoteSysPort) {
                sout << " -" << width2char(width) << "-" << speed2char(speed)
                    << "G-> "
                    << p_port->p_remoteSysPort->p_system->name << "/"
                    << p_port->p_remoteSysPort->name << endl;
            } else {
                sout << endl;
            }
        }
    }

    // Now dump system internals:
    for (map_str_psys::iterator sI = SystemByName.begin();
            sI != SystemByName.end();
            sI++) {
        IBSystem *p_system = (*sI).second;
        sout << "--------------- SYSTEM " << (*sI).first
                << " DUMP ----------------------" << endl;

        // go over all nodes of the system:
        for (map_str_pnode::iterator nI = p_system->NodeByName.begin();
                nI != p_system->NodeByName.end();
                nI++) {
            IBNode *p_node = (*nI).second;

            sout << "\nNode:" << p_node->name << " (" << p_node->type
                    << ", node guid: " << guid2str(p_node->guid_get()) << ", system guid: "
                    << guid2str(p_node->system_guid_get()) << ")" << endl;

            for (unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
                IBPort *p_port = p_node->getPort((phys_port_t)pn);

                if (! p_port) continue;

                // we do not report cross system connections:
                if (p_port->p_sysPort) {
                    sout << "   " << pn << " => SysPort:"
                            << p_port->p_sysPort->name << endl;
                } else if (p_port->p_remotePort) {
                    sout << "   "  << pn << " -" << width2char(p_port->get_common_width())
                            << "-" << speed2char(p_port->get_common_speed()) << "G-> "
                            << p_port->p_remotePort->getName() << endl;
                }
            }
        }
    }
}

// dump out the contents of the entire fabric as a topology file with
// required set of IBNL's if unknown
int
IBFabric::dumpTopology(const char *fileName, const char * ibnlDir)
{
    ofstream sout;
    string err_message;
    int rc = IBFabric::OpenFile(fileName, sout, false, err_message);

    if (rc) {
        cout << "-E- failed to open:" << fileName << " for writing." << endl;
        return rc;
    }

    /*
    sout.open(fileName);
    if (sout.fail()) {
        cout << "-E- failed to open:" << fileName << " for writing." << endl;
        return 1;
    }
    */
    sout << "# This topology file was automatically generated by IBDM" << endl;

    // we start with all systems at top level:
    for (map_str_psys::iterator sI = SystemByName.begin();
            sI != SystemByName.end();
            sI++) {
        IBSystem *p_system = (*sI).second;

        string sysType;
        if (p_system->newDef) {
            p_system->dumpIBNL(ibnlDir, sysType);
        } else {
            sysType = string(p_system->type);
        }
        sout << "\n" << sysType << " " << p_system->name
            << p_system->cfg << endl;
        for (map_str_psysport::iterator pI = p_system->PortByName.begin();
                pI != p_system->PortByName.end();
                pI++) {
            IBSysPort *p_sysPort = (*pI).second;
            IBLinkWidth width = IB_UNKNOWN_LINK_WIDTH;
            IBLinkSpeed speed = IB_UNKNOWN_LINK_SPEED;

            if (! p_sysPort) continue;

            // remote sys port?
            if ( p_sysPort->p_remoteSysPort) {
                width = p_sysPort->p_nodePort->get_common_width();
                speed = p_sysPort->p_nodePort->get_common_speed();

                if (p_sysPort->p_remoteSysPort->p_system->newDef) {
                    p_sysPort->p_remoteSysPort->p_system->dumpIBNL(ibnlDir, sysType);
                } else {
                    sysType = string(p_sysPort->p_remoteSysPort->p_system->type);
                }

                sout << "   " << p_sysPort->name
                        << " -" << width2char(width) << "-" << speed2char(speed)
                        << "G-> " << sysType.c_str() << " "
                        << p_sysPort->p_remoteSysPort->p_system->name << " "
                        << p_sysPort->p_remoteSysPort->name << endl;
            }
        }
    }
    sout.close();
    return 0;
}

// dump out the contents of the entire fabric as a lst file format
int
IBFabric::dumpLSTFile(char *fileName, bool write_with_lmc)
{
    ofstream sout;
    string err_message;
    int rc = OpenFile(fileName, sout, false, err_message);

    if (rc) {
        cout << "-E- failed to open:" << fileName << " for writing." << endl;
        return rc;
    }
    /*
    sout.open(fileName);
    if (sout.fail())
    {
        cout << "-E- failed to open:" << fileName << " for writing." << endl;
        return 1;
    }
    */

    //going over all ports in the fabric and initial some fields we need for this algorithm
    for (map_guid_pnode::iterator it = this->NodeByGuid.begin();
            it != this->NodeByGuid.end();
            ++it) {
        IBNode *p_curr_node = (*it).second;
        if (!p_curr_node) {
            cout << "-E- Reached null node in guid to node map (guid=" << (*it).first <<
                    ")." << endl;
            return 1;
        }
        for (unsigned int i = 1; i <= p_curr_node->numPorts; ++i) {    //start from port 1, no need to dump port 0
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            if (!p_curr_port)   //port not exist
                continue;
            p_curr_port->counter1 = 0;
        }
    }

    //going over all ports and dump to lst file the ones that are connected
    char lst_line[1024];
    for (map_guid_pnode::iterator it = this->NodeByGuid.begin();
            it != this->NodeByGuid.end();
            ++it) {
        IBNode *p_curr_node = (*it).second;
        if (!p_curr_node) {
            cout << "-E- Reached null node in guid to node map (guid=" << (*it).first <<
                    ")." << endl;
            return 1;
        }
        for (unsigned int i = 1; i <= p_curr_node->numPorts; ++i) {     //start from port 1, no need to dump port 0
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            if (!p_curr_port)   //port not exist
                continue;
            IBPort *p_remote_port = p_curr_port->p_remotePort;
            if (!p_remote_port) //port not connected
                continue;
            IBNode *p_remote_node = p_remote_port->p_node;
            if (!p_remote_node) {
                cout << "-E- Reached port with null node." << endl;
                return 1;
            }

            //don't write same connection more than one time
            if ((p_curr_port->counter1 == 1) || (p_remote_port->counter1 == 1))
                continue;
            p_curr_port->counter1 = 1;
            p_remote_port->counter1 = 1;

            stringstream lmc;
            lmc.str("");
            if (write_with_lmc && (p_curr_node->type == IB_CA_NODE) && p_curr_port->lmc)
                lmc << ":" << ((int)p_curr_port->lmc) << " ";
            else
                lmc << " ";
            //write connection in lst format
            memset(lst_line, 0, sizeof(lst_line));
            sprintf(lst_line, "{ "
                                "%s Ports:%02x" \
                                " SystemGUID:%016" PRIx64 " NodeGUID:%016" PRIx64 " PortGUID:%016" PRIx64   \
                                " VenID:%08x DevID:%x Rev:%08x"   \
                                " {%s}" \
                                " LID:%04x%sPN:%02x" \
                                " }",
                                nodetype2char(p_curr_node->type),
                                p_curr_node->numPorts,
                                p_curr_node->system_guid_get(),
                                p_curr_node->guid_get(),
                                p_curr_port->guid_get(),
                                p_curr_node->vendId,
                                p_curr_node->devId,
                                p_curr_node->revId,
                                p_curr_node->description.c_str(),
                                p_curr_port->base_lid,
                                lmc.str().c_str(),
                                p_curr_port->num);
            sout << lst_line << " ";

            lmc.str("");
            if (write_with_lmc && (p_remote_node->type == IB_CA_NODE) && p_remote_port->lmc)
                lmc << ":" << ((int)p_remote_port->lmc) << " ";
            else
                lmc << " ";
            memset(lst_line, 0, sizeof(lst_line));
            sprintf(lst_line, "{ "
                                "%s Ports:%02x" \
                                " SystemGUID:%016" PRIx64 " NodeGUID:%016" PRIx64 " PortGUID:%016" PRIx64   \
                                " VenID:%08x DevID:%x Rev:%08x"   \
                                " {%s}" \
                                " LID:%04x%sPN:%02x" \
                                " }",
                                (p_remote_node->type == IB_CA_NODE) ? "CA" : "SW",
                                p_remote_node->numPorts,
                                p_remote_node->system_guid_get(),
                                p_remote_node->guid_get(),
                                p_remote_port->guid_get(),
                                p_remote_node->vendId,
                                p_remote_node->devId,
                                p_remote_node->revId,
                                p_remote_node->description.c_str(),
                                p_remote_port->base_lid,
                                lmc.str().c_str(),
                                p_remote_port->num);
            sout << lst_line << " ";

            memset(lst_line, 0, sizeof(lst_line));
            sprintf(lst_line, "PHY=%s LOG=%s SPD=%s",
                                width2char(p_curr_port->get_common_width()),
                                portstate2char(p_curr_port->get_common_state()),
                                speed2char(p_curr_port->get_common_speed()));
            sout << lst_line << endl;
        }
    }
    return 0;
}

void IBFabric::setLidPort(lid_t lid, IBPort *p_port) {
  if (!lid) return;
  if ( PortByLid.empty() || (PortByLid.size() < (unsigned int)lid + 1))
    for (size_t i = PortByLid.size(); i < (size_t)(lid + 1); i++)
      PortByLid.push_back(NULL);
  if (PortByLid[lid]) {
    if (PortByLid[lid]->p_node != p_port->p_node) {
      cout << "-E- Overriding previous LID:" << lid
	   << " port: " << PortByLid[lid]->getName()
	   << " with new port: " << p_port->getName() << endl;
      PortByLid[lid] = p_port;
    }
  } else {
    PortByLid[lid] = p_port;
  }
  if (maxLid < lid)
    maxLid = lid;
}

void IBFabric::setLidVPort(lid_t lid, IBVPort *p_vport) {
  if (!lid) return;

  if ( VPortByLid.empty() || (VPortByLid.size() < (unsigned int)lid + 1))
      VPortByLid.resize((size_t)(lid + 1));

  if (VPortByLid[lid]) {
    if (VPortByLid[lid]->m_p_vnode != p_vport->m_p_vnode) {
      cout << "-E- Overriding previous LID:" << lid
       << " vport: " << VPortByLid[lid]->getName()
       << " with new vport: " << p_vport->getName() << endl;
      VPortByLid[lid] = p_vport;
    }
  } else {
    VPortByLid[lid] = p_vport;
  }

  if (maxLid < lid)
    maxLid = lid;
}

// dump out the name to guid and LID map of the given topolgy
// including all NODEs of the fabric
int
IBFabric::dumpNameMap(const char *fileName)
{
    ofstream sout;
    string err_message;
    int rc = OpenFile(fileName, sout, false, err_message);

    if (rc) {
        cout << "-E- failed to open:" << fileName << " for writing." << endl;
        return rc;
    }

    /*
    sout.open(fileName);
    if (sout.fail()) {
        cout << "-E- failed to open:" << fileName << " for writing." << endl;
        return 1;
    }
    */
    sout << "# This name map file was automaticlly generated by IBDM" << endl;
    sout << "# NodeGUID PortGUID PortLID NAME/SysImageGUID" << endl;

    // Go over all nodes of the fabric
    for (map_str_pnode::iterator nI = NodeByName.begin();
            nI != NodeByName.end();
            nI++) {
        IBNode *p_node = (*nI).second;
        phys_port_t fromPort;
        phys_port_t toPort;
        if (p_node->type == IB_SW_NODE) {
          fromPort = 0;
          toPort = 0;
        } else {
          fromPort = 1;
          toPort = p_node->numPorts;
        }
        for (unsigned int pn = fromPort; pn <= toPort; pn++) {
          IBPort *p_port = p_node->getPort((phys_port_t)pn);
          if (!p_port) continue;
          lid_t lid;
          uint8_t lmc;

          p_node->getLidAndLMC((phys_port_t)pn, lid, lmc);
          sout << guid2str(p_node->guid_get())
               << guid2str(p_port->guid_get())
               << " " << lid
               << " " << (*nI).first << endl;
        }
    }
    sout.close();
    return 0;
}

static int
convLinkWidthSpeed(string linkWS, IBLinkWidth &width, IBLinkSpeed &speed)
{
  regExp rexLW("^([0-9]+)x(\\S+)$");
  rexMatch *p_rexRes;
  if ((p_rexRes = rexLW.apply(linkWS.c_str()))) {
    string widthStr = p_rexRes->field(1);
    string speedStr = p_rexRes->field(2);
    if (!widthStr.compare("1")) {
      width = IB_LINK_WIDTH_1X;
    } else if (!widthStr.compare("4")) {
      width = IB_LINK_WIDTH_4X;
    } else if (!widthStr.compare("8")) {
      width = IB_LINK_WIDTH_8X;
    } else if (!widthStr.compare("12")) {
      width = IB_LINK_WIDTH_12X;
    } else if (!widthStr.compare("2")) {
      width = IB_LINK_WIDTH_2X;
    } else {
      width = IB_UNKNOWN_LINK_WIDTH;
    }

    if (!speedStr.compare("SDR")) {
      speed = IB_LINK_SPEED_2_5;
    } else if (!speedStr.compare("DDR")) {
      speed = IB_LINK_SPEED_5;
    } else if (!speedStr.compare("QDR")) {
      speed = IB_LINK_SPEED_10;
    } else if (!speedStr.compare("FDR")) {
      speed = IB_LINK_SPEED_14;
    } else if (!speedStr.compare("FDR10")) {
      speed = IB_LINK_SPEED_FDR_10;
    } else if (!speedStr.compare("EDR")) {
      speed = IB_LINK_SPEED_25;
    } else if (!speedStr.compare("EDR20")) {
        speed = IB_LINK_SPEED_EDR_20;
    } else if (!speedStr.compare("HDR")) {
       speed = IB_LINK_SPEED_50;
    } else {
      speed = IB_UNKNOWN_LINK_SPEED;
    }
    delete p_rexRes;
    return 0;
  }
  cout << "-W- Unknown format for Link Width Speed:" << linkWS << endl;
  speed = IB_UNKNOWN_LINK_SPEED;
  width = IB_UNKNOWN_LINK_WIDTH;
  return(1);
}

// Parse ibnetdiscover file and build the fabric from it.
int IBFabric::parseIBNetDiscover (string fn)
{
  ifstream f(fn.c_str());
  ios::fmtflags cout_flags(cout.flags());


  char sLine[1024];
  // 0 = idle -> 1 = base-port 2 = ports

  if (! f) {
    cout << "-E- Fail to open file:" << fn.c_str() << endl;
    return 1;
  }

  cout << "-I- Parsing ibnetdiscover file:" << fn.c_str() << endl;

  regExp argLine("^(vendid|devid|caguid|sysimgguid|switchguid)=(\\S+)$");
  //Switch  24 "S-005442ba00003080"         # "ISR9024 Voltaire" base port 0 lid 6 lmc 0
  //Ca      2 "H-0008f10403960558"          # "MT23108 InfiniHost Mellanox Technologies"
  regExp basePortLine("^(Ca|Switch)\\s+([0-9]+)\\s+\"(S|H|R)-([0-9a-fA-F]+)\"\\s+#\\s+\"([^\"]+)\""
                      "(\\s+(base|enhanced) port 0 lid\\s+([0-9a-fA-F]+)\\s+lmc\\s+([0-9]+))?$");

  //[22]    "H-0008f10403961354"[1]         # "MT23108 InfiniHost Mellanox Technologies" lid 4
  regExp swPortLine("^[[]([0-9]+)[]]\\s+\"(S|H|R)-([0-9a-fA-F]+)\"[[]([0-9]+)[]]([(]([0-9a-fA-F]+)[)])?"
                    "\\s+#\\s+\"([^\"]+)\"\\s+lid\\s+([0-9]+)\\s+(\\S+)\\s*$");

  //[1]     "S-005442ba00003080"[12]        # lid 10 lmc 1 "ISR9024 Voltaire" lid 6
  regExp caPortLine("^[[]([0-9]+)[]]([(]([0-9a-fA-F]+)[)])?\\s+\"(S|H|R)-([0-9a-fA-F]+)\""
                    "[[]([0-9]+)[]]\\s+#\\s+lid\\s+([0-9]+)\\s+lmc\\s+([0-9]+)\\s+\"([^\"]+)\""
                    "\\s+lid\\s+([0-9]+)\\s+(\\S+)\\s*$");
  regExp ignoreLine("^(#.*|Non-Chassis Nodes|)$");

  rexMatch *p_rexRes;
  unsigned int numErrs = 0;
  uint64_t nGuid = 0;
  uint64_t sGuid = 0;
  uint32_t vendid = 0;
  device_id_t devid = 0;
  IBSystem *pLocSys;
  IBNode   *pLocNode = NULL, *pRemNode = NULL;
  IBPort   *pLocPort = NULL, *pRemPort = NULL;
  bool mlx_nd_format = false;

  // since the number of ports is not known on remote port connections we
  // must do a double pass
  for (unsigned int pass = 0; pass < 2; pass++) {
    int state = 0;
    phys_port_t locPortNum = 0;
    uint64_t remNodeGuid = 0;
    phys_port_t remPortNum = 0;
    string linkWS;
    while (f.good()) {
      f.getline(sLine,1024);
      bool doConn = false;
      if ((p_rexRes = ignoreLine.apply(sLine))) {
        delete p_rexRes;
      } else if (((state == 0) || (state == 2)) && (p_rexRes = argLine.apply(sLine))) {
        if (!p_rexRes->field(1).compare("vendid")) {
          vendid = (uint32_t)strtoull(p_rexRes->field(2).c_str(), NULL, 16);
        } else if (!p_rexRes->field(1).compare("devid")) {
          devid = (device_id_t)strtoull(p_rexRes->field(2).c_str(), NULL, 16);
        } else if (!p_rexRes->field(1).compare("caguid")) {
          nGuid = strtoull(p_rexRes->field(2).c_str(), NULL, 16);
        } else if (!p_rexRes->field(1).compare("sysimgguid")) {
          sGuid = strtoull(p_rexRes->field(2).c_str(), NULL, 16);
        } else if (!p_rexRes->field(1).compare("switchguid")) {
          nGuid = strtoull(p_rexRes->field(2).c_str(), NULL, 16);
        } else {
          cout << "-W- ignoring unknown attribute: " << p_rexRes->field(1) << endl;
        }
        state = 0;
        delete p_rexRes;
      } else if ((state == 0) && (p_rexRes = basePortLine.apply(sLine))) {
        //regExp basePortLine("^(Ca|Switch)\\s+([0-9]+)\\s+\"(S|H|R)-([0-9a-fA-F]+)\"\\s+#\\s+\"([^\"]+)\""
        // "(\\s+(base|enhaced) port 0 lid\\s+([0-9a-fA-F]+)\\s+lmc\\s+([0-9]+))?$");
        phys_port_t numPorts = (phys_port_t)strtoul(p_rexRes->field(2).c_str(), NULL, 10);
        string nType = p_rexRes->field(3);
        uint64_t pGuid = strtoull(p_rexRes->field(4).c_str(), NULL, 16);
        string desc = p_rexRes->field(5);
        lid_t baseLid = (lid_t)strtoul(p_rexRes->field(8).c_str(), NULL, 10);
        IBNodeType nTypeCode =  IB_CA_NODE;
        if (!nType.compare("S")) {
          baseLid = (lid_t)strtoul(p_rexRes->field(8).c_str(), NULL, 10);
          nTypeCode = IB_SW_NODE;
        }

        // we only make new systems during pass 0
        if (pass == 0) {
          string sysType, sysName, nodeName;

          generateNodeAndSystemNames(nTypeCode, sGuid, nGuid, desc, false, nodeName, sysName, sysType,
                                     mlx_nd_format);

          pLocSys = getSystem(sysName);
          if (!pLocSys) {
            if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
              cout << "-V- Creating new sys:" << sysType << " name:" << sysName << endl;
            pLocSys = makeGenericSystem(sysName, sysType, mlx_nd_format);
          }

          pLocNode = getNodeByGuid(nGuid);
          if (! pLocNode) {
            if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
              cout << "-V- Creating new node: " << nType << " name:" << nodeName << endl;
            pLocNode = makeNode(nodeName, pLocSys, nTypeCode, numPorts);
            pLocNode->guid_set(nGuid);
            pLocNode->system_guid_set(sGuid);
            pLocNode->devId = devid;
            pLocNode->vendId = vendid;
            pLocNode->revId = 0;
            // for switches we need to assign base lid on all ports...
            if ( nTypeCode ==  IB_SW_NODE) {
              for (unsigned int pn = 1; pn <= numPorts; pn++) {
                pLocPort = pLocNode->getPort((phys_port_t)pn);
                pLocPort->base_lid = baseLid;
                pLocPort->guid_set(pGuid);
              }
            }
          }
        } else {
          pLocNode = getNodeByGuid(nGuid);
        }
        state = 1;
        delete p_rexRes;
      } else if ((state >= 1) && (p_rexRes = swPortLine.apply(sLine))) {
        // ^[[]([0-9]+)[]]\\s+\"(S|H|R)-([0-9a-fA-F]+)\"[[]([0-9]+)[]]([(]([0-9a-fA-F]+)[)])?
        // \\s+#\\s+\"([^\"]+)\"\\s+lid\\s+([0-9]+)\\s+(\\S+)\\s*$
        locPortNum = (phys_port_t)strtoul(p_rexRes->field(1).c_str(), NULL, 10);
        remNodeGuid = strtoull(p_rexRes->field(3).c_str(), NULL, 16);
        remPortNum = (phys_port_t)strtoul(p_rexRes->field(4).c_str(), NULL, 10);
        linkWS = p_rexRes->field(9);
        if (pass != 0) {
          if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {
            cout << "-V- Connecting SW port: " << (unsigned int)locPortNum << " to remNodeGuid: 0x";
            cout << setw(16) << setfill('0') << hex << remNodeGuid << dec << " port: "
                 << (unsigned int)remPortNum << endl;
          }
          doConn = true;
        }
        state = 2;
        delete p_rexRes;
      } else if ((state >= 1) && (p_rexRes = caPortLine.apply(sLine))) {
        // ^[[]([0-9]+)[]]([(]([0-9a-fA-F]+)[)])?\\s+\"(S|H|R)-([0-9a-fA-F]+)\"[[]([0-9]+)[]]\\s+#\\s+
        // lid\\s+([0-9]+)\\s+lmc\\s+([0-9]+)\\s+\"([^\"]+)\"
        // \\s+lid\\s+([0-9]+)\\s+(\\S+)\\s*$
        locPortNum = (phys_port_t)strtoul(p_rexRes->field(1).c_str(), NULL, 10);
        remNodeGuid = strtoull(p_rexRes->field(5).c_str(), NULL, 16);
        remPortNum = (phys_port_t)strtoul(p_rexRes->field(6).c_str(), NULL, 10);
        uint64_t pGuid = strtoull(p_rexRes->field(3).c_str(), NULL, 16);
        lid_t baseLid = (lid_t)strtoul(p_rexRes->field(7).c_str(), NULL, 10);
        // unsigned int lmc = strtoull(p_rexRes->field(8).c_str(), NULL, 10);
        linkWS = p_rexRes->field(11);
        if (pass != 0) {
          if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {
            cout << "-V- Connecting CA port: " << (unsigned int)locPortNum << " to remNodeGuid: 0x";
            cout << setw(16) << setfill('0') << hex << remNodeGuid << dec << " port: "
                 << (unsigned int)remPortNum << endl;
          }
          doConn = true;
        } else {
          if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE)
            cout << "-V- Setting PGUID: " << setw(16) << setfill('0') << hex << pGuid << dec << endl;
          pLocPort = pLocNode->getPort(locPortNum);
          if (!pLocPort) {
            cout << "-E- Could not get node: " << pLocNode->name
                 << " local port:" << (unsigned int)locPortNum << endl;
          } else {
            pLocPort->guid_set(pGuid);
            pLocPort->base_lid = baseLid;
          }
        }
        state = 2;
        delete p_rexRes;
      } else {
        cout << "-W- Ignoring line: " << sLine << endl;
      }

      // we may need to create a connection at this point
      if (doConn) {
        pLocPort = pLocNode->getPort(locPortNum);
        if (!pLocPort) {
          cout << "-E- Could not get local port:" << (unsigned int)locPortNum << endl;
          numErrs++;
        }

        pRemNode = getNodeByGuid(remNodeGuid);
        if (!pRemNode) {
          cout << "-E- Could not get remote node by NGUID:"
               << setw(16) << setfill('0') << hex << remNodeGuid << dec << endl;
          numErrs++;
	} else if (!pLocNode) {
          cout << "-E- Could not get local node by NGUID:"
               << setw(16) << setfill('0') << hex << nGuid << dec << endl;
          numErrs++;
        } else {
          pRemPort = pRemNode->getPort(remPortNum);
          if (!pRemPort) {
            cout << "-E- Could not get remote node: " << pRemNode->name
                 << " port:" << (unsigned int)remPortNum << endl;
            numErrs++;
          } else {
            // do the connection
            IBLinkWidth width;
            IBLinkSpeed speed;
            convLinkWidthSpeed(linkWS, width, speed);
            pLocPort->set_internal_data(speed, width, IB_PORT_STATE_ACTIVE);
            pRemPort->set_internal_data(speed, width, IB_PORT_STATE_ACTIVE);

            pLocPort->connect(pRemPort);

          }
        }
      }
    }
    // f.seekg(0);
    f.close();
    f.open(fn.c_str());
  }
  f.close();
  //restore cout flags
  cout.flags(cout_flags);
  constructSystems();
  return 0;
}

void IBFabric::constructSystems()
{
    map_str_psys systemTemplates;
    IBSystem *p_systemT;
    map_str_str mods;
    char sysTypeStr[16];
    bool isHCA = false;
    IBSystemsCollection *p_systemsCollection = theSysDefsCollection();
    //systemTemplates are added to SystemByName during execution
    //need to skip them
    map_str_psys  localSystemByName = SystemByName;

    for (map_str_psys::iterator tsI = localSystemByName.begin();
            tsI != localSystemByName.end();
            tsI++) {

        isHCA = false;
        IBSystem *p_system = (*tsI).second;

        // set system type
        if (p_system->type == "HCA") {
            if (p_system->sys_mlx_nd_format) {
                string max_mlx4, max_mlx5, sysTypeStr;
                for (map_str_pnode::iterator nI = p_system->NodeByName.begin();
                        nI != p_system->NodeByName.end() ; ++nI ) {
                    string nodeName = ((*nI).second)->name;
                    //find maximum i in mlx<4/5>_i for the current system
                    size_t inx = nodeName.find("mlx");
                    if (inx != std::string::npos) {
                        string cur_mlx = nodeName.substr(inx, nodeName.length()-1);
                        //the cur_mlx should include only mlx4 or mlx5
                        if((cur_mlx.find("mlx4") != std::string::npos) &&
                                (!max_mlx4.length() ||  cur_mlx > max_mlx4))
                            max_mlx4 = cur_mlx;
                        else if (!max_mlx5.length() || cur_mlx > max_mlx5)
                            max_mlx5 = cur_mlx;
                    }
                }
                //the result for sysTypeStr should inclue the mlx4/mlx5 and the
                //max number of HCAs for every mlx4/5. if there are none mlx4/5,
                //the sysTypeStr should not include mlx4/5.
                //so the sysTypeStr should be mlx4_<max_mlx4>_mlx5_<max_mlx5>
                sysTypeStr = p_system->type.c_str();
                sysTypeStr += (max_mlx4.length() ? ("_" + max_mlx4) : "");
                sysTypeStr += (max_mlx5.length() ? ("_" + max_mlx5) : "");
                p_system->type = string(sysTypeStr);
                size_t inx = max_mlx4.find("_");
                if (inx != std::string::npos) {
                    string mlx4_max_idx = max_mlx4.substr(inx+1);
                    p_system->max_mlx4 = atoi(mlx4_max_idx.c_str());
                }

                inx = max_mlx5.find("_");
                if (inx != std::string::npos) {
                    string mlx5_max_idx = max_mlx5.substr(inx+1);
                    p_system->max_mlx5 = atoi(mlx5_max_idx.c_str());
                }
            } else {
                int maxIdx = (int)p_system->NodeByName.size();
                int nodeIdx = 0;
                for (map_str_pnode::iterator nI = p_system->NodeByName.begin();
                        nI != p_system->NodeByName.end() ; ++nI ) {

                    nodeIdx = (int)strtol( nI->first.substr(
                                p_system->name.length()+2).c_str(), NULL, 0);

                    maxIdx = ( maxIdx > nodeIdx ? maxIdx : nodeIdx);
                }
                sprintf(sysTypeStr,"%s_%u", p_system->type.c_str(), maxIdx);
                p_system->type = string(sysTypeStr);
            }
            isHCA = true;
        }

        map_str_psys::iterator stI = systemTemplates.find(p_system->type);
        p_systemT = NULL;

        if (stI == systemTemplates.end()) {
            string name = p_system->type;
            name += "_Template";
            IBSysDef *p_sysDef = p_systemsCollection->getSysDef(p_system->type);
            if (p_sysDef) {
                p_systemT = p_systemsCollection->makeSystem(
                    this, name, p_system->type, mods);
                if (p_systemT)
                    systemTemplates[p_system->type] = p_systemT;
            } else {
                cout << "-W- Fail to find system definition for: " <<
                    p_system->type << "\n    (" <<
                    p_system->NodeByName.begin()->second->description << ")\n";
            }
        } else {
            p_systemT = stI->second;
        }


        if (p_systemT == NULL)
            constructGeneralSystem(p_system);
        else
            constructSystem(p_system, p_systemT, isHCA);
    }

    //clear systemTemplates
    for (map_str_psys::iterator stI = systemTemplates.begin();
            stI != systemTemplates.end();
            stI++) {
        delete((*stI).second);
    }

}

int IBFabric::constructGeneralSystem(IBSystem *p_system)
{
    p_system->newDef = true;

    // go over all nodes of the system:
    for (map_str_pnode::iterator nI = p_system->NodeByName.begin();
            nI != p_system->NodeByName.end();
            nI++) {
        IBNode *p_node = (*nI).second;

        if (constructGeneralSystemNode(p_system, p_node))
            return -1;
    }
    return 0;
}

int IBFabric::constructGeneralSystemNode(IBSystem *p_system, IBNode *p_node)
{
    bool isExternal = true;

    // go over all ports of the node:
    for (unsigned int pn = 1; pn <= p_node->numPorts; pn++) {
        IBPort *p_port = p_node->getPort((phys_port_t)pn);
        isExternal = true;

        if (!p_port || !p_port->p_remotePort) {
            isExternal = true;
        } else if (p_port->p_sysPort) {
            cout << "-E- Sys port: " << p_port->p_sysPort << "already exist for node: "
            << p_node->name << " port: " << pn << endl;
            continue;
        } else if (!p_port->p_remotePort->p_node ||
            !p_port->p_remotePort->p_node->p_system) {
            cout << "-E- Invalid remote port node or system for node: "
            << p_node->name << " port: " << pn << endl;
            isExternal = true;
        }
        else if (p_system == p_port->p_remotePort->p_node->p_system) {
            //single node SW do not supports internal links
            //all internal links ar loopbacks
            if (p_port->p_remotePort->p_node == p_port->p_node) {
                cout << "-W- Discovered loopback from: " << p_port->getName()
                << " to: " << p_port->p_remotePort->getName() << endl;
                isExternal = true;
            } else {
                isExternal = false;
            }
        }

        if (p_port && isExternal) {

            char buf[128];
            p_system->generateSysPortName(buf, p_node, pn);

            if (p_system->getSysPort(buf)) {
                cout << "-E- Sys port: " << buf << "already exist for node: "
                    << p_node->name << endl;
                return -1;
            }

            p_port->p_sysPort = new IBSysPort(buf, p_system);
            p_port->p_sysPort->p_nodePort = p_port;

            if (p_port->p_remotePort && p_port->p_remotePort->p_sysPort)
                p_port->p_sysPort->connectPorts(p_port->p_remotePort->p_sysPort);
        }
    }
    return 0;
}

int IBFabric::constructSystemNode(IBNode *p_node, IBNode *p_nodeT,
                                  set_nodes_ptr &missingNodesPtr,
                                  map_template_pnode &systemToTemplateMap,
                                  bool &isModSys, bool &isNewSys)
{
    bool isExternal = true;
    IBSystem *p_system = p_node->p_system;
    bool disconnectedInternal;

    for (unsigned int pn = 1; pn <= p_nodeT->numRealPorts(); pn++) {
        IBPort *p_port = p_node->getPort((phys_port_t)pn);
        IBPort *p_portT = p_nodeT->getPort((phys_port_t)pn);
        isExternal = true;
        disconnectedInternal = false;

        if (!p_port || !p_port->p_remotePort) {
            if (p_portT && p_portT->p_remotePort) {
                if (missingNodesPtr.find(p_portT->p_remotePort->p_node) ==
                    missingNodesPtr.end()) //mising System Boards reported separately
                    cout << "-E- Missing internal connectivity in node: "
                         << p_node->name << " port: " << pn
                         << endl;
                isModSys = true;
                disconnectedInternal = true;
            }
        }

        else if (p_port->p_sysPort) {
            cout << "-E- Sys port: " << p_port->p_sysPort
                << " already exist for node: " << p_node->name
                << " port: " << pn << endl;
            continue;
        }

        else if (!p_port->p_remotePort->p_node ||
                 !p_port->p_remotePort->p_node->p_system) {
            cout << "-E- Invalid remote port node or system for node: "
                << p_node->name << " port: " << pn << endl;

            if (p_portT && p_portT->p_remotePort)
                isModSys = true; // if internal in spec
        }

        //check internal link
        else if (p_portT && p_portT->p_remotePort) {

            map_template_pnode::iterator tI =
                systemToTemplateMap.find(p_port->p_remotePort->p_node);

            if (tI != systemToTemplateMap.end() &&
                p_portT->p_remotePort->p_node == tI->second &&
                p_port->p_remotePort->num == p_portT->p_remotePort->num) {
                isExternal = false; //identical to spec
            } else {
                //may be different from template
                if (p_system != p_port->p_remotePort->p_node->p_system) {
                    isNewSys = true;
                } else {
                    isExternal = false;
                }

                if (tI != systemToTemplateMap.end()) {
                    cout << "-W- Remote port: " << p_port->p_remotePort->getName()
                        << " different from template remote port: "
                        << p_portT->p_remotePort->getName()
                        << " from port: " << p_port->getName() << endl;

                    isNewSys = true;
                } //else will be checked by remote side
            }
        }

        // loopback
        else if (p_system == p_port->p_remotePort->p_node->p_system) {
            cout << "-W- Discovered loopback from: " << p_port->getName()
            << " to: " << p_port->p_remotePort->getName() << endl;
        }

        if (p_port && isExternal) {
            char buf[128];

            if (p_portT == NULL || p_portT->p_sysPort == NULL) {
                p_system->generateSysPortName(buf, p_node, pn);

                if (!p_portT) {
                    cout << "-E- failed to find node: "
                        << p_node->name << " port: " << pn
                        << " in system spec: " << p_system->type
                        << " using: "<< buf << endl;
                    isNewSys = true;
                } else if (!disconnectedInternal) {
                    cout << "-E- failed to find external port name for node: "
                        << p_node->name << " port: " << pn
                        << " using: "<< buf << endl;
                    isNewSys = true;
                }
            } else {
                sprintf(buf, "%s", p_portT->p_sysPort->name.c_str());
            }

            if (p_system->getSysPort(buf)) {
                cout << "-E- Sys port: " << buf << "already exist for node: "
                    << p_node->name << endl;
                continue;
            }

            p_port->p_sysPort = new IBSysPort(buf, p_system);
            p_port->p_sysPort->p_nodePort = p_port;

            if (p_port->p_remotePort && p_port->p_remotePort->p_sysPort)
                p_port->p_sysPort->connectPorts(p_port->p_remotePort->p_sysPort);
        }
    }

    return 0;
}

//create system ports acording to the nodes and ports
int IBFabric::constructSystem(IBSystem *p_system,
                              IBSystem *p_systemTemplate,
                              bool isHCA)
{
    p_system->newDef = false;

    // go over all nodes of the system:
    string instName;
    bool isNewSys = false;
    bool isModSys = false;
    map_template_pnode systemToTemplateMap;
    map_template_pnode templateToSystemMap;
    set_boards existingBoards;
    set_boards missingNodeBoards;
    set_nodes_ptr missingNodesPtr;

    regExp SwSpineLineExp("(\\S+/)([S,L])0*([0-9]+)(/\\S+)$");
    regExp NodeNameExp("\\S+/(\\S+)/(\\S+)$");
    rexMatch *p_rexRes;

    //map nodes - system to template and template to system

    for (map_str_pnode::iterator nI = p_system->NodeByName.begin();
            nI != p_system->NodeByName.end();
            nI++) {
        IBNode *p_node = (*nI).second;
        string nameWithoutSysName = p_node->name.substr(
            p_system->name.length(),
            p_node->name.length() - p_system->name.length());
        string nodeNameT = p_systemTemplate->name + nameWithoutSysName;

        IBNode *p_nodeT = p_systemTemplate->getNode(nodeNameT);
        if (p_nodeT == NULL &&
            (p_rexRes = SwSpineLineExp.apply(nodeNameT.c_str()))) {

            nodeNameT =  p_rexRes->field(1) +
                (p_rexRes->field(2) == "L" ? string("leaf") : string("spine")) +
                p_rexRes->field(3) + p_rexRes->field(4);
            p_nodeT = p_systemTemplate->getNode(nodeNameT);
            delete p_rexRes;
        }

        if (p_nodeT == NULL) {
            cout << "-E- Fail to find node name: " << nameWithoutSysName.substr(1)
            << " in system type: "<< p_systemTemplate->type << endl;

            if (p_system->p_fabric->constructGeneralSystemNode(p_system, p_node))
                return -1;

            isNewSys = true;
            continue;
        }

        systemToTemplateMap[p_node] = p_nodeT;
        templateToSystemMap[p_nodeT] = p_node;
    }

    //find Missing System Boards skip HCA_
    if (!isHCA) {
        for (map_str_pnode::iterator nI = p_systemTemplate->NodeByName.begin();
                nI != p_systemTemplate->NodeByName.end();
                nI++) {
            IBNode *p_nodeT = (*nI).second;
            string boardName;
            string nodeName;

            //Get Board Name
            if ((p_rexRes = NodeNameExp.apply(p_nodeT->name.c_str()))) {
                boardName =  p_rexRes->field(1);
                nodeName = p_rexRes->field(1) + "/" + p_rexRes->field(2);
                delete p_rexRes;
            } else {
                continue;
            }

            map_template_pnode::iterator tI =
                templateToSystemMap.find(p_nodeT);

            if (tI == templateToSystemMap.end()) {
                missingNodeBoards.insert(boardName);
                missingNodesPtr.insert(p_nodeT);
                isModSys = true;
            } else
                existingBoards.insert(boardName);
        }
    }

    for (map_template_pnode::iterator nI = systemToTemplateMap.begin();
            nI != systemToTemplateMap.end();
            nI++) {
        IBNode *p_node = (*nI).first;
        IBNode *p_nodeT = (*nI).second;

        constructSystemNode(p_node, p_nodeT, missingNodesPtr, systemToTemplateMap,
                            isModSys, isNewSys);
    }

    if (isNewSys) {
        cout << "-E- Sys: " << p_system->name << " does not match its IBNL: "
            << p_system->type << endl;
        p_system->type = p_system->type + string("_") + p_system->name;
        p_system->newDef = true;
    }
    else if (isModSys) {
        cout << "-I- Sys: " << p_system->name << " is modified.\n";
        //Boards that are missing comletely (missingNodeBoards - existingBoards)
        set_boards missingBoards;
        //Boards that are missing part ot thire nodes (missingNodeBoards - missingBoards)
        set_boards missingNodes;

        //generate system cfg string 
        //missingNodeBoards contains boards that misses at least one noad if the same board
        //do not exist in existingBoards, the board is missing completely
        if (!missingNodeBoards.empty()) {
            std::set_difference(
                missingNodeBoards.begin(),
                missingNodeBoards.end(),
                existingBoards.begin(),
                existingBoards.end(),
                std::insert_iterator<set_boards>(missingBoards, missingBoards.end()));

            std::set_difference(
                missingNodeBoards.begin(),
                missingNodeBoards.end(),
                missingBoards.begin(),
                missingBoards.end(),
                std::insert_iterator<set_boards>(missingNodes, missingNodes.end()));
        }

        if (!missingBoards.empty()) {
            stringstream sstream;
            set_boards::iterator bI = missingBoards.begin();
            cout << "-I- Missing Board: " << *bI
                << " in system: "<< p_system->name << endl;
            sstream << " CFG: " << *bI <<"=R";
            ++bI;

            for ( ; bI != missingBoards.end(); ++bI) {
                cout << "-I- Missing Board: " << *bI
                    << " in system: "<< p_system->name << endl;
                sstream << "," << *bI <<"=R";
            }
            p_system->cfg = sstream.str();
        }
        for ( set_boards::iterator nI = missingNodes.begin();
              nI != missingNodes.end(); ++nI)
            cout << "-W- Missing Node in Board " << *nI
                << " in system: "<< p_system->name << endl;

    }
    return 0;
}

//Returns: SUCCESS_CODE / ERR_CODE_IO_ERR
int IBFabric::OpenFile(const char *file_name,
                       ofstream& sout,
                       bool to_append,
                       string & err_message,
                       bool add_header,
                       ios_base::openmode mode)
{
    err_message.clear();

    if (to_append)
        sout.open(file_name, ios::app);
    else {

#ifndef WIN32 //Fix security vulnerability
        srand((unsigned int)time(NULL));
        int r = rand();
        char numstr[32]; // enough to hold all numbers up to 64-bits + 2 chars
        sprintf(numstr, "_%X", r);
        string tmp_path(file_name);
        tmp_path += numstr;

        remove(file_name);
        remove(tmp_path.c_str());

        sout.open(tmp_path.c_str(), mode);

        if (!sout.fail()) {
            if (rename(tmp_path.c_str(), file_name)) {
                sout.close();
                err_message = string("Failed to rename file to ") + file_name + ".";
                return 1;
            }
        }
#else
        sout.open(file_name, mode);
#endif
    }

    if (sout.fail()) {
        err_message = string("Failed to open file ") + file_name + " for writing.";
        return 1;
    }

    if (!to_append && add_header) {
        sout << "# This database file was automatically generated by IBDIAG" << endl;
        sout << endl << endl;
    }

    return 0;
}

int IBFabric::applySubCluster()
{
    //applySubCluster can be called twiced,
    //remember the whole fabric, before sub cluster
    if (FullNodeByName.empty())
        FullNodeByName = NodeByName;
    for (map_str_pnode::iterator nI = NodeByName.begin();
            nI != NodeByName.end();) {
        IBNode *p_curr_node = (*nI).second;

        if (!p_curr_node) {
            cout << "-E-  found null node in NodeByName map" << endl;
            return 1;
        }

        u_int8_t i =  (u_int8_t)(p_curr_node->type == IB_CA_NODE ? 1 : 0);

        //delete node and/or unrelevant ports to the path
        //delete (mark as NULL) this node's ports
        for (;i <= p_curr_node->numPorts; i++) {
            IBPort *p_curr_port = p_curr_node->getPort(i);
            if(!p_curr_port)
                continue;
            //node/port isn't in subcluster ,need to removed this node
            if (!p_curr_node->getInSubFabric() ||
                !p_curr_port->getInSubFabric()) {
                //remove the link that involves this port
                if (p_curr_port->p_remotePort)
                    p_curr_port->disconnect();
                //check if those entry exist first
                lid_t curr_lid = p_curr_port->base_lid;
                if (getPortByLid(curr_lid))
                    PortByLid[curr_lid] = NULL;
                u_int64_t curr_guid = p_curr_port->guid_get();
                if (getPortByGuid(curr_guid))
                    PortByGuid[curr_guid] = NULL;
            }
        }

        map_str_pnode::iterator toErase = nI;
        nI++;
        /* if node isn't in sub_node , need to erase from NodeByName */
        if(!p_curr_node->getInSubFabric())
            NodeByName.erase(toErase);
    }
    return 0;
}

int IBFabric::parseScopePortGuidsFile(const string& fn, bool include_in_scope)
{
    // get the extension of the file name
    size_t pos = fn.rfind(".");
    string ext;

    if (pos != string::npos) {
        ext = fn.substr(pos + 1);
        if (ext.compare("guid")) {
            cout << "-E- The scope file:" << fn.c_str()
                 << " has wrong extension " << ext
                 << ". Can't parse file." << endl;
            return 1;
        }
    }

    char sLine[1024] = {'\0'};
    //version: x.y
    regExp versionLine("^\\s*version:\\s*([0-9]+)\\.([0-9]+)\\s*$");
    //0x2c9000000002d
    //2c9000000002d
    //0x2c9000000002d@1/17/3  <-- meaning take ports 1,17,3 of this node
    regExp guidLine("^\\s*(0x)?([0-9a-fA-F]+)(@[0-9]+(/[0-9]+)*)?\\s*$");
    //ignore comments (#), white spaces
    regExp ignoreLine("^[ \t]*(#|[ \t]*$)");
    //key word for all switches, i.e take all switches
    regExp allSWsLine("^\\s*ALL_SWITCHES\\s*$");
    //key word for all CAs, i.e take all CAs
    regExp allCAsLine("^\\s*ALL_CAS\\s*$");
    bool beforeAnyData = true;

    ifstream fs(fn.c_str());
    rexMatch *p_rexRes;

    if (fs.fail()) {
        cout << "-E- Failed to open file:" << fn << endl;
        return 1;
    }

    cout << "-I- Parsing " << (include_in_scope ? "" : "exclude ")
         << "scope file: " << fn << endl;
    int lineNum = 0;

    unsigned int anyErr = 0, ver_major = 0, ver_minor = 0;
    bool is_all_sw = false, is_all_ca = false, ver_found = false;
    map_pnode_ports_bitset node_2_ports;
    IBNode *p_node = NULL;
    string guid_str, guid_ports, port_str;
    u_int64_t guid = 0;

    //get list of port guids from the file
    while (fs.good()) {
        lineNum++;
        fs.getline(sLine, sizeof(sLine));

        if ((p_rexRes = ignoreLine.apply(sLine)))
            ;
        else if ((p_rexRes = versionLine.apply(sLine))) {
            if (!beforeAnyData) {
                cout << "-E- File format version must be defined before "
                     << "any data line! Version found at line "
                     << lineNum << endl;
                anyErr++;
                break;
            }
            ver_major = atoi(p_rexRes->field(1).c_str());
            ver_minor = atoi(p_rexRes->field(2).c_str());
            ver_found = true;
            if (ver_major != 1 || ver_minor != 0) {
                cout << "-E- File format version " << ver_major << "."
                     << ver_minor << " is unsupported. "
                     << "The supported version is 1.0" << endl;
                anyErr++;
            }
        } else if ((p_rexRes = guidLine.apply(sLine))) {
            beforeAnyData = false;
            guid_str = p_rexRes->field(2);
            guid = strtoull(guid_str.c_str(), NULL, 16);
            p_node = NULL;
            p_node = getNodeByGuid(guid);
            if (p_node) {
                if ((is_all_sw && p_node->type == IB_SW_NODE) ||
                    (is_all_ca && p_node->type == IB_CA_NODE)) {
                    //skip node if its type is selected in all flag
                    delete p_rexRes;
                    p_rexRes = NULL;
                    continue;
                }
                port_str.clear();

                //if it's the first time we meet this node in the file
                //make entry
                if (node_2_ports.find(p_node) == node_2_ports.end()) {
                    PortsBitset node_ports;
                    node_2_ports.insert(pair<IBNode*, PortsBitset>(
                                            p_node, node_ports));
                }

                //check if the ports are mentioned
                //0x2c9000000002d@1/17/3
                if (p_rexRes->field(3).size()) {
                    //don't take @
                    guid_ports = p_rexRes->field(3).substr(1);
                    stringstream ss(guid_ports);

                    while (getline(ss, port_str, '/')) {
                        phys_port_t port = (phys_port_t)(atoi(port_str.c_str()));
                        if (port > IB_MAX_PHYS_NUM_PORTS ||
                            port < IB_MIN_PHYS_NUM_PORTS)
                            cout << "-W- guid " << guid_str
                                 << " has out of range ("
                                 << IB_MIN_PHYS_NUM_PORTS << "-"
                                 << IB_MAX_PHYS_NUM_PORTS
                                 << ") port number " << port_str
                                 << " at line " << lineNum << endl;
                        else if (port > p_node->numPorts) {
                            cout << "-E- guid " << guid_str
                                 << " has port " << port_str
                                 << " > number of ports for this node ("
                                 << (unsigned int)p_node->numPorts << ")"
                                 << " at line " << lineNum << endl;
                            anyErr++;
                        } else
                            node_2_ports[p_node].set(port);
                    }
                }
            } else
                cout << "-W- node guid at line " << lineNum
                     << " isn't found in the fabric" << endl;
        } else if ((p_rexRes = allSWsLine.apply(sLine)))
            is_all_sw = true;
        else if ((p_rexRes = allCAsLine.apply(sLine)))
            is_all_ca = true;
        else {
            cout << "-E- Unsupported syntax at line: " << lineNum << endl;
            anyErr++;
        }

        if (p_rexRes) {
            delete p_rexRes;
            p_rexRes = NULL;
        }
    }
    fs.close();

    // validate supported format
    if (!ver_found) {
        cout << "-E- version wasn't found" << endl;
        anyErr++;
    }

    if (anyErr) {
        cout << "-E- failed to parse file" << endl;
        return anyErr;
    }

    //check if was given an empty file and warn about it
    //empty scope file - nothing in the fabric
    //empty exclude scope file - everything in the fabric
    if (node_2_ports.empty() && !is_all_sw && !is_all_ca)
        cout << "-W- the " << (include_in_scope ? "" : "exclude ")
             << "scope file is empty" << endl;

    IBScope scope(node_2_ports, is_all_sw, is_all_ca);

    //mark nodes/ports to be left in the scope
    if (include_in_scope)
        markInScopeNodes(scope);
    else
        markOutScopeNodes(scope);

    applySubCluster();
    return 0;
}

int IBFabric::markInScopeNodes(IBScope &in_scope)
{
    //nothing to do, as all the nodes in NodesByName are marked as in scope
    if (in_scope.is_all_sw && in_scope.is_all_ca)
        return 0;

    //mark initially all node to be not in the scope
    for (map_str_pnode::iterator nI = NodeByName.begin();
         nI != NodeByName.end(); ++nI) {
        IBNode *p_node = (*nI).second;
        if (!p_node)
            continue;
        //except nodes which type is set by all flag
        if (!(p_node->type == IB_SW_NODE && in_scope.is_all_sw) &&
            !(p_node->type == IB_CA_NODE && in_scope.is_all_ca))
            p_node->setInSubFabric(false);
    }

    map_pnode_ports_bitset::iterator npI = in_scope.node_ports.begin();
    map_pnode_ports_bitset::iterator npEnd = in_scope.node_ports.end();
    list_phys_ports::iterator pI, pEnd;
    IBNode *p_curr_node = NULL;
    IBPort *p_curr_port = NULL;

    //go over scope nodes and their ports and mark them to be in the scope
    for (; npI != npEnd; npI++) {
        p_curr_node = npI->first;

        if (p_curr_node) {
            p_curr_node->setInSubFabric(true);
            //if node has only some ports in the scope
            if (npI->second.any()) {
                //go over the list of ports that are not in the scope
                PortsBitset ports_not_in_scope = ~npI->second;

                for (unsigned int i = 1; i <= p_curr_node->numPorts; i++) {
                    p_curr_port = p_curr_node->getPort((phys_port_t)i);
                    if (!p_curr_port)
                        continue;
                    //mark those that are not in scope
                    //check if the port is in the complementary list to scope
                    if (ports_not_in_scope.test(i))
                        p_curr_port->setInSubFabric(false);
                }
            }
        }
    }

    return 0;
}

int IBFabric::markOutScopeNodes(IBScope &out_scope)
{
    map_pnode_ports_bitset::iterator npI = out_scope.node_ports.begin();
    map_pnode_ports_bitset::iterator npEnd = out_scope.node_ports.end();
    list_phys_ports::iterator pI, pEnd;
    IBNode *p_curr_node = NULL;
    IBPort *p_curr_port = NULL;

    //assume that all the nodes/ports that are currently in the fabric
    //are marked correctly (true=in scope)
    for (; npI != npEnd; npI++) {
        p_curr_node = npI->first;

        if (p_curr_node) {
            //if node has ports, then mark only ports to be out of scope
            if (npI->second.any()) {
                u_int8_t max_port_set = (u_int8_t)npI->second.max_set_bit_pos();
                for (u_int8_t i = 0; i <= max_port_set; i++) {
                    p_curr_port = p_curr_node->getPort(i);
                    if (p_curr_port && npI->second.test(i))
                        p_curr_port->setInSubFabric(false);
                }
            } else
                p_curr_node->setInSubFabric(false);
        }
    }

    //ALL_CAS or ALL_SWITCHES is set, mark them out from scope
    if (out_scope.is_all_sw || out_scope.is_all_ca) {
        for (map_str_pnode::iterator nI = NodeByName.begin();
             nI != NodeByName.end(); ++nI) {
            IBNode *p_node = (*nI).second;
            if (!p_node)
                continue;
            if ((p_node->type == IB_SW_NODE && out_scope.is_all_sw) ||
                (p_node->type == IB_CA_NODE && out_scope.is_all_ca))
                p_node->setInSubFabric(false);
        }
    }

    return 0;
}

// make a new vnode if can not find it by guid.
// p_vport is the vport that we reached this vnode with.
// the local_vport_num is the vport local index num in the vnode.
IBVNode *
IBFabric::makeVNode(uint64_t guid, virtual_port_t num_vports,
                    IBVPort *p_vport, virtual_port_t local_vport_num)
{
    IBVNode *p_vnode;
    map_guid_pvnode::iterator nI = VNodeByGuid.find(guid);
    if (nI == VNodeByGuid.end()) {
        if (num_vports > IB_MAX_VIRT_NUM_PORTS) {
            cout << "-E- VNode " << guid << " has bad number of ports " << (unsigned int)num_vports << endl;
            return NULL;
        }
        p_vnode = new IBVNode(guid, this, num_vports);
    } else {
        p_vnode = (*nI).second;
    }

    //TBD: need to handle when the addVPort returns error
    p_vnode->addVPort(local_vport_num, p_vport);

    return p_vnode;
}

// make a new vport anyway even if this is duplicated vport.
// We will check for duplicated vports in the CheckVPortDuplicatedGuids
IBVPort *
IBFabric::makeVPort(IBPort *p_port, virtual_port_t vport_num,
                    uint64_t guid, IBPortState vport_state)
{
    IBVPort *p_vport;
    if (vport_num > IB_MAX_VIRT_NUM_PORTS) {
        cout << "-E- VPort " << guid << " has bad vport number "
             << (unsigned int)vport_num << endl;
        return NULL;
    }
    p_vport = new IBVPort(p_port, vport_num,
                          guid, vport_state, this);

    return p_vport;
}

//////////////////////////////////////////////////////////////////////////////
//
// CLASS IBVPort:
//

// constructor
IBVPort::IBVPort(IBPort *p_phys_portPtr, virtual_port_t number,
                 uint64_t guid, IBPortState vport_state, IBFabric *p_fabric)
{
    m_p_phys_port     = p_phys_portPtr;
    m_num             = number;
    m_guid            = guid;
    m_vport_state     = vport_state;
    m_p_fabric        = p_fabric;
    m_p_vnode         = NULL;
    m_vlid            = 0;

    m_p_fabric->VPortByGuid[m_guid] = this;

    //set vport index
    createIndex = ((p_phys_portPtr) && (p_phys_portPtr->p_node->p_fabric)) ?
                         p_phys_portPtr->p_node->p_fabric->getVPortIndex() : 0;
}

// destructor
IBVPort::~IBVPort()
{
    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {
        cout << "-I- Destructing VPort:" << this->getName()
                << "/" << m_p_phys_port->num << "/" << (unsigned int)m_num << endl;
    }

    // remove from the fabric VPortsByGuid:
    if (m_p_fabric) {
        map_guid_pvport::iterator nI = m_p_fabric->VPortByGuid.find(m_guid);
        if (nI != m_p_fabric->VPortByGuid.end()) {
            m_p_fabric->VPortByGuid.erase(nI);
        }
    }
}

// Set the guid in the fabric too:
void
IBVPort::guid_set(uint64_t g)
{
        m_guid = g;
}

string
IBVPort::getName()
{
    string name;
    if (!m_p_phys_port) {
        cerr << "Got a vport with no phys port" << endl;
        abort();
    }

    char buff[32];
    sprintf(buff,"/VP%u", m_num);
    name = m_p_phys_port->getName() + string(buff);
    return name;
}

void
IBVPort::setVNodePtr(IBVNode *p_vn)
{
    this->m_p_vnode = p_vn;
}

class IBVNode *
IBVPort::getVNodePtr()
{
    return this->m_p_vnode;
}

//////////////////////////////////////////////////////////////////////////////
//
// CLASS IBVNode:
//

// constructor
IBVNode::IBVNode(uint64_t g,
                 IBFabric *p_fab,
                 virtual_port_t np)
{
    p_fabric        = p_fab;
    numVPorts       = np;
    guid            = g;
    description     = "";

    p_fabric->VNodeByGuid[guid] = this;
    //set node index
    createIndex = (p_fab) ? p_fab->getVNodeIndex() : 0;
}

// Delete the node cleaning up all it's connections
IBVNode::~IBVNode()
{
    if (FabricUtilsVerboseLevel & FABU_LOG_VERBOSE) {
        cout << "-I- Destructing VNode:" << description << endl;
    }

    // remove from the fabric VNodesByGuid:
    if (p_fabric) {
        map_guid_pvnode::iterator nI = p_fabric->VNodeByGuid.find(guid);
        if (nI != p_fabric->VNodeByGuid.end()) {
            p_fabric->VNodeByGuid.erase(nI);
        }
    }
}

void
IBVNode::setDescription(string n) {
    description = n;
}

string
IBVNode::getDescription() {
    return description;
}

//Get the VNode name. Currently returns the vnode description.
//TBD:  need to change to be name of the vnode as the name of node.
//      The name should be geneate from the function "generateNodeAndSystemNames"
//      as function makeNode uses.
string
IBVNode::getName() {
    return description;
}

// add a vport
int
IBVNode::addVPort(virtual_port_t vportnum, IBVPort *p_vport)
{
    // The values for the vportnum and numVPorts was taken from VNodeInfo
    // So if we have an error, need to check why fw gives these values.
    if ((vportnum < 1) || (vportnum > numVPorts)) {
        cout << "-E- Given vport number out of range: 1 < " << (int)vportnum
            << " < " << numVPorts << endl;
        return 1;
    }
    map_vportnum_vport::iterator findVPort = VPorts.find(vportnum);
    if (findVPort != VPorts.end()) {
        cout << "-W- vport number "<< (int)vportnum << "already exist in DB "
            << endl;
        return 0;
    } else
        VPorts.insert(pair_vportnum_vport(vportnum, p_vport));

    return 0;
}

